/*
 * bootinfo.h 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Ru Yi <yi.ru@borqs.com>
 *
 * This file is designed for Borqs linux phone platform collecting boot info 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * Revision History:
 *                    Modification      Tracking
 * Author                 Date          Description of Changes
 * ----------------   ------------    -------------------------------
 * Leo He              Nov 9, 2009      Boot info refactoring
 *
 */
#ifndef __ASM_ARM_BOOT_INFO_H__
#define __ASM_ARM_BOOT_INFO_H__

#define BOOT_MODE_MASK  (0x0000FFFF)
#define BOOT_FLEX_MASK  (0xFFFF0000)

#define BOOT_MODE_TAG "androidboot.mode="
#define BOOT_FLEX_TAG "borqsboot.flex="

#define BOOT_MODE_LIST {                                \
	[BOOT_MODE_HARDWARE_RESET]  = "hardware_reset", \
	[BOOT_MODE_WATCHDOG]        = "watchdog",       \
	[BOOT_MODE_RTC_ALARM]       = "rtc_alarm",      \
	[BOOT_MODE_POWER_KEY]       = "power_key",      \
	[BOOT_MODE_USB_CABLE]       = "usb_cable",      \
	[BOOT_MODE_USB_FACTORY]     = "usb_factory",    \
	[BOOT_MODE_USB_CHARGER]     = "usb_charger",    \
	[BOOT_MODE_NORMAL]          = "normal",         \
	[BOOT_MODE_FACTORY]         = "factory",        \
	[BOOT_MODE_RECOVERY]        = "recovery",       \
	[BOOT_MODE_FOTA]            = "fota",           \
	[BOOT_MODE_CRITICAL]        = "critical",       \
	[BOOT_MODE_LOW_POWER_RESET] = "lowpower_reset",  \
	[BOOT_MODE_SDL_FLASHMODE]   = "sdl_flashmode",  \
}
#define BOOT_REASON_LIST {                   \
	[BOOT_REASON_NORMAL]   = "normal",   \
	[BOOT_REASON_FACTORY]  = "factory",  \
	[BOOT_REASON_FLASH]    = "flash",    \
	[BOOT_REASON_RECOVERY] = "recovery", \
	[BOOT_REASON_FOTA]     = "fota",     \
	[BOOT_REASON_CRITICAL] = "critical", \
	[BOOT_REASON_SDL_FLASHMODE] = "sdl_flashmode", \
}
#define BOOT_FLEX_LIST {                     \
	[BOOT_FLEX_USBNET]     = "usbnet",   \
	[BOOT_FLEX_FACTORY]    = "factory",  \
	[BOOT_FLEX_FOTA]       = "fota",     \
	[BOOT_FLEX_FTM]        = "ftm",      \
	[BOOT_FLEX_FLASH]      = "flash",    \
}

typedef enum {
	BOOT_MODE_HARDWARE_RESET,
	BOOT_MODE_WATCHDOG,
	BOOT_MODE_RTC_ALARM,
	BOOT_MODE_POWER_KEY,
	BOOT_MODE_USB_CABLE,
	BOOT_MODE_USB_FACTORY,
	BOOT_MODE_USB_CHARGER,
	BOOT_MODE_NORMAL,
	BOOT_MODE_FACTORY,
	BOOT_MODE_RECOVERY,
	BOOT_MODE_FOTA,
	BOOT_MODE_CRITICAL,
	BOOT_MODE_LOW_POWER_RESET,
	BOOT_MODE_SDL_FLASHMODE,
	BOOT_MODE_MAX,
} boot_mode_t;

typedef enum {
	BOOT_REASON_NORMAL,
	BOOT_REASON_FACTORY,
	BOOT_REASON_FLASH,
	BOOT_REASON_RECOVERY,
	BOOT_REASON_FOTA, 
	BOOT_REASON_CRITICAL,
	BOOT_REASON_SDL_FLASHMODE,
	BOOT_REASON_MAX,
} boot_reason_t;

typedef enum {
	BOOT_FLEX_USBNET,
	BOOT_FLEX_FOTA,
	BOOT_FLEX_FACTORY,
	BOOT_FLEX_FTM,
	BOOT_FLEX_FLASH,
	BOOT_FLEX_MAX,
} boot_flex_t;

boot_mode_t   get_boot_mode(void);
int           set_boot_reason(boot_reason_t br);
boot_reason_t get_boot_reason(void);
boot_flex_t   get_boot_flex(void);

#endif // __ASM_ARM_BOOT_INFO_H__
