/*
 *  linux/arch/arm/mach-pxa/beacon.c
 *
 *  Support for FBW Beacon Product.
 *
 *  Copyright (C) 2009 Borqs Ltd.
 *  Author:  Ru Yi <yi.ru@borqs.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 * 
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/input.h>
#include <linux/gpio_keys.h>
#include <linux/synaptics_i2c_rmi.h>
#include <linux/bq2407x_charger.h>
#include <linux/i2c-gpio.h>
#include <linux/console.h>
#include <linux/bootmem.h>
#include <linux/dpram.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <asm/irq.h>
#include <asm/bootinfo.h>
#include <asm/hci_ps.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/pxa3xx_nand.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/pxa27x_keypad.h>
#include <mach/gpio.h>
#include <mach/micco.h>
#include <mach/beacon.h>
#include <mach/irda.h>
#include <mach/pxa3xx_u2d.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/udc.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxafb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/pxa930_trkball.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa2xx_spi.h>
#include <mach/pxa-regs.h>
#include <mach/lis331.h>
#include <mach/akm8973.h>
#include <mach/ssp.h>
#include <mach/alsps.h>
#include <mach/micco_headset.h>
#include <mach/micco_vibrator.h>
#include <mach/gps.h>
#include <mach/i2c.h>
#include <mach/wlan.h>
#include <mach/ism2300s.h>
#include <mach/pxa3xx-aec-fm2010.h>
#include <mach/innodev.h>

#include "devices.h"
#include "generic.h"

#ifdef CONFIG_INPUT_KEYRESET 
#include <linux/keyreset.h>
#endif

#define MAX_SLOTS      3
struct platform_mmc_slot beacon_mmc_slot[MAX_SLOTS];
extern int is_android(void);
extern int is_comm_v75(void);

extern void enable_oscc_tout_s0(void);
extern void disable_oscc_tout_s0(void);

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

/* BEACON MFP configurations */
static mfp_cfg_t beacon_mfp_cfg[] __initdata = {
	/* GPIO FREE */
	GPIO61_FREE,
	GPIO62_FREE,
	GPIO83_FREE,
//	GPIO161_FREE,
//	GPIO163_FREE,
	/*GPIO111_FREE,*/
	GPIO112_FREE,
	GPIO118_FREE,
	GPIO119_FREE,
	GPIO120_FREE,
	GPIO175_FREE,	
//	GPIO121_FREE,
//	GPIO124_FREE,
//	GPIO127_FREE,

	/* IPC */
	BP_WD_OUT,
	DPRAM_INT_N,
	BP_READY,
	AP_READY,
	FF_CTS,
	FF_RTS,
	FF_RXD,
	FF_TXD,
	BP_ON,
	BP_SHUTDOWN,
	AP_RESET_BP,

	ONKEY,
	/* Keypad */
	KP_MKIN1,
	KP_MKOUT1,
	KP_MKIN2,
	KP_MKOUT2,
	KP_DKIN6,	/* Camear focus Key */
	KP_DKIN7,	/* Camera Key */

	/* OLED */	
	OLED_VSYNC,
	OLED_HSYNC,
	OLED_PCLK,
	OLED_DE,
	OLED_R0, 
	OLED_R1, 
	OLED_R2, 
	OLED_R3, 
	OLED_R4, 
	OLED_R5, 
	OLED_G0, 
	OLED_G1, 
	OLED_G2, 
	OLED_G3, 
	OLED_G4, 
	OLED_G5, 
	OLED_B0, 
	OLED_B1, 
	OLED_B2, 
	OLED_B3, 
	OLED_B4, 
	OLED_B5, 
	GPU_RESET_N,
	OLED_SPI_CLK,
	OLED_SPI_CS,
	OLED_SPI_D0,
	OLED_RESET_N,
	GPU_PWR_EN,
	
	/* gpio i2c bus */
	GPU_I2C_SCL,	/* OLED display buffer chip I2C */
	GPU_I2C_SDA,

	/* i2c bus */
	I2C_SCL,
	I2C_SDA,

	/* Misc */

	/* GPS */
	STRXD,
	STTXD,
	AGPS_RESET,
	AGPS_PON,
	GPS_PWR_EN,

	/* Audio */
	AEC_IRQ,
	I2C_SDA_ECHO,
	I2C_SCL_ECHO,
	AP_PCM_CLK_OFF,
	AP_PCM_FRM_OFF, 
	AP_PCM_TXD_OFF, 
	AP_PCM_RXD_OFF, 
	HIFI_CLK,
	HIFI_FRM,
	HIFI_TXD,
	AEC_RESET_N,
	EAR_POP,
	AUDIO_PA_EN,

	/* camera */	
	CAM_DATA7, 
	CAM_DATA6, 
	CAM_DATA5, 
	CAM_DATA4, 
	CAM_DATA3, 
	CAM_DATA2, 
	CAM_DATA1, 
	CAM_DATA0, 
	CAM_HSYNC, 
	CAM_VSYNC, 
	CAM_MCLK, 
	CAM_PCLK, 
	CAM_REAR_PDN,
	CAM_VGA_PDN,
	FLASH_EN,
	CAM_MCLK_SEL,
	CAM_REAR_RESET_N,
	CAM_PWR_EN,

	/*Pin for compass*/
	CSEN_RST_N,
	CSEN_DRDY,	
	/* PMIC */
	PMIC_INT_MICCO,

	/* BT/WIFI */
	WLAN_BT_RESET,
	AP_WAKEUP_BT,
	BT_RXD, 
	BT_TXD, 
	BT_CTS, 
	BT_RTS, 
	AP_WAKEUP_WLAN,
	WLAN_WAKEUP_AP,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	BT_WAKEUP_AP,
	WIFI_PMU_PON,
	WLAN_PON,

	/* MMC */
	MMC1_CD_N,
	MMC1_CMD, 
	MMC1_CLK, 
	MMC1_DAT0, 
	MMC1_DAT1, 
	MMC1_DAT2, 
	MMC1_DAT3,

	/* P/L sensor */
	PSEN_DOUT,
	PSEN_LEDON,
	PSEN_PD,

	/* G-Sensor */
	GSEN_INT,


	/* charger ic */
	CHARGE_EN2,
	CHARGE_EN1,
	CHARGE_PG_N,
	CHARGE_CE_N,
	CHARGE_STAT,

	/* CMMB */
	CMMB_IRQ,
	CMMB_RST_N,
	CMMB_SPI_CLK,
	CMMB_SPI_SFRM,
	CMMB_SPI_TXD,
	CMMB_SPI_RXD,
	CMMB_PWR_EN,
};
enum mach_version get_mach_version(void)
{
	extern unsigned int system_rev;

	printk("Hardware ID: 0x%x\n", system_rev);

	switch (system_rev) {
	case MACH_VERSION_W00:
	case MACH_VERSION_P00:
	case MACH_VERSION_P01:
	case MACH_VERSION_P02:
	case MACH_VERSION_P10:
	case MACH_VERSION_P11:
	case MACH_VERSION_P15:
	case MACH_VERSION_P20:
	case MACH_VERSION_P30:
	case MACH_VERSION_P40:
	case MACH_VERSION_M10:
		break;
	default:
		printk(KERN_ERR "Unknown Version, Change from 0x%x to 0x%x!\n",
				system_rev, MACH_VERSION_P01);
		system_rev = MACH_VERSION_P01;
	}

	return system_rev;
}
 
static struct platform_device micco_bl_device = {
	.name 		= "micco-bl",
	.id 		= -1,
};

static struct platform_device micco_ts_device = {
	.name 		= "micco-ts",
	.id 		= -1,
};

static struct micco_vibrator_platform_data beacon_vibrator_data = {
	.vib_strength = (0x33 << 1),
};

static struct platform_device micco_vibrator_device = {
	.name 		= "micco-vibrator",
	.id 		= -1,
	.dev		= {
		.platform_data = &beacon_vibrator_data,
	},
};

#ifdef CONFIG_INPUT_KEYRESET 
static int keys_up[] = {
	KEY_VOLUMEDOWN,
	0,
};

static struct keyreset_platform_data keyreset_pdata = {
	.keys_up	= keys_up,
	.keys_down	= {KEY_BACK, KEY_MENU, KEY_VOLUMEUP, KEY_CAMERA, 0},
};

static struct platform_device keyreset_device = {
	.name 		= "keyreset",
	.id 		= -1,
	.dev		= {
		.platform_data = &keyreset_pdata,
	},
};
#endif

#if defined(CONFIG_MICCO_HEADSET_DETECT)
#define HS_DET_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(HS_DET_N))
#define HS_DET_N_GPIO_3 mfp_to_gpio(MFP_CFG_PIN(HS_DET_N_3))
#define HS_UART_SW_GPIO	mfp_to_gpio(MFP_CFG_PIN(HS_UART_SW))
static struct mutex ear_pop_mutex;

static void hs_uart_switch(bool uart)
{
	if (uart)
		gpio_direction_output(HS_UART_SW_GPIO, 1);
	else
		gpio_direction_output(HS_UART_SW_GPIO, 0);
}

#define HEADSET_UART_ADC_DET 4
static int beacon_uart_detecting(void)
{	
	int val;
	int hs_gval, uart_gval, gpio_irq;
	u8 stu;
	static int adc_inited=0;

	switch (get_mach_version()) {
		case MACH_VERSION_W00:
		case MACH_VERSION_P00:
		case MACH_VERSION_P01:
		case MACH_VERSION_P02:
		case MACH_VERSION_P10:
		case MACH_VERSION_P11:
			if(gpio_get_value(HS_DET_N_GPIO))
				return 1;
			else
				return 0;

			break;
		case MACH_VERSION_P15:
			if( !adc_inited )
			{
				micco_auto_adc_poweron(PMIC_EVENT_ADC4);
				adc_inited=1;
			}
			val = micco_adc_read(HEADSET_UART_ADC_DET);
			printk("micco_adc_read value = %d\n",val);
			if (350 < val && val < 450) {
				hs_uart_switch(1);
				return 1;
			} else {
				hs_uart_switch(0);
				return 0;
			}
			break;
		default:
			if (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)))) {
				mutex_lock(&ear_pop_mutex);
				gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), 0);
				msleep(300);
				hs_gval = gpio_get_value(HS_DET_N_GPIO_3);
				gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), 1);
				pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EAR_POP), MFP_LPM_DRIVE_HIGH);
				mutex_unlock(&ear_pop_mutex);
			} else 
				hs_gval = gpio_get_value(HS_DET_N_GPIO_3);
			micco_read(MICCO_STATUS_B, &stu);
			gpio_irq = gpio_to_irq(HS_DET_N_GPIO_3);
			uart_gval = gpio_get_value((HS_UART_SW_GPIO));
			if ((stu & MICCO_STATUS_B_HEADSET) && hs_gval) {
				if (!uart_gval) {
					disable_irq(gpio_irq);
					hs_uart_switch(1);
				}
				return 1;
			} else {
				if (uart_gval) {
					hs_uart_switch(0);
					enable_irq(gpio_irq);
				}
				return 0;
			}
	}
}

static void beacon_headset_insert(void)
{
	mutex_lock(&ear_pop_mutex);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), 0);
	pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EAR_POP), MFP_LPM_DRIVE_LOW);
	mutex_unlock(&ear_pop_mutex);
}

static void beacon_headset_remove(void)
{
	mutex_lock(&ear_pop_mutex);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), 1);
	pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EAR_POP), MFP_LPM_DRIVE_HIGH);
	mutex_unlock(&ear_pop_mutex);
}

static struct micco_headset_platform_data beacon_headset_platform_data = {
	.gpio_detect = HS_DET_N_GPIO_3,
	.detecting_headset = beacon_uart_detecting,
	.insert	= beacon_headset_insert,
	.remove = beacon_headset_remove,
};

static int ear_pop_init(void)
{
    int ret = 0;

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), "EAR_POP");
    if (ret) {
        printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "EAR_POP", 
                                              mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), ret);
        return -1;
    }

    gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(EAR_POP)), 1);
    pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EAR_POP), MFP_LPM_DRIVE_HIGH);

    printk(KERN_INFO "<%s> GPIOs setup done\n", "EAR_POP");

    return 0;
}

static struct platform_device micco_headset_device = {
	.name		= "micco-headset",
	.id		= -1,
	.dev = {
		.platform_data = &beacon_headset_platform_data,
	},
};

static void __init beacon_headset_init(void)
{
	mutex_init(&ear_pop_mutex);
	ear_pop_init();
	platform_device_register(&micco_headset_device);
}
#else
static inline void beacon_headset_init(void) {}
#endif /* CONFIG_MICCO_HEADSET_DETECT */

static mfp_cfg_t beacon_W00_mfp_cfg[] __initdata = {
	HS_DET_N,
};

static mfp_cfg_t beacon_P15_mfp_cfg[] __initdata = {
	IPC_UART_OE,
	HS_DET_N_2,
	HS_UART_SW,
};

static mfp_cfg_t beacon_P20_mfp_cfg[] __initdata = {
	IPC_UART_OE,
	HS_DET_N_3,
	HS_UART_SW,
};

static void beacon_misc_mfp_config(void)
{
    	switch (get_mach_version()) {
		case MACH_VERSION_W00:
		case MACH_VERSION_P00:
		case MACH_VERSION_P01:
		case MACH_VERSION_P02:
		case MACH_VERSION_P10:
		case MACH_VERSION_P11:
			pxa3xx_mfp_config(ARRAY_AND_SIZE(beacon_W00_mfp_cfg));
			gpio_request(MFP_CFG_PIN(HS_DET_N),"headset");
			gpio_direction_input(MFP_CFG_PIN(HS_DET_N));
			break;
		case MACH_VERSION_P15:
			pxa3xx_mfp_config(ARRAY_AND_SIZE(beacon_P15_mfp_cfg));
			gpio_request(MFP_CFG_PIN(HS_UART_SW), "hs uart sw");
			gpio_direction_output(MFP_CFG_PIN(HS_UART_SW), 0);
			/* this is for disable BP uart. it is from P1.5 */
			gpio_request(MFP_CFG_PIN(IPC_UART_OE), "ipc uart oe");
			gpio_direction_output(MFP_CFG_PIN(IPC_UART_OE), 0);
			break;
		default:
			pxa3xx_mfp_config(ARRAY_AND_SIZE(beacon_P20_mfp_cfg));
			gpio_request(MFP_CFG_PIN(HS_UART_SW), "hs uart sw");
			gpio_direction_output(MFP_CFG_PIN(HS_UART_SW), 0);
			/* this is for disable BP uart. it is from P1.5 */
			gpio_request(MFP_CFG_PIN(IPC_UART_OE), "ipc uart oe");
			gpio_direction_output(MFP_CFG_PIN(IPC_UART_OE), 0);
	}

}

#ifdef CONFIG_IMM
static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		/* Moved up to offset 64KB due to collision with DVFM
		 * D2 entry code 
		 */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		/* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa935_resource_imm),
	.resource	= pxa935_resource_imm,
};
#endif

#ifdef CONFIG_I2C_GPIO
static struct i2c_gpio_platform_data i2c_gpio_data = {
	.sda_pin	= mfp_to_gpio(MFP_CFG_PIN(GPU_I2C_SDA)),
	.scl_pin	= mfp_to_gpio(MFP_CFG_PIN(GPU_I2C_SCL)),
	.udelay		= 5,
};
static struct platform_device i2c_gpio_device = {
	.name		= "i2c-gpio",
	.id		= 2, //this id is very important device=>adapter=>bus
	.dev		= {
		.platform_data = &i2c_gpio_data,
	},
};

static struct i2c_gpio_platform_data i2c_gpio2_data = {
	.sda_pin	= mfp_to_gpio(MFP_CFG_PIN(I2C_SDA_ECHO)),
	.scl_pin	= mfp_to_gpio(MFP_CFG_PIN(I2C_SCL_ECHO)),
	.udelay		= 5,
};
static struct platform_device i2c_gpio2_device = {
	.name		= "i2c-gpio2",
	.id		= 3, //this id is very important device=>adapter=>bus
	.dev		= {
		.platform_data = &i2c_gpio2_data,
	},
};

static void __init beacon_i2c_gpio_init(void)
{
	platform_device_register(&i2c_gpio_device);
	platform_device_register(&i2c_gpio2_device);
}
#else
static inline void beacon_i2c_gpio_init(void) {}
#endif

#if defined(CONFIG_BQ2407x_POWER)
#define CHARGE_EN1_GPIO		mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN1))
#define CHARGE_EN2_GPIO		mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN2))
#define CHARGE_PG_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_PG_N))
#define CHARGE_CE_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_CE_N))
#define CHARGE_STAT_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_STAT))

static int beacon_bq2407x_charger_init(void)
{
	int err;

	err = gpio_request(CHARGE_EN1_GPIO, "charger en1");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en1 error!\n");
		return err;
	}
	err = gpio_request(CHARGE_EN2_GPIO, "charger en2");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en2 error!\n");
		return err;
	}
	err = gpio_request(CHARGE_PG_N_GPIO, "charger poog");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger poog error!\n");
		return err;
	}
	err = gpio_request(CHARGE_CE_N_GPIO, "charger en");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en error!\n");
		return err;
	}
	err = gpio_request(CHARGE_STAT_GPIO, "charger det");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger det error!\n");
		return err;
	}

	gpio_direction_output(CHARGE_EN1_GPIO, 1);
	gpio_direction_output(CHARGE_EN2_GPIO, 1);
	gpio_direction_input(CHARGE_PG_N_GPIO);
	gpio_direction_output(CHARGE_CE_N_GPIO, 1);
	gpio_direction_input(CHARGE_STAT_GPIO);

	return 0;
}

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
#define WAITING_USB_READY_INTERVAL	200
#define WAITING_USB_READY_TIMES		5

extern void pxa_charger_decide_enable(int on);
extern int is_usb_driver_busy(void);
static struct completion charger_detect_complete;
#endif

static int beacon_bq2407x_charger_detect(void)
{
	int ret = BQ2407x_POWER_CHARGE_USB;
	int retry_times = 0;
	unsigned long timeout = 0;

	if (gpio_get_value(CHARGE_PG_N_GPIO)) {
		printk(KERN_DEBUG "%s: No charger detected\n", __func__);
		ret = BQ2407x_POWER_NO_CHARGER;
	} else {
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
		while (is_usb_driver_busy()) {
			printk(KERN_DEBUG "%s: waiting for usb status busy "
					  "(retry times=%d\n",
					  __func__, retry_times);

			msleep(WAITING_USB_READY_INTERVAL);
			if (WAITING_USB_READY_TIMES <= ++retry_times) {
				printk(KERN_DEBUG "%s: USB regisered\n",
					   __func__);
				goto out;
			}
		}
			
		INIT_COMPLETION(charger_detect_complete);

		pxa_charger_decide_enable(1);
		enable_irq(IRQ_U2P);
		timeout = wait_for_completion_timeout(&charger_detect_complete,
					msecs_to_jiffies(100));

		if (timeout) {
			ret = BQ2407x_POWER_CHARGE_AC;
			printk(KERN_DEBUG "%s: charger (AC) detected\n",
					   __func__);
		} else {
			disable_irq(IRQ_U2P);
			ret = BQ2407x_POWER_CHARGE_USB;
			printk(KERN_DEBUG "%s: charger (USB) detected\n",
					   __func__);
		}
		pxa_charger_decide_enable(0);
#endif
	}
out:
	return ret; 
}

static int beacon_bq2407x_is_charging(void)
{
	return (gpio_get_value(CHARGE_STAT_GPIO) ? 0 : 1);
}

static void beacon_bq2407x_set_charge(int enable)
{
	if (enable == BQ2407x_POWER_ENABLE) {
		printk(KERN_DEBUG "%s: charger enable\n", __func__);
		gpio_direction_output(CHARGE_CE_N_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_CE_N),
				   MFP_LPM_DRIVE_LOW);	
	}
	else {
		printk(KERN_DEBUG "%s: charger disable\n", __func__);
		gpio_direction_output(CHARGE_CE_N_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_CE_N),
				   MFP_LPM_DRIVE_HIGH);	
	}
}

static void beacon_bq2407x_set_current(int flags)
{
	switch(flags) {
	case CHARGING_SUSPEND:
		printk(KERN_DEBUG "%s: current suspend\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 1);
		gpio_direction_output(CHARGE_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	case CHARGING_CURRENT_100:
		printk(KERN_DEBUG "%s: current 100\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_500:
		printk(KERN_DEBUG "%s: current 500\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 1);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_FULL:
		printk(KERN_DEBUG "%s: current full\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	default:
		printk(KERN_DEBUG "%s: current default\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	}
}

static int bq2407x_is_cable_on(void)
{
	return gpio_get_value(CHARGE_PG_N_GPIO) ? 0 : 1;
}

static void beacon_bq2407x_charger_exit(void)
{
        gpio_free(CHARGE_EN1_GPIO);
        gpio_free(CHARGE_EN2_GPIO);
        gpio_free(CHARGE_PG_N_GPIO);
        gpio_free(CHARGE_CE_N_GPIO);
        gpio_free(CHARGE_STAT_GPIO);
}

static char *beacon_power_supplicants[] = {
	"battery",
};

static struct bq2407x_charger_pdata beacon_bq2407x_charger_pdata = {
	.init		= beacon_bq2407x_charger_init,
	.detect		= beacon_bq2407x_charger_detect,
	.is_charging	= beacon_bq2407x_is_charging,
	.is_cable_on	= bq2407x_is_cable_on,
	.set_charge	= beacon_bq2407x_set_charge,
	.set_current	= beacon_bq2407x_set_current,
	.exit		= beacon_bq2407x_charger_exit,
	.irq		= IRQ_GPIO(CHARGE_PG_N_GPIO),
	.det_irq	= IRQ_GPIO(CHARGE_STAT_GPIO),
	.supplied_to 	= beacon_power_supplicants,
	.num_supplicants = ARRAY_SIZE(beacon_power_supplicants),
};

static struct platform_device bq2407x_charger_device = {
	.name		= "charger",
	.id		= -1,
	.dev = {
		.platform_data = &beacon_bq2407x_charger_pdata,
	},
};

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
static irqreturn_t charger_detected_irq(int irq, void *dev_id)
{
	printk(KERN_DEBUG "charger detected via U2O D+/D- irq\n");
	disable_irq(IRQ_U2P);
	complete(&charger_detect_complete);
	return IRQ_HANDLED;
}
#endif

static void beacon_init_charger(void)
{
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	if(request_irq(IRQ_U2P, charger_detected_irq, 
		       IRQF_DISABLED | IRQF_TRIGGER_FALLING,
		       "charger_detect", NULL) < 0) {
        	printk("charger_detected_irq register failed\n");
	}
	disable_irq(IRQ_U2P);
	init_completion(&charger_detect_complete);
#endif

	platform_device_register(&bq2407x_charger_device);
}

#endif /* CONFIG_BQ2407x_POWER */

#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
#define BP_WD_OUT_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_WD_OUT))
#define BP_READY_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_READY))
#define BP_ON_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_ON))
#define BP_SHUTDOWN_GPIO	mfp_to_gpio(MFP_CFG_PIN(BP_SHUTDOWN))
#define AP_READY_GPIO		mfp_to_gpio(MFP_CFG_PIN(AP_READY))
#define AP_RESET_BP_GPIO	mfp_to_gpio(MFP_CFG_PIN(AP_RESET_BP))
#define DPRAM_INT_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(DPRAM_INT_N))

static int beacon_ipc_handshake_init(void)
{
	if (gpio_request(AP_RESET_BP_GPIO, "ap reset bp")) {
		pr_err("%s: can not get gpio %d\n", __func__, AP_RESET_BP_GPIO);
	} 
	gpio_direction_output(AP_RESET_BP_GPIO, 1);

	if (gpio_request(BP_ON_GPIO, "bp on")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_ON_GPIO);
	} 
	gpio_direction_output(BP_ON_GPIO, 0);

	if (gpio_request(BP_SHUTDOWN_GPIO, "bp shutdown")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_ON_GPIO);
	} 
	gpio_direction_output(BP_SHUTDOWN_GPIO, 0);

	if (gpio_request(AP_READY_GPIO, "ap ready")) {
		pr_err("%s: can not get gpio %d\n", __func__, AP_READY_GPIO);
	}
	gpio_direction_output(AP_READY_GPIO, 0);

	if (gpio_request(BP_READY_GPIO, "bp ready")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_READY_GPIO);
	}
	gpio_direction_input(BP_READY_GPIO);

	if (gpio_request(BP_WD_OUT_GPIO, "bp watch dog")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_WD_OUT_GPIO);
	}
	if (gpio_request(DPRAM_INT_N_GPIO, "ipc mailbox")) {
		pr_err("%s: can not get gpio %d\n", __func__, DPRAM_INT_N_GPIO);
	}
	gpio_direction_input(DPRAM_INT_N_GPIO);
	return 0;
}

static int beacon_ipc_handshake_exit(void)
{
	gpio_free(BP_READY_GPIO);
	gpio_free(DPRAM_INT_N_GPIO);
	gpio_free(AP_READY_GPIO);
	gpio_free(AP_READY_GPIO);
	gpio_free(BP_ON_GPIO);
	gpio_free(BP_WD_OUT_GPIO);
	return 0;
}

static struct clk *ipc_smc_clk = NULL;
static unsigned char *ipc_smc_membase = NULL;
static u16 *ipc_ram_addr = NULL;
static enum state_type beacon_config_ipc(enum config_ipc_cmd cmd)
{
	enum state_type ret = READY; 

	switch (cmd) {
		case GET_BP_READY:
			ret = gpio_get_value(BP_READY_GPIO) ? READY : NOT_READY;
			break;
		case RESET_BP:
			pr_info("%s: Reset BP\n", __func__);
			gpio_direction_output(AP_RESET_BP_GPIO, 1);

			gpio_direction_output(BP_ON_GPIO, 1);
			msleep(70);
			gpio_direction_output(BP_ON_GPIO, 0);

			gpio_direction_output(AP_RESET_BP_GPIO, 0);
			ret = READY;
			break;
		case SET_AP_READY:
			gpio_direction_output(AP_READY_GPIO, 1);
			ret = READY;
			break;
		case SET_AP_NOT_READY:
			gpio_direction_output(AP_READY_GPIO, 0);
			ret = NOT_READY;
			break;
		case SET_AP_SLEEP:
			break;
		case SET_AP_RESUME:
			break;
		case GET_MAILBOX_STATE:
			if (!gpio_get_value(DPRAM_INT_N_GPIO))
				ret = NOT_READY;
			break;
		case POWERON_BP:
			pr_info("%s: Power on BP\n", __func__);
			gpio_direction_output(BP_ON_GPIO, 1);
			msleep(70);
			gpio_direction_output(BP_ON_GPIO, 0);
			gpio_direction_output(AP_RESET_BP_GPIO, 0);
			
			break;
		case SHUTDOWN_BP:
			/* Power off BP */
			pr_info("%s: Power off BP\n", __func__);
			gpio_direction_output(BP_SHUTDOWN_GPIO, 1);
			udelay(50);
			gpio_direction_output(BP_SHUTDOWN_GPIO, 0);
			break;
		default:
			pr_err("%s: unknow config command: %d\n", __func__, cmd);
			break;
	}
	return ret;
}


/*****************************************************************************
 * The SMC registers read/write routines
 *****************************************************************************/
static unsigned int smc_reg_read(u8 *base, int offset)
{
	return __raw_readl(base + offset);
}

static inline void smc_reg_write(u8 *base, int offset, unsigned long val)
{
	unsigned long flags;

	local_irq_save(flags);
	writel(val, base + offset);
	__raw_readl(base + offset);
	local_irq_restore(flags);
}

#define SMC_ADDR_START		0x4A000000	/* SMC Registers*/
#define SMC_CSMSADRCFG		0xA0		/* CSMSADRCFG */
#define SMC_MEMCLKCFG		0x68		/* MEMCLKCFG */
#define SMC_CSADRCFG		0x8C		/* CSADRCFG */
#define SMC_MSC1		0xC		/* MSC1 */
#define SMC_ADDR_LEN		0xFF

#define DFI_CS3_IOBASE 	0x14000000
#define RAM_TOTAL	32*1024

static void beacon_ipc_smc_init(void)
{
	unsigned int val;

	if (!ipc_smc_clk || !ipc_smc_membase) {
		pr_err("%s: clk is NULL\n", __func__);
		return; 
	}

	clk_enable(ipc_smc_clk);

	/*CSMSADRCFG = 0x2  */
	smc_reg_write(ipc_smc_membase, SMC_CSMSADRCFG, 0x2);
	val = smc_reg_read(ipc_smc_membase, SMC_CSMSADRCFG);

	/*config CSADRCFG3  */
	smc_reg_write(ipc_smc_membase, SMC_CSADRCFG, 0x00364919);

	/*config MSC1  */
	val = smc_reg_read(ipc_smc_membase, SMC_MSC1);
	val &= ~0xffff0000;
	val |= 0x05590000;
	val &= ~0xf000f000;/* 31-28, 15-12 bits reserved, write 0 */
	smc_reg_write(ipc_smc_membase, SMC_MSC1, val);
}

static void beacon_ipc_smc_exit(void)
{

}

static struct dpram_platform_data cypress_dpram_data;

static int beacon_ipc_misc_init(void)
{
	int ret;
	if (!(ipc_smc_clk = clk_get(NULL, "SMCCLK"))) {
		pr_err("%s: get SMCCLK failed\n", __func__);
		ret = -EINVAL;
		return ret;
	}

	clk_enable(ipc_smc_clk);
	/*SMC controller registers address map*/
	ipc_smc_membase = ioremap(SMC_ADDR_START, SMC_ADDR_LEN);
	if (!ipc_smc_membase) {
		pr_err("%s: remap SMC address failed\n", __func__);
		ret = -ENOMEM;
		goto err0;
	}
	ipc_ram_addr = (u16 *)ioremap_nocache(DFI_CS3_IOBASE, RAM_TOTAL);
	if (!ipc_ram_addr) {
		pr_err("%s: remap ram address failed\n", __func__);
		ret = -ENOMEM;
		goto err1;
	}
	cypress_dpram_data.ram_addr = ipc_ram_addr;
	return 0;
err1:
	iounmap(ipc_smc_membase);
	ipc_smc_membase = NULL;
err0:
	clk_disable(ipc_smc_clk);
	clk_put(ipc_smc_clk);
	ipc_smc_clk = NULL;
	return ret;
}

static int beacon_ipc_misc_exit(void)
{	
	if (ipc_smc_membase) {
		iounmap(ipc_smc_membase);
		ipc_smc_membase = NULL;
	}
	if (ipc_ram_addr) {
		iounmap(ipc_ram_addr);
		cypress_dpram_data.ram_addr = NULL;
	}
	if (ipc_smc_clk) {
		clk_disable(ipc_smc_clk);
		clk_put(ipc_smc_clk);
	}
	return 0;
}

static struct dpram_platform_data cypress_dpram_data = {
	.bp_wd_out_irq = IRQ_GPIO(MFP_CFG_PIN(BP_WD_OUT)),
	.bp_rdy_irq = IRQ_GPIO(MFP_CFG_PIN(BP_READY)),
	.mailbox_irq = IRQ_GPIO(MFP_CFG_PIN(DPRAM_INT_N)),
	.config_ipc = beacon_config_ipc,
};

static struct platform_device cypress_dpram_device = {
	.name 		= "cypress_dpram",
	.id 		= -1,
	.dev = {
		.platform_data = &cypress_dpram_data,
	},
};

static void beacon_BP_init(void)
{
	beacon_ipc_handshake_init();
	beacon_ipc_misc_init();
	beacon_ipc_smc_init();
	beacon_config_ipc(POWERON_BP);

}
#endif

#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65_FOR_TD)
static void snd_ssp_mfp_cfg_init(void)
{
    extern mfp_cfg_t pcm_on_mfp_cfg[4];
    extern mfp_cfg_t pcm_off_mfp_cfg[4];
    extern mfp_cfg_t i2s_on_mfp_cfg[3];
    extern mfp_cfg_t i2s_off_mfp_cfg[3];

    pcm_on_mfp_cfg[0] = AP_PCM_CLK;
    pcm_on_mfp_cfg[1] = AP_PCM_FRM;
    pcm_on_mfp_cfg[2] = AP_PCM_TXD;
    pcm_on_mfp_cfg[3] = AP_PCM_RXD;

    pcm_off_mfp_cfg[0] = AP_PCM_CLK_OFF;
    pcm_off_mfp_cfg[1] = AP_PCM_FRM_OFF;
    pcm_off_mfp_cfg[2] = AP_PCM_TXD_OFF;
    pcm_off_mfp_cfg[3] = AP_PCM_RXD_OFF;

    i2s_on_mfp_cfg[0] = HIFI_CLK;
    i2s_on_mfp_cfg[1] = HIFI_FRM;
    i2s_on_mfp_cfg[2] = HIFI_TXD;

    i2s_off_mfp_cfg[0] = MFP_CFG_LPM(GPIO85, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[1] = MFP_CFG_LPM(GPIO87, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[2] = MFP_CFG_LPM(GPIO88, AF0, PULL_LOW);
}
#endif

#if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2010) || defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2010_MODULE)
typedef struct {
	int irq;
	int reset;    
} S_AEC_PIN;

static S_AEC_PIN g_aecPins;
static int g_aecInit = 0;

static int pxa3xx_aec_reset(void)
{
    printk("[AEC] aec reset\n");
	//gpio_direction_output(g_aecPins.reset, 0);
	//msleep(10);
	gpio_direction_output(g_aecPins.reset, 0);
	msleep(20);
	gpio_direction_output(g_aecPins.reset, 1);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.reset), MFP_LPM_DRIVE_HIGH);
	msleep(20);
	return 0;
}

static int pxa3xx_aec_wakeup(void)
{ 
	printk("[AEC] aec wakeup\n");
	pxa3xx_pmic_enable_voltage(VCC_AEC, 0);
	msleep(800);//Add for 13M clock 
	pxa3xx_pmic_enable_voltage(VCC_AEC, 1);
	msleep(20);

	gpio_direction_output(g_aecPins.reset, 0);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.reset), MFP_LPM_DRIVE_LOW);

	gpio_direction_output(g_aecPins.irq, 0);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_LOW);

	return 0; 
}

static int pxa3xx_aec_sleep(void)
{
	printk("[AEC] aec sleep...\n");

	msleep(20);
	pxa3xx_pmic_enable_voltage(VCC_AEC, 0);

	gpio_direction_output(g_aecPins.irq, 0);
	gpio_direction_output(g_aecPins.reset, 0);

	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_PULL_LOW);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.reset), MFP_LPM_PULL_LOW);

	return 0;
}

static int pxa3xx_aec_init(void)
{
	int ret;     
	printk("[AEC] init aec pins\n");

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AEC_IRQ)), "AEC_IRQ");
	BUG_ON(ret);

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N)), "AEC_RESET_N");
	BUG_ON(ret);

	g_aecPins.irq = mfp_to_gpio(MFP_CFG_PIN(AEC_IRQ));
	g_aecPins.reset = mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N));

	g_aecInit = 1;

	gpio_direction_output(g_aecPins.irq, 0);
	gpio_direction_output(g_aecPins.reset, 0);

	return 0;
}

static int pxa3xx_aec_set_mode(int mode)
{
	if(g_aecInit == 0) {
		printk("[AEC] device not init\n");
		return 0;
	}

	if (mode == AEC_BYPASS_MODE) {
		//gpio_direction_output(g_aecPins.irq, 0);
		//msleep(1);
		gpio_direction_output(g_aecPins.irq, 1);
	    pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_HIGH);
	} else if (mode == AEC_NORMAL_MODE) {
		//gpio_direction_output(g_aecPins.irq, 1);
		//msleep(1);
		gpio_direction_output(g_aecPins.irq, 0);
		pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_LOW);
	} else
		WARN_ON(1);
	return 0;
}

static const aec_reg_t g_bypass_tbl[] = {
     {0x0F84 ,0x0080}
    ,{0x1E30 ,0x0221}
    ,{0x1E41 ,0x0001}
    ,{0x1E4A ,0x000E}
    ,{0x1E52 ,0x000B}
    ,{0x1E60 ,0x0000}
    ,{0x1E70 ,0x05C0}
    ,{0x1ED4 ,0x3800}
    ,{0x1E3A ,0x0000}
    ,{0x0,0x0} /* End of table. */
};

static const aec_reg_t g_normal_tbl[] = {
     {0x0F84, 0x0080}
    ,{0x1E34, 0x00AA}
    ,{0x1E41, 0x0002}
    ,{0x1E45, 0x43DE}
    ,{0x1E46, 0x0035}
    ,{0x1E47, 0x3000}
    ,{0x1E48, 0x0800}
    ,{0x1E49, 0x0800}
    ,{0x1E52, 0x0013}
    ,{0x1E60, 0x0000}
    ,{0x1E63, 0x0001}
    ,{0x1E70, 0x05C0}
    ,{0x1E86, 0x0008}
    ,{0x1E87, 0x0004}
    ,{0x1E8C, 0x0380}
    ,{0x1E9A, 0x0900}
    ,{0x1E9B, 0x6800}
    ,{0x1E9C, 0x3200}
    ,{0x1EC5, 0x0800}
    ,{0x1ED4, 0x3800}
    ,{0x1ED6, 0x1800}
    ,{0x1ED7, 0x0600}
    ,{0x1EDA, 0x3600}
    ,{0x1EDF, 0x4000}
    ,{0x1EED, 0x0900}
    ,{0x1EEE, 0x3800}
    ,{0x1E3A, 0x0000}
    ,{0x0,0x0} /* End of table. */
};

static const aec_reg_t* pxa3xx_aec_get_parm(int mode)
{
	const aec_reg_t* ptr;

	switch (mode) {
		case AEC_BYPASS_MODE:
			ptr = g_bypass_tbl;
			break;

		case AEC_NORMAL_MODE:
		default:
			ptr = g_normal_tbl;
			break;
	}		

	return ptr;
}

static struct aec_platform_data g_aec_platform_data = {
	.init       = pxa3xx_aec_init,	
	.wakeup     = pxa3xx_aec_wakeup,
	.sleep      = pxa3xx_aec_sleep,
	.reset      = pxa3xx_aec_reset,
	.set_mode   = pxa3xx_aec_set_mode,
	.parmGet    = pxa3xx_aec_get_parm,
};

static struct i2c_board_info beacon_i2c_aec_info[] = {
	{
		.type       = "avp_fm2010",
		.addr       = 0x60,
		.platform_data  = &g_aec_platform_data,
	},
};

static void __init beacon_i2c_avp_fm2010_init(void)
{
	switch (get_mach_version()) 
    {
	    case MACH_VERSION_W00:
	    case MACH_VERSION_P00:
	    case MACH_VERSION_P01:
	    case MACH_VERSION_P02:
            i2c_register_board_info(0, ARRAY_AND_SIZE(beacon_i2c_aec_info));
		    break;
	    default:;
#ifdef CONFIG_I2C_GPIO
		    i2c_register_board_info(3, ARRAY_AND_SIZE(beacon_i2c_aec_info));
#endif
            break;
	}
}
#endif /* #if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2010)  */

struct platform_device usb_mass_storage_device = {
	.name		= "usb_mass_storage",
	.id		= -1,
};

static struct platform_device *devices[] __initdata = {
	&micco_bl_device,
	&micco_ts_device,
	&micco_vibrator_device,
#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
	&cypress_dpram_device,
#endif
#ifdef CONFIG_IMM
	&pxa3xx_device_imm,
#endif
#ifdef CONFIG_INPUT_KEYRESET
	&keyreset_device,
#endif
	&usb_mass_storage_device,
};

#if defined(CONFIG_SIRF_GSD3TW)
static struct sirf_platform_data beacon_sirf_platform_data = {
	.reset_pin = mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)),
	.on_off_pin = mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)),
};

static struct platform_device beacon_sirf_device = {
	.name = "sirf",
	.id = -1,
	.dev = {
		.platform_data = &beacon_sirf_platform_data,
	},
};

static void __init beacon_gps_sirf_init(void)
{
	printk(KERN_ERR"[%s:%d] gps pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)), "gps reset")) {
		printk(KERN_ERR "<SiRF> GPIO RST request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)), "gps on_off")) {
		printk(KERN_ERR "<SiRF> GPIO ON_OFF request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_PWR_EN)), "gps power on")) {
		printk(KERN_ERR "<SiRF> GPIO PWR_ON request error!!\n");
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_PWR_EN)), 1);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)), 1);
	printk(KERN_ERR"GPS set up finished\n");
	platform_device_register(&beacon_sirf_device);
}
#endif

#if defined(CONFIG_ALSPS)

#define PROXIMITY_SHUTDOWN 	MFP_CFG_PIN(PSEN_LEDON)
#define ALSPS_POWERDOWN 	MFP_CFG_PIN(PSEN_PD)
#define PROXIMITY_OUTPUT 	MFP_CFG_PIN(PSEN_DOUT)

static int beacon_alsps_init(void)
{
	int ret;

	/*config ambient light sensor switch pin, also as chip switch pin*/	
	ret = gpio_request(mfp_to_gpio(ALSPS_POWERDOWN), "ambient_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset ambient_powerdown failed\n");
		goto out;
	}
	gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);

	/*config proximity sensor switch pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_SHUTDOWN), "proximity_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_powerdown failed\n");
		goto out1;
	}
	gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);

	/*config proximity sensor output pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_OUTPUT), "proximity_output");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_output failed\n");
		goto out2;
	}
	gpio_direction_input(mfp_to_gpio(PROXIMITY_OUTPUT));

	return 0;

out2:
	gpio_free(mfp_to_gpio(PROXIMITY_SHUTDOWN));
out1:
	gpio_free(mfp_to_gpio(ALSPS_POWERDOWN));
out:
	printk("alsps: config gpio pins failed!\n");
	return ret;
}

static int beacon_alsps_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(PROXIMITY_OUTPUT)) ? 1 : 0;
}

static void beacon_alsps_setpower(int enable)
{
	if (enable & PROXIMITY_ENABLE) {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 0);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_HIGH);
	}

	if (enable) {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 0);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_HIGH);
	}
}

static int beacon_alsps_chip_verify(int* alsps, int size)
{
	if ((!alsps) || (size != 3))
		return -1;
	alsps[0] = gpio_get_value(PROXIMITY_SHUTDOWN);	/* proximity shutdown pin output */
	alsps[1] = gpio_get_value(PROXIMITY_OUTPUT);	/* proximity output value output */
	alsps[2] = gpio_get_value(ALSPS_POWERDOWN);	/* whole chip powerdown pin output */

	return 0;
}

static struct alsps_platform_data beacon_alsps_platform_data = {
	.irq		= IRQ_GPIO(mfp_to_gpio(PROXIMITY_OUTPUT)),
	.adc_event	= PMIC_EVENT_ADC5, 
	.init		= beacon_alsps_init, 
	.get_gpio_value	= beacon_alsps_get_gpio_value, 
	.setpower	= beacon_alsps_setpower,
	.chip_verify	= beacon_alsps_chip_verify,
};

static struct platform_device beacon_alsps_device = {
	.name = "alsps",
	.id = -1,
	.dev = {
		.platform_data = &beacon_alsps_platform_data,
	},
};

static void __init beacon_cm3601_alsps_init(void)
{
	platform_device_register(&beacon_alsps_device);
}
#endif

#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
static unsigned int beacon_matrix_key_map[] = {
	/* KEY(row, col, key_code) */
	KEY(1, 1, KEY_MENU),
	KEY(2, 1, KEY_BACK),
	KEY(2, 2, KEY_VOLUMEUP),
	KEY(1, 2, KEY_VOLUMEDOWN),
};

static unsigned int beacon_direct_key_map[] = {
	0,0,0,0,0,0,KEY_FOCUS,KEY_CAMERA,
};

static struct pxa27x_keypad_platform_data beacon_keypad_info = {
	.matrix_key_rows	= 3,
	.matrix_key_cols	= 3,
	.matrix_key_map		= beacon_matrix_key_map,
	.matrix_key_map_size	= ARRAY_SIZE(beacon_matrix_key_map),
	.direct_key_num		= ARRAY_SIZE(beacon_direct_key_map),
	.direct_key_map		= beacon_direct_key_map,
	.debounce_interval	= 30,
#ifdef CONFIG_DEBUG_TRIGGER_PANIC
    .trigger_key0       = KEY_MENU,
    .trigger_key1       = KEY_BACK,
#endif
};
#endif /* CONFIG_KEYBOARD_PXA27x || CONFIG_KEYBOARD_PXA27x_MODULE */

#if defined(CONFIG_MMC)
static int beacon_mci_init(struct device *dev,
		irq_handler_t beacon_detect_int,
		void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(beacon_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = beacon_mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for beacon MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, beacon_detect_int,
			IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			"MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void beacon_mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(beacon_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = beacon_mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static mfp_cfg_t sd_in_cfg[]={
	//SD_LDO_CE_ON,
	MMC1_CMD_ON,
	MMC1_DAT0_ON,
	MMC1_DAT1_ON,
	MMC1_DAT2_ON,
	MMC1_DAT3_ON,
	MMC1_CLK_ON,
};

static mfp_cfg_t sd_out_cfg[]={
	//SD_LDO_CE,
	MMC1_CMD,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	MMC1_CLK,
};

static void beacon_mci_setpower(struct device *dev, unsigned int vdd)
{
	struct pxamci_platform_data* p_d = dev->platform_data;

	if (( 1 << vdd) & p_d->ocr_mask) {
		pxa3xx_pmic_enable_voltage(VCC_SDIO, 1);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_in_cfg));
		msleep(1);
	} else {
        	//gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 0);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_out_cfg));
		pxa3xx_pmic_enable_voltage(VCC_SDIO, 0);
	}
}

static struct pxamci_platform_data beacon_mci_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
//	.gpio_detect    = mfp_to_gpio(MFP_CFG_PIN(MMC1_CD_N)),
	.init 		= beacon_mci_init,
	.setpower	= beacon_mci_setpower,
	.exit		= beacon_mci_exit,
};

static struct pxamci_platform_data beacon_mci2_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
};

static void __init beacon_init_mmc(void)
{
	if (device_is_enabled(device_mmc)) {
		pxa_set_mci_info(&beacon_mci_platform_data);
		pxa3xx_set_mci2_info(&beacon_mci2_platform_data);
	}
}
#else
static inline void beacon_init_mmc(void) {}
#endif

#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
static struct mtd_partition beacon_partitions[] = {
	[0] = {
		.name        = "Bootloader",
		.offset      = 0,
		.size        = 0x100000,
		.mask_flags  = MTD_WRITEABLE, /* force read-only */
	},
	[1] = {
		.name        = "security",
		.offset      = 0x100000,
		.size        = 0x040000,
		.mask_flags  = MTD_WRITEABLE, /* force read-only */
	},
	[2] = {
		.name        = "cpImage",
		.offset      = 0x140000, 
		.size        = 0x800000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},	
	[3] = {
		.name        = "NVM",
		.offset      = 0x940000, 
		.size        = 0x0c0000,
	},
	[4] = {
		.name        = "logo",
		.offset      = 0xa00000, 
		.size        = 0x040000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	[5] = {
		.name        = "Kernel",
		.offset      = 0xa40000,
		.size        = 0x300000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	[6] = {
		.name        = "system",
		.offset      = 0x0d40000,
		.size        = 0x9800000,     /* mount 128M fs */
	},
	[7] = {
		.name        = "opl",
		.offset      = 0xa540000,     /* mount 32M fs */
		.size        = 0x3000000,
	},
	[8] = {
		.name        = "local",
		.offset      = 0xd540000,
		.size        = 0x1000000,     /* mount 10M fs */
	},
	[9] = {
		.name        = "fota",
		.offset      = 0xe540000,
		.size        = 0x0600000,     /* mount 10M fs */
	},
	[10] = {
		.name        = "cache",       /* mount 8.725M fs */
		.offset      = 0xeb40000,
		.size        = 0x08c0000,
	},
	[11] = {
		.name        = "ftm",       /* mount 8.725M fs */
		.offset      = 0xf400000,
		.size        = 0x0600000,
	},
	[12] = {
		.name        = "panic",       /* mount 8.725M fs */
		.offset      = 0xfa00000,
		.size        = 0x0080000,
	},
	[13] = {
		.name        = "userdata",
		.offset      = 0x10000000,
		.size        = 0xf500000,     /* mount 54M fs */
	},

	[14] = {
		.name        = "BBT",
		.offset      = 0x1f580000,
		.size        = 0x0080000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	/* NOTES: We reserve some blocks for PXA3xx BBM at the end of NAND.
	 * And the max relocation blocks is not same on different platform.
	 * Please take care it when define the partition table.
	 */
};
static struct flash_platform_data beacon_onenand_info;
static void __init beacon_init_onenand(void)
{
	beacon_onenand_info.parts = beacon_partitions;
	beacon_onenand_info.nr_parts =	ARRAY_SIZE(beacon_partitions);

	pxa3xx_device_onenand.dev.platform_data = &beacon_onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#endif


#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info beacon_u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
static int micco_init_irq(void)
{
	gpio_direction_input(PMIC_INT_PIN);
	return 0;
}

static int micco_ack_irq(void)
{
	return 0;
}

static void beacon_micco_init(void)
{
	u8 value;

	/* Mask interrupts that are not needed */
	micco_write(MICCO_IRQ_MASK_A, 0xFC);
	micco_write(MICCO_IRQ_MASK_B, 0xFF);
	micco_write(MICCO_IRQ_MASK_C, 0x3F);
	micco_write(MICCO_IRQ_MASK_D, 0x3F);


	/* avoid SRAM power off during sleep*/
	value = MICCO_LDO2_ENABLE | MICCO_BUCK1_ENABLE;
	micco_write(MICCO_OVER1, value);

	/* enable all the GSM/WCDMA/EDGE related module */
	value = MICCO_LDO7_ENABLE | MICCO_LDO8_ENABLE | MICCO_LDO10_ENABLE |
		MICCO_LDO11_ENABLE;
	micco_write(MICCO_APP_OVER2, value);

	value = MICCO_LDO1_ENABLE | MICCO_LDO3_ENABLE | MICCO_LDO5_ENABLE | 
		MICCO_LDO13_ENABLE | MICCO_BUCK3_ENABLE;
	micco_write(MICCO_APP_OVER3, value);

	/* disable the ONKEY power down functionality */
	/* micco_write(MICCO_SYSCTRL_B, 0x20); */
	micco_write(MICCO_SYSCTRL_A, 0x60);

	/* Enable BUCK & LDO sleep setting */
	value = MICCO_BUCK1_nSLEEP_ENABLE  | MICCO_BUCK3_nSLEEP_ENABLE;
	micco_write(MICCO_BUCK_SLEEP, value);

	//value = MICCO_LDO8_nSLEEP_ENABLE;
	value = 0;
	micco_write(MICCO_LDO_nSLEEP1, value);

	value = 0;
	micco_write(MICCO_LDO_nSLEEP2, value);

	value = MICCO_LDO8_SLEEP_BIT_ENABLE;
	micco_write(MICCO_LDO_SLEEP_BIT1, value);

	value = 0;
	micco_write(MICCO_LDO_SLEEP_BIT2, value);

	value = MICCO_BUCK1_USE_LDO8_ENABLE | MICCO_BUCK3_USE_LDO8_ENABLE;
	micco_write(MICCO_BUCK_LDO8_EN, value);

	/* Set Buck3 as 2.2V to fix INNODEV stall bug */
	micco_write(MICCO_3DTV1, 0x04);
	micco_write(MICCO_VCC2, 0x01);

	/* IRQ is masked during the power-up sequence and will not be released
	 * until they have been read for the first time */
	micco_read(MICCO_EVENT_A, &value);
	printk("%s: MICCO_EVENT_A=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_B, &value);
	printk("%s: MICCO_EVENT_B=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_C, &value);
	printk("%s: MICCO_EVENT_C=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_D, &value);
	printk("%s: MICCO_EVENT_D=0x%x\n", __FUNCTION__, value);
}

/* micco_power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module miccoEA_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		0},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,		0},
	{VCC_SDIO,		LDO13},
	{VCC_CAMERA_ANA,	0},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO10},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,		LDO2},
	{VCC_1P8V,		0},
	{VCC_MEM,		LDO2},
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
	{VCC_BT,        	0},
	{VCC_AEC,       	LDO4},
};

static struct power_chip micco_chips[] = {
	{MICCO_EA_ID,	"miccoEA",	miccoEA_power_modules},
	{MICCO_EB_ID,	"miccoEB",	miccoEA_power_modules},
	{0,		NULL,		NULL},
};

static struct micco_platform_data micco_data = {
	.init_irq = micco_init_irq,
	.ack_irq = micco_ack_irq,
	.platform_init = beacon_micco_init,
	.power_chips = micco_chips,
};
#endif /* CONFIG_PXA3xx_MICCO || CONFIG_PXA3xx_MICCO_MODULE*/
#if	defined(CONFIG_SENSORS_AK8973)
#define COMPASS_INT_GPIO	mfp_to_gpio(MFP_CFG_PIN(CSEN_DRDY))
#define COMPASS_IRQ             IRQ_GPIO(COMPASS_INT_GPIO)
#define COMPASS_RESET_GPIO	mfp_to_gpio(MFP_CFG_PIN(CSEN_RST_N))

//We should add the other pin's request function

static int compass_init_irq(void)
{
	int ret = 0;
	ret = gpio_request(COMPASS_INT_GPIO, "compass irq");
	if (ret){
		printk("gpio_requset failed, return :%d\n",ret);
		goto err_request_cd;
	}
	ret=gpio_direction_input(COMPASS_INT_GPIO);
	if (ret) {
		printk("set gpio pin direction for compass failed, return :%d\n", ret);
		goto err_request_irq;
	}
err_request_irq:
	gpio_free(COMPASS_INT_GPIO);
err_request_cd:
	return ret;
}

static int compass_ack_irq(void)
{
	return 0;
}

static int compass_reset_pin(void)
{
	int ret = 0;
	ret = gpio_request(COMPASS_RESET_GPIO, "compass reset");
	if (ret) {
		printk("compass reset gpio_requset failed, return :%d\n",ret);
		return ret;	
	}
	gpio_direction_output(COMPASS_RESET_GPIO, 0);
	msleep(20);
	gpio_direction_output(COMPASS_RESET_GPIO, 1);
	msleep(20);

	return ret;
}

static void compass_fac_reset(void)
{
	printk(KERN_DEBUG "AKM8973 E-Compass factory test reset pin!\n");
	gpio_direction_output(COMPASS_RESET_GPIO, 0);
	msleep(20);
	gpio_direction_output(COMPASS_RESET_GPIO, 1);
	msleep(20);
}

static struct akm8973_platform_data akm8973_data = {
	.init_irq = compass_init_irq,
	.ack_irq = compass_ack_irq,
	.reset_pin = compass_reset_pin,
	.fac_reset = compass_fac_reset,
};
#endif

#if defined(CONFIG_SENSORS_LIS331)
static int lis331_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT)));

	return 0;
}

static int lis331_ack_irq(void)
{
	return 0;
}

static int lis331_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT))) ;
}

static struct lis331_platform_data lis331_data = {
	.init_irq = lis331_init_irq,
	.ack_irq = lis331_ack_irq,
	.get_gpio_value = lis331_get_gpio_value,
	.platform_init = NULL,
	.power_chips = NULL,
	.xyz_dir = (Y_POS | Z_POS | SWITCH_XY),
};
#endif

#if defined(CONFIG_IMAGIS_ISM2300S)
static void ism2300s_init(void)
{
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_PWR_EN)), 1);
#if !defined(CONFIG_OBM_LOGO_DISPLAY)
	msleep(10);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_RESET_N)), 0);
	msleep(10);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_RESET_N)), 1);
#endif
}
	
static struct imagis_platform_data ism2300s_data = {
	.platform_init = ism2300s_init,
};

static struct i2c_board_info beacon_i2c_ism2300s_info[] = {
	{
		.type           = "ism2300s",
		.addr           = 0x77, //SDAA LOW: 0x76; SDAA HIGH:0x77
		.platform_data  = &ism2300s_data,
	},
};

static void __init beacon_init_ism2300s(void)
{
	switch (get_mach_version()) {
	case MACH_VERSION_W00:
		i2c_register_board_info(0, ARRAY_AND_SIZE(beacon_i2c_ism2300s_info));
		break;
	default:;
#ifdef CONFIG_I2C_GPIO
		i2c_register_board_info(2, ARRAY_AND_SIZE(beacon_i2c_ism2300s_info));
#endif
	}
}
#else
static inline void beacon_init_ism2300s(void) {}
#endif

#if defined(CONFIG_PXA_CAMERA)

/* sensor init */
static int sensor_voltage_on(int flag)
{
	printk(KERN_ERR"[%s:%d] Camera power on\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN)), "CAM_PWR_EN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_PWR_EN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), "CAM_VGA_PDN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_VGA_PDN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), "CAM_REAR_PDN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_REAR_PDN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), "CAM_REAR_RESET_N")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_REAR_RESET_N));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), "CAM_MCLK_SEL")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_MCLK_SEL));
		return -EIO;
	}
	/* OV3640 power up with internal DVDD:
 	 *
	 * PWDN pin shall be high before power up.
	 * Before DOVDD is up, RESET# pin is better keep at low level.
	 * for PWDN to go low, power up must first become stable (AVDD to PWDN > 5 ms)
	 * H/W RESET. After power up stable , reset signal should be more than 3 mS.
	 */

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), 0);
	/* PWDN pin shall be high before power up. */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 1);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 1);
	mdelay(1);

	/* I/O & Analog power up */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN)), 1);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), 1);
	/* delay to wait power stable */
//	msleep(1);

	/* for PWDN to go low, power up must first become stable (AVDD to PWDN > 5 ms) */
	mdelay(1);
//	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 0);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 0);

	/* t1: from powers on to Reset pull high, >= 3ms */
	mdelay(1);
	/* assert RESET */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), 1);

	/* wait a moment then power down */
//	msleep(10);
//	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 1);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 1);

	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN)));	/* CAM_PWR_EN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)));	/* CAM_VGA_PDN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)));	/* CAM_REAR_PDN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)));	/* CAM_REAR_RESET_N */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)));	/* CAM_REAR_RESET_N */

	return 0;
}

static int sensor_power_on(int flag)
{
	printk(KERN_ERR"[%s:%d] Camera power on %d\n", __FILE__, __LINE__, flag);
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), "CAM_REAR_PDN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_REAR_PDN));
		return -EIO;
	}

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), "CAM_VGA_PDN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_VGA_PDN));
		return -EIO;
	}

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), "CAM_MCLK_SEL")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_MCLK_SEL));
		return -EIO;
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), 0);
//	msleep(10);
	if (!flag) {
		gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 0);
	} else {
		gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)), 1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 0);
	}
	msleep(5);
	return 0;
}

static int sensor_power_off(int flag)
{
	if (!flag) {
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN)), 1);

	} else {
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN)), 1);

	}
	msleep(5);
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL)));	/* CAM_MCLK_SEL */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN))); /* CAM_VGA_PDN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN))); /* CAM_REAR_PDN */
	return 0;
}

static int sensor_reset(int flag)
{
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), "CAM_REAR_RESET_N")) {
		printk( KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)) );
		return -EIO;
	}
	/* assert RESET */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), 0);
	msleep(5);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)), 1);
	msleep(5);
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N)));	/* CAM_REAR_RESET_N */
	return 0;
}

static int cam_flash_control(int mode)
{
	printk(KERN_INFO"spark: camera flash control(%d)\n", mode);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)), "FLASH_EN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(FLASH_EN));
		return -EIO;
	}

	if (mode)
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)), 1);
	else
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)), 0);
	msleep(5);
	gpio_free(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)));

	return 0;
}

static struct sensor_platform_data ov7690_sensor_data = {
	.id = SENSOR_LOW,
	.power_on = sensor_power_on,
	.power_off = sensor_power_off,
	/*.reset = sensor_reset,
	.flash_ctrl = cam_flash_control,*/
};
static struct sensor_platform_data ov5642_sensor_data = {
	.id = SENSOR_HIGH,
	.power_on = sensor_power_on,
	.power_off = sensor_power_off,
	.reset = sensor_reset,
	.flash_ctrl = cam_flash_control,
};

/* sensor init over */

/* camera platform data */
static mfp_cfg_t sync[] = {
	CAM_HSYNC,
	CAM_VSYNC,
};

static mfp_cfg_t sync_gpio[] = {
	CAM_HSYNC_GPIO,
	CAM_VSYNC_GPIO,
};

static void cam_sync_to_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync_gpio));
}

static void cam_sync_from_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync));
}

static int cam_init(void)
{
	printk(KERN_ERR"[%s:%d] %s()\n", __FILE__, __LINE__, __func__);
	return 0;
}

static void cam_deinit(void)
{
}

static void cam_suspend(void)
{
	sensor_power_off(0);
}

static void cam_resume(void)
{
	sensor_power_on(0);
}

static struct cam_platform_data cam_ops = {
	.vsync_gpio		= mfp_to_gpio(MFP_CFG_PIN(CAM_VSYNC_GPIO)),
	.init			= cam_init,
	.deinit			= cam_deinit,
	.suspend		= cam_suspend,
	.resume			= cam_resume,
	.sync_to_gpio		= cam_sync_to_gpio,
	.sync_from_gpio		= cam_sync_from_gpio,
};

static void __init beacon_init_cam(void)
{
	printk(KERN_ERR"[%s:%d] Camera pin init\n", __FILE__, __LINE__);

	if( gpio_request(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)), "FLASH_EN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(FLASH_EN));
		return;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)), 0);	/* FLASH_TORCH */

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(FLASH_EN)));	/* FLASH_TORCH */

	/* shall remove power on ov3640 if OBM do it when phone power up */
	sensor_voltage_on(0);

	pxa3xx_device_cam.dev.platform_data = &cam_ops;
	platform_device_register(&pxa3xx_device_cam);
}
/* QCI init over */
#endif

static struct i2c_board_info beacon_i2c_board_info[] = {
#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
	{
		.type		= "micco",
		.addr		= 0x34,
		.platform_data	= &micco_data,
		.irq		= IRQ_GPIO(PMIC_INT_PIN),
	},
#endif
#if defined(CONFIG_PXA_CAMERA)
	{
		.type		= "sensor_ov7690",
		.addr		= 0x21,
		.platform_data	= &ov7690_sensor_data,
	},
	{
		.type   = "sensor_ov5642",
		.addr           = 0x3C,
		.platform_data  = &ov5642_sensor_data,
	},
#endif
#if defined(CONFIG_SENSORS_AK8973)
	{
		.type	= "akm8973",
		.addr	= 0x1F,
		.platform_data	= &akm8973_data,
		.irq		= COMPASS_IRQ,
	},
#endif
#if defined(CONFIG_SENSORS_LIS331)
	{
		.type		= "lis331",
		.addr		= 0x1C,
		.platform_data	= &lis331_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT))),
	},
#endif
};

#if defined(CONFIG_MRVL8688) || defined(CONFIG_MRVL8688_MODULE)
static mfp_cfg_t wlan_lpm_gpio_on_mfp_cfg[] = {
    WLAN_PON_LPM_ON,
    WIFI_PMU_PON_LPM_ON,
    WLAN_BT_RESET_LPM_ON,
    WLAN_WAKEUP_AP_LPM_ON,
    AP_WAKEUP_BT_LPM_ON,
    BT_TXD_LPM_ON,
    BT_RXD_LPM_ON,
    BT_CTS_LPM_ON,
    BT_RTS_LPM_ON,
    BT_WAKEUP_AP_LPM_ON,
    WLAN_DAT3_LPM_ON,
    WLAN_DAT2_LPM_ON,
    WLAN_DAT1_LPM_ON,
    WLAN_DAT0_LPM_ON,
    WLAN_CMD_LPM_ON,
    WLAN_CLK_LPM_ON,
    AP_WAKEUP_WLAN_LPM_ON,
};

static mfp_cfg_t wlan_lpm_gpio_off_mfp_cfg[] = {
    WIFI_PMU_PON,
    WLAN_PON,
    WLAN_BT_RESET,
    WLAN_WAKEUP_AP,
    AP_WAKEUP_BT,
    BT_TXD,
    BT_RXD,
    BT_CTS,
    BT_RTS,
    BT_WAKEUP_AP,
    WLAN_DAT3,
    WLAN_DAT2,
    WLAN_DAT1,
    WLAN_DAT0,
    WLAN_CMD,
    WLAN_CLK,
    AP_WAKEUP_WLAN,
//    AP_PCM_CLK_OFF,
//    AP_PCM_FRM_OFF,
//    AP_PCM_TXD_OFF,
//    AP_PCM_RXD_OFF,
};

static int beacon_wlan_enabled = 0;
static void beacon_wlan_gpio_setpower(int enable)
{
	unsigned int pcm_clk, pcm_sync, pcm_rxd, pcm_txd;

	if (enable) {
		enable_oscc_tout_s0();
		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_on_mfp_cfg));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 1);
		msleep(1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
		msleep(10);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), 1);

		beacon_wlan_enabled = 1;
		printk(KERN_DEBUG "wlan_lpm_gpio_on\n");
	}
	else {
		pcm_clk = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_CLK));
		pcm_sync = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_FRM));
		pcm_txd = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_TXD));
		pcm_rxd = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_RXD));

		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_off_mfp_cfg));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)),0);
        	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)),0);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_CLK)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_FRM)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_TXD)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_RXD)),0);

 		msleep(10);

		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_CLK), pcm_clk);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_FRM), pcm_sync);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_TXD), pcm_txd);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_RXD), pcm_rxd);

		beacon_wlan_enabled = 0;
		disable_oscc_tout_s0();
		printk(KERN_DEBUG "wlan_lpm_gpio_off\n");
	}
}

static struct wlan_platform_data beacon_wlan_pdata = {
	.chip_gpio = 1,
	.chip_gpio_setpower = beacon_wlan_gpio_setpower,
        .bt_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP))),
        .wlan_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP))),
}; 

static struct platform_device wlan_device = {
	.name		= "mrvl8688",
	.id		= -1,
};

static void __init beacon_init_wlan(void)
{

	int err = 0;

	printk(KERN_ERR "WLAN set up\n");

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), "wlan pmic powerdown");	
	if (err) {
		printk(KERN_ERR "WIFI_PMU_PON Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), "wlan powerdown");
	if (err) {
		printk(KERN_ERR "WIFI_PMU_PON Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), "wlan bt reset");	
	if (err) {
		printk(KERN_ERR "WLAN_BT_RESET Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), "wlan mac wakeup");	
	if (err) {
		printk(KERN_ERR "AP_WAKEUP_WLAN Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)), "bt wakeup tavor");  
	if (err) {
		printk(KERN_ERR "BT_WAKEUP_AP Request error!\n");
		return;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), 0);

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_RXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_TXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_CTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_RTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT3)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT2)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT1)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT0)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CLK)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CMD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)),0);
//	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_ONE_WIRE)),0);

	wlan_device.dev.platform_data = &beacon_wlan_pdata;
	platform_device_register(&wlan_device);

	printk(KERN_ERR "WLAN set up finished\n");

	return;
}
#endif

#if defined(CONFIG_BT_HCIUART_PS)

#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
static void switch_bt_chip_lpm_state(int state)
{
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)), 
			      !!state);
        pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AP_WAKEUP_BT),
			   !!state ? MFP_LPM_DRIVE_HIGH : MFP_LPM_DRIVE_LOW);
}
#endif

static struct hci_ps_platform_data beacon_hci_ps_pdata = {
	.chip_type = CHIP_MRVL8688,
	.chip_setpower = NULL,
}; 

static struct platform_device hci_ps_device = {
	.name		= "hci_ps",
	.id		= -1,
};

static void __init beacon_hci_ps_init(void)
{
#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
	beacon_hci_ps_pdata.chip_setpower = switch_bt_chip_lpm_state;
#endif
	hci_ps_device.dev.platform_data = &beacon_hci_ps_pdata;
	platform_device_register(&hci_ps_device);
}
#endif

struct pxa3xx_freq_mach_info beacon_freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};

#define CMD0(x)		((0x0000) | ((x)))
#define DAT0(x)		((0x0100) | ((x)))
#define BEACON_OLED_DSS 9 	//Data size select for LCD spi
#define LUMI_LEVEL	10       //Brightness level setting
#define PARAMETER_NUM 	22

/* Initializing Sequence */
static uint32_t panel_set[] = {
	CMD0(0xF8),
	DAT0(0x01),
	DAT0(0x27),
	DAT0(0x27),
	DAT0(0x07),
	DAT0(0x07),
	DAT0(0x54),
	DAT0(0x9F),
	DAT0(0x63),
	DAT0(0x86),
	DAT0(0x1A),
	DAT0(0x33),
	DAT0(0x0D),
	DAT0(0x00),
	DAT0(0x00),
};

static uint32_t display_set[] = {
	CMD0(0xF2),
	DAT0(0x02),
	DAT0(0x03),
	DAT0(0x1C),
	DAT0(0x10),
	DAT0(0x10),

	CMD0(0xF7),
	DAT0(0x03), 	//Reverse scan display
	DAT0(0x00),
	DAT0(0x00),
};

static uint32_t gamma_setting[] = {
	/* 100 nit */
	CMD0(0xFA),
	DAT0(0x02),
	DAT0(0x18),
	DAT0(0x08),
	DAT0(0x24),
	DAT0(0x6D),
	DAT0(0x4A),
	DAT0(0x3C),
	DAT0(0xBF),
	DAT0(0xBF),
	DAT0(0xB0),
	DAT0(0xB7),
	DAT0(0xBB),
	DAT0(0xAB),
	DAT0(0xCC),
	DAT0(0xCE),
	DAT0(0xC3),
	DAT0(0x00),
	DAT0(0x69),
	DAT0(0x00),
	DAT0(0x68),
	DAT0(0x00),
	DAT0(0x89),
};

static uint32_t gamma_set_update[] = {
	CMD0(0xFA),
	DAT0(0x03),
};

/* Power On Setting Sequence */
static uint32_t etc_set[] = {
	CMD0(0xF6),
	DAT0(0x00),
	DAT0(0x8C),
	DAT0(0x07),

	CMD0(0xB3),
	DAT0(0x0C),

	CMD0(0xB5),
	DAT0(0x2C),
	DAT0(0x12),
	DAT0(0x0C),
	DAT0(0x0A),
	DAT0(0x10),
	DAT0(0x0E),
	DAT0(0x17),
	DAT0(0x13),
	DAT0(0x1F),
	DAT0(0x1A),
	DAT0(0x2A),
	DAT0(0x24),
	DAT0(0x1F),
	DAT0(0x1B),
	DAT0(0x1A),
	DAT0(0x17),
	DAT0(0x2B),
	DAT0(0x26),
	DAT0(0x22),
	DAT0(0x20),
	DAT0(0x3A),
	DAT0(0x34),
	DAT0(0x30),
	DAT0(0x2C),
	DAT0(0x29),
	DAT0(0x26),
	DAT0(0x25),
	DAT0(0x23),
	DAT0(0x21),
	DAT0(0x20),
	DAT0(0x1E),
	DAT0(0x1E),

	CMD0(0xB7),
	DAT0(0x2C),
	DAT0(0x12),
	DAT0(0x0C),
	DAT0(0x0A),
	DAT0(0x10),
	DAT0(0x0E),
	DAT0(0x17),
	DAT0(0x13),
	DAT0(0x1F),
	DAT0(0x1A),
	DAT0(0x2A),
	DAT0(0x24),
	DAT0(0x1F),
	DAT0(0x1B),
	DAT0(0x1A),
	DAT0(0x17),
	DAT0(0x2B),
	DAT0(0x26),
	DAT0(0x22),
	DAT0(0x20),
	DAT0(0x3A),
	DAT0(0x34),
	DAT0(0x30),
	DAT0(0x2C),
	DAT0(0x29),
	DAT0(0x26),
	DAT0(0x25),
	DAT0(0x23),
	DAT0(0x21),
	DAT0(0x20),
	DAT0(0x1E),
	DAT0(0x1E),

	CMD0(0xB6),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x11),
	DAT0(0x22),
	DAT0(0x33),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x55),
	DAT0(0x55),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),

	CMD0(0xB8),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x11),
	DAT0(0x22),
	DAT0(0x33),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x55),
	DAT0(0x55),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),

	CMD0(0xB9),
	DAT0(0x2C),
	DAT0(0x12),
	DAT0(0x0C),
	DAT0(0x0A),
	DAT0(0x10),
	DAT0(0x0E),
	DAT0(0x17),
	DAT0(0x13),
	DAT0(0x1F),
	DAT0(0x1A),
	DAT0(0x2A),
	DAT0(0x24),
	DAT0(0x1F),
	DAT0(0x1B),
	DAT0(0x1A),
	DAT0(0x17),
	DAT0(0x2B),
	DAT0(0x26),
	DAT0(0x22),
	DAT0(0x20),
	DAT0(0x3A),
	DAT0(0x34),
	DAT0(0x30),
	DAT0(0x2C),
	DAT0(0x29),
	DAT0(0x26),
	DAT0(0x25),
	DAT0(0x23),
	DAT0(0x21),
	DAT0(0x20),
	DAT0(0x1E),
	DAT0(0x1E),

	CMD0(0xBA),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x11),
	DAT0(0x22),
	DAT0(0x33),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x44),
	DAT0(0x55),
	DAT0(0x55),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
	DAT0(0x66),
};

static uint32_t standby_off[] = {
	CMD0(0x11),
};

static uint32_t display_on[] = {
	CMD0(0x29),
};

static uint32_t standby_on[] = {
	CMD0(0x10),
};

static uint32_t gamma_reg[] = {
	CMD0(0xFA),
};

static uint32_t gamma_val[LUMI_LEVEL][PARAMETER_NUM] = {
/*
{01st,02nd,03rd,04th,05th,06th,07th,08th,09th,10th,11th,12th,13th,14th,15th,16th,17th,18th,19th,20th,21th,22th},
 */
{0x02,0x18,0x08,0x24,0x86,0x00,0x61,0xC9,0xC3,0xBB,0xC3,0xC5,0xB5,0xD3,0xD6,0xCB,0x00,0x41,0x00,0x3C,0x00,0x58},
{0x02,0x18,0x08,0x24,0x78,0x35,0x52,0xC4,0xC3,0xB5,0xBC,0xC0,0xAF,0xCF,0xD3,0xC7,0x00,0x57,0x00,0x51,0x00,0x74},
{0x02,0x18,0x08,0x24,0x73,0x4C,0x4C,0xC1,0xC2,0xB3,0xB8,0xBC,0xAA,0xCB,0xCF,0xC2,0x00,0x67,0x00,0x61,0x00,0x8A},
{0x02,0x18,0x08,0x24,0x6D,0x50,0x44,0xBE,0xC0,0xB0,0xB5,0xBA,0xA8,0xCA,0xCD,0xC0,0x00,0x73,0x00,0x6D,0x00,0x9A},
{0x02,0x18,0x08,0x24,0x68,0x56,0x42,0xBC,0xBE,0xAD,0xB3,0xB8,0xA6,0xC8,0xCB,0xBE,0x00,0x7E,0x00,0x78,0x00,0xA9},
{0x02,0x18,0x08,0x24,0x66,0x58,0x40,0xBA,0xBD,0xAB,0xB1,0xB6,0xA3,0xC6,0xCA,0xBD,0x00,0x88,0x00,0x81,0x00,0xB6},
{0x02,0x18,0x08,0x24,0x66,0x58,0x3E,0xB7,0xBC,0xA9,0xB0,0xB4,0xA2,0xC3,0xC7,0xBA,0x00,0x92,0x00,0x8B,0x00,0xC3},
{0x02,0x18,0x08,0x24,0x62,0x57,0x3A,0xB5,0xBA,0xA7,0xAE,0xB3,0xA1,0xC1,0xC5,0xB7,0x00,0x9B,0x00,0x93,0x00,0xCF},
{0x02,0x18,0x08,0x24,0x5E,0x5A,0x39,0xB5,0xB8,0xA6,0xAC,0xB2,0x9F,0xC0,0xC3,0xB6,0x00,0xA2,0x00,0x9B,0x00,0xD9},
{0x02,0x18,0x08,0x24,0x5E,0x5A,0x36,0xB2,0xB8,0xA5,0xAC,0xB0,0x9E,0xBE,0xC1,0xB4,0x00,0xAA,0x00,0xA3,0x00,0xE4},
};

static uint32_t *gamma_changing(int brightness)
{
	int i;
	static uint32_t new_gamma_val[PARAMETER_NUM];

	for (i = 0; i < PARAMETER_NUM; i++)
	      new_gamma_val[i] = DAT0(gamma_val[brightness][i]);

	return new_gamma_val;
}

static void ssp_reconfig(struct ssp_dev *dev, int nparam)
{
	static int last_nparam = -1;
	int sscr0;

	/* check if it is necessary to re-config SSP */
	if (nparam == last_nparam)
		return;

	ssp_disable(dev);
	if( nparam >16 )
		sscr0 = 0x00107f80 | ((nparam-1) & 0xf);
	else
		sscr0 = 0x00007f80 | ((nparam-1) & 0xf);

	ssp_config(dev, sscr0, 0x18, 0, 0);
	last_nparam = nparam;
}

static void ssp_send_cmd(uint32_t *cmd, int num)
{
	static int ssp_initialized = 0;
	static struct ssp_dev ssp4;

	int i;

	if (!ssp_initialized) {
		ssp_init(&ssp4, 4, SSP_NO_IRQ);
		ssp_initialized = 1;
		clk_disable(ssp4.ssp->clk);
	}

	clk_enable(ssp4.ssp->clk);
	for (i = 0; i < num; i++, cmd++) {
		ssp_reconfig(&ssp4, BEACON_OLED_DSS);
		ssp_write_word(&ssp4, *cmd & 0x3fffffff);

		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp4);
	}
	clk_disable(ssp4.ssp->clk);
}

#if defined(CONFIG_IMAGIS_ISM2300S)
static void imagis_display_on(void)
{
	int cnt;
	u8 val;

	/* 1. XTAL Enable & PLL Set */
	imagis_write(0x00,0x00);
	imagis_write(0x03,0x80);
	imagis_write(0x0C,0x12);
	imagis_write(0x0D,0x10);
	udelay(500);

	/* 2. IO Config. */
	/* Input Mode Configuration */
	imagis_write(0x10,0x00); 	//RGB888
//	imagis_write(0x10,0x01); 	//RGB666
	imagis_write(0x11,0x06); 	//24bit bus
//	imagis_write(0x11,0x05); 	//18bit bus
	imagis_write(0x14,0x91); 	//???
//	imagis_write(0x16,0x04); 	//andy: input data RGB -> BGR
#if 0
	imagis_write(0x20,0x01); 	//Auto Size Calculation Enable Mode: 01 - enable
#else
	imagis_write(0x20,0x00); 	//Auto Size Calculation Enable Mode: 00 - disable
	imagis_write(0x21,0x30); 	//2:2:2:2-> XX:VAS:VBP:VFP
	imagis_write(0x22,0x20); 	//Input VAS: Vertical Active Size
	imagis_write(0x23,0x03); 	//VBP
	imagis_write(0x24,0x1C); 	//VFP
	imagis_write(0x25,0x10); 	//2:2:2:2-> XX:HAS:HBP:HFP
	imagis_write(0x26,0xE0); 	//HAS
	imagis_write(0x27,0x10); 	//HBP
	imagis_write(0x28,0x10); 	//HFP
#endif

	/* Output Mode Configuration */
	imagis_write(0x42,0x00);
	imagis_write(0x43,0x06); 	//24bit bus output
//	imagis_write(0x43,0x05); 	//18bit bus output
//	imagis_write(0x44,0x04); 	//andy: output data RGB -> BGR
	imagis_write(0x45,0x01); 	//andy

	imagis_write(0x00,0x01);
//	imagis_write(0x70,0x13);
	imagis_write(0xA6,0x30); 	//2:2:2:2-> XX:VAS:VBP:VFP
	imagis_write(0xA7,0x20); 	//Output VAS: Vertical Active Size
	imagis_write(0xA8,0x03); 	//Output VBP
	imagis_write(0xA9,0x1C); 	//Output VFP
	imagis_write(0xAA,0x10); 	//2:2:2:2-> XX:HAS:HBP:HFP
	imagis_write(0xAB,0xE0); 	//Output HAS: Horizontal Active Size
	imagis_write(0xAC,0x10); 	//Output HBP
	imagis_write(0xAD,0x10); 	//Output HFP
	imagis_write(0xAE,0x02); 	//Output VSYNC
	imagis_write(0xAF,0x02); 	//Output HSYNC
//?	imagis_write(0x17,0x66);
//?	imagis_write(0x18,0x60);

	/* SDRAM Config */
	imagis_write(0xBB,0x12);
	imagis_write(0xBD,0x01);
	imagis_write(0xBF,0x1F);
	imagis_write(0xC0,0x0C);
	imagis_write(0xC1,0x86);
#if 0
	/* xView OFF */
	imagis_write(0x0A,0x00);
	imagis_write(0x90,0x10);
#else
	/* xView ON */
	imagis_write(0x17,0x66);
	imagis_write(0x18,0x60);
	imagis_write(0x20,0x80);
	imagis_write(0x21,0x00);
	imagis_write(0x70,0x13);
	imagis_write(0x80,0xF4);
	imagis_write(0xD0,0x00);
	imagis_write(0xD1,0x00);
	imagis_write(0x00,0x02);
	imagis_write(0x01,0x02);
	imagis_write(0x32,0x77);
	imagis_write(0x33,0xE3);
	imagis_write(0x35,0x97);
	imagis_write(0x36,0x09);
	imagis_write(0x37,0xBA);
	imagis_write(0x38,0x48);
	imagis_write(0x39,0x00);
	imagis_write(0x5A,0x12);
	imagis_write(0xE3,0xFF);
#endif

//	imagis_write(0x00,0x02);
//	imagis_write(0x5A,0x12); 	//TD OFF

	/* 3. Sub Block Ordering */
	imagis_write(0x00,0x00);
	imagis_write(0xF4,0x61);
	imagis_write(0xF6,0x40);
	imagis_write(0xF7,0x57);

	/* 4. Clock Config. */
	imagis_write(0x03,0x41);
	imagis_write(0x04,0x40);
	imagis_write(0x05,0x04);
//	imagis_write(0x06,0x44); 	//Invert Pixel Clk
	imagis_write(0x06,0x00);
	imagis_write(0x07,0x02);
	imagis_write(0x02,0x5D);
	imagis_write(0x0B,0x0C);

	for (cnt = 0;cnt < 3; cnt ++)
	{
		/* 5. S/W Reset */
		imagis_write(0x00,0x00);
		imagis_write(0xFF,0x01);
		imagis_write(0xFF,0x00);

		/* 6. Clock On & Mute Off */
		imagis_write(0x02,0x01);
		udelay(800);

		/* Check SDRST */
		imagis_write(0x00,0x01);
		imagis_read(0xCB,&val);
		if (val == 0x01)
		{
			/* Clock MUTE */
			imagis_write(0x00,0x00);
			imagis_write(0x02,0x5D);
			continue;
		}
	}
	/* Output Mute Off */
	imagis_write(0x00,0x00);
	imagis_write(0x41,0x00);
}

static void imagis_display_off(void)
{
	/* 1. OLED OFF */
	imagis_write(0x00,0x00);
	imagis_write(0x0B,0x00);

	/* 2. Output Pad Configure */
	imagis_write(0x41,0x03);

	/* 3. Clock Configure */
	imagis_write(0x04,0x80);
	imagis_write(0x03,0xA0);
	imagis_write(0x02,0xFF);

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_PWR_EN)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_RESET_N)), 0);
}
#else
static void imagis_display_on(void){}
static void imagis_display_off(void){}
#endif

static void beacon_backlight_power( int on, int intensity)
{
	static uint32_t *gamma_brightness;

	if( on) {
		gamma_brightness = gamma_changing((intensity * LUMI_LEVEL) >> 8);
		ssp_send_cmd(ARRAY_AND_SIZE(gamma_reg));
		ssp_send_cmd(gamma_brightness,PARAMETER_NUM);
		ssp_send_cmd(ARRAY_AND_SIZE(gamma_set_update));
	} else {
		printk("%s: backlight off\n",__func__);
	}
}

static void beacon_lcd_power(int on, struct fb_var_screeninfo *var)
{
	static int pin_requested = 0;

	if( !pin_requested) {
		if( gpio_request( mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 
			"OLED controller reset"))
			return;
		pin_requested = 1;
	}

	if( on)
	{
		printk("lcd power on ...\n");
		pxa3xx_pmic_enable_voltage(VCC_LCD,1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_PWR_EN)), 1);
		msleep(25);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_RESET_N)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 0);
        	msleep(1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPU_RESET_N)), 1);

		imagis_display_on();

		ssp_send_cmd(ARRAY_AND_SIZE(panel_set));
		ssp_send_cmd(ARRAY_AND_SIZE(display_set));
//		ssp_send_cmd(ARRAY_AND_SIZE(gamma_setting));
//		ssp_send_cmd(ARRAY_AND_SIZE(gamma_set_update));
		ssp_send_cmd(ARRAY_AND_SIZE(etc_set));
		ssp_send_cmd(ARRAY_AND_SIZE(standby_off));
		msleep(120);
		ssp_send_cmd(ARRAY_AND_SIZE(display_on));
	}
	else
	{
		printk("lcd power off ...\n");
		ssp_send_cmd(ARRAY_AND_SIZE(standby_on));
		msleep(120);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 0);
		pxa3xx_pmic_enable_voltage(VCC_LCD,0);
		imagis_display_off();
	}
}

static void beacon_config_panel(int enable)
{
	if( enable )
	{
		/* Change to moving pictures */
		imagis_write(0x00, 0x01);
		imagis_write(0x13, 0x32);
		imagis_write(0x18, 0x64);
		imagis_write(0x19, 0x23);
		imagis_write(0x20, 0x81);
		imagis_write(0x22, 0x30);
		imagis_write(0x25, 0x13);
		imagis_write(0x31, 0x3F);
		imagis_write(0x33, 0x20);
		imagis_write(0x34, 0x20);
		imagis_write(0x35, 0x20);
		imagis_write(0x36, 0x20);
		imagis_write(0x3B, 0x10);
		imagis_write(0x93, 0xE9);
	}else
	{
		/* Change to still pictures */
		imagis_write(0x00, 0x01);
		imagis_write(0x13, 0x42);
		imagis_write(0x18, 0x60);
		imagis_write(0x19, 0x24);
		imagis_write(0x20, 0x80);
		imagis_write(0x93, 0xE2);
	}
}

static struct pxafb_mode_info samsung_ams353pcxx_modes[] = {
	/* VBP = 3; 	VFP = 28; 	VLW = 2;
	 * HBP = 16; 	HFP = 16; 	HLW = 2;
	 * 60fps - 25.5MHz
	 */
	[0] = {
//		.pixclock		= 52227,
		.pixclock		= 39170,
		.xres			= 480,
		.yres			= 800,
		.bpp			= 32,
		.hsync_len		= 2,
		.left_margin		= 12, //Set HBP as 14 to fix display offset bug
		.right_margin		= 16,
		.vsync_len		= 3,
		.upper_margin		= 1,
		.lower_margin		= 28,
	},
};

static struct pxafb_mach_info beacon_lcd_info = {
	.modes			= samsung_ams353pcxx_modes,
	.num_modes		= 1,
	.lccr0			= LCCR0_Act,// | LCCR0_DELAY_LBIAS,
	.lccr3			= LCCR3_PixRsEdg | LCCR3_OEP | LCCR3_HSP | LCCR3_VSP,
	.pxafb_lcd_power	= beacon_lcd_power,
	.pxafb_backlight_power	= beacon_backlight_power,
	.pxafb_config_panel	= beacon_config_panel,
};

static void __init beacon_init_lcd(void)
{
	set_pxa_fb_info(&beacon_lcd_info);
}

#if defined(CONFIG_LEDS_PXA)
static void beacon_leds_brightness_set(int value)
{
	if (value) {/* Turn keyboard backlight on */
		micco_write(MICCO_LED1_CONTROL, 0xFF); 
	} else { /* Turn keyboard backlight off */
		micco_write(MICCO_LED1_CONTROL, 0); 
	}
}

static struct leds_platform_data beacon_leds_info = {
	.set_button_brightness 	= beacon_leds_brightness_set,
};
#endif

#if defined(CONFIG_LEDS_DEBUG_SLEEP)
static void beacon_sleep_led_brightness_set (int value)
{
	return;
}

static struct sleep_led_platform_data beacon_sleep_led_info = {
	.set_brightness 	= beacon_sleep_led_brightness_set,
};
#endif

#ifdef CONFIG_PM
static int beacon_init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
/*	src->bits.uart1 = 1; */
	src->bits.uart2 = 1;
	src->bits.dkey = 1;
	src->bits.onkey = 1;
	src->bits.mmc2_dat1 = 1; //Enable mmc2 wakeup 
	src->bits.usbc = 1;
//	src->bits.cmwdt = 1;
	src->bits.psensor = 1;
	src->bits.modem = 1;

	return 0;
}

extern int pxa930_query_gwsr(int);
static int beacon_query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_PMIC)
		src->bits.onkey = 1;
	if (reg & PXA3xx_PM_WE_UART1)
		src->bits.uart1 = 1;
	if (reg & PXA3xx_PM_WE_BTUART)
		src->bits.uart2 = 1;
	if (reg & PXA3xx_PM_WE_DKEY)
		src->bits.dkey = 1;
	if (reg & PXA3xx_PM_WE_MKEY)
		src->bits.mkey = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GPIO) {
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(PSEN_DOUT))))
			src->bits.psensor = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(DPRAM_INT_N))) || pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(BP_READY))))
			src->bits.modem = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP))))
		        src->bits.uart2 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP))))
			src->bits.mmc2_dat1 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(CHARGE_PG_N))))
			src->bits.usbc = 1;
	}

	return 0;
}

static int beacon_ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int beacon_key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.dkey) {
			ret |= PXA3xx_PM_WE_DKEY;
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_PMIC;
		}
	} else {
		if (src.bits.dkey) {
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int beacon_mmc_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;

	if( enable )
	{
		if( src.bits.mmc2_dat1 && beacon_wlan_enabled )
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_AP), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}
	else
	{
		if( src.bits.mmc2_dat1 )
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_AP), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int beacon_psensor_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PSEN_DOUT),MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}else{
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PSEN_DOUT),MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int beacon_modem_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if( src.bits.modem) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(DPRAM_INT_N),MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BP_READY),MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}else{
		if( src.bits.modem) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(DPRAM_INT_N),MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BP_READY),MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int beacon_uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(FF_RXD), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_UART1;
		}
		if (src.bits.uart2 && beacon_wlan_enabled ) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_AP), MFP_EDGE_FALL);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_RXD), MFP_EDGE_FALL);
			/* note: on pxa930, uart2 use this bit */
			ret |= PXA3xx_PM_WE_BTUART|PXA3xx_PM_WE_GPIO;
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(FF_RXD), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_AP), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_RXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int beacon_usb_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;

	if (enable) {
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(CHARGE_PG_N), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	} else {	
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(CHARGE_PG_N), MFP_EDGE_NONE);
		}
	}
	return ret;

}

static int beacon_comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= beacon_init_wakeup,
	.query	= beacon_query_wakeup,
	.ext    = beacon_ext_wakeup,
	.key    = beacon_key_wakeup,
	.mmc    = beacon_mmc_wakeup,
	.uart   = beacon_uart_wakeup,
	.usb	= beacon_usb_wakeup,
	.cmwdt	= beacon_comm_wdt_wakeup,
	.psensor = beacon_psensor_wakeup,
	.modem = beacon_modem_wakeup,
};
#endif

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
	.flags		= PXA_I2C_STANDARD_MODE, 
};

#if defined(CONFIG_INNODEV_IF202) || defined(CONFIG_INNODEV_IF202_MODULE)

#define INNO_SPI_CLK_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_CLK))
#define INNO_SPI_CS_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_SFRM))
#define INNO_SPI_TXD_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_TXD))
#define INNO_SPI_RXD_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_RXD))
#define INNO_EN_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_PWR_EN))
#define INNO_INTR_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_IRQ))
#define INNO_RESET_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_RST_N))

static mfp_cfg_t if202_ssp_mfp_cfg[] = {
	CMMB_SPI_CLK,
	CMMB_SPI_SFRM,
	CMMB_SPI_TXD,
	CMMB_SPI_RXD,
};

static mfp_cfg_t if202_ssp_gpio_cfg[] = {
	CMMB_SPI_CLK_GPIO,
	CMMB_SPI_SFRM_GPIO,
	CMMB_SPI_TXD_GPIO,
	CMMB_SPI_RXD_GPIO,
};

static int inno_if202_ssp_gpio_input(void)
{
	/* config ssp pins to gpio input to save power */
	gpio_direction_output(INNO_SPI_CLK_PIN,0);
	gpio_direction_output(INNO_SPI_CS_PIN,0);
	gpio_direction_output(INNO_SPI_TXD_PIN,0);
	gpio_direction_output(INNO_SPI_RXD_PIN,0);

	pxa3xx_mfp_config(ARRAY_AND_SIZE(if202_ssp_gpio_cfg));

	return 0;
}

static int inno_if202_gpio_init(void)
{
        if (gpio_request(INNO_SPI_CLK_PIN, "INNO_CLK")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_CLK_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_CS_PIN, "INNO_CS")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_CS_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_TXD_PIN, "INNO_TXD")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_TXD_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_RXD_PIN, "INNO_RXD")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_RXD_PIN);
                return -EIO;
        }

	if (gpio_request(INNO_INTR_PIN, "INNO_INTR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_INTR_PIN);
		return -EIO;
	}

	if (gpio_request(INNO_RESET_PIN, "INNO_RESET")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_RESET_PIN);
		return -EIO;
	}
	/* config reset pin to gpio input to save power */

	if (gpio_request(INNO_EN_PIN, "INNO_ENABLE")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_EN_PIN);
		return -EIO;
	}

	inno_if202_ssp_gpio_input();

	gpio_direction_output(INNO_RESET_PIN,0);

	gpio_direction_output(INNO_INTR_PIN,0);

	gpio_direction_output(INNO_EN_PIN, 0);	/* power voltage off */

	return 0;
}

int inno_if202_voltage_on(void)
{
	/* The if101 chip may not power up correctly if the connected pins are high,
	 * so config ssp pins to gpio input before power up */
	inno_if202_ssp_gpio_input();
	gpio_direction_input(INNO_INTR_PIN);

	gpio_direction_output(INNO_RESET_PIN, 0);

	/* enable VDD (1.8V) & VIO (2.8V) */
	gpio_direction_output(INNO_EN_PIN, 1);

	udelay(50);

	gpio_direction_output(INNO_RESET_PIN, 1);

	msleep(400);	/* TBD */

	/* reconfig ssp pins to SSP after power up */
	pxa3xx_mfp_config(ARRAY_AND_SIZE(if202_ssp_mfp_cfg));

	return 0;
}

int inno_if202_voltage_off(void)
{
	gpio_direction_output(INNO_EN_PIN, 0);

	/* config reset pin to gpio input to save power */
	gpio_direction_output(INNO_RESET_PIN,0);

	gpio_direction_output(INNO_INTR_PIN,0);
	/* config ssp pins to gpio input to save power */
	inno_if202_ssp_gpio_input();

	return 0;
}

static int inno_if202_reset(void)
{
	gpio_direction_output(INNO_RESET_PIN, 0);
	msleep(1);
	gpio_direction_output(INNO_RESET_PIN, 1);
	msleep(1);
	return 0;
}

static struct innodev_platform_spi inno_spi_data = {
	.ssp_port = 1,
	.spi_cs_gpio = INNO_SPI_CS_PIN,
};

static struct platform_device inno_spi_device = {
	.name 		= "inno-spi",
	.id 		= -1,
};

static struct innodev_platform_data inno_if202_data = {
	.voltage_on = inno_if202_voltage_on,
	.voltage_off = inno_if202_voltage_off,
	.reset = inno_if202_reset,
	.fw_spi_download = 1,
	.irq = IRQ_GPIO(INNO_INTR_PIN),
};

static struct platform_device inno_if202_device = {
	.name 		= "inno-dev",
	.id 		= -1,
};

static void __init inno_if2xx_init(void)
{
	inno_if202_gpio_init();

	inno_if202_data.fw_name = "inno_if208.fw";

	inno_if202_device.dev.platform_data = &inno_if202_data;
	platform_device_register(&inno_if202_device);

	inno_spi_device.dev.platform_data = &inno_spi_data;
	platform_device_register(&inno_spi_device);
}
#else
static inline void inno_if2xx_init(void) {}
#endif /* #ifdef CONFIG_INNODEV_V2 */

/**************************************************
 ** beacon_speaker_read: read the speaker PA status
 ** 1: enable
 ** 0: disable
 ***************************************************/ 
static ssize_t beacon_speaker_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	int value;

	value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN))) ? 1 : 0;

	if ((count = sprintf(buf, "%d", value)) < 0)
		return -EFAULT;

	return count;
}

/**************************************************
 ** beacon_speaker_write: set the speaker PA status  
 ** 1: enable.
 ** 0: disable.
 ***************************************************/ 
static ssize_t beacon_speaker_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	char value[2];
	int ret;

	if ((count = sscanf(buf, "%1s", value)) < 0)
		return -EFAULT;

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), "AUDIO_PA_EN");
	if (ret)
	{
		gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));
		printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_PA_EN", 
                                                      mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), ret);
		return -1;
	}

	if (value[0] == '1')
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 1);
		pxa3xx_mfp_set_lpm(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), MFP_LPM_DRIVE_HIGH);
	}
	else if (value[0] == '0')
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 0);
		pxa3xx_mfp_set_lpm(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), MFP_LPM_DRIVE_LOW);
	}

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	return count;
}

static struct bin_attribute speaker_mode_attr = {
	.attr = {
		.name = "audio_pa_en",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = 4,
	.read = beacon_speaker_read,
	.write = beacon_speaker_write,
};

static int beacon_speaker_amp_init(void)
{
	int ret = 0;

	ret = sysfs_create_bin_file(kernel_kobj, &speaker_mode_attr);
	if (ret)
	{
		printk(KERN_ERR "<%s> Failed to create sys file\n", "AUDIO_PA_EN");
		return -ENOMEM;
	}

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), "AUDIO_PA_EN");
	if (ret)
	{
		gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));
		printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_PA_EN", 
                                              mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), ret);
		return -1;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 0);

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	printk(KERN_INFO "<%s> GPIO%d setup done\n", "AUDIO_PA_EN", mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	return 0;
}

#ifdef CONFIG_ANDROID_RAM_CONSOLE
#define RAM_CONSOLE_START   0xA8000000
#define RAM_CONSOLE_SIZE    0x20000
static struct resource ram_console_resource = {
	.start  = RAM_CONSOLE_START,
	.end    = (RAM_CONSOLE_START + RAM_CONSOLE_SIZE - 1),
	.flags  = IORESOURCE_MEM,
};

static struct platform_device ram_console_device = {
	.name = "ram_console",
	.id = 0,
	.num_resources  = 1,
	.resource       = &ram_console_resource,
};

static inline void beacon_ramconsole_init(void)
{
	platform_device_register(&ram_console_device);
}

static inline void pxa_ramconsole_reserve_sdram(void)
{
	reserve_bootmem(RAM_CONSOLE_START, RAM_CONSOLE_SIZE, 0);
}
#else
static inline void beacon_ramconsole_init(void) {}

static inline void pxa_ramconsole_reserve_sdram(void) {}
#endif

static void __init beacon_map_io(void)
{
	pxa_ramconsole_reserve_sdram();
	pxa_map_io();
}

static void __init beacon_init(void)
{
	/* initialize MFP configurations */
	pxa3xx_mfp_config(ARRAY_AND_SIZE(beacon_mfp_cfg));
	pxa3xx_mfp_set_lpm(MFP_PIN_DF_nCS1, MFP_LPM_DRIVE_HIGH);


	beacon_misc_mfp_config();
	/* register i2c device */
	pxa_set_i2c_info(&pxa3xx_i2c_platform_data);

	/* register gpio i2c device */
	beacon_i2c_gpio_init();

	/* dvfm device */
	set_pxa3xx_freq_info(&beacon_freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

	beacon_ramconsole_init();

	/* lcd */
	beacon_init_lcd();

	/* lcd gpu */
	beacon_init_ism2300s();

	/* GPS */
#if defined(CONFIG_SIRF_GSD3TW)
	beacon_gps_sirf_init();
#endif
	/* WLAN */
#if defined(CONFIG_MRVL8688) || defined(CONFIG_MRVL8688_MODULE)
	beacon_init_wlan();
#endif

	platform_add_devices(devices, ARRAY_SIZE(devices));

#ifdef CONFIG_ANDROID_PMEM
	/* PMEM size 24M */
	android_add_pmem("pmem", 0x01800000UL, 1, 1);
#endif

	/* uart */
	pxa_set_ffuart_info(NULL);
	pxa_set_btuart_info(NULL);
	pxa_set_stuart_info(NULL);

#if defined(CONFIG_BT_HCIUART_PS)
	beacon_hci_ps_init();
#endif

#if defined(CONFIG_LEDS_PXA)
	pxa_set_leds_info(&beacon_leds_info);
#endif
#if defined(CONFIG_LEDS_DEBUG_SLEEP)
	pxa_set_sleep_led_info(&beacon_sleep_led_info);
#endif
	/* onenand */
#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
	beacon_init_onenand();
#endif

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &beacon_u2o_info);
	}
#endif

#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
	pxa_set_keypad_info(&beacon_keypad_info);
#endif

	/* i2c devices */
	i2c_register_board_info(0, ARRAY_AND_SIZE(beacon_i2c_board_info));


#if defined(CONFIG_PXA_CAMERA)
	/* initialize camera, shall do it after enable buck3 */
	beacon_init_cam();
#endif

	inno_if2xx_init();

#if defined(CONFIG_MMC)
	/* MMC card detect & write protect for controller 0 */
	beacon_mmc_slot[0].gpio_cd  = mfp_to_gpio(MFP_CFG_PIN(MMC1_CD_N));
	beacon_init_mmc();
#endif

#if defined(CONFIG_ALSPS)
	/* cm3601_alsps proximity sensor register platform device */
	beacon_cm3601_alsps_init();
#endif

#if defined(CONFIG_BQ2407x_POWER)
	beacon_init_charger();
#endif

#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65_FOR_TD)
	snd_ssp_mfp_cfg_init();
#endif

#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif

	beacon_speaker_amp_init();

	beacon_headset_init();
#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
	beacon_BP_init();
#endif

	/*avp_fm2010 device init*/
#if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2010) || defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2010_MODULE)
	beacon_i2c_avp_fm2010_init();
#endif
	/* enable clock */
	OSCC |= OSCC_TENS0;
}

MACHINE_START(BEACON, "Beacon")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = beacon_map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = beacon_init,
MACHINE_END
