/*
 *  linux/arch/arm/mach-pxa/benzglb.c
 *
 *  Support for FBW Benzina Global Product.
 *  
 *  Copyright (C) 2008 Borqs Ltd.
 *  Author:  Ru Yi <yi.ru@borqs.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 *
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/input.h>
#include <linux/gpio_keys.h>
#include <linux/synaptics_i2c_rmi.h>
#include <linux/bq2407x_charger.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
//#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/hci_ps.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/pxa-regs.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/pxa3xx_nand.h>
#include <mach/leds-pxa.h>
#include <mach/leds-debug.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/gpio.h>
#include <mach/micco.h>
#include <mach/benzglb.h>
#include <mach/irda.h>
#include <mach/pxa3xx_u2d.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/udc.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxafb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/pxa930_trkball.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa2xx_spi.h>
#include <mach/pxa-regs.h>
#include <mach/lis331.h>
#include <mach/akm8973.h>
#include <mach/ssp.h>
#include <mach/alsps.h>
#include <mach/micco-vibrator.h>
#include <mach/micco_headset.h>
#include <mach/gps.h>
#include <mach/i2c.h>
#include <mach/wlan.h>

#include "devices.h"
#include "generic.h"

#define MAX_SLOTS      3
struct platform_mmc_slot benzglb_mmc_slot[MAX_SLOTS];
extern int is_android(void);
extern int is_comm_v75(void);

extern void enable_oscc_tout_s0(void);
extern void disable_oscc_tout_s0(void);

static struct gpio_keys_platform_data benzglb_gpio_keys_info;
#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

static mfp_cfg_t benzglb_mfp_free_cfg[] __initdata = {
	GPIO0_FREE,
	GPIO1_FREE,
	GPIO2_FREE,
	GPIO3_FREE,
	GPIO6_FREE,
	GPIO7_FREE,
	GPIO8_FREE,
	GPIO9_FREE,
	GPIO10_FREE,
	GPIO11_FREE,
	GPIO12_FREE,
	GPIO13_FREE,
	GPIO21_FREE,
	GPIO22_FREE,
	GPIO61_FREE,
	GPIO62_FREE,
	GPIO98_FREE,
	GPIO174_FREE,
};

/* BENZINA MFP configurations */
static mfp_cfg_t benzglb_mfp_cfg[] __initdata = {
	/* FFUART */
	FFRXD,
	FFTXD,

	/* GPS */
	GPS_RX,
	GPS_TX,
	GPS_RESET,
	GPS_ON,

	/* Keypad*/
	VR_KEY,

	/* max9724 headphone amplifier */
	HP_AMP_SD,

	/* i2c bus */
	I2C_SCL,
	I2C_SDA,
	
	/* SSP2/BSSP2 for audio */
	I2S_BITCLK,
	I2S_SYNC,
	I2S_SDATA_OUT,

	/* SSP */
	PCM_CLK_GSSP1, 
	PCM_SYNC_GSSP1, 
	PCM_TXD_GSSP1, 
	PCM_RXD_GSSP1, 

	/* camera */	
	CAM_DATA7, 
	CAM_DATA6, 
	CAM_DATA5, 
	CAM_DATA4, 
	CAM_DATA3, 
	CAM_DATA2, 
	CAM_DATA1, 
	CAM_DATA0, 
	CAM_HSYNC, 
	CAM_VSYNC, 
	CAM_MCLK, 
	CAM_PCLK, 
	CAM_PD_MAIN, 
	CAM_RESET_MAIN,
	CAMERA_ON,
	CAM_ID,
	VCM_PD,
	FLASH_TORCH_SEL,
	CAM_ON_KEY_N,
	CAM_SHOT_KEY_N, 

	/*Pin for compass*/
	CPS_RSTN,
	CPS_INT,

	/* LCD */
	LCD_VSYNC, 
	LCD_HSYNC, 
	LCD_PCLK, 
	LCD_DE, 
	LCD_R0, 
	LCD_R1, 
	LCD_R2, 
	LCD_R3, 
	LCD_R4, 
	LCD_R5, 
	LCD_G0, 
	LCD_G1, 
	LCD_G2, 
	LCD_G3, 
	LCD_G4, 
	LCD_G5, 
	LCD_B0, 
	LCD_B1, 
	LCD_B2, 
	LCD_B3, 
	LCD_B4, 
	LCD_B5, 
	LCD_LED_PWM, 
	LCD_RESET_N,


	/* LCD SPI */
	LCD_SPI_CLK, 
	LCD_SPI_CS, 
	LCD_SPI_DIN, 

	
	/* PMIC */
	PMIC_INT_MICCO,
	/* avoid GPIO83 confliction */
	GPIO83_FREE,

	/* 1 wire */
	TAVOR_ONE_WIRE,

	/* G-Sensor */
	MOT_INT,

	/* BT UART */
	BT_UART_RXD, 
	BT_UART_TXD, 
	BT_UART_CTS, 
	BT_UART_RTS, 
	TAVOR_WAKEUP_BT,
	BT_WAKEUP_TAVOR,

	/* ALSPS: CM3601 */
	PRO_DOUT,
	PRO_LEDON,
	PRO_ENB,

	/* MMC */
	MMC0_CD,
	MMC1_CMD, 
	MMC1_CLK, 
	MMC1_DAT0, 
	MMC1_DAT1, 
	MMC1_DAT2, 
	MMC1_DAT3,
	SD_LDO_CE,

	/* WLAN */
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	WLAN_BT_RESET,
	TAVOR_WAKEUP_WLAN,
	WLAN_PD_GH381,
	WLAN_PD_PW886,
	WLAN_WAKEUP_TAVOR,

	/* Touch Panel int */
	TP_INT,

	/* charger ic */
	EXT_CHG_SET_EN2,
	EXT_CHG_SET_EN1,
	EXT_CHG_POOG_N,
	EXT_CHG_EN_N,
	EXT_CHG_DET_N,

	CLK_32K_OUT_EN,
};

static struct platform_device micco_bl_device = {
	.name 		= "micco-bl",
	.id 		= -1,
};

static struct platform_device micco_kp_bl_device = {
	.name 		= "micco-kp-bl",
	.id 		= -1,
};

static struct platform_device micco_ts_device = {
	.name 		= "micco-ts",
	.id 		= -1,
};

static struct micco_vibrator_platform_data benzgbl_vibrator_data = {
	.vib_strength = (0x33 << 1),
};
static struct platform_device micco_vibrator_device = {
	.name 		= "micco-vibrator",
	.id 		= -1,
	.dev		= {
		.platform_data = &benzgbl_vibrator_data,
	},
};

#if defined(CONFIG_MICCO_HEADSET_DETECT) || defined(CONFIG_MICCO_HEADSET_DETECT_MODULE)
static struct platform_device micco_headset_device = {
	.name		= "micco-headset",
	.id		= -1,
};
#endif /* CONFIG_MICCO_HEADSET_DETECT || CONFIG_MICCO_HEADSET_DETECT_MODULE */

static struct resource pxa3xx_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code +
						  * The first 4KB are locked for pxa930 trusted */
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1, /* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa3xx_resource_imm),
	.resource	= pxa3xx_resource_imm,
};

static struct platform_device gpio_keys_device = {
	.name		= "gpio-keys",
	.id		= -1,
	.dev		= {
		.platform_data = &benzglb_gpio_keys_info,
	},
};

#if defined(CONFIG_BQ2407x_POWER)

static mfp_cfg_t charge_mfp_cfg[] __initdata = {
	EXT_CHG_SET_EN1,
	EXT_CHG_SET_EN2,
	EXT_CHG_POOG_N,
	EXT_CHG_EN_N,
	EXT_CHG_DET_N
};

#define EXT_CHG_SET_EN1_GPIO	mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_SET_EN1))
#define EXT_CHG_SET_EN2_GPIO	mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_SET_EN2))
#define EXT_CHG_POOG_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_POOG_N))
#define EXT_CHG_EN_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_EN_N))
#define EXT_CHG_DET_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_DET_N))

static int benzglb_bq2407x_charger_init(void)
{
	int err;

        pxa3xx_mfp_config(ARRAY_AND_SIZE(charge_mfp_cfg));

	err = gpio_request(EXT_CHG_SET_EN1_GPIO, "charger en1");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en1 error!\n");
		return err;
	}
	err = gpio_request(EXT_CHG_SET_EN2_GPIO, "charger en2");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en2 error!\n");
		return err;
	}
	err = gpio_request(EXT_CHG_POOG_N_GPIO, "charger poog");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger poog error!\n");
		return err;
	}
	err = gpio_request(EXT_CHG_EN_N_GPIO, "charger en");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en error!\n");
		return err;
	}
	err = gpio_request(EXT_CHG_DET_N_GPIO, "charger det");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger det error!\n");
		return err;
	}

	gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 1);
	gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 1);
	gpio_direction_input(EXT_CHG_POOG_N_GPIO);
	gpio_direction_input(EXT_CHG_DET_N_GPIO);

	/* set always charger enable for power consumption */
	gpio_direction_output(EXT_CHG_EN_N_GPIO, 0);
	pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_EN_N), MFP_LPM_DRIVE_LOW);	
	return 0;
}

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
#define WAITING_USB_READY_INTERVAL	200
#define WAITING_USB_READY_TIMES		5

extern void pxa_charger_decide_enable(int on);
extern int is_usb_driver_busy(void);
static struct completion charger_detect_complete;
#endif

static int benzglb_bq2407x_charger_detect(void)
{
	int ret = BQ2407x_POWER_CHARGE_USB;
	int retry_times = 0;
	unsigned long timeout = 0;

	if (gpio_get_value(EXT_CHG_POOG_N_GPIO)) {
		printk(KERN_DEBUG "%s: No charger detected\n", __func__);
		ret = BQ2407x_POWER_NO_CHARGER;
	} else {
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
		while (is_usb_driver_busy()) {
			printk(KERN_DEBUG "%s: waiting for usb status busy "
					  "(retry times=%d\n",
					  __func__, retry_times);

			msleep(WAITING_USB_READY_INTERVAL);
			if (WAITING_USB_READY_TIMES <= ++retry_times) {
				printk(KERN_DEBUG "%s: USB regisered\n",
					   __func__);
				goto out;
			}
		}
	
		INIT_COMPLETION(charger_detect_complete);

		pxa_charger_decide_enable(1);
		enable_irq(IRQ_U2P);
		timeout = wait_for_completion_timeout(&charger_detect_complete,
					msecs_to_jiffies(100));

		if (timeout) {
			ret = BQ2407x_POWER_CHARGE_AC;
			printk(KERN_DEBUG "%s: charger (AC) detected\n",
					   __func__);
		} else {
			disable_irq(IRQ_U2P);
			ret = BQ2407x_POWER_CHARGE_USB;
			printk(KERN_DEBUG "%s: charger (USB) detected\n",
					   __func__);
		}
		pxa_charger_decide_enable(0);
#endif
	}
out:
	return ret; 
}

static int benzglb_bq2407x_is_charging(void)
{
	return (gpio_get_value(EXT_CHG_DET_N_GPIO) ? 0 : 1);
}

static void benzglb_bq2407x_set_charge(int enable)
{
	if (enable == BQ2407x_POWER_ENABLE) {
		printk(KERN_DEBUG "%s: charger enable\n", __func__);
		gpio_direction_output(EXT_CHG_EN_N_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_EN_N),
				   MFP_LPM_DRIVE_LOW);	
	}
	else {
		printk(KERN_DEBUG "%s: charger disable\n", __func__);
#if 0	/* set always charger enable for power consumption */
		gpio_direction_output(EXT_CHG_EN_N_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_EN_N),
				   MFP_LPM_DRIVE_HIGH);	
#endif
	}
}

static void benzglb_bq2407x_set_current(int flags)
{
	switch(flags) {
	case CHARGING_SUSPEND:
		printk(KERN_DEBUG "%s: current suspend\n", __func__);
		gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 1);
		gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	case CHARGING_CURRENT_100:
		printk(KERN_DEBUG "%s: current 100\n", __func__);
		gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 0);
		gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_500:
		printk(KERN_DEBUG "%s: current 500\n", __func__);
		gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 1);
		gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_FULL:
		printk(KERN_DEBUG "%s: current full\n", __func__);
		gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 0);
		gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	default:
		printk(KERN_DEBUG "%s: current 0\n", __func__);
		gpio_direction_output(EXT_CHG_SET_EN1_GPIO, 0);
		gpio_direction_output(EXT_CHG_SET_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(EXT_CHG_SET_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	}
}

static int bq2407x_is_cable_on(void)
{
	return gpio_get_value(EXT_CHG_POOG_N_GPIO) ? 0 : 1;
}

static void benzglb_bq2407x_charger_exit(void)
{
        gpio_free(EXT_CHG_SET_EN1_GPIO);
        gpio_free(EXT_CHG_SET_EN2_GPIO);
        gpio_free(EXT_CHG_POOG_N_GPIO);
        gpio_free(EXT_CHG_EN_N_GPIO);
        gpio_free(EXT_CHG_DET_N_GPIO);
}

static char *benzglb_power_supplicants[] = {
	"battery",
};

static struct bq2407x_charger_pdata benzglb_bq2407x_charger_pdata = {
	.init		= benzglb_bq2407x_charger_init,
	.detect		= benzglb_bq2407x_charger_detect,
	.is_charging	= benzglb_bq2407x_is_charging,
	.is_cable_on	= bq2407x_is_cable_on,
	.set_charge	= benzglb_bq2407x_set_charge,
	.set_current	= benzglb_bq2407x_set_current,
	.exit		= benzglb_bq2407x_charger_exit,
	.irq		= IRQ_GPIO(EXT_CHG_POOG_N_GPIO),
	.det_irq	= IRQ_GPIO(EXT_CHG_DET_N_GPIO),
	.supplied_to 	= benzglb_power_supplicants,
	.num_supplicants = ARRAY_SIZE(benzglb_power_supplicants),
};

static struct platform_device bq2407x_charger_device = {
	.name		= "charger",
	.id		= -1,
	.dev = {
		.platform_data = &benzglb_bq2407x_charger_pdata,
	},
};

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
static irqreturn_t charger_detected_irq(int irq, void *dev_id)
{
	printk(KERN_DEBUG "charger detected via U2O D+/D- irq\n");
	disable_irq(IRQ_U2P);
	complete(&charger_detect_complete);
	return IRQ_HANDLED;
}
#endif

static void benzglb_init_charger(void)
{
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	init_completion(&charger_detect_complete);
	if(request_irq(IRQ_U2P, charger_detected_irq, 
		       IRQF_DISABLED | IRQF_TRIGGER_FALLING,
		       "charger_detect", NULL) < 0) {
        	printk("charger_detected_irq register failed\n");
	}
	disable_irq(IRQ_U2P);
#endif

	platform_device_register(&bq2407x_charger_device);
}

#endif /* CONFIG_BQ2407x_POWER */

/**************************************************
 ** hp_amp_read: read the amplifier status
 ** 0: disable
 ** 1: enable
 ***************************************************/ 
static ssize_t benzglb_hp_amp_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    int value;

    value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD))) ? 1 : 0;
                        
    printk(KERN_INFO "<%s>: [read] after copying to user:  value = %d\n", "MAX9724", value); 
                            
    if ((count = sprintf(buf, "%d", value)) < 0)
        return -EFAULT;
                                
    printk(KERN_INFO "<%s>: [read] after copying to user:  buf = %s\n", "MAX9724", buf);

    return count;
}

/**************************************************
 ** hp_amp_write: set the amplifier  
 ** 0: close headset amplifier.
 ** 1: open headset amplifier.
 ***************************************************/ 
static ssize_t benzglb_hp_amp_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    char value[2];
    int ret;

    if ((count = sscanf(buf, "%1s", value)) < 0)
        return -EFAULT;

    printk(KERN_INFO "<%s>: [write] after copying from user:  value=%s\n", "MAX9724", value); 

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)));
        printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "MAX9724", MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    if (value[0] == '1')
    {
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), 1);
        printk(KERN_INFO "<%s>: [write] defaultly setting GPIO logical HIGH \n", "MAX9724");
    }
    else if (value[0] == '0')
    {
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), 0);
        printk(KERN_INFO "<%s>: [write] defaultly setting GPIO logical LOW \n", "MAX9724");
    }

    gpio_free(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)));

    return count;
}

static struct bin_attribute headset_mode_attr = {
    .attr = {
            .name = "max9724",
            .owner = THIS_MODULE,
            .mode = S_IRUGO | S_IWUGO,
            },
            .size = 4,
            .read = benzglb_hp_amp_read,
            .write = benzglb_hp_amp_write,
};

static int benzglb_headset_amp_init(void)
{
    int ret = 0;

    ret = sysfs_create_bin_file(kernel_kobj, &headset_mode_attr);
    if (ret)
    {
        printk(KERN_ERR "<%s> Failed to create sys file\n", "MAX9724");
        return -ENOMEM;
    }

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(MFP_CFG_PIN(HP_AMP_SD));
        printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "MAX9724", MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), 0);

    gpio_free(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)));

    printk(KERN_INFO "<%s> GPIOs setup done\n", "MAX9724");

    return 0;
}

#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65)
static void snd_ssp_mfp_cfg_init(void)
{
    extern mfp_cfg_t pcm_gssp1_mfp_cfg[4];
    extern mfp_cfg_t pcm_ssp3_mfp_cfg[4];
    extern mfp_cfg_t pcm_off_mfp_cfg[4];
    extern mfp_cfg_t i2s_ssp2_mfp_cfg[3];
    extern mfp_cfg_t i2s_off_mfp_cfg[3];

    pcm_gssp1_mfp_cfg[0] = PCM_CLK_GSSP1;
    pcm_gssp1_mfp_cfg[1] = PCM_SYNC_GSSP1;
    pcm_gssp1_mfp_cfg[2] = PCM_TXD_GSSP1;
    pcm_gssp1_mfp_cfg[3] = PCM_RXD_GSSP1;

    pcm_ssp3_mfp_cfg[0] = PCM_CLK_SSP3;
    pcm_ssp3_mfp_cfg[1] = PCM_SYNC_SSP3;
    pcm_ssp3_mfp_cfg[2] = PCM_TXD_SSP3;
    pcm_ssp3_mfp_cfg[3] = PCM_RXD_SSP3;

    pcm_off_mfp_cfg[0] = MFP_CFG_LPM(GPIO79, AF0, PULL_HIGH);
    pcm_off_mfp_cfg[1] = MFP_CFG_LPM(GPIO80, AF0, PULL_HIGH);
    pcm_off_mfp_cfg[2] = MFP_CFG_LPM(GPIO81, AF0, PULL_LOW);
    pcm_off_mfp_cfg[3] = MFP_CFG_LPM(GPIO82, AF0, PULL_LOW);

    i2s_ssp2_mfp_cfg[0] = I2S_BITCLK;
    i2s_ssp2_mfp_cfg[1] = I2S_SYNC;
    i2s_ssp2_mfp_cfg[2] = I2S_SDATA_OUT;

    i2s_off_mfp_cfg[0] = MFP_CFG_LPM(GPIO85, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[1] = MFP_CFG_LPM(GPIO87, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[2] = MFP_CFG_LPM(GPIO88, AF0, PULL_LOW);
}
#endif

static struct platform_device *devices[] __initdata = {
	&micco_bl_device,
	&micco_kp_bl_device,
	&micco_ts_device,
	&micco_vibrator_device,
#if defined(CONFIG_MICCO_HEADSET_DETECT) || defined(CONFIG_MICCO_HEADSET_DETECT_MODULE)
	&micco_headset_device,
#endif
#ifdef CONFIG_IMM
	&pxa3xx_device_imm,
#endif
	&gpio_keys_device,
};

#ifdef CONFIG_KEYBOARD_GPIO
static struct gpio_keys_button benzglb_gpio_keys_button[] = {
	[0] = {
		.desc	= "back_key",
		.code	= KEY_BACK,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VR_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[1] = {
		.desc	= "cam_shot_key_n",
		.code	= KEY_CAMERA,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_SHOT_KEY_N)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[2] = {
		.desc	= "cam_on_key",
		.code	= KEY_FOCUS,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_ON_KEY_N)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[3] = {
		.desc	= "volume_up_key",
		.code	= KEY_VOLUMEUP,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_UP_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[4] = {
		.desc	= "volume_down_key",
		.code	= KEY_VOLUMEDOWN,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_DOWN_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[5] = {
		.desc	= "menu_key",
		.code	= KEY_MENU,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(OPTION_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
};

static struct gpio_keys_platform_data benzglb_gpio_keys_info = {
	.buttons = benzglb_gpio_keys_button,
	.nbuttons = (sizeof(benzglb_gpio_keys_button)
			/ sizeof(struct gpio_keys_button)),
	//.trigger_key0 = KEY_BACK,
	//.trigger_key1 = KEY_CAMERA,
};
#endif

#if defined(CONFIG_SIRF_GSD3TW)
static struct sirf_platform_data benzglb_sirf_platform_data = {
	.reset_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)),
	.on_off_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_ON)),
	//.power_pin = -1,
};

static struct platform_device benzglb_sirf_device = {
	.name = "sirf",
	.id = -1,
	.dev = {
		.platform_data = &benzglb_sirf_platform_data,
	},
};

static void __init benzglb_gps_sirf_init(void)
{
	printk(KERN_ERR"[%s:%d] gps pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), "gps reset")) {
		printk(KERN_ERR "<SiRF> GPIO RST request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), "gps on_off")) {
		printk(KERN_ERR "<SiRF> GPIO ON_OFF request error!!\n");
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), 0); /* GPS_ON_OFF */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), 1); /* GPS_RST */
	printk(KERN_ERR"GPS set up finished\n");
	platform_device_register(&benzglb_sirf_device);
}
#endif

#if defined(CONFIG_ALSPS)
#define PROXIMITY_SHUTDOWN 	MFP_CFG_PIN(PRO_LEDON)
#define ALSPS_POWERDOWN 	MFP_CFG_PIN(PRO_ENB)
#define PROXIMITY_OUTPUT 	MFP_CFG_PIN(PRO_DOUT)

static int benzglb_alsps_init(void)
{
	int ret;

	/*config ambient light sensor switch pin, also as chip switch pin*/	
	ret = gpio_request(mfp_to_gpio(ALSPS_POWERDOWN), "ambient_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset ambient_powerdown failed\n");
		goto out;
	}
	gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);

	/*config proximity sensor switch pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_SHUTDOWN), "proximity_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_powerdown failed\n");
		goto out1;
	}
	gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);

	/*config proximity sensor output pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_OUTPUT), "proximity_output");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_output failed\n");
		goto out2;
	}
	gpio_direction_input(mfp_to_gpio(PROXIMITY_OUTPUT));

	return 0;

out2:
	gpio_free(mfp_to_gpio(PROXIMITY_SHUTDOWN));
out1:
	gpio_free(mfp_to_gpio(ALSPS_POWERDOWN));
out:
	printk("alsps: config gpio pins failed!\n");
	return ret;
}

static int benzglb_alsps_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(PROXIMITY_OUTPUT)) ? 1 : 0;
}

static void benzglb_alsps_setpower(int enable)
{
	if (enable & PROXIMITY_ENABLE) {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 0);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_HIGH);
	}

	if (enable) {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 0);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_HIGH);
	}
}

static struct alsps_platform_data benzglb_alsps_platform_data = {
	.irq		= IRQ_GPIO(mfp_to_gpio(PROXIMITY_OUTPUT)),
	.adc_event	= PMIC_EVENT_ADC5, 
	.init		= benzglb_alsps_init, 
	.get_gpio_value	= benzglb_alsps_get_gpio_value, 
	.setpower	= benzglb_alsps_setpower,
};

static struct platform_device benzglb_alsps_device = {
	.name = "alsps",
	.id = -1,
	.dev = {
		.platform_data = &benzglb_alsps_platform_data,
	},
};

static void __init benzglb_cm3601_alsps_init(void)
{
	platform_device_register(&benzglb_alsps_device);
}
#endif


#if defined(CONFIG_MMC)
static int benzglb_mci_init(struct device *dev,
		irq_handler_t benzglb_detect_int,
		void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(benzglb_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = benzglb_mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for benzglb MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, benzglb_detect_int,
			IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			"MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void benzglb_mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(benzglb_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = benzglb_mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static mfp_cfg_t sd_in_cfg[]={
	SD_LDO_CE_ON,
	MMC1_CMD_ON,
	MMC1_DAT0_ON,
	MMC1_DAT1_ON,
	MMC1_DAT2_ON,
	MMC1_DAT3_ON,
	MMC1_CLK_ON,
};

static mfp_cfg_t sd_out_cfg[]={
	SD_LDO_CE,
	MMC1_CMD,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	MMC1_CLK,
};

static void benzglb_mci_setpower(struct device *dev, unsigned int vdd)
{
	struct pxamci_platform_data* p_d = dev->platform_data;

	if (( 1 << vdd) & p_d->ocr_mask) {
        	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 1);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_in_cfg));
	} else {
        	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 0);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_out_cfg));
	}
}

static struct pxamci_platform_data benzglb_mci_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
	.gpio_detect    = mfp_to_gpio(MFP_CFG_PIN(MMC0_CD)),
	.init 		= benzglb_mci_init,
	.setpower	= benzglb_mci_setpower,
	.exit		= benzglb_mci_exit,
};

static struct pxamci_platform_data benzglb_mci2_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
};

static void __init benzglb_init_mmc(void)
{
	if (device_is_enabled(device_mmc)) {
		pxa_set_mci_info(&benzglb_mci_platform_data);
		pxa3xx_set_mci2_info(&benzglb_mci2_platform_data);
	}
}
#else
static inline void benzglb_init_mmc(void) {}
#endif

#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
static struct flash_platform_data benzglb_onenand_info;
static void __init benzglb_init_onenand(void)
{
	/* Now, kernel get mtd partitions from commandline.*/
	if (is_android()) {
		benzglb_onenand_info.parts = android_256m_v75_partitions;
		benzglb_onenand_info.nr_parts = 
			ARRAY_SIZE(android_256m_v75_partitions);
	}
	else {
		benzglb_onenand_info.parts = pxa930_256m_v75_partitions;
		benzglb_onenand_info.nr_parts = 
			ARRAY_SIZE(pxa930_256m_v75_partitions);
	}

	pxa3xx_device_onenand.dev.platform_data = &benzglb_onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#endif


#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_PXA27X_UDC_MODULE)

int benzglb_udc_is_miniA(void)
{
	return 0;
}

static struct pxa2xx_udc_mach_info benzglb_udc_info = {
	.udc_is_miniA = benzglb_udc_is_miniA,
};
#endif

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info benzglb_u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
static int micco_init_irq(void)
{
	gpio_direction_input(PMIC_INT_PIN);

	return 0;
}

static int micco_ack_irq(void)
{
	return 0;
}

static void benzglb_micco_init(void)
{
	u8 value;

	/* Mask interrupts that are not needed */
	micco_write(MICCO_IRQ_MASK_A, 0xFC);
	micco_write(MICCO_IRQ_MASK_B, 0xFF);
	micco_write(MICCO_IRQ_MASK_C, 0x3F);
	micco_write(MICCO_IRQ_MASK_D, 0x3F);

	/* avoid SRAM power off during sleep*/
	value = MICCO_LDO2_ENABLE | MICCO_BUCK1_ENABLE;
	micco_write(MICCO_OVER1, value);

	/* enable all the GSM/WCDMA/EDGE related module */
	value = MICCO_LDO6_ENABLE | MICCO_LDO7_ENABLE | MICCO_LDO8_ENABLE |
		MICCO_LDO9_ENABLE | MICCO_LDO10_ENABLE | MICCO_LDO11_ENABLE |
		MICCO_LDO12_ENABLE;
	micco_write(MICCO_APP_OVER2, value);

	value = MICCO_LDO1_ENABLE | MICCO_LDO3_ENABLE | MICCO_LDO5_ENABLE | 
		MICCO_LDO13_ENABLE | MICCO_BUCK3_ENABLE | MICCO_RF_GP_ENABLE;
	micco_write(MICCO_APP_OVER3, value);

	/* disable the ONKEY power down functionality */
	/* micco_write(MICCO_SYSCTRL_B, 0x20); */
	micco_write(MICCO_SYSCTRL_A, 0x60);

	/* Enable BUCK & LDO sleep setting */
	value = MICCO_BUCK1_nSLEEP_ENABLE  | MICCO_BUCK3_nSLEEP_ENABLE;
	micco_write(MICCO_BUCK_SLEEP, value);

	//value = MICCO_LDO8_nSLEEP_ENABLE;
	value = 0;
	micco_write(MICCO_LDO_nSLEEP1, value);

	value = MICCO_RF_GP_nSLEEP_ENABLE;
	micco_write(MICCO_LDO_nSLEEP2, value);

	value = MICCO_LDO8_SLEEP_BIT_ENABLE;
	micco_write(MICCO_LDO_SLEEP_BIT1, value);

	value = 0;
	micco_write(MICCO_LDO_SLEEP_BIT2, value);

	value = MICCO_BUCK1_USE_LDO8_ENABLE | MICCO_BUCK3_USE_LDO8_ENABLE;
	micco_write(MICCO_BUCK_LDO8_EN, value);

	/* IRQ is masked during the power-up sequence and will not be released
	 * until they have been read for the first time */
	micco_read(MICCO_EVENT_A, &value);
	printk("%s: MICCO_EVENT_A=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_B, &value);
	printk("%s: MICCO_EVENT_B=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_C, &value);
	printk("%s: MICCO_EVENT_C=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_D, &value);
	printk("%s: MICCO_EVENT_D=0x%x\n", __FUNCTION__, value);
}

/* micco_power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module miccoEA_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		0},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,		0},
	{VCC_SDIO,		0},
	{VCC_CAMERA_ANA,	0},
	{VCC_USB,		LDO5},
	{VCC_LCD,		0},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,		LDO2},
	{VCC_1P8V,		0},
	{VCC_MEM,		LDO2},
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

#if	defined(CONFIG_SENSORS_AK8973)
#define COMPASS_INT_GPIO	mfp_to_gpio(MFP_CFG_PIN(CPS_INT))
#define COMPASS_IRQ             IRQ_GPIO(COMPASS_INT_GPIO)
#define COMPASS_RESET_GPIO	mfp_to_gpio(MFP_CFG_PIN(CPS_RSTN))

//We should add the other pin's request function

static int compass_init_irq(void)
{
	int ret = 0;
	ret = gpio_request(COMPASS_INT_GPIO, "compass irq");
	if (ret){
		printk("gpio_requset failed, return :%d\n",ret);
		goto err_request_cd;
	}
	ret=gpio_direction_input(COMPASS_INT_GPIO);
	if (ret) {
		printk("set gpio pin direction for compass failed, return :%d\n", ret);
		goto err_request_irq;
	}
err_request_irq:
	gpio_free(COMPASS_INT_GPIO);
err_request_cd:
	return ret;
}

static int compass_ack_irq(void)
{
	return 0;
}

static int compass_reset_pin(void)
{
	int ret = 0;
	ret = gpio_request(COMPASS_RESET_GPIO, "compass reset");
	if (ret) {
		printk("compass reset gpio_requset failed, return :%d\n",ret);
		return ret;	
	}
	gpio_direction_output(COMPASS_RESET_GPIO, 0);
	msleep(20);
	gpio_direction_output(COMPASS_RESET_GPIO, 1);
	msleep(20);

	return ret;
}

static void compass_fac_reset(void)
{
	printk(KERN_DEBUG "AKM8973 E-Compass factory test reset pin!\n");
	gpio_direction_output(COMPASS_RESET_GPIO, 0);
	msleep(20);
	gpio_direction_output(COMPASS_RESET_GPIO, 1);
	msleep(20);
}

static struct akm8973_platform_data akm8973_data = {
	.init_irq = compass_init_irq,
	.ack_irq = compass_ack_irq,
	.reset_pin = compass_reset_pin,
	.fac_reset = compass_fac_reset,
};
#endif
static int lis331_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(MOT_INT)));

	return 0;
}

static int lis331_ack_irq(void)
{
	return 0;
}

static int lis331_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))) ;
}

static struct lis331_platform_data lis331_data = {
	.init_irq = lis331_init_irq,
	.ack_irq = lis331_ack_irq,
	.get_gpio_value = lis331_get_gpio_value,
	.platform_init = NULL,
	.power_chips = NULL,
	.xyz_dir = Y_POS,
};

static struct power_chip micco_chips[] = {
	{MICCO_EA_ID,	"miccoEA",	miccoEA_power_modules},
	{MICCO_EB_ID,	"miccoEB",	miccoEA_power_modules},
	{0,	NULL,		NULL},
};

#if defined(CONFIG_TOUCHSCREEN_TM1175)
static int tm1175_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(TP_INT)));
	return 0;
}

static int tm1175_ack_irq(void)
{
	return 0;
}

static struct tm1175_platform_data tm1175_data = {
	.init_irq = tm1175_init_irq,
	.ack_irq = tm1175_ack_irq,
};

#endif
static struct micco_platform_data micco_data = {
	.init_irq = micco_init_irq,
	.ack_irq = micco_ack_irq,
	.platform_init = benzglb_micco_init,
	.power_chips = micco_chips,
/* you can choose the value of VIBRA_PWM<6:0> (bit7 ~ bit1) to change the freq of the vibrator 
 * 	00h : OFF
 * 	>= 55h: 100% alwayes on (code 84 decimal)
 */
};
#endif /* CONFIG_PXA3xx_MICCO || CONFIG_PXA3xx_MICCO_MODULE*/


#if defined(CONFIG_PXA_CAMERA)

/* sensor init */
static int sensor_voltage_on(int flag)
{
	printk(KERN_ERR"[%s:%d] Camera power on\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAMERA_ON)), "CAMERA_ON")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAMERA_ON));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), "CAD_PD_MAIN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_PD_MAIN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), "CAM_RESET_MAIN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_RESET_MAIN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(VCM_PD)), "VCM_PD")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(VCM_PD));
		return -EIO;
	}

	/* OV5642 power up with internal DVDD:
 	 *
	 * PWDN pin shall be high before power up.
	 * Before DOVDD is up, RESET# pin is better keep at low level.
	 * for PWDN to go low, power up must first become stable (AVDD to PWDN > 5 ms)
	 * H/W RESET. After power up stable , reset signal should be more than 3 mS.
	 */

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(VCM_PD)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAMERA_ON)), 0);
	/* PWDN pin shall be high before power up. */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), 1);
	mdelay(1);

	/* I/O & Analog power up */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAMERA_ON)), 1);
	/* delay to wait power stable */
	msleep(10);

	/* for PWDN to go low, power up must first become stable (AVDD to PWDN > 5 ms) */
	mdelay(5);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), 0);

	/* t1: from powers on to Reset pull high, >= 3ms */
	mdelay(3);
	/* assert RESET */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), 1);

	/* wait a moment then power down */
	msleep(10);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), 1);

	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAMERA_ON)));	/* CAMERA_ON */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)));	/* CAM_PD_MAIN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)));	/* CAM_RESET_MAIN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(VCM_PD)));		/* VCM_PD */

	return 0;
}

static int sensor_power_on(int flag)
{
	printk(KERN_ERR"[%s:%d] Camera power on\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), "CAD_PD_MAIN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_PD_MAIN));
		return -EIO;
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(VCM_PD)), "VCM_PD")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(VCM_PD));
		return -EIO;
	}
	msleep(10);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(VCM_PD)), 1);
	mdelay(1);

	return 0;
}

static int sensor_power_off(int flag)
{
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)), 1);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(VCM_PD)), 0);
	mdelay(1);

	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_PD_MAIN)));	/* CAM_PD_MAIN */
	gpio_free( mfp_to_gpio(MFP_CFG_PIN(VCM_PD)));		/* VCM_PD */

	return 0;
}

static int sensor_reset(int flag)
{
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), "CAM_RESET_MAIN")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_RESET_MAIN));
		return -EIO;
	}
	/* assert RESET */
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), 0);
	mdelay(1);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)), 1);

	gpio_free( mfp_to_gpio(MFP_CFG_PIN(CAM_RESET_MAIN)));	/* CAM_RESET_MAIN */
	return 0;
}


static int cam_flash_control(int mode)
{
	printk(KERN_INFO"spark: camera flash control(%d)\n", mode);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)), "FLASH_TORCH_SEL")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(FLASH_TORCH_SEL));
		return -EIO;
	}

	if (mode)
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)), 1);
	else
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)), 0);

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)));

	return 0;
}

static struct sensor_platform_data ov5642_sensor_data = {
	.id = SENSOR_HIGH,
	.power_on = sensor_power_on,
	.power_off = sensor_power_off,
	//.reset = sensor_reset,
	//.flash_ctrl = cam_flash_control,
};

/* sensor init over */

/* camera platform data */
static mfp_cfg_t sync[] = {
	CAM_HSYNC,
	CAM_VSYNC,
};

static mfp_cfg_t sync_gpio[] = {
	CAM_HSYNC_GPIO,
	CAM_VSYNC_GPIO,
};

static void cam_sync_to_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync_gpio));
}

static void cam_sync_from_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync));
}

static int cam_init(void)
{
	printk(KERN_ERR"[%s:%d] %s()\n", __FILE__, __LINE__, __func__);
	return 0;
}

static void cam_deinit(void)
{
}

static void cam_suspend(void)
{
	sensor_power_off(0);
}

static void cam_resume(void)
{
	sensor_power_off(0);
}

static struct cam_platform_data cam_ops = {
	.vsync_gpio		= mfp_to_gpio(MFP_CFG_PIN(CAM_VSYNC_GPIO)),
	.init			= cam_init,
	.deinit			= cam_deinit,
	.suspend		= cam_suspend,
	.resume			= cam_resume,
	.sync_to_gpio		= cam_sync_to_gpio,
	.sync_from_gpio		= cam_sync_from_gpio,
};

static void __init benzglb_init_cam(void)
{
	printk(KERN_ERR"[%s:%d] Camera pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(CAM_ID)), "CAM_ID")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(CAM_ID));
		return;
	}
	if( gpio_request(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)), "FLASH_TORCH_SEL")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", MFP_CFG_PIN(FLASH_TORCH_SEL));
		return;
	}

	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(CAM_ID)) );		/* CAM_ID */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)), 0);	/* FLASH_TORCH */

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(CAM_ID)));		/* CAM_ID */
	gpio_free(mfp_to_gpio(MFP_CFG_PIN(FLASH_TORCH_SEL)));	/* FLASH_TORCH */

	/* shall remove power on ov5642 if OBM do it when phone power up */
	sensor_voltage_on(0);

	pxa3xx_device_cam.dev.platform_data = &cam_ops;
	platform_device_register(&pxa3xx_device_cam);
}

/* QCI init over */

#endif

#ifdef CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI
static int benzglb_ts_power(int on)
{
	//gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(TP_INT)));
        return 0;
}

static struct synaptics_i2c_rmi_platform_data benzglb_ts_data[] = {
{
                .version = 0x0101,
                .power = benzglb_ts_power,
                .sensitivity_adjust = 7,
                .flags = SYNAPTICS_FLIP_X | SYNAPTICS_FLIP_Y | SYNAPTICS_SNAP_TO_INACTIVE_EDGE,
                .irqflags = IRQF_TRIGGER_FALLING,
                .inactive_left = 0,
                .inactive_right = 0,
                .inactive_top = 0,
                .inactive_bottom = 0,
                .snap_left_on = 0,
                .snap_left_off = 0,
                .snap_right_on = 0,
                .snap_right_off = 0,
                .snap_top_on = 0,
                .snap_top_off = 0,
                .snap_bottom_on = 0,
                .snap_bottom_off = 0,
                .display_width = 360,
                .display_height = 640,
        },
        {
                .flags = SYNAPTICS_FLIP_X | SYNAPTICS_FLIP_Y | SYNAPTICS_SNAP_TO_INACTIVE_EDGE,
                .inactive_left = 0,
                .inactive_right = 0,
                .inactive_top = 0,
                .inactive_bottom = 0,
                .display_width = 360,
                .display_height = 640,
        }
};
#endif

	static struct i2c_board_info benzglb_i2c_board_info[] = {
#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
		{
			.type	= "micco",
			.addr		= 0x34,
			.platform_data	= &micco_data,
			.irq		= IRQ_GPIO(PMIC_INT_PIN),
		},
#endif
#if defined(CONFIG_PXA_CAMERA)
		{
			.type	= "sensor_ov5642",
			.addr		= 0x3C,
			.platform_data	= &ov5642_sensor_data,
		},
#endif
#if defined(CONFIG_SENSORS_AK8973)
		{
			.type	= "akm8973",
			.addr		= 0x1e,
			.platform_data	= &akm8973_data,
			.irq		= COMPASS_IRQ,
		},
#endif

#if defined(CONFIG_SENSORS_LIS331)
		{
			.type	= "lis331",
			.addr		= 0x1C,
			.platform_data	= &lis331_data,
			.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))),
		},
#endif

#if defined(CONFIG_TOUCHSCREEN_TM1175)
		{
			.type	= "tm1175_touch",
			.addr		= 0x20,
			.platform_data	= &tm1175_data,
			.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
		},
#endif

#ifdef CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI
		{
			.type	= SYNAPTICS_I2C_RMI_NAME,
			.addr		= 0x20,
			.platform_data	= &benzglb_ts_data,
			.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
		},
#endif

	};

#if defined(CONFIG_MRVL8688) || defined(CONFIG_MRVL8688_MODULE)
static mfp_cfg_t wlan_lpm_gpio_on_mfp_cfg[] = {
    WLAN_PD_GH381_LPM_ON,
    WLAN_PD_PW886_LPM_ON,
    WLAN_BT_RESET_LPM_ON,
    WLAN_WAKEUP_TAVOR_LPM_ON,
    TAVOR_WAKEUP_BT_LPM_ON,
    BT_UART_TXD_LPM_ON,
    BT_UART_RXD_LPM_ON,
    BT_UART_CTS_LPM_ON,
    BT_UART_RTS_LPM_ON,
    BT_WAKEUP_TAVOR_LPM_ON,
    WLAN_DAT3_LPM_ON,
    WLAN_DAT2_LPM_ON,
    WLAN_DAT1_LPM_ON,
    WLAN_DAT0_LPM_ON,
    WLAN_CMD_LPM_ON,
    WLAN_CLK_LPM_ON,
    TAVOR_WAKEUP_WLAN_LPM_ON,
};

static mfp_cfg_t wlan_lpm_gpio_off_mfp_cfg[] = {
    WLAN_PD_PW886,
    WLAN_PD_GH381,
    WLAN_BT_RESET,
    WLAN_WAKEUP_TAVOR,
    TAVOR_WAKEUP_BT,
    BT_UART_TXD,
    BT_UART_RXD,
    BT_UART_CTS,
    BT_UART_RTS,
    BT_WAKEUP_TAVOR,
    WLAN_DAT3,
    WLAN_DAT2,
    WLAN_DAT1,
    WLAN_DAT0,
    WLAN_CMD,
    WLAN_CLK,
    TAVOR_WAKEUP_WLAN,
    PCM_CLK_OFF,
    PCM_SYNC_OFF,
    PCM_TXD_OFF,
    PCM_RXD_OFF,
};

static int benzglb_wlan_enabled=0;
static void benzglb_wlan_gpio_setpower(int enable)
{
	unsigned int pcm_clk, pcm_sync, pcm_rxd, pcm_txd;

	if (enable) {
		enable_oscc_tout_s0();
		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_on_mfp_cfg));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR)));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR)));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_PW886)), 1);
		msleep(1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_GH381)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
		msleep(10);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_WLAN)), 1);

		printk(KERN_DEBUG "wlan_lpm_gpio_on\n");
		benzglb_wlan_enabled=1;
	}
	else {
		pcm_clk = pxa3xx_mfp_read(MFP_CFG_PIN(PCM_CLK));
		pcm_sync = pxa3xx_mfp_read(MFP_CFG_PIN(PCM_SYNC));
		pcm_txd = pxa3xx_mfp_read(MFP_CFG_PIN(PCM_TXD));
		pcm_rxd = pxa3xx_mfp_read(MFP_CFG_PIN(PCM_RXD));

		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_off_mfp_cfg));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_PW886)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_GH381)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)),0);
        	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_BT)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_WLAN)),0);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(PCM_CLK)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(PCM_SYNC)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(PCM_TXD)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(PCM_RXD)),0);

 		msleep(10);

		pxa3xx_mfp_write(MFP_CFG_PIN(PCM_CLK), pcm_clk);
		pxa3xx_mfp_write(MFP_CFG_PIN(PCM_SYNC), pcm_sync);
		pxa3xx_mfp_write(MFP_CFG_PIN(PCM_TXD), pcm_txd);
		pxa3xx_mfp_write(MFP_CFG_PIN(PCM_RXD), pcm_rxd);

		disable_oscc_tout_s0();
		printk(KERN_DEBUG "wlan_lpm_gpio_off\n");
		benzglb_wlan_enabled=0;
	}
}

static struct wlan_platform_data benzglb_wlan_pdata = {
	.chip_gpio = 1,
	.chip_gpio_setpower = benzglb_wlan_gpio_setpower,
        .bt_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR))),
        .wlan_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR))),
}; 

static struct platform_device wlan_device = {
	.name		= "mrvl8688",
	.id		= -1,
};

static void __init benzglb_init_wlan(void)
{

	int err = 0;

	printk(KERN_ERR "WLAN set up\n");

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_PW886)), "wlan pmic powerdown");	
	if (err) {
		printk(KERN_ERR "WLAN_PD_PW886 Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_GH381)), "wlan powerdown");
	if (err) {
		printk(KERN_ERR "WLAN_PD_PW886 Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), "wlan bt reset");	
	if (err) {
		printk(KERN_ERR "WLAN_BT_RESET Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_WLAN)), "wlan mac wakeup");	
	if (err) {
		printk(KERN_ERR "TAVOR_WAKEUP_WLAN Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR)), "bt wakeup tavor");  
	if (err) {
		printk(KERN_ERR "BT_WAKEUP_TAVOR Request error!\n");
		return;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_PW886)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PD_GH381)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_WLAN)), 0);

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_UART_RXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_UART_TXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_UART_CTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_UART_RTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT3)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT2)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT1)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT0)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CLK)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CMD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_BT)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_ONE_WIRE)),0);

	wlan_device.dev.platform_data = &benzglb_wlan_pdata;
	platform_device_register(&wlan_device);

	printk(KERN_ERR "WLAN set up finished\n");

	return;
}
#endif

#if defined(CONFIG_BT_HCIUART_PS)

#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
static void switch_bt_chip_lpm_state(int state)
{
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TAVOR_WAKEUP_BT)), 
			      !!state);
        pxa3xx_mfp_set_lpm(MFP_CFG_PIN(TAVOR_WAKEUP_BT),
			   !!state ? MFP_LPM_DRIVE_HIGH : MFP_LPM_DRIVE_LOW);
}
#endif

static struct hci_ps_platform_data benzglb_hci_ps_pdata = {
	.chip_type = CHIP_MRVL8688,
	.chip_setpower = NULL,
}; 

static struct platform_device hci_ps_device = {
	.name		= "hci_ps",
	.id		= -1,
};

static void __init benzglb_hci_ps_init(void)
{
#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
	benzglb_hci_ps_pdata.chip_setpower = switch_bt_chip_lpm_state;
#endif
	hci_ps_device.dev.platform_data = &benzglb_hci_ps_pdata;
	platform_device_register(&hci_ps_device);
}
#endif


struct pxa3xx_freq_mach_info benzglb_freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};

static void benzglb_backlight_power( int on, int intensity)
{
	if( on) {
		CKENB |= 1<<(CKEN_PWM0-32);
		PWM_CTRL0 = 0x1;
		PWM_PWDUTY0 = intensity;
		PWM_PERVAL0 = 0xfe;
	} else {
		PWM_CTRL0 = 0x0;
		PWM_PWDUTY0 = 0x0;
		PWM_PERVAL0 = 0x0;
		CKENB &= ~(1<<(CKEN_PWM0-32));
	}
}

/* use bit 30, 31 as the indicator of command parameter number */
#define CMD0(x)		((0x00000000) | ((x)))
#define DAT0(x)		((0x00000100) | ((x)))
#define CMD1(x, x1)	((0x40000000) | ((x) << 9) | 0x100 | (x1))
#define CMD2(x, x1, x2)	((0x80000000) | ((x) << 18) | 0x20000 |\
			((x1) << 9) | 0x100 | (x2))
#define TPO_L1N304_LCD_DSS 	9 	//Data size select for LCD spi

static uint32_t sleep_out[] = {
	CMD0(0x11),
};

static uint32_t gamma_setting[] = {
	CMD0(0xC0),
	DAT0(0x0F),
	DAT0(0x0B),
	DAT0(0x0F),
	DAT0(0x0B),
	DAT0(0x6D),
	DAT0(0x55),
	DAT0(0x20),
	DAT0(0x08),
	DAT0(0x70),
	DAT0(0x57),
	DAT0(0x1E),
	DAT0(0x06),
	DAT0(0x08),
	DAT0(0x08),

	CMD0(0xE0),
	DAT0(0x0E),
	DAT0(0x09),
	DAT0(0x1A),
	DAT0(0x2C),
	DAT0(0x43),
	DAT0(0x4C),
	DAT0(0x47),
	DAT0(0x3F),
	DAT0(0x3D),
	DAT0(0x2C),
	DAT0(0x23),
	DAT0(0x25),
	DAT0(0x34),
	DAT0(0x1B),
	DAT0(0x2B),

	CMD0(0xE1),
	DAT0(0x1F),
	DAT0(0x0B),
	DAT0(0x20),
	DAT0(0x36),
	DAT0(0x53),
	DAT0(0x56),
	DAT0(0x47),
	DAT0(0x44),
	DAT0(0x3D),
	DAT0(0x36),
	DAT0(0x2F),
	DAT0(0x30),
	DAT0(0x35),
	DAT0(0x1D),
	DAT0(0x36),

	CMD0(0xE2),
	DAT0(0x0E),
	DAT0(0x09),
	DAT0(0x1A),
	DAT0(0x2C),
	DAT0(0x43),
	DAT0(0x4C),
	DAT0(0x47),
	DAT0(0x3F),
	DAT0(0x3D),
	DAT0(0x2C),
	DAT0(0x23),
	DAT0(0x25),
	DAT0(0x34),
	DAT0(0x1B),
	DAT0(0x2B),
};

static uint32_t initialize[] = {
	CMD0(0xFC),
	DAT0(0x00),
	DAT0(0x10),
	DAT0(0x00),
	DAT0(0x00),

	CMD0(0xD1),
	DAT0(0x84),
	DAT0(0x98),
	DAT0(0x86),
	DAT0(0x86),

	CMD0(0x3A),
	DAT0(0x60),

	CMD0(0xF3),
	DAT0(0xB0),

	CMD0(0xF2),
	DAT0(0x00),
	DAT0(0x09),

	CMD0(0xC1),
	DAT0(0x07),
};

static uint32_t disp_on[]={
	CMD0(0x11),

	CMD0(0x29),
};

static uint32_t disp_off[]={
	CMD0(0x28),
};

static uint32_t sleep_in[]={
	CMD0(0x10),
};

static void ssp_reconfig(struct ssp_dev *dev, int nparam)
{
	static int last_nparam = -1;
	int sscr0;

	/* check if it is necessary to re-config SSP */
	if (nparam == last_nparam)
		return;

	ssp_disable(dev);
	if( nparam >16 )
		sscr0 = 0x00107f80 | ((nparam-1) & 0xf);
	else
		sscr0 = 0x00007f80 | ((nparam-1) & 0xf);

	ssp_config(dev, sscr0, 0x18, 0, 0);
	last_nparam = nparam;
}

static void ssp_send_cmd(uint32_t *cmd, int num)
{
	static int ssp_initialized = 0;
	static struct ssp_dev ssp1;

	int i;

	if (!ssp_initialized) {
		ssp_init(&ssp1, 1, SSP_NO_IRQ);
		ssp_initialized = 1;
		clk_disable(ssp1.ssp->clk);
	}

	clk_enable(ssp1.ssp->clk);
	for (i = 0; i < num; i++, cmd++) {
		ssp_reconfig(&ssp1, TPO_L1N304_LCD_DSS);
		ssp_write_word(&ssp1, *cmd & 0x3fffffff);

		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp1);
	}
	clk_disable(ssp1.ssp->clk);
}

static void benzglb_lcd_power(int on, struct fb_var_screeninfo *var)
{
	static int pin_requested = 0;

	if( !pin_requested) {
		if( gpio_request( mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 
			"LCD controller reset"))
			return;
		pin_requested = 1;
	}

	if( on)
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 0);
        	udelay(100);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 1);
		msleep(120);
		ssp_send_cmd(ARRAY_AND_SIZE(initialize));
		ssp_send_cmd(ARRAY_AND_SIZE(gamma_setting));
		ssp_send_cmd(ARRAY_AND_SIZE(disp_on));
		msleep(100);
	}
	else
	{
		ssp_send_cmd(ARRAY_AND_SIZE(disp_off));
		ssp_send_cmd(ARRAY_AND_SIZE(sleep_in));
		msleep(60);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 0);
		msleep(120);
	}
}

static struct pxafb_mode_info novatek_l1N304_modes[] = {
	[0] = {
		//.flags			= PXAFB_AUTO_REFRESH,
		.pixclock		= 77000,
		.xres			= 640,
		.yres			= 360,
		.bpp			= 32,
		.hsync_len		= 2,
		.left_margin		= 4,
		.right_margin		= 24,
		.vsync_len		= 2,
		.upper_margin		= 3,
		.lower_margin		= 5,
		//		.sync			= 0,
		//		.update_framedata	= benzglb_lcd_update,
	},
};

static struct pxafb_mach_info benzglb_lcd_info = {
	.modes			= novatek_l1N304_modes,
	.num_modes		= 1,
	.lccr0			= LCCR0_Act | LCCR0_DELAY_LBIAS,
	.lccr3			= LCCR3_PixRsEdg | LCCR3_HSP | LCCR3_VSP,
	.pxafb_lcd_power	= benzglb_lcd_power,
	.pxafb_backlight_power	= benzglb_backlight_power,
};

static void __init benzglb_init_lcd(void)
{
	set_pxa_fb_info(&benzglb_lcd_info);
}

#if defined(CONFIG_LEDS_PXA)
static void benzglb_leds_brightness_set(int value)
{
	if (value) {/* Turn keyboard backlight on */
		micco_write(MICCO_LED1_CONTROL, 0xFF); 
	} else { /* Turn keyboard backlight off */
		micco_write(MICCO_LED1_CONTROL, 0); 
	}
}

static struct leds_platform_data benzglb_leds_info = {
	.set_button_brightness 	= benzglb_leds_brightness_set,
};
#endif

#ifdef CONFIG_PM
static int benzglb_init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
	src->bits.uart1 = 1;
	src->bits.uart2 = 1;
	src->bits.dkey = 1;
	src->bits.onkey = 1;
	src->bits.mmc2_dat1 = 1; //Enable mmc2 wakeup 
	src->bits.usbc = 1;
	src->bits.cmwdt = 1;
	src->bits.psensor = 1;

	return 0;
}

extern int pxa930_query_gwsr(int);
static int benzglb_query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_PMIC)
		src->bits.onkey = 1;
	if (reg & PXA3xx_PM_WE_UART1)
		src->bits.uart1 = 1;
	if (reg & PXA3xx_PM_WE_BTUART)
		src->bits.uart2 = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GPIO) {
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(VR_KEY)))
		    || pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(VOLUME_UP_KEY)))
		    || pxa930_query_gwsr(mfp_to_gpio((VOLUME_DOWN_KEY))))
			src->bits.dkey = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(PRO_DOUT))))
			src->bits.psensor = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_TAVOR))))
			src->bits.uart2 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR))))
			src->bits.mmc2_dat1 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(EXT_CHG_POOG_N))))
			src->bits.usbc = 1;
	}

	return 0;
}

static int benzglb_ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int benzglb_key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.dkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VR_KEY), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VOLUME_UP_KEY),
					    MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VOLUME_DOWN_KEY),
					    MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_PMIC;
		}
	} else {
		if (src.bits.dkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VR_KEY), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VOLUME_UP_KEY),
					    MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(VOLUME_DOWN_KEY),
					    MFP_EDGE_NONE);
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzglb_mmc_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;

	if( enable )
	{
		if( src.bits.mmc2_dat1 && benzglb_wlan_enabled)
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}
	else
	{
		if( src.bits.mmc2_dat1 )
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_TAVOR), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzglb_psensor_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PRO_DOUT),MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}else{
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PRO_DOUT),MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzglb_uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_UART1;
		}
		if (src.bits.uart2 && benzglb_wlan_enabled) {
			//pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_FALL);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_TAVOR), MFP_EDGE_FALL);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_UART_RXD), MFP_EDGE_FALL);

			/* note: on pxa930, uart2 use this bit */
			ret |= PXA3xx_PM_WE_BTUART|PXA3xx_PM_WE_GPIO;
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			//pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_TAVOR), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_UART_RXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzglb_usb_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;

	if (enable) {
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(EXT_CHG_POOG_N), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	} else {	
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(EXT_CHG_POOG_N), MFP_EDGE_NONE);
		}
	}
	return ret;

}

static int benzglb_comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= benzglb_init_wakeup,
	.query	= benzglb_query_wakeup,
	.ext    = benzglb_ext_wakeup,
	.key    = benzglb_key_wakeup,
	.mmc    = benzglb_mmc_wakeup,
	.uart   = benzglb_uart_wakeup,
	.usb	= benzglb_usb_wakeup,
	.cmwdt	= benzglb_comm_wdt_wakeup,
	.psensor = benzglb_psensor_wakeup,
};
#endif

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
	.flags		= PXA_I2C_STANDARD_MODE, 
};

static void __init benzglb_init(void)
{
	/* initialize MFP configurations */
	pxa3xx_mfp_config(ARRAY_AND_SIZE(benzglb_mfp_cfg));
	pxa3xx_mfp_config(ARRAY_AND_SIZE(benzglb_mfp_free_cfg));

	/* register i2c device */
	pxa_set_i2c_info(&pxa3xx_i2c_platform_data);

	/* update ISRAM configration for pxa935 */
	if (cpu_is_pxa935())
		pxa3xx_device_imm.resource = pxa935_resource_imm;

	/* dvfm device */
	set_pxa3xx_freq_info(&benzglb_freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

	/* lcd */
	benzglb_init_lcd();

	/* GPS */
#if defined(CONFIG_SIRF_GSD3TW)
	benzglb_gps_sirf_init();
#endif
	/* WLAN */
	benzglb_init_wlan();

	platform_add_devices(devices, ARRAY_SIZE(devices));
	
	/* PMEM size 16M */
	android_add_pmem("pmem", 0x01000000UL, 1, 1);

	/* uart */
	pxa_set_ffuart_info(NULL);
	pxa_set_btuart_info(NULL);
	pxa_set_stuart_info(NULL);

#if defined(CONFIG_BT_HCIUART_PS)
	benzglb_hci_ps_init();
#endif

#if defined(CONFIG_LEDS_PXA)
	pxa_set_leds_info(&benzglb_leds_info);
#endif
	/* onenand */
#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
	benzglb_init_onenand();
#endif

#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_PXA3XX_U2D_MODULE)
	/* u2d */
	if (device_is_enabled(device_u2d)) {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(benzglb_u2d_cfg));
		pxa_set_u2d_info(&benzglb_u2d_info);
	}
#endif

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &benzglb_u2o_info);
	}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_PXA27X_UDC_MODULE)
	if (device_is_enabled(device_udc)) {
		pxa_set_udc_info(&benzglb_udc_info);
	}
#endif

	/* i2c devices */
	i2c_register_board_info(0, ARRAY_AND_SIZE(benzglb_i2c_board_info));

#if defined(CONFIG_PXA_CAMERA)
	/* initialize camera, shall do it after enable buck3 */
	benzglb_init_cam();
#endif

#if defined(CONFIG_MMC)
	/* MMC card detect & write protect for controller 0 */
	benzglb_mmc_slot[0].gpio_cd  = mfp_to_gpio(MFP_CFG_PIN(MMC0_CD));
	benzglb_init_mmc();
#endif

#if defined(CONFIG_ALSPS)
	/* cm3601_alsps proximity sensor register platform device */
	benzglb_cm3601_alsps_init();
#endif

#if defined(CONFIG_BQ2407x_POWER)
	benzglb_init_charger();
#endif

	/*headset_amp_sd*/
	benzglb_headset_amp_init();

#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65)
	snd_ssp_mfp_cfg_init();
#endif

#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif

	/* enable clock */
	OSCC |= OSCC_TENS0;
}

MACHINE_START(BENZGLB, "Benzglb")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = pxa_map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = benzglb_init,
MACHINE_END

