/*
 *  linux/arch/arm/mach-pxa/benzina2.c
 *
 *  Support for the FBW BENZINA2.
 *
 *  2009-12-01   Song Zhiqiang <zhiqiang.song@borqs.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/onenand.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/i2c.h>
#include <linux/pda_power.h>
#include <linux/pda_battery.h>
#include <linux/gpio_keys.h>
#include <linux/switch_usbuart.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/pxa-regs.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/mfp-pxa3xx.h>
#include <mach/pxa3xx_nand.h>
#include <mach/pxa27x_keypad.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/ssp.h>
#include <mach/gpio.h>
#include <mach/benzina2.h>
#include <mach/micco.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/irda.h>
#include <mach/pxa3xx_u2d.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/ehci.h>
#include <mach/udc.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxa950fb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/i2c.h>
#include <mach/lis331.h>
#include <linux/proc_fs.h>
#include <mach/levante_micco_pmic_id.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <mach/tavor_common.h>
#include <mach/ssp.h>
#include <mach/gps.h>
#include <mach/pmem.h>
#include "devices.h"
#include "generic.h"

#define MAX_SLOTS      3

struct platform_mmc_slot benzina2_mmc_slot[MAX_SLOTS];

extern int is_android(void);
extern int is_comm_v75(void);

static struct gpio_keys_platform_data benzina2_gpio_keys_info;
#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

 static u8 g_pmic_id;

static mfp_cfg_t audio_mfp_cfg[] __initdata = {

	GPIO63_SSP3_CLK,
	GPIO64_SSP3_FRM,
	GPIO65_SSP3_TXD,
	GPIO66_SSP3_RXD,
	GPIO67_SSP3_SYSCLK,

	GPIO68_SSP2_RXD,
	GPIO69_SSP2_TXD,
	GPIO70_SSP2_SYSCLK,
	GPIO71_SSP2_FRM,
	GPIO72_SSP2_CLK,
};

static mfp_cfg_t wlan_on_mfp_cfg[]={
	WLAN_PD_GH381_LPM_ON,
	TAVOR_WAKEUP_BT_LPM_ON,
	BT_UART_RXD_LPM_ON,
	BT_UART_TXD_LPM_ON,
	BT_UART_CTS_LPM_ON,
	BT_UART_RTS_LPM_ON,
	WLAN_BT_RESET_LPM_ON,
	WLAN_DAT3_LPM_ON,
	WLAN_DAT2_LPM_ON,
	WLAN_DAT1_LPM_ON,
	WLAN_DAT0_LPM_ON,
	WLAN_CMD_LPM_ON,
	WLAN_CLK_LPM_ON,
	WLAN_PD_PW886_LPM_ON,
	WLAN_WAKEUP_TAVOR_LPM_ON,
	BT_WAKEUP_TAVOR_LPM_ON,
	TAVOR_WAKEUP_WLAN_LPM_ON,
};

static mfp_cfg_t wlan_off_mfp_cfg[]={
	WLAN_PD_GH381,
	TAVOR_WAKEUP_BT,
	BT_UART_RXD,
	BT_UART_TXD,
	BT_UART_CTS,
	BT_UART_RTS,
	WLAN_BT_RESET,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	WLAN_PD_PW886,
	WLAN_WAKEUP_TAVOR,
	BT_WAKEUP_TAVOR,
	TAVOR_WAKEUP_WLAN,
};

static mfp_cfg_t sd_in_cfg[]={
	MMC1_CLK_ON,
	MMC1_CMD_ON,
	MMC1_DAT0_ON,
	MMC1_DAT1_ON,
	MMC1_DAT2_ON,
	MMC1_DAT3_ON,
	SD_LDO_CE_ON,
};

static mfp_cfg_t sd_out_cfg[]={
	MMC1_CLK,
	MMC1_CMD,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	SD_LDO_CE,
};

static mfp_cfg_t benzina2_mfp_cfg[] __initdata = {
	GPIO8_FREE,
	GPIO43_FREE,
	GPIO49_FREE,
	GPIO50_FREE,
	GPIO62_FREE,
	GPIO67_FREE,
	GPIO70_FREE,
	GPIO80_FREE,
	GPIO83_FREE,
	GPIO92_FREE,
	GPIO93_FREE,
	G2_HW_ID,

	VOLUME_UP_KEY,			
	VOLUME_DOWN_KEY,
	OPTION_KEY,
	HOME_KEY,
	VR_KEY,
	CAM_ON_KEY,
	MOT_INT,
	CAM_SHOT_KEY,

	TP_LDO_EN,
	TP_INT,
	PRO_INT,
	USB2UART_SEL,
	USB2UART_EN,
	CPS_RSTN,
	CPS_INT,
	TAVOR_ONE_WIRE,

	OLED_SPI_CLK,
	OLED_SPI_CS,
	OLED_SPI_DIN,
	OLED_VSYNC,
	OLED_HSYNC,
	OLED_PCLK,
	OLED_RESET_N,
	OLED_DE,

	OLED_D0,
	OLED_D1,
	OLED_D2,
	OLED_D3,
	OLED_D4,
	OLED_D5,
	OLED_D6,
	OLED_D7,
	OLED_D8,
	OLED_D9,
	OLED_D10,
	OLED_D11,
	OLED_D12,
	OLED_D13,
	OLED_D14,
	OLED_D15,
	OLED_D16,
	OLED_D17,
	OLED_D18,
	OLED_D19,
	OLED_D20,
	OLED_D21,
	OLED_D22,
	OLED_D23,

	SD_LDO_CE,
	MMC1_CMD,
	MMC1_CLK,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	MMC0_CD,
/*
	FFRXD,
	FFTXD,
*/	GPS_RX,
	GPS_TX,
	GPS_RESET,
	GPS_ON,
	GPS_LNA_EN,

	HP_AMP_SD,
	AUDIO_AMP_PM_EN,
	PCM_SYNC_OFF,
	PCM_CLK_OFF,
	PCM_TXD_OFF,
	PCM_RXD_OFF,
	/*
	PCM_CLK,
	PCM_SYNC,
	PCM_TXD	,
	PCM_RXD,
	*/
	I2S_RXD,
	I2S_TXD,
	I2S_SYNC,
	I2S_BITCLK,

	CI2C_SCL,
	CI2C_SDA,
	I2C_SCL,
	I2C_SDA,


	TAVOR_WAKEUP_BT,
	WLAN_PD_GH381,
	WLAN_BT_RESET,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	WLAN_PD_PW886,
	WLAN_WAKEUP_TAVOR,
	BT_WAKEUP_TAVOR,
	TAVOR_WAKEUP_WLAN,
	BT_UART_RXD,
	BT_UART_TXD,
	BT_UART_CTS,
	BT_UART_RTS,

	FLASH_EN,
	VCM_PD,
	CAM_MCLK,
	CAMERA_ON,
	CAM_PD_MAIN,
	CAM_RESET_MAIN,
};

#ifdef	CONFIG_PROC_FS
#define	BENZINA2_PROC_FILE	"driver/benzina2"
static struct proc_dir_entry *benzina2_proc_file;
static int pmic_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
   //	struct pmic_ops *ops = file->private_data;
	void __user *uarg = (void __user *)arg;
	int ret = -EINVAL;
    u8 icotl_pmic_id;

	switch (cmd) {
	case IOCTL_PMIC_ID_GET:
	   	icotl_pmic_id = g_pmic_id;
		ret = put_user(icotl_pmic_id, (unsigned long __user *)uarg);
		printk(KERN_ALERT "-------> icotl_pmic_id %d \n",icotl_pmic_id);
		if (ret)
			ret = -EFAULT;
		break;


	default:
	  	//if (ops->ioctl)
	  	//	ret = ops->ioctl(cmd, arg);
		panic("BENZINA2-pmic_ioctl");
		break;
	}
	return ret;
}


static ssize_t benzina2_proc_read(struct file *filp,
        char __user *buffer, size_t length, loff_t *offset)
{
	printk(KERN_ALERT "benzina2_proc_read - no action done\n");

	return 0;
}

static ssize_t benzina2_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	printk(KERN_ALERT "benzina2_proc_write - no action done\n");

	return len;
}

static struct file_operations benzina2_proc_ops = {
	.read 	= benzina2_proc_read,
	.write 	= benzina2_proc_write,
	.ioctl	= pmic_ioctl,

};

static void create_benzina2_proc_file(void)
{
	benzina2_proc_file = create_proc_entry(BENZINA2_PROC_FILE, 0644, NULL);
	if (benzina2_proc_file) {
		benzina2_proc_file->owner = THIS_MODULE;
		benzina2_proc_file->proc_fops = &benzina2_proc_ops;
		printk(KERN_ALERT "BENZINA2 proc file created!\n");
	} else
		printk(KERN_ALERT "BENZINA2 proc file create failed!\n");
}

static void remove_benzina2_proc_file(void)
{
	//remove_proc_entry(BENZINA2_PROC_FILE, &proc_root);
}
#endif  //CONFIG_PROC_FS

static struct platform_device levante_ts_device = {
	.name 		= "levante-ts",
	.id 		= -1,
};

static struct platform_device levante_charger_device = {
	.name		= "levante-charger",
	.id		= -1,
};


static struct platform_device levante_kp_bl_device = {
	.name 		= "levante-kp-bl",
	.id 		= -1,
};

static struct platform_device levante_bl_device = {
       .name           = "levante-bl",
       .id             = -1,
};

static struct resource pxa3xx_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code +
						  * The first 4KB are locked for pxa930 trusted */
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1, /* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa3xx_resource_imm),
	.resource	= pxa3xx_resource_imm,
};

static struct platform_device gpio_keys_device = {
	.name		= "gpio-keys",
	.id		= -1,
	.dev		= {
		.platform_data = &benzina2_gpio_keys_info,
	},
};

static struct platform_device levante_hsdetect_device = {
	.name		= "levante-hsdetect",
	.id		= -1,
};

static struct platform_device *devices[] __initdata = {
   	&pxa3xx_device_imm,
	&gpio_keys_device,
};

static struct platform_device *devices_levante[] __initdata = {
#if defined(CONFIG_PXA9xx_LEVANTE)
 	&levante_bl_device,
    &levante_kp_bl_device,
 	&levante_ts_device,
 	&levante_charger_device,
	&levante_hsdetect_device
#endif
};

static struct android_pmem_platform_data android_pmem_pdata = {
	.name = "pmem",
	/* 
	 * Linux kernel uses 96M(total 128M) , 0xb8000000-0xbe000000
	 *                or 224M(total 256M) , 0xb0000000-0xbe000000 
	 * CP uses 12M, PMEM uses 16M
	 */
	.start = 0xbe000000,
	.size = 0x1000000,      /* 16M */
	.no_allocator = 1,
	.cached = 1,
};

#ifdef CONFIG_KEYBOARD_GPIO
static struct gpio_keys_button benzina2_gpio_keys_button[] = {
	[0] = {
		.desc	= "back_key",
		.code	= KEY_BACK,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VR_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[1] = {
		.desc	= "cam_shot_key_n",
		.code	= KEY_CAMERA,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_SHOT_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[2] = {
		.desc	= "cam_on_key",
		.code	= KEY_FOCUS,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_ON_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[3] = {
		.desc	= "volume_up_key",
		.code	= KEY_VOLUMEUP,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_UP_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[4] = {
		.desc	= "volume_down_key",
		.code	= KEY_VOLUMEDOWN,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_DOWN_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[5] = {
		.desc	= "menu_key",
		.code	= KEY_MENU,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(OPTION_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[6] = {
		.desc	= "home_key",
		.code	= KEY_HOME,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(HOME_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},

};

static struct gpio_keys_platform_data benzina2_gpio_keys_info = {
	.buttons = benzina2_gpio_keys_button,
	.nbuttons = (sizeof(benzina2_gpio_keys_button)
			/ sizeof(struct gpio_keys_button)),
	//.trigger_key0 = KEY_BACK,
	//.trigger_key1 = KEY_CAMERA,
};
#endif

#if defined(CONFIG_SIRF_GSD3TW)
static int benzina2_gps_set_power(int on_off)
{
	if( on_off )
	{
	}else
	{
	}
	return 0;
}
static struct sirf_platform_data benzina2_sirf_platform_data = {
	.reset_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)),
	.on_off_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_ON)),
	.set_power = benzina2_gps_set_power,
};

static struct platform_device benzina2_sirf_device = {
	.name = "sirf",
	.id = -1,
	.dev = {
		.platform_data = &benzina2_sirf_platform_data,
	},
};

static void __init benzina2_gps_sirf_init(void)
{
	printk(KERN_ERR"[%s:%d] gps pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), "gps reset")) {
		printk(KERN_ERR "<SiRF> GPIO RST request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), "gps on_off")) {
		printk(KERN_ERR "<SiRF> GPIO ON_OFF request error!!\n");
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), 0); /* GPS_ON_OFF */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), 1); /* GPS_RST */

	printk(KERN_ERR"GPS set up finished\n");
	platform_device_register(&benzina2_sirf_device);
}
#endif



#if defined(CONFIG_MMC) || defined(CONFIG_MMC_MODULE)
static int benzina2_mci_init(struct device *dev,
		irq_handler_t benzina2_detect_int,
			     void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(benzina2_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = benzina2_mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for benzina2 MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, benzina2_detect_int,
			  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			  "MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void benzina2_mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(benzina2_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = benzina2_mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static void benzina2_mci_setpower(struct device *dev, unsigned int vdd)
{
        struct pxamci_platform_data* p_d = dev->platform_data;

        if (( 1 << vdd) & p_d->ocr_mask) {
                gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 1);
                pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_in_cfg));
        } else {
                gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 0);
                pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_out_cfg));
        }
}

static struct pxamci_platform_data benzina2_mci_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
	.init 		= benzina2_mci_init,
	.exit		= benzina2_mci_exit,
	.setpower	= benzina2_mci_setpower,
	.gpio_detect	= mfp_to_gpio(MFP_CFG_PIN(MMC0_CD)),
};

static struct pxamci_platform_data benzina2_mci2_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34|MMC_VDD_165_195,
};

#if defined(CONFIG_EVB3_EMMC)
static mfp_cfg_t pxa9xx_mmc3_mfp_cfg[] __initdata = {
	/* MMC3 of Tavor PV2 */
	DF_ADDR0_MMC3_CMD,
	DF_ADDR1_MMC3_CLK,
	DF_ADDR2_MMC3_DAT0,
	DF_ADDR3_MMC3_DAT1,
	nXCVREN_MMC3_DAT2,
	nLUA_MMC3_DAT3,
	DF_nWE_MMC3_DAT4,
	nBE0_MMC3_DAT5,
	nBE1_MMC3_DAT6,
	DF_nRE_MMC3_DAT7,
};
#endif

static mfp_cfg_t pxa9xx_mmc5_mfp_cfg[] __initdata = {
	GPIO91_MMC5_CLK,
	GPIO90_MMC5_CMD,
};

static void __init benzina2_init_mmc(void)
{
#if defined(CONFIG_EVB3_EMMC)
		//pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa9xx_mmc3_mfp_cfg));
		pxa9xx_set_mci3_info(&tavor_evb3_mci2_platform_data);
#endif
	pxa9xx_set_mci4_info(&benzina2_mci_platform_data);
	pxa9xx_set_mci5_info(&benzina2_mci2_platform_data);
	/* MMC card detect for controller 1 of pxa9xx_mci */
        benzina2_mmc_slot[1].gpio_cd  = mfp_to_gpio(MFP_CFG_PIN(MMC0_CD));
}
#else
static inline void benzina2_init_mmc(void) {}
#endif

#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
static struct flash_platform_data benzina2_onenand_info;
static void __init benzina2_init_onenand(void)
{
	/* Now, kernel get mtd partitions from commandline.*/
	benzina2_onenand_info.parts = android_512m_v75_partitions;
	benzina2_onenand_info.nr_parts = ARRAY_SIZE(android_512m_v75_partitions);

	pxa3xx_device_onenand.dev.platform_data = &benzina2_onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_PXA27X_UDC_MODULE)
static mfp_cfg_t pxa930_otg_init_pins[] = {
	GPIO64_GPIO,
};

static mfp_cfg_t pxa930_otg_pins[] = {
	GPIO64_USB_P2_7,
};

int benzina2_udc_is_miniA(void)
{
	int otg_id = mfp_to_gpio(MFP_PIN_GPIO64);
	int id_value;
	int err;

	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_init_pins));
	err = gpio_request(otg_id, "OTG ID");
	if (err) {
		gpio_free(otg_id);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", otg_id, err);
		return 0;
	}
	gpio_direction_input(otg_id);
	id_value = gpio_get_value(otg_id);
	gpio_free(otg_id);
	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_pins));

	return (id_value == 0);
}

static struct pxa2xx_udc_mach_info benzina2_udc_info = {
	.udc_is_miniA = benzina2_udc_is_miniA,
};
#endif

#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)

void benzina2_reset_xcvr(void)
{
	int reset_pin;
	int err;

	reset_pin = mfp_to_gpio(MFP_PIN_GPIO43);
	err = gpio_request(reset_pin, "U2D Reset");
	if (err) {
		gpio_free(reset_pin);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", reset_pin, err);
		return;
	}
	gpio_direction_output(reset_pin, 0);
	mdelay(100);
	gpio_set_value(reset_pin, 1);

	gpio_free(reset_pin);
}

static struct pxa3xx_u2d_mach_info benzina2_u2d_info = {
	.reset_xcvr = benzina2_reset_xcvr,
};
#endif /* CONFIG_USB_PXA3XX_U2D || CONFIG_USB_PXA3XX_U2D_MODULE */

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info benzina2_u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#ifdef CONFIG_USB_EHCI_PXA9XX

static void __init benzina2_init_ehci(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_ehci_info(&benzina2_u2o_info);
}
#else
static inline void benzina2_init_ehci(void) {}
#endif

#if defined (CONFIG_USB_OTG) && defined(CONFIG_USB_PXA9XX_U2O)
static void __init benzina2_init_otg(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_u2otg_info(&benzina2_u2o_info);
}
#else
static inline void benzina2_init_otg(void) {}
#endif

static int lis331_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(MOT_INT)));

	return 0;
}

static int lis331_ack_irq(void)
{
	return 0;
}

static int lis331_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))) ;
}

static struct lis331_platform_data lis331_data = {
	.init_irq = lis331_init_irq,
	.ack_irq = lis331_ack_irq,
	.get_gpio_value = lis331_get_gpio_value,
	.platform_init = NULL,
	.power_chips = NULL,
	.xyz_dir = Y_POS,
};


#if defined(CONFIG_PXA9xx_LEVANTE)
static int levante_init_irq(void)
{
	pxa3xx_mfp_set_edge(MFP_CFG_PIN(MFP_PIN_GPIO83),
					MFP_EDGE_FALL);
	return 0;
}

static int levante_ack_irq(void)
{
	return 0;
}

static void benzina2_levante_init(void)
{
	u8 data;
	printk(KERN_ALERT "----->benzina2_levante_init\n");

	/* Set V_LDO12 to 2.8V */
	sanremo_write(SANREMO_LDO12, 0x33);
	sanremo_read(SANREMO_SUPPLIES_EN12, &data);
	data|= 0x20;
	sanremo_write(SANREMO_SUPPLIES_EN12,data);
}

/* power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module levante_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		BUCK2},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,	LDO3},
	{VCC_SDIO,		0},
	{VCC_CAMERA_ANA, LDO3},
	{VCC_USB,		LDO5},
	{VCC_LCD,		0},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,	BUCK2}, /* ? */
	{VCC_1P8V,		0}, 	/* ? */
	{VCC_MEM,		BUCK2}, /* ? */
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

static struct power_chip levante_chips[] = {
	{LEVATNE_A0_ID,	"levanteA0",	levante_power_modules},
	{LEVATNE_A1_ID,	"levanteA1",	levante_power_modules},
	{LEVATNE_END_ID,"levante",	levante_power_modules},
	{0,	NULL,		NULL},
};
static struct levante_platform_data  levante_data = {
    .platform_init = benzina2_levante_init
};

static struct sanremo_platform_data  sanremo_data = {
    .init_irq = levante_init_irq,
    .ack_irq = levante_ack_irq,
    .power_chips = levante_chips,
};

static struct portofino_platform_data  portofino_data = {
};

#endif /* CONFIG_PXA9xx_LEVANTE */

#if defined(CONFIG_PXA950_CAMERA)
static int cam_power_set(int flag)
{
	struct clk *axi_clk = NULL;
	struct clk *sci1_clk = NULL;
	struct clk *sci2_clk = NULL;

	sci1_clk = clk_get(NULL, "SCI1CLK");
	if (IS_ERR(sci1_clk)) {
		printk(KERN_ERR "unable to get SCI1CLK\n");
		return PTR_ERR(sci1_clk);
	}

	sci2_clk = clk_get(NULL, "SCI2CLK");
	if (IS_ERR(sci2_clk)) {
		printk(KERN_ERR "unable to get SCI2CLK\n");
		return PTR_ERR(sci2_clk);
	}

	if(flag){
		axi_clk = clk_get(NULL, "AXICLK");
		if (IS_ERR(axi_clk)) {
			printk(KERN_ERR "unable to get AXICLK\n");
			return PTR_ERR(axi_clk);
		}

		clk_enable(axi_clk);
		clk_enable(sci1_clk);
		clk_enable(sci2_clk);

		clk_put(axi_clk);
	}else{
		clk_disable(sci1_clk);
		clk_disable(sci2_clk);
	}

	clk_put(sci1_clk);
	clk_put(sci2_clk);

	return 0;
}
/* sensor init */
static int sensor_power_set(int res, int flag)
{
	/*
	 * RES, 0, LOW RESOLUTION, 1 HIGH RESOLUTION
	 * FLAG, 1: ON, 0: OFF
	 * GPIO93 LOW FOR HIGH
	 * GPIO14 LOW FOR LOW
	 */
	int high_pin = mfp_to_gpio(MFP_PIN_GPIO93);
	int low_pin = mfp_to_gpio(MFP_PIN_GPIO14);

	cam_power_set(flag);

	if (gpio_request(high_pin, "CAM_EANBLE_HI_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", high_pin);
		return -EIO;
	}

	if (res) {
		if (flag)
			gpio_direction_output(high_pin, 0);	/* enable */

		else
			gpio_direction_output(high_pin, 1);	/* disable */

	} else{
		if (gpio_request(low_pin, "CAM_EANBLE_LOW_SENSOR")) {
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d \n", low_pin);
			return -EIO;
		}
		if (flag) {
			/* high resolution is bridge */
			gpio_direction_output(high_pin, 0);	/* enable */
			gpio_direction_output(low_pin, 0);	/* enable */

		}
#if 0
		/* FIXME */
		else
			gpio_direction_output(low_pin, 1);	/* disable */
#endif
		gpio_free(low_pin);
	}

	gpio_free(high_pin);
	return 0;
}

static struct sensor_platform_data ov3640_sensor_data = {
	.id = SENSOR_HIGH,
	.power_set = sensor_power_set,
};

static struct sensor_platform_data ov7690_sensor_data = {
	.id = SENSOR_LOW,
	.power_set = sensor_power_set,
};

/* sensor init over */

static struct cam_platform_data cam_ops = {
	.power_set	= cam_power_set,
};

static mfp_cfg_t cam_mfp_cfg[] __initdata = {
	/* CAM */
};

static void __init benzina2_init_cam(void)
{
	pxa950_device_cam.dev.platform_data = &cam_ops;
	pxa3xx_mfp_config(ARRAY_AND_SIZE(cam_mfp_cfg));
	platform_device_register(&pxa950_device_cam);
	platform_device_register(&pxa950_device_csi);
}

#endif

static struct i2c_board_info benzina2_i2c_hs_info[] = {
	/* hs i2c device */
#if defined(CONFIG_SENSORS_LIS331)
	{
		.type		= "lis331",
		.addr		= 0x1C,
		.platform_data	= &lis331_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))),
	},
#endif
#if CONFIG_SENSORS_CYWEE 
 	{
		.type		= "cywee",
		.addr		= 0x78,
        },
#endif
#if CONFIG_SENSORS_YAS529 
{
		.type		= "fx_compass",
		.addr		= 0x2e,
        },
#endif    
#if defined(CONFIG_PXA950_CAMERA)
	{
		.type		= "ov3640",
		.addr		= 0x3c,
		.platform_data	= &ov3640_sensor_data,
	},
	{
		.type		= "ov7690",
		.addr		= 0x21,
		.platform_data	= &ov7690_sensor_data,
	},
#endif
};

static struct i2c_board_info benzina2_i2c_board_info_levante1[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "sanremo",
		.addr		= 0x34,
		.platform_data	= &sanremo_data,
	   	.irq		= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO83)),
	},
	{
		.type	= "portofino",
		.addr		= 0x10,
		.platform_data	= &portofino_data,
	},
#endif

};

static struct i2c_board_info benzina2_i2c_board_info_levante2[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "levante",
           .platform_data  = &levante_data,
	},
#endif
};

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_FAST_MODE, /* PXA_I2C_STANDARD_MODE - SLOW */
};

static struct i2c_pxa_platform_data pxa3xx_i2c_hs_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_STANDARD_MODE,
};

struct pxa3xx_freq_mach_info benzina2_freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};

#if defined(CONFIG_FB_PXA950)

#if 0  //TODO: Lincoln: move out the logic.
static void benzglb_backlight_power( int on, int intensity)
{
	if( on) {
		CKENB |= 1<<(CKEN_PWM0-32);
		PWM_CTRL0 = 0x1;
		PWM_PWDUTY0 = intensity;
		PWM_PERVAL0 = 0xfe;
	} else {
		PWM_CTRL0 = 0x0;
		PWM_PWDUTY0 = 0x0;
		PWM_PERVAL0 = 0x0;
		CKENB &= ~(1<<(CKEN_PWM0-32));
	}
}
#endif

/* use bit 30, 31 as the indicator of command parameter number */
#define CMD0(x)		((0x00000000) | ((x)))
#define DAT0(x)		((0x00000100) | ((x)))
#define CMD1(x, x1)	((0x40000000) | ((x) << 9) | 0x100 | (x1))
#define CMD2(x, x1, x2)	((0x80000000) | ((x) << 18) | 0x20000 |\
			((x1) << 9) | 0x100 | (x2))
#define BENZINA2_LCD_DSS 	9 	//Data size select for LCD spi

static uint32_t sleep_out[] = {
	CMD0(0x11),
};

static uint32_t gamma_setting[] = {
	CMD0(0xC0),
	DAT0(0x0F),
	DAT0(0x0B),
	DAT0(0x0F),
	DAT0(0x0B),
	DAT0(0x6D),
	DAT0(0x55),
	DAT0(0x20),
	DAT0(0x08),
	DAT0(0x70),
	DAT0(0x57),
	DAT0(0x1E),
	DAT0(0x06),
	DAT0(0x08),
	DAT0(0x08),

	CMD0(0xE0),
	DAT0(0x0E),
	DAT0(0x09),
	DAT0(0x1A),
	DAT0(0x2C),
	DAT0(0x43),
	DAT0(0x4C),
	DAT0(0x47),
	DAT0(0x3F),
	DAT0(0x3D),
	DAT0(0x2C),
	DAT0(0x23),
	DAT0(0x25),
	DAT0(0x34),
	DAT0(0x1B),
	DAT0(0x2B),

	CMD0(0xE1),
	DAT0(0x1F),
	DAT0(0x0B),
	DAT0(0x20),
	DAT0(0x36),
	DAT0(0x53),
	DAT0(0x56),
	DAT0(0x47),
	DAT0(0x44),
	DAT0(0x3D),
	DAT0(0x36),
	DAT0(0x2F),
	DAT0(0x30),
	DAT0(0x35),
	DAT0(0x1D),
	DAT0(0x36),

	CMD0(0xE2),
	DAT0(0x0E),
	DAT0(0x09),
	DAT0(0x1A),
	DAT0(0x2C),
	DAT0(0x43),
	DAT0(0x4C),
	DAT0(0x47),
	DAT0(0x3F),
	DAT0(0x3D),
	DAT0(0x2C),
	DAT0(0x23),
	DAT0(0x25),
	DAT0(0x34),
	DAT0(0x1B),
	DAT0(0x2B),
};

static uint32_t initialize[] = {
	CMD0(0xFC),
	DAT0(0x00),
	DAT0(0x10),
	DAT0(0x00),
	DAT0(0x00),

	CMD0(0xD1),
	DAT0(0x84),
	DAT0(0x98),
	DAT0(0x86),
	DAT0(0x86),

	CMD0(0x3A),
	DAT0(0x60),

	CMD0(0xF3),
	DAT0(0xB0),

	CMD0(0xF2),
	DAT0(0x00),
	DAT0(0x09),

	CMD0(0xC1),
	DAT0(0x07),
};

static uint32_t disp_on[]={
	CMD0(0x11),

	CMD0(0x29),
};

static uint32_t disp_off[]={
	CMD0(0x28),
};

static uint32_t sleep_in[]={
	CMD0(0x10),
};

static void ssp_reconfig(struct ssp_dev *dev, int nparam)
{
	static int last_nparam = -1;
	int sscr0;

	/* check if it is necessary to re-config SSP */
	if (nparam == last_nparam)
		return;

	ssp_disable(dev);
	if( nparam >16 )
		sscr0 = 0x00107f80 | ((nparam-1) & 0xf);
	else
		sscr0 = 0x00007f80 | ((nparam-1) & 0xf);

	ssp_config(dev, sscr0, 0x18, 0, 0);
	last_nparam = nparam;
}

static void ssp_send_cmd(uint32_t *cmd, int num)
{
	static int ssp_initialized = 0;
	static struct ssp_dev ssp1;

	int i;

	if (!ssp_initialized) {
		ssp_init(&ssp1, 1, SSP_NO_IRQ);
		ssp_initialized = 1;
		clk_disable(ssp1.ssp->clk);
	}

	clk_enable(ssp1.ssp->clk);
	for (i = 0; i < num; i++, cmd++) {
		ssp_reconfig(&ssp1, BENZINA2_LCD_DSS);
		ssp_write_word(&ssp1, *cmd & 0x3fffffff);

		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp1);
	}
	clk_disable(ssp1.ssp->clk);
}

static void benzina2_lcd_power(int on, struct fb_var_screeninfo *var)
{
	static int pin_requested = 0;

	if( !pin_requested) {
		if( gpio_request( mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 
			"LCD controller reset"))
			return;
		pin_requested = 1;
	}

	if( on)
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 0);
        	udelay(100);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 1);
		msleep(120);
		ssp_send_cmd(ARRAY_AND_SIZE(initialize));
		ssp_send_cmd(ARRAY_AND_SIZE(gamma_setting));
		ssp_send_cmd(ARRAY_AND_SIZE(disp_on));
		msleep(100);
	}
	else
	{
		ssp_send_cmd(ARRAY_AND_SIZE(disp_off));
		ssp_send_cmd(ARRAY_AND_SIZE(sleep_in));
		msleep(60);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(OLED_RESET_N)), 0);
		msleep(120);
	}
}

static void benzina2_reconfig_panel(int on)
{
	if( on)
	{
		ssp_send_cmd(ARRAY_AND_SIZE(sleep_out));
		msleep(120);
		ssp_send_cmd(ARRAY_AND_SIZE(initialize));
		ssp_send_cmd(ARRAY_AND_SIZE(disp_on));
		msleep(100);
	}
	else
	{
		ssp_send_cmd(ARRAY_AND_SIZE(disp_off));
		ssp_send_cmd(ARRAY_AND_SIZE(sleep_in));
		msleep(120);
	}
}

static struct fb_videomode novatek_l1N304_modes[] = {
	[0] = {
		//.flags			= PXAFB_AUTO_REFRESH,
		.pixclock		= 77000,
		.refresh        = 60,
		.xres			= 640,
		.yres			= 360,
//		.bpp			= 32,
		.hsync_len		= 2,
		.left_margin		= 4,
		.right_margin		= 24,
		.vsync_len		= 2,
		.upper_margin		= 3,
		.lower_margin		= 5,
		.sync			= 0,
		//		.update_framedata	= benzglb_lcd_update,
	},
};


//Lincoln: power control. follow benzglb.

//Lincoln: TODO: adding back light logic.
static struct pxa950fb_mach_info benzina2_lcd_info __initdata = {
	.id                     = "Base",
	.modes                  = novatek_l1N304_modes,
	.num_modes              = ARRAY_SIZE(novatek_l1N304_modes),
	.pix_fmt_in             = PIX_FMTIN_RGB_32,
//Lincoln: it is 24-bit Panel.
	.pix_fmt_out            = PIX_FMTOUT_18_RGB666,
	.panel_type             = LCD_Controller_Active,
	.mixer0                 = LCD_M2PARALELL_CONVERTER,
	.mixer1                 = LCD_MIXER_DISABLE,
	.active                 = 1,
	.pxa950fb_lcd_power     = benzina2_lcd_power,
//Lincoln: TODO: need check spec.
	.invert_pixclock        = 1,
};

static void __init benzina2_init_lcd(void)
{
	set_pxa950_fb_info(&benzina2_lcd_info);
//	set_pxa950_fb_ovly_info(&benzina2_lcd_ovly_info);
}

#endif

#ifdef CONFIG_PM
extern irqreturn_t uartrx_detect_irq(int irq, void *_dev);

static mfp_cfg_t benzina2_ffuart_enable_mfp_cfg[] = {
	/* FFUART */
	GPIO53_FFRXD,
	GPIO54_FFTXD,
	DF_ADDR0_GPIO_53,
};

static mfp_cfg_t benzina2_ffuart_disable_mfp_cfg[] = {
	/* FFUART */
	GPIO53_GPIO,
	GPIO54_GPIO,
	DF_ADDR0_NONE,
};

extern int is_uart_gpio(void);
static int benzina2_ffuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO53);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);
		/* if edge detection is not clear, the altfun switch will
		 * gpio interrupt is generated on this gpio pin when
		 * switching pins from UART to gpio. so clear it.
		 */
		GRER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		GFER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(benzina2_ffuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(benzina2_ffuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
		GRER(gpio_rx) |= GPIO_bit(gpio_rx);
		GFER(gpio_rx) |= GPIO_bit(gpio_rx);

		ret = gpio_request(gpio_rx, "FFUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info benzina2_ffuart_info = {
	.uart_pin_switch = benzina2_ffuart_switch,
};

static mfp_cfg_t benzina2_stuart_enable_mfp_cfg[] = {
	/* STUART */
	GPIO45_STRXD,
	GPIO46_STTXD,
};

static mfp_cfg_t benzina2_stuart_disable_mfp_cfg[] = {
	/* STUART */
	GPIO45_GPIO,
	GPIO46_GPIO,
};

static int benzina2_stuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO45);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);

		pxa3xx_mfp_config(ARRAY_AND_SIZE(benzina2_stuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(benzina2_stuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);

		ret = gpio_request(gpio_rx, "STUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info benzina2_stuart_info = {
	.uart_pin_switch = benzina2_stuart_switch,
};

static int benzina2_init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
	src->bits.uart1 = 1;
	src->bits.mkey = 1;
	src->bits.eth = 1;
	src->bits.tsi = 1;
	src->bits.cmwdt = 1;
	return 0;
}

extern int pxa930_query_gwsr(int);
static int benzina2_query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_KP)
		src->bits.mkey = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(3))
		src->bits.tsi = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(9))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart1 = 1;
		}
	}
	if (reg & PXA3xx_PM_WE_GENERIC(2))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart2 = 1;
		}
	}

	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(13)) {
		if (pxa930_query_gwsr(97))
			src->bits.eth = 1;
		if (pxa930_query_gwsr(53))
			src->bits.uart1 = 1;
	}
	return 0;
}

static int benzina2_ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int benzina2_key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_KP;
		}
	} else {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzina2_mmc_wakeup(pm_wakeup_src_t src, int enable)
{
#if 0
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mmc1_cd) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_NONE);
#endif
	return 0;
}

static int benzina2_uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(9);
			}
		}
		if (src.bits.uart2) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_FALL);
				/* note: on pxa930, uart2 use this bit */
				ret |= PXA3xx_PM_WE_GENERIC(2);
			}
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzina2_eth_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzina2_tsi_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(3);
		}
	} else {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int benzina2_comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= benzina2_init_wakeup,
	.query	= benzina2_query_wakeup,
	.ext    = benzina2_ext_wakeup,
	.key    = benzina2_key_wakeup,
	.mmc    = benzina2_mmc_wakeup,
	.uart   = benzina2_uart_wakeup,
	.eth	= benzina2_eth_wakeup,
	.tsi	= benzina2_tsi_wakeup,
	.cmwdt	= benzina2_comm_wdt_wakeup,
};
#else
static struct pxa_uart_mach_info benzina2_ffuart_info = NULL;
static struct pxa_uart_mach_info benzina2_stuart_info = NULL;
#endif

void benzina2_stereo_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO84, "STEREO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO84);
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d return :%d\n", MFP_PIN_GPIO84, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO84, value);
	gpio_free(MFP_PIN_GPIO84);
}

void benzina2_ext_mono_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO63, "MONO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO63);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", MFP_PIN_GPIO63, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO63, value);
	gpio_free(MFP_PIN_GPIO63);
}

static void benzina2_set_lowpower_mfp(void)
{
	/* It will change mfp value set by OBM to save more power,
	 * it should be merged into OBM mfp setting eventually. */

	/* these values have been merged int OBM
	 * pxa3xx_mfp_write(48, 0xa440);
	 * pxa3xx_mfp_write(49, 0xa440);
	 * pxa3xx_mfp_write(51, 0xa440);
	 */
}

#ifdef CONFIG_BATTERY_PDA
/*****************************************************************************/
/*                      pda_battery                                          */
/*****************************************************************************/
static struct pda_battery_pdata pmic_pda_battery_pdata = {
	.name             = "battery",
};

static struct platform_device pda_battery_device = {
	.name = "pda_battery",
	.id = -1,
	.dev = {
		.platform_data = &pmic_pda_battery_pdata,
	},
};

static void benzina2_init_pda_battery()
{
	platform_device_register(&pda_battery_device);
}
#endif


#ifdef CONFIG_PDA_POWER

static int pda_power_is_ac_online(void)
{
	return 0;
}

static int pda_power_is_usb_online(void)
{
	return 1;
}


static char *pda_power_supplicants[] = {
	"battery",
	"mon_battery",
};

static struct pda_power_pdata pda_power_supply_info = {
	.is_ac_online           = pda_power_is_ac_online,
	.is_usb_online          = pda_power_is_usb_online,
	.supplied_to            = pda_power_supplicants,
	.num_supplicants        = ARRAY_SIZE(pda_power_supplicants),
};

static struct platform_device pda_power_supply = {
	.name = "pda-power",
	.id = -1,
	.dev = {
		.platform_data = &pda_power_supply_info,
	},
};

static void benzina2_init_pda_power()
{
	platform_device_register(&pda_power_supply);
}
#endif

#ifdef CONFIG_SWITCH_USBUART
static unsigned int uu_event;
static enum usbuart_state uu_state = USBUART_AP_UART;
static int benzina2_usbuart_set_state(enum usbuart_event event, int on)
{
	if (on)
		uu_event |= event;
	else
		uu_event &= ~event; 

	if (uu_event) {
		console_ll_stop();
		gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 0);
		uu_state = USBUART_AP_USB;
	} else {
		gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 1);
		uu_state = USBUART_AP_UART; 
		if (get_boot_mode() != BOOT_MODE_SDL_FLASHMODE)
		      console_ll_start();
	}
	return 0;
}

static enum usbuart_state benzina2_usbuart_get_state(void)
{
	return uu_state;
}

static struct usbuart_switch_platform_data benzina2_usbuart_switch_pdata = {
	.name		= "switch-usbuart",
	.set_state	= benzina2_usbuart_set_state,
	.get_state	= benzina2_usbuart_get_state,
};

static struct platform_device usbuart_switch_device = {
	.name		= "switch-usbuart",
	.dev		= {
		.platform_data = &benzina2_usbuart_switch_pdata,
	},
};

static void __init benzina2_init_usbuart_switch(void)
{
	int err;
	err = gpio_request(MFP_CFG_PIN(USB2UART_EN), "usb2uart_en");
	BUG_ON(err);
	err = gpio_request(MFP_CFG_PIN(USB2UART_SEL),"usb2uart_sel");
	BUG_ON(err);

	gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL),1);
	gpio_direction_output(MFP_CFG_PIN(USB2UART_EN), 1);
	uu_state = USBUART_AP_UART;
	if (get_boot_mode() != BOOT_MODE_SDL_FLASHMODE)
	      console_ll_start();
	platform_device_register(&usbuart_switch_device);
	printk(KERN_CRIT "usbuart is swith to uart.\n");
}
#else
static inline void benzina2_init_usbuart_switch(void)
{
	int err;
	err = gpio_request(MFP_CFG_PIN(USB2UART_EN), "usb2uart_en");
	BUG_ON(err);
	err = gpio_request(MFP_CFG_PIN(USB2UART_SEL),"usb2uart_sel");
	BUG_ON(err);

	gpio_direction_output(MFP_CFG_PIN(USB2UART_EN), 1);
	gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 1);
}
#endif

/**************************************************
 ** benzina2_audio_amp_pm_en_read: read the speaker PA status
 ** 1: enable
 ** 0: disable
 ***************************************************/ 
static ssize_t benzina2_audio_amp_pm_en_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	int value;

	value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN))) ? 1 : 0;

	printk(KERN_INFO "<%s>: [read] after copying to user:  value = %d\n", "AUDIO_AMP_PM_EN", value); 

	if ((count = sprintf(buf, "%d", value)) < 0)
		return -EFAULT;

	printk(KERN_INFO "<%s>: [read] after copying to user:  buf = %s\n", "AUDIO_AMP_PM_EN", buf);

	return count;
}

/**************************************************
 ** benzina2_audio_amp_pm_en_write: set the speaker PA status  
 ** 1: enable.
 ** 0: disable.
 ***************************************************/ 
static ssize_t benzina2_audio_amp_pm_en_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	char value[2];
	int ret;

	if ((count = sscanf(buf, "%1s", value)) < 0)
		return -EFAULT;

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN)), "AUDIO_AMP_PM_EN");
	if (ret)
	{
		gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));
		printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_AMP_PM_EN", 
                                                                    MFP_CFG_PIN(AUDIO_AMP_PM_EN), ret);
		return -1;
	}

	if (value[0] == '1')
	{
		gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AUDIO_AMP_PM_EN), MFP_LPM_DRIVE_HIGH);
	}
	else if (value[0] == '0')
	{
		gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AUDIO_AMP_PM_EN), MFP_LPM_DRIVE_LOW);
	}

	gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	return count;
}

static struct bin_attribute speaker_mode_attr = {
	.attr = {
		.name = "audio_amp_pm_en",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = 4,
	.read = benzina2_audio_amp_pm_en_read,
	.write = benzina2_audio_amp_pm_en_write,
};

static int benzina2_audio_amp_pm_en_init(void)
{
	int ret = 0;

	ret = sysfs_create_bin_file(kernel_kobj, &speaker_mode_attr);
	if (ret)
	{
		printk(KERN_ERR "<%s> Failed to create sys file\n", "AUDIO_AMP_PM_EN");
		return -ENOMEM;
	}

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN)), "AUDIO_AMP_PM_EN");
	if (ret)
	{
		gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));
		printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_AMP_PM_EN", 
                                                            MFP_CFG_PIN(AUDIO_AMP_PM_EN), ret);
		return -1;
	}

	gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 0);

	gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	printk(KERN_INFO "<%s> GPIO%d setup done\n", "AUDIO_AMP_PM_EN", MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	return 0;
}

//beacon headset amplifier seting.

/**************************************************
 ** benzina2_hp_amp_sd_read: read the headset amplifier status.
 ** 0: disable
 ** 1: enable
 ***************************************************/ 
static ssize_t benzina2_hp_amp_sd_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    int value;

    value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD))) ? 1 : 0;
                        
    printk(KERN_INFO "<%s>: [read] after copying to user:  value = %d\n", "HP_AMP_SD", value); 
                            
    if ((count = sprintf(buf, "%d", value)) < 0)
        return -EFAULT;
                                
    printk(KERN_INFO "<%s>: [read] after copying to user:  buf = %s\n", "HP_AMP_SD", buf);

    return count;
}

/**************************************************
 ** benzina2_hp_amp_sd_write: set the headset amplifier.  
 ** 0: close headset amplifier.
 ** 1: open headset amplifier.
 ***************************************************/ 
static ssize_t benzina2_hp_amp_sd_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    char value[2];
    int ret;

    if ((count = sscanf(buf, "%1s", value)) < 0)
        return -EFAULT;

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(MFP_CFG_PIN(HP_AMP_SD));
        printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "HP_AMP_SD",                        
                                                                    MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    if (value[0] == '1')
    {
        gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 1);
    }
    else if (value[0] == '0')
    {
        gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 0);
    }

    gpio_free(MFP_CFG_PIN(HP_AMP_SD));

    return count;
}

static struct bin_attribute headset_mode_attr = {
    .attr = {
            .name = "hp_amp_sd",
            .owner = THIS_MODULE,
            .mode = S_IRUGO | S_IWUGO,
            },
            .size = 4,
            .read = benzina2_hp_amp_sd_read,
            .write = benzina2_hp_amp_sd_write,
};

static int benzina2_hp_amp_sd_init(void)
{
    int ret = 0;

    ret = sysfs_create_bin_file(kernel_kobj, &headset_mode_attr);
    if (ret)
    {
        printk(KERN_ERR "<%s> Failed to create sys file\n", "HP_AMP_SD");
        return -ENOMEM;
    }

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(MFP_CFG_PIN(HP_AMP_SD));
        printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "HP_AMP_SD", MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 0);

    gpio_free(MFP_CFG_PIN(HP_AMP_SD));

    printk(KERN_INFO "<%s> GPIOs setup done\n", "HP_AMP_SD");

    return 0;
}

static void __init benzina2_init(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(benzina2_mfp_cfg));

	/* register i2c device */
	pxa_set_i2c_info(&pxa3xx_i2c_platform_data);
	pxa_set_hs_i2c_info(&pxa3xx_i2c_hs_platform_data);

	/* update ISRAM configration for pxa935 */
	if (cpu_is_pxa935())
		pxa3xx_device_imm.resource = pxa935_resource_imm;

    g_pmic_id = pmic_id();

	/* dvfm device */
	set_pxa3xx_freq_info(&benzina2_freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

#if defined(CONFIG_SND_PXA3XX_SOC_BENZINA2)
	/* audio */
	//pxa3xx_mfp_config(ARRAY_AND_SIZE(audio_mfp_cfg));  /* mfp should NOT be configured here! */
#endif

	/* lcd */
#if defined(CONFIG_FB_PXA950)
	benzina2_init_lcd();
#endif

    if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_END_ID) ){
		platform_add_devices(devices_levante, ARRAY_SIZE(devices_levante));
		printk(KERN_ALERT "--------->platform_add_devices(devices_levante)!!!\n");
	}else
	   printk(KERN_ALERT "--------->platform_add_devices(no devices_micco & no devices_levante )!!!\n");

	platform_add_devices(devices, ARRAY_SIZE(devices));

	if(is_uart_gpio()) {
		pxa_set_ffuart_info(&benzina2_ffuart_info);
		pxa_set_stuart_info(&benzina2_stuart_info);
	} else {
		pxa_set_ffuart_info(NULL);
		pxa_set_stuart_info(NULL);
	}
	pxa_set_btuart_info(NULL);

	/* onenand */
	benzina2_init_onenand();

#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)
	/* u2d */
	if (device_is_enabled(device_u2d)) {
		pxa_set_u2d_info(&benzina2_u2d_info);
	}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_USB_PXA27X_UDC_MODULE)
	if (device_is_enabled(device_udc)) {
		pxa_set_udc_info(&benzina2_udc_info);
	}
#endif
	/* usb otg */
	benzina2_init_otg();

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &benzina2_u2o_info);
	}
#endif
#if defined(CONFIG_PXA950_CAMERA)
	/* initialize camera */
	benzina2_init_cam();
#endif


	/* i2c devices */


  if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_END_ID) ){

		i2c_register_board_info(0, ARRAY_AND_SIZE(benzina2_i2c_board_info_levante1));
		i2c_register_board_info(0, ARRAY_AND_SIZE(benzina2_i2c_board_info_levante2));
		printk(KERN_ALERT "i2c_register_board_info_levante register !!!\n");

	}else
		printk(KERN_ALERT "i2c_register_board_info PMIC *not* register !!!\n");

	i2c_register_board_info(2, ARRAY_AND_SIZE(benzina2_i2c_hs_info));

	/* GPS */
#if defined(CONFIG_SIRF_GSD3TW)
	benzina2_gps_sirf_init();
#endif

#if defined(CONFIG_PXA_IRDA) || defined(CONFIG_PXA_IRDA_MODULE)
	pxa_set_ficp_info(&benzina2_ficp_platform_data);
#endif
	benzina2_init_mmc();

	benzina2_init_ehci();

#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif
	benzina2_set_lowpower_mfp();

#ifdef	CONFIG_PROC_FS
	create_benzina2_proc_file();
#endif

#ifdef CONFIG_BATTERY_PDA
	benzina2_init_pda_battery();
#endif

#ifdef CONFIG_PDA_POWER
	benzina2_init_pda_power();
#endif
	android_set_pmem_info(&android_pmem_pdata);

}

MACHINE_START(BENZINA2, "BENZINA2")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = pxa_map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = benzina2_init,
MACHINE_END
