/*
 *  linux/arch/arm/mach-pxa/ignite.c
 *
 *  Support for the FBW IGNITE.
 *
 *  2009-12-01   Song Zhiqiang <zhiqiang.song@borqs.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/onenand.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/i2c.h>
#include <linux/pda_power.h>
#include <linux/pda_battery.h>
#include <linux/gpio_keys.h>
#include <linux/switch_usbuart.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/pxa-regs.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/mfp-pxa3xx.h>
#include <mach/pxa3xx_nand.h>
#include <mach/pxa27x_keypad.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/ssp.h>
#include <mach/gpio.h>
#include <mach/ignite.h>
#include <mach/micco.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/irda.h>
#include <mach/pxa3xx_u2d.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/ehci.h>
#include <mach/udc.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxa950fb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/i2c.h>
#include <mach/lis331.h>
#include <linux/proc_fs.h>
#include <mach/levante_micco_pmic_id.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <mach/tavor_common.h>
#include <mach/ssp.h>
#include <mach/gps.h>
#include <mach/pmem.h>
#include "devices.h"
#include "generic.h"
#include <mach/t1021.h>

#define MAX_SLOTS      3

struct platform_mmc_slot ignite_mmc_slot[MAX_SLOTS];

extern int is_android(void);
extern int is_comm_v75(void);

static struct gpio_keys_platform_data ignite_gpio_keys_info;
#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

 static u8 g_pmic_id;

static mfp_cfg_t audio_mfp_cfg[] __initdata = {

	GPIO63_SSP3_CLK,
	GPIO64_SSP3_FRM,
	GPIO65_SSP3_TXD,
	GPIO66_SSP3_RXD,
	GPIO67_SSP3_SYSCLK,

	GPIO68_SSP2_RXD,
	GPIO69_SSP2_TXD,
	GPIO70_SSP2_SYSCLK,
	GPIO71_SSP2_FRM,
	GPIO72_SSP2_CLK,
};

static mfp_cfg_t wlan_on_mfp_cfg[]={
	WLAN_PD_GH381_LPM_ON,
	TAVOR_WAKEUP_BT_LPM_ON,
	BT_UART_RXD_LPM_ON,
	BT_UART_TXD_LPM_ON,
	BT_UART_CTS_LPM_ON,
	BT_UART_RTS_LPM_ON,
	WLAN_BT_RESET_LPM_ON,
	WLAN_DAT3_LPM_ON,
	WLAN_DAT2_LPM_ON,
	WLAN_DAT1_LPM_ON,
	WLAN_DAT0_LPM_ON,
	WLAN_CMD_LPM_ON,
	WLAN_CLK_LPM_ON,
	WLAN_PD_PW886_LPM_ON,
	WLAN_WAKEUP_TAVOR_LPM_ON,
	BT_WAKEUP_TAVOR_LPM_ON,
	TAVOR_WAKEUP_WLAN_LPM_ON,
};

static mfp_cfg_t wlan_off_mfp_cfg[]={
	WLAN_PD_GH381,
	TAVOR_WAKEUP_BT,
	BT_UART_RXD,
	BT_UART_TXD,
	BT_UART_CTS,
	BT_UART_RTS,
	WLAN_BT_RESET,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	WLAN_PD_PW886,
	WLAN_WAKEUP_TAVOR,
	BT_WAKEUP_TAVOR,
	TAVOR_WAKEUP_WLAN,
};

static mfp_cfg_t sd_in_mfp_cfg[]={
	MMC1_CLK_ON,
	MMC1_CMD_ON,
	MMC1_DAT0_ON,
	MMC1_DAT1_ON,
	MMC1_DAT2_ON,
	MMC1_DAT3_ON,
	SD_LDO_CE_ON,
};

static mfp_cfg_t sd_out_mfp_cfg[]={
	MMC1_CLK,
	MMC1_CMD,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	SD_LDO_CE,
};

static mfp_cfg_t ignite_mfp_cfg[] __initdata = {
	GPIO8_FREE,
	GPIO12_FREE,
	GPIO13_FREE,
	GPIO15_FREE,
	GPIO17_FREE,
	GPIO18_FREE,
	GPIO19_FREE,
	GPIO21_FREE,
	GPIO23_FREE,
	GPIO24_FREE,
	GPIO25_FREE,
	GPIO26_FREE,
	GPIO27_FREE,
	GPIO28_FREE,
	GPIO29_FREE,
	GPIO30_FREE,
	GPIO31_FREE,
	GPIO32_FREE,
	GPIO33_FREE,
	GPIO34_FREE,
	GPIO35_FREE,
	GPIO36_FREE,
	GPIO37_FREE,
	GPIO38_FREE,
	GPIO42_FREE,
	GPIO47_FREE,
	GPIO49_FREE,
	GPIO50_FREE,
	GPIO62_FREE,
	GPIO67_FREE,
	GPIO79_FREE,
	GPIO80_FREE,
	GPIO83_FREE,
	GPIO92_FREE,
	GPIO93_FREE,
	GPIO95_FREE,
	GPIO100_FREE,
	GPIO101_FREE,
	GPIO102_FREE,
	GPIO103_FREE,
	GPIO104_FREE,
	GPIO105_FREE,

	VOLUME_UP_KEY,			
	VOLUME_DOWN_KEY,
	OPTION_KEY,
	HOME_KEY,
	VR_KEY,
	CAM_ON_KEY,
	MOT_INT,
	CAM_SHOT_KEY,

	TP_LDO_EN,
	TP_INT,
	PRO_INT,
	USB2UART_SEL,
	USB2UART_EN,
	CPS_RSTN,
	CPS_INT,
	TAVOR_ONE_WIRE,

	LCD_RESET_N,
	LCD_TE,

	SD_LDO_CE,
	MMC1_CMD,
	MMC1_CLK,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	MMC0_CD,

	FFRXD,
	FFTXD,
	GPS_RX,
	GPS_TX,
	GPS_RESET,
	GPS_ON,
	GPS_LNA_EN,

	HP_AMP_SD,
	AUDIO_AMP_PM_EN,
	PCM_SYNC_OFF,
	PCM_CLK_OFF,
	PCM_TXD_OFF,
	PCM_RXD_OFF,
	/*
	PCM_CLK,
	PCM_SYNC,
	PCM_TXD	,
	PCM_RXD,
	*/
	I2S_RXD,
	I2S_TXD,
	I2S_SYNC,
	I2S_BITCLK,

	CI2C_SCL,
	CI2C_SDA,
	I2C_SCL,
	I2C_SDA,


	TAVOR_WAKEUP_BT,
	WLAN_PD_GH381,
	WLAN_BT_RESET,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	WLAN_PD_PW886,
	WLAN_WAKEUP_TAVOR,
	BT_WAKEUP_TAVOR,
	TAVOR_WAKEUP_WLAN,
	BT_UART_RXD,
	BT_UART_TXD,
	BT_UART_CTS,
	BT_UART_RTS,

	VCM_PD,
	CAM_MCLK,
	CAMERA_ON,
	CAM_PD_MAIN,
	CAM_RESET_MAIN,
};

#ifdef	CONFIG_PROC_FS
#define	IGNITE_PROC_FILE	"driver/ignite"
static struct proc_dir_entry *ignite_proc_file;
static int pmic_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
   //	struct pmic_ops *ops = file->private_data;
	void __user *uarg = (void __user *)arg;
	int ret = -EINVAL;
    u8 icotl_pmic_id;

	switch (cmd) {
	case IOCTL_PMIC_ID_GET:
	   	icotl_pmic_id = g_pmic_id;
		ret = put_user(icotl_pmic_id, (unsigned long __user *)uarg);
		printk(KERN_ALERT "-------> icotl_pmic_id %d \n",icotl_pmic_id);
		if (ret)
			ret = -EFAULT;
		break;


	default:
	  	//if (ops->ioctl)
	  	//	ret = ops->ioctl(cmd, arg);
		panic("IGNITE-pmic_ioctl");
		break;
	}
	return ret;
}


static ssize_t ignite_proc_read(struct file *filp,
        char __user *buffer, size_t length, loff_t *offset)
{
	printk(KERN_ALERT "ignite_proc_read - no action done\n");

	return 0;
}

static ssize_t ignite_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	printk(KERN_ALERT "ignite_proc_write - no action done\n");

	return len;
}

static struct file_operations ignite_proc_ops = {
	.read 	= ignite_proc_read,
	.write 	= ignite_proc_write,
	.ioctl	= pmic_ioctl,

};

static void create_ignite_proc_file(void)
{
	ignite_proc_file = create_proc_entry(IGNITE_PROC_FILE, 0644, NULL);
	if (ignite_proc_file) {
		ignite_proc_file->owner = THIS_MODULE;
		ignite_proc_file->proc_fops = &ignite_proc_ops;
		printk(KERN_ALERT "IGNITE proc file created!\n");
	} else
		printk(KERN_ALERT "IGNITE proc file create failed!\n");
}

static void remove_ignite_proc_file(void)
{
	//remove_proc_entry(IGNITE_PROC_FILE, &proc_root);
}
#endif  //CONFIG_PROC_FS

static struct platform_device levante_ts_device = {
	.name 		= "levante-ts",
	.id 		= -1,
};

static struct platform_device levante_charger_device = {
	.name		= "levante-charger",
	.id		= -1,
};


static struct platform_device levante_kp_bl_device = {
	.name 		= "levante-kp-bl",
	.id 		= -1,
};

static struct platform_device levante_bl_device = {
       .name           = "levante-bl",
       .id             = -1,
};

static struct resource pxa3xx_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code +
						  * The first 4KB are locked for pxa930 trusted */
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1, /* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa3xx_resource_imm),
	.resource	= pxa3xx_resource_imm,
};

static struct platform_device gpio_keys_device = {
	.name		= "gpio-keys",
	.id		= -1,
	.dev		= {
		.platform_data = &ignite_gpio_keys_info,
	},
};

static struct platform_device levante_hsdetect_device = {
	.name		= "levante-hsdetect",
	.id		= -1,
};

static struct platform_device ispt_device = {
	.name		= "pxa-ispt",
	.id		= -1,
};

static struct platform_device *devices[] __initdata = {
   	&pxa3xx_device_imm,
   	&ispt_device,
	&gpio_keys_device,
};

static struct platform_device *devices_levante[] __initdata = {
#if defined(CONFIG_PXA9xx_LEVANTE)
 	&levante_bl_device,
    &levante_kp_bl_device,
 	&levante_ts_device,
 	&levante_charger_device,
	&levante_hsdetect_device
#endif
};

static struct android_pmem_platform_data android_pmem_pdata = {
	.name = "pmem",
	/* 
	 * Linux kernel uses 96M(total 128M) , 0xb8000000-0xbe000000
	 *                or 224M(total 256M) , 0xb0000000-0xbe000000 
	 * CP uses 12M, PMEM uses 16M
	 */
	.start = 0xbe000000,
	.size = 0x1000000,      /* 16M */
	.no_allocator = 1,
	.cached = 1,
};

#ifdef CONFIG_KEYBOARD_GPIO
static struct gpio_keys_button ignite_gpio_keys_button[] = {
	[0] = {
		.desc	= "back_key",
		.code	= KEY_BACK,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VR_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[1] = {
		.desc	= "cam_shot_key_n",
		.code	= KEY_CAMERA,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_SHOT_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[2] = {
		.desc	= "cam_on_key",
		.code	= KEY_FOCUS,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(CAM_ON_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[3] = {
		.desc	= "volume_up_key",
		.code	= KEY_VOLUMEUP,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_UP_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[4] = {
		.desc	= "volume_down_key",
		.code	= KEY_VOLUMEDOWN,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(VOLUME_DOWN_KEY)),
		.wakeup	= 1,
		.active_low = 1,
	},
	[5] = {
		.desc	= "menu_key",
		.code	= KEY_MENU,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(OPTION_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},
	[6] = {
		.desc	= "home_key",
		.code	= KEY_HOME,
		.type	= EV_KEY,
		.gpio	= mfp_to_gpio(MFP_CFG_PIN(HOME_KEY)),
		.wakeup	= 0,
		.active_low = 1,
	},

};

static struct gpio_keys_platform_data ignite_gpio_keys_info = {
	.buttons = ignite_gpio_keys_button,
	.nbuttons = (sizeof(ignite_gpio_keys_button)
			/ sizeof(struct gpio_keys_button)),
	//.trigger_key0 = KEY_BACK,
	//.trigger_key1 = KEY_CAMERA,
};
#endif

#if defined(CONFIG_SIRF_GSD3TW)
static int ignite_gps_set_power(int on_off)
{
	if( on_off )
	{
	}else
	{
	}
	return 0;
}
static struct sirf_platform_data ignite_sirf_platform_data = {
	.reset_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)),
	.on_off_pin = mfp_to_gpio(MFP_CFG_PIN(GPS_ON)),
	.set_power = ignite_gps_set_power,
};

static struct platform_device ignite_sirf_device = {
	.name = "sirf",
	.id = -1,
	.dev = {
		.platform_data = &ignite_sirf_platform_data,
	},
};

static void __init ignite_gps_sirf_init(void)
{
	printk(KERN_ERR"[%s:%d] gps pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), "gps reset")) {
		printk(KERN_ERR "<SiRF> GPIO RST request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), "gps on_off")) {
		printk(KERN_ERR "<SiRF> GPIO ON_OFF request error!!\n");
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_ON)), 0); /* GPS_ON_OFF */
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_RESET)), 1); /* GPS_RST */

	printk(KERN_ERR"GPS set up finished\n");
	platform_device_register(&ignite_sirf_device);
}
#endif



#if defined(CONFIG_MMC) || defined(CONFIG_MMC_MODULE)
static int ignite_mci_init(struct device *dev,
		irq_handler_t ignite_detect_int,
			     void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(ignite_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = ignite_mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for ignite MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, ignite_detect_int,
			  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			  "MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void ignite_mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(ignite_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = ignite_mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static struct pxamci_platform_data ignite_mci_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
	.init 		= ignite_mci_init,
	.exit		= ignite_mci_exit,
};

static struct pxamci_platform_data ignite_mci2_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
};

static mfp_cfg_t pxa9xx_mmc_mfp_cfg[] __initdata = {
	/* MMC4 of Tavor PV */
	GPIO47_GPIO,
	GPIO55_MMC4_CMD,
	GPIO56_MMC4_CLK,
	GPIO57_MMC4_DAT0,
	GPIO58_MMC4_DAT1,
	GPIO59_MMC4_DAT2,
	GPIO60_MMC4_DAT3,
};

#if 0
static int stamp_8688_wlan_poweron(void)
{
	int gpio_power = 0;
	int gpio_reset = 0;

	gpio_power = mfp_to_gpio(MFP_PIN_GPIO77);
	gpio_reset = mfp_to_gpio(MFP_PIN_GPIO67);

	if (gpio_request(gpio_power, "8688 wlan power down")) {
		printk(KERN_INFO "gpio %d request failed\n", gpio_power);
		return -1;
	}

	if(gpio_request(gpio_reset, "8688 wlan reset")) {
		printk(KERN_INFO "gpio %d request failed\n", gpio_reset);
		gpio_free(gpio_power);
		return -1;
	}

	gpio_direction_output(gpio_power, 1);
	gpio_direction_output(gpio_reset, 0);
	msleep(10);
	gpio_direction_output(gpio_reset, 1);

	gpio_free(gpio_power);
	gpio_free(gpio_reset);
	return 0;
}
#endif
static void __init ignite_init_mmc(void)
{
	//pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa9xx_mmc_mfp_cfg)); //no need config it
	pxa9xx_set_mci4_info(&ignite_mci_platform_data);
	pxa9xx_set_mci5_info(&ignite_mci2_platform_data);

	/* MMC card detect for controller 1 of pxa9xx_mci */
	ignite_mmc_slot[1].gpio_cd  = mfp_to_gpio(MMC0_CD);
}
#else
static inline void ignite_init_mmc(void) {}
#endif

#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
static struct flash_platform_data ignite_onenand_info;
static void __init ignite_init_onenand(void)
{
	/* Now, kernel get mtd partitions from commandline.*/
	if (is_android()) {
		ignite_onenand_info.parts = android_256m_v75_partitions;
		ignite_onenand_info.nr_parts = 
			ARRAY_SIZE(android_256m_v75_partitions);
	}
	else {
		ignite_onenand_info.parts = pxa930_256m_v75_partitions;
		ignite_onenand_info.nr_parts = 
			ARRAY_SIZE(pxa930_256m_v75_partitions);
	}

	pxa3xx_device_onenand.dev.platform_data = &ignite_onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_PXA27X_UDC_MODULE)
static mfp_cfg_t pxa930_otg_init_pins[] = {
	GPIO64_GPIO,
};

static mfp_cfg_t pxa930_otg_pins[] = {
	GPIO64_USB_P2_7,
};

int ignite_udc_is_miniA(void)
{
	int otg_id = mfp_to_gpio(MFP_PIN_GPIO64);
	int id_value;
	int err;

	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_init_pins));
	err = gpio_request(otg_id, "OTG ID");
	if (err) {
		gpio_free(otg_id);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", otg_id, err);
		return 0;
	}
	gpio_direction_input(otg_id);
	id_value = gpio_get_value(otg_id);
	gpio_free(otg_id);
	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_pins));

	return (id_value == 0);
}

static struct pxa2xx_udc_mach_info ignite_udc_info = {
	.udc_is_miniA = ignite_udc_is_miniA,
};
#endif

#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)

void ignite_reset_xcvr(void)
{
	int reset_pin;
	int err;

	reset_pin = mfp_to_gpio(MFP_PIN_GPIO43);
	err = gpio_request(reset_pin, "U2D Reset");
	if (err) {
		gpio_free(reset_pin);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", reset_pin, err);
		return;
	}
	gpio_direction_output(reset_pin, 0);
	mdelay(100);
	gpio_set_value(reset_pin, 1);

	gpio_free(reset_pin);
}

static struct pxa3xx_u2d_mach_info ignite_u2d_info = {
	.reset_xcvr = ignite_reset_xcvr,
};
#endif /* CONFIG_USB_PXA3XX_U2D || CONFIG_USB_PXA3XX_U2D_MODULE */

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info ignite_u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#ifdef CONFIG_USB_EHCI_PXA9XX

static void __init ignite_init_ehci(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_ehci_info(&ignite_u2o_info);
}
#else
static inline void ignite_init_ehci(void) {}
#endif

#if defined (CONFIG_USB_OTG) && defined(CONFIG_USB_PXA9XX_U2O)
static void __init ignite_init_otg(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_u2otg_info(&ignite_u2o_info);
}
#else
static inline void ignite_init_otg(void) {}
#endif

static int lis331_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(MOT_INT)));

	return 0;
}

static int lis331_ack_irq(void)
{
	return 0;
}

static int lis331_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))) ;
}

static struct lis331_platform_data lis331_data = {
	.init_irq = lis331_init_irq,
	.ack_irq = lis331_ack_irq,
	.get_gpio_value = lis331_get_gpio_value,
	.platform_init = NULL,
	.power_chips = NULL,
	.xyz_dir = Y_POS,
};

#if defined(CONFIG_PXA_IRDA) || defined(CONFIG_PXA_IRDA_MODULE)
static void ignite_irda_transceiver_mode(struct device *dev, int mode)
{
	unsigned long flags;
	int err;
	static int irda_mfp_init;
	int gpio_ir_shdn = ((struct pxaficp_platform_data *)
			dev->platform_data)->gpio_ir_shdn;

	if (!irda_mfp_init) {
		err = gpio_request(gpio_ir_shdn, "IRDA SHDN");
		if (err) {
			gpio_free(gpio_ir_shdn);
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d return :%d\n", gpio_ir_shdn, err);
			return;
		}
		gpio_direction_output(gpio_ir_shdn, 1);

		irda_mfp_init = 1;
	}

	local_irq_save(flags);
	if (mode & IR_SIRMODE) {
		gpio_set_value(gpio_ir_shdn, 0);
	} else if (mode & IR_FIRMODE) {
		/* do not support FIR */
	}
	if (mode & IR_OFF) {
		gpio_set_value(gpio_ir_shdn, 1);
	}
	local_irq_restore(flags);
}

static struct pxaficp_platform_data ignite_ficp_platform_data = {
	.transceiver_cap  = IR_SIRMODE | IR_OFF,
	.transceiver_mode = ignite_irda_transceiver_mode,
};
#endif /* (CONFIG_PXA_IRDA) || (CONFIG_PXA_IRDA_MODULE) */

#if defined(CONFIG_PXA9xx_LEVANTE)
static int levante_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_PIN_GPIO83));

	return 0;
}

static int levante_ack_irq(void)
{
	return 0;
}

static void ignite_levante_init(void)
{
	u8 data;
	printk(KERN_ALERT "----->ignite_levante_init\n");

	/* Set V_LDO12 to 2.8V */
	sanremo_write(SANREMO_LDO12, 0x33);
	sanremo_read(SANREMO_SUPPLIES_EN12, &data);
	data|= 0x20;
	sanremo_write(SANREMO_SUPPLIES_EN12,data);
}

/* power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module levanteA1_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		BUCK2},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,	LDO3},
	{VCC_SDIO,		LDO14},
	{VCC_CAMERA_ANA, LDO3},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO3},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,	BUCK2}, /* ? */
	{VCC_1P8V,		0}, 	/* ? */
	{VCC_MEM,		BUCK2}, /* ? */
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

static struct power_chip levante_chips[] = {
	{LEVATNE_A0_ID,	"levanteA0",	levanteA1_power_modules},
	{LEVATNE_A1_ID,	"levanteA1",	levanteA1_power_modules},
	{0,	NULL,		NULL},
};
static struct levante_platform_data  levante_data = {
    .platform_init = ignite_levante_init
};

static struct sanremo_platform_data  sanremo_data = {
    .init_irq = levante_init_irq,
    .ack_irq = levante_ack_irq,
    .power_chips = levante_chips,
};

static struct portofino_platform_data  portofino_data = {
};

#endif /* CONFIG_PXA9xx_LEVANTE */

#if defined(CONFIG_PXA950_CAMERA)
static int cam_power_set(int flag)
{
	struct clk *axi_clk = NULL;
	struct clk *sci1_clk = NULL;
	struct clk *sci2_clk = NULL;
	struct clk *cam_clk = NULL;

	sci1_clk = clk_get(NULL, "SCI1CLK");
	if (IS_ERR(sci1_clk)) {
		printk(KERN_ERR "unable to get SCI1CLK\n");
		return PTR_ERR(sci1_clk);
	}

	sci2_clk = clk_get(NULL, "SCI2CLK");
	if (IS_ERR(sci2_clk)) {
		printk(KERN_ERR "unable to get SCI2CLK\n");
		return PTR_ERR(sci2_clk);
	}

	if(flag){
		axi_clk = clk_get(NULL, "AXICLK");
		if (IS_ERR(axi_clk)) {
			printk(KERN_ERR "unable to get AXICLK\n");
			return PTR_ERR(axi_clk);
		}

		clk_enable(axi_clk);
		clk_enable(sci1_clk);
		clk_enable(sci2_clk);

		clk_put(axi_clk);
	}else{
		clk_disable(sci1_clk);
		clk_disable(sci2_clk);
	}

	clk_put(sci1_clk);
	clk_put(sci2_clk);

	return 0;
}
/* sensor init */
static int sensor_power_set(int res, int flag)
{
	/*
	 * RES, 0, LOW RESOLUTION, 1 HIGH RESOLUTION
	 * FLAG, 1: ON, 0: OFF
	 * GPIO93 LOW FOR HIGH
	 * GPIO14 LOW FOR LOW
	 */
	int high_pin = mfp_to_gpio(MFP_PIN_GPIO93);
	int low_pin = mfp_to_gpio(MFP_PIN_GPIO14);

	cam_power_set(flag);

	if (gpio_request(high_pin, "CAM_EANBLE_HI_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", high_pin);
		return -EIO;
	}

	if (res) {
		if (flag)
			gpio_direction_output(high_pin, 0);	/* enable */

		else
			gpio_direction_output(high_pin, 1);	/* disable */

	} else{
		if (gpio_request(low_pin, "CAM_EANBLE_LOW_SENSOR")) {
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d \n", low_pin);
			return -EIO;
		}
		if (flag) {
			/* high resolution is bridge */
			gpio_direction_output(high_pin, 0);	/* enable */
			gpio_direction_output(low_pin, 0);	/* enable */

		}
#if 0
		/* FIXME */
		else
			gpio_direction_output(low_pin, 1);	/* disable */
#endif
		gpio_free(low_pin);
	}

	gpio_free(high_pin);
	return 0;
}

static struct sensor_platform_data ov3640_sensor_data = {
	.id = SENSOR_HIGH,
	.power_set = sensor_power_set,
};

static struct sensor_platform_data ov7690_sensor_data = {
	.id = SENSOR_LOW,
	.power_set = sensor_power_set,
};

/* sensor init over */

static struct cam_platform_data cam_ops = {
	.power_set	= cam_power_set,
};

static mfp_cfg_t cam_mfp_cfg[] __initdata = {
	/* CAM */
};

static void __init ignite_init_cam(void)
{
	pxa950_device_cam.dev.platform_data = &cam_ops;
	pxa3xx_mfp_config(ARRAY_AND_SIZE(cam_mfp_cfg));
	platform_device_register(&pxa950_device_cam);
	platform_device_register(&pxa950_device_csi);
}

#endif

#define TP_GPIO_INT MFP_PIN_GPIO52
#define I2C_TRST MFP_PIN_GPIO7

static void i2c_hs_init(void)
{
	unsigned int i2c_trst;

	i2c_trst = mfp_to_gpio(I2C_TRST);
	if (gpio_request(i2c_trst, "I2C_TRST")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", i2c_trst);
	}
	gpio_direction_output(i2c_trst, 1);
	msleep(100);
	gpio_free(i2c_trst);
}
#ifdef CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI
static int benzina_ts_power(int on)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(TP_INT)));
        return 0;
}

static struct synaptics_i2c_rmi_platform_data benzina_ts_data[] = {
{
                .version = 0x0101,
                .power = benzina_ts_power,
                .sensitivity_adjust = 7,
                .flags = SYNAPTICS_FLIP_X | SYNAPTICS_FLIP_Y | SYNAPTICS_SNAP_TO_INACTIVE_EDGE,
                .irqflags = IRQF_TRIGGER_FALLING,
                .inactive_left = 0,
                .inactive_right = 0,
                .inactive_top = 0,
                .inactive_bottom = 0,
                .snap_left_on = 0,
                .snap_left_off = 0,
                .snap_right_on = 0,
                .snap_right_off = 0,
                .snap_top_on = 0,
                .snap_top_off = 0,
                .snap_bottom_on = 0,
                .snap_bottom_off = 0,
        },
        {
                .flags = SYNAPTICS_FLIP_X | SYNAPTICS_FLIP_Y | SYNAPTICS_SNAP_TO_INACTIVE_EDGE,
                .inactive_left = 0,
                .inactive_right = 0,
                .inactive_top = 0,
                .inactive_bottom = 0,
        }
};
#endif
#if defined(CONFIG_TOUCHSCREEN_T1021)
static int t1021_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(TP_INT)));
	return 0;
}

static int t1021_ack_irq(void)
{
	return 0;
}

static struct t1021_platform_data t1021_data= {
.init_irq = t1021_init_irq,
.ack_irq = t1021_ack_irq,
};

#endif

static struct i2c_board_info ignite_i2c_hs_info[] = {
	/* hs i2c device */
#if defined(CONFIG_SENSORS_LIS331)
	{
		.type		= "lis331",
		.addr		= 0x1C,
		.platform_data	= &lis331_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(MOT_INT))),
	},
#endif
#if CONFIG_SENSORS_CYWEE 
 	{
		.type		= "cywee",
		.addr		= 0x78,
        },
#endif
#if CONFIG_SENSORS_YAS529 
{
		.type		= "fx_compass",
		.addr		= 0x2e,
        },
#endif    
#if defined(CONFIG_PXA950_CAMERA)
	{
		.type		= "ov3640",
		.addr		= 0x3c,
		.platform_data	= &ov3640_sensor_data,
	},
	{
		.type		= "ov7690",
		.addr		= 0x21,
		.platform_data	= &ov7690_sensor_data,
	},
#endif
#if defined(CONFIG_TOUCHSCREEN_T1021)
	{
		.type 		= "t1021_touch",
		.addr		= 0x20,
		.platform_data	= &t1021_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
	},
#endif
#ifdef CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI
	{
		.type 		= SYNAPTICS_I2C_RMI_NAME,
		.addr		= 0x20,
		.platform_data	= &benzina_ts_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
	},
#endif
};

static struct i2c_board_info ignite_i2c_board_info_levante1[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "sanremo",
		.addr		= 0x34,
		.platform_data	= &sanremo_data,
	   	.irq		= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO83)),
	},
	{
		.type	= "portofino",
		.addr		= 0x10,
		.platform_data	= &portofino_data,
	},
#endif

};

static struct i2c_board_info ignite_i2c_board_info_levante2[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "levante",
           .platform_data  = &levante_data,
	},
#endif
};

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_FAST_MODE, /* PXA_I2C_STANDARD_MODE - SLOW */
};

static struct i2c_pxa_platform_data pxa3xx_i2c_hs_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_STANDARD_MODE,
};

struct pxa3xx_freq_mach_info ignite_freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};

#if defined(CONFIG_FB_PXA950)

/* Sumsung LMS350df03 HVGA LCD power on cmd */
static u32 lms_spi_cmdon1[] = {
	0x740007,
	0x760000,
};

/* wait 10 ms */
static u32 lms_spi_cmdon2[] = {
	0x740011,
	0x76222F,
	0x740012,
	0x760F00,
	0x740013,
	0x767FD9,
	0x740076,
	0x762213,
	0x740074,
	0x760001,
	0x740076,
	0x760000,
	0x740010,
	0x765604,
};

/* wait 6 frame, about 120 ms */
static u32 lms_spi_cmdon3[] = {
	0x740012,
	0x760c63,
};

/* wait 5 frame, about 100 ms */
static u32 lms_spi_cmdon4[] = {
	0x740001,
	0x76083B,
	0x740002,
	0x760300,
	0x740003,
	0x76F040,	/* sync without OE */
	0x740008,
	0x760002,
	0x740009,
	0x76000B,
	0x740076,
	0x762213,
	0x74000B,
	0x763340,
	0x74000C,
	0x760020,
	0x74001C,
	0x767770,
	0x740076,
	0x760000,
	0x74000D,
	0x760000,
	0x74000E,
	0x760500,
	0x740014,
	0x760000,
	0x740015,
	0x760803,
	0x740016,
	0x760000,
	0x740030,
	0x760003,
	0x740031,
	0x76070F,
	0x740032,
	0x760D05,
	0x740033,
	0x760405,
	0x740034,
	0x76090D,
	0x740035,
	0x760501,
	0x740036,
	0x760400,
	0x740037,
	0x760504,
	0x740038,
	0x760C09,
	0x740039,
	0x76010C,
};

/* Wait about 10ms */
static u32 lms_spi_cmdon5[] = {
	0x740007,
	0x760001,
};

/* Wait more than 1frame, about 40ms */
static u32 lms_spi_cmdon6[] = {
	0x740007,
	0x760101,
};

/* Wait more than 2frame, about 40ms */
static u32 lms_spi_cmdon7[] = {
	0x740007,
	0x760103,
};

static u32 lms_spi_cmdoff1[] = {
	0x740010,
	0x760001,
	0x74000B,
	0x7630E1,
	0x740007,
	0x760102,
};

/* Wait more than 2frame, about 30ms */
static u32 lms_spi_cmdoff2[] = {
	0x740007,
	0x760000,
	0x740012,
	0x760000,
	0x740010,
	0x760100,
};

static int lms_on_interval [] = {1, 10, 120, 100, 10, 40, 40};
static int lms_off_interval [] = {30};

static struct fb_videomode video_modes[] = {
	/* lms HVGA mode info */
	[0] = {
		.pixclock       = 100000,
		.refresh        = 60,
		.xres           = 320,
		.yres           = 480,
		.hsync_len      = 4,
		.left_margin    = 7,
		.right_margin   = 4,
		.vsync_len      = 2,
		.upper_margin   = 3,
		.lower_margin   = 4,
		.sync           = 0,
	},
};

void ignite_espon_lcd_panel_reset(void)
{
	if (gpio_request(MFP_PIN_GPIO45, "lcd reset"))
		printk(KERN_ERR "LTM 022A97A gpio_request: failed!\n");
	gpio_direction_output(MFP_PIN_GPIO45,1);
	msleep(10);
	gpio_direction_output(MFP_PIN_GPIO45,0);
	mdelay(2);
	gpio_direction_output(MFP_PIN_GPIO45,1);
	mdelay(2);
	gpio_free(MFP_PIN_GPIO45);
}

static void espon_ssp_config(struct ssp_dev *dev)
{
	ssp_disable(dev);
	ssp_config(dev, 0x001fff87, 0x18, 0, 0);
}
static void ssp_send_cmd_para(u32 *cmd,int num)
{
	static struct ssp_dev ssp;
	int i;
	static int ssp_initialized = 0;

	if (!ssp_initialized) {
		ssp_init(&ssp, 1, SSP_NO_IRQ);
		espon_ssp_config(&ssp);
		ssp_initialized = 1;
	}

	clk_enable(ssp.ssp->clk);
	for (i = 0; i < num; i++, cmd++) {
		ssp_write_word(&ssp, *cmd & 0xffffff);
		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp);
	}
	clk_disable(ssp.ssp->clk);
}
static void espon_lcd_power(int on)
{
	int err;
	/* turn on backlight */
	portofino_write(0x10,0x0e);
	portofino_write(0x15,0x02);
	portofino_write(0x16,0x10);
	portofino_write(0x00,0x0f);
	portofino_write(0x01,0x00);
	portofino_write(0x02,0x0f);
	portofino_write(0x03,0x19);

	if (on) {
		err = gpio_request(MFP_PIN_GPIO16, "LMS_LCD_LDO");
		if (err) {
			printk("%s: failed to request LCD LDO!\n", __func__);
			return;
		}
		gpio_direction_output(MFP_PIN_GPIO16, 1);
		gpio_free(MFP_PIN_GPIO16);

		ignite_espon_lcd_panel_reset();

		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon1));
		msleep(lms_on_interval[1]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon2));
		msleep(lms_on_interval[2]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon3));
		msleep(lms_on_interval[3]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon4));
		msleep(lms_on_interval[4]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon5));
		msleep(lms_on_interval[5]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon6));
		msleep(lms_on_interval[6]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdon7));

	} else {
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdoff1));
		msleep(lms_off_interval[0]);
		ssp_send_cmd_para(ARRAY_AND_SIZE(lms_spi_cmdoff2));	
	}
}

static struct pxa950fb_mach_info ignite_lcd_info __initdata = {
	.id                     = "Base",
	.modes                  = video_modes,
	.num_modes              = ARRAY_SIZE(video_modes),
	.pix_fmt_in             = PIX_FMTIN_RGB_32,
	.pix_fmt_out            = PIX_FMTOUT_18_RGB666,
	.panel_type             = LCD_Controller_Active,
	.mixer0                 = LCD_M2PARALELL_CONVERTER,
	.mixer1                 = LCD_MIXER_DISABLE,
	.active                 = 1,
	.pxa950fb_lcd_power     = espon_lcd_power,
	.invert_pixclock        = 1,
};

static struct pxa950fb_mach_info ignite_lcd_ovly_info __initdata = {
	.id                     = "Ovly",
	.modes                  = video_modes,
	.num_modes              = ARRAY_SIZE(video_modes),
	.pix_fmt_in             = PIX_FMTIN_YUV420,
	.pix_fmt_out            = PIX_FMTOUT_18_RGB666,
	.panel_type             = LCD_Controller_Active,
	.mixer0                 = LCD_M2DSI0,
	.mixer1                 = LCD_MIXER_DISABLE,
	.active                 = 1,
	.pxa950fb_lcd_power     = espon_lcd_power,
	.invert_pixclock        = 1,
};

static void __init ignite_init_lcd(void)
{
	set_pxa950_fb_info(&ignite_lcd_info);
	set_pxa950_fb_ovly_info(&ignite_lcd_ovly_info);
}

#endif

#ifdef CONFIG_PM
extern irqreturn_t uartrx_detect_irq(int irq, void *_dev);

static mfp_cfg_t ignite_ffuart_enable_mfp_cfg[] = {
	/* FFUART */
	GPIO53_FFRXD,
	GPIO54_FFTXD,
	DF_ADDR0_GPIO_53,
};

static mfp_cfg_t ignite_ffuart_disable_mfp_cfg[] = {
	/* FFUART */
	GPIO53_GPIO,
	GPIO54_GPIO,
	DF_ADDR0_NONE,
};

extern int is_uart_gpio(void);
static int ignite_ffuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO53);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);
		/* if edge detection is not clear, the altfun switch will
		 * gpio interrupt is generated on this gpio pin when
		 * switching pins from UART to gpio. so clear it.
		 */
		GRER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		GFER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(ignite_ffuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(ignite_ffuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
		GRER(gpio_rx) |= GPIO_bit(gpio_rx);
		GFER(gpio_rx) |= GPIO_bit(gpio_rx);

		ret = gpio_request(gpio_rx, "FFUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info ignite_ffuart_info = {
	.uart_pin_switch = ignite_ffuart_switch,
};

static mfp_cfg_t ignite_stuart_enable_mfp_cfg[] = {
	/* STUART */
	GPIO45_STRXD,
	GPIO46_STTXD,
};

static mfp_cfg_t ignite_stuart_disable_mfp_cfg[] = {
	/* STUART */
	GPIO45_GPIO,
	GPIO46_GPIO,
};

static int ignite_stuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO45);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);

		pxa3xx_mfp_config(ARRAY_AND_SIZE(ignite_stuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(ignite_stuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);

		ret = gpio_request(gpio_rx, "STUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info ignite_stuart_info = {
	.uart_pin_switch = ignite_stuart_switch,
};

static int ignite_init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
	src->bits.uart1 = 1;
	src->bits.mkey = 1;
	src->bits.eth = 1;
	src->bits.tsi = 1;
	src->bits.cmwdt = 1;
	return 0;
}

extern int pxa930_query_gwsr(int);
static int ignite_query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_KP)
		src->bits.mkey = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(3))
		src->bits.tsi = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(9))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart1 = 1;
		}
	}
	if (reg & PXA3xx_PM_WE_GENERIC(2))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart2 = 1;
		}
	}

	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(13)) {
		if (pxa930_query_gwsr(97))
			src->bits.eth = 1;
		if (pxa930_query_gwsr(53))
			src->bits.uart1 = 1;
	}
	return 0;
}

static int ignite_ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int ignite_key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_KP;
		}
	} else {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int ignite_mmc_wakeup(pm_wakeup_src_t src, int enable)
{
#if 0
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mmc1_cd) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_NONE);
#endif
	return 0;
}

static int ignite_uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(9);
			}
		}
		if (src.bits.uart2) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_FALL);
				/* note: on pxa930, uart2 use this bit */
				ret |= PXA3xx_PM_WE_GENERIC(2);
			}
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_FFRXD), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int ignite_eth_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int ignite_tsi_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(3);
		}
	} else {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int ignite_comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= ignite_init_wakeup,
	.query	= ignite_query_wakeup,
	.ext    = ignite_ext_wakeup,
	.key    = ignite_key_wakeup,
	.mmc    = ignite_mmc_wakeup,
	.uart   = ignite_uart_wakeup,
	.eth	= ignite_eth_wakeup,
	.tsi	= ignite_tsi_wakeup,
	.cmwdt	= ignite_comm_wdt_wakeup,
};
#else
static struct pxa_uart_mach_info ignite_ffuart_info = NULL;
static struct pxa_uart_mach_info ignite_stuart_info = NULL;
#endif

void ignite_stereo_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO84, "STEREO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO84);
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d return :%d\n", MFP_PIN_GPIO84, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO84, value);
	gpio_free(MFP_PIN_GPIO84);
}

void ignite_ext_mono_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO63, "MONO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO63);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", MFP_PIN_GPIO63, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO63, value);
	gpio_free(MFP_PIN_GPIO63);
}

static void ignite_set_lowpower_mfp(void)
{
	/* It will change mfp value set by OBM to save more power,
	 * it should be merged into OBM mfp setting eventually. */

	/* these values have been merged int OBM
	 * pxa3xx_mfp_write(48, 0xa440);
	 * pxa3xx_mfp_write(49, 0xa440);
	 * pxa3xx_mfp_write(51, 0xa440);
	 */
}

#ifdef CONFIG_BATTERY_PDA
/*****************************************************************************/
/*                      pda_battery                                          */
/*****************************************************************************/
static struct pda_battery_pdata pmic_pda_battery_pdata = {
	.name             = "battery",
};

static struct platform_device pda_battery_device = {
	.name = "pda_battery",
	.id = -1,
	.dev = {
		.platform_data = &pmic_pda_battery_pdata,
	},
};

static void ignite_init_pda_battery()
{
	platform_device_register(&pda_battery_device);
}
#endif


#ifdef CONFIG_PDA_POWER

static int pda_power_is_ac_online(void)
{
	return 0;
}

static int pda_power_is_usb_online(void)
{
	return 1;
}


static char *pda_power_supplicants[] = {
	"battery",
	"mon_battery",
};

static struct pda_power_pdata pda_power_supply_info = {
	.is_ac_online           = pda_power_is_ac_online,
	.is_usb_online          = pda_power_is_usb_online,
	.supplied_to            = pda_power_supplicants,
	.num_supplicants        = ARRAY_SIZE(pda_power_supplicants),
};

static struct platform_device pda_power_supply = {
	.name = "pda-power",
	.id = -1,
	.dev = {
		.platform_data = &pda_power_supply_info,
	},
};

static void ignite_init_pda_power()
{
	platform_device_register(&pda_power_supply);
}
#endif

#ifdef CONFIG_SWITCH_USBUART
static unsigned int uu_event;
static enum usbuart_state uu_state = USBUART_AP_UART;
static int ignite_usbuart_set_state(enum usbuart_event event, int on)
{
	if (on)
		uu_event |= event;
	else
		uu_event &= ~event; 

	if (uu_event) {
		console_ll_stop();
		gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 0);
		uu_state = USBUART_AP_USB;
	} else {
		gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 1);
		uu_state = USBUART_AP_UART; 
		if (get_boot_mode() != BOOT_MODE_SDL_FLASHMODE)
		      console_ll_start();
	}
	return 0;
}

static enum usbuart_state ignite_usbuart_get_state(void)
{
	return uu_state;
}

static struct usbuart_switch_platform_data ignite_usbuart_switch_pdata = {
	.name		= "switch-usbuart",
	.set_state	= ignite_usbuart_set_state,
	.get_state	= ignite_usbuart_get_state,
};

static struct platform_device usbuart_switch_device = {
	.name		= "switch-usbuart",
	.dev		= {
		.platform_data = &ignite_usbuart_switch_pdata,
	},
};

static void __init ignite_init_usbuart_switch(void)
{
	int err;
	err = gpio_request(MFP_CFG_PIN(USB2UART_EN), "usb2uart_en");
	BUG_ON(err);
	err = gpio_request(MFP_CFG_PIN(USB2UART_SEL),"usb2uart_sel");
	BUG_ON(err);

	gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL),1);
	gpio_direction_output(MFP_CFG_PIN(USB2UART_EN), 1);
	uu_state = USBUART_AP_UART;
	if (get_boot_mode() != BOOT_MODE_SDL_FLASHMODE)
	      console_ll_start();
	platform_device_register(&usbuart_switch_device);
	printk(KERN_CRIT "usbuart is swith to uart.\n");
}
#else
static inline void ignite_init_usbuart_switch(void)
{
	int err;
	err = gpio_request(MFP_CFG_PIN(USB2UART_EN), "usb2uart_en");
	BUG_ON(err);
	err = gpio_request(MFP_CFG_PIN(USB2UART_SEL),"usb2uart_sel");
	BUG_ON(err);

	gpio_direction_output(MFP_CFG_PIN(USB2UART_EN), 1);
	gpio_direction_output(MFP_CFG_PIN(USB2UART_SEL), 1);
}
#endif

/**************************************************
 ** ignite_audio_amp_pm_en_read: read the speaker PA status
 ** 1: enable
 ** 0: disable
 ***************************************************/ 
static ssize_t ignite_audio_amp_pm_en_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	int value;

	value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN))) ? 1 : 0;

	printk(KERN_INFO "<%s>: [read] after copying to user:  value = %d\n", "AUDIO_AMP_PM_EN", value); 

	if ((count = sprintf(buf, "%d", value)) < 0)
		return -EFAULT;

	printk(KERN_INFO "<%s>: [read] after copying to user:  buf = %s\n", "AUDIO_AMP_PM_EN", buf);

	return count;
}

/**************************************************
 ** ignite_audio_amp_pm_en_write: set the speaker PA status  
 ** 1: enable.
 ** 0: disable.
 ***************************************************/ 
static ssize_t ignite_audio_amp_pm_en_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	char value[2];
	int ret;

	if ((count = sscanf(buf, "%1s", value)) < 0)
		return -EFAULT;

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN)), "AUDIO_AMP_PM_EN");
	if (ret)
	{
		gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));
		printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_AMP_PM_EN", 
                                                                    MFP_CFG_PIN(AUDIO_AMP_PM_EN), ret);
		return -1;
	}

	if (value[0] == '1')
	{
		gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AUDIO_AMP_PM_EN), MFP_LPM_DRIVE_HIGH);
	}
	else if (value[0] == '0')
	{
		gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AUDIO_AMP_PM_EN), MFP_LPM_DRIVE_LOW);
	}

	gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	return count;
}

static struct bin_attribute speaker_mode_attr = {
	.attr = {
		.name = "audio_amp_pm_en",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = 4,
	.read = ignite_audio_amp_pm_en_read,
	.write = ignite_audio_amp_pm_en_write,
};

static int ignite_audio_amp_pm_en_init(void)
{
	int ret = 0;

	ret = sysfs_create_bin_file(kernel_kobj, &speaker_mode_attr);
	if (ret)
	{
		printk(KERN_ERR "<%s> Failed to create sys file\n", "AUDIO_AMP_PM_EN");
		return -ENOMEM;
	}

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_AMP_PM_EN)), "AUDIO_AMP_PM_EN");
	if (ret)
	{
		gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));
		printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_AMP_PM_EN", 
                                                            MFP_CFG_PIN(AUDIO_AMP_PM_EN), ret);
		return -1;
	}

	gpio_direction_output(MFP_CFG_PIN(AUDIO_AMP_PM_EN), 0);

	gpio_free(MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	printk(KERN_INFO "<%s> GPIO%d setup done\n", "AUDIO_AMP_PM_EN", MFP_CFG_PIN(AUDIO_AMP_PM_EN));

	return 0;
}

//beacon headset amplifier seting.

/**************************************************
 ** ignite_hp_amp_sd_read: read the headset amplifier status.
 ** 0: disable
 ** 1: enable
 ***************************************************/ 
static ssize_t ignite_hp_amp_sd_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    int value;

    value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD))) ? 1 : 0;
                        
    printk(KERN_INFO "<%s>: [read] after copying to user:  value = %d\n", "HP_AMP_SD", value); 
                            
    if ((count = sprintf(buf, "%d", value)) < 0)
        return -EFAULT;
                                
    printk(KERN_INFO "<%s>: [read] after copying to user:  buf = %s\n", "HP_AMP_SD", buf);

    return count;
}

/**************************************************
 ** ignite_hp_amp_sd_write: set the headset amplifier.  
 ** 0: close headset amplifier.
 ** 1: open headset amplifier.
 ***************************************************/ 
static ssize_t ignite_hp_amp_sd_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    char value[2];
    int ret;

    if ((count = sscanf(buf, "%1s", value)) < 0)
        return -EFAULT;

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(MFP_CFG_PIN(HP_AMP_SD));
        printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "HP_AMP_SD",                        
                                                                    MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    if (value[0] == '1')
    {
        gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 1);
    }
    else if (value[0] == '0')
    {
        gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 0);
    }

    gpio_free(MFP_CFG_PIN(HP_AMP_SD));

    return count;
}

static struct bin_attribute headset_mode_attr = {
    .attr = {
            .name = "hp_amp_sd",
            .owner = THIS_MODULE,
            .mode = S_IRUGO | S_IWUGO,
            },
            .size = 4,
            .read = ignite_hp_amp_sd_read,
            .write = ignite_hp_amp_sd_write,
};

static int ignite_hp_amp_sd_init(void)
{
    int ret = 0;

    ret = sysfs_create_bin_file(kernel_kobj, &headset_mode_attr);
    if (ret)
    {
        printk(KERN_ERR "<%s> Failed to create sys file\n", "HP_AMP_SD");
        return -ENOMEM;
    }

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HP_AMP_SD)), "HP_AMP_SD");
    if (ret)
    {
        gpio_free(MFP_CFG_PIN(HP_AMP_SD));
        printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "HP_AMP_SD", MFP_CFG_PIN(HP_AMP_SD), ret);
        return -1;
    }

    gpio_direction_output(MFP_CFG_PIN(HP_AMP_SD), 0);

    gpio_free(MFP_CFG_PIN(HP_AMP_SD));

    printk(KERN_INFO "<%s> GPIOs setup done\n", "HP_AMP_SD");

    return 0;
}

static void __init ignite_init(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(ignite_mfp_cfg));

	/* register i2c device */
	pxa_set_i2c_info   (&pxa3xx_i2c_platform_data);
	pxa_set_hs_i2c_info(&pxa3xx_i2c_hs_platform_data);

	/* update ISRAM configration for pxa935 */
	if (cpu_is_pxa935())
		pxa3xx_device_imm.resource = pxa935_resource_imm;

    g_pmic_id = pmic_id();

	/* dvfm device */
	set_pxa3xx_freq_info(&ignite_freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

#if defined(CONFIG_SND_PXA3XX_SOC_IGNITE)
	/* audio */
	//pxa3xx_mfp_config(ARRAY_AND_SIZE(audio_mfp_cfg));  /* mfp should NOT be configured here! */
#endif

	/* lcd */
#if defined(CONFIG_FB_PXA950)
	ignite_init_lcd();
#endif


    if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_A1_ID) ){
		platform_add_devices(devices_levante, ARRAY_SIZE(devices_levante));
		printk(KERN_ALERT "--------->platform_add_devices(devices_levante)!!!\n");
	}else
	   printk(KERN_ALERT "--------->platform_add_devices(no devices_micco & no devices_levante )!!!\n");

	platform_add_devices(devices, ARRAY_SIZE(devices));

	if(is_uart_gpio()) {
		pxa_set_ffuart_info(&ignite_ffuart_info);
#ifndef CONFIG_ISPT
		pxa_set_stuart_info(&ignite_stuart_info);
#endif
	} else {
		pxa_set_ffuart_info(NULL);
#ifndef CONFIG_ISPT
		pxa_set_stuart_info(NULL);
#endif
	}
	pxa_set_btuart_info(NULL);

	/* onenand */
	ignite_init_onenand();

#ifdef CONFIG_ISPT
	/*ignite_init_ispt(); FIXME: conflicted with lcd */
#endif
#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)
	/* u2d */
	if (device_is_enabled(device_u2d)) {
		pxa_set_u2d_info(&ignite_u2d_info);
	}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_USB_PXA27X_UDC_MODULE)
	if (device_is_enabled(device_udc)) {
		pxa_set_udc_info(&ignite_udc_info);
	}
#endif
	/* usb otg */
	ignite_init_otg();

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &ignite_u2o_info);
	}
#endif
#if defined(CONFIG_PXA950_CAMERA)
	/* initialize camera */
	ignite_init_cam();
#endif


	/* i2c devices */


  if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_A1_ID) ){

  	i2c_register_board_info(0, ARRAY_AND_SIZE(ignite_i2c_board_info_levante1));
    i2c_register_board_info(0, ARRAY_AND_SIZE(ignite_i2c_board_info_levante2));
   	printk(KERN_ALERT "i2c_register_board_info_levante register !!!\n");

  }else
	printk(KERN_ALERT "i2c_register_board_info PMIC *not* register !!!\n");

	i2c_hs_init();
	i2c_register_board_info(2, ARRAY_AND_SIZE(ignite_i2c_hs_info));

	/* GPS */
#if defined(CONFIG_SIRF_GSD3TW)
	ignite_gps_sirf_init();
#endif

#if defined(CONFIG_PXA_IRDA) || defined(CONFIG_PXA_IRDA_MODULE)
	pxa_set_ficp_info(&ignite_ficp_platform_data);
#endif
	ignite_init_mmc();

	ignite_init_ehci();

#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif
	ignite_set_lowpower_mfp();

#ifdef	CONFIG_PROC_FS
	create_ignite_proc_file();
#endif

#ifdef CONFIG_BATTERY_PDA
	ignite_init_pda_battery();
#endif

#ifdef CONFIG_PDA_POWER
	ignite_init_pda_power();
#endif
	android_set_pmem_info(&android_pmem_pdata);

    ignite_audio_amp_pm_en_init();

    ignite_hp_amp_sd_init();
}

MACHINE_START(IGNITE, "PXA950 handheld Platform (TAVOR EVB III)")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = pxa_map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = ignite_init,
MACHINE_END
