/*
 * Copyright (C) 2008 Borqs, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/device.h>

#define CS_BL_UPD_TH		0x4f /* baseline update threadold */
#define CS_SETL_TIME		0x50 /* settling time */
#define CS_OTH_SET		0x51 /* clock select */
#define CS_HYSTERESIS		0x52 /* hysteresis */
#define CS_DEBOUNCE		0x53 /* debounce */
#define CS_NEG_NOISE_TH		0x54 /* negative noise threshold */
#define CS_LOW_BL_RST		0x55 /* low baseline reset */
#define DEVICE_STATUS 		0x7b /* device status */
#define CS_SLID_MULM		0x77 /* slider resolution */
#define CS_SLID_MULL		0x78 /* slider resolution */
#define CS_SCAN_POS_00 		0x57 /* pin scan position */
#define CS_SCAN_POS_01 		0x58 /* pin scan position */
#define CS_SLID_CONFIG_REG	0x75 /* slider config register */
#define CS_SLEEP_CTRL_PIN	0x7e /* sleep control pin */
#define CS_SLEEP_CTRL		0x7f /* sleep control register */
#define CS_AWAKE_COUNTER	0x80 /* set awake counter */
#define DM_STRONG0		0x11 /* strong port0 */
#define OPEN_DRAIN0		0x13 /* open drain1 */

#define CS_READ_BUTTON 		0x81
#define CS_READ_BLM 		0x82
#define CS_READ_BLL 		0x83
#define CS_READ_DIFFM 		0x84
#define CS_READ_DIFFL 		0x85
#define CS_READ_RAWM 		0x86
#define CS_READ_RAWL 		0x87
#define CS_READ_CEN_POSM 	0x8a
#define CS_READ_CEN_POSL 	0x8b
#define CS_READ_CEN_PEAKM 	0x8c
#define CS_READ_CEN_PEAKL 	0x8d
#define CS_READ_STATUS0 	0x88
#define CS_READ_STATUS1 	0x89

#define INPUT_PORT0 		0x00
#define INPUT_PORT1 		0x01
#define STATUS_HOLD_MASK1 	0x0f

#define CS_ENABLE0		0x06
#define CS_ENABLE1		0x07
#define GPIO_ENABLE0		0x08
#define GPIO_ENABLE1		0x09
#define CS_FINGER_TH_00		0x61
#define CS_FINGER_TH_01		0x62
#define CS_IDAC_00		0x6b
#define CS_IDAC_01		0x6c

#define	COMMAND_REGISTER 	0xa0
#define	ENTER_NORMAL_MODE 	0x07
#define	ENTER_SETUP_MODE 	0x08
#define	DEVICE_RECONFIGURE 	0x06
#define	STORE_CONF_TO_NVRAM 	0x01
#define CS_EXIT_SLEEP_MODE	0x20
#define CS_STOP_SCAN		0x0a
#define CS_READ_FLASH		0x04
#define GPIO_DISABLE_PORT1	0x00
#define GPIO_ENABLE_PORT1_PIN0	0x01
#define GPIO_ENABLE_PORT0_PIN2	0x04
#define DISABLE_PORT0_STRONG_DRIVE	0x00
#define PORT0_OPEN_DRAIN_LOW	0x04
#define CS_SLEEP_CTRL_PIN_VAL 	0x04
#define CS_SLEEP_CTRL_PIN_VAL_SPARK 	0x81
#define CS_SLEEP_CTRL_VAL 	0x99
#define AWAKE_COUNTER_VAL 	0x01
#define DM_STRONG1 0x15
#define DISABLE_PORT1_STRONG_DRIVE 0x00
#define OPEN_DRAIN1 0x17
#define PORT1_OPEN_DRAIN_LOW 0x01


/**
 * chip unlock and lock, addr config
 */
#define I2C_ADDR_DEF		0x00
#define I2C_ADDR_NEW		0x03
#define I2C_CHIP_DM		0x80

#define	I2C_ADDR_DM_REG 	0x7c
#define	I2C_ADDR_DM_SET 	(I2C_CHIP_DM | I2C_ADDR_DEF)

/**
 * device id register
 */
#define DEVICE_ID		0x7a

#define SLEEP_PIN_INVALID	0
#define SLEEP_PIN_VALID		1

struct cy8c201a0_platform_data {
	int (*init_irq)(void);
	int (*ack_irq)(void);
	void (*platform_init)(void);
	void (*disable_chip)(void);
	void (*pre_deep_sleep)(int);
	int (*get_int_pin_value)(void);
	int sleep_pin;
	struct power_chip *power_chips;
};

struct capsense {
	struct i2c_client *client;
	int irq;
	int centroid;
};

