#ifndef _LEVANTE_H_
#define _LEVANTE_H_

struct levante_platform_data {
    int first_xy_sample;
   void    (*platform_init)(void);
};

#if 0
struct sanremo_power_supply_module {
	int command;
	int power_module;
    void (*platform_init)(void);
};

struct sanremo_power_chip {
	int chip_id;
	char *chip_name;
	struct sanremo_power_supply_module *power_supply_modules;
};
#endif

#ifndef _POWER_CHIP_
#define _POWER_CHIP_
struct power_supply_module {
	int command;
	int power_module;
};

struct power_chip {
#define MICCO_B0_ID	0x10
#define MICCO_EA_ID	0x30
#define MICCO_EB_ID	0x31
	int chip_id;
	char *chip_name;
	struct power_supply_module *power_supply_modules;
};
#endif  /* _POWER_CHIP_ */

struct sanremo_platform_data {
    int     (*init_irq)(void);
    int     (*ack_irq)(void);
//    void    (*platform_init)(void);
    spinlock_t  lock;
    struct mutex		m_lock;
    struct work_struct  work;
    struct power_chip   *power_chips;
};

struct portofino_platform_data {
    spinlock_t  lock;
};

/* General */
int sanremo_read(u8 reg, u8 *val);
int sanremo_read_block(u8 reg, u8 *pval, u8 length);
int sanremo_write(u8 reg, u8 val);
int sanremo_write_block(u8 reg,u8 length, const u8 *pval);
void sanremo_rw_lock(void);
void sanremo_rw_unlock(void);
int portofino_read(u8 reg, u8 *val);
int portofino_write(u8 reg, u8 val);

int levante_enable_pen_down_irq(int enable);
int levante_tsi_measurements_config(void);
int levante_tsi_disable_all_measurements(void);
int levante_tsi_disable_xy_measurements(void);
int levante_tsi_enable_xy_measurements(void);
int levante_tsi_readxy(u16 *x, u16 *y, u8 *pen_state);

int levante_enable_vibrator(unsigned char value);
int levante_set_vibrator(unsigned char value);
unsigned int levante_pxa_ticks_to_msec(unsigned int ticks);
void lcd_backlight_power_on(unsigned char intensity);
void lcd_backlight_power_off(void);
void kp_bl_power_on(unsigned char intensity);
void kp_bl_power_off(void);
void charger_on(void);
int levante_check_trim(void);
void levante_open_tsi(void);

/* For Audio */
int levante_codec_read(u8 reg, u8 *val);
int levante_codec_write(u8 reg, u8 val);
#ifdef CODEC_LEVANTE_TO_MICCO_NOT_DONE
int micco_codec_set_input_gain(int type, int gain);
int micco_codec_set_sample_rate(int port, int rate);
int micco_codec_enable_input(int type);
int micco_codec_disable_input(int type);
int micco_codec_enable_output(int type);
int micco_codec_disable_output(int type);
int micco_codec_enable_input(int type);
int micco_codec_disable_input(int type);
int micco_codec_set_output_vol(int type, int vol);
int micco_audio_init(void);
#endif //CODEC_LEVANTE_TO_MICCO_NOT_DONE



#define	LEVANTE_VBUCK1_BASE		0
#define	LEVANTE_VBUCK1_STEP		25
#define	LEVANTE_VBUCK1_MAX		1500
#define LVT_DEBUG               KERN_DEBUG
#define LVT_DEBUG_MUST          KERN_ALERT
#define LEVANTE_X_MES_TRIM             300 /*for debug*/
#define LEVANTE_Y_MES_TRIM      500 /*for debug*/
#define LEVANTE_X_AXIS_MAX             3500
#define LEVANTE_X_AXIS_MIN             0
#define LEVANTE_Y_AXIS_MAX             3500
#define LEVANTE_Y_AXIS_MIN             0
#define LEVANTE_PRESSURE_MAX   255
#define LEVANTE_PRESSURE_MIN   0
#define LEVANTE_TOOL_WIDTH_MAX  15
#define LEVANTE_TOOL_WIDTH_MIN  0
#define LEVNATE_XY_MES_CYC_MS  10

/*VBAT treshold*/
/* battery is empty in discharge state*/
#define VBAT_TS_DISCHARGE_DPOINT 3050
#define VBAT_TS_DISCHARGE_LOW 0000
#define VBAT_TS_DISCHARGE_UPP 3100

/* battery is low - limited functionalites*/
#define VBAT_TS_LOWBAT_DPOINT 3450
#define VBAT_TS_LOWBAT_LOW 3000
#define VBAT_TS_LOWBAT_UPP 3500

/* battery is in normal state - user can work */
#define VBAT_TS_NORMAL_DPOINT 4100
#define VBAT_TS_NORMAL_LOW 3400
#define VBAT_TS_NORMAL_UPP 4300

/* battery is in full state - user can work - high preformance */
#define VBAT_TS_FULL_LOW 4000
#define VBAT_TS_FULL_UPP 5400

/*TBAT treshold*/
/* table in 				mV	== C*/
#define TBAT_TEC_MIN		1472/*(-25)*/
#define TBAT_TEC_LOW_TH		1125/*(-20)*/
#define TBAT_TEC_DPONIT_TH	962	/*(-17)*/
#define TBAT_TEC_UPP_TH		868	/*(-15)*/
#define TBAT_C_LOW_TH		423	/*0*/
#define TBAT_C_DPONIT_TH   	386	/*2*/
#define TBAT_C_UPP_TH		338	/*5*/
#define TBAT_H_LOW_TH		87	/*40*/
#define TBAT_H_DPONIT_TH   	81	/*42*/
#define TBAT_H_UPP_TH		73	/*45*/
#define TBAT_EH_LOW_TH		53	/*55*/
#define TBAT_EH_DPONIT_TH  	49	/*57*/
#define TBAT_EH_UPP_TH		46	/*60*/
#define TBAT_EH_MAX			15	/*100*/
#endif  /* _LEVANTE_H_ */
