#ifndef _MICCO_BATTERY_H
#define _MICCO_BATTERY_H

#define NVM_DATA_SIZE				128
#define NVM_RESERVE_SIZE			(40 * 1024)
#define NVM_VOLT_OFFSET				0
#define NVM_TEMP_OFFSET				30

#define VBAT_ADC_FLOAT_MAX_NUM			2
#define READ_BAT_VOLT_MAX_NUM			3
#define MONITOR_VOLT_NUM			10

#define DEFAULT_BAT_ADC_MIN    			0x101	/* 10-bits */
#define DEFAULT_BAT_ADC_MAX    			0x229	/* 10-bits */
#define DEFAULT_BAT_VOLT_MIN   			3400	/* mV */
#define DEFAULT_BAT_VOLT_MAX   			4200	/* mV */

#define DEFAULT_TEMP_ADC_MIN			0x01
#define DEFAULT_TEMP_ADC_MAX			0x3c
#define DEFAULT_BAT_TEMP_MIN			1
#define DEFAULT_BAT_TEMP_MAX			60

#define BATTERY_TOOLOW_VOLTAGE			DEFAULT_BAT_VOLT_MIN
#define BATTERY_LOW_VOLTAGE			3480	/* mV */
#define VBATMON_NORMAL_TOOLOW_VOLTAGE		3150	/* mV */
#define VBATMON_NORMAL_LOW_VOLTAGE		3300	/* mV */
#define VBATMON_SUSPEND_TOOLOW_VOLTAGE		BATTERY_TOOLOW_VOLTAGE
#define VBATMON_SUSPEND_LOW_VOLTAGE		BATTERY_LOW_VOLTAGE
#define VBATMON_DISABLE_VOLTAGE			0

#define BATT_MONITOR_VOLT_LOW			(BATTERY_LOW_VOLTAGE + 100)
#define LOW_VOLT_MONITOR_INTERVAL		(HZ * 2)
#define BATT_MONITOR_VOLT_HIGH			(DEFAULT_BAT_VOLT_MAX - 100)
#define HIGH_VOLT_MONITOR_INTERVAL		(HZ * 10)

#define DEFAULT_BAT_MONITOR_INTERVAL		(HZ * 60)
#define BAT_MONITOR_QUICK_INTERVAL		(HZ / 10)
#define BAT_MONITOR_INIT_INTERVAL		(HZ * 1)
#define CHARGE_FULL_MIN_TIME			(30 * 60 * 1000) //ms
#define SYSTEM_INIT_TIME			(2 * 60 * 1000) //ms

#define BATTERY_TOOLOW_CAPACITY			1 	/* percent */
#define BATTERY_LOW_CAPACITY			14	/* percent */

#define DEFAULT_BAT_NORMAL_VOLT			3800	/* mV */
#define DEFAULT_BAT_NORMAL_CAPACITY		50	/* mV */

#define BAT_VOLT_FLOAT_BURST			300  	/* mV */
#define BAT_VOLT_FLOAT_MAX			100  	/* mV */
#define BAT_VOLT_FLOAT_MIN			20  	/* mV */

#define BAT_STATE_UNKNOW			0
#define BAT_STATE_NORMAL			1
#define BAT_STATE_LOW				2
#define BAT_STATE_TOOLOW			3
#define BAT_STATE_FULL				4

#define SYS_CALLING				(1 << 0)
#define SYS_AUDIO				(1 << 1)
#define SYS_CHARGING				(1 << 2)

#define BAT_CAPACITY_MAX_VOLT			DEFAULT_BAT_VOLT_MAX
#define BAT_CAPACITY_HIGH_VOLT			3850	/* mV */
#define BAT_CAPACITY_NORMAL_VOLT		3700	/* mV */
#define BAT_CAPACITY_LOW_VOLT			3500	/* mV */
#define BAT_CAPACITY_MIN_VOLT			3000	/* mV */

#define BAT_CAPACITY_MAX_MAH			1000	/* mAh */
#define BAT_CAPACITY_HIGH_MAH			800	/* mAh */
#define BAT_CAPACITY_NORMAL_MAH			200	/* mAh */
#define BAT_CAPACITY_LOW_MAH			80	/* mAh */
#define BAT_CAPACITY_MIN_MAH			0	/* mAh */

#define BAT_VOLT_CAPACITY_LEVEL_FULL		DEFAULT_BAT_VOLT_MAX		
#define BAT_VOLT_CAPACITY_LEVEL_HIGH		4000	/* mV */
#define BAT_VOLT_CAPACITY_LEVEL_NORMAL		3800	/* mV */
#define BAT_VOLT_CAPACITY_LEVEL_LOW		BATTERY_LOW_VOLTAGE	
#define BAT_VOLT_CAPACITY_LEVEL_CRITICAL	BATTERY_TOOLOW_VOLTAGE		
#define BAT_VOLT_CAPACITY_LEVEL_POWEROFF	3500	/* mV */

#define CDEV_AUDIO_LOW_VOL_THRESHOLD		0x10	/* volume */
#define CDEV_AUDIO_HIGH_VOL_THRESHOLD		0x20	/* volume */

#define BAT_CAPACITY_LEVEL_NUM			5

#define	CDEV_LCD_BL_MIN_BIT 		(1 << 0)
#define	CDEV_LCD_BL_MID_BIT		(1 << 1)
#define	CDEV_LCD_BL_MAX_BIT		(1 << 2)
#define	CDEV_CAMERA_BIT			(1 << 3)
#define	CDEV_MTV_BIT			(1 << 4)
#define	CDEV_BT_BIT			(1 << 5)
#define	CDEV_MMC_BIT			(1 << 6)
#define	CDEV_USB_BIT			(1 << 7)
#define	CDEV_GPS_BIT			(1 << 8)
#define	CDEV_AUDIO_NORMAL_BIT		(1 << 9)
#define	CDEV_AUDIO_SPEAK_MIN_BIT	(1 << 10)
#define	CDEV_AUDIO_SPEAK_MID_BIT	(1 << 11)
#define	CDEV_AUDIO_SPEAK_MAX_BIT	(1 << 12)
#define	CDEV_TEMP_D10_BIT		(1 << 13)
#define	CDEV_TEMP_0_BIT			(1 << 14)
#define	CDEV_TEMP_25_BIT		(1 << 15)
#define	CDEV_TEMP_50_BIT		(1 << 16)
#define	CDEV_OP_D0CS_BIT		(1 << 17)
#define	CDEV_OP_104M_BIT		(1 << 18)
#define	CDEV_OP_156M_BIT		(1 << 19)
#define	CDEV_OP_208M_BIT		(1 << 20)
#define	CDEV_OP_416M_BIT		(1 << 21)
#define	CDEV_OP_624M_BIT		(1 << 22)
#define	CDEV_WLAN_BIT			(1 << 23)
#define	CDEV_OP_D1_BIT			(1 << 24)
#define	CDEV_OP_D2_BIT			(1 << 25)
#define	CDEV_GSM_CALL_MAX_SIG_BIT	(1 << 26)
#define	CDEV_GSM_CALL_NORMAL_SIG_BIT	(1 << 27)
#define	CDEV_GSM_CALL_MIN_SIG_BIT	(1 << 28)
#define	CDEV_CHARGE_100_BIT		(1 << 29)
#define	CDEV_CHARGE_500_BIT		(1 << 30)
#define	CDEV_CHARGE_1000_BIT		(1 << 31)
#define CDEV_NUM			32


#define PM_STATUS_OFF	0
#define PM_STATUS_ON	1

#define	DEV_LCD_BIT 	(1 << 0)
#define	DEV_CAMERA_BIT	(1 << 1)
#define	DEV_MTV_BIT	(1 << 2)
#define	DEV_BT_BIT	(1 << 3)
#define	DEV_MMC_BIT	(1 << 4)
#define	DEV_USB_BIT	(1 << 5)
#define	DEV_GPS_BIT	(1 << 6)
#define	DEV_AUDIO_BIT	(1 << 7)

#define CHARGER_CURRENT_100	0x0
#define CHARGER_CURRENT_500	0x02
#define CHARGER_CURRENT_1000	0x01

struct cali_volt_data {
	int len;
	int min_volt;
	int min_adc;
	int max_volt;
	int max_adc;
};

struct cali_temp_data {
	int len;
	int min_temp;
	int min_adc;
	int max_temp;
	int max_adc;
};

int capa_volt[BAT_CAPACITY_LEVEL_NUM] = 
{4200, 3850, 3700, 3500, 3000}; 	/* mV */

int capa_mah[BAT_CAPACITY_LEVEL_NUM] = 
{1000, 800, 200, 80, 0}; 		/* mAh */

int volt_dev_table[][CDEV_NUM+1] = {
/*vol  0  1   2   3   4   5  6    7   8  9  10  11  12  13 14 15 16 17 18 19 20 21 22  23 24 25  26  27  28  29  30  31*/
{4295, 0, 0, 0,  0,  0,  0, 0, -60,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0,  0,  0,-60,-200},
{4290, 0, 0, 0,  0,  0,  0, 0, -60,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0,  0,  0,-60,-210},
{4280, 0, 0, 0,  0,  0,  0, 0, -80,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0,  0,  0,-80,-255},
{4270, 0, 0, 0,  0,  0,  0, 0, -100,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0,  0,  0,  0,  0,-100,-345},
{4260, 0, 0, 0,  0,  0,  0, 0, -120,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0,  0,  0,  0,  0,-120,-350},
{4250, 0, 0, 0,  0,  0,  0, 0, -130,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0,  0,  0,  0,  0,-130,-355},
{4240, 0, 0, 0,  0,  0,  0, 0, -140,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0,  0,  0,  0,  0,-140,-360},
{4220, 0, 0, 0,  0,  0,  0, 0, -150,  0, 0,  0,  0,  0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,  0,  0,  0,-150,-360},
{4210, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4200, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4180, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4160, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4140, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4120, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4100, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-150,-325},
{4090, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-180,-275},
{4080, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{4070, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-265},
{4060, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-260},
{4050, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-255},
{4040, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{4030, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{4020, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{4010, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{4000, 18, 8, 3, 33, 10, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{3980, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-235},
{3960, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-230},
{3940, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3920, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3900, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3880, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3860, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3840, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3820, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3800, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3780, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3760, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3740, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3730, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3720, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,-185,-270},
{3700, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3680, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3660, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3640, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3620, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3600, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3580, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3560, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3540, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3520, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3500, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3480, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3460, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3440, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3420, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{3400, 18, 8, 3, 33, 16, 18, 0,-160, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
{0,    18, 8, 3, 33, 16, 18, 0,   0, 10, 5, 10, 15, 20, 0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 5, 0, 0, 15, 10, 5, -50,   0,-270},
};

int volt2capa_table[][2] = {
{4180, 100},
{4171, 99},
{4163, 98},
{4155, 97},
{4147, 96},
{4138, 95},
{4128, 94},
{4112, 93},
{4103, 92},
{4094, 91},
{4085, 90},
{4077, 89},
{4068, 88},
{4060, 87},
{4052, 86},
{4046, 85},
{4043, 84},
{4036, 83},
{4028, 82},
{4019, 81},
{4011, 80},
{4002, 79},
{3994, 78},
{3985, 77},
{3977, 76},
{3974, 75},
{3970, 74},
{3962, 73},
{3953, 72},
{3945, 71},
{3938, 70},
{3935, 69},
{3930, 68},
{3924, 67},
{3920, 66},
{3914, 65},
{3905, 64},
{3901, 63},
{3898, 62},
{3890, 61},
{3882, 60},
{3873, 59},
{3865, 58},
{3857, 57},
{3848, 56},
{3844, 55},
{3841, 54},
{3832, 53},
{3824, 52},
{3818, 51},
{3814, 50},
{3810, 49},
{3807, 48},
{3802, 47},
{3798, 46},
{3795, 45},
{3791, 44},
{3789, 43},
{3786, 42},
{3784, 41},
{3781, 40},
{3778, 39},
{3775, 38},
{3772, 37},
{3770, 36},
{3768, 35},
{3766, 34},
{3763, 33},
{3761, 32},
{3760, 31},
{3759, 30},
{3757, 29},
{3755, 28},
{3752, 27},
{3749, 26},
{3747, 25},
{3745, 24},
{3743, 23},
{3741, 22},
{3740, 21},
{3737, 20},
{3733, 19},
{3730, 18},
{3727, 17},
{3722, 16},
{3720, 15},
{3716, 14},
{3712, 13},
{3706, 12},
{3697, 11},
{3688, 10},
{3680, 9},
{3677, 8},
{3672, 7},
{3669, 6},
{3650, 5},
{3610, 4},
{3560, 3},
{3420, 2},
{3400, 1},
};

#define BATTERY_VOLTAGE_MODIFICATION_20MA

#if defined(BATTERY_VOLTAGE_MODIFICATION_20MA)
int volt_modi_table[][2] = {
{4200, 0},
{4180, 0},
{4160, 0},
{4140, 0},
{4120, 0},
{4100, 0},
{4080, 0},
{4060, 0},
{4040, 0},
{4020, 0},
{4000, 0},
{3980, 0},
{3960, 0},
{3940, 0},
{3920, 0},
{3900, 0},
{3880, 0},
{3860, 0},
{3840, 0},
{3820, 0},
{3800, 0},
{3780, 0},
{3760, 0},
{3740, 0},
{3720, 0},
{3700, 0},
{3680, 0},
{3660, 0},
{3640, 0},
{3620, 0},
{3600, 0},
{3580, 0},
{3560, 0},
{3540, 0},
{3520, 0},
{3500, 0},
{3480, 0},
{3460, 0},
{3440, 0},
{3420, 0},
{3400, 0},
{0, 0},
};
#elif defined(BATTERY_VOLTAGE_MODIFICATION_GSM)
int volt_modi_table[][2] = {
{4200, 0},
{4180, 0},
{4160, 0},
{4140, 0},
{4120, 0},
{4100, 0},
{4080, 0},
{4060, 0},
{4040, 0},
{4020, 0},
{4000, 0},
{3980, 0},
{3960, 0},
{3940, 0},
{3920, 0},
{3900, 0},
{3880, 0},
{3860, 0},
{3840, 0},
{3820, 0},
{3800, 0},
{3780, 0},
{3760, 0},
{3740, 0},
{3720, 0},
{3700, 0},
{3680, 0},
{3660, 0},
{3640, 0},
{3620, 0},
{3600, 0},
{3580, 0},
{3560, 0},
{3540, 0},
{3520, 0},
{3500, 0},
{3480, 0},
{3460, 0},
{3440, 0},
{3420, 0},
{3400, 0},
{0, 0},
};
#endif

extern int change_dev_state(int type, int used, int level);

#endif

