#ifndef __PMIC_H__
#define __PMIC_H__

#include <linux/i2c.h>
#include <linux/interrupt.h>

/* this enum should be consistent with micco_power_module[]
 * in arch/arm/mach-pxa/xxx(platform).c */
enum {
	/* Set command > 0xFFFF0000 to avoid wrong
	 * parameter is used for pmic_set_voltage.
	 */
	VCC_CORE = 0xFFFF0000,
	VCC_SRAM,
	VCC_MVT,
	VCC_3V_APPS,
	VCC_SDIO,
	VCC_CAMERA_ANA,
	VCC_USB,
	VCC_LCD,
	VCC_TSI,
	VCC_CAMERA_IO,
	VCC_1P8V,
	VCC_MEM,
	HDMI_TX,
	TECH_3V,
	TECH_1P8V,

	/* add your command here */
	VCC_BT,
	VCC_AEC,

	/* max command */
	MAX_CMD,
};
#ifdef CONFIG_PXA3xx_MICCO
/* MICCO_EVENT_A bit(7:0)*/
#define	PMIC_EVENT_INIT		(0)
#define PMIC_EVENT_ONKEY 	(1)
#define	PMIC_EVENT_EXTON	(1 << 2)
#define	PMIC_EVENT_CHDET	(1 << 3)
#define	PMIC_EVENT_TBAT		(1 << 4)
#define	PMIC_EVENT_VBATMON	(1 << 5)
#define	PMIC_EVENT_REV_IOVER	(1 << 6)
#define	PMIC_EVENT_IOVER	(1 << 7)

/* MICCO_EVENT_B bit(15:8)*/
#define	PMIC_EVENT_CH_TCTO	(1 << 8)
#define	PMIC_EVENT_CH_CCTO	(1 << 9)
#define	PMIC_EVENT_USB_DEV	(1 << 10)
#define PMIC_EVENT_OTGCP_IOVER	(1 << 11)
#define PMIC_EVENT_VBUS_4P55	(1 << 12)
#define	PMIC_EVENT_VBUS_3P8	(1 << 13)
#define PMIC_EVENT_SESSION_1P8	(1 << 14)
#define	PMIC_EVENT_SRP_READY	(1 << 15)

/* MICCO_EVENT_C bit(23:16)*/
#define	PMIC_EVENT_ADC_MAN	(1 << 16)
#define	PMIC_EVENT_ADC_AUTO4	(1 << 17)
#define	PMIC_EVENT_ADC_AUTO5	(1 << 18)
#define	PMIC_EVENT_ADC_AUTO6	(1 << 19)
#define	PMIC_EVENT_PEN_DOWN	(1 << 20)
#define	PMIC_EVENT_TSI_READY	(1 << 21)
#define	PMIC_EVENT_DEFAULT		(0xFF)

/* MICCO_EVENT_D bit(31:24)*/
#define	PMIC_EVENT_TX_COMP	(1 << 24)
#define	PMIC_EVENT_RX_COMP	(1 << 25)
#define	PMIC_EVENT_TIMEOUT	(1 << 26)
#define	PMIC_EVENT_HEADSET	(1 << 27)
#define	PMIC_EVENT_HSDETECT	PMIC_EVENT_HEADSET
#define	PMIC_EVENT_HOOKSWITCH	(1 << 28)

/* redefine pmic events */
#define	PMIC_EVENT_TOUCH	PMIC_EVENT_PEN_DOWN
#define	PMIC_EVENT_ADC4		PMIC_EVENT_ADC_AUTO4
#define	PMIC_EVENT_ADC5		PMIC_EVENT_ADC_AUTO5
#define	PMIC_EVENT_ADC6		PMIC_EVENT_ADC_AUTO6

#define	PMIC_EVENT_VBUS		(PMIC_EVENT_USB_DEV |	\
				 PMIC_EVENT_VBUS_4P55 |	\
				 PMIC_EVENT_VBUS_3P8 |	\
				 PMIC_EVENT_SESSION_1P8)
#define	PMIC_EVENT_USB		(PMIC_EVENT_EXTON | PMIC_EVENT_VBUS)

#define	PMIC_EVENT_CHARGER	(PMIC_EVENT_TBAT |	\
				 PMIC_EVENT_REV_IOVER |	\
				 PMIC_EVENT_IOVER |	\
				 PMIC_EVENT_CHDET |	\
				 PMIC_EVENT_CH_CCTO |   \
				 PMIC_EVENT_CH_TCTO |   \
				 PMIC_EVENT_VBATMON)

#else

#define	PMIC_EVENT_INIT		(0)
#define	PMIC_EVENT_EXTON	(1 << 0)
#define	PMIC_EVENT_VBUS		(1 << 1)
#define	PMIC_EVENT_USB		(PMIC_EVENT_EXTON | PMIC_EVENT_VBUS)

#define	PMIC_EVENT_TOUCH	(1 << 2)

#define PMIC_EVENT_OTGCP_IOVER	(1 << 3)

#define	PMIC_EVENT_TBAT		(1 << 4)
#define	PMIC_EVENT_REV_IOVER	(1 << 5)
#define	PMIC_EVENT_IOVER	(1 << 6)
#define	PMIC_EVENT_CHDET	(1 << 7)
#define	PMIC_EVENT_VBATMON	(1 << 8)
#define PMIC_EVENT_HSDETECT	(1 << 9)
#define PMIC_EVENT_HOOKSWITCH   (1 << 10)

#define PMIC_EVENT_BAT_DET      (1 << 11)
#define PMIC_EVENT_VBAT_TS      (1 << 12)
#define PMIC_EVENT_VCHG_TS      (1 << 13)
#define PMIC_EVENT_VSYS         (1 << 14)
#define PMIC_EVENT_TINT         (1 << 15)
#define PMIC_EVENT_CAHRG_FAIL 	(1 << 16)
#define PMIC_EVENT_CHARG_COMP   (1 << 17)
#define PMIC_EVENT_MICIN        (1 << 18)
#define PMIC_EVENT_RF_TEMP      (1 << 19)
#define PMIC_EVENT_ONKEY        (1 << 20)
#define	PMIC_EVENT_DEFAULT		(0xFF)

/* only for pmic_callback_register and pmic_callback_unregister*/
#define	PMIC_EVENT_CHARGER	(PMIC_EVENT_TBAT |	\
				 PMIC_EVENT_REV_IOVER |	\
				 PMIC_EVENT_IOVER 	| \
				 PMIC_EVENT_CHDET 	| \
				 PMIC_EVENT_VBATMON | \
				 PMIC_EVENT_BAT_DET | \
				 PMIC_EVENT_VBAT_TS | \
				 PMIC_EVENT_VCHG_TS | \
				 PMIC_EVENT_VSYS    | \
				 PMIC_EVENT_TINT    | \
				 PMIC_EVENT_CAHRG_FAIL | \
				 PMIC_EVENT_CHARG_COMP)
#endif

struct pmic_ops {
	int (*get_voltage)(int cmd, int *pmv);
	int (*set_voltage)(int cmd, int mv);
	int (*enable_voltage)(int cmd, int enable);
	int (*check_voltage)(int cmd);

	int (*is_vbus_assert)(void);
	int (*is_avbusvld)(void);
	int (*is_asessvld)(void);
	int (*is_bsessvld)(void);
	int (*is_srp_ready)(void);
	int (*is_onkey_assert)(void);

	int (*set_pump)(int enable);
	int (*set_vbus_supply)(int enable, int srp);
	int (*set_usbotg_a_mask)(void);
	int (*set_usbotg_b_mask)(void);
	int (*is_usbpump_chg)(void);
	u8 (*get_rx_value)(void);

	int (*init)(struct device *dev);
	int (*deinit)(struct device *dev);

	struct list_head list;	/* callback list */
	spinlock_t cb_lock;	/* spinlock for callback list */
};

struct pmic_callback {
	unsigned long event;	/* the event which callback care about */
	void (*func)(unsigned long event); /*callback function */
	struct list_head list;
};

struct pxa3xx_pmic_regs {
	unsigned int data:8;
	unsigned int hit:1;
	unsigned int mask:1;
	unsigned int inited:1;
	unsigned int cacheable:1;
};

extern void start_calc_time(void);
extern void end_calc_time(void);

extern int pxa3xx_pmic_write(u8 reg, u8 val);
extern int pxa3xx_pmic_read(u8 reg, u8 *pval);

extern void pmic_set_ops(struct pmic_ops *ops);

extern int pmic_callback_register(unsigned long event,
		void (*func)(unsigned long event));
extern int pmic_callback_unregister(unsigned long event,
		void (*func)(unsigned long event));

extern int pmic_event_handle(unsigned long event);

extern int pxa3xx_pmic_get_voltage(int cmd, int *pval);
extern int pxa3xx_pmic_set_voltage(int cmd, int val);
extern int pxa3xx_pmic_check_voltage(int cmd);
extern int pxa3xx_pmic_enable_voltage(int cmd, int enable);
/* Check whether USB VBUS is asserted */
extern int pxa3xx_pmic_is_vbus_assert(void);
/* Check whether USB VBUS has gone above A-device VBUS valid threshold
 * Min: 4.4V	Max: N/A
 */
extern int pxa3xx_pmic_is_avbusvld(void);
/* Check whether VBUS has gone above A-device Session Valid threshold
 * Min: 0.8V	Max: 2.0V
 */
extern int pxa3xx_pmic_is_asessvld(void);
/* Check whether VBUS has gone above B-device Session Valid threshold
 * Min: 0.8V	Max: 4.0V
 */
extern int pxa3xx_pmic_is_bsessvld(void);
/* Check whether VBUS has gone above B-device Session End threshold
 * Min: 0.2V	Max: 0.8V
 */
extern int pxa3xx_pmic_is_srp_ready(void);
extern int pxa3xx_pmic_is_onkey_assert(void);
/* Initialize the USB PUMP */
extern int pxa3xx_pmic_set_pump(int enable);
/* Check the events change of PMIC */
extern int pxa3xx_pmic_event_change(void);
/* enable/disable VBUS supply */
extern int pxa3xx_pmic_set_vbus_supply(int enable, int srp);
/* Set events mask for USB A-device
 * A-device Sessino Valid event
 * A-device VBUS Valid event
 */
extern int pxa3xx_pmic_set_usbotg_a_mask(void);
/* Set events mask for USB B-device
 * B-device Session Valid event
 * B-device Session end event
 */
extern int pxa3xx_pmic_set_usbotg_b_mask(void);
extern int pxa3xx_pmic_is_onkey_assert(void);
extern u8 pxa3xx_pmic_get_rx_value(void);
#endif


