/*
 * arch/arm/mach-pxa/pm_check_beacon.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pm.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa3xx_pm_check.h>
#include <mach/micco.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/beacon.h>

struct gpio_check pxa3xx_beacon_mfp_setting[] = { 
	{MFP_PIN_GPIO0,		BP_WD_OUT	},
	{MFP_PIN_GPIO1,		ONKEY		},
	{MFP_PIN_GPIO2,		KP_MKIN1	},
	{MFP_PIN_GPIO3,		KP_MKOUT1	},
	{MFP_PIN_GPIO4,		KP_MKIN2	},
	{MFP_PIN_GPIO5,		KP_MKOUT2	},
	{MFP_PIN_GPIO6,		KP_DKIN6	},
	{MFP_PIN_GPIO7,		KP_DKIN7	},
	{MFP_PIN_GPIO8,		CMMB_IRQ	},
	{MFP_PIN_GPIO9,		MMC1_CD_N	},
	{MFP_PIN_GPIO10,	CMMB_RST_N	},
	{MFP_PIN_GPIO11,	DPRAM_INT_N	},
	{MFP_PIN_GPIO12,	CMMB_SPI_CLK_GPIO	},
	{MFP_PIN_GPIO13,	CMMB_SPI_SFRM_GPIO	},
	{MFP_PIN_GPIO14,	CMMB_SPI_TXD_GPIO	},
	{MFP_PIN_GPIO15,	CMMB_SPI_RXD_GPIO	},
	{MFP_PIN_GPIO16,	WLAN_BT_RESET	},
	{MFP_PIN_GPIO17,	OLED_VSYNC	},
	{MFP_PIN_GPIO18,	OLED_HSYNC	},
	{MFP_PIN_GPIO19,	OLED_PCLK	},
	{MFP_PIN_GPIO20,	OLED_DE		},
	{MFP_PIN_GPIO21,	GPU_I2C_SCL	},
	{MFP_PIN_GPIO22,	GPU_I2C_SDA	},
	{MFP_PIN_GPIO23,	OLED_R0		},
	{MFP_PIN_GPIO24,	OLED_R1		},
	{MFP_PIN_GPIO25,	OLED_R2		},
	{MFP_PIN_GPIO26,	OLED_R3		},
	{MFP_PIN_GPIO27,	OLED_R4     	},
	{MFP_PIN_GPIO28,	OLED_R5     	},
	{MFP_PIN_GPIO29,	OLED_G0     	},
	{MFP_PIN_GPIO30,	OLED_G1     	},
	{MFP_PIN_GPIO31,	OLED_G2     	},
	{MFP_PIN_GPIO32,	OLED_G3     	},
	{MFP_PIN_GPIO33,	OLED_G4     	},
	{MFP_PIN_GPIO34,	OLED_G5     	},
	{MFP_PIN_GPIO35,	OLED_B0     	},
	{MFP_PIN_GPIO36,	OLED_B1     	},
	{MFP_PIN_GPIO37,	OLED_B2     	},
	{MFP_PIN_GPIO38,	OLED_B3     	},
	{MFP_PIN_GPIO39,	OLED_B4     	},
	{MFP_PIN_GPIO40,	OLED_B5    	},
	{MFP_PIN_GPIO41,	GPU_RESET_N	},
	{MFP_PIN_GPIO42,	HS_DET_N	},
	{MFP_PIN_GPIO43,	OLED_SPI_CLK	},
	{MFP_PIN_GPIO44,	OLED_SPI_CS	},
	{MFP_PIN_GPIO45,	OLED_SPI_D0	},
	{MFP_PIN_GPIO46,	OLED_RESET_N	},
	{MFP_PIN_GPIO47,	FLAG_OVP	},
	{MFP_PIN_GPIO48,	BP_READY	},
	{MFP_PIN_GPIO49,	AP_READY	},
	{MFP_PIN_GPIO50,	PSEN_DOUT	},
	{MFP_PIN_GPIO51,	FF_CTS		},
	{MFP_PIN_GPIO52,	FF_RTS		},
	{MFP_PIN_GPIO53,	FF_RXD		},
	{MFP_PIN_GPIO54,	FF_TXD		},
	{MFP_PIN_GPIO55,	MMC1_CMD	},
	{MFP_PIN_GPIO56,	MMC1_CLK	},
	{MFP_PIN_GPIO57,	MMC1_DAT0	},
	{MFP_PIN_GPIO58,	MMC1_DAT1	},
	{MFP_PIN_GPIO59,	MMC1_DAT2	},
	{MFP_PIN_GPIO60,	MMC1_DAT3	},
	{MFP_PIN_GPIO61,	GPIO61_FREE	},
	{MFP_PIN_GPIO62,	GPIO62_FREE	},
	{MFP_PIN_GPIO63,	AEC_IRQ		},
	{MFP_PIN_GPIO64,	SLEEP_CLK_BT_A	},
	{MFP_PIN_GPIO65,	CAM_DATA7	},
	{MFP_PIN_GPIO66,	CAM_DATA6	},
	{MFP_PIN_GPIO67,	CAM_DATA5	},
	{MFP_PIN_GPIO68,	CAM_DATA4	},
	{MFP_PIN_GPIO69,	CAM_DATA3	},
	{MFP_PIN_GPIO70,	CAM_DATA2	},
	{MFP_PIN_GPIO71,	CAM_DATA1	},
	{MFP_PIN_GPIO72,	CAM_DATA0	},
	{MFP_PIN_GPIO73,	CAM_HSYNC	},
	{MFP_PIN_GPIO74,	CAM_VSYNC	},
	{MFP_PIN_GPIO75,	CAM_MCLK	},
	{MFP_PIN_GPIO76,	CAM_PCLK	},
	{MFP_PIN_GPIO77,	STRXD		},
	{MFP_PIN_GPIO78,	STTXD		},
	{MFP_PIN_GPIO79,	AP_PCM_CLK_OFF	},
	{MFP_PIN_GPIO80,	AP_PCM_FRM_OFF	},
	{MFP_PIN_GPIO81,	AP_PCM_TXD_OFF	},
	{MFP_PIN_GPIO82,	AP_PCM_RXD_OFF	},
	{MFP_PIN_GPIO83,	GPIO83_SSP3_RXD	},
	{MFP_PIN_GPIO84,	CMMB_PON	},
	{MFP_PIN_GPIO85,	HIFI_CLK	},
	{MFP_PIN_GPIO86,	AP_WAKEUP_BT	},
	{MFP_PIN_GPIO87,	HIFI_FRM	},
	{MFP_PIN_GPIO88,	HIFI_TXD	},
	{MFP_PIN_GPIO89,	I2C_SCL		},
	{MFP_PIN_GPIO90,	I2C_SDA		},
	{MFP_PIN_GPIO91,	BT_RXD		},
	{MFP_PIN_GPIO92,	BT_TXD		},
	{MFP_PIN_GPIO93,	CHARGE_STAT	},
	{MFP_PIN_GPIO94,	GSEN_INT	},
	{MFP_PIN_GPIO95,	BT_CTS		},
	{MFP_PIN_GPIO96,	BT_RTS		},
	{MFP_PIN_GPIO97,	AP_WAKEUP_WLAN	},
	{MFP_PIN_GPIO98,	AEC_RESET_N	},
	{MFP_PIN_GPIO99,	WLAN_WAKEUP_AP	},
	{MFP_PIN_GPIO100,	EAR_POP		},
	{MFP_PIN_GPIO101,	WLAN_DAT3	},
	{MFP_PIN_GPIO102,	WLAN_DAT2	},
	{MFP_PIN_GPIO103,	WLAN_DAT1	},
	{MFP_PIN_GPIO104,	WLAN_DAT0	},
	{MFP_PIN_GPIO105,	WLAN_CMD	},
	{MFP_PIN_GPIO106,	WLAN_CLK	},
	{MFP_PIN_GPIO107,	FLASH_EN	},
	{MFP_PIN_GPIO108,	PSEN_PD		},
	{MFP_PIN_GPIO109,	DBG_OE_N	},
	{MFP_PIN_GPIO110,	PSEN_LEDON	},
	{MFP_PIN_GPIO111,	CAM_PWR_EN	},
	{MFP_PIN_GPIO112,	GPIO112_FREE	},
	{MFP_PIN_GPIO113,	WIFI_PMU_PON	},
	{MFP_PIN_GPIO114,	GPU_PWR_EN	},
	{MFP_PIN_GPIO115,	AP_DBG_LED	},
	{MFP_PIN_GPIO116,	AP_CHG_LED	},
	{MFP_PIN_GPIO117,	GPS_PWR_EN	},
	{MFP_PIN_GPIO118,	GPIO118_FREE	},
	{MFP_PIN_GPIO119,	GPIO119_FREE	},
	{MFP_PIN_GPIO120,	GPIO120_FREE	},
	{MFP_PIN_GPIO121,	GPIO121_FREE	},
	{MFP_PIN_GPIO122,	CAM_MCLK_SEL	},
	{MFP_PIN_GPIO123,	CAM_REAR_RESET_N},
	{MFP_PIN_GPIO124,	CMMB_PWR_EN	},
	{MFP_PIN_GPIO125,	WLAN_PON	},
//	{MFP_PIN_GPIO126,	SYSCLK_26M	},
	{MFP_PIN_GPIO127,	AUDIO_PA_EN	},

	{MFP_PIN_GPIO159,	CHARGE_PG_N	},
	{MFP_PIN_GPIO160,	I2C_SCL_ECHO	},
	{MFP_PIN_GPIO161,	GPS_SLEEP_CLK_EN},
	{MFP_PIN_GPIO162,	CAM_REAR_PDN	},
	{MFP_PIN_GPIO163,	CSEN_DRDY	},
	{MFP_PIN_GPIO164,	AGPS_RESET	},
	{MFP_PIN_GPIO165,	I2C_SDA_ECHO	},
	{MFP_PIN_GPIO166,	BT_WAKEUP_AP	},
	{MFP_PIN_GPIO167,	BP_ON		},
	{MFP_PIN_GPIO168,	BP_SHUTDOWN	},
	{MFP_PIN_GPIO169,	AP_RESET_BP	},
	{MFP_PIN_GPIO170,	CAM_VGA_PDN	},
	{MFP_PIN_GPIO171,	CHARGE_CE_N	},
	{MFP_PIN_GPIO172,	CHARGE_EN1	},
	{MFP_PIN_GPIO173,	CHARGE_EN2	},
	{MFP_PIN_GPIO174,	AGPS_PON	},
	{MFP_PIN_GPIO175,	GPIO175_FREE	},
	{MFP_PIN_GPIO176,	CSEN_RST_N	},
	{-1,	0},
};

struct micco_check pxa3xx_beacon_micco_setting[] = {
//	{MICCO_CHIP_ID,			0	},	// 0x00
///	{MICCO_EVENT_A,			0x00	},	// 0x01
///	{MICCO_EVENT_B,			0x00	},	// 0x02	
///	{MICCO_EVENT_C,			0x00	},	// 0x03
///	{MICCO_EVENT_D,			0x00	},	// 0x04
//	{MICCO_STATUS_A,		0	},	// 0x05
//	{MICCO_STATUS_B,		0	},	// 0x06
	{MICCO_IRQ_MASK_A,		0xF6	},	// 0x07
	{MICCO_IRQ_MASK_B,		0xFF	},	// 0x08
	{MICCO_IRQ_MASK_C,		0x3F	},	// 0x09
	{MICCO_IRQ_MASK_D,		0xE7	},	// 0x0A
	{MICCO_SYSCTRL_A,		0x63	},	// 0x0B
	{MICCO_SYSCTRL_B,		0x00	},	// 0x0C
	{MICCO_FAULT_LOG,		0x00	},	// 0x0D
	{MICCO_OVER1,			0x05	},	// 0x10
	{MICCO_APP_OVER2,		0x7F	},	// 0x11
//	{MICCO_APP_OVER3,		0x39	},	// 0x12
	{MICCO_LDO643,			0x6A	},	// 0x13
	{MICCO_LDO987,			0x73	},	// 0x14
//	{MICCO_LDO1110,			0x00	},	// 0x15
//	{MICCO_LDO12,			0x00	},	// 0x16
	{MICCO_LDO13,			0x0A	},	// 0x17
	{MICCO_BUCK_SLEEP,		0x05	},	// 0x18
	{MICCO_LDO_nSLEEP1,		0x00	},	// 0x19
	{MICCO_LDO_nSLEEP2,		0x00	},	// 0x1A
	{MICCO_LDO_SLEEP_BIT1,		0x04	},	// 0x1B
	{MICCO_LDO_SLEEP_BIT2,		0x00	},	// 0x1C
	{MICCO_BUCK_LDO8_EN,		0x05	},	// 0x1D
//	{MICCO_VCC2,			0x00	},	// 0x1F
//	{MICCO_VCC1,			0x00	},	// 0x20
	{MICCO_COMM_OVER2,		0x00	},	// 0x21
	{MICCO_COMM_OVER3,		0x00	},	// 0x22
//	{MICCO_BUCK1_DVC1,		0x00	},	// 0x23
//	{MICCO_BUCK1_DVC2,		0x0B	},	// 0x24
//	{MICCO_APPS_AVRC,		0x18	},	// 0x25
//	{MICCO_COMM_CDTV1,		0x08	},	// 0x26
//	{MICCO_COMM_CDTV2,		0x0F	},	// 0x27
//	{MICCO_COMM_CVRC,		0x18	},	// 0x28
//	{MICCO_SRAM_DVC1,		0x1B	},	// 0x29
//	{MICCO_SRAM_DVC2,		0x0F	},	// 0x2A
//	{MICCO_SRAM_SVRC,		0x18	},	// 0x2B
//	{MICCO_3DTV1,			0x02	},	// 0x2C
//	{MICCO_3DTV2,			0x03	},	// 0x2D
//	{MICCO_3DVRC,			0x02	},	// 0x2E
//	{MICCO_BULK_AUTO_SKIP,		0x07	},	// 0x2F
//	{MICCO_LDO1_MDTV1,		0x94	},	// 0x32
//	{MICCO_LDO1_MDTV2,		0x14	},	// 0x33
//	{MICCO_LDO1_MVRC,		0x02	},	// 0x34
	{MICCO_LED1_CONTROL,		0x00	},	// 0x35
	{MICCO_LED2_CONTROL,		0x00	},	// 0x36
	{MICCO_LEDPC_CONTROL1,		0x00	},	// 0x37
	{MICCO_LEDPC_CONTROL2,		0x00	},	// 0x38
	{MICCO_LEDPC_CONTROL3,		0x00	},	// 0x39
	{MICCO_LEDPC_CONTROL4,		0x00	},	// 0x3A
	{MICCO_LEDPC_CONTROL5,		0x00	},	// 0x3B
	{MICCO_WLED_CONTROL1,		0x00	},	// 0x3C
	{MICCO_WLED_CONTROL2,		0x00	},	// 0x3D
	{MICCO_FLASH_CONTROL1,		0x00	},	// 0x3E
	{MICCO_FLASH_CONTROL2,		0x00	},	// 0x3F
	{MICCO_VIBRA_CONTROL,		0x00	},	// 0x40
	{MICCO_MISC,			0x05	},	// 0x41
	{MICCO_CHARGE_CONTROL,		0x22	},	// 0x42
	{MICCO_CCTR_CONTROL,		0x00	},	// 0x43
	{MICCO_TCTR_CONTROL,		0x00	},	// 0x44
//	{MICCO_CHARGE_PULSE,		0x07	},	// 0x45
//	{MICCO_DAC_MSB,			0x01	},	// 0x46
//	{MICCO_DAC_LSB,			0x00	},	// 0x47
	{MICCO_UART_ADDR,		0x08	},	// 0x48
	{MICCO_UART_CONTROL,		0x05	},	// 0x49
	{MICCO_UART_TX_DATA,		0x00	},	// 0x4A
	{MICCO_UART_RX_DATA,		0x00	},	// 0x4B
//	{MICCO_ADC_MAN_CONTROL,		0	},	// 0x50
//	{MICCO_ADC_AUTO_CONTROL_1,	0	},	// 0x51
//	{MICCO_ADC_AUTO_CONTROL_2,	0	},	// 0x52
//	{MICCO_TSI_CONTROL_1,		0	},	// 0x53
//	{MICCO_TSI_CONTROL_2,		0	},	// 0x54
//	{MICCO_VBATMON,			0	},	// 0x55
//	{MICCO_TBATHIGHP,		0	},	// 0x56
//	{MICCO_TBATHIGHN,		0	},	// 0x57
//	{MICCO_TBATLOW,			0	},	// 0x58
//	{MICCO_AUTO4HIGH,		0	},	// 0x59
//	{MICCO_AUTO4LOW,		0	},	// 0x5A
//	{MICCO_AUTO5HIGH,		0	},	// 0x5B
//	{MICCO_AUTO5LOW,		0	},	// 0x5C
//	{MICCO_AUTO6HIGH,		0	},	// 0x5D
//	{MICCO_AUTO6LOW,		0	},	// 0x5E
//	{MICCO_MAN_RES_LSB,		0	},	// 0x5F
//	{MICCO_MAN_RES_MSB,		0	},	// 0x60
//	{MICCO_VBAT_RES,		0	},	// 0x61
//	{MICCO_VBATMIN_RES,		0	},	// 0x62
//	{MICCO_ICHMAX_RES,		0	},	// 0x63
//	{MICCO_ICHMIN_RES,		0	},	// 0x64
//	{MICCO_ICHAVERAGE_RES,		0	},	// 0x65
//	{MICCO_VCHMAX_RES,		0	},	// 0x66
//	{MICCO_VCHMIN_RES,		0	},	// 0x67
//	{MICCO_TBAT_RES,		0	},	// 0x68
//	{MICCO_AUTO4_RES,		0	},	// 0x69
//	{MICCO_AUTO5_RES,		0	},	// 0x6A
//	{MICCO_AUTO6_RES,		0	},	// 0x6B
//	{MICCO_TSI_X_MSB,		0	},	// 0x6C
//	{MICCO_TSI_Y_MSB,		0	},	// 0x6D
//	{MICCO_TSI_XY_MSB,		0	},	// 0x6E
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_MONO,			0x00	},	// 0x70
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_BEAR,			0x00	},	// 0x71
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_LINE_OUT,		0x00	},	// 0x72
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH1,		0x00	},	// 0x73
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH2,		0x00	},	// 0x74
	{MICCO_AUDIO_REG_BASE + MICCO_AUDIO_LINE_AMP,		0x25	},	// 0x75
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH1,	0x3F	},	// 0x76
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH2,	0x3F	},	// 0x77
	{MICCO_AUDIO_REG_BASE + MICCO_HIFI_DAC_CONTROL,		0x00	},	// 0x78
	{MICCO_AUDIO_REG_BASE + MICCO_MONO_VOL,			0xBF	},	// 0x79
	{MICCO_AUDIO_REG_BASE + MICCO_BEAR_VOL,			0xBF	},	// 0x7A
///	{MICCO_AUDIO_REG_BASE + MICCO_I2S_CONTROL,		0x40	},	// 0x7B
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA,			0x00	},	// 0x7C
	{MICCO_AUDIO_REG_BASE + MICCO_MIC_PGA,			0x00	},	// 0x7D
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA_MUX,		0x00	},	// 0x7E
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_ADC_CONTROL,	0x02	},	// 0x7F
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_VDAC_CONTROL,	0x40	},	// 0x80
	{MICCO_AUDIO_REG_BASE + MICCO_SIDETONE,			0x00	},	// 0x81
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX1_2,		0x00	},	// 0x82
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX3,			0x00	},	// 0x83
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_DACS,			0x00	},	// 0x84
	{MICCO_AUDIO_REG_BASE + MICCO_SOFT_START_RAMP,		0xFF	},	// 0x85
	{MICCO_AUDIO_REG_BASE + MICCO_REM_IN_POLLING_TIME,	0x00	},	// 0x86
	{0xFF,	0},
};

static int __init pm_check_beacon_init(void)
{
	if( machine_is_beacon() ){
		register_pxa3xx_mfp_setting(&pxa3xx_beacon_mfp_setting);
		register_pxa3xx_micco_setting(&pxa3xx_beacon_micco_setting);
	}
	return 0;
}

static void __exit pm_check_beacon_exit(void)
{
}

module_init(pm_check_beacon_init);
module_exit(pm_check_beacon_exit);
