/*
 * arch/arm/mach-pxa/pm_check_jades.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pm.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <ma/pxa3xx-regs.h>
#include <mach/pxa3xx_pm_check.h>
#include <mach/micco.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/jades.h>

struct gpio_check pxa3xx_jades_mfp_setting[] = { 
	{MFP_PIN_GPIO0,		GPIO0_KP_MKIN_0		},
	{MFP_PIN_GPIO1,		GPIO1_KP_MKOUT_0	},
	{MFP_PIN_GPIO2,		GPIO2_KP_MKIN_1		},
	{MFP_PIN_GPIO3,		GPIO3_KP_MKOUT_1	},
	{MFP_PIN_GPIO4,		GPIO4_KP_DKIN_4		},
	{MFP_PIN_GPIO5,		GPIO5_GPIO_FREE		},
	{MFP_PIN_GPIO6,		GPIO6_CAM_PWDN_LPM	},
	{MFP_PIN_GPIO7,		GPIO7_IF101_RESET_LPM	},
	{MFP_PIN_GPIO8,		GPIO8_GPIO_MMC_DETECT	},
	{MFP_PIN_GPIO9,		GPIO9_IF101_INTR_LPM	},
	{MFP_PIN_GPIO10,	GPIO10_CHARGING_STATE	},
	{MFP_PIN_GPIO11,	GPIO11_CHARGE_CTR_EN1	},
	{MFP_PIN_GPIO12,	0			},
	{MFP_PIN_GPIO13,	0			},
	{MFP_PIN_GPIO14,	0			},
	{MFP_PIN_GPIO15,	0			},
	{MFP_PIN_GPIO16,	GPIO16_GPIO_FREE	},
	{MFP_PIN_GPIO17,	GPIO17_LCD_FCLK_RD	},
	{MFP_PIN_GPIO18,	GPIO18_LCD_LCLK_A0	},
	{MFP_PIN_GPIO19,	GPIO19_LCD_PCLK_WR	},
	{MFP_PIN_GPIO20,	GPIO20_LCD_BIAS		},
	{MFP_PIN_GPIO21,	GPIO21_GPIO_LCD_RESET	},
	{MFP_PIN_GPIO22,	GPIO22_GPIO_FREE	},
	{MFP_PIN_GPIO23,	GPIO23_LCD_DD0		},
	{MFP_PIN_GPIO24,	GPIO24_LCD_DD1		},
	{MFP_PIN_GPIO25,	GPIO25_LCD_DD2		},
	{MFP_PIN_GPIO26,	GPIO26_LCD_DD3		},
	{MFP_PIN_GPIO27,	GPIO27_LCD_DD4		},
	{MFP_PIN_GPIO28,	GPIO28_LCD_DD5		},
	{MFP_PIN_GPIO29,	GPIO29_LCD_DD6		},
	{MFP_PIN_GPIO30,	GPIO30_LCD_DD7		},
	{MFP_PIN_GPIO31,	GPIO31_LCD_DD8		},
	{MFP_PIN_GPIO32,	GPIO32_LCD_DD9		},
	{MFP_PIN_GPIO33,	GPIO33_LCD_DD10		},
	{MFP_PIN_GPIO34,	GPIO34_LCD_DD11		},
	{MFP_PIN_GPIO35,	GPIO35_LCD_DD12		},
	{MFP_PIN_GPIO36,	GPIO36_LCD_DD13		},
	{MFP_PIN_GPIO37,	GPIO37_LCD_DD14		},
	{MFP_PIN_GPIO38,	GPIO38_LCD_DD15		},
	{MFP_PIN_GPIO39,	GPIO39_LCD_DD16		},
	{MFP_PIN_GPIO40,	GPIO40_LCD_DD17		},
	{MFP_PIN_GPIO41,	(MFP_PIN_GPIO41 << 16)|(MFPR_DEFAULT & ~MFPR_LPM_MASK)|( 0x1  << 15)|(MFP_AF2)},
	{MFP_PIN_GPIO42,	(MFP_PIN_GPIO42 << 16)|(MFPR_DEFAULT & ~MFPR_LPM_MASK)|( 0x1  << 15)|(MFP_AF2)},
	{MFP_PIN_GPIO43,	GPIO43_SSP4_CLK		},
	{MFP_PIN_GPIO44,	GPIO44_SSP4_FRM		},
	{MFP_PIN_GPIO45,	GPIO45_SSP4_TXD		},
	{MFP_PIN_GPIO46,	GPIO46_SSP4_RXD		},		// free ???
	{MFP_PIN_GPIO47,	GPIO47_GPIO_FREE	},
	{MFP_PIN_GPIO48,	GPIO48_GPIO_FREE	},
	{MFP_PIN_GPIO49,	GPIO49_GPIO_FREE	},
	{MFP_PIN_GPIO50,	GPIO50_GPIO_BT_WAKE	},
	{MFP_PIN_GPIO51,	GPIO51_CAM_RESET_LPM	},
	{MFP_PIN_GPIO52,	GPIO52_GPIO_FREE	},
	{MFP_PIN_GPIO53,	GPIO53_FFRXD		},
	{MFP_PIN_GPIO54,	GPIO54_FFTXD		},
	{MFP_PIN_GPIO55,	GPIO55_MMC1_CMD		},
	{MFP_PIN_GPIO56,	GPIO56_MMC1_CLK		},
	{MFP_PIN_GPIO57,	GPIO57_MMC1_DAT0	},
	{MFP_PIN_GPIO58,	GPIO58_MMC1_DAT1	},
	{MFP_PIN_GPIO59,	GPIO59_MMC1_DAT2	},
	{MFP_PIN_GPIO60,	GPIO60_MMC1_DAT3	},
	{MFP_PIN_GPIO61,	GPIO61_CHARGER_IN	},
	{MFP_PIN_GPIO62,	0},
	{MFP_PIN_GPIO63,	GPIO63_CAM_I2C_OFF_LPM	},
	{MFP_PIN_GPIO64,	GPIO64_CHARGE_CTR_EN2	},
	{MFP_PIN_GPIO65,	GPIO65_CI_DD_7_LPM	},
	{MFP_PIN_GPIO66,	GPIO66_CI_DD_6_LPM	},
	{MFP_PIN_GPIO67,	GPIO67_CI_DD_5_LPM	},
	{MFP_PIN_GPIO68,	GPIO68_CI_DD_4_LPM	},
	{MFP_PIN_GPIO69,	GPIO69_CI_DD_3_LPM	},
	{MFP_PIN_GPIO70,	GPIO70_CI_DD_2_LPM	},
	{MFP_PIN_GPIO71,	GPIO71_CI_DD_1_LPM	},
	{MFP_PIN_GPIO72,	GPIO72_CI_DD_0_LPM	},
	{MFP_PIN_GPIO73,	0			},
	{MFP_PIN_GPIO74,	0			},
	{MFP_PIN_GPIO75,	0			},
	{MFP_PIN_GPIO76,	GPIO76_CI_PCLK_LPM	},
	{MFP_PIN_GPIO77,	GPIO77_STTXD		},
	{MFP_PIN_GPIO78,	GPIO78_STRXD		},
	{MFP_PIN_GPIO79,	0			},
	{MFP_PIN_GPIO80,	0			},
	{MFP_PIN_GPIO81,	0			},
	{MFP_PIN_GPIO82,	0			},
	{MFP_PIN_GPIO83,	GPIO83_SSP3_RXD		},
	{MFP_PIN_GPIO84,	GPIO84_CAM_VDD_EN_LPM	},
	{MFP_PIN_GPIO85,	0			},
	{MFP_PIN_GPIO86,	GPIO86_GPIO_BT_RESET	},
	{MFP_PIN_GPIO87,	0			},
	{MFP_PIN_GPIO88,	0			},
	{MFP_PIN_GPIO89,	GPIO89_CI2C_SCL		},
	{MFP_PIN_GPIO90,	GPIO90_CI2C_SDA		},
	{MFP_PIN_GPIO91,	GPIO91_BTRXD		},
	{MFP_PIN_GPIO92,	GPIO92_BTTXD		},
	{MFP_PIN_GPIO93,	GPIO93_BTCTS_N		},
	{MFP_PIN_GPIO94,	GPIO94_BTRTS_N		},
	{MFP_PIN_GPIO95,	GPIO95_GPIO_LED		},
	{MFP_PIN_GPIO96,	GPIO96_GPIO_SMB380_IRQ	},
	{MFP_PIN_GPIO97,	0			},
	{MFP_PIN_GPIO98,	0			},
	{MFP_PIN_GPIO99,	GPIO99_GPIO_FREE	},
	{MFP_PIN_GPIO100,	GPIO100_GPIO_FREE	},
	{MFP_PIN_GPIO101,	GPIO101_GPIO_FREE	},
	{MFP_PIN_GPIO102,	0},
	{MFP_PIN_GPIO103,	0},
	{MFP_PIN_GPIO104,	0},
	{MFP_PIN_GPIO105,	0},
	{MFP_PIN_GPIO106,	0},
	{-1,	0},
};

struct micco_check pxa3xx_jades_micco_setting[] = {
//	{MICCO_CHIP_ID,			0	},	// 0x00
///	{MICCO_EVENT_A,			0x00	},	// 0x01
///	{MICCO_EVENT_B,			0x00	},	// 0x02	
///	{MICCO_EVENT_C,			0x00	},	// 0x03
///	{MICCO_EVENT_D,			0x00	},	// 0x04
//	{MICCO_STATUS_A,		0	},	// 0x05
//	{MICCO_STATUS_B,		0	},	// 0x06
	{MICCO_IRQ_MASK_A,		0xDE	},	// 0x07
	{MICCO_IRQ_MASK_B,		0xFB	},	// 0x08
	{MICCO_IRQ_MASK_C,		0xEF	},	// 0x09
	{MICCO_IRQ_MASK_D,		0xE7	},	// 0x0A
	{MICCO_SYSCTRL_A,		0x63	},	// 0x0B
	{MICCO_SYSCTRL_B,		0x00	},	// 0x0C
	{MICCO_FAULT_LOG,		0x00	},	// 0x0D
	{MICCO_OVER1,			0x07	},	// 0x10
	{MICCO_APP_OVER2,		0x66	},	// 0x11
//	{MICCO_APP_OVER3,		0x39	},	// 0x12
	{MICCO_LDO643,			0x70	},	// 0x13
	{MICCO_LDO987,			0x53	},	// 0x14
	{MICCO_LDO1110,			0x00	},	// 0x15
	{MICCO_LDO12,			0x00	},	// 0x16
	{MICCO_LDO13,			0x00	},	// 0x17
	{MICCO_BUCK_SLEEP,		0x03	},	// 0x18
	{MICCO_LDO_nSLEEP1,		0x00	},	// 0x19
	{MICCO_LDO_nSLEEP2,		0x00	},	// 0x1A
	{MICCO_LDO_SLEEP_BIT1,		0x04	},	// 0x1B
	{MICCO_LDO_SLEEP_BIT2,		0x00	},	// 0x1C
	{MICCO_BUCK_LDO8_EN,		0x07	},	// 0x1D
	{MICCO_VCC2,			0x00	},	// 0x1F
	{MICCO_VCC1,			0x00	},	// 0x20
	{MICCO_COMM_OVER2,		0x00	},	// 0x21
	{MICCO_COMM_OVER3,		0x00	},	// 0x22
//	{MICCO_BUCK1_DVC1,		0x00	},	// 0x23
//	{MICCO_BUCK1_DVC2,		0x0B	},	// 0x24
//	{MICCO_APPS_AVRC,		0x18	},	// 0x25
//	{MICCO_COMM_CDTV1,		0x08	},	// 0x26
//	{MICCO_COMM_CDTV2,		0x0F	},	// 0x27
//	{MICCO_COMM_CVRC,		0x18	},	// 0x28
//	{MICCO_SRAM_DVC1,		0x1B	},	// 0x29
//	{MICCO_SRAM_DVC2,		0x0F	},	// 0x2A
//	{MICCO_SRAM_SVRC,		0x18	},	// 0x2B
//	{MICCO_3DTV1,			0x02	},	// 0x2C
//	{MICCO_3DTV2,			0x03	},	// 0x2D
//	{MICCO_3DVRC,			0x02	},	// 0x2E
//	{MICCO_BULK_AUTO_SKIP,		0x07	},	// 0x2F
//	{MICCO_LDO1_MDTV1,		0x94	},	// 0x32
//	{MICCO_LDO1_MDTV2,		0x14	},	// 0x33
//	{MICCO_LDO1_MVRC,		0x02	},	// 0x34
	{MICCO_LED1_CONTROL,		0x00	},	// 0x35
	{MICCO_LED2_CONTROL,		0x00	},	// 0x36
	{MICCO_LEDPC_CONTROL1,		0x00	},	// 0x37
	{MICCO_LEDPC_CONTROL2,		0x00	},	// 0x38
	{MICCO_LEDPC_CONTROL3,		0x00	},	// 0x39
	{MICCO_LEDPC_CONTROL4,		0x00	},	// 0x3A
	{MICCO_LEDPC_CONTROL5,		0x00	},	// 0x3B
//	{MICCO_WLED_CONTROL1,		0x00	},	// 0x3C
	{MICCO_WLED_CONTROL2,		0x00	},	// 0x3D
	{MICCO_FLASH_CONTROL1,		0x00	},	// 0x3E
	{MICCO_FLASH_CONTROL2,		0x00	},	// 0x3F
	{MICCO_VIBRA_CONTROL,		0x00	},	// 0x40
	{MICCO_MISC,			0x15	},	// 0x41
	{MICCO_CHARGE_CONTROL,		0x00	},	// 0x42
	{MICCO_CCTR_CONTROL,		0x00	},	// 0x43
	{MICCO_TCTR_CONTROL,		0x00	},	// 0x44
//	{MICCO_CHARGE_PULSE,		0x07	},	// 0x45
//	{MICCO_DAC_MSB,			0x01	},	// 0x46
//	{MICCO_DAC_LSB,			0x00	},	// 0x47
	{MICCO_UART_ADDR,		0x00	},	// 0x48
	{MICCO_UART_CONTROL,		0x00	},	// 0x49
	{MICCO_UART_TX_DATA,		0x00	},	// 0x4A
	{MICCO_UART_RX_DATA,		0x00	},	// 0x4B
//	{MICCO_ADC_MAN_CONTROL,		0	},	// 0x50
//	{MICCO_ADC_AUTO_CONTROL_1,	0	},	// 0x51
//	{MICCO_ADC_AUTO_CONTROL_2,	0	},	// 0x52
//	{MICCO_TSI_CONTROL_1,		0	},	// 0x53
//	{MICCO_TSI_CONTROL_2,		0	},	// 0x54
//	{MICCO_VBATMON,			0	},	// 0x55
//	{MICCO_TBATHIGHP,		0	},	// 0x56
//	{MICCO_TBATHIGHN,		0	},	// 0x57
//	{MICCO_TBATLOW,			0	},	// 0x58
//	{MICCO_AUTO4HIGH,		0	},	// 0x59
//	{MICCO_AUTO4LOW,		0	},	// 0x5A
//	{MICCO_AUTO5HIGH,		0	},	// 0x5B
//	{MICCO_AUTO5LOW,		0	},	// 0x5C
//	{MICCO_AUTO6HIGH,		0	},	// 0x5D
//	{MICCO_AUTO6LOW,		0	},	// 0x5E
//	{MICCO_MAN_RES_LSB,		0	},	// 0x5F
//	{MICCO_MAN_RES_MSB,		0	},	// 0x60
//	{MICCO_VBAT_RES,		0	},	// 0x61
//	{MICCO_VBATMIN_RES,		0	},	// 0x62
//	{MICCO_ICHMAX_RES,		0	},	// 0x63
//	{MICCO_ICHMIN_RES,		0	},	// 0x64
//	{MICCO_ICHAVERAGE_RES,		0	},	// 0x65
//	{MICCO_VCHMAX_RES,		0	},	// 0x66
//	{MICCO_VCHMIN_RES,		0	},	// 0x67
//	{MICCO_TBAT_RES,		0	},	// 0x68
//	{MICCO_AUTO4_RES,		0	},	// 0x69
//	{MICCO_AUTO5_RES,		0	},	// 0x6A
//	{MICCO_AUTO6_RES,		0	},	// 0x6B
//	{MICCO_TSI_X_MSB,		0	},	// 0x6C
//	{MICCO_TSI_Y_MSB,		0	},	// 0x6D
//	{MICCO_TSI_XY_MSB,		0	},	// 0x6E
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_MONO,			0x00	},	// 0x70
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_BEAR,			0x00	},	// 0x71
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_LINE_OUT,		0x00	},	// 0x72
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH1,		0x00	},	// 0x73
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH2,		0x00	},	// 0x74
	{MICCO_AUDIO_REG_BASE + MICCO_AUDIO_LINE_AMP,		0x25	},	// 0x75
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH1,	0x3F	},	// 0x76
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH2,	0x3F	},	// 0x77
	{MICCO_AUDIO_REG_BASE + MICCO_HIFI_DAC_CONTROL,		0x00	},	// 0x78
	{MICCO_AUDIO_REG_BASE + MICCO_MONO_VOL,			0xBF	},	// 0x79
	{MICCO_AUDIO_REG_BASE + MICCO_BEAR_VOL,			0xBF	},	// 0x7A
///	{MICCO_AUDIO_REG_BASE + MICCO_I2S_CONTROL,		0x40	},	// 0x7B
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA,			0x00	},	// 0x7C
	{MICCO_AUDIO_REG_BASE + MICCO_MIC_PGA,			0x00	},	// 0x7D
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA_MUX,		0x00	},	// 0x7E
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_ADC_CONTROL,	0x02	},	// 0x7F
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_VDAC_CONTROL,	0x40	},	// 0x80
	{MICCO_AUDIO_REG_BASE + MICCO_SIDETONE,			0x00	},	// 0x81
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX1_2,		0x00	},	// 0x82
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX3,			0x00	},	// 0x83
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_DACS,			0x00	},	// 0x84
	{MICCO_AUDIO_REG_BASE + MICCO_SOFT_START_RAMP,		0xFF	},	// 0x85
	{MICCO_AUDIO_REG_BASE + MICCO_REM_IN_POLLING_TIME,	0x00	},	// 0x86
	{0xFF,	0},
};

static int __init pm_check_jades_init(void)
{
	if( machine_is_jades()){
		register_pxa3xx_mfp_setting(&pxa3xx_jades_mfp_setting);
		register_pxa3xx_micco_setting(&pxa3xx_jades_micco_setting);
	}
	return 0;
}

static void __exit pm_check_jades_exit(void)
{
}

module_init(pm_check_jades_init);
module_exit(pm_check_jades_exit);
