/*
 * arch/arm/mach-pxa/pm_check_jades.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pm.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa3xx_pm_check.h>
#include <mach/micco.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/spark.h>

struct gpio_check pxa3xx_spark_mfp_setting[] = { 
	{MFP_PIN_GPIO0,		KP_MKIN0	},
	{MFP_PIN_GPIO1,		KP_MKOUT0	},
	{MFP_PIN_GPIO2,		KP_MKIN1	},
	{MFP_PIN_GPIO3,		KP_MKOUT1	},
	{MFP_PIN_GPIO4,		KP_MKIN2	},
	{MFP_PIN_GPIO5,		KP_MKOUT2	},
	{MFP_PIN_GPIO6,		KP_MKIN3	},
	{MFP_PIN_GPIO7,		KP_MKOUT3	},
	{MFP_PIN_GPIO8,		KP_MKIN4	},
	{MFP_PIN_GPIO9,		KP_MKOUT4	},
	{MFP_PIN_GPIO10,	KP_MKIN5	},
	{MFP_PIN_GPIO11,	KP_MKOUT5	},
	{MFP_PIN_GPIO12,	KP_DKIN0	},
	{MFP_PIN_GPIO13,	KP_MKOUT6	},
	{MFP_PIN_GPIO14,	EXT_CHG_SET_EN2	},
	{MFP_PIN_GPIO15,	EXT_CHG_SET_EN1	},
	{MFP_PIN_GPIO16,	MMC0_CD		},
	{MFP_PIN_GPIO17,	LCD_VSYNC	},
	{MFP_PIN_GPIO18,	LCD_HSYNC	},
	{MFP_PIN_GPIO19,	LCD_PCLK	},
	{MFP_PIN_GPIO20,	LCD_DE		},
	{MFP_PIN_GPIO21,	KEY_LED_EN_T	},
	{MFP_PIN_GPIO22,	GPIO22_FREE	},
	{MFP_PIN_GPIO23,	LCD_R0		},
	{MFP_PIN_GPIO24,	LCD_R1		},
	{MFP_PIN_GPIO25,	LCD_R2		},
	{MFP_PIN_GPIO26,	LCD_R3		},
	{MFP_PIN_GPIO27,	LCD_R4		},
	{MFP_PIN_GPIO28,	LCD_R5		},
	{MFP_PIN_GPIO29,	LCD_G0		},
	{MFP_PIN_GPIO30,	LCD_G1		},
	{MFP_PIN_GPIO31,	LCD_G2		},
	{MFP_PIN_GPIO32,	LCD_G3		},
	{MFP_PIN_GPIO33,	LCD_G4		},
	{MFP_PIN_GPIO34,	LCD_G5		},
	{MFP_PIN_GPIO35,	LCD_B0		},
	{MFP_PIN_GPIO36,	LCD_B1		},
	{MFP_PIN_GPIO37,	LCD_B2		},
	{MFP_PIN_GPIO38,	LCD_B3		},
	{MFP_PIN_GPIO39,	LCD_B4		},
	{MFP_PIN_GPIO40,	LCD_B5		},
	{MFP_PIN_GPIO41,	LCD_LED_PWM	},
	{MFP_PIN_GPIO42,	LOCK_KEY_N	},
	{MFP_PIN_GPIO43,	CAM_ON_KEY_N	},
	{MFP_PIN_GPIO44,	MOT_INT		},
	{MFP_PIN_GPIO45,	GPS_RX		},
	{MFP_PIN_GPIO46,	GPS_TX		},
	{MFP_PIN_GPIO47,	TP_INT		},
	{MFP_PIN_GPIO48,	GPIO48_FREE	},
	{MFP_PIN_GPIO49,	HALL_INT	},
	{MFP_PIN_GPIO50,	PRO_DOUT	},
	{MFP_PIN_GPIO51,	PRO_LEDON	},
	{MFP_PIN_GPIO52,	PRO_ENB		},
	{MFP_PIN_GPIO53,	FFRXD		},
	{MFP_PIN_GPIO54,	FFTXD		},
	{MFP_PIN_GPIO55,	MMC1_CMD	},
	{MFP_PIN_GPIO56,	MMC1_CLK	},
	{MFP_PIN_GPIO57,	MMC1_DAT0	},
	{MFP_PIN_GPIO58,	MMC1_DAT1	},
	{MFP_PIN_GPIO59,	MMC1_DAT2	},
	{MFP_PIN_GPIO60,	MMC1_DAT3	},
	{MFP_PIN_GPIO61,	GPIO61_FREE	},
	{MFP_PIN_GPIO62,	GPIO62_FREE	},
	{MFP_PIN_GPIO63,	EXT_CHG_POOG_N	},
	{MFP_PIN_GPIO64,	EXT_CHG_EN_N	},
	{MFP_PIN_GPIO65,	CAM_DATA7	},
	{MFP_PIN_GPIO66,	CAM_DATA6	},
	{MFP_PIN_GPIO67,	CAM_DATA5	},
	{MFP_PIN_GPIO68,	CAM_DATA4	},
	{MFP_PIN_GPIO69,	CAM_DATA3	},
	{MFP_PIN_GPIO70,	CAM_DATA2	},
	{MFP_PIN_GPIO71,	CAM_DATA1	},
	{MFP_PIN_GPIO72,	CAM_DATA0	},
	{MFP_PIN_GPIO73,	CAM_HSYNC	},
	{MFP_PIN_GPIO74,	CAM_VSYNC	},
	{MFP_PIN_GPIO75,	CAM_MCLK	},
	{MFP_PIN_GPIO76,	CAM_PCLK	},
	{MFP_PIN_GPIO77,	CAM_PD_MAIN	},
	{MFP_PIN_GPIO78,	GPS_RESET	},
	{MFP_PIN_GPIO79,	PCM_CLK_GSSP1 },
	{MFP_PIN_GPIO80,	PCM_SYNC_GSSP1 },
	{MFP_PIN_GPIO81,	PCM_TXD_GSSP1 },
	{MFP_PIN_GPIO82,	PCM_RXD_GSSP1 },
	{MFP_PIN_GPIO83,	GPIO83_SSP3_RXD	},
	{MFP_PIN_GPIO84,	WLAN_WAKEUP_TAVOR	},
	{MFP_PIN_GPIO85,	I2S_BITCLK	},
	{MFP_PIN_GPIO86,	HP_AMP_SD	},
	{MFP_PIN_GPIO87,	I2S_SYNC	},
	{MFP_PIN_GPIO88,	I2S_SDATA_OUT	},
	{MFP_PIN_GPIO89,	I2C_SCL		},
	{MFP_PIN_GPIO90,	I2C_SDA		},
	{MFP_PIN_GPIO91,	BT_UART_RXD	},
	{MFP_PIN_GPIO92,	BT_UART_TXD	},
	{MFP_PIN_GPIO93,	BT_UART_CTS	},
	{MFP_PIN_GPIO94,	BT_UART_RTS	},
	{MFP_PIN_GPIO95,	TAVOR_ONE_WIRE	},
	{MFP_PIN_GPIO96,	WLAN_BT_RESET	},
	{MFP_PIN_GPIO97,	TAVOR_WAKEUP_WLAN	},
	{MFP_PIN_GPIO98,	SIM_REMOVE_N	},
	{MFP_PIN_GPIO99,	CLK_32K_OUT_EN	},
	{MFP_PIN_GPIO100,	GPIO100_FREE	},
	{MFP_PIN_GPIO101,	WLAN_DAT3	},
	{MFP_PIN_GPIO102,	WLAN_DAT2	},
	{MFP_PIN_GPIO103,	WLAN_DAT1	},
	{MFP_PIN_GPIO104,	WLAN_DAT0	},
	{MFP_PIN_GPIO105,	WLAN_CMD	},
	{MFP_PIN_GPIO106,	WLAN_CLK	},
	{MFP_PIN_GPIO159,	FLASH_TORCH_SEL	},
	{MFP_PIN_GPIO160,	LCD_SPI_CLK	},
	{MFP_PIN_GPIO161,	LCD_SPI_CS	},
	{MFP_PIN_GPIO162,	SD_LDO_CE	},
	{MFP_PIN_GPIO163,	LCD_SPI_DIN	},
	{MFP_PIN_GPIO164,	CAM_ID		},
	{MFP_PIN_GPIO165,	CAM_SHOT_KEY_N	},
	{MFP_PIN_GPIO166,	EXT_CHG_DET_N	},
	{MFP_PIN_GPIO167,	WLAN_PD_GH381	},
	{MFP_PIN_GPIO168,	TAVOR_WAKEUP_BT	},
	{MFP_PIN_GPIO169,	LCD_RESET_N	},
	{MFP_PIN_GPIO170,	WLAN_PD_PW886	},
	{MFP_PIN_GPIO171,	VCM_PD		},
	{MFP_PIN_GPIO172,	GPS_ON		},
	{MFP_PIN_GPIO173,	CAMERA_ON	},
	{MFP_PIN_GPIO174,	GPIO174_FREE	},
	{MFP_PIN_GPIO175,	BT_WAKEUP_TAVOR	},
	{MFP_PIN_GPIO176,	CAM_RESET_MAIN	},
	{-1,	0},
};

struct micco_check pxa3xx_spark_micco_setting[] = {
//	{MICCO_CHIP_ID,			0	},	// 0x00
///	{MICCO_EVENT_A,			0x00	},	// 0x01
///	{MICCO_EVENT_B,			0x00	},	// 0x02	
///	{MICCO_EVENT_C,			0x00	},	// 0x03
///	{MICCO_EVENT_D,			0x00	},	// 0x04
//	{MICCO_STATUS_A,		0	},	// 0x05
//	{MICCO_STATUS_B,		0	},	// 0x06
	{MICCO_IRQ_MASK_A,		0xF6	},	// 0x07
	{MICCO_IRQ_MASK_B,		0xF8	},	// 0x08
	{MICCO_IRQ_MASK_C,		0xFF	},	// 0x09
	{MICCO_IRQ_MASK_D,		0xE7	},	// 0x0A
	{MICCO_SYSCTRL_A,		0x63	},	// 0x0B
	{MICCO_SYSCTRL_B,		0x00	},	// 0x0C
	{MICCO_FAULT_LOG,		0x00	},	// 0x0D
	{MICCO_OVER1,			0x05	},	// 0x10
	{MICCO_APP_OVER2,		0x7F	},	// 0x11
//	{MICCO_APP_OVER3,		0x39	},	// 0x12
	{MICCO_LDO643,			0x7A	},	// 0x13
	{MICCO_LDO987,			0x73	},	// 0x14
//	{MICCO_LDO1110,			0x00	},	// 0x15
//	{MICCO_LDO12,			0x00	},	// 0x16
	{MICCO_LDO13,			0x0A	},	// 0x17
	{MICCO_BUCK_SLEEP,		0x05	},	// 0x18
	{MICCO_LDO_nSLEEP1,		0x00	},	// 0x19
	{MICCO_LDO_nSLEEP2,		0x00	},	// 0x1A
	{MICCO_LDO_SLEEP_BIT1,		0x04	},	// 0x1B
	{MICCO_LDO_SLEEP_BIT2,		0x00	},	// 0x1C
	{MICCO_BUCK_LDO8_EN,		0x05	},	// 0x1D
//	{MICCO_VCC2,			0x00	},	// 0x1F
//	{MICCO_VCC1,			0x00	},	// 0x20
//	{MICCO_COMM_OVER2,		0x00	},	// 0x21
//	{MICCO_COMM_OVER3,		0x00	},	// 0x22
//	{MICCO_BUCK1_DVC1,		0x00	},	// 0x23
//	{MICCO_BUCK1_DVC2,		0x0B	},	// 0x24
//	{MICCO_APPS_AVRC,		0x18	},	// 0x25
//	{MICCO_COMM_CDTV1,		0x08	},	// 0x26
//	{MICCO_COMM_CDTV2,		0x0F	},	// 0x27
//	{MICCO_COMM_CVRC,		0x18	},	// 0x28
//	{MICCO_SRAM_DVC1,		0x1B	},	// 0x29
//	{MICCO_SRAM_DVC2,		0x0F	},	// 0x2A
//	{MICCO_SRAM_SVRC,		0x18	},	// 0x2B
//	{MICCO_3DTV1,			0x02	},	// 0x2C
//	{MICCO_3DTV2,			0x03	},	// 0x2D
//	{MICCO_3DVRC,			0x02	},	// 0x2E
//	{MICCO_BULK_AUTO_SKIP,		0x07	},	// 0x2F
//	{MICCO_LDO1_MDTV1,		0x94	},	// 0x32
//	{MICCO_LDO1_MDTV2,		0x14	},	// 0x33
//	{MICCO_LDO1_MVRC,		0x02	},	// 0x34
	{MICCO_LED1_CONTROL,		0x00	},	// 0x35
	{MICCO_LED2_CONTROL,		0x00	},	// 0x36
	{MICCO_LEDPC_CONTROL1,		0x00	},	// 0x37
	{MICCO_LEDPC_CONTROL2,		0x00	},	// 0x38
	{MICCO_LEDPC_CONTROL3,		0x00	},	// 0x39
	{MICCO_LEDPC_CONTROL4,		0x00	},	// 0x3A
	{MICCO_LEDPC_CONTROL5,		0x00	},	// 0x3B
	{MICCO_WLED_CONTROL1,		0x00	},	// 0x3C
	{MICCO_WLED_CONTROL2,		0x00	},	// 0x3D
	{MICCO_FLASH_CONTROL1,		0x00	},	// 0x3E
	{MICCO_FLASH_CONTROL2,		0x00	},	// 0x3F
	{MICCO_VIBRA_CONTROL,		0x00	},	// 0x40
	{MICCO_MISC,			0x05	},	// 0x41
	{MICCO_CHARGE_CONTROL,		0x22	},	// 0x42
	{MICCO_CCTR_CONTROL,		0x00	},	// 0x43
	{MICCO_TCTR_CONTROL,		0x00	},	// 0x44
//	{MICCO_CHARGE_PULSE,		0x07	},	// 0x45
//	{MICCO_DAC_MSB,			0x01	},	// 0x46
//	{MICCO_DAC_LSB,			0x00	},	// 0x47
	{MICCO_UART_ADDR,		0x08	},	// 0x48
	{MICCO_UART_CONTROL,		0x05	},	// 0x49
	{MICCO_UART_TX_DATA,		0x00	},	// 0x4A
	{MICCO_UART_RX_DATA,		0x00	},	// 0x4B
//	{MICCO_ADC_MAN_CONTROL,		0	},	// 0x50
//	{MICCO_ADC_AUTO_CONTROL_1,	0	},	// 0x51
//	{MICCO_ADC_AUTO_CONTROL_2,	0	},	// 0x52
//	{MICCO_TSI_CONTROL_1,		0	},	// 0x53
//	{MICCO_TSI_CONTROL_2,		0	},	// 0x54
//	{MICCO_VBATMON,			0	},	// 0x55
//	{MICCO_TBATHIGHP,		0	},	// 0x56
//	{MICCO_TBATHIGHN,		0	},	// 0x57
//	{MICCO_TBATLOW,			0	},	// 0x58
//	{MICCO_AUTO4HIGH,		0	},	// 0x59
//	{MICCO_AUTO4LOW,		0	},	// 0x5A
//	{MICCO_AUTO5HIGH,		0	},	// 0x5B
//	{MICCO_AUTO5LOW,		0	},	// 0x5C
//	{MICCO_AUTO6HIGH,		0	},	// 0x5D
//	{MICCO_AUTO6LOW,		0	},	// 0x5E
//	{MICCO_MAN_RES_LSB,		0	},	// 0x5F
//	{MICCO_MAN_RES_MSB,		0	},	// 0x60
//	{MICCO_VBAT_RES,		0	},	// 0x61
//	{MICCO_VBATMIN_RES,		0	},	// 0x62
//	{MICCO_ICHMAX_RES,		0	},	// 0x63
//	{MICCO_ICHMIN_RES,		0	},	// 0x64
//	{MICCO_ICHAVERAGE_RES,		0	},	// 0x65
//	{MICCO_VCHMAX_RES,		0	},	// 0x66
//	{MICCO_VCHMIN_RES,		0	},	// 0x67
//	{MICCO_TBAT_RES,		0	},	// 0x68
//	{MICCO_AUTO4_RES,		0	},	// 0x69
//	{MICCO_AUTO5_RES,		0	},	// 0x6A
//	{MICCO_AUTO6_RES,		0	},	// 0x6B
//	{MICCO_TSI_X_MSB,		0	},	// 0x6C
//	{MICCO_TSI_Y_MSB,		0	},	// 0x6D
//	{MICCO_TSI_XY_MSB,		0	},	// 0x6E
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_MONO,			0x00	},	// 0x70
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_BEAR,			0x00	},	// 0x71
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_LINE_OUT,		0x00	},	// 0x72
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH1,		0x00	},	// 0x73
	{MICCO_AUDIO_REG_BASE + MICCO_MUX_STEREO_CH2,		0x00	},	// 0x74
	{MICCO_AUDIO_REG_BASE + MICCO_AUDIO_LINE_AMP,		0x25	},	// 0x75
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH1,	0x3F	},	// 0x76
	{MICCO_AUDIO_REG_BASE + MICCO_STEREO_AMPLITUDE_CH2,	0x3F	},	// 0x77
	{MICCO_AUDIO_REG_BASE + MICCO_HIFI_DAC_CONTROL,		0x00	},	// 0x78
	{MICCO_AUDIO_REG_BASE + MICCO_MONO_VOL,			0xBF	},	// 0x79
	{MICCO_AUDIO_REG_BASE + MICCO_BEAR_VOL,			0xBF	},	// 0x7A
///	{MICCO_AUDIO_REG_BASE + MICCO_I2S_CONTROL,		0x40	},	// 0x7B
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA,			0x00	},	// 0x7C
	{MICCO_AUDIO_REG_BASE + MICCO_MIC_PGA,			0x00	},	// 0x7D
	{MICCO_AUDIO_REG_BASE + MICCO_TX_PGA_MUX,		0x00	},	// 0x7E
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_ADC_CONTROL,	0x02	},	// 0x7F
	{MICCO_AUDIO_REG_BASE + MICCO_VCODEC_VDAC_CONTROL,	0x40	},	// 0x80
	{MICCO_AUDIO_REG_BASE + MICCO_SIDETONE,			0x00	},	// 0x81
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX1_2,		0x00	},	// 0x82
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_AUX3,			0x00	},	// 0x83
	{MICCO_AUDIO_REG_BASE + MICCO_PGA_DACS,			0x00	},	// 0x84
	{MICCO_AUDIO_REG_BASE + MICCO_SOFT_START_RAMP,		0xFF	},	// 0x85
	{MICCO_AUDIO_REG_BASE + MICCO_REM_IN_POLLING_TIME,	0x00	},	// 0x86
	{0xFF,	0},
};

static int __init pm_check_spark_init(void)
{
	if( machine_is_spark() ){
		register_pxa3xx_mfp_setting(&pxa3xx_spark_mfp_setting);
		register_pxa3xx_micco_setting(&pxa3xx_spark_micco_setting);
	}
	return 0;
}

static void __exit pm_check_spark_exit(void)
{
}

module_init(pm_check_spark_init);
module_exit(pm_check_spark_exit);
