/*
 * PXA3xx DVFM Driver
 *
 * Copyright (C) 2007 Marvell Corporation
 * Haojian Zhuang <haojian.zhuang@marvell.com>
 *
 * This software program is licensed subject to the GNU General Public License
 * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html

 * (C) Copyright 2007 Marvell International Ltd.
 * All Rights Reserved
 */

#undef DEBUG

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sysdev.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/workqueue.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/hardware.h>
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#include <asm/io.h>
#include <mach/pxa_ispt.h>
#include <mach/mspm_prof.h>
#include <mach/micco.h>
#include <mach/pxa930_acipc.h>
#include <mach/hardware.h>
#ifdef CONFIG_PXA_MIPSRAM
#include <mach/pxa_mips_ram.h>
#endif
#include "devices.h"

/* Enter D2 before exiting D0CS */
#define DVFM_LP_SAFE
int ForceOP = 0;
int ForcedOPIndex = 0;
int ForceC0 = 0;
int ForceVCTCXO_EN = 0;
int ForceD2 = 0;
int EnableD2VoltageChange = 1;
unsigned int D2voltageLevelValue = 0x0D; //setting the default voltage level to 1.05V
extern struct info_head dvfm_trace_list;
#ifdef DVFM_LP_SAFE
static struct work_struct postpone_work;
#endif

/* Counter Structure for Debugging ENTER/EXIT D2/CGM/D0CS */
pxa3xx_DVFM_LPM_Global_Count DVFMLPMGlobalCount={0,0,0,0};
/*Offset for "zero" the data */
pxa3xx_DVFM_LPM_Global_Count DVFMLPMGlobalCountOffset={0,0,0,0}; 

struct pxa3xx_dvfm_info {
	/* flags */
	uint32_t 	       flags;

	/* CPU ID */
	uint32_t	       cpuid;

	/* LCD clock */
	struct clk		*lcd_clk;

	/* NAND clock */
	struct clk		*nand_clk;
	/* SMC clock */
	struct clk		*smc_clk;

	/* HSIO2 clock */
	struct clk		*hsio2_clk;

	/* clock manager register base */
	unsigned char __iomem	*clkmgr_base;

	/* service power management unit */
	unsigned char __iomem	*spmu_base;

	/* slave power management unit */
	unsigned char __iomem	*bpmu_base;

	/* dynamic memory controller register base */
	unsigned char __iomem	*dmc_base;

	/* static memory controller register base */
	unsigned char __iomem	*smc_base;
};

unsigned int min_safe_frequency = 624;

static void init_dvfm_block_frequency()
{
	/*
	 * PXA935H: in order to block all
	 * lower frequencies than 728MHz
	 * PXA935/PXA950/PXA968 the default
	 * parameter value is suited for both
	 * usage model.
	 */
	if (cpu_is_pxa935h()) {
		min_safe_frequency = 728;
	}
}

struct info_head pxa3xx_dvfm_op_list = {
	.list = LIST_HEAD_INIT(pxa3xx_dvfm_op_list.list),
	.lock = RW_LOCK_UNLOCKED,
};

#ifdef CONFIG_PXA3xx_DVFM_STATS

static unsigned int switch_lowpower_before, switch_lowpower_after;

static int pxa3xx_stats_notifier_freq(struct notifier_block *nb,
				unsigned long val, void *data);
static struct notifier_block notifier_freq_block = {
	.notifier_call = pxa3xx_stats_notifier_freq,
};
#endif

/* the operating point preferred by policy maker or user */
static int preferred_op;

extern unsigned int cur_op;		/* current operating point */
extern unsigned int def_op;		/* default operating point */

#ifdef CONFIG_PXA_MIPSRAM
int update_op_mips_ram(u32 old_pp,u32 new_pp);
#endif
#if defined EARLY_D0CS_WAKE
static unsigned int pxa3xx_ticks_to_msec(unsigned int ticks);
#endif
extern int enter_d0cs_a(volatile u32 *, volatile u32 *);
extern int exit_d0cs_a(volatile u32 *, volatile u32 *);
extern int md2fvinfo(struct pxa3xx_fv_info *, struct dvfm_md_opt *);
extern void set_idle_op(int, int);
extern u32 set_DDR_avail_flag(void);
extern u32 clear_DDR_avail_flag(void);
extern void acipc_set_constraint_no_op_change(void);
extern void acipc_disable_d0cs_no_change(void);
extern void acipc_disable_d0cs(void);
extern u32 get_ddr_avail_state(void);


#ifdef CONFIG_FB_PXA
extern void pxafb_set_pcd(void);
extern void pxafb_set_pcd2(unsigned int);
extern int pxafb_get_pcd2(unsigned int);
#else
static void __attribute__ ((unused)) pxafb_set_pcd(void) {}
static void pxafb_set_pcd2(unsigned int) {}
static int pxafb_get_pcd2(unsigned int) { return 0; }
#endif

static int dvfm_dev_id;
#define LPJ_D0CS (293888 * 100 / HZ)
#define LPJ_104M (517120 * 100 / HZ)
#define LPJ_156M (778128 * 100 / HZ)
#define LPJ_208M (1036288 * 100 / HZ)
#define LPJ_416M (2076672 * 100 / HZ)
#define LPJ_624M (3112960 * 100 / HZ)
#define LPJ_728M (3640000 * 100 / HZ)
#define LPJ_806M (4020906 * 100 / HZ)

static int d0cs_lpj = LPJ_D0CS;

//#define EARLY_D0CS_WAKE
#if defined(EARLY_D0CS_WAKE)
static int comm_wakeup_time, timer_expiry_time;
extern unsigned int AppsSyncEnabled;

#define D0CS_TIMER_THRESHOLD 40
#endif

static int boot_core_freq = 0;

/* define the operating point of S0D0 and S0D0CS mode */
static struct dvfm_md_opt pxa300_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 0,
		.dmcfs = 30, /* will be 60MHZ for PXA310 A2 and PXA935/PXA940 */
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS,
		.name = "D0CS",
	},
	/* 104MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 8,
		.xn = 1,
		.smcfs = 78,
		.sflfs = 104,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 260,
		/* Actually it's 19.5, not 19 */
		.df_clk = 19,
		.empi_clk = 19,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_104M,
		.name = "104M",
	},
	/* 208MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M,
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1375,
		.vcc_sram = 1400,
		.xl = 24,
		.xn = 2,
		.smcfs = 208,
		.sflfs = 312,
		.hss = 208,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
};

/* define the operating point of S0D0 and S0D0CS mode */
static struct dvfm_md_opt pxa310_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 60,
		.dmcfs = 30, /* will be 60MHZ for PXA310 A2 and PXA935/PXA940 */
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS,
		.name = "D0CS",
	},
	/* 104MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 8,
		.xn = 1,
		.smcfs = 78,
		.sflfs = 104,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 260,
		/* Actually it's 19.5, not 19 */
		.df_clk = 19,
		.empi_clk = 19,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_104M,
		.name = "104M",
	},
	/* 208MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M,
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 104,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1375,
		.vcc_sram = 1400,
		.xl = 24,
		.xn = 2,
		.smcfs = 208,
		.sflfs = 208,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
};

static struct dvfm_md_opt pxa320_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 0,
		.dmcfs = 30,
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS,
		.name = "D0CS",
	},
	/* 104MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 8,
		.xn = 1,
		.smcfs = 78,
		.sflfs = 104,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 260,
		/* Actually it's 19.5, not 19 */
		.df_clk = 19,
		.empi_clk = 19,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_104M,
		.name = "104M",
	},
	/* 208MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M,
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1375,
		.vcc_sram = 1400,
		.xl = 24,
		.xn = 2,
		.smcfs = 208,
		.sflfs = 312,
		.hss = 208,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
	/* 806MHz */
	{
		.vcc_core = 1400,
		.vcc_sram = 1400,
		.xl = 31,
		.xn = 2,
		.smcfs = 208,
		.sflfs = 312,
		.hss = 208,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_806M,
		.name = "806M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
};

static struct dvfm_md_opt pxa930_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 0,
		.dmcfs = 30,
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS, 
		.name = "D0CS",
	},
	/* 156MHz -- single PLL mode */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 12,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 208,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_156M, 
		.name = "156M",
	},
	/* 208MHz */
	{
		.vcc_core = 1000,
		.vcc_sram = 1100,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M, 
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M, 
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1375,
		.vcc_sram = 1400,
		.xl = 24,
		.xn = 2,
		.smcfs = 208,
		.sflfs = 312,
		.hss = 156,
		.vaufs = 0,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
	/* D2 mode */
	{
		.vcc_core = 1100,
		.vcc_sram = 1200,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
};

static struct dvfm_md_opt pxa935_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1250,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 60,
		.dmcfs = 30,
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS,
		.name = "D0CS",
	},
	/* 156MHz -- single PLL mode */
	{
		.vcc_core = 1250,
		.xl = 12,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 208,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_156M,
		.name = "156M",
	},
	/* 208MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M,
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 104,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1250,
		.xl = 24,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
	/* CG (clock gated) mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_CG,
		.flag = OP_FLAG_FACTORY,
		.name = "CG",
	},

};

static struct dvfm_md_opt pxa935h_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1250,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 60,
		.dmcfs = 30,
		.df_clk = 15,
		.empi_clk = 15,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_D0CS,
		.name = "D0CS",
	},
	/* 156MHz -- single PLL mode */
	{
		.vcc_core = 1250,
		.xl = 12,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 208,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_156M,
		.name = "156M",
	},
	/* 208MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_208M,
		.name = "208M",
	},
	/* 416MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 156,
		.vaufs = 104,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_416M,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1250,
		.xl = 24,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 260,
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_624M,
		.name = "624M",
	},
	/* 728MHz */
	{
		.vcc_core = 1350,
		.xl = 28,
		.xn = 2,
		.smcfs = 104,
		.sflfs = 208,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 260,  /* 312MHz will be achieved by GB_PLLR field */
		.df_clk = 52,
		.empi_clk = 52,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = LPJ_728M,
		.name = "728M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
	/* CG (clock gated) mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_CG,
		.flag = OP_FLAG_FACTORY,
		.name = "CG",
	},

};

static struct dvfm_md_opt pxa950_op_array[] = {
	/* 60MHz -- ring oscillator */
	{
		.vcc_core = 1250,
		.xl = 0,
		.xn = 0,
		.smcfs = 15,
		.sflfs = 60,
		.hss = 60,
		.vaufs = 60,
		.dmcfs = 30,
		.df_clk = 15,
		.empi_clk = 15,
		.dsi = 30,
		.power_mode = POWER_MODE_D0CS,
		.flag = OP_FLAG_FACTORY,
		.lpj = 293888*100/HZ,
		.name = "D0CS",
	},
	/* 156MHz -- single PLL mode */
	{
		.vcc_core = 1250,
		.xl = 12,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 208,
		.df_clk = 52,
		.empi_clk = 52,
		.dsi = 156,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 778128*100/HZ,
		.name = "156M",
	},
	/* 208MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 208,
		.df_clk = 52,
		.empi_clk = 52,
		.dsi = 156,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 1036288*100/HZ,
		.name = "208M",
	},
	/* 156MHz -- with HF DDR */
	{
		.vcc_core = 1250,
		.xl = 12,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 104,
		.vaufs = 78,
		.dmcfs = 400,
		.df_clk = 52,
		.empi_clk = 52,
		.dsi = 156,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 778128*100/HZ,
		.name = "156M_HF",
	},
	/* 208MHz -- with HF DDR */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 1,
		.smcfs = 104,
		.sflfs = 156,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 400,
		.df_clk = 52,
		.empi_clk = 52,
		.dsi = 312,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 1036288*100/HZ,
		.name = "208M_HF",
	},

	/* 416MHz */
	{
		.vcc_core = 1250,
		.xl = 16,
		.xn = 2,
		.smcfs = 156,
		.sflfs = 208,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 400,
		.df_clk = 78,
		.empi_clk = 52,
		.dsi = 312,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 2076672*100/HZ,
		.name = "416M",
	},
	/* 624MHz */
	{
		.vcc_core = 1250,
		.xl = 24,
		.xn = 2,
		.smcfs = 156,
		.sflfs = 312,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 400,
		.df_clk = 78,
		.empi_clk = 52,
		.dsi = 312,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 3112960*100/HZ,
		.name = "624M",
	},
	/* 832MHz */
	{
		.vcc_core = 1250,
		.xl = 32,
		.xn = 2,
		.smcfs = 156,
		.sflfs = 312,
		.hss = 208,
		.vaufs = 156,
		.dmcfs = 400,
		.df_clk = 78,
		.empi_clk = 52,
		.dsi = 312,
		.power_mode = POWER_MODE_D0,
		.flag = OP_FLAG_FACTORY,
		.lpj = 3112960*100/HZ,
		.name = "832M",
	},
#if 0
	/* D1 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D1,
		.flag = OP_FLAG_FACTORY,
		.name = "D1",
	},
#endif
	/* D2 mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_D2,
		.flag = OP_FLAG_FACTORY,
		.name = "D2",
	},
	/* CG (clock gated) mode */
	{
		.vcc_core = 1250,
		.power_mode = POWER_MODE_CG,
		.flag = OP_FLAG_FACTORY,
		.name = "CG",
	},

};


struct proc_op_array {
	unsigned int cpuid;
	char	*cpu_name;
	struct dvfm_md_opt *op_array;
	unsigned int nr_op;
};

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)
static struct proc_op_array proc_op_arrays[] = {
	{0x6880, "PXA300", ARRAY_AND_SIZE(pxa300_op_array)},
	{0x6890, "PXA310", ARRAY_AND_SIZE(pxa310_op_array)},
	{0x6820, "PXA320", ARRAY_AND_SIZE(pxa320_op_array)},
	{0x6830, "PXA930", ARRAY_AND_SIZE(pxa930_op_array)},
	{0x6930, "PXA935H", ARRAY_AND_SIZE(pxa935h_op_array)},
	{0x6930, "PXA935/PXA940", ARRAY_AND_SIZE(pxa935_op_array)},
	{0x6970, "PXA950", ARRAY_AND_SIZE(pxa950_op_array)},
};

extern void pxa_clkcfg_write(unsigned int);

static int prepare_dmc(void *driver_data, int flag);
static int polling_dmc(void *driver_data);

#ifdef CONFIG_ISPT
static int ispt_dvfm_op(int old, int new)
{
	return ispt_dvfm_msg(old, new);
}

static int ispt_block_dvfm(int enable, int dev_id)
{
	int ret;
	if (enable)
		ret = ispt_driver_msg(CT_P_DVFM_BLOCK_REQ, dev_id);
	else
		ret = ispt_driver_msg(CT_P_DVFM_BLOCK_REL, dev_id);
	return ret;
}


#else
static int ispt_dvfm_op(int old, int new) { return 0; }
static int ispt_block_dvfm(int enable, int dev_id) { return 0; }
#endif

unsigned int pxa3xx_clk_to_lpj(unsigned int clk)
{
	if (clk == 728000000)
		return LPJ_728M;
        if (clk == 624000000)
                return LPJ_624M;
        if (clk == 416000000)
                return LPJ_416M;
        if (clk == 208000000)
                return LPJ_208M;
        if (clk == 156000000)
                return LPJ_156M;
        if (clk == 104000000)
                return LPJ_104M;
	if (clk == 60000000) 
		return LPJ_D0CS;
	
	printk(KERN_CRIT "%s does not support clk (%d MHz)\n", 
		__FILE__, clk/1000000);

	return 0; 
}

/* #####################Debug Function######################## */
static int dump_op(void *driver_data, struct op_info *p, char *buf)
{
	int len, count, x;
	struct dvfm_md_opt *md = (struct dvfm_md_opt *)p->op;

	if (md == NULL)
		len = sprintf(buf, "Can't dump the op info\n");
	else {
		/* calculate how much bits is set in device word */
		x = p->device;
		for (count = 0; x; x = x & (x - 1), count++);
		len = sprintf(buf, "OP:%d name:%s [%s, %d]\n",
				p->index, md->name, (count)?"Disabled"
				:"Enabled", count);
		len += sprintf(buf + len, "vcore:%d vsram:%d xl:%d xn:%d "
				"smcfs:%d sflfs:%d hss:%d dmcfs:%d ",
				md->vcc_core, md->vcc_sram, md->xl, md->xn,
				md->smcfs, md->sflfs, md->hss, md->dmcfs);
		if (cpu_is_pxa310() || cpu_is_pxa935() || cpu_is_pxa950())
			len += sprintf(buf + len, "vaufs:%d ", md->vaufs);
		if (cpu_is_pxa950())
			len += sprintf(buf + len, "DSI:%d ", md->dsi);
		len += sprintf(buf + len, "df_clk:%d power_mode:%d flag:%d\n",
				md->df_clk, md->power_mode, md->flag);
	}
	return len;
}

static int dump_op_list(void *driver_data, struct info_head *op_table, int flag)
{
	struct op_info *p = NULL;
	struct dvfm_md_opt *q = NULL;
	struct list_head *list = NULL;
	struct pxa3xx_dvfm_info *info = driver_data;
	char buf[256];

	if (!op_table || list_empty(&op_table->list)) {
		printk(KERN_WARNING "op list is null\n");
		return -EINVAL;
	}
	memset(buf, 0, 256);
	list_for_each(list, &op_table->list) {
		p = list_entry(list, struct op_info, list);
		q = (struct dvfm_md_opt *)p->op;
		if (q->flag <= flag) {
			dump_op(info, p, buf);
			pr_debug("%s", buf);
		}
	}
	return 0;
}

/* ########################################################## */
static int freq2reg(struct pxa3xx_fv_info *fv_info, struct dvfm_md_opt *orig)
{
	int res = -EFAULT, tmp;

	if (orig && fv_info) {
		fv_info->vcc_core = orig->vcc_core;
		fv_info->vcc_sram = orig->vcc_sram;
		if (orig->power_mode == POWER_MODE_D0) {
			res = 0;
			fv_info->xl = orig->xl;
			fv_info->xn = orig->xn;
			fv_info->d0cs = 0;
			if (orig->smcfs == 78)
				fv_info->smcfs = 0;
			else if (orig->smcfs == 104)
				fv_info->smcfs = 2;
			else if (orig->smcfs == 156)
				fv_info->smcfs = 4;
			else if (orig->smcfs == 208)
				fv_info->smcfs = 5;
			else
				res = -EINVAL;
			if (orig->sflfs == 104)
				fv_info->sflfs = 0;
			else if (orig->sflfs == 156)
				fv_info->sflfs = 1;
			else if (orig->sflfs == 208)
				fv_info->sflfs = 2;
			else if (orig->sflfs == 312)
				fv_info->sflfs = 3;
			else
				res = -EINVAL;
			if (orig->hss == 104)
				fv_info->hss = 0;
			else if (orig->hss == 156)
				fv_info->hss = 1;
			else if (orig->hss == 208)
				fv_info->hss = 2;
			else
				res = -EINVAL;
			if (orig->vaufs == 78)
				fv_info->vaufs = 3;
			else if (orig->vaufs == 104)
				fv_info->vaufs = 0;
			else if (orig->vaufs == 156)
				fv_info->vaufs = 1;
			else if (orig->vaufs == 208)
				fv_info->vaufs = 2;
			else
				res = -EINVAL;
			if (orig->dmcfs == 26)
				fv_info->dmcfs = 0;
			else if (orig->dmcfs == 208)
				fv_info->dmcfs = 2;
			else if (cpu_is_pxa950()) {
				if (orig->dmcfs == 400)
				fv_info->dmcfs = 3;
			else
				res = -EINVAL;
			} else {
				if (orig->dmcfs == 260)
					fv_info->dmcfs = 3;
				else
					res = -EINVAL;
			}
		    if (cpu_is_pxa950()) {
				if (orig->dsi == 312)
					fv_info->dsi = 4;
				else if (orig->dsi == 208)
					fv_info->dsi = 3;
				else if (orig->dsi == 156)
					fv_info->dsi = 2;
				else if (orig->dsi == 104)
					fv_info->dsi = 1;
				else if (orig->dsi == 52)
					fv_info->dsi = 0;
				else if (orig->dsi == 624)
					fv_info->dsi = 6;
				else if (orig->dsi == 30)
					fv_info->dsi = 5;	// D0CS mode should use DDR PLL /2 setting.
				else
				{
					res = -EINVAL; 
				}
			}
			tmp = orig->smcfs / orig->df_clk;
			if (tmp == 2)
				fv_info->df_clk = 2;
			else if (tmp == 4)
				fv_info->df_clk = 3;
			fv_info->empi_clk = fv_info->df_clk;
		} else if (orig->power_mode == POWER_MODE_D0CS) {
			fv_info->d0cs = 1;
			res = 0;
		}
	}
	return res;
}

int md2fvinfo(struct pxa3xx_fv_info *fv_info, struct dvfm_md_opt *orig)
{
	return freq2reg(fv_info, orig);
}

static int reg2freq(void *driver_data, struct dvfm_md_opt *fv_info)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	int res = -EFAULT, tmp;
	uint32_t accr;

	if (fv_info) {
		res = 0;
		if (fv_info->power_mode == POWER_MODE_D0CS) {
			/* set S0D0CS operating pointer */
			fv_info->power_mode = POWER_MODE_D0CS;
			fv_info->xl = 0;
			fv_info->xn = 0;
			fv_info->smcfs = 15;
			fv_info->sflfs = 60;
			fv_info->hss = 60;
			/* PXA310 A2 or PXA935/PXA940 */
			accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
			if (accr & 0x80)
				fv_info->dmcfs = 60;
			else
				fv_info->dmcfs = 30;
			fv_info->df_clk = 15;
			fv_info->empi_clk = 15;
			if (cpu_is_pxa310() || cpu_is_pxa935() || cpu_is_pxa950())
				fv_info->vaufs = 60;
			else
				fv_info->vaufs = 0;
			if (cpu_is_pxa950())
				fv_info->dsi = 30;
		} else {
			/* set S0D0 operating pointer */
			fv_info->power_mode = POWER_MODE_D0;
			tmp = fv_info->smcfs;
			if (tmp == 0)
				fv_info->smcfs = 78;
			else if (tmp == 2)
				fv_info->smcfs = 104;
			else if (tmp == 4)
				fv_info->smcfs = 156;
			else if (tmp == 5)
				fv_info->smcfs = 208;
			else
				res = -EINVAL;
			tmp = fv_info->sflfs;
			if (tmp == 0)
				fv_info->sflfs = 104;
			else if (tmp == 1)
				fv_info->sflfs = 156;
			else if (tmp == 2)
				fv_info->sflfs = 208;
			else if (tmp == 3)
				fv_info->sflfs = 312;
			tmp = fv_info->hss;
			if (tmp == 0)
				fv_info->hss = 104;
			else if (tmp == 1)
				fv_info->hss = 156;
			else if (tmp == 2)
				fv_info->hss = 208;
			else
				res = -EINVAL;
			tmp = fv_info->dmcfs;
			if (tmp == 0)
				fv_info->dmcfs = 26;
			else if (tmp == 2)
				fv_info->dmcfs = 208;
			else if (tmp == 3) {
				if (cpu_is_pxa950())
				fv_info->dmcfs = 400;
				else
					fv_info->dmcfs = 260;
			}
			else
				res = -EINVAL;
			tmp = fv_info->df_clk;
			if (tmp == 1)
				fv_info->df_clk = fv_info->smcfs;
			else if (tmp == 2)
				fv_info->df_clk = fv_info->smcfs / 2;
			else if (tmp == 3)
				fv_info->df_clk = fv_info->smcfs / 4;
			fv_info->empi_clk = fv_info->df_clk;

			if (cpu_is_pxa310() || cpu_is_pxa935() || cpu_is_pxa950()) {
				tmp = fv_info->vaufs;
				if (tmp == 0)
					fv_info->vaufs = 104;
				else if (tmp == 1)
					fv_info->vaufs = 156;
				else if (tmp == 3)
					fv_info->vaufs = 78;
				else if (tmp == 2)
					fv_info->vaufs = 208;
				else
					res = -EINVAL;
			} else
				fv_info->vaufs = 0;
			if (cpu_is_pxa950()) {
				tmp = fv_info->dsi;
			if (tmp == 4)
					fv_info->dsi = 312;
				else if (tmp == 6)
					fv_info->dsi = 624;
				/*else
					res = -EINVAL; */
			}
		}
	}
	return res;
}

/* Get current setting, and record it in fv_info structure
 */
static int capture_op_info(void *driver_data, struct dvfm_md_opt *fv_info)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	int res = -EFAULT;
	uint32_t acsr, memclkcfg, accr1;

	if (fv_info) {
		memset(fv_info, 0, sizeof(struct dvfm_md_opt));
		acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
		fv_info->xl = (acsr >> ACCR_XL_OFFSET) & 0x3F;
		fv_info->xn = (acsr >> ACCR_XN_OFFSET) & 0x07;
		fv_info->smcfs = (acsr >> ACCR_SMCFS_OFFSET) & 0x07;
		fv_info->sflfs = (acsr >> ACCR_SFLFS_OFFSET) & 0x03;
		fv_info->hss = (acsr >> ACCR_HSS_OFFSET) & 0x03;
		fv_info->dmcfs = (acsr >> ACCR_DMCFS_OFFSET) & 0x03;
		fv_info->power_mode = (acsr >> ACCR_D0CS_OFFSET) & 0x01;
		memclkcfg = __raw_readl(info->smc_base + MEMCLKCFG_OFF);
		fv_info->df_clk = (memclkcfg >> MEMCLKCFG_DF_OFFSET) & 0x07;
		fv_info->empi_clk = (memclkcfg >> MEMCLKCFG_EMPI_OFFSET) & 0x07;
		if (cpu_is_pxa310() || cpu_is_pxa935() || cpu_is_pxa950())
			fv_info->vaufs = (acsr >> ACSR_VAUFS_OFFSET) & 0x03;
		else
			fv_info->vaufs = 0;
		if (cpu_is_pxa950()) {
			accr1 = __raw_readl(info->clkmgr_base + ACCR1_OFF);
			fv_info->dsi = (accr1 >> ACCR1_DSI_1_OFFSET) & 0x07;
		}
		/* Convert bits into frequency */
		res = reg2freq(info, fv_info);
		pxa3xx_pmic_get_voltage(VCC_CORE, &fv_info->vcc_core);
		if (cpu_is_pxa935() || cpu_is_pxa950()) {
			/* PXA935/PXA940 doesn't have VCC_SRAM */
			fv_info->vcc_sram = 0;
		} else {
			pxa3xx_pmic_get_voltage(VCC_SRAM, &fv_info->vcc_sram);
		}
		/* TODO: mix up the usage of struct dvfm_md_opt and struct pxa3xx_fv_info
		 * better to define reg2freq(struct dvfm_md_opt *md_info,
		 * struct pxa3xx_fv_info *fv_info)
		 */
	}
	return res;
}

/* return all op including user defined op, and boot op */
static int get_op_num(void *driver_data, struct info_head *op_table)
{
	struct list_head *entry = NULL;
	int num = 0;

	if (!op_table)
		goto out;
	read_lock(&op_table->lock);
	if (list_empty(&op_table->list)) {
		read_unlock(&op_table->lock);
		goto out;
	}
	list_for_each(entry, &op_table->list) {
		num++;
	}
	read_unlock(&op_table->lock);
out:
	return num;
}

/* return op name. */
static char *get_op_name(void *driver_data, struct op_info *p)
{
	struct dvfm_md_opt *q = NULL;
	if (p == NULL)
		return NULL;
	q = (struct dvfm_md_opt *)p->op;
	return q->name;
}

static int update_voltage(void *driver_data, struct dvfm_md_opt *old, struct dvfm_md_opt *new)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	/* we will disable power for mg1, for pv2 pm is enabled
	   only is the PowerDisabled flag is set (using proc) */
	if (PowerDisabled)
		return 0;

	if (!(info->flags & PXA3xx_USE_POWER_I2C)) {
		pxa3xx_pmic_set_voltage(VCC_CORE, new->vcc_core);
		pxa3xx_pmic_set_voltage(VCC_SRAM, new->vcc_sram);
	}
	return 0;
}

#if defined EARLY_D0CS_WAKE
static struct timer_list d0cs_prevent_timer;
static void pxa3xx_d0cs_prevent_cbk(unsigned long data)
{
	//preventing D0CS entry until the next comm wakeup.
	//printk(KERN_WARNING "preventing D0CS because of ap cp sync jiffies = %u\n", jiffies);
#if defined(CONFIG_PXA930_ACIPC)
	acipc_disable_d0cs();
#endif
	MIPS_RAM_ADD_PM_TRACE(EARLY_DDR_PREVENTING_D0CS);
	del_timer_sync(&d0cs_prevent_timer);
}
#endif

static void pxa3xx_enter_d0cs(void *driver_data)
{
	struct pxa3xx_dvfm_info *info = driver_data;

	unsigned int reg, spll = 0;
	uint32_t accr, acsr, mdrefr;

   acsr =  __raw_readl(info->clkmgr_base + ACSR_OFF);
   if (!(acsr & (1 << ACCR_D0CS_OFFSET)))   //making sure we are not in D0CS before we try to enter.
	{

     DVFMLPMGlobalCount.D0CS_Enter_count++;
#if defined EARLY_D0CS_WAKE
	if (AppsSyncEnabled)     //we will call the timer only if apps comm sync is enabled
	{
		timer_expiry_time = comm_wakeup_time - D0CS_TIMER_THRESHOLD;
		timer_expiry_time = msecs_to_jiffies(timer_expiry_time);	//setting the timer expiry time to be 30 mse before next comm wakeup
		//we will set a timer so we will exit D0CS TH msec before the next comm expected wakeup.
   		mod_timer(&d0cs_prevent_timer, jiffies + timer_expiry_time);
	}
#endif


	reg = (12 << ACCR_XL_OFFSET) | (1 << ACCR_XN_OFFSET);
	accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
	if (reg == (accr & (ACCR_XN_MASK | ACCR_XL_MASK))) {
		spll = 1;
	}

	prepare_dmc(info, DMEMC_D0CS_ENTER);
	clk_enable(info->hsio2_clk);
#if defined(CONFIG_PXA930_ACIPC)
	if (get_ddr_avail_state())
		printk("****EDDR - WARNING - GOING TO D0CS WITH DDR INDICATION TO COM SET - *****"); /*should not happen*/
#endif
	enter_d0cs_a((volatile u32 *)info->clkmgr_base, (volatile u32 *)info->dmc_base);
	clk_disable(info->hsio2_clk);
	polling_dmc(info);

	/* update to D0CS LPJ, it must be updated before udelay() */
	loops_per_jiffy = d0cs_lpj;
	if (cpu_is_pxa930())
		udelay(200);
	else
		udelay(100);

	/* disable PLL */

		accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
		if (spll) {
			/* core pll is already disabled.
			 * Only disable System PLL */
			accr |= (1 << ACCR_SPDIS_OFFSET);
		} else {
			/* Disable both System PLL and Core PLL */
			accr |= (1 << ACCR_XPDIS_OFFSET) | (1 << ACCR_SPDIS_OFFSET);
		}
		__raw_writel(accr, info->clkmgr_base + ACCR_OFF);


	mdrefr = __raw_readl(info->dmc_base + MDREFR_OFF);
	__raw_writel(mdrefr, info->dmc_base + MDREFR_OFF);
	}
	else
	{
	   printk(KERN_WARNING "Attempting to enter D0CS when we are already in D0CS\n");
	}
}

static void pxa3xx_exit_d0cs(void *driver_data)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	unsigned int spll = 0;
	uint32_t reg, accr, acsr, mdrefr;

	acsr =  __raw_readl(info->clkmgr_base + ACSR_OFF);
	if (acsr & (1 << ACCR_D0CS_OFFSET))   //making sure we are actually in D0CS before we try to exit,
	{
	DVFMLPMGlobalCount.D0CS_Exit_count++;
	reg = (12 << ACCR_XL_OFFSET) | (1 << ACCR_XN_OFFSET);
	accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
	if (reg == (accr & (ACCR_XN_MASK | ACCR_XL_MASK))) {
		spll = 1;
	}
	/* enable PLL */

		if (spll) {
			/* core PLL shound not be enabled.
			 * only enable System PLL */
			accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
			accr &= ~(1 << ACCR_SPDIS_OFFSET);
			__raw_writel(accr, info->clkmgr_base + ACCR_OFF);
			do {
				acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
			} while ((acsr & ACSR_SPLCK) == 0);
		} else {
			/* enable both System PLL and Core PLL */
			accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
			accr &= ~((1 << ACCR_XPDIS_OFFSET) |
					(1 << ACCR_SPDIS_OFFSET));
			__raw_writel(accr, info->clkmgr_base + ACCR_OFF);
			do {
				acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
			} while (((acsr & ACSR_XPLCK) == 0)
				|| ((acsr & ACSR_SPLCK) == 0));
		}

	/* This is a workaround for JIRA TPV2-2978 - fixed in B0
	This is due to a false indication of the system
	PLL lock indication (ACSR[SPLCK]) causing a crash in D0CS
	exit. the delay ensures that the PLL is stable */
	if (cpu_is_pxa950_a0())
		udelay(200);

	prepare_dmc(info, DMEMC_D0CS_EXIT);
	clk_enable(info->hsio2_clk);
	exit_d0cs_a((volatile u32 *)info->clkmgr_base, (volatile u32 *)info->dmc_base);
	clk_disable(info->hsio2_clk);
	polling_dmc(info);

	mdrefr = __raw_readl(info->dmc_base + MDREFR_OFF);
	__raw_writel(mdrefr, info->dmc_base + MDREFR_OFF);
#if defined(CONFIG_PXA930_ACIPC)
	set_DDR_avail_flag();
#endif

#if defined EARLY_D0CS_WAKE
	if (AppsSyncEnabled)
	{
		//we are checking the proximity to the next xomm wakeup to detrmine if this is a ap-cp-sync timer
		comm_wakeup_time = pxa3xx_ticks_to_msec(dvfm_get_next_comm_wakeup_time());

		//we should not disable the timer if we are not sure if this wake is because of the timer...
		if (comm_wakeup_time > D0CS_TIMER_THRESHOLD)    //if we are far enough from the comm wakeup we can disable the timer (it will be retriggered if we enter D0CS again/
		{
			del_timer_sync(&d0cs_prevent_timer);
		}
	}
#endif
	}
	else
	{
	   printk(KERN_WARNING "Attempting to exit D0CS when we are not in D0CS in the first place\n");
	}
}

/* Return 1 if Grayback PLL is on. */
static int check_grayback_pll(void *driver_data)
{
	struct pxa3xx_dvfm_info *info = driver_data;

	return (__raw_readl(info->clkmgr_base + OSCC_OFF) & (1 << 17));
}

static int set_grayback_pll(void *driver_data, int enable)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	uint32_t oscc, agenp;

	if (!cpu_is_pxa9xx()) {
		/* It's not PXA930/PXA935/PXA940*/
		return 0;
	}
	if (enable) {
		/*
		 * unmask enable/disable GBPLL commands initiated
		 * by AGENP[GBPLL_CTRL]
		 */
		oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		oscc &= ~OSCC_GPRM;
		__raw_writel(oscc, info->clkmgr_base + OSCC_OFF);
		do {
			oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		} while (oscc & OSCC_GPRM);

		/* turn on GB PLL by AGENP[GBPLL_CTRL] and AGENP[GBPLL_DATA] */
		agenp = __raw_readl(info->bpmu_base + AGENP_OFF);
		agenp |= (AGENP_GBPLL_CTRL | AGENP_GBPLL_DATA);
		__raw_writel(agenp, info->bpmu_base + AGENP_OFF);

		/* wait until GB PLL is ready for use as DMC clock source */
		do {
			oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		} while (!(oscc & OSCC_GPLS));
	} else {
		/* turn off Grayback PLL, only set AGENP[GBPLL_CTRL] */
		agenp = __raw_readl(info->bpmu_base + AGENP_OFF);
		agenp &= ~(AGENP_GBPLL_CTRL | AGENP_GBPLL_DATA);
		agenp |= AGENP_GBPLL_CTRL;
		__raw_writel(agenp, info->bpmu_base + AGENP_OFF);

		/* wait until GB PLL isn't ready to use */
		do {
			oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		} while (oscc & OSCC_GPLS);

		/*
		 * mask enable/disable GBPLL commands initiated
		 * by AGENP[GBPLL_CTRL]
		 */
		oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		oscc |= OSCC_GPRM;
		__raw_writel(oscc, info->clkmgr_base + OSCC_OFF);
		do {
			oscc = __raw_readl(info->clkmgr_base + OSCC_OFF);
		} while (!(oscc & OSCC_GPRM));
	}
	return 0;
}

/*
 * Return 2 if MTS should be changed to 2.
 * Return 1 if MTS should be changed to 1.
 * Return 0 if MTS won't be changed.
 * In this function, the maxium MTS is 2.
 */
static int check_mts(struct dvfm_md_opt *old, struct dvfm_md_opt *new)
{
	int ret = 0;
	if ((old->xn == 1) && (new->xn == 2))
		ret = 2;
	if ((old->xn == 2) && (new->xn == 1))
		ret = 1;
	return ret;
}

static int set_mts(void *driver_data, int mts)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	unsigned int ascr;

	ascr = __raw_readl(info->bpmu_base + ASCR_OFF);
	ascr &= ~(3 << ASCR_MTS_OFFSET);
	ascr |= (mts << ASCR_MTS_OFFSET);
	__raw_writel(ascr, info->bpmu_base + ASCR_OFF);

	/* wait MTS is set */
	do {
		ascr = __raw_readl(info->bpmu_base + ASCR_OFF);
	}while (((ascr >> ASCR_MTS_OFFSET) & 0x3)
		!= ((ascr >> ASCR_MTS_S_OFFSET) & 0x3));

	return 0;
}

static int prepare_dmc(void *driver_data, int flag)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	int data;
	uint32_t mdcnfg, ddr_hcal;

	if (flag == DMEMC_D0CS_ENTER || flag == DMEMC_FREQ_LOW) {
		/* we should not clear the phase detector if lowering freqeuncy in PV. */
		if (!((cpu_is_pxa935() || cpu_is_pxa950()) && (flag == DMEMC_FREQ_LOW))) {
			/* Clear HCAL_HCEN, enable phase detector */
			ddr_hcal = __raw_readl(info->dmc_base + DDR_HCAL_OFF);
			ddr_hcal &= ~(1 << HCAL_HCEN_OFFSET);
			__raw_writel(ddr_hcal, info->dmc_base + DDR_HCAL_OFF);
			do {
				ddr_hcal = __raw_readl(info->dmc_base + DDR_HCAL_OFF);
			} while ((ddr_hcal & (1 << HCAL_HCEN_OFFSET)) == 1);
		}

		/* Set MDCNFG_HWFREQ & MDCNFG_HWNOPHD */
		mdcnfg = __raw_readl(info->dmc_base + MDCNFG_OFF);
		data = (1 << MDCNFG_HWFREQ_OFFSET)
			| (1 << MDCNFG_HWNOPHD_OFFSET);
		mdcnfg |= data;
		__raw_writel(mdcnfg, info->dmc_base + MDCNFG_OFF);
		do {
			mdcnfg = __raw_readl(info->dmc_base + MDCNFG_OFF);
		} while ((mdcnfg & data) != data);


		return 0;
	}
	if (flag == DMEMC_D0CS_EXIT || flag == DMEMC_FREQ_HIGH) {
		/* we should not clear the phase detector if lowering freqeuncy in PV. */
		if (!((cpu_is_pxa935() || cpu_is_pxa950()) && (flag == DMEMC_FREQ_HIGH))) {
			/* Set HCAL_HCEN, enable phase detector */
			ddr_hcal = __raw_readl(info->dmc_base + DDR_HCAL_OFF);
			ddr_hcal |= (1 << HCAL_HCEN_OFFSET);
			__raw_writel(ddr_hcal, info->dmc_base + DDR_HCAL_OFF);
	   		do {
				ddr_hcal = __raw_readl(info->dmc_base + DDR_HCAL_OFF);
			} while ((ddr_hcal & (1 << HCAL_HCEN_OFFSET)) == 0);
		}

		/* Set MDCNFG_HWFREQ */
		mdcnfg = __raw_readl(info->dmc_base + MDCNFG_OFF);
		mdcnfg |= (1 << MDCNFG_HWFREQ_OFFSET);
		__raw_writel(mdcnfg, info->dmc_base + MDCNFG_OFF);
		do {
			mdcnfg = __raw_readl(info->dmc_base + MDCNFG_OFF);
		} while ((mdcnfg & (1 << MDCNFG_HWFREQ_OFFSET)) == 0);

		return 0;
	}

	return -EINVAL;
}

static int polling_dmc(void *driver_data)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	unsigned int mdcnfg;
	/* polling MDCNFG_HWFREQ cleared */
	do {
		mdcnfg = __raw_readl(info->dmc_base + MDCNFG_OFF);
	} while (mdcnfg & (1 << MDCNFG_HWFREQ_OFFSET));
	return 0;
}

static int set_dmc60(void *driver_data, int flag)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	uint32_t accr, reg;

	accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
	if (flag)
		accr |= 0x80;
	else
		accr &= ~0x80;
	__raw_writel(accr, info->clkmgr_base + ACCR_OFF);
	/* polling ACCR */
	do {
		reg = __raw_readl(info->clkmgr_base + ACCR_OFF);
	} while ((accr & 0x80) != (reg & 0x80));

	return 0;
}

/* set DF and EMPI divider */
/* TODO: why did not we see DF/EMPI clock as input here? If we want to set DFI_clock or
 * EMPI clock as other frequecy than 52, how can we do?
 */
static int set_df(void *driver_data, int smc)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	uint32_t memclkcfg;
	uint32_t df = 0, empi = 0;

	switch (smc) {
	case 208:
		/* divider -- 4 */
		df = empi = 3;
		break;
	case 104:
		/* divider -- 2 */
		df = empi = 2;
		break;
	case 78:
		/* divider -- 4 */
		df = empi = 3;
		break;
	case 15:
		/* divider -- 1 */
		df = empi = 1;
		break;
	}

	if (((info->cpuid > 0x6880) && (info->cpuid <= 0x6881))
		|| ((info->cpuid >= 0x6890) && (info->cpuid <= 0x6892)))
		/* It's PXA300 or PXA310 */
		empi = 3; /* set a fixed value */

	memclkcfg = __raw_readl(info->smc_base + MEMCLKCFG_OFF);
	memclkcfg &= ~((7 << MEMCLKCFG_DF_OFFSET) | (7 << MEMCLKCFG_EMPI_OFFSET));
	memclkcfg |= (df << MEMCLKCFG_DF_OFFSET) | (empi << MEMCLKCFG_EMPI_OFFSET);
	__raw_writel(memclkcfg, info->smc_base + MEMCLKCFG_OFF);
	memclkcfg = __raw_readl(info->smc_base + MEMCLKCFG_OFF);

	return 0;
}

/* TODO: sugguest to differentiate the operating point definition from
 * register info.And we can remove *reg_new here, and convert dvfm_md_opt to
 * it in the routine. That will make it much more clear.
 */
static int update_bus_freq(void *driver_data, struct dvfm_md_opt *old, struct dvfm_md_opt *new)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	struct pxa3xx_fv_info fv_info;
	uint32_t accr, acsr, accr1, mask = 0, mask2 = 0;
	unsigned int data = 0, data2 = 0;
	unsigned int pcd = 0;

	pcd = pxafb_get_pcd2(new->hss);
	if (old->hss < new->hss)
	      pxafb_set_pcd2(pcd);

	freq2reg(&fv_info, new);
	if (old->dmcfs < new->dmcfs)
		prepare_dmc(info, DMEMC_FREQ_HIGH);
	else if (old->dmcfs > new->dmcfs)
		prepare_dmc(info, DMEMC_FREQ_LOW);

	if (new->smcfs == 208 || new->smcfs == 78) {
		if (!cpu_is_pxa950())
			set_df(info, new->smcfs);
	}

	accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
	if (old->smcfs != new->smcfs) {
		data |= (fv_info.smcfs << ACCR_SMCFS_OFFSET);
		mask |= ACCR_SMCFS_MASK;
	}
	if (old->sflfs != new->sflfs) {
		data |= (fv_info.sflfs << ACCR_SFLFS_OFFSET);
		mask |= ACCR_SFLFS_MASK;
	}
	if (old->dmcfs != new->dmcfs) {
		data |= (fv_info.dmcfs << ACCR_DMCFS_OFFSET);
		mask |= ACCR_DMCFS_MASK;
	}
	if (old->hss != new->hss) {
		data |= (fv_info.hss << ACCR_HSS_OFFSET);
		mask |= ACCR_HSS_MASK;
	}
	if (cpu_is_pxa950()) {
	if (new->dmcfs == 400) {
		data &= ~(ACCR_DMCFS_312_OFFSET);
		mask |= ACCR_DMCFS_312_MASK;
		}
	}
	if (old->vaufs != new->vaufs) {
		data2 |= (fv_info.vaufs << ACCR_VAUFS_OFFSET);
		mask2 |= ACCR_VAUFS_MASK;
	}
	accr &= ~(mask | mask2);
	accr |= data | data2;
	__raw_writel(accr, info->clkmgr_base + ACCR_OFF);
	/* wait until ACSR is changed */
	do {
		accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
		acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
	} while ((((accr & mask) != data) || ((acsr & mask) != data))
		|| ((accr & ACCR_VAUFS_MASK) >> ACCR_VAUFS_OFFSET
		!= (acsr & ACSR_VAUFS_MASK) >> ACSR_VAUFS_OFFSET));
	/* set PCD just after HSS updated */
#if defined(CONFIG_FB_PXA)
	if (old->hss > new->hss)
	      pxafb_set_pcd2(pcd);
#endif

	if (old->dmcfs != new->dmcfs)
		polling_dmc(info);

	if ((new->smcfs == 104)  || (new->smcfs == 15)) {
		if (!cpu_is_pxa950())
			set_df(info, new->smcfs);
	}

	return 0;
}

static int set_freq(void *driver_data, struct dvfm_md_opt *old, struct dvfm_md_opt *new)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	int low_ddr;
	uint32_t accr, acsr;

	/* check whether new OP is single PLL mode */
	if (new->dmcfs == 208)
		low_ddr = 1;
	else
		low_ddr = 0;

	/* turn on Grayback PLL */
	if (!low_ddr & !check_grayback_pll(info))
		set_grayback_pll(info ,1);
	if (check_mts(old, new) == 2)
		set_mts(info, 2);

	accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
	accr &= ~(ACCR_XL_MASK | ACCR_XN_MASK | ACCR_XSPCLK_MASK);
	accr |= ((new->xl << ACCR_XL_OFFSET) | (new->xn << ACCR_XN_OFFSET)
			| (3 << ACCR_XSPCLK_OFFSET));
	__raw_writel(accr, info->clkmgr_base + ACCR_OFF);
	/* delay 2 cycles of 13MHz clock */
	udelay(1);

	if (check_mts(old, new) == 1)
		set_mts(info, 1);

	if ((new->xl == old->xl) && (new->xn != old->xn))
		/* set F and T bit */
		pxa_clkcfg_write(3);
	else
		/* set F bit */
		pxa_clkcfg_write(2);
	do {
		accr = __raw_readl(info->clkmgr_base + ACCR_OFF);
		acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
	} while ((accr & (ACCR_XL_MASK | ACCR_XN_MASK))
			!= (acsr & (ACCR_XL_MASK | ACCR_XN_MASK)));

	udelay(1);
	update_bus_freq(info, old, new);

	/* turn off Grayback PLL */
	if (low_ddr)
		set_grayback_pll(info, 0);
	return 0;
}


/* This function sets the GBPLL to 520MHz/624MHz. */
static set_grayback_pll_freq(int HighLowPLL)
{
	uint32_t gb_pll;

	gb_pll = GBPLLR;
	gb_pll &= ~GBPLLR_DMC_MASK;
	if (GBPLL_FREQ_HIGH == HighLowPLL) {
		gb_pll |= GBPLLR_312_DMC;
	} else { /*GBPLL_FREQ_LOW == HighLowPLL*/
		gb_pll |= GBPLLR_260_DMC;
	}
	GBPLLR = gb_pll;
}

static int update_freq(void *driver_data, struct dvfm_freqs *freqs)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	static struct dvfm_md_opt before_d0cs, pp1;
	struct dvfm_md_opt old, new;
	struct op_info *p = NULL;
	unsigned long flags;
	int found = 0, new_op = cur_op;
	static int found_pp1;

	/* we will disable power for mg1 and pv2. pm is enabled
	only is the PowerDisabled flag is cleared (using proc) */
	if (PowerDisabled)
		return 0;

	memset(&old, 0, sizeof(struct dvfm_md_opt));
	memset(&new, 0, sizeof(struct dvfm_md_opt));
	write_lock_irqsave(&pxa3xx_dvfm_op_list.lock, flags);
	if (!list_empty(&pxa3xx_dvfm_op_list.list)) {
		list_for_each_entry(p, &pxa3xx_dvfm_op_list.list, list) {
			if (!found_pp1) {
				if (p->index == 0x1) {
					found_pp1 = 1;
					memcpy(&pp1, (struct dvfm_md_opt *) p->op,
							sizeof(struct dvfm_md_opt));
				}
			}
			if (p->index == freqs->old) {
				found++;
				memcpy(&old, (struct dvfm_md_opt *)p->op,
					sizeof(struct dvfm_md_opt));
			}
			if (p->index == freqs->new) {
				found++;
				memcpy(&new, (struct dvfm_md_opt *)p->op,
					sizeof(struct dvfm_md_opt));
				new_op = p->index;
			}
			if (found == 2)
				break;
		}
	}
	write_unlock_irqrestore(&pxa3xx_dvfm_op_list.lock, flags);
	if (found != 2)
		return -EINVAL;

	if ((old.power_mode == POWER_MODE_D0)
		&& (new.power_mode == POWER_MODE_D0CS)) {
		memcpy(&before_d0cs, &old, sizeof(struct dvfm_md_opt));

		pxa3xx_enter_d0cs(info);
		update_voltage(info, &old, &new);
		set_df(info, new.smcfs); /* set_freq not called but need to update DF_CLK */
		cur_op = new_op;
		loops_per_jiffy = new.lpj;
		return 0;
	} else if ((old.power_mode == POWER_MODE_D0CS)
		&& (new.power_mode == POWER_MODE_D0)) {
		if (memcmp(&before_d0cs, &new, sizeof(struct dvfm_md_opt))) {
			/* exit d0cs and set new operating point */
			if ((before_d0cs.vcc_core < new.vcc_core) ||
				(before_d0cs.vcc_sram < new.vcc_sram)) {
				update_voltage(info, &old, &new);
			} else {
				update_voltage(info, &old, &before_d0cs);
			}
			pxa3xx_exit_d0cs(info);
			set_freq(info, &before_d0cs, &new);

			if ((before_d0cs.vcc_core > new.vcc_core) ||
				(before_d0cs.vcc_sram > new.vcc_sram))
				update_voltage(info, &before_d0cs, &new);
		} else {
			update_voltage(info, &old, &new);
			set_df(info, new.smcfs); /* set_freq not called but need to update DF_CLK */
			/* exit d0cs */
			pxa3xx_exit_d0cs(info);
		}
		cur_op = new_op;
		loops_per_jiffy = new.lpj;
		return 0;
	} else if ((old.power_mode == POWER_MODE_D0CS)
		&& (new.power_mode == POWER_MODE_D0CS)) {
		cur_op = new_op;
		return 0;
	}

	if (old.core < new.core) {
		update_voltage(info, &old, &new);
	}

if (cpu_is_pxa935h()) { /* This logic is for PXA935H (supports 728MHz) */
	if (28 == new.xl) {/*new == 728MHz */
		if ((12 != old.xl)) { /* Old != 156MHz */
			set_freq(info, &old, &pp1);
		}
		set_grayback_pll_freq(GBPLL_FREQ_HIGH);
		set_freq(info, &pp1, &new);
	} else if (28 == old.xl) {/*Old == PP5*/
		set_freq(info, &old, &pp1);
		set_grayback_pll_freq(GBPLL_FREQ_LOW);
		if ((12 != new.xl)) { /*new != PP1*/
			set_freq(info, &pp1, &new);
		}
	} else {
		set_freq(info, &old, &new);
	}
} else { /*not cpu_is_pxa935h()*/
	set_freq(info, &old, &new);
}

	if (old.core > new.core) {
		update_voltage(info, &old, &new);
	}
	cur_op = new_op;
	if ((new.power_mode == POWER_MODE_D0)
		|| (new.power_mode == POWER_MODE_D0CS))
		loops_per_jiffy = new.lpj;
	return 0;
}

/* function of entering low power mode */
extern void enter_lowpower_mode(int state);

static void do_freq_notify(void *driver_data, struct dvfm_freqs *freqs)
{
	struct pxa3xx_dvfm_info *info = driver_data;

	dvfm_notifier_frequency(freqs, DVFM_FREQ_PRECHANGE);
	update_freq(info, freqs);
	dvfm_notifier_frequency(freqs, DVFM_FREQ_POSTCHANGE);
#ifdef CONFIG_PXA_MIPSRAM
	update_op_mips_ram(freqs->old, freqs->new);
#endif
	ispt_dvfm_op(freqs->old, freqs->new);
}

static void do_lowpower_notify(void *driver_data, struct dvfm_freqs *freqs, unsigned int state)
{
	dvfm_notifier_frequency(freqs, DVFM_FREQ_PRECHANGE);

	/* we will disable power for mg1 and for pv2. pm is enabled
	only is the PowerDisabled flag is cleared (using proc)
	this is possible only for PV2*/
	if (!PowerDisabled)
	enter_lowpower_mode(state);
	dvfm_notifier_frequency(freqs, DVFM_FREQ_POSTCHANGE);

}

static int check_op(void *driver_data, struct dvfm_freqs *freqs, unsigned int new,
			unsigned int relation)
{
	struct op_info *p = NULL;
	struct dvfm_md_opt *q = NULL;
	int core, tmp_core = -1, found = 0;
	int first_op = 0;

	freqs->new = -1;
	if (!dvfm_find_op(new, &p)) {
		q = (struct dvfm_md_opt *)p->op;
		core = q->core;
	} else
		return -EINVAL;
	/*
	pr_debug("%s, old:%d, new:%d, core:%d\n", __FUNCTION__, freqs->old,
		new, core);
	*/
	read_lock(&pxa3xx_dvfm_op_list.lock);
	if (relation == RELATION_LOW) {
		/* Set the lowest frequency that is higher than specifed one */
		list_for_each_entry(p, &pxa3xx_dvfm_op_list.list, list) {
			q = (struct dvfm_md_opt *)p->op;
			if (core == 0) {
				/* Lowpower mode */
				if ((q->power_mode == POWER_MODE_D1)
					|| (q->power_mode == POWER_MODE_D2)
					|| (q->power_mode == POWER_MODE_CG)) {
					if (!p->device && (new == p->index)) {
						freqs->new = p->index;
						/*
						pr_debug("%s, found op%d\n",
							__FUNCTION__, p->index);
						*/
						break;
					}
				}
				continue;
			}

			if (!p->device && (q->core >= core)) {
				if (tmp_core == -1 || (tmp_core >= q->core)) {
					/*
					pr_debug("%s, found op%d, core:%d\n",
						__FUNCTION__, p->index,
						q->core);
					*/
					if (first_op == 0)
						first_op = p->index;
					freqs->new = p->index;
					tmp_core = q->core;
					found = 1;
				}
				if (found && (new == p->index))
					break;
			}
		}
		if (found && (first_op == 1) && (new != p->index))
			freqs->new = first_op;
	} else if (relation == RELATION_HIGH) {
		/* Set the highest frequency that is lower than specified one */
		list_for_each_entry(p, &pxa3xx_dvfm_op_list.list, list) {
			q = (struct dvfm_md_opt *)p->op;
			if (!p->device && (q->core <= core)) {
				if (tmp_core == -1 || tmp_core < q->core) {
					freqs->new = p->index;
					tmp_core = q->core;
				}
			}
		}
	} else if (relation == RELATION_STICK) {
		/* Set the specified frequency */
		list_for_each_entry(p, &pxa3xx_dvfm_op_list.list, list) {
			if (!p->device && (p->index == new)) {
				freqs->new = p->index;
				break;
			}
		}
	}
	read_unlock(&pxa3xx_dvfm_op_list.lock);
	if (freqs->new == -1) {
		/*
		pr_debug("%s, Can't find op\n", __FUNCTION__);
		pr_debug("%s, old:%d, new:%d, core:%d\n", __FUNCTION__,
			freqs->old, new, core);
		*/
		return -EINVAL;
	}
	return 0;
}

static int pxa3xx_get_freq(void *driver_data, struct op_info *p, struct op_freq *freq)
{
        struct dvfm_md_opt *q = (struct dvfm_md_opt *)p->op;
	freq->cpu_freq = q->core;
	return 0;
}

static int pxa3xx_check_active_op(void *driver_data, struct op_info *p)
{
	struct dvfm_md_opt *q = (struct dvfm_md_opt *)p->op;

	if ((!strcmp(q->name, "D0CS")) && (boot_core_freq >= q->core))
		return 0;

	if ((!strcmp(q->name, "156M")) && (boot_core_freq >= q->core))
		return 0;

	if ((!strcmp(q->name, "208M")) && (boot_core_freq >= q->core))
		return 0;

	if ((!strcmp(q->name, "416M")) && (boot_core_freq >= q->core))
		return 0;

	if ((!strcmp(q->name, "624M")) && (boot_core_freq >= q->core))
		return 0;

	if (!strcmp(q->name, "728M")  )
		return 0;

	return -EINVAL;
}

#if defined EARLY_D0CS_WAKE
struct dvfm_freqs freqs_tmp;
#endif
static int pxa3xx_set_op(void *driver_data, struct dvfm_freqs *freqs, unsigned int new,
			unsigned int relation)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	struct dvfm_md_opt *md = NULL, *old_md = NULL;
	struct op_info *p = NULL;
	unsigned long flags;
	int ret=-EFAULT;
	unsigned int ckena = 0;
#ifdef CONFIG_MTD_NAND_HSS_FIX
	extern atomic_t nand_in_cmd;
	int cnt;
#endif
#ifdef DVFM_LP_SAFE
	int out_d0cs = 0;
#endif

	local_fiq_disable();
	local_irq_save(flags);


/**************/
/* the following code will check if the op will like to go, is D0CS. if this is the case we verify that the PM will actually allow D0CS, then
   we try to clear the shared flag. in case the clear of the shared flag is not allowed we have to prevent D0CS. we do so by calling to
   acipc_set_constraint_no_op_change which will set the "comm awake" constraint and with no op change*/


	if(dvfm_find_op(freqs->old, &p))
		goto out;

	memcpy(&freqs->old_info, p, sizeof(struct op_info));

	if(dvfm_find_op(new, &p))
		goto out;

#if defined EARLY_D0CS_WAKE
	if (POWER_MODE_D0CS==((struct dvfm_md_opt *)(p->op))->power_mode)	/* verify that the requested PP is D0CS*/
	{
		memcpy(&freqs_tmp,freqs,sizeof(struct dvfm_freqs));
		if (dvfm_find_op(freqs_tmp.old, &p))
			goto out;

		memcpy(&freqs_tmp.old_info, p, sizeof(struct op_info));

		if (!check_op(info, &freqs_tmp, new, relation)) /* check if the PM will actually allow to go to D0CS*/
		{
			if (!dvfm_find_op(freqs_tmp.new, &p))
			{
				memcpy(&(freqs_tmp.new_info), p, sizeof(struct op_info));

				/* this check is to prevent an attempt to clear DDR_avail_flag if we already are in D0CS because if this attempt fails
				(meaning there is a pending DDR_REQ from the comm) we will set this flag although we are still in D0CS! this situation
				is no allowed since it means comm can from DDR although apps is in D0CS*/
				if (freqs->old_info.index == freqs_tmp.new_info.index)
					goto out;

				md = (struct dvfm_md_opt *)(freqs_tmp.new_info.op);
				if (md->power_mode == POWER_MODE_D0CS)
				{
#if defined(CONFIG_PXA930_ACIPC)
					if(clear_DDR_avail_flag()==1)
					{	/* if we are going to D0CS we need to clear the shared flag indicating that DDR available for Comm. if a DDR request is pending while clear_DDR_avail_flag is trying
						to clear the share flag, it will not change the shared flag state and return it. if the shared flag state was on we have to disable D0CS, so the decision made later (in check_op) will not decide to go to D0CS
						this should happen only if ddr_request arrive between disable interrupt in the beginning of pxa3xx_set_op, and the call to clear_DDR_avail_flag
						so it is fairy rare case.*/
						ret = -EBUSY;
						goto out;
						/* We comment out below lines in BPMD mechanism */
//						acipc_set_constraint_no_op_change();
					}
#endif
#if defined EARLY_D0CS_WAKE

#if defined(CONFIG_PXA930_ACIPC)
					else
#endif
					{
							/*	this is additional code to protect against D0CS entry near COMM	wakeups.
					  			This situation can cuase a delay in APPS DDR ack time if another interupt service routine is running
							 * 	Therfore, D0CS is not allowed within a defined thrrshold time of comm D2 wakeup	*/

						if (AppsSyncEnabled)
						{
							comm_wakeup_time = pxa3xx_ticks_to_msec(dvfm_get_next_comm_wakeup_time());

#if defined(CONFIG_PXA930_ACIPC)
							if (comm_wakeup_time < D0CS_TIMER_THRESHOLD)    //we should not be enetering d0cs.
							{
								set_DDR_avail_flag();   //if we are not going to D0CS we must set back the ddr_avail flag to 1.
								ret = -EBUSY;
								acipc_disable_d0cs_no_change(); //we will prevent D0CS entry because comm is soon to wake up.
								goto out;
								/* We comment out below lines in BPMD mechanism */
							}
#endif
						}
					}
#endif
				}
				md=NULL;
				p=NULL;
			}
			else
				goto out;
		}
		else
			goto out;
	}
#endif

	ret = check_op(info, freqs, new, relation);
	if (ret)
		goto out;

	if (!dvfm_find_op(freqs->new, &p)) {
		memcpy(&(freqs->new_info), p, sizeof(struct op_info));
		/* If find old op and new op is same, skip it.
		 * At here, ret should be zero.
		 */
		if (freqs->old_info.index == freqs->new_info.index)
			goto out;

#ifdef CONFIG_MTD_NAND_HSS_FIX 
		md = (struct dvfm_md_opt *)(freqs->new_info.op);
		old_md = (struct dvfm_md_opt *)(freqs->old_info.op);
		cnt = atomic_read(&nand_in_cmd);
		BUG_ON(cnt < 0 || cnt > 1);
		if (freqs->new < freqs->old && 
			atomic_read(&nand_in_cmd) && old_md->hss != md->hss) {
//			printk(KERN_WARNING "dec freq while nand is busying\n");
			ret=-EBUSY;
			goto out;
		}
#endif

#ifdef DVFM_LP_SAFE
		md = (struct dvfm_md_opt *)(freqs->new_info.op);
		old_md = (struct dvfm_md_opt *)(freqs->old_info.op);
		if ((old_md->power_mode == POWER_MODE_D0CS)
			&& ((md->power_mode == POWER_MODE_D1)
			|| (md->power_mode == POWER_MODE_D2))) {
			dvfm_disable_op_name("D0CS", dvfm_dev_id);
			out_d0cs = 1;
		}

		md = (struct dvfm_md_opt *)p->op;
		ckena = CKENA;
		CKENA |= (1 << CKEN_SMC) | (1 << CKEN_NAND);
		switch (md->power_mode) {
		case POWER_MODE_D0:
		case POWER_MODE_D0CS:
			if (ForceOP)  //this means that op is forced by user for debug
			{
					freqs->new = ForcedOPIndex;
					if (!dvfm_find_op(freqs->new, &p))
						memcpy(&(freqs->new_info), p, sizeof(struct op_info));
			}
			do_freq_notify(info, freqs);
			break;
		case POWER_MODE_D1:
		case POWER_MODE_D2:
		case POWER_MODE_CG:
			do_lowpower_notify(info, freqs, md->power_mode);
			break;
		}

		if (out_d0cs) {
			dvfm_enable_op_name_no_change("D0CS", dvfm_dev_id);
		}
#else
		md = (struct dvfm_md_opt *)p->op;
		ckena = CKENA;
		CKENA |= (1 << CKEN_SMC) | (1 << CKEN_NAND);
		switch (md->power_mode) {
		case POWER_MODE_D0:
		case POWER_MODE_D0CS:
			if (ForceOP)  //this means that op is forced by user for debug
			{
					freqs->new = ForcedOPIndex;
					if (!dvfm_find_op(freqs->new, &p))
						memcpy(&(freqs->new_info), p, sizeof(struct op_info));
			}
			do_freq_notify(info, freqs);
			break;
		case POWER_MODE_D1:
		case POWER_MODE_D2:
		case POWER_MODE_CG:
			do_lowpower_notify(info, freqs, md->power_mode);
			break;
		}
#endif
	}

	CKENA = ckena;
	local_irq_restore(flags);
	local_fiq_enable();
	return 0;
out:
	local_irq_restore(flags);
	local_fiq_enable();
	return ret;
}

/* Put the operation of switching back to D0CS in workqueue. */
static void d0cs_worker(struct work_struct *work)
{
	dvfm_enable_op_name("D0CS", dvfm_dev_id);
}

static int first_d2_entry = 1;

/*
This function has been entered as a temporary workaround for DDR access issue related to comm D2.
currenly it is not in use.
*/
static int __attribute__ ((unused)) pxa935_claim_d2_workaround(void)
{
	if (cpu_is_pxa935() || cpu_is_pxa950())
		dvfm_disable_op_name("156M", dvfm_dev_id);
	return 0;
}

int pxa935_rm_d2_workaround(void)
{
	if (first_d2_entry && (cpu_is_pxa935() || cpu_is_pxa950())) {
		dvfm_enable_op_name("156M", dvfm_dev_id);
		first_d2_entry = 0;
	}
	return 0;
}

static int pxa3xx_request_op(void *driver_data, int index)
{
	struct dvfm_freqs freqs;
	struct op_info *info = NULL;
	struct dvfm_md_opt *md = NULL;
	int relation, ret;
	ret = dvfm_find_op(index, &info);
	if (ret)
		goto out;
	freqs.old = cur_op;
	freqs.new = index;
	md = (struct dvfm_md_opt *)(info->op);
	switch (md->power_mode) {
	case POWER_MODE_D1:
	case POWER_MODE_D2:
	case POWER_MODE_CG:
		relation = RELATION_STICK;
		ret = pxa3xx_set_op(driver_data, &freqs, index, relation);
		break;
	default:
		relation = RELATION_LOW;
		/* only use non-low power mode as preferred op */
		ret = pxa3xx_set_op(driver_data, &freqs, index, relation);
		if (!ret)
			preferred_op = index;
		break;
	}
out:
	return ret;
}

static int is_d0cs(void *driver_data)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	unsigned int acsr;
	/* read ACSR */
	acsr = __raw_readl(info->clkmgr_base + ACSR_OFF);
	/* Check ring oscillator status */
	if (acsr & (1 << 26))
		return 1;
	return 0;
}

/* Produce a operating point table */
static int op_init(void *driver_data, struct info_head *op_table)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	unsigned long flags;
	int i, index;
	struct op_info *p = NULL, *q = NULL;
	struct dvfm_md_opt *md = NULL, *smd = NULL;
	struct proc_op_array *proc = NULL;

	write_lock_irqsave(&op_table->lock, flags);
	for (i = 0; i < ARRAY_SIZE(proc_op_arrays); i++){
		if (proc_op_arrays[i].cpuid == (info->cpuid & 0xfff0)) {
			if (0x6930 == proc_op_arrays[i].cpuid) {
				if (cpu_is_pxa935h()) {
					proc = &proc_op_arrays[i];
					break;
				} else {
					proc = &proc_op_arrays[i+1]; /*PXA935 is after PXA935H in table*/
					break;
				}
			}
			proc = &proc_op_arrays[i];
			break;
		}
	}
	if (proc == NULL) {
		printk(KERN_ERR "Failed to find op tables for cpu_id 0x%08x", info->cpuid);
		write_unlock_irqrestore(&op_table->lock, flags);
		return -EIO;
	} else {
		printk("initializing op table for %s\n", proc->cpu_name);
	}
	for (i = 0, index = 0; i < proc->nr_op; i++) {
		/* PXA310 A2 or PXA935/PXA940, dmcfs 60MHz in S0D0CS mode */
		if ((proc->op_array[i].power_mode == POWER_MODE_D0CS)
			&& (info->cpuid == 0x6892 || (info->cpuid & 0xFFF0) == 0x6930)) {
			set_dmc60(info, 1);
			proc->op_array[i].dmcfs = 60;
		}

		/* Set index of operating point used in idle */
		if (proc->op_array[i].power_mode != POWER_MODE_D0) {
			set_idle_op(index, proc->op_array[i].power_mode);
		}

		md = (struct dvfm_md_opt *)kzalloc(sizeof(struct dvfm_md_opt),
					GFP_KERNEL);
		p = (struct op_info *)kzalloc(sizeof(struct op_info),
					GFP_KERNEL);
		p->op = (void *)md;
		memcpy(p->op, &proc->op_array[i], sizeof(struct dvfm_md_opt));
		md->core = 13 * md->xl * md->xn;
		if (md->power_mode == POWER_MODE_D0CS)
			md->core = 60;
		p->index = index++;
		list_add_tail(&(p->list), &(op_table->list));
	}
	md = (struct dvfm_md_opt *)kzalloc(sizeof(struct dvfm_md_opt),
					GFP_KERNEL);
	p = (struct op_info *)kzalloc(sizeof(struct op_info), GFP_KERNEL);
	p->op = (void *)md;
	if (capture_op_info(info, md)) {
		printk(KERN_WARNING "Failed to get current op setting\n");
	} else {
		def_op = 0x5a5a;	/* magic number */
		list_for_each_entry(q, &(op_table->list), list) {
			smd = (struct dvfm_md_opt *)q->op;
			md->flag = smd->flag;
			md->lpj = smd->lpj;
			md->core = smd->core;
			if (memcmp(md, smd, sizeof(struct dvfm_md_opt)) == 0) {
				def_op = q->index;
				break;
			}
		}
	}
	if (is_d0cs(driver_data))
		md->core = 60;
	else
		md->core = 13 * md->xl * md->xn;
	md->lpj = loops_per_jiffy;
	md->flag = OP_FLAG_BOOT;
	sprintf(md->name, "BOOT OP");
	boot_core_freq = md->core;

#if 0   /* disable CUSTOM OP for borq platfrom */
	smd = (struct dvfm_md_opt *)kzalloc(sizeof(struct dvfm_md_opt),
					GFP_KERNEL);
	q = (struct op_info *)kzalloc(sizeof(struct op_info), GFP_KERNEL);
	memcpy(q, p, sizeof(struct op_info));
	memcpy(smd, md, sizeof(struct dvfm_md_opt));
	smd->core = md->core;
	smd->lpj = md->lpj;
	smd->flag = OP_FLAG_USER_DEFINED;
	sprintf(smd->name, "CUSTOM OP");
	q->op = (void *)smd;
	/* Add CUSTOM OP into op list */
	q->index = index++;
	list_add_tail(&q->list, &op_table->list);
#endif
	/* Add BOOT OP into op list */
	p->index = index++;
	preferred_op = p->index;
	list_add_tail(&p->list, &op_table->list);
	/* BOOT op */
	if (def_op == 0x5a5a) {
		cur_op = p->index;
		def_op = p->index;
	} else
		cur_op = def_op;
	pr_debug("%s, def_op:%d, cur_op:%d\n", __FUNCTION__, def_op, cur_op);

	op_nums = proc->nr_op + 1;	/* set the operating point number */

	pr_debug("Current Operating Point is %d\n", cur_op);
	dump_op_list(info, op_table, OP_FLAG_ALL);
	write_unlock_irqrestore(&op_table->lock, flags);

	return 0;
}

/*
 * The machine operation of dvfm_enable
 */
static int pxa3xx_enable_dvfm(void *driver_data, int dev_id)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	struct dvfm_md_opt *md = NULL;
	struct op_info *p = NULL;
	int i, num;
	num = get_op_num(info, &pxa3xx_dvfm_op_list);
	for (i = 0; i < num; i++) {
		if (!dvfm_find_op(i, &p)) {
			md = (struct dvfm_md_opt *)p->op;
/* 	For 930/935, min_safe_frequency=boot_core_freq=624
 *	For 935h, min_safe_frequency = 728, boot_core_freq = 624		
			if (md->core < boot_core_freq) */

			if (md->core < min_safe_frequency)
				dvfm_enable_op_name(md->name, dev_id);
		}
	}
	ispt_block_dvfm(0, dev_id);
	return 0;
}

/*
 * The mach operation of dvfm_disable
 */
static int pxa3xx_disable_dvfm(void *driver_data, int dev_id)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	struct dvfm_md_opt *md = NULL;
	struct op_info *p = NULL;
	int i, num;
	num = get_op_num(info, &pxa3xx_dvfm_op_list);
	for (i = 0; i < num; i++) {
		if (!dvfm_find_op(i, &p)) {
			md = (struct dvfm_md_opt *)p->op;
/* 	For 930/935, min_safe_frequency=boot_core_freq=624
 *	For 935h, min_safe_frequency = 728, boot_core_freq = 624		
			if (md->core < boot_core_freq) */

			if (md->core < min_safe_frequency)
				dvfm_disable_op_name(md->name, dev_id);
		}
	}
	ispt_block_dvfm(1, dev_id);
	return 0;
}

static int pxa3xx_enable_op(void *driver_data, int index, int relation)
{
	/*
	 * Restore preferred_op. Because this op is sugguested by policy maker
	 * or user.
	 */
	return pxa3xx_request_op(driver_data, preferred_op);
}

static int pxa3xx_disable_op(void *driver_data, int index, int relation)
{
	struct dvfm_freqs freqs;
	if (cur_op == index) {
		freqs.old = index;
		freqs.new = -1;
		dvfm_set_op(&freqs, freqs.old, relation);
	}
	return 0;
}

static int pxa3xx_volt_show(void *driver_data, char *buf)
{
	struct dvfm_md_opt new;
	int len = 0;

	memset(&new, 0, sizeof(struct dvfm_md_opt));
	pxa3xx_pmic_get_voltage(VCC_CORE, &new.vcc_core);
	pxa3xx_pmic_get_voltage(VCC_SRAM, &new.vcc_sram);
	len = sprintf(buf, "core voltage:%dmv, sram voltage:%dmv\n",
			new.vcc_core, new.vcc_sram);
	return len;
}

#ifdef CONFIG_PXA3xx_DVFM_STATS
/* Convert ticks from 32K timer to microseconds */
static unsigned int pxa3xx_ticks_to_usec(unsigned int ticks)
{
	return (ticks * 5 * 5 * 5 * 5 * 5 * 5) >> 9;
}

#if defined EARLY_D0CS_WAKE
static unsigned int pxa3xx_ticks_to_msec(unsigned int ticks)
{
	return (ticks * 5 * 5 * 5) >> 12;
}
#endif

static unsigned int pxa3xx_ticks_to_sec(unsigned int ticks)
{
	return (ticks >> 15);
}

static unsigned int pxa3xx_read_time(void)
{
	return OSCR4;
}

static int pxa3xx_core_freq_calc(int op_point)
{
	struct dvfm_md_opt *md = NULL;
	struct op_info *p = NULL;
	int freq = 0;

	if (!dvfm_find_op(op_point, &p)) {
		md = (struct dvfm_md_opt *)p->op;
		/* D0CS? */
		if ((md->xl == 0) &&
			(md->xn == 0))
			freq = 60;
		else
			freq = 13 * md->xl * md->xn;
	} else {
		/* unknown */
		freq = -1;
	}

	return freq;
}

static int pxa3xx_core_current_freq_get(void *driver_data)
{
	return pxa3xx_core_freq_calc(cur_op);
}

static int pxa3xx_core_freqs_table_get(void *driver_data,
					int *freq_table,
					int *num_pp,
					int table_sz)
{
	struct pxa3xx_dvfm_info *info = driver_data;
	int op_point;

	*num_pp = get_op_num(info, &pxa3xx_dvfm_op_list);
	if (table_sz < *num_pp)
		return -1;

	for (op_point = 0; op_point < *num_pp; op_point++)
		freq_table[op_point] = pxa3xx_core_freq_calc(op_point);

	return 0;
}

/* It's invoked by PM functions.
 * PM functions can store the accurate time of entering/exiting low power
 * mode.
 */
int calc_switchtime(unsigned int end, unsigned int start)
{
	switch_lowpower_before = end;
	switch_lowpower_after = start;
	return 0;
}

static int pxa3xx_stats_notifier_freq(struct notifier_block *nb,
				unsigned long val, void *data)
{
	struct dvfm_freqs *freqs = (struct dvfm_freqs *)data;
	struct op_info *info = &(freqs->new_info);
	struct dvfm_md_opt *md = NULL;
	unsigned int ticks;

	ticks = pxa3xx_read_time();
	md = (struct dvfm_md_opt *)(info->op);
	if (md->power_mode == POWER_MODE_D0 ||
		md->power_mode == POWER_MODE_D0CS) {
		switch (val) {
		case DVFM_FREQ_PRECHANGE:
			calc_switchtime_start(freqs->old, freqs->new, ticks);
			break;
		case DVFM_FREQ_POSTCHANGE:
			/* Calculate the costed time on switching frequency */
			calc_switchtime_end(freqs->old, freqs->new, ticks);
			dvfm_add_event(freqs->old, CPU_STATE_RUN,
					freqs->new, CPU_STATE_RUN);
			dvfm_add_timeslot(freqs->old, CPU_STATE_RUN);
			mspm_add_event(freqs->old, CPU_STATE_RUN);
			break;
		}
	} else if (md->power_mode == POWER_MODE_D1 ||
		md->power_mode == POWER_MODE_D2 ||
		md->power_mode == POWER_MODE_CG) {
		switch (val) {
		case DVFM_FREQ_PRECHANGE:
			calc_switchtime_start(freqs->old, freqs->new, ticks);
			/* Consider lowpower mode as idle mode */
			dvfm_add_event(freqs->old, CPU_STATE_RUN,
					freqs->new, CPU_STATE_IDLE);
			dvfm_add_timeslot(freqs->old, CPU_STATE_RUN);
			mspm_add_event(freqs->old, CPU_STATE_RUN);
			break;
		case DVFM_FREQ_POSTCHANGE:
			/* switch_lowpower_start before switch_lowpower_after
			 * is updated in calc_switchtime().
			 * It's invoked in pm function.
			 */
			calc_switchtime_end(freqs->old, freqs->new,
					switch_lowpower_before);
			calc_switchtime_start(freqs->new, freqs->old,
					switch_lowpower_after);
			calc_switchtime_end(freqs->new, freqs->old,
					ticks);
			dvfm_add_event(freqs->new, CPU_STATE_IDLE,
					freqs->old, CPU_STATE_RUN);
			dvfm_add_timeslot(freqs->new, CPU_STATE_IDLE);
			mspm_add_event(freqs->new, CPU_STATE_IDLE);
			break;
		}
	}
	return 0;
}
#else
#define pxa3xx_ticks_to_usec	NULL
#define pxa3xx_ticks_to_sec	NULL
#define pxa3xx_read_time	NULL
#endif

static struct dvfm_driver pxa3xx_driver = {
	.count	= get_op_num,
	.set	= pxa3xx_set_op,
	.dump	= dump_op,
	.name	= get_op_name,
	.request_set	= pxa3xx_request_op,
	.enable_dvfm	= pxa3xx_enable_dvfm,
	.disable_dvfm	= pxa3xx_disable_dvfm,
	.enable_op	= pxa3xx_enable_op,
	.disable_op	= pxa3xx_disable_op,
	.volt_show	= pxa3xx_volt_show,
	.ticks_to_usec	= pxa3xx_ticks_to_usec,
	.ticks_to_sec	= pxa3xx_ticks_to_sec,
	.read_time	= pxa3xx_read_time,
	.current_core_freq_get = pxa3xx_core_current_freq_get,
	.core_freqs_table_get = pxa3xx_core_freqs_table_get,
	.get_freq	= pxa3xx_get_freq,
	.check_active_op = pxa3xx_check_active_op,
};

#ifdef CONFIG_PM
static int pxa3xx_freq_suspend(struct platform_device *pdev, pm_message_t state)
{
	return 0;
}

static int pxa3xx_freq_resume(struct platform_device *pdev)
{
	return 0;
}
#else
#define pxa3xx_freq_suspend    NULL
#define pxa3xx_freq_resume     NULL
#endif

#ifdef CONFIG_PXA_MIPSRAM
int update_op_mips_ram(u32 old_pp,u32 new_pp)
{
	u32 data, d2_constraints, d0cs_constraints, ts;
	struct op_info *p = NULL;

	if ( old_pp != new_pp)
	{
		switch (new_pp) {
 		case 0:
			data = PP_D0CS_MIPS_RAM;
			break;
		case 1:
			data = PP_PP1_MIPS_RAM;
			break;
		case 2:
			data = PP_PP2_MIPS_RAM;
			break;
		case 3:
			data = PP_PP3_MIPS_RAM;
			break;
		case 4:
			data = PP_PP4_MIPS_RAM;
			break;
		default:
			data = INVALID_PP_NUMBER_MIPS_RAM;
			break;
		}
		MIPS_RAM_ADD_PP_CHANGE(data);
		dvfm_get_opinfo(0, &p);
		d0cs_constraints = p->device & 0x0000FFFF;
		ts = OSCR4;
		MIPS_RAM_ADD_32K_TIME_STAMP(ts);
		MIPS_RAM_ADD_TRACE(MIPSRAM_EVENT_RAW_DATA | d0cs_constraints);
		d0cs_constraints = (p->device & 0xFFFF0000) >> 16;
		MIPS_RAM_ADD_TRACE(MIPSRAM_EVENT_RAW_DATA | d0cs_constraints);
		dvfm_get_opinfo(5, &p);
		d2_constraints = p->device & 0x0000FFFF;
		MIPS_RAM_ADD_TRACE(MIPSRAM_EVENT_RAW_DATA | d2_constraints);
		d2_constraints = (p->device & 0xFFFF0000) >> 16;
		MIPS_RAM_ADD_TRACE(MIPSRAM_EVENT_RAW_DATA | d2_constraints);

	}
	return 0;

}

#endif

typedef struct
{
	unsigned long data_arg1;
	unsigned long data_arg2;
} data_struct;
static struct timer_list cpu_load_simulate_timer;
unsigned int load_cycle_time;
unsigned int first_call = 1;

static void __attribute__ ((unused)) cpu_simulate_load_cbk(unsigned long data)
{

	unsigned int timestamp1,timestamp2;
	unsigned int load_time = data;

	mod_timer(&cpu_load_simulate_timer, jiffies + load_cycle_time);
	timestamp1 = __raw_readl((void *)&(__REG(0x40A00040)));
    timestamp2 = timestamp1;

	while ((timestamp2 - timestamp1) < load_time)
	{
		 timestamp2 = __raw_readl((void *)&(__REG(0x40A00040)));
	}

}


static void pxa3xx_poweri2c_init(struct pxa3xx_dvfm_info *info)
{
	uint32_t avcr, svcr, cvcr, pcfr, pvcr, sdcr;

	if ((info->flags & PXA3xx_USE_POWER_I2C) &&
		(cpu_is_pxa935() || cpu_is_pxa950())) {
		/* set AVCR for PXA935/PXA940:
		 *	level 0: 1250mv, 0x15
		 *	level 1: 1250mv, 0x15
		 *	level 2: 1250mv, 0x15
		 *	level 3: 1250mv, 0x15
		 */
		avcr = __raw_readl(info->spmu_base + AVCR_OFF);
		/* Do not modify AVCR on PV2: OBM sets it so the chip works*/
		avcr &= 0xE0E0E0E0;
		if (cpu_is_pxa935h())
			avcr |= (0x19 << 24) | (0x15 << 16) | (0x15 << 8) | 0x11;
		else
			avcr |= (0x17 << 24) | (0x15 << 16) | (0x15 << 8) | 0x11;
		__raw_writel(avcr, info->spmu_base + AVCR_OFF);
		avcr = __raw_readl(info->spmu_base + AVCR_OFF);

		/* set delay */
		pcfr = __raw_readl(info->spmu_base + PCFR_OFF);



		pcfr &= 0xF00FFFFF;
		pcfr |= 0x00100000;
		/* Disable pullup/pulldown in PWR_SCL and PWR_SDA */
		pcfr |= 0x04;
		pcfr |= 0x100;

		pcfr |= 0x00028000; //this will enable the writes by setting WRM bits.

		__raw_writel(pcfr, info->spmu_base + PCFR_OFF);
		pcfr = __raw_readl(info->spmu_base + PCFR_OFF);



		/* enable FVE,PVE,TVE bit */
		__raw_writel(0xe0500034, info->spmu_base + PVCR_OFF);

		/* sdcr = __raw_readl(info->spmu_base + SDCR_OFF); */
		sdcr = 0x00000080;       /* turning on bit 7 (EN bit) and turing off bit 6 (VREN) and AVL. */
		__raw_writel(sdcr, info->spmu_base + SDCR_OFF);


	} else if (info->flags & PXA3xx_USE_POWER_I2C) {
		/* set AVCR for PXA300/PXA310/PXA320/PXA930
		 *	level 0: 1000mv, 0x0b
		 *	level 1: 1100mv, 0x0f
		 *	level 2: 1375mv, 0x1a
		 *	level 3: 1400mv, 0x1b
		 */
		avcr = __raw_readl(info->spmu_base + AVCR_OFF);
		avcr &= 0xE0E0E0E0;
		/*
		 * PXA930 B0(cpuid 0x6835) requires special setting
		 * PXA930 B1(cpuid 0x6837) requires 1100mv in 156MHz and 208MHz
		 * PXA930 B2(cpuid 0x6838) requires 1275mv in 624MHz
		 * PXA930 B1&B2 requires 1150mv in D2
		 */
		switch (info->cpuid) {
		case 0x6835:		/* PXA930 B0 */
			avcr |= (0x1b << 24) | (0x1a << 16) | (0x0f << 8) | 0xb;
			break;
		case 0x6837:		/* PXA930 B1 */
			avcr |= (0x11 << 24) | (0x1a << 16) | (0x0f << 8) | 0xf;
			break;
		case 0x6838:		/* PXA930 B2 */
			avcr |= (0x11 << 24) | (0x15 << 16) | (0x0f << 8) | 0xb;
			break;
		default:		/* PXA300/PXA310/PXA320 */
			avcr |= (0x0f << 24) | (0x1a << 16) | (0x0f << 8) | 0xb;
			break;
		}
		__raw_writel(avcr, info->spmu_base + AVCR_OFF);
		avcr = __raw_readl(info->spmu_base + AVCR_OFF);
		/* set SVCR:
		 * 	level 0: 1100mv, 0x0f
		 * 	level 1: 1200mv, 0x13
		 * 	level 2: 1400mv, 0x1b
		 * 	level 3: 1400mv, 0x1b
		 */
		svcr = __raw_readl(info->spmu_base + SVCR_OFF);
		svcr &= 0xE0E0E0E0;
		if (info->cpuid == 0x6835)
			svcr |= (0x1b << 24) | (0x1b << 16) | (0x13 << 8) | 0xf;
		else
			svcr |= (0x0f << 24) | (0x1b << 16) | (0x13 << 8) | 0xf;
		__raw_writel(svcr, info->spmu_base + SVCR_OFF);
		svcr = __raw_readl(info->spmu_base + SVCR_OFF);
		if (cpu_is_pxa930()) {
			/* set CVCR:
			 * 	level 0: 925mv, 0x08
			 * 	level 1: 1250mv, 0x15
			 * 	level 2: 1375mv, 0x1a
			 * 	level 3: 1400mv, 0x1b
			 */
			cvcr = __raw_readl(info->spmu_base + CVCR_OFF);
			cvcr &= 0xE0E0E0E0;
			switch (info->cpuid) {
			case 0x6835:		/* PXA930 B0 */
				cvcr |= (0x1b << 24) | (0x1a << 16)
					| (0x15 << 8) | 0x08;
				break;
			case 0x6837:		/* PXA930 B1 */
				cvcr |= (0x11 << 24) | (0x1a << 16)
					| (0x15 << 8) | 0x08;
				break;
			case 0x6838:		/* PXA930 B2 */
			default:
				cvcr |= (0x11 << 24) | (0x15 << 16)
					| (0x15 << 8) | 0x08;
				break;
			}
			__raw_writel(cvcr, info->spmu_base + CVCR_OFF);
			cvcr = __raw_readl(info->spmu_base + CVCR_OFF);
		}

		/* set delay */
		pcfr = __raw_readl(info->spmu_base + PCFR_OFF);
		pcfr &= 0x000FFFFF;
		pcfr |= 0x00100000;
		/* Disable pullup/pulldown in PWR_SCL and PWR_SDA */
		pcfr |= 0x04;
		pcfr |= 0x100;	//configuring bit 8 for I/O power saving
		__raw_writel(pcfr, info->spmu_base + PCFR_OFF);
		pcfr = __raw_readl(info->spmu_base + PCFR_OFF);

		/* enable FVE,PVE,TVE bit */
		__raw_writel(0xe0500034, info->spmu_base + PVCR_OFF);
	} else {
		/* disable FVE,PVE,TVE,FVC bit */
		pvcr = __raw_readl(info->spmu_base + PVCR_OFF);
		pvcr &= 0x0fffffff;
		__raw_writel(pvcr, info->spmu_base + PVCR_OFF);
	}
}

static void pxa3xx_df_init(struct pxa3xx_dvfm_info *info)
{
	 unsigned int memclkcfg;
//	 unsigned int accr;

	 /* initialize the DF-DIC to 2 -
	  * this is a constant value for TAVOR PV2 */
	 memclkcfg = __raw_readl(info->smc_base + MEMCLKCFG_OFF);
	 memclkcfg |= (2 << MEMCLKCFG_DF_OFFSET);
	 __raw_writel(memclkcfg, info->smc_base + MEMCLKCFG_OFF);

	 /*accr = __raw_readl(info->smc_base + ACCR_OFF);
	 accr &= ~(ACCR_DMCFS_312_OFFSET);
	 __raw_writel(accr, info->clkmgr_base + ACCR_OFF); */
}

static int pxa3xx_freq_probe(struct platform_device *pdev)
{
	struct resource *res;
	struct pxa3xx_freq_mach_info *pdata;
	struct pxa3xx_dvfm_info *info;

	/* initialize the information necessary to frequency/voltage change operation */
	pdata = pdev->dev.platform_data;
	info = kzalloc(sizeof(struct pxa3xx_dvfm_info), GFP_KERNEL);
	info->flags = pdata->flags;
	info->cpuid = read_cpuid(0) & 0xFFFF;

	if (!cpu_is_pxa950())
		info->lcd_clk = clk_get(&pxa_device_fb.dev, "LCDCLK");
	else
		info->lcd_clk = clk_get(&pxa950_device_fb.dev, "PXA950_LCDCLK");
	if (IS_ERR(info->lcd_clk)) goto err;

	info->hsio2_clk = clk_get(NULL, "HSIO2CLK");
	if (IS_ERR(info->hsio2_clk)) goto err;

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "clkmgr_regs");
	if (!res) goto err;
	info->clkmgr_base = ioremap(res->start, res->end - res->start + 1);

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "spmu_regs");
	if (!res) goto err;
	info->spmu_base = ioremap(res->start, res->end - res->start + 1);

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "bpmu_regs");
	if (!res) goto err;
	info->bpmu_base = ioremap(res->start, res->end - res->start + 1);

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "dmc_regs");
	if (!res) goto err;
	info->dmc_base = ioremap(res->start, res->end - res->start + 1);

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "smc_regs");
	if (!res) goto err;
	info->smc_base = ioremap(res->start, res->end - res->start + 1);

	pxa3xx_driver.priv = info;
	if (cpu_is_pxa950())
		pxa3xx_df_init(info);

	pxa3xx_poweri2c_init(info);
	op_init(info, &pxa3xx_dvfm_op_list);


#if defined EARLY_D0CS_WAKE
	init_timer(&d0cs_prevent_timer);
	d0cs_prevent_timer.function = pxa3xx_d0cs_prevent_cbk;
	d0cs_prevent_timer.data = 0;
#endif
	return dvfm_register_driver(&pxa3xx_driver, &pxa3xx_dvfm_op_list);
err:
	printk("pxa3xx_dvfm init failed\n");
	return -EIO;
}

static int pxa3xx_freq_remove(struct platform_device *pdev)
{
	kfree(pxa3xx_driver.priv);
	return dvfm_unregister_driver(&pxa3xx_driver);
}

static struct platform_driver pxa3xx_freq_driver = {
	.driver = {
		.name	= "pxa3xx-freq",
	},
	.probe		= pxa3xx_freq_probe,
	.remove		= pxa3xx_freq_remove,
#ifdef CONFIG_PM
	.suspend	= pxa3xx_freq_suspend,
	.resume		= pxa3xx_freq_resume,
#endif
};

static int __init pxa3xx_freq_init(void)
{
	int ret;

	ret = platform_driver_register(&pxa3xx_freq_driver);
	if (ret)
		goto out;
#ifdef CONFIG_PXA3xx_DVFM_STATS
	ret = dvfm_register_notifier(&notifier_freq_block,
				DVFM_FREQUENCY_NOTIFIER);
#endif
	ret = dvfm_register("DVFM", &dvfm_dev_id);
#ifdef DVFM_LP_SAFE
	INIT_WORK(&postpone_work, d0cs_worker);
#endif

	/* power is still disabled by default for PV2 & MG1 */
	if ((cpu_is_pxa950()) || (cpu_is_pxa968()))
		PowerDisabled = 1;
	init_dvfm_block_frequency();

out:
	return ret;
}

static void __exit pxa3xx_freq_exit(void)
{
#ifdef CONFIG_PXA3xx_DVFM_STATS
	dvfm_unregister_notifier(&notifier_freq_block,
				DVFM_FREQUENCY_NOTIFIER);
#endif
	dvfm_unregister("DVFM", &dvfm_dev_id);
	platform_driver_unregister(&pxa3xx_freq_driver);
}

module_init(pxa3xx_freq_init);
module_exit(pxa3xx_freq_exit);

