/*
 *  linux/arch/arm/mach-pxa/t8200.c
 *
 *  Support for  Huawei T8200 Product.
 *
 *  Copyright (C) 2010 Borqs Ltd.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 * 
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/input.h>
#include <linux/gpio_keys.h>
#include <linux/bq2407x_charger.h>
#include <linux/i2c-gpio.h>
#include <linux/console.h>
#include <linux/bootmem.h>
#include <linux/dpram.h>
#ifdef CONFIG_TOUCHSCREEN_ATMEL
#include <linux/atmel_i2c_rmi.h>
#elif CONFIG_TOUCHSCREEN_ATMELQT_OBP
#include <linux/maxtouch_obp_ts.h>
#endif
#ifdef CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539
#include <linux/synaptics_i2c_rmi_tm1539.h>
#endif
#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <asm/irq.h>
#include <asm/bootinfo.h>
#include <asm/hci_ps.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx_nand.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/pxa27x_keypad.h>
#include <mach/gpio.h>
#include <mach/micco.h>
#include <mach/t8200.h>
#include <mach/irda.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxafb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/pxa930_trkball.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa2xx_spi.h>
#include <mach/pxa-regs.h>
#include <mach/lis331.h>
#include <mach/akm8973.h>
#include <mach/ssp.h>
#include <mach/alsps.h>
#include <mach/micco_headset.h>
#include <mach/micco_vibrator.h>
#include <mach/gps.h>
#include <mach/i2c.h>
#include <mach/wlan.h>
#include <mach/pxa3xx-aec-fm2010.h>
#include <mach/innodev.h>
#include <mach/tavor_common.h>
#include <mach/leds-pxa.h>

#include "devices.h"
#include "generic.h"

#ifdef CONFIG_INPUT_KEYRESET 
#include <linux/keyreset.h>
#endif

#define MAX_SLOTS      3
static int g_touch_id;

static struct platform_mmc_slot mmc_slot[MAX_SLOTS];
extern int is_android(void);
extern int is_comm_v75(void);
static void cam_sync_from_gpio(void);
static void cam_sync_to_gpio(void);

extern void enable_oscc_tout_s0(void);
extern void disable_oscc_tout_s0(void);

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

/* MFP configurations */
static mfp_cfg_t mfp_cfg[] __initdata = {
	/* GPIO FREE */
	GPIO61_FREE,
	GPIO83_FREE,
//	GPIO161_FREE,
//	GPIO163_FREE,
	/*GPIO111_FREE,*/
	GPIO118_FREE,
	GPIO119_FREE,
	GPIO120_FREE,
	GPIO175_FREE,	
//	GPIO121_FREE,
//	GPIO124_FREE,
//	GPIO127_FREE,

	/* IPC */
	BP_WD_OUT,
	DPRAM_INT_N,
	BP_READY,
	AP_READY,
	FF_CTS,
	FF_RTS,
	FF_RXD,
	FF_TXD,
	BP_ON,
	BP_SHUTDOWN,
	AP_RESET_BP,

	ONKEY,
	/* Keypad */
	KP_MKIN1,
	KP_MKOUT1,
	KP_MKIN2,
	KP_MKOUT2,

	/* touch panel */
	TP_INT,
	TP_RESET,

	/* LCD */	
	LCD_VSYNC,
	LCD_HSYNC,
	LCD_PCLK,
	LCD_DE,
	LCD_R0, 
	LCD_R1, 
	LCD_R2, 
	LCD_R3, 
	LCD_R4, 
	LCD_R5, 
	LCD_G0, 
	LCD_G1, 
	LCD_G2, 
	LCD_G3, 
	LCD_G4, 
	LCD_G5, 
	LCD_B0, 
	LCD_B1, 
	LCD_B2, 
	LCD_B3, 
	LCD_B4, 
	LCD_B5, 
	LCD_SPI_CLK,
	LCD_SPI_CS,
	LCD_SPI_D0,
	LCD_RESET_N,

	/* i2c bus */
	I2C_SCL,
	I2C_SDA,

	/* Misc */

	/* GPS */
	STRXD,
	STTXD,
	AGPS_RESET,
	AGPS_PON,
	AGPS_PWR_ON,
    GPS_UART_SWITCH,

	/* Audio */
	AP_PCM_CLK, 
	AP_PCM_CLK_OFF,
	AP_PCM_FRM, 
	AP_PCM_FRM_OFF, 
	AP_PCM_TXD, 
	AP_PCM_TXD_OFF, 
	AP_PCM_RXD, 
	AP_PCM_RXD_OFF, 
	HIFI_CLK,
	HIFI_FRM,
	HIFI_TXD,
	AUDIO_PA_EN,
    	HEADPHONE_EN,

	/* camera */	
	CAM_DATA7, 
	CAM_DATA6, 
	CAM_DATA5, 
	CAM_DATA4, 
	CAM_DATA3, 
	CAM_DATA2, 
	CAM_DATA1, 
	CAM_DATA0, 
	CAM_HSYNC, 
	CAM_VSYNC, 
	CAM_MCLK, 
	CAM_PCLK, 
	CAM_REAR_PDN,
	CAM_VGA_PDN,
	CAM_MCLK_SEL,
	CAM_REAR_RESET_N,
	CAM_PWR_EN,

	/*Pin for compass*/
	CSEN_RST_N,
	CSEN_DRDY,	
	/* PMIC */
	PMIC_INT_MICCO,

	/* BT/WIFI */
	WLAN_BT_RESET,
	AP_WAKEUP_BT,
	BT_RXD, 
	BT_TXD, 
	BT_CTS, 
	BT_RTS, 
	AP_WAKEUP_WLAN,
	WLAN_WAKEUP_AP,
	WLAN_DAT3,
	WLAN_DAT2,
	WLAN_DAT1,
	WLAN_DAT0,
	WLAN_CMD,
	WLAN_CLK,
	BT_WAKEUP_AP,
	WIFI_PMU_PON,
	WLAN_PON,

	/* MMC */
	MMC1_CD_N,
	MMC1_CMD, 
	MMC1_CLK, 
	MMC1_DAT0, 
	MMC1_DAT1, 
	MMC1_DAT2, 
	MMC1_DAT3,

	/* P/L sensor */
	PSEN_DOUT,
	PSEN_LEDON,
	PSEN_PD,

	/* G-Sensor */
	GSEN_INT,


	/* charger ic */
	CHARGE_EN2,
	CHARGE_EN1,
	CHARGE_PG_N,
	CHARGE_CE_N,
	CHARGE_STAT,

	/* CMMB */
	CMMB_IRQ,
	CMMB_RST_N,
	CMMB_SPI_CLK,
	CMMB_SPI_SFRM,
	CMMB_SPI_TXD,
	CMMB_SPI_RXD,
	CMMB_PWR_EN,
};

enum mach_version get_mach_version(void)
{
	extern unsigned int system_rev;

	printk("Hardware ID: 0x%x\n", system_rev);

	switch (system_rev) {
	case MACH_VERSION_P10:
	case MACH_VERSION_P20:
	case MACH_VERSION_P21:
		break;
	default:
		printk(KERN_ERR "Unknown Version, Change from 0x%x to 0x%x!\n",
				system_rev, MACH_VERSION_P10);
		system_rev = MACH_VERSION_P10;
	}

	return system_rev;
}

static struct micco_vibrator_platform_data vibrator_data = {
	.vib_strength = (0x33 << 1),
};

static struct platform_device micco_vibrator_device = {
	.name 		= "micco-vibrator",
	.id 		= -1,
	.dev		= {
		.platform_data = &vibrator_data,
	},
};

#if defined(CONFIG_TOUCHSCREEN_ATMEL)

static int touch_power(int on)
{
	pr_info("touch_power in on = %d\n", on);
#if 0
	if (on)
	{
		pxa3xx_pmic_set_voltage(VCC_TS2V8, 2800);
		pxa3xx_pmic_set_voltage(VCC_TS1V8, 1800);
	}
	else
	{
		pxa3xx_pmic_set_voltage(VCC_TS2V8, 0);
		pxa3xx_pmic_set_voltage(VCC_TS1V8, 0);
	}
#endif	
	return 0;
}

static struct atmel_i2c_rmi_platform_data atmel_touch_data = {
	.power = touch_power,
};


#elif defined(CONFIG_TOUCHSCREEN_ATMELQT_OBP)

static int t8200_touch_reset(void)
{

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TP_RESET)), 1);
	msleep(1);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(TP_RESET)), 0);
	msleep(20);
	gpio_set_value(mfp_to_gpio(MFP_CFG_PIN(TP_RESET)), 1);
	msleep(45);

	return 0;
}

static struct maxtouch_ts_platform_data t8200_ts_platform_data;

static struct maxtouch_key t8200_maxtouch_keys[] = {
	{
		.channel	= 0,
		.code		= KEY_SEND,
	},
	{
		.channel	= 1,
		.code		= KEY_MENU,
	},
	{
		.channel	= 2,
		.code		= KEY_HOME,
	},
	{
		.channel	= 3,
		.code		= KEY_BACK,
	},
	{
		.channel	= 4,
		.code		= KEY_END,
	},
};

static struct mts_touch_keyarray_cfg t8200_key_array_data[] = {
	{
		.ctrl = 0x03,
		.x_origin = 0x03,
		.y_origin = 0x0b,
		.x_size = 0x05,
		.y_size = 0x01,
		.aks_cfg = 0x01,
		.burst_len = 0x10,
		.tch_det_thr = 0x28,
		.tch_det_int = 0x02,
		.rsvd1 = 0,
		.rsvd2 = 0,
	},
	{
		.ctrl		= 0,
		.x_origin	= 0,
		.y_origin	= 0,
		.x_size		= 0,
		.y_size		= 0,
		.aks_cfg	= 0,
		.burst_len	= 0,
		.tch_det_thr	= 0,
		.tch_det_int	= 0,
		.rsvd1		= 0,
		.rsvd2		= 0,
	},
};

static struct maxtouch_ts_platform_data t8200_ts_platform_data = {
	.flags		= (MAXTOUCH_USE_MULTITOUCH |
			   MAXTOUCH_USE_KEYARRAY |
			   MAXTOUCH_CFG_BACKUPNV |
			   MAXTOUCH_EEPROM_CHECKSUM),
	.irqflags	= IRQF_TRIGGER_FALLING,
	.abs_min_x	= 0,
	.abs_max_x	= 1024,
	.abs_min_y	= 0,
	.abs_max_y	= 1024,
	.abs_min_p	= 0,
	.abs_max_p	= 255,
	.abs_min_w	= 0,
	.abs_max_w	= 15,
	.x_delta	= 400,
	.y_delta	= 250,
	.nv_checksum	= 0xf429,
	.fuzz_x		= 0,
	.fuzz_y		= 0,
	.fuzz_p		= 2,
	.fuzz_w		= 2,
	.hw_reset	= t8200_touch_reset,
	.key_array = {
		.cfg		= t8200_key_array_data,
		.keys		= t8200_maxtouch_keys,
		.num_keys	= ARRAY_SIZE(t8200_maxtouch_keys),
	},
	.power_cfg	= {
		.idle_acq_int	= 50,
		.active_acq_int	= 10,
		.active_idle_to	= 50,
	},
	.acquire_cfg	= {
		.charge_time	= 0x06,
		.reserve0	= 0x00,
		.touch_drift	= 0x05,
		.drift_susp	= 0x05,
		.touch_autocal	= 0x00,
		.sync		= 0,
		.anti_cal_susp	= 0x0a,
		.anti_cal_sthr	= 0x14,
	},
	.multi_touch_cfg	= {
		.ctrl		= 0x0b,
		.x_origin	= 0,
		.y_origin	= 0,
		.x_size		= 0x10,
		.y_size		= 0x0b,
		.aks_cfg	= 0x01,
		.burst_len	= 0x20,
		.tch_det_thr	= 0x2d,
		.tch_det_int	= 0x07,
		.orient		= 0x03,
		.mrg_to		= 0x00,
		.mov_hyst_init	= 0x05,
		.mov_hyst_next	= 0x01,
		.mov_filter	= 0x00,
		.num_touch	= 0x02,
		.merge_hyst	= 0x0a,
		.merge_thresh	= 0x19,
		.amp_hyst	= 0x0a,
		.x_res		= 0x0000,
		.y_res		= 0x0000,
		.x_low_clip	= -20,
		.x_high_clip	= 0x00,
		.y_low_clip	= -15,
		.y_high_clip	= -15,
		.x_edge_ori	= 151,
		.x_edge_cdist	= 55,
		.y_edge_ori	= 140,
		.y_edge_cdist	= 82,
	},
	.linear_tbl_cfg = {
		.ctrl = 0x00,
		.x_offset = 0x0000,
		.x_segment = {
			0x00, 0x00, 0x00 , 0x00,
			0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00
		},
		.y_offset = 0x0000,
		.y_segment = {
			0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00
		},
	},
	.gpio_pwm_cfg = {
		.ctrl			= 0,
		.report_mask		= 0,
		.pin_direction		= 0,
		.internal_pullup	= 0,
		.output_value		= 0,
		.wake_on_change		= 0,
		.pwm_enable		= 0,
		.pwm_period		= 0,
		.duty_cycle_0		= 0,
		.duty_cycle_1		= 0,
		.duty_cycle_2		= 0,
		.duty_cycle_3		= 0,
		.trigger_0		= 0,
		.trigger_1		= 0,
		.trigger_2		= 0,
		.trigger_3		= 0,
	},
	.com_cfg = {
		.ctrl		= 0x00,
		.cmd		= 0x00,
	},
	.grip_suppression_cfg = {
		.ctrl		= 0x03,
		.xlogrip	= 0x00,
		.xhigrip	= 0x00,
		.ylogrip	= 0x00,
		.yhigrip	= 0x00,
		.maxtchs	= 0x00,
		.reserve0	= 0x00,
		.szthr1		= 0x55,
		.szthr2		= 0x28,
		.shpthr1	= 0x04,
		.shpthr2	= 0x0f,
		.supextto	= 0x00,
	},
	.noise_suppression_cfg = {
		.ctrl			= 0x05,
		.reserve0		= 0x0000,
		.gcaf_upper_limit	= 0x0000,
		.gcaf_lower_limit	= 0x0000,
		.gcaf_num_active	= 0x00,
		.noise_threshold	= 0x0f,
		.reserve1		= 0,
		.freq_hop_scale		= 0x00,
		.burst_freq_0		= 0x14,
		.burst_freq_1		= 0x1e,
		.burst_freq_2		= 0x28,
		.burst_freq_3		= 0xff,
		.burst_freq_4		= 0xff,
		.gcaf_num_idle		= 0x03,
	},
	.one_touch_gesture_proc_cfg = {
		.ctrl			= 0,
		.num_gest		= 0,
		.gesture_enable		= 0,
		.pres_proc		= 0,
		.tap_time_out		= 0,
		.flick_time_out		= 0,
		.drag_time_out		= 0,
		.short_press_time_out	= 0,
		.long_press_time_out	= 0,
		.repeat_press_time_out	= 0,
		.flick_threshold	= 0,
		.drag_threshold		= 0,
		.tap_threshold		= 0,
		.throw_threshold	= 0,
	},
	.self_test_cfg = {
		.ctrl			= 0,
		.command		= 0,
		.high_signal_limit_0	= 0,
		.low_signal_limit_0	= 0,
		.high_signal_limit_1	= 0,
		.low_signal_limit_1	= 0,
	},
	.two_touch_gesture_proc_cfg = {
		.ctrl			= 0,
		.reserved0		= 0,
		.reserved1		= 0,
		.gesture_enable		= 0,
		.rotate_threshold	= 0,
		.zoom_threshold		= 0,
	},
	.cte_config_cfg = {
		.ctrl			= 0x00,
		.command		= 0x00,
		.mode			= 0x00,
		.idle_gcaf_depth	= 0x04,
		.active_gcaf_depth	= 16,
		.voltage		= 0x0a,
	},
	.noise1_suppression_cfg = {
		.ctrl		= 0x00,
		.reserved	= 0x00,
		.atchthr	= 0x00,
		.duty_cycle	= 0x00,
	},
	.userdata = {
		.data_0		= 0x00,
		.data_1		= 0x00,
		.data_2		= 0x00,
		.data_3		= 0x00,
		.data_4		= 0x00,
		.data_5		= 0x00,
		.data_6		= 0x00,
		.data_7		= 0x00,
	},
	.vkeys			= {
		.count		= 0,
		.keys		= NULL,
	},
};

#endif

static void t8200_touch_init(void)
{
	gpio_request(mfp_to_gpio(MFP_CFG_PIN(TP_RESET)), "t8200 touch reset");
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(TP_RESET)), 1);


	gpio_request(mfp_to_gpio(MFP_CFG_PIN(TP_INT)), "t8200 touch irq");
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(TP_INT)));
    
    msleep(100);
}

#if defined(CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539)
static struct synaptics_key t8200_synaptics_keys[] = {
	{
		.channel = 0,
		.code = KEY_SEND,
		.start_position = 0,
		.end_position = 399,
	},
	{
		.channel = 1,
		.code = KEY_MENU,
		.start_position = 400,
		.end_position = 799,
	},
	{
		.channel = 2,
		.code = KEY_HOME,
		.start_position = 800,
		.end_position = 1199,
	},
	{
		.channel = 3,
		.code = KEY_BACK,
		.start_position = 1200,
		.end_position = 1599,
	},
	{
		.channel = 4,
		.code = KEY_END,
		.start_position = 1600,
		.end_position = 1997,
	},
};

static struct synaptics_i2c_rmi_platform_data t8200_ts_synaptics_platform_data = {
	.irqflags = IRQF_TRIGGER_FALLING,

	.abs_min_x = 0,
	.abs_max_x = 1997,
	.abs_min_y = 0,
	.abs_max_y = 2999,
	.abs_min_p = 0,
	.abs_max_p = 255,
	.abs_min_w = 0,
	.abs_max_w = 15,
	.fuzz_x = 0,
	.fuzz_y = 0,
	.fuzz_p = 2,
	.fuzz_w = 2,
	.keys_min_y = 3000,
	.keys_max_y = 3507,
	.touch_sensitivity = -16,
	.key_array = {
		.keys = t8200_synaptics_keys,
		.num_keys = ARRAY_SIZE(t8200_synaptics_keys),
	},
};

#endif

#ifdef CONFIG_INPUT_KEYRESET 
static int keys_up[] = {
	KEY_VOLUMEDOWN,
	0,
};

static struct keyreset_platform_data keyreset_pdata = {
	.keys_up	= keys_up,
	.keys_down	= {KEY_BACK, KEY_MENU, KEY_VOLUMEUP, KEY_CAMERA, 0},
};

static struct platform_device keyreset_device = {
	.name 		= "keyreset",
	.id 		= -1,
	.dev		= {
		.platform_data = &keyreset_pdata,
	},
};
#endif

#if defined(CONFIG_MICCO_HEADSET_DETECT)
#define HS_DET_GPIO mfp_to_gpio(MFP_CFG_PIN(HS_DET_N))
static struct micco_headset_platform_data t8200_headset_platform_data = {
	.gpio_detect = HS_DET_GPIO,
};

static struct platform_device micco_headset_device = {
	.name		= "micco-headset",
	.id		= -1,
	.dev		= {
		.platform_data = &t8200_headset_platform_data,
	},
};

static void __init headset_init(void)
{
	platform_device_register(&micco_headset_device);
}
#else
static inline void headset_init(void) {}
#endif /* CONFIG_MICCO_HEADSET_DETECT */

static mfp_cfg_t W00_mfp_cfg[] __initdata = {
	HS_DET_N,
};

static mfp_cfg_t P15_mfp_cfg[] __initdata = {
	IPC_UART_OE,
	HS_DET_N_2,
	HS_UART_SW,
};

static void misc_mfp_config(void)
{
    	switch (get_mach_version()) {
		case MACH_VERSION_P10:
			pxa3xx_mfp_config(ARRAY_AND_SIZE(W00_mfp_cfg));
			break;
		default:
			pxa3xx_mfp_config(ARRAY_AND_SIZE(P15_mfp_cfg));
			//gpio_request(MFP_CFG_PIN(HS_UART_SW), "hs uart sw");
			//gpio_direction_output(MFP_CFG_PIN(HS_UART_SW), 0);
			/* this is for disable BP uart. it is from P1.5 */
			gpio_request(MFP_CFG_PIN(IPC_UART_OE), "ipc uart oe");
			gpio_direction_output(MFP_CFG_PIN(IPC_UART_OE), 1);
	}

}

#ifdef CONFIG_IMM
static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		/* Moved up to offset 64KB due to collision with DVFM
		 * D2 entry code 
		 */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		/* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa935_resource_imm),
	.resource	= pxa935_resource_imm,
};
#endif
static inline void i2c_gpio_init(void) {}

#if defined(CONFIG_BQ2407x_POWER)
#define CHARGE_EN1_GPIO		mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN1))
#define CHARGE_EN2_GPIO		mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN2))
#define CHARGE_PG_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_PG_N))
#define CHARGE_CE_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_CE_N))
#define CHARGE_STAT_GPIO	mfp_to_gpio(MFP_CFG_PIN(CHARGE_STAT))

static int bq2407x_charger_init(void)
{
	int err;

	err = gpio_request(CHARGE_EN1_GPIO, "charger en1");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en1 error!\n");
		return err;
	}
	err = gpio_request(CHARGE_EN2_GPIO, "charger en2");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en2 error!\n");
		return err;
	}
	err = gpio_request(CHARGE_PG_N_GPIO, "charger poog");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger poog error!\n");
		return err;
	}
	err = gpio_request(CHARGE_CE_N_GPIO, "charger en");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger en error!\n");
		return err;
	}
	err = gpio_request(CHARGE_STAT_GPIO, "charger det");	
	if (err) {
		printk(KERN_ERR "gpio request ext charger det error!\n");
		return err;
	}

	gpio_direction_output(CHARGE_EN1_GPIO, 1);
	gpio_direction_output(CHARGE_EN2_GPIO, 1);
	gpio_direction_input(CHARGE_PG_N_GPIO);
	gpio_direction_output(CHARGE_CE_N_GPIO, 1);
	gpio_direction_input(CHARGE_STAT_GPIO);

	return 0;
}

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
#define WAITING_USB_READY_INTERVAL	200
#define WAITING_USB_READY_TIMES		5

extern void pxa_charger_decide_enable(int on);
extern int is_usb_driver_busy(void);
static struct completion charger_detect_complete;
int g_psy_type = BQ2407x_POWER_NO_CHARGER;
#endif

static int bq2407x_charger_detect(void)
{
	int ret = BQ2407x_POWER_CHARGE_USB;
	int retry_times = 0;
	unsigned long timeout = 0;

	if (gpio_get_value(CHARGE_PG_N_GPIO)) {
		printk(KERN_DEBUG "%s: No charger detected\n", __func__);
		ret = BQ2407x_POWER_NO_CHARGER;
		g_psy_type = BQ2407x_POWER_NO_CHARGER;
	} else {
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
		while (is_usb_driver_busy()) {
			printk(KERN_DEBUG "%s: waiting for usb status busy "
					  "(retry times=%d\n",
					  __func__, retry_times);

			msleep(WAITING_USB_READY_INTERVAL);
			if (WAITING_USB_READY_TIMES <= ++retry_times) {
				printk(KERN_DEBUG "%s: USB regisered\n",
					   __func__);
				goto out;
			}
		}
			
		INIT_COMPLETION(charger_detect_complete);

		pxa_charger_decide_enable(1);
		enable_irq(IRQ_U2P);
		timeout = wait_for_completion_timeout(&charger_detect_complete,
					msecs_to_jiffies(100));

		if (timeout) {
			ret = BQ2407x_POWER_CHARGE_AC;
			g_psy_type = BQ2407x_POWER_CHARGE_AC;
			printk(KERN_DEBUG "%s: charger (AC) detected\n",
					   __func__);
		} else {
			disable_irq(IRQ_U2P);
			ret = BQ2407x_POWER_CHARGE_USB;
			g_psy_type = BQ2407x_POWER_CHARGE_USB;
			printk(KERN_DEBUG "%s: charger (USB) detected\n",
					   __func__);
		}
		pxa_charger_decide_enable(0);
#endif
	}
out:
	return ret; 
}

static int bq2407x_is_charging(void)
{
	return (gpio_get_value(CHARGE_STAT_GPIO) ? 0 : 1);
}

static void bq2407x_set_charge(int enable)
{
	if (enable == BQ2407x_POWER_ENABLE) {
		printk(KERN_DEBUG "%s: charger enable\n", __func__);
		gpio_direction_output(CHARGE_CE_N_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_CE_N),
				   MFP_LPM_DRIVE_LOW);	
	}
	else {
		printk(KERN_DEBUG "%s: charger disable\n", __func__);
		gpio_direction_output(CHARGE_CE_N_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_CE_N),
				   MFP_LPM_DRIVE_HIGH);	
	}
}

static void bq2407x_set_current(int flags)
{
	switch(flags) {
	case CHARGING_SUSPEND:
		printk(KERN_DEBUG "%s: current suspend\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 1);
		gpio_direction_output(CHARGE_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	case CHARGING_CURRENT_100:
		printk(KERN_DEBUG "%s: current 100\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_500:
		printk(KERN_DEBUG "%s: current 500\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 1);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_HIGH);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	case CHARGING_CURRENT_FULL:
		printk(KERN_DEBUG "%s: current full\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 1);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_HIGH);	
		break;
	default:
		printk(KERN_DEBUG "%s: current default\n", __func__);
		gpio_direction_output(CHARGE_EN1_GPIO, 0);
		gpio_direction_output(CHARGE_EN2_GPIO, 0);
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN1),
				   MFP_LPM_DRIVE_LOW);	
		pxa3xx_mfp_set_lpm(MFP_CFG_PIN(CHARGE_EN2),
				   MFP_LPM_DRIVE_LOW);	
		break;
	}
}

static int bq2407x_is_cable_on(void)
{
	return gpio_get_value(CHARGE_PG_N_GPIO) ? 0 : 1;
}

static void bq2407x_charger_exit(void)
{
        gpio_free(CHARGE_EN1_GPIO);
        gpio_free(CHARGE_EN2_GPIO);
        gpio_free(CHARGE_PG_N_GPIO);
        gpio_free(CHARGE_CE_N_GPIO);
        gpio_free(CHARGE_STAT_GPIO);
}

static char *power_supplicants[] = {
	"battery",
};

static struct bq2407x_charger_pdata bq2407x_charger_pdata = {
	.init		= bq2407x_charger_init,
	.detect		= bq2407x_charger_detect,
	.is_charging	= bq2407x_is_charging,
	.is_cable_on	= bq2407x_is_cable_on,
	.set_charge	= bq2407x_set_charge,
	.set_current	= bq2407x_set_current,
	.exit		= bq2407x_charger_exit,
	.irq		= IRQ_GPIO(CHARGE_PG_N_GPIO),
	.det_irq	= IRQ_GPIO(CHARGE_STAT_GPIO),
	.supplied_to 	= power_supplicants,
	.num_supplicants = ARRAY_SIZE(power_supplicants),
};

static struct platform_device bq2407x_charger_device = {
	.name		= "charger",
	.id		= -1,
	.dev = {
		.platform_data = &bq2407x_charger_pdata,
	},
};

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
static irqreturn_t charger_detected_irq(int irq, void *dev_id)
{
	printk(KERN_DEBUG "charger detected via U2O D+/D- irq\n");
	disable_irq(IRQ_U2P);
	complete(&charger_detect_complete);
	return IRQ_HANDLED;
}
#endif

static void init_charger(void)
{
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	if(request_irq(IRQ_U2P, charger_detected_irq, 
		       IRQF_DISABLED | IRQF_TRIGGER_FALLING,
		       "charger_detect", NULL) < 0) {
        	printk("charger_detected_irq register failed\n");
	}
	disable_irq(IRQ_U2P);
	init_completion(&charger_detect_complete);
#endif

	platform_device_register(&bq2407x_charger_device);
}

#endif /* CONFIG_BQ2407x_POWER */

struct platform_device battery_device = {
	.name           = "battery",
	.id             = -1,
	.dev		= {
		.bus_id = "battery",
	},
};

#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
#define BP_WD_OUT_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_WD_OUT))
#define BP_READY_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_READY))
#define BP_ON_GPIO		mfp_to_gpio(MFP_CFG_PIN(BP_ON))
#define BP_SHUTDOWN_GPIO	mfp_to_gpio(MFP_CFG_PIN(BP_SHUTDOWN))
#define AP_READY_GPIO		mfp_to_gpio(MFP_CFG_PIN(AP_READY))
#define AP_RESET_BP_GPIO	mfp_to_gpio(MFP_CFG_PIN(AP_RESET_BP))
#define DPRAM_INT_N_GPIO	mfp_to_gpio(MFP_CFG_PIN(DPRAM_INT_N))

static int ipc_handshake_init(void)
{
	if (gpio_request(AP_RESET_BP_GPIO, "ap reset bp")) {
		pr_err("%s: can not get gpio %d\n", __func__, AP_RESET_BP_GPIO);
	} 
	gpio_direction_output(AP_RESET_BP_GPIO, 1);

	if (gpio_request(BP_ON_GPIO, "bp on")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_ON_GPIO);
	} 
	gpio_direction_output(BP_ON_GPIO, 0);

	if (gpio_request(BP_SHUTDOWN_GPIO, "bp shutdown")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_ON_GPIO);
	} 
	gpio_direction_output(BP_SHUTDOWN_GPIO, 0);

	if (gpio_request(AP_READY_GPIO, "ap ready")) {
		pr_err("%s: can not get gpio %d\n", __func__, AP_READY_GPIO);
	}
	gpio_direction_output(AP_READY_GPIO, 0);

	if (gpio_request(BP_READY_GPIO, "bp ready")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_READY_GPIO);
	}
	gpio_direction_input(BP_READY_GPIO);

	if (gpio_request(BP_WD_OUT_GPIO, "bp watch dog")) {
		pr_err("%s: can not get gpio %d\n", __func__, BP_WD_OUT_GPIO);
	}
	if (gpio_request(DPRAM_INT_N_GPIO, "ipc mailbox")) {
		pr_err("%s: can not get gpio %d\n", __func__, DPRAM_INT_N_GPIO);
	}
	gpio_direction_input(DPRAM_INT_N_GPIO);
	return 0;
}

static int ipc_handshake_exit(void)
{
	gpio_free(BP_READY_GPIO);
	gpio_free(DPRAM_INT_N_GPIO);
	gpio_free(AP_READY_GPIO);
	gpio_free(AP_READY_GPIO);
	gpio_free(BP_ON_GPIO);
	gpio_free(BP_WD_OUT_GPIO);
	return 0;
}

static struct clk *ipc_smc_clk = NULL;
static unsigned char *ipc_smc_membase = NULL;
static u16 *ipc_ram_addr = NULL;
static enum state_type config_ipc(enum config_ipc_cmd cmd)
{
	enum state_type ret = READY; 

	switch (cmd) {
		case GET_BP_READY:
			ret = gpio_get_value(BP_READY_GPIO) ? READY : NOT_READY;
			break;
		case RESET_BP:
			pr_info("%s: Reset BP\n", __func__);
			gpio_direction_output(AP_RESET_BP_GPIO, 1);

			gpio_direction_output(BP_ON_GPIO, 1);
			msleep(70);
			gpio_direction_output(BP_ON_GPIO, 0);

			gpio_direction_output(AP_RESET_BP_GPIO, 0);
			ret = READY;
			break;
		case SET_AP_READY:
			gpio_direction_output(AP_READY_GPIO, 1);
			ret = READY;
			break;
		case SET_AP_NOT_READY:
			gpio_direction_output(AP_READY_GPIO, 0);
			ret = NOT_READY;
			break;
		case SET_AP_SLEEP:
			break;
		case SET_AP_RESUME:
			break;
		case GET_MAILBOX_STATE:
			if (!gpio_get_value(DPRAM_INT_N_GPIO))
				ret = NOT_READY;
			break;
		case POWERON_BP:
			pr_info("%s: Power on BP\n", __func__);
			gpio_direction_output(BP_ON_GPIO, 1);
			msleep(70);
			gpio_direction_output(BP_ON_GPIO, 0);
			gpio_direction_output(AP_RESET_BP_GPIO, 0);
			
			break;
		case SHUTDOWN_BP:
			/* Power off BP */
			pr_info("%s: Power off BP\n", __func__);
			gpio_direction_output(BP_SHUTDOWN_GPIO, 1);
			udelay(50);
			gpio_direction_output(BP_SHUTDOWN_GPIO, 0);
			break;
		default:
			pr_err("%s: unknow config command: %d\n", __func__, cmd);
			break;
	}
	return ret;
}


/*****************************************************************************
 * The SMC registers read/write routines
 *****************************************************************************/
static unsigned int smc_reg_read(u8 *base, int offset)
{
	return __raw_readl(base + offset);
}

static inline void smc_reg_write(u8 *base, int offset, unsigned long val)
{
	unsigned long flags;

	local_irq_save(flags);
	writel(val, base + offset);
	__raw_readl(base + offset);
	local_irq_restore(flags);
}

#define SMC_ADDR_START		0x4A000000	/* SMC Registers*/
#define SMC_CSMSADRCFG		0xA0		/* CSMSADRCFG */
#define SMC_MEMCLKCFG		0x68		/* MEMCLKCFG */
#define SMC_CSADRCFG		0x8C		/* CSADRCFG */
#define SMC_MSC1		0xC		/* MSC1 */
#define SMC_ADDR_LEN		0xFF

#define DFI_CS3_IOBASE 	0x14000000
#define RAM_TOTAL	32*1024

static void ipc_smc_init(void)
{
	unsigned int val;

	if (!ipc_smc_clk || !ipc_smc_membase) {
		pr_err("%s: clk is NULL\n", __func__);
		return; 
	}

	clk_enable(ipc_smc_clk);

	/*CSMSADRCFG = 0x2  */
	smc_reg_write(ipc_smc_membase, SMC_CSMSADRCFG, 0x2);
	val = smc_reg_read(ipc_smc_membase, SMC_CSMSADRCFG);

	/*config CSADRCFG3  */
	smc_reg_write(ipc_smc_membase, SMC_CSADRCFG, 0x00364919);

	/*config MSC1  */
	val = smc_reg_read(ipc_smc_membase, SMC_MSC1);
	val &= ~0xffff0000;
	val |= 0x05590000;
	val &= ~0xf000f000;/* 31-28, 15-12 bits reserved, write 0 */
	smc_reg_write(ipc_smc_membase, SMC_MSC1, val);
}

static void ipc_smc_exit(void)
{

}

static struct dpram_platform_data cypress_dpram_data;

static int ipc_misc_init(void)
{
	int ret;
	if (!(ipc_smc_clk = clk_get(NULL, "SMCCLK"))) {
		pr_err("%s: get SMCCLK failed\n", __func__);
		ret = -EINVAL;
		return ret;
	}

	clk_enable(ipc_smc_clk);
	/*SMC controller registers address map*/
	ipc_smc_membase = ioremap(SMC_ADDR_START, SMC_ADDR_LEN);
	if (!ipc_smc_membase) {
		pr_err("%s: remap SMC address failed\n", __func__);
		ret = -ENOMEM;
		goto err0;
	}
	ipc_ram_addr = (u16 *)ioremap_nocache(DFI_CS3_IOBASE, RAM_TOTAL);
	if (!ipc_ram_addr) {
		pr_err("%s: remap ram address failed\n", __func__);
		ret = -ENOMEM;
		goto err1;
	}
	cypress_dpram_data.ram_addr = ipc_ram_addr;
	return 0;
err1:
	iounmap(ipc_smc_membase);
	ipc_smc_membase = NULL;
err0:
	clk_disable(ipc_smc_clk);
	clk_put(ipc_smc_clk);
	ipc_smc_clk = NULL;
	return ret;
}

static int ipc_misc_exit(void)
{	
	if (ipc_smc_membase) {
		iounmap(ipc_smc_membase);
		ipc_smc_membase = NULL;
	}
	if (ipc_ram_addr) {
		iounmap(ipc_ram_addr);
		cypress_dpram_data.ram_addr = NULL;
	}
	if (ipc_smc_clk) {
		clk_disable(ipc_smc_clk);
		clk_put(ipc_smc_clk);
	}
	return 0;
}

static struct dpram_platform_data cypress_dpram_data = {
	.bp_wd_out_irq = IRQ_GPIO(MFP_CFG_PIN(BP_WD_OUT)),
	.bp_rdy_irq = IRQ_GPIO(MFP_CFG_PIN(BP_READY)),
	.mailbox_irq = IRQ_GPIO(MFP_CFG_PIN(DPRAM_INT_N)),
	.config_ipc = config_ipc,
};

static struct platform_device cypress_dpram_device = {
	.name 		= "cypress_dpram",
	.id 		= -1,
	.dev = {
		.platform_data = &cypress_dpram_data,
	},
};

static void BP_init(void)
{
	ipc_handshake_init();
	ipc_misc_init();
	ipc_smc_init();
	config_ipc(POWERON_BP);

}
#endif

#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65_FOR_TD)
static void snd_ssp_mfp_cfg_init(void)
{
    extern mfp_cfg_t pcm_on_mfp_cfg[4];
    extern mfp_cfg_t pcm_off_mfp_cfg[4];
    extern mfp_cfg_t i2s_on_mfp_cfg[3];
    extern mfp_cfg_t i2s_off_mfp_cfg[3];

    pcm_on_mfp_cfg[0] = AP_PCM_CLK;
    pcm_on_mfp_cfg[1] = AP_PCM_FRM;
    pcm_on_mfp_cfg[2] = AP_PCM_TXD;
    pcm_on_mfp_cfg[3] = AP_PCM_RXD;

    pcm_off_mfp_cfg[0] = AP_PCM_CLK_OFF;
    pcm_off_mfp_cfg[1] = AP_PCM_FRM_OFF;
    pcm_off_mfp_cfg[2] = AP_PCM_TXD_OFF;
    pcm_off_mfp_cfg[3] = AP_PCM_RXD_OFF;

    i2s_on_mfp_cfg[0] = HIFI_CLK;
    i2s_on_mfp_cfg[1] = HIFI_FRM;
    i2s_on_mfp_cfg[2] = HIFI_TXD;

    i2s_off_mfp_cfg[0] = MFP_CFG_LPM(GPIO85, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[1] = MFP_CFG_LPM(GPIO87, AF0, PULL_HIGH);
    i2s_off_mfp_cfg[2] = MFP_CFG_LPM(GPIO88, AF0, PULL_LOW);
}
#endif

#if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2018) || defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2018_MODULE)
static mfp_cfg_t fm2018_b1_mfp_cfg[] = {
	AEC_RESET_N,
	AEC_IRQ,
	GPIO127_FREE_B1,
};

static mfp_cfg_t fm2018_b2_mfp_cfg[] = {
	AEC_RESET_N_B2,
	AEC_IRQ,
	GPIO98_FREE_B2,
};

typedef struct {
	int irq;
	int reset;    
} S_AEC_PIN;

static S_AEC_PIN g_aecPins;
static int g_aecInit = 0;

static int pxa3xx_aec_reset(void)
{
	gpio_direction_output(g_aecPins.reset, 0);
	msleep(10);
	gpio_direction_output(g_aecPins.reset, 1);
	msleep(20);
	return 0;
}

static int pxa3xx_aec_wakeup(void)
{ 
	printk("[AEC] aec wakeup\n");
	pxa3xx_pmic_enable_voltage(VCC_AEC, 0);
	msleep(100);
	pxa3xx_pmic_enable_voltage(VCC_AEC, 1);
	msleep(20);

	gpio_direction_output(g_aecPins.reset, 0);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.reset), MFP_LPM_DRIVE_HIGH);

	gpio_direction_output(g_aecPins.irq, 0);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_HIGH);

	return 0; 
}

static int pxa3xx_aec_sleep(void)
{
	printk("[AEC] aec sleep...\n");

	msleep(20);
	pxa3xx_pmic_enable_voltage(VCC_AEC, 0);

	gpio_direction_output(g_aecPins.irq, 0);
	gpio_direction_output(g_aecPins.reset, 0);

	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_LOW);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.reset), MFP_LPM_DRIVE_LOW);

	return 0;
}

static int pxa3xx_aec_init(void)
{
	int ret;     
	printk("[AEC] init aec pins\n");

	switch (get_mach_version()) {
		case MACH_VERSION_P10:
                	pxa3xx_mfp_config(ARRAY_AND_SIZE(fm2018_b1_mfp_cfg));
	        	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N)), "AEC_RESET_N");
	        	BUG_ON(ret);
	        	g_aecPins.reset = mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N));
			break;
		default:
                	pxa3xx_mfp_config(ARRAY_AND_SIZE(fm2018_b2_mfp_cfg));
	        	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N_B2)), "AEC_RESET_N_B2");
	        	BUG_ON(ret);
			g_aecPins.reset = mfp_to_gpio(MFP_CFG_PIN(AEC_RESET_N_B2));
            		break;
	}

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AEC_IRQ)), "AEC_IRQ");
	BUG_ON(ret);
	g_aecPins.irq = mfp_to_gpio(MFP_CFG_PIN(AEC_IRQ));

	g_aecInit = 1;

	gpio_direction_output(g_aecPins.irq, 0);
	gpio_direction_output(g_aecPins.reset, 0);

	return 0;
}

static int pxa3xx_aec_set_mode(int mode)
{
	if(g_aecInit == 0) {
		printk("[AEC] device not init\n");
		return 0;
	}

	if (mode == AEC_BYPASS_MODE) {
		//gpio_direction_output(g_aecPins.irq, 0);
		//msleep(1);
		gpio_direction_output(g_aecPins.irq, 1);
	    pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_HIGH);
	} else if (mode == AEC_NORMAL_MODE) {
		//gpio_direction_output(g_aecPins.irq, 1);
		//msleep(1);
		gpio_direction_output(g_aecPins.irq, 0);
		pxa3xx_mfp_set_lpm(gpio_to_mfp(g_aecPins.irq), MFP_LPM_DRIVE_LOW);
	} else
		WARN_ON(1);
	return 0;
}

static const aec_reg_t g_bypass_tbl[] = {
     {0x0F84 ,0x0080}
    ,{0x1E30 ,0x0221}
    ,{0x1E41 ,0x0001}
    ,{0x1E4A ,0x000E}
    ,{0x1E52 ,0x000B}
    ,{0x1E60 ,0x0000}
    ,{0x1E70 ,0x05C0}
    ,{0x1ED4 ,0x3800}
    ,{0x1E3A ,0x0000}
    ,{0x0,0x0} /* End of table. */
};

static const aec_reg_t g_normal_tbl[] = {
     {0x0F84, 0x0080}
    ,{0x1E34, 0x00AA}
    ,{0x1E41, 0x0002}
    ,{0x1E45, 0x43DE}
    ,{0x1E46, 0x0035}
    ,{0x1E47, 0x3000}
    ,{0x1E48, 0x0800}
    ,{0x1E49, 0x0800}
    ,{0x1E52, 0x0013}
    ,{0x1E60, 0x0000}
    ,{0x1E63, 0x0001}
    ,{0x1E70, 0x05C0}
    ,{0x1E86, 0x0007}
    ,{0x1E87, 0x0002}
    ,{0x1E8C, 0x0380}
    ,{0x1E9A, 0x0900}
    ,{0x1E9B, 0x6800}
    ,{0x1E9C, 0x2800}
    ,{0x1EC5, 0x1800}
    ,{0x1ED4, 0x3800}
    ,{0x1ED6, 0x1200}
    ,{0x1ED7, 0x0300}
    ,{0x1EDA, 0x4000}
    ,{0x1EDF, 0x4800}
    ,{0x1EED, 0x0500}
    ,{0x1EEE, 0x2800}
    ,{0x1E3A, 0x0000}
    ,{0x0,0x0} /* End of table. */
};

static const aec_reg_t* pxa3xx_aec_get_parm(int mode)
{
	const aec_reg_t* ptr;

	switch (mode) {
		case AEC_BYPASS_MODE:
			ptr = g_bypass_tbl;
			break;

		case AEC_NORMAL_MODE:
		default:
			ptr = g_normal_tbl;
			break;
	}		

	return ptr;
}

static struct aec_platform_data g_aec_platform_data = {
	.init       = pxa3xx_aec_init,	
	.wakeup     = pxa3xx_aec_wakeup,
	.sleep      = pxa3xx_aec_sleep,
	.reset      = pxa3xx_aec_reset,
	.set_mode   = pxa3xx_aec_set_mode,
	.parmGet    = pxa3xx_aec_get_parm,
};
#endif /* #if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2018)  */

struct platform_device usb_mass_storage_device = {
	.name		= "usb_mass_storage",
	.id		= -1,
};

static struct platform_device *devices[] __initdata = {
	&micco_vibrator_device,
#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
	&cypress_dpram_device,
#endif
#ifdef CONFIG_IMM
	&pxa3xx_device_imm,
#endif
#ifdef CONFIG_INPUT_KEYRESET
	&keyreset_device,
#endif
	&battery_device,
	&usb_mass_storage_device,
};

#if defined(CONFIG_SIRF_GSD3TW)
static struct sirf_platform_data sirf_platform_data = {
	.reset_pin = mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)),
	.on_off_pin = mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)),
};

static struct platform_device sirf_device = {
	.name = "sirf",
	.id = -1,
	.dev = {
		.platform_data = &sirf_platform_data,
	},
};

static void __init gps_sirf_init(void)
{
	printk(KERN_ERR"[%s:%d] gps pin init\n", __FILE__, __LINE__);

	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)), "gps reset")) {
		printk(KERN_ERR "<SiRF> GPIO RST request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)), "gps on_off")) {
		printk(KERN_ERR "<SiRF> GPIO ON_OFF request error!!\n");
	}
	if (gpio_request(mfp_to_gpio(MFP_CFG_PIN(AGPS_PWR_ON)), "gps power on")) {
		printk(KERN_ERR "<SiRF> GPIO PWR_ON request error!!\n");
	}
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AGPS_PWR_ON)), 1);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AGPS_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AGPS_RESET)), 1);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(GPS_UART_SWITCH)), 0);
	printk(KERN_ERR"GPS set up finished\n");
	platform_device_register(&sirf_device);
}
#endif

#if defined(CONFIG_ALSPS)

#define PROXIMITY_SHUTDOWN 	MFP_CFG_PIN(PSEN_LEDON)
#define ALSPS_POWERDOWN 	MFP_CFG_PIN(PSEN_PD)
#define PROXIMITY_OUTPUT 	MFP_CFG_PIN(PSEN_DOUT)

static int alsps_init(void)
{
	int ret;

	/*config ambient light sensor switch pin, also as chip switch pin*/	
	ret = gpio_request(mfp_to_gpio(ALSPS_POWERDOWN), "ambient_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset ambient_powerdown failed\n");
		goto out;
	}
	gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);

	/*config proximity sensor switch pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_SHUTDOWN), "proximity_power");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_powerdown failed\n");
		goto out1;
	}
	gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);

	/*config proximity sensor output pin*/
	ret = gpio_request(mfp_to_gpio(PROXIMITY_OUTPUT), "proximity_output");
	if (ret) {
		printk(KERN_ERR "gpio_requset proximity_output failed\n");
		goto out2;
	}
	gpio_direction_input(mfp_to_gpio(PROXIMITY_OUTPUT));

	return 0;

out2:
	gpio_free(mfp_to_gpio(PROXIMITY_SHUTDOWN));
out1:
	gpio_free(mfp_to_gpio(ALSPS_POWERDOWN));
out:
	printk("alsps: config gpio pins failed!\n");
	return ret;
}

static int alsps_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(PROXIMITY_OUTPUT)) ? 1 : 0;
}

static void alsps_setpower(int enable)
{
	if (enable & PROXIMITY_ENABLE) {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 0);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(PROXIMITY_SHUTDOWN), 1);
		pxa3xx_mfp_set_lpm(PROXIMITY_SHUTDOWN, MFP_LPM_DRIVE_HIGH);
	}

	if (enable) {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 0);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_LOW);
	}
	else {
		gpio_direction_output(mfp_to_gpio(ALSPS_POWERDOWN), 1);
		pxa3xx_mfp_set_lpm(ALSPS_POWERDOWN, MFP_LPM_DRIVE_HIGH);
	}
}

static int alsps_chip_verify(int* alsps, int size)
{
	if ((!alsps) || (size != 3))
		return -1;
	alsps[0] = gpio_get_value(PROXIMITY_SHUTDOWN);	/* proximity shutdown pin output */
	alsps[1] = gpio_get_value(PROXIMITY_OUTPUT);	/* proximity output value output */
	alsps[2] = gpio_get_value(ALSPS_POWERDOWN);	/* whole chip powerdown pin output */

	return 0;
}

static struct alsps_platform_data alsps_platform_data = {
	.irq		= IRQ_GPIO(mfp_to_gpio(PROXIMITY_OUTPUT)),
	.adc_event	= PMIC_EVENT_ADC5, 
	.init		= alsps_init, 
	.get_gpio_value	= alsps_get_gpio_value, 
	.setpower	= alsps_setpower,
	.chip_verify	= alsps_chip_verify,
};

static struct platform_device alsps_device = {
	.name = "alsps",
	.id = -1,
	.dev = {
		.platform_data = &alsps_platform_data,
	},
};

static void __init cm3601_alsps_init(void)
{
	platform_device_register(&alsps_device);
}
#endif

#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
static unsigned int matrix_key_map[] = {
	/* KEY(row, col, key_code) */
	KEY(1, 1, KEY_VOLUMEUP),
	KEY(2, 1, KEY_VOLUMEDOWN),
	KEY(2, 2, KEY_CAMERA),
	KEY(1, 2, KEY_FOCUS),
};

static struct pxa27x_keypad_platform_data keypad_info = {
	.matrix_key_rows	= 3,
	.matrix_key_cols	= 3,
	.matrix_key_map		= matrix_key_map,
	.matrix_key_map_size	= ARRAY_SIZE(matrix_key_map),
	.debounce_interval	= 30,
#ifdef CONFIG_DEBUG_TRIGGER_PANIC
    .trigger_key0 = KEY_VOLUMEUP,
    .trigger_key1 = KEY_CAMERA,
#endif
};
#endif /* CONFIG_KEYBOARD_PXA27x || CONFIG_KEYBOARD_PXA27x_MODULE */

#if defined(CONFIG_MMC)
static int mci_init(struct device *dev,
		irq_handler_t detect_int,
		void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	pxa3xx_pmic_set_voltage(VCC_SDIO, 2800);
	cd_irq = gpio_to_irq(mmc_slot[pdev->id].gpio_cd);
	gpio_cd = mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, detect_int,
			IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			"MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(mmc_slot[pdev->id].gpio_cd);
	gpio_cd = mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static mfp_cfg_t sd_in_cfg[]={
	//SD_LDO_CE_ON,
	MMC1_CMD_ON,
	MMC1_DAT0_ON,
	MMC1_DAT1_ON,
	MMC1_DAT2_ON,
	MMC1_DAT3_ON,
	MMC1_CLK_ON,
};

static mfp_cfg_t sd_out_cfg[]={
	//SD_LDO_CE,
	MMC1_CMD,
	MMC1_DAT0,
	MMC1_DAT1,
	MMC1_DAT2,
	MMC1_DAT3,
	MMC1_CLK,
};

static void mci_setpower(struct device *dev, unsigned int vdd)
{
	struct pxamci_platform_data* p_d = dev->platform_data;

	if (( 1 << vdd) & p_d->ocr_mask) {
        	//gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 1);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_in_cfg));
	} else {
        	//gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(SD_LDO_CE)), 0);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(sd_out_cfg));
	}
}

static struct pxamci_platform_data mci_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
//	.gpio_detect    = mfp_to_gpio(MFP_CFG_PIN(MMC1_CD_N)),
	.init 		= mci_init,
	.setpower	= mci_setpower,
	.exit		= mci_exit,
};

static struct pxamci_platform_data mci2_platform_data = {
	.detect_delay	= 20,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
};

static void __init init_mmc(void)
{
	if (device_is_enabled(device_mmc)) {
		pxa_set_mci_info(&mci_platform_data);
		pxa3xx_set_mci2_info(&mci2_platform_data);
	}
}
#else
static inline void init_mmc(void) {}
#endif

#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
static struct mtd_partition partitions[] = {
	[0] = {
		.name        = "Bootloader",
		.offset      = 0,
		.size        = 0x100000,
		.mask_flags  = MTD_WRITEABLE, /* force read-only */
	},
	[1] = {
		.name        = "security",
		.offset      = 0x100000,
		.size        = 0x040000,
		.mask_flags  = MTD_WRITEABLE, /* force read-only */
	},
	[2] = {
		.name        = "cpImage",
		.offset      = 0x140000, 
		.size        = 0x800000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},	
	[3] = {
		.name        = "NVM",
		.offset      = 0x940000, 
		.size        = 0x0c0000,
	},
	[4] = {
		.name        = "logo",
		.offset      = 0xa00000, 
		.size        = 0x040000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	[5] = {
		.name        = "Kernel",
		.offset      = 0xa40000,
		.size        = 0x300000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	[6] = {
		.name        = "system",
		.offset      = 0x0d40000,
		.size        = 0x9800000,     /* mount 128M fs */
	},
	[7] = {
		.name        = "opl",
		.offset      = 0xa540000,     /* mount 32M fs */
		.size        = 0x3000000,
	},
	[8] = {
		.name        = "local",
		.offset      = 0xd540000,
		.size        = 0x1000000,     /* mount 10M fs */
	},
	[9] = {
		.name        = "fota",
		.offset      = 0xe540000,
		.size        = 0x0600000,     /* mount 10M fs */
	},
	[10] = {
		.name        = "cache",       /* mount 8.725M fs */
		.offset      = 0xeb40000,
		.size        = 0x08c0000,
	},
	[11] = {
		.name        = "ftm",       /* mount 8.725M fs */
		.offset      = 0xf400000,
		.size        = 0x0600000,
	},
	[12] = {
		.name        = "panic",       /* mount 8.725M fs */
		.offset      = 0xfa00000,
		.size        = 0x0080000,
	},
	[13] = {
		.name        = "userdata",
		.offset      = 0x10000000,
		.size        = 0xf500000,     /* mount 54M fs */
	},

	[14] = {
		.name        = "BBT",
		.offset      = 0x1f580000,
		.size        = 0x0080000,
		.mask_flags  = MTD_WRITEABLE,  /* force read-only */
	},
	/* NOTES: We reserve some blocks for PXA3xx BBM at the end of NAND.
	 * And the max relocation blocks is not same on different platform.
	 * Please take care it when define the partition table.
	 */
};
static struct flash_platform_data onenand_info;
static void __init init_onenand(void)
{
	onenand_info.parts = partitions;
	onenand_info.nr_parts =	ARRAY_SIZE(partitions);

	pxa3xx_device_onenand.dev.platform_data = &onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#endif


#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
static int micco_init_irq(void)
{
	gpio_direction_input(PMIC_INT_PIN);
	return 0;
}

static int micco_ack_irq(void)
{
	return 0;
}

static void micco_init(void)
{
	u8 value;

	/* Mask interrupts that are not needed */
	micco_write(MICCO_IRQ_MASK_A, 0xFC);
	micco_write(MICCO_IRQ_MASK_B, 0xFF);
	micco_write(MICCO_IRQ_MASK_C, 0x3F);
	micco_write(MICCO_IRQ_MASK_D, 0x3F);

	/* avoid SRAM power off during sleep*/
	value = MICCO_LDO2_ENABLE | MICCO_BUCK1_ENABLE;
	micco_write(MICCO_OVER1, value);

	/* enable all the GSM/WCDMA/EDGE related module */
	value = MICCO_LDO7_ENABLE | MICCO_LDO8_ENABLE | MICCO_LDO10_ENABLE |
		MICCO_LDO11_ENABLE | MICCO_LDO12_ENABLE;
	micco_write(MICCO_APP_OVER2, value);

	value = MICCO_LDO1_ENABLE | MICCO_LDO3_ENABLE | MICCO_LDO5_ENABLE | 
		MICCO_LDO13_ENABLE | MICCO_BUCK3_ENABLE | MICCO_RF_GP_ENABLE;
	micco_write(MICCO_APP_OVER3, value);

	/* disable the ONKEY power down functionality */
	/* micco_write(MICCO_SYSCTRL_B, 0x20); */
	micco_write(MICCO_SYSCTRL_A, 0x60);

	/* Enable BUCK & LDO sleep setting */
	value = MICCO_BUCK1_nSLEEP_ENABLE  | MICCO_BUCK3_nSLEEP_ENABLE;
	micco_write(MICCO_BUCK_SLEEP, value);

	//value = MICCO_LDO8_nSLEEP_ENABLE;
	value = 0;
	micco_write(MICCO_LDO_nSLEEP1, value);

	value = MICCO_RF_GP_nSLEEP_ENABLE;
	micco_write(MICCO_LDO_nSLEEP2, value);

	value = MICCO_LDO8_SLEEP_BIT_ENABLE;
	micco_write(MICCO_LDO_SLEEP_BIT1, value);

	value = 0;
	micco_write(MICCO_LDO_SLEEP_BIT2, value);

	value = MICCO_BUCK1_USE_LDO8_ENABLE | MICCO_BUCK3_USE_LDO8_ENABLE;
	micco_write(MICCO_BUCK_LDO8_EN, value);

	/* Set Buck3 as 2.2V to fix INNODEV stall bug */
	micco_write(MICCO_3DTV1, 0x04);
	micco_write(MICCO_VCC2, 0x01);

	/* IRQ is masked during the power-up sequence and will not be released
	 * until they have been read for the first time */
	micco_read(MICCO_EVENT_A, &value);
	printk("%s: MICCO_EVENT_A=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_B, &value);
	printk("%s: MICCO_EVENT_B=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_C, &value);
	printk("%s: MICCO_EVENT_C=0x%x\n", __FUNCTION__, value);
	micco_read(MICCO_EVENT_D, &value);
	printk("%s: MICCO_EVENT_D=0x%x\n", __FUNCTION__, value);
}

/* micco_power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module miccoEA_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		0},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,		0},
	{VCC_SDIO,		LDO13},
	{VCC_CAMERA_ANA,	LDO7},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO12},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,		0},
	{VCC_1P8V,		0},
	{VCC_MEM,		LDO2},
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
	{VCC_BT,        	0},
	{VCC_AEC,       	LDO4},
};
static struct power_chip micco_chips[] = {
	{MICCO_EA_ID,	"miccoEA",	miccoEA_power_modules},
	{MICCO_EB_ID,	"miccoEB",	miccoEA_power_modules},
	{0,		NULL,		NULL},
};

static struct micco_platform_data micco_data = {
	.init_irq = micco_init_irq,
	.ack_irq = micco_ack_irq,
	.platform_init = micco_init,
	.power_chips = micco_chips,
};
#endif /* CONFIG_PXA3xx_MICCO || CONFIG_PXA3xx_MICCO_MODULE*/
#if	defined(CONFIG_SENSORS_AK8973)
#define COMPASS_INT_GPIO	mfp_to_gpio(MFP_CFG_PIN(CSEN_DRDY))
#define COMPASS_IRQ             IRQ_GPIO(COMPASS_INT_GPIO)
#define COMPASS_RESET_GPIO	mfp_to_gpio(MFP_CFG_PIN(CSEN_RST_N))

//We should add the other pin's request function

static int compass_init_irq(void)
{
	int ret = 0;
	ret = gpio_request(COMPASS_INT_GPIO, "compass irq");
	if (ret){
		printk("gpio_requset failed, return :%d\n",ret);
		goto err_request_cd;
	}
	ret=gpio_direction_input(COMPASS_INT_GPIO);
	if (ret) {
		printk("set gpio pin direction for compass failed, return :%d\n", ret);
		goto err_request_irq;
	}
err_request_irq:
	gpio_free(COMPASS_INT_GPIO);
err_request_cd:
	return ret;
}

static int compass_ack_irq(void)
{
	return 0;
}

static int compass_reset_pin(void)
{
	int ret;
	ret = gpio_request(COMPASS_RESET_GPIO, "compass reset");
	if (ret) {
		printk("compass reset gpio_requset failed, return :%d\n",ret);
		return ret;
	}
	gpio_direction_output(COMPASS_RESET_GPIO, 0);
	msleep(20);
	gpio_direction_output(COMPASS_RESET_GPIO, 1);
	msleep(20);

	return ret;
}

static struct akm8973_platform_data akm8973_data = {
	.init_irq = compass_init_irq,
	.ack_irq = compass_ack_irq,
	.reset_pin = compass_reset_pin,
};
#endif

#if defined(CONFIG_SENSORS_LIS331)
static int lis331_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT)));

	return 0;
}

static int lis331_ack_irq(void)
{
	return 0;
}

static int lis331_get_gpio_value(void)
{
	return gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT))) ;
}

static struct lis331_platform_data lis331_data = {
	.init_irq = lis331_init_irq,
	.ack_irq = lis331_ack_irq,
	.get_gpio_value = lis331_get_gpio_value,
	.platform_init = NULL,
	.power_chips = NULL,
	.xyz_dir = SWITCH_XY,
};
#endif

#if defined(CONFIG_PXA_CAMERA)

#define QCI_CAM_REAR_PWDN_PIN	mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_PDN))
#define QCI_CAM_FRONT_RESET_PIN	mfp_to_gpio(MFP_CFG_PIN(CAM_VGA_PDN))
#define QCI_CAM_REAR_RESET_PIN	mfp_to_gpio(MFP_CFG_PIN(CAM_REAR_RESET_N))
#define QCI_POWER_LDO_EN	mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN))
#define QCI_MCLK_SWITCH		mfp_to_gpio(MFP_CFG_PIN(CAM_MCLK_SEL))

static int cam_sensor_set_powerdown_pin(int flag, int value)
{
	//printk("debug: jades.c cam_sensor_set_reset_pin(%d)\n", value);
	if (gpio_request(QCI_CAM_REAR_PWDN_PIN, "CAM_rear_powerdown_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", QCI_CAM_REAR_PWDN_PIN);
		return -EIO;
	}

	if (gpio_request(QCI_CAM_FRONT_RESET_PIN, "CAM_front_reset_SENSOR")){
		gpio_free(QCI_CAM_REAR_PWDN_PIN);
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d\n", QCI_CAM_FRONT_RESET_PIN);
		return -EIO;
	}

	if (flag)
	{
		gpio_direction_output(QCI_CAM_FRONT_RESET_PIN, 0);	/* LOW */
		if ( value==0 )
			gpio_direction_output(QCI_CAM_REAR_PWDN_PIN, 0);	/* LOW */
		else
			gpio_direction_output(QCI_CAM_REAR_PWDN_PIN, 1);	/* HIGH */
	}
	else
	{
		/*for sub camera, power down rear camera*/
		gpio_direction_output(QCI_CAM_REAR_PWDN_PIN, 1);	/* HIGH */
	}
	
	gpio_free(QCI_CAM_FRONT_RESET_PIN);
	gpio_free(QCI_CAM_REAR_PWDN_PIN);

	return 0;
}

/* set sensor reset pin*/
static int cam_sensor_set_reset_pin(int flag, int value)
{
	if (flag)
	{
		if (gpio_request(QCI_CAM_REAR_RESET_PIN, "CAM_rear_RESET_SENSOR")) {
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d \n", QCI_CAM_REAR_RESET_PIN);
			return -EIO;
		}
		
		if ( value==0 )
			gpio_direction_output(QCI_CAM_REAR_RESET_PIN, 0);	/* LOW */
		else
			gpio_direction_output(QCI_CAM_REAR_RESET_PIN, 1);	/* HIGH */
		gpio_free(QCI_CAM_REAR_RESET_PIN);
	}
	else
	{
		if (gpio_request(QCI_CAM_FRONT_RESET_PIN, "CAM_front_reset_SENSOR")){
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d\n", QCI_CAM_FRONT_RESET_PIN);
			return -EIO;
		}
		
		if ( value==0 )
			gpio_direction_output(QCI_CAM_FRONT_RESET_PIN, 0);	/* LOW */
		else
			gpio_direction_output(QCI_CAM_FRONT_RESET_PIN, 1);	/* HIGH */
		gpio_free(QCI_CAM_FRONT_RESET_PIN);
	}
	

	return 0;
}

/* sensor init */
static int sensor_power_on(int flag)
{
	pxa3xx_pmic_set_voltage(VCC_CAMERA_ANA,2800);
	pxa3xx_pmic_enable_voltage(VCC_CAMERA_ANA,1);
	msleep(5);
	/*
	 * flag, 0, low resolution
	 * flag, 1, high resolution
	 */
	if (gpio_request(QCI_POWER_LDO_EN, "CAM_EANBLE_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d\n", QCI_POWER_LDO_EN);
		return -EIO;
	}

	if (gpio_request(QCI_MCLK_SWITCH, "CAM_MCLK_SWITCH")){
		gpio_free(QCI_POWER_LDO_EN);
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d\n", QCI_MCLK_SWITCH);
		return -EIO;
	}

	gpio_direction_output(QCI_POWER_LDO_EN, 0);
	cam_sync_from_gpio();
	
	if(flag){
		gpio_direction_output(QCI_MCLK_SWITCH, 1);
	}else{
		gpio_direction_output(QCI_MCLK_SWITCH, 0);
	}
	cam_sensor_set_reset_pin(flag, 0);
	cam_sensor_set_powerdown_pin(flag, 0);
	
	gpio_direction_output(QCI_POWER_LDO_EN, 1);
	msleep(20);

	cam_sensor_set_reset_pin(flag, 1);
	msleep(20);	

	gpio_free(QCI_POWER_LDO_EN);
	gpio_free(QCI_MCLK_SWITCH);

	return 0;
}

static int sensor_power_off(int flag)
{
	/*
	 * flag, 0, low resolution
	 * flag, 1, high resolution
	 */
	cam_sensor_set_reset_pin(flag, 0);

	if (gpio_request(QCI_POWER_LDO_EN, "CAM_EANBLE_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", QCI_POWER_LDO_EN);
		return -EIO;
	}

	gpio_direction_output(QCI_POWER_LDO_EN, 0);
	cam_sync_to_gpio();
	
	gpio_free(QCI_POWER_LDO_EN);
	
	pxa3xx_pmic_enable_voltage(VCC_CAMERA_ANA,0);
	return 0;
}

#if defined(CONFIG_PXA_CAMERA_MICRON_MT9P111)
static struct sensor_platform_data mt9p111_sensor_data = {
	.id = SENSOR_HIGH,
	.power_on = sensor_power_on,
	.power_off = sensor_power_off,
};
#endif
#if defined(CONFIG_PXA_CAMERA_MICRON_MT9V113)
static struct sensor_platform_data mt9v113_sensor_data = {
	.id = SENSOR_LOW,
	.power_on = sensor_power_on,
	.power_off = sensor_power_off,
};
#endif

/* sensor init over */

/* camera platform data */
static mfp_cfg_t sync[] = {
	CAM_HSYNC,
	CAM_VSYNC,
};

static mfp_cfg_t sync_gpio[] = {
	CAM_HSYNC_GPIO,
	CAM_VSYNC_GPIO,
};

static void cam_sync_to_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync_gpio));
}

static void cam_sync_from_gpio(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(sync));
}

static int cam_init(void)
{
	printk(KERN_ERR"[%s:%d] %s()\n", __FILE__, __LINE__, __func__);
	return 0;
}

static void cam_deinit(void)
{
}

static void cam_suspend(void)
{
	sensor_power_off(0);
}

static void cam_resume(void)
{
	sensor_power_on(0);
}

static struct cam_platform_data cam_ops = {
	.vsync_gpio		= mfp_to_gpio(MFP_CFG_PIN(CAM_VSYNC_GPIO)),
	.init			= cam_init,
	.deinit			= cam_deinit,
	.suspend		= cam_suspend,
	.resume			= cam_resume,
	.sync_to_gpio		= cam_sync_to_gpio,
	.sync_from_gpio		= cam_sync_from_gpio,
};

static void __init init_cam(void)
{
	pxa3xx_device_cam.dev.platform_data = &cam_ops;
	platform_device_register(&pxa3xx_device_cam);
}
/* QCI init over */
#endif

static struct i2c_board_info i2c_board_info[] = {
#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
	{
		.type		= "micco",
		.addr		= 0x34,
		.platform_data	= &micco_data,
		.irq		= IRQ_GPIO(PMIC_INT_PIN),
	},
#endif
#if defined(CONFIG_PXA_CAMERA)
#if defined(CONFIG_PXA_CAMERA_MICRON_MT9P111)
	{
		.type	= "sensor_mt9p111",
		.addr		= 0x3d,
		.platform_data	= &mt9p111_sensor_data,
	}, 
#endif 
#if defined(CONFIG_PXA_CAMERA_MICRON_MT9V113) 
	{
		.type	= "sensor_mt9v113",
		.addr		= 0x3c,
		.platform_data	= &mt9v113_sensor_data,
	}, 
#endif 
#endif
#if defined(CONFIG_TOUCHSCREEN_ATMEL)
	{
		.type	= ATMEL_I2C_RMI_NAME,
		.addr		= 0x4a,
		.platform_data	= &atmel_touch_data,
		.irq = IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
	}, 
#endif

#if defined(CONFIG_SENSORS_AK8973)
	{
		.type	= "akm8973",
		.addr	= 0x1F,
		.platform_data	= &akm8973_data,
		.irq		= COMPASS_IRQ,
	},
#endif
#if defined(CONFIG_SENSORS_LIS331)
	{
		.type		= "lis331",
		.addr		= 0x1C,
		.platform_data	= &lis331_data,
		.irq	= IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(GSEN_INT))),
	},
#endif
#if defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2018) || defined(CONFIG_SND_PXA3XX_SOC_AEC_FM2018_MODULE)
    {
        .type       = "fm2018", 
        .addr       = 0x60, 
        .platform_data  = &g_aec_platform_data, 
    },
#endif
};

#if defined(CONFIG_TOUCHSCREEN_ATMELQT_OBP)
static struct i2c_board_info i2c_board_info_atmelqt_obp[] = {
	{
		I2C_BOARD_INFO(MAXTOUCH_TS_NAME, 0x4A),
		.platform_data = &t8200_ts_platform_data,
		.irq = IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
	},
};
#endif

#if defined(CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539)
static struct i2c_board_info i2c_board_info_synaptics[] = {

	{
		I2C_BOARD_INFO(SYNAPTICS_I2C_RMI_NAME, 0x4A),
		.platform_data = &t8200_ts_synaptics_platform_data,
		.irq = IRQ_GPIO(mfp_to_gpio(MFP_CFG_PIN(TP_INT))),
	},

};
#endif

#if defined(CONFIG_MRVL8688) || defined(CONFIG_MRVL8688_MODULE)
static mfp_cfg_t wlan_lpm_gpio_on_mfp_cfg[] = {
    WLAN_PON_LPM_ON,
    WLAN_BT_RESET_LPM_ON,
    WLAN_WAKEUP_AP_LPM_ON,
    AP_WAKEUP_BT_LPM_ON,
    BT_TXD_LPM_ON,
    BT_RXD_LPM_ON,
    BT_CTS_LPM_ON,
    BT_RTS_LPM_ON,
    BT_WAKEUP_AP_LPM_ON,
    WLAN_DAT3_LPM_ON,
    WLAN_DAT2_LPM_ON,
    WLAN_DAT1_LPM_ON,
    WLAN_DAT0_LPM_ON,
    WLAN_CMD_LPM_ON,
    WLAN_CLK_LPM_ON,
    AP_WAKEUP_WLAN_LPM_ON,
};

static mfp_cfg_t wlan_lpm_gpio_off_mfp_cfg[] = {
    WLAN_PON,
    WLAN_BT_RESET,
    WLAN_WAKEUP_AP,
    AP_WAKEUP_BT,
    BT_TXD,
    BT_RXD,
    BT_CTS,
    BT_RTS,
    BT_WAKEUP_AP,
    WLAN_DAT3,
    WLAN_DAT2,
    WLAN_DAT1,
    WLAN_DAT0,
    WLAN_CMD,
    WLAN_CLK,
    AP_WAKEUP_WLAN,
//    AP_PCM_CLK_OFF,
//    AP_PCM_FRM_OFF,
//    AP_PCM_TXD_OFF,
//    AP_PCM_RXD_OFF,
};

static void wlan_gpio_setpower(int enable)
{
	unsigned int pcm_clk, pcm_sync, pcm_rxd, pcm_txd;

	if (enable) {
		enable_oscc_tout_s0();
		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_on_mfp_cfg));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)));
		gpio_direction_input(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 1);
		msleep(1);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
		msleep(10);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 1);
		msleep(1);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), 1);

		printk(KERN_DEBUG "wlan_lpm_gpio_on\n");
	}
	else {
		pcm_clk = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_CLK));
		pcm_sync = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_FRM));
		pcm_txd = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_TXD));
		pcm_rxd = pxa3xx_mfp_read(MFP_CFG_PIN(AP_PCM_RXD));

		pxa3xx_mfp_config(ARRAY_AND_SIZE(wlan_lpm_gpio_off_mfp_cfg));
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)),0);
        	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)),0);

		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_CLK)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_FRM)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_TXD)),0);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_PCM_RXD)),0);

 		msleep(10);

		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_CLK), pcm_clk);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_FRM), pcm_sync);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_TXD), pcm_txd);
		pxa3xx_mfp_write(MFP_CFG_PIN(AP_PCM_RXD), pcm_rxd);

		disable_oscc_tout_s0();
		printk(KERN_DEBUG "wlan_lpm_gpio_off\n");
	}
}

static struct wlan_platform_data wlan_pdata = {
	.chip_gpio = 1,
	.chip_gpio_setpower = wlan_gpio_setpower,
        .bt_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP))),
        .wlan_chip_wakeup_irq = gpio_to_irq(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP))),
}; 

static struct platform_device wlan_device = {
	.name		= "mrvl8688",
	.id		= -1,
};

static void __init init_wlan(void)
{

	int err = 0;

	printk(KERN_ERR "WLAN set up\n");

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), "wlan pmic powerdown");	
	if (err) {
		printk(KERN_ERR "WIFI_PMU_PON Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), "wlan powerdown");
	if (err) {
		printk(KERN_ERR "WIFI_PMU_PON Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), "wlan bt reset");	
	if (err) {
		printk(KERN_ERR "WLAN_BT_RESET Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), "wlan mac wakeup");	
	if (err) {
		printk(KERN_ERR "AP_WAKEUP_WLAN Request error!\n");
		return;
	}

	err = gpio_request(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)), "bt wakeup tavor");  
	if (err) {
		printk(KERN_ERR "BT_WAKEUP_AP Request error!\n");
		return;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_BT_RESET)), 0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_WLAN)), 0);

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_RXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_TXD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_CTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_RTS)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT3)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT2)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT1)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_DAT0)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CLK)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_CMD)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP)),0);
	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP)),0);
//	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_ONE_WIRE)),0);

	wlan_device.dev.platform_data = &wlan_pdata;
	platform_device_register(&wlan_device);

	printk(KERN_ERR "WLAN set up finished\n");

	return;
}
#endif

#if defined(CONFIG_BT_HCIUART_PS)

#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
static void switch_bt_chip_lpm_state(int state)
{
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT)), 
			      !!state);
        pxa3xx_mfp_set_lpm(MFP_CFG_PIN(AP_WAKEUP_BT),
			   !!state ? MFP_LPM_DRIVE_HIGH : MFP_LPM_DRIVE_LOW);
}
#endif

static struct hci_ps_platform_data hci_ps_pdata = {
	.chip_type = CHIP_MRVL8688,
	.chip_setpower = NULL,
}; 

static struct platform_device hci_ps_device = {
	.name		= "hci_ps",
	.id		= -1,
};

static void __init hci_ps_init(void)
{
#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
	hci_ps_pdata.chip_setpower = switch_bt_chip_lpm_state;
#endif
	hci_ps_device.dev.platform_data = &hci_ps_pdata;
	platform_device_register(&hci_ps_device);
}
#endif

struct pxa3xx_freq_mach_info freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};


/* use bit 30, 31 as the indicator of command parameter number */
#define CMD0(x)     ((0x00000000) | ((x)))
#define DAT0(x)		((0x00000100) | ((x)))
#define TPO_L1N304_LCD_DSS 	9 	//Data size select for LCD spi

static uint32_t initialize[] = {
	CMD0(0xd0),
	DAT0(0x02),
	DAT0(0x41),
	DAT0(0x16),

	CMD0(0xd1),
	DAT0(0x00),
	DAT0(0x1e),
	DAT0(0x1f),

	CMD0(0xd2),
	DAT0(0x01),
	DAT0(0x22),

	CMD0(0xc0),
	DAT0(0x00),
	DAT0(0x3b),
	DAT0(0x00),
	DAT0(0x02),
	DAT0(0x11),

	CMD0(0xc5),
	DAT0(0x03),

	CMD0(0xc8),
	DAT0(0x02),
	DAT0(0x14),
	DAT0(0x30),
	DAT0(0x60),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x76),
	DAT0(0x36),
	DAT0(0x75),
	DAT0(0x00),
	DAT0(0x0f),
	DAT0(0x01),

	CMD0(0xf8),
	DAT0(0x01),

	CMD0(0xfe),
	DAT0(0x00),
	DAT0(0x02),

	CMD0(0x36),
	DAT0(0x0a),

	CMD0(0xc6),
	DAT0(0x02),
};

static uint32_t initialize2[] = {
	CMD0(0x2a),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x01),
	DAT0(0x3f),
    
	CMD0(0x2b),
	DAT0(0x00),
	DAT0(0x00),
	DAT0(0x01),
	DAT0(0xdf),

	CMD0(0x29),

	CMD0(0x2c),
};

static uint32_t rgb_setting[] = {
	CMD0(0x0c),
	DAT0(0x00),
	DAT0(0x66),

	CMD0(0x3a),
	DAT0(0x66),

	CMD0(0xb4),
	DAT0(0x11),
};

static uint32_t sleep_out[] = {
	CMD0(0x11),
};

static uint32_t sleep_in[]={
	CMD0(0x10),
};

static uint32_t disp_on[]={
	CMD0(0x11),

	CMD0(0x29),
};

static uint32_t disp_off[]={
	CMD0(0x28),
};


static void ssp_reconfig(struct ssp_dev *dev, int nparam)
{
	static int last_nparam = -1;
	int sscr0;

	/* check if it is necessary to re-config SSP */
	if (nparam == last_nparam)
		return;

	ssp_disable(dev);
	if( nparam >16 )
		sscr0 = 0x00107f80 | ((nparam-1) & 0xf);
	else
		sscr0 = 0x00007f80 | ((nparam-1) & 0xf);

	ssp_config(dev, sscr0, 0x18, 0, 0);
	last_nparam = nparam;
}

static void ssp_send_cmd(uint32_t *cmd, int num)
{
	static int ssp_initialized = 0;
	static struct ssp_dev ssp4;

	int i;

	if (!ssp_initialized) {
		ssp_init(&ssp4, 4, SSP_NO_IRQ);
		ssp_initialized = 1;
		clk_disable(ssp4.ssp->clk);
	}

	clk_enable(ssp4.ssp->clk);
	for (i = 0; i < num; i++, cmd++) {
		ssp_reconfig(&ssp4, TPO_L1N304_LCD_DSS);
		ssp_write_word(&ssp4, *cmd & 0x3fffffff);

		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp4);
	}
	clk_disable(ssp4.ssp->clk);
}

static void backlight_power(int on, int intensity)
{
	unsigned int val;
	u8 wled[3];

	micco_read_block(MICCO_LEDPC_CONTROL5,wled, 3);
	val = intensity * 3 / 8 +1;
	if( on )
	{/* Turn lcd backlight on */
		printk(KERN_DEBUG "[10]Backlight is on, intensity=%d\n",val);
		if( !val )
			return;
		wled[0] = (wled[0] & 0x0f)| 0x40;	/* set boost converter load regulation to 9.17 */
		wled[1] = val;
		wled[2] = 0x30;
		micco_write_block(MICCO_LEDPC_CONTROL5, wled, 3);
	}else
	{/* Turn lcd backlight off */
		if( !(wled[2] & 0x20 ) )
		{
			printk(KERN_WARNING "[10]Backlight is off while request to turn off\n");
			return;
		}
		wled[1]=wled[2]=0;
		wled[0] &= 0x0f;
		micco_write_block(MICCO_LEDPC_CONTROL5, wled, 3);
		printk(KERN_DEBUG "[10]Backlight is off\n");
	}
}

#ifdef CONFIG_LCD_RAMP
void micco_ramp_enable(int on, int enable, int intensity)
{
	int wlv;
	micco_read(MICCO_WLED_CONTROL1, &wlv);

	if(enable == 1)
		micco_write(MICCO_WLED_CONTROL1, wlv | (1<<7));
	else
		micco_write(MICCO_WLED_CONTROL1, wlv & 0x7f);

	backlight_power(on, intensity);
}
#endif

static void lcd_power(int on, struct fb_var_screeninfo *var)
{
	static int pin_requested = 0;

	if( !pin_requested) {
		if( gpio_request( mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 
			"LCD controller reset"))
			return;
		pin_requested = 1;
	}

	if( on)
	{
		pxa3xx_pmic_set_voltage(VCC_LCD, 1800);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 0);
        	udelay(100);
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(LCD_RESET_N)), 1);
        	ssp_send_cmd(ARRAY_AND_SIZE(sleep_out));
		msleep(120);
		ssp_send_cmd(ARRAY_AND_SIZE(initialize));
		ssp_send_cmd(ARRAY_AND_SIZE(initialize2));
		ssp_send_cmd(ARRAY_AND_SIZE(rgb_setting));
	}
	else
	{
		ssp_send_cmd(ARRAY_AND_SIZE(disp_off));
		ssp_send_cmd(ARRAY_AND_SIZE(sleep_in));
	}
}

static struct pxafb_mode_info RM68040_hvga_mode[] = {
    [0] = {
        .pixclock               = 98000,//modify 2010/0205
        .xres                   = 320,
        .yres                   = 480,
        .bpp                    = 32,

        .hsync_len              = 8,
        .left_margin            = 5,
        .right_margin           = 10,
        .vsync_len              = 1,
        .upper_margin           = 2,
        .lower_margin           = 5,
        .sync                   = 0,
	},
};

static struct pxafb_mach_info lcd_info = {
	.modes			= RM68040_hvga_mode,
	.num_modes		= 1,
	.lccr0			= LCCR0_Act,
	.lccr3			= LCCR3_PixRsEdg | LCCR3_VSP,
	.pxafb_lcd_power	= lcd_power,
	.pxafb_backlight_power	= backlight_power,
#ifdef CONFIG_LCD_RAMP
	.ramp_enable		= micco_ramp_enable,
#endif
};

static void __init init_lcd(void)
{
	set_pxa_fb_info(&lcd_info);
}

#if defined(CONFIG_LEDS_PXA)
static void leds_brightness_set(int value)
{
	if (value) {/* Turn keyboard backlight on */
		micco_write(MICCO_LED1_CONTROL, 0x01);
	} else { /* Turn keyboard backlight off */
		micco_write(MICCO_LED1_CONTROL, 0); 
	}
}

static struct leds_platform_data leds_info = {
	.set_button_brightness 	= leds_brightness_set,
};
#endif

#if defined(CONFIG_LEDS_DEBUG_SLEEP)
static void sleep_led_brightness_set (int value)
{
	return;
}

static struct sleep_led_platform_data sleep_led_info = {
	.set_brightness 	= sleep_led_brightness_set,
};
#endif

#ifdef CONFIG_PM
static int init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
/*	src->bits.uart1 = 1; */
	src->bits.uart2 = 1;
	src->bits.dkey = 1;
	src->bits.onkey = 1;
	src->bits.mmc2_dat1 = 1; //Enable mmc2 wakeup 
	src->bits.usbc = 1;
//	src->bits.cmwdt = 1;
	src->bits.psensor = 1;
	src->bits.modem = 1;

	return 0;
}

extern int pxa930_query_gwsr(int);
static int query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_PMIC)
		src->bits.onkey = 1;
	if (reg & PXA3xx_PM_WE_UART1)
		src->bits.uart1 = 1;
	if (reg & PXA3xx_PM_WE_BTUART)
		src->bits.uart2 = 1;
	if (reg & PXA3xx_PM_WE_DKEY)
		src->bits.dkey = 1;
	if (reg & PXA3xx_PM_WE_MKEY)
		src->bits.mkey = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GPIO) {
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(PSEN_DOUT))))
			src->bits.psensor = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(DPRAM_INT_N))) || pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(BP_READY))))
			src->bits.modem = 1;
		if (pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(BT_WAKEUP_AP))))
		        src->bits.uart2 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(WLAN_WAKEUP_AP))))
			src->bits.mmc2_dat1 = 1;
		if ( pxa930_query_gwsr(mfp_to_gpio(MFP_CFG_PIN(CHARGE_PG_N))))
			src->bits.usbc = 1;
	}

	return 0;
}

static int ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.dkey) {
			ret |= PXA3xx_PM_WE_DKEY;
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_PMIC;
		}
	} else {
		if (src.bits.dkey) {
		}
		if (src.bits.onkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_MICCO),
					    MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int mmc_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
    int value;

    value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(WIFI_PMU_PON))) ? 1 : 0;
	if( enable )
	{
		if( src.bits.mmc2_dat1 && 1 == value )
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_AP), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}
	else
	{
		if( src.bits.mmc2_dat1 && 1 == value )
		{
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(WLAN_WAKEUP_AP), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int psensor_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PSEN_DOUT),MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}else{
		if( src.bits.psensor) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PSEN_DOUT),MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int modem_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if( src.bits.modem) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(DPRAM_INT_N),MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BP_READY),MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	}else{
		if( src.bits.modem) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(DPRAM_INT_N),MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BP_READY),MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(FF_RXD), MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_UART1;
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_AP), MFP_EDGE_FALL);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_RXD), MFP_EDGE_FALL);
			/* note: on pxa930, uart2 use this bit */
			ret |= PXA3xx_PM_WE_BTUART|PXA3xx_PM_WE_GPIO;
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(FF_RXD), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_WAKEUP_AP), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(BT_RXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int usb_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;

	if (enable) {
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(CHARGE_PG_N), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GPIO;
		}
	} else {	
		if (src.bits.usbc) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(CHARGE_PG_N), MFP_EDGE_NONE);
		}
	}
	return ret;

}

static int comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= init_wakeup,
	.query	= query_wakeup,
	.ext    = ext_wakeup,
	.key    = key_wakeup,
	.mmc    = mmc_wakeup,
	.uart   = uart_wakeup,
	.usb	= usb_wakeup,
	.cmwdt	= comm_wdt_wakeup,
	.psensor = psensor_wakeup,
	.modem = modem_wakeup,
};
#endif

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
	.flags		= PXA_I2C_STANDARD_MODE, 
};

#if defined(CONFIG_INNODEV_IF202) || defined(CONFIG_INNODEV_IF202_MODULE)

#define INNO_SPI_CLK_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_CLK))
#define INNO_SPI_CS_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_SFRM))
#define INNO_SPI_TXD_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_TXD))
#define INNO_SPI_RXD_PIN	mfp_to_gpio(MFP_CFG_PIN(CMMB_SPI_RXD))
#define INNO_EN_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_PWR_EN))
#define INNO_INTR_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_IRQ))
#define INNO_RESET_PIN		mfp_to_gpio(MFP_CFG_PIN(CMMB_RST_N))

static mfp_cfg_t if202_ssp_mfp_cfg[] = {
	CMMB_SPI_CLK,
	CMMB_SPI_SFRM,
	CMMB_SPI_TXD,
	CMMB_SPI_RXD,
};

static mfp_cfg_t if202_ssp_gpio_cfg[] = {
	CMMB_SPI_CLK_GPIO,
	CMMB_SPI_SFRM_GPIO,
	CMMB_SPI_TXD_GPIO,
	CMMB_SPI_RXD_GPIO,
};

static int inno_if202_ssp_gpio_input(void)
{
	/* config ssp pins to gpio input to save power */
	gpio_direction_output(INNO_SPI_CLK_PIN,0);
	gpio_direction_output(INNO_SPI_CS_PIN,0);
	gpio_direction_output(INNO_SPI_TXD_PIN,0);
	gpio_direction_output(INNO_SPI_RXD_PIN,0);

	pxa3xx_mfp_config(ARRAY_AND_SIZE(if202_ssp_gpio_cfg));

	return 0;
}

static int inno_if202_gpio_init(void)
{
        if (gpio_request(INNO_SPI_CLK_PIN, "INNO_CLK")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_CLK_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_CS_PIN, "INNO_CS")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_CS_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_TXD_PIN, "INNO_TXD")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_TXD_PIN);
                return -EIO;
        }
        if (gpio_request(INNO_SPI_RXD_PIN, "INNO_RXD")) {
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d \n", INNO_SPI_RXD_PIN);
                return -EIO;
        }

	if (gpio_request(INNO_INTR_PIN, "INNO_INTR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_INTR_PIN);
		return -EIO;
	}

	if (gpio_request(INNO_RESET_PIN, "INNO_RESET")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_RESET_PIN);
		return -EIO;
	}
	/* config reset pin to gpio input to save power */

	if (gpio_request(INNO_EN_PIN, "INNO_ENABLE")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", INNO_EN_PIN);
		return -EIO;
	}

	inno_if202_ssp_gpio_input();

	gpio_direction_output(INNO_RESET_PIN,0);

	gpio_direction_output(INNO_INTR_PIN,0);

	gpio_direction_output(INNO_EN_PIN, 0);	/* power voltage off */

	return 0;
}

int inno_if202_voltage_on(void)
{
	/* The if101 chip may not power up correctly if the connected pins are high,
	 * so config ssp pins to gpio input before power up */
	inno_if202_ssp_gpio_input();
	gpio_direction_input(INNO_INTR_PIN);

	gpio_direction_output(INNO_RESET_PIN, 0);


	/* enable VDD (1.8V) & VIO (2.8V) */
	gpio_direction_output(INNO_EN_PIN, 1);

	udelay(50);

	gpio_direction_output(INNO_RESET_PIN, 1);

	msleep(400);	/* TBD */

	/* reconfig ssp pins to SSP after power up */
	pxa3xx_mfp_config(ARRAY_AND_SIZE(if202_ssp_mfp_cfg));

	return 0;
}

int inno_if202_voltage_off(void)
{
	gpio_direction_output(INNO_EN_PIN, 0);

	/* config reset pin to gpio input to save power */
	gpio_direction_output(INNO_RESET_PIN,0);

	gpio_direction_output(INNO_INTR_PIN,0);
	/* config ssp pins to gpio input to save power */
	inno_if202_ssp_gpio_input();

	return 0;
}

static int inno_if202_reset(void)
{
	gpio_direction_output(INNO_RESET_PIN, 0);
	mdelay(1);
	gpio_direction_output(INNO_RESET_PIN, 1);
	msleep(1);
	return 0;
}

static struct innodev_platform_spi inno_spi_data = {
	.ssp_port = 1,
	.spi_cs_gpio = INNO_SPI_CS_PIN,
};

static struct platform_device inno_spi_device = {
	.name 		= "inno-spi",
	.id 		= -1,
};

static struct innodev_platform_data inno_if202_data = {
	.voltage_on = inno_if202_voltage_on,
	.voltage_off = inno_if202_voltage_off,
	.reset = inno_if202_reset,
	.fw_spi_download = 1,
	.irq = IRQ_GPIO(INNO_INTR_PIN),
};

static struct platform_device inno_if202_device = {
	.name 		= "inno-dev",
	.id 		= -1,
};

static void __init inno_if2xx_init(void)
{
	inno_if202_gpio_init();

	inno_if202_data.fw_name = "inno_if208.fw";

	inno_if202_device.dev.platform_data = &inno_if202_data;
	platform_device_register(&inno_if202_device);

	inno_spi_device.dev.platform_data = &inno_spi_data;
	platform_device_register(&inno_spi_device);
}
#else
static inline void inno_if2xx_init(void) {}
#endif /* #ifdef CONFIG_INNODEV_V2 */

/**************************************************
 ** speaker_read: read the speaker PA status
 ** 1: enable
 ** 0: disable
 ***************************************************/ 
static ssize_t speaker_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	int value;

	value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN))) ? 1 : 0;

	if ((count = sprintf(buf, "%d", value)) < 0)
		return -EFAULT;

	return count;
}

/**************************************************
 ** speaker_write: set the speaker PA status  
 ** 1: enable.
 ** 0: disable.
 ***************************************************/ 
static ssize_t speaker_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
	char value[2];
	int ret;

	if ((count = sscanf(buf, "%1s", value)) < 0)
		return -EFAULT;

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), "AUDIO_PA_EN");
	if (ret)
	{
		gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));
		printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_PA_EN", 
                                                      mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), ret);
		return -1;
	}

	if (value[0] == '1')
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 1);
		pxa3xx_mfp_set_lpm(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), MFP_LPM_DRIVE_HIGH);
	}
	else if (value[0] == '0')
	{
		gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 0);
		pxa3xx_mfp_set_lpm(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), MFP_LPM_DRIVE_LOW);
	}

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	return count;
}

static struct bin_attribute speaker_mode_attr = {
	.attr = {
		.name = "audio_pa_en",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = 4,
	.read = speaker_read,
	.write = speaker_write,
};

static int speaker_amp_init(void)
{
	int ret = 0;

	ret = sysfs_create_bin_file(kernel_kobj, &speaker_mode_attr);
	if (ret)
	{
		printk(KERN_ERR "<%s> Failed to create sys file\n", "AUDIO_PA_EN");
		return -ENOMEM;
	}

	ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), "AUDIO_PA_EN");
	if (ret)
	{
		gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));
		printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "AUDIO_PA_EN", 
                                              mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), ret);
		return -1;
	}

	gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)), 0);

	gpio_free(mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	printk(KERN_INFO "<%s> GPIO%d setup done\n", "AUDIO_PA_EN", mfp_to_gpio(MFP_CFG_PIN(AUDIO_PA_EN)));

	return 0;
}

//headset enable seting.

/**************************************************
 ** headphone_enable_read: read the headset status.
 ** 0: disable
 ** 1: enable
 ***************************************************/ 
static ssize_t headphone_enable_read(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    int value;

    value = gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN))) ? 1 : 0;
                        
    if ((count = sprintf(buf, "%d", value)) < 0)
        return -EFAULT;
                                
    return count;
}

/**************************************************
 ** headphone_enable_write: set the headset enable.  
 ** 0: close headset ear pop.
 ** 1: open headset ear pop.
 ***************************************************/ 
static ssize_t headphone_enable_write(struct kobject *kobj, struct bin_attribute *attr, char *buf, loff_t off, size_t count)
{
    char value[2];
    int ret;

    if ((count = sscanf(buf, "%1s", value)) < 0)
        return -EFAULT;

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), "HEADPHONE_EN");
    if (ret)
    {
        gpio_free(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)));
        printk(KERN_ERR "<%s> [write] Request GPIO failed,""gpio: %d return :%d\n", "HEADPHONE_EN", 
                                                      mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), ret);
        return -1;
    }

    if (value[0] == '1')
    {
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), 1);
        pxa3xx_mfp_set_lpm(MFP_CFG_PIN(HEADPHONE_EN), MFP_LPM_DRIVE_HIGH);
    }
    else if (value[0] == '0')
    {
        gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), 0);
        pxa3xx_mfp_set_lpm(MFP_CFG_PIN(HEADPHONE_EN), MFP_LPM_DRIVE_LOW);
    }

    gpio_free(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)));

    return count;
}

static struct bin_attribute headset_mode_attr = {
    .attr = {
            .name = "headphone_en",
            .owner = THIS_MODULE,
            .mode = S_IRUGO | S_IWUGO,
            },
            .size = 4,
            .read = headphone_enable_read,
            .write = headphone_enable_write,
};

static int headphone_enable_init(void)
{
    int ret = 0;

    ret = sysfs_create_bin_file(kernel_kobj, &headset_mode_attr);
    if (ret)
    {
        printk(KERN_ERR "<%s> Failed to create sys file\n", "HEADPHONE_EN");
        return -ENOMEM;
    }

    ret = gpio_request(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), "HEADPHONE_EN");
    if (ret)
    {
        gpio_free(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)));
        printk(KERN_ERR "<%s> Request GPIO failed,""gpio: %d return :%d\n", "HEADPHONE_EN", 
                                              mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), ret);
        return -1;
    }

    gpio_direction_output(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)), 0);

    gpio_free(mfp_to_gpio(MFP_CFG_PIN(HEADPHONE_EN)));

    printk(KERN_INFO "<%s> GPIOs setup done\n", "HEADPHONE_EN");

    return 0;
}

#ifdef CONFIG_ANDROID_RAM_CONSOLE
#define RAM_CONSOLE_START   0xA8000000
#define RAM_CONSOLE_SIZE    0x20000
static struct resource ram_console_resource = {
	.start  = RAM_CONSOLE_START,
	.end    = (RAM_CONSOLE_START + RAM_CONSOLE_SIZE - 1),
	.flags  = IORESOURCE_MEM,
};

static struct platform_device ram_console_device = {
	.name = "ram_console",
	.id = 0,
	.num_resources  = 1,
	.resource       = &ram_console_resource,
};

static inline void ramconsole_init(void)
{
	platform_device_register(&ram_console_device);
}

static inline void pxa_ramconsole_reserve_sdram(void)
{
	reserve_bootmem(RAM_CONSOLE_START, RAM_CONSOLE_SIZE, 0);
}
#else
static inline void ramconsole_init(void) {}

static inline void pxa_ramconsole_reserve_sdram(void) {}
#endif

static void __init map_io(void)
{
	pxa_ramconsole_reserve_sdram();
	pxa_map_io();
}

static void __init init_machine(void)
{
	/* initialize MFP configurations */
	pxa3xx_mfp_config(ARRAY_AND_SIZE(mfp_cfg));
	pxa3xx_mfp_set_lpm(MFP_PIN_DF_nCS1, MFP_LPM_DRIVE_HIGH);

	misc_mfp_config();
	/* register i2c device */
	pxa_set_i2c_info(&pxa3xx_i2c_platform_data);

	t8200_touch_init();

	/* register gpio i2c device */
	i2c_gpio_init();

	/* dvfm device */
	set_pxa3xx_freq_info(&freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

	ramconsole_init();

	/* lcd */
	init_lcd();

	/* GPS */
#if defined(CONFIG_SIRF_GSD3TW)
	gps_sirf_init();
#endif
	/* WLAN */
#if defined(CONFIG_MRVL8688) || defined(CONFIG_MRVL8688_MODULE)
	init_wlan();
#endif
	platform_add_devices(devices, ARRAY_SIZE(devices));
#ifdef CONFIG_ANDROID_PMEM
	/* PMEM size 24M */
	android_add_pmem("pmem", 0x01800000UL, 1, 1);
#endif
	/* uart */
	pxa_set_ffuart_info(NULL);
	pxa_set_btuart_info(NULL);
	pxa_set_stuart_info(NULL);
#if defined(CONFIG_BT_HCIUART_PS)
	hci_ps_init();
#endif
#if defined(CONFIG_LEDS_PXA)
	pxa_set_leds_info(&leds_info);
#endif
#if defined(CONFIG_LEDS_DEBUG_SLEEP)
	pxa_set_sleep_led_info(&sleep_led_info);
#endif
	/* onenand */
#if defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE)
	init_onenand();
#endif
#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &u2o_info);
	}
#endif
#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
	pxa_set_keypad_info(&keypad_info);
#endif

#if defined(CONFIG_TOUCHSCREEN_ATMELQT_OBP) && defined(CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539)
	g_touch_id = touch_id();
#endif
	/* i2c devices */
	i2c_register_board_info(0, ARRAY_AND_SIZE(i2c_board_info));
#if defined(CONFIG_PXA_CAMERA)
	/* initialize camera, shall do it after enable buck3 */
	init_cam();
#endif

#if defined(CONFIG_TOUCHSCREEN_ATMELQT_OBP) && defined(CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539)
	if (0x80 == g_touch_id)
	{
		i2c_register_board_info(0, ARRAY_AND_SIZE(i2c_board_info_atmelqt_obp));
	}
	else
	{
		i2c_register_board_info(0, ARRAY_AND_SIZE(i2c_board_info_synaptics));
	}
#else
#if defined(CONFIG_TOUCHSCREEN_ATMELQT_OBP)
	i2c_register_board_info(0, ARRAY_AND_SIZE(i2c_board_info_atmelqt_obp));
#endif
#if defined(CONFIG_TOUCHSCREEN_SYNAPTICS_I2C_RMI_TM1539)
	i2c_register_board_info(0, ARRAY_AND_SIZE(i2c_board_info_synaptics));
#endif
#endif

	inno_if2xx_init();
#if defined(CONFIG_MMC)
	/* MMC card detect & write protect for controller 0 */
	mmc_slot[0].gpio_cd  = mfp_to_gpio(MFP_CFG_PIN(MMC1_CD_N));
	init_mmc();
#endif
#if defined(CONFIG_ALSPS)
	/* cm3601_alsps proximity sensor register platform device */
	cm3601_alsps_init();
#endif
#if defined(CONFIG_BQ2407x_POWER)
	init_charger();
#endif
#if defined(CONFIG_SND_PXA3XX_SOC_FBW_TAVORP65_FOR_TD)
	snd_ssp_mfp_cfg_init();
#endif
#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif

    /*Audio part*/
	speaker_amp_init();
	headphone_enable_init();
	headset_init();

#if defined(CONFIG_AP_BP_MUX_IPC) || defined(CONFIG_AP_BP_MUX_IPC_MODULE) 
	BP_init();
#endif
	/* enable clock */
	OSCC |= OSCC_TENS0;
}

MACHINE_START(T8200, "T8200")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = init_machine,
MACHINE_END
