/*
 *  linux/arch/arm/mach-pxa/tavor_common.c
 *
 *  Support for the Marvell Tavor Development Platforms.
 *
 *  Author:	Joseph Hanin (largely modified code)
 *  Created:	May 21, 2009
 *  Copyright:	(C) Copyright 2006 Marvell International Ltd.
 *
*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 */
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>


#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/i2c.h>

#define I2C_CLK		(1UL << (CKEN_I2C - 32))
#define I2C_PMIC_ADDR	0x34


static void __init pmic_i2c_init(void)
{
	/* Enable i2c clock */
	CKENB |= I2C_CLK;

	/* Initialize I2C Unit */
	ISAR = 1;
	ICR = ICR_UR;
	ICR = ICR_SCLE | ICR_IUE;
}

static void __init pmic_i2c_deinit(void)
{
	ICR = ICR_UR;
	CKENB &= ~(I2C_CLK);
}

static int __init wait_i2c_rx_full(int timeout)
{
	u32 temp;
	while (timeout--) {
		temp = ISR;
		if ((temp & ISR_IRF) == ISR_IRF) {
			ISR = temp | ISR_IRF;
			return 0;
		}
		 msleep(20);
	}
	return 1;
}


static int __init wait_i2c_tx_empty(int timeout)
{
	u32 temp;

	while (timeout--) {
		temp = ISR;
		if ((temp & ISR_ITE) == ISR_ITE) {
			ISR = temp | ISR_ITE;
			if ((temp & ISR_ALD) == ISR_ALD)
				ISR |= ISR_ALD;
			return 0;
		}
		msleep(20);
	}
	return 1;
}

static int __init pmic_i2c_write(int sreg, unsigned char *buf, int len)
{
	int icr;

	IDBR = (sreg << 1);

	icr = ICR & ~(ICR_STOP | ICR_ALDIE);
	ICR = icr | ICR_START | ICR_TB;

	if (wait_i2c_tx_empty(20))
		return -1;

	while (len--) {
		IDBR = (unsigned int) (*buf++);
		icr = (ICR & ~ICR_START) | ICR_ALDIE | ICR_TB;
		if (len == 0)
			icr |= ICR_STOP;

		ICR = icr;
		if (wait_i2c_tx_empty(250))
			return -1;
	}

	ICR &= ~ICR_STOP;
	return 0;
}

static int  __init pmic_i2c_read(int sreg, char *data_buf, int data_count)
{
	u32 reg;

	IDBR = (sreg << 1) | 1;
	reg = ICR;
	reg |= (ICR_START | ICR_TB);
	reg &= ~(ICR_STOP | ICR_ALDIE);
	ICR = reg;

	if (wait_i2c_tx_empty(20))
		return -1;

	while (data_count--) {
		reg = ICR;
		reg &= ~ICR_START;
		reg |= ICR_ALDIE | ICR_TB;
		if (data_count == 0)
			reg |= ICR_ACKNAK | ICR_STOP;
		else
			reg &= ~ICR_ACKNAK;

		ICR = reg;
		if (wait_i2c_rx_full(60))
			return -1;
		reg = IDBR & 0xFF;
		*data_buf++ = (u8) reg;
	}
	ICR &= ~(ICR_STOP | ICR_ACKNAK);
	return 0;
}

int __init pmic_id(void)
{
	static u8 cur_chip_id = 0;
	u8 reg;

	if(cur_chip_id == 0){

	pmic_i2c_init();

	/* set the reg address of the micco version */
	reg = 0x0; /* chip id reg */
	if (pmic_i2c_write(I2C_PMIC_ADDR, &reg, 1)) {
		pr_err("ERROR: cannot write the address of pmic reg!\n");
		return 0;
	}

	/* read the reg address of the micco version */
	if (pmic_i2c_read(I2C_PMIC_ADDR, &cur_chip_id, 1)) {
		pr_err("ERROR: cannot write the address of pmic reg!\n");
		cur_chip_id = 0;
		return 0;
	}

	pmic_i2c_deinit();

	pr_info("pmic id %x!\n", cur_chip_id);

	}

	return (cur_chip_id);
}

int __init touch_id(void)
{
	static u8 cur_chip_id = 0;
	u16 reg;

	pmic_i2c_init();

	/* set the reg address of the micco version */
	reg = 0x0; /* chip id reg */
	if (pmic_i2c_write(0x4A, &reg, 2)) {
		pr_err("ERROR: cannot write the address of pmic reg!\n");
		return 0;
	}

	/* read the reg address of the micco version */
	if (pmic_i2c_read(0x4A, &cur_chip_id, 1)) {
		pr_err("ERROR: cannot write the address of pmic reg!\n");
		cur_chip_id = 0;
		return 0;
	}

	pmic_i2c_deinit();

	pr_info("pmic id %x!\n", cur_chip_id);

	return (cur_chip_id);
}
