/*
 *  linux/arch/arm/mach-pxa/tavor_evb3.c
 *
 *  Support for the Marvell EVB3 Development Platform.
 *
 *  Author:	Jason Chagas (largely modified code)
 *  Created:	Nov 20, 2006
 *  Copyright:	(C) Copyright 2006 Marvell International Ltd.
 *
 *  2007-11-22  modified to align with latest kernel
 *              eric miao <eric.miao@marvell.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  publishhed by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/onenand.h>
#include <mtd/mtd-abi.h>
#include <linux/irq.h>
#include <linux/i2c.h>
#include <linux/i2c/si4703.h>
#include <linux/pda_power.h>
#include <linux/pda_battery.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <mach/pxa-regs.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/mfp-pxa3xx.h>
#include <mach/pxa3xx_nand.h>
#include <mach/pxa27x_keypad.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/ssp.h>
#include <mach/gpio.h>
#include <mach/micco.h>
#include <mach/tavor_evb3.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/saar.h>
#include <mach/irda.h>
#include <mach/pxa3xx_u2d.h>
#include <mach/pxa9xx_u2o.h>
#include <mach/ehci.h>
#include <mach/udc.h>
#include <mach/uart.h>
#include <mach/mmc.h>
#include <mach/pxa3xx_dvfm.h>
#include <mach/imm.h>
#include <mach/pxa950fb.h>
#include <mach/pxa3xx_pm.h>
#include <mach/pxa930_trkball.h>
#include <mach/camera.h>
#include <mach/pmu.h>
#include <mach/part_table.h>
#include <mach/pxa3xx-regs.h>
#include <mach/i2c.h>
#include <linux/proc_fs.h>
#include <mach/levante_micco_pmic_id.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <mach/tavor_common.h>
#include <mach/ssp.h>
#include "devices.h"
#include "generic.h"
#ifdef CONFIG_SD8XXX_RFKILL
#include <linux/sd8x_rfkill.h>
#endif

#define MAX_SLOTS      3

struct platform_mmc_slot_evb3 tavor_evb3_mmc_slot[MAX_SLOTS];

extern int is_android(void);
extern int is_comm_v75(void);

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

 static u8 g_pmic_id;

static mfp_cfg_t __attribute__ ((unused)) audio_mfp_cfg[] __initdata = {

	GPIO63_SSP3_CLK,
	GPIO64_SSP3_FRM,
	GPIO65_SSP3_TXD,
	GPIO66_SSP3_RXD,
	GPIO67_SSP3_SYSCLK,

	GPIO68_SSP2_RXD,
	GPIO69_SSP2_TXD,
	GPIO70_SSP2_SYSCLK,
	GPIO71_SSP2_FRM,
	GPIO72_SSP2_CLK,
};

static mfp_cfg_t lcd_mfp_cfg[] __initdata = {
	GPIO16_GPIO,

#if defined(TAVOR_EVB3_V2)
	GPIO20_LCD_BIAS,
#else
	GPIO20_LCD_RST,
	GPIO93_LCD_CLK_TOUT,
#endif

	GPIO23_LCD_DD0,
	GPIO24_LCD_DD1,
	GPIO25_LCD_DD2,
	GPIO26_LCD_DD3,
	GPIO27_LCD_DD4,
	GPIO28_LCD_DD5,
	GPIO29_LCD_DD6,
	GPIO30_LCD_DD7,
	GPIO31_LCD_DD8,
	GPIO32_LCD_DD9,
	GPIO33_LCD_DD10,
	GPIO34_LCD_DD11,
	GPIO35_LCD_DD12,
	GPIO36_LCD_DD13,
	GPIO37_LCD_DD14,
	GPIO38_LCD_DD15,
	GPIO39_LCD_DD16,
	GPIO40_LCD_DD17,

	GPIO17_LCD_FCLK_RD,
	GPIO18_LCD_LCLK_A0,
	GPIO19_LCD_PCLK_WR,

	GPIO12_SSP1_CLK,
	GPIO13_SSP1_FRM,
	GPIO15_SSP1_TXD,
};

#ifdef	CONFIG_PROC_FS
#define	EVB3_PROC_FILE	"driver/evb3"
static struct proc_dir_entry *evb3_proc_file;
static int pmic_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
   //	struct pmic_ops *ops = file->private_data;
	void __user *uarg = (void __user *)arg;
	int ret = -EINVAL;
    u8 icotl_pmic_id;

	switch (cmd) {
	case IOCTL_PMIC_ID_GET:
	   	icotl_pmic_id = g_pmic_id;
		ret = put_user(icotl_pmic_id, (unsigned long __user *)uarg);
		printk(KERN_ALERT "-------> icotl_pmic_id %d \n",icotl_pmic_id);
		if (ret)
			ret = -EFAULT;
		break;


	default:
	  	//if (ops->ioctl)
	  	//	ret = ops->ioctl(cmd, arg);
		panic("EVB3-pmic_ioctl");
		break;
	}
	return ret;
}


static ssize_t evb3_proc_read(struct file *filp,
        char __user *buffer, size_t length, loff_t *offset)
{
	printk(KERN_ALERT "evb3_proc_read - no action done\n");

	return 0;
}

static ssize_t evb3_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	printk(KERN_ALERT "evb3_proc_write - no action done\n");

	return len;
}

static struct file_operations evb3_proc_ops = {
	.read 	= evb3_proc_read,
	.write 	= evb3_proc_write,
	.ioctl	= pmic_ioctl,

};

static void create_evb3_proc_file(void)
{
	evb3_proc_file = create_proc_entry(EVB3_PROC_FILE, 0644, NULL);
	if (evb3_proc_file) {
		evb3_proc_file->owner = THIS_MODULE;
		evb3_proc_file->proc_fops = &evb3_proc_ops;
		printk(KERN_ALERT "EVB3 proc file created!\n");
	} else
		printk(KERN_ALERT "EVB3 proc file create failed!\n");
}

//static void remove_evb3_proc_file(void)
//{
	//remove_proc_entry(EVB3_PROC_FILE, &proc_root);
//}
#endif  //CONFIG_PROC_FS


#if defined(CONFIG_MOUSE_PXA930_TRKBALL) || \
	defined(CONFIG_MOUSE_PXA930_TRKBALL_MODULE)
static struct pxa930_trkball_platform_data tavor_evb3_mouse_info = {
	.x_filter	= 2,
	.y_filter	= 2,
};

static void __init tavor_evb3_init_trackball(void)
{
	pxa_register_device(&pxa930_device_trkball, &tavor_evb3_mouse_info);
}
#else
static inline void tavor_evb3_init_trackball(void) {}
#endif

static struct resource smc91x_resources[] = {
	[0] = {
		.start	= (TAVOREVB3_ETH_PHYS + 0x300),
		.end	= (TAVOREVB3_ETH_PHYS + 0xfffff),
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO42)),
		.end	= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO42)),
		.flags	= IORESOURCE_IRQ | IORESOURCE_IRQ_HIGHEDGE,
	}
};

static struct platform_device smc91x_device = {
	.name		= "smc91x",
	.id		= 0,
	.num_resources	= ARRAY_SIZE(smc91x_resources),
	.resource	= smc91x_resources,

};

static struct platform_device sensor_input_device = {
	.name 		= "sensor_input",
	.id   		= -1,
};

static struct platform_device micco_bl_device = {
	.name 		= "micco-bl",
	.id 		= -1,
};

static struct platform_device micco_kp_bl_device = {
	.name 		= "micco-kp-bl",
	.id 		= -1,
};

static struct platform_device micco_ts_device = {
	.name 		= "micco-ts",
	.id 		= -1,
};

static struct platform_device levante_ts_device = {
	.name 		= "levante-ts",
	.id 		= -1,
};

static struct platform_device levante_charger_device = {
	.name		= "levante-charger",
	.id		= -1,
};


static struct platform_device levante_kp_bl_device = {
	.name 		= "levante-kp-bl",
	.id 		= -1,
};

static struct platform_device levante_bl_device = {
       .name           = "levante-bl",
       .id             = -1,
};

#ifdef CONFIG_IMM
static struct resource pxa3xx_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code +
						  * The first 4KB are locked for pxa930 trusted */
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct resource pxa935_resource_imm[] = {
	[0] = {
		.name   = "phy_sram",
		.start	= PHYS_SRAM_START + 0x10000, /* Moved up to offset 64KB due to collision with DVFM
						  * D2 entry code */
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1, /* 2 banks on pxa935 */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.name   = "imm_sram",
		.start	= PHYS_SRAM_START + 0x10000,
		.end	= PHYS_SRAM_START + 2 * PHYS_SRAM_BANK_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device pxa3xx_device_imm = {
	.name 		= "pxa3xx-imm",
	.id 		= -1,
	.num_resources	= ARRAY_SIZE(pxa3xx_resource_imm),
	.resource	= pxa3xx_resource_imm,
};
#endif

static struct platform_device micco_charger_device = {
	.name		= "micco-charger",
	.id		= -1,
};

static struct platform_device micco_hsdetect_device = {
	.name		= "micco-hsdetect",
	.id		= -1,
};

static struct platform_device levante_hsdetect_device = {
	.name		= "levante-hsdetect",
	.id		= -1,
};

static struct platform_device ispt_device = {
	.name		= "pxa-ispt",
	.id		= -1,
};


static struct platform_device vnc_device = {
	.name		= "vnc-ts",
	.id		= -1,
};

static struct platform_device sfh7225_nled_device = {
	.name 		= "sfh7225_nled",
	.id		= -1,
};

static struct platform_device sfh7225_ls_device = {
	.name           = "sfh7225_ls",
	.id             = -1,
};

static struct platform_device *devices[] __initdata = {
#ifdef CONFIG_IMM
   	&pxa3xx_device_imm,
#endif
   	&ispt_device,
	&vnc_device,
   	&sfh7225_nled_device,
   	&sfh7225_ls_device,
};

static struct platform_device *devices_levante[] __initdata = {
#if defined(CONFIG_PXA9xx_LEVANTE)
 	&levante_bl_device,
    &levante_kp_bl_device,
 	&levante_ts_device,
 	&levante_charger_device,
	&levante_hsdetect_device
#endif
};

static struct platform_device *devices_micco[] __initdata = {
#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
  	&micco_bl_device,
  	&micco_kp_bl_device,
  	&micco_ts_device,
  	&micco_charger_device,
 	&micco_hsdetect_device,
#endif
};


#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
static unsigned int tavor_evb3_matrix_key_map[] = {
	/* KEY(row, col, key_code) */
	KEY(1, 3, KEY_0), KEY(0, 0, KEY_1), KEY(1, 0, KEY_UP), KEY(2, 0, KEY_3),
	KEY(0, 1, KEY_LEFT), KEY(1, 1, KEY_5), KEY(2, 1, KEY_RIGHT), KEY(0, 2, KEY_7),
	KEY(1, 2, KEY_DOWN), KEY(2, 2, KEY_9),

	KEY(0, 3, KEY_KPASTERISK), 	/* * */
	KEY(2, 3, KEY_KPDOT), 		/* # */

	KEY(4, 0, KEY_HOME),
	KEY(3, 3, KEY_END),
	KEY(4, 1, KEY_BACK),

	KEY(0, 4, KEY_ENTER),		/* scroll push */

	KEY(3, 2, KEY_SEND),
	KEY(4, 2, KEY_VOLUMEUP),
	KEY(4, 3, KEY_VOLUMEDOWN),

	KEY(3, 0, KEY_MENU),	/* soft1 */
	KEY(3, 1, KEY_REPLY),	/* soft2 */

	KEY(1, 4, KEY_CAMERA),
};

static struct pxa27x_keypad_platform_data tavor_evb3_keypad_info = {
	.matrix_key_rows	= 5,
	.matrix_key_cols	= 5,
	.matrix_key_map		= tavor_evb3_matrix_key_map,
	.matrix_key_map_size	= ARRAY_SIZE(tavor_evb3_matrix_key_map),
	.debounce_interval	= 30,
};
#endif /* CONFIG_KEYBOARD_PXA27x || CONFIG_KEYBOARD_PXA27x_MODULE */

#ifdef CONFIG_ISPT
static mfp_cfg_t pxa930_ispt_pins[] __initdata = {
	/* SSP4 */
	GPIO51_SSP4_CLK,
	GPIO52_SSP4_FRM,
	GPIO45_SSP4_TXD,
};

static mfp_cfg_t pxa950_ispt_pins[] __initdata = {
	/* SSP3 */
	GPIO43_SSP3_CLK,
	GPIO44_SSP3_FRM,
	GPIO46_SSP3_TXD,
};

static void __init tavor_evb3_init_ispt(void)
{
	if(cpu_is_pxa950())
		pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa950_ispt_pins));
	else
	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_ispt_pins));
}
#endif

#if defined(CONFIG_MMC) || defined(CONFIG_MMC_MODULE)
static int tavor_evb3_mci_init(struct device *dev,
		irq_handler_t tavor_evb3_detect_int,
			     void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int err, cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(tavor_evb3_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = tavor_evb3_mmc_slot[pdev->id].gpio_cd;

	/*
	 * setup GPIO for tavor_evb3 MMC controller
	 */
	err = gpio_request(gpio_cd, "mmc card detect");
	if (err)
		goto err_request_cd;
	gpio_direction_input(gpio_cd);

	err = request_irq(cd_irq, tavor_evb3_detect_int,
			  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			  "MMC card detect", data);
	if (err) {
		printk(KERN_ERR "%s: MMC/SD/SDIO: "
				"can't request card detect IRQ\n", __func__);
		goto err_request_irq;
	}

	return 0;

err_request_irq:
	gpio_free(gpio_cd);
err_request_cd:
	return err;
}

static void tavor_evb3_mci_exit(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	int cd_irq, gpio_cd;

	cd_irq = gpio_to_irq(tavor_evb3_mmc_slot[pdev->id].gpio_cd);
	gpio_cd = tavor_evb3_mmc_slot[pdev->id].gpio_cd;

	free_irq(cd_irq, data);
	gpio_free(gpio_cd);
}

static struct pxamci_platform_data tavor_evb3_mci4_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34,
	.init 		= tavor_evb3_mci_init,
	.exit		= tavor_evb3_mci_exit,
};

static struct pxamci_platform_data tavor_evb3_mci3_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34|MMC_VDD_165_195,
};

static struct pxamci_platform_data tavor_evb3_mci5_platform_data = {
	.detect_delay	= 50,
	.ocr_mask	= MMC_VDD_32_33|MMC_VDD_33_34|MMC_VDD_165_195,
#ifdef CONFIG_SD8XXX_RFKILL
	.init		= sd8x_sdh_init,
#endif
};

static mfp_cfg_t __attribute__ ((unused)) pxa9xx_mmc_mfp_cfg[] __initdata = {
	/* MMC4 of Tavor PV */
	GPIO47_GPIO,
	GPIO55_MMC4_CMD,
	GPIO56_MMC4_CLK,
	GPIO57_MMC4_DAT0,
	GPIO58_MMC4_DAT1,
	GPIO59_MMC4_DAT2,
	GPIO60_MMC4_DAT3,
};

static mfp_cfg_t __attribute__ ((unused)) pxa9xx_mmc3_mfp_cfg[] __initdata = {
	/* MMC3 of Tavor PV2 */
	DF_ADDR0_MMC3_CMD,
	DF_ADDR1_MMC3_CLK,
	DF_ADDR2_MMC3_DAT0,
	DF_ADDR3_MMC3_DAT1,
	nXCVREN_MMC3_DAT2,
	nLUA_MMC3_DAT3,
	DF_nWE_MMC3_DAT4,
	nBE0_MMC3_DAT5,
	nBE1_MMC3_DAT6,
	DF_nRE_MMC3_DAT7,
};

static mfp_cfg_t pxa9xx_mmc5_mfp_cfg[] __initdata = {
	GPIO91_MMC5_CLK,
	GPIO90_MMC5_CMD,
};

static int sd8xxx_poweron(void)
{
	int gpio_pd = 0;
	int gpio_reset = 0;

	gpio_pd = mfp_to_gpio(MFP_PIN_GPIO77);
	gpio_reset = mfp_to_gpio(MFP_PIN_GPIO67);

	if (gpio_request(gpio_pd, "sd8xxx wlan power down")) {
		printk(KERN_INFO "gpio %d request failed\n", gpio_pd);
		return -1;
	}

	if(gpio_request(gpio_reset, "sd8xxx wlan reset")) {
		printk(KERN_INFO "gpio %d request failed\n", gpio_reset);
		gpio_free(gpio_pd);
		return -1;
	}

	gpio_direction_output(gpio_pd, 1);
	gpio_direction_output(gpio_reset, 0);
	mdelay(500);
	gpio_direction_output(gpio_reset, 1);

	gpio_free(gpio_pd);
	gpio_free(gpio_reset);
	return 0;
}

static void __init tavor_evb3_init_mmc(void)
{
	if (device_is_enabled(device_mmc)) {
		pxa_set_mci_info(&tavor_evb3_mci4_platform_data);
#if defined(CONFIG_EVB3_EMMC)
		pxa3xx_set_mci2_info(&tavor_evb3_mci3_platform_data);
#endif
		/* MMC card detect for controller 0 of pxamci */
		tavor_evb3_mmc_slot[0].gpio_cd  = mfp_to_gpio(MFP_PIN_GPIO20);
	}
	if (device_is_enabled(device_mmc_new)) {


	pxa9xx_set_mci3_info(&tavor_evb3_mci3_platform_data);

	/* pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa9xx_mmc_mfp_cfg)); */
	pxa9xx_set_mci4_info(&tavor_evb3_mci4_platform_data);
	/* MMC card detect for controller 1 of pxa9xx_mci */
	tavor_evb3_mmc_slot[1].gpio_cd  = mfp_to_gpio(MFP_PIN_GPIO47);

	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa9xx_mmc5_mfp_cfg));
	pxa9xx_set_mci5_info(&tavor_evb3_mci5_platform_data);
	}
}
#else
static inline void tavor_evb3_init_mmc(void) {}
#endif

#if (defined(CONFIG_MTD_ONENAND) || defined(CONFIG_MTD_ONENAND_MODULE))
static unsigned char __iomem *smc_mbase;
static void __attribute__ ((unused)) tavor_onenand_mmcontrol(struct mtd_info *mtd, int sync_read)
{
	struct onenand_chip *this = mtd->priv;
	static unsigned int csadrcfg2, msc1, sxcnfg;
	unsigned int syscfg;

	if (sync_read) {
		csadrcfg2 = readl(smc_mbase + CSADRCFG2_OFF);
		msc1 = readl(smc_mbase + MSC1_OFF);
		sxcnfg = readl(smc_mbase + SXCNFG_OFF);
		writel(0x0032091d, smc_mbase + CSADRCFG2_OFF);
		writel(((msc1 & 0xffff0000) | 0x7f18), smc_mbase + MSC1_OFF);
		writel(((sxcnfg & 0xffff) | 0x30110000), smc_mbase + SXCNFG_OFF);
		writel(0x04, smc_mbase + CLK_RET_DEL_OFF);
		writel(0x00, smc_mbase + ADV_RET_DEL_OFF);
		syscfg = this->read_word(this->base + ONENAND_REG_SYS_CFG1);
		this->write_word((syscfg | sync_read),
				this->base + ONENAND_REG_SYS_CFG1);
	} else {
		syscfg = this->read_word(this->base + ONENAND_REG_SYS_CFG1);
		this->write_word((syscfg & ~sync_read),
				this->base + ONENAND_REG_SYS_CFG1);
		writel(sxcnfg, smc_mbase + SXCNFG_OFF);
		writel(msc1, smc_mbase + MSC1_OFF);
		writel(csadrcfg2, smc_mbase + CSADRCFG2_OFF);
	}

}

static struct flash_platform_data tavor_evb3_onenand_info;
static void __init tavor_evb3_init_onenand(void)
{
	if (is_android()) {
		if (is_comm_v75()) {
			tavor_evb3_onenand_info.parts = android_256m_v75_partitions;
			tavor_evb3_onenand_info.nr_parts = ARRAY_SIZE(android_256m_v75_partitions);
		} else {
			tavor_evb3_onenand_info.parts = android_256m_partitions;
			tavor_evb3_onenand_info.nr_parts = ARRAY_SIZE(android_256m_partitions);
		}
	} else {
		if (is_comm_v75()) {
			tavor_evb3_onenand_info.parts = pxa930_256m_v75_partitions;
			tavor_evb3_onenand_info.nr_parts = ARRAY_SIZE(pxa930_256m_v75_partitions);
		} else {
			tavor_evb3_onenand_info.parts = pxa930_256m_partitions;
			tavor_evb3_onenand_info.nr_parts = ARRAY_SIZE(pxa930_256m_partitions);
		}
	}

	smc_mbase = ioremap(SMC_START, SMC_END - SMC_START + 1);
	//tavor_evb3_onenand_info.mmcontrol = tavor_onenand_mmcontrol;

	pxa3xx_device_onenand.dev.platform_data = &tavor_evb3_onenand_info;
	platform_device_register(&pxa3xx_device_onenand);
}
#else
static inline void tavor_evb3_init_onenand(void) {}
#endif

#if defined(CONFIG_MTD_NAND_PXA3xx) || defined(CONFIG_MTD_NAND_PXA3xx_MODULE)
static struct pxa3xx_nand_platform_data tavor_evb3_nand_info;
static void __init tavor_evb3_init_nand(void)
{
	if (is_android()) {
		tavor_evb3_nand_info.parts = android_128m_partitions;
		tavor_evb3_nand_info.nr_parts = ARRAY_SIZE(android_128m_partitions);
	} else {
		if (is_comm_v75()) {
			tavor_evb3_nand_info.parts = pxa930_128m_v75_partitions;
			tavor_evb3_nand_info.nr_parts = ARRAY_SIZE(pxa930_128m_v75_partitions);
		} else {
			tavor_evb3_nand_info.parts = pxa930_128m_partitions;
			tavor_evb3_nand_info.nr_parts = ARRAY_SIZE(pxa930_128m_partitions);
		}
	}

	pxa3xx_device_nand.dev.platform_data = &tavor_evb3_nand_info;
	platform_device_register(&pxa3xx_device_nand);
}
#else
static inline void tavor_evb3_init_nand(void) {}
#endif /* CONFIG_MTD_NAND_PXA3xx || CONFIG_MTD_NAND_PXA3xx_MODULE */

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_PXA27X_UDC_MODULE)
static mfp_cfg_t pxa930_otg_init_pins[] = {
	GPIO64_GPIO,
};

static mfp_cfg_t pxa930_otg_pins[] = {
	GPIO64_USB_P2_7,
};

int tavor_evb3_udc_is_miniA(void)
{
	int otg_id = mfp_to_gpio(MFP_PIN_GPIO64);
	int id_value;
	int err;

	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_init_pins));
	err = gpio_request(otg_id, "OTG ID");
	if (err) {
		gpio_free(otg_id);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", otg_id, err);
		return 0;
	}
	gpio_direction_input(otg_id);
	id_value = gpio_get_value(otg_id);
	gpio_free(otg_id);
	pxa3xx_mfp_config(ARRAY_AND_SIZE(pxa930_otg_pins));

	return (id_value == 0);
}

static struct pxa2xx_udc_mach_info tavor_evb3_udc_info = {
	.udc_is_miniA = tavor_evb3_udc_is_miniA,
};
#endif

#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)

void tavor_evb3_reset_xcvr(void)
{
	int reset_pin;
	int err;

	reset_pin = mfp_to_gpio(MFP_PIN_GPIO43);
	err = gpio_request(reset_pin, "U2D Reset");
	if (err) {
		gpio_free(reset_pin);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", reset_pin, err);
		return;
	}
	gpio_direction_output(reset_pin, 0);
	mdelay(100);
	gpio_set_value(reset_pin, 1);

	gpio_free(reset_pin);
}

static struct pxa3xx_u2d_mach_info tavor_evb3_u2d_info = {
	.reset_xcvr = tavor_evb3_reset_xcvr,
};
#endif /* CONFIG_USB_PXA3XX_U2D || CONFIG_USB_PXA3XX_U2D_MODULE */

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_PXA9XX_U2O_MODULE)
extern void pxa9xx_u2o_xcvr_init(unsigned *base);
extern struct resource pxa9xx_u2o_resources[];
static struct pxa9xx_u2o_mach_info tavor_evb3_u2o_info = {
	.xcvr_init = pxa9xx_u2o_xcvr_init,
};
#endif

#ifdef CONFIG_USB_EHCI_PXA9XX

static void __init tavor_evb3_init_ehci(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_ehci_info(&tavor_evb3_u2o_info);
}
#else
static inline void tavor_evb3_init_ehci(void) {}
#endif

#if defined (CONFIG_USB_OTG) && defined(CONFIG_USB_PXA9XX_U2O)
static void __init tavor_evb3_init_otg(void)
{
	if (!device_is_enabled(device_u2o))
		return;

	pxa_set_u2otg_info(&tavor_evb3_u2o_info);
}
#else
static inline void tavor_evb3_init_otg(void) {}
#endif

#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
static int micco_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_PIN_GPIO83));

	return 0;
}

static int micco_ack_irq(void)
{
	return 0;
}

static void tavor_evb3_micco_init(void)
{
	u8 value;

	/* Mask interrupts that are not needed */
	micco_write(MICCO_IRQ_MASK_A, 0xFE);
	micco_write(MICCO_IRQ_MASK_B, 0xFF);
	micco_write(MICCO_IRQ_MASK_C, 0xFF);
	micco_write(MICCO_IRQ_MASK_D, 0xFF);

	/* avoid SRAM power off during sleep*/
	micco_write(0x10, 0x07);
	micco_write(0x11, 0xff);
	micco_write(0x12, 0xbf); /*never enable LDO4: CP controls it via COMM_OVER3 reg (0x22)*/

	/* Enable the ONKEY power down functionality */
	/*micco_write(MICCO_SYSCTRL_B, 0x20);*/
	micco_write(MICCO_SYSCTRL_A, 0x60);

	/* IRQ is masked during the power-up sequence and will not be released
	 * until they have been read for the first time */
	micco_read(MICCO_EVENT_A, &value);
	micco_read(MICCO_EVENT_B, &value);
	micco_read(MICCO_EVENT_C, &value);
	micco_read(MICCO_EVENT_D, &value);
}

/* micco_power_module[] should be consistent with enum
 * in arch/arm/mach-pxa/include/mach/pxa3xx_pmic.h */
static struct power_supply_module miccoB0_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		BUCK2},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,		LDO3},
	{VCC_SDIO,		LDO13},
	{VCC_CAMERA_ANA,	LDO3},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO3},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,		BUCK3},
	{VCC_1P8V,		LDO4},
	{VCC_MEM,		BUCK3},
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

/* micco_power_module[] should be consistent with enum
 * in arch/arm/mach-pxa/include/mach/pxa3xx_pmic.h */
static struct power_supply_module miccoEA_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		BUCK2},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,		LDO3},
	{VCC_SDIO,		LDO13},
	{VCC_CAMERA_ANA,	LDO3},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO3},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,		LDO2},
	{VCC_1P8V,		LDO4},
	{VCC_MEM,		LDO2},
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

static struct power_chip micco_chips[] = {
	{MICCO_B0_ID,	"miccoB0",	miccoB0_power_modules},
	{MICCO_EA_ID,	"miccoEA",	miccoEA_power_modules},
	{MICCO_EB_ID,	"miccoEB",	miccoEA_power_modules},
	{0,	NULL,		NULL},
};

static struct micco_platform_data micco_data = {
	.init_irq = micco_init_irq,
	.ack_irq = micco_ack_irq,
	.platform_init = tavor_evb3_micco_init,
	.power_chips = micco_chips,
};
#endif /* CONFIG_PXA3xx_MICCO || CONFIG_PXA3xx_MICCO_MODULE*/

#if defined(CONFIG_PXA_IRDA) || defined(CONFIG_PXA_IRDA_MODULE)
static void tavor_evb3_irda_transceiver_mode(struct device *dev, int mode)
{
	unsigned long flags;
	int err;
	static int irda_mfp_init;
	int gpio_ir_shdn = ((struct pxaficp_platform_data *)
			dev->platform_data)->gpio_ir_shdn;

	if (!irda_mfp_init) {
		err = gpio_request(gpio_ir_shdn, "IRDA SHDN");
		if (err) {
			gpio_free(gpio_ir_shdn);
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d return :%d\n", gpio_ir_shdn, err);
			return;
		}
		gpio_direction_output(gpio_ir_shdn, 1);

		irda_mfp_init = 1;
	}

	local_irq_save(flags);
	if (mode & IR_SIRMODE) {
		gpio_set_value(gpio_ir_shdn, 0);
	} else if (mode & IR_FIRMODE) {
		/* do not support FIR */
	}
	if (mode & IR_OFF) {
		gpio_set_value(gpio_ir_shdn, 1);
	}
	local_irq_restore(flags);
}

static struct pxaficp_platform_data tavor_evb3_ficp_platform_data = {
	.transceiver_cap  = IR_SIRMODE | IR_OFF,
	.transceiver_mode = tavor_evb3_irda_transceiver_mode,
};
#endif /* (CONFIG_PXA_IRDA) || (CONFIG_PXA_IRDA_MODULE) */

#if defined(CONFIG_PXA9xx_LEVANTE)
static int levante_init_irq(void)
{
	gpio_direction_input(mfp_to_gpio(MFP_PIN_GPIO83));
	pxa3xx_mfp_set_edge(MFP_CFG_PIN(MFP_PIN_GPIO83),
					MFP_EDGE_FALL);
	return 0;
}

static int levante_ack_irq(void)
{
	return 0;
}

static void tavorevb3_levante_init(void)
{

	printk(KERN_ALERT "----->evb3_levante_init ( all is done in levante.c only platfrom specific should be here )\n");
}

/* power_module[] should be consistent with enum
 * in include/asm-arm/arch-pxa/pxa3xx_pmic.h */
static struct power_supply_module levante_power_modules[] = {
	/* {command,		power_module}, */
	{VCC_CORE,		BUCK1},
	{VCC_SRAM,		BUCK2},
	{VCC_MVT,		LDO1},
	{VCC_3V_APPS,	LDO3},
	{VCC_SDIO,		LDO14},
	{VCC_CAMERA_ANA, LDO3},
	{VCC_USB,		LDO5},
	{VCC_LCD,		LDO3},
	{VCC_TSI,		0},
	{VCC_CAMERA_IO,	BUCK2}, /* ? */
	{VCC_1P8V,		0}, 	/* ? */
	{VCC_MEM,		BUCK2}, /* ? */
	{HDMI_TX,		0},
	{TECH_3V,		0},
	{TECH_1P8V,		0},
};

static struct power_chip levante_chips[] = {
	{LEVATNE_A0_ID,	"levanteA0",	levante_power_modules},
	{LEVATNE_A1_ID,	"levanteA1",	levante_power_modules},
	{LEVATNE_END_ID,"levante",	levante_power_modules},
	{0,	NULL,		NULL},
};
static struct levante_platform_data  levante_data = {
    .platform_init = tavorevb3_levante_init
};

static struct sanremo_platform_data  sanremo_data = {
    .init_irq = levante_init_irq,
    .ack_irq = levante_ack_irq,
    .power_chips = levante_chips,
};

static struct portofino_platform_data  portofino_data = {
};

#endif /* CONFIG_PXA9xx_LEVANTE */

#if defined(CONFIG_PXA950_CAMERA)
static int cam_power_set(int flag)
{
	struct clk *axi_clk = NULL;
	struct clk *sci1_clk = NULL;
	struct clk *sci2_clk = NULL;

	sci1_clk = clk_get(NULL, "SCI1CLK");
	if (IS_ERR(sci1_clk)) {
		printk(KERN_ERR "unable to get SCI1CLK\n");
		return PTR_ERR(sci1_clk);
	}

	sci2_clk = clk_get(NULL, "SCI2CLK");
	if (IS_ERR(sci2_clk)) {
		printk(KERN_ERR "unable to get SCI2CLK\n");
		return PTR_ERR(sci2_clk);
	}

	if(flag){
		axi_clk = clk_get(NULL, "AXICLK");
		if (IS_ERR(axi_clk)) {
			printk(KERN_ERR "unable to get AXICLK\n");
			return PTR_ERR(axi_clk);
		}

		clk_enable(axi_clk);
		clk_enable(sci1_clk);
		clk_enable(sci2_clk);

		clk_put(axi_clk);
	}else{
		clk_disable(sci1_clk);
		clk_disable(sci2_clk);
	}

	clk_put(sci1_clk);
	clk_put(sci2_clk);

	return 0;
}
/* sensor init */
static int sensor_power_set(int res, int flag)
{
	/*
	 * RES, 0, LOW RESOLUTION, 1 HIGH RESOLUTION
	 * FLAG, 1: ON, 0: OFF
	 * GPIO93 LOW FOR HIGH
	 * GPIO14 LOW FOR LOW
	 */
	int high_pin = 0;
	int low_pin = mfp_to_gpio(MFP_PIN_GPIO14);

	if (cpu_is_pxa950_z0()){
		high_pin = mfp_to_gpio(MFP_PIN_GPIO93);
	}
	else
		high_pin = mfp_to_gpio(MFP_PIN_GPIO41);

	cam_power_set(flag);

	if (gpio_request(high_pin, "CAM_EANBLE_HI_SENSOR")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", high_pin);
		return -EIO;
	}

	if (res) {
		if (flag){
			gpio_direction_output(high_pin, 0);	/* enable */
			msleep(1);
		} else {
			gpio_direction_output(high_pin, 1);	/* disable */
		}
	} else{
		if (gpio_request(low_pin, "CAM_EANBLE_LOW_SENSOR")) {
			printk(KERN_ERR "Request GPIO failed,"
					"gpio: %d \n", low_pin);
			return -EIO;
		}
		if (flag) {
			/* high resolution is bridge */
			gpio_direction_output(high_pin, 0);	/* enable */
			gpio_direction_output(low_pin, 0);	/* enable */

		}
#if 0
		/* FIXME */
		else
			gpio_direction_output(low_pin, 1);	/* disable */
#endif
		gpio_free(low_pin);
		msleep(100);
	}

#ifdef CONFIG_CPU_PXA968
#define MFP_GPIO111         __REG(0x40E10570)
#define GPDR3         __REG(0x40E0010C)
#define GPDR0         __REG(0x40E0000C)
#define GPSR0         __REG(0x40E00018)
        MFP_GPIO111 = 0x0000B0C2;
        GPDR3 |=0x00008000;
        GPDR0 |=0x00004000;
        GPSR0 |=0x00004000;
#endif

	gpio_free(high_pin);
	return 0;
}

static struct sensor_platform_data ov3640_sensor_data = {
	.id = SENSOR_HIGH,
	.power_set = sensor_power_set,
};

static struct sensor_platform_data ov5642_sensor_data = {
	.id = SENSOR_HIGH,
	.power_set = sensor_power_set,
};

static struct sensor_platform_data ov7690_sensor_data = {
	.id = SENSOR_LOW,
	.power_set = sensor_power_set,
};

/* sensor init over */
static struct cam_platform_data cam_ops = {
	.power_set	= cam_power_set,
};

static mfp_cfg_t cam_mfp_cfg[] __initdata = {
	/* CAM */
};

static void __init tavor_evb3_init_cam(int id)
{
	static bool binit = false;

	switch(id)// TavorPV2 supports 2 camera interface controller
	{
		case CCIC_0:
			pxa950_device_cam0.dev.platform_data = &cam_ops;
			pxa3xx_mfp_config(ARRAY_AND_SIZE(cam_mfp_cfg));
			platform_device_register(&pxa950_device_cam0);
			break;

		case CCIC_1:
			pxa950_device_cam1.dev.platform_data = &cam_ops;
			pxa3xx_mfp_config(ARRAY_AND_SIZE(cam_mfp_cfg));
			platform_device_register(&pxa950_device_cam1);
			break;
	}
	if (!binit)// TavorPV2 supports 1 CSI interface controller
	{
		platform_device_register(&pxa950_device_csi);
		binit = true;
	}
}

#endif

#if defined(CONFIG_RADIO_SI4703) || defined(CONFIG_RADIO_SI4703_MODULE)
#if 0
int si4703_setup(struct i2c_client *client, void *context)
{
	int reset_pin = mfp_to_gpio(MFP_PIN_GPIO86);
	int irq_pin = mfp_to_gpio(MFP_PIN_GPIO96);

	gpio_request(reset_pin, "si4703 FM radio reset");
	gpio_request(irq_pin, "si4703 FM radio interrupt");

	/* clear GPIO96 edge detect */
	pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO96_GPIO), MFP_EDGE_NONE);

	/* configure interrupt pin as input */
	gpio_direction_input(reset_pin);

	/* assert reset for 100 ms*/
	gpio_direction_output(reset_pin, 0);
	mdelay(100);

	/* deassert reset */
	gpio_set_value(reset_pin, 0);

	gpio_free(reset_pin);
	gpio_free(irq_pin);

	return 0;
}

static struct si4703_platform_data si4703_data = {
	.setup = si4703_setup,
};
#endif
#endif


static struct i2c_board_info tavor_evb3_i2c_hs_info[] = {
	/* hs i2c device */
#if defined(CONFIG_PXA950_CAMERA)
	{
		.type		= "ov5642",
		.addr		= 0x3c,
		.platform_data	= &ov5642_sensor_data,
	},
	{
		.type		= "ov3640",
		.addr		= 0x3c,
		.platform_data	= &ov3640_sensor_data,
	},
	{
		.type		= "ov7690",
		.addr		= 0x21,
		.platform_data	= &ov7690_sensor_data,
	},
#endif

#if defined(CONFIG_SENSORS_LIS311DL)
	{
		.type		= "lis311dl",
		.addr		= 0x1c,
	},
#endif
};
static struct i2c_board_info evb3_i2c_board_info_micco[] = {

#if defined(CONFIG_PXA3xx_MICCO) || defined(CONFIG_PXA3xx_MICCO_MODULE)
	{
		.type	= "micco",
		.addr		= 0x34,
		.platform_data	= &micco_data,
		.irq		= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO83)),
	},
#endif

#if defined(CONFIG_RADIO_SI4703) || defined(CONFIG_RADIO_SI4703_MODULE)
	#if 0
		{
			.type	= "si4703",
			.addr		= 0x10,
			.platform_data	= &si4703_data,
			.irq   		= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO96)),
		},
	#endif
#endif
};

static struct i2c_board_info evb3_i2c_board_info_levante1[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "sanremo",
		.addr		= 0x34,
		.platform_data	= &sanremo_data,
	   	.irq		= IRQ_GPIO(mfp_to_gpio(MFP_PIN_GPIO83)),
	},
	{
		.type	= "portofino",
		.addr		= 0x10,
		.platform_data	= &portofino_data,
	},
#endif

};

static struct i2c_board_info evb3_i2c_board_info_levante2[] = {

#if defined(CONFIG_PXA9xx_LEVANTE)
	{
		.type	= "levante",
           .platform_data  = &levante_data,
	},
#endif
};

static struct i2c_pxa_platform_data pxa3xx_i2c_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_FAST_MODE, /* PXA_I2C_STANDARD_MODE - SLOW */
};

static struct i2c_pxa_platform_data pxa3xx_i2c_hs_platform_data = {
	.use_pio	= 0,
 	.flags		= PXA_I2C_FAST_MODE,
};

struct pxa3xx_freq_mach_info tavor_evb3_freq_mach_info = {
	.flags = PXA3xx_USE_POWER_I2C,
};

#if defined(CONFIG_FB_PXA950)

#define ESPON_LCD_COMMAND 0x0 << 9
#define ESPON_LCD_PARAMETERS 0x1 << 9
#define ESPON_LCD_COMMAND_SLPOUT 0x11
#define ESPON_LCD_COMMAND_SLPIN 0x10
#define ESPON_LCD_COMMAND_DISON 0x29
#define ESPON_LCD_COMMAND_DISOFF 0x28

#define ESPON_CMD(x) (ESPON_LCD_COMMAND | x)
#define ESPON_PARA(x) (ESPON_LCD_PARAMETERS | x)

static u16 espon_slpout[] = {
	ESPON_CMD(ESPON_LCD_COMMAND_SLPOUT),
};

static u16 espon_dison[] = {
	ESPON_CMD(ESPON_LCD_COMMAND_DISON),
};

static u16 espon_disoff[] = {
	ESPON_CMD(ESPON_LCD_COMMAND_DISOFF),
};

static u16 espon_slpin[] = {
	ESPON_CMD(ESPON_LCD_COMMAND_SLPIN),
};

void espon_lcd_panel_reset(void)
{
	msleep(1);
	if (gpio_request(MFP_PIN_GPIO20, "lcd reset"))
		printk(KERN_ERR "LTM 022A97A gpio_request: failed!\n");
	gpio_direction_output(MFP_PIN_GPIO20,0);
	msleep(1);
	gpio_direction_output(MFP_PIN_GPIO20,1);
	gpio_free(MFP_PIN_GPIO20);
	msleep(10);
}

static void espon_ssp_config(struct ssp_dev *dev)
{
	ssp_disable(dev);
	ssp_config(dev, 0x00000588, 0x18, 0, 0);
}
static void ssp_send_cmd_para(u16 *cmd,int num)
{
	static struct ssp_dev ssp;
	int i;
	static int ssp_initialized = 0;

	if (!ssp_initialized) {
		ssp_init(&ssp, 1, SSP_NO_IRQ);
		espon_ssp_config(&ssp);
		ssp_initialized = 1;
	}

	clk_enable(ssp.ssp->clk);

	for (i = 0; i < num; i++, cmd++) {
		ssp_write_word(&ssp, *cmd & 0x1ff);

		/* FIXME: ssp_flush() is mandatory here to work */
		ssp_flush(&ssp);
	}
	clk_disable(ssp.ssp->clk);
}
static void espon_lcd_power(int on)
{
	if (on) {
#if defined(TAVOR_EVB3_V2)
		espon_lcd_panel_reset();
#endif

		/* SLPOUT*/
		ssp_send_cmd_para(ARRAY_AND_SIZE(espon_slpout));
		mdelay(60);

		/* DISON */
		ssp_send_cmd_para(ARRAY_AND_SIZE(espon_dison));

	} else {
		/* DISOFF  */
		ssp_send_cmd_para(ARRAY_AND_SIZE(espon_disoff));

		mdelay(60);

		/* SLPIN */
		ssp_send_cmd_para(ARRAY_AND_SIZE(espon_slpin));
	}
}

static void tavorpv2_evb3_dsi_reset(void)
{
        int reset_pin;
        int err;

#if defined(TAVOR_EVB3_V2)
        reset_pin = mfp_to_gpio(MFP_PIN_GPIO61);
#else
        reset_pin = mfp_to_gpio(MFP_PIN_GPIO20);
#endif
        err = gpio_request(reset_pin, "DSI Reset");
        if (err) {
                gpio_free(reset_pin);
                printk(KERN_ERR "Request GPIO failed,"
                                "gpio: %d return :%d\n", reset_pin, err);
                return;
        }
        gpio_direction_output(reset_pin, 1);
        mdelay(1);
        gpio_direction_output(reset_pin, 0);
        mdelay(1);
        gpio_direction_output(reset_pin, 1);
        mdelay(10);
        gpio_free(reset_pin);
}

static struct fb_videomode video_modes[] = {
	[0] = {
		.pixclock       = 41701,
		.refresh        = 60,
		.xres           = 480,
		.yres           = 640,
		.hsync_len      = 19,
		.left_margin    = 40,
		.right_margin   = 59,
		.vsync_len      = 9,
		.upper_margin   = 4,
		.lower_margin   = 9,
		.sync           = 0,
	},
};

static struct pxa950fb_mach_info tavor_evb3_lcd_info __initdata = {
	.id                     = "Base",
	.modes                  = video_modes,
	.num_modes              = ARRAY_SIZE(video_modes),
	.pix_fmt_in             = PIX_FMTIN_RGB_16,
	.pix_fmt_out            = PIX_FMTOUT_16_RGB565,
	.panel_type             = LCD_Controller_Active,
	.mixer0                 = LCD_M2DSI0,
	/* choose this if you use PARALEEL */
	/* .mixer0                 = LCD_M2PARALELL_CONVERTER,*/
	.mixer1                 = LCD_MIXER_DISABLE,
	.active                 = 1,
	.pxa950fb_lcd_power     = espon_lcd_power,
	.pxa950fb_dsi_reset	= tavorpv2_evb3_dsi_reset,
	.invert_pixclock        = 1,
};

static struct pxa950fb_mach_info tavor_evb3_lcd_ovly_info __initdata = {
	.id                     = "Ovly",
	.modes                  = video_modes,
	.num_modes              = ARRAY_SIZE(video_modes),
	.pix_fmt_in             = PIX_FMTIN_YUV420,
	.pix_fmt_out            = PIX_FMTOUT_16_RGB565,
	.panel_type             = LCD_Controller_Active,
	.mixer0                 = LCD_M2DSI0,
	.mixer1                 = LCD_MIXER_DISABLE,
	.active                 = 1,
	.pxa950fb_lcd_power     = espon_lcd_power,
	.invert_pixclock        = 1,
};

static void __init tavor_evb3_init_lcd(void)
{
	set_pxa950_fb_info(&tavor_evb3_lcd_info);
	set_pxa950_fb_ovly_info(&tavor_evb3_lcd_ovly_info);
}

#endif

extern int is_uart_gpio(void);
#ifdef CONFIG_PM
extern irqreturn_t uartrx_detect_irq(int irq, void *_dev);

static mfp_cfg_t tavor_evb3_ffuart_enable_mfp_cfg[] = {
	/* FFUART */
	/* using MFP_CFG_LPM macro instead of calling the
	 * definition from mfp-pxa9xx.h since this is a
	 * platform specific configuration */
	MFP_CFG_LPM(GPIO54, AF1, PULL_HIGH),
	DF_ADDR0_GPIO_53,
};

static mfp_cfg_t tavor_evb3_ffuart_disable_mfp_cfg[] = {
	/* FFUART */
	GPIO53_GPIO,
	GPIO54_GPIO,
	DF_ADDR0_NONE,
};

static int tavor_evb3_ffuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO53);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);
		/* if edge detection is not clear, the altfun switch will
		 * gpio interrupt is generated on this gpio pin when
		 * switching pins from UART to gpio. so clear it.
		 */
		GRER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		GFER(gpio_rx) &= ~GPIO_bit(gpio_rx);
		pxa3xx_mfp_config(ARRAY_AND_SIZE(tavor_evb3_ffuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(tavor_evb3_ffuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
		GRER(gpio_rx) |= GPIO_bit(gpio_rx);
		GFER(gpio_rx) |= GPIO_bit(gpio_rx);

		ret = gpio_request(gpio_rx, "FFUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info tavor_evb3_ffuart_info = {
	.uart_pin_switch = tavor_evb3_ffuart_switch,
};

static mfp_cfg_t tavor_evb3_stuart_enable_mfp_cfg[] = {
	/* STUART */
	GPIO45_STRXD,
	GPIO46_STTXD,
};

static mfp_cfg_t tavor_evb3_stuart_disable_mfp_cfg[] = {
	/* STUART */
	GPIO45_GPIO,
	GPIO46_GPIO,
};

static int tavor_evb3_stuart_switch(int enable)
{
	int gpio_rx, irq_rx;
	int ret;
	gpio_rx = mfp_to_gpio(MFP_PIN_GPIO45);
	irq_rx = gpio_to_irq(gpio_rx);

	if (enable) {
		gpio_free(gpio_rx);

		pxa3xx_mfp_config(ARRAY_AND_SIZE(tavor_evb3_stuart_enable_mfp_cfg));
	} else {
		pxa3xx_mfp_config(ARRAY_AND_SIZE(tavor_evb3_stuart_disable_mfp_cfg));
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);

		ret = gpio_request(gpio_rx, "STUART RX");
		/* clear the MFPR */
		gpio_direction_input(gpio_rx);
	}
	return irq_rx;
}

static struct pxa_uart_mach_info tavor_evb3_stuart_info = {
	.uart_pin_switch = tavor_evb3_stuart_switch,
};

static int tavor_evb3_init_wakeup(pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	src->bits.rtc = 1;
	src->bits.ost = 1;
#ifdef CONFIG_PXA930_ACIPC
	src->bits.msl = 1;
#endif
	src->bits.ext0 = 1;
	src->bits.uart1 = 1;
	src->bits.mkey = 1;
	src->bits.eth = 1;
	src->bits.tsi = 1;
	src->bits.cmwdt = 1;
	return 0;
}

extern int pxa930_query_gwsr(int);
static int tavor_evb3_query_wakeup(unsigned int reg, pm_wakeup_src_t *src)
{
	memset(src, 0, sizeof(pm_wakeup_src_t));
	if (reg & PXA3xx_PM_WE_RTC)
		src->bits.rtc = 1;
	if (reg & PXA3xx_PM_WE_OST)
		src->bits.ost = 1;
	if (reg & PXA3xx_PM_WE_MSL0)
		src->bits.msl = 1;
	if (reg & PXA3xx_PM_WE_EXTERNAL0)
		src->bits.ext0 = 1;
	if (reg & PXA3xx_PM_WE_KP)
		src->bits.mkey = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(3))
		src->bits.tsi = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(9))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart1 = 1;
		}
	}
	if (reg & PXA3xx_PM_WE_GENERIC(2))
	{
		if (!is_uart_gpio())
		{
			src->bits.uart2 = 1;
		}
	}

	if (reg & PXA3xx_PM_WE_GENERIC(12))
		src->bits.cmwdt = 1;
	if (reg & PXA3xx_PM_WE_GENERIC(13)) {
		if (pxa930_query_gwsr(97))
			src->bits.eth = 1;
		if (pxa930_query_gwsr(53))
			src->bits.uart1 = 1;
	}
	return 0;
}

static int tavor_evb3_ext_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.ext0)
			ret |= PXA3xx_PM_WE_EXTERNAL0;
		if (src.bits.ext1)
			ret |= PXA3xx_PM_WE_EXTERNAL1;
	}
	return ret;
}

static int tavor_evb3_key_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_BOTH);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_KP;
		}
	} else {
		if (src.bits.mkey) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO0_KP_MKIN_0), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO2_KP_MKIN_1), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO4_KP_MKIN_2), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO6_KP_MKIN_3), MFP_EDGE_NONE);
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO8_KP_MKIN_4), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int tavor_evb3_mmc_wakeup(pm_wakeup_src_t src, int enable)
{
#if 0
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.mmc1_cd) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else
		pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO20_GPIO), MFP_EDGE_NONE);
#endif
	return 0;
}

static int tavor_evb3_uart_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.uart1) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO53_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(MFP_CFG_LPM(GPIO54, AF1, PULL_HIGH)), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(9);
			}
		}
		if (src.bits.uart2) {
			if (is_uart_gpio()) {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_GPIO), MFP_EDGE_FALL);
				ret |= PXA3xx_PM_WE_GENERIC(13);
			} else {
				pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_FALL);
				/* note: on pxa930, uart2 use this bit */
				ret |= PXA3xx_PM_WE_GENERIC(2);
			}
		}
	} else {
		if (src.bits.uart1) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(MFP_CFG_LPM(GPIO54, AF1, PULL_HIGH)), MFP_EDGE_NONE);
		}
		if (src.bits.uart2) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO45_STRXD), MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int tavor_evb3_eth_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_BOTH);
			ret |= PXA3xx_PM_WE_GENERIC(13);
		}
	} else {
		if (src.bits.eth) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(GPIO97_ETHER_IRQ),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int tavor_evb3_tsi_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_FALL);
			ret |= PXA3xx_PM_WE_GENERIC(3);
		}
	} else {
		if (src.bits.tsi) {
			pxa3xx_mfp_set_edge(MFP_CFG_PIN(PMIC_INT_GPIO_83),
					MFP_EDGE_NONE);
		}
	}
	return ret;
}

static int tavor_evb3_comm_wdt_wakeup(pm_wakeup_src_t src, int enable)
{
	unsigned int ret = 0;
	if (enable) {
		if (src.bits.cmwdt)
			ret |= PXA3xx_PM_WE_GENERIC(12);
	}
	return ret;
}

static struct pxa3xx_peripheral_wakeup_ops wakeup_ops = {
	.init	= tavor_evb3_init_wakeup,
	.query	= tavor_evb3_query_wakeup,
	.ext    = tavor_evb3_ext_wakeup,
	.key    = tavor_evb3_key_wakeup,
	.mmc    = tavor_evb3_mmc_wakeup,
	.uart   = tavor_evb3_uart_wakeup,
	.eth	= tavor_evb3_eth_wakeup,
	.tsi	= tavor_evb3_tsi_wakeup,
	.cmwdt	= tavor_evb3_comm_wdt_wakeup,
};
#else
static struct pxa_uart_mach_info tavor_evb3_ffuart_info = {
	.uart_pin_switch = NULL,
};
static struct pxa_uart_mach_info tavor_evb3_stuart_info = {
	.uart_pin_switch = NULL,
};
#endif

void tavor_evb3_stereo_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO84, "STEREO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO84);
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d return :%d\n", MFP_PIN_GPIO84, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO84, value);
	gpio_free(MFP_PIN_GPIO84);
}

void tavor_evb3_ext_mono_amp_en(int value)
{
	int err;

	err = gpio_request(MFP_PIN_GPIO63, "MONO_AMP_EN");
	if (err) {
		gpio_free(MFP_PIN_GPIO63);
		printk(KERN_ERR "Request GPIO failed,"
		       "gpio: %d return :%d\n", MFP_PIN_GPIO63, err);
		return;
	}
	gpio_direction_output(MFP_PIN_GPIO63, value);
	gpio_free(MFP_PIN_GPIO63);
}

static void tavor_evb3_set_lowpower_mfp(void)
{
	/* It will change mfp value set by OBM to save more power,
	 * it should be merged into OBM mfp setting eventually. */

	/* these values have been merged int OBM
	 * pxa3xx_mfp_write(48, 0xa440);
	 * pxa3xx_mfp_write(49, 0xa440);
	 * pxa3xx_mfp_write(51, 0xa440);
	 */
}


#define SMC_RESOURCE_START	(TAVOREVB3_ETH_PHYS + 0x300)
#define SMC_IO_EXTENT		16
#define BANK_SELECT		14

static int smc91x_exist(void)
{
	unsigned char __iomem *smc_membase;
	unsigned char __iomem *smc91x_membase;
	unsigned int csadrcfg3, temp;

	smc91x_membase = ioremap(SMC_RESOURCE_START, SMC_IO_EXTENT);
	smc_membase = ioremap(SMC_START, SMC_END - SMC_START + 1);
	csadrcfg3 = readl(smc_membase + CSADRCFG3_OFF);
	writel((csadrcfg3 & 0xFFFFFFF0) | 0xD, smc_membase + CSADRCFG3_OFF);
	while((readl(smc_membase + CSADRCFG3_OFF) & 0xF) != 0xD)
		;
	temp = *(unsigned short *)(smc91x_membase + BANK_SELECT);
	writel(csadrcfg3, smc_membase + CSADRCFG3_OFF);
	while(readl(smc_membase + CSADRCFG3_OFF) != csadrcfg3)
		;
	iounmap(smc_membase);
	iounmap(smc91x_membase);

	if (temp == 0x3300)
		return 1;
	else
		return 0;
}


#ifdef CONFIG_BATTERY_PDA
/*****************************************************************************/
/*                      pda_battery                                          */
/*****************************************************************************/
static struct pda_battery_pdata micco_pda_battery_pdata = {
	.name		  = "battery",
};

static struct platform_device pda_battery_device = {
	.name = "pda_battery",
	.id = -1,
	.dev = {
		.platform_data = &micco_pda_battery_pdata,
	},
};

static void tavor_evb3_init_pda_battery()
{
	platform_device_register(&pda_battery_device);
}
#endif


#ifdef CONFIG_PDA_POWER

static int pda_power_is_ac_online(void)
{
	return 0;
}

static int pda_power_is_usb_online(void)
{
	return 1;
}


static char *pda_power_supplicants[] = {
	"battery",
	"mon_battery",
};

static struct pda_power_pdata pda_power_supply_info = {
	.is_ac_online 		= pda_power_is_ac_online,
	.is_usb_online 		= pda_power_is_usb_online,
	.supplied_to 		= pda_power_supplicants,
	.num_supplicants 	= ARRAY_SIZE(pda_power_supplicants),
};

static struct platform_device pda_power_supply = {
	.name = "pda-power",
	.id = -1,
	.dev = {
		.platform_data = &pda_power_supply_info,
	},
};

static void tavor_evb3_init_pda_power()
{
	platform_device_register(&pda_power_supply);
}
#endif

static void __init evb3_init(void)
{

	gpio_direction_output(MFP_PIN_GPIO92,0); /* turn off the torch (flash light)*/
	/* register i2c device */
	pxa_set_i2c_info   (&pxa3xx_i2c_platform_data);
	pxa_set_hs_i2c_info(&pxa3xx_i2c_hs_platform_data);

#ifdef CONFIG_IMM
	/* update ISRAM configration for pxa935 */
	if (cpu_is_pxa935())
		pxa3xx_device_imm.resource = pxa935_resource_imm;
#endif

	/* check whether the platform is 2.6 by micco's version currently */
	/* saar pxa935-1.1, same as saar v2.6 board for MMC and IrDA */

	g_pmic_id = pmic_id();

	/* dvfm device */
	set_pxa3xx_freq_info(&tavor_evb3_freq_mach_info);

	/* performance monitor unit */
	pxa3xx_set_pmu_info(NULL);

#if defined(CONFIG_SND_PXA3XX_SOC_TAVOR_EVB3)
	/* audio */
	//pxa3xx_mfp_config(ARRAY_AND_SIZE(audio_mfp_cfg));  /* mfp should NOT be configured here! */
#endif

	/* lcd */
#if defined(CONFIG_FB_PXA950)
	pxa3xx_mfp_config(ARRAY_AND_SIZE(lcd_mfp_cfg));
	tavor_evb3_init_lcd();
#endif
	/* ethernet */
	gpio_direction_input(mfp_to_gpio(MFP_PIN_GPIO97));
	if(smc91x_exist())
		platform_device_register(&smc91x_device);

    if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_END_ID) ){
		platform_add_devices(devices_levante, ARRAY_SIZE(devices_levante));
		printk(KERN_ALERT "--------->platform_add_devices(devices_levante)!!!\n");
	}else if( (g_pmic_id >= MICCO_EA_ID ) && (g_pmic_id <= MICCO_EB_ID) ){
 		platform_add_devices(devices_micco, ARRAY_SIZE(devices_micco));
		printk(KERN_ALERT "--------->platform_add_devices(devices_micco)!!!\n");
	}else
	   printk(KERN_ALERT "--------->platform_add_devices(no devices_micco & no devices_levante )!!!\n");

	platform_add_devices(devices, ARRAY_SIZE(devices));

	if(is_uart_gpio()) {
		pxa_set_ffuart_info(&tavor_evb3_ffuart_info);
#ifndef CONFIG_ISPT
		pxa_set_stuart_info(&tavor_evb3_stuart_info);
#endif
	} else {
		pxa_set_ffuart_info(NULL);
#ifndef CONFIG_ISPT
		pxa_set_stuart_info(NULL);
#endif
	}
	pxa_set_btuart_info(NULL);

#if defined(CONFIG_KEYBOARD_PXA27x) || defined(CONFIG_KEYBOARD_PXA27x_MODULE)
	pxa_set_keypad_info(&tavor_evb3_keypad_info);
#endif

	/* trackball */
	tavor_evb3_init_trackball();

	/* nand */
	tavor_evb3_init_nand();

	/* onenand */
	tavor_evb3_init_onenand();

#ifdef CONFIG_ISPT
	tavor_evb3_init_ispt();
#endif
#if defined(CONFIG_USB_PXA3XX_U2D) || defined(CONFIG_USB_USB_PXA3XX_U2D_MODULE)
	/* u2d */
	if (device_is_enabled(device_u2d)) {
		pxa_set_u2d_info(&tavor_evb3_u2d_info);
	}
#endif

#if defined(CONFIG_USB_PXA27X_UDC) || defined(CONFIG_USB_USB_PXA27X_UDC_MODULE)
	if (device_is_enabled(device_udc)) {
		pxa_set_udc_info(&tavor_evb3_udc_info);
	}
#endif
	/* usb otg */
	tavor_evb3_init_otg();

#if defined(CONFIG_USB_PXA9XX_U2O) || defined(CONFIG_USB_USB_PXA9XX_U2O_MODULE)
	/* u2o */
	if (device_is_enabled(device_u2o)) {
		pxa_set_u2o_info(pxa9xx_u2o_resources, &tavor_evb3_u2o_info);
	}
#endif
#if defined(CONFIG_PXA950_CAMERA)
	/* initialize camera for both two camera interface */
	tavor_evb3_init_cam(CCIC_0);
	tavor_evb3_init_cam(CCIC_1);
#endif

	/* i2c devices */
#if defined(CONFIG_SENSORS_LIS311DL)
	/*dummy driver init*/
	platform_device_register(&sensor_input_device);
#endif
  if  ((g_pmic_id >= LEVATNE_A0_ID) && (g_pmic_id <=  LEVATNE_END_ID) ){

  	i2c_register_board_info(0, ARRAY_AND_SIZE(evb3_i2c_board_info_levante1));
    i2c_register_board_info(0, ARRAY_AND_SIZE(evb3_i2c_board_info_levante2));
   	printk(KERN_ALERT "i2c_register_board_info_levante register !!!\n");

  }else if( (g_pmic_id >= MICCO_EA_ID ) && (g_pmic_id <= MICCO_EB_ID) ){

  	i2c_register_board_info(0, ARRAY_AND_SIZE(evb3_i2c_board_info_micco));
	printk(KERN_ALERT "i2c_register_board_info_micco register!!!\n");

  }else
	printk(KERN_ALERT "i2c_register_board_info PMIC *not* register !!!\n");

	i2c_register_board_info(2, ARRAY_AND_SIZE(tavor_evb3_i2c_hs_info));

#if defined(CONFIG_PXA_IRDA) || defined(CONFIG_PXA_IRDA_MODULE)
	pxa_set_ficp_info(&tavor_evb3_ficp_platform_data);
#endif
	tavor_evb3_init_mmc();

	tavor_evb3_init_ehci();

#ifdef CONFIG_PM
	pxa3xx_wakeup_register(&wakeup_ops);
#endif
	tavor_evb3_set_lowpower_mfp();

#ifdef	CONFIG_PROC_FS
	create_evb3_proc_file();
#endif

#ifdef CONFIG_BATTERY_PDA
	tavor_evb3_init_pda_battery();
#endif

#ifdef CONFIG_PDA_POWER
	tavor_evb3_init_pda_power();
#endif
}

MACHINE_START(TAVOREVB3, "PXA950 handheld Platform (TAVOR EVB III)")
	.phys_io        = 0x40000000,
	.boot_params    = 0xa0000100,
	.io_pg_offst    = (io_p2v(0x40000000) >> 18) & 0xfffc,
	.map_io         = pxa_map_io,
	.init_irq       = pxa3xx_init_irq,
	.timer          = &pxa_timer,
	.init_machine   = evb3_init,
MACHINE_END
