/*
 * linux/arch/arm/mach-pxa/test_bpm.c
 *
 * Provide the test cases for bpm.c
 *
 * Copyright (C) 2008 Borqs Corporation.
 *
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This software program is licensed subject to the GNU General Public License
 * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html
 *
 */

#include <mach/mfp-pxa9xx.h>

struct kernel_test {
        const char *name;
        int enabled;
        int (*pre_test) (void *data);
        int (*test) (void *data);
        int (*post_test) (void *data);
        int result;
};

enum {
        D0CS = 0,
        M156,
        M208,
        M416,
        M624,
        D2,
        BOOTOP,
};

static struct dvfm_op g_ut_dyn_ops[BPM_MAX_OP_NUM];
static int g_ut_active_ops_map[BPM_MAX_OP_NUM];
static int g_ut_active_ops_num;
static int g_ut_active_cur_idx;
static int g_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1];
static struct bpm_freq_policy g_ut_active_policy[BPM_MAX_OP_NUM];

static int save_global_variables(void *data)
{
	memcpy(g_ut_dyn_ops, g_dyn_ops, sizeof(g_dyn_ops));
	memcpy(g_ut_active_ops_map, g_active_ops_map, sizeof(g_active_ops_map));
	g_ut_active_ops_num = g_active_ops_num;
	g_ut_active_cur_idx = g_active_cur_idx;
	memcpy(&g_ut_active_bonus, &g_active_bonus, sizeof(g_active_bonus));
	memcpy(g_ut_active_policy, g_active_policy, sizeof(g_active_policy));
	return 0;
}

static int restore_global_variables(void *data)
{
	memcpy(g_active_policy, g_ut_active_policy, sizeof(g_active_policy));
	memcpy(&g_active_bonus, &g_ut_active_bonus, sizeof(g_active_bonus));
	g_active_cur_idx = g_ut_active_cur_idx;
	g_active_ops_num = g_ut_active_ops_num;
	memcpy(g_active_ops_map, g_ut_active_ops_map, sizeof(g_active_ops_map));
	memcpy(g_dyn_ops, g_ut_dyn_ops, sizeof(g_dyn_ops));
	return 0;
}

static int build_dyn_ops_test0(void *data)
{
        int i, ret = 0;
	unsigned long flag;

        struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
                {D0CS, 1, 60, "D0CS"},
                {M156, 0, 156, "156M"},
                {M208, 0, 208, "208M"},
                {M416, 0, 416, "416M"},
                {M624, 0, 624, "624M"},
                {D2, 1, 0, "D2"},
                {BOOTOP, 1, 624, "BOOT OP"},
                {-1, 0, 0, NULL}
        };

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

        build_dyn_ops();

        for (i = 0; i < BPM_MAX_OP_NUM; ++i) {
		if ((ut_dyn_ops[i].index == -1) && (g_dyn_ops[i].index == -1))
			break;

                if (ut_dyn_ops[i].index != g_dyn_ops[i].index) {
                        ret = -1;
                        break;
                }

                if (ut_dyn_ops[i].cpu_freq != g_dyn_ops[i].cpu_freq) {
                        ret = -1;
                        break;
                }

                if (strcmp(ut_dyn_ops[i].name, g_dyn_ops[i].name)) {
                        ret = -1;
                        break;
                }

        }

        if (ret != 0) {
                bpm_dump_policy();
        }

        restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

        return ret;
}

/*
 * Case: only M208 and M416 are enabled.
 */
static int build_active_ops_test0(void *data)
{
	int tmp_ret, ret = 0;
	unsigned long flag;

	/* input */
	static struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
		{D0CS,   1,  60, "D0CS"   },
		{M156,   1, 156, "156M"   },
		{M208,   0, 208, "208M"   },
		{M416,   0, 416, "416M"   },
		{M624,   1, 624, "624M"   },
		{D2,     1,   0, "D2"     },
		{BOOTOP, 1, 624, "BOOT OP"},
                {-1, 0, 0, NULL},
                {-1, 0, 0, NULL},
                {-1, 0, 0, NULL},
	};

	/* output */
	static int res_ut_active_ops_map[BPM_MAX_OP_NUM] =
	    	{ 2, 3, -1, -1, -1, -1, -1, -1, -1, -1};
	static int res_ut_active_ops_num = 2;
	static int res_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1] = {
		{ 0,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  1,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
	};

	static struct bpm_freq_policy res_ut_active_policy[BPM_MAX_OP_NUM] = {
		{{0,   0,  0}, {80, 95, 100}},
		{{0,   0, 40}, {100, 100, 100}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}}
	};

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

	/* fake the input */
	memcpy(g_dyn_ops, ut_dyn_ops, sizeof(g_dyn_ops));

	build_active_ops();

	/* check the result */
	tmp_ret =
	    memcmp(res_ut_active_ops_map, g_active_ops_map,
		   sizeof(g_active_ops_map));
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_map is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_bonus, g_active_bonus, sizeof(g_active_bonus));
	if (tmp_ret) {
		printk(KERN_ERR "active_bonus is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_policy, g_active_policy,
		   sizeof(g_active_policy));
	if (tmp_ret) {
		printk(KERN_ERR "active_policy is not correct!\n");
		ret = -1;
	}

	tmp_ret = (res_ut_active_ops_num == g_active_ops_num ? 0 : -1);
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_num is not correct!\n");
		ret = -1;
	}

	if (ret != 0) {
		bpm_dump_policy();
	}

	restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

	return ret;
}

/*
 * M204, M416, M624 are enabled.
 */
static int build_active_ops_test1(void *data)
{
	int tmp_ret, ret = 0;
	unsigned long flag;

	/* input */
	static struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
		{D0CS, 1, 60, "D0CS"},
		{M156, 1, 156, "156M"},
		{M208, 0, 208, "208M"},
		{M416, 0, 416, "416M"},
		{M624, 0, 624, "624M"},
		{D2, 1, 0, "D2"},
		{BOOTOP, 1, 624, "BOOT OP"},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
	};

	/* output */
	static int res_ut_active_ops_map[BPM_MAX_OP_NUM] =
	    { 2, 3, 4, -1, -1, -1, -1, -1, -1, -1};
	static int res_ut_active_ops_num = 3;
	static int res_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1] = {
		{ 0,  0,  0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  1,  2,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  1,  2,  2,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
	};

	static struct bpm_freq_policy res_ut_active_policy[BPM_MAX_OP_NUM] = {
		{{0,  0,   0}, { 80, 100, 100}},
		{{0,  0,  40}, { 96, 100, 100}},
		{{0,  26, 64}, {100, 100, 100}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}}
	};

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

	/* fake the input */
	memcpy(g_dyn_ops, ut_dyn_ops, sizeof(g_dyn_ops));

	build_active_ops();

	/* check the result */
	tmp_ret =
	    memcmp(res_ut_active_ops_map, g_active_ops_map,
		   sizeof(g_active_ops_map));
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_map is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_bonus, g_active_bonus, sizeof(g_active_bonus));
	if (tmp_ret) {
		printk(KERN_ERR "active_bonus is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_policy, g_active_policy,
		   sizeof(g_active_policy));
	if (tmp_ret) {
		printk(KERN_ERR "active_policy is not correct!\n");
		ret = -1;
	}

	tmp_ret = (res_ut_active_ops_num == g_active_ops_num ? 0 : -1);
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_num is not correct!\n");
		ret = -1;
	}

	if (ret != 0) {
		bpm_dump_policy();
	}

        restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

	return ret;
}

/*
 * D0CS, M208, M416, M624 are enabled.
 */
static int build_active_ops_test2(void *data)
{
	int tmp_ret, ret = 0;
	unsigned long flag;

	/* input */
	static struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
		{D0CS, 0, 60, "D0CS"},
		{M156, 1, 156, "156M"},
		{M208, 0, 208, "208M"},
		{M416, 0, 416, "416M"},
		{M624, 0, 624, "624M"},
		{D2, 1, 0, "D2"},
		{BOOTOP, 1, 624, "BOOTOP"},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
	};

	/* output */
	static int res_ut_active_ops_map[BPM_MAX_OP_NUM] =
	    { 0, 2, 3, 4, -1, -1, -1, -1, -1, -1};
	static int res_ut_active_ops_num = 4;
	static int res_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1] = {
		{ 0,  0,  0,  0,  1,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  0,  1,  2,  3,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  1,  2,  3,  3,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  1,  2,  3,  3,  3,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
	};

	static struct bpm_freq_policy res_ut_active_policy[BPM_MAX_OP_NUM] = {
		{{0, 0, 0},  { 90, 95, 100}},
		{{0, 0, 25}, { 80, 100, 100}},
		{{0, 12, 40}, { 96, 100, 100}},
		{{0, 26, 64}, {100, 100, 100}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}}
	};

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

	/* fake the input */
	memcpy(g_dyn_ops, ut_dyn_ops, sizeof(g_dyn_ops));

	build_active_ops();

	/* check the result */
	tmp_ret =
	    memcmp(res_ut_active_ops_map, g_active_ops_map,
		   sizeof(g_active_ops_map));
	if (0 != tmp_ret) {
		printk(KERN_ERR "active_ops_map is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_bonus, g_active_bonus, sizeof(g_active_bonus));
	if (0 != tmp_ret) {
		printk(KERN_ERR "active_bonus is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_policy, g_active_policy,
		   sizeof(g_active_policy));
	if (0 != tmp_ret) {
		printk(KERN_ERR "active_policy is not correct!\n");
		ret = -1;
	}

	tmp_ret = (res_ut_active_ops_num == g_active_ops_num ? 0 : -1);
	if (0 != tmp_ret) {
		printk(KERN_ERR "active_ops_num is not correct!\n");
		ret = -1;
	}

	if (ret != 0) {
		bpm_dump_policy();
	}

        restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

	return ret;
}

/*
 * D0CS, M156, M208, M416, M624 are enabled.
 */
static int build_active_ops_test3(void *data)
{
	int tmp_ret, ret = 0;
	unsigned long flag;

	/* input */
	static struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
		{D0CS, 0, 60, "D0CS"},
		{M156, 0, 156, "156M"},
		{M208, 0, 208, "208M"},
		{M416, 0, 416, "416M"},
		{M624, 0, 624, "624M"},
		{D2, 1, 0, "D2"},
		{BOOTOP, 1, 624, "BOOT OP"},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
	};

	/* output */
	static int res_ut_active_ops_map[BPM_MAX_OP_NUM] =
	    { 0, 1, 2, 3, 4, -1, -1, -1, -1, -1 };
	static int res_ut_active_ops_num = 5;
	static int res_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1] = {
		{ 0,  0,  0,  0,  0,  1,  2,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  0,  0,  1,  2,  3,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  0,  1,  2,  3,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  1,  2,  3,  4,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  1,  2,  3,  4,  4,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
	};

	static struct bpm_freq_policy res_ut_active_policy[BPM_MAX_OP_NUM] = {
		{{0, 0, 0}, {90, 95, 100}},
		{{0, 0, 34}, {80, 95, 100}},
		{{0, 25, 60}, {80, 100, 100}},
		{{0, 30, 40}, {96, 100, 100}},
		{{0, 26, 64}, {100, 100, 100}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
	};

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

	/* fake the input */
	memcpy(g_dyn_ops, ut_dyn_ops, sizeof(g_dyn_ops));

	build_active_ops();

	/* check the result */
	tmp_ret =
	    memcmp(res_ut_active_ops_map, g_active_ops_map,
		   sizeof(g_active_ops_map));
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_map is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_bonus, g_active_bonus, sizeof(g_active_bonus));
	if (tmp_ret) {
		printk(KERN_ERR "active_bonus is not correct!\n");
		ret = -1;
	}

	tmp_ret =
	    memcmp(res_ut_active_policy, g_active_policy,
		   sizeof(g_active_policy));
	if (tmp_ret) {
		printk(KERN_ERR "active_policy is not correct!\n");
		ret = -1;
	}

	tmp_ret = (res_ut_active_ops_num == g_active_ops_num ? 0 : -1);
	if (tmp_ret) {
		printk(KERN_ERR "active_ops_num is not correct!\n");
		ret = -1;
	}

	if (ret != 0) {
		bpm_dump_policy();
	}

        restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

	return ret;
}

static int handle_profiler_arg_test0(void *data)
{
	struct bpm_freq_bonus_arg bonus_arg;
	int bonus;
	int new_idx;
	int old_dyn_idx, new_dyn_idx;
	int ret = 0;
        unsigned long flag;

	static struct dvfm_op ut_dyn_ops[BPM_MAX_OP_NUM] = {
		{D0CS, 0, 60, "D0CS"},
		{M156, 0, 156, "156M"},
		{M208, 0, 208, "208M"},
		{M416, 0, 416, "416M"},
		{M624, 0, 624, "624M"},
		{D2, 1, 0, "D2"},
		{BOOTOP, 1, 624, "BOOT OP"},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
		{-1, 0, 0, NULL},
	};

	/* output */
	static int res_ut_active_ops_map[BPM_MAX_OP_NUM] =
	    { 0, 1, 2, 3, 4, -1, -1, -1, -1, -1 };
	static int res_ut_active_ops_num = 5;
	static int res_ut_active_bonus[BPM_MAX_OP_NUM][BPM_MAX_OP_NUM * 2 - 1] = {
		{ 0,  0,  0,  0,  0,  1,  2,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  0,  0,  1,  2,  3,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  0,  1,  2,  3,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  0,  1,  2,  3,  4,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{ 0,  1,  2,  3,  4,  4,  4,  4,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
	};

	static struct bpm_freq_policy res_ut_active_policy[BPM_MAX_OP_NUM] = {
		{{0, 0, 0}, {90, 95, 100}},
		{{0, 0, 34}, {80, 95, 100}},
		{{0, 25, 60}, {80, 100, 100}},
		{{0, 30, 40}, {96, 100, 100}},
		{{0, 26, 64}, {100, 100, 100}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
		{{-1, -1, -1}, {-1, -1, -1}},
	};

        spin_lock_irqsave(&g_dyn_ops_lock, flag);
        save_global_variables(NULL);

	/* fake the input */
	memcpy(g_dyn_ops, ut_dyn_ops, sizeof(g_dyn_ops));
	memcpy(g_active_ops_map, res_ut_active_ops_map,
	       sizeof(g_active_ops_map));
	g_active_ops_num = res_ut_active_ops_num;
	memcpy(g_active_bonus, res_ut_active_bonus, sizeof(g_active_bonus));
	memcpy(g_active_policy, res_ut_active_policy, sizeof(g_active_policy));

	g_active_cur_idx = g_active_ops_num - 1;

	/* sub case 0: */
	bonus_arg.mips = 90 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);
	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != 0 && new_idx != 4 && old_dyn_idx != M624
	    && new_dyn_idx != M624) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	/* sub case 1: */
	bonus_arg.mips = 100 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != 0 && new_idx != 4 && old_dyn_idx != M624
	    && new_dyn_idx != M624) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	/* sub case 2: */
	bonus_arg.mips = 53 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != 0 && new_idx != 4 && old_dyn_idx != M624
	    && new_dyn_idx != M624) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	/* sub case 3: */
	bonus_arg.mips = 52 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != -1 && new_idx != 3 && old_dyn_idx != M624
	    && new_dyn_idx != M416) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	/* sub case 4: */
	bonus_arg.mips = 4 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != -2 && new_idx != 2 && old_dyn_idx != M624
	    && new_dyn_idx != M208) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	g_active_cur_idx = 1;

	/* sub case 5: */
	bonus_arg.mips = 4 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != -2 && new_idx != 0 && old_dyn_idx != M156
	    && new_dyn_idx != D0CS) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	/* sub case 6: */
	bonus_arg.mips = 96 * get_cur_freq() / 100;
	bonus_arg.mem_stall = 0;
	bonus = calc_bonus(&bonus_arg);
	new_idx = calc_new_idx(bonus);

	old_dyn_idx = get_dyn_idx(g_active_cur_idx);
	new_dyn_idx = get_dyn_idx(new_idx);

	if (bonus != 2 && new_idx != 3 && old_dyn_idx != M156
	    && new_dyn_idx != M416) {
		printk(KERN_ERR
		       "line %d calc error (bonus: %d new_idx: %d old_dyn_idx: %d new_dyn_idx: %d)\n",
		       __LINE__, bonus, new_idx, old_dyn_idx, new_dyn_idx);
		ret = -1;
	}

	if (0 != ret) {
		printk(KERN_ERR "calc bonus or new idx failed\n");
	}

        restore_global_variables(NULL);
        spin_unlock_irqrestore(&g_dyn_ops_lock, flag);

	return ret;
}

static int bpm_dis_enb_op_name_test0(void *data)
{
	int ret = 0;
#if 0	
	ret = bpm_disable_op_name("D0CS", "ut");
	if (ret != 0) 
		goto dis_enb_op_out;
	ret = bpm_disable_op_name("D2", "ut");
	if (ret != 0)
		goto dis_enb_op_out;

	ret = bpm_disable_op_name("156M", "ut");
	if (ret != 0)
		goto dis_enb_op_out;

        ret = bpm_enable_op_name("156M", "ut");
        if (ret != 0)
		goto dis_enb_op_out;

        ret = bpm_enable_op_name("D2", "ut");
        if (ret != 0)
		goto dis_enb_op_out;

        ret = bpm_enable_op_name("D0CS", "ut");
        if (ret != 0)
		goto dis_enb_op_out;
dis_enb_op_out:
#endif
	return ret;
}

static int bpm_change_op_test0(void *data)
{
	return 0;
}

static int dvfm_set_op_test0(void *data)
{
	extern int cur_op;
        struct dvfm_freqs freqs;
	int i, j, m;
	int ret;
	int freq_old = cur_op;

	for (m = 0; m < 100; ++m) {
		for (i = 0; i <= 4; ++i) {
			printk(KERN_ERR "m:%d change op to (%d)\n", m, i);
                        freqs.old = cur_op;
                        freqs.new = i;
			ret = dvfm_set_op(&freqs, freqs.new, RELATION_STICK);
			
			for (j = 0; j <= 4; ++j) {
				memset(&freqs, 0, sizeof(struct dvfm_freqs));

				freqs.old = cur_op;
				freqs.new = j;
				printk(KERN_ERR "m:%d change op form (%d) to (%d)\n", m, freqs.old, freqs.new);
				ret = dvfm_set_op(&freqs, freqs.new, RELATION_STICK);
			
				freqs.old = cur_op;
				freqs.new = i;
				printk(KERN_ERR "m:%d change op form (%d) to (%d)\n", m, freqs.old, freqs.new);
				ret = dvfm_set_op(&freqs, freqs.new, RELATION_STICK);
			}
		}
	}

        freqs.old = cur_op;
        freqs.new = freq_old;
        printk(KERN_ERR "go back: change op form (%d) to (%d)\n", freqs.old, freqs.new);
        ret = dvfm_set_op(&freqs, freqs.new, RELATION_STICK);

	return 0;
}

static int dvfm_set_op_test1(void *data)
{
	extern int cur_op;
        struct dvfm_freqs freqs;
	int m;
	int ret = 0;
	int freq_old;

	freqs.new = cur_op;
	freq_old = cur_op;

        for (m = 0; m < 1000; ++m) {
       		freqs.old = freqs.new;
                freqs.new = OSCR % 5;
                printk(KERN_ERR "m:%d change op form (%d) to (%d)\n", m, freqs.old, freqs.new);
		
                if (dvfm_set_op(&freqs, freqs.new, RELATION_STICK)) {
			freqs.new = freqs.old;
		}
        }

        freqs.old = freqs.new;
        freqs.new = freq_old;
        printk(KERN_ERR "go back: change op form (%d) to (%d)\n", freqs.old, freqs.new);
        ret = dvfm_set_op(&freqs, freqs.new, RELATION_STICK);

        return ret;
}

static int bpm_current_test0(void *data)
{
        extern int cur_op;
        int i;
        int ret = 0;
	int cur_idx, new_idx;
	unsigned long oscr;

	if (g_active_cur_idx == -1) {
		change_to_active_op();
	}

	for (i = 0; i < 5; ++i) {
		printk(KERN_ERR "prefix...%d\n", i);
		cur_idx = get_dyn_idx(g_active_cur_idx);
		new_idx = get_dyn_idx(g_active_ops_num - 1);
		bpm_change_op(cur_idx, new_idx);
		g_active_cur_idx = g_active_ops_num - 1;
		
	        oscr = OSCR;
        	while(time_before((unsigned long)OSCR, oscr + 3250 * 3 * 1000));

                cur_idx = get_dyn_idx(g_active_cur_idx);
                new_idx = get_dyn_idx(0);
                bpm_change_op(cur_idx, new_idx);
                g_active_cur_idx = 0;

                oscr = OSCR;
                while(time_before((unsigned long)OSCR, oscr + 3250 * 3 * 1000));
	}

	for (i = g_active_ops_num - 1; i >= 0; --i) {
                cur_idx = get_dyn_idx(g_active_cur_idx);
                new_idx = get_dyn_idx(i);
		printk(KERN_ERR "change op %d -> %d\n", cur_idx, new_idx);
                bpm_change_op(cur_idx, new_idx);
                g_active_cur_idx = i;

                oscr = OSCR;
                while(time_before((unsigned long)OSCR, oscr + 3250 * 30 * 1000));
	}

        for (i = 0; i < 5; ++i) {
		printk(KERN_ERR "postfix...%d\n", i);
                cur_idx = get_dyn_idx(g_active_cur_idx);
                new_idx = get_dyn_idx(g_active_ops_num - 1);
                bpm_change_op(cur_idx, new_idx);
                g_active_cur_idx = g_active_ops_num - 1;

                oscr = OSCR;
                while(time_before((unsigned long)OSCR, oscr + 3250 * 3 * 1000));

                cur_idx = get_dyn_idx(g_active_cur_idx);
                new_idx = get_dyn_idx(0);
                bpm_change_op(cur_idx, new_idx);
                g_active_cur_idx = 0;

                oscr = OSCR;
                while(time_before((unsigned long)OSCR, oscr + 3250 * 3 * 1000));
        }

        return ret;
}

static int dvfm_dis_enb_test0(void *data)
{
	//dvfm_disable();
	
	schedule_timeout(1000);
	
	//dvfm_enable();
	
        return 0;
}

static struct kernel_test g_bpm_tests[] = {
        {
         "build_dyn_ops_test0",
         0,
	 NULL,
         build_dyn_ops_test0,
	 NULL,
         -2,
         },
	{
	 "build_active_ops_test0",
	 0,
	 NULL,
	 build_active_ops_test0,
	 NULL,
	 -2,
	 },
	{
	 "build_active_ops_test1",
	 0,
	 NULL,
	 build_active_ops_test1,
	 NULL,
	 -2,
	 },
	{
	 "build_active_ops_test2",
	 0,
	 NULL,
	 build_active_ops_test2,
	 NULL,
	 -2,
	 },
	{
	 "build_active_ops_test3",
	 0,
	 NULL,
	 build_active_ops_test3,
	 NULL,
	 -2,
	 },
	{
	 "handle_profiler_arg_test0",
	 0,
	 NULL,
	 handle_profiler_arg_test0,
	 NULL,
	 -2,
	 },
	{
	 "bpm_dis_enb_op_name_test0",
	 0,
	 NULL,
	 bpm_dis_enb_op_name_test0,
	 NULL,
	 -2,
	 },
        {
         "bpm_change_op_test0",
         0,
	 NULL,
         bpm_change_op_test0,
	 NULL,
         -2,
         },

        {
         "dvfm_set_op_test0",
         0,
	 NULL,
	 dvfm_set_op_test0,
	 NULL,
         -2,
         },

        {
         "dvfm_set_op_test1",
         0,
	 NULL,
	 dvfm_set_op_test1,
	 NULL,
         -2,
         },
        {
         "bpm_current_test0",
         0,
	 NULL,
         bpm_current_test0,
	 NULL,
         -2,
         },
        {
         "dvfm_dis_enb_test0",
         0,
         NULL,
         dvfm_dis_enb_test0,
         NULL,
         -2,
         },
};

#define BPM_TEST_CASE_NUM (sizeof(g_bpm_tests) / sizeof(struct kernel_test)) 

/*****************************************************************************/
/*                                                                           */
/*                        IPMD TEST SYS Interface                            */
/*                                                                           */
/*****************************************************************************/

static ssize_t cpuload_mode_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
	char *s = buf;

	s += sprintf(s, "0: %s; 1: %s; 2:%s; 3:%s 4:%s 5:%s\n",
		     "really", "everytime", "random", "random everytime", 
		     "low usage", "high usage");

	s += sprintf(s, "def: %d\n", g_cpuload_mode);

	return (s - buf);
}

static ssize_t cpuload_mode_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
	sscanf(buf, "%u", &g_cpuload_mode);

	if (g_cpuload_mode > 5 || g_cpuload_mode < 0) {
		pr_info("Error Input\n");
		g_cpuload_mode = 0;
	}
	pr_info("Done, Change cpuload mode to %d\n", g_cpuload_mode);

	return count;
}

SYSDEV_ATTR(cpuload_mode, 0644, cpuload_mode_show, cpuload_mode_store);

static ssize_t ut_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
	char *s = buf;
	int i;
	struct kernel_test *ptest = NULL;

	for (i = 0; i < BPM_TEST_CASE_NUM; ++i) {
		ptest = &g_bpm_tests[i];
		s += sprintf(s, "id: %3d:    enable: %d;    result: %2d    desc: %s\n",
			     i, ptest->enabled, ptest->result, ptest->name);
	}

	return (s - buf);
}

static ssize_t ut_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
	int i;
	struct kernel_test *ptest = NULL;
	unsigned long flags;

	if (count >= strlen("run") && strncmp(buf, "run", strlen("run")) == 0) {
		for (i = 0; i < BPM_TEST_CASE_NUM; ++i) {
			ptest = &g_bpm_tests[i];
			local_irq_save(flags);
			if (1 == ptest->enabled) {
				printk(KERN_ERR "Running %s\n", ptest->name);
				if (ptest->pre_test)
					ptest->pre_test(NULL);
				if (ptest->test)
					ptest->result = ptest->test(NULL);
				if (ptest->post_test)
					ptest->post_test(NULL);
			}
			local_irq_restore(flags);
		}
		return count;
	}

	if (count >= strlen("clear") && strncmp(buf, "clear", strlen("clear")) == 0) {
		for (i = 0; i < BPM_TEST_CASE_NUM; ++i) {
			ptest = &g_bpm_tests[i];
			ptest->result = -2;
		}
		return count;
	}

	printk(KERN_ERR "invaid input, please try \"run\" or \"clear\"\n");
	return count;
}

SYSDEV_ATTR(ut, 0644, ut_show, ut_store);

static ssize_t ut_config_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
	char *s = buf;
	int i;
	struct kernel_test *ptest = NULL;

	for (i = 0; i < BPM_TEST_CASE_NUM; ++i) {
		ptest = &g_bpm_tests[i];
		s += sprintf(s, "%4d: %4d; \t%s\n", i, ptest->enabled,
			     ptest->name);
	}

	return (s - buf);
}

static ssize_t ut_config_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
	int idx;
	int enabled;
	struct kernel_test *ptest = NULL;

	sscanf(buf, "%u %u", &idx, &enabled);

	if ((idx >= 0 && idx < BPM_TEST_CASE_NUM)
	    && (enabled >= 0 && enabled <= 1)) {
		ptest = &g_bpm_tests[idx];
		ptest->enabled = enabled;
	} else {
		printk(KERN_ERR
		       "bpm config error input (case id: %d, enabled: %d) error!\n",
		       idx, enabled);
	}

	return count;
}

SYSDEV_ATTR(ut_config, 0644, ut_config_show, ut_config_store);

static ssize_t mfp_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
        int mfp;
        unsigned int val;
	static int tag = 1;

        char *s = buf;

	if (tag % 2) {
        	for (mfp = 0; mfp < 150; ++mfp) {
                	val = pxa3xx_mfp_read(mfp);
	                s += sprintf(s, "%d 0x%08x\n", mfp, val);
        	}
	} else {
		for (mfp = 150; mfp < MFP_PIN_MAX; ++mfp) {
                        val = pxa3xx_mfp_read(mfp);
                        s += sprintf(s, "%d 0x%08x\n", mfp, val);
		}
	}
	++tag;

        return (s - buf);
}

static ssize_t mfp_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
        int mfp;
        unsigned int val;

        sscanf(buf, "%d %x", &mfp, &val);

        pxa3xx_mfp_write(mfp, val);

        return count;
}

SYSDEV_ATTR(mfp, 0644, mfp_show, mfp_store);

static ssize_t idlectl_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
	return 0;
}

#define OST_C4		(1 << 4)
extern void pxa3xx_cpu_idle(void);
static ssize_t idlectl_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
	unsigned int ctl = 0;
	unsigned long flags;
	unsigned long next;

	sscanf(buf, "%d", &ctl);

	printk("before cpu idle\n");
	local_irq_save(flags);
	switch (ctl) {
		case 0:
			//OIER &= ~OST_C4;
			pxa3xx_cpu_idle();
			next = OSCR4;
			//OSSR = OST_C4;
			//OSMR4 = next + HZ;
			//OIER |= OST_C4;
		default:;
	}
	local_irq_restore(flags);
	printk("after cpu idle\n");
	return count;
}

SYSDEV_ATTR(idlectl, 0644, idlectl_show, idlectl_store);


static ssize_t lpm_show(struct sys_device *dev, struct sysdev_attribute *attr, char *buf)
{
        int mfp;
        unsigned int val;

        char *s = buf;

        for (mfp = 0; mfp < MFP_PIN_MAX; ++mfp) {
                val = pxa3xx_mfp_get_lpm(mfp);
                s += sprintf(s, "%d 0x%x\n", mfp, val);
        }

        return (s - buf);
}

static ssize_t lpm_store(struct sys_device *dev, struct sysdev_attribute *attr, const char *buf, size_t count)
{
        int mfp;
        unsigned int val = 0x1f;

        sscanf(buf, "%d %x", &mfp, &val);

        if (val == 0x1f) {
                val = pxa3xx_mfp_get_lpm(mfp);
        } else {
                pxa3xx_mfp_set_lpm(mfp, val);
        }

        printk("gpio %d <lpm = 0x%x>\n", mfp, val);

        return count;
}

SYSDEV_ATTR(lpm, 0644, lpm_show, lpm_store);

static struct attribute *bpm_test_attr[] = {
	&attr_cpuload_mode.attr,
	&attr_ut_config.attr,
	&attr_ut.attr,
	&attr_mfp.attr,
	&attr_lpm.attr,
	&attr_idlectl.attr,
};

static int bpm_test_add(struct sys_device *sys_dev)
{
	int i, n, ret;
	n = ARRAY_SIZE(bpm_test_attr);
	for (i = 0; i < n; ++i) {
		ret = sysfs_create_file(&(sys_dev->kobj), bpm_test_attr[i]);
		if (ret)
			return ret;
	}
	return 0;
}

static int bpm_test_rm(struct sys_device *sys_dev)
{
	int i, n;
	n = ARRAY_SIZE(bpm_attr);
	for (i = 0; i < n; i++) {
		sysfs_remove_file(&(sys_dev->kobj), bpm_test_attr[i]);
	}
	return 0;
}

static struct sysdev_driver bpm_test_driver = {
	.add = bpm_test_add,
	.remove = bpm_test_rm,
};
