/*
 * drivers/char/levante_charger.c
 *
 *
 * Copyright (2006) Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/uio.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>
#include <linux/workqueue.h>
#include <linux/power_supply.h>


#include <mach/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mman.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/bitfield.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/tavor_common.h>


DEFINE_MUTEX(charger_input_lock);

/* low thresholds are not working so w/a it*/
/*#define LOW_TS_WORKAROUND*/
/*when working with charger dis pen-det */
/*#define DISABLE_PEN_DET_WORKAROUND*/

enum charge_state_t{
	NO_CHARGE,
	PRE_CHARGE,
	FAST_CHARGE
};

enum{
	UNKNOWN_CHARGER = 0,
	USB_CHARGER,
	WALL_CHARGER,
	NO_CHARGER,
};

enum tbat_state_t {
	CHARGE_NOT_ALLOW,
	CHARGE_ALLOW,
	PHONE_OFF
};

enum charge_enable_state_t {
	CHARGER_DISABLE,
	BATTERY_NO_CHARGER,
	CHARGER_AND_BATTERY,
};

static int	plug_dev = UNKNOWN_CHARGER;
/*set only in ChargingStateMachine*/
static enum charge_state_t ChargeState = NO_CHARGE;
static struct	work_struct	charger_readings_work;
static struct	work_struct	charger_machine_work;
static struct work_struct chg_set_current_work;
static struct timer_list	charger_timer1;
static struct timer_list	charger_timer2;
static u8 pmicId;
static u8 chargerOn = 1;


static void start_pre_charging(void);
static void start_charging(void);
static void stop_charging(void);

static int  charger_release(struct inode *inode, struct file *file);
static int  charger_open(struct inode *inode, struct file *file);
static void __attribute__((unused)) charger_timer_modify1(void);
static void charger_timer_modify2(u16 delay);

/* for development only!!! - for real phone chargerOn should be true always*/
/* for real phone SANREMO_INT_EN_BAT should be set*/
void charger_on(void)
{
  u8 val;
  chargerOn = 1;
  sanremo_read(SANREMO_INT_EN_1, &val);
  val |= (SANREMO_INT_EN_BAT);
  sanremo_write(SANREMO_INT_EN_1, val);
}
static u16 levanteAligenVbatMeasurmentValue(u8 measRegister)
{
	u16 mesurementValue;
	u8 regValue[2];

    if (sanremo_read_block(measRegister, regValue, 2) == 0)    // read success
	{
		// Read two registers, the alignment will be done as follows:
		// Register 1 - bits 7:0 => 8 MSB bits <11:4> of measurement value
		// Register 2 - bits 3:0 => 4 LSB bits <3:0> of measurement value
		mesurementValue = ((regValue[0] << 4) | (regValue[1] & 0x0F));
	}
	else
	{
		return 0;
	}
	return (mesurementValue);
}

static u16 levanteAligenIbatMeasurmentValue(u8 measRegister)
{
	#define IBAT_NUM_OF_MEAS 1
	u16 measTotal = 0;
	u8 regValue[2], val;
	u8 i;

	for(i=0;i<IBAT_NUM_OF_MEAS;i++)
	{
		if (sanremo_read_block(measRegister, regValue, 2) == 0)    // read success
		{
			// Read two registers, the alignment will be done as follows:
			sanremo_read(SANREMO_IBAT_MEAS1, &val);
			measTotal = (val << 8);
			sanremo_read(SANREMO_IBAT_MEAS2, &val);
			measTotal |= val;
		}
		else
		{
			return 0;
		}
	}

	return (measTotal/IBAT_NUM_OF_MEAS);
}

static void readVbat(u16 *vbat)
{
	u32 measmentRead;
	measmentRead = levanteAligenVbatMeasurmentValue (SANREMO_VBAT_MEAS1);
	*vbat = (u16)(((u32)measmentRead * 3 * 18 * 1000 )>> 12)/10 ; /* voltage in miliVolt*/
}

void readIbat(u16 *ibat, u8 *sign)
{
	u16 measmentRead;
	measmentRead = levanteAligenIbatMeasurmentValue (SANREMO_IBAT_MEAS1);
	/* negative number*/
	if (measmentRead & 0x8000) {
		*sign = 1;
		measmentRead = ~measmentRead;
		measmentRead &= 0x7FFF;/*turn off the sign bit*/
		measmentRead += 1;
	} else {
	 /* positve number */
	 *sign = 0;
	}
	*ibat = (measmentRead >> 2); /* ibat in mA -  * 0.25*/
}

static void readTbat(u16 *tbat)
{
	u32 measmentRead;
	measmentRead = levanteAligenVbatMeasurmentValue(SANREMO_GPADC1_MEAS1);
	*tbat = (u16)(((u32)measmentRead * 18 * 1000) >> 12)/10;
}

static void readTint(u16 *tint)
{
	u32 measmentRead, tint_mv;
	measmentRead = levanteAligenVbatMeasurmentValue(SANREMO_TINT_MEAS1);
	tint_mv = (u16)(((u32)measmentRead * 18 * 1000) >> 12)/10;
	*tint = (tint_mv - 884)*1000/3611;
}


/****************************************************************************
 *
 *Added by Michael for  supporting power supply interface @ 20091015
 *
 ****************************************************************************/
static int levante_power_get_property(struct power_supply *psy,
				  enum power_supply_property psp,
				  union power_supply_propval *val)
{
	int power_dev;
	val->intval = 0;

	if (psy->type == POWER_SUPPLY_TYPE_MAINS)
		power_dev = WALL_CHARGER;
	else if (psy->type == POWER_SUPPLY_TYPE_USB)
		power_dev = USB_CHARGER;

	switch (psp) {
	case POWER_SUPPLY_PROP_ONLINE:
		if ( power_dev == plug_dev ) 
			val->intval = 1;
		else
			val->intval = 0;
		break;
	case POWER_SUPPLY_PROP_STATUS:
		if ( power_dev == plug_dev && ChargeState != NO_CHARGE )
			val->intval = POWER_SUPPLY_STATUS_CHARGING;
		else
			val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static enum power_supply_property levante_power_props[] = {
	POWER_SUPPLY_PROP_ONLINE,
	POWER_SUPPLY_PROP_STATUS,
};

static char *levante_power_supplied_to[] = {
	"battery",
};

static struct power_supply levante_power_supplies[] = {
	{
		.name = "ac",
		.type = POWER_SUPPLY_TYPE_MAINS,
		.supplied_to = levante_power_supplied_to,
		.num_supplicants = ARRAY_SIZE(levante_power_supplied_to),
		.properties = levante_power_props,
		.num_properties = ARRAY_SIZE(levante_power_props),
		.get_property = levante_power_get_property,
	},
	{
		.name = "usb",
		.type = POWER_SUPPLY_TYPE_USB,
		.supplied_to = levante_power_supplied_to,
		.num_supplicants = ARRAY_SIZE(levante_power_supplied_to),
		.properties = levante_power_props,
		.num_properties = ARRAY_SIZE(levante_power_props),
		.get_property = levante_power_get_property,
	},
};

static void setVbatThresholds(u16 lower,u16 upper)
{
  u8 val;
  val = ((lower * 0xFF)/5400);
  sanremo_write(SANREMO_VBAT_LOW_TH, val);
  val = ((upper * 0xFF)/5400);
  sanremo_write(SANREMO_VBAT_UPP_TH, val);
}

static void setTbatThresholds(u16 lower, u16 upper)
{
  u8 val;
  /* the upper temprture is translte to lower value
   * and stored into the LOW_TH register*/
  val = ((upper * 0xFF)/1800);
  sanremo_write(SANREMO_GPADC1_LOW_TH, val);
  /* the lower temprture is translte to higher value
   * and stored into the UPP_TH register*/
  val = ((lower * 0xFF)/1800);
  sanremo_write(SANREMO_GPADC1_UPP_TH, val);
}


static enum charge_enable_state_t ChargeEnable(void)
{
	u8 val;

	printk(KERN_ALERT \
	"------->ChargeEnable \n");

	sanremo_read(SANREMO_STATUS, &val);

	/* Check for battery - currently battery detection is not working */

	if (chargerOn) {

		printk(KERN_ALERT \
		"------->ChargeEnable->chargerOn SANREMO_STATUS[%x] \n", val);
		if (levante_check_trim() == 57) {
			panic \
			("Chipset fault for charger only-repalce levante!!!");
		}

		if ((val & SANREMO_STS_BAT) && (val & SANREMO_STS_CHG)) {
			printk(KERN_ALERT \
			"------->ChargeEnable-detect battery and charger \n");
		  return CHARGER_AND_BATTERY;
		} else if (val & SANREMO_STS_BAT) {
			printk(KERN_ALERT \
			"------->ChargeEnable-detect battery but no charger\n");
		  return BATTERY_NO_CHARGER;
		} else {
			printk(KERN_ALERT \
			"------->ChargeEnable - no battery no charger \n");
		  return CHARGER_DISABLE;
		}
	}else
	   printk(KERN_ALERT \
	   "------->ChargeEnable- charger disable \n");
	return CHARGER_DISABLE;
}
static void ChargerTestPageWorkAroundForA1( void)
{
	if((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){

   	/* ---> enter test page*/
   		sanremo_write(0xFA, 0x00);
   		sanremo_write(0xFB, 0x00);
   		sanremo_write(0xFF ,0x00 );
		/*--no Bias current and Yn at the mux during pen detection--*/
   		sanremo_write(0xD3 ,0x56 );
   		sanremo_write(0xD4 ,0x0B );
   		sanremo_write(0xD5 ,0x01 );
   		sanremo_write(0xD6 ,0xEA );
   		sanremo_write(0xD7 ,0x01 );
		/*-----------------------*/
   		/*<--- exit test page*/
   		sanremo_write(0xFE, 0x00);
   		sanremo_write(0xFC, 0x00);
		printk(LVT_DEBUG_MUST "levante_charger test page W/A D3-D7 \n");
	}
}

static enum tbat_state_t TbatState(void)
{
	u16 tbatIn_mV;
	enum tbat_state_t tstate = CHARGE_NOT_ALLOW;

	readTbat(&tbatIn_mV);
	printk(KERN_ALERT "-------TbatState - tbatIn_mV = [%d]mV\n", tbatIn_mV);

	/***** The higher the mV the lower the temprature and vise versa *****/
	/*MIN temprature - */
	if (tbatIn_mV > TBAT_TEC_DPONIT_TH)/* {tbat < -17C}*/
	{
		setTbatThresholds(TBAT_TEC_MIN, TBAT_TEC_UPP_TH);
		tstate = PHONE_OFF;

	  printk(KERN_ALERT \
	  "----TbatState[MIN]{tbat<-17C}tbatIn_mV>%d tbatTS((%dmV)-(%dmV))\n", \
		TBAT_TEC_DPONIT_TH, TBAT_TEC_MIN, TBAT_TEC_UPP_TH);
	  /* notify user space - */
	  printk(KERN_ALERT \
	   "------->TbatState[MIN] very low temprature phone should be off\n");
	/*LOW temprature*/
	} else if
	(tbatIn_mV < TBAT_TEC_DPONIT_TH && tbatIn_mV > TBAT_C_DPONIT_TH) {
	/*LOW temprature { tbat > -17C && tbat < 2C } - charge not allow*/
	setTbatThresholds(TBAT_TEC_LOW_TH, TBAT_C_UPP_TH);
	tstate = CHARGE_NOT_ALLOW;

	printk(KERN_ALERT "------->{ tbat > -17C && tbat < 2C }\n");
	printk(KERN_ALERT \
	"--->TbatState[LOW] %dmV>tbatIn_mV>%dmV  tbatTS((%dmV)-(%dmV))\n", \
		TBAT_TEC_DPONIT_TH, TBAT_C_DPONIT_TH, \
		TBAT_TEC_LOW_TH, TBAT_C_UPP_TH);
	/* notify user space - low temprature charge not allow */
	printk(KERN_ALERT \
	"------->TbatState[LOW] low temprature charge not allow\n");
	/*NORMAL temprature*/
	} else if
	(tbatIn_mV < TBAT_C_DPONIT_TH && tbatIn_mV > TBAT_H_DPONIT_TH) {
	 /*NORMAL temprature { tbat > 2C && tbat < 42C }- charge allow*/

	setTbatThresholds(TBAT_C_LOW_TH, TBAT_H_UPP_TH);
	tstate = CHARGE_ALLOW;

	printk(KERN_ALERT "------->{ tbat > 2C && tbat < 42C }\n");
	printk(KERN_ALERT \
	"--->TbatState[NORMAL] %dmV>tbatIn_mV>%dmV tbatTS((%dmV)-(%dmV))\n",
		TBAT_C_DPONIT_TH, TBAT_H_DPONIT_TH, \
		TBAT_C_LOW_TH, TBAT_H_UPP_TH);
	/* notify user space - charge allow */
	printk(KERN_ALERT "------->TbatState[NORMAL] charge allow\n");

	} else if
	(tbatIn_mV < TBAT_H_DPONIT_TH && tbatIn_mV > TBAT_EH_DPONIT_TH) {
	/*HIGH temprature { tbat > 42C && tbat < 57C }- charge not allow*/

	setTbatThresholds(TBAT_H_LOW_TH, TBAT_EH_UPP_TH);
	tstate = CHARGE_NOT_ALLOW;

	printk(KERN_ALERT "------->{ tbat > 42C && tbat < 57C }\n");
	printk(KERN_ALERT \
	"------->TbatState[HIGH] %dmV > tbatIn_mV > %dmV ((%dmV)-(%dmV))\n",
		TBAT_H_DPONIT_TH, TBAT_EH_DPONIT_TH, \
		TBAT_H_LOW_TH, TBAT_EH_UPP_TH);
	/* notify user space - high temprature charge not allow */
	printk(KERN_ALERT \
	"------->TbatState[HIGH] high temprature charge not allow\n");
	} else if
	(tbatIn_mV < TBAT_EH_DPONIT_TH) {
	/*MAX -Extra High temprature { tbat > 57C }- phone off*/

	setTbatThresholds(TBAT_EH_LOW_TH, TBAT_EH_MAX);
	tstate = PHONE_OFF;

	printk(KERN_ALERT "------->{ tbat > 57C }\n");
	printk(KERN_ALERT \
	"------->TbatState[MAX] tbatIn_mV < %dmV tbatTS((%dmV)-(%dmV))\n", \
		TBAT_EH_DPONIT_TH, TBAT_EH_LOW_TH, TBAT_EH_MAX);
	/* notify user space - Extra High temprature - phone off */
	printk(KERN_ALERT \
	"---->TbatState[MAX] Extra High temprature - phone should be off\n");

	}
	return tstate;
}


static void ChargingStateMachine(unsigned long event)
{
	u16 vbat;
	enum charge_enable_state_t charger_status;
	enum tbat_state_t temprature_status;

	mutex_lock(&charger_input_lock);

	/* gather charger data*/
	charger_status 		= ChargeEnable();
	temprature_status 	= TbatState();
	readVbat(&vbat);

	printk(KERN_ALERT \
	"------->ChargingStateMachine-vbat[%d]mV,chargerSts[%d],tempSts[%d]\n" \
				, vbat, charger_status, temprature_status);

	if (event & (PMIC_EVENT_IOVER 	|
			PMIC_EVENT_TINT 	|
			PMIC_EVENT_VBATMON 	|
			PMIC_EVENT_VSYS 	|
			PMIC_EVENT_VCHG_TS 	|
			PMIC_EVENT_CAHRG_FAIL	|
			PMIC_EVENT_CHARG_COMP)) {

		stop_charging();
		ChargeState = NO_CHARGE;
		printk(KERN_ALERT \
		"CSM[all other event] stop charge\n");

	} else if ((charger_status == BATTERY_NO_CHARGER ||
			charger_status == CHARGER_AND_BATTERY) &&
				 (temprature_status == CHARGE_ALLOW)) {
		/*EMPTY - pre-charge*/
		/* if battery < 3050mv then pre-charge and
		wait till we reach 3200mV*/
		if (vbat < VBAT_TS_DISCHARGE_DPOINT) {
			if (ChargeState != NO_CHARGE) {
				/*if we are charging
				then stop it before switching.*/
				stop_charging();
				printk(KERN_ALERT \
				"ChargingStateMachine[PRECH] stop charge\n");
			}

			setVbatThresholds(VBAT_TS_DISCHARGE_LOW,
						VBAT_TS_DISCHARGE_UPP);
			printk(KERN_ALERT \
			"CSM[PRECH]vbat<%dmV vbatTS(%d-%d) \n", \
				VBAT_TS_DISCHARGE_DPOINT, \
				VBAT_TS_DISCHARGE_LOW,    \
				VBAT_TS_DISCHARGE_UPP);

			if (charger_status == CHARGER_AND_BATTERY) {
				start_pre_charging();
				ChargeState = PRE_CHARGE;
				/* notify user space - app should block
				all phone functionalties since we are
				low on battery */
				printk(KERN_ALERT \
				"CSM[PRECH]app should block all phone func.\n");

		  } else
			printk(KERN_ALERT \
			"CSM[PRECH] not start_pre_charging\n");

		} else if
			(vbat > VBAT_TS_DISCHARGE_DPOINT &&
			vbat < VBAT_TS_LOWBAT_DPOINT) {
			/*LOW - fast-charge*/

			/*if we came from pre-charge stop it before switching.*/
			if (ChargeState == PRE_CHARGE) {
				stop_charging();
				printk(KERN_ALERT \
				"ChargingStateMachine[LOW] stop charge\n");
			}

		setVbatThresholds(VBAT_TS_LOWBAT_LOW, VBAT_TS_LOWBAT_UPP);

		printk(KERN_ALERT \
		"CSM[LOW]fast charge vbat>%dmV && vbat<%dmV vbatTS(%d-%d)\n", \
		VBAT_TS_DISCHARGE_DPOINT, VBAT_TS_LOWBAT_DPOINT, \
		VBAT_TS_LOWBAT_UPP, VBAT_TS_LOWBAT_LOW);

		/*start fast charge only if we are not in it already.*/
		if (ChargeState != FAST_CHARGE) {
			if (charger_status == CHARGER_AND_BATTERY) {
				start_charging();
				ChargeState = FAST_CHARGE;
				/* notify user space - app should block all
				 * phone functionalties since we are still
				 *  low on battery */
				printk(KERN_ALERT \
				"CSM[LOW] app should block PS func.\n");
				} else
					printk(KERN_ALERT \
					"CSM[LOW] not start charging\n");
			}
		} else if
		(vbat > VBAT_TS_LOWBAT_DPOINT && vbat < VBAT_TS_NORMAL_DPOINT) {
		/*NORMAL - fast-charge*/
			/*couldn't be , but for safety:
			 * if we came from pre-charge stop
			 * it before switching.*/
			if (ChargeState == PRE_CHARGE) {
				stop_charging();
				printk(KERN_ALERT \
				"ChargingStateMachine[NORMAL] stop charge\n");
			}

		setVbatThresholds(VBAT_TS_NORMAL_LOW, VBAT_TS_NORMAL_UPP);

		printk(KERN_ALERT \
		"[NORMAL]fast charge vbat>%dmV && vbat<%dmV vbatTS(%d-%d)\n",\
			VBAT_TS_LOWBAT_DPOINT, VBAT_TS_NORMAL_DPOINT,\
			VBAT_TS_NORMAL_LOW, VBAT_TS_NORMAL_UPP);

		  /*start fast charge only if we are not in it already.*/
		  if (ChargeState != FAST_CHARGE)
		  {
			if (charger_status == CHARGER_AND_BATTERY) {
				start_charging();
				ChargeState = FAST_CHARGE;
				/* notify user space -
				 * app should display charger icon */
				printk(KERN_ALERT \
				"[NORMAL] app should display charger icon\n");
			} else
				printk(KERN_ALERT \
				"CSM[NORMAL] not start charging\n");
		  }
		} else if (vbat > VBAT_TS_NORMAL_DPOINT) {
			/*FULL - no-charge*/
			stop_charging();
			printk(KERN_ALERT \
			"------->ChargingStateMachine[FULL] stop charge\n");
			ChargeState = NO_CHARGE;
			setVbatThresholds(VBAT_TS_FULL_LOW, VBAT_TS_FULL_UPP);

			printk(KERN_ALERT \
			"CSM[FULL] set no charge vbat > %dmV vbatTS(%d-%d)\n", \
			VBAT_TS_NORMAL_DPOINT, VBAT_TS_FULL_LOW,
			VBAT_TS_FULL_UPP);
			/* notify user space -
			 * app should display full battery */
			printk(KERN_ALERT \
			"CSM[FULL] app should display full battery\n");
		}
	} else {
		/* charger or battery are not detected or
		 * charger is disable or
		 * temprature is not in range*/
		stop_charging();
		printk(KERN_ALERT \
		"ChargingStateMachine[charg or bat or temp not] stop charge\n");
		ChargeState = NO_CHARGE;

		if ((vbat < VBAT_TS_DISCHARGE_DPOINT) ||
			(vbat > VBAT_TS_DISCHARGE_DPOINT && /* discharge bat*/
			vbat < VBAT_TS_LOWBAT_DPOINT)) {/* low vbat state */
		printk(KERN_ALERT \
		"CSM - vbat = [%d]mV notify user and enter S3 \n", vbat);
		  /* notify user space */
		  /* enter S3 */
		} else if (vbat > VBAT_TS_LOWBAT_DPOINT &&
				vbat < VBAT_TS_NORMAL_DPOINT) {/*normal vbat*/
		printk(KERN_ALERT \
		"CSM[NORMAL]vbat=[%d]mV normal working voltage(COM can work)\n"\
		, vbat);
		/* if COM side is in reset mode, release it*/
		}
	}

	mutex_unlock(&charger_input_lock);

}

static void stop_charging(void)
{
	u8 val;

   	sanremo_read(SANREMO_CHG_CTRL1, &val);

	if ((val & 0x01 ) || (val & 0x02))
	{
		printk(LVT_DEBUG_MUST "------->levante_stop_charging \n");
		/* this is done auto by Levante -
		 *  we need it in case we want to stop manualy*/
		val &= ~0x03;
		sanremo_write(SANREMO_CHG_CTRL1, val);

		/*for Levante B0 and previous -
		* work in continues mode
		* to avoid false VBAT measurement (~2705mV)*/
		/* should be fixed in C0*/
		if  (pmicId > LEVATNE_B0_ID) {
			sanremo_read(SANREMO_MEAS_ENABLE3, &val);
			val &= ~(SANREMO_MEAS_EN3_COULOMB_COUNTER | \
					 SANREMO_MEAS_EN3_IBAT);
			sanremo_write(SANREMO_MEAS_ENABLE3, val);
		}
	#if defined(DISABLE_PEN_DET_WORKAROUND)
		if (((pmicId >= LEVATNE_A0_ID) &&
			(pmicId <=  LEVATNE_A1_ID)) ||
			(pmicId ==  LEVATNE_B0_ID)) {
			/*W/A for A0/A1/B0 since the MEAS_EN3_COULOMB_COUNTER
			is enabled should be fixed in B1*/
			/*before enabling it make sure it not harm charging -
			this is not needed for JIL demo.*/
			levante_enable_pen_down_irq(1);
		}
	#endif
		/*for real phone we can remove the traces
		when charger is not working
		need to un-comment */
		/*del_timer(&charger_timer1);*/
		/*del_timer(&charger_timer2);*/
	} else
	   printk(LVT_DEBUG_MUST \
		"--->levante_stop_charging(doing nothing already stopped)\n");

	return;
}

static void start_pre_charging(void)
{
  	u8 val;

  	printk(KERN_ALERT "------->levante_start_pre-charging \n");

	/* interrupt are working fine*/
	portofino_write(PORTOFINO_PREREG_REG1,0x0E); /*set pre-regulator to 1500mA & Vsys to 4.5v*/

	sanremo_write(SANREMO_CHG_CTRL1, 0x00); /* make sure we stop charge */
	sanremo_write(SANREMO_CHG_CTRL3, 0xF0); /* 0xF0 Charger Timeout = disable - there is mistake in doc - 0xF0 will disable timeout */
	sanremo_write(SANREMO_CHG_CTRL4, 0x75); /*Ipre = 20ma , Vpre term 3.2v*/

	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
	   	//	sanremo_write(SANREMO_CHG_CTRL6, /*0x0B*/ 0x0C); /*under voltage TS=3.4v | over V TS=4.4v| enable under | enable over | Debounce |  */
		sanremo_write(SANREMO_CHG_CTRL7, 0x40); /* disable detection of UV_SYS -if we don;t get over VBAT then we need to change here*/
	}
	if  ( pmicId >= LEVATNE_B0_ID ){
	/*
	 * 6.7	Charger Initialization
		Only valid for SR B0- Write to SanRemo register 0x4D, CHG_CTRL6 the value 0x01
		The field BD_MSK='01' (BAT TEMP is active, BAT_ID is masked)
		Only valid for SR B0- Write to SanRemo register 0x4E, CHG_CTRL7 the value 0x80
		The field ILIM_LONGTMREN='1' (ILIM_TMR_SET multiplied by factor of 128)
		The field IFSM_EN='1' (enables charge current thermal control )
		Total ILIM timer set is 512mSec (ILIM_TMR_SET='00' and ILIM_LONGTMREN='1')
 	 */
		/*sanremo_write(SANREMO_CHG_CTRL6, 0x01);*/
		sanremo_write(SANREMO_CHG_CTRL7, 0x80);
	}

	/*for Levante B0 and previous -
	 * work in continues mode
	 * to avoid false VBAT measurement (~2705mV)*/
	/* should be fixed in C0*/
	if  (pmicId > LEVATNE_B0_ID) {
		sanremo_read(SANREMO_MEAS_ENABLE3, &val);
		val |= (SANREMO_MEAS_EN3_COULOMB_COUNTER |
			SANREMO_MEAS_EN3_IBAT);
		sanremo_write(SANREMO_MEAS_ENABLE3, val);
	}

      #if defined(DISABLE_PEN_DET_WORKAROUND)
	if  ( ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID)) || (pmicId ==  LEVATNE_B0_ID)){
		/*W/A for A0/A1/B0 since the MEAS_EN3_COULOMB_COUNTER is enabled should be fixed in B1*/
		levante_enable_pen_down_irq(0);
		levante_tsi_disable_all_measurements();
	}
      #endif
	sanremo_write(SANREMO_CHG_CTRL1, 0x01); /* pre-charge */
	charger_timer_modify2(200);
}

static void start_charging(void)
{
	u8 val;

	printk(KERN_ALERT "------->levante_start_charging \n");

	/* interrupt are working fine*/
	portofino_write(PORTOFINO_PREREG_REG1,0x0E); /*set pre-regulator to 1500mA & Vsys to 4.5v*/

	sanremo_write(SANREMO_CHG_CTRL1, 0x00); /*0xEE IFSM | enable ILMIT | 4.5V | Iterm = 100ma| fast */
	sanremo_write(SANREMO_CHG_CTRL2, 0x0F/*0x09*/); /*TOFFMAX_SET=0|PREREG_OFF=0|LRSW_EN = 0|ICHG_SET= 500ma*/
	sanremo_write(SANREMO_CHG_CTRL3, 0xF0); /* 0x00 Charger Timeout = disable - there is mistake in doc - 0xF0 will disable timeout */
	sanremo_write(SANREMO_CHG_CTRL4, 0x40); /* 0x40 IBAT is monitored | TBAT is not monitored not connected ( 0xC0 eanble TBAT monitor)| VSYS according to state machine*/
	sanremo_write(SANREMO_CHG_CTRL5, 0x00); /*as defualt - relavnt to charge pump limiter*/
   	//sanremo_write(SANREMO_CHG_CTRL6, 0xCC); /* don't enable since it will do reset if un-plug -> plug the USB */
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
	   	//	sanremo_write(SANREMO_CHG_CTRL6, /*0x0B*/ 0x0C); /*under voltage TS=3.4v | over V TS=4.4v| enable under | enable over | Debounce |  */
		sanremo_write(SANREMO_CHG_CTRL7, 0x40); /* disable detection of UV_SYS -if we don;t get over VBAT then we need to change here*/
	}
	if  ( pmicId >= LEVATNE_B0_ID ){
	/*
	 * 6.7	Charger Initialization
		Only valid for SR B0- Write to SanRemo register 0x4D, CHG_CTRL6 the value 0x01
		The field BD_MSK='01' (BAT TEMP is active, BAT_ID is masked)
		Only valid for SR B0- Write to SanRemo register 0x4E, CHG_CTRL7 the value 0x80
		The field ILIM_LONGTMREN='1' (ILIM_TMR_SET multiplied by factor of 128)
		The field IFSM_EN='1' (enables charge current thermal control )
		Total ILIM timer set is 512mSec (ILIM_TMR_SET='00' and ILIM_LONGTMREN='1')
 	 */
		/*sanremo_write(SANREMO_CHG_CTRL6, 0x01);*/
		sanremo_write(SANREMO_CHG_CTRL7, 0x80);
	}

	/*for Levante B0 and previous -
	 * work in continues mode
	 * to avoid false VBAT measurement (~2705mV)*/
	/* should be fixed in C0*/
	if  (pmicId > LEVATNE_B0_ID) {
		sanremo_read(SANREMO_MEAS_ENABLE3, &val);
		val |= (SANREMO_MEAS_EN3_COULOMB_COUNTER |
			SANREMO_MEAS_EN3_IBAT);
		sanremo_write(SANREMO_MEAS_ENABLE3, val);
	}

      #if defined(DISABLE_PEN_DET_WORKAROUND)
	if  ( ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID)) || (pmicId ==  LEVATNE_B0_ID)){
		/*W/A for A0/A1/B0 since the MEAS_EN3_COULOMB_COUNTER is enabled should be fixed in B1*/
		levante_enable_pen_down_irq(0);
		levante_tsi_disable_all_measurements();
	}
      #endif

	if (levante_check_trim() == 57)  /* in case we don't have trim part*/
		sanremo_write(SANREMO_CHG_CTRL1, 0x6E);
		/*4.05V|Iterm= 100mA|fast = b10 - charger on */
	else
		sanremo_write(SANREMO_CHG_CTRL1, 0x92);
		/* 4.2V | Iterm = 20mA| fast = b10 - charger on */
	charger_timer_modify2(200);

	return;
}


static unsigned int usb_alloc_current = 100;
void charger_set_current(struct work_struct *pdata)
{
	pr_info("%s: usb enumeration success, current %d \n",__func__, usb_alloc_current);
}

// change USB charging current, irq safe
void usb_callback_set_current(unsigned icurrent)
{
	usb_alloc_current = icurrent;
	schedule_work(&chg_set_current_work);
	return;
}

EXPORT_SYMBOL(usb_callback_set_current);

/****************************************************************************
 * Interrupt Handler
 ***************************************************************************/
/* Currently we don't enable the timeout for charging. So Just Event A is
 * enough.
 */
void levante_charger_interrupt(unsigned long event)
{

	if (event & PMIC_EVENT_CHDET) {
		/* Charge detection */
		printk(KERN_ALERT \
		"------->levante_charger::PMIC_EVENT_CHDET \n");
	} else if (event & PMIC_EVENT_VCHG_TS) {
		/* VCHG over or lower voltage */
		/* TODO: Notify user space appliction*/
		printk(KERN_ALERT \
		"-->levante_charger::PMIC_EVENT_VCHG_TS-over or lower VCHG \n");
	} else if (event & PMIC_EVENT_VBAT_TS) {
		/* VBAT over or lower voltage */
		/* TODO: Notify user space appliction*/
		printk(KERN_ALERT \
		"levante_charger::PMIC_EVENT_VBAT_TS - over or under VBAT \n");
	} else if (event & PMIC_EVENT_TBAT) {
		/* if Battery over or under temperature*/
		/* TODO: Notify user space appliction*/
		printk(LVT_DEBUG_MUST \
		"---->levante_charger::PMIC_EVENT_TBAT-over or under TBAT \n");
	} else if (event & PMIC_EVENT_IOVER) {
		/* if over or under IOVER*/
		/* TODO: Notify user space appliction*/
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_IOVER \n");
	} else if (event & PMIC_EVENT_CAHRG_FAIL) {
		/* TODO: Notify user space appliction: Charger fail */
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_CAHRG_FAIL \n");
	} else if (event & PMIC_EVENT_TINT) {
		/* levante temprature is too high or
		charger fail from some reason */
		/* TODO: Notify user space appliction: Charger fail */
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_TINT \n");
	} else if (event & (PMIC_EVENT_CHARG_COMP)) {
		/* TODO: Notify user space appliction: Charger done */
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_CHARG_COMP \n");
	} else if (event & PMIC_EVENT_VBATMON) {
		/* Charge voltage too low */
		/* TODO: Notify user space appliction: Need charging */
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_VBATMON \n");
	} else if (event & PMIC_EVENT_VSYS) {
		/* vsys over or lower voltage */
		/* TODO: Notify user space appliction: Need charging */
		/* same case as vbat over /lower ? */
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::PMIC_EVENT_VSYS \n");
	} else {
		/*this is a WA for working with SAAR/EVB3
		no need for real phone*/
		if (event == PMIC_EVENT_BAT_DET) {
			u8 val;
			static u8 i;
			if (i > 0) {
				i = 0;
				sanremo_read(SANREMO_INT_EN_1, &val);
				val &= ~(SANREMO_INT_EN_BAT);
				sanremo_write(SANREMO_INT_EN_1, val);
			}
			i++;
		}
		printk(LVT_DEBUG_MUST \
		"------->levante_charger::UN HANDLE EVENT \n");
		/* NO SUCH EVENT IN LEVANTE*/
	}

	ChargingStateMachine(event);
}

static void __attribute__((unused)) charger_timer_modify1(void)
{

}
static void charger_timer_modify2(u16 delay)
{
	if (mod_timer(&charger_timer2, jiffies + delay)) {
		/* timer is active already */
		printk(LVT_DEBUG_MUST \
		"charger_timer_modify2:timer is active already \n");
	}

}
static void charger_timer_event1 ( unsigned long data )
{
  schedule_work(&charger_machine_work);
}

static void charger_timer_event2 ( unsigned long data )
{
  charger_timer_modify2(3000);
  schedule_work(&charger_readings_work);
}

static void levante_charger_machine_worker(struct work_struct *work)
{


}

static void levante_charger_readings_worker(struct work_struct *work)
{
	u16 ibat, vbat, tbat, tint;
	u8 sign;

   /* read the battery level and decide if need
    * charge also need to read temprature*/
	readVbat(&vbat);
	readIbat(&ibat, &sign);
	readTbat(&tbat);
	readTint(&tint);
	if (sign)
		printk(KERN_ALERT \
		 "------->levante_charger_worker ibat[%d]mA\n"\
		, ibat);
	else
		printk(KERN_ALERT \
		"------->levante_charger_worker ibat[-%d]mA\n"\
		, ibat);
	printk(KERN_ALERT
	"------->Vbat =[%d]mV tbat[%d]mV tint[%d]C,ChrgSts[%d]\n"\
		, vbat, tbat, tint, ChargeState);
	/* for saftey reasons , if we will not get the interrupt
	 * - remove if LOW TS interrupts are working*/
  #if defined(LOW_TS_WORKAROUND)
   	if (((ChargeState == NO_CHARGE) && (ChargeEnable())) &&
		(vbat < (VBAT_TS_LOWBAT_LOW-50) ||
		vbat < (VBAT_TS_NORMAL_LOW-50) ||
		vbat < (VBAT_TS_FULL_LOW-50))) {
			printk(LVT_DEBUG_MUST \
			"------->levante_charger_worker::LOW_TS_WORKAROUND\n");
			ChargingStateMachine(PMIC_EVENT_VBAT_TS);
   	}
   #endif
}

static void init_charger_timer(void)
{
	init_timer(&charger_timer1);
	init_timer(&charger_timer2);
	charger_timer1.function = charger_timer_event1;
	charger_timer2.function = charger_timer_event2;
	charger_timer1.data = (unsigned long)NULL;
	charger_timer2.data = (unsigned long)NULL;
}

/****************************************************************************
 * File System I/O operations
 ***************************************************************************/

static int charger_open(struct inode *inode, struct file *file)
{
	printk(LVT_DEBUG_MUST "------->levante_charger_open stub TBD \n");
	return 0;
}

static ssize_t charger_read(struct file *file, char __user *buf, size_t count,
			    loff_t *ppos)
{
	printk(KERN_ALERT "------->levante_charger_read stub TBD \n");
	return 0;
}

/* We assume that the user buf will not larger than kbuf size */
static ssize_t charger_write(struct file *file, const char __user *buf,
			      size_t count, loff_t *ppos)
{
	printk(KERN_ALERT "------->levante_charger_write stub TBD \n");
	return count;
}

static int charger_release(struct inode *inode, struct file *file)
{
	printk(KERN_ALERT "------->charger_release stub TBD\n");
	return 0;
}

static struct file_operations charger_fops = {
	.owner		= THIS_MODULE,
	.open		= charger_open,
	.release	= charger_release,
	.write		= charger_write,
	.read		= charger_read,
};

static struct miscdevice charger_miscdev = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= "levante_charger",
	.fops		= &charger_fops,
};

/****************************************************************************
 * Initialization / Registeration / Removal
 ***************************************************************************/
static int charger_probe(struct platform_device *pdev)
{
	u8 val;
	int ret;
	int i = 0;

	printk(LVT_DEBUG_MUST "------->levante_charger_probe - start\n");

	ret = misc_register(&charger_miscdev);
	if (ret < 0)
		return ret;

	/* Charger should handle EVENT_CHARGER and VBUS */
	ret = pmic_callback_register((PMIC_EVENT_CHARGER | PMIC_EVENT_VBUS)
						, levante_charger_interrupt);

	if (ret)  {  /*on fail de-register the device*/
		misc_deregister(&charger_miscdev);

		return ret;
	}

	INIT_WORK(&charger_readings_work, levante_charger_readings_worker);
	INIT_WORK(&charger_machine_work, levante_charger_machine_worker);
	INIT_WORK(&chg_set_current_work,charger_set_current);
	
	init_charger_timer();

	pxa3xx_pmic_set_pump(1); /* set_pump implemented in sanremo.c -
				* this is basicly for enabling the USB detect.*/
	/* Enable interrupt:
	 * the over current on charger reverse,
	 * the VBAT less then VBATMON,
	 * the Battery over temperature,
	 * the Charger detection/removal.
	 */

	pmicId = pmic_id();

	sanremo_read(SANREMO_INT_EN_1, &val);
	val |= (/*SANREMO_INT_EN_BAT |*/ SANREMO_INT_EN_CHG  \
		   /*| SANREMO_INT_EN_EXTON */);
	sanremo_write(SANREMO_INT_EN_1, val);

	sanremo_read(SANREMO_INT_EN_2, &val);
	val |= (SANREMO_INT_EN_TBAT | 	\
			SANREMO_INT_EN_TINT | 	\
			SANREMO_INT_EN_VSYS |	\
			SANREMO_INT_EN_VCHG | 	\
			SANREMO_INT_EN_VBAT);
	sanremo_write(SANREMO_INT_EN_2, val);

	sanremo_read(SANREMO_INT_EN_3, &val);
	val |= (SANREMO_INT_EN_CHG_IOVER | SANREMO_INT_EN_CHG_DONE | \
			 SANREMO_INT_EN_CHG_FAIL) ;
	sanremo_write(SANREMO_INT_EN_3, val);

	/* Enable all Auto measurments */
	/*-----------------------------*/
	sanremo_read(SANREMO_MEAS_ENABLE1, &val);
	val |= (SANREMO_MEAS_EN1_VSYS | SANREMO_MEAS_EN1_VCHG | \
			SANREMO_MEAS_EN1_VBAT | SANREMO_MEAS_EN1_TBAT);
	sanremo_write(SANREMO_MEAS_ENABLE1, val);

	sanremo_read(SANREMO_MEAS_ENABLE3, &val);
	val |= (SANREMO_MEAS_EN3_IBAT);
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		val |= (SANREMO_MEAS_EN3_IBAT_COMP );
	}else if(pmicId >= LEVATNE_B0_ID){
		val |= (SANREMO_MEAS_EN3_BAT_DET_EN );
	}
	sanremo_write(SANREMO_MEAS_ENABLE3, val);

	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		sanremo_read(SANREMO_MEAS_OFF_TIME1, &val);
		val |= (SANREMO_MEASOFFTIME1_BAT_DET_EN_A1);
    	sanremo_write(SANREMO_MEAS_OFF_TIME1, val);
	}

	sanremo_write(SANREMO_CHG_CTRL6, 0x03); /*to enable battery detection*/


	/*current src level (90=50uA/20=15uA)for battery temprature*/
	sanremo_write(SANREMO_GP_BAIS_REG2, 0x20);

	/*change setting if charger exist- set for Temprature mesurmants*/
	if  (pmicId >= LEVATNE_B0_ID) {
		sanremo_read(SANREMO_GPADC_MISC2, &val);
		val &= ~(SANREMO_GPADC1_GP_BIAS_A1);
		sanremo_write(SANREMO_GPADC_MISC2, val);
	}
	/* IBAT AVG for Ibat 2 comp. calc see readIbat*/
	sanremo_write(SANREMO_CC_READ, 0x20);

	/* for Levante B0 and previous -
	 * work in continues mode to avoid false VBAT measurement (~2705mV)*/
	if  (pmicId > LEVATNE_B0_ID) {
		sanremo_read(SANREMO_MEAS_ENABLE3, &val);
		val |= (SANREMO_MEAS_EN3_COULOMB_COUNTER | \
			SANREMO_MEAS_EN3_IBAT);
		sanremo_write(SANREMO_MEAS_ENABLE3, val);
	}

	ChargerTestPageWorkAroundForA1();
	/* register power supply */
	for (i = 0; i < ARRAY_SIZE(levante_power_supplies); i++) {
		ret = power_supply_register(&pdev->dev, &levante_power_supplies[i]);
		if (ret) {
			printk("failed to register %s power supply\n", levante_power_supplies[i].name);
			goto supply_failed;
		}
	}	
	/* activate charging state machine -
	this will set the TBAT & VBAT treshold*/
	ChargingStateMachine(PMIC_EVENT_INIT);

	printk(LVT_DEBUG_MUST "------->levante_charger_probe - done\n");

	return 0;

supply_failed:
	while (i--)
		power_supply_unregister(&levante_power_supplies[i]);   

	return ret;
}

static int charger_remove(struct platform_device *pdev)
{
	/* Disable TBAT, VCH, ICH and VBAT all auto measurements */
	u8 val;
	int ret, i = 0;

    printk(LVT_DEBUG_MUST "------->levante_charger_remove \n");

	/* Mask IRQ */
	sanremo_read(SANREMO_INT_EN_2, &val);
	val &= ~( SANREMO_INT_EN_TBAT | SANREMO_INT_EN_TINT | SANREMO_INT_EN_VSYS | SANREMO_INT_EN_VCHG | SANREMO_INT_EN_VBAT  ) ;
	sanremo_write(SANREMO_INT_EN_2, val);

	sanremo_read(SANREMO_INT_EN_3, &val);
	val &= ~( SANREMO_INT_EN_CHG_IOVER | SANREMO_INT_EN_CHG_DONE | SANREMO_INT_EN_CHG_FAIL) ;
	sanremo_write(SANREMO_INT_EN_3, val);

	sanremo_read(SANREMO_MEAS_ENABLE1, &val);
	val &= ~( SANREMO_MEAS_EN1_TBAT | SANREMO_MEAS_EN1_TINT | SANREMO_MEAS_EN1_VSYS | SANREMO_MEAS_EN1_VCHG | SANREMO_MEAS_EN1_VBAT);
	sanremo_write(SANREMO_MEAS_ENABLE1, val);

	sanremo_read(SANREMO_MEAS_ENABLE3, &val);
	val &= ~( SANREMO_MEAS_EN3_IBAT_COMP | SANREMO_MEAS_EN3_IBAT);
	sanremo_write(SANREMO_MEAS_ENABLE3, val);

	pxa3xx_pmic_set_pump(0);

	del_timer(&charger_timer1);
	del_timer(&charger_timer2);

	for (i = 0; i < ARRAY_SIZE(levante_power_supplies); i++) {
		power_supply_unregister(&levante_power_supplies[i]);
	}

	pmic_callback_unregister((PMIC_EVENT_CHARGER | PMIC_EVENT_VBUS), levante_charger_interrupt);
	misc_deregister(&charger_miscdev);
	return 0;
}

static struct platform_driver charger_driver = {
	.driver = {
		.name	= "levante-charger",
	},
	.probe		= charger_probe,
	.remove		= charger_remove,
};

static int __devinit charger_init(void)
{
	return platform_driver_register(&charger_driver);
}

static void __exit charger_exit(void)
{
	platform_driver_unregister(&charger_driver);
}

module_init(charger_init);
module_exit(charger_exit);

MODULE_AUTHOR("Yossef,Hanin <yhanin@marvell.com>");
MODULE_DESCRIPTION("Levante Charger Driver");
MODULE_LICENSE("GPL");

