/*
 * drivers/char/levante_hsdetect.c
 *
 * The headset detect driver based on levante
 *
 * Copyright (2008) Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/uio.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>

#include <mach/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mman.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/bitfield.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/tavor_common.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/levante_hsdetect.h>



static u8 pmicId = 0;
static struct device *hsdetect_dev;
static struct LEVANTE_HS_IOCTL hs_detect;

static int  levante_hsdetect_release(struct inode *inode, struct file *file);
static int  levante_hsdetect_open(struct inode *inode, struct file *file);

static void levante_hook_detect_enable_irq(void)
{
	u8 val;

	sanremo_read(SANREMO_INT_EN_3, &val);
	val |= (SANREMO_INT_EN_HEADSET | SANREMO_INT_EN_HOOK);
	sanremo_write(SANREMO_INT_EN_3, val);

	printk(LVT_DEBUG_MUST "------->levante_hook_detect_enable_irq \n");
}

static void levante_hook_detect_enable(u8 period,u8 enable)
{
	u8 val;

	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		sanremo_read(SANREMO_HEADSET_DETECTION, &val);
	}else if (pmicId >= LEVATNE_B0_ID){
		sanremo_read(SANREMO_HEADSET_DETECTION_B0, &val);
	}

	if(SANREMO_HEADSET_DET_ENA & enable){
		val &= period;
		val |= enable;
	}else{
		val &= period;
		val &= ~SANREMO_HEADSET_DET_ENA;
	}

		if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		  	sanremo_write(SANREMO_HEADSET_DETECTION, val);
		}else if (pmicId >= LEVATNE_B0_ID){
		  	sanremo_write(SANREMO_HEADSET_DETECTION_B0, val);
		}
	printk(LVT_DEBUG_MUST "------->levante_hook_detect_enable \n");
}

/****************************************************************************
 * Interrupt Handler
 ***************************************************************************/
void levante_hsdetect_interrupt(unsigned long event)
{
	u8 val;
	u8 mask;

	printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt event=0x%x \n",(unsigned int)event);

	/*this is workaround for levante A0 and A1 in B0 it should be off*/
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
	  //mask = SANREMO_STS_MICIN;
	  mask = SANREMO_STS_HEADSET;
	}else {//Levante B0
	  mask = SANREMO_STS_HEADSET;
  	}

	if ((event & PMIC_EVENT_HSDETECT))
	 {
		/* Headset detection */
   		if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
				sanremo_read(SANREMO_STATUS, &val);
		else if(pmicId >= LEVATNE_B0_ID)
				sanremo_read(SANREMO_STATUS_B0, &val);

		printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt SANREMO_STATUS =0x%x \n",val);

		if (val & mask){ //detect

				kobject_uevent(&hsdetect_dev->kobj, KOBJ_ADD);
				hs_detect.hsdetect_status = LEVANTE_HEADSET_ADD;
				levante_hook_detect_enable(SANREMO_HEADSET_PERIOD250MS,SANREMO_HEADSET_DET_ENA);//enable + 250ms
				printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt insert period =0x%x \n",val);

					sanremo_read(SANREMO_ADC_ANALOG_PROGRAM1, &val);
					val |= (SANREMO_ADC_ANA_PRO1_EN_MICBIAS_2_PAD_BITS );
					sanremo_write(SANREMO_ADC_ANALOG_PROGRAM1, val);


		 }else{//remove

				kobject_uevent(&hsdetect_dev->kobj, KOBJ_REMOVE);
				hs_detect.hsdetect_status = LEVANTE_HEADSET_REMOVE;
				levante_hook_detect_enable(SANREMO_HEADSET_PERIOD1000MS,SANREMO_HEADSET_DET_ENA);//enable + 1000ms
			   	printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt un-plug period =0x%x \n",val);

					sanremo_read(SANREMO_ADC_ANALOG_PROGRAM1, &val);
					val &= ~(SANREMO_ADC_ANA_PRO1_EN_MICBIAS_2_PAD_BITS );
					sanremo_write(SANREMO_ADC_ANALOG_PROGRAM1, val);

			}
    } else if ((event & PMIC_EVENT_HOOKSWITCH)){
		/* Hook switch */
		sanremo_read(SANREMO_STATUS, &val);
		if (val & SANREMO_STS_HOOK){//detect
			kobject_uevent(&hsdetect_dev->kobj, KOBJ_ONLINE);
			hs_detect.hookswitch_status = LEVANTE_HOOKSWITCH_PRESSED;
			printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt HOOK insert\n");

		}else{//remove
			kobject_uevent(&hsdetect_dev->kobj, KOBJ_OFFLINE);
			hs_detect.hookswitch_status = LEVANTE_HOOKSWITCH_RELEASED;
			printk(LVT_DEBUG_MUST "------->levante_hsdetect_interrupt HOOK un-plug\n");
		}
	}

	/*this is workaround for levante A0 and A1 in B0 it should be off*/
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		sanremo_read(SANREMO_INT_STATUS_3, &val);
		if (( val & SANREMO_INT_STS3_MICIN) || ( val & SANREMO_INT_STS3_HEADSET) || ( val & SANREMO_INT_STS3_HOOK)){
			val = SANREMO_INT_STS3_MICIN | SANREMO_INT_STS3_HEADSET | SANREMO_INT_STS3_HOOK;
			sanremo_write(SANREMO_INT_STATUS_3, val);
		}
 	}


}
/****************************************************************************
 * File System I/O operations
 ***************************************************************************/

static int levante_hsdetect_open(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t levante_hsdetect_read(struct file *file, char __user *buf, size_t count,
			    loff_t *ppos)
{
	return 0;
}

/* We assume that the user buf will not larger than kbuf size */
static ssize_t levante_hsdetect_write(struct file *file, const char __user *buf,
			      size_t count, loff_t *ppos)
{
	return count;
}

static int levante_hsdetect_release(struct inode *inode, struct file *file)
{
	return 0;
}

static int levante_hsdetect_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg)
{
	struct LEVANTE_HS_IOCTL hs_ioctl;
	if (copy_from_user(&hs_ioctl, (void *)arg, sizeof(struct LEVANTE_HS_IOCTL) ))
		return -EFAULT;

	switch(cmd)
	{
		case LEVANTE_HSDETECT_STATUS:
			hs_ioctl.hsdetect_status = hs_detect.hsdetect_status;
			hs_ioctl.hookswitch_status = hs_detect.hookswitch_status;
			printk(LVT_DEBUG_MUST "------->levante_hsdetect_ioctl LEVANTE_HSDETECT_STATUS \n");
			break;

		case LEVANTE_HOOKSWITCH_STATUS:
			hs_ioctl.hookswitch_status = hs_detect.hookswitch_status;
			hs_ioctl.hsdetect_status = hs_detect.hsdetect_status;
			printk(LVT_DEBUG_MUST "------->levante_hsdetect_ioctl LEVANTE_HOOKSWITCH_STATUS \n");
			break;

		default:
			 return -ENOTTY;

	}
	return copy_to_user((void *)arg,&hs_ioctl, sizeof(struct LEVANTE_HS_IOCTL));
}

static struct file_operations levante_hsdetect_fops = {
	.owner		= THIS_MODULE,
	.open		= levante_hsdetect_open,
	.release	= levante_hsdetect_release,
	.ioctl		= levante_hsdetect_ioctl,
	.write		= levante_hsdetect_write,
	.read		= levante_hsdetect_read,
};

static struct miscdevice levante_hsdetect_miscdev = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= "levante_hsdetect",
	.fops		= &levante_hsdetect_fops,
};

/****************************************************************************
 * Initialization / Registeration / Removal
 ***************************************************************************/
static int levante_hsdetect_probe(struct platform_device *pdev)
{
	u8 val;
	int ret;

	printk(LVT_DEBUG_MUST "------->levante_hsdetect_probe- start \n");

	ret = misc_register(&levante_hsdetect_miscdev);
	if (ret < 0)
		return ret;

	/* Charger should handle EVENT_CHARGER and VBUS */
	ret = pmic_callback_register((PMIC_EVENT_HSDETECT | PMIC_EVENT_HOOKSWITCH /*| PMIC_EVENT_MICIN*/),
			levante_hsdetect_interrupt);
	if (ret)  {
		printk(LVT_DEBUG_MUST "------->levante_hsdetect_prob - deregister - \n");
		misc_deregister(&levante_hsdetect_miscdev);
		return ret;
	}

	hsdetect_dev = &pdev->dev;

	hs_detect.hsdetect_status =0;
	hs_detect.hookswitch_status =0;

	printk(LVT_DEBUG_MUST "------->LEVANTE hsdetect initialized!\n");

	pmicId = pmic_id();

	/* enable headset and hook interrupts*/
	/* need to activate audio section in Levante*/
    /*this is workaround for levante A0 and A1 in B0 it should be off*/
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){

		sanremo_read(SANREMO_MIS_REG2, &val);
		val |= (SANREMO_MIS_REG2_AUDIO_PU | SANREMO_MIS_REG2_AUDIO_RSTB | SANREMO_MIS_REG2_PLL_PU);
		sanremo_write(SANREMO_MIS_REG2, val);

		sanremo_read(SANREMO_SHORT_REG, &val);
		val |= (SANREMO_SHORT_REG_HS1_SHORT_P_BIT |
				SANREMO_SHORT_REG_HS2_SHORT_P_BIT |
				SANREMO_SHORT_LO1_SHORT_P_BIT 	  |
			    SANREMO_SHORT_LO2_SHORT_P_BIT);
		sanremo_write(SANREMO_SHORT_REG, val);

		sanremo_read(SANREMO_AUDIO_SUPPLIES2, &val);
		val |= (SANREMO_REG27_MPU);
		sanremo_write(SANREMO_AUDIO_SUPPLIES2, val);

		sanremo_read(SANREMO_MIC_BUTTON_DETECTION, &val);
		val |= (SANREMO_EN_MIC_DET );
		val |= (SANREMO_BTN_DEBOUNCE );
		//val &= (~SANREMO_BTN_DEBOUNCE);
		sanremo_write(SANREMO_MIC_BUTTON_DETECTION, val);

		//Set Bias for hook switch and ramIn
		sanremo_read(SANREMO_ADC_ANALOG_PROGRAM1, &val);
		val |= (SANREMO_ADC_ANA_PRO1_EN_MICBIAS_1_2_BIT );
		sanremo_write(SANREMO_ADC_ANALOG_PROGRAM1, val);
		printk(LVT_DEBUG_MUST "------->levante_hook_detect_enable - open audio - A1 workaround \n");
	}else
		printk(LVT_DEBUG_MUST "------->levante_hook_detect_enable - open audio - A1 workaround *not* active \n");

	levante_hook_detect_enable_irq();
	levante_hook_detect_enable(SANREMO_HEADSET_PERIOD1000MS,SANREMO_HEADSET_DET_ENA);

	levante_hsdetect_interrupt(PMIC_EVENT_HSDETECT);
    levante_hsdetect_interrupt(PMIC_EVENT_HOOKSWITCH);


	printk(LVT_DEBUG_MUST "------->levante_hsdetect_probe - done\n");

	return 0;
}

static int levante_hsdetect_remove(struct platform_device *pdev)
{
	u8 val;


	/* Mask IRQ */
	sanremo_read(SANREMO_INT_EN_3, &val);
	val &= ~(SANREMO_INT_EN_HEADSET | SANREMO_INT_EN_HOOK);
	sanremo_write(SANREMO_INT_EN_3, val);

	pmic_callback_unregister(PMIC_EVENT_HSDETECT, levante_hsdetect_interrupt);
	misc_deregister(&levante_hsdetect_miscdev);
	return 0;
}

static struct platform_driver levante_hsdetect_driver = {
	.driver = {
		.name	= "levante-hsdetect",
	},
	.probe		= levante_hsdetect_probe,
	.remove		= levante_hsdetect_remove,
};

static int __devinit levante_hsdetect_init(void)
{
	return platform_driver_register(&levante_hsdetect_driver);
}

static void __exit levante_hsdetect_exit(void)
{
	platform_driver_unregister(&levante_hsdetect_driver);
}

module_init(levante_hsdetect_init);
module_exit(levante_hsdetect_exit);
MODULE_LICENSE("GPL");
