/*
 * Copyright (C) 2008 Borqs, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <asm/ioctl.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/gpio.h>
#include <mach/mfp.h>
#include <asm/mach-types.h>
#include <linux/input.h>
#include <linux/sensors.h>
#include <linux/time.h>
#include <linux/workqueue.h>
#include <linux/completion.h>
#include <mach/cy8c201a0.h>
#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

#define CEN_POS_MIN	0
#define CEN_POS_MID	25
#define CEN_POS_MAX     50

typedef struct __cy8c201a0_stat__ {
	struct list_head list;
	volatile unsigned long status;
#define CY8C201A0_DATA_READY 0
#define CY8C201A0_FIRST_POLL 1
} cy8c201a0_stat_t;

static struct capsense *cy8c201a0;
static struct input_dev *input_dev;
static struct semaphore sem_list;
static LIST_HEAD(process_list);
static wait_queue_head_t workq_head;
static struct delayed_work cap_wq;
static unsigned int centroid;
static struct proc_dir_entry *cy8c201a0_proc_file;

#if 0
static u8 chip_unlock[] = {
	0x79, 0x3C, 0xA5, 0x69
};

static u8 chip_lock[] = {
	0x79, 0x96, 0x5A, 0xC3
};
#endif

#ifdef CONFIG_GPIO_SENSORS_CY8C201A0
static u8 gpio_conf[] = {
	0x08, 0x00, 0x01
};
#else
static u8 gpio_conf[] = {
	0x08, 0x00, 0x00
};
#endif



static const u8 capsense_conf[] = {
	0x06, 0x0F, 0x08, 0x00, 0x01,
	0x00, 0x01, 0x00, 0x01, 0x1F,
	0x09, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00
};

static u8 capsense_conf_epr15[] = {
	0x06, 0x0F, 0x08, 0x00, 0x01,
	0x00, 0x01, 0x00, 0x01, 0x1F,
	0x09, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00
};

static u8 output_port_conf[] = {
	0x04, 0x00, 0x00
};

static u8 operation_select[] = {
	0x1C, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x80, 0x03, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
	0x00
};

static u8 noise_threshold[] = {
	0x4E, 0x32, 0x64, 0xC8, 0x02,
	0x0A, 0x03, 0x14, 0x50, 0x00
};

static u8 cs_slide_conf[] = {
	0x75, 0x01, 0x00, 0x32, 0x00
};

#if 0
static u8 scan_pos_benz[] = {
	0x57, 0x00, 0x01, 0x02, 0x03,
	0x04, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0x31, 0x31, 0xFF, 0xFF,
	0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x02, 0x02, 0x22, 0x22,
	0x22, 0x00, 0x00, 0x00, 0x00,
	0x00
};
#endif

static u8 scan_pos_benz_epr15[] = { 	
	0x57, 0x00, 0x01, 0x02, 0x03,
	0x04, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0x45, 0x45, 0xFF, 0xFF,
	0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x02, 0x02, 0x22, 0x22,
	0x22, 0x00, 0x00, 0x00, 0x00,
	0x00
};

static u8 scan_pos_spark[] = {
	0x57, 0x00, 0x01, 0x02, 0x03,
	0xFF, 0xFF, 0xFF, 0xFF, 0x04,
	0xFF, 0x54, 0x54, 0x64, 0x64,
	0x00, 0x00, 0x00, 0x00, 0x64,
	0x00, 0x05, 0x05, 0x0A, 0x0A,
	0x00, 0x00, 0x00, 0x00, 0x0A,
	0x00
};

static u8 sleep_conf[] = {
	0x7E, 0x00, 0x20, 0x00
};

static u8 sleep_conf_epr15[] = {
	0x7E, 0x81, 0x20, 0x00
};

static void __attribute__ ((unused)) cy8c201a0_wakeup(void)
{
	int ret;
	struct cy8c201a0_platform_data *pdata = NULL;

	if ((cy8c201a0) && (cy8c201a0->client)) {
		pdata = cy8c201a0->client->dev.platform_data;
	}
	if (!pdata) {
		return;
	}
	if ( pdata->pre_deep_sleep )
		pdata->pre_deep_sleep(0);
	/* exit sleep mode */
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, CS_SLEEP_CTRL, CS_EXIT_SLEEP_MODE);
	if (ret < 0) {
		printk("cy8c201a0: failed to exit sleep mode\n");
		printk("cy8c201a0: failed to wakeup\n");
		return;
	}
	/* chip reset */
	if (pdata->disable_chip) {
		pdata->disable_chip();
	}
	if (pdata->platform_init) {
		pdata->platform_init();
	}
	msleep(100);
	printk("capsense: enable chip, enable irq!\n");
	enable_irq(cy8c201a0->client->irq);
	msleep(20);
}

static void __attribute__ ((unused)) cy8c201a0_enter_deep_sleep(void)
{
	int ret;
	struct cy8c201a0_platform_data *pdata;

	if ((cy8c201a0) && (cy8c201a0->client)) {
		pdata = cy8c201a0->client->dev.platform_data;
		if ((pdata) && (pdata->pre_deep_sleep)) {
			/* mask chip irq */
			disable_irq(cy8c201a0->client->irq);
			/* stop scan */
			ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, CS_STOP_SCAN);
			if (ret < 0) {
				printk("cy8c201a0: failed to stop scan\n");
				printk("cy8c201a0: failed to enter deep sleep mode\n");
				return;
			}
			msleep(5);

			/*for benzina EPR-1.5*/
			{
				/* enter setup mode */

				ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, ENTER_SETUP_MODE);
				if (ret < 0) {
					printk("cy8c201a0: failed to enter setup mode\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				msleep(10);

				/* disable GP[1] */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, GPIO_ENABLE1, GPIO_DISABLE_PORT1);
				if (ret < 0) {
					printk("cy8c201a0: failed to disable gpio pin\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* enable GP[1] */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, GPIO_ENABLE1, GPIO_ENABLE_PORT1_PIN0);
				if (ret < 0) {
					printk("cy8c201a0: failed to enable gpio pin\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* pre-deep sleep */
				if (pdata->pre_deep_sleep) {
					pdata->pre_deep_sleep(0);
				}
				msleep(10);
				/* disable strong drive port-1 */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, DM_STRONG1, DISABLE_PORT1_STRONG_DRIVE);
				if (ret < 0) {
					printk("cy8c201a0: failed to disable strong drive port-1\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* set open drain low port-1 */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, OPEN_DRAIN1, PORT1_OPEN_DRAIN_LOW);
				if (ret < 0) {
					printk("cy8c201a0: failed to disable open drain low\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* set sleep control pin */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, CS_SLEEP_CTRL_PIN, CS_SLEEP_CTRL_PIN_VAL_SPARK);
				if (ret < 0) {
					printk("cy8c201a0: failed to set sleep control pin\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* set sleep control register */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, CS_SLEEP_CTRL, CS_SLEEP_CTRL_VAL);
				if (ret < 0) {
					printk("cy8c201a0: failed to set sleep control register\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* set awake counter */
				ret = i2c_smbus_write_byte_data(cy8c201a0->client, CS_AWAKE_COUNTER, AWAKE_COUNTER_VAL);
				if (ret < 0) {
					printk("cy8c201a0: failed to set awake counter\n");
					printk("cy8c201a0: failed to enter deep sleep mode\n");
					return;
				}
				/* post-deep sleep */
				if (pdata->pre_deep_sleep) {
					pdata->pre_deep_sleep(1);
				}
				msleep(10);

			}

			printk("cy8c201a0: enter into deep sleep mode ok\n");
		}
	}
}

static void cy8c201a0_chip_reset(void)
{
	struct cy8c201a0_platform_data *pdata;

	pdata = NULL;
	if ((cy8c201a0) && (cy8c201a0->client)) {
		pdata = cy8c201a0->client->dev.platform_data;
		if ((pdata) && (pdata->disable_chip) && (pdata->platform_init)) {
			pdata->disable_chip();
			msleep(10);
			pdata->platform_init();
			msleep(10);
		}
	}
}

static void slider_handler(struct work_struct *work)
{
	unsigned int val;
	unsigned char msb, lsb;
	struct cy8c201a0_platform_data *pdata = NULL;

	if (!(cy8c201a0->client) || !(cy8c201a0)) {
		return;
	}

	pdata = cy8c201a0->client->dev.platform_data;
	if (!(pdata->get_int_pin_value) || !(pdata)) {
		return;
	}

	if ( pdata->get_int_pin_value()) {
		printk("press up! centroid: %d\n", centroid);
		input_report_abs(input_dev, ABS_TILT_Y, centroid);
		input_report_abs(input_dev, ABS_VOLUME, 0);
		input_sync(input_dev);
		centroid = -1;
		return;
	}else{
		/**
		 * don't delete this! we wait for the chip is stable,
		 * then get centroid. is this interval ok? not sure...
		 */
		msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_POSM);
		lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_POSL);
		val = (msb << 8) | lsb;

		if ((val >= CEN_POS_MIN) && (val <= CEN_POS_MAX)) {
			if (machine_is_spark()) {
				val = CEN_POS_MAX - val;
			}
			if (val != centroid) {
				centroid = val;
				printk("press down! centroid: %d\n", centroid);
				input_report_abs(input_dev, ABS_TILT_Y, centroid);
				input_report_abs(input_dev, ABS_VOLUME, 1);
				input_sync(input_dev);
			}
		} else {
			printk("centroid out of range: %d\n", val);
		}
		schedule_delayed_work(&cap_wq, msecs_to_jiffies(80));
	}
	return;
}


static irqreturn_t cy8c201a0_isr(int irq, void *dev_id)
{
	printk(KERN_DEBUG"capsense triggered!\n");
	schedule_delayed_work(&cap_wq, 0);
	return IRQ_HANDLED;
}

static int detect_cy8c201a0(void)
{
	int ret;

	if (cy8c201a0) {
		ret = i2c_smbus_read_byte_data(cy8c201a0->client, DEVICE_ID);
		printk("capsense detected, chip id: 0x%02x\n", ret);
		if (ret < 0) {
			return -ENXIO;
		}
		return ret;
	}
	return -ENXIO;
}

static int __attribute__ ((unused)) cy8c201a0_is_stored(void)
{
	int idac[2];
	int finger_threshold[2];

	memset(finger_threshold, 0, 2);
	memset(idac, 0, 2);

	finger_threshold[0] = i2c_smbus_read_byte_data(cy8c201a0->client, CS_FINGER_TH_00);
	if (finger_threshold[0] < 0) {
		printk("capsense: read finger threshold failed! at %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	finger_threshold[1] = i2c_smbus_read_byte_data(cy8c201a0->client, CS_FINGER_TH_01);
	if (finger_threshold[1] < 0) {
		printk("capsense: read finger threshold failed! at %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	idac[0] = i2c_smbus_read_byte_data(cy8c201a0->client, CS_IDAC_00);
	if (idac[0] < 0) {
		printk("capsense: read finger threshold failed! at %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	idac[1] = i2c_smbus_read_byte_data(cy8c201a0->client, CS_IDAC_01);
	if (idac[1] < 0) {
		printk("capsense: read finger threshold failed! at %s:%d\n", __FILE__, __LINE__);
		return -1;
	}
	if ((machine_is_benzina()) && (
				(finger_threshold[0]== scan_pos_benz_epr15[11])&&
				(finger_threshold[1]== scan_pos_benz_epr15[12])&&
				(idac[0]== scan_pos_benz_epr15[21])&&
				(idac[1]== scan_pos_benz_epr15[22]))) {
		printk("benzina: register configure is already stored to nvram!\n");
		return 0;
	} else if ((machine_is_spark()) && (
				(finger_threshold[0]== scan_pos_spark[11])&&
				(finger_threshold[1]== scan_pos_spark[12])&&
				(idac[0]== scan_pos_spark[21])&&
				(idac[1]== scan_pos_spark[22]))) {
		printk("spark: register configure is already stored to nvram!\n");
		return 0;
	} else if ((machine_is_blaze()) && (
				(finger_threshold[0]== scan_pos_spark[11])&&
				(finger_threshold[1]== scan_pos_spark[12])&&
				(idac[0]== scan_pos_spark[21])&&
				(idac[1]== scan_pos_spark[22]))) {
		printk("blaze: register configure is already stored to nvram!\n");
		return 0;
	}

}

static int cy8c201a0_initchip(void)
{
	int ret;
	struct cy8c201a0_platform_data *pdata = NULL;

	if ((cy8c201a0) && (cy8c201a0->client)) {
		pdata = cy8c201a0->client->dev.platform_data;
	}

	if (!pdata) {
		return -EINVAL;
	}

#if 0
	ret = cy8c201a0_is_stored();
	if (ret < 0) {
		goto failed_initchip;
	} else if (ret == 0) {
		return 0;
	}
#endif

	/* first enter setup mode */
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, ENTER_SETUP_MODE);
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}

#if 0
	/* chip unlock */
	ret = i2c_master_send(cy8c201a0->client, chip_unlock, ARRAY_SIZE(chip_unlock));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* set chip address & DM */
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, I2C_ADDR_DM_REG, I2C_ADDR_DM_SET);
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* chip lock */
	ret = i2c_master_send(cy8c201a0->client, chip_lock, ARRAY_SIZE(chip_lock));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
#endif

	/* config gpio function: disable GP0, GP1 gpio function */
	ret = i2c_master_send(cy8c201a0->client, gpio_conf, ARRAY_SIZE(gpio_conf));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* configure capsense function, GP0:pin-0, pin-1 as capsense;
	 * GP1: pin-0 as gpio
	 */
	if (pdata->sleep_pin) {
		ret = i2c_master_send(cy8c201a0->client, capsense_conf_epr15, ARRAY_SIZE(capsense_conf_epr15));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	} else {
		ret = i2c_master_send(cy8c201a0->client, capsense_conf, ARRAY_SIZE(capsense_conf));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	}
	/* config output port */
	ret = i2c_master_send(cy8c201a0->client,output_port_conf, ARRAY_SIZE(output_port_conf));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* config capsense operations */
	ret = i2c_master_send(cy8c201a0->client,operation_select, ARRAY_SIZE(operation_select));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* config noise threshold */
	ret = i2c_master_send(cy8c201a0->client,noise_threshold, ARRAY_SIZE(noise_threshold));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* config capsense slide function */
	ret = i2c_master_send(cy8c201a0->client,cs_slide_conf, ARRAY_SIZE(cs_slide_conf));
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	/* because of hardware design, different products have different parameters,
	 * so do this, too ugly!
	 */
	if (machine_is_benzina()) {
		ret = i2c_master_send(cy8c201a0->client, scan_pos_benz_epr15, ARRAY_SIZE(scan_pos_benz_epr15));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	} else if (machine_is_spark() || machine_is_blaze() ) {
		ret = i2c_master_send(cy8c201a0->client, scan_pos_spark, ARRAY_SIZE(scan_pos_spark));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	}

	/* config chip sleep control pin, sleep mode and awake counter */
	if (pdata->sleep_pin) {
		ret = i2c_master_send(cy8c201a0->client, sleep_conf_epr15, ARRAY_SIZE(sleep_conf_epr15));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	} else {
		ret = i2c_master_send(cy8c201a0->client, sleep_conf, ARRAY_SIZE(sleep_conf));
		if (ret < 0) {
			printk("exit at %d line\n", __LINE__);
			goto failed_initchip;
		}
	}
	/* store chip configure to nvram, must delay 125ms at least */
	printk("delay 125ms to write parameters to nvram ...\n");
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, STORE_CONF_TO_NVRAM);
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	msleep(125);

	/* chip reset for not causing i2c error */
	cy8c201a0_chip_reset();

	/* reconfigure chip, if effect, must delay 5ms at least */
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, DEVICE_RECONFIGURE);
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	msleep(25);

	/* enter normal mode, this mode can reduce chip ACk timing on i2c bus */
	ret = i2c_smbus_write_byte_data(cy8c201a0->client, COMMAND_REGISTER, ENTER_NORMAL_MODE);
	if (ret < 0) {
		printk("exit at %d line\n", __LINE__);
		goto failed_initchip;
	}
	printk("cy8c210a0: initial chip ok\n");
	return 0;
failed_initchip:
	printk("cy8c210a0: initial chip failed\n");
	return ret;
}

#ifdef CONFIG_PROC_FS
static int cy8c201a0_seq_show(struct seq_file *file, void *p)
{
	int ret = 0;
	int msb = 0;
	int lsb = 0;

	if (!cy8c201a0) {
		return -EFAULT;
	}

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, DEVICE_STATUS);
	seq_printf(file, "device status: 0x%02x\n", ret);
	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_BLM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_BLL);
	seq_printf(file, "baseline: %d\n", (msb<<8)|lsb);

	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_DIFFM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_DIFFL);
	seq_printf(file, "diff: %d\n", (msb<<8)|lsb);

	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_RAWM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_RAWL);
	seq_printf(file, "raw: %d\n", (msb<<8)|lsb);

	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SLID_MULM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SLID_MULL);
	seq_printf(file, "resolution msb: %d\n", msb);
	seq_printf(file, "resolution lsb: %d\n", lsb);

	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_PEAKM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_PEAKL);
	seq_printf(file, "centroid peak: %d\n", (msb<<8)|lsb);

	msb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_POSM);
	lsb = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_CEN_POSL);
	seq_printf(file, "centroid position: %d\n", (msb<<8)|lsb);

	seq_printf(file, "--------------------\n");
	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_STATUS0);
	seq_printf(file, "port[0] read status: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_READ_STATUS1);
	seq_printf(file, "port[1] read status: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, INPUT_PORT0);
	seq_printf(file, "port[0] input status: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, INPUT_PORT1);
	seq_printf(file, "port[1] input status: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, STATUS_HOLD_MASK1);
	seq_printf(file, "port[1] hold mask: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_ENABLE0);
	seq_printf(file, "enable port[0]: 0x%02x\n", ret);
	ret = i2c_smbus_read_byte_data(cy8c201a0->client, GPIO_ENABLE1);
	seq_printf(file, "enable port[1]: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SCAN_POS_00);
	seq_printf(file, "scan position[0]: 0x%02x\n", ret);
	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SCAN_POS_01);
	seq_printf(file, "scan position[1]: 0x%02x\n", ret);

	seq_printf(file, "--------------------\n");
	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_BL_UPD_TH);
	seq_printf(file, "baseline update threshold: %d\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_DEBOUNCE);
	seq_printf(file, "debounce: %d\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SETL_TIME);
	seq_printf(file, "settling time: %d\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_OTH_SET);
	seq_printf(file, "cs oth select: 0x%02x\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_LOW_BL_RST);
	seq_printf(file, "low baseline reset: %d\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_HYSTERESIS);
	seq_printf(file, "hysteresis: %d\n", ret);

	ret = i2c_smbus_read_byte_data(cy8c201a0->client, CS_SLID_CONFIG_REG);
	seq_printf(file, "slider config: %d\n", ret);

	return 0;
}

static int cy8c201a0_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, &cy8c201a0_seq_show, NULL);
}

static struct file_operations cy8c201a0_seq_fops = {
	.owner = THIS_MODULE,
	.open = cy8c201a0_seq_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

static int create_cy8c201a0_proc_file(void)
{
	cy8c201a0_proc_file = create_proc_entry("driver/cy8c201a0", 0444, NULL);
	if (!cy8c201a0_proc_file) {
		printk("cy8c201a0: create proc file for cy8c201a0 failed!\n");
		return -ENOMEM;
	}
	cy8c201a0_proc_file->proc_fops = &cy8c201a0_seq_fops;
	return 0;
}

static void remove_cy8c201a0_proc_file(void)
{
	remove_proc_entry("driver/cy8c201a0", NULL);
}
#endif

#ifdef CONFIG_ANDROID_POWER
static void cy8c201a0_early_suspend(android_early_suspend_t *h)
{
	cy8c201a0_enter_deep_sleep();
}

static void cy8c201a0_late_resume(android_early_suspend_t *h)
{
	cy8c201a0_wakeup();
}

static android_early_suspend_t cy8c201a0_early_suspend_desc = {
	.level = 1,
	.suspend = cy8c201a0_early_suspend,
	.resume = cy8c201a0_late_resume,
};
#endif

static int __devexit cy8c201a0_remove(struct i2c_client *client)
{
	struct capsense *cy8c201a0;

	cy8c201a0 = i2c_get_clientdata(client);

	if (((client->irq) > 0) && (!cy8c201a0)) {
		free_irq(client->irq, cy8c201a0);
	}
	memset(cy8c201a0, 0, sizeof(struct capsense));
	kfree(cy8c201a0);
	i2c_set_clientdata(client, NULL);

#ifdef CONFIG_ANDROID_POWER
	android_unregister_early_suspend(&cy8c201a0_early_suspend_desc);
#endif

#ifdef	CONFIG_PROC_FS
	remove_cy8c201a0_proc_file();
#endif

	input_unregister_device(input_dev);
	input_free_device(input_dev);

	return 0;
}

static void cy8c201a0_shutdown(struct i2c_client *dev)
{
	struct cy8c201a0_platform_data *pdata;

	pdata = NULL;
	if (dev) {
		pdata = (struct cy8c201a0_platform_data *)(dev->dev.platform_data);
		if ((pdata) && (pdata->disable_chip)) {
			pdata->disable_chip();
			msleep(10);
		}
	}
}

static int cy8c201a0_probe(struct i2c_client *client, const struct i2c_device_id *devid)
{
	struct cy8c201a0_platform_data *pdata;
	int ret = -EINVAL;

	init_waitqueue_head(&workq_head);
	init_MUTEX(&sem_list);
	INIT_DELAYED_WORK(&cap_wq, slider_handler);

	pdata = (struct cy8c201a0_platform_data *)(client->dev.platform_data);
	if (!pdata) {
		return ret;
	}

	cy8c201a0 = kzalloc(sizeof(struct capsense), GFP_KERNEL);
	if (!cy8c201a0) {
		return -ENOMEM;
	}
	memset(cy8c201a0, 0, sizeof(struct capsense));
	cy8c201a0->client = client;
	dev_set_drvdata(&client->dev, cy8c201a0);

	pdata->platform_init();
	msleep(100);

	ret = detect_cy8c201a0();
	if (ret < 0) {
		printk("cy8c201a0 is undetected!\n");
		ret = -ENXIO;
		goto failed;
	}
#if 1
	ret = cy8c201a0_initchip();
	if (ret < 0) {
		printk("capsense: initialize cy8c201a0 failed!\n");
		ret = -ENXIO;
		goto failed;
	}
#endif
	ret = request_irq(client->irq, cy8c201a0_isr, IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, "capsense", cy8c201a0);
	if (ret) {
		dev_err(&client->dev, "cy8c201a0: can't allocate irq %d!\n", client->irq);
		goto failed;
	}

#ifdef CONFIG_PROC_FS
	ret = create_cy8c201a0_proc_file();
	if (ret) {
		printk(KERN_INFO "cy8c201: failed to create cy8c201a0 procfs entry!\n");
		goto failed_reg_procfs;
	}
#endif

	input_dev = input_allocate_device();
	if (!input_dev) {
		printk("cy8c201a0: failed to allocate input_dev!\n");
		ret = -ENOMEM;
		goto failed_alloc_input_dev;
	}
	input_dev->name = "capsensor-key";
	input_dev->id.bustype = BUS_HOST;
	set_bit(EV_ABS, input_dev->evbit);
	set_bit(ABS_VOLUME, input_dev->absbit);
	set_bit(ABS_TILT_Y, input_dev->absbit);
	ret = input_register_device(input_dev);
	if (ret < 0) {
		printk("cy8c201a0: unable to register input device!\n");
		ret = -EFAULT;
		goto failed_reg_input_driver;
	}
#ifdef CONFIG_ANDROID_POWER
	android_register_early_suspend(&cy8c201a0_early_suspend_desc);
#endif
	return 0;

failed_reg_input_driver:
	input_free_device(input_dev);
failed_alloc_input_dev:
	remove_cy8c201a0_proc_file();
failed_reg_procfs:
	free_irq(client->irq, NULL);
failed:
	memset(cy8c201a0, 0, sizeof(struct capsense));
	kfree(cy8c201a0);
	return ret;
}

static struct i2c_driver cy8c201a0_driver = {
	.driver = {
		.name = "cy8c201a0",
	},
	.probe = cy8c201a0_probe,
	.remove = cy8c201a0_remove,
	.shutdown = cy8c201a0_shutdown,
};

static int __init cy8c201a0_init(void)
{
	int ret;
	ret = i2c_add_driver(&cy8c201a0_driver);
	if (ret) {
		printk(KERN_WARNING "cy8c201a0: driver register failed!\n");
		return ret;
	}
	return ret;
}

static void __exit cy8c201a0_exit(void)
{
	i2c_del_driver(&cy8c201a0_driver);
}

module_init(cy8c201a0_init);
module_exit(cy8c201a0_exit);
MODULE_LICENSE("GPL");
