/*
 *  drivers/i2c/chips/cywee.c
 *
 * Copyright (C) 2009 Borqs, Inc.
 * 
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <linux/gpio.h>
#include <linux/input.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/sensors.h>

extern void gsensor_sensors(char *x,char *y, char *z);
extern int cywee_input_key(int keycode, int pressed);
static void mcu_to_key(char *mcu_data);
struct semaphore cywee_mutex;

struct gsensor_xyz {
	s16 x;
	s16 y;
	s16 z;
};

struct gsensor_xyz adxl_xyz = {
	.x = 0,
	.y = 0,
	.z = 0
};
#if defined(CONFIG_SENSORS_ADXL345)
extern int adxl345_get_xyz(struct adxl345_xyz *p);
//add yamaha-yas529 function to get compass data
#endif
#define CYWEE_INPUT_DEV
//#define CYWEE_KEYPAD_DEVICE
#define CYWEE_BRINGUP_TEST  0 //when bring up,open this macro.
#define CYWEE_DEBUG //add debug level for printk information
#define CYWEE_ID	0
#define MCU_KEY_STATUS 0 

static struct input_dev *cywee_input;

static int cywee_dev_open(struct inode *inode, struct file *file)
{
	//Null now.need add some actions here later.
	return 0;
}

static ssize_t cywee_dev_read(struct file *file, char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	return 0;
}

static unsigned int cywee_dev_poll(struct file *filep, poll_table * pt)
{
	return 0;
}

static int cywee_dev_release(struct inode *inode, struct file *filp)
{
	return 0;
}

#define MCU_LENTH  18
#define CYWEE_IOCTL_WRITE 0x1 //carry Akm's iocommnad operation
char rwbuf[MCU_LENTH];
static int  cywee_ioctl(struct inode *inode, struct file *file,
        unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int i = 0;

	if (down_interruptible(&cywee_mutex))
		return -ERESTARTSYS;

	memset(rwbuf, 0, sizeof(rwbuf));
	switch (cmd) {
		case CYWEE_IOCTL_WRITE:
			if (copy_from_user(rwbuf, argp, sizeof(rwbuf)))
				return -EFAULT;
			break;
		default:
			printk(KERN_INFO "Only support write command now!\n");
			break;
	}

#if CYWEE_BRINGUP_TEST 
	for(i = 0; i < sizeof(rwbuf); i++) {
		if(rwbuf[i])
			printk(KERN_INFO "MCU bytes[%d] = %d\n", i, (int)rwbuf[i]);
	}
#endif
	//need process these data now
	mcu_to_key(rwbuf);

	up(&cywee_mutex);

	return 0;
}

static struct file_operations cywee_dev_fops = {
	.owner = THIS_MODULE,
	.open = cywee_dev_open,
	.read = cywee_dev_read,
	.poll = cywee_dev_poll,
	.release = cywee_dev_release,
	.ioctl = cywee_ioctl,
};

struct sensors_dev cywee_dev = {
	.id = SENSORS_ORIENTATION,
	.name = "cywee",
	//.num_properties = ARRAY_SIZE(cywee_properties),
	//.properties = cywee_properties,
	//.get_property = cywee_get_property,
	//.put_property = cywee_put_property,
	.fops = &cywee_dev_fops
};

/*
 * from cywee engineer,status 1=left,2=right,4=up,8=down.
 * this has conflict with our phone.Reserve it as backup.Use
 * phone's keymap by default.
 */
static unsigned short cywee_keymap_backup[9] = {
  [1] = KEY_LEFT,
  [2] = KEY_RIGHT,
  [4] = KEY_DOWN,
  [8] = KEY_UP,
};

//add different GAME_MODE,then use different key_maps,as a feature?
int map_size = 5;
static unsigned short cywee_keymap[5] = {
  [0] = KEY_UP,
  [1] = KEY_DOWN,
  [2] = KEY_LEFT,
  [3] = KEY_RIGHT,
  [4] = KEY_REPLY,
};

enum{
	LENGTH,
	CMD,
	OPCODE,
};

/*
 * this function will read gsensor/compass data,
 * create the packet and send to cywee mcu.
 * data packet stype need further change.
 */
static int cywee_write(struct i2c_client *this_client) {

	unsigned char send_data[0x12], check_sum = 0;
	int i;
	s16 x,y,z;
	int ret;

	memset(send_data, 0, sizeof(send_data));
	send_data[LENGTH] = 0x12;
	send_data[CMD] = 0x1;
	send_data[OPCODE] = 0xf4;
	//check_sum = 0x12 + 0x1 + 0xf4;
	
	//action:data type here
#if 	defined(CONFIG_SENSORS_ADXL345)
	if(adxl345_get_xyz)
		adxl345_get_xyz(&adxl_xyz);
#endif
	send_data[3] = adxl_xyz.x;
	send_data[5] = adxl_xyz.y;
	send_data[7] = adxl_xyz.z;
#if 0
	gsensor_sensors(&x, &y, &z);
	send_data[3] = x;
	send_data[5] = y;
	send_data[7] = z;
#endif
#if CYWEE_BRINGUP_TEST 
	send_data[3] = x<<4;
	send_data[4] = x>>4;
	send_data[5] = y<<4;
	send_data[6] = y>>4;
	send_data[7] = z<<4;
	send_data[8] = z>>4;
#endif
	//check_sum += (x +y +z);
	//printk("gsensor:x:%d,y:%d,z%d\n",x,y,z);
	for (i =3; i < 7;i++) {
		send_data[2*i+3] = 0x08;
		send_data[2*i+4] = 0x00;
		//check_sum += 0x08;
	}
	for (i =0; i< 0x11;i++)
		check_sum += send_data[i];

	send_data[0x11] = check_sum;

	struct i2c_msg msgs[] = {
		{
		 .addr = this_client->addr,
		 .flags = 0,
		 .len = 0x12,
		 .buf = send_data,
		 },
	};

	if (i2c_transfer(this_client->adapter, msgs, 1) < 0) {
		printk(KERN_ERR "AKM8976 AKI2C_RxData: transfer error\n");
		ret =  -EIO;
	} else
		ret = 0;

	//printk(KERN_INFO "write successful,check_sum = %d\n",check_sum);

}

/*
 * 1. this function translate mcu data to key status 
 * up/down/left/right.then we need report key event 
 * to app. We can use keypad input-dev or cywee independent
 * input device by macro
 * ADD: we only process mcu_data[0] for key status now.
 * 2. reserve somebytes for future usage
 * 3. already hold the cywee_mutex,this will protect mcu-data coocurence
 */
int key_curr = 0;
static void mcu_to_key(char *mcu_data)
{
	char status;
	int key_code, index = 0;

	status = mcu_data[MCU_KEY_STATUS]; //only care byte-0.

	if(!status && !key_curr) //no key pressed
		return;

	while(index < map_size) {
		if(test_bit(index, &status)) {
			key_code = cywee_keymap[index];
			input_report_key(cywee_input, key_code, 1);
		} else if(test_bit(index, &key_curr)) {
			key_code = cywee_keymap[index];
			input_report_key(cywee_input, key_code, 0);
		}
		index++;
	//printk("status:%d,key-code=%d\n", status,key_code);
	}

#ifdef CYWEE_KEYPAD_DEVICE
	cywee_input_key(key_code, !!status);
#endif
	key_curr = status;
	//Need Process Other Bytes Gotten from User-space
}
static int cywee_read(struct i2c_client *this_client)
{
	char rx_cywee[15];
	int ret,i;
	memset(rx_cywee, 0, sizeof(rx_cywee));
	//printk(KERN_INFO "AKM8976A stands for cywee drivers\n");
	//put write/read msgs in the same array.
	struct i2c_msg msgs[] = {
		{
		 .addr = this_client->addr,
		 .flags = I2C_M_RD,
		 .len = 15,
		 .buf = rx_cywee,
		 },
	};

	if (i2c_transfer(this_client->adapter, msgs, 1) < 0) {
		printk(KERN_ERR "cywee read data: transfer error\n");
		ret =  -EIO;
	} else
		ret = 0;

#if CYWEE_BRINGUP_TEST 
	printk("0x%x  ", rx_cywee[0]);
	printk("0x%x  ", rx_cywee[1]);
	printk("%d  ", rx_cywee[3]<<8 | rx_cywee[2]);
	printk("%d  ", rx_cywee[5]<<8 | rx_cywee[4]);
	for(i = 6; i < 15; i++) {
		//printk("read values 0x%x  ", rx_cywee[i]);
		printk("0x%x  ", rx_cywee[i]);
	}
#endif	
	mcu_to_key(rx_cywee);

#if CYWEE_BRINGUP_TEST 
	printk("cywee read i2c values`\n");
	if(!this_client) {
		printk("client error\n");
		return -EIO;
	}
	for(i = 0; i < 10; i++) {
		ret = i2c_smbus_read_byte_data(this_client, CYWEE_ID+10);
		if (ret < 0)
			printk("cywee read fail\n");
		else
			printk("read values %d\n", ret);
	}
#endif
	return 0;
}

static struct i2c_client *cywee;
//thread will grab g-sensor and ecompass values and send them to MCU/Library,
//it will also read/write with Cywee for App games by sensor/input way.
static void cywee_thread(void)
{
	//Note:
	//1. add wakelock to prevent suspend
	//2. add a waitqueue for wakeup from somewhere else
	//3. add read/write fail solution,such as break to kill thread.
	
	while(1) {
		cywee_write(cywee);
		mdelay(100);
		cywee_read(cywee);
		//printk("\n");
		msleep(5000);
	}
	do_exit(0);
}

static int cywee_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	int i, err;

	if(!client) {
		printk("client NULL error\n");
		return -EIO;
	}
	printk("cywee address: 0x%x\n", client->addr);

	cywee = client;

	cywee_input = input_allocate_device(); //we'd better use cywee device as boss
	if (cywee_input == NULL) {
		err = -ENOMEM;
		printk("gpio_event_probe: Failed to allocate input device\n");
	 	goto err_allocate_fai;
	}
	cywee_input->name = "cywee-input-device";

	err = input_register_device(cywee_input);
	if (err) {
		printk("gpio_event_probe: Unable to register %s input device\n",
			cywee_input->name);
	 	goto err_register_fai;
	}

	set_bit(EV_KEY, cywee_input->evbit);

	set_bit(KEY_LEFT, cywee_input->keybit);
	set_bit(KEY_RIGHT, cywee_input->keybit);
	set_bit(KEY_UP, cywee_input->keybit);
	set_bit(KEY_DOWN, cywee_input->keybit);
	set_bit(KEY_REPLY, cywee_input->keybit);
	//add other key mapping here
	
	//kthread_run(cywee_thread, NULL, "cywee_sensor");

	/* Do some work for future integration with BORQS */
	err = sensors_dev_register(&cywee_dev);
	if (err) {
		printk(KERN_ERR "cywee register sensors core fail\n");
		goto err_sensor_register_fai;
	}

	init_MUTEX(&cywee_mutex);

	return 0;
err_sensor_register_fai:
	sensors_dev_unregister(&cywee_dev);
err_register_fai:
err_allocate_fai:
	input_free_device(cywee_input);
	return err;
}

//unregister input/sensor device
static int cywee_remove(struct i2c_client *client)
{
	sensors_dev_unregister(&cywee_dev);
	input_free_device(cywee_input);
	return 0;
}

static const struct i2c_device_id cywee_id[] = {
	{"cywee", 0 },
	{ }
};

//add suspend/resume field,maybe control cywee chip 
static struct i2c_driver cywee_driver = {
	.driver = {
		.name = "cywee",
		.owner = THIS_MODULE,
	 },
	.probe 	= cywee_probe,
	.remove = cywee_remove,
	.id_table = cywee_id,
	//.suspend = cywee_suspend,
	//.resume	= cywee_resume,
};


static int __init cywee_init(void)
{
	int ret;

	printk(KERN_WARNING "Cywee Kernel Driver: Init.\n");

	return i2c_add_driver(&cywee_driver);
}

static void __exit cywee_exit(void)
{
	sensors_dev_unregister(&cywee_dev);
	return i2c_del_driver(&cywee_driver);
}

module_init(cywee_init);
module_exit(cywee_exit);

