/* fx_compass.c - YAHAMA compass driver
 *
 * Copyright (C) 2008-2009 Foxconn Tech. Corporation.
 * Author: Ben Tsai <ben.ch.tsai@foxconn.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/irq.h>
#include <linux/miscdevice.h>
#include <linux/gpio.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/workqueue.h>
#include <linux/freezer.h>
#include <linux/kthread.h>
#include <linux/dmi.h>
#include <linux/random.h>
#include <linux/earlysuspend.h>
#include <linux/pm.h>

#define D1 printk
#define D2 printk

static atomic_t open_count = ATOMIC_INIT(0);

struct vreg  *power_vreg;
struct vreg  *compass_power_vreg;

#if defined(FEATURE_BCK_S0)
  #define GPIO_RESET 83
#else
  #define GPIO_RESET 84  //change
#endif

  #define GPIO_RESET 97  //change
static int compass_aot_open(struct inode *inode, struct file *file);
static int compass_aot_release(struct inode *inode, struct file *file);

static struct file_operations compass_aot_fops = {
    .owner = THIS_MODULE,
    .open  = compass_aot_open,
    .release = compass_aot_release,
};

static struct miscdevice compass_aot_device = {
    .minor = MISC_DYNAMIC_MINOR,
    .name = "fox_compass_aot",
    .fops = &compass_aot_fops,
};

#if 0
  #define D3(msg, ...) printk("[compass] %s : " msg "\n", __FUNCTION__, ## __VA_ARGS__)
#else
  #define D3(...) 
#endif

static struct i2c_client *this_client;

#ifdef CONFIG_HAS_EARLYSUSPEND
struct early_suspend early_suspend;
#endif

struct fox_compass {
    char    init;           /* has it been initialized ? */
    int     bus;            /* i2c bus */
    int     address;        /* i2c address */
    struct i2c_client       client;     /* i2c client */
    struct platform_device  *pdev;
};

static struct fox_compass fox_compass;

#define    ECOMPASS_PXA950

#ifndef    ECOMPASS_PXA950
static void compass_poweron(void){
  vreg_set_level(power_vreg, 2600);
  vreg_set_level(compass_power_vreg, 2850);
	vreg_enable(power_vreg);
	vreg_enable(compass_power_vreg);

  gpio_set_value(GPIO_RESET,0);
	mdelay(5);
	gpio_set_value(GPIO_RESET,1);
  mdelay(5);

  printk("%s()\n", __func__);
}

static void compass_poweroff(void){
	vreg_disable(power_vreg);
	vreg_disable(compass_power_vreg);
  printk("%s()\n", __func__);
}
#else
static void compass_poweron(void){ }
static void compass_poweroff(void){ }
#endif
static int compass_aot_open(struct inode *inode, struct file *file)
{
    D1(KERN_DEBUG "[GSR] compass_aot_open\n");
    if (atomic_cmpxchg(&open_count, 0, 1) == 0) {
        compass_poweron();
    }
    else{
        atomic_inc(&open_count);
    }

    return 0;
}

static int compass_aot_release(struct inode *inode, struct file *file)
{
    D1(KERN_DEBUG "[GSR] compass_aot_release\n");

    if (atomic_read(&open_count) == 0) {
        D1(KERN_WARNING "[GSR] Release more than open!\n");
    } else {
        atomic_dec(&open_count);

        if (atomic_read(&open_count) == 0) {
            compass_poweroff();
        } 
    }
    return 0;
}
#define MS3CDRV_CMD_MEASURE_ROUGHOFFSET 0x1
#define MS3CDRV_CMD_MEASURE_XY1Y2T      0x2

#define MS3CDRV_RDSEL_MEASURE     0xc0
#define MS3CDRV_RDSEL_CALREGISTER 0xc8

//verify hardware on Mg1 platform,follow Library operations
static void  hardware_verify(void)
{
	int ret;

	if(!this_client)
		return;

	//prepare write 0x2 --> 0xc0
	//
	ret = i2c_smbus_read_byte_data(this_client, 0xc0);
	printk("reg 0xc0: %d\n", ret);

	ret = i2c_smbus_read_byte_data(this_client, 0x2);
	printk("reg 0x2: %d\n", ret);
}

static int fox_compass_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    int err ,ret;

    printk("******************\n");
    printk("%s()\n", __func__);
    printk("******************\n");

    //compass_power_vreg = vreg_get(NULL, "gp6");
    //power_vreg = vreg_get(NULL, "gp1");
    
    if (fox_compass.init)
        return 0;

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        err = -ENODEV;
        goto exit_check_functionality_failed;
    }

    fox_compass.client = *client;
    fox_compass.client.addr = fox_compass.address;    
    strcpy(fox_compass.client.name, "Yamaha G-Sensor");

    i2c_set_clientdata(client, &fox_compass);
    this_client = client;

    ret = gpio_request(GPIO_RESET, "compass_reset_pin");
    printk("gpio_request %d\n",ret);
    if (ret)
    {
        printk("[fox_compass_probe] gpio_request failed on pin %d (ret=%d)\n", GPIO_RESET, ret);
        return ret;
    }

    ret = gpio_direction_output(GPIO_RESET, 1);
    printk("gpio_direction_output %d\n",ret);
    if (ret)
    {
        printk("[fox_compass_probe] gpio_direction_output failed on pin %d (ret=%d)\n",GPIO_RESET, ret);
        return ret;
    }

    err = misc_register(&compass_aot_device);
    if (0 > err) {
        D2(KERN_ERR "fox_compass_probe: compass_aot_device register failed\n");
        goto exit_misc_device_register_failed;
    }
    else
    {
        D2(KERN_DEBUG "fox_gsr_probe: compass_aot_device register succeed\n");
    }
    compass_poweron();

    //hardware_verify();

    return 0;

exit_misc_device_register_failed:
exit_check_functionality_failed:
    return err;
}

static int fox_compass_remove(struct i2c_client *client)
{
    struct fox_compass_data *fox_compass = i2c_get_clientdata(client);

    //i2c_detach_client(client);
    kfree(fox_compass);
    return 0;
}

static int fox_compass_suspend(struct i2c_client *client, pm_message_t mesg)
{
    D1(KERN_DEBUG "[compass_PWR] compass suspend done. \n");
    if (atomic_read(&open_count) > 0)
    {
        compass_poweroff();
    } 

    //call i2c suspend function here;

    return 0;
}

static int fox_compass_resume(struct i2c_client *client)
{
    D1(KERN_DEBUG "[compass_PWR] compass resume done. \n");
    if (atomic_read(&open_count) > 0)
    {
        compass_poweron();
    }

    //call i2c resume function here        

    return 0;
}

#ifdef CONFIG_HAS_EARLYSUSPEND
static void fox_compass_early_suspend(struct early_suspend *h)
{
    fox_compass_suspend(this_client, PMSG_SUSPEND);
}
static void fox_compass_early_resume(struct early_suspend *h)
{
    fox_compass_resume(this_client);
}
#endif

static const struct i2c_device_id fox_compass_id[] = {
	{ "fx_compass", 0 },
	{ }
};

static struct i2c_driver fox_compass_driver = {
    .driver = 
    {
        .name   = "fx_compass",
        .owner  = THIS_MODULE,
    },
    .probe      = fox_compass_probe,
    .remove     = fox_compass_remove,
    .id_table   = fox_compass_id,
#ifndef CONFIG_HAS_EARLYSUSPEND
    //.suspend    = fox_compass_suspend,
    //.resume     = fox_compass_resume,
#endif
};

static int __init fx_compass_init(void)
{
    printk("==================\n");
    printk("%s()\n", __func__);
    printk("==================\n");

    fox_compass.bus = 0;
    fox_compass.address = 0x2e;
    fox_compass.client.driver = &fox_compass_driver;

#ifdef CONFIG_HAS_EARLYSUSPEND
    early_suspend.level = EARLY_SUSPEND_LEVEL_DISABLE_FB;
    early_suspend.suspend = fox_compass_early_suspend;
    early_suspend.resume = fox_compass_early_resume;
    register_early_suspend(&early_suspend);
#endif

    return i2c_add_driver(&fox_compass_driver);
}

static void __exit fx_compass_exit(void)
{
    i2c_del_driver(&fox_compass_driver);
}

module_init(fx_compass_init);
module_exit(fx_compass_exit);

MODULE_AUTHOR("Ben Tsai <ben.ch.tsai@foxconn.com>");
MODULE_DESCRIPTION("fx_compass driver");
MODULE_LICENSE("GPL");
