/*
 * Monahans Micco PMIC Management Routines
 *
 *
 * Copyright (C) 2009, Borqs Ltd.
 * Copyright (C) 2006, Marvell Corporation(fengwei.yin@marvell.com).
 *
 * This software program is licensed subject to the GNU General Public License
 * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html
 */
//#define DEBUG

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/suspend.h>
#include <asm/ioctl.h>
#include <asm/irq.h>
#include <mach/gpio.h>
#include <mach/hardware.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa3xx_pm.h>
#include <mach/mfp.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/micco.h>
#include <mach/tavor_common.h>
#include <linux/wakelock.h>
#include <linux/earlysuspend.h>

/*
 * FIXME(EL): This macro should be improved!
 */
#define EXT0_WAKEUP_INTERRUPT	1	

#define DUMP_ALL_MICCO_REGS	1

#define MICCO_REG_NUM		(0xC0)

extern int get_pm_state(void);
static struct pxa3xx_pmic_regs micco_regs[MICCO_REG_NUM];
static int micco_cacheable_array[] __initdata = {
	MICCO_CHIP_ID,
	MICCO_IRQ_MASK_A,
	MICCO_IRQ_MASK_B,
	MICCO_IRQ_MASK_C,
	MICCO_IRQ_MASK_D,
	MICCO_SYSCTRL_A,
	MICCO_SYSCTRL_B,
	MICCO_FAULT_LOG,
	MICCO_OVER1,
	MICCO_APP_OVER2,
//	MICCO_APP_OVER3,
//	MICCO_LDO643,
	MICCO_LDO987,
	MICCO_LDO1110,
	MICCO_LDO12,
	MICCO_LDO13,
	MICCO_BUCK_SLEEP,
	MICCO_LDO_nSLEEP1,
	MICCO_LDO_nSLEEP2,
	MICCO_LDO_SLEEP_BIT1,
	MICCO_LDO_SLEEP_BIT2,
	MICCO_BUCK_LDO8_EN,
	MICCO_VCC2,
	MICCO_VCC1,
	MICCO_COMM_OVER2,
	MICCO_COMM_OVER3,
	MICCO_BUCK1_DVC1,
//	MICCO_BUCK1_DVC2,
	MICCO_APPS_AVRC,
	MICCO_COMM_CDTV1,
//	MICCO_COMM_CDTV2,
	MICCO_COMM_CVRC,
	MICCO_SRAM_DVC1,
//	MICCO_SRAM_DVC2,
	MICCO_SRAM_SVRC,
	MICCO_3DTV1,
//	MICCO_3DTV2,
	MICCO_3DVRC,
	MICCO_BULK_AUTO_SKIP,
	MICCO_LDO1_MDTV1,
//	MICCO_LDO1_MDTV2,
	MICCO_LDO1_MVRC,
	MICCO_LED1_CONTROL,
	MICCO_LED2_CONTROL,
	MICCO_LEDPC_CONTROL1,
	MICCO_LEDPC_CONTROL2,
	MICCO_LEDPC_CONTROL3,
	MICCO_LEDPC_CONTROL4,
	MICCO_LEDPC_CONTROL5,
	MICCO_WLED_CONTROL1,
	MICCO_WLED_CONTROL2,
	MICCO_FLASH_CONTROL1,
	MICCO_FLASH_CONTROL2,
	MICCO_VIBRA_CONTROL,
	MICCO_MISC,
	MICCO_CHARGE_CONTROL,
	MICCO_CCTR_CONTROL,
	MICCO_TCTR_CONTROL,
	MICCO_CHARGE_PULSE,
	MICCO_DAC_MSB,
	MICCO_DAC_LSB,
	MICCO_UART_ADDR,
//	MICCO_UART_CONTROL,
	MICCO_UART_TX_DATA,
//	MICCO_ADC_MAN_CONTROL,
	MICCO_ADC_AUTO_CONTROL_1,
	MICCO_ADC_AUTO_CONTROL_2,
	MICCO_TSI_CONTROL_1,
	MICCO_TSI_CONTROL_2,
	MICCO_VBATMON,
	MICCO_TBATHIGHP,
	MICCO_TBATHIGHN,
	MICCO_TBATLOW,
};
static struct power_supply_module *micco_power_module;

/* Make sure that Power I2C has been initialized before invoke this function */
extern int pxa_i2c_set_speed(int speed);
extern int pxa_i2c_get_speed(void);

/* Unique ID allocation */
struct micco_data {
	struct i2c_client 	*g_client;
	struct workqueue_struct *micco_wq;
};

static struct micco_data *micco_d;

static struct wake_lock  micco_suspend_lock = {
        .name = "micco"
};

static struct wake_lock  ext0_suspend_lock = {
	.name = "micco-ext0"
};

int micco_read_cache(u8 reg, u8 *pval)
{
	BUG_ON(reg >= MICCO_REG_NUM);

	*pval = micco_regs[reg].data;
	return 0;
}

int micco_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	wake_lock(&micco_suspend_lock);
	if (micco_d->g_client == NULL) {	/* No global client pointer? */
		wake_unlock(&micco_suspend_lock);
		return -1;
	}
	/* Cache read of the Micco register. Disabled temporary */
#if	1
	if (micco_regs[reg].inited && micco_regs[reg].cacheable) {
		*pval = micco_regs[reg].data;
#define MICCO_REGISTER_CACHE_CHECK
#ifndef MICCO_REGISTER_CACHE_CHECK
		wake_unlock(&micco_suspend_lock);
		return 0;
#endif
	}
#endif

	ret = i2c_smbus_read_byte_data(micco_d->g_client, reg);

	if( (ret <= 0) && (reg == MICCO_VBAT_RES) )  {
		// VBAT is used and read by APPS and by COMM processors
		// So there could be 2 consecutive read-accesses to VBAT.
		// In that case MICCO could return value ZERO.
		// VBAT = 0 may cause urgent shutdown!
		// To prevent this let's return previous "cache" value or valid 0x7F
		// Do not update the cache for this case. RETURN OK
		*pval = (micco_regs[reg].data == 0) ? 0x7F : micco_regs[reg].data;
		status = 0;
	}
	else	if (ret >= 0) {
#ifdef MICCO_REGISTER_CACHE_CHECK
		if (micco_regs[reg].inited && micco_regs[reg].cacheable) {
			if( *pval != ret )
				printk(KERN_ERR "Micco register 0x%x cache mismatch, (%x,%x)!\n",
					reg, ret,*pval);
		}
#endif
		*pval = ret;
		micco_regs[reg].inited = 1;
		micco_regs[reg].data = ret;
		status = 0;
	} else {
		printk("Micco read error!\n");
		status = -EIO;
	}

	wake_unlock(&micco_suspend_lock);
	return status;
}

int micco_write_block(u8 reg, u8 *pval, u8 len)
{
	int ret;
	int tmp = len;
	u8 *mul_pval;

	if (len < 1)
		return -EINVAL;

       	mul_pval = kmalloc(2 * len - 1, GFP_KERNEL);

	if (mul_pval == NULL)
		return -ENOMEM;

	while(--tmp > 0) {
		mul_pval[2 * tmp] = pval[tmp];
		mul_pval[2 * tmp -1] = reg + tmp;
	}
	mul_pval[0] = pval[0];

#if 0
	for(tmp =0; tmp < (2*len-1);tmp++)
		printk("mul_pval[%d] = %d\n", tmp, mul_pval[tmp]);
#endif
	wake_lock(&micco_suspend_lock);
	if (micco_d->g_client == NULL) {	/* No global client pointer? */
		wake_unlock(&micco_suspend_lock);
		return -1;
	}

	ret = i2c_smbus_write_i2c_block_data(micco_d->g_client, reg, 2*len -1, mul_pval);

	if (ret == 0) {
		tmp = reg;
		for (; reg < (len + tmp); reg ++) {
			if( micco_regs[reg].cacheable && !micco_regs[reg].inited)
				micco_regs[reg].inited = 1;
			micco_regs[reg].data = pval[reg - tmp];
		}
		reg = tmp;
	} else {
		printk("Micco block write error!\n");
	        ret = -EIO;
	}

	wake_unlock(&micco_suspend_lock);

	kfree(mul_pval);
	return ret;
}

int micco_read_block(u8 reg, u8 *pval, u8 len)
{
	int ret, tmp;

	wake_lock(&micco_suspend_lock);
	if (micco_d->g_client == NULL) {	/* No global client pointer? */
		wake_unlock(&micco_suspend_lock);
		return -1;
	}

	ret = i2c_smbus_read_i2c_block_data(micco_d->g_client, reg, len, pval);

	if (ret == len) {
		tmp = reg;
		for (; reg < (len + tmp); reg ++) {
#ifdef MICCO_REGISTER_CACHE_CHECK
			if (micco_regs[reg].inited && micco_regs[reg].cacheable) {
				if( pval[reg-tmp] != micco_regs[reg].data ) {
					//micco_regs[reg].data = pval[reg-tmp];
					printk(KERN_ERR "Micco register 0x%x cache mismatch, (%x,%x)!\n",
						reg, ret,*pval);
				}
			}
#endif
			micco_regs[reg].data = pval[reg-tmp];
			micco_regs[reg].inited = 1;
		}
		reg = tmp;
	} else {
		printk("Micco block read error!\n");
		ret = -EIO;
	}

	wake_unlock(&micco_suspend_lock);
	return ret;

}

int micco_write(u8 reg, u8 val)
{
	int ret;
	int status;

	wake_lock(&micco_suspend_lock);
	if (micco_d->g_client == NULL) {	/* No global client pointer? */
		wake_unlock(&micco_suspend_lock);
		return -1;
	}

	ret = i2c_smbus_write_byte_data(micco_d->g_client, reg, val);
	if (ret == 0) {
		if( micco_regs[reg].cacheable && !micco_regs[reg].inited)
			micco_regs[reg].inited = 1;
		micco_regs[reg].data = val;
		status = 0;
	} else {
		printk("Micco write error!\n");
		status = -EIO;
	}

	wake_unlock(&micco_suspend_lock);
	return status;
}

static int __init micco_initchip(void)
{
	int i;

	memset(&micco_regs, 0,
			(sizeof(struct pxa3xx_pmic_regs) * MICCO_REG_NUM));
	/* TODO: Mask all micco registers uncacheable now.
	 * We can do some optimization here later.
	 */
	for (i = 0; i < ARRAY_SIZE(micco_cacheable_array); i++)
	{
		micco_regs[micco_cacheable_array[i]].cacheable = 1;
	}

	return micco_write(MICCO_SYSCTRL_A, 0xE8);
}

int micco_enable_ldo_adc(int enable)
{
	int status;
	u8 val;
	static atomic_t ldo_adc_count=ATOMIC_INIT(0);

	status = micco_read(MICCO_ADC_MAN_CONTROL, &val);
	if (status)
		return -EIO;

	if( enable )
	{
		atomic_inc(&ldo_adc_count);
		val |= MICCO_ADC_MAN_LDO_ADC_ENABLE;
	}else
	{
		if( atomic_dec_and_test(&ldo_adc_count) )
			val &= ~MICCO_ADC_MAN_LDO_ADC_ENABLE;
		else
			return 0;
	}
	status = micco_write(MICCO_ADC_MAN_CONTROL, val);
	if (status)
		return -EIO;
	return 0;
}

/* Micco TSI functions */
int micco_enable_pen_down_irq(int enable)
{
	int ret;
	u8 val;

	if (enable) {
		/* enable pen down IRQ */
		ret = micco_read(MICCO_IRQ_MASK_C, &val);
		val &= ~0x10;
		ret = micco_write(MICCO_IRQ_MASK_C, val);
	} else {
		/* disable pen down IRQ */
		ret = micco_read(MICCO_IRQ_MASK_C, &val);
		if (!(val & IRQ_MASK_C_PEN_DOWN)) {
			val |= 0x10;
			ret = micco_write(MICCO_IRQ_MASK_C, val);
		}
	}
	return ret;
}

int micco_tsi_poweron(void)
{
	int status;
	u8 val;
	u8 tsi_val[2]={0x33, 0x00};

	status = micco_enable_ldo_adc(1);
	if (status)
		return -EIO;

	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (!(val & MICCO_ADC_AUTO_2_PENDET_EN)) {
		val |= MICCO_ADC_AUTO_2_PENDET_EN;
		status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);
	}

	/* TSI_DEABY: 3 slot. TSI_SKIP: 3 slot */
	status = micco_write_block(MICCO_TSI_CONTROL_1, tsi_val, 2);

	if (status)
		return -EIO;

	return 0;
}

int micco_tsi_poweroff(void)
{
	int status;
	u8 val;

	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;

	val &= ~(MICCO_ADC_AUTO_2_TSI_EN | MICCO_ADC_AUTO_2_PENDET_EN);
	status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);

	if (status)
		return -EIO;

	status = micco_enable_ldo_adc(0);
	if (status)
		return -EIO;

	return 0;
}

int micco_tsi_readxy(u16 *x, u16 *y, int pen_state)
{
	int status;
	u8 val;
	u16 mx, my, lxy;

	status = micco_read(MICCO_TSI_X_MSB, &val);
	if (status)
		return -EIO;
	mx = val;

	status = micco_read(MICCO_TSI_Y_MSB, &val);
	if (status)
		return -EIO;
	my = val;

	status = micco_read(MICCO_TSI_XY_MSB, &val);
	if (status)
		return -EIO;
	lxy = val;

	*x = ((mx << 2) & 0x3fc) + (lxy & 0x03);
	*y = ((my << 2) & 0x3fc) + ((lxy & 0x0c) >> 2);

	return 0;
}

int micco_tsi_enable_pen(int pen_en)
{
	int status;
	u8 val;

	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;

	if (pen_en)
		val |= MICCO_ADC_AUTO_2_PENDET_EN;
	else
		val &= ~MICCO_ADC_AUTO_2_PENDET_EN;

	status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);
	if (status)
		return -EIO;

	return 0;
}

int micco_tsi_enable_tsi(int tsi_en)
{
	int status;
	u8 val;

	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;

	if (tsi_en)
		val |= MICCO_ADC_AUTO_2_TSI_EN;
	else
		val &= ~MICCO_ADC_AUTO_2_TSI_EN;

	status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);
	if (status)
		return -EIO;

	return 0;
}
/* Micco TSI functions end */

/* Micco auto ADC functions */
int micco_enable_adc_irq(int adc, int enable)
{
	int ret;
	u8 val, mask;

	if (adc == PMIC_EVENT_ADC4)
		mask = IRQ_MASK_C_ADC_AUTO4;
	else if (adc == PMIC_EVENT_ADC5)
		mask = IRQ_MASK_C_ADC_AUTO5;
	else if (adc == PMIC_EVENT_ADC6)
		mask = IRQ_MASK_C_ADC_AUTO6;
	else
		return -EINVAL;

	if (enable) {
		/* enable pen down IRQ */
		ret = micco_read(MICCO_IRQ_MASK_C, &val);
		val &= ~mask;
		ret = micco_write(MICCO_IRQ_MASK_C, val);
	} else {
		/* disable pen down IRQ */
		ret = micco_read(MICCO_IRQ_MASK_C, &val);
		if (!(val & mask)) {
			val |= mask;
			ret = micco_write(MICCO_IRQ_MASK_C, val);
		}
	}
	return ret;
}

int micco_auto_adc_poweron(int adc)
{
	int status;
	u8 val, mask;

	if (adc == PMIC_EVENT_ADC4)
		mask = MICCO_ADC_AUTO_2_ADC4_EN;
	else if (adc == PMIC_EVENT_ADC5)
		mask = MICCO_ADC_AUTO_2_ADC5_EN;
	else if (adc == PMIC_EVENT_ADC6)
		mask = MICCO_ADC_AUTO_2_ADC6_EN;
	else
		return -EINVAL;

	status = micco_enable_ldo_adc(1);
	if( status )
		return -EIO;

	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;
	if (!(val & mask)) {
		val |= mask;
		status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);
		if (status)
			return -EIO;
	}

	return 0;
}

int micco_auto_adc_poweroff(int adc)
{
	int status;
	u8 val, mask;

	if (adc == PMIC_EVENT_ADC4)
		mask = MICCO_ADC_AUTO_2_ADC4_EN;
	else if (adc == PMIC_EVENT_ADC5)
		mask = MICCO_ADC_AUTO_2_ADC5_EN;
	else if (adc == PMIC_EVENT_ADC6)
		mask = MICCO_ADC_AUTO_2_ADC6_EN;
	else
		return -EINVAL;


	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;

	val &= ~(mask);
	status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);

	if (status)
		return -EIO;

	/*TODO: we know CP will operation MICCO_ADC_MAN_CONTROL */
	status = micco_enable_ldo_adc(0);
	if (status)
		return -EIO;
	
	return 0;
}

int micco_auto_adc_read(int adc, u8 *pval)
{
	int status;
	u8 reg;

	if (adc == PMIC_EVENT_ADC4)
		reg = MICCO_AUTO4_RES;
	else if (adc == PMIC_EVENT_ADC5)
		reg = MICCO_AUTO5_RES;
	else if (adc == PMIC_EVENT_ADC6)
		reg = MICCO_AUTO6_RES;
	else
		return -EINVAL;

	status = micco_read(reg, pval);
	if (status)
		return -EIO;

	return 0;
}

int micco_auto_adc_set_trigger(int adc, u8 high, u8 low)
{
	int status;
	u8 reghigh, reglow;

	if (adc == PMIC_EVENT_ADC4) {
		reghigh = MICCO_AUTO4HIGH;
		reglow = MICCO_AUTO4LOW;
	}
	else if (adc == PMIC_EVENT_ADC5) {
		reghigh = MICCO_AUTO5HIGH;
		reglow = MICCO_AUTO5LOW;
	}
	else if (adc == PMIC_EVENT_ADC6) {
		reghigh = MICCO_AUTO6HIGH;
		reglow = MICCO_AUTO6LOW;
	}
	else {
		return -EINVAL;
	}

	status = micco_write(reghigh, high);
	if (status)
		return -EIO;
	status = micco_write(reglow, low);
	if (status)
		return -EIO;

	return 0;
}

int micco_auto_adc_enable(int adc, int enable)
{
	int status;
	u8 val, mask;

	if (adc == PMIC_EVENT_ADC4)
		mask = MICCO_ADC_AUTO_2_ADC4_EN;
	else if (adc == PMIC_EVENT_ADC5)
		mask = MICCO_ADC_AUTO_2_ADC5_EN;
	else if (adc == PMIC_EVENT_ADC6)
		mask = MICCO_ADC_AUTO_2_ADC6_EN;
	else
		return -EINVAL;


	status = micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	if (status)
		return -EIO;

	if (enable)
		val |= mask;
	else
		val &= ~mask;

	status = micco_write(MICCO_ADC_AUTO_CONTROL_2, val);
	if (status)
		return -EIO;

	return 0;
}
/* Micco auto ADC functions end */

/* Micco manual measurement functions --------PaulKwong*/
int micco_read_ACD_IN4(void)
{
	u8 val;

	/*Do ACD_IN4 measurement (manual)*/

	micco_write(MICCO_ADC_MAN_CONTROL, 0x1C);
	msleep(2);

	micco_read(MICCO_ADC_MAN_CONTROL, &val);
	while (val & MICCO_ADC_MAN_CONT_CONV) {
		msleep(2);
		micco_read(MICCO_ADC_MAN_CONTROL, &val);
	}
	micco_read(MICCO_MAN_RES_MSB, &val);
	return (int)val;
}
EXPORT_SYMBOL(micco_read_ACD_IN4);

int micco_read_ACD_IN5(void)
{
	u8 val;

	/*Do ACD_IN5 measurement (manual)*/

	micco_write(MICCO_ADC_MAN_CONTROL, 0x1D);
	msleep(2);

	micco_read(MICCO_ADC_MAN_CONTROL, &val);
	while (val & MICCO_ADC_MAN_CONT_CONV) {
		msleep(2);
		micco_read(MICCO_ADC_MAN_CONTROL, &val);
	}
	micco_read(MICCO_MAN_RES_MSB, &val);
	return (int)val;
}
EXPORT_SYMBOL(micco_read_ACD_IN5);

int micco_read_ACD_IN6(void)
{
	u8 val;

	/*Do ACD_IN6 measurement (manual)*/

	micco_write(MICCO_ADC_MAN_CONTROL, 0x1E);
	msleep(2);

	micco_read(MICCO_ADC_MAN_CONTROL, &val);
	while (val & MICCO_ADC_MAN_CONT_CONV) {
		msleep(2);
		micco_read(MICCO_ADC_MAN_CONTROL, &val);
	}
	micco_read(MICCO_MAN_RES_MSB, &val);
	return (int)val;
}
EXPORT_SYMBOL(micco_read_ACD_IN6);
/* Micco manual measurement functions end*/


/* Micco Audio functions */
static volatile u8 micco_audio_regs[MICCO_AUDIO_REGS_NUM];

int micco_codec_read(u8 reg, u8 *val)
{
	return micco_read(MICCO_AUDIO_REG_BASE + reg, val);
}

int micco_codec_write(u8 reg, u8 val)
{
	return micco_write(MICCO_AUDIO_REG_BASE + reg, val);
}

void micco_read_codec(void)
{
	int i;
	u8 val;

	for (i = 0; i < MICCO_AUDIO_REGS_NUM; i++) {
		micco_codec_read(i, &val);
		micco_audio_regs[i] = val;
	}
}

void micco_dump_codec(void)
{
	int i;

	micco_read_codec();

	for (i = 0; i < MICCO_AUDIO_REGS_NUM; i++) {
		printk(KERN_ALERT "%s: Micco_audio_reg[%d] = 0x%x\n",
				__func__, i, micco_audio_regs[i]);
	}
}
EXPORT_SYMBOL(micco_dump_codec);

int micco_audio_init(void)
{
	int i;

	/* The default setting for Micco */
	micco_audio_regs[MICCO_MUX_MONO] = 0x00; /* MONO */
	micco_audio_regs[MICCO_MUX_BEAR] = 0x00; /* BEAR */
	micco_audio_regs[MICCO_MUX_LINE_OUT] = 0x00; /* LINE OUT */
	micco_audio_regs[MICCO_MUX_STEREO_CH1] =
		0x00;	/* STEREO_CH1 */
	micco_audio_regs[MICCO_MUX_STEREO_CH2] =
		0x00;	/* STEREO_CH2 */

	micco_audio_regs[MICCO_AUDIO_LINE_AMP] =
		MICCO_AUDIO_LINE_AMP_EN;	/* Enable the Line AMP */

	/* Gain for both channel controlled separately. Enable Setero */
	micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] =
		MICCO_STEREO_GAIN_SEPARATE | MICCO_STEREO_EN;

	/* Soft startup of the Stereo amplifiers */
	micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH2] = 0x0;

	micco_audio_regs[MICCO_HIFI_DAC_CONTROL] =
		MICCO_HIFI_DAC_ON;

	micco_audio_regs[MICCO_MONO_VOL] =
		MICCO_MONO_EN | 0xa;
	micco_audio_regs[MICCO_BEAR_VOL] =
		MICCO_BEAR_EN | 0xa;

	/* Micco as I2S slave. Use I2S MSB normal mode */
	micco_audio_regs[MICCO_I2S_CONTROL] =
		MICCO_I2S_MSB_JU_MODE;

	micco_audio_regs[MICCO_TX_PGA] =
		0x0c;	/* 0 dB */
	micco_audio_regs[MICCO_MIC_PGA] =
		MICCO_MIC_PGA_EXT_EN | MICCO_MIC_PGA_INT_EN |
		MICCO_MIC_PGA_SELMIC_2 | MICCO_MIC_PGA_AMP_EN |
		0x7;	/* 30 dB*/

	micco_audio_regs[MICCO_TX_PGA_MUX] =
		0xFF;

	micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] =
		MICCO_VCODEC_ADC_ON_EN | 0x08;
	/* PCM_SDI normal operation, PCM_SDO enabled */
	micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] =
		MICCO_VDAC_ON | MICCO_VDAC_HPF_BYPASS;

	micco_audio_regs[MICCO_SIDETONE] =
		MICCO_SIDETONE_EN | MICCO_SIDETONE_GAIN_STEREO | 0x08;

	/* Enable AUX1,2. AUX1, 2 gain 0dB */
	micco_audio_regs[MICCO_PGA_AUX1_2] =
		MICCO_PGA_AUX1_EN | MICCO_PGA_AUX2_EN;

	/* Enable AUX3. AUX3 gain 0 dB */
	micco_audio_regs[MICCO_PGA_AUX3] =
		MICCO_PGA_AUX3_EN;

	/* DAC1, 2, 3 gain 0dB */
	micco_audio_regs[MICCO_PGA_DACS] =
		0x00;

	/*Soft start for MONO, BEAR LINE and STEREO is 61.5ms */
	micco_audio_regs[MICCO_SOFT_START_RAMP] =
		0x00;

	for (i = 0; i < MICCO_AUDIO_REGS_NUM; i++)
		micco_codec_write(i, micco_audio_regs[i]);

	return 0;
}
EXPORT_SYMBOL(micco_audio_init);

/* FIXME: The Stereo have left and right channel. Need add it later */
int micco_codec_enable_output(int type)
{
	switch (type) {
		case CODEC_BEAR:
			if (!(micco_audio_regs[MICCO_BEAR_VOL] & MICCO_BEAR_EN)) {
				micco_audio_regs[MICCO_BEAR_VOL] |=  MICCO_BEAR_EN;
				micco_codec_write(MICCO_BEAR_VOL,
						micco_audio_regs[MICCO_BEAR_VOL]);
			}
			break;

		case CODEC_MONO:
			if (!(micco_audio_regs[MICCO_MONO_VOL] & MICCO_MONO_EN)) {
				micco_audio_regs[MICCO_MONO_VOL] |= MICCO_MONO_EN;
				micco_codec_write(MICCO_MONO_VOL,
						micco_audio_regs[MICCO_MONO_VOL]);
			}
			break;

		case CODEC_STEREO:
			if (!(micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] &
						MICCO_STEREO_EN)){
				micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] |=
					MICCO_STEREO_EN;
				micco_codec_write(MICCO_STEREO_AMPLITUDE_CH1,
						micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1]);
				micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH2] |=
					MICCO_STEREO_EN;
				micco_codec_write(MICCO_STEREO_AMPLITUDE_CH2,
						micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH2]);
			}
			break;

		case CODEC_LINE_OUT:
			if (!(micco_audio_regs[MICCO_AUDIO_LINE_AMP] &
						MICCO_AUDIO_LINE_AMP_EN)) {
				micco_audio_regs[MICCO_AUDIO_LINE_AMP] |=
					MICCO_AUDIO_LINE_AMP_EN;
				micco_codec_write(MICCO_AUDIO_LINE_AMP,
						micco_audio_regs[MICCO_AUDIO_LINE_AMP]);
			}
			break;

		case CODEC_ADC:
			if (!(micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] &
						MICCO_VCODEC_ADC_ON_EN)) {
				micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] |=
					MICCO_VCODEC_ADC_ON_EN;
				micco_codec_write(MICCO_VCODEC_ADC_CONTROL,
						micco_audio_regs[MICCO_VCODEC_ADC_CONTROL]);
			}
			break;
		default:
			return -EINVAL;
	}

	return 0;
}
EXPORT_SYMBOL(micco_codec_enable_output);

int micco_codec_disable_output(int type)
{
	switch (type) {
	case CODEC_BEAR:
		if (micco_audio_regs[MICCO_BEAR_VOL] & MICCO_BEAR_EN) {
			micco_audio_regs[MICCO_BEAR_VOL] &=  ~MICCO_BEAR_EN;
			micco_codec_write(MICCO_BEAR_VOL,
				 micco_audio_regs[MICCO_BEAR_VOL]);
		}
		break;

	case CODEC_MONO:
		if (micco_audio_regs[MICCO_MONO_VOL] & MICCO_MONO_EN) {
			micco_audio_regs[MICCO_MONO_VOL] &= ~MICCO_MONO_EN;
			micco_codec_write(MICCO_MONO_VOL,
				micco_audio_regs[MICCO_MONO_VOL]);
		}
		break;

	case CODEC_STEREO:
		if (micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] &
				 MICCO_STEREO_EN){
			micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] &=
				~MICCO_STEREO_EN;
			micco_codec_write(MICCO_STEREO_AMPLITUDE_CH1,
				micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1]);
		}
		break;

	case CODEC_LINE_OUT:
		if (micco_audio_regs[MICCO_AUDIO_LINE_AMP] &
				MICCO_AUDIO_LINE_AMP_EN) {
			micco_audio_regs[MICCO_AUDIO_LINE_AMP] &=
				~MICCO_AUDIO_LINE_AMP_EN;
			micco_codec_write(MICCO_AUDIO_LINE_AMP,
				micco_audio_regs[MICCO_AUDIO_LINE_AMP]);
		}
		break;

	case CODEC_ADC:
		if (micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] &
				MICCO_VCODEC_ADC_ON_EN) {
			micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] &=
				~MICCO_VCODEC_ADC_ON_EN;
			micco_codec_write(MICCO_VCODEC_ADC_CONTROL,
				micco_audio_regs[MICCO_VCODEC_ADC_CONTROL]);
		}
		break;
	default:
		return -EINVAL;
	}

	return 0;
}
EXPORT_SYMBOL(micco_codec_disable_output);

int micco_codec_check_output(int type)
{
	int ret = 0;
	switch (type) {
	case CODEC_BEAR:
		if (micco_audio_regs[MICCO_BEAR_VOL] & MICCO_BEAR_EN) {
			ret = 1;
		}
		break;

	case CODEC_MONO:
		if (micco_audio_regs[MICCO_MONO_VOL] & MICCO_MONO_EN) {
			ret = 1;
		}
		break;

	case CODEC_STEREO:
		if (micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] &
				 MICCO_STEREO_EN){
			ret = 1;
		}
		break;

	case CODEC_LINE_OUT:
		if (micco_audio_regs[MICCO_AUDIO_LINE_AMP] &
				MICCO_AUDIO_LINE_AMP_EN) {
			ret = 1;
		}
		break;

	case CODEC_ADC:
		if (micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] &
				MICCO_VCODEC_ADC_ON_EN) {
			ret = 1;
		}
		break;
	default:
		break;
	}
	return ret;
}
EXPORT_SYMBOL(micco_codec_check_output);

/* We don't check the paramater. The caller need make sure
 * that the vol is in legal range.
 */
int micco_codec_set_output_vol(int type, int vol)
{
	switch (type) {
	case CODEC_BEAR:
		micco_audio_regs[MICCO_BEAR_VOL] = (~0x3F &
			micco_audio_regs[MICCO_BEAR_VOL]) | vol;
		micco_codec_write(MICCO_BEAR_VOL,
			micco_audio_regs[MICCO_BEAR_VOL]);
		break;

	case CODEC_MONO:
		micco_audio_regs[MICCO_MONO_VOL] = (~0x3F &
			micco_audio_regs[MICCO_MONO_VOL]) | vol;
		micco_codec_write(MICCO_MONO_VOL,
			micco_audio_regs[MICCO_MONO_VOL]);

	case CODEC_STEREO:
		micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1] =
			(~0x3F & micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1]) |
			vol;
		micco_codec_write(MICCO_STEREO_AMPLITUDE_CH1,
			micco_audio_regs[MICCO_STEREO_AMPLITUDE_CH1]);
		break;

	case CODEC_LINE_OUT:
		micco_audio_regs[MICCO_AUDIO_LINE_AMP] =
			(~0x0F & micco_audio_regs[MICCO_AUDIO_LINE_AMP]) |
			vol;
		micco_codec_write(MICCO_AUDIO_LINE_AMP,
			micco_audio_regs[MICCO_AUDIO_LINE_AMP]);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}
EXPORT_SYMBOL(micco_codec_set_output_vol);

int micco_codec_enable_input(int type)
{
	switch (type) {
	case CODEC_AUX1:
		if (!(micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX1_EN)) {
			micco_audio_regs[MICCO_PGA_AUX1_2] |= MICCO_PGA_AUX1_EN;
			micco_codec_write(MICCO_PGA_AUX1_2,
				micco_audio_regs[MICCO_PGA_AUX1_2]);
		}
		break;

	case CODEC_AUX2:
		if (!(micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX2_EN)) {
			micco_audio_regs[MICCO_PGA_AUX1_2] |= MICCO_PGA_AUX2_EN;
			micco_codec_write(MICCO_PGA_AUX1_2,
				micco_audio_regs[MICCO_PGA_AUX1_2]);
		}
		break;

	case CODEC_AUX3:
		if (!(micco_audio_regs[MICCO_PGA_AUX3] & MICCO_PGA_AUX3_EN)) {
			micco_audio_regs[MICCO_PGA_AUX3] |= MICCO_PGA_AUX3_EN;
			micco_codec_write(MICCO_PGA_AUX3,
				micco_audio_regs[MICCO_PGA_AUX3]);
		}
		break;

	case CODEC_MIC1:
		micco_audio_regs[MICCO_MIC_PGA] &= ~MICCO_MIC_PGA_SELMIC_2;
		micco_audio_regs[MICCO_MIC_PGA] =
		       micco_audio_regs[MICCO_MIC_PGA] |
		       MICCO_MIC_PGA_EXT_EN | MICCO_MIC_PGA_INT_EN |
		       MICCO_MIC_PGA_AMP_EN;
		micco_codec_write(MICCO_MIC_PGA,
			micco_audio_regs[MICCO_MIC_PGA]);
		break;

	case CODEC_MIC2:
		micco_audio_regs[MICCO_MIC_PGA] |= MICCO_MIC_PGA_SELMIC_2;
		micco_audio_regs[MICCO_MIC_PGA] =
		       micco_audio_regs[MICCO_MIC_PGA] |
		       MICCO_MIC_PGA_EXT_EN | MICCO_MIC_PGA_INT_EN |
		       MICCO_MIC_PGA_AMP_EN;
		micco_codec_write(MICCO_MIC_PGA,
			micco_audio_regs[MICCO_MIC_PGA]);
		break;

	case CODEC_PCM:
		if (!(micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] &
				MICCO_VDAC_ON)) {
			micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] |=
				MICCO_VDAC_ON;
			micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] &=
				~MICCO_VDAC_HPF_MUTE;
			micco_codec_write(MICCO_VCODEC_VDAC_CONTROL,
				micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL]);
		}
		break;

	case CODEC_HIFI:
		if (!(micco_audio_regs[MICCO_HIFI_DAC_CONTROL] &
				MICCO_HIFI_DAC_ON)) {
			micco_audio_regs[MICCO_HIFI_DAC_CONTROL] |=
				MICCO_HIFI_DAC_ON;
			micco_codec_write(MICCO_HIFI_DAC_CONTROL,
				micco_audio_regs[MICCO_HIFI_DAC_CONTROL]);
		}
		break;

	default:
		return -EINVAL;
	}

	return 0;
}
EXPORT_SYMBOL(micco_codec_enable_input);

int micco_codec_disable_input(int type)
{
	switch (type) {
		case CODEC_AUX1:
			if (micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX1_EN) {
				micco_audio_regs[MICCO_PGA_AUX1_2] &=
					~MICCO_PGA_AUX1_EN;
				micco_codec_write(MICCO_PGA_AUX1_2,
						micco_audio_regs[MICCO_PGA_AUX1_2]);
			}
			break;

		case CODEC_AUX2:
			if (micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX2_EN) {
				micco_audio_regs[MICCO_PGA_AUX1_2] &=
					~MICCO_PGA_AUX2_EN;
				micco_codec_write(MICCO_PGA_AUX1_2,
						micco_audio_regs[MICCO_PGA_AUX1_2]);
			}
			break;

		case CODEC_AUX3:
			if (micco_audio_regs[MICCO_PGA_AUX3] & MICCO_PGA_AUX3_EN) {
				micco_audio_regs[MICCO_PGA_AUX3] &= ~MICCO_PGA_AUX3_EN;
				micco_codec_write(MICCO_PGA_AUX3,
						micco_audio_regs[MICCO_PGA_AUX3]);
			}
			break;

		case CODEC_MIC1:
			micco_audio_regs[MICCO_MIC_PGA] =
				micco_audio_regs[MICCO_MIC_PGA] &
				~(MICCO_MIC_PGA_EXT_EN | MICCO_MIC_PGA_INT_EN |
						MICCO_MIC_PGA_AMP_EN);
			micco_codec_write(MICCO_MIC_PGA,
					micco_audio_regs[MICCO_MIC_PGA]);
			break;

		case CODEC_MIC2:
			micco_audio_regs[MICCO_MIC_PGA] =
				micco_audio_regs[MICCO_MIC_PGA] &
				~(MICCO_MIC_PGA_EXT_EN | MICCO_MIC_PGA_INT_EN |
						MICCO_MIC_PGA_AMP_EN);
			micco_codec_write(MICCO_MIC_PGA,
					micco_audio_regs[MICCO_MIC_PGA]);
			break;

		case CODEC_PCM:
			if (micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] &
					MICCO_VDAC_ON) {
				micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] &=
					~MICCO_VDAC_ON;
				micco_codec_write(MICCO_VCODEC_VDAC_CONTROL,
						micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL]);
			}
			break;

		case CODEC_HIFI:
			if (micco_audio_regs[MICCO_HIFI_DAC_CONTROL] &
					MICCO_HIFI_DAC_ON) {
				micco_audio_regs[MICCO_HIFI_DAC_CONTROL] &=
					~MICCO_HIFI_DAC_ON;
				micco_codec_write(MICCO_HIFI_DAC_CONTROL,
						micco_audio_regs[MICCO_HIFI_DAC_CONTROL]);
			}
			break;

		default:
			return -EINVAL;
	}
	return 0;
}
EXPORT_SYMBOL(micco_codec_disable_input);

int micco_codec_check_input(int type)
{
	int ret = 0;
	switch (type) {
	case CODEC_AUX1:
		if (micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX1_EN) {
			ret = 1;
		}
		break;

	case CODEC_AUX2:
		if (micco_audio_regs[MICCO_PGA_AUX1_2] & MICCO_PGA_AUX2_EN) {
			ret = 1;
		}
		break;

	case CODEC_AUX3:
		if (micco_audio_regs[MICCO_PGA_AUX3] & MICCO_PGA_AUX3_EN) {
			ret = 1;
		}
		break;

	case CODEC_MIC1:
		if (!(micco_audio_regs[MICCO_MIC_PGA] & 
			MICCO_MIC_PGA_SELMIC_2)) {
			if(micco_audio_regs[MICCO_MIC_PGA] &
				MICCO_MIC_PGA_EXT_EN & MICCO_MIC_PGA_INT_EN &
				MICCO_MIC_PGA_AMP_EN) {
				ret = 1;
			}
		}
		break;

	case CODEC_MIC2:
		if ((micco_audio_regs[MICCO_MIC_PGA] & 
			MICCO_MIC_PGA_SELMIC_2)) {
			if(micco_audio_regs[MICCO_MIC_PGA] &
				MICCO_MIC_PGA_EXT_EN & MICCO_MIC_PGA_INT_EN &
				MICCO_MIC_PGA_AMP_EN) {
				ret = 1;
			}
		}

		break;

	case CODEC_PCM:
		if (micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] &
				MICCO_VDAC_ON) {
			ret = 1;
		}
		break;

	case CODEC_HIFI:
		if (micco_audio_regs[MICCO_HIFI_DAC_CONTROL] &
				MICCO_HIFI_DAC_ON) {
			ret = 1;
		}
		break;

	default:
		break;
	}
	return ret;
}
EXPORT_SYMBOL(micco_codec_check_input);

/* We don't check the paramater. The caller need make sure
 * that the vol is in legal range.
 */
int micco_codec_set_input_gain(int type, int gain)
{
	switch (type) {
		case CODEC_AUX1:
			micco_audio_regs[MICCO_PGA_AUX1_2] = ((~0x03 &
						micco_audio_regs[MICCO_PGA_AUX1_2]) | gain);
			micco_codec_write(MICCO_PGA_AUX1_2,
					micco_audio_regs[MICCO_PGA_AUX1_2]);
			break;

		case CODEC_AUX2:
			micco_audio_regs[MICCO_PGA_AUX1_2] = ((~0x30 &
						micco_audio_regs[MICCO_PGA_AUX1_2]) | (gain << 4));
			micco_codec_write(MICCO_PGA_AUX1_2,
					micco_audio_regs[MICCO_PGA_AUX1_2]);
			break;

		case CODEC_AUX3:
			micco_audio_regs[MICCO_PGA_AUX3] = ((~0x03 &
						micco_audio_regs[MICCO_PGA_AUX3]) | gain);
			micco_codec_write(MICCO_PGA_AUX3,
					micco_audio_regs[MICCO_PGA_AUX3]);
			break;

		case CODEC_MIC1:
		case CODEC_MIC2:
			micco_audio_regs[MICCO_MIC_PGA] = ((~0x07 &
						micco_audio_regs[MICCO_MIC_PGA]) | gain);
			micco_codec_write(MICCO_MIC_PGA,
					micco_audio_regs[MICCO_MIC_PGA]);
			break;

		case CODEC_PCM:	/* Need check whether HIFI and PCM support input? */
		case CODEC_HIFI:
			break;
		default:
			return -EINVAL;
	}

	return 0;
}

int micco_codec_set_sample_rate(int port, int rate)
{
	switch (port) {
		case MICCO_VOICE_PORT:
			switch (rate) {
				case 8000:
					micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] =
						(~(0x03 << 3) &
						 micco_audio_regs[MICCO_VCODEC_ADC_CONTROL]);
					break;

				case 16000:
					micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] =
						(~(0x03 << 3) &
						 micco_audio_regs[MICCO_VCODEC_ADC_CONTROL]) |
						(0x01 << 3);
					break;

				case 32000:
					micco_audio_regs[MICCO_VCODEC_ADC_CONTROL] |=
						(0x3 << 3);
					break;
				default:
					return -EINVAL;
			}
			micco_codec_write(MICCO_VCODEC_ADC_CONTROL,
					micco_audio_regs[MICCO_VCODEC_ADC_CONTROL]);
			break;

		case MICCO_HIFI_PORT:
			switch (rate) {
				case 8000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					break;
				case 11025:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x01;
					break;
				case 12000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x02;
					break;
				case 16000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x03;
					break;
				case 22050:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x04;
					break;
				case 24000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x05;
					break;
				case 32000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x06;
					break;
				case 44100:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x07;
					break;
				case 48000:
					micco_audio_regs[MICCO_I2S_CONTROL] &= 0xF0;
					micco_audio_regs[MICCO_I2S_CONTROL] |= 0x0F;
					break;
				default:
					return -EINVAL;
			}
			micco_codec_write(MICCO_I2S_CONTROL,
					micco_audio_regs[MICCO_I2S_CONTROL]);
			break;
		default:
			return -EINVAL;
	}

	return 0;
}
EXPORT_SYMBOL(micco_codec_set_sample_rate);

/* micco_set_vibrator parameter:
 *    SWITCH mode, repetition frequency controled by VIBRA_PWM<6:0>
 * 	00h : OFF
 * 	>= 54h: 100% alwayes on (code 84 decimal)
 */
int micco_set_vibrator(unsigned char value)
{
	micco_write(MICCO_VIBRA_CONTROL, value);
	return 0;
}
EXPORT_SYMBOL(micco_set_vibrator);

int micco_sidetone_enable(void)
{
	if (!(micco_audio_regs[MICCO_SIDETONE] & MICCO_SIDETONE_EN)) {
		micco_audio_regs[MICCO_SIDETONE] |= MICCO_SIDETONE_EN;
		micco_codec_write(MICCO_SIDETONE,
				micco_audio_regs[MICCO_SIDETONE]);
	}

	return 0;
}

int micco_sidetone_disable(void)
{
	if (micco_audio_regs[MICCO_SIDETONE] & MICCO_SIDETONE_EN) {
		micco_audio_regs[MICCO_SIDETONE] &= ~MICCO_SIDETONE_EN;
		micco_codec_write(MICCO_SIDETONE,
				micco_audio_regs[MICCO_SIDETONE]);
	}

	return 0;
}

/* The Micco route rule:
 * 1. DAC3 can't be route to TX_PGA.
 * 2. MIC1 and MIC2 can't be enabled at the same time.
 */
int micco_check_route(u16 *routemap)
{
	/* TODO */
	return 0;
}

int micco_set_route(u16 *rmap, u16 *current_rmap)
{
	int ret;

	ret = micco_check_route(current_rmap);
	if (ret)
		return ret;

	return 0;
}

/*	Test functions for 
 *	mono amplifier testing
 *	input: 	MIC1
 *	output:	MONO
 * */
void mic1_to_mono(u8 level)
{
	u8 val, i;	

	/*Enable mic1 but disable mic amplifier  0x7D*/
	micco_codec_enable_input(CODEC_MIC1);
	micco_audio_regs[MICCO_MIC_PGA] &= ~MICCO_MIC_PGA_AMP_EN;
	micco_codec_write(MICCO_MIC_PGA,micco_audio_regs[MICCO_MIC_PGA]);


	//micco_codec_disable_input(CODEC_MIC2);
	micco_codec_disable_input(CODEC_AUX1);
	micco_codec_disable_input(CODEC_AUX2);
	micco_codec_disable_input(CODEC_AUX3);
	micco_codec_disable_input(CODEC_PCM);
	micco_codec_disable_input(CODEC_HIFI);

	//MIC-PGA 0x7D
	micco_codec_set_input_gain(CODEC_MIC1, 0);	// 0~7 (0dB~33dB)

	//TX-PGA MUX 0x7E 
	micco_audio_regs[MICCO_TX_PGA_MUX] =0x0C; 	//MIC only
	micco_codec_write(MICCO_TX_PGA_MUX, micco_audio_regs[MICCO_TX_PGA_MUX]);

	//TX-PGA 0x7C
	micco_audio_regs[MICCO_TX_PGA] =0x0C; // 0x1E;
	micco_codec_write(MICCO_TX_PGA, micco_audio_regs[MICCO_TX_PGA]);

	//PCM loop 0x80
	micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL] = MICCO_VDAC_PCM_LOOP|
		MICCO_VDAC_ON | MICCO_VDAC_HPF_BYPASS;
	micco_codec_write(MICCO_VCODEC_VDAC_CONTROL  , micco_audio_regs[MICCO_VCODEC_VDAC_CONTROL]);


	//SIDETONE
	//	micco_sidetone_enable();					//SIDE TONE eanble
	//SIDETONE-PGA
	//micco_audio_regs[MICCO_SIDETONE] &= 0xE0;	//0dB sidetone 
	//micco_codec_write(MICCO_SIDETONE, micco_audio_regs[MICCO_SIDETONE]);

	//DAC3 0x84
	micco_audio_regs[MICCO_PGA_DACS] =0x00; 	//ALL DACs
	micco_codec_write(MICCO_PGA_DACS, micco_audio_regs[MICCO_PGA_DACS]);

	//DAC3 to MONO 0x70
	micco_codec_write(MICCO_MUX_MONO, 0x20);	//Only DAC3

	//Hanson add
	//DAC1 DAC2 DAC3 to MONO STEREO
	micco_codec_write(MICCO_MUX_MONO, 0x38);
	micco_codec_write(MICCO_MUX_STEREO_CH1, 0x38);
	micco_codec_write(MICCO_MUX_STEREO_CH2, 0x38);

	/*Set output*/
	micco_codec_enable_output(CODEC_MONO);
	//Hanson add
	micco_codec_enable_output(CODEC_STEREO);
	micco_codec_enable_output(CODEC_ADC);

	micco_codec_disable_output(CODEC_LINE_OUT);
	micco_codec_disable_output(CODEC_BEAR);
	//Hanson add
	//micco_codec_disable_output(CODEC_STEREO);

	//Amplifier Enable 0x75
	micco_audio_regs[MICCO_AUDIO_LINE_AMP] |= LINE_OUT_GPO;
	micco_codec_write(MICCO_AUDIO_LINE_AMP, micco_audio_regs[MICCO_AUDIO_LINE_AMP]);


	printk(KERN_INFO "%s: Input volume level = %d\n", __func__, level );

	//MONO VOLUME 0x79
	micco_codec_set_output_vol(CODEC_MONO, (int)level);

	//Hanson add
	//STEREO VOLUMN 0x76 0x77
	micco_codec_set_output_vol(CODEC_STEREO, 0x0c);

	//0x78
	micco_codec_write(MICCO_HIFI_DAC_CONTROL, 0x80);

	//Show all reg info
	for (i=0; i <0x17; i++){
		micco_codec_read(i, &val);
		printk(KERN_INFO " reg %x = 0x%x\n",MICCO_AUDIO_REG_BASE +i , val);
	}
}
/* Micco Audio primitive end here */

/* Initialization functions for system */
/* LDO12 enabled for IO */
void micco_enable_LDO12(void)
{
	u8 val;

	micco_read(MICCO_LDO12, &val);
	val = (0x0F & val);
	micco_write(MICCO_LDO12, val);

	return;
}

/* USB Device/OTG functions */
static int micco_set_pump(int enable)
{
	int status;
	u8 val;
	unsigned long flags;

	local_irq_save(flags);
	if (enable) {
		status = micco_read(MICCO_MISC, &val);
		if (status)
			goto out;
		val |= MICCO_MISC_VBUS_COMPS_EN;
		status = micco_write(MICCO_MISC, val);
		if (status)
			goto out;

		/* FIXME: Littleton didn't connect EXTON as cable detection
		 * signal. We use USB_DEV detection in event B as cable
		 * detection.
		 */
		/* TODO: This is a platform related code. Need split
		 * to platform related code.
		 */
		status = micco_read(MICCO_IRQ_MASK_B, &val);
		if (status)
			goto out;
		val &= ~IRQ_MASK_B_USB_DEV;
		status = micco_write(MICCO_IRQ_MASK_B, val);
		if (status)
			goto out;
	} else {
		status = micco_read(MICCO_MISC, &val);
		if (status)
			goto out;
		val &= ~MICCO_MISC_VBUS_COMPS_EN;
		status = micco_write(MICCO_MISC, val);
		if (status)
			goto out;

		status = micco_read(MICCO_IRQ_MASK_B, &val);
		if (status)
			goto out;
		val |= IRQ_MASK_B_USB_DEV;
		status = micco_write(MICCO_IRQ_MASK_B, val);
		if (status)
			goto out;
	}
out:
	local_irq_restore(flags);
	return status;
}

/* 1. enable: 1, srp: 0, 100ma mode
 * 2. enable: 1, srp: 1, 10ma mode
 *
 * Micco before e: 1:s: 1, must e:0, s:1.
 */
static int micco_set_vbus_supply(int enable, int srp)
{
	int ret;
	u8 val, tmp;

	ret = micco_read(MICCO_MISC, &val);
	if (ret) {
		printk(KERN_ERR "%s: I2C operation read error 0x%x\n", __FUNCTION__, ret);
		return ret;
	}

	if (enable) {
		/* When enable the USB pump, need check the OTGCP_IOVER. */
		micco_read(MICCO_IRQ_MASK_B, &tmp);
		tmp &= ~IRQ_MASK_B_OTGCP_IOVER;
		micco_write(MICCO_IRQ_MASK_B, tmp);

		val |= MICCO_MISC_USBCP_EN;
		if (srp) {
			val |= MICCO_MISC_USBSR_EN;
		} else {
			val &= ~MICCO_MISC_USBSR_EN;
		}
	} else {
		/* When disable the USB pump, needn't check the OTGCP_IOVER. */
		micco_read(MICCO_IRQ_MASK_B, &tmp);
		tmp |= IRQ_MASK_B_OTGCP_IOVER;
		micco_write(MICCO_IRQ_MASK_B, tmp);

		val &= ~(MICCO_MISC_USBCP_EN | MICCO_MISC_USBSR_EN);
	}
	ret = micco_write(MICCO_MISC, val);
	if (ret)
		printk(KERN_ERR "%s: I2C operation write error 0x%x\n", __FUNCTION__, ret);

	return ret;
}

static int micco_set_usbotg_a_mask(void)
{
	int ret;
	u8 val;

	ret = micco_read(MICCO_IRQ_MASK_B, &val);
	if (ret) {
		printk(KERN_ERR "%s: I2C operation read error 0x%x\n", __FUNCTION__, ret);
		return ret;
	}

	/* Enable the interrupt for VUBS > 4.4V and Session valid
	 * which A device care about.
	 *
	 * NOTESSSSSSSSS: We care about the SRP detection signal (0.8v ~ 2.0v)
	 * on Micco. Which map to SESSION_VALID_1_8.
	 */
	val |= (IRQ_MASK_B_VBUS_VALID_3_8 | IRQ_MASK_B_SRP_READY_0_6);
	val &= ~(IRQ_MASK_B_VBUS_VALID_4_55 | IRQ_MASK_B_SESSION_VALID_1_8);

	ret = micco_write(MICCO_IRQ_MASK_B, val);
	return ret;
}

static int micco_set_usbotg_b_mask(void)
{
	int ret;
	u8 val;

	ret = micco_read(MICCO_IRQ_MASK_B, &val);
	if (ret) {
		printk(KERN_ERR "%s: I2C operation read error 0x%x\n", __FUNCTION__, ret);
		return ret;
	}

	/* Mask all USB VBUS interrupt for B device */
	val |= (IRQ_MASK_B_VBUS_VALID_3_8 | IRQ_MASK_B_SESSION_VALID_1_8 | \
			IRQ_MASK_B_VBUS_VALID_4_55 | IRQ_MASK_B_SRP_READY_0_6);

	ret = micco_write(MICCO_IRQ_MASK_B, val);

	return ret;
}

static int is_micco_vbus_assert(void)
{
	int ret;
	u8 val;

	ret = micco_read(MICCO_STATUS_B, &val);
	if (ret) {
		printk(KERN_ERR "%s: I2C operation read error 0x%x\n", __FUNCTION__, ret);
		return ret;
	}
	
	/* USBDEV interrupt is not so reliable for vbus assert check,
 	 * replace with 4.4V check 
 	 * 
	if (val & MICCO_STATUS_B_USBDEV) */
	if (val & MICCO_STATUS_B_VBUS_V_4P4)
		return 1;
	else
		return 0;
}

#define	MICCO_EVENT_LEN		4
#define EVENT_A_OFF		0
#define EVENT_B_OFF		1
#define EVENT_C_OFF		2
#define EVENT_D_OFF		3

static int micco_event_change(unsigned int *event)
{
	u8 v[MICCO_EVENT_LEN] = {0, 0, 0, 0};
	int ret = 0, i = 0, retry = 0;

read_block_event:
	/* read event[A-D] data once instead of four times */		
	ret = micco_read_block(MICCO_EVENT_A, v, MICCO_EVENT_LEN);
	if (ret < 0) {
		/* read block error, try three times */
		mdelay(10); 
		if (retry < 3) {
			retry++;
			goto read_block_event;
		}
			
		/* read block error, try to read single address again */
		for (i = 0; i < MICCO_EVENT_LEN; i++) {
			ret = micco_read(MICCO_EVENT_A+i, &v[i]);
			if (ret < 0)
				return ret;
		}
	}

	*event = (v[3] << 24) | (v[2] << 16) | (v[1] << 8) | v[0];
	return 0;
}

static int is_micco_avbusvld(void)
{
	u8 val;

	micco_read(MICCO_IRQ_MASK_B, &val);
	if (val & IRQ_MASK_B_VBUS_VALID_4_55)
		return 0;

	micco_read(MICCO_STATUS_B, &val);
	if (val & MICCO_STATUS_B_VBUS_V_4P4)
		return 1;
	else
		return 0;
}

static int is_micco_asessvld(void)
{
	u8 val;

	micco_read(MICCO_IRQ_MASK_B, &val);
	if (val & IRQ_MASK_B_SESSION_VALID_1_8)
		return 0;

	micco_read(MICCO_STATUS_B, &val);
	if (val & MICCO_STATUS_B_SESS_V_1P8)
		return 1;
	else
		return 0;
}

static int is_micco_bsessvld(void)
{
	u8 val;

	micco_read(MICCO_IRQ_MASK_B, &val);
	if (val & IRQ_MASK_B_VBUS_VALID_3_8)
		return 0;

	micco_read(MICCO_STATUS_B, &val);
	if (val & MICCO_STATUS_B_VBUS_V_3P8)
		return 1;
	else
		return 0;
}

static int is_micco_srp_ready(void)
{
	u8 val;

	micco_read(MICCO_IRQ_MASK_B, &val);
	if (val & IRQ_MASK_B_SESSION_VALID_1_8)
		return 0;

	/* FIXME: When cable detached, the SESS Valid
	 * of STATUA B can not change to 0 immediately.
	 * That will cause potential problems.
	 */
	micco_read(MICCO_STATUS_B, &val);
	if (val & MICCO_STATUS_B_SRP_READY)
		return 1;
	else
		return 0;
}

static int is_micco_onkey_assert(void)
{
	u8 val;

	micco_read(MICCO_STATUS_A, &val);

	return (!(val & MICCO_STATUS_A_NONKEY));
}

static int get_power_supply_module(int cmd)
{
	int command, power_module;

	if (cmd < VCC_CORE || cmd >= MAX_CMD) {
		printk(KERN_WARNING "input wrong command: %d\n", cmd);
		return -EINVAL;
	}
	command = micco_power_module[cmd - VCC_CORE].command;
	if (command != cmd) {
		printk(KERN_WARNING "micco_power_module[] is error: %d\n", cmd);
		return -EINVAL;
	}
	power_module = micco_power_module[cmd - VCC_CORE].power_module;
	if (power_module == 0) {
		printk(KERN_WARNING "the command not supported: %d\n", cmd);
		return -EINVAL;
	}

	return power_module;
}

/* USB Device/OTG functions end here*/
static int get_micco_voltage(int cmd, int *pmv)
{
	int power_module, status = 0;
	u8 val;

	*pmv = 0;

	power_module = get_power_supply_module(cmd);

	start_calc_time();
	switch (power_module) {
	case BUCK1:
		status = micco_read(MICCO_BUCK1_DVC2, &val);
		if (status)
			return status;

		val &= 0x1f;
		*pmv = val * MICCO_VBUCK1_STEP + MICCO_VBUCK1_BASE;
		break;
	case BUCK2: /* TODO */
		*pmv = 1800;
		break;
	case BUCK3: /* TODO */
		*pmv = 1800;
		break;
	case LDO1:
		status = micco_read(MICCO_LDO1_MDTV1, &val);
		if (status)
			return status;

		val &= 0x0f;
		*pmv = val * MICCO_VLDO1_STEP + MICCO_VLDO1_BASE;
		break;
	case LDO2:
		status = micco_read(MICCO_SRAM_DVC2, &val);
		if (status)
			return status;

		val = val & 0x1f;
		*pmv = val * MICCO_VSRAM_STEP + MICCO_VSRAM_BASE;
		break;
	case LDO3:
		status = micco_read(MICCO_LDO643, &val);
		if (status)
			return status;

		val &= 0x0f;
		*pmv = val * MICCO_VLDO3_STEP + MICCO_VLDO3_BASE;
		break;
	case LDO4:
		status = micco_read(MICCO_LDO643, &val);
		if (status)
			return status;

		if (val & 0x10)
			*pmv = 2900;
		else
			*pmv = 1800;
		break;
	case LDO5: /* TODO */
		*pmv = 3100;
		break;
	case LDO6:
		status = micco_read(MICCO_LDO643, &val);
		if (status)
			return status;

		val = (val & 0xE0) >> 5;
		*pmv = val * MICCO_VLDO6_STEP + MICCO_VLDO6_BASE;
		break;
	case LDO7: /* TODO */
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
	case LDO8: /* TODO */
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
	case LDO9:
		status = micco_read(MICCO_LDO987, &val);
		if (status)
			return status;

		val = (val & 0xE0) >> 5;
		*pmv = val * MICCO_VLDO9_STEP + MICCO_VLDO9_BASE;
		break;
	case LDO10:
		status = micco_read(MICCO_LDO987, &val);
		if (status)
			return status;

		val &= 0x07;
		*pmv = val * MICCO_VLDO10_STEP + MICCO_VLDO10_BASE;
		break;
	case LDO11:
		status = micco_read(MICCO_LDO1110, &val);
		if (status)
			return status;

		val = (val & 0xf0) >> 4;
		*pmv = val * MICCO_VLDO11_STEP + MICCO_VLDO11_BASE;
		break;
	case LDO12:
		status = micco_read(MICCO_LDO12, &val);
		if (status)
			return status;

		val &= 0x07;
		if (val & 0x08)
			*pmv = val * MICCO_VLDO12_STEP + MICCO_VLDO12_BASE1;
		else
			*pmv = val * MICCO_VLDO12_STEP + MICCO_VLDO12_BASE;
		break;
	case LDO13:
		status = micco_read(MICCO_LDO13, &val);
		if (status)
			return status;

		val &= 0x0f;
		*pmv = val * MICCO_VLDO13_STEP + MICCO_VLDO13_BASE;
		break;
	case LDO14:
	case LDO15:
	case USB_OTG: /* TODO */
	case LDO_GPADC: /* TODO */
	case LDO_AUDIO: /* TODO */
	case LDO_PMCORE: /* TODO */
	case LDO_BBAT: /* TODO */
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
	default:
		printk(KERN_WARNING "input wrong command: %d\n", cmd);
		return -EINVAL;
	}
	end_calc_time();
	return status;
}

static int set_micco_voltage(int cmd, int mv)
{
	int power_module, status = 0;
	u8 val;

	power_module = get_power_supply_module(cmd);

	start_calc_time();
	switch (power_module) {
	case BUCK1:
		if ((mv < MICCO_VBUCK1_BASE) || (mv > MICCO_VBUCK1_MAX))
			return -EINVAL;
		status = micco_read(MICCO_BUCK1_DVC2, &val);
		if (status)
			return status;

		val &= 0xe0;
		val = val | ((mv - MICCO_VBUCK1_BASE) / MICCO_VBUCK1_STEP);
		status = micco_write(MICCO_BUCK1_DVC2, val);
		/* Update the voltage output. Otherwise, voltage won't change */
		status = micco_write(MICCO_VCC1, 0x33);
		break;
	case BUCK2: /* TODO */
		return 0;
	case BUCK3: /* TODO */
		return 0;
	case LDO1:
		if ((mv < MICCO_VLDO1_BASE) || (mv > MICCO_VLDO1_MAX))
			return -EINVAL;
		status = micco_read(MICCO_LDO1_MDTV1, &val);
		if (status)
			return status;

		val &= 0xf0;
		val = val | ((mv - MICCO_VLDO1_BASE) / MICCO_VLDO1_STEP);
		status = micco_write(MICCO_LDO1_MDTV1, val);
		break;
	case LDO2:
		if ((mv < MICCO_VSRAM_BASE) || mv > (MICCO_VSRAM_MAX))
			return -EINVAL;
		status = micco_read(MICCO_SRAM_DVC2, &val);
		if (status)
			return status;

		val &= 0xe0;
		val = val | ((mv - MICCO_VSRAM_BASE) / MICCO_VSRAM_STEP);
		status = micco_write(MICCO_SRAM_DVC2, val);
		/* Update the voltage output. Otherwise, voltage won't change */
		status = micco_write(MICCO_VCC1, 0x33);
		break;
	case LDO3:
		if ((mv < MICCO_VLDO3_BASE) || (mv > MICCO_VLDO3_MAX))
			return -EINVAL;

		status = micco_read(MICCO_LDO643, &val);
		if (status)
			return status;

		val &= 0xf0;
		val = val | ((mv - MICCO_VLDO3_BASE) / MICCO_VLDO3_STEP);
		status = micco_write(MICCO_LDO643, val);
		break;
	case LDO4:
		if ((mv != 2900) && (mv != 1800))
			return -EINVAL;
		status = micco_read(MICCO_LDO643, &val);
		if (2900 == mv)
			val |= 0x10;
		else
			val &= ~0x10;
		status = micco_write(MICCO_LDO643, val);
		break;
	case LDO5: /* TODO */
		return 0;
	case LDO6:
		if ((mv < MICCO_VLDO6_BASE) || (mv > MICCO_VLDO6_MAX))
			return -EINVAL;

		status = micco_read(MICCO_LDO643, &val);
		if (status)
			return status;

		val &= 0x1f;
		val |= ((mv - MICCO_VLDO6_BASE) / MICCO_VLDO6_STEP) << 5;
		status = micco_write(MICCO_LDO643, val);
		break;
	case LDO7:
#if defined(CONFIG_COPYDATA)
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
#else
		if( (mv < MICCO_VLDO7_BASE) || (mv > MICCO_VLDO7_MAX))
			return -EINVAL;
		status = micco_read(MICCO_LDO987, &val);
		if (status)
			return status;

		val &= 0xf8;
		val |= ((mv-MICCO_VLDO7_BASE) / MICCO_VLDO7_STEP);
		status = micco_write(MICCO_LDO987, val);
		micco_read(MICCO_LDO987, &val);
		printk("setting LDO7 to %d, val=%d\n",mv, val);
		break;
#endif
	case LDO8:
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
	case LDO9:
		if ( mv == 0 )
		{
			/* disable LDO9 */
			status = micco_read(MICCO_APP_OVER2, &val);
			if (status)
				return status;
			val &= ~MICCO_LDO9_ENABLE;
			micco_write(MICCO_APP_OVER2, val);
			break;
		}

		if ((mv < MICCO_VLDO9_BASE) || (mv > MICCO_VLDO9_MAX))
			return -EINVAL;
		status = micco_read(MICCO_LDO987, &val);
		if (status)
			return status;

		val &= 0x1f;
		val |= ((mv - MICCO_VLDO9_BASE) / MICCO_VLDO9_STEP) << 5;
		status = micco_write(MICCO_LDO987, val);

		/* enable LDO9 */
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		val |= MICCO_LDO9_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO10:
		if ((mv < MICCO_VLDO10_BASE) || (mv > MICCO_VLDO10_MAX))
			return -EINVAL;

		status = micco_read(MICCO_LDO1110, &val);
		if (status)
			return status;

		val &= 0xf8;
		val = val | ((mv - MICCO_VLDO10_BASE) / MICCO_VLDO10_STEP);
		status = micco_write(MICCO_LDO1110, val);
		break;
	case LDO11:
		if ((mv < MICCO_VLDO11_BASE) || (mv > MICCO_VLDO11_MAX))
			return -EINVAL;

		status = micco_read(MICCO_LDO1110, &val);
		if (status)
			return status;

		val &= 0x0f;
		val |= ((mv - MICCO_VLDO11_BASE) / MICCO_VLDO11_STEP) << 4;
		status = micco_write(MICCO_LDO1110, val);
		break;
	case LDO12:
		if ((mv < MICCO_VLDO12_BASE) || (mv > MICCO_VLDO12_MAX))
			return -EINVAL;


		status = micco_read(MICCO_LDO12, &val);
		if (status)
			return status;

		val &= 0xf8;
		if (mv > MICCO_VLDO12_BASE1) {
			val = val | ((mv - MICCO_VLDO12_BASE1) /	\
				MICCO_VLDO12_STEP);
			val |= 0x08;
		} else {
			val = val | ((mv - MICCO_VLDO12_BASE) /		\
				MICCO_VLDO12_STEP);
			val &= ~0x08;
		}
		status = micco_write(MICCO_LDO12, val);
		break;
	case LDO13:
		if ((mv < MICCO_VLDO13_BASE) || (mv > MICCO_VLDO13_MAX))
			return -EINVAL;

		status = micco_read(MICCO_LDO13, &val);
		if (status)
			return status;

		val &= 0xf0;
		val |= ((mv - MICCO_VLDO13_BASE) / MICCO_VLDO13_STEP);
		status = micco_write(MICCO_LDO13, val);
		break;
	case LDO14:
	case LDO15:
	case USB_OTG: /* TODO */
	case LDO_GPADC: /* TODO */
	case LDO_AUDIO: /* TODO */
	case LDO_PMCORE: /* TODO */
	case LDO_BBAT: /* TODO */
		printk(KERN_WARNING "power supply module TODO: %d\n", cmd);
		return -EINVAL;
	default:
		printk(KERN_INFO "error command\n");
		return -EINVAL;
	}

	if (status != 0)
		return status;

	end_calc_time();
	return status;
}

#if defined(CONFIG_COPYDATA) || defined(CONFIG_PXA930_ACIPC) || defined(CONFIG_BATTERY_BQ27541)
/*
  This function configures the ADC as requires for CP implementation.
  CP does not "own" the ADC configuration registers and relies on AP:
  Reason: enable automatic ADC measurements needed for CP to get VBAT and RF temperature readings.
- Set and keep Micco register 0x50 (MICCO_ADC_MAN_CONTROL) bit 4 (LDO_ADC_enable) set at all times.
- Set and keep Micco register 0x51 (MICCO_ADC_AUTO_CONTROL_1) bits 0 (AUTO_ADC_sleep_enable) set at all times.
- Set and keep Micco register 0x51 (MICCO_ADC_AUTO_CONTROL_1) bits 1 (AUTO_VBAT_enable) set at all times.
- Set and keep Micco register 0x52 (MICCO_ADC_AUTO_CONTROL_2) bits 0 (AUTO4_enable) set at all times.
  ADC channel 4 is connected to RF temperature sensor.
  
 */
static void micco_enable_adc_for_cp(void)
{
	u8 val;

	micco_enable_ldo_adc(1);

	micco_read(MICCO_ADC_AUTO_CONTROL_1, &val);
	val |= (MICCO_ADC_AUTO_1_AUTOADC_SLEEP_EN|MICCO_ADC_AUTO_1_VBAT_EN);
	micco_write(MICCO_ADC_AUTO_CONTROL_1, val); 

#if defined(CONFIG_COPYDATA) || defined(CONFIG_PXA930_ACIPC)
	micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	val |= MICCO_ADC_AUTO_2_ADC4_EN;
	micco_write(MICCO_ADC_AUTO_CONTROL_2, val); 
#endif
}
#endif

static int check_micco_voltage(int cmd)
{
	int power_module, status = 0;
	u8 val, ret = 0;

	power_module = get_power_supply_module(cmd);
	switch (power_module) {
	case BUCK1:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (val & MICCO_BUCK1_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case BUCK2:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (val & MICCO_BUCK2_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case BUCK3:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_BUCK3_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO1:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_LDO1_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO2:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (val & MICCO_LDO2_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO3:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_LDO3_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO4:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_LDO4_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO5:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_LDO5_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO6:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO6_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO7:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO7_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO8:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO8_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO9:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO9_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO10:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO10_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO11:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO11_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO12:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (val & MICCO_LDO12_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO13:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (val & MICCO_LDO13_ENABLE)
			ret = 1;	
		else
			ret = 0;	
		break;
	case LDO14:
	default:
		printk(KERN_INFO "error command\n");
		return -EINVAL;
	}

	return ret;
}

static int enable_micco_voltage(int cmd, int enable)
{
	int power_module, status = 0;
	u8 val;

	power_module = get_power_supply_module(cmd);
	switch (power_module) {
	case BUCK1:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_BUCK1_ENABLE;
		else
			val &= ~MICCO_BUCK1_ENABLE;
		micco_write(MICCO_OVER1, val);
		break;
	case BUCK2:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_BUCK2_ENABLE;
		else
			val &= ~MICCO_BUCK2_ENABLE;
		micco_write(MICCO_OVER1, val);
		break;
	case BUCK3:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_BUCK3_ENABLE;
		else
			val &= ~MICCO_BUCK3_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO1:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO1_ENABLE;
		else
			val &= ~MICCO_LDO1_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO2:
		status = micco_read(MICCO_OVER1, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO2_ENABLE;
		else
			val &= ~MICCO_LDO2_ENABLE;
		micco_write(MICCO_OVER1, val);
		break;
	case LDO3:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO3_ENABLE;
		else
			val &= ~MICCO_LDO3_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO4:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO4_ENABLE;
		else
			val &= ~MICCO_LDO4_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO5:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO5_ENABLE;
		else
			val &= ~MICCO_LDO5_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO6:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO6_ENABLE;
		else
			val &= ~MICCO_LDO6_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO7:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO7_ENABLE;
		else
			val &= ~MICCO_LDO7_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO8:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO8_ENABLE;
		else
			val &= ~MICCO_LDO8_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO9:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO9_ENABLE;
		else
			val &= ~MICCO_LDO9_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO10:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO10_ENABLE;
		else
			val &= ~MICCO_LDO10_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO11:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO11_ENABLE;
		else
			val &= ~MICCO_LDO11_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO12:
		status = micco_read(MICCO_APP_OVER2, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO12_ENABLE;
		else
			val &= ~MICCO_LDO12_ENABLE;
		micco_write(MICCO_APP_OVER2, val);
		break;
	case LDO13:
		status = micco_read(MICCO_APP_OVER3, &val);
		if (status)
			return status;
		if (enable)
			val |= MICCO_LDO13_ENABLE;
		else
			val &= ~MICCO_LDO13_ENABLE;
		micco_write(MICCO_APP_OVER3, val);
		break;
	case LDO14:
	default:
		printk(KERN_INFO "error command\n");
		return -EINVAL;
	}

	return status;
}

/*
 * Micco interrupt service routine.
 * In the ISR we need to check the Status bits in Micco and according
 * to those bits to check which kind of IRQ had happened.
 */
static irqreturn_t micco_irq_handler(int irq, void *dev_id)
{
	struct i2c_client *client = dev_id;
	struct micco_platform_data *pdata = client->dev.platform_data;

	/* clear the irq */
	pdata->ack_irq();
	pr_debug("Micco: micco irq handler\n");

	queue_work(micco_d->micco_wq, &pdata->work);
	
	wake_lock_timeout(&micco_suspend_lock, 5 * HZ);
#ifdef CONFIG_PM
#if defined(CONFIG_PXA930_ACIPC)
        pxa3xx_lock_suspend_cancel();
#endif
#endif
	return IRQ_HANDLED;
}

#if EXT0_WAKEUP_INTERRUPT
static irqreturn_t ext0_irq_handler(int irq, void *dev_id)
{
	struct i2c_client *client = dev_id;
	struct micco_platform_data *pdata = client->dev.platform_data;
	PECR |= PECR_E0IS;

	pr_debug("Micco ext0: ext0 irq handler\n");
	queue_work(micco_d->micco_wq, &pdata->work);

	wake_lock_timeout(&ext0_suspend_lock, HZ / 2);
#ifdef CONFIG_PM
#if defined(CONFIG_PXA930_ACIPC)
        pxa3xx_lock_suspend_cancel();
#endif
#endif
	return IRQ_HANDLED;
}
#endif

#ifdef CONFIG_PM
void micco_codec_low_power_mode(void)
{
	u8 micco_audio_lpm[MICCO_AUDIO_REGS_NUM] = 
	{ 0x0 , 0x0 , 0x0 , 0x0 ,   0x0 , 0x0 , 0x0 , 0x40,
	  0x60, 0x80, 0x80, 0x17,   0x0 , 0x0 , 0x0 , 0x0 ,
	  0x0 , 0x0 , 0x0 , 0x0 ,   0x0 , 0xff };

	int i; 
	for (i = 0; i < MICCO_AUDIO_REGS_NUM; i++)
		micco_codec_write(i, micco_audio_lpm[i]);
}

void micco_codec_restore(void)
{
	int i; 
	for (i = 0; i < MICCO_AUDIO_REGS_NUM; i++)
		micco_codec_write(i, micco_audio_regs[i]);
}

void micco_android_suspend_handler(struct early_suspend *h)
{
	printk(KERN_DEBUG "micco_android_suspend_handler reached\n");

	micco_write(MICCO_SYSCTRL_A, 0x63);

	printk(KERN_DEBUG "micco_android_suspend_handler returned\n");
}

void micco_android_resume_handler(struct early_suspend *h)
{
	printk(KERN_DEBUG "micco_android_resume_handler reached\n");

	micco_write(MICCO_SYSCTRL_A, 0x60);

	printk(KERN_DEBUG "micco_android_resume_handler returned\n");
}

static struct early_suspend micco_android_suspend = {
	.level = 98,
	.suspend = micco_android_suspend_handler,
	.resume = micco_android_resume_handler,
};

/*
 * Suspend the micco interface.
 * Add register save/restore here.
 */
static int micco_suspend(struct i2c_client *client, pm_message_t state)
{
	unsigned char val;

	disable_irq(client->irq);
	if ((get_pm_state() == PM_SUSPEND_MEM)) {
		micco_read(MICCO_OVER1, &val);
		micco_regs[MICCO_OVER1].data = val;
		micco_read(MICCO_APP_OVER2, &val);
		micco_regs[MICCO_APP_OVER2].data = val;
		micco_read(MICCO_APP_OVER3, &val);
		micco_regs[MICCO_APP_OVER3].data = val;
		micco_read(MICCO_BUCK_SLEEP, &val);
		micco_regs[MICCO_BUCK_SLEEP].data = val;
		micco_read(MICCO_SYSCTRL_A, &val);
		micco_regs[MICCO_SYSCTRL_A].data = val;

		micco_write(MICCO_BUCK_SLEEP, 0x6D);
		micco_write(MICCO_SYSCTRL_A, 0xff);

		/*		micco_write(MICCO_OVER1, 0x04);
				micco_write(MICCO_APP_OVER2, 0x30);
				micco_write(MICCO_APP_OVER3, 0x7c);
				*/
	}
	return 0;
}

/*
 * Resume the Micco interface.
 */
static int micco_resume(struct i2c_client *client)
{
	int i;

	for (i = 0; i < MICCO_REG_NUM; i++)
		micco_regs[i].hit = 0;

	enable_irq(client->irq);
	micco_irq_handler(client->irq, client);
	if ((get_pm_state() == PM_SUSPEND_MEM)) {
		micco_write(MICCO_OVER1, micco_regs[MICCO_OVER1].data);
		micco_write(MICCO_APP_OVER2, micco_regs[MICCO_APP_OVER2].data);
		micco_write(MICCO_APP_OVER3, micco_regs[MICCO_APP_OVER3].data);
		micco_write(MICCO_BUCK_SLEEP, micco_regs[MICCO_BUCK_SLEEP].data);
		micco_write(MICCO_SYSCTRL_A, micco_regs[MICCO_SYSCTRL_A].data);
	}
	return 0;
}

#else				/* CONFIG_PM */
#define	micco_suspend		NULL
#define	micco_resume		NULL
#endif				/* CONFIG_PM */

static u8 rx_value = 0;
static u8 micco_get_rx_value(void) {
	u8 val = rx_value;
	rx_value = 0;
	return val;
}

static struct pmic_ops micco_pmic_ops = {
	.get_voltage		= get_micco_voltage,
	.set_voltage		= set_micco_voltage,
	.enable_voltage		= enable_micco_voltage,
	.check_voltage		= check_micco_voltage,

	.is_vbus_assert		= is_micco_vbus_assert,
	.is_avbusvld		= is_micco_avbusvld,
	.is_asessvld		= is_micco_asessvld,
	.is_bsessvld		= is_micco_bsessvld,
	.is_srp_ready		= is_micco_srp_ready,

	.set_pump		= micco_set_pump,
	.set_vbus_supply	= micco_set_vbus_supply,
	.set_usbotg_a_mask	= micco_set_usbotg_a_mask,
	.set_usbotg_b_mask	= micco_set_usbotg_b_mask,
	
	.is_onkey_assert	= is_micco_onkey_assert,
	.get_rx_value		= micco_get_rx_value,
};

static void micco_worker(struct work_struct *work)
{
	unsigned int event = 0;
	u8 val;

	while (1) {
		if (micco_event_change(&event))
			break;

		if (event == 0)
			break;
#if 0
		if (event != PMIC_EVENT_TOUCH)
			printk(KERN_DEBUG "Micco event = 0x%x\n", event);
#endif
		if (event & PMIC_EVENT_RX_COMP)
			micco_read(MICCO_UART_RX_DATA, &rx_value);
			
		pmic_event_handle(event);

		/* We don't put these codes to USB specific code because
		 * we need handle it even when no USB callback registered.
		 */
		if (event & PMIC_EVENT_OTGCP_IOVER) {
			/* According to Micco spec, when OTGCP_IOVER happen,
			 * Need clean the USBPCP_EN in MISC. and then set
			 * it again.
			 */
			micco_read(MICCO_MISC, &val);
			val &= ~MICCO_MISC_USBCP_EN;
			micco_write(MICCO_MISC, val);
			val |= MICCO_MISC_USBCP_EN;
			micco_write(MICCO_MISC, val);
		}
	}
}

#ifdef	CONFIG_PROC_FS
#define	MICCO_PROC_FILE	"driver/micco"
static struct proc_dir_entry *micco_proc_file;
static int index;

static ssize_t micco_proc_read(struct file *filp, char __user *buffer, size_t length, loff_t *off)
{
	u8 reg_val;
        const char fmt[]="register 0x%02x: 0x%02x\n";
        char buf[sizeof(fmt)];/* %2x is 4 chars while it yields 2 chars in output */
        size_t len;

	if ((index < 0) || (index > MICCO_REG_NUM))
		return 0;

        /* It would be correct to first produce the full output again on the second call with *off>0 
           However, the total size required is so small that re-reading micco reg again is redundant */
        if(*off) return 0;

	micco_read(index, &reg_val);
        
	len=sprintf(buf, fmt, (unsigned)index, (unsigned)reg_val);
        len-=*off;
        if(length<len) return -EFAULT;/*unreasonably short buffer supplied*/
        if(len>length) len=length;
        if (copy_to_user(buffer, buf+(*off), len))
		return -EFAULT;
        *off+=len;
	return len;
}

#define VIBRA_OFF_VALUE	0
#define VIBRA_ON_VALUE	0xfe	/* you can choose the value of bit7 ~ bit1 to change the freq of the vibrator */
static ssize_t micco_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if (strncmp(messages, "dump", 4) == 0){
		int i;
		for (i=1; i <0xE; i++){
			micco_read(i, &reg_val);
			printk(KERN_INFO " reg %x = 0x%x\n",i , reg_val);
		}	

	}
	else if (strncmp(messages, "spk", 3) == 0){
		u8 ampVol;
		/* set the volume value spk0x11 : vol = 0x11 */
		memcpy(vol, messages+3, len-1);

		ampVol = (u8)simple_strtoul(vol, NULL, 16);
		ampVol = (ampVol > MONO_MAX_VOL ) ? MONO_MAX_VOL : ampVol;		
		printk(KERN_INFO "Speaker Amplifier gain : 0x%x!\n", ampVol);
		//MIC ON
		micco_audio_regs[MICCO_MONO_VOL] |= MICCO_MONO_EN;
		micco_codec_write(MICCO_MONO_VOL,micco_audio_regs[MICCO_MONO_VOL]);	
		mic1_to_mono(ampVol);		
	}
	else if (strncmp(messages, "viboff", 6) == 0)
		micco_set_vibrator(VIBRA_OFF_VALUE);
	else if (strncmp(messages, "vibon", 5) == 0)
		micco_set_vibrator(VIBRA_ON_VALUE);
	else if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
	} else {
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		micco_write(index, reg_val & 0xFF);
	}

	return len;
}

static struct file_operations micco_proc_ops = {
	.read = micco_proc_read,
	.write = micco_proc_write,
};

#ifdef DUMP_ALL_MICCO_REGS
#define	MICCO_ALL_PROC_FILE	"driver/micco_all"
static ssize_t micco_all_proc_read(struct file *filp,
                char *buffer, size_t length, loff_t *offset)
{
        u8 reg_val;
	int micco_reg;

	for (micco_reg = 0; micco_reg < MICCO_REG_NUM; ++micco_reg) {
        	micco_read(micco_reg, &reg_val);
        	printk(KERN_INFO "register 0x%x: 0x%x\n", micco_reg, reg_val);
	}	
        return 0;
}

static struct file_operations micco_all_proc_ops = {
        .read = micco_all_proc_read,
};
#endif

static void create_micco_proc_file(void)
{
	micco_proc_file = create_proc_entry(MICCO_PROC_FILE, 0644, NULL);
	if (micco_proc_file) {
		micco_proc_file->owner = THIS_MODULE;
		micco_proc_file->proc_fops = &micco_proc_ops;
	} else
		printk(KERN_INFO "proc file create failed!\n");
#ifdef DUMP_ALL_MICCO_REGS
        micco_proc_file = create_proc_entry(MICCO_ALL_PROC_FILE, 0444, NULL);
        if (micco_proc_file) {
                micco_proc_file->owner = THIS_MODULE;
                micco_proc_file->proc_fops = &micco_all_proc_ops;
        } else
                printk(KERN_INFO "proc file create failed!\n");
#endif
}

static void remove_micco_proc_file(void)
{
#ifdef DUMP_ALL_MICCO_REGS
	remove_proc_entry(MICCO_ALL_PROC_FILE, NULL);
#endif
	remove_proc_entry(MICCO_PROC_FILE, NULL);
}

#endif

static void micco_adc_init(void)
{
#if 0
	u8 val;

	/* Enable reading temperature from ADC */
	micco_read(MICCO_MISC, &val);
	micco_write(MICCO_MISC, val | MICCO_MISC_I_TBAT_ON);

	/* Enable the TBAT, VCH, ICH and VBAT auto measurements */
	micco_read(MICCO_ADC_AUTO_CONTROL_1, &val);
	micco_write(MICCO_ADC_AUTO_CONTROL_1, val | 0x7F);
	micco_read(MICCO_ADC_AUTO_CONTROL_2, &val);
	micco_write(MICCO_ADC_AUTO_CONTROL_2, val | 0x07);
#endif
}

int micco_adc_aread(u8 mux)
{
	u8 data;

	micco_read(mux, &data);
	return data << 2;
}

int micco_adc_read(u8 mux)
{
	u8 i, val, adc_res[2];
	unsigned short data;

	micco_enable_ldo_adc(1);
	micco_write(MICCO_ADC_MAN_CONTROL, 
		MICCO_ADC_MAN_CONT_CONV | MICCO_ADC_MAN_LDO_ADC_ENABLE | mux);

	for (i = 0; i < 10; i++) {
		udelay(80);
		micco_read(MICCO_ADC_MAN_CONTROL, &val);
		if (val & MICCO_ADC_MAN_CONT_CONV) {
			printk("read adc timeout\n");
			continue;
		}
		micco_read_block(MICCO_MAN_RES_LSB, adc_res, 2);
		data = (adc_res[1] << 2) | (adc_res[0] & 0x03);
		micco_enable_ldo_adc(0);
		return data;
	}
	micco_enable_ldo_adc(0);
	return 0;
}

int get_adc_average(u8 mux)
{
	int i, sum = 0;

	for (i = 0; i < READ_ADC_NUM; i++) 
		sum += micco_adc_read(mux);
	return (sum / READ_ADC_NUM);
}

int get_adc_auto_average(u8 mux)
{
	int i, sum = 0;

	for (i = 0; i < READ_ADC_NUM; i++) 
		sum += micco_adc_aread(mux);
	return (sum / READ_ADC_NUM);
}

static void precharge_check(void)
{
	u8 val;

        micco_read(MICCO_OTP_ADDR_A4, &val);
	if (val & MICCO_USB_PRE_CHARGING_EN)
		printk("Micco: supports USB pre-charging 0x%02x\n", val);
	else
		printk("Micco: not support USB pre-charging 0x%02x\n", val);
}


static int __devinit micco_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct micco_platform_data *pdata;
	int ret;
	int chip_id;
	int i;

	chip_id = i2c_smbus_read_byte_data(client, MICCO_CHIP_ID);
	if (chip_id < 0) {
		printk(KERN_WARNING "micco unavailable!\n");
		micco_d->g_client = NULL;
		return -ENXIO;
	} else {
		printk(KERN_INFO "micco(chip id:0x%02x) detected.\n", chip_id);
	}

	micco_d->g_client = client;

	ret = micco_initchip();
	if (ret != 0)
		printk(KERN_WARNING "Initialize Micco failed with ret 0x%x\n",
				ret);

#if defined(CONFIG_COPYDATA) || defined(CONFIG_PXA930_ACIPC) || defined(CONFIG_BATTERY_BQ27541)
        /* Configure the automatic ADC required for CP operation */
	micco_enable_adc_for_cp();
#endif

	/* create micco work queue thread */
	micco_d->micco_wq = create_singlethread_workqueue("micco-wq");
	if (!micco_d->micco_wq) {
                printk(KERN_WARNING "failed to create work queue.\n");
		return -ESRCH;
	}

	/* init irq */
	pdata = client->dev.platform_data;
	pdata->init_irq();
	ret = request_irq(client->irq, micco_irq_handler, IRQF_TRIGGER_FALLING,
			"Micco", client);
	if (ret) {
		printk(KERN_WARNING "Request IRQ for Micco failed, return:%d\n",
			ret);
		return ret;
	}

#if EXT0_WAKEUP_INTERRUPT
	PECR |= PECR_E0IE;
	PECR &= ~PECR_DIR0;

	ret = request_irq(IRQ_WAKEUP0, ext0_irq_handler, IRQF_DISABLED,
			"Micco-ext0", client);
	if (ret) {
		printk(KERN_WARNING "Request IRQ for ext0 failed, return:%d\n",
				ret);
		return ret;
	}
#endif

	/* init workqueue */
#ifdef CONFIG_WORKQUEUE_PRIO
	INIT_WORK_PRIO(&pdata->work, micco_worker, 98);
#else
	INIT_WORK(&pdata->work, micco_worker);
#endif

	pdata->platform_init();

	for (i = 0; pdata->power_chips[i].chip_id != 0; i++) {
		if (pdata->power_chips[i].chip_id == chip_id) {
			micco_power_module =
				pdata->power_chips[i].power_supply_modules;
			break;
		}
	}
	if (pdata->power_chips[i].chip_id == 0)
	{/* if not found a match version, then use default */
		micco_power_module = pdata->power_chips[0].power_supply_modules;
		printk(KERN_WARNING "This Micco chip is not supported");
	}


#ifdef	CONFIG_PROC_FS
	create_micco_proc_file();
#endif
	micco_adc_init();
	precharge_check();
	pmic_set_ops(&micco_pmic_ops);
	//pmic_sysfs_register(&client->dev);

	return 0;
}

static int micco_remove(struct i2c_client *client)
{
	//pmic_sysfs_unregister();

	pmic_set_ops(NULL);

#ifdef CONFIG_PROC_FS
	remove_micco_proc_file();
#endif
	free_irq(client->irq, NULL);

#if EXT0_WAKEUP_INTERRUPT
	PECR &= ~PECR_E0IE;
	free_irq(IRQ_WAKEUP0, NULL);
#endif
	return 0;
}

static const struct i2c_device_id micco_id[] = {
	{ "micco", 0 },
	{ }
};

static struct i2c_driver micco_driver = {
	.driver = {
		.name	= "micco",
	},
	.probe		= micco_probe,
	.remove		= micco_remove,
	.id_table	= micco_id,
	.suspend	= micco_suspend,
	.resume		= micco_resume,
};

static int __init micco_init(void)
{
	wake_lock_init(&micco_suspend_lock, WAKE_LOCK_SUSPEND, "micco");
	wake_lock_init(&ext0_suspend_lock, WAKE_LOCK_SUSPEND, "micco-ext0");
	register_early_suspend(&micco_android_suspend);
	micco_d = kmalloc(sizeof(struct micco_data), GFP_KERNEL);
        if (!micco_d) {
                return -ENOMEM;
        }

	return i2c_add_driver(&micco_driver);
}

static void __exit micco_exit(void)
{
	unregister_early_suspend(&micco_android_suspend);
	wake_lock_destroy(&ext0_suspend_lock);
	wake_lock_destroy(&micco_suspend_lock);
	flush_scheduled_work();
	i2c_del_driver(&micco_driver);
	kfree(micco_d);
}

subsys_initcall(micco_init);
module_exit(micco_exit);

EXPORT_SYMBOL(micco_adc_read);
EXPORT_SYMBOL(micco_adc_aread);
EXPORT_SYMBOL(get_adc_average);
EXPORT_SYMBOL(get_adc_auto_average);

MODULE_DESCRIPTION("Micco Driver");
MODULE_LICENSE("GPL");
