/*
 * drivers/input/misc/levante_keys.c
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/ctype.h>

#include <mach/pxa27x_keypad.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/levante.h>

static struct input_dev *input=NULL;
static int keyevent_enable = 0;

#ifdef CONFIG_PXA3xx_PMIC
static void onkey_interrupt(unsigned long event)
{
	int status = pxa3xx_pmic_is_onkey_assert(); 
//	printk(KERN_DEBUG "On Key is triggered status = %d\n", status);
	WARN_ON(!(event | PMIC_EVENT_ONKEY));
	input_report_key(input, KEY_POWER, status);
}
#endif

#ifdef CONFIG_USB_CAN_SLEEP_TEST
#include <linux/delay.h>

void dummy_onkey_keyevent()
{
	input_report_key(input,KEY_POWER,1);
	msleep(150);
	input_report_key(input,KEY_POWER,0);
}
EXPORT_SYMBOL(dummy_onkey_keyevent);

#endif

void hookswitch_keyevent(int stat)
{
	printk("hookswitch stat = %d\n", stat);
	if (keyevent_enable)
		input_report_key(input,KEY_HOOKSWITCH,stat);
}
EXPORT_SYMBOL(hookswitch_keyevent);

static int __init pxa3xx_powerkey_init(void)
{
	int error;

        input = input_allocate_device();
        if (!input)
                return -ENOMEM;

	set_bit(EV_KEY, input->evbit);
	set_bit(EV_REL, input->evbit);

        input->name = "power-keys";

        input->id.bustype = BUS_HOST;
        input->id.vendor = 0x0001;
        input->id.product = 0x0001;
        input->id.version = 0x0100;

	error = input_register_device(input);
	if (error) {
		pr_err("power-keys: Unable to register input device, "
			"error: %d\n", error);
		goto fail;
	}

#ifdef CONFIG_PXA3xx_PMIC
	pmic_callback_register(PMIC_EVENT_ONKEY, onkey_interrupt);
#endif
	set_bit(KEY_POWER, input->keybit);//Power button
	set_bit(KEY_HOOKSWITCH, input->keybit);//Headset hookswitch

	keyevent_enable=1;
	return 0;
fail:
	input_free_device(input);
	return error;
}

static void __exit pxa3xx_powerkey_exit(void)
{
#ifdef CONFIG_PXA3xx_PMIC
	pmic_callback_unregister(PMIC_EVENT_ONKEY, onkey_interrupt);
#endif
	keyevent_enable=0;
	input_unregister_device(input);
	input_free_device(input);
	input = NULL;
}

late_initcall(pxa3xx_powerkey_init);
module_exit(pxa3xx_powerkey_exit);
