/*
 * * Copyright (C) 2009, Marvell Corporation(bin.yang@marvell.com).
 * *
 * * Author: Bin Yang <bin.yang@marvell.com> 
 * *
 * * This software program is licensed subject to the GNU General Public License
 * * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html
 * */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/suspend.h>
#include <linux/input.h>
#include <asm/uaccess.h>
#include <mach/gpio.h>

#define BIGSTREAM_TOUCH_REG_NUM		0x1C

static struct i2c_client *g_client;
static struct input_dev *bigstream_ts_input_dev;
struct work_struct	bigstream_ts_work;
struct task_struct	*bigstream_ts_kthread;  
static struct completion	thread_exit_complete;
static int thread_exit;
static int int_gpio;
static int TS_print;

extern int is_android(void);

int bigstream_touch_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;
	ret = i2c_smbus_read_byte_data(g_client, reg);
	if (ret >= 0) {
		*pval = ret;
		status = 0;
	} else {
		status = -EIO;
	}

	return status;
}

int bigstream_touch_write(u8 reg, u8 val)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;
	ret = i2c_smbus_write_byte_data(g_client, reg, val);
	if (ret == 0)
		status = 0;
	else
		status = -EIO;

	return status;
}

static int bigstream_touch_initchip(void)
{
	return 0;
}

static int bigstream_touch_workarround_thread(void *data)
{
	u8 tmp;

	daemonize("bigstream touch");
	while(!thread_exit) {
		bigstream_touch_read(0, &tmp);
		//schedule_work(&bigstream_ts_work);
		msleep(800);
	}
	complete_and_exit(&thread_exit_complete, 0);
	return 0;
}

static void bigstream_touch_work(struct work_struct *work)
{
	struct i2c_client *client = g_client;
	static u16 tem_x = 0xffff;
	static u16 tem_y = 0xffff;
	u8 tmp;
	static int need_pen_up;
	static u16 last_x = 0xffff;
	static u16 last_y = 0xffff;

	disable_irq(client->irq);
__irq_loop:
	bigstream_touch_read(0, &tmp);
	if(tmp == 1) {
		need_pen_up = 1;
		bigstream_touch_read(1, &tmp);
		tem_x = tmp<<8;
		bigstream_touch_read(2, &tmp);
		tem_x += tmp;
		bigstream_touch_read(3, &tmp);
		tem_y = tmp<<8;
		bigstream_touch_read(4, &tmp);
		tem_y += tmp;

		tem_x = tem_x;
		tem_y = tem_y;

		if(last_x != 0xffff) {
			input_report_abs(bigstream_ts_input_dev,
					ABS_X, last_x);
			input_report_abs(bigstream_ts_input_dev,
					ABS_Y, last_y);
			if (!is_android())
				input_report_abs(bigstream_ts_input_dev,
						ABS_PRESSURE, 255);
			if (is_android()) {
				input_report_key(bigstream_ts_input_dev,
						BTN_TOUCH, 1);
				input_sync(bigstream_ts_input_dev);
			}
			if(TS_print) printk("======TSI=> X=%d  Y=%d\n",last_x,last_y);
		}
		last_x = tem_x;
		last_y = tem_y;
	}else if(tmp == 2) {
		bigstream_touch_read(1, &tmp);
		tem_x = tmp;
		bigstream_touch_read(2, &tmp);
		tem_y = tmp;
		bigstream_touch_read(3, &tmp);
		tem_x += tmp<<8;
		bigstream_touch_read(4, &tmp);
		tem_y += tmp<<8;
		bigstream_touch_read(5, &tmp);
		last_x = last_y = 0xffff;
		/*
		   printk(KERN_INFO " Two fingers detected:\n"
		   "x1:%d, y1:%d, x2:%d, y2:%d\n"
		   "distance of two fingers: %d\n"
		   "Zoom in: %d, Zoom out: %d\n",
		   tem_x&0xff, tem_y&0xff, tem_x>>8, tem_y>>8,
		   tmp&0x3f, tmp&0x40, tmp&0x80);
		   */
	}else {
#if 0
		bigstream_touch_read(1, &tmp);
		bigstream_touch_read(2, &tmp);
		bigstream_touch_read(3, &tmp);
		bigstream_touch_read(4, &tmp);
		bigstream_touch_read(5, &tmp);
#endif
		last_x = last_y = 0xffff;
		if(need_pen_up) {
			need_pen_up = 0;
			if (!is_android()) {
				input_report_abs(bigstream_ts_input_dev,
						ABS_PRESSURE, 0);
				input_report_abs(bigstream_ts_input_dev,
						ABS_TOOL_WIDTH, 1);
			}
			if (is_android()) {
				input_report_key(bigstream_ts_input_dev,
						BTN_TOUCH, 0);
				input_sync(bigstream_ts_input_dev);
			}
		}
	}

	if(!gpio_get_value(int_gpio))
		goto __irq_loop;
	enable_irq(client->irq);
}

static irqreturn_t bigstream_touch_irq_handler(int irq, void *dev_id)
{
	schedule_work(&bigstream_ts_work);
	return IRQ_HANDLED;
}

#ifdef	CONFIG_PM
static int bigstream_touch_suspend(struct i2c_client *client, pm_message_t state)
{
	return 0;
}

static int bigstream_touch_resume(struct i2c_client *client)
{
	return 0;
}
#else
#define	bigstream_touch_suspend		NULL
#define	bigstream_touch_resume		NULL
#endif

#ifdef	CONFIG_PROC_FS
#define	BIGSTREAM_TOUCH_PROC_FILE	"driver/bigstream_touch"
static struct proc_dir_entry *bigstream_touch_proc_file;
static int index;

static ssize_t bigstream_touch_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	u8 reg_val;

	if ((index < 0) || (index > BIGSTREAM_TOUCH_REG_NUM))
		return 0;

	bigstream_touch_read(index, &reg_val);
	printk(KERN_INFO "register 0x%x: 0x%x\n", index, reg_val);
	return 0;
}

static ssize_t bigstream_touch_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
	} else if ('?' == messages[0]) {
		printk("  -   tson  tsoff \n");
	} else if (strncmp(messages, "tson", 4) == 0) {
		TS_print = 1;
	} else if (strncmp(messages, "tsoff", 4) == 0) {
		TS_print = 0;
	} else {
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		bigstream_touch_write(index, reg_val & 0xFF);
	}

	return len;
}

static struct file_operations bigstream_touch_proc_ops = {
	.read = bigstream_touch_proc_read,
	.write = bigstream_touch_proc_write,
};

static void create_bigstream_touch_proc_file(void)
{
	bigstream_touch_proc_file = create_proc_entry(BIGSTREAM_TOUCH_PROC_FILE, 0644, NULL);
	if (bigstream_touch_proc_file) {
		bigstream_touch_proc_file->owner = THIS_MODULE;
		bigstream_touch_proc_file->proc_fops = &bigstream_touch_proc_ops;
	} else
		printk(KERN_INFO "proc file create failed!\n");
}

static void remove_bigstream_touch_proc_file(void)
{
	extern struct proc_dir_entry proc_root;
	remove_proc_entry(BIGSTREAM_TOUCH_PROC_FILE, &proc_root);
}

#endif

static int bigstream_touch_open(struct input_dev *idev)
{
	thread_exit = 0;
	kernel_thread(bigstream_touch_workarround_thread,
			NULL, 0);
	return 0;
}

static void bigstream_touch_close(struct input_dev *idev)
{
	init_completion(&thread_exit_complete);
	thread_exit = 1;
	wait_for_completion(&thread_exit_complete);
	return;
}

static int __devinit bigstream_touch_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	int ret;
	int chip_id;

	chip_id = i2c_smbus_read_byte_data(client, 0x7);
	if (chip_id < 0) {
		printk(KERN_WARNING "bigstream_touch unavailable!\n");
		g_client = NULL;
		return -ENXIO;
	} else {
		printk(KERN_INFO "bigstream_touch(chip id:0x%02x) detected.\n", chip_id);
	}

	g_client = client;

	ret = bigstream_touch_initchip();
	if (ret != 0)
		printk(KERN_WARNING "Initialize Portofino failed with ret 0x%x\n",
				ret);

	/* register input device */
	bigstream_ts_input_dev = input_allocate_device();
	if (bigstream_ts_input_dev == NULL) {
		printk(KERN_ERR "%s: failed to allocate input dev\n",
				__FUNCTION__);
		return -ENOMEM;
	}

	bigstream_ts_input_dev->name = "bigstream-ts";

	bigstream_ts_input_dev->phys = "bigstream-ts/input0";
	//	bigstream_ts_input_dev->dev.parent = &pdev->dev;

	bigstream_ts_input_dev->open = bigstream_touch_open;

	bigstream_ts_input_dev->close = bigstream_touch_close;

	__set_bit(EV_ABS, bigstream_ts_input_dev->evbit);
	__set_bit(ABS_X, bigstream_ts_input_dev->absbit);
	__set_bit(ABS_Y, bigstream_ts_input_dev->absbit);
	__set_bit(ABS_PRESSURE, bigstream_ts_input_dev->absbit);

	__set_bit(EV_SYN, bigstream_ts_input_dev->evbit);
	__set_bit(EV_KEY, bigstream_ts_input_dev->evbit);
	__set_bit(BTN_TOUCH, bigstream_ts_input_dev->keybit);

	input_set_abs_params(bigstream_ts_input_dev, ABS_X, 0, 2050, 0, 0);
	input_set_abs_params(bigstream_ts_input_dev, ABS_Y, 0, 2050, 0, 0);
	input_set_abs_params(bigstream_ts_input_dev, ABS_PRESSURE, 0, 255, 0, 0);
	input_set_abs_params(bigstream_ts_input_dev, ABS_TOOL_WIDTH, 0, 15, 0, 0);

	ret = input_register_device(bigstream_ts_input_dev);
	if (ret) {
		printk(KERN_ERR
				"%s: unabled to register input device, ret = %d\n",
				__FUNCTION__, ret);
		return ret;
	}

	INIT_WORK(&bigstream_ts_work, bigstream_touch_work);
	ret = request_irq(client->irq, bigstream_touch_irq_handler, IRQF_DISABLED | IRQF_TRIGGER_FALLING,
			"Bigstream touch", client);

	if (ret) {
		printk(KERN_WARNING "Request IRQ for Bigstream touch failed, return:%d\n",
				ret);
		return ret;
	}

	int_gpio = irq_to_gpio(client->irq);

#ifdef	CONFIG_PROC_FS
	create_bigstream_touch_proc_file();
#endif
	return 0;
}

static int bigstream_touch_remove(struct i2c_client *client)
{
#ifdef	CONFIG_PROC_FS
	remove_bigstream_touch_proc_file();
#endif
	input_unregister_device(bigstream_ts_input_dev);
	return 0;
}

static const struct i2c_device_id bigstream_touch_id[] = {
	{ "bigstream_touch", 0 },
	{ }
};

static struct i2c_driver bigstream_touch_driver = {
	.driver = {
		.name	= "bigstream_touch",
	},
	.id_table 	= bigstream_touch_id,
	.probe		= bigstream_touch_probe,
	.remove		= bigstream_touch_remove,
	.suspend	= bigstream_touch_suspend,
	.resume		= bigstream_touch_resume,
};

static int __init bigstream_touch_init(void)
{
	return i2c_add_driver(&bigstream_touch_driver);
}

static void __exit bigstream_touch_exit(void)
{
	i2c_del_driver(&bigstream_touch_driver);
}

module_init(bigstream_touch_init);
module_exit(bigstream_touch_exit);

MODULE_DESCRIPTION("Bigstream touch Driver");
MODULE_LICENSE("GPL");

