/*
 * drivers/input/touchscreen/t1021.c
 *
 * For multi-touch support, touch device: Synaptics device.
 * kewei.jing@borqs.com
 * 
 */

#include <linux/init.h>
#include <linux/input.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/suspend.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/ctype.h>

#include <asm/ioctl.h>
#include <asm/irq.h>
//#include <asm/hardware.h>
#include <asm/uaccess.h>
#include <asm/mach-types.h>
//#include <asm/arch/pxa-regs.h>
//#include <asm/arch/mfp.h>
//#include <asm/arch/pxa3xx_pmic.h>
//#include <asm/arch/pxa3xx_pm.h>
//#include <asm/arch/pmic_sysfs.h>
#include <mach/t1021.h>

#ifdef CONFIG_PXA3xx_DVFM
//#include <asm/arch/dvfm.h>
//#include <asm/arch/pxa3xx_dvfm.h>
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

#ifdef CONFIG_HAS_EARLYSUSPEND
#include <linux/earlysuspend.h>
#endif

#define SENSITIVITY_DEBUG
//#define REDUCE_REPORTING_ON

//Product related setting
#define LCD_WIDTHX 639
#define LCD_HEIGHTY 359
#define VIRT_TOUCH_WIDTHX 6336
#define VIRT_TOUCH_HEIGHTY 3663

#define TOUCH_ABS_PIXEL_X 6336
#define TOUCH_ABS_PIXEL_Y 3663

#define TOUCH_ABS_PIXEL_Y_TM1227 6336
#define TOUCH_ABS_PIXEL_X_TM1227 3663

// input devie setting
#define DEVICE_PATH "micco-ts/input0"	// should modify the path in released version
#define DEVICE_NAME "cleartouch"
#define FINGER_WIDTH 15

#define CLEARTOUCH_COUNTER 3

#define TOUCH_DATA_LEN 15
static u8 *touch_data = NULL;

enum {
        CLEARPAD_FINGER_UNKOWN = 0,
        CLEARPAD_FINGER_UP = 1,
        CLEARPAD_FINGER_DOWN = 2
};

enum {
	NO_FINGER = 0,
	ONE_FINGER = 1,
	TWO_FINGER = 2,
	OVER3_FINGER = 3	// more than 3 finger touching the cleanpad
};

enum {
	NO_GESTURE = 0,
	ZOOM_GESTURE = 1,
	FLICK_GESTURE = 2,
	PRESS_GESTURE = 3,
	TAP_GESTURE = 4

};

enum touch_lock_status {
	UNLOCKED = 0,
	LOCKED = 1,
};

enum touch_power_mode {
	ACTIVING = 0,
	SLEEPING = 1,
};

struct t1021_device_data {
	int chip_id;
	int usage;
	int locked;
	int power_mode;

	spinlock_t touch_lock;
	struct completion open_tm_completion;
	struct completion close_tm_completion;
	struct task_struct *thread;

	int pos_z;
	int pos_w;
	int pos_x1;
	int pos_y1;
	int finger_state1;
	int pos_x2;
	int pos_y2;
	int finger_state2;
	int finger_state_changed;	// 1: finger state has been changed;  0: No change
	int finger_counter;
	int finger_lock_process;

	int gesture_type;
	int tap_code;

	struct i2c_client *g_client;
	wait_queue_head_t touch_wait_queue;
#ifdef CONFIG_PXA3xx_DVFM
	int dvfm_dev_idx;
	struct dvfm_lock dvfm_lock;
	struct timer_list ui_timer;
	unsigned int ui_timeout_val;
#endif
};

static struct t1021_device_data *touchp_dev;
struct input_dev *touch_input_dev;
static int thread_interrupt_signal = 0;

#ifdef CONFIG_PXA3xx_DVFM
static void set_dvfm_constraint(struct t1021_device_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_disable_op_name("416M", p->dvfm_dev_idx);
	dvfm_disable_op_name("208M", p->dvfm_dev_idx);
	dvfm_disable_op_name("156M", p->dvfm_dev_idx);
	dvfm_disable_op_name("D0CS", p->dvfm_dev_idx);

	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void unset_dvfm_constraint(struct t1021_device_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_enable_op_name("D0CS",p->dvfm_dev_idx);
	dvfm_enable_op_name("156M",p->dvfm_dev_idx);
	dvfm_enable_op_name("208M",p->dvfm_dev_idx);
	dvfm_enable_op_name("416M",p->dvfm_dev_idx);
	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void ui_timer_handler(unsigned long data)
{
        struct t1021_device_data *p = (struct t1021_device_data *)data;
       	unset_dvfm_constraint(p);
}
#endif

/*
 * Read one byte from cleartouch device
 * Return the reg value if success, else return < 0
 *
 */
static int cleartouch_byte_read(u8 reg)
{
	volatile int val;

	if (NULL == (touchp_dev->g_client))
		return -EINVAL;

	val = i2c_smbus_read_byte_data(touchp_dev->g_client, reg);

	return val;
}

static int cleartouch_byte_write(u8 reg, u8 val)
{
	int ret;

	if (NULL == (touchp_dev->g_client))
		return -EINVAL;

	ret = i2c_smbus_write_byte_data(touchp_dev->g_client, reg, val);

	return ret;
}

static int cleartouch_watch_thread(void *tmdata)
{
	int ret = 0;
	int reg_value;
	int z_temp = 0;

	DEFINE_WAIT(wait);
	daemonize("touch_daemon");
	touchp_dev->thread = current;

	thread_interrupt_signal = 0;

	cleartouch_chip_config();

	complete(&touchp_dev->open_tm_completion);
	
	while (1) {

		prepare_to_wait(&(touchp_dev->touch_wait_queue), &wait, TASK_INTERRUPTIBLE);
		if (!thread_interrupt_signal)
			schedule();
		finish_wait(&(touchp_dev->touch_wait_queue), &wait);

		if ( LOCKED == touchp_dev->locked ) {
			if ( 1 == touchp_dev->finger_lock_process ) {
				touchp_dev->finger_lock_process = 0;

				printk(KERN_INFO "will reset touch point...\n");
				printk(KERN_INFO "before lock: x1: %d, y1: %d, finger_st1: %d; x2: %d, y2: %d, finger_st2: %d\n",
					touchp_dev->pos_x1, touchp_dev->pos_y1, touchp_dev->finger_state1,
					touchp_dev->pos_x2, touchp_dev->pos_y2, touchp_dev->finger_state2);

				// reset point when touch is locked.
				if ( CLEARPAD_FINGER_DOWN == touchp_dev->finger_state1 ) {
					touchp_dev->finger_state1 = CLEARPAD_FINGER_UP;
					input_report_abs(touch_input_dev, ABS_X, touchp_dev->pos_x1);
					input_report_abs(touch_input_dev, ABS_Y, touchp_dev->pos_y1);
					input_report_key(touch_input_dev, BTN_TOUCH, (touchp_dev->finger_state1 - 1));
				}

				if ( CLEARPAD_FINGER_UNKOWN != touchp_dev->finger_state2 ) {
					touchp_dev->finger_state2 = CLEARPAD_FINGER_UP;
					input_report_abs(touch_input_dev, ABS_RX, touchp_dev->pos_x2);
					input_report_abs(touch_input_dev, ABS_RY, touchp_dev->pos_y2);
					input_report_key(touch_input_dev, BTN_TOUCH2, (touchp_dev->finger_state2 - 1));
					touchp_dev->finger_state2 = CLEARPAD_FINGER_UNKOWN;
				}
				input_sync(touch_input_dev);

				touchp_dev->finger_state_changed = 0;
				touchp_dev->finger_counter = NO_FINGER;
				printk(KERN_INFO "after lock: x1: %d, y1: %d, finger_st1: %d; x2: %d, y2: %d, finger_st2: %d\n",
					touchp_dev->pos_x1, touchp_dev->pos_y1, touchp_dev->finger_state1,
					touchp_dev->pos_x2, touchp_dev->pos_y2, touchp_dev->finger_state2);

			}
			goto touch_continue_loop;
		}

		if (thread_interrupt_signal > 0) {
			thread_interrupt_signal--;

			ret = i2c_smbus_read_i2c_block_data(touchp_dev->g_client, REG_SMB_DATA_REG0, TOUCH_DATA_LEN, touch_data);
			if ( (ret < 0) || (ret != TOUCH_DATA_LEN) ) {
				goto touch_continue_loop;
			}

			reg_value = touch_data[0];
			if ( (0 == z_temp) && (0 == reg_value) ) {
				goto touch_continue_loop;
			}
			z_temp = reg_value;
			reg_value &= SENSOR_STATE_FINGER;	// low 3-bits.
			if ( SENSOR_STATE_FINGER == reg_value ) {
				goto touch_continue_loop;
			}

			touchp_dev->pos_z = touch_data[1];
			touchp_dev->pos_w = touch_data[0] >> 4;
			touchp_dev->pos_x1 = ((touch_data[2] & 0x1f) << 8) + touch_data[3];
			touchp_dev->pos_y1 = ((touch_data[4] & 0x1f) << 8) + touch_data[5];

			if (reg_value < 2) {
				touchp_dev->pos_x2 = 0;
				touchp_dev->pos_y2 = 0;
			} else {
				touchp_dev->pos_x2 = ((touch_data[8] & 0x1f) << 8) + touch_data[9];
				touchp_dev->pos_y2 = ((touch_data[10] & 0x1f) << 8) + touch_data[11];

			}

			/* Finger state */
			switch (reg_value) {

				case 0:	// No touch
					if (CLEARPAD_FINGER_DOWN == touchp_dev->finger_state1) {
						touchp_dev->finger_state_changed = 1;
					}
					touchp_dev->finger_state1 = CLEARPAD_FINGER_UP;

					if (CLEARPAD_FINGER_DOWN == touchp_dev->finger_state2) {
						touchp_dev->finger_state2 = CLEARPAD_FINGER_UP;
						touchp_dev->finger_state_changed = 1;
					} else {
						touchp_dev->finger_state2 = CLEARPAD_FINGER_UNKOWN;
					}

					touchp_dev->finger_counter = NO_FINGER;
					break;

				case 1:
					if ( (CLEARPAD_FINGER_UP == touchp_dev->finger_state1) || ((reg_value != touchp_dev->finger_counter))){
						touchp_dev->finger_state_changed = 1;
					}
					touchp_dev->finger_state1 = CLEARPAD_FINGER_DOWN;

					if ( 1 < touchp_dev->finger_counter ){
						touchp_dev->finger_state2 = CLEARPAD_FINGER_UP;
						touchp_dev->finger_state_changed = 1;
					} else {
						touchp_dev->finger_state2 = CLEARPAD_FINGER_UNKOWN;
					}

					touchp_dev->finger_counter = reg_value;

					break;
				case 2:
				case 3:
					if ( (CLEARPAD_FINGER_UP == touchp_dev->finger_state1) || ((reg_value != touchp_dev->finger_counter))){
						touchp_dev->finger_state_changed = 1;
					}
					touchp_dev->finger_state1 = CLEARPAD_FINGER_DOWN;

					if (CLEARPAD_FINGER_DOWN != touchp_dev->finger_state2) {
						touchp_dev->finger_state2 = CLEARPAD_FINGER_DOWN;
						touchp_dev->finger_state_changed = 1;
					}

					touchp_dev->finger_counter = reg_value;
					break;

				default:
					break;

			}

			// filter abnormal points
			if ( reg_value > 0x01 ) {
				if ( ((0x0==touchp_dev->pos_y2)&&(0x0==touchp_dev->pos_x2)) || ((TOUCH_ABS_PIXEL_X==touchp_dev->pos_x2)&&(TOUCH_ABS_PIXEL_Y==touchp_dev->pos_y2)) ) {
					goto touch_continue_loop;
				}
			}

			// rotate x, y to meet LCD's reqirement
			if ( !(machine_is_spark())) {
				transform_xy_blaze(&touchp_dev->pos_x1, &touchp_dev->pos_y1);
				transform_xy_blaze(&touchp_dev->pos_x2, &touchp_dev->pos_y2);
			}

#if 0
			printk(KERN_DEBUG "x1: %d, y1: %d, finger_st1: %d; x2: %d, y2: %d, finger_st2: %d, sensor_st: %d, z: 0x%x\n",
				touchp_dev->pos_x1, touchp_dev->pos_y1, touchp_dev->finger_state1,
				touchp_dev->pos_x2, touchp_dev->pos_y2, touchp_dev->finger_state2, reg_value, touch_data[1]);
#endif

			// report point to application
			input_report_abs(touch_input_dev, ABS_X, touchp_dev->pos_x1);
			input_report_abs(touch_input_dev, ABS_Y, touchp_dev->pos_y1);
			input_report_abs(touch_input_dev, ABS_PRESSURE, touchp_dev->pos_z);
			input_report_abs(touch_input_dev, ABS_TOOL_WIDTH, touchp_dev->pos_w);
			input_report_key(touch_input_dev, BTN_TOUCH, (touchp_dev->finger_state1 - 1));
//			input_report_abs(touch_input_dev, ABS_GESTRUE, (int)((pos_y-pos_x))/100);
//			input_report_abs(touch_input_dev, ABS_GESTRUE, gesture_range);

			if (CLEARPAD_FINGER_UNKOWN != touchp_dev->finger_state2) {
				if (CLEARPAD_FINGER_DOWN == touchp_dev->finger_state2) {
					input_report_abs(touch_input_dev, ABS_RX, touchp_dev->pos_x2);
					input_report_abs(touch_input_dev, ABS_RY, touchp_dev->pos_y2);
				}
				input_report_key(touch_input_dev, BTN_TOUCH2, (touchp_dev->finger_state2 - 1));
			}
			input_sync(touch_input_dev);

touch_continue_loop:
			if ( !(touch_data[14] & CHIP_CONFIGURED) ) {
				cleartouch_chip_config();
			}
			schedule_timeout_interruptible(1);

		}

		if (!touchp_dev->thread) {
			break;
		}

	} // end of while(1)

	touchp_dev->thread = NULL;
	complete(&touchp_dev->close_tm_completion);
	return 0;
}

int transform_xy_blaze(int *x, int *y)
{

	if (*x > TOUCH_ABS_PIXEL_X)
		*x = TOUCH_ABS_PIXEL_X;
	if (*y > TOUCH_ABS_PIXEL_Y)
		*y = TOUCH_ABS_PIXEL_Y;

        *x = TOUCH_ABS_PIXEL_X - *x;
        *y = TOUCH_ABS_PIXEL_Y - *y;

        return 0;

}

static irqreturn_t touch_irq_handler(int irq, void *dev_id)
{

	thread_interrupt_signal++;
	if (mod_timer(&touchp_dev->ui_timer, jiffies + touchp_dev->ui_timeout_val))
        { /* timer is already active */}
        else
		set_dvfm_constraint(touchp_dev);

	wake_up_interruptible(&touchp_dev->touch_wait_queue);

	return IRQ_HANDLED;
}

int cleartouch_chip_config(void)
{
	int ret, temp, config_counter;

	config_counter = CLEARTOUCH_COUNTER;
config_start:
	if ( 0 > (ret = cleartouch_byte_read(REG_SMB_DEVICE_STATE)) ) {
		printk("cleartouch: read REG_SMB_DEVICE_STATE failed!\n");
		goto error_config;
	}
	temp = ret;

	if (ret & CHIP_CONFIGURED) {
#ifdef CLEARPAD_DEBUG
		printk(KERN_INFO "cleartouch has been configed, 0x%x\n", ret);
#endif
		return 0;
	}
	
	printk(KERN_INFO "cleartouch: device status: 0x%x\n", ret);

	config_counter--;
	if (config_counter < 0) {
		printk(KERN_INFO "cleartouch: config failed finnally, pls check it...\n");
		WARN_ON(config_counter < 0);
		return -1;
	}
	
	// set the configured bit to YES, 0x47 - default
	temp |= CHIP_CONFIGURED;
	printk(KERN_INFO "configure device status to: 0x%x\n", temp);
	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_DEVICE_STATE, temp)) ) {
		goto error_config;
	}

#ifdef REDUCE_REPORTING_ON
	// Reduce the report times. 0x08 - default
	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_2D_CTRL, 0x08)) ) {
		 printk(KERN_INFO "REG_SMB_2D_CTRL failed!!!\n");
		goto error_config;
	}
#endif

	// Report rate: 0x41 - 40/s;  0x81 - 80/s
	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_DEVICE_CTRL, 0x81)) ) {
		printk(KERN_INFO "REG_SMB_DEVICE_CTRL failed!!!\n");
		goto error_config;
	}

	// Sensitivity adjust, -128 ~ +127
	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_SENSITIVITY, 10)) ) {
		printk(KERN_INFO "REG_SMB_SENSITIVITY failed!\n");
		goto error_config;
	}
	 // enable abs int 
	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_INT_ENABLE, 0x01)) ) {
		printk(KERN_INFO "REG_SMB_INT_ENABLE failed!\n");
		goto error_config;
	}
	

	goto config_start;

error_config:
	return ret;
}

static int cleartouch_chip_init(void)
{
	int ret;

	ret = cleartouch_chip_reset();

	return ret;
}

int cleartouch_chip_reset(void)
{
	int ret;

	ret = cleartouch_byte_write(REG_SMB_RESET, RESET);

	printk(KERN_INFO "reset value: %d\n", ret);
	if (ret) {
		printk(KERN_INFO "cleartouch reset failed!\n");
	}

	return ret;
}

static int cleartouch_open(struct input_dev *input_dev)
{
	int ret = 0;
	unsigned long flags;

	printk(KERN_INFO "cleartouch was opened......................................\n");

	spin_lock_irqsave(&touchp_dev->touch_lock, flags);

	touchp_dev->usage++;
	if (touchp_dev->thread) {
		printk(KERN_INFO "ClearPad: watch thread has been running...\n");
		spin_unlock_irqrestore(&touchp_dev->touch_lock, flags);
		return 0;
	}

	ret = kernel_thread(cleartouch_watch_thread, touchp_dev, 0);
	if(0 > ret) {
		spin_unlock_irqrestore(&touchp_dev->touch_lock, flags);
		printk(KERN_INFO "ClearPad: Create watch thread failed...\n");
		return ret;
	}
	wait_for_completion(&touchp_dev->open_tm_completion);
	spin_unlock_irqrestore(&touchp_dev->touch_lock, flags);
	
	return 0;
}

static void cleartouch_close(struct input_dev *input_dev)
{
	unsigned long flags;

	printk(KERN_INFO "cleartouch was closed......................................\n");
	touchp_dev->usage--;

	spin_lock_irqsave(&touchp_dev->touch_lock, flags);
	if (touchp_dev->thread) {
		touchp_dev->thread = NULL;
		wake_up_interruptible(&touchp_dev->touch_wait_queue);
		wait_for_completion(&touchp_dev->close_tm_completion);
	}
	spin_unlock_irqrestore(&touchp_dev->touch_lock, flags);
}

static ssize_t cleartouch_sensitivity_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	char *after;
	unsigned long state = simple_strtoul(buf, &after, 10);
	size_t count = after - buf;
	s8 sensitivity = 0;

	if (*after && isspace(*after))
		count++;

	if (count == size) {
		printk(KERN_INFO "sensitivity input: %ld\n", state);
		if ( state < 0 || state > 255) {
			printk(KERN_INFO "input range is: 0-255 <-> (-128) - 127 \n");
			return -1;
		}

		sensitivity = (s8) (state - 128);
		printk(KERN_INFO "value written to touch: %d\n", sensitivity);

		if ( 0 > (cleartouch_byte_write(REG_SMB_SENSITIVITY, sensitivity)) ) {
			printk(KERN_ERR "set sensitivity failed!\n");
			return -1;
		}
	}

	return 0;
}

static DEVICE_ATTR(sensitivity, S_IWUSR, NULL, cleartouch_sensitivity_store);

static ssize_t cleartouch_lock_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "lock status: %d\npower mode: %s\n",
		touchp_dev->locked, ((1==touchp_dev->power_mode)?"sleeping":"activing" ));
}

static ssize_t cleartouch_lock_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	char *after;
	unsigned long state = simple_strtoul(buf, &after, 10);
	size_t count = after - buf;

	if (*after && isspace(*after))
		count++;

	if (count == size) {

#if 0
		switch(state) {
		case UNLOCKED:
			if ( 0 > cleartouch_byte_write(REG_SMB_INT_ENABLE, 0x03) ) {
				printk("touch panel unlock error.\n");
				return -1;
			}
			touchp_dev->locked = UNLOCKED;
			break;
		case LOCKED:
			if ( 0 > cleartouch_byte_write(REG_SMB_INT_ENABLE, 0x00) ) {
				printk("touch panel lock error.\n");
				return -1;
			}
			touchp_dev->locked = LOCKED;
			break;
		}
#else
		switch(state) {
		case UNLOCKED:
			touchp_dev->locked = UNLOCKED;
			touchp_dev->finger_lock_process = 0;
			break;
		case LOCKED:
			touchp_dev->locked = LOCKED;
			touchp_dev->finger_lock_process = 1;
			break;
		default:
			printk(KERN_ERR "cleartouch: wrong lock/unlock request from Java...\n");
			break;
		}
#endif

	}
	return 0;
}

static DEVICE_ATTR(lock, S_IRUGO | S_IWUSR, cleartouch_lock_show, cleartouch_lock_store);

#ifdef CONFIG_PXA3xx_DVFM
static ssize_t cleartouch_uitimeout_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", touchp_dev->ui_timeout_val);
}

static ssize_t cleartouch_uitimeout_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned int ui_timeout = simple_strtoul(buf, NULL, 10);
	if( ui_timeout > 0 && ui_timeout <= HZ )
	{
		touchp_dev->ui_timeout_val = ui_timeout;
		return size;
	}else
		printk(KERN_ERR "Error: range must be at (0,HZ)\n");
	
	return -EINVAL;
}

static DEVICE_ATTR(uitimeout, S_IRUGO | S_IWUSR, cleartouch_uitimeout_show, cleartouch_uitimeout_store);
#endif

static int cleartouch_power_mode(int mode)
{
	int power_property, reg;

	power_property = cleartouch_byte_read(REG_SMB_PRODUCT_PROP);

	if(0 > power_property) {
		printk(KERN_ERR "cleartouch: power set failed, cannot read power property\n");
		return -1;
	}

	if(power_property & 0x02) {
		reg = cleartouch_byte_read(REG_SMB_DEVICE_CTRL);
		if(0 > reg) {
			printk(KERN_ERR "cleartouch: power set failed, read error\n");
			return -1;
		}
		reg &= 0xf0;
		reg |= mode;
		if ( 0 > (cleartouch_byte_write(REG_SMB_DEVICE_CTRL, reg)) ) {
			printk(KERN_ERR "cleartouch: power set failed, write error\n");
			return -1;
		}
		printk(KERN_DEBUG "cleartouch: enter %s mode...\n", ((mode>0x01)?"sleep":"active") );
	} else {
		printk(KERN_ERR "cleartouch: doesn't support sleep mode\n");
		return -1;
	}

	return 0;
}

#ifdef CONFIG_HAS_EARLYSUSPEND
void cleartouch_android_suspend_handler(struct early_suspend *h)
{
	int ret;

	printk(KERN_DEBUG "cleartouch suspend...\n");
	touchp_dev->locked = LOCKED;
	touchp_dev->finger_lock_process = 1;

	ret = cleartouch_power_mode(DEEP_SLEEP_MODE);
	if ( 0==ret ) {
		touchp_dev->power_mode = SLEEPING;
	}
}

void cleartouch_android_resume_handler(struct early_suspend *h)
{
	int ret;

	printk(KERN_DEBUG "cleartouch resume...\n");
	ret = cleartouch_power_mode(NORMAL_MODE);
	if ( 0==ret ) {
		touchp_dev->power_mode = ACTIVING;
	}

	if ( 0 > (ret = cleartouch_byte_write(REG_SMB_DEVICE_STATE, 0)) ) {
		printk(KERN_ERR "cleartouch error...\n");
	}
	cleartouch_chip_reset();

	touchp_dev->locked = UNLOCKED;
	touchp_dev->finger_lock_process = 0;
}

static struct early_suspend cleartouch_android_suspend = {
	.level = 5,
	.suspend = cleartouch_android_suspend_handler,
	.resume = cleartouch_android_resume_handler,
};
#endif

static int __devinit t1021_probe(struct i2c_client *client)
{
	struct t1021_platform_data *pdata;
	int ret;

	if (!touch_data) {
		touch_data = kzalloc(TOUCH_DATA_LEN, GFP_KERNEL);
		if (!touch_data) {
			return -ENOMEM;
		}
	}
	touchp_dev->g_client = NULL;
	init_waitqueue_head(&(touchp_dev->touch_wait_queue));
	
	touchp_dev->chip_id = i2c_smbus_read_byte_data(client, REG_SMB_PID);
	if ((touchp_dev->chip_id) < 0) {
		printk(KERN_ERR "touch unavailable!\n");
		return -ENXIO;
	} else {
		printk(KERN_INFO "t1021(chip id: 0x%02x) detected. \n", touchp_dev->chip_id);
	}

	spin_lock_init(&touchp_dev->touch_lock);
	init_completion(&touchp_dev->open_tm_completion);
	init_completion(&touchp_dev->close_tm_completion);
	touchp_dev->thread = NULL;

	touchp_dev->g_client = client;
	touchp_dev->usage = 0;
	touchp_dev->locked = UNLOCKED;
	touchp_dev->power_mode = ACTIVING;
	touchp_dev->finger_state1 = CLEARPAD_FINGER_UP;
	touchp_dev->finger_state2 = CLEARPAD_FINGER_UNKOWN;
	touchp_dev->finger_state_changed = 0;
	touchp_dev->finger_counter = NO_FINGER;
	touchp_dev->finger_lock_process = 0;

	touchp_dev->gesture_type = NO_GESTURE;

	/* init input device */
	touch_input_dev = input_allocate_device();
	if (!touch_input_dev) {
		dev_err(&(touchp_dev->g_client->dev), "fail to allocate the input device!\n");
		return -ENOMEM;
	}

//	touch_input_dev->name = touchp_dev->g_client->name;
	touch_input_dev->name = DEVICE_NAME;
	touch_input_dev->id.bustype = BUS_I2C;
	touch_input_dev->phys = DEVICE_PATH;
	touch_input_dev->open = cleartouch_open;
	touch_input_dev->close = cleartouch_close;

	__set_bit(EV_SYN, touch_input_dev->evbit);
	__set_bit(EV_ABS, touch_input_dev->evbit);
	__set_bit(ABS_X, touch_input_dev->absbit);
	__set_bit(ABS_Y, touch_input_dev->absbit);
	__set_bit(ABS_RX, touch_input_dev->absbit);
	__set_bit(ABS_RY, touch_input_dev->absbit);
	//__set_bit(ABS_GESTRUE, touch_input_dev->absbit);
	__set_bit(ABS_PRESSURE, touch_input_dev->absbit);
	__set_bit(ABS_TOOL_WIDTH, touch_input_dev->absbit);

	// for finger state report
	__set_bit(EV_KEY, touch_input_dev->evbit);
	__set_bit(BTN_TOUCH, touch_input_dev->keybit);
	__set_bit(BTN_TOUCH2, touch_input_dev->keybit);

	input_set_abs_params(touch_input_dev, ABS_X, 0, VIRT_TOUCH_WIDTHX, 0, 0);
	input_set_abs_params(touch_input_dev, ABS_Y, 0, VIRT_TOUCH_HEIGHTY, 0, 0);
	input_set_abs_params(touch_input_dev, ABS_RX, 0, VIRT_TOUCH_WIDTHX, 0, 0);
	input_set_abs_params(touch_input_dev, ABS_RY, 0, VIRT_TOUCH_HEIGHTY, 0, 0);
	//input_set_abs_params(touch_input_dev, ABS_GESTRUE, -128, 127, 0, 0);
	input_set_abs_params(touch_input_dev, ABS_PRESSURE, 0, 255, 0, 0);
	input_set_abs_params(touch_input_dev, ABS_TOOL_WIDTH, 0, FINGER_WIDTH, 0, 0);

	ret = input_register_device(touch_input_dev);
	if (ret < 0) {
		dev_err(&(touchp_dev->g_client->dev), "unable to register input device!\n");
		input_free_device(touch_input_dev);
		return ret;
	}

	dev_set_drvdata(&touch_input_dev->dev, touchp_dev);

	ret = cleartouch_chip_init();
	if (-1 == ret) {
		printk(KERN_ERR "cleartouch init failed!\n");
		goto error_return;
	}

#if defined(CONFIG_PXA3xx_DVFM)
	touchp_dev->dvfm_dev_idx = -1;
        touchp_dev->dvfm_lock.lock = SPIN_LOCK_UNLOCKED;

	if(!dvfm_register(DEVICE_NAME, &touchp_dev->dvfm_dev_idx))
	{
		printk("registered dvfm_dev_idx=%d\n",touchp_dev->dvfm_dev_idx);
		init_timer(&touchp_dev->ui_timer);
		touchp_dev->ui_timer.function = ui_timer_handler;
		touchp_dev->ui_timer.data = (unsigned long)touchp_dev;
		touchp_dev->ui_timeout_val = HZ/2;
		ret = device_create_file( &touch_input_dev->dev, &dev_attr_uitimeout);
		if (ret) {
			printk(KERN_ERR "register ui_timeout interface error.\n");
		}
	}
#endif

	/* init irq: GPIO62*/
	pdata = client->dev.platform_data;
	pdata->init_irq();
	ret = request_irq(client->irq, touch_irq_handler, IRQF_TRIGGER_FALLING, "tp_cleartouch", client);
//	ret = request_irq(client->irq, touch_irq_handler, IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, "tp_cleartouch", client);
	if (ret) {
		printk(KERN_ERR "Request IRQ for cleartouch failed, return:%d\n", ret);
		goto error_return;
	}

#ifdef SENSITIVITY_DEBUG
	ret = device_create_file( &touch_input_dev->dev, &dev_attr_sensitivity);
	if (ret) {
		printk(KERN_ERR "register sensitivity interface error.\n");
		goto error_return;
	}
#endif

	ret = device_create_file( &touch_input_dev->dev, &dev_attr_lock);
	if (ret) {
		printk(KERN_ERR "register lock interface error.\n");
		goto error_return;
	} else {
		printk(KERN_INFO "touchpanel lock creates successfully.\n");
	}

#ifdef CONFIG_HAS_EARLYSUSPEND
	register_early_suspend(&cleartouch_android_suspend);
#endif

	return 0;

error_return:
	input_unregister_device(touch_input_dev);
	return ret;
	
	
}

static int t1021_remove(struct i2c_client *client)
{
#ifdef CONFIG_PXA3xx_DVFM
	if( touchp_dev->dvfm_dev_idx >=0 )
	{
		del_timer(&touchp_dev->ui_timer);
		dvfm_unregister(DEVICE_NAME, &touchp_dev->dvfm_dev_idx);
		touchp_dev->dvfm_dev_idx = -1;
	}
#endif
#ifdef CONFIG_HAS_EARLYSUSPEND
	unregister_early_suspend(&cleartouch_android_suspend);
#endif
	free_irq(client->irq, NULL);
	input_unregister_device(touch_input_dev);
	return 0;
}

static void t1021_shutdown(struct i2c_client *client)
{
	cleartouch_power_mode(DEEP_SLEEP_MODE);
}

struct i2c_driver t1021_touch_driver = {
	.driver = {
		   .name = "t1021_touch",
		   },
	.probe = t1021_probe,
	.remove = t1021_remove,
	.shutdown = t1021_shutdown,
};

static int __init t1021_init(void)
{

	touchp_dev = kmalloc(sizeof(struct t1021_device_data), GFP_KERNEL);
	if (!touchp_dev) {
		printk(KERN_ERR "Cannot malloc the cleartouch memory!\n");
		return -ENOMEM;
	}
	return i2c_add_driver(&t1021_touch_driver);
}

static void __exit t1021_exit(void)
{
	i2c_del_driver(&t1021_touch_driver);
	kfree(touchp_dev);
}

module_init(t1021_init);
module_exit(t1021_exit);

MODULE_DESCRIPTION("Tm1175 touch driver");
MODULE_LICENSE("GPL");
