/*
 * linux/drivers/input/touchscreen/vnc_touch.c
 *
 * touch screen driver for pxa935 - vnc
 *
 * Copyright (C) 2006, Marvell Corporation (fengwei.yin@Marvell.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/freezer.h>
#include <linux/proc_fs.h>

#include <asm/irq.h>
#include <mach/hardware.h>

#include <mach/pxa-regs.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/pxa3xx_pmic.h>
#include <linux/uaccess.h>


enum {
 TSI_PEN_UNKNOW = 0,
 TSI_PEN_DOWN = 1,
 TSI_PEN_UP = 2
};

static u16 tem_x_pre = 0xffff;
static u16 tem_y_pre = 0xffff;


struct vnc_ts_data{
 spinlock_t ts_lock;
 struct task_struct *thread;
 int suspended;
 int pen_state;
 int use_count;
 int step;
};

 struct ts_data {
	int i1;
	int i2;
	int i3;
 };

static struct vnc_ts_data *vnc_td;
static struct input_dev *vnc_ts_input_dev;
static void vnc_ts_handler(u8 pen_state, u16 tem_x, u16 tem_y);

extern int is_android(void);

static int vnc_ts_proc_write(struct file *file, const char __user *buffer,
 unsigned long count, void *data)
{
 struct ts_data ts;

 if (count != sizeof(struct ts_data)) {
	/*printk("size error %d of %d\n", count, sizeof(struct ts_data));*/
	return -EINVAL;
 }

 if (copy_from_user(&ts, buffer, sizeof(struct ts_data)))
	return -EFAULT;



 /*vnc_ts_handler*/
 /*sscanf(kbuf, "%d", &arg);*/
 /*printk("vnc_ts_proc_write=%d:%d:%d\n", ts.i1, ts.i2, ts.i3);*/
 vnc_ts_handler(ts.i1, ts.i2, ts.i3);
 return count;
}

int vnc_ts_read(u16 *x, u16 *y, int *pen_state)
{
 int ret=0;

 /*ret = vnc_tsi_readxy(x, y, *pen_state); idaviden*/
 return ret;
}

/*(xd*xd + yd*yd) < (Diameter*Diameter)*/
static int __attribute__ ((unused)) vnc_evaluate_point(int iX_Pre, int iY_Pre, int iX, int iY)
{
 int tempx, tempy, diff;
 int fRet = 0;

 if ((iX_Pre < 0) || (iX < 0)) /* not valid point*/
 return -EINVAL;

 if (tem_x_pre == 0xffff && tem_y_pre == 0xffff)
	/*the first time no need to evaluate*/
	return 0;


 tempx = abs(iX - iX_Pre);
 tempy = abs(iY - iY_Pre);

 diff = ((tempx ^ 2) + (tempy ^ 2));

 fRet = (((0 < diff) &&  (diff < (200 * 200))) ? 0 : -EINVAL);
 /*if the dis between 2 points are less then 200 but
 * greater the zero ( zero means its the same point which was reported)
 * then it's valid */
 /*printk(LVT_DEBUG "-------->vnc_evaluate_point2 ok - (diff) = %d \n", diff);*/

 /*if (fRet < 0)
	printk(LVT_DEBUG_MUST "-------->!!!vnc_evaluate_point skip!!! - \
	diff [%d] pre[x=%d y=%d] current[x=%d y=%d]\n",
	diff, iX_Pre, iY_Pre, iX, iY);*/

 return fRet;
}

static void vnc_ts_handler(u8 pen_state, u16 tem_x, u16 tem_y)
{
 static int iReportUp;


 if (pen_state) {
	/*printk("-------->vnc_ts_handler - pen down event\n");*/

	/*check current location vs. previous - if too far then measure again */
	/*don't send to user.*/
	/*valid_point =
		vnc_evaluate_point(tem_x_pre, tem_y_pre, tem_x, tem_y);*/

	if (1)/*(!valid_point)*/ {
		iReportUp = 1;
		/*send measure to user here.*/
		input_report_abs(vnc_ts_input_dev,
		ABS_X, tem_x & 0xfff/*only 12b vaild*/);
		input_report_abs(vnc_ts_input_dev,
		ABS_Y, tem_y & 0xfff/*only 12b vaild*/);
		if (!is_android())
			input_report_abs(vnc_ts_input_dev,
			ABS_PRESSURE, 255);
		if (is_android()) {
			input_report_key(vnc_ts_input_dev,
			BTN_TOUCH, 1);
		input_sync(vnc_ts_input_dev);
	}
	/*printk("-------->vnc_ts_handler - report pen down x[%d],y[%d]\n",
		tem_x,tem_y);*/
 } /*else
 printk("-------->vnc_ts_handler - NOT VALID point don't report\n");*/


 tem_x_pre = tem_x;/* update for next points evaluate*/
 tem_y_pre = tem_y;/* update for next points evaluate*/

 } else if (iReportUp == 1) {
	iReportUp = 0;
	/* Report a pen up event */

	if (!is_android()) {
		input_report_abs(vnc_ts_input_dev,
		ABS_PRESSURE, 0);
		input_report_abs(vnc_ts_input_dev,
		ABS_TOOL_WIDTH, 1);
	}
	if (is_android()) {
		input_report_key(vnc_ts_input_dev,
		BTN_TOUCH, 0);
		input_sync(vnc_ts_input_dev);
	}
	/*printk(LVT_DEBUG_MUST "-------->vnc_ts_handler - \
		reported pen up event \n");*/
	vnc_td->pen_state = TSI_PEN_UP;
 }
 return;
}

static int vnc_ts_open(struct input_dev *idev)
{
 unsigned long flags;

 pr_debug("%s: enter", __func__);

 if (vnc_td->suspended) {
	/*printk(LVT_DEBUG_MUST "touch has been suspended!\n");*/
	return -1;
 }

 spin_lock_irqsave(&vnc_td->ts_lock, flags);
 if (vnc_td->use_count++ == 0)
	spin_unlock_irqrestore(&vnc_td->ts_lock, flags);
 else
	spin_unlock_irqrestore(&vnc_td->ts_lock, flags);

 return 0;
}

/* Kill the touchscreen thread and stop the touch digitiser. */
static void vnc_ts_close(struct input_dev *idev)
{
 unsigned long flags;

 pr_debug("%s: enter with use count = %d", __func__,
	vnc_td->use_count);

 spin_lock_irqsave(&vnc_td->ts_lock, flags);
 if (--vnc_td->use_count == 0)
	spin_unlock_irqrestore(&vnc_td->ts_lock, flags);
 else
	spin_unlock_irqrestore(&vnc_td->ts_lock, flags);
}

static int vnc_ts_probe(struct platform_device *pdev)
{
 int ret;
 struct proc_dir_entry *vnc_ts_proc_entry;


 /*printk("------->vnc_ts_probe - start\n");*/

 vnc_td = kzalloc(sizeof(struct vnc_ts_data), GFP_KERNEL);
 if (!vnc_td) {
	ret = -ENOMEM;
 /*printk(LVT_DEBUG_MUST "------->vnc_ts_probe - kzalloc fail \n");*/
 goto vnc_ts_out;
 }
 spin_lock_init(&vnc_td->ts_lock);
 vnc_td->pen_state = TSI_PEN_UP;
 vnc_td->suspended = 0;
 vnc_td->use_count = 0;

 /* register input device */
 vnc_ts_input_dev = input_allocate_device();
 if (vnc_ts_input_dev == NULL) {
	/*printk(LVT_DEBUG_MUST "------->vnc_ts_probe - \
		fail to allocate device\n");*/
	return -ENOMEM;
 }

 vnc_ts_input_dev->name = "vnc-ts";
 vnc_ts_input_dev->phys = "vnc-ts/input0";
 vnc_ts_input_dev->dev.parent = &pdev->dev;

 vnc_ts_input_dev->open = vnc_ts_open;
 vnc_ts_input_dev->close = vnc_ts_close;

 __set_bit(EV_ABS, vnc_ts_input_dev->evbit);
 __set_bit(ABS_X, vnc_ts_input_dev->absbit);
 __set_bit(ABS_Y, vnc_ts_input_dev->absbit);
 __set_bit(ABS_PRESSURE, vnc_ts_input_dev->absbit);

 __set_bit(EV_SYN, vnc_ts_input_dev->evbit);
 __set_bit(EV_KEY, vnc_ts_input_dev->evbit);
 __set_bit(BTN_TOUCH, vnc_ts_input_dev->keybit);

 input_set_abs_params(vnc_ts_input_dev, ABS_X, 0, 240, 0, 0);
 input_set_abs_params(vnc_ts_input_dev, ABS_Y, 0, 320, 0, 0);
 input_set_abs_params(vnc_ts_input_dev, ABS_PRESSURE,
	LEVANTE_PRESSURE_MIN, LEVANTE_PRESSURE_MAX, 0, 0);
 input_set_abs_params(vnc_ts_input_dev, ABS_TOOL_WIDTH,
	LEVANTE_TOOL_WIDTH_MIN, LEVANTE_TOOL_WIDTH_MAX, 0, 0);

 ret = input_register_device(vnc_ts_input_dev);
 if (ret) {
	/*printk(LVT_DEBUG_MUST "------->vnc_ts_probe - \
	fail to register device\n");*/
	return ret;
 }

 vnc_ts_proc_entry = create_proc_entry("driver/vnc_ts", 0, NULL);
 if (vnc_ts_proc_entry)
	vnc_ts_proc_entry->write_proc = vnc_ts_proc_write;


 /*printk(LVT_DEBUG_MUST "------->vnc_ts_probe - end \n");*/

 return 0;

//pmic_cb_out:
 input_unregister_device(vnc_ts_input_dev);
vnc_ts_out:
 kfree(vnc_td);

 /*printk(LVT_DEBUG_MUST "------->vnc_ts_probe - end \n");*/
 return ret;
}

static int vnc_ts_remove(struct platform_device *pdev)
{
 input_unregister_device(vnc_ts_input_dev);

 return 0;
}

#if CONFIG_PM
static int vnc_ts_resume(struct platform_device *pdev)
{
 vnc_td->suspended = 0;
 return 0;
}

static int vnc_ts_suspend(struct platform_device *pdev, pm_message_t state)
{
 vnc_td->suspended = 1;

 return 0;
}
#else
#define vnc_ts_resume NULL
#define vnc_ts_suspend NULL
#endif

static struct platform_driver vnc_ts_driver = {
 .driver = {
 .name = "vnc-ts",
 },
 .probe = vnc_ts_probe,
 .remove = vnc_ts_remove,
 .resume = vnc_ts_resume,
 .suspend = vnc_ts_suspend,
};

static int __init vnc_ts_init(void)
{
 /*printk("vnc_ts_init\n");*/
 return platform_driver_register(&vnc_ts_driver);
}

static void __exit vnc_ts_exit(void)
{
 /*printk("vnc_ts_exit\n");*/
 platform_driver_unregister(&vnc_ts_driver);
}

module_init(vnc_ts_init);
module_exit(vnc_ts_exit);

MODULE_AUTHOR("Israel, Davidenko <israel.davidenko@marvell.com>");
MODULE_DESCRIPTION("vnc touch screen driver");
MODULE_LICENSE("GPL");
