#include "mux_buff.h"
#include <linux/slab.h>
#include <linux/module.h>

static struct kmem_cache *buff_cachep;

struct mux_buff *mux_buff_alloc(gfp_t flags)
{
	struct mux_buff *mbuf;
	if (!buff_cachep)
		return 0;
	if (!(mbuf = kmem_cache_zalloc(buff_cachep, flags))) {
		pr_err("%s: malloc mbuf failed\n", __func__);
		return 0;
	}
	INIT_LIST_HEAD(&mbuf->list);
	return mbuf;
}

void mux_buff_free(struct mux_buff *mbuf)
{
	if (!mbuf)
		return;

	if (buff_cachep)
		kmem_cache_free(buff_cachep, mbuf);
}

EXPORT_SYMBOL(mux_buff_alloc);
EXPORT_SYMBOL(mux_buff_free);

static int __init mux_buff_init(void)
{
	buff_cachep = kmem_cache_create("mux_buff", 
			sizeof(struct mux_buff), 0, 
			SLAB_PANIC, NULL);
	if (!buff_cachep) {
		pr_err("%s: kmem_cache_create() failed\n", __func__);
		return -ENOMEM;
	}

	return 0;
}
core_initcall(mux_buff_init);
