/**********************************************
 * FileName:	inno-i2c.c
 * Description:  
 *		I2C driver for Innofidei IF101 chip.
 * Author:	
 *		Jimmy.li <lizhengming@innofidei.com>
 * Created:	
 *		Oct. 15, 2007
 *
 * Copyright (c) 2007, Innofidei Inc
 *
 ***********************************************/
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/i2c.h>


static int i2c_inited = 0;

static struct i2c_client* g_client = NULL;

static int __init i2c_innodev_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	g_client = client;
	return 0;
}

static int i2c_innodev_remove(struct i2c_client *client)
{
	g_client = NULL;
	return 0;
}

static struct i2c_driver i2c_innodev_driver = {
	.driver = {
 		.name = "inno_i2c",
	},
	.probe = i2c_innodev_probe,
	.remove = i2c_innodev_remove,
};

int inno_i2c_write_byte(u8 command, u8 value)
{
	if(g_client==NULL)
	{
		printk(KERN_ERR "inno_i2c_write_byte: cmd=%x val=%x\n", command, value);
		return -EINVAL;
	}
	return i2c_smbus_write_byte_data(g_client, command, value);
}

int inno_i2c_read_byte(u8 command, u8* value)
{
	int ret;
	if(g_client==NULL)
	{
		printk(KERN_ERR "inno_i2c_read_byte: cmd=%x\n", command);
		return -EINVAL;
	}
	ret = i2c_smbus_read_byte_data(g_client, command);
	if(ret >= 0){
		*value = ret;
	}else
		return -EIO;

	return 0;
}

int inno_i2c_init(u8 slave_addr)
{
	if(g_client==NULL)
	{
		printk(KERN_ERR "inno_i2c_init: i2c is not avaible\n");
		return -EINVAL;
	}
	i2c_inited = 1;
	return 0;
}

int inno_i2c_exit(void)
{
	i2c_inited = 0;
	return 0;
}

/*=========================
 * module init / exit entry
 *========================*/

static int __init inno_i2c_module_init(void)
{
	int ret;

	if ( (ret = i2c_add_driver(&i2c_innodev_driver)) ) {
		printk(KERN_ERR "if101: I2C Driver registration failed,"
			"module not inserted.\n");
		return ret;
	}

	return 0;
}

static void __exit inno_i2c_module_deinit(void)
{
	return i2c_del_driver(&i2c_innodev_driver);
}

module_init(inno_i2c_module_init);
module_exit(inno_i2c_module_deinit);

EXPORT_SYMBOL(inno_i2c_init);
EXPORT_SYMBOL(inno_i2c_write_byte);
EXPORT_SYMBOL(inno_i2c_read_byte);
EXPORT_SYMBOL(inno_i2c_exit);

MODULE_AUTHOR("jimmy.li <lizhengming@innofidei.com>");
MODULE_DESCRIPTION("Innofidei IF101 chip I2C driver");
MODULE_LICENSE("GPL");

