/**********************************************
 * FileName:	inno-i2c.c
 * Description:  
 *		I2C driver for Innofidei IF101 chip.
 * Author:	
 *		Jimmy.li <lizhengming@innofidei.com>
 * Created:	
 *		Oct. 15, 2007
 *
 * Copyright (c) 2007, Innofidei Inc
 *
 ***********************************************/
#include <linux/types.h>
#include "if208.h"

/* Calculating the register bank id, x is 10-bit register address */
#define BANK(x)		(unsigned char)(((x)&0x0300)>>8)	/* register's bank address */
#define OFFSET(x)	(unsigned char)((x)&0x00ff)		/* register's offset to this bank */

/* Update the bank address */
#define SELECT_BANK_CMD		(0xFF)


static unsigned char selected_bank = 0xFF;	//0xff represent not select bank.

/**********************************************
 *
 * these functions implements by users
 *
 * return 0 represent success, otherwize failed.
 *********************************************/
extern int inno_i2c_init(u8 slave_addr);
extern int inno_i2c_write_byte(u8 command, u8 value);
extern int inno_i2c_read_byte(u8 command, u8* value);
extern int inno_i2c_exit(void);

static int i2c_select_bank(u8 bankid)
{
	int ret = 0;
 	
	ret = inno_i2c_write_byte(SELECT_BANK_CMD, (bankid&0x03));

	if(ret != 0){
		ret =  -1;
	}else{
		selected_bank = bankid ;
	}

	return ret;
}

static int i2c_write(u16 reg, u8 data)
{
	unsigned char bank = BANK(reg);
	unsigned char offset = OFFSET(reg);
	
	if((reg & 0x03FF) >  0x0400)
	{
		printk(KERN_ERR "%s : register address error, should [0x0000-0x3fff]", __FUNCTION__);
		return -1;
	}

	if( (bank!=selected_bank) && i2c_select_bank(bank))	//select register bank
		return -1;
	

	return	inno_i2c_write_byte(offset, data);
}

static int i2c_read(u16 reg, u8* data)
{
	unsigned char bank = BANK(reg);
	unsigned char offset = OFFSET(reg);

	if( (bank!=selected_bank) && i2c_select_bank(bank) )	// select register bank
		return -1;
	
	return inno_i2c_read_byte(offset, data);
}

static int i2c_init(void)
{
	selected_bank = 0xFF;

	return inno_i2c_init(INNO_I2C_SLAVE_ADDR);
}

static int i2c_deinit(void)
{
	selected_bank = 0xFF;
	
	return inno_i2c_exit();
}

inno_i2c_driver_t inno_i2c_driver = {
	.init			=	i2c_init,
	.read			=	i2c_read,
	.write			=	i2c_write,
	.exit			=	i2c_deinit,
};
