/************************************************
 * FileName:	inno_lgxchar.c
 *
 * Description:
 *		driver for innofidei if101 chip.
 * Author:		
 *		Jimmy.li <lizhengming@innofidei.com>
 * Created:
 *		Oct.15, 2007
 * History:
 * 		Dec.13, 2007	--	base on if101 v1 driver, modify for IF101 V2 Version by jimmy.li
 *
 * Copyright (c) 2007, Innofidei Inc
 *
 ***********************************************/
 
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <asm/page.h>
#include <linux/poll.h>
#include <linux/kdev_t.h>
#include <linux/device.h>
#include <linux/semaphore.h>
#include <mach/pxa-regs.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include "if208.h"
#include "if101_control_proto.h"

extern struct inno_device g_inno_dev;
extern struct semaphore inno_spi_mutex;
extern struct semaphore inno_iic_mutex;

extern int innodev_startup(inno_device_t* inno_dev);
extern int innodev_shutdown(inno_device_t* inno_dev);
extern int inno_spi_r(unsigned char cmd, unsigned char *rsp);

static struct inno_lgx_ids lgx_ids_table[] = {
	{"LG0", LG0},
	{"LG1", LG1},
	{"UAM", UAM},
	{NULL,}
};

static unsigned int uam_cmd_len = 0;
struct inno_uam_parameter uam_para;
//static unsigned int response = 0;

static int lgx_alloc_buffer(struct inno_lgx *lgx)
{
	int i,j;
	int page_num;
	struct page *page = NULL;
	struct inno_buffer *inno_buf = &lgx->inno_buffer;
	int ret = 0;

	memset(inno_buf, 0, sizeof(struct inno_buffer));
	init_MUTEX(&inno_buf->sem);
	down(&inno_buf->sem);

	// alloc buffer
	page_num = PAGE_ALIGN(INNO_BUFFER_SIZE) / PAGE_SIZE;
	inno_buf->pages = (struct page **)kzalloc(page_num * sizeof(struct page *), GFP_KERNEL);
	if (!inno_buf->pages) {
		inno_err("No enough memory");
		ret = -ENOMEM;
		goto alloc_node_error;
	}

	for (i = 0; i < page_num; i++) {
		page = alloc_page(GFP_KERNEL);

		if (!page) {
			inno_err("No enough page");
			ret = -ENOMEM;
			goto alloc_pages_error;
		}
		//SetPageReserved(page);
		inno_buf->pages[i] = page;
	}

	inno_buf->page_num = page_num;

	inno_buf->bufsize = page_num * PAGE_SIZE;
	inno_buf->vaddr = vmap(inno_buf->pages, page_num, VM_MAP, PAGE_KERNEL);

	/* check if the memory map is OK. */
	if (!inno_buf->vaddr) {
		inno_err("vmap() failure");
		ret = -EFAULT;
		goto vmap_error;
	}

	memset(inno_buf->vaddr, 0, inno_buf->bufsize);

	inno_buf->start = inno_buf->vaddr;
	up(&inno_buf->sem);

	return 0;

vmap_error:
alloc_pages_error:
	for (j = 0; j < i; j++) {
		page = inno_buf->pages[j];
		//ClearPageReserved(page);
		__free_page(page);
	}
	kfree(inno_buf->pages);
alloc_node_error:
	return ret;
}

static void lgx_free_buffer(struct inno_lgx *lgx)
{
	int i;
	struct page *page = NULL;
	struct inno_buffer *inno_buf = &lgx->inno_buffer;

	/*
	 * vunmap will do TLB flush for us.
	 */
	vunmap(inno_buf->vaddr);
	inno_buf->vaddr = NULL;
	
	for (i = 0; i < inno_buf->page_num; i++) {
		page = inno_buf->pages[i];
		//ClearPageReserved(page);
		__free_page(page);
	}

	kfree(inno_buf->pages);

	memset(inno_buf, 0, sizeof(struct inno_buffer));
}

/*====================================
 * character device file operations
 *===================================*/
static int lgx_open(struct inode* inode, struct file* filp)
{
	struct inno_lgx* lgx;
	unsigned int lg_num = iminor(inode);
	int i=0;

	inno_msg("+++++++++++++++++++++++++++++++++++++++++");
	inno_msg("CMMB IF208 open [%s]", lgx_ids_table[lg_num].name);
	inno_msg("+++++++++++++++++++++++++++++++++++++++++");

	if(lg_num >= LG_END){
		inno_err("minor error, should be [%d-%d]", LG0, UAM);
		return -1;
	}

	if(g_inno_dev.lgxs[lg_num]){ 	//already opened, just increase reference count
		down(&g_inno_dev.lgxs[lg_num]->sem);
		g_inno_dev.lgxs[lg_num]->use_count++;
		//g_inno_dev.lgxs[lg_num]->use_count = 1;
		inno_msg("reopen");
		up(&g_inno_dev.lgxs[lg_num]->sem);
		filp->private_data = g_inno_dev.lgxs[lg_num];
		return 0;
	}
	

	/* OK, it's not open, Create it */
	lgx = kmalloc(sizeof(struct inno_lgx), GFP_KERNEL);
	if(!lgx)
		return -ENOMEM;

	memset(lgx, 0, sizeof(struct inno_lgx));
	for(i=0; lgx_ids_table[i].name!=NULL; i++){
		if(lg_num == lgx_ids_table[i].id){
				lgx->ids = &lgx_ids_table[i];
			break;
		}
	}

	if (lgx_alloc_buffer(lgx))
	{
		kfree(lgx);
		return -ENOMEM;
	}

	lgx->use_count = 1;
	init_MUTEX(&lgx->sem);
	init_waitqueue_head(&lgx->read_wait);
	init_waitqueue_head(&lgx->uam_wait);
	g_inno_dev.lgxs[lg_num] = lgx;

	//inno_msg("open file for inno_num: %d\n", lg_num);

	if( innodev_startup(&g_inno_dev) )
	{
		inno_err("error: startup fail, free lgx");
		g_inno_dev.lgxs[lg_num] = NULL;
		lgx_free_buffer(lgx);
		kfree(lgx);
		return -EIO;
	}

	filp->private_data = lgx;
	return 0;
}

static ssize_t lgx_read(struct file* filp, char* buffer, size_t count, loff_t* offset)
{
	DECLARE_WAITQUEUE(wait, current);
	int ret = 0;
	int err = 0;
	struct inno_lgx* lgx = (struct inno_lgx*)filp->private_data;
	struct inno_buffer* inno_buf = &lgx->inno_buffer;

	add_wait_queue(&lgx->read_wait, &wait);
	while(1){
		err = -ERESTARTSYS;	
		if(signal_pending(current))
			break;
		if(inno_buf->valid && inno_buf->own_id==lgx->ids->id)
		{
			down(&inno_buf->sem);
			ret = min(inno_buf->valid_size, count);
			err = copy_to_user(buffer, inno_buf->start, ret);
			inno_buf->valid = 0;
			up(&inno_buf->sem);
			break;
		}
		set_current_state(TASK_INTERRUPTIBLE);
		err = -EAGAIN;
		if(filp->f_flags & O_NONBLOCK)
			break;
		schedule();
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&lgx->read_wait, &wait);

	return ret;
}

static unsigned int lgx_poll(struct file* filp, poll_table *wait)
{
	unsigned int ret = 0;
	struct inno_lgx* lgx = (struct inno_lgx*)filp->private_data;
	struct inno_buffer* inno_buf = &lgx->inno_buffer;

	poll_wait(filp, &lgx->read_wait, wait);
	if(inno_buf->valid && inno_buf->own_id==lgx->ids->id)
		ret = POLLIN | POLLRDNORM;

	return ret;
}
static int lgx_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg)
{
	long retval = 0;
	long val;
	inno_device_t* inno_dev = &g_inno_dev;

	if(_IOC_TYPE(cmd) != INNO_IOC_MAGIC)
		return -ENOTTY;
	
	switch(cmd){
	case INNO_GET_INTR_TYPE:
		val = inno_dev->cfg.intr_type;
		put_user(val, (long*)arg);
		break;
	case INNO_READ_REG:
	{
		struct inno_reg_data* reg_data;
		reg_data = kmalloc(sizeof(struct inno_reg_data), GFP_KERNEL);
		retval = copy_from_user(reg_data, (void*)arg, sizeof(struct inno_reg_data));
		if(retval == 0){
			down(&inno_iic_mutex);
			g_inno_dev.i2c_driver->read(reg_data->reg, (unsigned char*)&reg_data->data);	
			up(&inno_iic_mutex);
			retval = copy_to_user((void*)arg, reg_data, sizeof(struct inno_reg_data));
		}
		kfree(reg_data);
		break;
	}
	case INNO_WRITE_REG:
	{
		struct inno_reg_data* reg_data;
		reg_data = kmalloc(sizeof(struct inno_reg_data), GFP_KERNEL);
		retval = copy_from_user(reg_data, (void*)arg, sizeof(struct inno_reg_data));
		if(retval == 0){
			down(&inno_iic_mutex);
			g_inno_dev.i2c_driver->write(reg_data->reg, reg_data->data);
			up(&inno_iic_mutex);
			retval = copy_to_user((void*)arg, reg_data, sizeof(struct inno_reg_data));
		}
		kfree(reg_data);
		break;
	}
	case INNO_MMIS_READ:	//pass value
	{
		u8 reg  = (u8)arg;
		u8 value;
		down(&inno_spi_mutex);
		g_inno_dev.spi_driver->write(&reg, 1);
		g_inno_dev.spi_driver->read(&value, 1);
		retval = value;
		up(&inno_spi_mutex);
		break;
	}
	case INNO_MMIS_CMD:	//pass value 
	{
		u8 value = (u8)arg;
		down(&inno_spi_mutex);
		g_inno_dev.spi_driver->write(&value,1);
		up(&inno_spi_mutex);
		break;
	}
	case INNO_MMIS_WRITE:	//pass value	mmis_cmd|value
	{
		u8 cmd, value;
		val = arg;
		cmd = (val & 0xFF00)>>8;
		value = (val & 0x00FF);
		down(&inno_spi_mutex);
		g_inno_dev.spi_driver->write(&cmd, 1);
		g_inno_dev.spi_driver->write(&value, 1);
		up(&inno_spi_mutex);
		break;
	}
	case INNO_SCAN_FREQUENCY:
	{
		struct inno_freq_scan_area* 	scan_area;
		scan_area = kmalloc(sizeof(struct inno_freq_scan_area), GFP_KERNEL);
		retval = copy_from_user(scan_area, (void*)arg, sizeof(struct inno_freq_scan_area));

		if(retval == 0)
			retval = inno_scan_frequency(scan_area->start, scan_area->end);
		kfree(scan_area);
		break;
	}
	case INNO_SCAN_FREQUENCY_DOT:
	{
		struct inno_freq_scan_area*		scan_area;
		scan_area = kmalloc(sizeof(struct inno_freq_scan_area), GFP_KERNEL);
		retval = copy_from_user(scan_area, (void*)arg, sizeof(struct inno_freq_scan_area));
		if(retval == 0)
			retval = inno_scan_frequency_dot((u8)scan_area->start, (u8)scan_area->end);
		kfree(scan_area);
		break;
	}
	case INNO_SET_FREQUENCY:	// unit:MHZ
	{
		retval = copy_from_user(&val, (void*)arg, sizeof(unsigned long));
		inno_msg("[inno-ioctl]: INNO_SET_FREQUENCY, val=%ld\n", val);
		
		if(retval == 0)
			retval = inno_set_frequency((u16)val);

		break;
	}
	case INNO_SET_FREQUENCY_DOT:
	{
		struct inno_sys_state  sys_state;
		unsigned char status = 0;

		retval = copy_from_user(&val, (void*)arg, sizeof(unsigned long));
		inno_msg("[inno-ioctl]: INNO_SET_FREQUENCY_DOT, val=%ld\n", val);
		if(retval == 0)
			retval = inno_set_frequency_dot((u8)val);
// For debug 1029 AAAA
msleep(1000);
inno_spi_r(0x04, &status);
//printk("AAAAAAAA: %d%\n", status);

sys_state.stattype = STATTYPE_SYNC_STATE;
retval = inno_get_system_state(&sys_state);
//printk("***********************stattype: 0x%x, statdata: 0x%x\n", sys_state.stattype, sys_state.statdata);
sys_state.stattype = STATTYPE_SIGNAL_STRENGTH;
retval = inno_get_system_state(&sys_state);
//printk("************************** stattype: 0x%x, statdata: 0x%x\n", sys_state.stattype, sys_state.statdata);
sys_state.stattype = STATTYPE_LDPC_ERR_COUNT;
retval = inno_get_system_state(&sys_state);
//printk("************************** stattype: 0x%x, statdata: 0x%x\n", sys_state.stattype, sys_state.statdata);
sys_state.stattype = STATTYPE_LDPC_TOTAL_COUNT;
retval = inno_get_system_state(&sys_state);
//printk("************************** stattype: 0x%x, statdata: 0x%x\n", sys_state.stattype, sys_state.statdata);

		break;
	}
	case INNO_SET_CP_TYPE:
	{
		retval = copy_from_user(&val, (void*)arg, sizeof(unsigned long));
		if(retval == 0)
			retval = inno_set_cp_type((u8)val);
		break;
	}
	case INNO_GET_CHANNEL_CONFIG:
	{
		struct inno_channel_config* 	cfg;
		cfg = kmalloc(sizeof(struct inno_channel_config), GFP_KERNEL);
		inno_msg("[inno-ioctl]: INNO_GET_CHANNEL_CONFIG\n");
		retval = copy_from_user(cfg, (void*)arg, sizeof(struct inno_channel_config));
		
		if(retval == 0)
			retval = inno_get_channel_config(cfg);
		if(!retval)
			retval = copy_to_user((void*)arg, cfg, sizeof(struct inno_channel_config));
		kfree(cfg);
		break;	
	}
	case INNO_SET_CHANNEL_CONFIG:
	{
		struct inno_channel_config* 	cfg;
		cfg = kmalloc(sizeof(struct inno_channel_config), GFP_KERNEL);
		inno_msg("[inno-ioctl]: INNO_SET_CHANNEL_CONFIG\n");
		retval = copy_from_user(cfg, (void*)arg, sizeof(struct inno_channel_config));
		
		if(retval == 0)
			retval = inno_set_channel_config(cfg);
		kfree(cfg);
		break;	
	}
	case INNO_GET_SYS_STATE:
	{
		struct inno_sys_state*	sys_state;
		sys_state = kmalloc(sizeof(struct inno_sys_state), GFP_KERNEL);
		retval = copy_from_user(sys_state, (void*)arg, sizeof(struct inno_sys_state));

		if(retval == 0)
			retval = inno_get_system_state(sys_state);
		if(!retval)
			retval = copy_to_user((void*)arg, sys_state, sizeof(struct inno_sys_state));
		kfree(sys_state);
		break;
	}
	case INNO_SET_PM:
	{
		struct inno_pm_control* pmctl;
		pmctl = kmalloc(sizeof(struct inno_pm_control), GFP_KERNEL);
		retval = copy_from_user(pmctl, (void*)arg, sizeof(struct inno_pm_control));

		if(retval == 0)
			retval = inno_set_pm(pmctl);
		kfree(pmctl);
		break;
	}
	case INNO_GET_FW_VER:
	{	
		struct inno_fw_ver ver = {0, 0};
		inno_get_fireware_version(&ver);
		retval = copy_to_user((void*)arg, &ver, sizeof(struct inno_fw_ver));
		break;
	}
	
	case INNO_FW_DOWNLOAD:
	{
		struct inno_fw_bin* fw_bin = kmalloc(sizeof(struct inno_fw_bin), GFP_KERNEL);
		memset(fw_bin, 0, sizeof(struct inno_fw_bin));
		retval = copy_from_user(fw_bin, (void*)arg, sizeof(struct inno_fw_bin));
		if(retval == 0){
			inno_msg("IOCTL: INNO_FW_DOWNLOAD");
			retval = inno_download_fireware(fw_bin);
		}
		kfree(fw_bin);
		break;
	}
	
	case INNO_UPDATE_FW:
	{
		inno_update_fireware();
		break;
	}
	case INNO_ENABLE_IRQ:
		break;
	case INNO_DISABLE_IRQ:
		break;
	case INNO_SHUTDOWN_IRQ:
	{	//NOTE: cleanup all (i2c, spi, power down) mannually while, NOT the device itself
        if(!g_inno_dev.use_count)
    	{
            inno_msg("shutdown innodev. %d", g_inno_dev.use_count);
    	    innodev_shutdown(&g_inno_dev);
    	    g_inno_dev.use_count = 0; //unrefer all reference counter
    	}
	    break;
	}
	case INNO_GET_BUFFERLEN:
		val = INNO_BUFFER_SIZE;
		put_user(val, (long*)arg);
		break;
	case INNO_UAM_SET_CMD:
		//printk("we are in kernel ioctl^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
		retval = copy_from_user(&uam_para.pBufIn, (void*)arg, uam_cmd_len);
		if(retval == 0){
			inno_dbg("IOCTL: INNO_UAM_TRANSFER");
		}
		uam_para.bufInLen = uam_cmd_len;
		inno_uam_transfer(&uam_para);
		break;

	case INNO_UAM_SET_CMDL:
		retval = copy_from_user((void *)&uam_cmd_len, (void*)arg, sizeof(unsigned int));
		inno_dbg("INNO_UAM_SET_CMDL - uam_cmd_len: %d\n", uam_cmd_len);
		break;

	case INNO_UAM_READ_RES:
		inno_dbg("INNO_UAM_READ_RES - return response len: %d\n", uam_para.pBufOutLen);
		return copy_to_user((void *)arg, uam_para.pBufOut, uam_para.pBufOutLen);

	case INNO_UAM_READ_STATUS:
		inno_dbg("INNO_UAM_READ_STATUS - return status: 0x%x\n", uam_para.sw);
		return put_user(uam_para.sw, (unsigned long *)arg);

	case INNO_UAM_READ_STATUS_LEN:
		inno_dbg("INNO_UAM_READ_STATUS_LEN - return response len: 0x%x\n", uam_para.pBufOutLen);
		return put_user(uam_para.pBufOutLen, (unsigned long *)arg);

	default:
		return -ENOTTY;
	}
	return retval;
}

static int lgx_close(struct inode* inode, struct file* filp)
{
	int lg_num = iminor(inode);
	struct inno_lgx* lgx = g_inno_dev.lgxs[lg_num];

	inno_msg("------------------------------------------");
	inno_msg("CMMB IF208 close [%s], user count[%d]", lgx->ids->name, lgx->use_count);
	inno_msg("------------------------------------------");

	down(&inno_spi_mutex);
	if(!(--lgx->use_count)){
		innodev_shutdown(&g_inno_dev);
		g_inno_dev.lgxs[lg_num] = NULL;
		lgx_free_buffer(lgx);
		kfree(lgx);
	}
	up(&inno_spi_mutex);
	return 0;
}

static struct file_operations lgx_ops = {
	.owner		= THIS_MODULE,
	.open		= lgx_open,
	.read		= lgx_read,
	.poll		= lgx_poll,
	.ioctl		= lgx_ioctl,
	.release	= lgx_close,
};

static struct class *innodev_class;


/*=====================================
 * Module load/unload entry point
 *=====================================*/
extern int innodev_init(void);
extern void innodev_uninit(void);

// module parameter option
static unsigned int major = INNODEV_MAJOR;

static int __init init_lgxchar(void)
{
	int ret;
	
	innodev_init();

	ret = register_chrdev(major, INNODEV_CHRDEV_NAME, &lgx_ops);
	if(ret)
		goto out;

	innodev_class = class_create(THIS_MODULE, "innodev");
	if(!innodev_class)
	{
		unregister_chrdev(major, INNODEV_CHRDEV_NAME);
		innodev_uninit();
		goto out;
	}

	//device_create(innodev_class, NULL, MKDEV(major, 0), NULL, "innodev");
	device_create(innodev_class, NULL, MKDEV(major, 0), NULL, "innodev0");
	device_create(innodev_class, NULL, MKDEV(major, 1), NULL, "innodev1");
	device_create(innodev_class, NULL, MKDEV(major, 2), NULL, "innodev2");

	return 0;
out:
	inno_err("install module failed.");
	return ret;
}

static void __exit cleanup_lgxchar(void)
{
	device_destroy(innodev_class, MKDEV(major, 0));
	device_destroy(innodev_class, MKDEV(major, 1));
	device_destroy(innodev_class, MKDEV(major, 2));

	class_destroy(innodev_class);

	unregister_chrdev(major, INNODEV_CHRDEV_NAME);

	innodev_uninit();
}

module_init(init_lgxchar);
module_exit(cleanup_lgxchar);

module_param(major, uint, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
MODULE_PARM_DESC(major, "if101 character device id");

MODULE_AUTHOR("jimmy.li <lizhengming@innofidei.com>");
MODULE_DESCRIPTION("innofidei IF101 chip driver");
MODULE_LICENSE("GPL");
