/***********************************************
 * FileName:	inno-spi.c
 *
 * Description:  
 *		SPI driver for Innofidei IF101 chip.
 * Author:	
 *		Jimmy.li <lizhengming@innofidei.com>
 * Date:	
 *		Oct. 15, 2007
 * Histroy:
 * 		Jan. 4, 2008,  add inno_spi_is_busy() function.
 * Copyright (c) 2007, Innofidei Inc
 ***********************************************/
#include <linux/types.h> 
#include "if208.h"

/*======================================
 * These functions implements by users
 *
 * return 0 reprements sucess, otherwise failed.
 *=====================================*/
#define INNO_SPI_MASTER		0
#define INNO_SPI_SLAVE		1
extern int inno_spi_init(u8 type);
extern int inno_spi_uninit(void);
extern int inno_spi_write_bytes(u8* data, u32 length);
extern int inno_spi_read_bytes(u8* data, u32 length);
extern int inno_spi_sync_cs(void);
extern int inno_spi_drive_cs(int);

static int spi_init(u8 type)
{
	return inno_spi_init(type);
}

static int spi_uninit(void)
{
	return inno_spi_uninit();
}

static int spi_write(u8* pdata, u32 len)
{
	return inno_spi_write_bytes(pdata, len);	
}

/* note: if101 chip need ncs keep low staus when read until datas read end.
 * if we use spi controller to control nCS, the nCS will pull high after read every byte, 
 * so we should manual control nCS pin
 */
static int spi_read(u8* pdata, u32 len)
{
	return inno_spi_read_bytes(pdata, len);
}

static int spi_sync_cs(void)
{
	return inno_spi_sync_cs();
}

static int spi_drive_cs(int drive)
{
	return inno_spi_drive_cs(drive);
}

inno_spi_driver_t inno_spi_driver = {
	.init		=	spi_init,
	.read		=	spi_read,
	.write		=	spi_write,
	.uninit		=	spi_uninit,
	.sync_cs	=	spi_sync_cs,
	.drive_cs	=	spi_drive_cs
};
