/**********************************************
 * FileName:	mmis_cmdset.c
 * Description:
 *		MMIS COMMANS
 * Author:
 *		Jimmy.li <lizhengming@innofidei.com>
 * Created:
 *		Oct. 15, 2007
 * History:
 * 		Dec. 13, 2007 	---- 	rewrite inno_cmdset struct, and implement the functions by jimmy.li
 * Copyright (c) 2007, Innofidei Inc
 *
 **********************************************/
#include <linux/init.h>
#include <linux/kernel.h>
#include "inno_regs.h"
#include "mmis_cmdlist_v2.h"
#include "if208.h"

extern struct inno_device g_inno_dev;
extern struct semaphore inno_spi_mutex;

/*======================================
 *
 *  SPI MMIS COMMAND SET
 *
 *=====================================*/
static void mmis_send_recv_data_cmd(u8 channelID)
{
	u8 cmd;

	switch(channelID)
	{
		case LG0:
			cmd = FETCH_LG0_DATA;
			break;
		case LG1:
			cmd = FETCH_LG1_DATA;
			break;
		case UAM:
			cmd = FETCH_UAM_DATA;
			break;
		default:
			return;
	}
	g_inno_dev.spi_driver->write(&cmd, 1);
}

static u32 mmis_get_lgx_length(u8 channelID)
{
	u32 length = 0;
	u8  data;
	u8  cmd_h, cmd_m, cmd_l;

	switch(channelID)
	{
		case LG0:
			cmd_h = READ_LG0_LEN_HIGH;
			cmd_m = READ_LG0_LEN_MID;
			cmd_l = READ_LG0_LEN_LOW;
			break;
		case LG1:
			cmd_h = READ_LG1_LEN_HIGH;
			cmd_m = READ_LG1_LEN_MID;
			cmd_l = READ_LG1_LEN_LOW;
			break;
		case UAM:
			cmd_h = READ_UAM_LEN_HIGH;
			cmd_m = READ_UAM_LEN_MID;
			cmd_l = READ_UAM_LEN_LOW;
			break;
		default:
			return 0;
	}

	down(&inno_spi_mutex);
	g_inno_dev.spi_driver->write(&cmd_h, 1);
	g_inno_dev.spi_driver->read(&data, 1);
	length |= (data<<16);

	g_inno_dev.spi_driver->write(&cmd_m, 1);
	g_inno_dev.spi_driver->read(&data, 1);
	length |= (data<<8);

	g_inno_dev.spi_driver->write(&cmd_l, 1);
	g_inno_dev.spi_driver->read(&data, 1);
	length |= data;
	up(&inno_spi_mutex);

	return length;
}

static u8 mmis_get_intr_status(void)
{
	u8 intr = 0;
	u8 cmd = READ_INT_STATUS;

	down(&inno_spi_mutex);
	g_inno_dev.spi_driver->write(&cmd, 1);
	g_inno_dev.spi_driver->read(&intr, 1);
	up(&inno_spi_mutex);
	return intr;
}


static void mmis_fetch_lgx_data(u8 channelID, u8* pbuff, u32 len)
{ 
	down(&inno_spi_mutex);
    switch(g_inno_dev.cfg.trans_size)
    {
    	case TRANS_BURST:
    		mmis_send_recv_data_cmd(channelID);
    		g_inno_dev.spi_driver->read(pbuff, len);
    		break;
    	case TRANS_4BYTES:
    		{
    			u32 times = (len+3)>>2;
    			u8* p = pbuff;
    			int i;
    			for(i=0; i<times; i++){
    				mmis_send_recv_data_cmd(channelID);
    				g_inno_dev.spi_driver->read(p, 4);
    				p += 4;
    			}
    		}
    		break;
    	case TRANS_2BYTES:
    		{
    			u32 times = (len+1)>>1;
    			u8* p = pbuff;
    			int i;
    			for(i=0; i<times; i++){
	    			mmis_send_recv_data_cmd(channelID);
  		  			g_inno_dev.spi_driver->read(p, 2);
    				p += 2;
    			}
    		}
    		break;
    	case TRANS_1BYTE:
    		{
    			u8 *p = pbuff;
    			int i;
    			for(i=0; i<len; i++){
    				mmis_send_recv_data_cmd(channelID);
    				g_inno_dev.spi_driver->read(p, 1);
    				p++;
    			}
    		}
    		break;
    	default:
    		break;
    }
	up(&inno_spi_mutex);
}

struct inno_cmdset inno_cmdset_ops = {
	.get_intr_status			=	mmis_get_intr_status,
	.get_lgx_length				= 	mmis_get_lgx_length,
	.fetch_lgx_data				=	mmis_fetch_lgx_data
};

