/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for MT9V113
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/
#include "camera.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
//#include <asm/arch/mfp.h>
//#include <asm/arch/gpio.h>
//#include <asm/hardware.h>
#include <asm/uaccess.h>
#include <linux/err.h>
#include <linux/proc_fs.h>

#include <linux/clk.h>
#include <linux/i2c.h>

/*****************************************************************************
 * Constants & Registers
 *****************************************************************************/
/* CICLK for QCI */
#define CICLK	1300

/* Revision constants */
#define PID_VERH		0x2280
#define PID_ADRRH            0x0000
#define REG_PLL_CONTROL		0x0014
#define REG_STANDBY_CONTROL		0x0018
#define REG_MCU_ADDRESS		0x098C
#define REG_MCU_DATA		0x0990
#define MCU_REQ_CMD			0xA103
#define PLL_LOCK_STATUS		(0x1 << 15)
#define STANDBY_DONE		(0x1 << 14)

#define AF_SETTING		0
/* Output Size & Format */
#define MT9V113_SIZE_QVGA	1		//320x240
#define MT9V113_SIZE_VGA		2		//640x480
#define MT9V113_SIZE_SVAG	3		//800x600
#define MT9V113_SIZE_D1		4		//864x600
#define MT9V113_SIZE_XGA		5		//1024x768
#define MT9V113_SIZE_SXGA	6		//1280x1024
#define MT9V113_SIZE_UXGA	7		//1600X1200
#define MT9V113_SIZE_QXGA	8		//2048x1536
#define MT9V113_SIZE_QSXGA	9		//2592X1944
#define MT9V113_SIZE_YGA		10		//480X360

#define MT9V113_SIZE_SQCIF	100		//128x96
#define MT9V113_SIZE_QCIF	101		//176x144
#define MT9V113_SIZE_CIF		102		//352x288
#define MT9V113_SIZE_JPEG_VGA	103		//640x480
#define MT9V113_SIZE_256X192	200		//256X192 for 4X zoom
#define MT9V113_SIZE_JPEG_QVGA	201		//320X240
#define MT9V113_SIZE_JPEG_256X192 202		//256X192

#define MT9V113_FORMAT_YUV_422	( 1 << 8)
#define MT9V113_FORMAT_RAW10	( 2 << 8)
#define MT9V113_FORMAT_RAW8	( 3 << 8)
#define MT9V113_FORMAT_JPG	( 4 << 8)
#define MT9V113_FORMAT_RGB565	( 5 << 8)

#define MT9V113_QVGA_WIDTH	320
#define MT9V113_QVGA_HEIGHT	240
#define MT9V113_VGA_WIDTH	640
#define MT9V113_VGA_HEIGHT	480
#define MT9V113_YGA_WIDTH	480
#define MT9V113_YGA_HEIGHT	360
#define MT9V113_SQCIF_WIDTH	128
#define MT9V113_SQCIF_HEIGHT	96
#define MT9V113_QCIF_WIDTH	176
#define MT9V113_QCIF_HEIGHT	144
#define MT9V113_CIF_WIDTH	352
#define MT9V113_CIF_HEIGHT	288
#define MT9V113_XGA_WIDTH	1024
#define MT9V113_XGA_HEIGHT	768
#define MT9V113_UXGA_WIDTH	1600
#define MT9V113_UXGA_HEIGHT	1200
#define MT9V113_QXGA_WIDTH	2048
#define MT9V113_QXGA_HEIGHT	1536

#define MT9V113_REGEND		(0xFFFF)

#define AF_STATE_STEP_STATE_NO		0x80
#define AF_STATE_MASK_MODECHANGE	0x40
#define AF_STATE_MASK_CAPTURECMD	0x20
#define AF_STATE_RESERVED_INVALID	0x10
#define AF_STATE_MODE_IDLE		0x00
#define AF_STATE_MODE_SINGLE		0x04
#define AF_STATE_MODE_CONTINUE		0x08
#define AF_STATE_MODE_STEP		0x0c
#define AF_STATE_MODE_STEP_INSTRUCTION	0x00
#define AF_STATE_MODE_STEP_FOCUSING	0x01
#define AF_STATE_MODE_STEP_FOCUSED	0x02
#define AF_STATE_MODE_STEP_CAPTURE	0x03

#define	AF_STATE_INF		(AF_STATE_MODE_IDLE|AF_STATE_MODE_STEP_INSTRUCTION)
#define AF_STATE_SINGLE		(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSING|AF_STATE_MASK_MODECHANGE|AF_STATE_MASK_CAPTURECMD)
#define AF_STATE_SUCCESS_S	(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSED|AF_STATE_MASK_MODECHANGE)
#define AF_STATE_FAIL_S		(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSED|AF_STATE_MASK_MODECHANGE|AF_STATE_STEP_STATE_NO)
#define AF_STATE_CAPTURE_S	(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_CAPTURE|AF_STATE_MASK_MODECHANGE)

/* Unique ID allocation */
static struct i2c_client *g_client;

#define I2C_DEBUG
//#undef I2C_DEBUG

#ifdef  I2C_DEBUG
#define ovI2C_dbg(fmt, arg...)    printk(KERN_INFO "%s(line %d): " fmt "\n",  \
                __FUNCTION__, __LINE__, ##arg)
#else 
#define ovI2C_dbg(fmt, arg...)        do {} while (0)
#endif

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

/****************************************************************************

   Gloable variable
   
*****************************************************************************/
static int G_mt9v113_mode_change = CAMERA_MODE_VIDEO;
static int g_mt9v113_size = -1;
static int g_mt9v113_format = -1;
#if 0
static u8  mt9v113_gaincode = 0;
static int  mt9v113_expcode = 0;
static u8  mt9v113_recRgain =0,mt9v113_recBgain=0,mt9v113_recGgain=0;
#define QXGAMAXEXPLIN 1568
#define XGAMAXEXPLIN    779
const static u8 gSensorSlaveAddr = 0x3c;
#endif

static int af_flag= 0;
static int reg_format= 0, reg_size= 0;

static int mt9v113set_fps = 0;

static DECLARE_MUTEX(mt9v113_af_lock);

static DECLARE_MUTEX(mt9v113_i2c_lock);

static inline int read_sensor_reg( const u16 subAddress, u16 *bufP );
static inline int write_sensor_reg( const u16 subAddress, u16 bufP );

static unsigned int atoi( char* str)
{
	unsigned int i= 0, radius= 10, val= 0;

	// hex
	if( ( str[ i] == '0') && ( str[ i+ 1] == 'x'))
	{
		radius= 16;
		i+= 2;
	}

	while( str[ i] != '\0')
	{
		val*= radius;

		if( ( str[ i] >= '0') && ( str[ i] <= '9'))
			val+= str[ i]- '0';
		else if( ( str[ i] >= 'A') && ( str[ i] <= 'F'))
			val+= 10+ str[ i]- 'A';
		else if( ( str[ i] >= 'a') && ( str[ i] <= 'f'))
			val+= 10+ str[ i]- 'a';
		else			// not a recognized char
		{
			val/= radius;
			break;
		}
		i++;
	}

	return val;
}

/*****************************************************************************

 *  Register Settings Function

 *****************************************************************************/

/* include AF firmware register tabs */

typedef struct{
	u16	func;
	u16	reg;
	u16	val;
} MT9V113_REG_TABLE;

#if 1//MCLK 13M
static const MT9V113_REG_TABLE mt9v113_init1[] =
{
	{0, 0x0018, 0x4028},  // STANDBY_CONTROL
};

static const MT9V113_REG_TABLE mt9v113_init2[] =
{
	{0, 0x001A, 0x0011},  // RESET_AND_MISC_CONTROL
	{1, 0x0000, 0x000A},  // delay 10ms
	{0, 0x001A, 0x0010},  // RESET_AND_MISC_CONTROL
	{1, 0x0000, 0x000A},  // delay 10ms
	{0, 0x0018, 0x4028},  // STANDBY_CONTROL
};

static const MT9V113_REG_TABLE mt9v113_init3[] =
{
	{0, 0x098C, 0x02F0},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x02F2},  // MCU_ADDRESS
	{0, 0x0990, 0x0210},  // MCU_DATA_0
	{0, 0x098C, 0x02F4},  // MCU_ADDRESS
	{0, 0x0990, 0x001A},  // MCU_DATA_0
	{0, 0x098C, 0x2145},  // MCU_ADDRESS [RESERVED_SEQ_45]
	{0, 0x0990, 0x02F4},  // MCU_DATA_0
	{0, 0x098C, 0xA134},  // MCU_ADDRESS [RESERVED_SEQ_34]
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	//[PLL]
	{0, 0x31E0, 0x0001},  // RESERVED_CORE_31E0
	{0, 0x001A, 0x0210},  // RESET_AND_MISC_CONTROL
	{0, 0x001E, 0x0777},  // PAD_SLEW
	{0, 0x0016, 0x42DF},  // CLOCKS_CONTROL
	{0, 0x001A, 0x0210},
	{0, 0x0014, 0xB04B},  // PLL_CONTROL
	{0, 0x0014, 0xB049},  // PLL_CONTROL
	{0, 0x0010, 0x0010},  // PLL_DIVIDERS
	{0, 0x0012, 0x0000},  // PLL_P_DIVIDERS
	{0, 0x0014, 0x244B},  // PLL_CONTROL
	{1, 0x0000, 0x000A},  // delay 10ms
	{0, 0x0014, 0x304B},  // PLL_CONTROL
};

static const MT9V113_REG_TABLE mt9v113_init4[] =
{
	//[Preview & Capture Init]
	{0, 0x001E, 0x0004},
	{0, 0x0014, 0xB04A},  // PLL_CONTROL
	{0, 0x098C, 0x2703},  // MCU_ADDRESS
	{0, 0x0990, 0x00B0},  // MCU_DATA_0
	{0, 0x098C, 0x2705},  // MCU_ADDRESS
	{0, 0x0990, 0x0090},  // MCU_DATA_0
	{0, 0x098C, 0x2707},  // MCU_ADDRESS
	{0, 0x0990, 0x0280},  // MCU_DATA_0
	{0, 0x098C, 0x2709},  // MCU_ADDRESS
	{0, 0x0990, 0x01E0},  // MCU_DATA_0
	{0, 0x098C, 0x270D},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x270F},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x2711},  // MCU_ADDRESS
	{0, 0x0990, 0x01E7},  // MCU_DATA_0
	{0, 0x098C, 0x2713},  // MCU_ADDRESS
	{0, 0x0990, 0x0287},  // MCU_DATA_0
	{0, 0x098C, 0x2715},  // MCU_ADDRESS
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	{0, 0x098C, 0x2717},  // MCU_ADDRESS
	{0, 0x0990, 0x0026},  // MCU_DATA_0
	{0, 0x098C, 0x2719},  // MCU_ADDRESS
	{0, 0x0990, 0x001A},  // MCU_DATA_0
	{0, 0x098C, 0x271B},  // MCU_ADDRESS
	{0, 0x0990, 0x006B},  // MCU_DATA_0
	{0, 0x098C, 0x271D},  // MCU_ADDRESS
	{0, 0x0990, 0x006B},  // MCU_DATA_0
	{0, 0x098C, 0x271F},  // MCU_ADDRESS
	{0, 0x0990, 0x041E},  // MCU_DATA_0
	{0, 0x098C, 0x2721},  // MCU_ADDRESS
	{0, 0x0990, 0x034A},  // MCU_DATA_0
	{0, 0x098C, 0x2723},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x2725},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x2727},  // MCU_ADDRESS
	{0, 0x0990, 0x01E7},  // MCU_DATA_0
	{0, 0x098C, 0x2729},  // MCU_ADDRESS
	{0, 0x0990, 0x0287},  // MCU_DATA_0
	{0, 0x098C, 0x272B},  // MCU_ADDRESS
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	{0, 0x098C, 0x272D},  // MCU_ADDRESS
	{0, 0x0990, 0x0026},  // MCU_DATA_0
	{0, 0x098C, 0x272F},  // MCU_ADDRESS
	{0, 0x0990, 0x001A},  // MCU_DATA_0
	{0, 0x098C, 0x2731},  // MCU_ADDRESS
	{0, 0x0990, 0x006B},  // MCU_DATA_0
	{0, 0x098C, 0x2733},  // MCU_ADDRESS
	{0, 0x0990, 0x006B},  // MCU_DATA_0
	{0, 0x098C, 0x2735},  // MCU_ADDRESS
	{0, 0x0990, 0x041E},  // MCU_DATA_0
	{0, 0x098C, 0x2737},  // MCU_ADDRESS
	{0, 0x0990, 0x034A},  // MCU_DATA_0
	{0, 0x098C, 0x2739},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x273B},  // MCU_ADDRESS
	{0, 0x0990, 0x027F},  // MCU_DATA_0
	{0, 0x098C, 0x273D},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x273F},  // MCU_ADDRESS
	{0, 0x0990, 0x01DF},  // MCU_DATA_0
	{0, 0x098C, 0x2747},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x2749},  // MCU_ADDRESS
	{0, 0x0990, 0x027F},  // MCU_DATA_0
	{0, 0x098C, 0x274B},  // MCU_ADDRESS
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0x274D},  // MCU_ADDRESS
	{0, 0x0990, 0x01DF},  // MCU_DATA_0
	{0, 0x098C, 0x222D},  // MCU_ADDRESS
	{0, 0x0990, 0x0081},  // MCU_DATA_0
	{0, 0x098C, 0xA408},  // MCU_ADDRESS
	{0, 0x0990, 0x001E},  // MCU_DATA_0
	{0, 0x098C, 0xA409},  // MCU_ADDRESS
	{0, 0x0990, 0x0021},  // MCU_DATA_0
	{0, 0x098C, 0xA40A},  // MCU_ADDRESS
	{0, 0x0990, 0x0025},  // MCU_DATA_0
	{0, 0x098C, 0xA40B},  // MCU_ADDRESS
	{0, 0x0990, 0x0028},  // MCU_DATA_0
	{0, 0x098C, 0x2411},  // MCU_ADDRESS
	{0, 0x0990, 0x0081},  // MCU_DATA_0
	{0, 0x098C, 0x2413},  // MCU_ADDRESS
	{0, 0x0990, 0x009A},  // MCU_DATA_0
	{0, 0x098C, 0x2415},  // MCU_ADDRESS
	{0, 0x0990, 0x0081},  // MCU_DATA_0
	{0, 0x098C, 0x2417},  // MCU_ADDRESS
	{0, 0x0990, 0x009A},  // MCU_DATA_0
	{0, 0x098C, 0xA404},  // MCU_ADDRESS
	{0, 0x0990, 0x0010},  // MCU_DATA_0
	{0, 0x098C, 0xA40D},  // MCU_ADDRESS
	{0, 0x0990, 0x0002},  // MCU_DATA_0
	{0, 0x098C, 0xA40E},  // MCU_ADDRESS
	{0, 0x0990, 0x0003},  // MCU_DATA_0
	{0, 0x098C, 0xA410},  // MCU_ADDRESS
	{0, 0x0990, 0x000A},  // MCU_DATA_0
	{0, 0x098C, 0xA103},  // MCU_ADDRESS
	{0, 0x0990, 0x0006},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_init5[] =
{
	{0, 0x098C, 0xA103},  // MCU_ADDRESS
	{0, 0x0990, 0x0005},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_AWBCCM[] =
{
	//[AWB & CCM]
	{0, 0x098C, 0x2306},  // MCU_ADDRESS [AWB_CCM_L_0]
	{0, 0x0990, 0x0315},  // MCU_DATA_0
	{0, 0x098C, 0x2308},  // MCU_ADDRESS [AWB_CCM_L_1]
	{0, 0x0990, 0xFDDC},  // MCU_DATA_0
	{0, 0x098C, 0x230A},  // MCU_ADDRESS [AWB_CCM_L_2]
	{0, 0x0990, 0x003A},  // MCU_DATA_0
	{0, 0x098C, 0x230C},  // MCU_ADDRESS [AWB_CCM_L_3]
	{0, 0x0990, 0xFF58},  // MCU_DATA_0
	{0, 0x098C, 0x230E},  // MCU_ADDRESS [AWB_CCM_L_4]
	{0, 0x0990, 0x02B7},  // MCU_DATA_0
	{0, 0x098C, 0x2310},  // MCU_ADDRESS [AWB_CCM_L_5]
	{0, 0x0990, 0xFF31},  // MCU_DATA_0
	{0, 0x098C, 0x2312},  // MCU_ADDRESS [AWB_CCM_L_6]
	{0, 0x0990, 0xFF4C},  // MCU_DATA_0
	{0, 0x098C, 0x2314},  // MCU_ADDRESS [AWB_CCM_L_7]
	{0, 0x0990, 0xFE4C},  // MCU_DATA_0
	{0, 0x098C, 0x2316},  // MCU_ADDRESS [AWB_CCM_L_8]
	{0, 0x0990, 0x039E},  // MCU_DATA_0
	{0, 0x098C, 0x2318},  // MCU_ADDRESS [AWB_CCM_L_9]
	{0, 0x0990, 0x001C},  // MCU_DATA_0
	{0, 0x098C, 0x231A},  // MCU_ADDRESS [AWB_CCM_L_10]
	{0, 0x0990, 0x0039},  // MCU_DATA_0
	{0, 0x098C, 0x231C},  // MCU_ADDRESS [AWB_CCM_RL_0]
	{0, 0x0990, 0x007F},  // MCU_DATA_0
	{0, 0x098C, 0x231E},  // MCU_ADDRESS [AWB_CCM_RL_1]
	{0, 0x0990, 0xFF77},  // MCU_DATA_0
	{0, 0x098C, 0x2320},  // MCU_ADDRESS [AWB_CCM_RL_2]
	{0, 0x0990, 0x000A},  // MCU_DATA_0
	{0, 0x098C, 0x2322},  // MCU_ADDRESS [AWB_CCM_RL_3]
	{0, 0x0990, 0x0020},  // MCU_DATA_0
	{0, 0x098C, 0x2324},  // MCU_ADDRESS [AWB_CCM_RL_4]
	{0, 0x0990, 0x001B},  // MCU_DATA_0
	{0, 0x098C, 0x2326},  // MCU_ADDRESS [AWB_CCM_RL_5]
	{0, 0x0990, 0xFFC6},  // MCU_DATA_0
	{0, 0x098C, 0x2328},  // MCU_ADDRESS [AWB_CCM_RL_6]
	{0, 0x0990, 0x0086},  // MCU_DATA_0
	{0, 0x098C, 0x232A},  // MCU_ADDRESS [AWB_CCM_RL_7]
	{0, 0x0990, 0x00B5},  // MCU_DATA_0
	{0, 0x098C, 0x232C},  // MCU_ADDRESS [AWB_CCM_RL_8]
	{0, 0x0990, 0xFEC3},  // MCU_DATA_0
	{0, 0x098C, 0x232E},  // MCU_ADDRESS [AWB_CCM_RL_9]
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	{0, 0x098C, 0x2330},  // MCU_ADDRESS [AWB_CCM_RL_10]
	{0, 0x0990, 0xFFEF},  // MCU_DATA_0
	{0, 0x098C, 0xA348},  // MCU_ADDRESS [AWB_GAIN_BUFFER_SPEED]
	{0, 0x0990, 0x0008},  // MCU_DATA_0
	{0, 0x098C, 0xA349},  // MCU_ADDRESS [AWB_JUMP_DIVISOR]
	{0, 0x0990, 0x0002},  // MCU_DATA_0
	{0, 0x098C, 0xA34A},  // MCU_ADDRESS [AWB_GAIN_MIN]
	{0, 0x0990, 0x0090},  // MCU_DATA_0
	{0, 0x098C, 0xA34B},  // MCU_ADDRESS [AWB_GAIN_MAX]
	{0, 0x0990, 0x00FF},  // MCU_DATA_0
	{0, 0x098C, 0xA34C},  // MCU_ADDRESS [AWB_GAINMIN_B]
	{0, 0x0990, 0x0075},  // MCU_DATA_0
	{0, 0x098C, 0xA34D},  // MCU_ADDRESS [AWB_GAINMAX_B]
	{0, 0x0990, 0x00EF},  // MCU_DATA_0
	{0, 0x098C, 0xA351},  // MCU_ADDRESS [AWB_CCM_POSITION_MIN]
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0xA352},  // MCU_ADDRESS [AWB_CCM_POSITION_MAX]
	{0, 0x0990, 0x007F},  // MCU_DATA_0
	{0, 0x098C, 0xA354},  // MCU_ADDRESS [AWB_SATURATION]
	{0, 0x0990, 0x0043},  // MCU_DATA_0
	{0, 0x098C, 0xA355},  // MCU_ADDRESS [AWB_MODE]
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	{0, 0x098C, 0xA35D},  // MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MIN]
	{0, 0x0990, 0x0078},  // MCU_DATA_0
	{0, 0x098C, 0xA35E},  // MCU_ADDRESS [AWB_STEADY_BGAIN_OUT_MAX]
	{0, 0x0990, 0x0086},  // MCU_DATA_0
	{0, 0x098C, 0xA35F},  // MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MIN]
	{0, 0x0990, 0x007E},  // MCU_DATA_0
	{0, 0x098C, 0xA360},  // MCU_ADDRESS [AWB_STEADY_BGAIN_IN_MAX]
	{0, 0x0990, 0x0082},  // MCU_DATA_0
	{0, 0x098C, 0x2361},  // MCU_ADDRESS [RESERVED_AWB_61]
	{0, 0x0990, 0x0040},  // MCU_DATA_0
	{0, 0x098C, 0xA363},  // MCU_ADDRESS [RESERVED_AWB_63]
	{0, 0x0990, 0x00D2},  // MCU_DATA_0
	{0, 0x098C, 0xA364},  // MCU_ADDRESS [RESERVED_AWB_64]
	{0, 0x0990, 0x00F6},  // MCU_DATA_0
	{0, 0x098C, 0xA302},  // MCU_ADDRESS [AWB_WINDOW_POS]
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0xA303},  // MCU_ADDRESS [AWB_WINDOW_SIZE]
	{0, 0x0990, 0x00EF},  // MCU_DATA_0
	{0, 0x098C, 0xAB20},  // MCU_ADDRESS [HG_LL_SAT1]
	{0, 0x0990, 0x0024},  // MCU_DATA_0
	{0, 0x098C, 0xA103},  // MCU_ADDRESS
	{0, 0x0990, 0x0006},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_AWBCCM2[] =
{
	{0, 0x098C, 0xA103},  // MCU_ADDRESS
	{0, 0x0990, 0x0005},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_lens[] =
{	
	//[lens shading]
	{0, 0x3658, 0x0030},  // P_RD_P0Q0
	{0, 0x365A, 0xB82B},  // P_RD_P0Q1
	{0, 0x365C, 0x42F2},  // P_RD_P0Q2
	{0, 0x365E, 0xD1F0},  // P_RD_P0Q3
	{0, 0x3660, 0xBB33},  // P_RD_P0Q4
	{0, 0x3680, 0x084C},  // P_RD_P1Q0
	{0, 0x3682, 0xA0AC},  // P_RD_P1Q1
	{0, 0x3684, 0xFD6F},  // P_RD_P1Q2
	{0, 0x3686, 0xF031},  // P_RD_P1Q3
	{0, 0x3688, 0xB92E},  // P_RD_P1Q4
	{0, 0x36A8, 0x0893},  // P_RD_P2Q0
	{0, 0x36AA, 0xF792},  // P_RD_P2Q1
	{0, 0x36AC, 0xB255},  // P_RD_P2Q2
	{0, 0x36AE, 0x0D17},  // P_RD_P2Q3
	{0, 0x36B0, 0x65B8},  // P_RD_P2Q4
	{0, 0x36D0, 0x88F0},  // P_RD_P3Q0
	{0, 0x36D2, 0x5D6E},  // P_RD_P3Q1
	{0, 0x36D4, 0xFC53},  // P_RD_P3Q2
	{0, 0x36D6, 0x91D7},  // P_RD_P3Q3
	{0, 0x36D8, 0x9719},  // P_RD_P3Q4
	{0, 0x36F8, 0xF7B3},  // P_RD_P4Q0
	{0, 0x36FA, 0x6FB4},  // P_RD_P4Q1
	{0, 0x36FC, 0xA1D7},  // P_RD_P4Q2
	{0, 0x36FE, 0x163A},  // P_RD_P4Q3
	{0, 0x3700, 0x4FBD},  // P_RD_P4Q4
	{0, 0x364E, 0x0030},  // P_GR_P0Q0
	{0, 0x3650, 0xF32C},  // P_GR_P0Q1
	{0, 0x3652, 0x41F2},  // P_GR_P0Q2
	{0, 0x3654, 0xDCF0},  // P_GR_P0Q3
	{0, 0x3656, 0x8194},  // P_GR_P0Q4
	{0, 0x3676, 0xDDEA},  // P_GR_P1Q0
	{0, 0x3678, 0x050E},  // P_GR_P1Q1
	{0, 0x367A, 0x14B0},  // P_GR_P1Q2
	{0, 0x367C, 0x9073},  // P_GR_P1Q3
	{0, 0x367E, 0x8193},  // P_GR_P1Q4
	{0, 0x369E, 0x04B3},  // P_GR_P2Q0
	{0, 0x36A0, 0x97D3},  // P_GR_P2Q1
	{0, 0x36A2, 0xCFD5},  // P_GR_P2Q2
	{0, 0x36A4, 0x2157},  // P_GR_P2Q3
	{0, 0x36A6, 0x5118},  // P_GR_P2Q4
	{0, 0x36C6, 0xC14E},  // P_GR_P3Q0
	{0, 0x36C8, 0x9C92},  // P_GR_P3Q1
	{0, 0x36CA, 0xB6B1},  // P_GR_P3Q2
	{0, 0x36CC, 0x86D7},  // P_GR_P3Q3
	{0, 0x36CE, 0xBCD9},  // P_GR_P3Q4
	{0, 0x36EE, 0xA454},  // P_GR_P4Q0
	{0, 0x36F0, 0x7CB5},  // P_GR_P4Q1
	{0, 0x36F2, 0xE1D7},  // P_GR_P4Q2
	{0, 0x36F4, 0x2B19},  // P_GR_P4Q3
	{0, 0x36F6, 0x5AFD},  // P_GR_P4Q4
	{0, 0x3662, 0x0030},  // P_BL_P0Q0
	{0, 0x3664, 0x89ED},  // P_BL_P0Q1
	{0, 0x3666, 0x3AB2},  // P_BL_P0Q2
	{0, 0x3668, 0xFFCF},  // P_BL_P0Q3
	{0, 0x366A, 0xEB93},  // P_BL_P0Q4
	{0, 0x368A, 0x654A},  // P_BL_P1Q0
	{0, 0x368C, 0xE6CE},  // P_BL_P1Q1
	{0, 0x368E, 0xB6F0},  // P_BL_P1Q2
	{0, 0x3690, 0x0531},  // P_BL_P1Q3
	{0, 0x3692, 0x6792},  // P_BL_P1Q4
	{0, 0x36B2, 0x6CD2},  // P_BL_P2Q0
	{0, 0x36B4, 0xA7F3},  // P_BL_P2Q1
	{0, 0x36B6, 0x8536},  // P_BL_P2Q2
	{0, 0x36B8, 0x36D7},  // P_BL_P2Q3
	{0, 0x36BA, 0x7978},  // P_BL_P2Q4
	{0, 0x36DA, 0x928C},  // P_BL_P3Q0
	{0, 0x36DC, 0x4392},  // P_BL_P3Q1
	{0, 0x36DE, 0x8FD3},  // P_BL_P3Q2
	{0, 0x36E0, 0x82B8},  // P_BL_P3Q3
	{0, 0x36E2, 0xC1F9},  // P_BL_P3Q4
	{0, 0x3702, 0xBCF3},  // P_BL_P4Q0
	{0, 0x3704, 0x4036},  // P_BL_P4Q1
	{0, 0x3706, 0x8DB6},  // P_BL_P4Q2
	{0, 0x3708, 0x42D7},  // P_BL_P4Q3
	{0, 0x370A, 0x3B9D},  // P_BL_P4Q4
	{0, 0x366C, 0x0150},  // P_GB_P0Q0
	{0, 0x366E, 0xC90C},  // P_GB_P0Q1
	{0, 0x3670, 0x4332},  // P_GB_P0Q2
	{0, 0x3672, 0xBD10},  // P_GB_P0Q3
	{0, 0x3674, 0xF2B3},  // P_GB_P0Q4
	{0, 0x3694, 0x8F2C},  // P_GB_P1Q0
	{0, 0x3696, 0x118E},  // P_GB_P1Q1
	{0, 0x3698, 0x494F},  // P_GB_P1Q2
	{0, 0x369A, 0x9B93},  // P_GB_P1Q3
	{0, 0x369C, 0xA753},  // P_GB_P1Q4
	{0, 0x36BC, 0x02B3},  // P_GB_P2Q0
	{0, 0x36BE, 0x9B13},  // P_GB_P2Q1
	{0, 0x36C0, 0xBDD5},  // P_GB_P2Q2
	{0, 0x36C2, 0x1CD7},  // P_GB_P2Q3
	{0, 0x36C4, 0x2538},  // P_GB_P2Q4
	{0, 0x36E4, 0x98CE},  // P_GB_P3Q0
	{0, 0x36E6, 0xDDD2},  // P_GB_P3Q1
	{0, 0x36E8, 0xD971},  // P_GB_P3Q2
	{0, 0x36EA, 0xEF76},  // P_GB_P3Q3
	{0, 0x36EC, 0xA279},  // P_GB_P3Q4
	{0, 0x370C, 0x8A94},  // P_GB_P4Q0
	{0, 0x370E, 0x06D6},  // P_GB_P4Q1
	{0, 0x3710, 0xF457},  // P_GB_P4Q2
	{0, 0x3712, 0x5DB9},  // P_GB_P4Q3
	{0, 0x3714, 0x6DFD},  // P_GB_P4Q4
	{0, 0x3644, 0x015C},  // POLY_ORIGIN_C
	{0, 0x3642, 0x00F4},  // POLY_ORIGIN_R
	{0, 0x3210, 0x09B8},  // COLOR_PIPELINE_CONTROL
	{0, 0x3644, 0x0160},  // POLY_ORIGIN_C
	{0, 0x3642, 0x00F4},  // POLY_ORIGIN_R
	{0, 0x3644, 0x0160},  // POLY_ORIGIN_C
	{0, 0x3642, 0x00F4},  // POLY_ORIGIN_R
};

static const MT9V113_REG_TABLE mt9v113_saturation[] =
{	
	//[saturation and AP]
	//saturation//
	{0, 0x098C, 0xAB20},  // MCU_ADDRESS [HG_LL_SAT1]
	{0, 0x0990, 0x0060},  // MCU_DATA_0
	{0, 0x098C, 0xAB24},  // MCU_ADDRESS [HG_LL_SAT2]
	{0, 0x0990, 0x0040},  // MCU_DATA_0
	
	
	{0, 0x098C, 0xAB22},  // MCU_ADDRESS [HG_LL_APCORR1]
	{0, 0x0990, 0x0002},  // MCU_DATA_0
	{0, 0x098C, 0xAB26},  // MCU_ADDRESS [HG_LL_APCORR2]
	{0, 0x0990, 0x0001},  // MCU_DATA_0
	{0, 0x098C, 0xAB27},  // MCU_ADDRESS [HG_LL_APTHRESH2]
	{0, 0x0990, 0x0008},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_denoise[] =
{	
	//[denoise]
	
	{0, 0x098C, 0xA20E},  // MCU_ADDRESS [AE_MAX_VIRTGAIN]
	{0, 0x0990, 0x008C},  // MCU_DATA_0
};

static const MT9V113_REG_TABLE mt9v113_AEstable[] =
{	
	//[AE stable]
	{0, 0x098C, 0xAB1F},  // MCU_ADDRESS [HG_LLMODE]
	{0, 0x0990, 0x00C8},  // MCU_DATA_0
	
	{0, 0x098C, 0x2B28},  // MCU_ADDRESS [HG_LL_BRIGHTNESSSTART]
	{0, 0x0990, 0x03E8},  // MCU_DATA_0
	{0, 0x098C, 0x2B2A},  // MCU_ADDRESS [HG_LL_BRIGHTNESSSTOP]
	{0, 0x0990, 0x07D0},  // MCU_DATA_0
	{0, 0x098C, 0x2B38},  // MCU_ADDRESS [HG_GAMMASTARTMORPH]
	{0, 0x0990, 0x03E8},  // MCU_DATA_0
	{0, 0x098C, 0x2B3A},  // MCU_ADDRESS [HG_GAMMASTOPMORPH]
	{0, 0x0990, 0x07D0},  // MCU_DATA_0
	
	
	{0, 0x098C, 0xA244},  // MCU_ADDRESS [AE_DRTFEATURECTRL]
	{0, 0x0990, 0x00B3},  // MCU_DATA_0
	{0, 0x098C, 0xA202},  //// AE_WINDOW_POS [1]	
	{0, 0x0990, 0x0044},	
	{0, 0x098C, 0xA203},  //// AE_WINDOW_SIZE [1]	
	{0, 0x0990, 0x0088},	
	{0, 0x098C, 0xA207},  ////(2) AE_GATE	
	{0, 0x0990, 0x0010},	
	{0, 0x098C, 0xA24C},  ////(1) AE_TARGETBUFFERSPEED	
	{0, 0x0990, 0x0010},  
	{0, 0x098C, 0xA24F},  ////(5) AE_BASETARGET 
	{0, 0x0990, 0x0042},	
	{0, 0x098C, 0xA109},  ////(2) SEQ_AE_FASTBUFF	
	{0, 0x0990, 0x001E},	
	{0, 0x098C, 0xA10A},  ////(2) SEQ_AE_FASTSTEP	
	{0, 0x0990, 0x0002},	
	{0, 0x098C, 0xA20D},  ////(1) AE_MIN_VIRTGAIN	
	{0, 0x0990, 0x0030},	
	{0, 0x098C, 0xA24A},  //// AE_TARGETMIN [1] 
	{0, 0x0990, 0x0028},	
	{0, 0x098C, 0xA24B},  //// AE_TARGETMAX [1] 
	{0, 0x0990, 0x0096},  
};

static const MT9V113_REG_TABLE mt9v113_Capture[] =
{	
	//[Capture] //640*480
	{0, 0x098C, 0xA115},  // MCU_ADDRESS [SEQ_CAP_MODE]
	{0, 0x0990, 0x0002},  // MCU_DATA_0
	{0, 0x098C, 0xA103},  // MCU_ADDRESS [SEQ_CMD]
	{0, 0x0990, 0x0002},  // MCU_DATA_0
};
		
static const MT9V113_REG_TABLE mt9v113_Preview[] =
{
	//[Preview] //320*240
	{0, 0x098C, 0xA115},  // MCU_ADDRESS [SEQ_CAP_MODE]
	{0, 0x0990, 0x0000},  // MCU_DATA_0
	{0, 0x098C, 0xA103},  // MCU_ADDRESS [SEQ_CMD]
	{0, 0x0990, 0x0001},  // MCU_DATA_0
};
#endif

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_on)
		pdata->power_on(pdata->id);
}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_off)
		pdata->power_off(pdata->id);
}

static void sensor_hw_reset(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->reset)
		pdata->reset(pdata->id);
}

static
void mt9v113hw_power_down( u8 powerMode )
{
	/* MT9V113 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();
	msleep(10);
}

static
void mt9v113hw_reset(void)
{
	printk(KERN_ERR"[%s:%d]sensor reset\n", __FILE__, __LINE__);
	write_sensor_reg(0x3012, 0x80);	
	mdelay(5);
	return;
}

static void mt9v113hw_resume_return(void)
{
#if 0
	write_sensor_reg(0x3086, 0x08);
	write_sensor_reg(0x308D, 0x14);
	wite_sensor_reg(0x300E, 0x32);
#endif
}

static
inline void mt9v113_write_regs( const MT9V113_REG_TABLE *tab, const unsigned len)
{
	unsigned i= 0;

	for( i= 0; i< len; i++)
	{
		if (1 == tab[ i].func)
		{
			msleep(tab[ i].val);
		}
		else
		{
			write_sensor_reg( tab[ i].reg, tab[ i].val);
		}
	}
}

static
int mt9v113_set_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != MT9V113_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		status = (write_sensor_reg( regP[curReg << 1], regVal) == 0) ?
			0 : -EIO;
		curReg++;
	}

	return status;
}

static
void mt9v113hw_update_setting(void)
{
	u16 value1=0, value2=0;
	int carry= 0;
	read_sensor_reg( 0x3002, &value1);
	read_sensor_reg( 0x3003, &value2);

	if( value2 > 127)
		carry= 1;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x3002, value1);
	write_sensor_reg( 0x3003, value2);

	read_sensor_reg( 0x302d, &value1);
	read_sensor_reg( 0x302e, &value2);

	if( value2 > 127)
		carry= 1;
	else
		carry= 0;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x302d, value1);
	write_sensor_reg( 0x302e, value2);
}

static
int mt9v113_sensor_reg_init(int format)
{	
//	if( ( format == reg_format) || ( format == reg_size))
//		return 0;

	switch(format)
	{
	case MT9V113_SIZE_QVGA:
		reg_size= format;
		break;
	case MT9V113_SIZE_YGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_YGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_256X192:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_256X192\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_VGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_VGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_SQCIF:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_SQCIF\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_QCIF:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_QCIF\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_CIF:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_CIF\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_QXGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_QXGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_UXGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_UXGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_XGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_XGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_JPEG_VGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_JPEG_VGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_JPEG_QVGA:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_JPEG_QVGA\n", __FILE__, __LINE__);
		reg_size= format;
		break;
	case MT9V113_SIZE_JPEG_256X192:
printk(KERN_ERR"[%s:%d] MT9V113_SIZE_JPEG_256X192\n", __FILE__, __LINE__);
		reg_size= format;
		break;

	case MT9V113_FORMAT_JPG:
printk(KERN_ERR"[%s:%d] MT9V113_FORMAT_JPG\n", __FILE__, __LINE__);
		//mt9v113hw_update_setting();
		reg_format= format;
		break;
	case MT9V113_FORMAT_YUV_422:
printk(KERN_ERR"[%s:%d] MT9V113_FORMAT_YUV_422\n", __FILE__, __LINE__);
		reg_format= format;
		break;

	case AF_SETTING:
		break;
	default:
		printk("No such init table\n");
		return -1;
	}
	return 0;
}

static
int mt9v113_sensorID_detect(void)
{
	u16 val_h= 0;
	read_sensor_reg(PID_ADRRH, &val_h);
	
	printk(KERN_ERR"[%s:%d] val_h= 0x%02X\n", __FILE__, __LINE__, val_h);
		
	if( (val_h == PID_VERH ))
		return 0;
	else
		return -1;
}

static
int mt9v113hw_read_sensor_reg( const u16 subAddress, u16 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

static
int mt9v113hw_write_sensor_reg( const u16 subAddress, u16 *bufP )
{
	return write_sensor_reg(subAddress, *bufP);
}

/*****************************************************************************
 *  Settings
 *****************************************************************************/
static
int mt9v113hw_version_revision(u16 * pCmRevision)
{
	read_sensor_reg( PID_ADRRH, pCmRevision );
	return 0;
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
static int mt9v113_sensor_stop_preview(void)
{
	u16 reg = 0;

	/* Stop AG/AE */
	read_sensor_reg(0x3013,&reg);
	reg &= 0xfa;
	write_sensor_reg(0x3013,reg);

	printk("mt9v113_sensor_stop_preview\n");
	return 0;
}

static int mt9v113_sensor_prepare_snap(void)
{
	mt9v113hw_update_setting();

	printk("mt9v113_sensor_prepare_snap\n");
	return 0;
}

static int mt9v113_sensor_stop_snap(void)
{
	printk("mt9v113_sensor_stop_snap\n");
	return 0;
}

static int mt9v113_sensor_back2preview(void)
{
	//mt9v113_write_regs( ARRAY_AND_SIZE( preview_to_HVGA));

	printk("mt9v113_sensor_back2preview\n");
	return 0;
}

/*****************************************************************************
 *  Format
 *****************************************************************************/

static int mt9v113hw_set_framerate(int fps)
{
	printk(KERN_INFO "mt9v113: set video default fps.\n");
	
	return 0;
}

static
int mt9v113hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{
	if(G_mt9v113_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			mt9v113_sensor_stop_preview();
		else
			mt9v113_sensor_stop_snap();
	}

#if 1
	mt9v113_sensor_reg_init( colorFormat);
	mt9v113_sensor_reg_init( captureSizeFormat);
#else
	mt9v113_sensor_reg_init( captureSizeFormat);
	mt9v113_sensor_reg_init( colorFormat);
#endif

	if( colorFormat == MT9V113_FORMAT_YUV_422)
	{
		//printk("wait focus-----------\n");
		//mt9v113_sensor_reg_init( AF_SETTING);//af function
		//printk("wait focus done-----------\n");
		mt9v113hw_set_framerate(mt9v113set_fps);
	}
/*
	{
	u8 value;
	int addr;
	addr= 0x3000;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	}
*/
	if(G_mt9v113_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			mt9v113_sensor_prepare_snap();
		else
			mt9v113_sensor_back2preview();
		G_mt9v113_mode_change = mode;
	}

	return 0;
}

/*
int mode: sensor output mode, MODE_STILL for QXGA, MODE_VIDEO for XGA;
int iZoomLevel: zoom level, 100-1X, 110-1.1X, ... etc;
unsigned int iWidthOut: output width of the sensor;
unsigned int iHeightOut: output height of the sensor;
*/
static int mt9v113hw_set_resolution_zoom(u32 iZoomLevel)
{
	return 0;
}

static
int mt9v113hw_set_gamma( u32 value)
{
	return 0;
}

//cyj, need modify
static
int mt9v113hw_set_brightness( u32 value)
{
	return 0;
}

static
int mt9v113hw_set_contrast( u32 value)
{
	return 0;
}

static
int mt9v113hw_set_exposure( u32 value)
{
	return 0;
}

static
int mt9v113_poll_MCU_register(u16 reg, u16 mask, u16 value, int times)
{
	int ret = 0;
	u16 sensor_value;

	while(times > 0)
	{
		ret = read_sensor_reg(reg, &sensor_value);
		//printk("polling reg = 0x%x value = 0x%x\n", reg, sensor_value);
		if (value == (sensor_value & mask))
			break;
		msleep(10);
		times--;
	}
	
	if (0 == times)
	{
		printk(KERN_ERR"[mt9v113_poll_MCU_register] failed to poll reg: 0x%x\n", reg);
	}
	return ret;
}

static
int mt9v113_poll_sensor_register(u16 reg, u16 mask, u16 value, int times)
{
	int ret = 0;
	u16 sensor_value;

	while(times > 0)
	{
		ret = write_sensor_reg(0x098C, reg);
		ret = read_sensor_reg(0x0990, &sensor_value);
		//printk("polling reg = 0x%x value = 0x%x\n", reg, sensor_value);
		if (value == (sensor_value & mask))
			break;
		msleep(10);
		times--;
	}
	
	if (0 == times)
	{
		printk(KERN_ERR"[mt9v113_poll_sensor_register] failed to poll reg: 0x%x\n", reg);
	}
	return ret;
}

static int mt9v113hw_sensor_init(void)
{
	int ret = 0;
	G_mt9v113_mode_change = CAMERA_MODE_VIDEO;
	g_mt9v113_size = -1;
	g_mt9v113_format = -1;

	mt9v113set_fps = 0;

	printk(KERN_INFO "mt9v113: sensor init\n");
	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_init1));
	//  POLL  STANDBY_CONTROL::STANDBY_DONE =>  0x01, ..., 0x0
	mt9v113_poll_MCU_register(0x0018, 0x4000, 0, 20);
	//  POLL  RESERVED_CORE_301A::[0x0004] =>  0x01
	mt9v113_poll_MCU_register(0x301A, 0x0004, 0x0004, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_init2));
	//  POLL  STANDBY_CONTROL::STANDBY_DONE =>  0x01, ..., 0x0
	mt9v113_poll_MCU_register(0x0018, 0x4000, 0, 20);
	//  POLL  RESERVED_CORE_301A::[0x0004] =>  0x01
	mt9v113_poll_MCU_register(0x301A, 0x0004, 0x0004, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_init3));
	//  POLL  PLL_CONTROL::PLL_LOCK_STATUS =>  0x01
	mt9v113_poll_MCU_register(0x0014, 0x8000, 0x8000, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_init4));
	//  POLL  SEQ_CMD =>  0x06, ..., 0x00
	mt9v113_poll_sensor_register(0xA103, 0xFFFF, 0, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_init5));
	//  POLL  SEQ_CMD =>  0x05, ..., 0x00
	mt9v113_poll_sensor_register(0xA103, 0xFFFF, 0, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_AWBCCM));
	//  POLL  SEQ_CMD =>  0x06, ..., 0x00
	mt9v113_poll_sensor_register(0xA103, 0xFFFF, 0, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_AWBCCM2));
	//  POLL  SEQ_CMD =>  0x05, ..., 0x00
	mt9v113_poll_sensor_register(0xA103, 0xFFFF, 0, 20);

	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_lens));
	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_saturation));
	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_denoise));
	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_AEstable));
	mt9v113_write_regs( ARRAY_AND_SIZE( mt9v113_Preview));
	return 0;
}

/*****************************************************************************
 * MT9V113 Functions
 *****************************************************************************/
static
int mt9v113_init( p_camera_context_t camera_context )
{
	u16 cm_pid;
	u32 timeout;
	int status;
	
	printk("[MT9V113] mt9v113_init in\n");

	printk(KERN_ERR"mt9v113_init\n");

	af_flag= reg_format= reg_size= 0;

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE | /* SENSOR_CAP_MANUAL_BRIGHTNESS | */
		SENSOR_CAP_MANUAL_EXPOSURE | SENSOR_CAP_MANUAL_ZOOM;

	/* Configure CI according to MT9V113's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);
	
	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 48MHZ
	 */
	/* software work around*/
	ci_set_clock(1, 1, CICLK);

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* MT9V113 Power on sequence
	 * Take out of Power down mode (GPIO_176), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	mt9v113hw_power_down(CAMERA_POWER_FULL);
	mt9v113hw_resume_return();
	//mt9v113hw_reset();

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = 0;
		status = mt9v113hw_version_revision(&cm_pid);

		/* Check to make sure we are working with an MT9V113 */
		if( cm_pid != PID_VERH  )
		{
			printk("mt9v113: pid(%x) is not correct, try again!\n", cm_pid);
			mt9v113hw_power_down( CAMERA_POWER_OFF );
			mt9v113hw_power_down( CAMERA_POWER_FULL );
			sensor_hw_reset();
			mdelay(1);
		}
		else
		{
			printk("mt9v113: pid(%x) is correct!\n", cm_pid);			
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_VERH );

	mt9v113hw_sensor_init();

	return 0;
}

static
int mt9v113_deinit(p_camera_context_t camera_context)
{
	/* power off the external module */
	mt9v113hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

static
int mt9v113_sleep(  p_camera_context_t camera_context )
{
	return mt9v113_deinit( camera_context );
}

static
int mt9v113_wake(  p_camera_context_t camera_context )
{
	return mt9v113_init( camera_context );
}

static
int mt9v113_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 mtSizeFormat, mtFormat;
	int capture_width, capture_height;
	int zoom, zoom_width, zoom_height;

	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
//			printk(KERN_ERR"[%s:%d] output format yuv422\n", __FILE__, __LINE__);
			mtFormat = MT9V113_FORMAT_YUV_422;
			break;
			
		case V4L2_PIX_FMT_SRGGB10:
//			printk(KERN_ERR"[%s:%d] output format raw10\n", __FILE__, __LINE__);
			mtFormat = MT9V113_FORMAT_RAW10;		
			break;
			
		case V4L2_PIX_FMT_SRGGB8:
//			printk(KERN_ERR"[%s:%d] output format raw8\n", __FILE__, __LINE__);
			mtFormat = MT9V113_FORMAT_RAW8;		
			break;

			
		case V4L2_PIX_FMT_JPEG:
//			printk(KERN_ERR"[%s:%d] output format jpeg\n", __FILE__, __LINE__);
			mtFormat = MT9V113_FORMAT_JPG;
			break;
		
		case V4L2_PIX_FMT_RGB565X:
//			printk(KERN_ERR"[%s:%d] output format rgb565\n", __FILE__, __LINE__);
			 mtFormat = MT9V113_FORMAT_RGB565;
		
			 break;
		default:
//			printk(KERN_ERR"[%s:%d] output format default yuv422\n", __FILE__, __LINE__);
			mtFormat = MT9V113_FORMAT_YUV_422;
			break;
	}

	capture_width	 = camera_context->capture_input_width;
	capture_height = camera_context->capture_input_height;

	zoom = camera_context->sensor_status.digital_zoom_value;
	if(zoom<100)
		zoom = 100;

	printk(KERN_INFO "mt9v113: capture %d:%d zoom %d\n", capture_width, capture_height, zoom);

	if( mtFormat == MT9V113_FORMAT_JPG )
	{	/* JPEG */
		zoom_width = MT9V113_QXGA_WIDTH*100/zoom;
		zoom_height = MT9V113_QXGA_HEIGHT*100/zoom;

		if(capture_width>zoom_width)
			capture_width = zoom_width;
		if(capture_height>zoom_height)
			capture_height = zoom_height;

		if( (capture_width >= MT9V113_QXGA_WIDTH) && (capture_height >= MT9V113_QXGA_HEIGHT) ) {
			mtSizeFormat = MT9V113_SIZE_QXGA;
			capture_width = MT9V113_QXGA_WIDTH;
			capture_height = MT9V113_QXGA_HEIGHT;
		}
		else if( (capture_width >= MT9V113_UXGA_WIDTH) && (capture_height >= MT9V113_UXGA_HEIGHT) ) {
			mtSizeFormat = MT9V113_SIZE_UXGA;
			capture_width = MT9V113_UXGA_WIDTH;
			capture_height = MT9V113_UXGA_HEIGHT;
		}
		else if( (capture_width >= MT9V113_XGA_WIDTH) && (capture_height >= MT9V113_XGA_HEIGHT) ) {
			mtSizeFormat = MT9V113_SIZE_XGA;
			capture_width = MT9V113_XGA_WIDTH;
			capture_height = MT9V113_XGA_HEIGHT;
		}
		else if( (capture_width >= MT9V113_VGA_WIDTH) && (capture_height >= MT9V113_VGA_HEIGHT) ) {
			mtSizeFormat = MT9V113_SIZE_JPEG_VGA;
			capture_width = MT9V113_VGA_WIDTH;
			capture_height = MT9V113_VGA_HEIGHT;
		}
		else if( (capture_width >= 320) && (capture_height>= 240) ) {
			mtSizeFormat = MT9V113_SIZE_JPEG_QVGA;
			capture_width = 320;
			capture_height = 240;
		}
		else if( (capture_width >= 256) && (capture_height>= 192) ) {
			mtSizeFormat = MT9V113_SIZE_JPEG_256X192;
			capture_width = 256;
			capture_height = 192;
		}
		else {
			mtSizeFormat = MT9V113_SIZE_JPEG_VGA;
			capture_width = MT9V113_VGA_WIDTH;
			capture_height = MT9V113_VGA_HEIGHT;
			printk(KERN_ERR "mt9v113: can not capture %d:%d at zoom %d\n", capture_width, capture_height, zoom);
		}

		printk(KERN_DEBUG "mt9v113: jpeg capture %d:%d, zoom=%d, zoomed %d:%d\n",
			capture_width, capture_height, zoom, zoom_width, zoom_height);

		camera_context->capture_input_width = capture_width;
		camera_context->capture_input_height = capture_height;

	} else /* YUV422 */
	if( ( camera_context->capture_input_width <= MT9V113_SQCIF_WIDTH) 
			&& (camera_context->capture_input_height <= MT9V113_SQCIF_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_SQCIF;
	} else if( ( camera_context->capture_input_width <= MT9V113_QCIF_WIDTH)
			&& (camera_context->capture_input_height <= MT9V113_QCIF_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_QCIF;
	} else if( ( camera_context->capture_input_width <= 256)
			&& (camera_context->capture_input_height <= 192)) {
		mtSizeFormat = MT9V113_SIZE_256X192;
	} else if( ( camera_context->capture_input_width <= MT9V113_QVGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9V113_QVGA_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_QVGA;
	} else if( ( camera_context->capture_input_width <= MT9V113_CIF_WIDTH)
			&& (camera_context->capture_input_height <= MT9V113_CIF_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_CIF;
	} else if( ( camera_context->capture_input_width <= MT9V113_YGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9V113_YGA_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_YGA;
	} else if( ( camera_context->capture_input_width <= MT9V113_VGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9V113_VGA_HEIGHT)) {
		mtSizeFormat = MT9V113_SIZE_VGA;
	} else {
		mtSizeFormat = MT9V113_SIZE_VGA;
	}

	if( (mtSizeFormat!=g_mt9v113_size)||(mtFormat!=g_mt9v113_format)||(camera_context->capture_mode!=G_mt9v113_mode_change) ) {
		status = mt9v113hw_set_format( mtSizeFormat, mtFormat, camera_context->capture_mode);
		g_mt9v113_size = mtSizeFormat;
		g_mt9v113_format = mtFormat;
	}

	/* set digital zoom after size changed */
	mt9v113hw_set_resolution_zoom(zoom);

	timing.BFW = 0x01;
	timing.BLW = 0x00;

	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);
	return 0;
}

static
int mt9v113_start_capture( p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;
	if(camera_context->capture_mode == CAMERA_MODE_STILL)
		/* wait about 2 frame before capture */
		camera_context->capture_drop_frames = 2;
	else
		camera_context->capture_drop_frames = 0;

	return 0;
}

static
int mt9v113_stop_capture(  p_camera_context_t camera_context )
{
	return 0;
}

static
int mt9v113_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	mt9v113hw_power_down( mode );
	return 0;
}

static
int mt9v113_read_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;

	mt9v113hw_read_sensor_reg( reg_addr, reg_val );
	return 0;
}

static
int mt9v113_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	u16 buffer;
	int status;

	(void) camera_context;

	buffer = reg_val;
	status = (mt9v113hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

static int mt9v113_set_digital_zoom( p_camera_context_t camera_context, u32 value)
{
	if ((value<100) || (value>699))
		return -EINVAL;

	/* set digital zoom */
	return mt9v113hw_set_resolution_zoom(value);
}

/**get digital zoom max level*/
static int mt9v113_get_digital_zoom_max( p_camera_context_t cam_ctx, u32 *value)
{
	int w, h, mw, mh;

	w = cam_ctx->mode[CAMERA_MODE_VIDEO].width;
	h = cam_ctx->mode[CAMERA_MODE_VIDEO].height;

	if( (w<=0) || (h<=0) )
		return -EINVAL;

	mw = 1024 * 100 / w;
	mh = 768 * 100 / h;
	if (mw<mh)
		*value = mw;
	else
		*value = mh;
	return 0;
}

/**get digital zoom step*/
static int mt9v113_get_digital_zoom_step( p_camera_context_t camera_context, u32 *value)
{
	*value = 100;
	return 0;
}

static
int mt9v113_set_gamma( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 2)
		value= 2;

	return mt9v113hw_set_gamma( value);
}

static
int mt9v113_set_brightness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 8)
		value= 8;

	return mt9v113hw_set_brightness( value);
}

static
int mt9v113_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int mtindex;
	(void) camera_context;

	if( value == SENSOR_CONTRAST_LOWEST )
		mtindex = 1;	/* M4 */
	else if( value == SENSOR_CONTRAST_LOW )
		mtindex = 3;	/* M2 */
	else if( value == SENSOR_CONTRAST_MIDDLE )
		mtindex = 0;	/* default */
	else if( value == SENSOR_CONTRAST_HIGH )
		mtindex = 6;	/* 2 */
	else if( value == SENSOR_CONTRAST_HIGHEST )
		mtindex = 8;	/* 4 */
	else
		mtindex = 0;

	return mt9v113hw_set_contrast( mtindex);
}

static
int mt9v113_set_exposure( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int mtindex;
	(void) camera_context;

	{
		if( value == SENSOR_EXPOSURE_LOWEST )
			mtindex = 1;	/* -1.7EV */
		else if( value == SENSOR_EXPOSURE_LOW )
			mtindex = 3;	/* -1.0EV */
		else if( value == SENSOR_EXPOSURE_MIDDLE )
			mtindex = 0;	/* default */
		else if( value == SENSOR_EXPOSURE_HIGH )
			mtindex = 8;	/* 1.0EV */
		else if( value == SENSOR_EXPOSURE_HIGHEST )
			mtindex = 10;	/* 1.7EV */
		else
			mtindex = 0;
	}

	return mt9v113hw_set_exposure( mtindex);
}

static
int mt9v113_get_framerate(u32 format, u32 width, u32 height, u32 *numerator, u32 *denominator)
{
	int ret = 0;
	switch (format) {
		case V4L2_PIX_FMT_SRGGB10:
			if ((width <= 800) && (height <= 600)) {
				/* CIF */
				*numerator = 15;	
				*denominator = 1;
			} else if ((width <= 1600) && (height <= 1200)) {
				/* UXGA */
				*numerator = 15;	
				*denominator = 2;
				
			} else if ((width <= 2560) && (height <= 1920)) {
				/* FULL */
				*numerator = 5;	
				*denominator = 2;				
			}else{
				//ret = -1;
			}
			break;
			
		default:
			//ret = -1;
			break;
		}

	return ret;
}

static int mt9v113_set_framerate(u32 numerator, u32 denominator)
{
	int fps;
	if (0<denominator && 0<=numerator) {
		fps = numerator / denominator;

		if( fps<=0 )
			mt9v113set_fps = 0;
		else if( fps>25 )
			mt9v113set_fps = 30;
		else if( fps>15 )
			mt9v113set_fps = 25;
		else
			mt9v113set_fps = 15;

		return mt9v113hw_set_framerate(mt9v113set_fps);
	}
	return 0;
}

static char name[] = "Micron_mt9v113";

static p_camera_function_t camera_functions;

static int format_list[] = {

	V4L2_PIX_FMT_YUV422P,	
	-1	
};

static
int mt9v113_register(int id)
{
	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max = 640;
	camera_functions->height_max = 480;
	camera_functions->width_min = 48;
	camera_functions->height_min = 32;

	camera_functions->v_power = 0; //2800;
	camera_functions->v_io = 0; //1800;
		

	camera_functions->format_list = format_list;

	camera_functions->init = mt9v113_init;     	
	camera_functions->deinit = mt9v113_deinit;     
	camera_functions->set_capture_format = mt9v113_set_capture_format; 	
	camera_functions->start_capture = mt9v113_start_capture;	 		
	camera_functions->stop_capture = mt9v113_stop_capture; 			
	camera_functions->sleep = mt9v113_sleep;	
	camera_functions->wakeup = mt9v113_wake;	
	camera_functions->read_8bit = NULL;			
	camera_functions->write_8bit = NULL;			
	camera_functions->read_16bit = mt9v113_read_16bit;		
	camera_functions->write_16bit = mt9v113_write_16bit;		
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode = mt9v113_set_power_mode;			
	camera_functions->set_contrast = mt9v113_set_contrast;			
	camera_functions->set_whitebalance = NULL;		
	camera_functions->set_exposure = mt9v113_set_exposure;			
	camera_functions->set_gamma= mt9v113_set_gamma;
	camera_functions->set_zoom = NULL;

	camera_functions->get_framerate = mt9v113_get_framerate;
	camera_functions->set_framerate =	mt9v113_set_framerate;
	camera_functions->set_autofocus =	NULL;
	camera_functions->set_flashmode =	NULL;
	camera_functions->set_digital_zoom = 	mt9v113_set_digital_zoom;
	camera_functions->get_digital_zoom_max 	= mt9v113_get_digital_zoom_max;
	camera_functions->get_digital_zoom_step	= mt9v113_get_digital_zoom_step;
	
	camera_functions->name = name;
	camera_functions->id = id;

	/*register mt9v113 as high resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static
int mt9v113_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *                              MT9V113 I2C Client Driver
 *****************************************************************************/
static
int read_sensor_reg_16bit_dat_16bit(u16 addr, u16 *pvalue)
{    
	int ret;
	u8 dataH;
	u8 dataL;
	u8 address[2];
	if( g_client == NULL )	/* Not initialize the client */	
	{
		ovI2C_dbg("not initialize the client\n");
		return -1;
	}
	address[0] = addr>>8;	
	address[1] = addr;
	ret = i2c_smbus_write_byte_data(g_client,address[0],address[1]);
	if(ret) {
		printk(KERN_ERR "mt9v113: read reg [%04x], write address failed %d\n", addr, ret);
		return ret;
	}
	ret = i2c_smbus_read_byte(g_client);
	if(ret<0) {
		printk(KERN_ERR "mt9v113: read reg [%04x], read byte failed %d\n", addr, ret);
		return ret;
	}
	dataH = (u8)ret;
	ret = i2c_smbus_read_byte(g_client);
	if(ret<0) {
		printk(KERN_ERR "mt9v113: read reg [%04x], read byte failed %d\n", addr, ret);
		return ret;
	}
	dataL = (u8)ret;
	*pvalue = (dataH << 8) | dataL;

	return 0; 
}

static
int write_sensor_reg_16bit_dat_16bit(u16 addr, u16 *value)
{
	int ret;
	u8 data[4];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;			
	data[0] = addr>>8;
	data[1] = addr;	
	data[2]=  (*value) >> 8;
	data[3]=  (*value) & 0xFF;
	

	ret = i2c_master_send(g_client, data, 4);
	if(ret<0) {
		printk(KERN_ERR "mt9v113: write reg [%04x:%02x] failed %d\n", addr, *value, ret);
		return ret;
	}
	
	return 0;
}

static
int write_sensor_reg_16bit_dat_16bit_m(u16 addr, u16 value, u16 mask)
{
	int ret = 0;
	u16 regval;
	
	read_sensor_reg_16bit_dat_16bit(addr, &regval);
	
	regval  &=  (~mask);
	value   &=  mask;
	
	regval|=value;
	
	ret = write_sensor_reg_16bit_dat_16bit(addr,&regval);

	return ret;
}

static int write_sensor_reg(u16 addr, u16 value)
{
	int ret;
	down(&mt9v113_i2c_lock);
	ret = write_sensor_reg_16bit_dat_16bit(addr, &value);
	up(&mt9v113_i2c_lock);
	return ret;
}

static int read_sensor_reg(u16 addr, u16 *pvalue)
{
	int ret;
	down(&mt9v113_i2c_lock);
	ret = read_sensor_reg_16bit_dat_16bit(addr,pvalue);
	up(&mt9v113_i2c_lock);
	return ret;
}

static int polling_MCU_TEST(u32 u32timeout)
{
	u16 u16value = 0;
	
	read_sensor_reg(0x001A, &u16value);
	printk("$%d", u16value);
	read_sensor_reg(0x0014, &u16value);
	printk("$%d", u16value);
	read_sensor_reg(0x0018, &u16value);
	printk("$%d", u16value);
	
	write_sensor_reg(REG_MCU_ADDRESS, 0x2703);
	read_sensor_reg(REG_MCU_DATA, &u16value);
	printk("$%d", u16value);
	write_sensor_reg(REG_MCU_ADDRESS, 0x2705);
	read_sensor_reg(REG_MCU_DATA, &u16value);
	printk("$%d", u16value);
	write_sensor_reg(REG_MCU_ADDRESS, 0x2707);
	read_sensor_reg(REG_MCU_DATA, &u16value);
	printk("$%d", u16value);
	write_sensor_reg(REG_MCU_ADDRESS, 0x2709);
	read_sensor_reg(REG_MCU_DATA, &u16value);
	printk("$%d", u16value);
	return 0;
}

static int __devinit sensor_probe(struct i2c_client *client, const struct i2c_device_id *dev_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	int id = 0;
	
	pdata = client->dev.platform_data;
	id = pdata->id;
	printk(KERN_ERR"[%s:%d] id= %d\n", __FILE__, __LINE__, id);

	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		printk("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);
	
	ci_set_clock(1, 1, CICLK);
	
	g_client = client;

	if( pdata->power_on && (pdata->power_on( id) != 0) )
		printk(KERN_ERR"[%s:%d] Camera power on failed!!!!!!!\n", __FILE__, __LINE__);

	if (mt9v113_sensorID_detect()<0)
		goto err;

	if(mt9v113_register(id)<0)
		goto err;
	printk("MT9V113 detected.\n");
	
	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	clk_disable(clk);
	clk_put(clk);

	return 0;
err:
	printk("MT9V113 detect failed.\n");
	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	printk("[MT9V113] sensor_remove in\n");
	
	mt9v113_unregister(pdata->id);
	g_client = NULL;

	return 0;
}

static const struct i2c_device_id mt9v113_id[] = {
	{"sensor_mt9v113", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "sensor_mt9v113",
	},
	.id_table	= mt9v113_id,
	.probe		= sensor_probe,
	.remove		= sensor_remove,
};

static int __init sensor_init(void)
{
	int ret = 0;

	//mt9v113hw_reset();

	ret = i2c_add_driver(&sensor_driver);
	printk( KERN_INFO "cyj test ret = 0x%x\n", ret);
	return ret;
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}

module_init(sensor_init);
module_exit(sensor_exit);

MODULE_DESCRIPTION("MT9V113 I2C Client driver");
MODULE_LICENSE("GPL");
