/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for OV3640
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/
#include "camera.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <mach/mfp.h>
#include <mach/gpio.h>
#include <mach/hardware.h>
#include <linux/err.h>

#include <linux/clk.h>
#include <linux/i2c.h>

/*****************************************************************************
 * Constants & Registers
 *****************************************************************************/
/* CICLK for QCI */
#define CICLK	2600

/* Revision constants */
#define PID_VERH		0x36
#define PID_VERL	       0x40
#define PID_ADRRH            0x300a
#define PID_ADRRL            0x300b

/* Output Size & Format */
#define OV3640_SIZE_QVGA	       1          //320x240
#define OV3640_SIZE_VGA               2          //640x480
#define OV3640_SIZE_SVAG             3          //800x600
#define OV3640_SIZE_D1                 4          //864x600
#define OV3640_SIZE_XGA               5         //1024x768
#define OV3640_SIZE_SXGA		6          //1280x1024
#define OV3640_SIZE_UXGA             7         //1600X1200
#define OV3640_SIZE_QXGA		8          //2048x1536
#define OV3640_SIZE_QSXGA           9         //2592X1944

#define OV3640_FORMAT_YUV_422  1
#define OV3640_FORMAT_RAW10     2
#define OV3640_FORMAT_RAW8       3
#define OV3640_FORMAT_JPG           4
#define OV3640_FORMAT_RGB565     5

#define OV3640_PREVIEW_WIDTH     320
#define OV3640_PREVIEW_HEIGHT    240

#define OV3640_REGEND		(0xFFFF)
/* Unique ID allocation */
static struct i2c_client *g_client;

#define I2C_DEBUG
//#undef I2C_DEBUG

#ifdef  I2C_DEBUG
#define ovI2C_dbg(fmt, arg...)    printk(KERN_INFO "%s(line %d): " fmt "\n",  \
                __FUNCTION__, __LINE__, ##arg)
#else 
#define ovI2C_dbg(fmt, arg...)        do {} while (0)
#endif

/****************************************************************************

   Gloable variable
   
*****************************************************************************/
int G_ov3640_mode_change = CAMERA_MODE_VIDEO;
u8  ov3640_gaincode = 0;
int  ov3640_expcode = 0;
u8  ov3640_recRgain =0,ov3640_recBgain=0,ov3640_recGgain=0;
#define QXGAMAXEXPLIN 1568
#define XGAMAXEXPLIN    779
const static u8 gSensorSlaveAddr = 0x3c;

static int read_sensor_reg( const u16 subAddress, u8 *bufP );
static int write_sensor_reg( const u16 subAddress, u8 bufP );
/*****************************************************************************

 *  Register Settings Function

 *****************************************************************************/
const static u16 QVGA_yuv_init_tab[] = 
{
	0x3012 ,0x80,
	0x304d ,0x45,
	0x3087 ,0x16,
	0x30aa ,0x45,
	0x30b0 ,0xff,
	0x30b1 ,0xff,
	0x30b2 ,0x10,
	0x300e ,0x32, /* 15fps */
	0x300f ,0x21,
	0x3010 ,0x20,
	0x3011 ,0x00,
	0x304c ,0x82,
	0x3018 ,0x38,
	0x3019 ,0x30,
	0x301a ,0x61,
	0x307d ,0x00,
	0x3087 ,0x02,
	0x3082 ,0x20,
	0x3015 ,0x12,
	0x3014 ,0x04,
	0x3013 ,0xf7,
	0x303c ,0x08,
	0x303d ,0x18,
	0x303e ,0x06,
	0x303F ,0x0c,
	0x3030 ,0x62,
	0x3031 ,0x26,
	0x3032 ,0xe6,
	0x3033 ,0x6e,
	0x3034 ,0xea,
	0x3035 ,0xae,
	0x3036 ,0xa6,
	0x3037 ,0x6a,
	0x3104 ,0x02,
	0x3105 ,0xfd,
	0x3106 ,0x00,
	0x3107 ,0xff,
	0x3300 ,0x12,
	0x3301 ,0xde,
	0x3302 ,0xef,
	0x3316 ,0xff,
	0x3317 ,0x00,
	0x3308 ,0xa5,
	0x3312 ,0x26,
	0x3314 ,0x42,
	0x3313 ,0x2b,
	0x3315 ,0x42,
	0x3310 ,0xd0,
	0x3311 ,0xbd,
	0x330c ,0x18,
	0x330d ,0x18,
	0x330e ,0x56,
	0x330f ,0x5c,
	0x330b ,0x1c,
	0x3306 ,0x5c,
	0x3307 ,0x11,
	0x336a ,0x52,
	0x3370 ,0x46,
	0x3376 ,0x38,
	0x30b8 ,0x20,
	0x30b9 ,0x17,
	0x30ba ,0x04,
	0x30bb ,0x08,
	0x3100 ,0x02,
	0x3301 ,0xde,
	0x3304 ,0xfc,
	0x3400 ,0x00,
	0x3404 ,0x02,
	0x3012 ,0x10,
	0x3023 ,0x06,
	0x3026 ,0x03,
	0x3027 ,0x04,
	0x302a ,0x03,
	0x302b ,0x10,
	0x3075 ,0x24,
	0x300d ,0x01,
	0x30d7 ,0x80,
	0x3069 ,0x04,
	0x303e ,0x00,
	0x303f ,0xc0,
	0x335f ,0x34,
	0x3360 ,0x0c,
	0x3361 ,0x04,
	0x3362 ,0x01,
	0x3363 ,0x48,
	0x3364 ,0xf4,
	0x3403 ,0x42,
	0x3088 ,0x01,
	0x3089 ,0x40,
	0x308a ,0x00,
	0x308b ,0xf0,
	0x3600 ,0x02, /* VSYNC to HREF is decrease to almost 0*/
	0x304c,0x89,
	0x3011,0x00,

	OV3640_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 VGA_yuv_init_tab[] = 
{
	//24Mhz Mclk, 35Mhz PCLK, 30fps
	0x3012, 0x80,
	0x304d, 0x45, 
	0x30A7, 0x5E,
	0x3087, 0x16,
	0x309C, 0x1a,
	0x30a2, 0xe4,
	0x30aa, 0x42,
	0x30b0, 0xff,
	0x30b1, 0xff,
	0x30b2, 0x10,
	0x300e, 0x32,
	0x300f, 0x21,
	0x3010, 0x20,
	0x3011, 0x00,
	0x304c, 0x82,
	0x30d9, 0x0d,
	0x30db, 0x08,
	0x3016, 0x82,
	0x30D7, 0x10,
	0x3018, 0x38,
	0x3019, 0x30,
	0x301a, 0x61,
	0x307d, 0x00,
	0x3087, 0x02,
	0x3082, 0x20,
	0x3015, 0x12,
	0x3014, 0x04,
	0x3013, 0xf7,
	0x303c, 0x08,
	0x303d, 0x18 ,
	0x303e, 0x06,
	0x303f, 0x0c,
	0x3030, 0x62,
	0x3031, 0x26,
	0x3032, 0xe6,
	0x3033, 0x6e,
	0x3034, 0xea,
	0x3035, 0xae,
	0x3036, 0xa6,
	0x3037, 0x6a,
	0x3104, 0x02,
	0x3105, 0xfd,
	0x3106, 0x00,
	0x3107, 0xff,
	0x3300, 0x12,
	0x3301, 0xde,
	0x3302, 0xef, //dcw
	0x3316, 0xff,
	0x3317, 0x00,    
	0x3308, 0xa5,
	0x3312, 0x26,
	0x3314, 0x42,
	0x3313, 0x2b,
	0x3315, 0x42,
	0x3310, 0xd0,
	0x3311, 0xbd,
	0x330c, 0x18,
	0x330d, 0x18,
	0x330e, 0x56,
	0x330f, 0x5c,
	0x330b, 0x1c,
	0x3306, 0x5c,
	0x3307, 0x11,
	0x336a, 0x52,
	0x3370, 0x46,
	0x3376, 0x38,
	0x3300, 0x13,
	0x30b8, 0x20,
	0x30b9, 0x17,
	0x30ba, 0x04,
	0x30bb, 0x08,
	0x3507, 0x06,
	0x350a, 0x4f,

	// yuv  
	0x3100, 0x02,
	0x3301, 0xde,
	0x3304, 0xfc,
	0x3400, 0x00,
	0x3404, 0x02,
	0x3012, 0x10, 
	0x3023, 0x06,
	0x3026, 0x03,
	0x3027, 0x04,
	0x302a, 0x03,
	0x302b, 0x10,
	0x3075, 0x24,
	0x300d, 0x01,
	0x30d7, 0x80,
	0x3069, 0x04,
	0x303e, 0x00,
	0x303f, 0xc0,
	0x335f, 0x34,
	0x3360, 0x0c,
	0x3361, 0x04,
	0x3362, 0x34,
	0x3363, 0x08, 
	0x3364, 0x04, 
	0x3403, 0x42,
	0x3088, 0x04,
	0x3089, 0x00, 
	0x308a, 0x03,
	0x308b, 0x00, 
	0x308d, 0x04,
	0x3086, 0x03,
	0x3086, 0x00,
	0x3600, 0xc4,

	//XGA->VGA
	0x3302, 0xef,
	0x335f, 0x34,
	0x3360, 0x0c,
	0x3361, 0x04,
	0x3362, 0x12,
	0x3363, 0x88,
	0x3364, 0xe4,
	0x3403, 0x42,
	0x3088, 0x12,
	0x3089, 0x80,
	0x308a, 0x01,
	0x308b, 0xe0,
	0x304c, 0x83,

	OV3640_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 QXGA_yuv_init_tab[] = 
{
	0x304d ,0x45,
	0x3087 ,0x16,
	0x30aa ,0x45,
	0x30b0 ,0xff,
	0x30b1 ,0xff,
	0x30b2 ,0x10,
	0x300e ,0x32,
	0x300f ,0x21,
	0x3010 ,0x20,
	0x304c ,0x82,
	0x30a9 ,0xbd,
	0x3018 ,0x38,
	0x3019 ,0x30,
	0x301a ,0x61,
	0x307d ,0x00,
	0x3087 ,0x02,
	0x3082 ,0x20,
	0x3015 ,0x12,
	0x3014 ,0x0c,
	0x3013 ,0xf7,
	0x303c ,0x08,
	0x303d ,0x18,
	0x303e ,0x06,
	0x303F ,0x0c,
	0x3030 ,0x62,
	0x3031 ,0x26,
	0x3032 ,0xe6,
	0x3033 ,0x6e,
	0x3034 ,0xea,
	0x3035 ,0xae,
	0x3036 ,0xa6,
	0x3037 ,0x6a,
	0x3104 ,0x02,
	0x3105 ,0xfd,
	0x3106 ,0x00,
	0x3107 ,0xff,
	0x3300 ,0x13,
	0x3301 ,0xde,
	0x3302 ,0xef,
	0x336a ,0x52,
	0x3370 ,0x46,
	0x3376 ,0x38,
	0x30b8 ,0x20,
	0x30b9 ,0x17,
	0x30ba ,0x04,
	0x30bb ,0x08,
	0x3507 ,0x06,
	0x350a ,0x4f,
	0x3100 ,0x32,
	0x3400 ,0x02,
	0x3404 ,0x22,
	0x3500 ,0x00,
	0x3600 ,0xc4,
	0x3610 ,0x60,
	0x350a ,0x4f,
	0x3088 ,0x08,
	0x3089 ,0x00,
	0x308a ,0x06,
	0x308b ,0x00,
	0x308d ,0x04,
	0x3086 ,0x03,
	0x3086 ,0x00,
	0x3012 ,0x00,
	0x3020 ,0x01,
	0x3021 ,0x1d,
	0x3022 ,0x00,
	0x3023 ,0x0a,
	0x3024 ,0x08,
	0x3025 ,0x18,
	0x3026 ,0x06,
	0x3027 ,0x0c,
	0x302a ,0x06,
	0x302b ,0x20,
	0x3075 ,0x44,
	0x300d ,0x00,
	0x30d7 ,0x00,
	0x3069 ,0x44,
	0x303e ,0x01,
	0x303f ,0x80,
	0x3302 ,0xef,
	0x335f ,0x68,
	0x3360 ,0x18,
	0x3361 ,0x0c,
	0x3362 ,0x34,
	0x3363 ,0x08,
	0x3364 ,0x06,
	0x3403 ,0x42,
	0x3088 ,0x08,
	0x3089 ,0x00,
	0x308a ,0x06,
	0x308b ,0x00,
	0x3100 ,0x02,
	0x3304 ,0xfc,
	0x3400 ,0x00,
	0x3404 ,0x02,
	0x3600 ,0xc4,
	0x3014 ,0x04,
	0x302e ,0x00,
	0x302d ,0x00,
	0x3015 ,0x12,
	0x3014 ,0x0c,
	0x3010 ,0x10,
	0x335f ,0x68,
	0x3360 ,0x18,
	0x3361 ,0x0c,
	0x3362 ,0x68,
	0x3363 ,0x08,
	0x3364 ,0x04,
	0x3403 ,0x42,
	0x3088 ,0x08,
	0x3089 ,0x00,
	0x308a ,0x06,
	0x308b ,0x00,
	0x304c,0x81,
	0x3011,0x01,
	
	OV3640_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 QXGA_jpg_init_tab[] = 
{
	0x3012 ,0x80,
	0x304d ,0x45,
	0x3087 ,0x16,
	0x30aa ,0x45,
	0x30b0 ,0xff,
	0x30b1 ,0xff,
	0x30b2 ,0x13,
	0x300e ,0x33,
	0x300f ,0x21,
	0x3010 ,0x20,
	0x3011 ,0x00,
	0x304c ,0x84,
	0x3018 ,0x38,
	0x3019 ,0x30,
	0x301a ,0x61,
	0x307d ,0x00,
	0x3087 ,0x02,
	0x3082 ,0x20,
	0x3015 ,0x12,
	0x3014 ,0x04,
	0x3013 ,0xf7,
	0x303c ,0x08,
	0x303d ,0x18,
	0x303e ,0x06,
	0x303f ,0x0c,
	0x3030 ,0x62,
	0x3031 ,0x26,
	0x3032 ,0xe6,
	0x3033 ,0x6e,
	0x3034 ,0xea,
	0x3035 ,0xae,
	0x3036 ,0xa6,
	0x3037 ,0x6a,
	0x3104 ,0x02,
	0x3105 ,0xfd,
	0x3106 ,0x00,
	0x3107 ,0xff,
	0x3300 ,0x12,
	0x3301 ,0xde,
	0x3302 ,0xcf,
	0x3312 ,0x26,
	0x3314 ,0x42,
	0x3313 ,0x2b,
	0x3315 ,0x42,
	0x3310 ,0xd0,
	0x3311 ,0xbd,
	0x330c ,0x18,
	0x330d ,0x18,
	0x330e ,0x56,
	0x330f ,0x5c,
	0x330b ,0x1c,
	0x3306 ,0x5c,
	0x3307 ,0x11,
	0x336a ,0x52,
	0x3370 ,0x46,
	0x3376 ,0x38,
	0x30b8 ,0x20,
	0x30b9 ,0x17,
	0x30ba ,0x04,
	0x30bb ,0x08,
	0x3507 ,0x06,
	0x350a ,0x4f,
	0x3100 ,0x32,
	0x3304 ,0x00,
	0x3400 ,0x02,
	0x3404 ,0x22,
	0x3500 ,0x00, 
	0x3600 ,0xc4, 
	OV3640_REGEND, 0x00	    /* End of list delimiter */
	
};

const static u16 AF_init_tab[] = 
{
	0x308c,0x00,
	0x3104,0x02,
	0x3105,0xff,
	0x3106,0x00,
	0x3107,0xff,
	0x8000,0x02,
	0x8001,0x06,
	0x8002,0xF6,
	0x8003,0x02,
	0x8004,0x08,
	0x8005,0xA3,
	0x8006,0x02,
	0x8007,0x1B,
	0x8008,0x00,
	0x8009,0xA0,
	0x800A,0x00,
	0x800B,0x02,
	0x800C,0x0A,
	0x800D,0xAE,
	0x800E,0xE5,
	0x800F,0x23,
	0x8010,0x14,
	0x8011,0x60,
	0x8012,0x72,
	0x8013,0x14,
	0x8014,0x70,
	0x8015,0x03,
	0x8016,0x02,
	0x8017,0x01,
	0x8018,0x6F,
	0x8019,0x14,
	0x801A,0x70,
	0x801B,0x03,
	0x801C,0x02,
	0x801D,0x02,
	0x801E,0x2F,
	0x801F,0x14,
	0x8020,0x70,
	0x8021,0x03,
	0x8022,0x02,
	0x8023,0x03,
	0x8024,0x18,
	0x8025,0x24,
	0x8026,0x04,
	0x8027,0x60,
	0x8028,0x03,
	0x8029,0x02,
	0x802A,0x03,
	0x802B,0x2D,
	0x802C,0xE4,
	0x802D,0xF5,
	0x802E,0x36,
	0x802F,0xF5,
	0x8030,0x37,
	0x8031,0x12,
	0x8032,0x08,
	0x8033,0x6C,
	0x8034,0x8E,
	0x8035,0x24,
	0x8036,0x8F,
	0x8037,0x25,
	0x8038,0xE4,
	0x8039,0xF5,
	0x803A,0x36,
	0x803B,0x75,
	0x803C,0x37,
	0x803D,0x01,
	0x803E,0x12,
	0x803F,0x08,
	0x8040,0x6C,
	0x8041,0x8E,
	0x8042,0x2A,
	0x8043,0x8F,
	0x8044,0x2B,
	0x8045,0xC3,
	0x8046,0xE5,
	0x8047,0x1A,
	0x8048,0x94,
	0x8049,0x58,
	0x804A,0xE5,
	0x804B,0x19,
	0x804C,0x94,
	0x804D,0x02,
	0x804E,0x50,
	0x804F,0x08,
	0x8050,0x75,
	0x8051,0x22,
	0x8052,0x01,
	0x8053,0x75,
	0x8054,0x23,
	0x8055,0x01,
	0x8056,0x80,
	0x8057,0x06,
	0x8058,0x75,
	0x8059,0x22,
	0x805A,0xFF,
	0x805B,0x75,
	0x805C,0x23,
	0x805D,0x03,
	0x805E,0xAF,
	0x805F,0x22,
	0x8060,0xEF,
	0x8061,0x33,
	0x8062,0x95,
	0x8063,0xE0,
	0x8064,0xFE,
	0x8065,0x7C,
	0x8066,0x00,
	0x8067,0x7D,
	0x8068,0x14,
	0x8069,0x12,
	0x806A,0x0A,
	0x806B,0x6D,
	0x806C,0xEF,
	0x806D,0x25,
	0x806E,0x1A,
	0x806F,0xF5,
	0x8070,0x1A,
	0x8071,0xEE,
	0x8072,0x35,
	0x8073,0x19,
	0x8074,0xF5,
	0x8075,0x19,
	0x8076,0xE5,
	0x8077,0x1A,
	0x8078,0x25,
	0x8079,0x1C,
	0x807A,0xF5,
	0x807B,0x37,
	0x807C,0xE5,
	0x807D,0x19,
	0x807E,0x35,
	0x807F,0x1B,
	0x8080,0xF5,
	0x8081,0x36,
	0x8082,0x02,
	0x8083,0x09,
	0x8084,0xD1,
	0x8085,0xE4,
	0x8086,0xF5,
	0x8087,0x36,
	0x8088,0xF5,
	0x8089,0x37,
	0x808A,0x12,
	0x808B,0x08,
	0x808C,0x6C,
	0x808D,0x8E,
	0x808E,0x28,
	0x808F,0x8F,
	0x8090,0x29,
	0x8091,0x8E,
	0x8092,0x2E,
	0x8093,0x8F,
	0x8094,0x2F,
	0x8095,0x30,
	0x8096,0x0C,
	0x8097,0x02,
	0x8098,0x80,
	0x8099,0x40,
	0x809A,0xE4,
	0x809B,0xF5,
	0x809C,0x36,
	0x809D,0x75,
	0x809E,0x37,
	0x809F,0x01,
	0x80A0,0x12,
	0x80A1,0x08,
	0x80A2,0x6C,
	0x80A3,0x8E,
	0x80A4,0x2C,
	0x80A5,0x8F,
	0x80A6,0x2D,
	0x80A7,0xC3,
	0x80A8,0xE5,
	0x80A9,0x2D,
	0x80AA,0x95,
	0x80AB,0x2B,
	0x80AC,0xFF,
	0x80AD,0xE5,
	0x80AE,0x2C,
	0x80AF,0x95,
	0x80B0,0x2A,
	0x80B1,0xFE,
	0x80B2,0x12,
	0x80B3,0x0A,
	0x80B4,0x7F,
	0x80B5,0xC0,
	0x80B6,0x06,
	0x80B7,0xC0,
	0x80B8,0x07,
	0x80B9,0xC3,
	0x80BA,0xE5,
	0x80BB,0x2F,
	0x80BC,0x95,
	0x80BD,0x25,
	0x80BE,0xFF,
	0x80BF,0xE5,
	0x80C0,0x2E,
	0x80C1,0x95,
	0x80C2,0x24,
	0x80C3,0xFE,
	0x80C4,0x12,
	0x80C5,0x0A,
	0x80C6,0x7F,
	0x80C7,0xD0,
	0x80C8,0x05,
	0x80C9,0xD0,
	0x80CA,0x04,
	0x80CB,0xD3,
	0x80CC,0xEF,
	0x80CD,0x9D,
	0x80CE,0xEC,
	0x80CF,0x64,
	0x80D0,0x80,
	0x80D1,0xF8,
	0x80D2,0xEE,
	0x80D3,0x64,
	0x80D4,0x80,
	0x80D5,0x98,
	0x80D6,0x40,
	0x80D7,0x0A,
	0x80D8,0xD2,
	0x80D9,0x0C,
	0x80DA,0x85,
	0x80DB,0x24,
	0x80DC,0x30,
	0x80DD,0x85,
	0x80DE,0x25,
	0x80DF,0x31,
	0x80E0,0x80,
	0x80E1,0x0C,
	0x80E2,0x85,
	0x80E3,0x2C,
	0x80E4,0x2E,
	0x80E5,0x85,
	0x80E6,0x2D,
	0x80E7,0x2F,
	0x80E8,0x85,
	0x80E9,0x2A,
	0x80EA,0x30,
	0x80EB,0x85,
	0x80EC,0x2B,
	0x80ED,0x31,
	0x80EE,0xD3,
	0x80EF,0xE5,
	0x80F0,0x2F,
	0x80F1,0x95,
	0x80F2,0x31,
	0x80F3,0xE5,
	0x80F4,0x2E,
	0x80F5,0x95,
	0x80F6,0x30,
	0x80F7,0x40,
	0x80F8,0x3A,
	0x80F9,0xE5,
	0x80FA,0x1A,
	0x80FB,0x94,
	0x80FC,0x58,
	0x80FD,0xE5,
	0x80FE,0x19,
	0x80FF,0x94,
	0x8100,0x02,
	0x8101,0x50,
	0x8102,0x2D,
	0x8103,0xAF,
	0x8104,0x22,
	0x8105,0xEF,
	0x8106,0x33,
	0x8107,0x95,
	0x8108,0xE0,
	0x8109,0xFE,
	0x810A,0x7C,
	0x810B,0x00,
	0x810C,0x7D,
	0x810D,0x14,
	0x810E,0x12,
	0x810F,0x0A,
	0x8110,0x6D,
	0x8111,0xEF,
	0x8112,0x25,
	0x8113,0x1A,
	0x8114,0xF5,
	0x8115,0x1A,
	0x8116,0xEE,
	0x8117,0x35,
	0x8118,0x19,
	0x8119,0xF5,
	0x811A,0x19,
	0x811B,0xE5,
	0x811C,0x1A,
	0x811D,0x25,
	0x811E,0x1C,
	0x811F,0xF5,
	0x8120,0x37,
	0x8121,0xE5,
	0x8122,0x19,
	0x8123,0x35,
	0x8124,0x1B,
	0x8125,0xF5,
	0x8126,0x36,
	0x8127,0x12,
	0x8128,0x09,
	0x8129,0xD1,
	0x812A,0x75,
	0x812B,0x23,
	0x812C,0x02,
	0x812D,0x02,
	0x812E,0x02,
	0x812F,0x14,
	0x8130,0x02,
	0x8131,0x03,
	0x8132,0x14,
	0x8133,0x75,
	0x8134,0x22,
	0x8135,0xFF,
	0x8136,0xD3,
	0x8137,0xE5,
	0x8138,0x1A,
	0x8139,0x94,
	0x813A,0x14,
	0x813B,0xE5,
	0x813C,0x19,
	0x813D,0x94,
	0x813E,0x00,
	0x813F,0xAF,
	0x8140,0x22,
	0x8141,0xEF,
	0x8142,0x40,
	0x8143,0x28,
	0x8144,0x33,
	0x8145,0x95,
	0x8146,0xE0,
	0x8147,0xFE,
	0x8148,0x7C,
	0x8149,0x00,
	0x814A,0x7D,
	0x814B,0x28,
	0x814C,0x12,
	0x814D,0x0A,
	0x814E,0x6D,
	0x814F,0xEF,
	0x8150,0x25,
	0x8151,0x1A,
	0x8152,0xF5,
	0x8153,0x1A,
	0x8154,0xEE,
	0x8155,0x35,
	0x8156,0x19,
	0x8157,0xF5,
	0x8158,0x19,
	0x8159,0xE5,
	0x815A,0x1A,
	0x815B,0x25,
	0x815C,0x1C,
	0x815D,0xF5,
	0x815E,0x37,
	0x815F,0xE5,
	0x8160,0x19,
	0x8161,0x35,
	0x8162,0x1B,
	0x8163,0xF5,
	0x8164,0x36,
	0x8165,0x12,
	0x8166,0x09,
	0x8167,0xD1,
	0x8168,0x75,
	0x8169,0x23,
	0x816A,0x03,
	0x816B,0x22,
	0x816C,0x02,
	0x816D,0x02,
	0x816E,0xF0,
	0x816F,0xE4,
	0x8170,0xF5,
	0x8171,0x36,
	0x8172,0xF5,
	0x8173,0x37,
	0x8174,0x12,
	0x8175,0x08,
	0x8176,0x6C,
	0x8177,0x8E,
	0x8178,0x28,
	0x8179,0x8F,
	0x817A,0x29,
	0x817B,0x8E,
	0x817C,0x2E,
	0x817D,0x8F,
	0x817E,0x2F,
	0x817F,0x30,
	0x8180,0x0C,
	0x8181,0x02,
	0x8182,0x80,
	0x8183,0x40,
	0x8184,0xE4,
	0x8185,0xF5,
	0x8186,0x36,
	0x8187,0x75,
	0x8188,0x37,
	0x8189,0x01,
	0x818A,0x12,
	0x818B,0x08,
	0x818C,0x6C,
	0x818D,0x8E,
	0x818E,0x2C,
	0x818F,0x8F,
	0x8190,0x2D,
	0x8191,0xC3,
	0x8192,0xE5,
	0x8193,0x2D,
	0x8194,0x95,
	0x8195,0x2B,
	0x8196,0xFF,
	0x8197,0xE5,
	0x8198,0x2C,
	0x8199,0x95,
	0x819A,0x2A,
	0x819B,0xFE,
	0x819C,0x12,
	0x819D,0x0A,
	0x819E,0x7F,
	0x819F,0xC0,
	0x81A0,0x06,
	0x81A1,0xC0,
	0x81A2,0x07,
	0x81A3,0xC3,
	0x81A4,0xE5,
	0x81A5,0x2F,
	0x81A6,0x95,
	0x81A7,0x25,
	0x81A8,0xFF,
	0x81A9,0xE5,
	0x81AA,0x2E,
	0x81AB,0x95,
	0x81AC,0x24,
	0x81AD,0xFE,
	0x81AE,0x12,
	0x81AF,0x0A,
	0x81B0,0x7F,
	0x81B1,0xD0,
	0x81B2,0x05,
	0x81B3,0xD0,
	0x81B4,0x04,
	0x81B5,0xD3,
	0x81B6,0xEF,
	0x81B7,0x9D,
	0x81B8,0xEC,
	0x81B9,0x64,
	0x81BA,0x80,
	0x81BB,0xF8,
	0x81BC,0xEE,
	0x81BD,0x64,
	0x81BE,0x80,
	0x81BF,0x98,
	0x81C0,0x40,
	0x81C1,0x0A,
	0x81C2,0xD2,
	0x81C3,0x0C,
	0x81C4,0x85,
	0x81C5,0x24,
	0x81C6,0x30,
	0x81C7,0x85,
	0x81C8,0x25,
	0x81C9,0x31,
	0x81CA,0x80,
	0x81CB,0x0C,
	0x81CC,0x85,
	0x81CD,0x2C,
	0x81CE,0x2E,
	0x81CF,0x85,
	0x81D0,0x2D,
	0x81D1,0x2F,
	0x81D2,0x85,
	0x81D3,0x2A,
	0x81D4,0x30,
	0x81D5,0x85,
	0x81D6,0x2B,
	0x81D7,0x31,
	0x81D8,0xD3,
	0x81D9,0xE5,
	0x81DA,0x2F,
	0x81DB,0x95,
	0x81DC,0x31,
	0x81DD,0xE5,
	0x81DE,0x2E,
	0x81DF,0x95,
	0x81E0,0x30,
	0x81E1,0x40,
	0x81E2,0x43,
	0x81E3,0xE5,
	0x81E4,0x1A,
	0x81E5,0x94,
	0x81E6,0x58,
	0x81E7,0xE5,
	0x81E8,0x19,
	0x81E9,0x94,
	0x81EA,0x02,
	0x81EB,0x50,
	0x81EC,0x36,
	0x81ED,0xAF,
	0x81EE,0x22,
	0x81EF,0xEF,
	0x81F0,0x33,
	0x81F1,0x95,
	0x81F2,0xE0,
	0x81F3,0xFE,
	0x81F4,0x7C,
	0x81F5,0x00,
	0x81F6,0x7D,
	0x81F7,0x14,
	0x81F8,0x12,
	0x81F9,0x0A,
	0x81FA,0x6D,
	0x81FB,0xEF,
	0x81FC,0x25,
	0x81FD,0x1A,
	0x81FE,0xF5,
	0x81FF,0x1A,
	0x8200,0xEE,
	0x8201,0x35,
	0x8202,0x19,
	0x8203,0xF5,
	0x8204,0x19,
	0x8205,0xE5,
	0x8206,0x1A,
	0x8207,0x25,
	0x8208,0x1C,
	0x8209,0xF5,
	0x820A,0x37,
	0x820B,0xE5,
	0x820C,0x19,
	0x820D,0x35,
	0x820E,0x1B,
	0x820F,0xF5,
	0x8210,0x36,
	0x8211,0x12,
	0x8212,0x09,
	0x8213,0xD1,
	0x8214,0x85,
	0x8215,0x24,
	0x8216,0x26,
	0x8217,0x85,
	0x8218,0x25,
	0x8219,0x27,
	0x821A,0x85,
	0x821B,0x28,
	0x821C,0x24,
	0x821D,0x85,
	0x821E,0x29,
	0x821F,0x25,
	0x8220,0x02,
	0x8221,0x02,
	0x8222,0xE1,
	0x8223,0x02,
	0x8224,0x03,
	0x8225,0x14,
	0x8226,0x75,
	0x8227,0x22,
	0x8228,0xFF,
	0x8229,0xAF,
	0x822A,0x22,
	0x822B,0xEF,
	0x822C,0x02,
	0x822D,0x02,
	0x822E,0xF0,
	0x822F,0xE4,
	0x8230,0xF5,
	0x8231,0x36,
	0x8232,0xF5,
	0x8233,0x37,
	0x8234,0x12,
	0x8235,0x08,
	0x8236,0x6C,
	0x8237,0x8E,
	0x8238,0x26,
	0x8239,0x8F,
	0x823A,0x27,
	0x823B,0x8E,
	0x823C,0x2E,
	0x823D,0x8F,
	0x823E,0x2F,
	0x823F,0x30,
	0x8240,0x0C,
	0x8241,0x02,
	0x8242,0x80,
	0x8243,0x40,
	0x8244,0xE4,
	0x8245,0xF5,
	0x8246,0x36,
	0x8247,0x75,
	0x8248,0x37,
	0x8249,0x01,
	0x824A,0x12,
	0x824B,0x08,
	0x824C,0x6C,
	0x824D,0x8E,
	0x824E,0x2C,
	0x824F,0x8F,
	0x8250,0x2D,
	0x8251,0xC3,
	0x8252,0xE5,
	0x8253,0x2D,
	0x8254,0x95,
	0x8255,0x2B,
	0x8256,0xFF,
	0x8257,0xE5,
	0x8258,0x2C,
	0x8259,0x95,
	0x825A,0x2A,
	0x825B,0xFE,
	0x825C,0x12,
	0x825D,0x0A,
	0x825E,0x7F,
	0x825F,0xC0,
	0x8260,0x06,
	0x8261,0xC0,
	0x8262,0x07,
	0x8263,0xC3,
	0x8264,0xE5,
	0x8265,0x2F,
	0x8266,0x95,
	0x8267,0x25,
	0x8268,0xFF,
	0x8269,0xE5,
	0x826A,0x2E,
	0x826B,0x95,
	0x826C,0x24,
	0x826D,0xFE,
	0x826E,0x12,
	0x826F,0x0A,
	0x8270,0x7F,
	0x8271,0xD0,
	0x8272,0x05,
	0x8273,0xD0,
	0x8274,0x04,
	0x8275,0xD3,
	0x8276,0xEF,
	0x8277,0x9D,
	0x8278,0xEC,
	0x8279,0x64,
	0x827A,0x80,
	0x827B,0xF8,
	0x827C,0xEE,
	0x827D,0x64,
	0x827E,0x80,
	0x827F,0x98,
	0x8280,0x40,
	0x8281,0x0A,
	0x8282,0xD2,
	0x8283,0x0C,
	0x8284,0x85,
	0x8285,0x24,
	0x8286,0x30,
	0x8287,0x85,
	0x8288,0x25,
	0x8289,0x31,
	0x828A,0x80,
	0x828B,0x0C,
	0x828C,0x85,
	0x828D,0x2C,
	0x828E,0x2E,
	0x828F,0x85,
	0x8290,0x2D,
	0x8291,0x2F,
	0x8292,0x85,
	0x8293,0x2A,
	0x8294,0x30,
	0x8295,0x85,
	0x8296,0x2B,
	0x8297,0x31,
	0x8298,0xD3,
	0x8299,0xE5,
	0x829A,0x2F,
	0x829B,0x95,
	0x829C,0x31,
	0x829D,0xE5,
	0x829E,0x2E,
	0x829F,0x95,
	0x82A0,0x30,
	0x82A1,0x40,
	0x82A2,0x47,
	0x82A3,0xD3,
	0x82A4,0xE5,
	0x82A5,0x1A,
	0x82A6,0x94,
	0x82A7,0x00,
	0x82A8,0xE5,
	0x82A9,0x19,
	0x82AA,0x94,
	0x82AB,0x00,
	0x82AC,0x40,
	0x82AD,0x3A,
	0x82AE,0xAF,
	0x82AF,0x22,
	0x82B0,0xEF,
	0x82B1,0x33,
	0x82B2,0x95,
	0x82B3,0xE0,
	0x82B4,0xFE,
	0x82B5,0x7C,
	0x82B6,0x00,
	0x82B7,0x7D,
	0x82B8,0x14,
	0x82B9,0x12,
	0x82BA,0x0A,
	0x82BB,0x6D,
	0x82BC,0xEF,
	0x82BD,0x25,
	0x82BE,0x1A,
	0x82BF,0xF5,
	0x82C0,0x1A,
	0x82C1,0xEE,
	0x82C2,0x35,
	0x82C3,0x19,
	0x82C4,0xF5,
	0x82C5,0x19,
	0x82C6,0xE5,
	0x82C7,0x1A,
	0x82C8,0x25,
	0x82C9,0x1C,
	0x82CA,0xF5,
	0x82CB,0x37,
	0x82CC,0xE5,
	0x82CD,0x19,
	0x82CE,0x35,
	0x82CF,0x1B,
	0x82D0,0xF5,
	0x82D1,0x36,
	0x82D2,0x12,
	0x82D3,0x09,
	0x82D4,0xD1,
	0x82D5,0x85,
	0x82D6,0x24,
	0x82D7,0x28,
	0x82D8,0x85,
	0x82D9,0x25,
	0x82DA,0x29,
	0x82DB,0x85,
	0x82DC,0x26,
	0x82DD,0x24,
	0x82DE,0x85,
	0x82DF,0x27,
	0x82E0,0x25,
	0x82E1,0x85,
	0x82E2,0x2C,
	0x82E3,0x2A,
	0x82E4,0x85,
	0x82E5,0x2D,
	0x82E6,0x2B,
	0x82E7,0x22,
	0x82E8,0x80,
	0x82E9,0x2A,
	0x82EA,0x75,
	0x82EB,0x22,
	0x82EC,0x01,
	0x82ED,0xAF,
	0x82EE,0x22,
	0x82EF,0xEF,
	0x82F0,0x33,
	0x82F1,0x95,
	0x82F2,0xE0,
	0x82F3,0xFE,
	0x82F4,0x7C,
	0x82F5,0x00,
	0x82F6,0x7D,
	0x82F7,0x14,
	0x82F8,0x12,
	0x82F9,0x0A,
	0x82FA,0x6D,
	0x82FB,0xEF,
	0x82FC,0x25,
	0x82FD,0x1A,
	0x82FE,0xF5,
	0x82FF,0x1A,
	0x8300,0xEE,
	0x8301,0x35,
	0x8302,0x19,
	0x8303,0xF5,
	0x8304,0x19,
	0x8305,0xE5,
	0x8306,0x1A,
	0x8307,0x25,
	0x8308,0x1C,
	0x8309,0xF5,
	0x830A,0x37,
	0x830B,0xE5,
	0x830C,0x19,
	0x830D,0x35,
	0x830E,0x1B,
	0x830F,0xF5,
	0x8310,0x36,
	0x8311,0x12,
	0x8312,0x09,
	0x8313,0xD1,
	0x8314,0x75,
	0x8315,0x23,
	0x8316,0x04,
	0x8317,0x22,
	0x8318,0xE5,
	0x8319,0x1A,
	0x831A,0x45,
	0x831B,0x19,
	0x831C,0x70,
	0x831D,0x07,
	0x831E,0xF5,
	0x831F,0x36,
	0x8320,0xF5,
	0x8321,0x37,
	0x8322,0x12,
	0x8323,0x09,
	0x8324,0xD1,
	0x8325,0xC2,
	0x8326,0x0C,
	0x8327,0xE4,
	0x8328,0xF5,
	0x8329,0x23,
	0x832A,0x75,
	0x832B,0x18,
	0x832C,0x04,
	0x832D,0x22,
	0x832E,0xE5,
	0x832F,0x3A,
	0x8330,0xD3,
	0x8331,0x94,
	0x8332,0x04,
	0x8333,0x40,
	0x8334,0x03,
	0x8335,0xE4,
	0x8336,0xF5,
	0x8337,0x3A,
	0x8338,0xE5,
	0x8339,0x3A,
	0x833A,0x25,
	0x833B,0xE0,
	0x833C,0x25,
	0x833D,0xE0,
	0x833E,0xF5,
	0x833F,0x3A,
	0x8340,0xE5,
	0x8341,0x37,
	0x8342,0xAE,
	0x8343,0x36,
	0x8344,0x78,
	0x8345,0x05,
	0x8346,0xCE,
	0x8347,0xC3,
	0x8348,0x13,
	0x8349,0xCE,
	0x834A,0x13,
	0x834B,0xD8,
	0x834C,0xF9,
	0x834D,0xFB,
	0x834E,0xAA,
	0x834F,0x06,
	0x8350,0x7C,
	0x8351,0x00,
	0x8352,0x7D,
	0x8353,0x03,
	0x8354,0xFF,
	0x8355,0x12,
	0x8356,0x0A,
	0x8357,0x6D,
	0x8358,0xEF,
	0x8359,0x78,
	0x835A,0x03,
	0x835B,0xCE,
	0x835C,0xC3,
	0x835D,0x13,
	0x835E,0xCE,
	0x835F,0x13,
	0x8360,0xD8,
	0x8361,0xF9,
	0x8362,0xF5,
	0x8363,0x41,
	0x8364,0xE5,
	0x8365,0x39,
	0x8366,0xAE,
	0x8367,0x38,
	0x8368,0x78,
	0x8369,0x03,
	0x836A,0xCE,
	0x836B,0xC3,
	0x836C,0x13,
	0x836D,0xCE,
	0x836E,0x13,
	0x836F,0xD8,
	0x8370,0xF9,
	0x8371,0xFF,
	0x8372,0x7C,
	0x8373,0x00,
	0x8374,0x7D,
	0x8375,0x03,
	0x8376,0x12,
	0x8377,0x0A,
	0x8378,0x6D,
	0x8379,0xEF,
	0x837A,0x78,
	0x837B,0x03,
	0x837C,0xCE,
	0x837D,0xC3,
	0x837E,0x13,
	0x837F,0xCE,
	0x8380,0x13,
	0x8381,0xD8,
	0x8382,0xF9,
	0x8383,0xF9,
	0x8384,0x7C,
	0x8385,0x00,
	0x8386,0x7D,
	0x8387,0x05,
	0x8388,0xAF,
	0x8389,0x03,
	0x838A,0xAE,
	0x838B,0x02,
	0x838C,0x12,
	0x838D,0x0A,
	0x838E,0x6D,
	0x838F,0xEF,
	0x8390,0x78,
	0x8391,0x03,
	0x8392,0xCE,
	0x8393,0xC3,
	0x8394,0x13,
	0x8395,0xCE,
	0x8396,0x13,
	0x8397,0xD8,
	0x8398,0xF9,
	0x8399,0xFB,
	0x839A,0xE5,
	0x839B,0x39,
	0x839C,0xAE,
	0x839D,0x38,
	0x839E,0x78,
	0x839F,0x03,
	0x83A0,0xCE,
	0x83A1,0xC3,
	0x83A2,0x13,
	0x83A3,0xCE,
	0x83A4,0x13,
	0x83A5,0xD8,
	0x83A6,0xF9,
	0x83A7,0xFF,
	0x83A8,0x7C,
	0x83A9,0x00,
	0x83AA,0x7D,
	0x83AB,0x05,
	0x83AC,0x12,
	0x83AD,0x0A,
	0x83AE,0x6D,
	0x83AF,0xEF,
	0x83B0,0x78,
	0x83B1,0x03,
	0x83B2,0xCE,
	0x83B3,0xC3,
	0x83B4,0x13,
	0x83B5,0xCE,
	0x83B6,0x13,
	0x83B7,0xD8,
	0x83B8,0xF9,
	0x83B9,0xFD,
	0x83BA,0xE5,
	0x83BB,0x41,
	0x83BC,0x75,
	0x83BD,0xF0,
	0x83BE,0x20,
	0x83BF,0xA4,
	0x83C0,0x85,
	0x83C1,0xF0,
	0x83C2,0x3B,
	0x83C3,0xF5,
	0x83C4,0x3C,
	0x83C5,0xE9,
	0x83C6,0x75,
	0x83C7,0xF0,
	0x83C8,0x08,
	0x83C9,0xA4,
	0x83CA,0x85,
	0x83CB,0xF0,
	0x83CC,0x3D,
	0x83CD,0xF5,
	0x83CE,0x3E,
	0x83CF,0xEB,
	0x83D0,0x75,
	0x83D1,0xF0,
	0x83D2,0x20,
	0x83D3,0xA4,
	0x83D4,0x85,
	0x83D5,0xF0,
	0x83D6,0x3F,
	0x83D7,0xF5,
	0x83D8,0x40,
	0x83D9,0xAF,
	0x83DA,0x05,
	0x83DB,0xEF,
	0x83DC,0x75,
	0x83DD,0xF0,
	0x83DE,0x08,
	0x83DF,0xA4,
	0x83E0,0xFD,
	0x83E1,0xAC,
	0x83E2,0xF0,
	0x83E3,0xE5,
	0x83E4,0x3B,
	0x83E5,0x54,
	0x83E6,0x0F,
	0x83E7,0xFB,
	0x83E8,0xE5,
	0x83E9,0x3D,
	0x83EA,0x54,
	0x83EB,0x07,
	0x83EC,0xC4,
	0x83ED,0xF8,
	0x83EE,0x54,
	0x83EF,0xF0,
	0x83F0,0xC8,
	0x83F1,0xE4,
	0x83F2,0xC4,
	0x83F3,0x54,
	0x83F4,0x0F,
	0x83F5,0x48,
	0x83F6,0x2B,
	0x83F7,0xFB,
	0x83F8,0xE5,
	0x83F9,0x3F,
	0x83FA,0x54,
	0x83FB,0x0F,
	0x83FC,0xFA,
	0x83FD,0xEC,
	0x83FE,0x54,
	0x83FF,0x07,
	0x8400,0xC4,
	0x8401,0xF8,
	0x8402,0x54,
	0x8403,0xF0,
	0x8404,0xC8,
	0x8405,0xE4,
	0x8406,0xC4,
	0x8407,0x54,
	0x8408,0x0F,
	0x8409,0x48,
	0x840A,0x2A,
	0x840B,0xFF,
	0x840C,0x90,
	0x840D,0x33,
	0x840E,0x5F,
	0x840F,0xE0,
	0x8410,0x90,
	0x8411,0x39,
	0x8412,0x0A,
	0x8413,0xB4,
	0x8414,0x68,
	0x8415,0x11,
	0x8416,0x74,
	0x8417,0x18,
	0x8418,0xF0,
	0x8419,0xA3,
	0x841A,0x74,
	0x841B,0x48,
	0x841C,0xF0,
	0x841D,0xA3,
	0x841E,0x74,
	0x841F,0x28,
	0x8420,0xF0,
	0x8421,0xA3,
	0x8422,0x74,
	0x8423,0x78,
	0x8424,0xF0,
	0x8425,0x80,
	0x8426,0x0F,
	0x8427,0x74,
	0x8428,0x0C,
	0x8429,0xF0,
	0x842A,0xA3,
	0x842B,0x74,
	0x842C,0x24,
	0x842D,0xF0,
	0x842E,0xA3,
	0x842F,0x74,
	0x8430,0x14,
	0x8431,0xF0,
	0x8432,0xA3,
	0x8433,0x74,
	0x8434,0x3C,
	0x8435,0xF0,
	0x8436,0x90,
	0x8437,0x33,
	0x8438,0xAA,
	0x8439,0xEB,
	0x843A,0xF0,
	0x843B,0xE5,
	0x843C,0x3C,
	0x843D,0xA3,
	0x843E,0xF0,
	0x843F,0xE5,
	0x8440,0x3E,
	0x8441,0xA3,
	0x8442,0xF0,
	0x8443,0xA3,
	0x8444,0xEF,
	0x8445,0xF0,
	0x8446,0xE5,
	0x8447,0x40,
	0x8448,0xA3,
	0x8449,0xF0,
	0x844A,0xED,
	0x844B,0xA3,
	0x844C,0xF0,
	0x844D,0x22,
	0x844E,0x90,
	0x844F,0x33,
	0x8450,0x62,
	0x8451,0xE0,
	0x8452,0x54,
	0x8453,0x0F,
	0x8454,0xFE,
	0x8455,0xA3,
	0x8456,0xE0,
	0x8457,0x7C,
	0x8458,0x00,
	0x8459,0x24,
	0x845A,0x00,
	0x845B,0xF5,
	0x845C,0x33,
	0x845D,0xEC,
	0x845E,0x3E,
	0x845F,0xF5,
	0x8460,0x32,
	0x8461,0x90,
	0x8462,0x33,
	0x8463,0x62,
	0x8464,0xE0,
	0x8465,0x54,
	0x8466,0x70,
	0x8467,0x75,
	0x8468,0xF0,
	0x8469,0x10,
	0x846A,0xA4,
	0x846B,0xFF,
	0x846C,0xAE,
	0x846D,0xF0,
	0x846E,0x90,
	0x846F,0x33,
	0x8470,0x64,
	0x8471,0xE0,
	0x8472,0x2F,
	0x8473,0xF5,
	0x8474,0x35,
	0x8475,0xEC,
	0x8476,0x3E,
	0x8477,0xF5,
	0x8478,0x34,
	0x8479,0xE5,
	0x847A,0x09,
	0x847B,0xB5,
	0x847C,0x33,
	0x847D,0x11,
	0x847E,0xE5,
	0x847F,0x08,
	0x8480,0xB5,
	0x8481,0x32,
	0x8482,0x0C,
	0x8483,0xE5,
	0x8484,0x0B,
	0x8485,0x65,
	0x8486,0x35,
	0x8487,0x70,
	0x8488,0x04,
	0x8489,0xE5,
	0x848A,0x0A,
	0x848B,0x65,
	0x848C,0x34,
	0x848D,0x60,
	0x848E,0x1E,
	0x848F,0x85,
	0x8490,0x32,
	0x8491,0x08,
	0x8492,0x85,
	0x8493,0x33,
	0x8494,0x09,
	0x8495,0x85,
	0x8496,0x34,
	0x8497,0x0A,
	0x8498,0x85,
	0x8499,0x35,
	0x849A,0x0B,
	0x849B,0x85,
	0x849C,0x32,
	0x849D,0x36,
	0x849E,0x85,
	0x849F,0x33,
	0x84A0,0x37,
	0x84A1,0x85,
	0x84A2,0x34,
	0x84A3,0x38,
	0x84A4,0x85,
	0x84A5,0x35,
	0x84A6,0x39,
	0x84A7,0xE4,
	0x84A8,0xF5,
	0x84A9,0x3A,
	0x84AA,0x12,
	0x84AB,0x03,
	0x84AC,0x2E,
	0x84AD,0x90,
	0x84AE,0x33,
	0x84AF,0x65,
	0x84B0,0xE0,
	0x84B1,0x30,
	0x84B2,0xE3,
	0x84B3,0x35,
	0x84B4,0x90,
	0x84B5,0x33,
	0x84B6,0xA4,
	0x84B7,0xE0,
	0x84B8,0x54,
	0x84B9,0x0F,
	0x84BA,0xFC,
	0x84BB,0xA3,
	0x84BC,0xE0,
	0x84BD,0x75,
	0x84BE,0xF0,
	0x84BF,0x04,
	0x84C0,0xA4,
	0x84C1,0xAE,
	0x84C2,0xF0,
	0x84C3,0x24,
	0x84C4,0x00,
	0x84C5,0xF5,
	0x84C6,0x33,
	0x84C7,0xEE,
	0x84C8,0x3C,
	0x84C9,0xF5,
	0x84CA,0x32,
	0x84CB,0x90,
	0x84CC,0x33,
	0x84CD,0xA4,
	0x84CE,0xE0,
	0x84CF,0x54,
	0x84D0,0x70,
	0x84D1,0x75,
	0x84D2,0xF0,
	0x84D3,0x10,
	0x84D4,0xA4,
	0x84D5,0xFD,
	0x84D6,0xAC,
	0x84D7,0xF0,
	0x84D8,0x90,
	0x84D9,0x33,
	0x84DA,0xA6,
	0x84DB,0xE0,
	0x84DC,0x75,
	0x84DD,0xF0,
	0x84DE,0x04,
	0x84DF,0xA4,
	0x84E0,0xAE,
	0x84E1,0xF0,
	0x84E2,0x2D,
	0x84E3,0xF5,
	0x84E4,0x35,
	0x84E5,0xEE,
	0x84E6,0x3C,
	0x84E7,0xF5,
	0x84E8,0x34,
	0x84E9,0xE5,
	0x84EA,0x0D,
	0x84EB,0xB5,
	0x84EC,0x33,
	0x84ED,0x11,
	0x84EE,0xE5,
	0x84EF,0x0C,
	0x84F0,0xB5,
	0x84F1,0x32,
	0x84F2,0x0C,
	0x84F3,0xE5,
	0x84F4,0x0F,
	0x84F5,0x65,
	0x84F6,0x35,
	0x84F7,0x70,
	0x84F8,0x04,
	0x84F9,0xE5,
	0x84FA,0x0E,
	0x84FB,0x65,
	0x84FC,0x34,
	0x84FD,0x60,
	0x84FE,0x1B,
	0x84FF,0x85,
	0x8500,0x32,
	0x8501,0x0C,
	0x8502,0x85,
	0x8503,0x33,
	0x8504,0x0D,
	0x8505,0x85,
	0x8506,0x34,
	0x8507,0x0E,
	0x8508,0x85,
	0x8509,0x35,
	0x850A,0x0F,
	0x850B,0x85,
	0x850C,0x32,
	0x850D,0x36,
	0x850E,0x85,
	0x850F,0x33,
	0x8510,0x37,
	0x8511,0x85,
	0x8512,0x34,
	0x8513,0x38,
	0x8514,0x85,
	0x8515,0x35,
	0x8516,0x39,
	0x8517,0x12,
	0x8518,0x05,
	0x8519,0x1B,
	0x851A,0x22,
	0x851B,0xAE,
	0x851C,0x36,
	0x851D,0xAF,
	0x851E,0x37,
	0x851F,0x7C,
	0x8520,0x00,
	0x8521,0x7D,
	0x8522,0x03,
	0x8523,0x12,
	0x8524,0x0A,
	0x8525,0x6D,
	0x8526,0xEF,
	0x8527,0x78,
	0x8528,0x03,
	0x8529,0xCE,
	0x852A,0xC3,
	0x852B,0x13,
	0x852C,0xCE,
	0x852D,0x13,
	0x852E,0xD8,
	0x852F,0xF9,
	0x8530,0xF5,
	0x8531,0x3B,
	0x8532,0x8E,
	0x8533,0x3A,
	0x8534,0xAE,
	0x8535,0x38,
	0x8536,0xAF,
	0x8537,0x39,
	0x8538,0x7C,
	0x8539,0x00,
	0x853A,0x7D,
	0x853B,0x03,
	0x853C,0x12,
	0x853D,0x0A,
	0x853E,0x6D,
	0x853F,0xEF,
	0x8540,0x78,
	0x8541,0x03,
	0x8542,0xCE,
	0x8543,0xC3,
	0x8544,0x13,
	0x8545,0xCE,
	0x8546,0x13,
	0x8547,0xD8,
	0x8548,0xF9,
	0x8549,0xF5,
	0x854A,0x3D,
	0x854B,0x8E,
	0x854C,0x3C,
	0x854D,0xAE,
	0x854E,0x36,
	0x854F,0xAF,
	0x8550,0x37,
	0x8551,0x7C,
	0x8552,0x00,
	0x8553,0x7D,
	0x8554,0x05,
	0x8555,0x12,
	0x8556,0x0A,
	0x8557,0x6D,
	0x8558,0xEF,
	0x8559,0x78,
	0x855A,0x03,
	0x855B,0xCE,
	0x855C,0xC3,
	0x855D,0x13,
	0x855E,0xCE,
	0x855F,0x13,
	0x8560,0xD8,
	0x8561,0xF9,
	0x8562,0xF5,
	0x8563,0x3F,
	0x8564,0x8E,
	0x8565,0x3E,
	0x8566,0xAE,
	0x8567,0x38,
	0x8568,0xAF,
	0x8569,0x39,
	0x856A,0x7C,
	0x856B,0x00,
	0x856C,0x7D,
	0x856D,0x05,
	0x856E,0x12,
	0x856F,0x0A,
	0x8570,0x6D,
	0x8571,0xEF,
	0x8572,0x78,
	0x8573,0x03,
	0x8574,0xCE,
	0x8575,0xC3,
	0x8576,0x13,
	0x8577,0xCE,
	0x8578,0x13,
	0x8579,0xD8,
	0x857A,0xF9,
	0x857B,0xFD,
	0x857C,0xAC,
	0x857D,0x06,
	0x857E,0xE5,
	0x857F,0x3A,
	0x8580,0x54,
	0x8581,0x0F,
	0x8582,0xFB,
	0x8583,0xE5,
	0x8584,0x3C,
	0x8585,0x54,
	0x8586,0x07,
	0x8587,0xC4,
	0x8588,0xF8,
	0x8589,0x54,
	0x858A,0xF0,
	0x858B,0xC8,
	0x858C,0xE4,
	0x858D,0xC4,
	0x858E,0x54,
	0x858F,0x0F,
	0x8590,0x48,
	0x8591,0x2B,
	0x8592,0xFB,
	0x8593,0xE5,
	0x8594,0x3E,
	0x8595,0x54,
	0x8596,0x0F,
	0x8597,0xFA,
	0x8598,0xEC,
	0x8599,0x54,
	0x859A,0x07,
	0x859B,0xC4,
	0x859C,0xF8,
	0x859D,0x54,
	0x859E,0xF0,
	0x859F,0xC8,
	0x85A0,0xE4,
	0x85A1,0xC4,
	0x85A2,0x54,
	0x85A3,0x0F,
	0x85A4,0x48,
	0x85A5,0x2A,
	0x85A6,0xFF,
	0x85A7,0x90,
	0x85A8,0x33,
	0x85A9,0xAA,
	0x85AA,0xEB,
	0x85AB,0xF0,
	0x85AC,0xE5,
	0x85AD,0x3B,
	0x85AE,0xA3,
	0x85AF,0xF0,
	0x85B0,0xE5,
	0x85B1,0x3D,
	0x85B2,0xA3,
	0x85B3,0xF0,
	0x85B4,0xA3,
	0x85B5,0xEF,
	0x85B6,0xF0,
	0x85B7,0xE5,
	0x85B8,0x3F,
	0x85B9,0xA3,
	0x85BA,0xF0,
	0x85BB,0xAF,
	0x85BC,0x05,
	0x85BD,0xEF,
	0x85BE,0xA3,
	0x85BF,0xF0,
	0x85C0,0x22,
	0x85C1,0xE5,
	0x85C2,0x17,
	0x85C3,0x70,
	0x85C4,0x0A,
	0x85C5,0xF5,
	0x85C6,0x1B,
	0x85C7,0xF5,
	0x85C8,0x1C,
	0x85C9,0x12,
	0x85CA,0x0A,
	0x85CB,0xA0,
	0x85CC,0x05,
	0x85CD,0x17,
	0x85CE,0x22,
	0x85CF,0xE5,
	0x85D0,0x17,
	0x85D1,0xC3,
	0x85D2,0x94,
	0x85D3,0x01,
	0x85D4,0x50,
	0x85D5,0x03,
	0x85D6,0x05,
	0x85D7,0x17,
	0x85D8,0x22,
	0x85D9,0xE5,
	0x85DA,0x17,
	0x85DB,0xB4,
	0x85DC,0x01,
	0x85DD,0x19,
	0x85DE,0x12,
	0x85DF,0x08,
	0x85E0,0x67,
	0x85E1,0x8E,
	0x85E2,0x15,
	0x85E3,0x8F,
	0x85E4,0x16,
	0x85E5,0x75,
	0x85E6,0x1B,
	0x85E7,0x00,
	0x85E8,0x75,
	0x85E9,0x1C,
	0x85EA,0x05,
	0x85EB,0x85,
	0x85EC,0x1B,
	0x85ED,0x36,
	0x85EE,0x85,
	0x85EF,0x1C,
	0x85F0,0x37,
	0x85F1,0x12,
	0x85F2,0x09,
	0x85F3,0xD1,
	0x85F4,0x05,
	0x85F5,0x17,
	0x85F6,0x22,
	0x85F7,0x12,
	0x85F8,0x08,
	0x85F9,0x67,
	0x85FA,0xC3,
	0x85FB,0xEF,
	0x85FC,0x95,
	0x85FD,0x16,
	0x85FE,0xFF,
	0x85FF,0xEE,
	0x8600,0x95,
	0x8601,0x15,
	0x8602,0xFE,
	0x8603,0x12,
	0x8604,0x0A,
	0x8605,0x7F,
	0x8606,0xD3,
	0x8607,0xEF,
	0x8608,0x94,
	0x8609,0x20,
	0x860A,0xEE,
	0x860B,0x64,
	0x860C,0x80,
	0x860D,0x94,
	0x860E,0x80,
	0x860F,0x40,
	0x8610,0x18,
	0x8611,0xC2,
	0x8612,0x0B,
	0x8613,0xE4,
	0x8614,0xF5,
	0x8615,0x17,
	0x8616,0x74,
	0x8617,0xFB,
	0x8618,0x25,
	0x8619,0x1C,
	0x861A,0xF5,
	0x861B,0x1C,
	0x861C,0x74,
	0x861D,0xFF,
	0x861E,0x35,
	0x861F,0x1B,
	0x8620,0xF5,
	0x8621,0x1B,
	0x8622,0xE4,
	0x8623,0xF5,
	0x8624,0x36,
	0x8625,0xF5,
	0x8626,0x37,
	0x8627,0x80,
	0x8628,0x22,
	0x8629,0xC3,
	0x862A,0xE5,
	0x862B,0x1C,
	0x862C,0x94,
	0x862D,0x96,
	0x862E,0xE5,
	0x862F,0x1B,
	0x8630,0x94,
	0x8631,0x00,
	0x8632,0x50,
	0x8633,0x1A,
	0x8634,0x74,
	0x8635,0x05,
	0x8636,0x25,
	0x8637,0x1C,
	0x8638,0xF5,
	0x8639,0x1C,
	0x863A,0xE4,
	0x863B,0x35,
	0x863C,0x1B,
	0x863D,0xF5,
	0x863E,0x1B,
	0x863F,0xE5,
	0x8640,0x1C,
	0x8641,0x24,
	0x8642,0xFF,
	0x8643,0xF5,
	0x8644,0x37,
	0x8645,0xE5,
	0x8646,0x1B,
	0x8647,0x34,
	0x8648,0xFF,
	0x8649,0xF5,
	0x864A,0x36,
	0x864B,0x02,
	0x864C,0x09,
	0x864D,0xD1,
	0x864E,0xC2,
	0x864F,0x0B,
	0x8650,0xE4,
	0x8651,0xF5,
	0x8652,0x17,
	0x8653,0xF5,
	0x8654,0x36,
	0x8655,0xF5,
	0x8656,0x37,
	0x8657,0x12,
	0x8658,0x09,
	0x8659,0xD1,
	0x865A,0x75,
	0x865B,0x1B,
	0x865C,0x00,
	0x865D,0x75,
	0x865E,0x1C,
	0x865F,0xC8,
	0x8660,0x22,
	0x8661,0x90,
	0x8662,0x3F,
	0x8663,0x00,
	0x8664,0xE0,
	0x8665,0x54,
	0x8666,0x1F,
	0x8667,0xFF,
	0x8668,0xE0,
	0x8669,0x54,
	0x866A,0xE0,
	0x866B,0xF0,
	0x866C,0xEF,
	0x866D,0x14,
	0x866E,0xB4,
	0x866F,0x09,
	0x8670,0x00,
	0x8671,0x40,
	0x8672,0x03,
	0x8673,0x02,
	0x8674,0x06,
	0x8675,0xF5,
	0x8676,0x90,
	0x8677,0x06,
	0x8678,0x84,
	0x8679,0x75,
	0x867A,0xF0,
	0x867B,0x03,
	0x867C,0xA4,
	0x867D,0xC5,
	0x867E,0x83,
	0x867F,0x25,
	0x8680,0xF0,
	0x8681,0xC5,
	0x8682,0x83,
	0x8683,0x73,
	0x8684,0x02,
	0x8685,0x06,
	0x8686,0x9F,
	0x8687,0x02,
	0x8688,0x06,
	0x8689,0xA4,
	0x868A,0x02,
	0x868B,0x06,
	0x868C,0xAA,
	0x868D,0x02,
	0x868E,0x06,
	0x868F,0xAE,
	0x8690,0x02,
	0x8691,0x06,
	0x8692,0xB3,
	0x8693,0x02,
	0x8694,0x06,
	0x8695,0xBB,
	0x8696,0x02,
	0x8697,0x06,
	0x8698,0xC3,
	0x8699,0x02,
	0x869A,0x06,
	0x869B,0xDD,
	0x869C,0x02,
	0x869D,0x06,
	0x869E,0xF1,
	0x869F,0x75,
	0x86A0,0x32,
	0x86A1,0x01,
	0x86A2,0x80,
	0x86A3,0x03,
	0x86A4,0xE4,
	0x86A5,0xF5,
	0x86A6,0x32,
	0x86A7,0x02,
	0x86A8,0x0A,
	0x86A9,0x91,
	0x86AA,0xD2,
	0x86AB,0x0D,
	0x86AC,0x80,
	0x86AD,0x02,
	0x86AE,0xC2,
	0x86AF,0x0D,
	0x86B0,0x02,
	0x86B1,0x0A,
	0x86B2,0x3B,
	0x86B3,0xC2,
	0x86B4,0x09,
	0x86B5,0x75,
	0x86B6,0x18,
	0x86B7,0x01,
	0x86B8,0xC2,
	0x86B9,0x08,
	0x86BA,0x22,
	0x86BB,0xC2,
	0x86BC,0x09,
	0x86BD,0xE4,
	0x86BE,0xF5,
	0x86BF,0x18,
	0x86C0,0xD2,
	0x86C1,0x08,
	0x86C2,0x22,
	0x86C3,0xE4,
	0x86C4,0xF5,
	0x86C5,0x18,
	0x86C6,0xC2,
	0x86C7,0x08,
	0x86C8,0xC2,
	0x86C9,0x0B,
	0x86CA,0x90,
	0x86CB,0x3F,
	0x86CC,0x05,
	0x86CD,0xF0,
	0x86CE,0xA3,
	0x86CF,0xF0,
	0x86D0,0xF5,
	0x86D1,0x36,
	0x86D2,0xF5,
	0x86D3,0x37,
	0x86D4,0x12,
	0x86D5,0x09,
	0x86D6,0xD1,
	0x86D7,0xE4,
	0x86D8,0xF5,
	0x86D9,0x19,
	0x86DA,0xF5,
	0x86DB,0x1A,
	0x86DC,0x22,
	0x86DD,0xC2,
	0x86DE,0x09,
	0x86DF,0xE4,
	0x86E0,0xF5,
	0x86E1,0x18,
	0x86E2,0xC2,
	0x86E3,0x08,
	0x86E4,0xF5,
	0x86E5,0x19,
	0x86E6,0xF5,
	0x86E7,0x1A,
	0x86E8,0x85,
	0x86E9,0x19,
	0x86EA,0x36,
	0x86EB,0x85,
	0x86EC,0x1A,
	0x86ED,0x37,
	0x86EE,0x02,
	0x86EF,0x09,
	0x86F0,0xD1,
	0x86F1,0xC2,
	0x86F2,0x09,
	0x86F3,0xD2,
	0x86F4,0x0B,
	0x86F5,0x22,
	0x86F6,0x78,
	0x86F7,0x7F,
	0x86F8,0xE4,
	0x86F9,0xF6,
	0x86FA,0xD8,
	0x86FB,0xFD,
	0x86FC,0x75,
	0x86FD,0x81,
	0x86FE,0x41,
	0x86FF,0x02,
	0x8700,0x07,
	0x8701,0x3D,
	0x8702,0x02,
	0x8703,0x09,
	0x8704,0x16,
	0x8705,0xE4,
	0x8706,0x93,
	0x8707,0xA3,
	0x8708,0xF8,
	0x8709,0xE4,
	0x870A,0x93,
	0x870B,0xA3,
	0x870C,0x40,
	0x870D,0x03,
	0x870E,0xF6,
	0x870F,0x80,
	0x8710,0x01,
	0x8711,0xF2,
	0x8712,0x08,
	0x8713,0xDF,
	0x8714,0xF4,
	0x8715,0x80,
	0x8716,0x29,
	0x8717,0xE4,
	0x8718,0x93,
	0x8719,0xA3,
	0x871A,0xF8,
	0x871B,0x54,
	0x871C,0x07,
	0x871D,0x24,
	0x871E,0x0C,
	0x871F,0xC8,
	0x8720,0xC3,
	0x8721,0x33,
	0x8722,0xC4,
	0x8723,0x54,
	0x8724,0x0F,
	0x8725,0x44,
	0x8726,0x20,
	0x8727,0xC8,
	0x8728,0x83,
	0x8729,0x40,
	0x872A,0x04,
	0x872B,0xF4,
	0x872C,0x56,
	0x872D,0x80,
	0x872E,0x01,
	0x872F,0x46,
	0x8730,0xF6,
	0x8731,0xDF,
	0x8732,0xE4,
	0x8733,0x80,
	0x8734,0x0B,
	0x8735,0x01,
	0x8736,0x02,
	0x8737,0x04,
	0x8738,0x08,
	0x8739,0x10,
	0x873A,0x20,
	0x873B,0x40,
	0x873C,0x80,
	0x873D,0x90,
	0x873E,0x00,
	0x873F,0x06,
	0x8740,0xE4,
	0x8741,0x7E,
	0x8742,0x01,
	0x8743,0x93,
	0x8744,0x60,
	0x8745,0xBC,
	0x8746,0xA3,
	0x8747,0xFF,
	0x8748,0x54,
	0x8749,0x3F,
	0x874A,0x30,
	0x874B,0xE5,
	0x874C,0x09,
	0x874D,0x54,
	0x874E,0x1F,
	0x874F,0xFE,
	0x8750,0xE4,
	0x8751,0x93,
	0x8752,0xA3,
	0x8753,0x60,
	0x8754,0x01,
	0x8755,0x0E,
	0x8756,0xCF,
	0x8757,0x54,
	0x8758,0xC0,
	0x8759,0x25,
	0x875A,0xE0,
	0x875B,0x60,
	0x875C,0xA8,
	0x875D,0x40,
	0x875E,0xB8,
	0x875F,0xE4,
	0x8760,0x93,
	0x8761,0xA3,
	0x8762,0xFA,
	0x8763,0xE4,
	0x8764,0x93,
	0x8765,0xA3,
	0x8766,0xF8,
	0x8767,0xE4,
	0x8768,0x93,
	0x8769,0xA3,
	0x876A,0xC8,
	0x876B,0xC5,
	0x876C,0x82,
	0x876D,0xC8,
	0x876E,0xCA,
	0x876F,0xC5,
	0x8770,0x83,
	0x8771,0xCA,
	0x8772,0xF0,
	0x8773,0xA3,
	0x8774,0xC8,
	0x8775,0xC5,
	0x8776,0x82,
	0x8777,0xC8,
	0x8778,0xCA,
	0x8779,0xC5,
	0x877A,0x83,
	0x877B,0xCA,
	0x877C,0xDF,
	0x877D,0xE9,
	0x877E,0xDE,
	0x877F,0xE7,
	0x8780,0x80,
	0x8781,0xBE,
	0x8782,0xE5,
	0x8783,0x18,
	0x8784,0xD3,
	0x8785,0x94,
	0x8786,0x02,
	0x8787,0x40,
	0x8788,0x03,
	0x8789,0x15,
	0x878A,0x18,
	0x878B,0x22,
	0x878C,0xE5,
	0x878D,0x18,
	0x878E,0xB4,
	0x878F,0x02,
	0x8790,0x0B,
	0x8791,0xE4,
	0x8792,0xF5,
	0x8793,0x18,
	0x8794,0x12,
	0x8795,0x08,
	0x8796,0x68,
	0x8797,0x8E,
	0x8798,0x12,
	0x8799,0x8F,
	0x879A,0x13,
	0x879B,0x22,
	0x879C,0x12,
	0x879D,0x08,
	0x879E,0x67,
	0x879F,0x8E,
	0x87A0,0x32,
	0x87A1,0x8F,
	0x87A2,0x33,
	0x87A3,0xC3,
	0x87A4,0xE5,
	0x87A5,0x33,
	0x87A6,0x95,
	0x87A7,0x13,
	0x87A8,0xFF,
	0x87A9,0xE5,
	0x87AA,0x32,
	0x87AB,0x95,
	0x87AC,0x12,
	0x87AD,0xFE,
	0x87AE,0x12,
	0x87AF,0x0A,
	0x87B0,0x7F,
	0x87B1,0x8E,
	0x87B2,0x34,
	0x87B3,0x8F,
	0x87B4,0x35,
	0x87B5,0xE5,
	0x87B6,0x10,
	0x87B7,0xD3,
	0x87B8,0x94,
	0x87B9,0x02,
	0x87BA,0x40,
	0x87BB,0x2E,
	0x87BC,0xD3,
	0x87BD,0xE5,
	0x87BE,0x35,
	0x87BF,0x94,
	0x87C0,0x40,
	0x87C1,0xE5,
	0x87C2,0x34,
	0x87C3,0x94,
	0x87C4,0x00,
	0x87C5,0x50,
	0x87C6,0x05,
	0x87C7,0x12,
	0x87C8,0x0A,
	0x87C9,0x5B,
	0x87CA,0x50,
	0x87CB,0x0A,
	0x87CC,0x85,
	0x87CD,0x32,
	0x87CE,0x12,
	0x87CF,0x85,
	0x87D0,0x33,
	0x87D1,0x13,
	0x87D2,0xE4,
	0x87D3,0xF5,
	0x87D4,0x11,
	0x87D5,0x22,
	0x87D6,0x05,
	0x87D7,0x11,
	0x87D8,0xE5,
	0x87D9,0x11,
	0x87DA,0xD3,
	0x87DB,0x94,
	0x87DC,0x02,
	0x87DD,0x40,
	0x87DE,0x21,
	0x87DF,0xE4,
	0x87E0,0xF5,
	0x87E1,0x10,
	0x87E2,0xF5,
	0x87E3,0x11,
	0x87E4,0x75,
	0x87E5,0x18,
	0x87E6,0x01,
	0x87E7,0x02,
	0x87E8,0x00,
	0x87E9,0x0E,
	0x87EA,0xD3,
	0x87EB,0xE5,
	0x87EC,0x35,
	0x87ED,0x94,
	0x87EE,0x40,
	0x87EF,0xE5,
	0x87F0,0x34,
	0x87F1,0x94,
	0x87F2,0x00,
	0x87F3,0x50,
	0x87F4,0x05,
	0x87F5,0x12,
	0x87F6,0x0A,
	0x87F7,0x5B,
	0x87F8,0x50,
	0x87F9,0x03,
	0x87FA,0x05,
	0x87FB,0x10,
	0x87FC,0x22,
	0x87FD,0xE4,
	0x87FE,0xF5,
	0x87FF,0x10,
	0x8800,0x22,
	0x8801,0x7E,
	0x8802,0x08,
	0x8803,0xE5,
	0x8804,0x3A,
	0x8805,0x33,
	0x8806,0x92,
	0x8807,0x01,
	0x8808,0x90,
	0x8809,0x30,
	0x880A,0xB4,
	0x880B,0xE5,
	0x880C,0x20,
	0x880D,0xF0,
	0x880E,0xE5,
	0x880F,0x3A,
	0x8810,0x25,
	0x8811,0xE0,
	0x8812,0xF5,
	0x8813,0x3A,
	0x8814,0xD2,
	0x8815,0x00,
	0x8816,0xE5,
	0x8817,0x20,
	0x8818,0xF0,
	0x8819,0x7F,
	0x881A,0x01,
	0x881B,0xAD,
	0x881C,0x07,
	0x881D,0x1F,
	0x881E,0xED,
	0x881F,0x70,
	0x8820,0xFA,
	0x8821,0xC2,
	0x8822,0x00,
	0x8823,0x90,
	0x8824,0x30,
	0x8825,0xB4,
	0x8826,0xE5,
	0x8827,0x20,
	0x8828,0xF0,
	0x8829,0xDE,
	0x882A,0xD8,
	0x882B,0x90,
	0x882C,0x30,
	0x882D,0xB1,
	0x882E,0xE0,
	0x882F,0x54,
	0x8830,0x7F,
	0x8831,0xF0,
	0x8832,0xD2,
	0x8833,0x00,
	0x8834,0x90,
	0x8835,0x30,
	0x8836,0xB4,
	0x8837,0xE5,
	0x8838,0x20,
	0x8839,0xF0,
	0x883A,0x7F,
	0x883B,0x01,
	0x883C,0xAE,
	0x883D,0x07,
	0x883E,0x1F,
	0x883F,0xEE,
	0x8840,0x70,
	0x8841,0xFA,
	0x8842,0x90,
	0x8843,0x30,
	0x8844,0xB5,
	0x8845,0xE0,
	0x8846,0x54,
	0x8847,0x02,
	0x8848,0xFE,
	0x8849,0xC2,
	0x884A,0x00,
	0x884B,0x90,
	0x884C,0x30,
	0x884D,0xB4,
	0x884E,0xE5,
	0x884F,0x20,
	0x8850,0xF0,
	0x8851,0x90,
	0x8852,0x30,
	0x8853,0xB1,
	0x8854,0xE0,
	0x8855,0x44,
	0x8856,0x80,
	0x8857,0xF0,
	0x8858,0x7F,
	0x8859,0x01,
	0x885A,0xAD,
	0x885B,0x07,
	0x885C,0x1F,
	0x885D,0xED,
	0x885E,0x70,
	0x885F,0xFA,
	0x8860,0xBE,
	0x8861,0x02,
	0x8862,0x02,
	0x8863,0xC3,
	0x8864,0x22,
	0x8865,0xD3,
	0x8866,0x22,
	0x8867,0xE4,
	0x8868,0xF5,
	0x8869,0x36,
	0x886A,0xF5,
	0x886B,0x37,
	0x886C,0xE5,
	0x886D,0x37,
	0x886E,0x70,
	0x886F,0x0A,
	0x8870,0xE5,
	0x8871,0x36,
	0x8872,0x75,
	0x8873,0xF0,
	0x8874,0x08,
	0x8875,0xA4,
	0x8876,0x24,
	0x8877,0x80,
	0x8878,0x80,
	0x8879,0x0E,
	0x887A,0xE5,
	0x887B,0x37,
	0x887C,0x64,
	0x887D,0x01,
	0x887E,0x70,
	0x887F,0x17,
	0x8880,0xE5,
	0x8881,0x36,
	0x8882,0x75,
	0x8883,0xF0,
	0x8884,0x08,
	0x8885,0xA4,
	0x8886,0x24,
	0x8887,0x84,
	0x8888,0x90,
	0x8889,0x3A,
	0x888A,0x00,
	0x888B,0xF0,
	0x888C,0x90,
	0x888D,0x3A,
	0x888E,0x03,
	0x888F,0xE0,
	0x8890,0xF5,
	0x8891,0x38,
	0x8892,0xE0,
	0x8893,0xFC,
	0x8894,0xE0,
	0x8895,0xFD,
	0x8896,0xE0,
	0x8897,0xED,
	0x8898,0xFE,
	0x8899,0xEC,
	0x889A,0x7A,
	0x889B,0x00,
	0x889C,0x24,
	0x889D,0x00,
	0x889E,0xFF,
	0x889F,0xEA,
	0x88A0,0x3E,
	0x88A1,0xFE,
	0x88A2,0x22,
	0x88A3,0xC0,
	0x88A4,0xE0,
	0x88A5,0xC0,
	0x88A6,0xF0,
	0x88A7,0xC0,
	0x88A8,0x83,
	0x88A9,0xC0,
	0x88AA,0x82,
	0x88AB,0xC0,
	0x88AC,0xD0,
	0x88AD,0x75,
	0x88AE,0xD0,
	0x88AF,0x00,
	0x88B0,0xC0,
	0x88B1,0x00,
	0x88B2,0xC0,
	0x88B3,0x01,
	0x88B4,0xC0,
	0x88B5,0x02,
	0x88B6,0xC0,
	0x88B7,0x03,
	0x88B8,0xC0,
	0x88B9,0x04,
	0x88BA,0xC0,
	0x88BB,0x05,
	0x88BC,0xC0,
	0x88BD,0x06,
	0x88BE,0xC0,
	0x88BF,0x07,
	0x88C0,0x12,
	0x88C1,0x0A,
	0x88C2,0x1A,
	0x88C3,0xD0,
	0x88C4,0x07,
	0x88C5,0xD0,
	0x88C6,0x06,
	0x88C7,0xD0,
	0x88C8,0x05,
	0x88C9,0xD0,
	0x88CA,0x04,
	0x88CB,0xD0,
	0x88CC,0x03,
	0x88CD,0xD0,
	0x88CE,0x02,
	0x88CF,0xD0,
	0x88D0,0x01,
	0x88D1,0xD0,
	0x88D2,0x00,
	0x88D3,0xD0,
	0x88D4,0xD0,
	0x88D5,0xD0,
	0x88D6,0x82,
	0x88D7,0xD0,
	0x88D8,0x83,
	0x88D9,0xD0,
	0x88DA,0xF0,
	0x88DB,0xD0,
	0x88DC,0xE0,
	0x88DD,0x32,
	0x88DE,0x90,
	0x88DF,0x31,
	0x88E0,0x00,
	0x88E1,0xE0,
	0x88E2,0x54,
	0x88E3,0xFE,
	0x88E4,0xF0,
	0x88E5,0xE0,
	0x88E6,0x54,
	0x88E7,0xFD,
	0x88E8,0xF0,
	0x88E9,0xA3,
	0x88EA,0xE4,
	0x88EB,0xF0,
	0x88EC,0x90,
	0x88ED,0x33,
	0x88EE,0xB0,
	0x88EF,0xF0,
	0x88F0,0xA3,
	0x88F1,0x04,
	0x88F2,0xF0,
	0x88F3,0xA3,
	0x88F4,0xF0,
	0x88F5,0x90,
	0x88F6,0x33,
	0x88F7,0x00,
	0x88F8,0x74,
	0x88F9,0x17,
	0x88FA,0xF0,
	0x88FB,0x12,
	0x88FC,0x09,
	0x88FD,0xF8,
	0x88FE,0x90,
	0x88FF,0x39,
	0x8900,0x01,
	0x8901,0x74,
	0x8902,0x34,
	0x8903,0xF0,
	0x8904,0x90,
	0x8905,0x39,
	0x8906,0x00,
	0x8907,0x74,
	0x8908,0x20,
	0x8909,0xF0,
	0x890A,0x90,
	0x890B,0x37,
	0x890C,0x00,
	0x890D,0xE0,
	0x890E,0x54,
	0x890F,0xF7,
	0x8910,0xF0,
	0x8911,0xE0,
	0x8912,0x54,
	0x8913,0xDF,
	0x8914,0xF0,
	0x8915,0x22,
	0x8916,0xC2,
	0x8917,0x8C,
	0x8918,0x75,
	0x8919,0x89,
	0x891A,0x03,
	0x891B,0x12,
	0x891C,0x08,
	0x891D,0xDE,
	0x891E,0x75,
	0x891F,0xA8,
	0x8920,0x87,
	0x8921,0x12,
	0x8922,0x0A,
	0x8923,0xA0,
	0x8924,0x12,
	0x8925,0x04,
	0x8926,0x4E,
	0x8927,0x12,
	0x8928,0x06,
	0x8929,0x61,
	0x892A,0x30,
	0x892B,0x0A,
	0x892C,0xF7,
	0x892D,0xC2,
	0x892E,0x0A,
	0x892F,0x30,
	0x8930,0x0B,
	0x8931,0x05,
	0x8932,0x12,
	0x8933,0x05,
	0x8934,0xC1,
	0x8935,0x80,
	0x8936,0xED,
	0x8937,0xE5,
	0x8938,0x18,
	0x8939,0xB4,
	0x893A,0x01,
	0x893B,0x07,
	0x893C,0x12,
	0x893D,0x00,
	0x893E,0x0E,
	0x893F,0xC2,
	0x8940,0x0A,
	0x8941,0x80,
	0x8942,0xE1,
	0x8943,0x30,
	0x8944,0x08,
	0x8945,0xDE,
	0x8946,0x12,
	0x8947,0x07,
	0x8948,0x82,
	0x8949,0x80,
	0x894A,0xD9,
	0x894B,0xC2,
	0x894C,0x01,
	0x894D,0x90,
	0x894E,0x30,
	0x894F,0xB4,
	0x8950,0xE5,
	0x8951,0x20,
	0x8952,0xF0,
	0x8953,0x7F,
	0x8954,0x01,
	0x8955,0xAE,
	0x8956,0x07,
	0x8957,0x1F,
	0x8958,0xEE,
	0x8959,0x70,
	0x895A,0xFA,
	0x895B,0xD2,
	0x895C,0x00,
	0x895D,0x90,
	0x895E,0x30,
	0x895F,0xB4,
	0x8960,0xE5,
	0x8961,0x20,
	0x8962,0xF0,
	0x8963,0x7F,
	0x8964,0x01,
	0x8965,0xAE,
	0x8966,0x07,
	0x8967,0x1F,
	0x8968,0xEE,
	0x8969,0x70,
	0x896A,0xFA,
	0x896B,0xD2,
	0x896C,0x01,
	0x896D,0x90,
	0x896E,0x30,
	0x896F,0xB4,
	0x8970,0xE5,
	0x8971,0x20,
	0x8972,0xF0,
	0x8973,0x7F,
	0x8974,0x01,
	0x8975,0xAE,
	0x8976,0x07,
	0x8977,0x1F,
	0x8978,0xEE,
	0x8979,0x70,
	0x897A,0xFA,
	0x897B,0x22,
	0x897C,0xD2,
	0x897D,0x01,
	0x897E,0xD2,
	0x897F,0x00,
	0x8980,0x90,
	0x8981,0x30,
	0x8982,0xB4,
	0x8983,0xE5,
	0x8984,0x20,
	0x8985,0xF0,
	0x8986,0x7F,
	0x8987,0x02,
	0x8988,0xAE,
	0x8989,0x07,
	0x898A,0x1F,
	0x898B,0xEE,
	0x898C,0x70,
	0x898D,0xFA,
	0x898E,0xC2,
	0x898F,0x01,
	0x8990,0x90,
	0x8991,0x30,
	0x8992,0xB4,
	0x8993,0xE5,
	0x8994,0x20,
	0x8995,0xF0,
	0x8996,0x7F,
	0x8997,0x01,
	0x8998,0xAE,
	0x8999,0x07,
	0x899A,0x1F,
	0x899B,0xEE,
	0x899C,0x70,
	0x899D,0xFA,
	0x899E,0xC2,
	0x899F,0x00,
	0x89A0,0x90,
	0x89A1,0x30,
	0x89A2,0xB4,
	0x89A3,0xE5,
	0x89A4,0x20,
	0x89A5,0xF0,
	0x89A6,0x22,
	0x89A7,0x12,
	0x89A8,0x09,
	0x89A9,0x7C,
	0x89AA,0x75,
	0x89AB,0x3A,
	0x89AC,0x18,
	0x89AD,0x12,
	0x89AE,0x08,
	0x89AF,0x01,
	0x89B0,0x40,
	0x89B1,0x02,
	0x89B2,0x80,
	0x89B3,0x14,
	0x89B4,0xE5,
	0x89B5,0x38,
	0x89B6,0xF5,
	0x89B7,0x3A,
	0x89B8,0x12,
	0x89B9,0x08,
	0x89BA,0x01,
	0x89BB,0x40,
	0x89BC,0x02,
	0x89BD,0x80,
	0x89BE,0x09,
	0x89BF,0xE5,
	0x89C0,0x39,
	0x89C1,0xF5,
	0x89C2,0x3A,
	0x89C3,0x12,
	0x89C4,0x08,
	0x89C5,0x01,
	0x89C6,0x40,
	0x89C7,0x05,
	0x89C8,0x12,
	0x89C9,0x09,
	0x89CA,0x4B,
	0x89CB,0x80,
	0x89CC,0xDA,
	0x89CD,0x12,
	0x89CE,0x09,
	0x89CF,0x4B,
	0x89D0,0x22,
	0x89D1,0xE5,
	0x89D2,0x36,
	0x89D3,0x54,
	0x89D4,0x03,
	0x89D5,0xFE,
	0x89D6,0xE5,
	0x89D7,0x37,
	0x89D8,0xC4,
	0x89D9,0xF8,
	0x89DA,0x54,
	0x89DB,0x0F,
	0x89DC,0xC8,
	0x89DD,0x68,
	0x89DE,0xFF,
	0x89DF,0xEE,
	0x89E0,0xC4,
	0x89E1,0x54,
	0x89E2,0xF0,
	0x89E3,0x48,
	0x89E4,0xFE,
	0x89E5,0xE5,
	0x89E6,0x37,
	0x89E7,0x45,
	0x89E8,0x36,
	0x89E9,0x70,
	0x89EA,0x03,
	0x89EB,0x7E,
	0x89EC,0x80,
	0x89ED,0xFF,
	0x89EE,0x43,
	0x89EF,0x07,
	0x89F0,0x08,
	0x89F1,0x8E,
	0x89F2,0x38,
	0x89F3,0x8F,
	0x89F4,0x39,
	0x89F5,0x02,
	0x89F6,0x09,
	0x89F7,0xA7,
	0x89F8,0x90,
	0x89F9,0x30,
	0x89FA,0xB2,
	0x89FB,0xE0,
	0x89FC,0x44,
	0x89FD,0x18,
	0x89FE,0xF0,
	0x89FF,0x90,
	0x8A00,0x30,
	0x8A01,0xB0,
	0x8A02,0xE0,
	0x8A03,0x44,
	0x8A04,0x03,
	0x8A05,0xF0,
	0x8A06,0xA3,
	0x8A07,0xE0,
	0x8A08,0x44,
	0x8A09,0xC0,
	0x8A0A,0xF0,
	0x8A0B,0x90,
	0x8A0C,0x30,
	0x8A0D,0xB4,
	0x8A0E,0xE0,
	0x8A0F,0x44,
	0x8A10,0x0F,
	0x8A11,0xF0,
	0x8A12,0xE0,
	0x8A13,0x54,
	0x8A14,0xF7,
	0x8A15,0xF0,
	0x8A16,0xE0,
	0x8A17,0xF5,
	0x8A18,0x20,
	0x8A19,0x22,
	0x8A1A,0x90,
	0x8A1B,0x37,
	0x8A1C,0x08,
	0x8A1D,0xE0,
	0x8A1E,0x30,
	0x8A1F,0xE5,
	0x8A20,0x0D,
	0x8A21,0x90,
	0x8A22,0x3F,
	0x8A23,0x01,
	0x8A24,0xE0,
	0x8A25,0x04,
	0x8A26,0xF0,
	0x8A27,0x90,
	0x8A28,0x37,
	0x8A29,0x08,
	0x8A2A,0x74,
	0x8A2B,0x20,
	0x8A2C,0xF0,
	0x8A2D,0x22,
	0x8A2E,0x90,
	0x8A2F,0x37,
	0x8A30,0x08,
	0x8A31,0xE0,
	0x8A32,0x30,
	0x8A33,0xE3,
	0x8A34,0x05,
	0x8A35,0x74,
	0x8A36,0x08,
	0x8A37,0xF0,
	0x8A38,0xD2,
	0x8A39,0x0A,
	0x8A3A,0x22,
	0x8A3B,0x90,
	0x8A3C,0x30,
	0x8A3D,0x14,
	0x8A3E,0xE0,
	0x8A3F,0x30,
	0x8A40,0x0D,
	0x8A41,0x10,
	0x8A42,0x54,
	0x8A43,0xF7,
	0x8A44,0xF0,
	0x8A45,0xA3,
	0x8A46,0xE0,
	0x8A47,0x54,
	0x8A48,0xBF,
	0x8A49,0xF0,
	0x8A4A,0x90,
	0x8A4B,0x30,
	0x8A4C,0x2D,
	0x8A4D,0xE4,
	0x8A4E,0xF0,
	0x8A4F,0xA3,
	0x8A50,0xF0,
	0x8A51,0x22,
	0x8A52,0x44,
	0x8A53,0x80,
	0x8A54,0xF0,
	0x8A55,0xA3,
	0x8A56,0xE0,
	0x8A57,0x44,
	0x8A58,0x40,
	0x8A59,0xF0,
	0x8A5A,0x22,
	0x8A5B,0x90,
	0x8A5C,0x30,
	0x8A5D,0x1B,
	0x8A5E,0xE0,
	0x8A5F,0xFF,
	0x8A60,0x65,
	0x8A61,0x14,
	0x8A62,0x70,
	0x8A63,0x01,
	0x8A64,0xE4,
	0x8A65,0x60,
	0x8A66,0x04,
	0x8A67,0x8F,
	0x8A68,0x14,
	0x8A69,0xD3,
	0x8A6A,0x22,
	0x8A6B,0xC3,
	0x8A6C,0x22,
	0x8A6D,0xEF,
	0x8A6E,0x8D,
	0x8A6F,0xF0,
	0x8A70,0xA4,
	0x8A71,0xA8,
	0x8A72,0xF0,
	0x8A73,0xCF,
	0x8A74,0x8C,
	0x8A75,0xF0,
	0x8A76,0xA4,
	0x8A77,0x28,
	0x8A78,0xCE,
	0x8A79,0x8D,
	0x8A7A,0xF0,
	0x8A7B,0xA4,
	0x8A7C,0x2E,
	0x8A7D,0xFE,
	0x8A7E,0x22,
	0x8A7F,0xC3,
	0x8A80,0xEE,
	0x8A81,0x64,
	0x8A82,0x80,
	0x8A83,0x94,
	0x8A84,0x80,
	0x8A85,0x40,
	0x8A86,0x02,
	0x8A87,0x80,
	0x8A88,0x07,
	0x8A89,0xC3,
	0x8A8A,0xE4,
	0x8A8B,0x9F,
	0x8A8C,0xFF,
	0x8A8D,0xE4,
	0x8A8E,0x9E,
	0x8A8F,0xFE,
	0x8A90,0x22,
	0x8A91,0xE5,
	0x8A92,0x32,
	0x8A93,0x90,
	0x8A94,0x33,
	0x8A95,0x00,
	0x8A96,0x60,
	0x8A97,0x04,
	0x8A98,0x74,
	0x8A99,0x57,
	0x8A9A,0xF0,
	0x8A9B,0x22,
	0x8A9C,0x74,
	0x8A9D,0x17,
	0x8A9E,0xF0,
	0x8A9F,0x22,
	0x8AA0,0xE4,
	0x8AA1,0xF5,
	0x8AA2,0x19,
	0x8AA3,0xF5,
	0x8AA4,0x1A,
	0x8AA5,0x85,
	0x8AA6,0x19,
	0x8AA7,0x36,
	0x8AA8,0x85,
	0x8AA9,0x1A,
	0x8AAA,0x37,
	0x8AAB,0x02,
	0x8AAC,0x09,
	0x8AAD,0xD1,
	0x8AAE,0x32,
	0x3104,0x00,
	0x3f00  ,0x05,
	OV3640_REGEND, 0x00	    /* End of list delimiter */
};

int ov3640_record_previewcode(void)
{
	u8 reg;
	write_sensor_reg(0x33a0,0);//awb r
	read_sensor_reg(0x33ca,&ov3640_recRgain);

	write_sensor_reg(0x33a0,1);//awb g
	read_sensor_reg(0x33ca,&ov3640_recGgain);


	
	write_sensor_reg(0x33a0,2);//awb b
	read_sensor_reg(0x33ca,&ov3640_recBgain);	

	read_sensor_reg(0x3002,&reg);	
	ov3640_expcode = (int)reg;
	ov3640_expcode <<= 8;
	read_sensor_reg(0x3003,&reg);	
	ov3640_expcode += reg;
	read_sensor_reg(0x3001,&ov3640_gaincode);
	return 0;
}

int ov3640_sensor_irq_subfunc(unsigned int int_state)
{	
	u8 CurveValue0, CurveValue1, CurveValue2, CurveValue3;
	/*****read antishake value*****/
	write_sensor_reg(0x3a00, 0x80);
	read_sensor_reg(0x3a03,&CurveValue0);				
	read_sensor_reg(0x3a03,&CurveValue1);				
	read_sensor_reg(0x3a03,&CurveValue2);
	read_sensor_reg(0x3a03,&CurveValue3);	

	return ((CurveValue2 << 8) + CurveValue1);		
}

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_on(pdata->id);

}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_off(pdata->id);
}

void ov3640hw_power_down( u8 powerMode )
{
	/* OV3640 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();

	mdelay(100);
}

void ov3640hw_reset(void)
{
	printk("sensor reset\n");
	write_sensor_reg(0x3012,0x80);	
	printk("reset over\n");
	mdelay(50);
	return;
}

int ov3640_set_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != OV3640_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		status = (write_sensor_reg( regP[curReg << 1], regVal) == 0) ?
			0 : -EIO;
		curReg++;
	}

	return status;
}

int ov3640_sensor_reg_init(int format)
{	
	switch(format)
	{
	case 0:
		ov3640_set_regs(QVGA_yuv_init_tab);
		printk("ov3640 set QVGA YUV setting\n");
		break;
	case 1:
		ov3640_set_regs(VGA_yuv_init_tab);
		printk("ov3640 set VGA YUV setting\n");
		break;
        case 2:
                ov3640_set_regs(QXGA_yuv_init_tab);
                printk("ov3640 set QXGA YUV setting\n");
		break;
	case 3:
		ov3640_set_regs(QXGA_jpg_init_tab);
		printk("ov3640 set QXGA JPEG setting\n");
		break;
	case 4:
		ov3640_set_regs(AF_init_tab);
		break;
	case 5:
		ov3640hw_reset( );
		break;
	default:
		printk("No such format\n");
		return -1;
	}
	return 0;
}
  int ov3640_sensor_preview2snap(void)
{
	// qxga->1568 xga->779
	u8 reg;
	/* write awb*/
	write_sensor_reg(0x33a7,ov3640_recRgain);
	write_sensor_reg(0x33a8,ov3640_recGgain);
	write_sensor_reg(0x33a9,ov3640_recBgain);
	/*stop auto awb*/
	read_sensor_reg(0x332b,&reg);
	write_sensor_reg(0x332b,reg|0x08); 
	/*stop AGC*/
	read_sensor_reg(0x3013,&reg);
	write_sensor_reg(0x3013,reg&0xfa);
	/*write gain*/
	write_sensor_reg(0x3001,ov3640_gaincode);
	/*write exposure line */
	reg = ov3640_expcode &0xff;
	write_sensor_reg(0x3003,reg);
	ov3640_expcode >>= 8;
	reg = ov3640_expcode & 0xff;
	write_sensor_reg(0x3002,reg);
	printk("gain code =%x\n",ov3640_gaincode);
	printk("ov3640_sensor_preview2snap\n");
	return 0;
}
 int ov3640_sensor_snap2preview(void)
{
	u8 reg;
       /*write back gain code*/
	write_sensor_reg(0x3001,ov3640_gaincode);
	 /*write back exposure line*/
	reg = ov3640_expcode &0xff;
	write_sensor_reg(0x3003,reg);
	ov3640_expcode >>= 8;
	reg = ov3640_expcode & 0xff;
	write_sensor_reg(0x3002,reg);
	/*enable AGC*/
	read_sensor_reg(0x3013,&reg);
	write_sensor_reg(0x3013,reg|0x5);
	/*write back awb*/
	write_sensor_reg(0x33a0,0);//awb r
	write_sensor_reg(0x33ca,ov3640_recRgain);

	write_sensor_reg(0x33a0,1);//awb g
	write_sensor_reg(0x33ca,ov3640_recGgain);


	
	write_sensor_reg(0x33a0,2);//awb b
	write_sensor_reg(0x33ca,ov3640_recBgain);	
	/* enable awb*/
	read_sensor_reg(0x332b,&reg);
	write_sensor_reg(0x332b,reg&0xf7);
	printk("ov3640_sensor_snap2preview\n");
	return 0;
}


int ov3640_sensorID_detect(void)
{
	u8 val_h,val_l;
	read_sensor_reg(PID_ADRRH, &val_h);
	read_sensor_reg(PID_ADRRL, &val_l);
	
	printk("val_h=%x, val_l=%x\n", val_h, val_l);
	
	if( (val_h == PID_VERH ))
		return 0;
	else
		return -1;
}



int ov3640hw_read_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

int ov3640hw_write_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return write_sensor_reg(subAddress, *bufP);
}


/*****************************************************************************
 *  Settings
 *****************************************************************************/
int ov3640hw_version_revision(u8 * pCmRevision, u8 *pSensorRevision)
{
	read_sensor_reg( PID_ADRRH, pCmRevision );
	read_sensor_reg( PID_ADRRL, pSensorRevision );
	return 0;
}

void ov3640hw_set_hsync(void)
{

}

void ov3640hw_auto_function_on(void)
{
	
}

void ov3640hw_auto_function_off(void)
{
	
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
int ov3640hw_viewfinder_on(void)
{


	return 0;
}


int ov3640hw_viewfinder_off(void)
{
	

	return 0;
}


/*****************************************************************************
 *  Format
 *****************************************************************************/
int ov3640hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{
//	ov3640_record_previewcode();
	switch(colorFormat)
	{
	case OV3640_FORMAT_YUV_422:
	{
		if(captureSizeFormat == OV3640_SIZE_QVGA){
//			ov3640_sensor_reg_init(5);
			ov3640_sensor_reg_init(0);			
		} else if (captureSizeFormat == OV3640_SIZE_VGA) {
			ov3640_sensor_reg_init(1);
		} else if (captureSizeFormat == OV3640_SIZE_QXGA) {
			ov3640_sensor_reg_init(2);
		} else
			return -EPERM;
	}
		break;
	case OV3640_FORMAT_JPG:
	{
		ov3640_sensor_reg_init(3);
	}
		break;
	default:
		return -EPERM;
		break;
	}

	printk("wait focus-----------\n");
	ov3640_sensor_reg_init(4);//af function
/*
	printk("G_ov3640_mode_change=%d, mode=%d\n", G_ov3640_mode_change, mode);	
	if(G_ov3640_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			ov3640_sensor_preview2snap();
		else
			ov3640_sensor_snap2preview();
		G_ov3640_mode_change = mode;
	}

*/
	return 0;
}


int ov3640hw_set_contrast(u32 value)
{
	
	return 0;
}



int ov3640hw_set_exposure(u32 value)
{

	return 0;
}

/*****************************************************************************
 * Auto White Balance
 *****************************************************************************/


int ov3640hw_set_white_balance(u32 value)
{
	
	return 0;
}


/*****************************************************************************
 * OV3640 Functions
 *****************************************************************************/
int ov3640_init( p_camera_context_t camera_context )
{
	u8 cm_rev, cm_pid;
	u32 timeout;
	int status;

	printk("ov3640_init\n");

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to OV3640's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);
	
	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	/* software work around*/
	ci_set_clock(1, 1, CICLK);

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* OV3640 Power on sequence
	 * Take out of Power down mode (GPIO_103), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	ov3640hw_power_down(CAMERA_POWER_FULL);
	ov3640hw_reset();

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = cm_rev = 0;
		status = ov3640hw_version_revision(&cm_pid, &cm_rev);

		/* Check to make sure we are working with an OV3640 */
		if( cm_pid != PID_VERH  )
		{
			ov3640hw_power_down( CAMERA_POWER_OFF );
			ov3640hw_power_down( CAMERA_POWER_FULL );
			ov3640hw_reset();
			mdelay(1);
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_VERH );

	/* turn sensor output off */
	ov3640hw_viewfinder_off();
	
	return 0;
}

int ov3640_deinit(p_camera_context_t camera_context)
{
	/* power off the external module */

	ov3640hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

int ov3640_sleep(  p_camera_context_t camera_context )
{
	return ov3640_deinit( camera_context );
}

int ov3640_wake(  p_camera_context_t camera_context )
{
	return ov3640_init( camera_context );
}

int ov3640_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 ovSizeFormat, ovFormat;
	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
			printk("para defaut yuv 422\n");
			ovFormat = OV3640_FORMAT_YUV_422;
			break;
			
		case V4L2_PIX_FMT_SRGGB10:
			ovFormat = OV3640_FORMAT_RAW10;		
			break;
			
		case V4L2_PIX_FMT_SRGGB8:
			ovFormat = OV3640_FORMAT_RAW8;		
			break;

			
		case V4L2_PIX_FMT_JPEG:
			ovFormat = OV3640_FORMAT_JPG;
			printk("para OV3640_FORMAT_JPG\n");
			break;
		
		case V4L2_PIX_FMT_RGB565X:
			 ovFormat = OV3640_FORMAT_RGB565;
		
			 break;
		default:
			ovFormat = OV3640_FORMAT_YUV_422;
			printk("para defaut yuv 422\n");
			break;
	}

	printk("camera_context->capture_input_width=%d\n", camera_context->capture_input_width);
	printk("camera_context->capture_input_height=%d\n", camera_context->capture_input_height);
	
		
	if ((camera_context->capture_input_width <= OV3640_PREVIEW_WIDTH)
                        && (camera_context->capture_input_height <= OV3640_PREVIEW_HEIGHT ))
	{
		ovSizeFormat = OV3640_SIZE_QVGA;
	} else if ((camera_context->capture_input_width <= 640)
                        && (camera_context->capture_input_height <= 480 ))
	{
		ovSizeFormat = OV3640_SIZE_VGA;	
	} else {
		ovSizeFormat = OV3640_SIZE_QXGA;
	}
/*
	if((camera_context->capture_input_width > OV3640_PREVIEW_WIDTH) 
			&& (camera_context->capture_input_height > OV3640_PREVIEW_HEIGHT ))
	{
		ovSizeFormat = OV3640_SIZE_QXGA;
	}else{
		ovSizeFormat = OV3640_SIZE_QVGA;	
	}
*/
	status = ov3640hw_set_format( ovSizeFormat,ovFormat, camera_context->capture_mode);

	timing.BFW = 0x00;	/* FIXME: green line */
	timing.BLW = 0x00;
	
	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);
	return 0;
}

int ov3640_start_capture(p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;

	return 0;
}

int ov3640_stop_capture(  p_camera_context_t camera_context )
{
	(void) camera_context;

	return 0;
}

int ov3640_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	ov3640hw_power_down( mode );
	return 0;
}

int ov3640_read_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;

	ov3640hw_read_sensor_reg( reg_addr, (u8 *)reg_val );
	return 0;
}

int ov3640_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	u8 buffer;
	int status;

	(void) camera_context;

	buffer = (u8)reg_val;
	status = (ov3640hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

int ov3640_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	(void) camera_context;

		return 0;
}

int ov3640_set_exposure( p_camera_context_t camera_context, u8 mode, u32 value)
{
	(void) camera_context;

		return 0;
}

int ov3640_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
	(void) camera_context;

		return 0;
}

int ov3640_get_framerate(u32 format, u32 width, u32 height, u32 *numerator, u32 *denominator)
{
	int ret = 0;
	switch (format) {
		case V4L2_PIX_FMT_SRGGB10:
			if ((width <= 800) && (height <= 600)) {
				/* CIF */
				*numerator = 15;	
				*denominator = 1;
			} else if ((width <= 1600) && (height <= 1200)) {
				/* UXGA */
				*numerator = 15;	
				*denominator = 2;
				
			} else if ((width <= 2560) && (height <= 1920)) {
				/* FULL */
				*numerator = 5;	
				*denominator = 2;				
			}else{
				ret = -1;
			}
			break;
			
		default:
			ret = -1;
			break;
		}

	return ret;
}

static char name[] = "Omnivision3640";

static p_camera_function_t camera_functions;

static int format_list[] = {

	V4L2_PIX_FMT_YUV422P,
	//V4L2_PIX_FMT_SRGGB10,

	/* to be realized later */
	/*V4L2_PIX_FMT_SRGGB8,*/
	/*V4L2_PIX_FMT_RGB565X,*/

	V4L2_PIX_FMT_JPEG, 
	
	-1	
};

int ov3640_register(int id)
{

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max = 2048;
	camera_functions->height_max = 1536;
	camera_functions->width_min = 2;
	camera_functions->height_min = 2;

	camera_functions->v_power = 0; //2800;
	camera_functions->v_io = 0; //1800;
		

	camera_functions->format_list = format_list;

	camera_functions->init = ov3640_init;     	
	camera_functions->deinit = ov3640_deinit;     
	camera_functions->set_capture_format = ov3640_set_capture_format; 	
	camera_functions->start_capture =	ov3640_start_capture;	 		
	camera_functions->stop_capture =	ov3640_stop_capture; 			
	camera_functions->sleep = ov3640_sleep;	
	camera_functions->wakeup = ov3640_wake;	
	camera_functions->read_8bit = NULL;			
	camera_functions->write_8bit = NULL;			
	camera_functions->read_16bit = ov3640_read_16bit;		
	camera_functions->write_16bit = ov3640_write_16bit;		
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode = ov3640_set_power_mode;			
	camera_functions->set_contrast =	ov3640_set_contrast;			
	camera_functions->set_whitebalance =	ov3640_set_white_balance;		
	camera_functions->set_exposure =	 ov3640_set_exposure;			
	camera_functions->set_zoom = NULL;

	camera_functions->get_framerate = ov3640_get_framerate;
	
	camera_functions->name = name;
	camera_functions->id = id;

	/*register 3640 as high resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

int ov3640_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *                              OV3640 I2C Client Driver
 *****************************************************************************/
int read_sensor_reg_16bit(u16 addr, u8 *pvalue)
{    
	u8 data;        
	u8 address[2];
	if( g_client == NULL )	/* Not initialize the client */	
	{
		ovI2C_dbg("not initialize the client\n");
		return -1;
	}
	address[0] = addr>>8;	
	address[1] = addr;
	i2c_smbus_write_byte_data(g_client,address[0],address[1]);
	data = i2c_smbus_read_byte(g_client);
	*pvalue = data;   

	return 0;
}
int write_sensor_reg_16bit(u16 addr, u8 *value)
{
	u8 data[3];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;			
	data[0] = addr>>8;
	data[1] = addr;	
	data[2]=  *value;
	

	i2c_master_send(g_client, data, 3);
	
	return 0;
}

int write_sensor_reg_16bit_m(u16 addr, u8 value, u8 mask)
{
	int ret = 0;
	u8 regval;
	
	read_sensor_reg_16bit(addr, &regval);
	
	regval  &=  (~mask);
	value   &=  mask;
	
	regval|=value;
	
	ret = write_sensor_reg_16bit(addr,&regval);

	return ret;
}

int write_sensor_reg(u16 addr, u8 value)
{
	return write_sensor_reg_16bit(addr, &value);	
}

int read_sensor_reg(u16 addr, u8 *pvalue)
{
	return read_sensor_reg_16bit(addr,pvalue);
}


static int __devinit sensor_probe(struct i2c_client *client,
	const struct i2c_device_id *_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	int id = 0;
	
	pdata = client->dev.platform_data;
	id = pdata->id;

	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		printk("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);
	
	ci_set_clock(1, 1, CICLK);
	
	g_client = client;

	pdata->power_on(id);

	mdelay(1);
	
	if (ov3640_sensorID_detect()<0){
		goto err;
	}

	if(ov3640_register(id)<0)
		goto err;


	printk("OV3640 detected.\n");
	
	ci_set_clock(0, 0, CICLK);
	pdata->power_off(id);

	clk_disable(clk);
	clk_put(clk);

	return 0;
err:
	printk("OV3640 detect failed.\n");
	ci_set_clock(0, 0, CICLK);
	pdata->power_off(id);

	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	ov3640_unregister(pdata->id);
	g_client = NULL;
	return 0;
}

static const struct i2c_device_id sensor_ov3640_id[] = {
	{ "sensor_ov3640", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "sensor_ov3640",
	},
	.probe		= sensor_probe,
	.remove		= sensor_remove,
	.id_table	= sensor_ov3640_id,
};

static int __init sensor_init(void)
{
	return i2c_add_driver(&sensor_driver);
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}

module_init(sensor_init);
module_exit(sensor_exit);

MODULE_DESCRIPTION("OV3643 I2C Client driver");
MODULE_LICENSE("GPL");
