
/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */
/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for OV7690
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include "camera.h"
#include <linux/delay.h>

#include <mach/mfp.h>
#include <mach/hardware.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <mach/zylonite.h>

#include <linux/i2c.h>
#include <mach/gpio.h>
#include <mach/camera.h>

/*****************************************************************************
 * register
 *****************************************************************************/
#define CICLK	2400


/* Revision constants */

#define PID_OV76XX			0x76
#define PID_7690			0x91

/* Output Size & Format */
#define OV_SIZE_QQVGA			0x01
#define OV_SIZE_QVGA			( OV_SIZE_QQVGA << 1 )
#define OV_SIZE_VGA				( OV_SIZE_QQVGA << 2 )
#define OV_SIZE_QQCIF			0x10
#define OV_SIZE_QCIF			( OV_SIZE_QQCIF << 1 )
#define OV_SIZE_CIF				( OV_SIZE_QQCIF << 2 )
#define OV_FORMAT_YUV_422		1
#define OV_FORMAT_RGB_565		2
#define OV_FORMAT_RAW8   		3
#define OV_FORMAT_JPEG   		4

/* OV7690 Register Definitions */
#define OV7690_PID		0x000A	/* Product ID */
#define OV7690_VER		0x000B	/* Version */
#define OV7690_REG0C	0x0c
#define OV7690_REG12	0x12
#define OV7690_REG13	0x13
#define OV7690_REG81	0x81
#define OV7690_REGDC	0xdc
#define OV7690_REG_PWD  0x49
#define OV7690_PWD_1_8  0x0C
/* End of OV7690 register */
#define OV7690_REGEND		( 0xE1 + 1 )

/* Unique ID allocation */
static struct i2c_client *g_client;

/*****************************************************************************
 *  Register Settings
 *****************************************************************************/
const static u8 gYUV_INITIALIZE[] =
{
  0x12, 0x80,
  0x0c, 0x46,
  0x48, 0x42,
  0x41, 0x43,
  0x4c, 0x73,
  0x81, 0xef,
  0x21, 0x44,
  0x16, 0x03,
  0x39, 0x80,
  0x1e, 0xb1,
  0x12, 0x00,
  0x82, 0x03,
  0xd0, 0x48,
  0x80, 0x7f,
  0x3e, 0x30,
  0x22, 0x00,
  0x17, 0x69,
  0x18, 0xa4,
  0x19, 0x08,
  0x1a, 0xf6,
  0xc8, 0x02,
  0xc9, 0x80,
  0xca, 0x01,
  0xcb, 0xe1,
  0xcc, 0x02,
  0xcd, 0x80,
  0xce, 0x01,
  0xcf, 0xe1,
  0x85, 0x90,
  0x86, 0x18,
  0x87, 0x00,
  0x88, 0x10,
  0x89, 0x18,
  0x8a, 0x10,
  0x8b, 0x14,
  0xBB, 0x20,
  0xBC, 0x40,
  0xBD, 0x60,
  0xBE, 0x58,
  0xBF, 0x48,
  0xC0, 0x10,
  0xC1, 0x33,
  0xc2, 0x02,
  0xb7, 0x02,
  0xb8, 0x0b,
  0xb9, 0x00,
  0xba, 0x18,
  0x5A, 0x4A,
  0x5B, 0x9F,
  0x5C, 0x48,
  0x5d, 0x32,
  0x24, 0x88,
  0x25, 0x78,
  0x26, 0xb3,
  0xa3, 0x10,
  0xa4, 0x1c,
  0xa5, 0x30,
  0xa6, 0x58,
  0xa7, 0x68,
  0xa8, 0x76,
  0xa9, 0x81,
  0xaa, 0x8a,
  0xab, 0x94,
  0xac, 0x9c,
  0xad, 0xa8,
  0xae, 0xb4,
  0xaf, 0xc6,
  0xb0, 0xd7,
  0xb1, 0xe8,
  0xb2, 0x20,
  0x8e, 0x92,
  0x96, 0xff,
  0x97, 0x00,
  0x50, 0x4d,
  0x51, 0x3f,
  0x21, 0x57,
  0x20, 0x00,
  0x14, 0x49,
  0x13, 0xf7,
  0x11, 0x01,
  0x68, 0xb0,
  0x42, 0x0d,
  //0x11, 0x00,
  //0x29, 0x50,
  //0x2a, 0x30,
  //0x2b, 0x08,
  //0x2c, 0x2b,
  //0x15, 0x00,
  //0x2d, 0x00,
  //0x2e, 0x00,
  0x49, 0x0C,
  
  0x80, 0x7f,
  0x85, 0x90,
  0x86, 0x00,
  0x87, 0x00,
  0x88, 0x10,
  0x89, 0x30,
  0x8a, 0x29,
  0x8b, 0x26,
  
	OV7690_REGEND, 0x00	    /* End of list delimiter */
};

const static u8 gYUV_QCIF_INITIALIZE[] =
{
	0x16, 0x03,
  0x17, 0x83,
  0x18, 0x97,
  0x19, 0x08,
  0x1a, 0xf6,
  0x22, 0x10,
  0xc8, 0x02,
  0xc9, 0x4c,
  0xca, 0x01,
  0xcb, 0xe2,
  0xcc, 0x00,
  0xcd, 0xb0,
  0xce, 0x00,
  0xcf, 0x91,
	OV7690_REGEND, 0x00     /* End of list delimiter */
};


const static u8 gYUV_CIF_INITIALIZE[] =
{
	0x16, 0x03,
  0x17, 0x83,
  0x18, 0x97,
  0x19, 0x08,
  0x1a, 0xf6,
  0x22, 0x00,
  0xc8, 0x02,
  0xc9, 0x4c,
  0xca, 0x01,
  0xcb, 0xe2,
  0xcc, 0x01,
  0xcd, 0x60,
  0xce, 0x01,
  0xcf, 0x21,
	OV7690_REGEND, 0x00	    /* End of list delimiter */
};

const static u8 gYUV_QVGA_INITIALIZE[] =
{
	0x16, 0x03,
  0x17, 0x69,
  0x18, 0xa4,
  0x19, 0x08,
  0x1a, 0xf6,
  0x22, 0x10,
  0xc8, 0x02,
  0xc9, 0x80,
  0xca, 0x01,
  0xcb, 0xe2,
  0xcc, 0x01,
  0xcd, 0x40,
  0xce, 0x00,
  0xcf, 0xf1,
	OV7690_REGEND, 0x00	    /* End of list delimiter */
}; 
const static u8 gYUV_VGA_INITIALIZE[] =
{
	0x16, 0x03,
	0x17, 0x69,
  0x18, 0xa4,
  0x19, 0x0a/*0x0c*/,
  0x1a, 0xf6,
  0x22, 0x00,
  0xc8, 0x02,
  0xc9, 0x80,
  0xca, 0x01,
  0xcb, 0xe1,
  0xcc, 0x02,
  0xcd, 0x80,
  0xce, 0x01,
  0xcf, 0xe1,
	OV7690_REGEND, 0x00	    /* End of list delimiter */
};
//------------<<
const static u8 gSensorSlaveAddr = 0x21;
static int read_sensor_reg( const u8 subAddress, u8 *bufP );
static int write_sensor_reg( const u8 subAddress, u8 *bufP );

struct OV7690_SETTING
{
	int   format;
	int   resolution;
	const u8 *setting;
};

struct OV7690_SETTING ov7690_setting_table[] =
{
	{OV_FORMAT_YUV_422, OV_SIZE_QVGA , gYUV_QVGA_INITIALIZE},
	{OV_FORMAT_YUV_422, OV_SIZE_VGA  , gYUV_VGA_INITIALIZE},
	{OV_FORMAT_YUV_422, OV_SIZE_QCIF , gYUV_QCIF_INITIALIZE},
	{OV_FORMAT_YUV_422, OV_SIZE_CIF  , gYUV_CIF_INITIALIZE},
	{0, 0, 0}
} ;



/*****************************************************************************
 *  Sensor read/write
 *****************************************************************************/
static int ov7690hw_read_sensor_reg( const u8 subAddress, u8 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

static int ov7690hw_write_sensor_reg( const u8 subAddress, u8 *bufP )
{
	return write_sensor_reg(subAddress, bufP);
}

static int ov7690hw_set_regs( const u8 *regP )
{
	u32	curReg = 0;
	int	status = 0;

	/* The list is a register number followed by the value */
	while( regP[curReg << 1] < OV7690_REGEND )
	{
		u8 regVal = regP[(curReg << 1) + 1];

		status = (write_sensor_reg( regP[curReg << 1], &regVal ) == 0) ?
			0 : -EIO;
		if( curReg == 0 )
			mdelay(5);

		curReg++;
	}

	return status;
}

static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_on(pdata->id);

}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_off(pdata->id);
}

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void ov7690hw_power_down( u8 powerMode )
{
	/* OV7690 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();

//	mdelay(100);
}

/*****************************************************************************
 *  Settings
 *****************************************************************************/
static int ov7690hw_version_revision(u8 * pCmRevision, u8 *pSensorRevision)
{
	read_sensor_reg( OV7690_PID, pCmRevision );
	read_sensor_reg( OV7690_VER, pSensorRevision );
	return 0;
}


static void ov7690hw_auto_function_on(void)
{
	u8 val;
	read_sensor_reg( OV7690_REG13, &val );
	val |= 0x07;	/* don't disturb AWB */
	write_sensor_reg( OV7690_REG13, &val );
}

static void ov7690hw_auto_function_off(void)
{
	u8 val;
	read_sensor_reg( OV7690_REG13, &val );
	val &= ~0x07;	/* don't disturb AWB */
	write_sensor_reg( OV7690_REG13, &val );
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
static int ov7690hw_viewfinder_on(void)
{
	u8	com3;

	read_sensor_reg( OV7690_REG0C, &com3 );
	com3 &= ~0x01;
	write_sensor_reg( OV7690_REG0C, &com3 );

	return 0;
}


static int ov7690hw_viewfinder_off(void)
{
	u8	com3;

	read_sensor_reg( OV7690_REG0C, &com3 );
	com3 |= 0x01;
	write_sensor_reg( OV7690_REG0C, &com3 );

	return 0;
}
/*****************************************************************************
 *  Format, ov7690 only support YUV422 now,RAW and RGB have not been implemented
 *****************************************************************************/
static int ov7690hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{

	int	status;
	u8	*regsP;
	const u8	*defaultDataP;
	struct OV7690_SETTING  *ov7690_setting;
	defaultDataP  = NULL;
	for(ov7690_setting = ov7690_setting_table;
		ov7690_setting->format != 0;ov7690_setting++) {
		if ((colorFormat == ov7690_setting->format) &&
			(captureSizeFormat == ov7690_setting->resolution)) {
			defaultDataP = ov7690_setting->setting;
			break;
		}
	}

	if (defaultDataP == NULL)
		return -EPERM;

	/* Get the pointer to the basic setting.
	 * The pointer must be freed after exiting.
	 */
	regsP = (u8 *) defaultDataP;

	/* Blast the entire parameter tree into the part */
	status = ov7690hw_set_regs( regsP );

	return 0;
}

/*****************************************************************************
 * Contrast
 *****************************************************************************/
const static u8 ContrastLowestSettings[] = {
	0x81, 0xff,
	0xd5, 0x20,
	0xd4, 0x10,
	0xd3, 0xd0,
	0xd2, 0x04,
	0xdc, 0x01,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 ContrastLowSettings[] = {
	0x81, 0xff,
	0xd5, 0x20,
	0xd4, 0x18,
	0xd3, 0x48,
	0xd2, 0x04,	
	0xdc, 0x01,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};
//workaroud, the midddle is same as the defualt
const static u8 ContrastMiddleSettings[] = {
	0x81, 0xff,
	0x81, 0xef,
	0xd5, 0x00,
	0xd4, 0x20,
	0xd3, 0x00,
	0xD2, 0x00,
	0xdc, 0x01,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 ContrastHighSettings[] = {
	0x81, 0xff,
	0xd5, 0x20,
	0xd4, 0x28,
	0xd3, 0x00,
	0xd2, 0x04,		
	0xdc, 0x01,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 ContrastHighestSettings[] = {
	0x81, 0xff,
	0xd5, 0x20,
	0xd4, 0x30,
	0xd3, 0x00,
	0xd2, 0x04,
	0xdc, 0x01,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

static int ov7690hw_set_contrast(u32 value)
{
	const u8 *regP;
	u8	val;

	regP = NULL;
	switch(value) {
		case SENSOR_CONTRAST_LOWEST:
			regP = ContrastLowestSettings;
			break;
		case SENSOR_CONTRAST_LOW:
			regP = ContrastLowSettings;
			break;
		case SENSOR_CONTRAST_MIDDLE:
			regP = ContrastMiddleSettings;
			break;
		case SENSOR_CONTRAST_HIGH:
			regP = ContrastHighSettings;
			break;
		case SENSOR_CONTRAST_HIGHEST:
			regP = ContrastHighestSettings;
			break;
		default:
			regP = ContrastMiddleSettings;
			break;
	}

	if (regP)
		ov7690hw_set_regs(regP);

	printk(KERN_INFO "VGA Camera contrast is %d now\n",value);
	return 0;
}

/*****************************************************************************
 * Exposure
 *****************************************************************************/
const static u8 ExposureSettings[] = {
	0x50, 0x40, 0x81, /* EV-2 */
	0x70, 0x60, 0x92, /* EV-1 */
	0x90, 0x80, 0xb4, /* EV0 */
	0xa8, 0xa0, 0xc4, /* EV+1 */
	0xc0, 0xb8, 0xf8, /* EV+2 */
};

static int ov7690hw_set_exposure(u32 value)
{
	u8 aew, aeb, vpt;
	int index = -1;

	switch(value) {
		case SENSOR_EXPOSURE_LOWEST:
			index = 0;
			break;
		case SENSOR_EXPOSURE_LOW:
			index = 3;
			break;
		case SENSOR_EXPOSURE_MIDDLE:
			index = 6;
			break;
		case SENSOR_EXPOSURE_HIGH:
			index = 9;
			break;
		case SENSOR_EXPOSURE_HIGHEST:
			index = 12;
			break;
		default:
			break;
	}

	aew = aeb = vpt = 0;

	if (index != -1){
		aew = ExposureSettings[index];
		aeb = ExposureSettings[index + 1];
		vpt  = ExposureSettings[index + 2];
	}

	/* set hw */
	if (aew || aeb || vpt)
	{
		ov7690hw_write_sensor_reg( 0x24, &aew);
		ov7690hw_write_sensor_reg( 0x25, &aeb);
		ov7690hw_write_sensor_reg( 0x26, &vpt);
	}

	return 0;
}
/**ev compesate**/
static int ov7690_set_ev_offset( p_camera_context_t camera_context, int value)
{
    if (value==0)		
        ov7690hw_set_exposure(SENSOR_EXPOSURE_MIDDLE);
    else if (value>0) {
        if (value<=33)
            ov7690hw_set_exposure(SENSOR_EXPOSURE_HIGH);
        else
            ov7690hw_set_exposure(SENSOR_EXPOSURE_HIGHEST);
    }
    else {
        if (value>=-33)
            ov7690hw_set_exposure(SENSOR_EXPOSURE_LOW);
        else	
            ov7690hw_set_exposure(SENSOR_EXPOSURE_LOWEST);
    }
    printk(KERN_INFO "VGA Camera ev_offset is %d now\n",value);
    return 0;
}

/*****************************************************************************
 * Auto White Balance
 *****************************************************************************/
const static u8 AWBAuto[] = {
	0x13, 0xf7,
	0x15, 0x00,	
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 AWBFluorescent[] = {
	0x13, 0xf5,
	0x01, 0x5c,
	0x02, 0x56,
	0x15, 0x00,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 AWBOutdoor[] = {
	0x13, 0xf5,
	0x01, 0x58,
	0x02, 0x56,
	0x15, 0x00,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};

const static u8 AWBIncandescent[] = {
	0x13, 0xf5,
	0x01, 0x67,
	0x02, 0x40,
	0x15, 0x00,
	OV7690_REGEND, 	0x00		/* End of list delimiter */
};


static int ov7690hw_set_white_balance(u32 value)
{
	const u8 *regP;

	regP = NULL;
	switch(value) {
		case SENSOR_WHITEBALANCE_AUTO:		/* Auto */
			regP = AWBAuto;
			break;
		case SENSOR_WHITEBALANCE_INCANDESCENT:	/* Incandescent */
			regP = AWBIncandescent;
			break;
		case SENSOR_WHITEBALANCE_SUNNY:		/* Sunny */
			regP = AWBOutdoor;
			break;
		case SENSOR_WHITEBALANCE_FLUORESCENT:	/* Fluorescent */
			regP = AWBFluorescent;
			break;
		default:
			break;
	}

	/* set hw */
	if (regP) {
		ov7690hw_set_regs(regP);
	}
	return 0;
}
/*****************************************************************************
 * OV7690 Functions
 *****************************************************************************/
static int ov7690_init( p_camera_context_t camera_context )
{
	u8 cm_rev, cm_pid, regValue; 
	u32 timeout;
	int status;

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to OV7690's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);

	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	ci_set_clock(1, 1, CICLK);
	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* OV7690 Power on sequence
	 * Take out of Power down mode (GPIO_103), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	ov7690hw_power_down(CAMERA_POWER_FULL );

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = cm_rev = 0;
		status = ov7690hw_version_revision(&cm_pid, &cm_rev);

		/* Check to make sure we are working with an OV7690 */
		if( cm_pid != PID_OV76XX || cm_rev != PID_7690 )
		{
			ov7690hw_power_down( CAMERA_POWER_OFF );
			ov7690hw_power_down( CAMERA_POWER_FULL );
			mdelay(1);
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_OV76XX );
	/* reset */
	regValue = 0x80;
	write_sensor_reg(OV7690_REG12, &regValue);
    mdelay(3);
    status = ov7690hw_set_regs(  gYUV_INITIALIZE );
	/* turn sensor output off */
	ov7690hw_viewfinder_off();

	return 0;
}

static int ov7690_deinit(  p_camera_context_t camera_context )
{
	/* power off the external module */
	ov7690hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

static int ov7690_sleep(  p_camera_context_t camera_context )
{
	return ov7690_deinit( camera_context );
}

static int ov7690_wake(  p_camera_context_t camera_context )
{
	return ov7690_init( camera_context );
}

static int ov7690_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 ovSizeFormat, ovFormat;

	/* Set the current mode */
	switch(camera_context->capture_input_format) {
		case V4L2_PIX_FMT_YUV422P:
			ovFormat = OV_FORMAT_YUV_422;
			break;
		case V4L2_PIX_FMT_RGB565X:
			ovFormat = OV_FORMAT_RGB_565;
			break;
		case V4L2_PIX_FMT_SRGGB8:
			ovFormat = OV_FORMAT_RAW8;
			break;
		case V4L2_PIX_FMT_JPEG:
			ovFormat = OV_FORMAT_JPEG;
			break;
		default:
			ovFormat = OV_FORMAT_YUV_422;
			break;
	}
	if ( camera_context->capture_input_width == 88 &&
			camera_context->capture_input_height == 72 )
		ovSizeFormat = OV_SIZE_QQCIF;
	else if ( camera_context->capture_input_width == 176 &&
			camera_context->capture_input_height == 144 )
		ovSizeFormat = OV_SIZE_QCIF;
	else if ( camera_context->capture_input_width == 352 &&
			camera_context->capture_input_height == 288 )
		ovSizeFormat = OV_SIZE_CIF;
	else if ( camera_context->capture_input_width == 160 &&
			camera_context->capture_input_height == 120 )
		ovSizeFormat = OV_SIZE_QQVGA;
	else if ( camera_context->capture_input_width == 320 &&
			camera_context->capture_input_height == 240 )
		ovSizeFormat = OV_SIZE_QVGA;
	else if ( camera_context->capture_input_width == 640 &&
			camera_context->capture_input_height == 480 )
		ovSizeFormat = OV_SIZE_VGA;
	else
		ovSizeFormat = OV_SIZE_VGA;

	if(ovFormat != OV_FORMAT_YUV_422){
		printk(KERN_ERR "err capture format,ov7690 only support YUV422 now\n");
		return -EPERM;
	}
	status = ov7690hw_set_format( ovSizeFormat,
			ovFormat, camera_context->capture_mode);

	/* set capture width/height and timing */
	if (ovFormat == OV_FORMAT_RAW8) {
		timing.BFW = 0x01;
		timing.BLW = 0x01; /*GRBG to RGGB*/
	} else {
		timing.BFW = 0x01;
		timing.BLW = 0x00;
	}
	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);

	if (cpu_is_pxa310())
		ci_set_ycbcr_420_down_sample (camera_context->ycbcr_ds);

	return 0;
}

static int ov7690_start_capture(p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;

	/* turn auto function on only doing continues capture */
	if (frames == 0) {
		ov7690hw_auto_function_on();
	} else {
		//ov7690hw_auto_function_on();
		ov7690hw_auto_function_off();
	}

	/* turn viewfinder on */
	ov7690hw_viewfinder_on();
	return 0;
}

static int ov7690_stop_capture(  p_camera_context_t camera_context )
{
	(void) camera_context;

	/* turn auto function off */
	ov7690hw_auto_function_off();

	/* turn viewfinder off */
	ov7690hw_viewfinder_off();

	return 0;
}

static int ov7690_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	ov7690hw_power_down( mode );
	return 0;
}

static int ov7690_read_8bit( p_camera_context_t camera_context,
		u8 reg_addr,  u8 *reg_val)
{
	(void) camera_context;

	ov7690hw_read_sensor_reg( reg_addr, reg_val );
	return 0;
}

static int ov7690_write_8bit( p_camera_context_t camera_context,
		u8 reg_addr,  u8 reg_val)
{
	u8 buffer;
	int status;

	(void) camera_context;

	buffer = reg_val;
	status = (ov7690hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

static int ov7690_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	(void) camera_context;

	if (mode == SENSOR_MANUAL_CONTRAST)
		return ov7690hw_set_contrast(value);
	else
		return 0;
}

static int ov7690_set_exposure( p_camera_context_t camera_context, u8 mode, u32 value)
{
	(void) camera_context;

	if (mode == SENSOR_MANUAL_EXPOSURE)
		return ov7690hw_set_exposure(value);
	else
		return 0;
}

static int ov7690_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
	(void) camera_context;

	if (mode == SENSOR_MANUAL_WHITEBALANCE)
		return ov7690hw_set_white_balance(value);
	else
		return 0;
}

static int ov7690_get_framerate(u32 format, u32 width, u32 height, u32 *numerator, u32 *denominator)
{
	int ret = 0;
	switch (format) {
		case V4L2_PIX_FMT_YUV422P:
		case V4L2_PIX_FMT_RGB565X:
		case V4L2_PIX_FMT_SRGGB8:
		case V4L2_PIX_FMT_SRGGB10:
			if ((width <= 640) && (height <= 480)) {
				
				*numerator = 15;	
				*denominator = 2;				
			}else{
				ret = -1;
			}
			break;		
			
		default:
			ret = -1;
			break;
		}

	return ret;
}

static char name[] = "Omnivision7690";

static int format_list[] = {
	V4L2_PIX_FMT_YUV422P,
	V4L2_PIX_FMT_RGB565X,
	V4L2_PIX_FMT_SRGGB8,
	-1	
};

static p_camera_function_t camera_functions;

static int ov7690_register(int id)
{

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->format_list = format_list;

	camera_functions->width_max = 640;
	camera_functions->height_max = 480;
	camera_functions->width_min = 88;
	camera_functions->height_min = 72;

	camera_functions->v_power = 2500;
	camera_functions->v_io = 1800;
	
	camera_functions->init = ov7690_init;     	
	camera_functions->deinit = ov7690_deinit;     
	camera_functions->set_capture_format = ov7690_set_capture_format; 	
	camera_functions->start_capture =	ov7690_start_capture;	 		
	camera_functions->stop_capture =	ov7690_stop_capture; 			
	camera_functions->sleep = ov7690_sleep;	
	camera_functions->wakeup = ov7690_wake;	
	camera_functions->read_8bit = ov7690_read_8bit;			
	camera_functions->write_8bit = ov7690_write_8bit;			
	camera_functions->read_16bit = NULL;		
	camera_functions->write_16bit = NULL;		
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode = ov7690_set_power_mode;			
	camera_functions->set_contrast =	ov7690_set_contrast;			
	camera_functions->set_whitebalance =	ov7690_set_white_balance;		
	camera_functions->set_exposure =	 ov7690_set_exposure;			
    camera_functions->set_ev_offset	= ov7690_set_ev_offset;
	camera_functions->set_zoom = NULL;

	camera_functions->get_framerate = ov7690_get_framerate;
	camera_functions->name = name;
	camera_functions->id = id;

	/*register 7690 as LOW resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static int ov7690_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}


/*****************************************************************************
 *                              OV7690 I2C Client Driver
 *****************************************************************************/
//#define	OV7690_ADDRESS	0x21
static int read_sensor_reg( const u8 subAddress, u8 *bufP )
{
	int ret = 0;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;
	
	ret = i2c_smbus_write_byte(g_client, subAddress);
	if(ret){
		printk("sensor i2c read error\n");
		printk("ret = %x,chip addr = %x,subaddr = %x,g_client->name = %s\n",ret,g_client->addr,subAddress,g_client->name);
		return ret;
	}

	ret = i2c_smbus_read_byte(g_client);
	if (ret >= 0) {
		*bufP = ret;
	}
	return ret;
}

static int write_sensor_reg( const u8 subAddress, u8 *bufP )
{
	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	return i2c_smbus_write_byte_data(g_client, subAddress, *bufP);
}
static int i2c_read(struct i2c_client *client, u8 reg)
{
	int err = 0;

	err = i2c_smbus_write_byte(client, reg);
	if(err){
		printk("sensor i2c read error\n");
		printk("err = %x,chip addr = %x,reg = %x,client->name = %s\n",err,client->addr,reg,client->name);
		return err;
	}
		
	return i2c_smbus_read_byte(client);
}
static int __devinit sensor_probe(struct i2c_client *client, const struct i2c_device_id *dev_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	u8 pidh = 0, pidl = 0;
	int id = 0;
	int temp;

	pdata = client->dev.platform_data;
	id = pdata->id;
	
	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		printk("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);

	ci_set_clock(1, 1, CICLK);
	
	pdata->power_on(id);
	//mdelay(1);
	
	pidh = i2c_read(client, OV7690_PID);
	pidl = i2c_read(client, OV7690_VER);
	printk("ov7690 pidh = %x,pidl = %x\n",pidh,pidl);
	if ((pidh != PID_OV76XX) || (pidl != PID_7690)){
		goto err;
	}
    /*
       Before OV7690 power down, 0x49 should be written with different values
       according to DOVDD voltage. Please see OV5642_OVM7690 Dual Camera Module
       Application Notes_R1.0.pdf 2.2.1 step 5.
       */
    temp = i2c_read(client, OV7690_REG_PWD);
    temp &= 0xf0;
    temp |= OV7690_PWD_1_8;
    write_sensor_reg(0x49,&temp);

	if(ov7690_register(id)<0)
		goto err;

	g_client = client;

	printk("OV7690 detected.\n");
	
	ci_set_clock(0, 0, CICLK);
    pdata->power_off(id);

	clk_disable(clk);
	clk_put(clk);

	return 0;
err:
	printk("OV7690 detect failed.\n");
	ci_set_clock(0, 0, CICLK);
	pdata->power_off(id);
	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	ov7690_unregister(pdata->id);
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "sensor_ov7690", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "sensor_ov7690",
	},
	.probe		= sensor_probe,
	.remove		= sensor_remove,
	.id_table       = sensor_id,
};

static int __init sensor_init(void)
{
	return i2c_add_driver(&sensor_driver);
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}
MODULE_DESCRIPTION("OV7690 I2C Client driver");
MODULE_LICENSE("GPL");

module_init(sensor_init);
module_exit(sensor_exit);
