/***************************************************************************
 ***
 ***     Copyright 2006 Foxconn Electronics Inc.
 ***     All Rights Reserved
 ***     No portions of this material may be reproduced in any form
 ***     without the written permission of:
 ***
 ***                      Foxconn Electronics Inc.
 ***
 ***     All information contained in this document is Foxconn Electronics
 ***     Corporation company private, proprietary, and trade secret.
 ***
 ****************************************************************************
 ***
 ***    FILE NAME :  siv120a.h
 ***
 ***    DESCRIPTION :
 ***     This file contains the header file for SIV120A driver
 ***
 ***    HISTORY :
 ***     Jenn-Dar 2009/3/18 file create
 ***
 ***************************************************************************/

#ifndef __SIV120_H
#define __SIV120_H

/*==========================================================================
 *
 *      Library Inclusion Segment
 *
 *==========================================================================
 */

/*==========================================================================
 *
 *      Constant Definition Segment
 *
 *==========================================================================
 */
#define FXCI_CAMERA_SIV_I2C_READ        0xA
#define FXCI_CAMEAR_SIV_I2C_WRITE       0xB
#define FXCI_CAMERA_EXPOSURE_SET        0xC
#define FXCI_CAMERA_WHITE_BALANCE_SET   0xD

/*siv120 chip id & version*/
#define SIV120A_CHIP_ID		0x12
#define SIV120A_CHIP_VER	0x10
#define SIV120B_CHIP_ID     	0x12
#define SIV120B_CHIP_VER    	0x11

#define SIV120_REG_END      	0xFF

#define BLK_SEL 0x00
#define CICLK   2600        	/* 26 Mhz */

#define CTRL_BLK    0
#define AE_BLK      1
#define AWB_BLK     2
#define IDP_BLK     3

#define SIV120A_REG_END 	0xFF
#define SIV120_CHIP_ID 		SIV120A_CHIP_ID

typedef struct {
	unsigned char addr;
	unsigned char val;
} S_REG_SET;

struct siv120hw_data {
	int ae_mode;
	int wb_mode;
	int fps;
};

struct sivCDev {
	struct cdev cdev;
};

typedef struct {
	u8  exposureMode;
	u32 exposureValue;
	u8  whiteBalanceMode;
	u32 whiteBalanceValue;
} CAMERA_SETTINGS;

typedef struct {
	u8 addr;
	u8 val;
} S_INIT_TBL;

struct siv120_reg_config {
	const S_REG_SET *initTable;
	const S_REG_SET *autoExposureTable;
	const S_REG_SET *exposureLowestTable;
	const S_REG_SET *exposureLowTable;
	const S_REG_SET *exposureMiddleTable;
	const S_REG_SET *exposureHighTable;
	const S_REG_SET *exposureHighestTable;
	const S_REG_SET *autoWBTable;
	const S_REG_SET *wbIncandescentTable;
	const S_REG_SET *wbSunnyTable;
	const S_REG_SET *wbFluorescentTable;
	const S_REG_SET *wbCloudyTable;
	const S_REG_SET *outFmtYUV422Table;
	const S_REG_SET *outFmtRGB565Table;
	const S_REG_SET *outSize640_480_Table;
	const S_REG_SET *outSize352_288_Table;
	const S_REG_SET *outSize320_240_Table;
	const S_REG_SET *outSize176_144_Table;
	const S_REG_SET *outSize160_120_Table;
	const S_REG_SET *outSize88_72_Table;
	const S_REG_SET *stillModeTable;
	const S_REG_SET *opadTable;
};

#endif 
