#include <linux/workqueue.h>
#include <linux/platform_device.h>
#include <mach/micco.h>
#include <mach/micco_vibrator.h>
#include <linux/wakelock.h>

static struct wake_lock vibrator_suspend_lock;
struct micco_vibrator_platform_data *pdata;
extern int micco_set_vibrator(unsigned char value);

#ifdef CONFIG_SYSFS 

#define	VIBON 	 -1
#define	VIBOFF 	 0


#define MAX_TIMEOUT 10  /* Max vibrate time is 10secs */
#define VIBRATORING  1

#define BASE 	2
#define BIGGER	3
#define BIGGEST	4

/*******************************
 value -> 0:off
 value ->0:time_out vibrate value
 add a wakelock
********************************/

struct vibrator {
	struct delayed_work dwork;		/* the delay work to close vibrator */
	int over; 				/* vibrator status 1:vibrating 0:vibrator is over */
	long int val;				/* current vibrator value */
	struct device	*vibrator_dev;
};

static struct vibrator micco_vibrator;


#ifdef CONFIG_VIBRATOR_TIMEOUT

static void vibrator_set_resume (void)
{
	micco_vibrator.over = !VIBRATORING;
}

static void vibrator_off(struct work_struct *work)
{
	micco_set_vibrator(VIBOFF);
	vibrator_set_resume();
}

/* save current vibrator time and value,used for next vibrator valid check */
static void update(struct vibrator *vb)
{
	if(!vb->over) {
     		micco_set_vibrator(pdata->vib_strength);
		vb->over = VIBRATORING;
		schedule_delayed_work(&vb->dwork, msecs_to_jiffies(vb->val));
	} else {
		cancel_delayed_work_sync(&vb->dwork);
		schedule_delayed_work(&vb->dwork, msecs_to_jiffies(vb->val));
	}

	wake_lock_timeout(&vibrator_suspend_lock, msecs_to_jiffies(vb->val));
}

static void vibrator_init(struct vibrator *vb)
{
	INIT_DELAYED_WORK(&vb->dwork, vibrator_off);
	vb->over = !VIBRATORING;
}

#else
#define vibrator_set_resume() NULL 
#endif	/* CONFIG_VIBRATOR_TIMEOUT */
 
static inline void vibrator_set_status(long value)
{
	printk(KERN_DEBUG "------>>>>>>>>>>the vibrator value = %ld\n", value);
	if (value == VIBOFF) {
		micco_set_vibrator(VIBOFF);
		vibrator_set_resume();
	}
	else if (value == VIBON) {
		if (micco_vibrator.over == 1)
			cancel_delayed_work_sync(&micco_vibrator.dwork);
		vibrator_set_resume();
		micco_set_vibrator(pdata->vib_strength);
	}
#ifdef CONFIG_VIBRATOR_TIMEOUT
	else if (value > 0) {
		if(value/1000 > MAX_TIMEOUT)
			value = MAX_TIMEOUT * 1000;
		micco_vibrator.val = value;
		update(&micco_vibrator);
	} 
#if 0	
	else if (value < -1) {
		value = -value;
		switch (value) {
			case BASE	:
				pdata->vib_strength = (0x33 << 1);
				break;
			case BIGGER	:
				pdata->vib_strength = (0xAA << 1);
				break;
			case BIGGEST	:
				pdata->vib_strength = (0xFF << 1);
				break;

			default:	break;
		}
	}
#endif
#endif
	else 
		printk( KERN_INFO "vibrator getting error data");
}

static ssize_t vibrator_status_store(struct device *dev,
			struct device_attribute *attr, 
			const char *buf, size_t size)
{
	int value;
	sscanf(buf, "%d", &value);
	vibrator_set_status(value);
	return size;
}

static DEVICE_ATTR(enable, 0644, NULL, vibrator_status_store);

struct class output_class = {
        .name		= "timed_output",
};

static int creat_vibrator_sysfs_file(void)
{
	int err;

	err = class_register(&output_class);
	if (err) 
	{
		printk(KERN_ERR "timed_output: unable to register timed_output class\n");
		return err;
	}
	micco_vibrator.vibrator_dev = device_create(&output_class, NULL, 0, "%s", "vibrator");
	err = device_create_file(micco_vibrator.vibrator_dev, &dev_attr_enable);
	if (err)
	{
		device_unregister(micco_vibrator.vibrator_dev);
		return err;
	}
	return 0;
}

static void remove_vibrator_sysfs_file(void)
{
	device_remove_file(micco_vibrator.vibrator_dev, &dev_attr_enable);
	device_unregister(micco_vibrator.vibrator_dev);
	class_unregister(&output_class);
}

#endif

static int pxa_vibrator_probe(struct platform_device *pdev)
{

	pdata = pdev->dev.platform_data;

	if(!pdata)
		printk(KERN_INFO "pdata null error");

	wake_lock_init(&vibrator_suspend_lock, WAKE_LOCK_SUSPEND, "vibrator-lock");

#ifdef 	CONFIG_SYSFS 
	creat_vibrator_sysfs_file();
#endif

#ifdef CONFIG_VIBRATOR_TIMEOUT
	vibrator_init(&micco_vibrator);
#endif
	printk(KERN_INFO ">>>>>Vibrator register ok\n");
	return 0;
}

static struct platform_driver pxa_vibrator_drv = {
	.driver = {
		.name = "micco-vibrator",
	},
	.probe = pxa_vibrator_probe,
};


static int __init pxa_vibrator_init(void)
{
	return platform_driver_register(&pxa_vibrator_drv);
}

static void __exit pxa_vibrator_exit(void)
{
#ifdef 	CONFIG_SYSFS 
	remove_vibrator_sysfs_file();
#endif
	platform_driver_unregister(&pxa_vibrator_drv);
}

module_init(pxa_vibrator_init);
module_exit(pxa_vibrator_exit);

MODULE_DESCRIPTION("Vibrator Controller Driver");
MODULE_LICENSE("GPL");
