/*
 *  drivers/misc/misc_pxaregs.c
 *
 *  Misc driver for TCMD read/write tavor registers
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 * (C) Copyright 2008 Borqs Ltd.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <mach/mfp.h>
#include <mach/gpio.h>
#include <asm/uaccess.h>
#include <asm/io.h>


#define	PXAREG_MIN_ADDR  0x40000000
#define	PXAREG_MAX_ADDR  0x58100000
#define PXAREG_MAX_SIZE	 0x00001000

static int phys_addr = PXAREG_MIN_ADDR;
static int addr_size = 4;

static ssize_t read_pxaregs(struct kobject *kobj,
			    struct bin_attribute *attr, 
			    char *buf, loff_t off, size_t count)
{
	char *kbuf;
	void *map_addr;
	int buf_count;
	int kbuf_offset;
	int reg_val;


	map_addr = ioremap(phys_addr, addr_size);
	if( !map_addr )
	{
		printk("PXAREG: Error remap regs\n");
		return -EFAULT;
	}

	kbuf=kmalloc((addr_size>>2)*10,GFP_KERNEL);
	if(!kbuf)
	{
		iounmap(map_addr);
		return -ENOMEM;
	}

	kbuf_offset=0;
	for(buf_count=0;buf_count<addr_size;buf_count+=4) 
	{ 
		reg_val=__raw_readl(map_addr+buf_count);
		kbuf_offset += sprintf(kbuf+kbuf_offset,"%8.8x ",reg_val);
		if( ((buf_count>>2)%8) == 7 )
			kbuf_offset += sprintf(kbuf+kbuf_offset,"\n");
	}
	if( off > kbuf_offset )
		buf_count=0;
	else
		buf_count= min(count,(unsigned int)(kbuf_offset-(int)off));

	memcpy(buf, kbuf+off,buf_count);
	kfree(kbuf);
	iounmap(map_addr);
	return buf_count;
}

static ssize_t write_pxaregs(struct kobject *kobj,
			      struct bin_attribute *attr,
			      char *buf, loff_t off, size_t count)
{
	char messages[256];
	void *map_addr;
	int reg_val,ret;

	if( count>255)
		count=255;

	memcpy(messages, buf, count);

	messages[count]='\0';
	if ('-' == messages[0]) {
		/* set the register index */
		ret = sscanf(messages+1,"%x %x", &phys_addr, &addr_size );
		if(ret == 1){
			addr_size = 4;
		}else if( ret == 2){
		}else{
			printk("phys_addr error format!\n");
			return -EFAULT;
		}
		if( (phys_addr & 0x3) != 0 )
			phys_addr = (phys_addr+3) & 0xfffffffc;
		if( (addr_size & 0x3) != 0 )
			addr_size = (addr_size+3) & 0xfffffffc;
		if( addr_size > PXAREG_MAX_SIZE )
			addr_size = PXAREG_MAX_SIZE;
		if ( (phys_addr + addr_size) > PXAREG_MAX_ADDR )
			addr_size = PXAREG_MAX_ADDR - phys_addr;

		printk("phys_addr = 0x%8.8x, size=0x%8.8x\n",phys_addr, addr_size);
	} else {
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		printk("set reg[0x%8.8x]=0x%8.8x\n", phys_addr, reg_val);
		map_addr = ioremap(phys_addr, 4);
		if( !map_addr )
		{
			printk("PXAREG: Error remap regs\n");
			return -EFAULT;
		}
		__raw_writel(reg_val, map_addr);
		iounmap(map_addr);
	}
	return count;
}

static struct bin_attribute misc_pxaregs_attr = {
	.attr = {
		.name = "pxa-regs",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUSR,
	},
	.read = read_pxaregs,
	.write = write_pxaregs,
};

static int __init misc_pxaregs_init(void)
{
	int ret;

	ret = sysfs_create_bin_file(kernel_kobj, &misc_pxaregs_attr);
	if (ret) {
		printk(KERN_ERR "PXAREGS: Failed to create sys file\n");
		return -ENOMEM;
	}

	return 0;
}

static void __exit misc_pxaregs_exit(void)
{
	sysfs_remove_bin_file(kernel_kobj, &misc_pxaregs_attr);
}

module_init(misc_pxaregs_init);
module_exit(misc_pxaregs_exit);

MODULE_AUTHOR("Borqs Ltd.");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("PXAREGS SYSFS Interface");
