/*
 *  drivers/misc/sirf.h
 *
 *  Supports for SiRF AGPS chip - GSC3LT
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 * (C) Copyright 2008 Borqs Ltd.
 */

#ifndef __SIRF_H
#define __SIRF_H

#define SIRF_MODE_SIZE 4

#ifdef CONFIG_SIRF_GSC3LT
/* GSC3LT */
/* download firmware to flash */
#define FLASH_DOWN_MODE 0
/* run firmware from flash */
#define FLASH_RUN_MODE 1
/* run firmware from ROM */
#define ROM_RUN_MODE 2
/* stopped */
#define STOP_MODE 3

/* hibernation */
#define HIBERNATION 0
/* full power */
#define FULL_POWER 1
/* power off */
#define POWER_OFF 2

#elif CONFIG_SIRF_GSD3TW
/* GSD3tw */
#define RUN_MODE 1
#define STOP_MODE 3

#define FULL_POWER 1
/* power off */
#define POWER_OFF 2
#endif

/* chip operation */
#define OP_POWER_ON 100
#define OP_POWER_OFF 101
#define OP_RESET 102
#define OP_STATUS_QUERY 103
#define OP_PM 104
#define OP_ONOFF 104
#define OP_FLASH_DOWN 105
#define OP_FLASH_RUN 106
#define OP_ROM_RUN 107
#define OP_CLK_ENABLE 200
#define OP_CLK_DISABLE 201

#ifdef CONFIG_HAS_WAKELOCK
#define OP_PM_LOCK 300
#define OP_PM_UNLOCK 301
#endif

#define GPIO_LEVEL_LOW	0
#define GPIO_LEVEL_HIGH	1

/* define change GPIO status macro */
#define GPIO_SET_VALUE(pin, signal)         gpio_set_value(mfp_to_gpio(pin), signal)
#define GPIO_GET_VALUE(pin)                 gpio_get_value(mfp_to_gpio(pin))
#define GPIO_DIRECTION_INPUT(pin)           gpio_direction_input(mfp_to_gpio(pin))
#define GPIO_DIRECTION_OUTPUT(pin, signal)  gpio_direction_output(mfp_to_gpio(pin), signal)
#define GPIO_REQUEST(pin, value)            gpio_request(mfp_to_gpio(pin), value)


struct sirf {
	struct sirf_platform_data *pdata;
};
#endif
