/* 
 * drivers/omux/omux_debug.h 
 *
 * Support the debug function of Open MUX Framework.
 *
 * Copyright (C) 2009 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef _OMUX_DEBUG_H_
#define _OMUX_DEBUG_H_

#define OMUX_DEBUG
#ifdef OMUX_DEBUG
#define OMUX_BUG_ON(condition)                                                \
        do {                                                                  \
                if (unlikely(condition)) {                                    \
                        printk(KERN_ERR "BUG: failure at %s:%d/%s()!\n",      \
                                        __FILE__, __LINE__, __FUNCTION__);    \
			omux_assert("omux or spi error\n");		      \
                }                                                             \
        } while(0)
#define DPRINTK(fmt,args...)                                                  \
        do {                                                                  \
                if (OMUX_DEBUG_MASK & OMUX_DEBUG_LOG)             	      \
                        printk(KERN_ERR "%s: " fmt, __FUNCTION__ , ## args);  \
        } while (0)
#define DENTER(fmt,args...)                                                   \
        do {                                                                  \
                if (OMUX_DEBUG_MASK & OMUX_DEBUG_FUNC)            	      \
                        printk(KERN_ERR "[OM] Enter %s: " fmt,		      \
					__FUNCTION__ , ## args); 	      \
        } while (0)
#define DLEAVE(fmt,args...)                                                   \
        do {                                                                  \
                if (OMUX_DEBUG_MASK & OMUX_DEBUG_FUNC)            	      \
                        printk(KERN_ERR "[OM] Leave %s: " fmt, 		      \
					__FUNCTION__ , ## args);              \
        } while (0)

#else /* OMUX_DEBUG */
#define OMUX_BUG_ON(condition)                                                \
        do {                                                                  \
                if (unlikely(condition)) {                                    \
                        printk(KERN_ERR "BUG: failure at %s:%d/%s()!\n",      \
                                        __FILE__, __LINE__, __FUNCTION__);    \
                }                                                             \
        } while(0)
#define DPRINTK(fmt,args...)                                                  \
        do {} while (0)
#define DENTER(fmt,args...)                                                   \
        do {} while (0)
#define DLEAVE(fmt,args...)                                                   \
        do {} while (0)
#define OMUX_BUG_ON(fmt,args...)                                              \
        do {} while (0)
#endif /* OMUX_DEBUG */

void omux_dump_data(const unsigned char *buf, int count);

#endif /* _OMUX_DEBUG_H_ */
