/**
 *  omux modem management module  
 *
 *  Copyright (C) 2009~2010 Borqs Inc. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>

/* To be handled in fbcon! */
#ifdef CONFIG_FBCON_DRAW_PANIC_TEXT
extern int draw_panic_text(const char *panic_text, size_t panic_len,
			   int do_timer, int clear_screen);
#else
static inline int draw_panic_text(const char *panic_text, size_t panic_len,
				  int do_timer, int clear_screen)
{
	return 0;
}
#endif

#define OMUX_MMM_INFO_LEN 	(64)
struct omux_mmm {
	struct delayed_work work;
	unsigned char ms_info[OMUX_MMM_INFO_LEN];
};

static struct omux_mmm *mmm;
static int mmm_enable = 0;
module_param_named(mmm_enable, mmm_enable, int, 0644);

void omux_assert(unsigned char *info)
{
	printk("mmm=%p, mmm_disable=%d, info=%s\n", mmm, mmm_enable, info);
	if (mmm && mmm_enable) {
		memset(mmm->ms_info, 0, OMUX_MMM_INFO_LEN);
		strncpy(mmm->ms_info, info, OMUX_MMM_INFO_LEN - 1);
		schedule_delayed_work(&mmm->work, HZ / 10);
	}
}

EXPORT_SYMBOL(omux_assert);

static void mmm_work(struct work_struct *work)
{
	printk("modem assert info: %s\n", mmm->ms_info);
	draw_panic_text(mmm->ms_info, strlen(mmm->ms_info), 1, 0);
}

static int __init omux_mmm_init(void)
{
	mmm = kzalloc(sizeof(*mmm), GFP_KERNEL);
	if (!mmm)
		return -ENOMEM;

	INIT_DELAYED_WORK_DEFERRABLE(&mmm->work, mmm_work);
	return 0;
}

static void __exit omux_mmm_exit(void)
{
	cancel_delayed_work_sync(&mmm->work);
	kfree(mmm);
}

module_init(omux_mmm_init);
module_exit(omux_mmm_exit);

MODULE_DESCRIPTION("OMUX MMM");
MODULE_LICENSE("GPL");
