#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/miscdevice.h>

#include <linux/platform_device.h>
#include <linux/irq.h>
#include <linux/clk.h>
#include <linux/gpio.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/omux.h>

#include <mach/hardware.h>
#include <mach/mfp.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/mfp-pxa3xx.h>
#include <asm/io.h>

#define CONFIG_DPRAM_BPLOG

#ifdef CONFIG_DPRAM_BPLOG
#include <linux/cdev.h>
#include <linux/device.h>
#include <asm/uaccess.h>
#include <linux/poll.h>   
#endif
#include <linux/wakelock.h>

#include "../core/omux_debug.h"
#include <linux/dpram.h>

/* Circular Buffer */
struct ring_buf {
	unsigned		buf_size;
	char			*buf_buf;
	char			*buf_get;
	char			*buf_put;
};

/*
 * Allocate a circular buffer and all associated memory.
 */
static int ring_buf_alloc(struct ring_buf *rb, unsigned size)
{
	rb->buf_buf = kmalloc(size, GFP_KERNEL);
	if (rb->buf_buf == NULL)
		return -ENOMEM;

	rb->buf_size = size;
	rb->buf_put = rb->buf_buf;
	rb->buf_get = rb->buf_buf;

	return 0;
}

/*
 * Free the buffer and all associated memory.
 */
static void ring_buf_free(struct ring_buf *rb)
{
	kfree(rb->buf_buf);
	rb->buf_buf = NULL;
}
#if 0
/*
 * Clear out all data in the circular buffer.
 */
static void ring_buf_clear(struct ring_buf *rb)
{
	rb->buf_get = rb->buf_put;
	/* equivalent to a get of all data available */
}
#endif

/*
 * Return the number of bytes of data written into the circular
 * buffer.
 */
static unsigned ring_buf_data_avail(struct ring_buf *rb)
{
	return (rb->buf_size + rb->buf_put - rb->buf_get) % rb->buf_size;
}

/*
 * Return the number of bytes of space available in the circular
 * buffer.
 */
static unsigned ring_buf_space_avail(struct ring_buf *rb)
{
	return (rb->buf_size + rb->buf_get - rb->buf_put - 1) % rb->buf_size;
}

/*
 * Copy data data from a user buffer and put it into the circular buffer.
 * Restrict to the amount of space available.
 *
 * Return the number of bytes copied.
 */
static unsigned
ring_buf_put(struct ring_buf *rb, const char *buf, unsigned count)
{
	unsigned len;

	len  = ring_buf_space_avail(rb);
	if (count > len)
		count = len;

	if (count == 0)
		return 0;

	len = rb->buf_buf + rb->buf_size - rb->buf_put;
	if (count > len) {
		memcpy(rb->buf_put, buf, len);
		memcpy(rb->buf_buf, buf+len, count - len);
		rb->buf_put = rb->buf_buf + count - len;
	} else {
		memcpy(rb->buf_put, buf, count);
		if (count < len)
			rb->buf_put += count;
		else /* count == len */
			rb->buf_put = rb->buf_buf;
	}

	return count;
}

#ifdef CONFIG_DPRAM_BPLOG
static unsigned
ring_buf_copy_from_user(struct ring_buf *rb, const char __user *buf, unsigned count)
{
	unsigned len;
	int ret;

	len  = ring_buf_space_avail(rb);
	if (count > len)
		count = len;

	if (count == 0)
		return 0;

	len = rb->buf_buf + rb->buf_size - rb->buf_put;
	if (count > len) {
		if((ret = copy_from_user(rb->buf_put, buf, len))) {
			rb->buf_put += len - ret;
			return len - ret;
		}
		if ((ret = copy_from_user(rb->buf_buf, buf+len, count - len))) {
			rb->buf_put = rb->buf_buf + ((count - len) - ret);
			return count - ret;
		} else {
			rb->buf_put = rb->buf_buf + count - len;
		}
	} else {
		if ((ret = copy_from_user(rb->buf_put, buf, count))) {
			rb->buf_put += count - ret;
			return count - ret;
		}
		if (count < len)
			rb->buf_put += count;
		else /* count == len */
			rb->buf_put = rb->buf_buf;
	}

	return count;
}
#endif

#if 0
/*
 * Get data from the circular buffer and copy to the given buffer.
 * Restrict to the amount of data available.
 *
 * Return the number of bytes copied.
 */
static unsigned
ring_buf_get(struct ring_buf *rb, char *buf, unsigned count)
{
	unsigned len;

	len = ring_buf_data_avail(rb);
	if (count > len)
		count = len;

	if (count == 0)
		return 0;

	len = rb->buf_buf + rb->buf_size - rb->buf_get;
	if (count > len) {
		memcpy(buf, rb->buf_get, len);
		memcpy(buf+len, rb->buf_buf, count - len);
		rb->buf_get = rb->buf_buf + count - len;
	} else {
		memcpy(buf, rb->buf_get, count);
		if (count < len)
			rb->buf_get += count;
		else /* count == len */
			rb->buf_get = rb->buf_buf;
	}

	return count;
}
#endif
struct logic_dev; 
static u32 copy_to_dpram(const unsigned char *buf, u32 len, struct logic_dev *l_dev);

static unsigned
ring_buf_copy_to_dpram(struct ring_buf *rb, unsigned count, struct logic_dev *l_dev)
{
	unsigned len;
	int ret;

	len = ring_buf_data_avail(rb);
	if (count > len)
		count = len;

	if (count == 0)
		return 0;

	len = rb->buf_buf + rb->buf_size - rb->buf_get;
	if (count > len) {
		if ((ret = copy_to_dpram(rb->buf_get, len, l_dev)) < len) {
			rb->buf_get +=ret;
			return ret;
		}

		if ((copy_to_dpram(rb->buf_buf, count - len, l_dev)) < count - len) {
			rb->buf_get = rb->buf_buf + ret;
			return len + ret;
		}
		rb->buf_get = rb->buf_buf + count - len;
	} else {
		copy_to_dpram(rb->buf_get, count, l_dev);
		if (count < len)
			rb->buf_get += count;
		else /* count == len */
			rb->buf_get = rb->buf_buf;
	}

	return count;
}
#ifdef CONFIG_DPRAM_BPLOG
static unsigned
ring_buf_copy_to_user(struct ring_buf *rb, char __user *buf, unsigned count)
{
	unsigned len;
	int ret;

	len = ring_buf_data_avail(rb);
	if (count > len)
		count = len;

	if (count == 0)
		return 0;

	len = rb->buf_buf + rb->buf_size - rb->buf_get;
	if (count > len) {
		if ((ret = copy_to_user(buf, rb->buf_get, len))) {
			rb->buf_get += len - ret;
			return len - ret;
		}
		if ((ret = copy_to_user(buf+len, rb->buf_buf, count - len))) {
			rb->buf_get = rb->buf_buf + (count - len) - ret;
			return count - ret;
		}
		rb->buf_get = rb->buf_buf + count - len;
	} else {
		if ((ret = copy_to_user(buf, rb->buf_get, count))) {
			rb->buf_get += count - ret;
			return count - ret;
		}
		if (count < len)
			rb->buf_get += count;
		else /* count == len */
			rb->buf_get = rb->buf_buf;
	}

	return count;
}
#endif

/* circuit buffer definitions */ 
struct cir_buf {
	u16 *start_addr;	/* still need 32 bit aligned */
	u16 *end_addr;		/* still need 32 bit aligned */
	u16 *in_ptr;
	u16 *extra_info;
	u16 *out_ptr;
};

/* receive callback */
typedef int (*dpram_recv_callback)(unsigned char *chars, unsigned int size);

/* logical device definitions */ 
struct logic_dev {
	struct cir_buf 	recv_zone;
	struct cir_buf 	send_zone;
	dpram_recv_callback 	recv_callback;
	s32 flag_first_send;
};

#ifdef CONFIG_DPRAM_BPLOG
struct bplog_dev {
	struct ring_buf *logbuf;
	struct ring_buf	*write_buf;
	wait_queue_head_t wait;
	struct class *class;
	struct cdev cdev;
	struct device *dev;
	dev_t devid;
	struct logic_dev *l_dev;
};
#endif


struct dpram_dev {
	struct dpram_platform_data *pdata;
	struct device		*dev;
	struct clk		*smc_clk;
	enum state_type	 	ap_state;
	struct delayed_work 	rx_work;
	struct delayed_work 	tx_work;
	struct workqueue_struct *wqueue;
	struct ring_buf 	*write_buf;
	struct timer_list	recv_timer;
	struct logic_dev 	*l_dev;		/* logic device */
	u32 			l_dev_num; /* logic dev number */
	u16 			*send_mailbox_addr;
	u16 			*recv_mailbox_addr;
	u16 			*ram_addr;
	unsigned char		*smc_membase;
	struct wake_lock 	wake_lock;
	struct timer_list	wake_lock_timer;
	unsigned long		flag;
#define LOGIC_DEV_INIT	1
#define TX_RUNNING	2
#ifdef CONFIG_DPRAM_BPLOG
	struct bplog_dev	*bplog;
#endif
};


static struct dpram_dev *g_dev = NULL;
#define RECV_TIMER_EXPIRES	(HZ/10)		/* 1 second/10 = 100 ms */

#ifdef CONFIG_DPRAM_BPLOG
#define DPRAM_TOTAL_LOGIC_DEV	2
#else
#define DPRAM_TOTAL_LOGIC_DEV	1
#endif

#define WRITE_BUF_LEN		(16*1024)
#define TRUE	1
#define FALSE	0

/* events definitions */ 
enum state_event { 
	IPC_SEND, 
	IPC_BP_RDY, 
	IPC_AP_RDY, 
	IPC_DP,
	IPC_SLEEP 
};

static int debug_mask = 0;
module_param_named(debug_mask, debug_mask, int, S_IRUGO | S_IWUSR);
#define DEBUG_SEND		(1 << 0)
#define DEBUG_INTERRUPT 	(1 << 1)
#define DEBUG_SOFT_INTERRUPT 	(1 << 2)
#define DEBUG_BPLOG		(1 << 3)

/* dpram state machine */
static void update_ap_state(struct dpram_dev *dev, enum state_event event);
static int logic_dev_init(struct dpram_dev *dev);

typedef u32 logic_dev_id;
/* return free room of a send buffer of a logical device to clients in BYTE */
static u32 get_send_buf_free_len(logic_dev_id dev_id);

static void ap_sleep(unsigned long _dev)
{
	struct dpram_dev *dev = (struct dpram_dev *)_dev;

	update_ap_state(dev, IPC_SLEEP);
}

static void dpram_wake_lock(struct dpram_dev *dev)
{
	wake_lock_timeout(&dev->wake_lock, 5 * HZ);
	mod_timer(&dev->wake_lock_timer, jiffies + 3 * HZ);
}

/* force read back a 16 bit word to ensure write to DRPAM  */
static inline int ram_write_back_word(u16 data, volatile u16 *addr)
{
	int count = 10;

	do {
		*addr = data;
	} while (*addr != data && count--);

	if (*addr != data) {
		pr_err("%s: can not write a short to dpram\n", __func__);
		return -EFAULT;
	}

	return 0;
}

static inline int ram_write_back_byte(u8 data, volatile u8 *addr)
{
	int count = 10;

	do {
		*addr = data;
	} while (*addr != data && count--);

	if (*addr != data) {
		pr_err("%s: can not write a short to dpram\n", __func__);
		return -EFAULT;
	}

	return 0;
}

static inline u16 ram_read_word(volatile u16 *addr)
{
	return *addr;
}

static inline u8 ram_read_byte(volatile u8 *addr)
{
	return *addr;
}

static ssize_t ipc_cypress_dpram_tx_buf(unsigned char *buf, ssize_t len)
{
	unsigned long flag;
	int ret;
	struct dpram_dev *dev = g_dev;
	if (!dev || !dev->write_buf) {
		pr_err("%s: dev or write_buf is NULL\n", __func__);
		return 0;
	}
	if (len) {
		local_irq_save(flag);
		if (((ret = ring_buf_space_avail(dev->write_buf))) < len) {
			local_irq_restore(flag);
			pr_err("%s: space %d less then len %d", __func__, ret, len);
			return -EAGAIN;
		}
		ret = ring_buf_put(dev->write_buf, buf, len);
		if (ret != len)
			pr_err("%s: space %d len %d", __func__, ret, len);
		local_irq_restore(flag);
	}
	queue_delayed_work(dev->wqueue, &dev->tx_work, 0);
	return len;
}

static struct omux_device omux_cypress_dpram_dev = {
	.owner = THIS_MODULE,
	.name = "ts27010",
	.tx_buf = ipc_cypress_dpram_tx_buf,
};

static int 
ipc_cypress_dpram_rx_buf_handler(unsigned char *buf, unsigned int len)
{
	return omux_rx_buf_handler(&omux_cypress_dpram_dev, buf, len);
}

#ifdef CONFIG_DPRAM_BPLOG 
static u16 get_recv_buf_extra_info(logic_dev_id dev_id)
{
	struct dpram_dev *dev = g_dev;
	struct logic_dev *l_dev;
	if (dev) {
		l_dev = dev->l_dev + dev_id;
		return *l_dev->recv_zone.extra_info;
	}
	
	return 0;
}

#if 0
static void set_send_buf_extra_info(u16 info, logic_dev_id dev_id)
{

	struct dpram_dev *dev = g_dev;
	struct logic_dev *l_dev;
	if (dev) {
		l_dev = dev->l_dev + dev_id;
		*l_dev->send_zone.extra_info = info;
	}
}

static u16 get_send_buf_extra_info(logic_dev_id dev_id)
{
	struct dpram_dev *dev = g_dev;
	struct logic_dev *l_dev;
	if (dev) {
		l_dev = dev->l_dev + dev_id;
		return *l_dev->send_zone.extra_info;
	}
	return 0;
}
#endif

#endif /* CONFIG_DPRAM_BPLOG */

static u32 get_send_buf_free_len(logic_dev_id dev_id)
{
	u16 out, in, total;
	u16 *start, *end;
	u32 ret;
	struct dpram_dev *dev = g_dev;
	struct logic_dev *l_dev;

	if (!dev)
		return 0;
	if (dev_id < 0 || dev_id >= dev->l_dev_num)
		return 0;
	l_dev = dev->l_dev + dev_id;
	start = l_dev->send_zone.start_addr;
	end = l_dev->send_zone.end_addr;
	/* get in byte location of send zone */
	in = ram_read_word(l_dev->send_zone.in_ptr);
	/* get out byte location of send zone */
	out = ram_read_word(l_dev->send_zone.out_ptr);
	total = (u32) (end - start + 1) * 2;
	/* the first byte in ptrs is 0... */
	ret = (in >= out ? total - (in - out) - 1 : (out - in - 1));
	return ret;
}

void update_ap_state(struct dpram_dev *dev, enum state_event event)
{
	unsigned long  irq_flag;
	struct dpram_platform_data *pdata;

	if (!dev) {
		pr_err("%s: dev is NUL\n", __func__);
		return; 
	}
	pdata = dev->pdata;
	local_irq_save(irq_flag);
	switch (event)	{
		case IPC_SEND :
		case IPC_BP_RDY :
		case IPC_DP : //for race condition.
		case IPC_AP_RDY :
			if (dev->ap_state == NOT_READY) {
				pdata->config_ipc(SET_AP_READY);
				dev->ap_state = READY;
			}
			break;
		case IPC_SLEEP :
			if (dev->ap_state == READY) {
				pdata->config_ipc(SET_AP_NOT_READY);
				dev->ap_state = NOT_READY;					
			}
			break;
	}
	local_irq_restore(irq_flag);
}

static u32 copy_to_dpram(const unsigned char *buf, u32 len, struct logic_dev *l_dev)
{
	u16 i, tmp_val;
	u16 *in, *out;
	u16 total;
	volatile u8 *tmp, *next_in_addr, *start, *end;
	struct dpram_dev *dev = g_dev;

	if (dev == NULL) {
		pr_err("%s: dev is NULL\n", __func__);
		return 0;
	}

	/* add data verification at first data send. */
	if (l_dev->flag_first_send == TRUE) {
		if (ram_read_word(l_dev->send_zone.out_ptr) == 0) {
			l_dev->flag_first_send = FALSE;
		} else {
			printk(KERN_ERR"IPC not READY! out %d\n",
				ram_read_word(l_dev->send_zone.out_ptr));
			return 0;
		}
	}

	tmp = (u8 *)buf;

	start = (u8 *)(l_dev->send_zone.start_addr);
	end = (u8 *)(l_dev->send_zone.end_addr);
	in = l_dev->send_zone.in_ptr;
	out = l_dev->send_zone.out_ptr;

	/* how many bytes to be sent ? */
	total = (u16)((end - start + 1) * 2);
	if ((ram_read_word(in) >= total) || (ram_read_word(out) >= total)) {
		pr_err("%s: %d in %d out %d total %d\n", __func__, __LINE__,
			ram_read_word(in), ram_read_word(out), total);
		return 0;
	}

	if (debug_mask & DEBUG_SEND) {
		pr_info("%s:\n", __func__);
		omux_dump_data(buf, len);
	}

	next_in_addr = (u8 *)((u32)start + *in + 1);
	for (i = 0; i < len; i++) {
		if (next_in_addr > (u8 *)((u32)end + 1))
			next_in_addr = (u8 *)start;
		/*
		 * we should write 16bit to DPRAM if current location 
		 * in DPRAM is 16 bit aligned. save bus cycles.
		 */
		if ((u32)next_in_addr & 1) {	
			/* 
			 * target not aligned in 16 bit
			 * just write 1 byte, next time target will be 
			 * aligned.
			 */
			*next_in_addr = *tmp;
			tmp++, next_in_addr++;

		} else {/* target aligned, check source */
			if (((len - i) >= 2) && (!((next_in_addr + 1) > (u8 *)((u32)end + 1)))) {
				/* 
				 * we have at least 2 bytes or 1 word left to 
				 * write and we will not over-boundary
				 */
				if ((u32)tmp & 1)
					tmp_val =
						*tmp | ((u16)*(tmp + 1) << 8);
				else
					tmp_val = *(u16 *)tmp;
				*(u16 *)next_in_addr = tmp_val;
				tmp += 2, next_in_addr += 2, i++;

			} else {
				/* 
				 * less than 1 word to write or we will 
				 * over-boundary
				 */
				*next_in_addr = *tmp;
				tmp++, next_in_addr++;
			}
		}
	}
	/* update send zone's in_ptr after data written... */
	*in = (u16)((u32)next_in_addr - (u32)start - 1);
	/* trigger mailbox interrupt */
	ram_write_back_word(0x1234, dev->send_mailbox_addr);

	return len;
}



static inline int single_zone(struct cir_buf recv_zone,
		dpram_recv_callback recv_callback)
{
	u8 *buff_ptr;
	u32 buff_len;
	u32 read_len;
	int ret = FALSE;

	u16 in = ram_read_word(recv_zone.in_ptr);
	u16 out = ram_read_word(recv_zone.out_ptr);
	u16 total = (u16)((recv_zone.end_addr - recv_zone.start_addr + 1) * 2);
	/* in byte */

	if (recv_callback == NULL) {
		pr_err("%s: recv_callback is NULL\n", __func__);
		return TRUE;
	}

	if ((out >= total) || (in >= total)) {
		pr_info("%s: out %d, in %d, total %d\n", __func__, out, in, total);
		return ret;
	}

	if (out <= in) {
		buff_ptr = ((u8 *)recv_zone.start_addr) + out + 1;
		buff_len = in - out;
		read_len = recv_callback(buff_ptr, buff_len);
		if (read_len < buff_len) {
			ret = TRUE;
		} else {
			read_len = buff_len;
		}
		ram_write_back_word(out + read_len, recv_zone.out_ptr);
	} else {
		/*
		 * (out + 1) should be in range 0, 1, 2, ..., total - 1
		 */
		if (out + 1 <= total - 1) {
			buff_ptr = (u8 *)(recv_zone.start_addr) + out + 1;
			buff_len = (total - 1) - out;
			read_len = recv_callback(buff_ptr, buff_len);
			if (read_len < buff_len) {
				/* new_out = out + read_len; */
				ram_write_back_word(out + read_len, recv_zone.out_ptr);
				return TRUE;
			} else {
				read_len = buff_len;
				/* new_out = out + read_len; */
				ram_write_back_word(out + read_len,
						recv_zone.out_ptr);
			}
		}

		buff_ptr = (u8 *)recv_zone.start_addr;
		buff_len = in + 1;
		read_len = recv_callback(buff_ptr, buff_len);
		if (read_len < buff_len) {
			ret = TRUE;
		} else {
			read_len = buff_len;
		}
		/* new_out = read_len - 1; */
		if (read_len > 0) {
			ram_write_back_word(read_len - 1, recv_zone.out_ptr);
		}
	}

	return ret;
}

void schedule_rx_work(unsigned long _dev)
{
	struct dpram_dev *dev = (struct dpram_dev *)_dev;
	queue_delayed_work(dev->wqueue, &dev->rx_work, 1);
}

static void rx_worker(struct work_struct *work)
{
	struct dpram_dev *dev = container_of(work, struct dpram_dev, rx_work.work);
	struct logic_dev *l_dev;
	int i, count = 5;

	if (!dev) {
		pr_err("%s: dev is NUL\n", __func__);
		return; 
	}

	/* make sure mailbox is cleaned. */
       	ram_read_word(dev->recv_mailbox_addr);
	if (!dev->pdata || !dev->pdata->config_ipc) {
		pr_err("%s: pdata or config_ipc is NULL\n", __func__);
		/* clean it again. */
       		ram_read_word(dev->recv_mailbox_addr);
	} else {
		while (dev->pdata->config_ipc(GET_MAILBOX_STATE) != READY) {
			if (!count--)
				break;
       			ram_read_word(dev->recv_mailbox_addr);
		}
	}
	update_ap_state(dev, IPC_DP);
	dpram_wake_lock(dev);

	/* loop to check the zones */
	for (i = 0; i < dev->l_dev_num; i++) {
		l_dev = dev->l_dev + i;	
		if (!l_dev->recv_callback) {
			pr_err("%s: callback function is NULL\n", __func__);
			continue;
		}
		if (ram_read_word(l_dev->recv_zone.in_ptr) == ram_read_word(l_dev->recv_zone.out_ptr)) {
			continue;
		}
		if (!single_zone(l_dev->recv_zone, l_dev->recv_callback))
			continue;

		mod_timer(&dev->recv_timer, jiffies + RECV_TIMER_EXPIRES);
	}

}

static void tx_worker(struct work_struct *work)
{
	unsigned long flag;
	int ret;
        u32 freelen;
	struct dpram_dev *dev = container_of(work, struct dpram_dev, tx_work.work);
	struct dpram_platform_data *pdata;

	if (!(pdata = dev->pdata) || !(pdata->config_ipc)) {
		pr_err("%s: some NULL pointers\n", __func__);
		queue_delayed_work(dev->wqueue, &dev->tx_work, HZ);
		return; 
	}
	if (test_bit(TX_RUNNING, &dev->flag))
		return;
	else
		set_bit(TX_RUNNING, &dev->flag);

	if (pdata->config_ipc(GET_BP_READY) == NOT_READY) {
		if (debug_mask & DEBUG_SOFT_INTERRUPT)
			printk(KERN_ERR"%s: %d \n", __func__, __LINE__);
		pdata->config_ipc(SET_AP_NOT_READY);
		udelay(50);
		pdata->config_ipc(SET_AP_READY);
		goto out;
	}
	
	if (!test_bit(LOGIC_DEV_INIT, &dev->flag)) {
		pr_info("%s: initial logic device\n", __func__);
		if (logic_dev_init(dev)) {
			pdata->config_ipc(RESET_BP);
			pr_err("%s: initial logic device failed\n", __func__);
			queue_delayed_work(dev->wqueue, &dev->tx_work, HZ * 2);
			goto out;
		}
	}

	update_ap_state(dev, IPC_SEND);
	dpram_wake_lock(dev);

	freelen = get_send_buf_free_len(0);
	local_irq_save(flag);
	ret = ring_buf_copy_to_dpram(dev->write_buf, freelen, dev->l_dev);
#ifdef CONFIG_DPRAM_BPLOG
	if (dev->bplog) {
		freelen = get_send_buf_free_len(1);
		ring_buf_copy_to_dpram(dev->bplog->write_buf, freelen, dev->l_dev + 1);
	}
#endif
	local_irq_restore(flag);
	if (ring_buf_data_avail(dev->write_buf) 
		|| ring_buf_data_avail(dev->bplog->write_buf))
		queue_delayed_work(dev->wqueue, &dev->tx_work, 0);

out:
	clear_bit(TX_RUNNING, &dev->flag);
	return;
}

#ifdef CONFIG_DPRAM_BPLOG

#define BUFFER_SIZE     128*1024

static int bplog_callback(unsigned char *buf, unsigned int len)
{
	unsigned int ret;
	unsigned int freebuflen;
#define MAGIC1	0x55
#define MAGIC2	0xAA
	struct {
		u8 magic1;
		u8 magic2;
		u8 magic3;
		u8 magic4;
	} magic = {MAGIC1, MAGIC2, MAGIC1, MAGIC2};
	struct {
		u32 type;
		u32 len;
	}  __attribute__ ((packed)) header;
	struct bplog_dev *dev = g_dev->bplog;

	if (!dev) {
		pr_err("%s: dev is NULL\n", __func__);
		return 0;
	}
	/* make sure we can put at least one data into buffer */
	freebuflen = ring_buf_space_avail(dev->logbuf);
	if (freebuflen < sizeof(magic) + sizeof(header) + 1)
		return 0;

	/* magic number */
	ret = ring_buf_put(dev->logbuf, (unsigned char *)&magic, sizeof(magic));
	if (ret != sizeof(magic)) {
		pr_err("%s: copy log magic number to buffer error\n", __func__);
		BUG();
	}

	len = min(len, freebuflen - sizeof(magic) - sizeof(header));
	header.type = get_recv_buf_extra_info(1);
	/* get the len that can put into buffer */
	header.len = len;
		
	if (debug_mask & DEBUG_BPLOG) {
		pr_info("%s:\n", __func__);
		omux_dump_data(buf, len);
	}
	ret = ring_buf_put(dev->logbuf, (unsigned char *)&header, sizeof(header));
	if( ret != sizeof(header)){
		pr_err("%s: copy log header to buffer error\n", __func__);
		BUG();
	}
		
	/* buffer return acurate put in len */
	ret = ring_buf_put(dev->logbuf, buf, len);
	if(ret != len){
		pr_err("%s: copy data into buffer error\n", __func__);
		BUG();
	}

        wake_up_interruptible(&dev->wait);
	return ret;
}

static int bplog_open(struct inode *inode, struct file *file)
{
	if (g_dev)
		file->private_data = g_dev;
	else
		return -ENODEV;
	return 0;
}

static int bplog_release(struct inode *inode, struct file *file)
{
	file->private_data = NULL;
	return 0;
}

static ssize_t 
bplog_read(struct file *file, char __user *buffer, size_t len, loff_t *ppos)
{
	int ret;
	struct dpram_dev *dev = file->private_data;

	if (!dev || !dev->bplog) {
		pr_err("%s: dev is NULL\n", __func__);
		return -EIO;
	}
       	/* use min value of reqired or available in buffer */
	ret =  min(len, ring_buf_data_avail(dev->bplog->logbuf));
	ret = ring_buf_copy_to_user(dev->bplog->logbuf, buffer, ret);
	return ret;
}

static ssize_t 
bplog_write(struct file *file, const char __user *buffer, size_t len, loff_t *ppos)
{
	int ret;
	struct dpram_dev *dev = file->private_data;
	size_t old_len = len;

	if (!dev || !dev->bplog) {
		pr_err("%s: dev is NULL\n", __func__);
		return -EIO;
	}
	if (!(ret = ring_buf_copy_from_user(dev->bplog->write_buf, buffer, len))) {
		pr_info("%s: from user %d wanted %d copied\n", __func__, len, ret);
		return ret;
	}
	tx_worker(&dev->tx_work.work);
		
	return old_len;
}

static unsigned int bplog_poll(struct file *file, poll_table *wait)
{
	unsigned int mask = 0;
	
	struct dpram_dev *dev = file->private_data;

	if (!dev || !dev->bplog) {
		pr_err("%s: dev is NULL\n", __func__);
		return -EIO;
	}

	poll_wait(file, &dev->bplog->wait , wait);
	if (ring_buf_data_avail(dev->bplog->logbuf))
		mask |= POLLIN | POLLRDNORM;
	return mask;

}

static const struct file_operations bplog_fops = {
	.owner = THIS_MODULE,
	.open  = bplog_open,
	.release = bplog_release,
	.read  = bplog_read,
	.write = bplog_write,
	.poll  = bplog_poll,
};

static __init int bplog_init(void)
{
	int ret;
	struct bplog_dev *dev;
	
	dev = (struct bplog_dev *)kzalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		pr_err("%s: alloc log_buf failed\n", __func__);
		return -ENOMEM;
	}
	dev->logbuf = (struct ring_buf *)kzalloc(sizeof(*(dev->logbuf)), GFP_KERNEL);
	if (dev->logbuf == NULL) {
		pr_err("%s: alloc log_buf failed\n", __func__);
		ret = -ENOMEM;
		goto err1;
	}
	ret = ring_buf_alloc(dev->logbuf, BUFFER_SIZE);
	if (ret < 0) {
		pr_err("%s: alloc ring_buf failed\n", __func__);
		ret = -ENOMEM;
		goto err2;
	}
	dev->write_buf = (struct ring_buf *)kzalloc(sizeof(*(dev->write_buf)), GFP_KERNEL);
	if (dev->write_buf == NULL) {
		pr_err("%s: alloc log_buf failed\n", __func__);
		ret = -ENOMEM;
		goto err1;
	}
	ret = ring_buf_alloc(dev->write_buf, 2048);
	if (ret < 0) {
		pr_err("%s: alloc ring_buf failed\n", __func__);
		ret = -ENOMEM;
		goto err2;
	}
	init_waitqueue_head(&dev->wait);

	ret = alloc_chrdev_region(&dev->devid, 0, 1, "bplog");
	if (ret < 0){
		pr_err("%s: Can not get device ID\n", __func__);
		goto err3;
	}

	cdev_init(&dev->cdev, &bplog_fops);
	dev->cdev.owner = THIS_MODULE;
	
	ret = cdev_add(&dev->cdev, MKDEV(MAJOR(dev->devid), 0), 1);
	if (ret){
		pr_err("%s: Can not register cdev driver to system\n", __func__);
		goto err4;
	}
	
	dev->class = class_create(THIS_MODULE, "bplog");
	if (IS_ERR(dev->class)){
		pr_err("%s: Can not registe class\n", __func__);
		ret = PTR_ERR(dev->class);
		goto err5;
	}

	dev->dev = device_create(dev->class, NULL, MKDEV(MAJOR(dev->devid), 0), NULL, "bplog");
	if (IS_ERR(dev)) {
		pr_err("%s: Can not add device to system\n", __func__);
		ret = PTR_ERR(dev->dev);
		goto err6;
	}
	g_dev->bplog = dev;	
	return 0;

err6:
	class_destroy(dev->class);
err5:
	cdev_del(&dev->cdev);
err4:
	unregister_chrdev_region(dev->devid, 1);
err3:
	ring_buf_free(dev->logbuf);
err2:
	kfree(dev->logbuf);
err1:
	kfree(dev);

	return ret;
}

static void bplog_exit(void)
{

	struct bplog_dev *dev;
	if (!g_dev)
		return;
	if (!(dev = g_dev->bplog))
		return;

	class_destroy(dev->class);
	cdev_del(&dev->cdev);
	unregister_chrdev_region(dev->devid, 1);
	ring_buf_free(dev->logbuf);
	kfree(dev->logbuf);
	kfree(dev);
}
#endif

static unsigned int bp_assert;

static int lpr_open(struct inode *inode, struct file *file)
{
	bp_assert = 0;
	return 0;
}

static DECLARE_WAIT_QUEUE_HEAD(lpr_wait);

static unsigned int lpr_poll(struct file *file, poll_table *wait)
{
	unsigned int mask = 0;
	
	poll_wait(file, &lpr_wait , wait);
	if (bp_assert) 
		mask |= POLLIN | POLLRDNORM;

	return mask;

}

static ssize_t 
lpr_read(struct file *file, char __user *buffer, size_t len, loff_t *ppos)
{
	if(copy_to_user(buffer, &bp_assert, sizeof(bp_assert))) {
		pr_err("%s: copy_to_user failed\n", __func__);
		return -EFAULT;
	}
	bp_assert = 0;
	return sizeof(bp_assert);
}

static const struct file_operations lpr_fops = {
	.owner  = THIS_MODULE,
	.open  = lpr_open,
	.poll  = lpr_poll,
	.read  = lpr_read,
};
	
static struct miscdevice lpr_device = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "lpr",
	.fops = &lpr_fops,
};

static irqreturn_t bp_wd_out_handler(int irq, void *_dev)
{
	struct dpram_dev *dev = (struct dpram_dev *)_dev;

	if (!dev) {
		pr_err("%s: dev is NUL\n", __func__);
		return IRQ_HANDLED; 
	}
	bp_assert = ~0;
	pr_info("%s: BP assert 0x%x\n", __func__, bp_assert);
	wake_up_interruptible(&lpr_wait);

	return IRQ_HANDLED;
}


static int lpr_init(struct dpram_dev *dev)
{
	int ret;

	if (!dev || !dev->pdata) {
		pr_err("%s: dev or pdata is NULL\n", __func__);
		return -EFAULT;
	}
	if ((ret = misc_register(&lpr_device))) {
		printk("misc_register returned %d in %s\n", ret, __func__);
		goto err0;
	}
	if ((ret = request_irq(dev->pdata->bp_wd_out_irq, bp_wd_out_handler,
				IRQF_TRIGGER_FALLING, "bp watch dog", dev))) {
		/* register irq error handler */
		printk("BP Watch dog interrupt request error\n");
		goto err1;
	}

	return 0;
err1:
	misc_deregister(&lpr_device);
err0:
	return ret;
}

static void lpr_exit(struct dpram_dev *dev)
{
	if (!dev || !dev->pdata) {
		pr_err("%s: dev or pdata is NULL\n", __func__);
		return;
	}
	free_irq(dev->pdata->bp_wd_out_irq, dev);
	if (misc_deregister(&lpr_device)) {
		printk("misc_register failed in %s\n", __func__);
	}
}

/*
 * Interrupter handler for BP_RDY
 */
static irqreturn_t ipc_bp_rdy_handler(int irq, void *_dev)
{
	struct dpram_dev *dev = (struct dpram_dev *)_dev;

	if (!dev) {
		pr_err("%s: dev is NUL\n", __func__);
		return IRQ_HANDLED;
	}
	/* Call State Machine (BP_RDY) */
	if (debug_mask & DEBUG_INTERRUPT)
		pr_info("%s: BP ready\n", __func__);
	update_ap_state(dev, IPC_BP_RDY);
	dpram_wake_lock(dev);
	if (ring_buf_data_avail(dev->write_buf))
		queue_delayed_work(dev->wqueue, &dev->tx_work, 0);

	return IRQ_HANDLED;
}

static irqreturn_t mailbox_interrupt(int irq, void *_dev)
{
	struct dpram_dev *dev = (struct dpram_dev *)_dev;

	if (!dev) {
		pr_err("%s: dev is NUL\n", __func__);
		return IRQ_HANDLED; 
	}
	ram_read_word(dev->recv_mailbox_addr);
	queue_delayed_work(dev->wqueue, &dev->rx_work, 0);

	return IRQ_HANDLED;
}

/*
 * cypress dpram logic device, now 2 nodes, one for data, one for BP log.
 */

/* dpram logic device: 0 recv & send region offset */
#define LOGIC_DEV_0_RECV_START_OFFSET 		(0x0)
#define LOGIC_DEV_0_RECV_END_OFFSET 		(0x1BFF)
#define LOGIC_DEV_0_RECV_IN_PTR_OFFSET		(0x3FEE)
#define LOGIC_DEV_0_RECV_EXTRA_INFO_OFFSET	(0x3FEF)
#define LOGIC_DEV_0_RECV_OUT_PTR_OFFSET		(0x3FF0)	

#define LOGIC_DEV_0_SEND_START_OFFSET 		(0x1C00)
#define LOGIC_DEV_0_SEND_END_OFFSET 		(0x35FF)
#define LOGIC_DEV_0_SEND_IN_PTR_OFFSET		(0x3FF2)
#define LOGIC_DEV_0_SEND_EXTRA_INFO_OFFSET	(0x3FF3)
#define LOGIC_DEV_0_SEND_OUT_PTR_OFFSET		(0x3FF4)

#ifdef CONFIG_DPRAM_BPLOG
/* dpram logic device: bp log, recv & send region offset */
#define LOGIC_DEV_BP_LOG_RECV_START_OFFSET	(0x3600)
#define LOGIC_DEV_BP_LOG_RECV_END_OFFSET	(0x3DFF)
#define LOGIC_DEV_BP_LOG_RECV_IN_PTR_OFFSET	(0x3FF6)
#define LOGIC_DEV_BP_LOG_RECV_EXTRA_INFO_OFFSET	(0x3FF7)
#define LOGIC_DEV_BP_LOG_RECV_OUT_PTR_OFFSET	(0x3FF8)

#define LOGIC_DEV_BP_LOG_SEND_START_OFFSET	(0x3E00)
#define LOGIC_DEV_BP_LOG_SEND_END_OFFSET	(0x3FED)
#define LOGIC_DEV_BP_LOG_SEND_IN_PTR_OFFSET	(0x3FFA)
#define LOGIC_DEV_BP_LOG_SEND_EXTRA_INFO_OFFSET	(0x3FFB)
#define LOGIC_DEV_BP_LOG_SEND_OUT_PTR_OFFSET	(0x3FFC)
#endif

/* mailbox address */
#define SEND_MAILBOX_OFFSET			(0x3FFF)
#define RECV_MAILBOX_OFFSET			(0x3FFE)

static int logic_dev_init(struct dpram_dev *dev)
{
	struct logic_dev *l_dev = dev->l_dev;
	if (!l_dev) {
		pr_err("%s: l_dev is NULL\n", __func__);
	}
	/* init dpram logic device 0 */
	l_dev->recv_zone.start_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_RECV_START_OFFSET);
	l_dev->recv_zone.end_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_RECV_END_OFFSET);
	l_dev->recv_zone.in_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_RECV_IN_PTR_OFFSET);
	l_dev->recv_zone.extra_info = (u16 *)(dev->ram_addr + LOGIC_DEV_0_RECV_EXTRA_INFO_OFFSET);
	l_dev->recv_zone.out_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_RECV_OUT_PTR_OFFSET);
	if (ram_write_back_word(0, l_dev->recv_zone.out_ptr)) {
		pr_err("%s: can not initial recv_zone.out_ptr\n", __func__);
		return -EFAULT;
	}
	if (ram_write_back_word(0, l_dev->recv_zone.in_ptr)) {
		pr_err("%s: can not initial recv_zone.in_ptr\n", __func__);
		return -EFAULT;
	}

	l_dev->send_zone.start_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_SEND_START_OFFSET);
	l_dev->send_zone.end_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_SEND_END_OFFSET);
	l_dev->send_zone.in_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_SEND_IN_PTR_OFFSET);
	l_dev->send_zone.extra_info = (u16 *)(dev->ram_addr + LOGIC_DEV_0_SEND_EXTRA_INFO_OFFSET);
	l_dev->send_zone.out_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_0_SEND_OUT_PTR_OFFSET);
	l_dev->flag_first_send = TRUE;
	l_dev->recv_callback = ipc_cypress_dpram_rx_buf_handler;
	if (ram_write_back_word(0, l_dev->send_zone.in_ptr)) {
		pr_err("%s: can not initial send_zone.in_ptr\n", __func__);
		return -EFAULT;
	}
	if (ram_write_back_word(0, l_dev->send_zone.out_ptr)) {
		pr_err("%s: can not initial send_zone.out_ptr\n", __func__);
		return -EFAULT;
	}
#ifdef CONFIG_DPRAM_BPLOG
	/* if we have bp log logic device ? */
	l_dev++;
	/* init bp log logic device address */
	l_dev->recv_zone.start_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_RECV_START_OFFSET);
	l_dev->recv_zone.end_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_RECV_END_OFFSET);
	l_dev->recv_zone.in_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_RECV_IN_PTR_OFFSET);
	l_dev->recv_zone.extra_info = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_RECV_EXTRA_INFO_OFFSET);
	l_dev->recv_zone.out_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_RECV_OUT_PTR_OFFSET);
	l_dev->recv_callback = bplog_callback;
	if (ram_write_back_word(0, l_dev->recv_zone.out_ptr)) {
		pr_err("%s: can not initial bp log recv_zone.out_ptr\n", __func__);
		return -EFAULT;
	}

	l_dev->send_zone.start_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_SEND_START_OFFSET);
	l_dev->send_zone.end_addr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_SEND_END_OFFSET);
	l_dev->send_zone.in_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_SEND_IN_PTR_OFFSET);
	l_dev->send_zone.extra_info = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_SEND_EXTRA_INFO_OFFSET);
	l_dev->send_zone.out_ptr = (u16 *)(dev->ram_addr + LOGIC_DEV_BP_LOG_SEND_OUT_PTR_OFFSET);
	l_dev->flag_first_send = TRUE;
	if (ram_write_back_word(0, l_dev->send_zone.in_ptr)) {
		pr_err("%s: can not initial bp log send_zone.in_ptr\n", __func__);
		return -EFAULT;
	}
#endif
	set_bit(LOGIC_DEV_INIT, &dev->flag);
	return 0;
}

static int __init cypress_dpram_probe(struct device *_dev)
{
	struct dpram_dev *dev;
	struct dpram_platform_data *pdata;
	struct logic_dev *l_dev;
	int ret = 0;

	if (!_dev->platform_data) {
		pr_err("%s: platform data is NULL\n", __func__);
		return -EINVAL;
	}

	pdata = _dev->platform_data;
	dev = (struct dpram_dev *)kzalloc(sizeof(*dev), GFP_ATOMIC);
	if (IS_ERR(dev)) {
		pr_err("%s: kzalloc failed\n", __func__);
		ret = -ENOMEM;
		goto err0;
	}

	dev->pdata = pdata;
	dev->l_dev_num = DPRAM_TOTAL_LOGIC_DEV;
	dev->ap_state = NOT_READY;
	dev->dev = _dev;

	l_dev = kzalloc(sizeof(*l_dev) * dev->l_dev_num, GFP_ATOMIC);
	if (!l_dev) {
		pr_err("%s: kzalloc failed\n", __func__);
		ret = -ENOMEM;
		goto err1;
	}
	dev->l_dev = l_dev;

	dev->write_buf = (struct ring_buf *)kzalloc(sizeof(*(dev->write_buf)), GFP_ATOMIC);
	if (!dev->write_buf) {
		pr_err("%s: alloc write_buf failed\n", __func__);
		goto err2;
	}
	ret = ring_buf_alloc(dev->write_buf, WRITE_BUF_LEN);
	if (ret < 0) {
		pr_err("%s: logic device initial failed\n", __func__);
		goto err3;
	}

	if (!(dev->ram_addr = pdata->ram_addr)) {
		pr_err("%s: ram addr is NULL\n", __func__);
		ret = -EINVAL;
		goto err3;
	}
	dev->send_mailbox_addr = (u16 *)(dev->ram_addr + SEND_MAILBOX_OFFSET);
	dev->recv_mailbox_addr = (u16 *)(dev->ram_addr + RECV_MAILBOX_OFFSET);

	if ((ret = omux_register_device(&omux_cypress_dpram_dev))) {
		pr_err("%s: Failed to register omux dev(%d).\n", __func__, ret);
		goto err3;
	}

        if ((ret = request_irq(pdata->mailbox_irq, mailbox_interrupt,
			IRQF_TRIGGER_FALLING, "ipc mailbox", dev))) {
                pr_err("%s: mailbox irq: ret = %d\n", __func__, ret);
		goto err4;
	}
	if ((ret = request_irq(pdata->bp_rdy_irq, ipc_bp_rdy_handler,
				IRQF_TRIGGER_RISING, "ipc hdshk", dev))) {
		/* register irq error handler */
		printk("BP_RDY interrupt request error\n");
		goto err5;
	}

	dev->wqueue = create_singlethread_workqueue("ipc");
	if (!dev->wqueue) {
		pr_err("%s: create workqueue failed\n", __func__);
		goto err6;
	}
	INIT_DELAYED_WORK(&dev->tx_work, tx_worker);
	init_timer(&dev->recv_timer);
	dev->recv_timer.function = schedule_rx_work;
	dev->recv_timer.data = (unsigned long)dev;
	init_timer(&dev->wake_lock_timer);
	dev->wake_lock_timer.function = ap_sleep;
	dev->wake_lock_timer.data = (unsigned long)dev;
	wake_lock_init(&dev->wake_lock, WAKE_LOCK_SUSPEND, "ipc dpram");
	INIT_DELAYED_WORK(&dev->rx_work, rx_worker);
	dev_set_drvdata(_dev, dev);
	g_dev = dev;
	/* bplog_init() will refer to g_dev. */
#ifdef CONFIG_DPRAM_BPLOG
	bplog_init();
#endif
	lpr_init(dev);

        ram_read_word(dev->recv_mailbox_addr);
	pr_info("%s: %d, dpram probe successed\n", __func__, __LINE__);
	return 0;
err6:
	free_irq(pdata->bp_rdy_irq, dev);
err5:
	free_irq(pdata->mailbox_irq, dev);
err4:
	if (omux_unregister_device(&omux_cypress_dpram_dev)) {
		pr_err("%s: unregister omux device failed\n", __func__);
	}
err3:
	kfree(dev->write_buf);
	dev->write_buf = NULL;
err2:
	kfree(l_dev);
	dev->l_dev = NULL;
err1:
	kfree(dev);
err0:
	return ret;
}

static int __exit cypress_dpram_remove(struct device *_dev)
{
	struct dpram_dev *dev;
	struct dpram_platform_data *pdata;
	
#ifdef CONFIG_DPRAM_BPLOG
	bplog_exit();
#endif
	g_dev = NULL;
	if (!_dev->platform_data) {
		pr_err("%s: platform data is NULL\n", __func__);
		return -EINVAL;
	}
	pdata = _dev->platform_data;

	dev = dev_get_drvdata(_dev);
	if (!dev) {
		pr_err("%s: dev is NULL\n", __func__);
		return -EINVAL;
	}

	lpr_exit(dev);
	dev_set_drvdata(_dev, NULL);
	cancel_delayed_work(&dev->tx_work);
    	wake_lock_destroy(&dev->wake_lock);
	cancel_delayed_work(&dev->rx_work);

	free_irq(pdata->mailbox_irq, dev);
	free_irq(pdata->bp_rdy_irq, dev);
	if (dev->write_buf) {
		ring_buf_free(dev->write_buf);
		kfree(dev->write_buf);
		dev->write_buf = NULL;
	}
	if (omux_unregister_device(&omux_cypress_dpram_dev)) {
		pr_err("%s: unregister omux device failed\n", __func__);
	}
	if (dev->l_dev)
		kfree(dev->l_dev);

	kfree(dev);
	return 0;
}

static void cypress_dpram_shutdown(struct device *_dev)
{
	struct dpram_platform_data *pdata;
	if (!_dev->platform_data) {
		pr_err("%s: platform data is NULL\n", __func__);
		return;
	}
	pdata = _dev->platform_data;
	if (pdata->config_ipc)
		pdata->config_ipc(SHUTDOWN_BP);

}

static struct device_driver cypress_dpram_driver = 
{
    .name       = (char *)"cypress_dpram",
    .bus        = &platform_bus_type,
    .probe      = cypress_dpram_probe,
    .remove     = __exit_p(cypress_dpram_remove), 
    .shutdown   = cypress_dpram_shutdown,
};

static int __init cypress_dpram_init(void)
{
	return driver_register(&cypress_dpram_driver); 
}

static void __exit cypress_dpram_exit(void)
{
	driver_unregister(&cypress_dpram_driver);
}

module_init(cypress_dpram_init);
module_exit(cypress_dpram_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Cypress DPRAM Driver");
