/*
 * simple ring buffer
 *
 * supports a concurrent reader and writer without locking
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/slab.h>

#include "ringbuf.h"
#if 0
static struct ringbuf *ringbuf_alloc(int len)
{
	struct ringbuf *rbuf;

	rbuf = kzalloc(sizeof(*rbuf) + len, GFP_KERNEL);
	if (rbuf == NULL)
		return NULL;

	rbuf->len = len;
	rbuf->head = 0;
	rbuf->tail = 0;

	return rbuf;
}

static void ringbuf_free(struct ringbuf *rbuf)
{
	kfree(rbuf);
}
#endif
static struct ringbuf *ringbuf_init(unsigned char *data, int len)
{
	struct ringbuf *rbuf;

	rbuf = (struct ringbuf *)data;

	rbuf->len = len - sizeof(*rbuf);
	rbuf->head = 0;
	rbuf->tail = 0;

	return rbuf;
}

static void ringbuf_fini(struct ringbuf *rbuf)
{
}

static int ringbuf_level(struct ringbuf *rbuf)
{
	int level = rbuf->head - rbuf->tail;

	if (level < 0)
		level = rbuf->len + level;

	return level;
}

static int ringbuf_room(struct ringbuf *rbuf)
{
	return rbuf->len - ringbuf_level(rbuf) - 1;
}

static u8 ringbuf_peek(struct ringbuf * rbuf, int i)
{
	return rbuf->buf[(rbuf->tail + i) % rbuf->len];
}

static int ringbuf_consume(struct ringbuf *rbuf, int count)
{
	count = min(count, ringbuf_level(rbuf));

	rbuf->tail = (rbuf->tail + count) % rbuf->len;

	return count;
}

static int ringbuf_push(struct ringbuf *rbuf, u8 datum)
{
	if (ringbuf_room(rbuf) == 0)
		return 0;

	rbuf->buf[rbuf->head] = datum;
	rbuf->head = (rbuf->head + 1) % rbuf->len;

	return 1;
}

static int ringbuf_write(struct ringbuf *rbuf,
			  const u8 * data, int len)
{
	int count = 0;
	int i;

	for (i = 0; i < len; i++)
		count += ringbuf_push(rbuf, data[i]);

	return count;
}
