/* 
 * drivers/omux/mux/ts27010_omux.c 
 *
 * TS27010mux's omux abstract layer.
 *
 * Copyright (C) 2010 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/tty.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/poll.h>
#include <linux/omux.h>

#include "ts27010_mux.h"
#include "ringbuf.h"

#include "../core/omux_debug.h"

#include "ts27010_mux.c"


static int debug_mask = 0;
module_param_named(debug_mask, debug_mask, int, S_IRUGO | S_IWUSR);
#define DEBUG_SEND	(1 << 0)
#define DEBUG_RECV	(1 << 1)

static struct mutex send_lock;

static int ts27010_rx_buf_handler(unsigned char *buf, ssize_t len)
{
	if (debug_mask & DEBUG_RECV) {
		pr_info("%s:\n", __func__);
		omux_dump_data(buf, len);
	}

	return ts27010_mux_rx(buf, len);
}

static struct omux_driver omux_ts27010_driver = {
        .name = "ts27010",
        .rx_buf_handler = ts27010_rx_buf_handler,
};

int ts27010_mux_send(unsigned char *buf, ssize_t len)
{
	ssize_t count;

	if (debug_mask & DEBUG_SEND) {
		pr_info("%s:\n", __func__);
		omux_dump_data(buf, len);
	}

	mutex_lock(&send_lock);
	count = omux_tx_buf(&omux_ts27010_driver, buf, len);
	BUG_ON(count > len || count < 0);
	mutex_unlock(&send_lock);

	return count;
}


static int __init ts27010_omux_init(void)
{
	int err;

	mutex_init(&send_lock);

	err = ts27010_mux_init();
	if (err < 0)
		pr_err("ts27010: unable to init mux driver\n");

	err = omux_register_driver(&omux_ts27010_driver);
	if (err < 0)
		pr_err("ts27010: unable to register omux driver\n");

	return err;
}

static void __exit ts27010_omux_exit(void)
{
        int err;

        err = omux_unregister_driver(&omux_ts27010_driver);
        if (err < 0)
                pr_err("ts27010: unable to unregister omux driver\n");

	OMUX_BUG_ON(err < 0);

	ts27010_mux_remove();
	mutex_destroy(&send_lock);
}

module_init(ts27010_omux_init)
module_exit(ts27010_omux_exit)
