#include <linux/module.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/workqueue.h>
#include <linux/pm.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <mach/micco.h>
#include <mach/micco_battery.h>
#include <mach/mfp-pxa9xx.h>
#include <asm/gpio.h>
#include <mach/dvfm.h>
#include <mach/pxa3xx_pmic.h>
#include <linux/mtd/mtd.h>
#include <mach/t8200.h>
#include <linux/bq2407x_charger.h>

#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

static short g_bat_adc_min = DEFAULT_BAT_ADC_MIN;
static short g_bat_adc_max = DEFAULT_BAT_ADC_MAX;
static short g_bat_volt_min = DEFAULT_BAT_VOLT_MIN;
static short g_bat_volt_max = DEFAULT_BAT_VOLT_MAX;
static short g_temp_adc_min = DEFAULT_TEMP_ADC_MIN;
static short g_temp_adc_max = DEFAULT_TEMP_ADC_MAX;
static short g_bat_temp_min = DEFAULT_BAT_TEMP_MIN;
static short g_bat_temp_max = DEFAULT_BAT_TEMP_MAX;

static int g_charge_state = 0, old_bat_voltage = 0;
static int g_battery_state = BAT_STATE_UNKNOW;
static int g_adc_is_normal = 1;
static int adc_max_float_value = 0;
static struct pxa_micco_battery_info *bat_di;
static int full_jiff = 0, init_jiff = 0;
static int charge_is_full = 0;
static int g_is_call = 0, g_call_signal = 0;
static char call_status[5];
static int g_adc_calibrated = 0;

static char *low_voltage_env[]={"LOW_VOLTAGE=TRUE"};

extern unsigned int cur_op;

struct pxa_micco_battery_info {
	struct device *dev;

	unsigned long update_time;	/* jiffies when data read */
	int charge_status;		/* POWER_SUPPLY_STATUS_* */
	int health;			/* */
	int present;               	/* */
	int vol_adc;                     /* BYD used for NPI read adc value*/
	int capacity;               	/* in percents! */
	int life_sec;			/* units of seconds */
	int temp_C;			/* C */
	int voltage_mV;			/* mV */
	int voltage_savege_mV;		/* mV */
	int technology;			/* */
	int vol_toolow;

	int full_counter;
	struct power_supply bat;
	struct workqueue_struct *monitor_wqueue;
	struct delayed_work monitor_work;
};

static unsigned int cache_time = 1000;
module_param(cache_time, uint, 0644);
MODULE_PARM_DESC(cache_time, "cache time in milliseconds");


#ifdef CONFIG_ANDROID_POWER
static android_suspend_lock_t pxa_battery_lock = {
        .name = "pxa_battery"
};
#endif 

static int bat_fvolt2adc(int volt)
{
	return (((g_bat_adc_max - g_bat_adc_min) * volt) /\
				(g_bat_volt_max - g_bat_volt_min));
}

static int bat_adc2volt(int adc)
{
	return (g_bat_volt_min + (int)((g_bat_volt_max -\
			g_bat_volt_min) * (adc - g_bat_adc_min) /\
		 	(g_bat_adc_max - g_bat_adc_min)));
}

static int bat_volt2adc(int volt)
{
	return (g_bat_adc_min + (int)((volt - g_bat_volt_min) \
			* (g_bat_adc_max - g_bat_adc_min) /\
		 	(g_bat_volt_max - g_bat_volt_min)));
}

static int read_vbat_adc(u16 *adc)
{
	int data[READ_ADC_NUM], index[2]; 
	int tmp, result = 0;
	int i, pass, swapped = 0, flag = 0; 

	for (i = 0; i < READ_ADC_NUM; i++)
		result += data[i] = micco_adc_read(MICCO_ADC_MUX_VBAT);

	for (pass = 1; pass <= READ_ADC_NUM - 1; pass++) {
		for (i = 1; i <= (READ_ADC_NUM - 1) - (pass - 1); i++)
			if (data[i-1] > data[i]) {
				tmp = data[i-1];
				data[i-1] = data[i];
				data[i] = tmp;
				swapped++;
			}
		if (!swapped)
			break;

		swapped = 0;
	}

	for (i = 1; i < READ_ADC_NUM; i++) 
		if ((data[i] - data[i-1]) > adc_max_float_value) {
			if (++flag > VBAT_ADC_FLOAT_MAX_NUM) {
				TWARN("ignore adc value, flag: %d\n", flag); 
				return -1;
			}
			else 
				index[flag-1] = i;
		}

	if (flag == 1) {
		if (index[0] == 1) 
			result -= data[0];
		else if (index[0] == 2) {
			result -= (data[0] + data[1]);
			flag++;
		}
		else if (index[0] == 8) {
			result -= (data[8] + data[9]);
			flag++;
		}
		else if (index[0] == 9) 
			result -= data[9];
		else {
			TWARN("ignore adc value, flag: %d\n", flag); 
			return -1;
		}
	} else if (flag == 2) {
		if ((index[0] == 1) && (index[1] == 2))
			result -= (data[0] + data[1]);
		else if ((index[0] == 8) && (index[1] == 9)) 
			result -= (data[8] + data[9]);
		else
			return -1;
	}

	if (flag != 0)
		TINFO("normal adc data, flag: %d\n", flag); 

  	*adc = result / (READ_ADC_NUM - flag);
	return 0;
}

static int bat_adc2temp(int adc)
{
	return adc;
}

static int get_temper(void)
{
	return bat_adc2temp(get_adc_average(MICCO_ADC_MUX_TBAT));
}

/* For T8200 pm status collecting */
static int lcd_power_check(void)
{
	int vol;

	pxa3xx_pmic_get_voltage(VCC_LCD, &vol);
	if(vol == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

static int camera_power_check(void)
{
	if(gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(CAM_PWR_EN))) == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

static int mtv_power_check(void)
{
	if(gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(CMMB_PWR_EN))) == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

static int bt_power_check(void)
{
	if (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_BT))) == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

static int mmc_power_check(void)
{
	if (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(MMC1_CD_N))) == 0)
	//int vol;
	//pxa3xx_pmic_get_voltage(VCC_SDIO, &vol);
	//if (vol == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

static int usb_power_check(void)
{
	extern int g_psy_type;

	if (g_charge_state == POWER_SUPPLY_STATUS_CHARGING && g_psy_type == BQ2407x_POWER_CHARGE_USB)
		return PM_STATUS_ON;
	else
		return PM_STATUS_OFF;
}

static int gps_power_check(void)
{
	if (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(AGPS_PWR_ON))) == 0)
		return PM_STATUS_OFF;
	else
		return PM_STATUS_ON;
}

extern void micco_read_codec(void);
static int audio_power_check(void)
{
	int i, ret = 0;

	micco_read_codec();

	for (i = 0; i < MICCO_CODEC_INPUT_NUMBER; i++) {
		ret = micco_codec_check_input(i);
		if (ret)
			return ret;
	}
	for (i = 0; i < MICCO_CODEC_OUPUT_NUMBER; i++) {
		ret = micco_codec_check_output(i);
		if (ret)
			return ret;
	}
	return ret;
}

static void get_pstatus(unsigned long *s)
{
	*s = (lcd_power_check() ? DEV_LCD_BIT : 0) 
	   | (camera_power_check() ? DEV_CAMERA_BIT : 0) 
	   | (mtv_power_check() ? DEV_MTV_BIT : 0) 
	   | (bt_power_check() ? DEV_BT_BIT : 0) 
	   | (mmc_power_check() ? DEV_MMC_BIT : 0) 
	   | (usb_power_check() ? DEV_USB_BIT : 0) 
	   | (gps_power_check() ? DEV_GPS_BIT : 0) 
	   | (audio_power_check() ? DEV_AUDIO_BIT : 0);
}

static void  set_audio_cali_bit(unsigned long long *cdev_s)
{
	u8 val;
	int speak_vol; 
	unsigned long long status = *cdev_s;

	micco_read(MICCO_AUDIO_LINE_AMP, &val);
	if (!(val & MICCO_AUDIO_LINE_AMP_GPO))
		status |= CDEV_AUDIO_NORMAL_BIT;
	else {
		micco_read(MICCO_MONO_VOL, &val);
		speak_vol = val && 0x3F;
		if (speak_vol < CDEV_AUDIO_LOW_VOL_THRESHOLD)
			status |= CDEV_AUDIO_SPEAK_MIN_BIT;
		else if (speak_vol > CDEV_AUDIO_HIGH_VOL_THRESHOLD)
			status |= CDEV_AUDIO_SPEAK_MAX_BIT;
		else
			status |= CDEV_AUDIO_SPEAK_MID_BIT;
	}

	*cdev_s = status;
}

static void set_lcd_cali_bit(unsigned long dev_s, 
				unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;
        u8 wled1,wled2;

        micco_read(MICCO_WLED_CONTROL1,&wled1);
        micco_read(MICCO_WLED_CONTROL2,&wled2);
	if ((wled2 & 0x20) && (dev_s & DEV_LCD_BIT)) {
		if (wled1 <= 30)
			status |= CDEV_LCD_BL_MIN_BIT;
		else if ((wled1 > 30) && (wled1 <= 70))
			status |= CDEV_LCD_BL_MID_BIT;
		else if ((wled1 > 70))
			status |= CDEV_LCD_BL_MAX_BIT;
	}

	*cdev_s = status;
}

static void  set_temp_cali_bit(unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;
	int temp = get_temper();

	if (temp <= -5)
		status |= CDEV_TEMP_D10_BIT;
	else if (temp > -5 && temp <= 10) 
		status |= CDEV_TEMP_0_BIT;
	else if (temp > 10 && temp <= 40) 
		status |= CDEV_TEMP_25_BIT;
	else if (temp > 40)
		status |= CDEV_TEMP_50_BIT;

	*cdev_s = status;
}

static void set_wlan_cali_bit(unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;
	
	if(gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(WLAN_PON))) > 0)
		status |= CDEV_WLAN_BIT;

	*cdev_s = status;
}

static void set_charger_bit(unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;
	u8 charger_h, charger_l, charger;

	charger_h = (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN1))) > 0 ? 1 : 0) << 1;
	charger_l = (gpio_get_value(mfp_to_gpio(MFP_CFG_PIN(CHARGE_EN2))) > 0 ? 1 : 0) << 0;
	charger = (charger_h | charger_l) & 0x03;

	if (charger == CHARGER_CURRENT_100) {
		status |= CDEV_CHARGE_100_BIT;
	}
	else if (charger == CHARGER_CURRENT_500) {
		status |= CDEV_CHARGE_500_BIT;
	}
	else if (charger == CHARGER_CURRENT_1000) {
		status |= CDEV_CHARGE_1000_BIT;
	}

	*cdev_s = status;
}

static void  set_op_cali_bit(unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;
	const char *op_name = dvfm_get_op_name(cur_op);

	if (op_name) {
		if (!strncmp(op_name, "D0CS", 4))
			status |= CDEV_OP_D0CS_BIT;
		else if (!strncmp(op_name, "104M", 4))
			status |= CDEV_OP_104M_BIT;
		else if (!strncmp(op_name, "156M", 4))
			status |= CDEV_OP_156M_BIT;
		else if (!strncmp(op_name, "208M", 4))
			status |= CDEV_OP_208M_BIT;
		else if (!strncmp(op_name, "416M", 4))
			status |= CDEV_OP_416M_BIT;
		else if (!strncmp(op_name, "624M", 4))
			status |= CDEV_OP_624M_BIT;
		else if (!strncmp(op_name, "D1", 2))
			status |= CDEV_OP_D1_BIT;
		else if (!strncmp(op_name, "D2", 2))
			status |= CDEV_OP_D2_BIT;
		else if (!strncmp(op_name, "BOOT OP", 7))
			status |= CDEV_OP_624M_BIT;
	}

	*cdev_s = status;
}

static void  set_call_cali_bit(unsigned long long *cdev_s)
{
	unsigned long long status = *cdev_s;

	if (g_is_call) {
		if ((g_call_signal >= 0) && (g_call_signal < 10))
			status |= CDEV_GSM_CALL_MIN_SIG_BIT;
		else if ((g_call_signal >= 10) && (g_call_signal < 20))
			status |= CDEV_GSM_CALL_NORMAL_SIG_BIT;
		else if (g_call_signal >= 20)
			status |= CDEV_GSM_CALL_MAX_SIG_BIT;
	}

	*cdev_s = status;
}

static void get_cali_dev_status(unsigned long dev_s, 
				unsigned long long *cdev_s)
{
	unsigned long long status = 0;
	extern int g_psy_type;

	if (dev_s & DEV_CAMERA_BIT)
		status |= CDEV_CAMERA_BIT;
	if (dev_s & DEV_MTV_BIT)
		status |= CDEV_MTV_BIT;
	if (dev_s & DEV_BT_BIT)
		status |= CDEV_BT_BIT;
	if (dev_s & DEV_MMC_BIT)
		status |= CDEV_MMC_BIT;
	if (dev_s & DEV_USB_BIT)
		status |= CDEV_USB_BIT;
	if (dev_s & DEV_GPS_BIT)
		status |= CDEV_GPS_BIT;
	if (dev_s & DEV_AUDIO_BIT)
		set_audio_cali_bit(&status);

	if (g_charge_state == POWER_SUPPLY_STATUS_CHARGING && 
		g_psy_type == BQ2407x_POWER_CHARGE_AC) {
			set_charger_bit(&status);
	}

	set_lcd_cali_bit(dev_s, &status);
	set_temp_cali_bit(&status); 
	set_op_cali_bit(&status);
	set_call_cali_bit(&status);
	set_wlan_cali_bit(&status);
	*cdev_s = status;
}

static int get_actual_volt(int volt_mv) 
{ 
	unsigned long dev_s; 
	unsigned long long cdev_s;
	int volt0 = volt_mv - 65, volt1, volt2, i = 1, j;
	char dev_buf[100];

	volt1 = volt0;
	memset(dev_buf, 0, sizeof(dev_buf));

	/* the calibration of the voltage */
	get_pstatus(&dev_s);
	get_cali_dev_status(dev_s, &cdev_s);

	do {
		if (volt_mv >= volt_dev_table[i][0]) {
			for (j = 0; j < CDEV_NUM; j++) 
				if (cdev_s & (1 << j)) {
					volt1 += volt_dev_table[i][j+1];
					if (strlen(dev_buf))
                        			sprintf(dev_buf, "%s,%d", dev_buf, j);
                			else
                        			sprintf(dev_buf, "%d", j);
				}
			break;
		}
	} while (volt_dev_table[i++][0]);

	/* the modification of the voltage */
	i = 0;
	do {
		if ((volt2 = volt1) >= volt_modi_table[i][0]) {
			volt2 += volt_modi_table[i][1];
			break;
		}
	} while (volt_modi_table[i++][0]);

	TDBG("v = %d, v0 = %d, v1 = %d, v2 = %d, dev = 0x%lx, [%s]\n",
		volt_mv, volt0, volt1, volt2, dev_s, dev_buf);

	return ((volt2 >= g_bat_volt_max) ? g_bat_volt_max : volt2);
}

static int read_vbat_adcs(u16 *adc)
{
	int i;

	for (i = 0; i < READ_BAT_VOLT_MAX_NUM; i++) 
		if (!read_vbat_adc(adc)) 
			return 0;
		else {
			TWARN("ignore adc value, times: %d\n", i);
			udelay(800);
		}

	return -1;
}

static int read_NVM(void)
{
	void *addr = NULL;
	struct mtd_info *mtd = NULL;
	int retlen;

	mtd = get_mtd_device_nm("NVM");
	if (mtd == NULL)
		return -ENODEV;
	addr = kmalloc(NVM_DATA_SIZE, GFP_KERNEL);
	if (!addr)
		return -ENOMEM;

	/* NVM data begin address from 40k */
	if (mtd->writesize == 2048)
		mtd->read(mtd, NVM_RESERVE_SIZE, NVM_DATA_SIZE, &retlen, addr);

	TINFO("load nvm into address: 0x%x, retlen: %d\n", (unsigned int)addr, retlen);
	return (int)addr;
}

static void get_NVM_cali_value(void)
{
	struct cali_volt_data cali_volt_data;
	struct cali_temp_data cali_temp_data;
	void *nvm = (void *)read_NVM();

	memset(&cali_volt_data, 0x00, sizeof(cali_volt_data));
	memcpy(&cali_volt_data, nvm + NVM_VOLT_OFFSET, sizeof(cali_volt_data));
	TDBG("voltage calibration: %d, %d, %d, %d, %d\n", 
		cali_volt_data.len,
		cali_volt_data.min_volt,
		cali_volt_data.min_adc,
		cali_volt_data.max_volt,
		cali_volt_data.max_adc);

	if ((cali_volt_data.len == 16)
	    && (cali_volt_data.min_volt >= 3300)
	    && (cali_volt_data.min_volt <= 3600)
	    && (cali_volt_data.min_adc > 100)
	    && (cali_volt_data.min_adc < 300)
	    && (cali_volt_data.max_volt >= 4100)
	    && (cali_volt_data.max_volt <= 4200)
	    && (cali_volt_data.max_adc > 500)
	    && (cali_volt_data.max_adc < 700)
	    && (cali_volt_data.min_adc < cali_volt_data.max_adc)
	    && (cali_volt_data.min_volt < cali_volt_data.max_volt)) {
		g_bat_adc_min = cali_volt_data.min_adc;
		g_bat_adc_max = cali_volt_data.max_adc;
		g_bat_volt_max = cali_volt_data.max_volt;
		g_bat_volt_min = cali_volt_data.min_volt;
	}

	memset(&cali_temp_data, 0x00, sizeof(cali_temp_data));
	memcpy(&cali_temp_data, nvm + NVM_TEMP_OFFSET, sizeof(cali_temp_data));
	TDBG("temperature calibration: %d, %d, %d, %d, %d\n", 
		cali_temp_data.len,
		cali_temp_data.min_temp,
		cali_temp_data.min_adc,
		cali_temp_data.max_temp,
		cali_temp_data.max_adc);

	if ((cali_temp_data.len == 16)
	    && (cali_temp_data.min_temp >= -20)
	    && (cali_temp_data.min_temp <= 10)
	    && (cali_temp_data.max_temp >= 50)
	    && (cali_temp_data.max_temp <= 80)
	    && (cali_temp_data.min_adc < cali_temp_data.max_adc)
	    && (cali_temp_data.min_temp < cali_temp_data.max_temp)) {
		g_temp_adc_min = cali_temp_data.min_adc;
		g_temp_adc_max = cali_temp_data.max_adc;
		g_bat_temp_max = cali_temp_data.max_temp;
		g_bat_temp_min = cali_temp_data.min_temp;
	}
}

static int get_voltage(void)
{
	int ret, volt_mv; 
	u16 adc = 0;

	/* If need to get the more accurate voltage value of battery when 
 	 * charger is pluged in, the charge IC should be suspended. But the 
 	 * frequent charging/discharging operation is harmful to the life 
 	 * of battery, so if battery is charging then set charge current to 
 	 * 100mA in order to read original voltage data. */
	ret = read_vbat_adcs(&adc);

	if (ret) {
		if (old_bat_voltage) {
			TWARN("Warning: failed to read voltage, use last voltage\n");
			return old_bat_voltage;
		}
		else {
			TWARN("failed to read voltage, use average voltage\n");
			adc = get_adc_average(MICCO_ADC_MUX_VBAT);
		}
	}

	if (time_after(jiffies, init_jiff + msecs_to_jiffies(SYSTEM_INIT_TIME))
			&& (!g_adc_calibrated)) {
		get_NVM_cali_value();
		g_adc_calibrated = 1;
	}

	volt_mv = bat_adc2volt(adc);
	volt_mv = get_actual_volt(volt_mv);

	if (((old_bat_voltage - volt_mv) >= BAT_VOLT_FLOAT_BURST) && (g_is_call)) {
		TWARN("burst voltage, ignored\n");
		return old_bat_voltage;
	}

	if (((old_bat_voltage - volt_mv) > BAT_VOLT_FLOAT_MAX) 
	 		|| ((volt_mv - old_bat_voltage) > BAT_VOLT_FLOAT_MAX)) 
		TWARN("voltage, new: %d, old: %d\n", volt_mv, old_bat_voltage);

	/* If system is booting, or don't get ADC calibration data,
 	   we don't compare voltage value */
	if( time_after(jiffies, init_jiff + msecs_to_jiffies(SYSTEM_INIT_TIME))
			&& g_adc_calibrated) {
		if ((POWER_SUPPLY_STATUS_DISCHARGING == g_charge_state) 
				&& (old_bat_voltage != 0)
				&& ( volt_mv > old_bat_voltage))
			volt_mv = old_bat_voltage;
		else if ((POWER_SUPPLY_STATUS_CHARGING == g_charge_state) 
				&& ( volt_mv < old_bat_voltage))
			volt_mv = old_bat_voltage;
	}
	
	old_bat_voltage = volt_mv;
	return volt_mv;
}

static int set_vbat_mon(unsigned int vol, unsigned int bat_state)
{
	if (!g_adc_is_normal)
		return 0;

	TDBG("vol = %d, bat_state = %d\n", vol, bat_state);
	if (bat_state != BAT_STATE_UNKNOW)
		g_battery_state = bat_state;

	if (!vol)
		return micco_write(MICCO_VBATMON, vol);
	else
		return micco_write(MICCO_VBATMON, (bat_volt2adc(vol) >> 2));
}

static int bat_volt2mah(int volt_mv)
{
	int i;

	if (volt_mv >= g_bat_volt_max)
		return BAT_CAPACITY_MAX_MAH;

	for (i = 1; i < BAT_CAPACITY_LEVEL_NUM; i++) 
		if (volt_mv >= capa_volt[i])
			return (capa_mah[i-1] + (int)((volt_mv - capa_volt[i-1])
				* (capa_mah[i] - capa_mah[i-1]) /
			 	(capa_volt[i] - capa_volt[i-1])));

	return BAT_CAPACITY_MIN_MAH;
}

static int bat_mah2capacity(int mah)
{
	int cap = (mah * 100) / BAT_CAPACITY_MAX_MAH;

	if ((g_charge_state == POWER_SUPPLY_STATUS_CHARGING) && (cap == 100))
		if (!charge_is_full) {
			if (!full_jiff) {
				if(!time_after(jiffies, full_jiff + msecs_to_jiffies(CHARGE_FULL_MIN_TIME)))
					cap = 99;
				else
					charge_is_full = 1;
			} else
				full_jiff = jiffies;
		}
	return cap;
}

static int bat_volt2capacity(int volt)
{
 	int i = 0;

        do {
                if (volt >= volt2capa_table[i][0])
			return volt2capa_table[i][1];
        } while (volt2capa_table[++i][0]);

	return volt2capa_table[i][1];
}

static int pxa_micco_battery_read_status(struct pxa_micco_battery_info *di)
{
	u16 adc = 0;
        
	if (di->update_time && time_before(jiffies, di->update_time +
					   msecs_to_jiffies(cache_time)))
		return 0;

	di->update_time = jiffies;
	di->temp_C = get_temper();
	di->health = POWER_SUPPLY_HEALTH_GOOD ;
	di->present = 1;

	// Add for NPI read adc.    
	read_vbat_adcs(&adc);
	di->vol_adc = (int)adc;

	if (!g_adc_is_normal) {
			di->voltage_mV = DEFAULT_BAT_NORMAL_VOLT;
			di->capacity = DEFAULT_BAT_NORMAL_CAPACITY;
	} else {
		unsigned long last_jiff = jiffies;	
		//int mah;

		di->voltage_mV = get_voltage();
#if 0
		mah = bat_volt2mah(di->voltage_mV);
		di->capacity = bat_mah2capacity(mah);
#endif
		di->capacity = bat_volt2capacity(di->voltage_mV);

		if (g_charge_state != POWER_SUPPLY_STATUS_CHARGING) {
			if ((BAT_STATE_TOOLOW == g_battery_state) || 
                   	   		(di->voltage_mV <= BATTERY_TOOLOW_VOLTAGE)) {
				di->voltage_mV = BATTERY_TOOLOW_VOLTAGE;
				di->capacity = BATTERY_TOOLOW_CAPACITY;
			} else if ((di->voltage_mV > BATTERY_TOOLOW_VOLTAGE) && 
		        		(di->voltage_mV <= BATTERY_LOW_VOLTAGE)) {
				if (BAT_STATE_NORMAL == g_battery_state) 
					set_vbat_mon(VBATMON_NORMAL_TOOLOW_VOLTAGE, BAT_STATE_LOW);
			}
		}

		pr_debug("battery: voltage= %d, charge_state= %d, bat_state: %d, capacity= %d, temp: %d, jiffies= %lu\n", 
				di->voltage_mV, di->charge_status, g_battery_state, di->capacity, di->temp_C, jiffies - last_jiff);
	}
	return 0;
}

static void pxa_micco_battery_update_status(struct pxa_micco_battery_info *di)
{
	int last_capacity = di->capacity;

	pxa_micco_battery_read_status(di);

	if (power_supply_am_i_supplied(&di->bat)) {
		di->charge_status = POWER_SUPPLY_STATUS_CHARGING;
		if (g_charge_state != POWER_SUPPLY_STATUS_CHARGING)
			set_vbat_mon(VBATMON_DISABLE_VOLTAGE, BAT_STATE_NORMAL);
	} else {
		full_jiff = 0;
		charge_is_full = 0;
		di->charge_status = POWER_SUPPLY_STATUS_DISCHARGING;
		if (POWER_SUPPLY_STATUS_CHARGING == g_charge_state)
			set_vbat_mon(VBATMON_NORMAL_LOW_VOLTAGE, BAT_STATE_NORMAL);
	}

	g_charge_state = di->charge_status;
	if (last_capacity!=di->capacity)
		power_supply_changed(&di->bat);
}

static void pxa_micco_battery_work(struct work_struct *work)
{
	int interval;
	struct pxa_micco_battery_info *di = container_of(work,
			struct pxa_micco_battery_info, monitor_work.work);

	pxa_micco_battery_update_status(di);
	if ((di->voltage_mV <= BATT_MONITOR_VOLT_LOW) &&
			(di->charge_status == POWER_SUPPLY_STATUS_DISCHARGING))
		interval = LOW_VOLT_MONITOR_INTERVAL;
	else if ((di->voltage_mV >= BATT_MONITOR_VOLT_HIGH) &&  
			(di->voltage_mV  != DEFAULT_BAT_VOLT_MAX) &&
			(di->charge_status == POWER_SUPPLY_STATUS_CHARGING))
		interval = HIGH_VOLT_MONITOR_INTERVAL;
	else
		interval = DEFAULT_BAT_MONITOR_INTERVAL;

	cancel_delayed_work(&di->monitor_work);
	queue_delayed_work(di->monitor_wqueue, &di->monitor_work, interval);
}

#define to_pxa_micco_battery_info(x) container_of((x), struct pxa_micco_battery_info, \
					      bat);

static void pxa_micco_battery_external_power_changed(struct power_supply *psy)
{
	struct pxa_micco_battery_info *di = to_pxa_micco_battery_info(psy);

	cancel_delayed_work(&di->monitor_work);
	queue_delayed_work(di->monitor_wqueue, 
		&di->monitor_work, BAT_MONITOR_QUICK_INTERVAL);
}

static int pxa_call_status_put_property(struct power_supply *psy,
				       enum power_supply_property psp,
				       union power_supply_propval *val)
{
	struct pxa_micco_battery_info *di = to_pxa_micco_battery_info(psy);

	switch (psp) {
	case POWER_SUPPLY_PROP_CALL_STATUS:
		sscanf(val->strval, "%d,%d", &g_is_call, &g_call_signal);
		if (((g_is_call != 0) && (g_is_call != 1)) || 
				(g_call_signal < 0) || (g_call_signal > 100))
			return -EINVAL;
		else
			return 0;
#if 0
	case POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW:
		di->vol_toolow = !!val->intval;
		break;
#endif
	default:
		break;
	}

	return 0;
}

static int pxa_micco_battery_get_property(struct power_supply *psy,
				       enum power_supply_property psp,
				       union power_supply_propval *val)
{
	struct pxa_micco_battery_info *di = to_pxa_micco_battery_info(psy);

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
		if (power_supply_am_i_supplied(&di->bat))
			val->intval = POWER_SUPPLY_STATUS_CHARGING;
		else
			val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
		return 0;
	default:
		break;
	}

	//pxa_micco_battery_read_status(di);

	switch (psp) {
	case POWER_SUPPLY_PROP_HEALTH:
		val->intval = di->health;
		break;
	case POWER_SUPPLY_PROP_PRESENT:
		val->intval = di->present;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_AVG:   // Add for NPI read adc.
		val->intval = di->vol_adc;
		break;
	case POWER_SUPPLY_PROP_CAPACITY:
		val->intval = di->capacity;
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:
		val->intval = di->technology;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
		val->intval = di->voltage_mV;
		break;
#if 0
	case POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW:
		val->intval = di->vol_toolow;
		break;
#endif
	case POWER_SUPPLY_PROP_TEMP:
		val->intval = di->temp_C;
		break;
	case POWER_SUPPLY_PROP_CALL_STATUS:
		sprintf(call_status, "%d,%d", g_is_call, g_call_signal);
		val->strval = call_status;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static void battery_low_interrupt(unsigned long event)
{
	u8 val;
	//union power_supply_propval propval;

#ifdef CONFIG_ANDROID_POWER
	android_lock_suspend_auto_expire(&pxa_battery_lock, 5 * HZ);
#endif
	if (!(event & PMIC_EVENT_VBATMON)) 
		return ;

	micco_read(MICCO_VBATMIN_RES, &val);
	TINFO("VBATMIN_RES = 0x%0x\n", val);
	micco_read(MICCO_STATUS_A, &val);
	if (!(val & MICCO_STATUS_A_VBATMON)) {
		return ;
	}
	TDBG("STATUS_A = 0x%0x\n", val);

	switch (g_battery_state) {
	case BAT_STATE_NORMAL:
		TDBG("Battery voltage is low.\n");
		set_vbat_mon(VBATMON_NORMAL_TOOLOW_VOLTAGE, BAT_STATE_LOW);
		break;
#if 0
	case BAT_STATE_LOW:
		TWARN("%s: Battery voltage is too low, force power down.\n", __func__);
		propval.intval = 1;
		bat_di->bat.put_property(&bat_di->bat,POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW,&propval);
		power_supply_changed(&bat_di->bat);
		break;
#endif
	default:
		break;
	}

	cancel_delayed_work(&bat_di->monitor_work);
	queue_delayed_work(bat_di->monitor_wqueue, 
		&bat_di->monitor_work, BAT_MONITOR_QUICK_INTERVAL);
	return ;
}

static enum power_supply_property pxa_micco_battery_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_VOLTAGE_AVG,   // Add for NPI read adc.
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_CAPACITY,
	POWER_SUPPLY_PROP_TECHNOLOGY,
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
//	POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW,
	POWER_SUPPLY_PROP_TEMP,
	POWER_SUPPLY_PROP_CALL_STATUS,
};

static void micco_vbatmon_init(void)
{
	u8 val;
	micco_read(MICCO_VBATMIN_RES, &val);

	micco_read(MICCO_IRQ_MASK_A, &val);
	val &= ~IRQ_MASK_A_VBATMON;
	micco_write(MICCO_IRQ_MASK_A, val);

	set_vbat_mon(VBATMON_NORMAL_LOW_VOLTAGE, BAT_STATE_NORMAL);
}

static int __init pxa_micco_battery_probe(struct platform_device *pdev)
{
	int retval = 0;
	struct pxa_micco_battery_info *di;
	struct micco_battery_pdata *pdata;

	init_jiff = jiffies;

	adc_max_float_value = bat_fvolt2adc(BAT_VOLT_FLOAT_MIN);
	TDBG("adc_max_float_value is 0x%x\n", adc_max_float_value);

	di = kzalloc(sizeof(*di), GFP_KERNEL);
	if (!di) {
		retval = -ENOMEM;
		goto di_alloc_failed;
	}

	platform_set_drvdata(pdev, di);

	pdata = pdev->dev.platform_data;
	di->dev	= &pdev->dev;
	di->bat.name = pdev->dev.bus_id;
	di->bat.type = POWER_SUPPLY_TYPE_BATTERY;
	di->bat.properties = pxa_micco_battery_props;
	di->bat.num_properties = ARRAY_SIZE(pxa_micco_battery_props);
	di->bat.get_property = pxa_micco_battery_get_property;
	di->bat.put_property = pxa_call_status_put_property;
	di->bat.external_power_changed = pxa_micco_battery_external_power_changed;
	di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;
	di->technology = POWER_SUPPLY_TECHNOLOGY_LION;
	di->vol_toolow = 0;
	memset(call_status, 0, sizeof(call_status));

	if (!micco_adc_read(MICCO_ADC_MUX_VBAT)) {
		TWARN("read voltage value error from micco adc\n");
		g_adc_is_normal = 0;
	}

	retval = power_supply_register(&pdev->dev, &di->bat);
	if (retval) {
		TERR("failed to register battery\n");
		goto batt_failed;
	}
	
	micco_vbatmon_init();

	retval = pmic_callback_register(PMIC_EVENT_VBATMON, battery_low_interrupt);
	if (retval < 0) {
                TERR("Cann't callback register vbatmon.\n");
		goto pmic_callback_failed;
	}

	INIT_DELAYED_WORK(&di->monitor_work, pxa_micco_battery_work);
	di->monitor_wqueue = create_singlethread_workqueue(pdev->dev.bus_id);
	if (!di->monitor_wqueue) {
                TERR("failed to create work queue.\n");
		retval = -ESRCH;
		goto workqueue_failed;
	}
	bat_di = di;
	queue_delayed_work(di->monitor_wqueue, 
		&di->monitor_work, BAT_MONITOR_INIT_INTERVAL);

	goto success;

workqueue_failed:
	pmic_callback_unregister(PMIC_EVENT_VBATMON, battery_low_interrupt);
pmic_callback_failed:
	power_supply_unregister(&di->bat);
batt_failed:
	kfree(di);
di_alloc_failed:
success:
	return retval;
}

static int pxa_micco_battery_remove(struct platform_device *pdev)
{
	struct pxa_micco_battery_info *di = platform_get_drvdata(pdev);

	cancel_rearming_delayed_workqueue(di->monitor_wqueue,
					  &di->monitor_work);
	destroy_workqueue(di->monitor_wqueue);
	power_supply_unregister(&di->bat);
	pmic_callback_unregister(PMIC_EVENT_VBATMON, battery_low_interrupt);

	return 0;
}

#ifdef CONFIG_PM
#ifdef CONFIG_ANDROID_POWER
void battery_android_suspend_handler(android_early_suspend_t *h)
{
	printk(KERN_DEBUG "%s enter.\n", __func__);
	if (g_charge_state != POWER_SUPPLY_STATUS_CHARGING) {
		cancel_delayed_work(&bat_di->monitor_work);
		if (BAT_STATE_NORMAL == g_battery_state) 
			set_vbat_mon(VBATMON_SUSPEND_LOW_VOLTAGE, BAT_STATE_NORMAL);
		else if ((BAT_STATE_LOW == g_battery_state) ||
				(BAT_STATE_TOOLOW == g_battery_state))
			set_vbat_mon(VBATMON_SUSPEND_TOOLOW_VOLTAGE, BAT_STATE_LOW);
	}
}

void battery_android_resume_handler(android_early_suspend_t *h)
{
	printk(KERN_DEBUG "%s enter.\n", __func__);
//	bat_di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;
//	power_supply_changed(&bat_di->bat);

	if (g_charge_state != POWER_SUPPLY_STATUS_CHARGING) {
		if (BAT_STATE_NORMAL == g_battery_state) 
			set_vbat_mon(VBATMON_NORMAL_LOW_VOLTAGE, BAT_STATE_NORMAL);
		else if (BAT_STATE_LOW == g_battery_state) 
			set_vbat_mon(VBATMON_NORMAL_TOOLOW_VOLTAGE, BAT_STATE_LOW);
		queue_delayed_work(bat_di->monitor_wqueue, &bat_di->monitor_work, HZ);
	}
}

static android_early_suspend_t battery_android_suspend = {
        .level = 5,
	.suspend = battery_android_suspend_handler,
	.resume = battery_android_resume_handler,
};
#endif
static int pxa_micco_battery_suspend(struct platform_device *pdev,
				  pm_message_t state)
{
	struct pxa_micco_battery_info *di = platform_get_drvdata(pdev);

	di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;

	return 0;
}

static int pxa_micco_battery_resume(struct platform_device *pdev)
{
	struct pxa_micco_battery_info *di = platform_get_drvdata(pdev);

	di->charge_status = POWER_SUPPLY_STATUS_UNKNOWN;
	power_supply_changed(&di->bat);

	cancel_delayed_work(&di->monitor_work);
	queue_delayed_work(di->monitor_wqueue, &di->monitor_work, HZ);

	return 0;
}

#else

#define pxa_micco_battery_suspend NULL
#define pxa_micco_battery_resume NULL

#endif /* CONFIG_PM */

static struct platform_driver pxa_battery_driver = {
	.driver = {
		.name = "battery",
	},
	.probe	  = pxa_micco_battery_probe,
	.remove   = pxa_micco_battery_remove,
	.suspend  = pxa_micco_battery_suspend,
	.resume	  = pxa_micco_battery_resume,
};

static int __init pxa_micco_battery_init(void)
{
#ifdef CONFIG_ANDROID_POWER
	android_init_suspend_lock(&pxa_battery_lock);
	//android_register_early_suspend(&battery_android_suspend);
#endif
	return platform_driver_register(&pxa_battery_driver);
}

static void __exit pxa_micco_battery_exit(void)
{
#ifdef CONFIG_ANDROID_POWER
	android_uninit_suspend_lock(&pxa_battery_lock);
	//android_unregister_early_suspend(&battery_android_suspend);
#endif
	platform_driver_unregister(&pxa_battery_driver);
}

module_init(pxa_micco_battery_init);
module_exit(pxa_micco_battery_exit);

MODULE_LICENSE("GPL");
