#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/spinlock.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/proc_fs.h>

#include "udc_sysfs.h"

static struct class *udc_class;
static struct device *udc_dev;
static struct udc_sysfs_data udc_sysfs_data;

static ssize_t switch_mode_show(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	return sprintf(buf, "%d\n", udc_sysfs_data.mode);
}
static ssize_t switch_testmode_store(struct device *dev,
				           struct device_attribute *attr,
				           const char *buf, size_t count)
{
	
	if(count<3){
	
		switch(buf[0]){
		case '1':	//input "echo 1 > testmode " in the console will switch to usbnet mode
			udc_sysfs_data.testmode=1;
			udc_sysfs_data.mode=SWITCH_T0_ETHER;
			kobject_uevent(&udc_dev->kobj, KOBJ_CHANGE);
			break;
		case '2': 	//input "echo 2 > testmode " in the console will switch to usbserial with acm mode
        	        udc_sysfs_data.testmode=2;
                	udc_sysfs_data.mode=SWITCH_TO_SERIAL_ACM;
	                kobject_uevent(&udc_dev->kobj, KOBJ_CHANGE);
			break;
        
	        case '3':	//input "echo 3 > testmode " in the console will switch to usbserial without acn mode
        	        udc_sysfs_data.testmode=3;
	        	udc_sysfs_data.mode=SWITCH_TO_SERIAL_NOACM;
	                kobject_uevent(&udc_dev->kobj, KOBJ_CHANGE);
			break;

	        case '4':	//input "echo 4 > testmode " in the console will switch to mtp mode
	                udc_sysfs_data.testmode=4;
        	        udc_sysfs_data.mode=SWITCH_TO_MTP;
                	kobject_uevent(&udc_dev->kobj, KOBJ_CHANGE);
			break;
		}
	}
	printk("count=%d",count);	
	return count;
}
static ssize_t switch_net_show(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	return sprintf(buf, "%d\n", udc_sysfs_data.net_enum);
}
static ssize_t switch_testmode_show(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	return sprintf(buf, "%d\n", udc_sysfs_data.testmode);
}
DEVICE_ATTR(mode, S_IRUGO, switch_mode_show, NULL);
DEVICE_ATTR(net_enum, S_IRUGO, switch_net_show, NULL);
DEVICE_ATTR(testmode, 644, switch_testmode_show, switch_testmode_store);
static int udc_uevent(struct device *dev, struct kobj_uevent_env *env)
{
	if (add_uevent_var(env, "MODE=%d", udc_sysfs_data.mode))
            return -ENOMEM;
	if (add_uevent_var(env, "NET_ENUM=%d", udc_sysfs_data.net_enum))
            return -ENOMEM;
	if (add_uevent_var(env, "TESTMODE=%d", udc_sysfs_data.testmode))
            return -ENOMEM;
	return 0;
}

void udc_kobject_uevent(struct udc_sysfs_data *data, 
			enum kobject_action action,bool uevent)
{
	udc_sysfs_data.mode = data->mode;
	udc_sysfs_data.net_enum = data->net_enum;
	if(data->net_enum!=0)
		udc_sysfs_data.testmode=1;
	else
		udc_sysfs_data.testmode=0;
	if(uevent)
		kobject_uevent(&udc_dev->kobj, action);
}

int udc_sysfs_register(struct device *dev,char *name)
{
	int ret = 0;
	int rc = 0;
	if(name==NULL)
		goto device_create_failed;
	udc_dev = device_create(udc_class, dev, 0, "%s", name);
         if (IS_ERR(udc_dev)) {
                ret = PTR_ERR(udc_dev);
		printk("device_create failed\n");
		goto device_create_failed;
        }
       	rc = device_create_file(udc_dev,&dev_attr_mode);
	if (rc)
		goto groups_create_failed;
	rc = device_create_file(udc_dev,&dev_attr_net_enum);
	if (rc)
		goto groups_create_failed;
	rc = device_create_file(udc_dev,&dev_attr_testmode);
	if (rc)
		goto groups_create_failed;
	goto success;
	
groups_create_failed:
	device_unregister(udc_dev);
device_create_failed:
success:
	return ret;
}

void udc_sysfs_unregister(void)
{
	
	device_remove_file(udc_dev,&dev_attr_mode);
	device_remove_file(udc_dev,&dev_attr_net_enum);
	device_remove_file(udc_dev,&dev_attr_testmode);
	device_unregister(udc_dev);

}

static int __init udc_class_init(void)
{
	udc_class = class_create(THIS_MODULE, "udc");
        printk(KERN_DEBUG "udc class init\n");
	if (IS_ERR(udc_class))
		return PTR_ERR(udc_class);

	udc_class->dev_uevent = udc_uevent;
	memset(&udc_sysfs_data, 0, sizeof(udc_sysfs_data));

	return 0;
}

static void __exit udc_class_exit(void)
{
	class_destroy(udc_class);
}
EXPORT_SYMBOL_GPL(udc_sysfs_register);
EXPORT_SYMBOL_GPL(udc_sysfs_unregister);
EXPORT_SYMBOL_GPL(udc_kobject_uevent);

subsys_initcall(udc_class_init);
module_exit(udc_class_exit);
