/*
 *  linux/drivers/video/pxa950fb.c
 */

#include "pxa950fb.h"

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/uaccess.h>
#include <linux/kthread.h>

#include <mach/io.h>
#include <mach/irqs.h>
#include <mach/hardware.h>
#include <mach/pxa3xx-regs.h>

#include <mach/pxa950fb.h>
#include <mach/micco.h>

#ifdef CONFIG_PXA3xx_DVFM
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

extern void enable_oscc_tout_s0(void);
static volatile LCD_Controller_Configurations LCD_Conf;
static void controller_enable_disable(struct pxa950fb_info *fbi, int onoff);

static int display_enabled = 0;
static int display_update_ongoing = 0;

u32 *dma_addr_cpu;
u32 dma_addr_dma;

pxa950fb_registers pxa950fb_regs;

#define MAX_DSI_DESC 30
/*b- base address (of display controller); c - LCD_Controller_Converter*/
#define	CONVERTER_BASE_ADDRESS(b, c)	(b + converter_offset[c])
/*LCD_Controller_Converter enum to actual converter offset LUT*/
u32 converter_offset[] = {
0x1000,/*LCD_Controller_PARALLEL*/
0x2000,/*LCD_Controller_DSI0*/
0x3000,/*LCD_Controller_DSI1*/
};
#define DSI_CLOCK_VALUE(d)	(dsi_clock_val[d])
/*LCD_Controller_DSI_Clock enum to actual frequency value LUT*/
u32 dsi_clock_val[] = {
52,	/*LCD_Controller_DSI_52MHZ*/
104,/*LCD_Controller_DSI_104MHZ*/
156,/*LCD_Controller_DSI_156MHZ*/
208,/*LCD_Controller_DSI_208MHZ*/
312,/*LCD_Controller_DSI_312MHZ*/
0,/*LCD_Controller_DSI_UNDEFINEDMHZ*/
624,/*LCD_Controller_DSI_624MHZ*/
};
u8 dsi_video_mode;
u16 DSI_COMMAND_BUFFER[128];
u32 LCD_DMAD_DSI[4][MAX_DSI_DESC];    /* For DSI - DMA*/
u32 ber_count;
u8 BER_ARRAY[100];
u32 dsi_dcs_long_write[10];

static void *MMCUbase;
static u8 dsi_clk_shadow[3];
static volatile u8 dsi_lanes_number;
static volatile u8 dsi_power_mode;
static volatile u8 dsi_rgb_mode;
static volatile u8 dsi_video_run_mode = 0;

static volatile DSI_Controller_Configurations DSI_Conf;
static volatile FRAME_Controller_Configurations FRAME_Conf;

#if defined(CONFIG_FBCON_DRAW_PANIC_TEXT)
extern int (*fbcon_draw_panic_text_callback)(void);
extern int panic_text_displayed;
#endif

u8 Init_Board [][BOARD_INIT_MAX_DATA_BYTES] = {
	/*wait, command,0,  1st,  2nd,  3rd,  4th,  5th  ,  6th  , 7th  , 8th,  9th,  10th*/
	{ 0xFF,LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER,LCD_Controller_DCS_SOFT_RESET}, /*Soft reset*/
	{ 10,LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER,LCD_Controller_DCS_NOP}, /*Nop*/
	/*wait 10ms*/
	{ 0xFF,LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER,LCD_Controller_DCS_EXIT_SLEEP_MODE}, /* Exit sleep*/
	{ 20,LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER,LCD_Controller_DCS_NOP}, /*Nop*/
	/*wait 20ms*/
	{ 1,LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER,LCD_Controller_DCS_SET_OUTPUT_PIXEL_CLOCK_FREQUENCY,0x52}, /* SET_OUTPUT_CLOCK_FREQUENCY*/
	{ 1,LCD_Controller_DCS_LONG_WRITE,0x05,0x00,LCD_Controller_DCS_ENABLE_SET_SPECIAL_COMMAND,0x03,0x7F,0x5C,0x33}, /* Enable_Set_Special_command*/
	{ 1,LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER,LCD_Controller_DCS_DISPLAY_BUFFER_IO_CONTROL,0x05}, /*2 THSSI channels and RGB16*/
	/* { 1,LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER,LCD_Controller_DCS_SET_TEAR_ON,0x00}, //Set TEAR ON */
	{ 1,LCD_Controller_DCS_LONG_WRITE,0x07,0x00,LCD_Controller_DCS_SET_OUTPUT_VERTICAL_TIMINGS,0x04,0x09,0x09,0x00,0x02,0x80}, /*set_output_vertical_timing*/
	{ 1,LCD_Controller_DCS_LONG_WRITE,0x07,0x00,LCD_Controller_DCS_SET_OUTPUT_HORIZONTAL_TIMINGS,0x28,0x13,0x3B,0x00,0x01,0xE0}, /*set_output_horizontal_timing*/
	/*wait 1ms*/
	{ 1, LCD_Controller_DCS_LONG_WRITE,0x05,0x00 , LCD_Controller_DCS_WRITE_EDISCO_REGISTER , 0xCC, 0x00, 0x00, 0x0E}, /* MRESET*/
	{ 1, LCD_Controller_DCS_LONG_WRITE,0x05,0x00 , LCD_Controller_DCS_WRITE_EDISCO_REGISTER , 0xC8, 0x00, 0x72, 0x42}, /* IOCTRL*/
	{ 1, LCD_Controller_DCS_LONG_WRITE,0x05,0x00 , LCD_Controller_DCS_WRITE_EDISCO_REGISTER , 0x44, 0x00, 0xFF, 0x00}, /* SSITIM1*/
	{ 1, LCD_Controller_DCS_LONG_WRITE,0x05,0x00 , LCD_Controller_DCS_WRITE_EDISCO_REGISTER , 0x40, 0x14, 0x00, 0x1A}, /* SSICTL*/
	/*wait 1ms*/
	{ 1,LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER, LCD_Controller_DCS_GET_PIXEL_FORMAT , 0x05}, /* RGB565*/
	{ 1,LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER,LCD_Controller_DCS_SET_DISPLAY_ON}, /* Display on*/
	{ 1,LCD_Controller_DCS_LONG_WRITE,0x05,0x00,LCD_Controller_DCS_WRITE_EDISCO_REGISTER,0x20,0x00,0x00,0x24}, /*LCDCTRL*/
	/*wait 1ms*/
	{0xFF,0xFF},
};

void pxa950fb_writel(unsigned int val, unsigned int *reg_base, unsigned int offset)
{
	/* if display update is ongoing, register write should be avoided */
	if (display_enabled) {
		while(display_update_ongoing)
			msleep(1);
	}
	if (LCD_Conf.LCD_Mixer[0] != LCD_MIXER_DISABLE) {
		if ((offset == LCD_MIXER0_CTL0) && (val & LCD_MIXERx_CTL0_DISP_UPDATE) && display_enabled)
			display_update_ongoing = 1;
	}
	if (LCD_Conf.LCD_Mixer[1] != LCD_MIXER_DISABLE) {
		if ((offset == LCD_MIXER1_CTL0) && (val & LCD_MIXERx_CTL0_DISP_UPDATE) && display_enabled)
			display_update_ongoing = 1;
	}
	writel(val, (unsigned int *)((unsigned int)reg_base + offset));
	return;
}

#ifdef CONFIG_PXA3xx_DVFM
static int dvfm_dev_idx;
static void set_dvfm_constraint(void)
{
	/* Disable Lowpower mode */
	dvfm_disable_op_name("D0CS", dvfm_dev_idx);
	dvfm_disable_op_name("D1", dvfm_dev_idx);
	dvfm_disable_op_name("D2", dvfm_dev_idx);
	if (cpu_is_pxa935() || cpu_is_pxa950())
		dvfm_disable_op_name("CG", dvfm_dev_idx);
	/* Disable below frequency since they only support DSI clock up to 156Mhz */
	dvfm_disable_op_name("156M", dvfm_dev_idx);
	dvfm_disable_op_name("208M", dvfm_dev_idx);
	dvfm_disable_op_name("156M_HF", dvfm_dev_idx);

}

static void unset_dvfm_constraint(void)
{
	/* Enable Lowpower mode */
	dvfm_enable_op_name("D0CS", dvfm_dev_idx);
	dvfm_enable_op_name("D1", dvfm_dev_idx);
	dvfm_enable_op_name("D2", dvfm_dev_idx);
	if (cpu_is_pxa935() || cpu_is_pxa950())
		dvfm_enable_op_name("CG", dvfm_dev_idx);
	/* Enable frequency */
	dvfm_enable_op_name("156M", dvfm_dev_idx);
	dvfm_enable_op_name("208M", dvfm_dev_idx);
	dvfm_enable_op_name("156M_HF", dvfm_dev_idx);
}

#else
static void set_dvfm_constraint(void) {}
static void unset_dvfm_constraint(void) {}
#endif

/*PLL formulas
  DSIbitclock  = PLL(CKOUT)/OUTDIV
 PLL(CKOUT) = (13MHz �FBDIV + 2))/(3�CKSEL+1))
 */
u32 dsi_Calc_fbdiv(u16 dsi_clk, u8 outdiv, u8 cksel)
{
    u32 ckout, fbdiv;

	/*calc actual outdiv*/
	outdiv = 16 - outdiv;

	/*calc actual PLL(CKOUT)*/
	ckout = dsi_clk/outdiv;

    /*Calc fbdiv*/
	fbdiv = ((ckout * (3*(cksel + 1)))/13) - 2;

	return (u32)fbdiv;
}

void lcd_set_dsi_clock(u8 dsi_clk, u8 converter)
{
	if (cpu_is_pxa950_z0() || cpu_is_pxa950_a0() || cpu_is_pxa968_Ax()) {
		if (converter == LCD_Controller_DSI1) {
			/*Clear previous state*/
			ACCR1 &= ~(ACCR1_REG_DSI_BIT2(ACCR1_REG_DSI_BIT_MASK));
			/*Set the wanted clk*/
			ACCR1 |= ACCR1_REG_DSI_BIT2(dsi_clk);
		} else {
			/*Clear previous state*/
			ACCR1 &= ~(ACCR1_REG_DSI_BIT1(ACCR1_REG_DSI_BIT_MASK));
			/*Set the wanted clk*/
			ACCR1 |= ACCR1_REG_DSI_BIT1(dsi_clk);
		}
	} else {
		struct MCCUDSIRegisters  *p_MCCURegs;
		u32 fbdiv, refdiv, outdiv, cksel, dsi_clk_val;

		/*Calc MCCU address according to DSI*/
		p_MCCURegs = (struct MCCUDSIRegisters  *)\
			(MMCUbase+(converter-1)*MCCU_ADDRESS_SPACE);

		/*
		 * This code only allow DSI clock frequencies from the
		 * values defined in LCD_Controller_DSI_Clock enum.
		 * It is possible to remove this limitation and work
		 * directly with continuous numerical values. HOWEVER,
		 * if doing so PHY timing configuration MUST ALSO
		 * BE MODIFIED with the needed calcultaion to comply
		 * with non-fixed values.
		 * */
		/*Make sure dsi_clk is within range*/
		if (dsi_clk > LCD_Controller_DSI_624MHZ) {
			printk(KERN_ERR
			"%s: BAD clock value used\n", __func__);
			return;
		}
		/*Save dsi_clk enum value for later use as DSI PLL
		 registers are RO*/
		dsi_clk_shadow[converter] = dsi_clk;
		/*Translate enum to numerical value for calculations*/
		dsi_clk_val = DSI_CLOCK_VALUE(dsi_clk);

		/*Disable PLL with internal timer*/
		p_MCCURegs->MCCU_DSI_REG1 =
			MCCU_DSI_REG1_DSIx_PLL_LOCK_SEL;

		/*Set default values*/
		refdiv = 0x1;/*Always Set to 00001*/
		if (dsi_clk_val > 200) {
			outdiv = 0xF;
			if (dsi_clk_val < 400)
				cksel = 0x1;
			else
				cksel = 0x0;
		} else {
			outdiv = 16 - (200/dsi_clk_val);
			cksel = 0x0;
		}

		/*Calculate FBDIV according to dsi_clk_val*/
		fbdiv = dsi_Calc_fbdiv(dsi_clk_val, outdiv, cksel);

		/*Assemble REG2 value*/
		p_MCCURegs->MCCU_DSI_REG2 = MCCU_DSI_REG2_DEFAULT
		| MCCU_DSI_REG2_CK_SEL(cksel)
		| MCCU_DSI_REG2_OUTDIV(outdiv)
		| MCCU_DSI_REG2_FBDIV(fbdiv)
		| MCCU_DSI_REG2_REFDIV(refdiv);

		/*Enable PLL with internal timer*/
		p_MCCURegs->MCCU_DSI_REG1 =
				MCCU_DSI_REG1_DSIx_PLL_LOCK_SEL
				| MCCU_DSI_REG1_DSIx_PLL_ON;

		/*Need to at least wait 100us*/
		mdelay(1);
	}
	return;
}
void lcd_enable_dsi_clock(struct pxa950fb_info * fbi, u8 dsi_clk_enable, u8 converter)
{
	if (cpu_is_pxa950_z0() || cpu_is_pxa950_a0() || cpu_is_pxa968_Ax()) {
		if (dsi_clk_enable)
			clk_enable( converter == LCD_Controller_DSI0? fbi->clk_dsi0 : fbi->clk_dsi1);
		else
			clk_disable( converter == LCD_Controller_DSI0? fbi->clk_dsi0 : fbi->clk_dsi1);
	} else {
		struct MCCUDSIRegisters  *p_MCCURegs;

		/*Calc MCCU address according to DSI*/
		p_MCCURegs = (struct MCCUDSIRegisters  *)\
			(MMCUbase+(converter-1)*MCCU_ADDRESS_SPACE);

		if (dsi_clk_enable)
		{
			/*Enable PLL with internal timer*/
			p_MCCURegs->MCCU_DSI_REG1 =
				MCCU_DSI_REG1_DSIx_PLL_LOCK_SEL
				| MCCU_DSI_REG1_DSIx_PLL_ON;
			mdelay(1);
			clk_enable( converter == LCD_Controller_DSI0? fbi->clk_dsi0 : fbi->clk_dsi1);

		}
		else
		{
			clk_disable( converter == LCD_Controller_DSI0? fbi->clk_dsi0 : fbi->clk_dsi1);
			/*Disable PLL with internal timer*/
			p_MCCURegs->MCCU_DSI_REG1 =
				MCCU_DSI_REG1_DSIx_PLL_LOCK_SEL;
			mdelay(1);

		}
	}
	return;
}

u32 lcd_get_dsi_clock(u8 converter)
{
	u32 temp;

	if (cpu_is_pxa950_z0() || cpu_is_pxa950_a0() || cpu_is_pxa968_Ax()) {
		/*Clear previous state*/
		if (converter == LCD_Controller_DSI1) {
			temp = ACCR1&ACCR1_REG_DSI_BIT2(ACCR1_REG_DSI_BIT_MASK);
			temp = temp >> ACCR1_REG_DSI_BIT2_OFFSET;
		} else {
			temp = ACCR1&ACCR1_REG_DSI_BIT1(ACCR1_REG_DSI_BIT_MASK);
			temp = temp >> ACCR1_REG_DSI_BIT1_OFFSET;
		}
	} else {
		/*DSI PLL Registers are Read-Only so we use the shadow copy*/
		temp = dsi_clk_shadow[converter];
	}

	return temp;
}

u8 dsi_enable_disable(struct pxa950fb_info *fbi, int converter, int enable_disable)
{
	u8 ret = 1;
	u32 x;
	u32 *calibration_p;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	switch(enable_disable)
	{
		case DSI_DISABLE:
			x = readl(conv_base + LCD_DSI_DxSCR1_OFFSET) &
						(~LCD_DSI_DxSCR1_DSI_EN);
			writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);
			break;
		case DSI_ENABLE:
			calibration_p = ioremap_nocache(0x42404078, 4);
			if(calibration_p != NULL){
				x = *calibration_p | (5<<8);
				x &= ~(1<<9);
				*calibration_p = x;
				iounmap(calibration_p);
				x = readl(conv_base + LCD_DSI_DxPHY_CAL_OFFSET)
									| 0x1;
				writel(x, conv_base + LCD_DSI_DxPHY_CAL_OFFSET);
			}
			x = readl(conv_base + LCD_DSI_DxSCR1_OFFSET)
							|LCD_DSI_DxSCR1_DSI_EN;
			writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);
			break;
		default:
			ret = 0;
			break;
	}

	return ret;
}

u8 dsi_set_lanes_number(struct pxa950fb_info *fbi, int converter, int nol)
{
	u8 ret = 1;
	u32 x;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	if(nol>LCD_Controller_DSI_4LANE)
	{
		/*only 4 lanes available*/
		ret = 0;
	}
	else
	{
		/*Clear previous state*/
		x = readl(conv_base + LCD_DSI_DxSCR1_OFFSET)
					& (~LCD_DSI_DxSCR1_NOL_MASK);
		writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);
		/*Set number of lanes*/
		x |= LCD_DSI_DxSCR1_NOL(nol);
		writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);

	}

	return ret;
}

u8 dsi_set_power_mode(struct pxa950fb_info *fbi, int converter, int power_mode)
{
	u8 ret = 1;
	u8 rihs_mode,ulps_mode,lpdt_mode,bta_mode;
	u32 x;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

#ifndef CONFIG_CPU_PXA968
	rihs_mode = power_mode&LCD_Controller_DSI_RIHS;
#else
	rihs_mode = 0;
#endif
	ulps_mode = (power_mode&LCD_Controller_DSI_ULPS)>>1;
	lpdt_mode = (power_mode&LCD_Controller_DSI_LPDT)>>2;
	bta_mode = (power_mode&LCD_Controller_DSI_BTA)>>3;

	/*Clear previous state*/
	x = readl(conv_base + LCD_DSI_DxSCR1_OFFSET)
					& (~LCD_DSI_DxSCR1_POWER_MASK);
	writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);

	/*Set power mode*/
	x |= LCD_DSI_DxSCR1_RIHS(rihs_mode)|LCD_DSI_DxSCR1_ULPS(ulps_mode)
		|LCD_DSI_DxSCR1_LPDT(lpdt_mode)|LCD_DSI_DxSCR1_BTA(bta_mode);
	writel(x, conv_base + LCD_DSI_DxSCR1_OFFSET);

	return ret;
}

u32 dsi_perform_bta(struct pxa950fb_info *fbi, u8 converter)
{
	u32 x;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	dsi_set_power_mode(fbi, converter, LCD_Controller_DSI_BTA);
	mdelay(1);
	dsi_set_power_mode(fbi, converter, dsi_power_mode);

	/*Check for error*/
	x = readl(conv_base + LCD_DSI_DxINST1_OFFSET);
	if (x & LCD_DSI_DxINST1_BTA_TO_ERR)
	{
		writel(x | LCD_DSI_DxINST1_BTA_TO_ERR,
				conv_base + LCD_DSI_DxINST1_OFFSET);
		return 0;
	}
	/*Check for trigger*/
	if (x & LCD_DSI_DxINST1_TRIG_ST)
	{
		/*Get T_IN*/
		writel(x | LCD_DSI_DxINST1_TRIG_ST,
					conv_base + LCD_DSI_DxINST1_OFFSET);
		/*Return with Trig bit for parse*/
		x = ((readl(conv_base + LCD_DSI_DxTRIG_OFFSET)>>24)&0xFF)
								|DSI_BTA_TRIG;
		return x;
	}
	/*No trigger - return ACK_PAYLOAD*/
	x = (readl(conv_base + LCD_DSI_DxADAT_OFFSET)&0xFFFF)|DSI_BTA_ACK;
	return x;

}

void dsi_parse_bta_value(u32 bta_value,u32 ber_test)
{
	u32 temp_value;

	/*Check for ber_test*/
	if(ber_test)
	{
		/*Get BTA result identifier*/
		temp_value =  bta_value&0xF0000000;
		if(temp_value == DSI_BTA_TRIG)
		{
			temp_value = bta_value&0xFF;
			if(temp_value != 0x04)/*TBD enum*/
			{
				BER_ARRAY[ber_count] = temp_value;
				ber_count++;
			}
		}
		else if(temp_value == DSI_BTA_ACK)
		{
			temp_value = bta_value&0xFFFF;
			BER_ARRAY[ber_count] = temp_value;
			ber_count++;
		}

	}
	else
	{
		/*Get BTA result identifier*/
		temp_value =  bta_value&0xF0000000;
		if(temp_value == DSI_BTA_TRIG)
		{
			printk("%s: Trigger received\n", __func__);
		}
		else if(temp_value == DSI_BTA_ACK)
		{
			printk("%s: ACK received\n", __func__);
		}
		else
		{
			printk("%s: No Response from the Panel!\n", __func__);
		}
	}

	return;
}

u8 dsi_wait_interrupt(struct pxa950fb_info *fbi, u8 converter,u32 time_to_wait)
{
	u8 ret = 1;
	u32 x;
	int i = 0;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	/*Wait until we received interrupt from the command fifo*/
	x = readl(conv_base + LCD_DSI_DxINST0_OFFSET);
	while ((x & LCD_DSI_DxINST0_SP_GEN_INT_STS) !=
						LCD_DSI_DxINST0_SP_GEN_INT_STS)
	{
		i++;
		msleep(100);
		x = readl(conv_base + LCD_DSI_DxINST0_OFFSET);
		if (i>50){
			ret = 0;
			break;
		}
	}
	if (i > 50)
		printk("%s: failed to receive DSI interrupt!!!\n", __func__);
	mdelay(time_to_wait);
	writel(x|LCD_DSI_DxINST0_SP_GEN_INT_STS,
				conv_base + LCD_DSI_DxINST0_OFFSET);
	return ret;

}

void dsi_enable_mved_dma_channel(u8 converter,u32 desc_address)
{
	u32  *mved_dma;

	mved_dma = (u32 *)MVED_DADR_ADDRESS;	/*TODO*/

	/*Set the descriptor address*/
	*(mved_dma) = desc_address;

	/*Set the desired channel*/
	if (converter == LCD_Controller_DSI1)
	{
		mved_dma = (u32 *)MVED_DRCMR3_ADDRESS;
		*(mved_dma) = 0x81;
	}
	else
	{
		mved_dma = (u32 *)MVED_DRCMR2_ADDRESS;
		*(mved_dma) = 0x81;
	}

	mved_dma = (u32 *)MVED_DCSR_ADDRESS;

	/*Run dma*/
	*(mved_dma) = 0x80000000;

	return;
}

u8 dsi_create_dma_desc(u32 *desc_addr, u32 src_addr, u32 dest_addr, u32 mem_size, u32 len)
{
	u8 ret = 1;
	int i;
	int Number_of_descriptors;

	/*Get number of descriptors*/
	Number_of_descriptors = mem_size/len;

	/*The Descriptors number is limited - return error if exceeded*/
	if(Number_of_descriptors>MAX_DSI_DESC)
	{
		printk("%s: Descriptors number is exceeds limit!!\n", __func__);
		return 0;
	}

	/*Build the descriptors*/
	for(i=0;i<(Number_of_descriptors*4);i=i+4)
	{
		desc_addr[i] =  (u32)(&desc_addr[i+4]);		/*Next descriptor*/
		desc_addr[i+1] =  src_addr+(u32)((i/4)*len);	/*Buffer for LCD command DMA*/
		desc_addr[i+2] =  dest_addr;			/*FIFO address*/
		desc_addr[i+3] =  len|(0x9002C000);		/*Width = 0x11, Size = 16 bytes, FLOWTRG=1, INCSRCADDR=1*/
	}

	/*Build the last descriptor*/
	desc_addr[i] =    (u32)desc_addr;			/*Next descriptor*/
	desc_addr[i+1] =  src_addr+(u32)((i/4)*len);		/*Buffer for LCD command DMA*/
	desc_addr[i+2] =  dest_addr;				/*FIFO address*/
	desc_addr[i+3] =  (mem_size-(i/4)*len)|(0x9002C000);	/*Width = 0x11, Size = 16 bytes, FLOWTRG=1, INCSRCADDR=1*/

	return ret;
}

u8 dsi_send_dma_cmd(struct pxa950fb_info *fbi, u8 converter, u32 size_of_desc)
{
	u8 ret = 1;
	u32 x;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	if(!dsi_create_dma_desc((u32 *)&(LCD_DMAD_DSI[0][0]),
				COMMAND_DMA_BUFFER,
	CONVERTER_BASE_ADDRESS(DISPLAY_CONTROLLER_BASE_ADDRESS, converter)
				+LCD_DSI_DxCFIF_OFFSET,
				size_of_desc,
				0x1F00))
	{
		printk("%s: dsi_create_dma_desc failed!!\n", __func__);
		return 0;
	}



	/*Clear the Flow_dis bit to allow fetching from DMA*/
	x = readl(conv_base + LCD_DSI_DxSCR0_OFFSET)
			& (~LCD_DSI_DxSCR0_FLOW_DIS);
	writel(x, conv_base + LCD_DSI_DxSCR0_OFFSET);

	/*Insert a big delay to FIFO to allow time for the DMA commands to be loaded*/
	writel(0x0AFF0908, conv_base + LCD_DSI_DxCFIF_OFFSET);

	/*enable the DMA control*/
	dsi_enable_mved_dma_channel(converter,(u32)&(LCD_DMAD_DSI[0][0]));

	return ret;
}

u8 dsi_send_cmd(struct pxa950fb_info *fbi, u8 converter)
{
	int i;
	u8 ret = 1;

	/*send commands to the FIFO to enable frame readings*/
	for(i=1;i<DSI_COMMAND_BUFFER[0]+1;i=i+2)
	{
		writel(((u32)DSI_COMMAND_BUFFER[i])|((u32)DSI_COMMAND_BUFFER[i+1]<<16),
			CONVERTER_BASE_ADDRESS(fbi->reg_base, converter) + LCD_DSI_DxCFIF_OFFSET);
	}

	return ret;

}

u8 dsi_send_cmd_ber(struct pxa950fb_info *fbi, u8 converter)
{
	int i;
	u8 ret = 1;
	u32 bta_value;

	/*send commands to the FIFO to enable frame readings*/
	for(i=1;i<DSI_COMMAND_BUFFER[0]+1;i=i+2)
	{
		writel(((u32)DSI_COMMAND_BUFFER[i])|((u32)DSI_COMMAND_BUFFER[i+1]<<16),
			CONVERTER_BASE_ADDRESS(fbi->reg_base, converter) + LCD_DSI_DxCFIF_OFFSET);
		/*Check if we received interrupt*/
		if(dsi_wait_interrupt(fbi, converter,0))
		{
			bta_value =  dsi_perform_bta(fbi, converter);
			if(!bta_value)
			{
				printk("%s: dsi_perform_bta failed!!\n", __func__);
				return 0;
			}
			dsi_parse_bta_value(bta_value,1);
		}

	}
	return ret;
}

int dsi_add_cmd_dma(int dsi_index, u8 dsi_cmd, u8 dsi_data)
{
	u16 *cmd_dma;

	cmd_dma = (u16*)COMMAND_DMA_BUFFER;

	cmd_dma[dsi_index++] = LCD_DSI_COMMAND_DATA(dsi_data)
		|LCD_DSI_COMMAND_CMD(dsi_cmd);

	return dsi_index;
}

int dsi_add_cmd(int dsi_index, u8 dsi_cmd, u8 dsi_data, int dsi_loop)
{
	DSI_COMMAND_BUFFER[dsi_index++] = LCD_DSI_COMMAND_DATA(dsi_data)
		|LCD_DSI_COMMAND_CMD(dsi_cmd)
		|LCD_DSI_COMMAND_LOOP(dsi_loop);

	return dsi_index;
}

int dsi_add_dcs_read_cmd(int dsi_index, u8 parm, int dsi_loop)
{
	int cmd_index=dsi_index;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_DCS_READ_NO_PARAMETER, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, 0, dsi_loop);

	return cmd_index;
}

int dsi_add_dcs_short_cmd(int dsi_index, u8 parm, int dsi_loop)
{
	int cmd_index=dsi_index;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, 0, dsi_loop);

	return cmd_index;
}



int dsi_add_short_cmd(int dsi_index,u8 Command, u8 parm1, u8 parm2, int dsi_loop)
{
	int cmd_index=dsi_index;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,Command, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm1, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm2, dsi_loop);

	return cmd_index;
}

int dsi_add_dcs_short_cmd_arg(int dsi_index, u8 parm, u8 Argument, int dsi_loop)
{
	int cmd_index=dsi_index;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, Argument, dsi_loop);

	return cmd_index;
}

int dsi_add_dcs_long_cmd(int dsi_index, u8 *parm, int dsi_loop)
{
	int i=0, cmd_index=dsi_index;
	u16 param_count;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_DCS_LONG_WRITE, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);

	param_count = ((u16)parm[0])|((u16)(parm[1]<<8));
	for(;i<param_count+2;)
	{
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);
	}

	return cmd_index;
}

int dsi_add_long_cmd(int dsi_index,u8 Command, u8 *parm, int dsi_loop)
{
	int i=0, cmd_index=dsi_index;
	u16 param_count;

	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, Command, dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);

	param_count = ((u16)parm[0])|((u16)(parm[1]<<8));
	for(;i<param_count+2;)
	{
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm[i++], dsi_loop);
	}

	return cmd_index;
}

void dsi_calc_lp_hs_time(struct pxa950fb_info *fbi, u8 converter)
{
	u16 DSI_Clock_Values[] = {52,104,156,204,312,0,624};
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	/*Get DSI bit Clock*/
	DSI_Conf.DSI_BitClock = DSI_Clock_Values[lcd_get_dsi_clock(converter)];
	DSI_Conf.DSI_LPClock = DSI_Clock_Values[0];

	/*Get HS and LP clocks*/
	DSI_Conf.DSI_HSTRAIL =
		(u8)(readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET) & 0xff);
	DSI_Conf.DSI_HSZERO =
		(u8)((readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)>>8) & 0xff);
	DSI_Conf.DSI_HSPREP =
		(u8)((readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)>>16) & 0xff);
	DSI_Conf.DSI_LPX =
	(u8)((readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)>>24) & 0xff);

	DSI_Conf.DSI_CLTRAIL =
		(u8)(readl(conv_base + LCD_DSI_DxPHY_TIM1_OFFSET) & 0xff);
	DSI_Conf.DSI_CLZERO =
		(u8)((readl(conv_base + LCD_DSI_DxPHY_TIM1_OFFSET)>>8) & 0xff);
	DSI_Conf.DSI_HSEXIT =
		(u8)((readl(conv_base + LCD_DSI_DxPHY_TIM1_OFFSET)>>24) & 0xff);

	/*Calculate HSExitTime*/
	DSI_Conf.DSI_HSExitTime_LPClocks = DSI_Conf.DSI_HSTRAIL+DSI_Conf.DSI_HSEXIT+2;
	DSI_Conf.DSI_HSExitTime_DSIClocks = (DSI_Conf.DSI_BitClock/DSI_Clock_Values[0])
		*DSI_Conf.DSI_HSExitTime_LPClocks;
	DSI_Conf.DSI_HSExitTime_nsec = ((1000)*(DSI_Conf.DSI_HSExitTime_DSIClocks))/(DSI_Conf.DSI_BitClock);

	/*Calculate Single NOP delay*/
	DSI_Conf.DSI_SingleNOP_LPClocks = 1;
	DSI_Conf.DSI_SingleNOP_DSIClocks = DSI_Conf.DSI_BitClock/DSI_Clock_Values[0];
	DSI_Conf.DSI_SingleNOP_nsec = 1000/DSI_Clock_Values[0];

	/*Calculate HSEntryTime*/
	DSI_Conf.DSI_HSEntryTime_LPClocks = DSI_Conf.DSI_HSPREP+DSI_Conf.DSI_HSZERO+DSI_Conf.DSI_LPX;
	DSI_Conf.DSI_HSEntryTime_DSIClocks = (DSI_Conf.DSI_BitClock/DSI_Clock_Values[0])
		*DSI_Conf.DSI_HSEntryTime_LPClocks+28;
	DSI_Conf.DSI_HSEntryTime_nsec = (1000*DSI_Conf.DSI_HSEntryTime_DSIClocks)/DSI_Conf.DSI_BitClock;

	return;
}

LCD_Controller_Configurations *get_lcd_conf(void)
{
	return (LCD_Controller_Configurations *)&LCD_Conf;
}

void dsi_calc_frame_time(void)
{
	LCD_Controller_Configurations *pLCD_Conf;

	/*Get LCD_Conf*/
	pLCD_Conf = get_lcd_conf();

	/*HSYNC freq in khz*/
	FRAME_Conf.FRAME_HSYNC_freq = (pLCD_Conf->LCD_FRAME_Rate)
		*(pLCD_Conf->LCD_BFW+pLCD_Conf->LCD_Vertical+pLCD_Conf->LCD_EFW)/1000;

	/*HSYNC in nsec*/
	FRAME_Conf.FRAME_HSYNC_nsec = 1000000/FRAME_Conf.FRAME_HSYNC_freq;

	/*Get General information*/
	FRAME_Conf.FRAME_nol = dsi_lanes_number + 1;
	switch(dsi_rgb_mode)
	{
		case LCD_Controller_RGB_888_PACKET :
			FRAME_Conf.FRAME_bpp = 24;
			FRAME_Conf.FRAME_bytes = 3;
			break;
		case LCD_Controller_RGB_565_PACKET :
			FRAME_Conf.FRAME_bpp = 16;
			FRAME_Conf.FRAME_bytes = 2;
			break;
		case LCD_Controller_RGB_666_PACKET :
			FRAME_Conf.FRAME_bpp = 18;
			FRAME_Conf.FRAME_bytes = 2.25;
			break;
		case LCD_Controller_RGB_666_LOOSELY_PACKET :
			FRAME_Conf.FRAME_bpp = 18;
			FRAME_Conf.FRAME_bytes = 3;
			break;
		default:
			FRAME_Conf.FRAME_bpp = 24;
			FRAME_Conf.FRAME_bytes = 3;
			break;
	}

	/*Bits per frame*/
	FRAME_Conf.FRAME_bpf = (u32)((pLCD_Conf->LCD_BLW+pLCD_Conf->LCD_Horizontal+pLCD_Conf->LCD_ELW)
		*(pLCD_Conf->LCD_BFW+pLCD_Conf->LCD_Vertical+pLCD_Conf->LCD_EFW)*FRAME_Conf.FRAME_bpp);

	/*Bits per second*/
	FRAME_Conf.FRAME_bps = (u32)(pLCD_Conf->LCD_FRAME_Rate)*FRAME_Conf.FRAME_bpf;

	/*Pixels per second*/
	FRAME_Conf.FRAME_pps = FRAME_Conf.FRAME_bps/FRAME_Conf.FRAME_bpp;

	/*High Speed bit time*/
	FRAME_Conf.FRAME_HS_bit_time = 1000/(FRAME_Conf.FRAME_nol*DSI_Conf.DSI_BitClock);

	/*Pixel time*/
	FRAME_Conf.FRAME_pixel_time = FRAME_Conf.FRAME_bpp * FRAME_Conf.FRAME_HS_bit_time;

	return;
}

u8 dsi_find_dly(u32 parm1,u32 parm2)
{
	int i,power = 1;
	int temp_dev;

	if(parm1<parm2)
	{
		temp_dev = parm2/256;
	}
	else
	{
		temp_dev = parm1/256;
	}

	for(i=0;i<16;i++)
	{
		if(power >= temp_dev)
			break;
		power *= 2;
	}

	return i;
}

u8 dsi_find_mult(u32 parm)
{
	int i,power = 1;

	for(i=0;i<parm;i++)
	{
		power *= 2;
	}

	return power;
}

void dsi_send_frame_cmd(void)
{
	int cmd_index=1;
	u8 parm1,parm2;
	u32 parm_temp;
	LCD_Controller_Configurations *pLCD_Conf;

	/*Get LCD_Conf*/
	pLCD_Conf = get_lcd_conf();

	/*Set the commands to be sent to the DSI*/
	/*Two commands can be send each clock cycle*/

	/*Set delay for Nop - for line*/
	parm1 = DSI_DCS_DLY;/*Time base set*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_SET_DLY_MULT,parm1,0);


	/*##############  IMPORTANT NOTE   ###############*/
	/*The right way to send the pixels command is with a DCS LONG command where WRITE_MEMORY_START or*/
	/*WRITE_MEMORY_CONTINUE is the first byte followed by a line of pixels. As for now, The DSI controller*/
	/*does not support this method. There for a Workaround is introduced - Send the WRITE_MEMORY_START or*/
	/*WRITE_MEMORY_CONTINUE as DCS short command and then a packet of pixels. This method works on specific*/
	/*panels and is not recommended as the default method.*/
	/*################################################*/

	/*WRITE_MEMORY_START with line from the LCD controller*/
	cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_START,0);
	/*Send another WRITE_MEMORY_START for 2 lanes*/
	if(dsi_lanes_number == LCD_Controller_DSI_2LANE)
	{
		cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_START,0);
	}
#if 0	/* Fix the bottom white line issue.*/
	/*RGB packet with loop*/
	parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
	parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);

	/*Line data write with loop - send a line from memory*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1,0);
#endif
	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*WRITE_MEMORY_CONTINUE with line from the LCD controller*/
	cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_CONTINUE,1);

	/*RGB packet with loop*/
	parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
	parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,1);

	/*Line data write with loop - send a line from memory*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1,1);

	/*Execute loop according to Vertical resolution - 1 (above)*/
	parm_temp = pLCD_Conf->LCD_Vertical-1;
	while(parm_temp>0x100)
	{
		parm1 = 0xFF;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);
		parm_temp -= 0x100;
	}
	parm1 = (u8)(parm_temp-1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);

	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*Send end of transmition*/
	parm1 = 0x0f;
	parm2 = 0x0f;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_END_OF_TRANSMITION,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);


	parm1 = 10;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
	}

	/*Keep how many commands we have for later*/
	DSI_COMMAND_BUFFER[0] = cmd_index - 1;

	return;
}

void dsi_send_frame_ber(void)
{
	int cmd_index=1;
	u8 parm1,parm2;
	u32 parm_temp;
	LCD_Controller_Configurations *pLCD_Conf;

	/*Get LCD_Conf*/
	pLCD_Conf = get_lcd_conf();

	/*Set the commands to be sent to the DSI*/
	/*Two commands can be send each clock cycle*/

	/*Set delay for Nop - for line*/
	parm1 = DSI_DCS_DLY;/*Time base set*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_SET_DLY_MULT,parm1,0);

	/*Interrupt the processor*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_INTERRUPT_THE_PROCESSOR,parm1,1);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,1);
	}


	/*WRITE_MEMORY_START with line from the LCD controller*/
	cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_START,0);
	/*Send another WRITE_MEMORY_START for 2 lanes*/
	if(dsi_lanes_number == LCD_Controller_DSI_2LANE)
	{
		cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_START,0);
	}

	/*RGB packet with loop*/
	parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
	parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);

	/*Line data write with loop - send a line from memory*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1,0);

	/*Interrupt the processor*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_INTERRUPT_THE_PROCESSOR,parm1,0);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
	}

	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*WRITE_MEMORY_CONTINUE with line from the LCD controller*/
	cmd_index = dsi_add_dcs_short_cmd(cmd_index,LCD_Controller_DCS_WRITE_MEMORY_CONTINUE,1);

	/*RGB packet with loop*/
	parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
	parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,1);

	/*Line data write with loop - send a line from memory*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1,1);

	/*Interrupt the processor*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_INTERRUPT_THE_PROCESSOR,parm1,1);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,1);
	}


	/*Execute loop according to Vertical resolution - 2 (above)*/
	parm_temp = pLCD_Conf->LCD_Vertical-2;
	while(parm_temp>0x100)
	{
		parm1 = 0xFF;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);
		parm_temp -= 0x100;
	}
	parm1 = (u8)(parm_temp-1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);

	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*Send end of transmition*/
	parm1 = 0x0f;
	parm2 = 0x0f;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_END_OF_TRANSMITION,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);


	parm1 = 10;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
	}

	/*Keep how many commands we have for later*/
	DSI_COMMAND_BUFFER[0] = cmd_index - 1;

	return;
}

/*The values for the timing in this function are board specific*/
u32 dsi_send_video_nonburst(void)
{
	int cmd_index=0,i,j;
	u8 parm1,parm2;
	u8 mult=3,nop1=90,nop2=10,nop3=1,Blank=40;
	u32 parm_temp;
	LCD_Controller_Configurations *pLCD_Conf;

	/*Get LCD_Conf*/
	pLCD_Conf = get_lcd_conf();


	/*Set the commands to be sent to the DSI*/
	/*Two commands can be send each clock cycle*/

	/* delay mlt*/
	parm1 = 8;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

	/*Put a big Nop at the start to fill the FIFO*/
	parm1 = 30;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);

	/* delay mlt*/
	parm1 = 2;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

	/*Assert VSYNC*/

	/*VSYNC start == HSYNC start*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_VSYNC_START);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	/*HSW delay*/
	parm1 = Blank*FRAME_Conf.FRAME_bytes;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_BLANKING_PACKET);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
	{
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
	}

	/*HSYNC end*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,LCD_Controller_HSYNC_END);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	/*HBP delay*/
	/* delay mlt*/
	parm1 = mult;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

	parm1 = nop1;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);

	/*loop according to VSW*/
	parm_temp = (u32)pLCD_Conf->LCD_VSW-2;
	for(i=0;i<parm_temp;i++)
	{
		/*HSYNC start*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_START);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HSW delay*/
		parm1 = Blank*FRAME_Conf.FRAME_bytes;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_BLANKING_PACKET);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
		{
			cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
		}

		/*HSYNC end*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_END);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HBP delay*/
		/* delay mlt*/
		parm1 = mult;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

		parm1 = nop1;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);
	}

	/*VSYNC END*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,LCD_Controller_VSYNC_END);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	/*HSW delay*/
	parm1 = Blank*FRAME_Conf.FRAME_bytes;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_BLANKING_PACKET);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
	{
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
	}

	/*HSYNC end*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,LCD_Controller_HSYNC_END);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	/*HBP delay*/
	/* delay mlt*/
	parm1 = mult;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

	parm1 = nop1;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);

	/*loop according to VBP*/
	parm_temp = (u32)(pLCD_Conf->LCD_BFW-pLCD_Conf->LCD_VSW);
	for(i=0;i<parm_temp;i++)
	{
		/*HSYNC start*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_START);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HSW delay*/
		parm1 = Blank*FRAME_Conf.FRAME_bytes;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_BLANKING_PACKET);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
		{
			cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
		}

		/*HSYNC end*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_END);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HBP delay*/
		/* delay mlt*/
		parm1 = mult;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

		parm1 = nop1;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);
	}

	/*loop according to Vertical resolution*/
	parm_temp = (u32)pLCD_Conf->LCD_Vertical;
	for(i=0;i<parm_temp;i++)
	{
		/*HSYNC start*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_START);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HSW delay*/
		parm1 = Blank*FRAME_Conf.FRAME_bytes;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_BLANKING_PACKET);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
		{
			cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
		}

		/*HSYNC end*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_END);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HBP delay*/
		/* delay mlt*/
		parm1 = mult;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

		parm1 = nop2;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);

		/*Send Line*/
		parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
		parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*Line data write - send a line from memory*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1);

		/*HFP delay*/
		/* delay mlt*/
		parm1 = mult;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

		parm1 = nop3;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);
	}


	/*loop according to VFP*/
	parm_temp = (u32)pLCD_Conf->LCD_EFW;
	for(i=0;i<parm_temp;i++)
	{
		/*HSYNC start*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_START);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HSW delay*/
		parm1 = Blank*FRAME_Conf.FRAME_bytes;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_BLANKING_PACKET);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		for(j=0;j<(Blank*FRAME_Conf.FRAME_bytes);j++)
		{
			cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,0x00);
		}

		/*HSYNC end*/
		parm1 = 0;
		parm2 = 0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
			LCD_Controller_HSYNC_END);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

		/*HBP delay*/
		/* delay mlt*/
		parm1 = mult;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_SET_DLY_MULT,parm1);

		parm1 = nop1;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);
	}

	/*Send end of transmition*/
	parm1 = 0x0f;
	parm2 = 0x0f;
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_END_OF_TRANSMITION);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm1);
	cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_COMMAND_WRITE,parm2);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd_dma(cmd_index,LCD_Controller_NO_OPERATION,parm1);
	}

	return (cmd_index*2);
}

void dsi_send_video_burst(void)
{
	int cmd_index=1;
	u8 parm1,parm2;
	u32 parm_temp;
	LCD_Controller_Configurations *pLCD_Conf;

	/*Get LCD_Conf*/
	pLCD_Conf = get_lcd_conf();

	/*Set the commands to be sent to the DSI*/
	/*Two commands can be send each clock cycle*/

	/*Begin Subroutine*/
	parm1 = 0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_START_LABEL,parm1,0);

	/*Set delay for Nop - for line*/
	parm1 = 20;/*Time base set*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_SET_DLY_MULT,parm1,0);

	/*Assert VSYNC*/
	/*VSYNC start*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_VSYNC_START,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);

	/*VSYNC end*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_VSYNC_END,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,0);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,0);

	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*HSYNC start*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_HSYNC_START,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,1);

	/*HSYNC end*/
	parm1 = 0;
	parm2 = 0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,
		LCD_Controller_HSYNC_END,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,1);

	/*RGB packet with loop*/
	parm1 = (u8)(pLCD_Conf->LCD_Horizontal & 0xff);			/*LSB of Horizontal resolution*/
	parm2 = (u8)((pLCD_Conf->LCD_Horizontal & 0xff00) >> 0x8);	/*MSB of Horizontal resolution*/
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,dsi_rgb_mode,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm1,1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE,parm2,1);

	/*Line data write with loop - send a line from memory*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_LINE_DATA_WRITE,parm1,1);

	/*Execute loop according to Vertical resolution - 1 (above)*/
	parm_temp = pLCD_Conf->LCD_Vertical-1;
	while(parm_temp>0x100)
	{
		parm1 = 0xFF;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);
		parm_temp -= 0x100;
	}
	parm1 = (u8)(parm_temp-1);
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_EXECUTE_LOOP_BUFFER,parm1,0);

	/*Invalidate all loop buffer contents*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_FLUSH_LOOP_BUFFER,parm1,0);

	/*END - go to start for a new frame*/
	parm1 = 0x0;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_GOTO_START,parm1,0);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
	}

	/*Keep how many commands we have for later*/
	DSI_COMMAND_BUFFER[0] = cmd_index - 1;

	return;
}

u32 dsi_send_frame(struct pxa950fb_info *fbi, u32 dsi_converter)
{
	u32 dma_size;

	/*Set number of lanes and power mode*/
	dsi_set_lanes_number(fbi, dsi_converter, dsi_lanes_number);

	dsi_set_power_mode(fbi, dsi_converter, dsi_power_mode);

	/*Send the active Video from the processor*/
	/*Check for video mode*/
	if(dsi_video_mode)
	{
		/*Video mode should run only one time*/
		if(!dsi_video_run_mode)
		{
			if(dsi_video_mode == DSI_VIDEO_MODE_NON_BURST)
			{

				/*Quick disable the LCD controller*/
				controller_enable_disable(fbi, LCD_Controller_Quick_Disable);

				dsi_calc_lp_hs_time(fbi, dsi_converter);

				dsi_calc_frame_time();

				dma_size = dsi_send_video_nonburst();

				/*Set the DMA descriptor and run*/
				if(!dsi_send_dma_cmd(fbi, dsi_converter,dma_size))
				{
					printk("%s: dsi_send_dma_cmd failed!!\n", __func__);
					return 0;
				}
				/*Enable the LCD controller*/
				controller_enable_disable(fbi, LCD_Controller_Enable);
			}
			else
			{
				dsi_send_video_burst();

				/*Send the commands*/
				if(!dsi_send_cmd(fbi, dsi_converter))
				{
					printk("%s: dsi_send_cmd failed!!\n", __func__);
					return 0;
				}
			}

			dsi_video_run_mode = 1;
		}
		else/*Video mode already started - skip sending frame*/
		{
			return 1;
		}
	}
	else
	{
		dsi_send_frame_cmd();

		/*Send the commands*/
		if(!dsi_send_cmd(fbi, dsi_converter))
		{
			printk("%s: dsi_send_cmd failed!!\n", __func__);
			return 0;
		}
	}
	return 1;
}

u32 dsi_send_cmd_wait(struct pxa950fb_info *fbi, u8 converter,u32 time_to_wait)
{

	/*Set number of lanes and power mode for Low Power*/
	dsi_set_lanes_number(fbi, converter, dsi_lanes_number);

	dsi_set_power_mode(fbi, converter, dsi_power_mode);

	/*Send the commands*/
	if(!dsi_send_cmd(fbi, converter))
	{
		printk("%s: dsi_send_cmd failed!!\n", __func__);
		return 0;
	}

	/*wait until the processor finish the DSI fifo sequence*/
	if(!dsi_wait_interrupt(fbi, converter,time_to_wait))
	{
		printk("%s: dsi_wait_interrupt failed!!\n", __func__);
		return 0;
	}

	return 1;
}

u8 *dsi_get_board_parm(u32 board,int index)
{
	switch(board)
	{
		case DSI_BOARD_ZERO:
			return Init_Board[index];
		case DSI_BOARD_ONE:
			/*Here you can add new board init array*/
			break;
		default:
			break;
	}

	return 0;
}

u32 dsi_init_board(struct pxa950fb_info *fbi, u32 board,u32 dsi_converter)
{
	u8 ret = 1;
	int cmd_index=1,j=1,k=0,i=0,l=0;
	u8 parm1,parm[BOARD_INIT_MAX_DATA_BYTES];
	u8 Temp_command;
	u8 *pInitBoard;
	u16 param_count;

	/*Get board parameters*/
	pInitBoard = dsi_get_board_parm(board,i);

	Temp_command = pInitBoard[j++];
	while(Temp_command != 0xFF)
	{
		switch(Temp_command)
		{
			case LCD_Controller_DCS_SHORT_WRITE_NO_PARAMETER:
				/*Built command*/
				cmd_index = dsi_add_dcs_short_cmd(cmd_index,pInitBoard[j],0);
				j=1;
				break;
			case LCD_Controller_DCS_SHORT_WRITE_WITH_PARAMETER:
				/*Built command*/
				cmd_index = dsi_add_dcs_short_cmd_arg(cmd_index,pInitBoard[j],
					pInitBoard[j+1],0);
				j=1;
				break;
			case LCD_Controller_GENERIC_SHORT_WRITE_TWO_PARAMETERS:
			case LCD_Controller_SET_MAXIMUM_RETURN_PACKET_SIZE:
			case LCD_Controller_TURN_ON_PERIPHERAL:
				/*Built command*/
				cmd_index = dsi_add_short_cmd(cmd_index,Temp_command,pInitBoard[j],
					pInitBoard[j+1],0);
				j=1;
				break;
			case LCD_Controller_DCS_LONG_WRITE:
				/*Built command*/
				param_count = ((u16)pInitBoard[j])|((u16)(pInitBoard[j+1]<<8));
				for(l=0;l<param_count+2;l++)
				{
					parm[k] = pInitBoard[j];
					k++;
					j++;
				}
				cmd_index = dsi_add_dcs_long_cmd(cmd_index,parm,0);
				k=0;
				j=1;
				break;
			case LCD_Controller_GENERIC_LONG_WRITE:
			case LCD_Controller_LONG_PACKET_TBD:
				/*Built command*/
				param_count = ((u16)pInitBoard[j])|((u16)(pInitBoard[j+1]<<8));
				for(l=0;l<param_count+2;l++)
				{
					parm[k] = pInitBoard[j];
					k++;
					j++;
				}
				cmd_index = dsi_add_long_cmd(cmd_index,Temp_command,parm,0);
				k=0;
				j=1;
				break;

			default:
				break;
		}

		/*Get timeout - if not 0 - add interrupt and exit. if 0xFF - dont add a nop*/
		if(!pInitBoard[0])
		{
			parm1 = 10;
			cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
		}
		else if (pInitBoard[0]!=0xFF)
		{
			/*Interrupt the proccesor for finishing*/
			parm1 = 0x00;
			cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_INTERRUPT_THE_PROCESSOR,parm1,0);

			/*We must pad to 2 commands for each cycle with a NOP*/
			if((cmd_index-1)%2==1)
			{
				/*Nop*/
				parm1 = 0x0;
				cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
			}

			/*Keep how many commands we have for later*/
			DSI_COMMAND_BUFFER[0] = cmd_index - 1;

			if(!dsi_send_cmd_wait(fbi, dsi_converter,pInitBoard[0]))
			{
				printk("%s: Send command failed!!\n", __func__);
				return 0;
			}

			cmd_index = 1;
		}

		/*Get the next command*/
		i++;
		pInitBoard = dsi_get_board_parm(board,i);
		Temp_command = pInitBoard[j++];
	}

	return ret;
}

void lcd_set_dsi_time(struct pxa950fb_info *fbi, u8 converter, u8 freq)
{
	u32 x;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	/*Set timing configuration according to PXA950 docs*/

	/*Clear previous states*/
	writel(0, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
	writel(0, conv_base + LCD_DSI_DxPHY_TIM1_OFFSET);
	writel(0, conv_base + LCD_DSI_DxPHY_TIM2_OFFSET);

	/*general settings*/
	writel(LCD_DSI_DxPHY_TIM0_LPX(3),
				conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
	writel(LCD_DSI_DxPHY_TIM1_HSEXIT(7)|LCD_DSI_DxPHY_TIM1_CLTRAIL(5)
		|LCD_DSI_DxPHY_TIM1_CLZERO(17)|LCD_DSI_DxPHY_TIM1_TAGO(12),
		conv_base + LCD_DSI_DxPHY_TIM1_OFFSET);
	writel(LCD_DSI_DxPHY_TIM2_TAGET(15)|LCD_DSI_DxPHY_TIM2_WAKEUP(0xFFFF),
		conv_base + LCD_DSI_DxPHY_TIM2_OFFSET);

	/*specific setting*/
	switch(freq)
	{
		case LCD_Controller_DSI_52MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(8)
				|LCD_DSI_DxPHY_TIM0_HSZERO(12)
				|LCD_DSI_DxPHY_TIM0_HSPREP(7);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			break;
		case LCD_Controller_DSI_104MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(5)
				|LCD_DSI_DxPHY_TIM0_HSZERO(9)
				|LCD_DSI_DxPHY_TIM0_HSPREP(5);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			x = readl(conv_base + LCD_DSI_DxPHY_TIM1_OFFSET)
				|LCD_DSI_DxPHY_TIM1_CLTRAIL(10)
				|LCD_DSI_DxPHY_TIM1_CLZERO(3);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM1_OFFSET);
			break;
		case LCD_Controller_DSI_156MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(10)
				|LCD_DSI_DxPHY_TIM0_HSZERO(8)
				|LCD_DSI_DxPHY_TIM0_HSPREP(4);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			break;
		case LCD_Controller_DSI_208MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(12)
				|LCD_DSI_DxPHY_TIM0_HSZERO(7)
				|LCD_DSI_DxPHY_TIM0_HSPREP(3);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			break;
		case LCD_Controller_DSI_312MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(12)
				|LCD_DSI_DxPHY_TIM0_HSZERO(6)
				|LCD_DSI_DxPHY_TIM0_HSPREP(3);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			break;
		case LCD_Controller_DSI_UNDEFINEDMHZ:
			break;
		case LCD_Controller_DSI_624MHZ:
			x = readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET)
				|LCD_DSI_DxPHY_TIM0_HSTRAIL(12)
				|LCD_DSI_DxPHY_TIM0_HSZERO(6)
				|LCD_DSI_DxPHY_TIM0_HSPREP(4);
			writel(x, conv_base + LCD_DSI_DxPHY_TIM0_OFFSET);
			break;
		default:
			break;
	}
	return;
}

static void dsi_command(struct pxa950fb_info *fbi, int dsi_converter,
			int *value1, int *value2, int *value3, int *value4)
{
	int i;
	int cmd_index=1;
	u32  user_value, user_error, bta_value;
	u32  *P_value, *P_error;
	u16  param_count;
	u8   parm1,parm2,parm[10];

	P_value = &user_value;
	P_error = &user_error;
	*P_value = *value1;

	/*Handle command*/
	switch(*P_value)
	{
		case DSI_COMMANDS_FREQ:
			printk("%s: Change DSI frequency selected.\n", __func__);
			printk("%s: 0 - 52Mhz, 1 - 104Mhz, 2 - 156Mhz, 3 - 208Mhz, 4 - 312Mhz\n", __func__);

			/*Set new frequency and configure timings*/
			lcd_set_dsi_clock(*value2, dsi_converter);

			lcd_set_dsi_time(fbi, dsi_converter, *P_value);

			break;
		case DSI_COMMANDS_POWER:
			printk("%s: Change power mode selected.\n", __func__);
			printk("%s: 1 - High Speed, 2 - Ultra Low, 4 - Low Power, 8 BTA or combined.\n", __func__);

			/*Set power mode*/
			if(*P_value == LCD_Controller_DSI_BTA)
			{
				bta_value =  dsi_perform_bta(fbi,
								dsi_converter);
				if(!bta_value)
				{
					printk("%s: dsi_perform_bta failed!!\n", __func__);
					return;
				}
				dsi_parse_bta_value(bta_value,0);
			}
			else
			{
				dsi_power_mode = *value2;
			}
			dsi_set_power_mode(fbi, dsi_converter, dsi_power_mode);

			break;
		case DSI_COMMANDS_NOL:
			printk("%s: Change number of lanes selected.\n", __func__);
			printk("%s: 0 - one lane, 1 - two lanes\n", __func__);

			/*Set number of lanes*/
			dsi_lanes_number = *value2;
			dsi_set_lanes_number(fbi, dsi_converter,
							 dsi_lanes_number);

			break;

		case DSI_COMMANDS_FIFO_COMMAND:
			printk("%s: Send FIFO command selected.\n", __func__);

			/*Keep parameter*/
			parm1 = (u8)((*value2) & 0xff);

			parm2 = (u8)((*value3) & 0xff);

			cmd_index = dsi_add_cmd(cmd_index,parm1, parm2, 0);

			break;

		case DSI_COMMANDS_DSI_COMMAND:
			printk("%s: Send DSI command selected.\n", __func__);

			parm1 = (u8)((*value2) & 0xff);
			cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm1, 0);

			parm1 = (u8)((*value3) & 0xff);
			cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm1, 0);

			parm1 = (u8)((*value4) & 0xff);
			cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_COMMAND_WRITE, parm1, 0);

			break;

		case DSI_COMMANDS_DCS_SHORT_WRITE_NO_PARAMETER:
			printk("%s: Send DCS short write command with no parameter selected.\n", __func__);

			/*Send the command*/
			parm1 = (u8)((*value2) & 0xff);
			cmd_index = dsi_add_dcs_short_cmd(cmd_index,parm1,0);

			break;
		case DSI_COMMANDS_DCS_SHORT_WRITE_WITH_PARAMETER:
			printk("%s: Send DCS short write command with parameter selected.\n", __func__);

			/*Keep parameter*/
			parm1 = (u8)((*value2) & 0xff);

			/*Send the command*/
			parm2 = (u8)((*value3) & 0xff);

			cmd_index = dsi_add_dcs_short_cmd_arg(cmd_index,parm1,parm2,0);

			break;

		case DSI_COMMANDS_DCS_LONG_WRITE:
			printk("%s: Send DCS long write command selected.\n", __func__);

			parm[0] = (u8)((*value2) & 0xff); /*LSB*/
			parm[1] = (u8)(((*value3) & 0xff00) >> 0x8); /*MSB*/
			param_count = ((u16)parm[0])|((u16)(parm[1]<<8));
			for(i=2;i<param_count+2;i++)
			{
				parm[i] = (u8)((dsi_dcs_long_write[i]) & 0xff);
			}

			/*Send the command*/
			cmd_index = dsi_add_dcs_long_cmd(cmd_index,parm,0);

			break;
		case DSI_COMMANDS_INIT_BOARD:
			printk("%s: Init_board command selected.\n", __func__);

			/*Handle board*/
			*P_value = 0;
			switch(*P_value)
			{
				case DSI_BOARD_ZERO:		/* only support this currently */
					if (!dsi_init_board(fbi, *P_value,
								dsi_converter))
					{
						printk("%s: dsi_init_board failed!!\n", __func__);
						return;
					}
					/*Set to high speed and 2 lanes and RGB565 packet*/
					dsi_lanes_number = LCD_Controller_DSI_2LANE;
					dsi_power_mode = LCD_Controller_DSI_RIHS;
					dsi_rgb_mode = LCD_Controller_RGB_565_PACKET;
					return;
				default:
					break;
			}
			break;
		case DSI_COMMANDS_SEND_FRAME:
			printk("%s: Send frame selected\n", __func__);
			printk("%s: 0 - Command mode, 1 - Video mode burst, 2 - Video mode non burst\n", __func__);

			/*Set the mode of the video*/
			dsi_video_mode = *value2;

			/*Start the frame only for Video mode.*/
			/*In command mode - the frame has been sent in DUMB_LCD_COLOR function*/
			if(dsi_video_mode != DSI_COMMAND_MODE)
			{
				mdelay(10);
				dsi_send_frame(fbi, dsi_converter);
			}

			return;
		case DSI_COMMANDS_BER_TEST:
			printk("%s: BER test selected. Frame will be sent 100 times.\n", __func__);

			/*Set the mode of the video*/
			dsi_video_mode = DSI_COMMAND_MODE;

			/*Prepare frame with interrupts after each line*/
			dsi_send_frame_ber();

			for(i=0;i<100;i++)
			{
				/*Send the commands with interrupts*/
				if (!dsi_send_cmd_ber(fbi, dsi_converter))
				{
					printk("%s: dsi_send_cmd_ber failed!!\n", __func__);
					return;
				}
			}

			return;

		default:
			printk("%s: Undefined input!!\n", __func__);
			return;
	}

	/*Interrupt the proccesor for finishing*/
	parm1 = 0x00;
	cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_INTERRUPT_THE_PROCESSOR,parm1,0);

	/*We must pad to 2 commands for each cycle with a NOP*/
	if((cmd_index-1)%2==1)
	{
		/*Nop*/
		parm1 = 0x0;
		cmd_index = dsi_add_cmd(cmd_index,LCD_Controller_NO_OPERATION,parm1,0);
	}

	/*Keep how many commands we have for later*/
	DSI_COMMAND_BUFFER[0] = cmd_index - 1;

	if (!dsi_send_cmd_wait(fbi, dsi_converter, 1))
		printk("%s: Send command failed!!\n", __func__);
	else
		printk("%s: DSI Command has been issued!!\n", __func__);

	return;
}

u32 dsi_init(struct pxa950fb_info *fbi, u32 dsi_converter)
{
	u8 ret = 1;
	int x;

	/*Set default values - Low Power and 2 lanes*/
	dsi_lanes_number = LCD_Controller_DSI_2LANE;
	dsi_power_mode = LCD_Controller_DSI_LPDT;

	/*Set number of lanes and power mode*/
	dsi_set_lanes_number(fbi, dsi_converter, dsi_lanes_number);

	dsi_set_power_mode(fbi, dsi_converter, dsi_power_mode);

	/*Enable the DSI*/
	if(!dsi_enable_disable(fbi, dsi_converter, DSI_ENABLE))
	{
		printk("%s: dsi_enable_disable failed!!\n", __func__);
		return 0;
	}

	/*Enable the LCD controller*/
	controller_enable_disable(fbi, LCD_Controller_Enable);
	mdelay(100);

	/*End if inialization - go to select commands*/
	x = DSI_COMMANDS_INIT_BOARD;
	dsi_command(fbi, dsi_converter, &x, NULL, NULL, NULL);
	return ret;

}

static void dump_regs_base(struct pxa950fb_info *fbi)
{
        printk("LCD_CTL = %x\n", readl(fbi->reg_base + LCD_CTL));
        printk("LCD_CTL_STS = %x\n", readl(fbi->reg_base + LCD_CTL_STS));
        printk("LCD_NXT_DESC_ADDR0 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR0));
        printk("LCD_FETCH_CTL0 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL0));
        printk("LCD_WIN0_CTL = %x\n", readl(fbi->reg_base + LCD_WIN0_CTL));
        printk("LCD_CH0_ALPHA = %x\n", readl(fbi->reg_base + LCD_CH0_ALPHA));
        printk("LCD_MIXER0_CTL0 = %x\n", readl(fbi->reg_base + LCD_MIXER0_CTL0));
        printk("LCD_MIXER0_STS = %x\n", readl(fbi->reg_base + LCD_MIXER0_STS));
        printk("LCD_MIXER0_CTL1 = %x\n", readl(fbi->reg_base + LCD_MIXER0_CTL1));
        printk("LCD_MIXER0_CTL2 = %x\n", readl(fbi->reg_base + LCD_MIXER0_CTL2));
        printk("LCD_MIXER0_OL1_CFG0 = %x\n", readl(fbi->reg_base + LCD_MIXER0_OL1_CFG0));
        printk("LCD_MIXER0_OL1_CFG1 = %x\n", readl(fbi->reg_base + LCD_MIXER0_OL1_CFG1));
	printk("DSI DxSCR0 = 0x%x\n", readl(fbi->reg_base + 0x2000));
	printk("DSI DxSCR1 = 0x%x\n", readl(fbi->reg_base + 0x2100));
	printk("DSI DxSSR  = 0x%x\n", readl(fbi->reg_base + 0x2104));

        printk("LCD_MIXER1_CTL0 = %x\n", readl(fbi->reg_base + LCD_MIXER1_CTL0));
        printk("LCD_MIXER1_STS = %x\n", readl(fbi->reg_base + LCD_MIXER1_STS));
        printk("LCD_MIXER1_CTL1 = %x\n", readl(fbi->reg_base + LCD_MIXER1_CTL1));
        printk("LCD_MIXER1_CTL2 = %x\n", readl(fbi->reg_base + LCD_MIXER1_CTL2));
        printk("LCD_MIXER1_OL1_CFG0 = %x\n", readl(fbi->reg_base + LCD_MIXER1_OL1_CFG0));
        printk("LCD_MIXER1_OL1_CFG1 = %x\n", readl(fbi->reg_base + LCD_MIXER1_OL1_CFG1));
	printk("DSI D1SCR0 = 0x%x\n", readl(fbi->reg_base + 0x3000));
	printk("DSI D1SCR1 = 0x%x\n", readl(fbi->reg_base + 0x3100));
	printk("DSI D1SSR  = 0x%x\n", readl(fbi->reg_base + 0x3104));
}

static u32 pxa168fb_base_set_colorkeyalpha(struct pxa950fb_info *fbi)
{
	unsigned int x = 0;
	struct _sColorKeyNAlpha *color_a = &fbi->ckey_alpha;

	dev_dbg(fbi->fb_info->dev, "Enter %s\n", __FUNCTION__);

	if (color_a->win_alpha_en)
		/* YCbCr formate can only select window specific alpha */
		x = LCD_CHx_ALPHA_W_ALPHA_EN | LCD_CHx_ALPHA_CH_ALPHA(color_a->default_alpha_val);
	else {
		/*
		 * color_a->alpha_method = 0: color match alpha method (only for RGB16 and RGB24)
		 * color_a->alpha_method = 1: lower bit alpha method (only for RGB16 and RGB24)
		 * color_a->alpha_method = 2: source alpha method (only for RGB32)
		 */
		if ((LCD_CHx_ALPHA_CLR_KEY_EN(color_a->alpha_method)) == 0) {
			x = LCD_CHx_ALPHA_CH_ALPHA(color_a->default_alpha_val);
			pxa950fb_writel(color_a->color_match, fbi->reg_base, LCD_CH0_CLR_MATCH);
		} else
			x = LCD_CHx_ALPHA_CLR_KEY_EN(color_a->alpha_method);
	}
	pxa950fb_writel(x, fbi->reg_base, LCD_CH0_ALPHA);

	return 0;
}

static int pxa950fb_graphic_ioctl(struct fb_info *fi, unsigned int cmd,
		unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	u32 x;
	int gfx_on = 1;
	/*This API will work on both Mixers, if enabled*/
	switch (cmd) {
		case FB_IOCTL_SWITCH_GRA_OVLY:
			if (copy_from_user(&gfx_on, argp, sizeof(int))) {
				return -EFAULT;
			}
			if (0 == gfx_on) {
				if (LCD_Conf.LCD_Mixer[0] !=
							LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base
							+ LCD_MIXER0_CTL0);
					x |= LCD_MIXERx_CTL0_OL1_ID_MASK
						| LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base
							+ LCD_MIXER1_CTL0);
					x |= LCD_MIXERx_CTL0_OL1_ID_MASK
						| LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							LCD_MIXER1_CTL0);
				}
			} else {
				printk(KERN_INFO "%s: FB_IOCTL_SWITCH_GRA_OVLY\n",
					 __func__);
				if (LCD_Conf.LCD_Mixer[0] !=
							LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base
						+ LCD_MIXER0_CTL0);
					x &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
					x |=
					LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID)
					|LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x,
					 fbi->reg_base, LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base
						+ LCD_MIXER1_CTL0);
					x &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
					x |=
					 LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID)
					|LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
						 LCD_MIXER1_CTL0);
				}
			}
			break;
		case FB_IOCTL_SET_COLORKEYnALPHA:
			if (copy_from_user(&fbi->ckey_alpha, argp,
						sizeof(struct _sColorKeyNAlpha)))
				return -EFAULT;

			pxa168fb_base_set_colorkeyalpha(fbi);
			break;
		case FB_IOCTL_GET_COLORKEYnALPHA:
			if (copy_to_user(argp, &fbi->ckey_alpha,
						sizeof(struct _sColorKeyNAlpha)))
				return -EFAULT;
			break;
		default:
			break;
	}
	return 0;
}

static void controller_enable_disable(struct pxa950fb_info *fbi, int onoff)
{
	int ctrl, stat;
	ctrl = readl(fbi->reg_base + LCD_CTL);
	if (!(ctrl & LCD_CTL_AXI32_EN))
		printk("%s: AXI32_EN is zero which should not happen!!!\n", __func__);
	switch(onoff)
	{
		case LCD_Controller_Enable:
			ctrl |= LCD_CTL_LCD_EN | LCD_CTL_GMIX_INT_EN | LCD_CTL_AXI32_EN;
			pxa950fb_writel(ctrl, fbi->reg_base, LCD_CTL);
			stat = readl(fbi->reg_base + LCD_CTL_STS);
			while (!(stat & LCD_CTL_STS_LCD_EN_STS))
				stat = readl(fbi->reg_base + LCD_CTL_STS);
			display_enabled = 1;
			break;
		case LCD_Controller_Disable:
			ctrl = readl(fbi->reg_base + LCD_CTL);
			ctrl &= ~(LCD_CTL_LCD_EN);
			pxa950fb_writel(ctrl, fbi->reg_base, LCD_CTL);
			display_enabled = 0;
			/* after disable, the LCD_CTL_STS[LCD_CTL_STS_LCD_EN_STS] is cleared only after
			 * send a frame to DSI. So cannot polling status bit here.
			 * */
			break;
		case LCD_Controller_Quick_Disable:
			ctrl = readl(fbi->reg_base + LCD_CTL);
			ctrl |= LCD_CTL_LCD_QD;
			pxa950fb_writel(ctrl, fbi->reg_base, LCD_CTL);
			stat = readl(fbi->reg_base + LCD_CTL_STS);
			while (!(stat & LCD_CTL_STS_LCD_QD_STS))
				stat = readl(fbi->reg_base + LCD_CTL_STS);
			display_enabled = 0;
			break;
		default:
			printk("%s: invalid parameter\n", __func__);
			break;
	}
}

static void set_fetch(struct pxa950fb_info *fbi, int channel)
{
	/* set DMA descriptor register */
	DisplayControllerFrameDescriptor *dmadesc_cpu;
	uint dmadesc_dma;
	dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)dma_addr_cpu - 16);
	dmadesc_dma = dma_addr_dma - 16;
	dmadesc_cpu->LCD_FR_IDx = 0;
	dmadesc_cpu->LCD_CH_CMDx = (u32)(LCD_Conf.LCD_Horizontal*
		LCD_Conf.LCD_Vertical*(u16)LCD_Conf.LCD_BPP);
	dmadesc_cpu->LCD_NXT_DESC_ADDRx = dmadesc_dma;
	dmadesc_cpu->LCD_FR_ADDRx = dma_addr_dma;
	pxa950fb_regs.descregs[channel].LCD_NXT_DESC_ADDRx = dmadesc_dma;

	/* set fetch registers */
	pxa950fb_regs.fetchregs[channel].LCD_FETCH_CTLx &= ~(LCD_FETCH_CTLx_SRC_FOR(0x7));
	/* set format */
	pxa950fb_regs.fetchregs[channel].LCD_FETCH_CTLx |=
		LCD_FETCH_CTLx_CHAN_EN|LCD_FETCH_CTLx_SRC_FOR(fbi->pix_fmt);


	pxa950fb_writel(pxa950fb_regs.descregs[0].LCD_NXT_DESC_ADDRx,
			fbi->reg_base, LCD_NXT_DESC_ADDR0);
	pxa950fb_writel(pxa950fb_regs.fetchregs[0].LCD_FETCH_CTLx,
			fbi->reg_base, LCD_FETCH_CTL0);

}

static void set_window(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var)
{
	/* set input image resolution */
	pxa950fb_regs.wgregs[0].LCD_WINx_CTL &= ~(LCD_WINx_CTL_WIN_XRES(0x1ff)
		|LCD_WINx_CTL_WIN_YRES(0x1ff));
	pxa950fb_regs.wgregs[0].LCD_WINx_CTL = LCD_WINx_CTL_WIN_XRES(var->xres/4)
		|LCD_WINx_CTL_WIN_YRES(var->yres/4);
	//Lincoln:old config
	//pxa950fb_regs.wgregs[0].LCD_CHx_ALPHA = 0x800000ff;

	pxa950fb_writel(pxa950fb_regs.wgregs[0].LCD_WINx_CTL,
			fbi->reg_base, LCD_WIN0_CTL);
	//Lincoln:old config
	//pxa950fb_writel(pxa950fb_regs.wgregs[0].LCD_CHx_ALPHA,
	//		fbi->reg_base, LCD_CH0_ALPHA);

	/* TODO - scale setting */
}

static void set_scale(struct pxa950fb_info *fbi)
{
	u32 scale_lut[] = {0x00004000, 0xff0940f8, 0xfe0f40f3, 0xfc163ef0, 0xfa1c3cee,
		0xf7223aed, 0xf42837ed, 0xf22c34ee, 0xf03030f0, 0xee342cf2, 0xed3728f4,
		0xed3a22f7, 0xee3c1cfa, 0xf03e16fc, 0xf3400ffe, 0xf84009ff};
	int i;
	for (i = 0; i < 16; i++)
	{
		pxa950fb_writel(i, fbi->reg_base, LCD_WIN4_SCALE_PTR);
		pxa950fb_writel(scale_lut[i], fbi->reg_base, LCD_WIN4_SCALE_RW);
	}
}

static void set_gamma(struct pxa950fb_info *fbi, u32 converter)
{
	int i, x;
	u32 RedLinearLUT[] = {0x30201000, 0x70605040, 0xb0a09080, 0xf0e0d0c0, 0x000000ff};
	u32 GreenLinearLUT[] = {0x30201000, 0x70605040, 0xb0a09080, 0xf0e0d0c0, 0x000000ff};
	u32 BlueLinearLUT[] = {0x30201000, 0x70605040, 0xb0a09080, 0xf0e0d0c0, 0x000000ff};
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	for(i=0;i<5;i++)
	{
		pxa950fb_writel(LCD_CONVx_G_CTL_ADD_PTR_R(i)
			 | LCD_CONVx_G_CTL_ADD_PTR_G(i)
			| LCD_CONVx_G_CTL_ADD_PTR_B(i),
			 conv_base, LCD_DSI_Dx_G_CTL_OFFSET);
		pxa950fb_writel(RedLinearLUT[i],
			 conv_base, LCD_DSI_Dx_G_DAT_RED_OFFSET);
		pxa950fb_writel(GreenLinearLUT[i],
			 conv_base, LCD_DSI_Dx_G_DAT_GREEN_OFFSET);
		pxa950fb_writel(BlueLinearLUT[i],
			 conv_base, LCD_DSI_Dx_G_DAT_BLUE_OFFSET);
	}
	x = readl(conv_base + LCD_DSI_Dx_G_CTL_OFFSET);
	x |= LCD_CONVx_G_CTL_Q4_C(0)|LCD_CONVx_G_CTL_Q3_C(0)
		|LCD_CONVx_G_CTL_Q2_C(0)|LCD_CONVx_G_CTL_Q1_C(0);
	pxa950fb_writel(x, conv_base, LCD_DSI_Dx_G_CTL_OFFSET);
}

static void set_parallel_converter(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var)
{
	struct pxa950fb_mach_info *mi;
	mi = fbi->dev->platform_data;

	pxa950fb_regs.convregs.LCD_MIXERx_TIM0 = ((var->sync & FB_SYNC_VERT_HIGH_ACT) ? 0 : (1 << 2))
		|((var->sync & FB_SYNC_HOR_HIGH_ACT) ? 0 : (1 << 3))
		|(1 << 6)	/* VSYNC delay to remove the bottom white line */
		|(mi->invert_pixclock ? 1 : 0)	/* only can be called from probe since mi->invert_pixclock is not saved*/
		|LCD_MIXERx_TIM0_HSW(var->hsync_len)
		|LCD_MIXERx_TIM0_ELW(var->left_margin)
		|LCD_MIXERx_TIM0_BLW(var->right_margin);
	pxa950fb_regs.convregs.LCD_MIXERx_TIM1 = LCD_MIXERx_TIM1_VSW(var->vsync_len)
		|LCD_MIXERx_TIM1_EFW(var->upper_margin)
		|LCD_MIXERx_TIM1_BFW(var->lower_margin);
	/* set converter registers */
	pxa950fb_regs.convregs.LCD_CONVx_CTL = LCD_CONVx_CTL_TV_FOR
		|LCD_CONVx_CTL_OP_FOR(LCD_Conf.LCD_OP_FOR)
		|LCD_CONVx_CTL_DISP_TYPE(LCD_Conf.LCD_Display);

	pxa950fb_writel(pxa950fb_regs.convregs.LCD_MIXERx_TIM0,
			fbi->reg_base, LCD_MIXER0_TIM0);
	pxa950fb_writel(pxa950fb_regs.convregs.LCD_MIXERx_TIM1,
			fbi->reg_base, LCD_MIXER0_TIM1);
	pxa950fb_writel(pxa950fb_regs.convregs.LCD_CONVx_CTL,
			fbi->reg_base, LCD_CONV0_CTL);

}

static void set_dsi_converter(struct pxa950fb_info *fbi, u8 converter)
{
	u32 dsi_clock;
	void *conv_base = CONVERTER_BASE_ADDRESS(fbi->reg_base, converter);

	/*Set Gamma correction*/
	set_gamma(fbi, converter);
#if 0
	if(!LCD_Controller_Set_Gamma(LCD_DSI))
	{
		printk("\n\rLCD_Controller_Gamma failed!!\n\r");
		ret = 0;
	}
#endif

	/*Get the DSI Clock*/
	dsi_clock = lcd_get_dsi_clock(converter);

	/*Set DSI timings according to the frequency*/
	lcd_set_dsi_time(fbi, converter, dsi_clock);

	/*Enable the DSI converter with primary channel*/
	pxa950fb_writel(LCD_DSI_DxSCR0_PRIM_VC|LCD_DSI_DxSCR0_FLOW_DIS,
				 conv_base, LCD_DSI_DxSCR0_OFFSET);

	/*Set to bidirectional panel*/
	pxa950fb_writel(LCD_DSI_DxSCR1_PT, conv_base, LCD_DSI_DxSCR1_OFFSET);

	/*Send DSI commands and enable dsi*/
	if(!dsi_init(fbi, converter))
	{
		printk("%s: dsi_init failed!!\n", __func__);
	}
}

static void set_mixer(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var, u8 overlay)
{
	uint LCD_MIXERx_CTL0_Value;
	int index;
	/* set mixer registers */
	switch (LCD_Conf.LCD_Mixer[0]) {
		case LCD_MIXER_DISABLE:
			break;
		case LCD_M2PARALELL_CONVERTER:
			pxa950fb_regs.mixerregs[0].LCD_MIXERx_CTL2 =
				LCD_MIXERx_CTL2_CONV_ID(LCD_Controller_P_CONVERTER0);
			break;
		case LCD_M2DSI0:
			pxa950fb_regs.mixerregs[0].LCD_MIXERx_CTL2 = 
				LCD_MIXERx_CTL2_CONV_ID(LCD_Controller_S_CONVERTER0);
			break;
		default:
			printk("%s: invalid LCD_Conf.LCD_Mixer[0] !\n", __func__);
			break;
	}
	switch (LCD_Conf.LCD_Mixer[1]) {
		case LCD_MIXER_DISABLE:
			break;
		case LCD_M2DSI1:
			pxa950fb_regs.mixerregs[1].LCD_MIXERx_CTL2 = 
				LCD_MIXERx_CTL2_CONV_ID(LCD_Controller_S_CONVERTER1);
			break;
		default:
			printk("%s: invalid LCD_Conf.LCD_Mixer[1] !\n", __func__);
			break;
	}
	for (index = 0; index < 2; index++) {
		if ((index == 0) && (LCD_Conf.LCD_Mixer[0] == LCD_MIXER_DISABLE))
			continue;
		if ((index == 1) && (LCD_Conf.LCD_Mixer[1] == LCD_MIXER_DISABLE))
			continue;
		pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 =
			readl(fbi->reg_base + ((index == 0) ? LCD_MIXER0_CTL0 : LCD_MIXER1_CTL0));

		pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL1 &= 
			~(LCD_MIXERx_CTL1_DISP_XRES(0x7ff) | LCD_MIXERx_CTL1_DISP_YRES(0x7ff));
		pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL1 |= 
			LCD_MIXERx_CTL1_DISP_XRES(var->xres) | LCD_MIXERx_CTL1_DISP_YRES(var->yres);

		switch (overlay) {
			case LCD_Controller_Base_Plane:	
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG0 &= 
					~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG0 |= 
					LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) 
					| LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG1 &= 
					~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG1 |= 
					LCD_MIXERx_CFG1_XRES(var->xres) | LCD_MIXERx_CFG1_YRES(var->yres);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_BP_ID(LCD_Conf.BP_ID);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_BP_ID_MASK;
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 |= 
					LCD_MIXERx_CTL0_Value | LCD_MIXERx_CTL0_MIX_EN;
				pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG0,
					fbi->reg_base,
					((index == 0) ? LCD_MIXER0_BP_CFG0 : LCD_MIXER1_BP_CFG0));
				pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_BP_CFG1,
					fbi->reg_base,
					((index == 0) ? LCD_MIXER0_BP_CFG1 : LCD_MIXER1_BP_CFG1));
				break;
			case LCD_Controller_Overlay1:
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG0 &= 
					~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG0 |= 
					LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) 
					| LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG1 &= 
					~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG1 |= 
					LCD_MIXERx_CFG1_XRES(var->xres)	| LCD_MIXERx_CFG1_YRES(var->yres);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 &= 
					~LCD_MIXERx_CTL0_OL1_ID_MASK;
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 |= 
					LCD_MIXERx_CTL0_Value | LCD_MIXERx_CTL0_MIX_EN;
				pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG0,
					fbi->reg_base,
					((index == 0) ? LCD_MIXER0_OL1_CFG0 : LCD_MIXER1_OL1_CFG0));
				pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL1_CFG1,
					fbi->reg_base,
					((index == 0) ? LCD_MIXER0_OL1_CFG1 : LCD_MIXER1_OL1_CFG1));
				break;
			case LCD_Controller_Overlay2:
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL2_CFG0 &= 
					~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL2_CFG0 |= 
					LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) 
					| LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL2_CFG1 &= 
					~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL2_CFG1 |= 
					LCD_MIXERx_CFG1_XRES(LCD_Conf.LCD_Horizontal) 
					| LCD_MIXERx_CFG1_YRES(LCD_Conf.LCD_Vertical);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL2_ID(LCD_Conf.OL2_ID);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_OL2_ID_MASK;
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 |= 
					LCD_MIXERx_CTL0_Value | LCD_MIXERx_CTL0_MIX_EN;
				break;
			case LCD_Controller_Overlay3:
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL3_CFG0 &= 
					~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL3_CFG0 |= 
					LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) 
					| LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL3_CFG1 &= 
					~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_OL3_CFG1 |= 
					LCD_MIXERx_CFG1_XRES(LCD_Conf.LCD_Horizontal) 
					| LCD_MIXERx_CFG1_YRES(LCD_Conf.LCD_Vertical);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL3_ID(LCD_Conf.OL3_ID);
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
				pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 |= 
					LCD_MIXERx_CTL0_Value | LCD_MIXERx_CTL0_MIX_EN;
				break;

		}
		pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL1 | LCD_MIXERx_CTL1_DISP_UPDATE_INT_EN,
			fbi->reg_base,
			((index == 0) ? LCD_MIXER0_CTL1 : LCD_MIXER1_CTL1));
		pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL2,
			fbi->reg_base,
			((index == 0) ? LCD_MIXER0_CTL2 : LCD_MIXER1_CTL2));
		/* LCD_MIXERx_CTL0_DISP_UPDATE should be set as the last step */
		pxa950fb_writel(pxa950fb_regs.mixerregs[index].LCD_MIXERx_CTL0 | LCD_MIXERx_CTL0_DISP_UPDATE,
			fbi->reg_base,
			((index == 0) ? LCD_MIXER0_CTL0 : LCD_MIXER1_CTL0));
	}
}

/* only can be called from probe since mi->modes is not saved*/
static void set_configuration(struct pxa950fb_info *fbi)
{
	struct pxa950fb_mach_info *mi = fbi->dev->platform_data;

	/* Modify Configuration structre */
	LCD_Conf.LCD_FRAME_BUFFER = (u32)dma_addr_dma;	/* TODO */

	LCD_Conf.LCD_Horizontal = mi->modes->xres;
	LCD_Conf.LCD_Vertical = mi->modes->yres;

	LCD_Conf.LCD_Position_x = 0;
	LCD_Conf.LCD_Position_y = 0;
	
	/* 
	 * parallel setting don't depend on below LCD_Conf.LCD_XXW
	 * since mi->modes are copied to var.
	 * TODO: need to think about the using mode for DSI
	 */
	LCD_Conf.LCD_HSW = mi->modes->hsync_len;
	LCD_Conf.LCD_VSW = mi->modes->vsync_len;
	LCD_Conf.LCD_BLW = mi->modes->right_margin;
	LCD_Conf.LCD_ELW = mi->modes->left_margin;
	LCD_Conf.LCD_BFW = mi->modes->lower_margin;
	LCD_Conf.LCD_EFW = mi->modes->upper_margin;

	LCD_Conf.LCD_BPP = BYTES_PER_PIXEL;
	LCD_Conf.LCD_Display = mi->panel_type;
	LCD_Conf.LCD_OP_FOR = mi->pix_fmt_out;
	LCD_Conf.LCD_SRC_FOR = mi->pix_fmt_in;

	/* 
	 * NOTE:
	 * currently set WGe0 (base layer) as OV1_ID in mixer
	 * to make sure base layer is on top of overlay (WGe3/WGe4)
	 */
	LCD_Conf.LCD_Overlay = LCD_Controller_Overlay1;
	LCD_Conf.OL1_ID = LCD_Controller_WGe0;

	LCD_Conf.LCD_Mixer[0] = mi->mixer0;
	LCD_Conf.LCD_Mixer[1] = mi->mixer1;
	return;
}


#define MAX_HWC_SIZE		(64*64*2)
#define MAX_REFRESH_RATE	100	/* Hz */
#define MIN_REFRESH_RATE	47	/* Hz */
#define DEFAULT_REFRESH		60	/* Hz */

static int determine_best_pix_fmt(struct fb_var_screeninfo *var)
{
	/*
	 * Pseudocolor mode?
	 */
	if (var->bits_per_pixel == 8)
		return PIX_FMT_PSEUDOCOLOR;

	/*
	 * Check for 565/1555.
	 */
	if (var->bits_per_pixel == 16 && var->red.length <= 5 &&
			var->green.length <= 6 && var->blue.length <= 5) {
		if (var->transp.length == 0) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_16;
			else
				printk("%s: BGR565 format is not supported!\n", __func__);
		}
	}

	/*
	 * Check for 888/A888.
	 */
	if (var->bits_per_pixel <= 32 && var->red.length <= 8 &&
			var->green.length <= 8 && var->blue.length <= 8) {
		if (var->bits_per_pixel == 24 && var->transp.length == 0) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_24_PACK;
			else
				printk("%s: BGR888 packed format is not supported!\n", __func__);
		}

		if (var->bits_per_pixel == 32 && var->transp.length == 8) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_32;
			else
				printk("%s: BGRA888 format is not supported!\n", __func__);
		} else {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_24;
			else
				printk("%s: BGR888 format is not supported!\n", __func__);;
		}

		/* fall through */
	}

	return -EINVAL;


}

static void set_pix_fmt(struct fb_var_screeninfo *var, int pix_fmt)
{
	switch (pix_fmt) {
		case PIX_FMTIN_RGB_16:
			var->bits_per_pixel = 16;
			var->red.offset = 11;    var->red.length = 5;
			var->green.offset = 5;   var->green.length = 6;
			var->blue.offset = 0;    var->blue.length = 5;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_RGB_24:
			var->bits_per_pixel = 32;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_RGB_32:
			var->bits_per_pixel = 32;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 24; var->transp.length = 8;
			break;
		case PIX_FMTIN_RGB_24_PACK:
			var->bits_per_pixel = 24;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMT_PSEUDOCOLOR:
			var->bits_per_pixel = 8;
			var->red.offset = 0;     var->red.length = 8;
			var->green.offset = 0;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
	}
}
static void set_mode(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var,
		struct fb_videomode *mode, int pix_fmt, int ystretch)
{
	set_pix_fmt(var, pix_fmt);

	var->xres = mode->xres;
	var->yres = mode->yres;
	var->xres_virtual = max(var->xres, var->xres_virtual);
	if (ystretch)
		var->yres_virtual = fbi->fb_size /
			(var->xres_virtual * (var->bits_per_pixel >> 3));
	else
		var->yres_virtual = max(var->yres, var->yres_virtual);
	var->grayscale = 0;
	var->accel_flags = FB_ACCEL_NONE;
	var->pixclock = mode->pixclock;
	var->left_margin = mode->left_margin;
	var->right_margin = mode->right_margin;
	var->upper_margin = mode->upper_margin;
	var->lower_margin = mode->lower_margin;
	var->hsync_len = mode->hsync_len;
	var->vsync_len = mode->vsync_len;
	var->sync = mode->sync;
	var->vmode = FB_VMODE_NONINTERLACED;
	var->rotate = FB_ROTATE_UR;
}

static int pxa950fb_check_var(struct fb_var_screeninfo *var,
		struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;

	if (var->bits_per_pixel == 8)
		return -EINVAL;
	/*
	 * Basic geometry sanity checks.
	 */
	if (var->xoffset + var->xres > var->xres_virtual)
		return -EINVAL;
	if (var->yoffset + var->yres > var->yres_virtual)
		return -EINVAL;
	if (var->xres + var->right_margin +
			var->hsync_len + var->left_margin > 2048)
		return -EINVAL;
	if (var->yres + var->lower_margin +
			var->vsync_len + var->upper_margin > 2048)
		return -EINVAL;

	/*
	 * Check size of framebuffer.
	 */
	if (var->xres_virtual * var->yres_virtual *
			(var->bits_per_pixel >> 3) > fbi->fb_size)
		return -EINVAL;

	return 0;
}

/*
 * The hardware clock divider has an integer and a fractional
 * stage:
 *
 *	clk2 = clk_in / integer_divider
 *	clk_out = clk2 * (1 - (fractional_divider >> 12))
 *
 * Calculate integer and fractional divider for given clk_in
 * and clk_out.
 */
static void set_clock_divider(struct pxa950fb_info *fbi,
		const struct fb_videomode *m)
{
	int divider_int;
	int needed_pixclk;
	int hsio_clk;
	u64 div_result;
	u32 x = 0;

	/*
	 * Notice: The field pixclock is used by linux fb
	 * is in pixel second. E.g. struct fb_videomode &
	 * struct fb_var_screeninfo
	 */

	/*
	 * Check input values.
	 */
	if (!m || !m->pixclock || !m->refresh) {
		printk(KERN_ERR "Input refresh or pixclock is wrong.\n");
		return;
	}

	/*
	 * Calc divider according to refresh rate.
	 */
	div_result = 1000000000000ll;
	do_div(div_result, m->pixclock);
	needed_pixclk = (u32)div_result;

	switch ((ACSR & 0xC000) >> 14)	/* HSS */
	{
		case LCD_Controller_104:
			hsio_clk = 104000000ll;
			break;
		case LCD_Controller_156:
			hsio_clk = 156000000ll;
			break;
		case LCD_Controller_208:
			hsio_clk = 208000000ll;
			break;
		default:
			hsio_clk = 104000000ll;
			break;
	}

	divider_int = hsio_clk / needed_pixclk;

	/* check whether divisor is too small. */
	if (divider_int < 2) {
		printk(KERN_WARNING "Warning: clock source is too slow."
				"Try smaller resolution\n");
		divider_int = 2;
	}

	/*
	 * Set setting to reg. undocumented? - TODO
	 */
	x = readl(fbi->reg_base + LCD_CTL);	
	x |= divider_int;
	pxa950fb_writel(x, fbi->reg_base, LCD_CTL);
}

static void set_graphics_start(struct fb_info *info, int xoffset, int yoffset)
{
	struct pxa950fb_info *fbi = info->par;
	DisplayControllerFrameDescriptor *dmadesc_cpu;
	int pixel_offset;
	struct fb_var_screeninfo *var = &info->var;

	dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)dma_addr_cpu - 16);
	pixel_offset = ((yoffset * var->xres_virtual) + xoffset) * var->bits_per_pixel / 8;

	info->screen_base = fbi->fb_start+pixel_offset;

#if defined(CONFIG_FBCON_DRAW_PANIC_TEXT)
	if( unlikely(panic_text_displayed) )
	{
		if( fbcon_draw_panic_text_callback )
			fbcon_draw_panic_text_callback();
	}
#endif

	dmadesc_cpu->LCD_FR_ADDRx = dma_addr_dma + pixel_offset;
	return;
}

static int pxa950fb_set_par(struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;
	struct fb_var_screeninfo *var = &info->var;
	const struct fb_videomode *m = 0;
	struct fb_videomode mode;
	int pix_fmt;
	struct pxa950fb_mach_info *mi;
	mi = fbi->dev->platform_data;
	/*
	 * Determine which pixel format we're going to use.
	 */
	pix_fmt = determine_best_pix_fmt(&info->var);
	if (pix_fmt < 0)
		return pix_fmt;
	fbi->pix_fmt = pix_fmt;

	/*
	 * Set additional mode info.
	 */
	if (pix_fmt == PIX_FMT_PSEUDOCOLOR)
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
	else
		info->fix.visual = FB_VISUAL_TRUECOLOR;
	info->fix.line_length = var->xres_virtual * var->bits_per_pixel / 8;
	info->fix.ypanstep = var->yres;

	/*
	 * Configure global panel parameters.
	 */
	/*
	 * convet var to video mode
	 */
	fb_var_to_videomode(&mode, &info->var);
	m = &mode;

	/* set fetch registers */
	set_fetch(fbi, 0);	/* only use channel 0 for base frame */
	/* set window registers */
	set_window(fbi, var);
	/* set pixclock */

	set_clock_divider(fbi, &mode);
	/* configure graphics dma */
	set_graphics_start(info, info->var.xoffset, info->var.yoffset);

	/* set mixer registers */
	set_mixer(fbi, var, LCD_Conf.LCD_Overlay);

	return 0;
}

static unsigned int chan_to_field(unsigned int chan, struct fb_bitfield *bf)
{
	return ((chan & 0xffff) >> (16 - bf->length)) << bf->offset;
}

static u32 to_rgb(u16 red, u16 green, u16 blue)
{
	red >>= 8;
	green >>= 8;
	blue >>= 8;

	return (red << 16) | (green << 8) | blue;
}

	static int
pxa950fb_setcolreg(unsigned int regno, unsigned int red, unsigned int green,
		unsigned int blue, unsigned int trans, struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;
	u32 val;

	if (info->fix.visual == FB_VISUAL_TRUECOLOR && regno < 16) {
		val =  chan_to_field(red,   &info->var.red);
		val |= chan_to_field(green, &info->var.green);
		val |= chan_to_field(blue , &info->var.blue);
		fbi->pseudo_palette[regno] = val;
	}

	if (info->fix.visual == FB_VISUAL_PSEUDOCOLOR && regno < 256) {
		val = to_rgb(red, green, blue);
		/* TODO */
	}

	return 0;
}

static int pxa950fb_blank(int blank, struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;

	mutex_lock(&fbi->access_ok);

	fbi->is_blanked = (blank == FB_BLANK_UNBLANK) ? 0 : 1;
	mutex_unlock(&fbi->access_ok);

	return 0;	/* TODO */
}

static int pxa950fb_pan_display(struct fb_var_screeninfo *var,
		struct fb_info *info)
{
	set_graphics_start(info, var->xoffset, var->yoffset);

	return 0;
}

static irqreturn_t pxa950fb_handle_irq(int irq, void *dev_id)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)dev_id;
	u32     isr;

	if (LCD_Conf.LCD_Mixer[0] != LCD_MIXER_DISABLE) {
		isr = readl(fbi->reg_base + LCD_MIXER0_INT_STS);
		writel(isr, fbi->reg_base + LCD_MIXER0_INT_STS);
		if ((isr & LCD_MIXERx_CTL1_DISP_UPDATE_INT_EN)) {
			/* wake up queue. */
			//atomic_set(&fbi->w_intr, 1);
			//wake_up(&fbi->w_intr_wq);
			display_update_ongoing = 0;
		}
	}

	if (LCD_Conf.LCD_Mixer[1] != LCD_MIXER_DISABLE) {
		isr = readl(fbi->reg_base + LCD_MIXER1_INT_STS);
		writel(isr, fbi->reg_base + LCD_MIXER1_INT_STS);
		if ((isr & LCD_MIXERx_CTL1_DISP_UPDATE_INT_EN)) {
			/* wake up queue. */
			//atomic_set(&fbi->w_intr, 1);
			//wake_up(&fbi->w_intr_wq);
			display_update_ongoing = 0;
		}
	}
	return IRQ_HANDLED;
}

static struct fb_ops pxa950fb_ops = {
	.owner		= THIS_MODULE,
	.fb_check_var	= pxa950fb_check_var,
	.fb_set_par	= pxa950fb_set_par,
	.fb_setcolreg	= pxa950fb_setcolreg,	/* TODO */
	.fb_blank	= pxa950fb_blank,
	.fb_pan_display	= pxa950fb_pan_display,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_ioctl	= pxa950fb_graphic_ioctl,
};

static int pxa950fb_init_mode(struct fb_info *info,
		struct pxa950fb_mach_info *mi)
{
	struct pxa950fb_info *fbi = info->par;
	struct fb_var_screeninfo *var = &info->var;
	int ret = 0;
	u32 total_w, total_h, refresh;
	u64 div_result;
	const struct fb_videomode *m;

	/*
	 * Set default value
	 */
	refresh = DEFAULT_REFRESH;

	/*
	 * If has bootargs, apply it first.
	 */
	if (fbi->dft_vmode.xres && fbi->dft_vmode.yres &&
			fbi->dft_vmode.refresh) {
		/* set data according bootargs */
		var->xres = fbi->dft_vmode.xres;
		var->yres = fbi->dft_vmode.yres;
		refresh = fbi->dft_vmode.refresh;
	}

	/* try to find best video mode. */
	m = fb_find_best_mode(&info->var, &info->modelist);
	if (m)
		fb_videomode_to_var(&info->var, m);

	/* Init settings. */
	var->xres_virtual = var->xres;
	var->yres_virtual = fbi->fb_size /
		(var->xres_virtual * (var->bits_per_pixel >> 3));
	printk("pxa950fb: find best mode: res = %dx%d\n", var->xres, var->yres);

	/* correct pixclock. */
	total_w = var->xres + var->left_margin + var->right_margin +
		var->hsync_len;
	total_h = var->yres + var->upper_margin + var->lower_margin +
		var->vsync_len;

	div_result = 1000000000000ll;
	do_div(div_result, total_w * total_h * refresh);
	var->pixclock = (u32)div_result;

	return ret;
}

static void __attribute__ ((unused)) dump_regs(struct pxa950fb_info *fbi)
{
	/*TBD - pass converter as an argumnet for dynamic choice*/
	void *conv_base =
		CONVERTER_BASE_ADDRESS(fbi->reg_base, LCD_Controller_DSI0);

	printk(KERN_INFO "LCD_DSI_D0SCR0 = %x\n",
	readl(conv_base + LCD_DSI_DxSCR0_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0SCR1 = %x\n",
	readl(conv_base + LCD_DSI_DxSCR1_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0SSR = %x\n",
	readl(conv_base + LCD_DSI_DxSSR_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0TRIG = %x\n",
	readl(conv_base + LCD_DSI_DxTRIG_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0INEN = %x\n",
	readl(conv_base + LCD_DSI_DxINEN_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0INST1 = %x\n",
	readl(conv_base + LCD_DSI_DxINST1_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0TEINTCNT = %x\n",
	readl(conv_base + LCD_DSI_DxTEINTCNT_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0_G_CTL = %x\n",
	readl(conv_base + LCD_DSI_Dx_G_CTL_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0_G_DAT_RED = %x\n",
	readl(conv_base + LCD_DSI_Dx_G_DAT_RED_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0_G_DAT_GREEN = %x\n",
	readl(conv_base + LCD_DSI_Dx_G_DAT_GREEN_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0_G_DAT_BLUE = %x\n",
	readl(conv_base + LCD_DSI_Dx_G_DAT_BLUE_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0PRSR = %x\n",
	readl(conv_base + LCD_DSI_DxPRSR_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0INST0 = %x\n",
	readl(conv_base + LCD_DSI_DxINST0_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0ADAT = %x\n",
	readl(conv_base + LCD_DSI_DxADAT_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0CFIF = %x\n",
	readl(conv_base + LCD_DSI_DxCFIF_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0TIM0 = %x\n",
	readl(conv_base + LCD_DSI_DxTIM0_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0TIM1 = %x\n",
	readl(conv_base + LCD_DSI_DxTIM1_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0PHY_TIM0 = %x\n",
	readl(conv_base + LCD_DSI_DxPHY_TIM0_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0PHY_TIM1 = %x\n",
	readl(conv_base + LCD_DSI_DxPHY_TIM1_OFFSET));
	printk(KERN_INFO "LCD_DSI_D0PHY_TIM2 = %x\n",
	readl(conv_base + LCD_DSI_DxPHY_TIM2_OFFSET));
}

static int dsi_sendframe_thread(void *arg)
{
	struct pxa950fb_info *fbi = arg;
	int saved_blank_status=fbi->is_blanked;	// 1-> display disabled

	while (1) {
		/*		dsi_video_mode = DSI_VIDEO_MODE;*/
		mutex_lock(&fbi->access_ok);

		if (fbi->is_blanked == 0)
		{
			if (fbi->is_blanked != saved_blank_status)
			{
				/* to turn on the display */
				saved_blank_status = fbi->is_blanked;

				set_dvfm_constraint();

				clk_enable(fbi->clk_axi);
				clk_enable(fbi->clk_lcd);

				if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
					lcd_enable_dsi_clock(fbi, 1, LCD_Controller_DSI0);
				if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
					lcd_enable_dsi_clock(fbi, 1, LCD_Controller_DSI1);

				/* enable LCD controller clock */
				controller_enable_disable(fbi, LCD_Controller_Enable);
			}
			mutex_unlock(&fbi->access_ok);
			dsi_video_mode = DSI_COMMAND_MODE;
			if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
				dsi_send_frame(fbi, LCD_Controller_DSI0);
			if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
				dsi_send_frame(fbi, LCD_Controller_DSI1);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(30 * HZ / 1000);

		}
		else
		{
			if (fbi->is_blanked != saved_blank_status)
			{
				controller_enable_disable(fbi, LCD_Controller_Disable);
				if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
					dsi_send_frame(fbi, LCD_Controller_DSI0);
				if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
					dsi_send_frame(fbi, LCD_Controller_DSI1);

				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(30 * HZ / 1000);

				if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
					lcd_enable_dsi_clock(fbi, 0, LCD_Controller_DSI0);
				if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
					lcd_enable_dsi_clock(fbi, 0, LCD_Controller_DSI1);

				saved_blank_status = fbi->is_blanked;
				clk_disable(fbi->clk_lcd);
				clk_disable(fbi->clk_axi);

				unset_dvfm_constraint();
			}
			mutex_unlock(&fbi->access_ok);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(100 * HZ / 1000);
		}
	}
	return 0;
}


static int __init pxa950fb_probe(struct platform_device *pdev)
{
	struct pxa950fb_mach_info *mi;
	struct fb_info *info = 0;
	struct pxa950fb_info *fbi = 0;
	struct resource *res;
	struct clk *clk_lcd, *clk_dsi0, *clk_dsi1, *clk_axi;
	int irq, ret;

	struct task_struct *kthread_lcd;

	mi = pdev->dev.platform_data;
	if (mi == NULL) {
		dev_err(&pdev->dev, "no platform data defined\n");
		return -EINVAL;
	}

	clk_lcd = clk_get(&pdev->dev, "PXA950_LCDCLK");
	if (IS_ERR(clk_lcd)) {
		dev_err(&pdev->dev, "unable to get LCDCLK");
		return PTR_ERR(clk_lcd);
	}

	clk_axi = clk_get(NULL, "AXICLK");
	if (IS_ERR(clk_axi)) {
		dev_err(&pdev->dev, "unable to get axi bus clock");
		return PTR_ERR(clk_axi);
	}

	clk_dsi0 = clk_get(&pdev->dev, "PXA950_DSI0CLK");
	if (IS_ERR(clk_dsi0)) {
		dev_err(&pdev->dev, "unable to get DSI0 CLK");
		return PTR_ERR(clk_dsi0);
	}

	clk_dsi1 = clk_get(&pdev->dev, "PXA950_DSI1CLK");
	if (IS_ERR(clk_dsi1)) {
		dev_err(&pdev->dev, "unable to get DSI1 CLK");
		return PTR_ERR(clk_dsi1);
	}

	enable_oscc_tout_s0();

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "no IO memory defined\n");
		return -ENOENT;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "no IRQ defined\n");
		return -ENOENT;
	}

	info = framebuffer_alloc(sizeof(struct pxa950fb_info), &pdev->dev);
	if (info == NULL)
		return -ENOMEM;

	/* Initialize private data */
	fbi = info->par;
	platform_set_drvdata(pdev, fbi);
	fbi->clk_lcd = clk_lcd;
	fbi->clk_axi = clk_axi;
	fbi->clk_dsi0 = clk_dsi0;
	fbi->clk_dsi1 = clk_dsi1;
	fbi->dev = &pdev->dev;
	fbi->panel_rbswap = mi->panel_rbswap;
	fbi->is_blanked = 0;
	fbi->debug = 0;
	fbi->active = mi->active;

	/* Initialize boot setting */
	fbi->dft_vmode.xres = mi->modes->xres;
	fbi->dft_vmode.yres = mi->modes->yres;
	fbi->dft_vmode.refresh = mi->modes->refresh;

	mutex_init(&fbi->access_ok);

	init_waitqueue_head(&fbi->w_intr_wq);

	/*
	 * Initialise static fb parameters.
	 */
	info->flags = FBINFO_DEFAULT | FBINFO_PARTIAL_PAN_OK |
		FBINFO_HWACCEL_XPAN | FBINFO_HWACCEL_YPAN;
	info->node = -1;
	strcpy(info->fix.id, mi->id);
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.type_aux = 0;
	info->fix.xpanstep = 0;
	info->fix.ypanstep = 0;
	info->fix.ywrapstep = 0;
	info->fix.mmio_start = res->start;
	info->fix.mmio_len = res->end - res->start + 1;
	info->fix.accel = FB_ACCEL_NONE;
	info->fbops = &pxa950fb_ops;
	info->pseudo_palette = fbi->pseudo_palette;

	/*
	 * Map LCD controller registers.
	 */
	fbi->reg_base = ioremap_nocache(res->start, res->end - res->start);
	if (fbi->reg_base == NULL) {
		ret = -ENOMEM;
		goto failed;
	}


	/* Map DSI PLL registers*/
	MMCUbase = ioremap_nocache(MCCU_DSI_BASE_ADDRESS,
		MCCU_DSI_REGS_MEM_SIZE);
	if (MMCUbase == NULL) {
		ret = -ENOMEM;
		goto failed;
	}

	/*
	 * Allocate framebuffer memory.
	 */
	//fbi->fb_size = PAGE_ALIGN(DEFAULT_FB_SIZE);
	
fbi->fb_size = PAGE_ALIGN(mi->modes->xres * mi->modes->yres * PXA950FB_FB_NUM * 4 + PAGE_SIZE);

	fbi->fb_start = dma_alloc_writecombine(fbi->dev, fbi->fb_size,
			&fbi->fb_start_dma,
			GFP_KERNEL);
	if (fbi->fb_start == NULL) {
		ret = -ENOMEM;
		goto failed;
	}
	memset(fbi->fb_start, 0, fbi->fb_size);	
	fbi->fb_start = fbi->fb_start + PAGE_SIZE;
	fbi->fb_start_dma = fbi->fb_start_dma + PAGE_SIZE;
	fbi->fb_size = fbi->fb_size - PAGE_SIZE;

	dma_addr_dma = fbi->fb_start_dma;
	dma_addr_cpu = fbi->fb_start;
	info->fix.smem_start = fbi->fb_start_dma;
	info->fix.smem_len = fbi->fb_size;
	info->screen_base = fbi->fb_start;
	info->screen_size = fbi->fb_size/2;

	/* save machine info to LCD_Conf */
	set_configuration(fbi);

	/*
	 * Set video mode according to platform data.
	 */
	set_mode(fbi, &info->var, mi->modes, LCD_Conf.LCD_SRC_FOR, 1);
	fb_videomode_to_modelist(mi->modes, mi->num_modes, &info->modelist);
	/*
	 * init video mode data.
	 */
	pxa950fb_init_mode(info, mi);
	/*
	 * enable controller clock
	 */
	if ((LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0) || (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1))
		mi->pxa950fb_dsi_reset();

	/* CCU setting for LCD */
	clk_enable(fbi->clk_axi);
	clk_enable(fbi->clk_lcd);

	if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
	{
		lcd_enable_dsi_clock(fbi, 1, LCD_Controller_DSI0);
		lcd_set_dsi_clock(LCD_Controller_DSI_208MHZ,
				LCD_Controller_DSI0);
	}
	else
		lcd_enable_dsi_clock(fbi, 0, LCD_Controller_DSI0);
	if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
	{
		lcd_enable_dsi_clock(fbi, 1, LCD_Controller_DSI1);
		lcd_set_dsi_clock(LCD_Controller_DSI_208MHZ,
				LCD_Controller_DSI1);
	}
	else
		lcd_enable_dsi_clock(fbi, 0, LCD_Controller_DSI1);


	/* Enable AXI32 before modifying the controller registers */
	pxa950fb_writel(LCD_CTL_AXI32_EN, fbi->reg_base, LCD_CTL);

	/* read register value as default */
	pxa950fb_regs.fetchregs[0].LCD_FETCH_CTLx = 
		readl(fbi->reg_base + LCD_FETCH_CTL0);
	pxa950fb_regs.wgregs[0].LCD_WINx_CTL = 
		readl(fbi->reg_base + LCD_WIN0_CTL);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_CTL0 = 
		readl(fbi->reg_base + LCD_MIXER0_CTL0);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_CTL1 = 
		readl(fbi->reg_base + LCD_MIXER0_CTL1);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_CTL2 = 
		readl(fbi->reg_base + LCD_MIXER0_CTL2);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_BP_CFG0 = 
		readl(fbi->reg_base + LCD_MIXER0_BP_CFG0);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_BP_CFG1 = 
		readl(fbi->reg_base + LCD_MIXER0_BP_CFG1);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_OL1_CFG0 = 
		readl(fbi->reg_base + LCD_MIXER0_OL1_CFG0);
	pxa950fb_regs.mixerregs[0].LCD_MIXERx_OL1_CFG1 = 
		readl(fbi->reg_base + LCD_MIXER0_OL1_CFG1);

	pxa950fb_regs.mixerregs[1].LCD_MIXERx_CTL0 =
		readl(fbi->reg_base + LCD_MIXER1_CTL0);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_CTL1 =
		readl(fbi->reg_base + LCD_MIXER1_CTL1);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_CTL2 =
		readl(fbi->reg_base + LCD_MIXER1_CTL2);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_BP_CFG0 =
		readl(fbi->reg_base + LCD_MIXER1_BP_CFG0);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_BP_CFG1 =
		readl(fbi->reg_base + LCD_MIXER1_BP_CFG1);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_OL1_CFG0 = 
		readl(fbi->reg_base + LCD_MIXER1_OL1_CFG0);
	pxa950fb_regs.mixerregs[1].LCD_MIXERx_OL1_CFG1 = 
		readl(fbi->reg_base + LCD_MIXER1_OL1_CFG1);


	pxa950fb_regs.convregs.LCD_CONVx_CTL = 
		readl(fbi->reg_base + LCD_CONV0_CTL);

	controller_enable_disable(fbi, LCD_Controller_Disable);

	/*
	 * Register irq handler.
	 */
	ret = request_irq(irq, pxa950fb_handle_irq, IRQF_SHARED, mi->id, fbi);
	if (ret < 0) {
		dev_err(&pdev->dev, "unable to request IRQ\n");
		ret = -ENXIO;
		goto failed_free_cmap;
	}

	/* set scale registers */
	set_scale(fbi);

	pxa950fb_set_par(info);

	if (LCD_Conf.LCD_Mixer[0] == LCD_M2PARALELL_CONVERTER) {
		set_gamma(fbi, LCD_Controller_PARALLEL); /*set gamma registers*/
		set_parallel_converter(fbi, &info->var);
		controller_enable_disable(fbi, LCD_Controller_Enable);
	} else if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
		set_dsi_converter(fbi, LCD_Controller_DSI0);

	if (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)
		set_dsi_converter(fbi, LCD_Controller_DSI1);

	/*
	 * Allocate color map.
	 */
	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		ret = -ENOMEM;
		goto failed_free_clk;
	}

	/*
	 * Enable GFX interrupt
	 */

	/*
	 * Register framebuffer.
	 */
	ret = register_framebuffer(info);
	if (ret < 0) {
		dev_err(&pdev->dev, "Failed to register pxa950-fb: %d\n", ret);
		ret = -ENXIO;
		goto failed_free_irq;
	}
	printk(KERN_INFO "pxa950fb: frame buffer device was loaded"
			" to /dev/fb%d <%s>.\n", info->node, info->fix.id);

	set_dvfm_constraint();
	mi->pxa950fb_lcd_power(1);

	if ((LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0) || (LCD_Conf.LCD_Mixer[1] == LCD_M2DSI1)) {
		kthread_lcd = kthread_run(dsi_sendframe_thread, fbi,
				"lcd_refresh");
		if (IS_ERR(kthread_lcd)) {
			pr_err("%s: unable to create kernel thread\n", __func__);
			return PTR_ERR(kthread_lcd);
		}
	}
	return 0;

failed_free_irq:
	free_irq(irq, fbi);
failed_free_cmap:
	fb_dealloc_cmap(&info->cmap);
failed_free_clk:
	clk_disable(fbi->clk_lcd);
	clk_disable(fbi->clk_dsi0);
	clk_disable(fbi->clk_dsi1);
	clk_disable(fbi->clk_axi);
	clk_put(fbi->clk_lcd);
	clk_put(fbi->clk_dsi0);
	clk_put(fbi->clk_dsi1);
	clk_put(fbi->clk_axi);
failed:
	pr_err("pxa950-fb: frame buffer device init failed\n");
	platform_set_drvdata(pdev, NULL);
	fb_dealloc_cmap(&info->cmap);

	if (fbi && fbi->reg_base) {
		iounmap(fbi->reg_base);
		kfree(fbi);
	}

	return ret;
}

static struct platform_driver pxa950fb_driver = {
	.driver		= {
		.name	= "pxa950-fb",
		.owner	= THIS_MODULE,
	},
	.probe		= pxa950fb_probe,
};

static int __devinit pxa950fb_init(void)
{
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("pxa950-fb", &dvfm_dev_idx);
#endif
	return platform_driver_register(&pxa950fb_driver);
}
module_init(pxa950fb_init);

MODULE_AUTHOR("Lennert Buytenhek <buytenh@marvell.com>");
MODULE_DESCRIPTION("Framebuffer driver for PXA950");
MODULE_LICENSE("GPL");
