/*
 * drivers/video/pxa950fb.h
 *
 *  Copyright (C) 2009 Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef LCD_CONTROLLER_H
#define LCD_CONTROLLER_H

#include <linux/types.h>

#define BYTES_PER_PIXEL 2
#define PXA950FB_FB_NUM 2
/*   registers address */

#define LCD_CTL			0x0
#define LCD_CTL_STS		0x0004

#define LCD_FETCH_STS0			0x0204

#define LCD_NXT_DESC_ADDR0      0x0208
#define LCD_NXT_DESC_ADDR1      0x0248
#define LCD_NXT_DESC_ADDR2      0x0288
#define LCD_NXT_DESC_ADDR3      0x02C8
#define LCD_NXT_DESC_ADDR4      0x0308
#define LCD_NXT_DESC_ADDR5      0x0348
#define LCD_NXT_DESC_ADDR6      0x0388

#define LCD_FETCH_CTL0          0x0200
#define LCD_FETCH_CTL1          0x0240
#define LCD_FETCH_CTL2          0x0280
#define LCD_FETCH_CTL3          0x02C0
#define LCD_FETCH_CTL4          0x0300
#define LCD_FETCH_CTL5          0x0340
#define LCD_FETCH_CTL6          0x0380

#define LCD_WIN0_CTL            0x0600
#define LCD_WIN1_CTL            0x0640
#define LCD_WIN2_CTL            0x0680
#define LCD_WIN3_CTL            0x06C0
#define LCD_WIN4_CTL            0x07C0
#define LCD_WIN0_CFG            0x0604
#define LCD_WIN1_CFG            0x0644
#define LCD_WIN2_CFG            0x0684
#define LCD_WIN3_CFG            0x06C4
#define LCD_WIN4_CFG            0x07C4

#define LCD_WIN0_SCALE_RW	0x0618
#define LCD_WIN1_SCALE_RW       0x0658
#define LCD_WIN2_SCALE_RW       0x0698
#define LCD_WIN3_SCALE_RW       0x06D8
#define LCD_WIN4_SCALE_RW       0x07D8

#define LCD_WIN0_SCALE_PTR	0x0614
#define LCD_WIN1_SCALE_PTR      0x0654
#define LCD_WIN2_SCALE_PTR      0x0694
#define LCD_WIN3_SCALE_PTR      0x06D4
#define LCD_WIN4_SCALE_PTR      0x07D4

#define LCD_MIXER0_CTL0         0x0800
#define LCD_MIXER0_CTL1         0x0804
#define LCD_MIXER0_CTL2     	0x0844
#define LCD_MIXER0_STS			0x082C
#define LCD_MIXER0_STS1			0x0848
#define LCD_MIXER0_BP_CFG0      0x0808
#define LCD_MIXER0_BP_CFG1      0x080C
#define LCD_MIXER0_OL1_CFG0	0x0810
#define LCD_MIXER0_OL1_CFG1     0x0814
#define LCD_MIXER0_OL2_CFG0     0x0818
#define LCD_MIXER0_OL2_CFG1     0x081C


#define LCD_MIXER1_CTL0         0x0900
#define LCD_MIXER1_CTL1         0x0904
#define LCD_MIXER1_CTL2     	0x0944
#define LCD_MIXER1_STS			0x092C
#define LCD_MIXER1_STS1			0x0948
#define LCD_MIXER1_BP_CFG0      0x0908
#define LCD_MIXER1_BP_CFG1      0x090C
#define LCD_MIXER1_OL1_CFG0     0x0910
#define LCD_MIXER1_OL1_CFG1     0x0914

#define LCD_CH_ERR_ADDR         0x05A0
#define LCD_CH0_ALPHA       	0x05C0
#define LCD_CH1_ALPHA       	0x05C4
#define LCD_CH2_ALPHA           0x05C8
#define LCD_CH3_ALPHA           0x05CC
#define LCD_CH4_ALPHA           0x05D0

#define LCD_CH0_CLR_MATCH      	0x05E0
#define LCD_CH1_CLR_MATCH      	0x05E4
#define LCD_CH2_CLR_MATCH      	0x05E8
#define LCD_CH3_CLR_MATCH      	0x05EC
#define LCD_CH4_CLR_MATCH      	0x05F0

#define LCD_CONV0_CTL    	0x1000
#define LCD_CONV0_INT_STS	0x1028

#define LCD_MIXER0_TIM0		0x1004
#define LCD_MIXER0_TIM1  	0x1008

#define LCD_MIXERx_G_CTL        0x1014
#define LCD_MIXERx_G_DAT_RED    0x1018
#define LCD_MIXERx_G_DAT_GREEN  0x101C
#define LCD_MIXERx_G_DAT_BLUE   0x1020

#define LCD_FETCH_INT_STS0	0x0580
#define LCD_FETCH_INT_STS1      0x0584
#define LCD_GWIN_INT_STS        0x010c
#define LCD_GMIX_INT_STS        0x0110
#define LCD_WIN0_INT_STS        0x061c
#define LCD_MIXER0_INT_STS      0x0840
#define LCD_MIXER1_INT_STS      0x0940

/*  bits definition */

/* LCD_MIXERx_G_CTL bits */
#define LCD_MIXERx_G_CTL_ADD_PTR_R(n)   ((n)<<0)
#define LCD_MIXERx_G_CTL_ADD_PTR_G(n)   ((n)<<4)
#define LCD_MIXERx_G_CTL_ADD_PTR_B(n)   ((n)<<8)
#define LCD_MIXERx_G_CTL_Q4_C(n)        ((n)<<16)
#define LCD_MIXERx_G_CTL_Q3_C(n)        ((n)<<18)
#define LCD_MIXERx_G_CTL_Q2_C(n)        ((n)<<20)
#define LCD_MIXERx_G_CTL_Q1_C(n)        ((n)<<22)
#define LCD_MIXERx_G_CTL_G_VAL(n)       ((n)<<24)

#define LCD_CHx_ALPHA_CH_ALPHA(n)       ((n & 0xff)<<0)
#define LCD_CHx_ALPHA_CLR_KEY_EN(n)     ((n & 0x3)<<29)
#define LCD_CHx_ALPHA_W_ALPHA_EN        (1<<31)

/* ACCR1_REG dsi bits */
#define ACCR1_REG_DSI_BIT1_OFFSET   13
#define ACCR1_REG_DSI_BIT2_OFFSET   16
#define ACCR1_REG_DSI_BIT1(n)	    ((n)<<ACCR1_REG_DSI_BIT1_OFFSET)
#define ACCR1_REG_DSI_BIT2(n)	    ((n)<<ACCR1_REG_DSI_BIT2_OFFSET)
#define ACCR1_REG_DSI_BIT_MASK      (0x7)


/* LCD_CTL bits */
#define LCD_CTL_PCLK_DIV(n)	    ((n))
#define LCD_CTL_GFETCH_INT_EN   (0x1u<<15)
#define LCD_CTL_GWIN_INT_EN	(0x1u<<16)
#define LCD_CTL_GWIN_INT_EN	(0x1u<<16)
#define LCD_CTL_GMIX_INT_EN  	(0x1u<<17)
#define LCD_CTL_AXI32_EN    	(0x1u<<23)
#define LCD_CTL_LCD_QD_INT_EN	(0x1u<<27)
#define LCD_CTL_LCD_DIS_INT_EN	(0x1u<<28)
#define LCD_CTL_LCD_EN_INT_EN	(0x1u<<29)
#define LCD_CTL_LCD_QD      	(0x1u<<30)
#define LCD_CTL_LCD_EN      	(0x1u<<31)

/* LCD_CTL_STS bits */
#define LCD_CTL_STS_PCLK_DIV_STS (0x1ffu << 0)
#define LCD_CTL_STS_LCD_QD_STS   (0x1u << 27)
#define LCD_CTL_STS_LCD_EN_STS   (0x1u << 29)

/* LCD_CTL_INT_STS bits */
#define LCD_CTL_INT_STS_LCD_QD_INT_STS  (0x1u<<0)
#define LCD_CTL_INT_STS_LCD_DIS_INT_STS (0x1u<<1)
#define LCD_CTL_INT_STS_LCD_EN_INT_STS  (0x1u<<2)
#define LCD_CTL_INT_STS_GFETCH_INT_STS  (0x1u<<8)
#define LCD_CTL_INT_STS_GWIN_INT_STS    (0x1u<<9)
#define LCD_CTL_INT_STS_GMIX_INT_STS    (0x1u<<10)

/* LCD_FETCH_CTLx bits */
#define LCD_FETCH_CTLx_CHAN_EN            (0x1u<<0)
#define LCD_FETCH_CTLx_SRC_FOR(n)         ((n)<<2)
#define LCD_FETCH_CTLx_END_FR_INT_EN      (0x1u<<6)
#define LCD_FETCH_CTLx_START_FR_INT_EN    (0x1u<<7)
#define LCD_FETCH_CTLx_BUS_ERR_INT_EN     (0x1u<<8)
#define LCD_FETCH_CTLx_SP_MODE            (0x1u<<10)
#define LCD_FETCH_CTLx_SP_MODIFIED_FRAME  (0x1u<<11)

/* LCD_FETCH_STSx bits */
#define LCD_FETCH_STSx_CHAN_STS           (0x1u<<0)
#define LCD_FETCH_STSx_SRC_FOR_STS        (0x7u<<2)

/* LCD_NXT_DESC_ADDRx */
#define LCD_NXT_DESC_ADDRx_NXT_DESC_ADDR(n)	    ((n)<<4)

/* LCD_FR_IDx */
#define LCD_FR_IDx_FR_ID(n) ((n))

/* LCD_FR_ADDRx */
#define LCD_FR_ADDRx_FR_ADDR(n) ((n)<<4)

/* LCD_FETCH_INT_STS0 bits */
#define LCD_FETCH_INT_STS0_START_FR0_INT_STS (0x1u<<0)
#define LCD_FETCH_INT_STS0_START_FR1_INT_STS (0x1u<<1)
#define LCD_FETCH_INT_STS0_START_FR2_INT_STS (0x1u<<2)
#define LCD_FETCH_INT_STS0_START_FR3_INT_STS (0x1u<<3)
#define LCD_FETCH_INT_STS0_START_FR4_INT_STS (0x1u<<4)
#define LCD_FETCH_INT_STS0_START_FR5_INT_STS (0x1u<<5)
#define LCD_FETCH_INT_STS0_START_FR6_INT_STS (0x1u<<6)
#define LCD_FETCH_INT_STS0_START_FR7_INT_STS (0x1u<<7)
#define LCD_FETCH_INT_STS0_END_FR0_INT_STS   (0x1u<<16)
#define LCD_FETCH_INT_STS0_END_FR1_INT_STS   (0x1u<<17)
#define LCD_FETCH_INT_STS0_END_FR2_INT_STS   (0x1u<<18)
#define LCD_FETCH_INT_STS0_END_FR3_INT_STS   (0x1u<<19)
#define LCD_FETCH_INT_STS0_END_FR4_INT_STS   (0x1u<<20)
#define LCD_FETCH_INT_STS0_END_FR5_INT_STS   (0x1u<<21)
#define LCD_FETCH_INT_STS0_END_FR6_INT_STS   (0x1u<<22)
#define LCD_FETCH_INT_STS0_END_FR7_INT_STS   (0x1u<<23)
#define LCD_FETCH_INT_STS0_OVERFLOW_INT_STS  (0x1u<<31)

/* LCD_WINx_CTL bits */
#define LCD_WINx_CTL_WIN_EN_STS      (0x1u<<0)
#define LCD_WINx_CTL_WIN_URUN_INT_EN (0x1u<<1)
#define LCD_WINx_CTL_WIN_EOF_INT_EN  (0x1u<<2)
#define LCD_WINx_CTL_WIN_SOF_INT_EN  (0x1u<<3)
#define LCD_WINx_CTL_WIN_XRES(n)     ((n)<<8)
#define LCD_WINx_CTL_WIN_YRES(n)     ((n)<<20)

/* LCD_WINx_CFG bits */
#define LCD_WINx_CFG_WIN_XSCALE(n)	((n)<<19)
#define LCD_WINx_CFG_WIN_YSCALE(n)      ((n)<<13)


/* LCD_MIXERx_CTL0 bits */
#define LCD_MIXERx_CTL0_MIX_EN           (0x1u<<0)
#define LCD_MIXERx_CTL0_BP_COLOR_USE     (0x1u<<1)
#define LCD_MIXERx_CTL0_BP_ID_MASK       (0x1fu<<2)
#define LCD_MIXERx_CTL0_BP_ID(n)         ((n) <<2)
#define LCD_MIXERx_CTL0_OL1_COLOR_USE    (0x1u<<7)
#define LCD_MIXERx_CTL0_OL1_ID_MASK      (0x1fu<<8)
#define LCD_MIXERx_CTL0_OL1_ID(n)        ((n) <<8)
#define LCD_MIXERx_CTL0_OL2_COLOR_USE    (0x1u<<13)
#define LCD_MIXERx_CTL0_OL2_ID_MASK      (0x1fu<<14)
#define LCD_MIXERx_CTL0_OL2_ID(n)        ((n) <<14)
#define LCD_MIXERx_CTL0_OL3_COLOR_USE    (0x1u<<19)
#define LCD_MIXERx_CTL0_OL3_ID_MASK      (0x1fu<<20)
#define LCD_MIXERx_CTL0_OL3_ID(n)        ((n) <<20)
#define LCD_MIXERx_CTL0_CRSR_ID_MASK     (0x1fu<<26)
#define LCD_MIXERx_CTL0_CRSR_ID(n)       ((n) <<26)
#define LCD_MIXERx_CTL0_DISP_UPDATE      (0x1u<<31)

/* LCD_MIXERx_CTL1 bits */
#define LCD_MIXERx_CTL1_DISP_EN_INT_EN     (0x1u<<0)
#define LCD_MIXERx_CTL1_DISP_UPDATE_INT_EN (0x1u<<1)
#define LCD_MIXERx_CTL1_BLINK_INT_EN       (0x1u<<2)
#define LCD_MIXERx_CTL1_DISP_XRES(n)       ((n) <<5)
#define LCD_MIXERx_CTL1_DISP_YRES(n)       ((n) <<17)

/* LCD_MIXERx_CTL2 bits */
#define LCD_MIXERx_CTL2_CONV_ID(n)       ((n) <<26)

/* LCD_MIXERx_CFG0 bits */
#define LCD_MIXERx_CFG0_XPOS(n)       ((n) <<0)
#define LCD_MIXERx_CFG0_YPOS(n)       ((n) <<12)

/* LCD_MIXERx_CFG1 bits */
#define LCD_MIXERx_CFG1_XRES(n)       ((n) <<0)
#define LCD_MIXERx_CFG1_YRES(n)       ((n) <<12)

/* LCD_MIXERx_STS0 bits */
#define LCD_MIXERx_STS0_MIX_EN_STS    (0x1u<<0)

/* LCD_CONVx_CTL bits */
#define LCD_CONVx_CTL_DISP_URUN_INT_EN (0x1u<<0)
#define LCD_CONVx_CTL_DISP_EOF_INT_EN  (0x1u<<1)
#define LCD_CONVx_CTL_DISP_SOF_INT_EN  (0x1u<<2)
#define LCD_CONVx_CTL_SMART_INT_EN     (0x1u<<3)
#define LCD_CONVx_CTL_SP_RD_INT_EN     (0x1u<<4)
#define LCD_CONVx_CTL_SP_UNDRUN_INT_EN (0x1u<<5)
#define LCD_CONVx_CTL_SP_BREAK_INT_EN  (0x1u<<7)
#define LCD_CONVx_CTL_FLOW_DIS         (0x1u<<8)
#define LCD_CONVx_CTL_BREAK            (0x1u<<9)
#define LCD_CONVx_CTL_CONV_EN          (0x1u<<10)
#define LCD_CONVx_CTL_TV_FOR           (0x1u<<15)
#define LCD_CONVx_CTL_OP_FOR(n)        ((n) <<16)
#define LCD_CONVx_CTL_END_BIT          (0x1u<<19)
#define LCD_CONVx_CTL_SYNC_CNT(n)      ((n) <<20)
#define LCD_CONVx_CTL_GPIO0_CTL        (0x1u<<28)
#define LCD_CONVx_CTL_GPIO1_CTL        (0x1u<<29)
#define LCD_CONVx_CTL_DISP_TYPE(n)     ((n) <<30)

/* LCD_MIXERx_TIM0 bits */
#define LCD_MIXERx_TIM0_PCP           (0x1u<<0)
#define LCD_MIXERx_TIM0_OEP           (0x1u<<1)
#define LCD_MIXERx_TIM0_VSP           (0x1u<<2)
#define LCD_MIXERx_TIM0_HSP           (0x1u<<3)
#define LCD_MIXERx_TIM0_VSYNC_OUT     (0x1u<<4)
#define LCD_MIXERx_TIM0_MODE          (0x1u<<5)
#define LCD_MIXERx_TIM0_HSW(n)        ((n) <<8)
#define LCD_MIXERx_TIM0_ELW(n)        ((n) <<16)
#define LCD_MIXERx_TIM0_BLW(n)        ((n) <<24)

/* LCD_MIXERx_TIM1 bits */
#define LCD_MIXERx_TIM1_VSW(n)        ((n)<<8)
#define LCD_MIXERx_TIM1_EFW(n)        ((n)<<16)
#define LCD_MIXERx_TIM1_BFW(n)        ((n)<<24)

/* LCD_MIXERx_TIM2 bits */
#define LCD_MIXERx_TIM2_GSUP(n)       ((n)<<0)
#define LCD_MIXERx_TIM2_SPW(n)        ((n)<<8)
#define LCD_MIXERx_TIM2_SPSH(n)       ((n)<<13)
#define LCD_MIXERx_TIM2_GSW(n)        ((n)<<24)

/* LCD_MIXERx_TIM3 bits */
#define LCD_MIXERx_TIM3_REVDEL(n)     ((n) <<0)
#define LCD_MIXERx_TIM3_GSP           (0x1u<<8)
#define LCD_MIXERx_TIM3_SPP           (0x1u<<9)
#define LCD_MIXERx_TIM3_PSP           (0x1u<<10)
#define LCD_MIXERx_TIM3_PSH(n)        ((n) <<24)

/* LCD_MIXERx_TIM4 bits */
#define LCD_MIXERx_TIM4_TSH(n)        ((n)<<0)

/*LCD DSI Dx register offsets*/
#define LCD_DSI_DxSCR0_OFFSET		0x0000
#define LCD_DSI_DxSCR1_OFFSET          	0x0100
#define LCD_DSI_DxSSR_OFFSET		0x0104
#define LCD_DSI_DxTRIG_OFFSET          	0x0108
#define LCD_DSI_DxINEN_OFFSET		0x010C
#define LCD_DSI_DxINST1_OFFSET         	0x0110
#define LCD_DSI_DxTEINTCNT_OFFSET	0x0114
#define LCD_DSI_Dx_G_CTL_OFFSET		0x0014
#define LCD_DSI_Dx_G_DAT_RED_OFFSET	0x0018
#define LCD_DSI_Dx_G_DAT_GREEN_OFFSET  	0x001C
#define LCD_DSI_Dx_G_DAT_BLUE_OFFSET   	0x0020
#define LCD_DSI_DxPRSR_OFFSET          	0x0024
#define LCD_DSI_DxINST0_OFFSET		0x0028
#define LCD_DSI_DxADAT_OFFSET		0x0128
#define LCD_DSI_DxCFIF_OFFSET		0x002C
#define LCD_DSI_DxTIM0_OFFSET		0x0130
#define LCD_DSI_DxTIM1_OFFSET		0x0134
#define LCD_DSI_DxPHY_TIM0_OFFSET	0x013C
#define LCD_DSI_DxPHY_TIM1_OFFSET	0x0140
#define LCD_DSI_DxPHY_TIM2_OFFSET	0x0144
#define LCD_DSI_DxPHY_CAL_OFFSET	0x0148

/* LCD_DSI_DxSCR0 bits */
#define LCD_DSI_DxSCR0_DISP_URUN_INT_EN (0x1u<<0)
#define LCD_DSI_DxSCR0_DISP_EOF_INT_EN  (0x1u<<1)
#define LCD_DSI_DxSCR0_DISP_SOF_INT_EN  (0x1u<<2)
#define LCD_DSI_DxSCR0_SMART_INT_EN     (0x1u<<3)
#define LCD_DSI_DxSCR0_SP_RD_INT_EN     (0x1u<<4)
#define LCD_DSI_DxSCR0_SP_UNDRUN_INT_EN (0x1u<<5)
#define LCD_DSI_DxSCR0_SP_BREAK_INT_EN  (0x1u<<7)
#define LCD_DSI_DxSCR0_FLOW_DIS         (0x1u<<8)
#define LCD_DSI_DxSCR0_BREAK            (0x1u<<9)
#define LCD_DSI_DxSCR0_CONV_EN          (0x1u<<10)
#define LCD_DSI_DxSCR0_PRIM_VC          (0x1u<<11)
#define LCD_DSI_DxSCR0_SYNC_CNT(n)      ((n) <<20)

/* LCD_DSI_DxSCR1 bits */
#define LCD_DSI_DxSCR1_DSI_EN     (0x1u<<0)
#define LCD_DSI_DxSCR1_NOL_MASK   (0x3u<<2)
#define LCD_DSI_DxSCR1_NOL(n)     ((n) <<2)
#define LCD_DSI_DxSCR1_PT         (0x1u<<4)
#define LCD_DSI_DxSCR1_VC0(n)     ((n) <<6)
#define LCD_DSI_DxSCR1_VC1(n)     ((n) <<8)
#define LCD_DSI_DxSCR1_VC_EN      (0x1u<<10)
#define LCD_DSI_DxSCR1_VSYNC_MODE (0x1u<<12)
#define LCD_DSI_DxSCR1_POWER_MASK (0x119u<<23)
#define LCD_DSI_DxSCR1_RIHS(n)    ((n) <<23)
#define LCD_DSI_DxSCR1_TRIG       (0x1u<<24)
#define LCD_DSI_DxSCR1_ULPS(n)    ((n) <<26)
#define LCD_DSI_DxSCR1_LPDT(n)    ((n) <<27)
#define LCD_DSI_DxSCR1_BTA(n)     ((n) <<31)

/* LCD_DSI_DxSSR bits */
#define LCD_DSI_DxSSR_DSI_EN_STAT (0x1u<<0)
#define LCD_DSI_DxSSR_L0_PM       (0x1u<<24)
#define LCD_DSI_DxSSR_L1_PM       (0x1u<<25)
#define LCD_DSI_DxSSR_L2_PM       (0x1u<<26)
#define LCD_DSI_DxSSR_L3_PM       (0x1u<<27)
#define LCD_DSI_DxSSR_STOP_ST     (0x1u<<28)
#define LCD_DSI_DxSSR_DIR         (0x1u<<31)

/* LCD_DSI_DxTIM0 bits */
#define LCD_DSI_DxTIM0_HS_TX_TO(n)   ((n)<<0)
#define LCD_DSI_DxTIM0_LP_RX_TO(n)   ((n)<<16)

/* LCD_DSI_DxTIM1 bits */
#define LCD_DSI_DxTIM1_BTA_TO(n)    ((n)<<0)
#define LCD_DSI_DxTIM1_G_TIMER(n)   ((n)<<16)

/* LCD_DSI_DxPHY_TIM0 bits */
#define LCD_DSI_DxPHY_TIM0_HSTRAIL(n)   ((n)<<0)
#define LCD_DSI_DxPHY_TIM0_HSZERO(n)    ((n)<<8)
#define LCD_DSI_DxPHY_TIM0_HSPREP(n)    ((n)<<16)
#define LCD_DSI_DxPHY_TIM0_LPX(n)       ((n)<<24)

/* LCD_DSI_DxPHY_TIM1 bits */
#define LCD_DSI_DxPHY_TIM1_CLTRAIL(n)   ((n)<<0)
#define LCD_DSI_DxPHY_TIM1_CLZERO(n)    ((n)<<8)
#define LCD_DSI_DxPHY_TIM1_TAGO(n)      ((n)<<16)
#define LCD_DSI_DxPHY_TIM1_HSEXIT(n)    ((n)<<24)

/* LCD_DSI_DxPHY_TIM2 bits */
#define LCD_DSI_DxPHY_TIM2_WAKEUP(n)   ((n)<<0)
#define LCD_DSI_DxPHY_TIM2_TAGET(n)    ((n)<<16)

/* LCD_CONVx_G_CTL bits */
#define LCD_CONVx_G_CTL_ADD_PTR_R(n)   ((n)<<0)
#define LCD_CONVx_G_CTL_ADD_PTR_G(n)   ((n)<<4)
#define LCD_CONVx_G_CTL_ADD_PTR_B(n)   ((n)<<8)
#define LCD_CONVx_G_CTL_Q4_C(n)        ((n)<<16)
#define LCD_CONVx_G_CTL_Q3_C(n)        ((n)<<18)
#define LCD_CONVx_G_CTL_Q2_C(n)        ((n)<<20)
#define LCD_CONVx_G_CTL_Q1_C(n)        ((n)<<22)
#define LCD_CONVx_G_CTL_G_VAL(n)       ((n)<<24)

/* LCD_DSI_COMMAND bits */
#define LCD_DSI_COMMAND_DATA(n)  ((n)<<0)
#define LCD_DSI_COMMAND_CMD(n)   ((n)<<8)
#define LCD_DSI_COMMAND_LOOP(n)  ((n)<<13)
#define LCD_DSI_COMMAND_VC(n)    ((n) <<14)

/* LCD_DSI_DxINST0 bits */
#define LCD_DSI_DxINST0_SP_GEN_INT_STS    (0x1u<<6)
#define LCD_DSI_DxINST0_BREAK_INT_STS     (0x1u<<7)

//LCD_DSI_DxINST1 bits
#define LCD_DSI_DxINST1_BTA_TO_ERR    (0x1u<<14)
#define LCD_DSI_DxINST1_TRIG_ST       (0x1u<<19)

/*MCCU_DSI_REG1 bits*/
#define MCCU_DSI_REG1_DSIx_PLL_LOCK_SEL   (0x1u<<0)
#define MCCU_DSI_REG1_DSIx_PLL_ON         (0x1u<<1)

/*MCCU_DSI_REG2 bits*/
#define MCCU_DSI_REG2_CK_SEL(n)    ((n)<<0)
#define MCCU_DSI_REG2_OUTDIV(n)    ((n)<<13)
#define MCCU_DSI_REG2_FBDIV(n)     ((n)<<17)
#define MCCU_DSI_REG2_REFDIV(n)    ((n)<<26)
#define MCCU_DSI_REG2_DEFAULT      0x178A


typedef struct DisplayControllerFetchRegisters {
	u32    LCD_FETCH_CTLx;   	/* Fetch unit control register  */
	u32    LCD_FETCH_STSx;     	/* Fetch unit status register */
} DisplayControllerFetchRegisters;

typedef struct DisplayControllerFrameDescriptor {
	u32    LCD_NXT_DESC_ADDRx;   /* Next Descriptor Address Register */
	u32    LCD_FR_ADDRx;
	u32    LCD_FR_IDx;           /* Frame ID Register */
	u32    LCD_CH_CMDx;
} DisplayControllerFrameDescriptor;

typedef struct DisplayControllerMixerRegisters {
	u32    LCD_MIXERx_CTL0;   	/* Mixer Control 0 Register  */
	u32    LCD_MIXERx_CTL1;     	/* Mixer Control 1 Register */
	u32    LCD_MIXERx_BP_CFG0;   /* Base Plane Configuration Register 0  */
	u32    LCD_MIXERx_BP_CFG1;   /* Base Plane Configuration Register 1 */
	u32    LCD_MIXERx_OL1_CFG0;  /* Overlay 1 Configuration Register 0 */
	u32    LCD_MIXERx_OL1_CFG1;  /* Overlay 1 Configuration Register 1 */
	u32    LCD_MIXERx_OL2_CFG0;  /* Overlay 2 Configuration Register 0 */
	u32    LCD_MIXERx_OL2_CFG1;  /* Overlay 2 Configuration Register 1 */
	u32    LCD_MIXERx_OL3_CFG0;  /* Overlay 3 Configuration Register 0 */
	u32    LCD_MIXERx_OL3_CFG1;  /* Overlay 3 Configuration Register 1 */
	u32    LCD_MIXERx_CRSR_CFG;  /* Cursor Configuration Register */
	u32    LCD_MIXERx_STS0;      /* Mixer Status Register 0 */
	u32    LCD_MIXERx_BP_CLR;   	/* Base Plane Color Register */
	u32    LCD_MIXERx_OL1_CLR;   /* Overlay 1 Color Register */
	u32    LCD_MIXERx_OL2_CLR;   /* Overlay 2 Color Register */
	u32    LCD_MIXERx_OL3_CLR;   /* Overlay 3 Color Register */
	u32    LCD_MIXERx_INT_STS;   /* Mixer Interrupt Status Register  */
	u32    LCD_MIXERx_CTL2;      /* Mixer Control 2 Register */
	u32    LCD_MIXERx_STS1;      /* Mixer Status Register 1 */

} DisplayControllerMixerRegisters;

typedef struct DisplayControllerConverterRegisters {
	u32    LCD_CONVx_CTL;     	 /* Converter Control register */
	u32    LCD_MIXERx_TIM0;     	 /* Timing Control 0 Register  */
	u32    LCD_MIXERx_TIM1;       /* Timing Control 1 Register  */
	u32    LCD_MIXERx_TIM2;       /* Timing Control 2 Register  */
	u32    LCD_MIXERx_TIM3;       /* Timing Control 3 Register  */
	u32    LCD_CONVx_G_CTL;       /* Converter Gamma Correction Control Register  */
	u32    LCD_CONVx_G_DAT_RED;   /* Gamma Correction Register - Red */
	u32    LCD_CONVx_G_DAT_GREEN; /* Gamma Correction Register - Green */
	u32    LCD_CONVx_G_DAT_BLUE;  /* Gamma Correction Register - Blue */
	u32    LCD_MIXERx_PRSR;       /* Smart Panel Control Register */
	u32    LCD_CONVx_INT_STS;     /* Converter Interrupt Status Register */
	u32    LCD_CONVx_SCFIF;       /* LCD Smart Panel Command FIFO Register */
	u32    LCD_MIXERx_TIM4;       /* Timing Control 4 Register  */
} DisplayControllerConverterRegisters;

typedef struct DisplayControllerDSI0Registers {
	u32    LCD_DSI_DxSCR0;     	 /* DSI Converter Control Register 0 */
	u32    RESERVE0;          	 /* Reserve0 Register  */
	u32    RESERVE1;          	 /* Reserve1 Register  */
	u32    RESERVE2;          	 /* Reserve2 Register  */
	u32    RESERVE3;          	 /* Reserve3 Register  */
	u32    LCD_CONVx_G_CTL;       /* DSI Gamma Correction Control Register */
	u32    LCD_CONVx_G_DAT_RED;   /* DSI Gamma Correction Register - Red */
	u32    LCD_CONVx_G_DAT_GREEN; /* DSI Gamma Correction Register - Green */
	u32    LCD_CONVx_G_DAT_BLUE;  /* DSI Gamma Correction Register - Blue */
	u32    LCD_DSI_DxPRSR;        /* DSI Panel Read Status Register */
	u32    LCD_DSI_DxINST0;       /* DSI Interrupt Status Register 0 */
	u32    LCD_DSI_DxCFIF;        /* DSI Command FIFO Register */
} DisplayControllerDSI0Registers;

typedef struct DisplayControllerDSI1Registers {
	u32    LCD_DSI_DxSCR1;     	 /* DSI Control Register 1 */
	u32    LCD_DSI_DxSSR;      	 /* DSI Status Register */
	u32    LCD_DSI_DxTRIG;      	 /* DSI Trigger Register */
	u32    LCD_DSI_DxINEN;      	 /* DSI Interrupt Enable Register */
	u32    LCD_DSI_DxINST1;     	 /* DSI Interrupt Status Register 1 */
	u32    LCD_DSI_DxTEINTCNT;    /* DSI TE Interrupt Count */
	u32    RESERVE0;          	 /* Reserve0 Register  */
	u32    RESERVE1;          	 /* Reserve1 Register  */
	u32    RESERVE2;          	 /* Reserve2 Register  */
	u32    RESERVE3;          	 /* Reserve3 Register  */
	u32    LCD_DSI_DxADAT;        /* DSI ACK_DATA Register */
	u32    RESERVE4;          	 /* Reserve4 Register  */
	u32    LCD_DSI_DxTIM0;        /* DSI TIMER Register 0 */
	u32    LCD_DSI_DxTIM1;        /* DSI TIMER Register 1 */
	u32    LCD_DSI_DxTEST;        /* DSI Test Register */
	u32    LCD_DSI_DxPHY_TIM0;    /* DSI-DPHY TIMER Register 0 */
	u32    LCD_DSI_DxPHY_TIM1;    /* DSI-DPHY TIMER Register 1 */
	u32    LCD_DSI_DxPHY_TIM2;    /* DSI-DPHY TIMER Register 2 */
	u32    LCD_DSI_DSxDPHY_CAL;   /* DSI PHY Calibration */
} DisplayControllerDSI1Registers;


typedef struct DisplayControllerWindowRegisters {
	u32     LCD_WINx_CTL;         /* Window Control Register */
	u32     LCD_WINx_CFG;         /* Window Configuration Register */
	u32     LCD_CHx_ALPHA;
	u32     LCD_WINx_CROP0;
	u32     LCD_WINx_CROP1;
	u32     LCD_WINx_SCALE_PTR;
	u32     LCD_WINx_SCALE_RW;
} DisplayControllerWindowRegisters;

typedef struct pxa950fb_registers {
	DisplayControllerFetchRegisters		fetchregs[7];
	DisplayControllerFrameDescriptor	descregs[7];
	DisplayControllerMixerRegisters		mixerregs[2];
	DisplayControllerWindowRegisters	wgregs[5];
	DisplayControllerConverterRegisters	convregs;
	DisplayControllerDSI0Registers		dsi0regs;
	DisplayControllerDSI1Registers		dsi1regs;
} pxa950fb_registers;


struct MCCUDSIRegisters {
	u32    MCCU_DSI_REG1;
	u32    MCCU_DSI_REG2;
	u32    MCCU_DSI_REG3;
};


#define MCCU_DSI_BASE_ADDRESS 0x42404154L
#define MCCU_DSI_REGS_MEM_SIZE	0x18
#define MCCU_DSI_REG1_OFFSET	0x0
#define MCCU_DSI_REG2_OFFSET	0x4

#define BYTES_PER_PIXEL 2
#define COMMAND_DMA_BUFFER  0x82000000L

#define DMA_ADDRESS_SPACE 0x40
#define WGE_ADDRESS_SPACE 0x40
#define CONVERTER_ADDRESS_SPACE 0x1000
#define ALPHA_ADDRESS_SPACE 0x4
#define MIXER_ADDRESS_SPACE 0x100
#define MCCU_ADDRESS_SPACE  0x10

#define BP_DMA_CHANNEL 0
#define OL1_DMA_CHANNEL 1

#define DSI_CONTROLLER_CLOCK 104  //104Mhz - TBD

#define DISPLAY_CONTROLLER_BASE_ADDRESS      0x44100000L
#define DISPLAY_CONTROLLER_STATUS_ADDRESS    0x44100100L
#define DISPLAY_CONTROLLER_FETCH_ADDRESS     0x44100200L
#define DISPLAY_CONTROLLER_FRAME_ADDRESS     0x44100250L
#define DISPLAY_CONTROLLER_WINDOW_ADDRESS    0x44100600L
#define DISPLAY_CONTROLLER_MIXER_ADDRESS     0x44100800L
#define DISPLAY_CONTROLLER_CONVERTER_ADDRESS 0x44101000L
#define DISPLAY_CONTROLLER_DSI0_ADDRESS      0x44102000L
#define DISPLAY_CONTROLLER_DSI0_CFIF_ADDRESS 0x4410202CL
#define DISPLAY_CONTROLLER_DSI1_ADDRESS      0x44102100L
#define DISPLAY_CONTROLLER_CHx_ALPHA_REG     0x441005C0L


#endif

