/*
 * include/linux/maxtouch_obp_ts.h - platform/protocol data for Atmel maxTouch IC
 *
 * Copyright (C) 2009 Borqs Beijing Ltd.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *
 */

#ifndef _LINUX_MAXTOUCH_OBP_TS_H
#define _LINUX_MAXTOUCH_OBP_TS_H

#define MAXTOUCH_TS_NAME "maxtouch-obp-ts"

#define MTS_OBP_ID_INFO_ADDR		0

enum {
	MTS_OBJ_RESERVED0		= 0,
	MTS_OBJ_RESERVED1		= 1,
	MTS_OBJ_DBG_DELTAS		= 2,
	MTS_OBJ_DBG_REFS		= 3,
	MTS_OBJ_DBG_SIGS		= 4,
	MTS_OBJ_GEN_MSG_PROC		= 5,
	MTS_OBJ_GEN_CMD_PROC		= 6,
	MTS_OBJ_GEN_PWR_CONF		= 7,
	MTS_OBJ_GEN_ACQUIRE_CONF	= 8,
	MTS_OBJ_TOUCH_MULTI		= 9,
	MTS_OBJ_TOUCH_SINGLE		= 10,
	MTS_OBJ_TOUCH_XSLIDER		= 11,
	MTS_OBJ_TOUCH_SLIDER		= 12,
	MTS_OBJ_TOUCH_XWHEEL		= 13,
	MTS_OBJ_TOUCH_YWHEEL		= 14,
	MTS_OBJ_TOUCH_KEYARRAY		= 15,
	MTS_OBJ_PROCG_SIG_FILTER	= 16,
	MTS_OBJ_PROCI_LINEAR_TBL	= 17,
	MTS_OBJ_SPT_COM_CONFIG		= 18,
	MTS_OBJ_SPT_GPIO_PWM		= 19,
	MTS_OBJ_PROCI_GRIPFACESUPPRESSION = 20,
	MTS_OBJ_RESERVED3		= 21,
	MTS_OBJ_PROCG_NOISE_SUPPRESSION	= 22,
	MTS_OBJ_TOUCH_PROXIMITY		= 23,
	MTS_OBJ_PROCI_ONE_TOUCH_GESTURE_PROC = 24,
	MTS_OBJ_SPT_SELF_TEST		= 25,
	MTS_OBJ_DEBUG_CTE_RANGE		= 26,
	MTS_OBJ_PROCI_TWO_TOUCH_GESTURE_PROC = 27,
	MTS_OBJ_SPT_CTE_CONFIG		= 28,
	MTS_OBJ_NOISESUPPRESSION_1  = 36,
	MTS_OBJ_DEBUG_DIAGNOSTIC        = 37,
	MTS_OBJ_CPT_USERDATA            = 38,

	/* Max number of objects currently defined */
	MTS_OBP_MAX_OBJECT_NUM = MTS_OBJ_CPT_USERDATA + 1,
};

/* OBP structures as defined by the wire protocol. */

/* Note: Not all the structures below need an explicit packed attribute since
 * many of them just contain uint8_t's. However, the protocol is defined in
 * such a way that the structures may expand in the future, with
 * potential multi-byte fields. Thus, we will mark them all as packed to
 * minimize silly bugs in the future.
 */

/* part of the info block */
struct mts_id_info {
	uint8_t			family_id;
	uint8_t			variant_id;
	uint8_t			version;
	uint8_t			build;
	uint8_t			matrix_x_size;
	uint8_t			matrix_y_size;
	uint8_t			num_objs;
} __attribute__ ((packed));

/* an entry in the ote table */
struct mts_obj_entry {
	uint8_t			type;
	uint16_t		addr;
	uint8_t			size;
	uint8_t			num_inst;
	uint8_t			num_rids;
} __attribute__ ((packed));


/*******************************/
/*********** messages **********/
/*******************************/

/* generic message received from the message_processor object. size/buffer
 * defined at runtime after reading the info block */
struct mts_obj_message {
	uint8_t			report_id;
	uint8_t			msg[0];
} __attribute__ ((packed));

/* status message sent by the command processor - T6 */
#define MTS_CMD_PROC_STATUS_RESET	(1 << 7)
#define MTS_CMD_PROC_STATUS_OFL		(1 << 6)
#define MTS_CMD_PROC_STATUS_SIGERR	(1 << 5)
#define MTS_CMD_PROC_STATUS_CAL		(1 << 4)
#define MTS_CMD_PROC_STATUS_CFGERR	(1 << 3)
struct mts_cmd_proc_msg {
	uint8_t			report_id;
	uint8_t			status;
	uint16_t		checksum;
} __attribute__ ((packed));

/* status message sent by the mutlitouch touch object - T9*/
#define MTS_TOUCH_MULTI_STATUS_TOUCH		(1 << 7)
#define MTS_TOUCH_MULTI_STATUS_PRESS		(1 << 6)
#define MTS_TOUCH_MULTI_STATUS_RELEASE		(1 << 5)
#define MTS_TOUCH_MULTI_STATUS_MOVE		(1 << 4)
#define MTS_TOUCH_MULTI_STATUS_VECTOR		(1 << 3)
#define MTS_TOUCH_MULTI_STATUS_AMPLITUDE	(1 << 2)
struct mts_touch_multi_msg {
	uint8_t			report_id;
	uint8_t			status;
	uint8_t			xpos_msb;
	uint8_t			ypos_msb;
	uint8_t			xypos_lsb;
	uint8_t			touch_area;
	uint8_t			touch_amp;
	uint8_t			touch_vect;
} __attribute__ ((packed));

/* status message sent by the keyarray touch object - T15 */
#define MTS_TOUCH_KEYARRAY_STATUS_TOUCH		(1 << 7)
struct mts_touch_keyarray_msg {
	uint8_t			report_id;
	uint8_t			status;
	uint32_t		keystate;
} __attribute__ ((packed));



/*******************************/
/**** configuration objects ****/
/*******************************/

/* GEN_COMMANDPROCESSOR_T6 */
struct mts_gen_cmd_proc {
	uint8_t			reset;
	uint8_t			backupnv;
	uint8_t			calibrate;
	uint8_t			reportall;
	uint8_t			reserved;
	uint8_t			debugctrl;
} __attribute__ ((packed));

/* GEN_POWERCONFIG_T7 */
struct mts_gen_power_cfg {
	uint8_t			idle_acq_int;      /* in ms */
	uint8_t			active_acq_int;    /* in ms */
	uint8_t			active_idle_to;    /* in 200ms */
} __attribute__ ((packed));

/* GEN_ACQUIRECONFIG_T8 */
struct mts_gen_acquire_cfg {
	uint8_t			charge_time;       /* in 250ns */
	uint8_t			reserve0;          /* Reserved */
	uint8_t			touch_drift;       /* in 200ms */
	uint8_t			drift_susp;        /* in 200ms */
	uint8_t			touch_autocal;     /* in 200ms */
	uint8_t			sync;
	uint8_t			anti_cal_susp;
	uint8_t			anti_cal_sthr;
} __attribute__ ((packed));

/* TOUCH_MULTITOUCHSCREEN_T9 */
struct mts_touch_multi_cfg {
	uint8_t			ctrl;
	uint8_t			x_origin;
	uint8_t			y_origin;
	uint8_t			x_size;
	uint8_t			y_size;
	uint8_t			aks_cfg;
	uint8_t			burst_len;
	uint8_t			tch_det_thr;
	uint8_t			tch_det_int;
	uint8_t			orient;
	uint8_t			mrg_to;
	uint8_t			mov_hyst_init;
	uint8_t			mov_hyst_next;
	uint8_t			mov_filter;
	uint8_t			num_touch;
	uint8_t			merge_hyst;
	uint8_t			merge_thresh;
	uint8_t			amp_hyst;
	uint16_t		x_res;
	uint16_t		y_res;
	uint8_t			x_low_clip;
	uint8_t			x_high_clip;
	uint8_t			y_low_clip;
	uint8_t			y_high_clip;
	uint8_t			x_edge_ori;
	uint8_t			x_edge_cdist;
	uint8_t			y_edge_ori;;
	uint8_t			y_edge_cdist;
} __attribute__ ((packed));

/* TOUCH_KEYARRAY_T15 */
struct mts_touch_keyarray_cfg {
	uint8_t			ctrl;
	uint8_t			x_origin;
	uint8_t			y_origin;
	uint8_t			x_size;
	uint8_t			y_size;
	uint8_t			aks_cfg;
	uint8_t			burst_len;
	uint8_t			tch_det_thr;
	uint8_t			tch_det_int;
	uint8_t			rsvd1;
	uint8_t			rsvd2;
} __attribute__ ((packed));

/* PROCG_SIGNALFILTER_T16 */
struct mts_procg_sig_filter_cfg {
	uint8_t			slew;
	uint8_t			median;
	uint8_t			iir;
} __attribute__ ((packed));

/* PROCI_LINEARIZATIONTABLE_T17 */
struct mts_proci_linear_tbl_cfg {
	uint8_t			ctrl;
	uint16_t		x_offset;
	uint8_t			x_segment[16];
	uint16_t		y_offset;
	uint8_t			y_segment[16];
} __attribute__ ((packed));

/* SPT_COM_CONFIG_T18*/
struct mts_spt_com_cfg {
	uint8_t			ctrl;
	uint8_t			cmd;
} __attribute__ ((packed));

/* SPT_GPIOPWM_T19*/
struct mts_spt_gpio_pwm_cfg {
	uint8_t			ctrl;
	uint8_t			report_mask;
	uint8_t			pin_direction;
	uint8_t			internal_pullup;
	uint8_t			output_value;
	uint8_t			wake_on_change;
	uint8_t			pwm_enable;
	uint8_t			pwm_period;
	uint8_t			duty_cycle_0;
	uint8_t			duty_cycle_1;
	uint8_t			duty_cycle_2;
	uint8_t			duty_cycle_3;
	uint8_t			trigger_0;
	uint8_t			trigger_1;
	uint8_t			trigger_2;
	uint8_t			trigger_3;
} __attribute__ ((packed));

/* PROCI_GRIPFACESUPPRESSION_T20 */
struct mts_proci_grip_suppression_cfg {
	uint8_t			ctrl;
	uint8_t			xlogrip;
	uint8_t			xhigrip;
	uint8_t			ylogrip;
	uint8_t			yhigrip;
	uint8_t			maxtchs;
	uint8_t			reserve0;
	uint8_t			szthr1;
	uint8_t			szthr2;
	uint8_t			shpthr1;
	uint8_t			shpthr2;
	uint8_t			supextto;
} __attribute__ ((packed));

/* PROCG_NOISESUPPRESSION_T22 */
struct mts_procg_noise_suppression_cfg {
	uint8_t			ctrl;
	uint16_t		reserve0;
	uint16_t		gcaf_upper_limit;
	uint16_t		gcaf_lower_limit;
	uint8_t			gcaf_num_active;
	uint8_t			noise_threshold;
	uint8_t			reserve1;
	uint8_t			freq_hop_scale;
	uint8_t			burst_freq_0;
	uint8_t			burst_freq_1;
	uint8_t			burst_freq_2;
	uint8_t			burst_freq_3;
	uint8_t			burst_freq_4;
	uint8_t			gcaf_num_idle;
} __attribute__ ((packed));

/* PROCI_ONETOUCHGESTUREPROCESSOR_T24 */
struct mts_proci_one_touch_gesture_proc_cfg {
	uint8_t			ctrl;
	uint8_t			num_gest;
	uint16_t		gesture_enable;
	uint8_t			pres_proc;
	uint8_t			tap_time_out;
	uint8_t			flick_time_out;
	uint8_t			drag_time_out;
	uint8_t			short_press_time_out;
	uint8_t			long_press_time_out;
	uint8_t			repeat_press_time_out;
	uint16_t		flick_threshold;
	uint16_t		drag_threshold;
	uint16_t		tap_threshold;
	uint16_t		throw_threshold;
} __attribute__ ((packed));

/* SPT_SELFTEST_T25 */
struct mts_spt_self_test_cfg {
	uint8_t			ctrl;
	uint8_t			command;
	uint16_t		high_signal_limit_0;
	uint16_t		low_signal_limit_0;
	uint16_t		high_signal_limit_1;
	uint16_t		low_signal_limit_1;
} __attribute__ ((packed));

/* PROCI_TWOTOUCHGESTUREPROCESSOR_T27 */
struct mts_proci_two_touch_gesture_proc_cfg {
	uint8_t			ctrl;
	uint8_t			reserved0;
	uint8_t			reserved1;
	uint8_t			gesture_enable;
	uint8_t			rotate_threshold;
	uint16_t		zoom_threshold;
} __attribute__ ((packed));

/* SPT_CTECONFIG_T28 */
struct mts_spt_cte_config_cfg {
	uint8_t			ctrl;
	uint8_t			command;
	uint8_t			mode;
	uint8_t			idle_gcaf_depth;
	uint8_t			active_gcaf_depth;
	uint8_t			voltage;
} __attribute__ ((packed));

/* MTS_OBJ_NOISESUPPRESSION_1 */
struct mts_proci_noise1_suppression_cfg {
	uint8_t			ctrl;
	uint8_t			reserved;
	uint8_t			atchthr;
	uint8_t			duty_cycle;
} __attribute__ ((packed));

/* MTS_DEBUG_DIAGNOSTIC_T37 */
struct mts_diagnostic_debug {
	uint8_t		mode;
	uint8_t		page;
	uint16_t	data[64];	
} __attribute__ ((packed));

/* MTS_OBJ_CPT_USERDATA_T38 */
struct mts_spt_userdata {
	uint8_t			data_0;
	uint8_t			data_1;
	uint8_t			data_2;
	uint8_t			data_3;
	uint8_t			data_4;
	uint8_t			data_5;
	uint8_t			data_6;
	uint8_t			data_7;
} __attribute__ ((packed));

/*******************************/
/******** platform data ********/
/*******************************/

struct vkey {
	int     code;
	int     center_x;
	int     center_y;
	int     width;
	int     height;
};

struct virt_keys {
	struct vkey		*keys;
	int			count;
	int			start;
};

struct maxtouch_key {
	uint8_t				channel;
	int				code;
};

struct maxtouch_key_array {
	struct mts_touch_keyarray_cfg	*cfg;
	struct maxtouch_key		*keys;
	int				num_keys;
};

#define MAXTOUCH_FLIP_X		(1 << 0)
#define MAXTOUCH_FLIP_Y		(1 << 1)
#define MAXTOUCH_SWAP_XY		(1 << 2)
#define MAXTOUCH_USE_MULTITOUCH	(1 << 3)
#define MAXTOUCH_USE_KEYARRAY	(1 << 4)
#define MAXTOUCH_CFG_BACKUPNV	(1 << 5)
#define MAXTOUCH_EEPROM_CHECKSUM  (1 << 6)

struct maxtouch_ts_platform_data {
	uint32_t		flags;
	unsigned long		irqflags;

	uint32_t		abs_min_x;
	uint32_t		abs_max_x;
	uint32_t		abs_min_y;
	uint32_t		abs_max_y;
	uint32_t		abs_min_p;
	uint32_t		abs_max_p;
	uint32_t		abs_min_w;
	uint32_t		abs_max_w;

	uint32_t		x_delta;
	uint32_t		y_delta;

	uint16_t		nv_checksum;

	uint32_t		fuzz_x;
	uint32_t		fuzz_y;
	uint32_t		fuzz_p;
	uint32_t		fuzz_w;

	int			(*hw_reset)(void);

	struct virt_keys	vkeys;

	/* TODO: allow multiple key arrays */
	struct maxtouch_key_array			key_array;

	/* object configuration information from board */
	struct mts_gen_power_cfg		power_cfg;
	struct mts_gen_acquire_cfg		acquire_cfg;
	struct mts_touch_multi_cfg		multi_touch_cfg;
	struct mts_touch_keyarray_cfg   key_array_cfg;
	struct mts_procg_sig_filter_cfg		sig_filter_cfg;
	struct mts_proci_linear_tbl_cfg		linear_tbl_cfg;
	struct mts_proci_grip_suppression_cfg	    grip_suppression_cfg;
	struct mts_spt_com_cfg			        com_cfg;
	struct mts_spt_gpio_pwm_cfg			        gpio_pwm_cfg;
	struct mts_procg_noise_suppression_cfg	    noise_suppression_cfg;
	struct mts_proci_one_touch_gesture_proc_cfg	\
					one_touch_gesture_proc_cfg;
	struct mts_spt_self_test_cfg			    self_test_cfg;
	struct mts_proci_two_touch_gesture_proc_cfg	two_touch_gesture_proc_cfg;
	struct mts_spt_cte_config_cfg			    cte_config_cfg;
	struct mts_proci_noise1_suppression_cfg     noise1_suppression_cfg;
	struct mts_spt_userdata     userdata;
};

#endif /* _LINUX_MAXTOUCH_OBP_TS_H */

