/*
    pxa_camera - PXA camera driver header file

    Copyright (C) 2003, Intel Corporation

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __LINUX_PXA_CAMERA_H_
#define __LINUX_PXA_CAMERA_H_

#include <asm/ioctl.h>
#include <linux/videodev2.h>




/*
 * histogram feature
 */

typedef enum {
    HST_COLOR_RED,
    HST_COLOR_BLUE,
    HST_COLOR_GREEN1,
    HST_COLOR_GREEN2
} hst_color_type;


/*
 * private pxa-camera register I/O control
 */
struct reg_set_s {
    int  val1;
    int  val2;
};

#define WCAM_VIDIOCSCAMREG       _IOW('v', 211, struct reg_set_s)
#define WCAM_VIDIOCGCAMREG       _IOWR('v', 212, struct reg_set_s)
#define WCAM_VIDIOCSCIREG        _IOW('v', 213, struct reg_set_s)
#define WCAM_VIDIOCGCIREG        _IOWR('v', 214, struct reg_set_s)
#define WCAM_VIDIOCSINFOR        _IOW('v', 215, struct reg_set_s)
#define WCAM_VIDIOCGINFOR        _IOWR('v', 216, struct reg_set_s)

/*
 *Image format definition
 */


/*
 * SSU scaling parameter
 */
typedef enum {
    CI_SSU_SCALE_DISABLE   = 0,    // disabled
    CI_SSU_SCALE_HALF      = 1,    // 2:1
    CI_SSU_SCALE_QUARTER   = 2     // 4:1
} CI_SSU_SCALE;



/*
 * extended pixel format for V4l2
 */
#define V4L2_PIX_FMT_SRGGB8  v4l2_fourcc('R','A','8','1') /*  8  RGRG.. GBGB.. */
#define V4L2_PIX_FMT_SRGGB10 v4l2_fourcc('R','A','A','1') /*  10  RGRG.. GBGB.. */


/*  Borqs/OMS private definition */
#define V4L2_MAX_AF_WINDOWS_COUNT	16

/* Defines a auto focus window */
struct v4l2_ext_af_window
{
	struct v4l2_rect win;		/* window rectangle */
	int weight;			/* window weight, 0-100 */
};

/* Defines auto focus windows configuration */
struct v4l2_ext_af_windows_config
{
	__u32 count;			/* valid numbers of auto focus window */
	struct v4l2_ext_af_window wins[V4L2_MAX_AF_WINDOWS_COUNT];	/* windows array */
	__u32 reserved[4];
};

/*  Borqs/OMS private camera I/O controls */
#define VIDIOC_OMS_PRIVATE_BASE			(BASE_VIDIOC_PRIVATE+32)

#define VIDIOC_S_AF_WINDOWS_CONFIG		_IOWR ('V', VIDIOC_OMS_PRIVATE_BASE+1, struct v4l2_ext_af_windows_config)
#define VIDIOC_G_AF_WINDOWS_CONFIG		_IOWR ('V', VIDIOC_OMS_PRIVATE_BASE+2, struct v4l2_ext_af_windows_config)

/*  Borqs/OMS private class control IDs */
#define V4L2_CID_EXT_FLASH_MODE			(V4L2_CID_PRIVATE_BASE+1)
#define V4L2_CID_EXT_SCENE			(V4L2_CID_PRIVATE_BASE+2)
#define V4L2_CID_EXT_COLORTONE			(V4L2_CID_PRIVATE_BASE+3)
#define V4L2_CID_EXT_SHARPNESS			(V4L2_CID_PRIVATE_BASE+4)
#define V4L2_CID_EXT_JPEG_QUALITY_MODE		(V4L2_CID_PRIVATE_BASE+5)
#define V4L2_CID_EXT_EV_OFFSET			(V4L2_CID_PRIVATE_BASE+6)
#define V4L2_CID_EXT_ISOSPEED			(V4L2_CID_PRIVATE_BASE+7)
#define V4L2_CID_EXT_DIGITAL_ZOOM		(V4L2_CID_PRIVATE_BASE+8)
#define V4L2_CID_EXT_FOCUS_MODE			(V4L2_CID_PRIVATE_BASE+9)
#define V4L2_CID_EXT_FOCUS_MACRO		(V4L2_CID_PRIVATE_BASE+10)
#define V4L2_CID_EXT_DIGITAL_ZOOM_MAX		(V4L2_CID_PRIVATE_BASE+11)
#define V4L2_CID_EXT_DIGITAL_ZOOM_STEP		(V4L2_CID_PRIVATE_BASE+12)
#define V4L2_CID_EXT_FLASH_TORCH		(V4L2_CID_PRIVATE_BASE+13)
#define V4L2_CID_EXT_WB_MODE			(V4L2_CID_PRIVATE_BASE+14)
#define V4L2_CID_EXT_EFFECT			(V4L2_CID_PRIVATE_BASE+15)
#define V4L2_CID_EXT_JPEG_QUALITY_PERCENT	(V4L2_CID_PRIVATE_BASE+16)

#endif /* __LINUX_PXA_CAMERA_H_ */
