/* kernel/power/latesuspend.c
 *
 * Copyright (C) 2009 Borqs, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/latesuspend.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/rtc.h>
#include <linux/syscalls.h> /* sys_sync */
#include <linux/wakelock.h>
#include <linux/workqueue.h>

#include "power.h"

static DEFINE_MUTEX(late_suspend_lock);
static LIST_HEAD(late_suspend_handlers);

#ifdef CONFIG_PXA3xx
extern int pxa3xx_enter_standby_mode(void);
#endif

/*add register_early/late_suspend function for Fakesleep */
void register_late_suspend(struct late_suspend *handler)
{
	struct list_head *pos;

	mutex_lock(&late_suspend_lock);
	list_for_each(pos, &late_suspend_handlers) {
		struct late_suspend *e = list_entry(pos, struct late_suspend, link);
		if(e->level > handler->level)
			break;
	}
	list_add_tail(&handler->link, pos);
	mutex_unlock(&late_suspend_lock);
}
EXPORT_SYMBOL(register_late_suspend);

void unregister_late_suspend(struct late_suspend *handler)
{
	mutex_lock(&late_suspend_lock);
	list_del(&handler->link);
	mutex_unlock(&late_suspend_lock);
}
EXPORT_SYMBOL(unregister_late_suspend);

/* 
 * Call late suspend handlers and Marvell pxa platform pm_suspend function
 */
int pm_late_suspend(void)
{
        int ret = 0;
	struct late_suspend *pos;

	mutex_lock(&late_suspend_lock);
	list_for_each_entry(pos, &late_suspend_handlers, link) {
		if (pos->suspend != NULL)
			pos->suspend(pos);
	}
	mutex_unlock(&late_suspend_lock);
	
        /*Add Marvell specific pm suspend function here*/
        printk(KERN_ERR "Enter Standby Mode\n");
#if !(defined(CONFIG_MACH_BENZINA2) || defined(CONFIG_MACH_IGNITE))
        ret = pxa3xx_enter_standby_mode();
#endif
        /*touch watch dog after resume from standby mode, or watch dog would bark*/
        touch_softlockup_watchdog();

	mutex_lock(&late_suspend_lock);
	list_for_each_entry_reverse(pos, &late_suspend_handlers, link)
		if (pos->resume != NULL)
			pos->resume(pos);
	mutex_unlock(&late_suspend_lock);
	return ret;
}
