#!/bin/sh

# 
# Copyright (C) 2008~2009 Borqs, Inc.
#
# This software is licensed under the terms of the GNU General Public
# License version 2, as published by the Free Software Foundation, and
# may be copied, distributed, and modified under those terms.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#


release_base=${PWD}
release_root=${PWD}/mkrelease_root

build_check=`echo $* | grep BUILD_CHECK > /dev/null; echo $?`
products=`echo $* | sed -e 's/BUILD_CHECK//g'`

svn_google_ropo=svn://192.168.1.250/linux/pool/android-2.6.29-goldfish

release_sources_tmp=${release_root}/release_sources.tmp
release_includes_tmp=${release_root}/release_includes.tmp
release_sources=${release_root}/release_sources.txt
release_includes=${release_root}/release_includes.txt


func_build_maintree_kernel()
{
	#
	# build base release, make sure our internal svn can build successfully! 
	#
	PRODUCT=$1
	mkdir -p $release_root
	echo "====== Begin to build $PRODUCT......"
	./setup.sh PRODUCT=$PRODUCT root=ramdisk 2>&1 | tee $release_root/base.$PRODUCT.makedefconfig
	make -j4 2>&1 | tee $release_root/base.$PRODUCT.makezImage
	cat .config | grep ^CONFIG | sort -u > $release_root/base.$PRODUCT.config
	cp .config $release_root/base.$PRODUCT.defconfig
	cp ./vmlinux $release_root/base.$PRODUCT.vmlinux
	cp ./System.map $release_root/base.$PRODUCT.System.map
	cp ./arch/arm/boot/zImage $release_root/base.$PRODUCT.zImage
	echo "====== Build $PRODUCT successfully"
}

func_check_build_files()
{
	PRODUCT=$1
	# FIXME: we only support pxa and opxa3 now!
	ARCH=`grep -e "CONFIG_ARCH_PXA=" -e "CONFIG_ARCH_OPXA3=" -e "CONFIG_ARCH_GOLDFISH=" $release_root/base.$PRODUCT.config | sed -e 's/CONFIG_ARCH_//g' -e 's/=y//g' | tr 'A-Z' 'a-z'`
	if [ "x$ARCH" != "xpxa" -a "x$ARCH" != "xopxa3" -a "x$ARCH" != "xgoldfish" ]; then
		echo "Unknown arch, exit!"
		exit 0
	fi
	echo "release $ARCH"

	> $release_sources_tmp
	> $release_includes_tmp
	> $release_sources
	> $release_includes

        find ./include/asm-generic/ -name "*" | sed -e "s/\.\///g" |
        while read line
        do
                echo $line >> $release_includes_tmp
        done
	find ./ -name "*.o" | sed -e "s/\.\///g" -e "s/\.o//g" | 
	while read line
	do 
	for i in c s S h
	do
		if [ -f $line.$i ]
		then
			echo $line.$i >> $release_sources_tmp
			dirpath=`dirname $line` 
			perl scripts/printincludes.pl $line.$i |
			while read includes
			do
				if [ -f include/$includes -o -f arch/arm/include/$includes -o -f arch/arm/mach-$ARCH/include/$includes ]
				then
				echo include/$includes | sed -e "s/include\/asm-generic\//gggg\//g" -e "s/include\/asm\//aaaa\//g" -e "s/gggg\//include\/asm-generic\//g" -e  "s/aaaa\//arch\/arm\/include\/asm\//g" -e "s/include\/mach\//arch\/arm\/mach-$ARCH\/include\/mach\//g" >> $release_includes_tmp
				else
				if [ -f $dirpath/$includes ]
				then 
					echo $dirpath/$includes >> $release_includes_tmp
				fi
				fi
			done
		fi
	done
	done
	
	find . -name "Kconfig*" >> $release_sources_tmp
	find . -name "Makefile*" >> $release_sources_tmp
	echo "arch/arm/tools/mach-types" >> $release_sources_tmp

	perl  scripts/puresource.pl $release_sources_tmp > $release_sources
	perl  scripts/puresource.pl $release_includes_tmp > $release_includes

#	make distclean 2>&1 | tee $release_root/base.distclean
	rm -f $release_sources_tmp
	rm -f $release_includes_tmp
}

func_generate_release_source()
{
	PRODUCT=$1
	release_product=$release_root/lnx2629pxa_release

	if [ ! -d $release_product ]; then
		svn export $svn_google_ropo $release_product
		find $release_product -name "*.git*" | xargs rm -rf
	fi

	cp $release_root/base.${PRODUCT}.defconfig $release_product/arch/arm/configs/`echo ${PRODUCT} | tr 'A-Z' 'a-z'`_defconfig -f

	cat $release_sources |
	while read line 
	do
		filepath=$release_product/$line
		dirpath=`dirname $filepath`
		if [ ! -d $dirpath ]
		then
			mkdir -p $dirpath
		fi
		cp $release_base/$line $filepath -rf
	done

	cat $release_includes |
	while read line 
	do
		filepath=$release_product/$line
		dirpath=`dirname $filepath`
		if [ ! -d $dirpath ]
		then
			mkdir -p $dirpath
		fi
		cp $release_base/$line $filepath -rf
	done

	(cd $release_product; make mrproper 2>&1 | tee $release_root/product.mrproper)
	(cd $release_base; make mrproper 2>&1 | tee $release_root/base.mrproper)
	rm -rf $release_product/mkrelease_root   #for multi-product release temproarily 
	rm -f $release_sources
	rm -f $release_includes

}

func_build_release_kernel()
{
	PRODUCT=$1
	release_product=$release_root/lnx2629pxa_release
(cd $release_product
	make `echo ${PRODUCT} | tr 'A-Z' 'a-z'`_defconfig
        make -j4 2>&1 | tee $release_root/release.$PRODUCT.makezImage
        cat .config | grep ^CONFIG | sort -u > $release_root/release.$PRODUCT.config
        cp ./vmlinux $release_root/release.$PRODUCT.vmlinux
        cp ./System.map $release_root/release.$PRODUCT.System.map
        cp ./arch/arm/boot/zImage $release_root/release.$PRODUCT.zImage
        make mrproper 2>&1 | tee $release_root/release.$PRODUCT.mrproper
)
}

func_check_release()
{
	PRODUCT=$1
	if [ ! -f $release_root/release.$PRODUCT.vmlinux ]; then
		echo "Build error of internal svn, please check it manually!!!"
		exit 0	
	fi
}

###############################################################################
#
# GO GO GO!
#
###############################################################################

if rev=`svn info 2>/dev/null | grep '^Revision'`; then
	echo "Please use 'svn export svn_ropo' to run this script"
	exit
fi

for p in $products
do
	echo "make release: $p"

	func_build_maintree_kernel $p
	func_check_build_files $p
	func_generate_release_source $p
done

for p in $products
do
	if [ $build_check -eq 0 ]; then
		func_build_release_kernel $p 
		func_check_release $p 
	fi
done
