/*
 * linux/sound/soc/codecs/levante.c
 * Base on linux/sound/soc/codecs/wm9712.c
 *
 * Copyright (C) 2007 Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
//#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>

#define DA9034_ERRATA_17	1

#define	MICCO_MUX_MONO_V		0x16
#define	MICCO_MUX_BEAR_V		0x17
#define	MICCO_MUX_LINE_OUT_V		0x18
#define	MICCO_MUX_STEREO_CH1_V		0x19
#define	MICCO_MUX_STEREO_CH2_V		0x1a
#define	MICCO_MUX_TX_V			0x1b
#define MICCO_MIXER0_SELECT		0x1c
#define MICCO_MIXER1_SELECT		0x1d

#define LEVANTE_ID_V			0x16

/* debug */
#define LEVANTE_DEBUG 0
#if LEVANTE_DEBUG
#define dbg(format, arg...) printk(KERN_INFO "levante: " format "\n", ## arg)
#else
#define dbg(format, arg...)
#endif

/* Temporary */
#define MICCO_HIFI_DAC_CONTROL 1
#define MICCO_VCODEC_VDAC_CONTROL 2
#define MICCO_VCODEC_ADC_CONTROL 3
#define MICCO_MONO_VOL 4
#define MICCO_BEAR_VOL 5
#define MICCO_AUDIO_LINE_AMP 6
#define MICCO_STEREO_AMPLITUDE_CH1 7
#define MICCO_MIC_PGA 8
#define MICCO_PGA_AUX1_2 9
#define MICCO_PGA_AUX3 10
#define MICCO_STEREO_AMPLITUDE_CH2 11
#define MICCO_TX_PGA 12
#define MICCO_SIDETONE 13
/* End of Temporary part */

static unsigned int levante_soc_read(struct snd_soc_codec *codec,
	unsigned int reg);
static int levante_soc_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int val);

static const u8 levante_audio_regs[] =
{
	0x00, 0x00, 0x00, 0x00,	0x00,     /*0x42 ~ 0x46*/
	0x00, 0x00, 0x10, 0x00,		      /*0xb0 ~ 0xb3*/
	0x00, 0x00, 0x00, 0x00,		      /*0xb4 ~ 0xb7*/
	0x00, 0x00, 0x00, 0x00,		      /*0xb8 ~ 0xbb*/
	0x00, 0x10, 0xc8, 0x00,		      /*0xbc ~ 0xbf*/
	0x00, 0x00, 0x00, 0x00,		      /*0xc0 ~ 0xc3*/
	0x00, 0x3f, 0x3f, 0x00,		      /*0xc4 ~ 0xc7*/
	0x3f, 0x3f, 0x00, 0x00,		      /*0xc8 ~ 0xcb*/
	0x00, 0x00, 0x00, 0x00,		      /*0xcc ~ 0xcf*/
	0x7e, 0x84, 0x1b, 0x11,		      /*0xd0 ~ 0xd3*/
	0x70, 0x4c, 0x4c, 0x00,		      /*0xd4 ~ 0xd7*/
	0x00, 0x34, 0x05, 0x00,		      /*0xd8 ~ 0xdb*/
	0x0b, 0x05, 0x3f, 0x23,		      /*0xdc ~ 0xdf*/
	0x00, 0x39, 0x00, 0x00,		      /*0xe0 ~ 0xe3*/
	0x00, 0x00, 0x00, 0x09,		      /*0xe4 ~ 0xe7*/
	0x01, 0x00, 0x00, 0x0a		      /*0xe8 ~ 0xeb*/
};

static const struct snd_kcontrol_new levante_direct_access[] = {
	SOC_SINGLE("LEVANTE_MISC2", LEVANTE_MISC2, 0, 0xff, 0),                                /* 0x42 */
	SOC_SINGLE("LEVANTE_PLL_CTRL1", LEVANTE_PLL_CTRL1, 0, 0xff, 0),                        /* 0x43 */
	SOC_SINGLE("LEVANTE_PLL_FRAC1", LEVANTE_PLL_FRAC1, 0, 0xff, 0),                        /* 0x44 */
	SOC_SINGLE("LEVANTE_PLL_FRAC2", LEVANTE_PLL_FRAC2, 0, 0xff, 0),                        /* 0x45 */
	SOC_SINGLE("LEVANTE_PLL_FRAC3", LEVANTE_PLL_FRAC3, 0, 0xff, 0),                        /* 0x46 */
	SOC_SINGLE("LEVANTE_PCM_INTERFACE1", LEVANTE_PCM_INTERFACE1, 0, 0xff, 0),              /* 0xB0 */
	SOC_SINGLE("LEVANTE_PCM_INTERFACE2", LEVANTE_PCM_INTERFACE2, 0, 0xff, 0),              /* 0xB1 */
	SOC_SINGLE("LEVANTE_PCM_INTERFACE3", LEVANTE_PCM_INTERFACE3, 0, 0xff, 0),              /* 0xB2 */
	SOC_SINGLE("LEVANTE_PCM_RATE", LEVANTE_PCM_RATE, 0, 0xff, 0),                          /* 0xB3 */
	SOC_SINGLE("LEVANTE_ECHO_CANCEL_PATH", LEVANTE_ECHO_CANCEL_PATH, 0, 0xff, 0),          /* 0xB4 */
	SOC_SINGLE("LEVANTE_SIDETONE_GAIN1", LEVANTE_SIDETONE_GAIN1, 0, 0xff, 0),              /* 0xB5 */
	SOC_SINGLE("LEVANTE_SIDETONE_GAIN2", LEVANTE_SIDETONE_GAIN2, 0, 0xff, 0),              /* 0xB6 */
	SOC_SINGLE("LEVANTE_SIDETONE_GAIN3", LEVANTE_SIDETONE_GAIN3, 0, 0xff, 0),              /* 0xB7 */
	SOC_SINGLE("LEVANTE_ADC_OFFSET1", LEVANTE_ADC_OFFSET1, 0, 0xff, 0),                    /* 0xB8 */
	SOC_SINGLE("LEVANTE_ADC_OFFSET2", LEVANTE_ADC_OFFSET2, 0, 0xff, 0),                    /* 0xB9 */
	SOC_SINGLE("LEVANTE_DMIC_DELAY", LEVANTE_DMIC_DELAY, 0, 0xff, 0),                      /* 0xBA */
	SOC_SINGLE("LEVANTE_I2S_INTERFACE1", LEVANTE_I2S_INTERFACE1, 0, 0xff, 0),              /* 0xBB */
	SOC_SINGLE("LEVANTE_I2S_INTERFACE2", LEVANTE_I2S_INTERFACE2, 0, 0xff, 0),              /* 0xBC */
	SOC_SINGLE("LEVANTE_I2S_INTERFACE3", LEVANTE_I2S_INTERFACE3, 0, 0xff, 0),              /* 0xBD */
	SOC_SINGLE("LEVANTE_I2S_INTERFACE4", LEVANTE_I2S_INTERFACE4, 0, 0xff, 0),              /* 0xBE */
	SOC_SINGLE("LEVANTE_EQUALIZER_N0_1", LEVANTE_EQUALIZER_N0_1, 0, 0xff, 0),              /* 0xBF */
	SOC_SINGLE("LEVANTE_EQUALIZER_N0_2", LEVANTE_EQUALIZER_N0_2, 0, 0xff, 0),              /* 0xC0 */
	SOC_SINGLE("LEVANTE_EQUALIZER_N1_1", LEVANTE_EQUALIZER_N1_1, 0, 0xff, 0),              /* 0xC1 */
	SOC_SINGLE("LEVANTE_EQUALIZER_N1_2", LEVANTE_EQUALIZER_N1_2, 0, 0xff, 0),              /* 0xC2 */
	SOC_SINGLE("LEVANTE_EQUALIZER_D1_1", LEVANTE_EQUALIZER_N1_1, 0, 0xff, 0),              /* 0xC3 */
	SOC_SINGLE("LEVANTE_EQUALIZER_D1_2", LEVANTE_EQUALIZER_N1_2, 0, 0xff, 0),              /* 0xC4 */
	SOC_SINGLE("LEVANTE_SIDE_TONE1", LEVANTE_SIDE_TONE1, 0, 0xff, 0),                      /* 0xC5 */
	SOC_SINGLE("LEVANTE_SIDE_TONE2", LEVANTE_SIDE_TONE2, 0, 0xff, 0),                      /* 0xC6 */
	SOC_SINGLE("LEVANTE_LEFT_GAIN1", LEVANTE_LEFT_GAIN1, 0, 0xff, 0),                      /* 0xC7 */
	SOC_SINGLE("LEVANTE_LEFT_GAIN2", LEVANTE_LEFT_GAIN2, 0, 0xff, 0),                      /* 0xC8 */
	SOC_SINGLE("LEVANTE_RIGHT_GAIN1", LEVANTE_RIGHT_GAIN1, 0, 0xff, 0),                    /* 0xC9 */
	SOC_SINGLE("LEVANTE_RIGHT_GAIN2", LEVANTE_RIGHT_GAIN2, 0, 0xff, 0),                    /* 0xCA */
	SOC_SINGLE("LEVANTE_DAC_OFFSET", LEVANTE_DAC_OFFSET, 0, 0xff, 0),                      /* 0xCB */
	SOC_SINGLE("LEVANTE_OFFSET_LEFT1", LEVANTE_OFFSET_LEFT1, 0, 0xff, 0),                  /* 0xCC */
	SOC_SINGLE("LEVANTE_OFFSET_LEFT2", LEVANTE_OFFSET_LEFT2, 0, 0xff, 0),                  /* 0xCD */
	SOC_SINGLE("LEVANTE_OFFSET_RIGHT1", LEVANTE_OFFSET_RIGHT1, 0, 0xff, 0),                /* 0xCE */
	SOC_SINGLE("LEVANTE_OFFSET_RIGHT2", LEVANTE_OFFSET_RIGHT2, 0, 0xff, 0),                /* 0xCF */
	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM1", LEVANTE_ADC_ANALOG_PROGRAM1, 0, 0xff, 0),    /* 0xD0 */
	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM2", LEVANTE_ADC_ANALOG_PROGRAM2, 0, 0xff, 0),    /* 0xD1 */
	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM3", LEVANTE_ADC_ANALOG_PROGRAM3, 0, 0xff, 0),    /* 0xD2 */
	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM4", LEVANTE_ADC_ANALOG_PROGRAM4, 0, 0xff, 0),    /* 0xD3 */
	SOC_SINGLE("LEVANTE_ANALOG_TO_ANALOG", LEVANTE_ANALOG_TO_ANALOG, 0, 0xff, 0),          /* 0xD4 */
	SOC_SINGLE("LEVANTE_HS1_CTRL", LEVANTE_HS1_CTRL, 0, 0xff, 0),                          /* 0xD5 */
	SOC_SINGLE("LEVANTE_HS2_CTRL", LEVANTE_HS2_CTRL, 0, 0xff, 0),                          /* 0xD6 */
	SOC_SINGLE("LEVANTE_LO1_CTRL", LEVANTE_LO1_CTRL, 0, 0xff, 0),                          /* 0xD7 */
	SOC_SINGLE("LEVANTE_LO2_CTRL", LEVANTE_LO2_CTRL, 0, 0xff, 0),                          /* 0xD8 */
	SOC_SINGLE("LEVANTE_EAR_SPKR_CTRL1", LEVANTE_EAR_SPKR_CTRL1, 0, 0xff, 0),              /* 0xD9 */
	SOC_SINGLE("LEVANTE_EAR_SPKR_CTRL2", LEVANTE_EAR_SPKR_CTRL2, 0, 0xff, 0),              /* 0xDA */
	SOC_SINGLE("LEVANTE_AUDIO_SUPPLIES1", LEVANTE_AUDIO_SUPPLIES1, 0, 0xff, 0),            /* 0xDB */
	SOC_SINGLE("LEVANTE_AUDIO_SUPPLIES2", LEVANTE_AUDIO_SUPPLIES2, 0, 0xff, 0),            /* 0xDC */
	SOC_SINGLE("LEVANTE_ADC_ENABLES1", LEVANTE_ADC_ENABLES1, 0, 0xff, 0),                  /* 0xDD */
	SOC_SINGLE("LEVANTE_ADC_ENABLES2", LEVANTE_ADC_ENABLES2, 0, 0xff, 0),                  /* 0xDE */
	SOC_SINGLE("LEVANTE_DAC_ENABLES1", LEVANTE_DAC_ENABLES1, 0, 0xff, 0),                  /* 0xDF */
	SOC_SINGLE("LEVANTE_DAC_DUMMY", LEVANTE_DUMMY, 0, 0xff, 0),                            /* 0xE0 */
	SOC_SINGLE("LEVANTE_DAC_ENABLES2", LEVANTE_DAC_ENABLES2, 0, 0xff, 0),                  /* 0xE1 */
	SOC_SINGLE("LEVANTE_AUDIO_CAL1", LEVANTE_AUDIO_CAL1, 0, 0xff, 0),                      /* 0xE2 */
	SOC_SINGLE("LEVANTE_AUDIO_CAL2", LEVANTE_AUDIO_CAL2, 0, 0xff, 0),                      /* 0xE3 */
	SOC_SINGLE("LEVANTE_AUDIO_CAL3", LEVANTE_AUDIO_CAL3, 0, 0xff, 0),                      /* 0xE4 */
	SOC_SINGLE("LEVANTE_AUDIO_CAL4", LEVANTE_AUDIO_CAL4, 0, 0xff, 0),                      /* 0xE5 */
	SOC_SINGLE("LEVANTE_AUDIO_CAL5", LEVANTE_AUDIO_CAL5, 0, 0xff, 0),                      /* 0xE6 */
	SOC_SINGLE("LEVANTE_ANALOG_INPUT_SEL1", LEVANTE_ANALOG_INPUT_SEL1, 0, 0xff, 0),        /* 0xE7 */
	SOC_SINGLE("LEVANTE_ANALOG_INPUT_SEL2", LEVANTE_ANALOG_INPUT_SEL2, 0, 0xff, 0),        /* 0xE8 */
	SOC_SINGLE("LEVANTE_MIC_BUTTON_DETECTION", LEVANTE_MIC_BUTTON_DETECTION, 0, 0xff, 0),  /* 0xE9 */
	SOC_SINGLE("LEVANTE_HEADSET_DETECTION", LEVANTE_HEADSET_DETECTION, 0, 0xff, 0),        /* 0xEA */
	SOC_SINGLE("LEVANTE_SHORTS", LEVANTE_SHORTS, 0, 0xff, 0),                              /* 0xEB */
};

static int levante_id_get(struct snd_kcontrol *kcontrol,struct snd_ctl_elem_value *ucontrol)
{
	int ret;
	unsigned char value;

	ret = levante_codec_read(0x00, &value);
	if (ret !=0 )
		return -1;

	ucontrol->value.integer.value[0] = value;
	return 0;
}


static const struct snd_kcontrol_new levante_id_read =
	SOC_SINGLE_EXT("LEVANTE_ID", LEVANTE_ID_V, 0, 0xff, 0, levante_id_get,NULL);

static int levante_add_direct_access(struct snd_soc_codec *codec)
{
	int err, i;
	for (i = 0; i < ARRAY_SIZE(levante_direct_access); i++)
	{
		err = snd_ctl_add(codec->card, snd_soc_cnew(&levante_direct_access[i], codec, NULL));
		if (err < 0)
			return err;
	}
	return 0;
}

static int levante_id_access(struct snd_soc_codec *codec)
{
	int err;
	printk(KERN_ERR "levante_id_access\n");
	err = snd_ctl_add(codec->card, snd_soc_cnew(&levante_id_read, codec, NULL));
	if (err < 0)
		return err;

	return 0;
}

static int do_post_event(struct snd_soc_dapm_widget *w, struct snd_kcontrol *k, int event)
{
	return 0;
}

static int do_mixer0_event(struct snd_soc_dapm_widget *w, struct snd_kcontrol *k, int event)
{
	return 0;
}

static int do_mixer1_event(struct snd_soc_dapm_widget *w, struct snd_kcontrol *k, int event)
{
	return 0;
}

#ifdef DA9034_ERRATA_17
#define MUX_INPUT	\
		"AUX1", "AUX2", "AUX3", "DAC1", "DAC2", "DAC3", "AUX2inv",\
		"NULL"
#else
#define MUX_INPUT	\
		"AUX1", "AUX2", "AUX2inv", "AUX3", "DAC1", "DAC2", "DAC3",\
		"NULL"
#endif

#define MIXER0_SELECT	\
		"CLOSE", "MONO", "BEAR", "STEREO", "MONO+BEAR", \
		"MONO+STEREO", "BEAR+STEREO", "MONO+BEAR+STEREO"

#define MIXER1_SELECT   \
		"CLOSE", "MIC ENABLE", "AUX1 ENABLE", "AUX2 ENABLE",\
		"AUX3 ENABLE", "VOICE ADC ENABLE", "VOICE DAC ENABLE",\
		"HIFI DAC ENABLE", "LINE AMP ENABLE", "SIDETONE ENABLE"

static const char *levante_mono_mux[] = {MUX_INPUT};
static const char *levante_bear_mux[] = {MUX_INPUT};
static const char *levante_line_out_mux[] = {"AUX1", "AUX2", "AUX3", "DAC1",
	"DAC2", "DAC3", "AUX2inv", "MIC_P", "NULL"};
static const char *levante_stereo_ch1_mux[] = {MUX_INPUT};
static const char *levante_stereo_ch2_mux[] = {MUX_INPUT};
static const char *levante_tx_mux[] = {"AUX2", "AUX1", "Mic1", "Mic2", "AUX2inv",
	"AUX3", "DAC1", "DAC2", "NULL"};

static const char *levante_select_mixer0[] = {MIXER0_SELECT};
static const char *levante_select_mixer1[] = {MIXER1_SELECT};

static const struct soc_enum levante_enum[] = {
	SOC_ENUM_SINGLE(MICCO_MUX_MONO_V,	0, 8, levante_mono_mux),
	SOC_ENUM_SINGLE(MICCO_MUX_BEAR_V,	0, 8, levante_bear_mux),
	SOC_ENUM_SINGLE(MICCO_MUX_LINE_OUT_V,	0, 9, levante_line_out_mux),
	SOC_ENUM_SINGLE(MICCO_MUX_STEREO_CH1_V, 0, 8, levante_stereo_ch1_mux),
	SOC_ENUM_SINGLE(MICCO_MUX_STEREO_CH2_V, 0, 8, levante_stereo_ch2_mux),
	SOC_ENUM_SINGLE(MICCO_MUX_TX_V,		0, 9, levante_tx_mux),
	SOC_ENUM_SINGLE(MICCO_MIXER0_SELECT,	0, 8, levante_select_mixer0),
	SOC_ENUM_SINGLE(MICCO_MIXER1_SELECT,	0, 10, levante_select_mixer1),
};

static const struct snd_kcontrol_new levante_mono_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[0]);
static const struct snd_kcontrol_new levante_bear_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[1]);
static const struct snd_kcontrol_new levante_line_out_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[2]);
static const struct snd_kcontrol_new levante_stereo_ch1_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[3]);
static const struct snd_kcontrol_new levante_stereo_ch2_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[4]);
static const struct snd_kcontrol_new levante_tx_mux_controls =
	SOC_DAPM_ENUM("Route", levante_enum[5]);
static const struct snd_kcontrol_new levante_select_mixer0_controls =
	SOC_DAPM_ENUM("Route", levante_enum[6]);
static const struct snd_kcontrol_new levante_select_mixer1_controls =
	SOC_DAPM_ENUM("Route", levante_enum[7]);

static const struct snd_soc_dapm_widget levante_dapm_widgets[] = {
	SND_SOC_DAPM_DAC("DAC1", "Left HiFi Playback", MICCO_HIFI_DAC_CONTROL,
			7, 0),
	SND_SOC_DAPM_DAC("DAC2", "Right HiFi Playback", MICCO_HIFI_DAC_CONTROL,
			7, 0),
	SND_SOC_DAPM_DAC("DAC3", "Voice Playback", MICCO_VCODEC_VDAC_CONTROL,
			3, 0),
	SND_SOC_DAPM_ADC("ADC", "Voice Capture", MICCO_VCODEC_ADC_CONTROL,
			0, 0),

	SND_SOC_DAPM_MUX_E("MIXER0 Select", SND_SOC_NOPM,
			0, 0, &levante_select_mixer0_controls,
		do_mixer0_event, SND_SOC_DAPM_POST_REG),

	SND_SOC_DAPM_MUX_E("MIXER1 Select", SND_SOC_NOPM,
			0, 0, &levante_select_mixer1_controls,
		do_mixer1_event, SND_SOC_DAPM_POST_REG),

	SND_SOC_DAPM_MUX_E("MONO Mux", SND_SOC_NOPM,
			0, 0, &levante_mono_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),
	SND_SOC_DAPM_MUX_E("BEAR Mux", SND_SOC_NOPM,
			0, 0, &levante_bear_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),
	SND_SOC_DAPM_MUX_E("LINE_OUT Mux", SND_SOC_NOPM, 0,
			0, &levante_line_out_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),
	SND_SOC_DAPM_MUX_E("STEREO_CH1 Mux", SND_SOC_NOPM,
			0, 0, &levante_stereo_ch1_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),
	SND_SOC_DAPM_MUX_E("STEREO_CH2 Mux", SND_SOC_NOPM,
			0, 0, &levante_stereo_ch2_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),
	SND_SOC_DAPM_MUX_E("TX Mux", SND_SOC_NOPM,
			0, 0, &levante_tx_mux_controls,
		do_post_event, SND_SOC_DAPM_POST_REG),

	SND_SOC_DAPM_PGA("MONO PGA", MICCO_MONO_VOL, 6, 0, NULL, 0),
	SND_SOC_DAPM_PGA("BEAR PGA", MICCO_BEAR_VOL, 6, 0, NULL, 0),

	SND_SOC_DAPM_PGA("LINE_OUT PGA", MICCO_AUDIO_LINE_AMP, 4, 0, NULL, 0),
	SND_SOC_DAPM_PGA("STEREO_CH1 PGA", MICCO_STEREO_AMPLITUDE_CH1,
			7, 0, NULL, 0),
	SND_SOC_DAPM_PGA("STEREO_CH2 PGA", MICCO_STEREO_AMPLITUDE_CH1,
			6, 0, NULL, 0),
	SND_SOC_DAPM_PGA("MIC PGA", MICCO_MIC_PGA, 3, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AUX1 PGA", MICCO_PGA_AUX1_2, 3, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AUX2 PGA", MICCO_PGA_AUX1_2, 7, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AUX3 PGA", MICCO_PGA_AUX3, 3, 0, NULL, 0),

	SND_SOC_DAPM_OUTPUT("BEAR1"),
	SND_SOC_DAPM_OUTPUT("BEAR2"),
	SND_SOC_DAPM_OUTPUT("MONO1"),
	SND_SOC_DAPM_OUTPUT("MONO2"),
	SND_SOC_DAPM_OUTPUT("LINE_OUT"),
	SND_SOC_DAPM_OUTPUT("STEREO_CH1"),
	SND_SOC_DAPM_OUTPUT("STEREO_CH2"),
	SND_SOC_DAPM_INPUT("AUX1"),
	SND_SOC_DAPM_INPUT("AUX2"),
	SND_SOC_DAPM_INPUT("AUX3"),
	SND_SOC_DAPM_INPUT("MIC1"),
	SND_SOC_DAPM_INPUT("MIC2"),
};

static const char *levante_audio_map[][3] = {
	/*MONO Mux*/
	{"MONO Mux", "AUX1", "AUX1"},
	{"MONO Mux", "AUX2", "AUX2"},
	{"MONO Mux", "AUX3", "AUX3"},
	{"MONO Mux", "DAC1", "DAC1"},
	{"MONO Mux", "DAC2", "DAC2"},
	{"MONO Mux", "DAC3", "DAC3"},

	{"MONO PGA", NULL, "MONO Mux"},
	{"MONO1", NULL, "MONO PGA"},
	{"MONO2", NULL, "MONO PGA"},

	/*BEAR Mux*/
	{"BEAR Mux", "AUX1", "AUX1"},
	{"BEAR Mux", "AUX2", "AUX2"},
	{"BEAR Mux", "AUX3", "AUX3"},
	{"BEAR Mux", "DAC1", "DAC1"},
	{"BEAR Mux", "DAC2", "DAC2"},
	{"BEAR Mux", "DAC3", "DAC3"},

	{"BEAR PGA", NULL, "BEAR Mux"},
	{"BEAR1", NULL, "BEAR PGA"},
	{"BEAR2", NULL, "BEAR PGA"},

	/*LINE OUT Mux Mux*/
	{"LINE_OUT Mux", "AUX1", "AUX1"},
	{"LINE_OUT Mux", "AUX2", "AUX2"},
	{"LINE_OUT Mux", "AUX3", "AUX3"},
	{"LINE_OUT Mux", "DAC1", "DAC1"},
	{"LINE_OUT Mux", "DAC2", "DAC2"},
	{"LINE_OUT Mux", "DAC3", "DAC3"},
	{"LINE_OUT Mux", "MIC_P", "MIC_P"},

	{"LINE_OUT PGA", NULL, "LINE_OUT Mux"},
	{"LINE_OUT", NULL, "LINE_OUT PGA"},


	/*STEREO_CH1 Mux*/
	{"STEREO_CH1 Mux", "AUX1", "AUX1"},
	{"STEREO_CH1 Mux", "AUX2", "AUX2"},
	{"STEREO_CH1 Mux", "AUX3", "AUX3"},
	{"STEREO_CH1 Mux", "DAC1", "DAC1"},
	{"STEREO_CH1 Mux", "DAC2", "DAC2"},
	{"STEREO_CH1 Mux", "DAC3", "DAC3"},

	{"STEREO_CH1 PGA", NULL, "STEREO_CH1 Mux"},
	{"STEREO_CH1", NULL, "STEREO_CH1 PGA"},

	/*STEREO_CH2 Mux*/
	{"STEREO_CH2 Mux", "AUX1", "AUX1"},
	{"STEREO_CH2 Mux", "AUX2", "AUX2"},
	{"STEREO_CH2 Mux", "AUX3", "AUX3"},
	{"STEREO_CH2 Mux", "DAC1", "DAC1"},
	{"STEREO_CH2 Mux", "DAC2", "DAC2"},
	{"STEREO_CH2 Mux", "DAC3", "DAC3"},

	{"STEREO_CH2 PGA", NULL, "STEREO_CH2 Mux"},
	{"STEREO_CH2", NULL, "STEREO_CH2 PGA"},

	/*TX Mux*/
	{"TX Mux", "AUX1", "AUX1"},
	{"TX Mux", "AUX2", "AUX2"},
	{"TX Mux", "AUX3", "AUX3"},
	{"TX Mux", "DAC1", "DAC1"},
	{"TX Mux", "DAC2", "DAC2"},
	{"TX Mux", "Mic1", "MIC PGA"},
	{"TX Mux", "Mic2", "MIC PGA"},

	{"ADC", NULL, "TX Mux"},
	{"MIC PGA", NULL, "MIC1"},
	{"MIC PGA", NULL, "MIC2"},

	{NULL, NULL, NULL},
};

static const struct snd_kcontrol_new levante_snd_controls[] = {
	SOC_SINGLE("LEVANTE_MISC2",
	LEVANTE_MISC2, 0, 0xff, 0),                 /* 0x42 */

	SOC_SINGLE("LEVANTE_PLL_CTRL1",
	LEVANTE_PLL_CTRL1, 0, 0xff, 0),             /* 0x43 */

	SOC_SINGLE("LEVANTE_PLL_FRAC1",
	LEVANTE_PLL_FRAC1, 0, 0xff, 0),             /* 0x44 */

	SOC_SINGLE("LEVANTE_PLL_FRAC2",
	LEVANTE_PLL_FRAC2, 0, 0xff, 0),             /* 0x45 */

	SOC_SINGLE("LEVANTE_PLL_FRAC3",
	LEVANTE_PLL_FRAC3, 0, 0xff, 0),             /* 0x46 */

	SOC_SINGLE("LEVANTE_PCM_INTERFACE1",
	LEVANTE_PCM_INTERFACE1, 0, 0xff, 0),        /* 0xB0 */

	SOC_SINGLE("LEVANTE_PCM_INTERFACE2",
	LEVANTE_PCM_INTERFACE2, 0, 0xff, 0),        /* 0xB1 */

	SOC_SINGLE("LEVANTE_PCM_INTERFACE3",
	LEVANTE_PCM_INTERFACE3, 0, 0xff, 0),        /* 0xB2 */

	SOC_SINGLE("LEVANTE_PCM_RATE",
	LEVANTE_PCM_RATE, 0, 0xff, 0),              /* 0xB3 */

	SOC_SINGLE("LEVANTE_ECHO_CANCEL_PATH",
	LEVANTE_ECHO_CANCEL_PATH, 0, 0xff, 0),      /* 0xB4 */

	SOC_SINGLE("LEVANTE_SIDETONE_GAIN1",
	LEVANTE_SIDETONE_GAIN1, 0, 0xff, 0),        /* 0xB5 */

	SOC_SINGLE("LEVANTE_SIDETONE_GAIN2",
	LEVANTE_SIDETONE_GAIN2, 0, 0xff, 0),        /* 0xB6 */

	SOC_SINGLE("LEVANTE_SIDETONE_GAIN3",
	LEVANTE_SIDETONE_GAIN3, 0, 0xff, 0),        /* 0xB7 */

	SOC_SINGLE("LEVANTE_ADC_OFFSET1",
	LEVANTE_ADC_OFFSET1, 0, 0xff, 0),           /* 0xB8 */

	SOC_SINGLE("LEVANTE_ADC_OFFSET2",
	LEVANTE_ADC_OFFSET2, 0, 0xff, 0),           /* 0xB9 */

	SOC_SINGLE("LEVANTE_DMIC_DELAY",
	LEVANTE_DMIC_DELAY, 0, 0xff, 0),            /* 0xBA */

	SOC_SINGLE("LEVANTE_I2S_INTERFACE1",
	LEVANTE_I2S_INTERFACE1, 0, 0xff, 0),        /* 0xBB */

	SOC_SINGLE("LEVANTE_I2S_INTERFACE2",
	LEVANTE_I2S_INTERFACE2, 0, 0xff, 0),        /* 0xBC */

	SOC_SINGLE("LEVANTE_I2S_INTERFACE3",
	LEVANTE_I2S_INTERFACE3, 0, 0xff, 0),        /* 0xBD */

	SOC_SINGLE("LEVANTE_I2S_INTERFACE4",
	LEVANTE_I2S_INTERFACE4, 0, 0xff, 0),        /* 0xBE */

	SOC_SINGLE("LEVANTE_EQUALIZER_N0_1",
	LEVANTE_EQUALIZER_N0_1, 0, 0xff, 0),        /* 0xBF */

	SOC_SINGLE("LEVANTE_EQUALIZER_N0_2",
	LEVANTE_EQUALIZER_N0_2, 0, 0xff, 0),        /* 0xC0 */

	SOC_SINGLE("LEVANTE_EQUALIZER_N1_1",
	LEVANTE_EQUALIZER_N1_1, 0, 0xff, 0),        /* 0xC1 */

	SOC_SINGLE("LEVANTE_EQUALIZER_N1_2",
	LEVANTE_EQUALIZER_N1_2, 0, 0xff, 0),        /* 0xC2 */

	SOC_SINGLE("LEVANTE_EQUALIZER_D1_1",
	LEVANTE_EQUALIZER_N1_1, 0, 0xff, 0),        /* 0xC3 */

	SOC_SINGLE("LEVANTE_EQUALIZER_D1_2",
	LEVANTE_EQUALIZER_N1_2, 0, 0xff, 0),        /* 0xC4 */

	SOC_SINGLE("LEVANTE_SIDE_TONE1",
	LEVANTE_SIDE_TONE1, 0, 0xff, 0),            /* 0xC5 */

	SOC_SINGLE("LEVANTE_SIDE_TONE2",
	LEVANTE_SIDE_TONE2, 0, 0xff, 0),            /* 0xC6 */

	SOC_SINGLE("LEVANTE_LEFT_GAIN1",
	LEVANTE_LEFT_GAIN1, 0, 0xff, 0),            /* 0xC7 */

	SOC_SINGLE("LEVANTE_LEFT_GAIN2",
	LEVANTE_LEFT_GAIN2, 0, 0xff, 0),            /* 0xC8 */

	SOC_SINGLE("LEVANTE_RIGHT_GAIN1",
	LEVANTE_RIGHT_GAIN1, 0, 0xff, 0),           /* 0xC9 */

	SOC_SINGLE("LEVANTE_RIGHT_GAIN2",
	LEVANTE_RIGHT_GAIN2, 0, 0xff, 0),           /* 0xCA */

	SOC_SINGLE("LEVANTE_DAC_OFFSET",
	LEVANTE_DAC_OFFSET, 0, 0xff, 0),            /* 0xCB */

	SOC_SINGLE("LEVANTE_OFFSET_LEFT1",
	LEVANTE_OFFSET_LEFT1, 0, 0xff, 0),          /* 0xCC */

	SOC_SINGLE("LEVANTE_OFFSET_LEFT2",
	LEVANTE_OFFSET_LEFT2, 0, 0xff, 0),          /* 0xCD */

	SOC_SINGLE("LEVANTE_OFFSET_RIGHT1",
	LEVANTE_OFFSET_RIGHT1, 0, 0xff, 0),         /* 0xCE */

	SOC_SINGLE("LEVANTE_OFFSET_RIGHT2",
	LEVANTE_OFFSET_RIGHT2, 0, 0xff, 0),         /* 0xCF */

	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM1",
	LEVANTE_ADC_ANALOG_PROGRAM1, 0, 0xff, 0),   /* 0xD0 */

	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM2",
	LEVANTE_ADC_ANALOG_PROGRAM2, 0, 0xff, 0),   /* 0xD1 */

	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM3",
	LEVANTE_ADC_ANALOG_PROGRAM3, 0, 0xff, 0),   /* 0xD2 */

	SOC_SINGLE("LEVANTE_ADC_ANALOG_PROGRAM4",
	LEVANTE_ADC_ANALOG_PROGRAM4, 0, 0xff, 0),   /* 0xD3 */

	SOC_SINGLE("LEVANTE_ANALOG_TO_ANALOG",
	LEVANTE_ANALOG_TO_ANALOG, 0, 0xff, 0),      /* 0xD4 */

	SOC_SINGLE("LEVANTE_HS1_CTRL",
	LEVANTE_HS1_CTRL, 0, 0xff, 0),              /* 0xD5 */

	SOC_SINGLE("LEVANTE_HS2_CTRL",
	LEVANTE_HS2_CTRL, 0, 0xff, 0),              /* 0xD6 */

	SOC_SINGLE("LEVANTE_LO1_CTRL",
	LEVANTE_LO1_CTRL, 0, 0xff, 0),              /* 0xD7 */

	SOC_SINGLE("LEVANTE_LO2_CTRL",
	LEVANTE_LO2_CTRL, 0, 0xff, 0),              /* 0xD8 */

	SOC_SINGLE("LEVANTE_EAR_SPKR_CTRL1",
	LEVANTE_EAR_SPKR_CTRL1, 0, 0xff, 0),        /* 0xD9 */

	SOC_SINGLE("LEVANTE_EAR_SPKR_CTRL2",
	LEVANTE_EAR_SPKR_CTRL2, 0, 0xff, 0),        /* 0xDA */

	SOC_SINGLE("LEVANTE_AUDIO_SUPPLIES1",
	LEVANTE_AUDIO_SUPPLIES1, 0, 0xff, 0),       /* 0xDB */

	SOC_SINGLE("LEVANTE_AUDIO_SUPPLIES2",
	LEVANTE_AUDIO_SUPPLIES2, 0, 0xff, 0),       /* 0xDC */

	SOC_SINGLE("LEVANTE_ADC_ENABLES1",
	LEVANTE_ADC_ENABLES1, 0, 0xff, 0),          /* 0xDD */

	SOC_SINGLE("LEVANTE_ADC_ENABLES2",
	LEVANTE_ADC_ENABLES2, 0, 0xff, 0),          /* 0xDE */

	SOC_SINGLE("LEVANTE_DAC_ENABLES1",
	LEVANTE_DAC_ENABLES1, 0, 0xff, 0),          /* 0xDF */

	SOC_SINGLE("LEVANTE_DAC_DUMMY",
	LEVANTE_DUMMY, 0, 0xff, 0),                 /* 0xE0 */

	SOC_SINGLE("LEVANTE_DAC_ENABLES2",
	LEVANTE_DAC_ENABLES2, 0, 0xff, 0),          /* 0xE1 */

	SOC_SINGLE("LEVANTE_AUDIO_CAL1",
	LEVANTE_AUDIO_CAL1, 0, 0xff, 0),            /* 0xE2 */

	SOC_SINGLE("LEVANTE_AUDIO_CAL2",
	LEVANTE_AUDIO_CAL2, 0, 0xff, 0),            /* 0xE3 */

	SOC_SINGLE("LEVANTE_AUDIO_CAL3",
	LEVANTE_AUDIO_CAL3, 0, 0xff, 0),            /* 0xE4 */

	SOC_SINGLE("LEVANTE_AUDIO_CAL4",
	LEVANTE_AUDIO_CAL4, 0, 0xff, 0),            /* 0xE5 */

	SOC_SINGLE("LEVANTE_AUDIO_CAL5",
	LEVANTE_AUDIO_CAL5, 0, 0xff, 0),            /* 0xE6 */

	SOC_SINGLE("LEVANTE_ANALOG_INPUT_SEL1",
	LEVANTE_ANALOG_INPUT_SEL1, 0, 0xff, 0),     /* 0xE7 */

	SOC_SINGLE("LEVANTE_ANALOG_INPUT_SEL2",
	LEVANTE_ANALOG_INPUT_SEL2, 0, 0xff, 0),     /* 0xE8 */

	SOC_SINGLE("LEVANTE_MIC_BUTTON_DETECTION",
	LEVANTE_MIC_BUTTON_DETECTION, 0, 0xff, 0),  /* 0xE9 */

	SOC_SINGLE("LEVANTE_HEADSET_DETECTION",
	LEVANTE_HEADSET_DETECTION, 0, 0xff, 0),     /* 0xEA */

	SOC_SINGLE("LEVANTE_SHORTS",
	LEVANTE_SHORTS, 0, 0xff, 0)                 /* 0xEB */
};

static int levante_add_controls(struct snd_soc_codec *codec)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(levante_snd_controls); i++) {
		err = snd_ctl_add(codec->card,
			snd_soc_cnew(&levante_snd_controls[i], codec, NULL));
		if (err < 0)
			return err;
	}
	return 0;
}

static int levante_add_widgets(struct snd_soc_codec *codec)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(levante_dapm_widgets); i++)
		snd_soc_dapm_new_control(codec, &levante_dapm_widgets[i]);

	/* set up audio path audio_map nects */
	for (i = 0; levante_audio_map[i][0] != NULL; i++) {
		snd_soc_dapm_connect_input(codec, levante_audio_map[i][0],
			levante_audio_map[i][1], levante_audio_map[i][2]);
	}

	snd_soc_dapm_new_widgets(codec);
	return 0;
}

static unsigned int levante_soc_read(struct snd_soc_codec *codec,
	unsigned int reg)
{
	u8 *cache = codec->reg_cache;

	if (reg > (ARRAY_SIZE(levante_audio_regs)))
		return -EIO;
	return cache[reg];
}

static int levante_soc_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int val)
{
	u8 *cache = codec->reg_cache;

	if (reg > (ARRAY_SIZE(levante_audio_regs))) {
		printk(KERN_ERR "ERROR! levante_soc_write; reg= %d\n", reg);
		return -EIO;
	}

	levante_codec_write(reg, val);
	cache[reg] = val;

	return 0;
}

static int levante_reset(struct snd_soc_codec *codec, int try_warm)
{
	return 0;
}

static int levante_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}

static int levante_cp_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}

static int levante_hifi_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}

static void levante_hifi_shutdown(struct snd_pcm_substream *substream)
{
}

#define MICCO_HIFI_RATES (SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 |\
		SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 | \
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 | \
		SNDRV_PCM_RATE_48000)

#define MICCO_VOICE_RATES (SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | \
		SNDRV_PCM_RATE_32000)

#define MICCO_PCM_FORMATS (SNDRV_PCM_FMTBIT_S16_LE)

struct snd_soc_dai levante_dai[] = {
{
	.name = "I2S HiFi",
	.playback = {
		.stream_name = "HiFi Playback",
		.channels_min = 2,
		.channels_max = 2,
		.rates = MICCO_HIFI_RATES,
		.formats = MICCO_PCM_FORMATS,},
	.ops = {
		.prepare = levante_hifi_prepare,
		.shutdown = levante_hifi_shutdown,},
	},
	{
	.name = "PCM Voice",
	.playback = {
		.stream_name = "Voice Playback",
		.channels_min = 1,
		.channels_max = 1,
		.rates = MICCO_VOICE_RATES,
		.formats = MICCO_PCM_FORMATS,},
	.capture = {
		.stream_name = "Voice Capture",
		.channels_min = 1,
		.channels_max = 1,
		.rates = MICCO_VOICE_RATES,
		.formats = MICCO_PCM_FORMATS,},
	.ops = {
		.prepare = levante_voice_prepare,},
	},
	{
	.name = "CP PCM Voice",
	.playback = {
		.stream_name = "CP Voice Playback",
		.channels_min = 1,
		.channels_max = 1,
		.rates = MICCO_VOICE_RATES,
		.formats = MICCO_PCM_FORMATS,},
	.capture = {
		.stream_name = "CP Voice Capture",
		.channels_min = 1,
		.channels_max = 1,
		.rates = MICCO_VOICE_RATES,
		.formats = MICCO_PCM_FORMATS,},
	.ops = {
		.prepare = levante_cp_voice_prepare,},
	},
};
EXPORT_SYMBOL_GPL(levante_dai);

static int levante_dapm_event(struct snd_soc_codec *codec, int event)
{
	return 0;
}

static int levante_soc_suspend(struct platform_device *pdev,
	pm_message_t state)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

	levante_dapm_event(codec, SNDRV_CTL_POWER_D3cold);
	return 0;
}

static int levante_soc_resume(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;
	u8 *cache = codec->reg_cache;
	unsigned int reg;


	for (reg = 0; reg < ARRAY_SIZE(levante_audio_regs); reg++)
		levante_soc_write(codec, reg, cache[reg]);

	levante_dapm_event(codec, SNDRV_CTL_POWER_D3hot);

	return 0;
}

static ssize_t levante_proc_read(char *page, char **start, off_t off,
			  int count, int *eof, void *data)
{
	char			*buf = page;
	char			*next = buf;
	unsigned		size = count;
	int 			t;
	int 			i;
	int 			reg;
	struct snd_soc_codec 	*codec = data;

	t = scnprintf(next, size, "Levante regs: \n");
	size -= t;
	next += t;

	for (i = 0; i < ARRAY_SIZE(levante_audio_regs); i++) {
		reg = levante_soc_read(codec, i);
		t = scnprintf(next, size, "[0x%02x]=0x%02x  \n", i, reg);
		size -= t;
		next += t;
	}

	*eof = 1;
	return count - size;
}

static int levante_proc_write(struct file *file, const char __user *buffer,
			   unsigned long count, void *data)
{
	static char kbuf[4096];
	char *buf = kbuf;
	struct snd_soc_codec 	*codec = data;
	unsigned int	i, reg, reg2;
	char cmd;

	if (count >= 4096)
		return -EINVAL;
	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	sscanf(buf, "%c 0x%x 0x%x", &cmd, &i, &reg);

	if ('r' == cmd) {
		if (i > 0x16 || (i & 1)) {
			printk(KERN_ERR "invalid index!\n");
			goto error;
		}
		reg = levante_soc_read(codec, i);
		printk(KERN_INFO "0x[%2x]=0x%2x\n", i, reg);
	} else if ('w' == cmd) {
		if (i > 0x16) {
			printk(KERN_ERR "invalid index!\n");
			goto error;
		}
		if (reg > 0xff) {
			printk(KERN_ERR "invalid value!\n");
			goto error;
		}
		levante_soc_write(codec, i, reg);
		reg2 = levante_soc_read(codec, i);
		printk(KERN_INFO
			"write 0x%2x to 0x[%2x], read back 0x%2x\n",
			reg, i, reg2);
	} else {
		printk(KERN_ERR "unknow opt!\n");
		goto error;
	}

	return count;
error:
	printk(KERN_INFO "r/w index(0x%%2x) value(0x%%2x)\n");
	return count;
}

static int levante_soc_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec;
	int ret = 0;
	struct proc_dir_entry *levante_proc_entry;

	printk(KERN_INFO "Levante(DA9034) SoC Audio Codec\n");

	socdev->codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL);
	if (socdev->codec == NULL)
		return -ENOMEM;

	codec = socdev->codec;
	mutex_init(&codec->mutex);

	codec->reg_cache = kzalloc(sizeof(u16) * ARRAY_SIZE(levante_audio_regs),
				GFP_KERNEL);
	if (codec->reg_cache == NULL) {
		ret = -ENOMEM;
		goto cache_err;
	}
	memcpy(codec->reg_cache, levante_audio_regs,
		sizeof(u8) * ARRAY_SIZE(levante_audio_regs));
	codec->reg_cache_size = sizeof(u8) * ARRAY_SIZE(levante_audio_regs);
	codec->reg_cache_step = 1;
	codec->private_data = NULL;
	codec->name = "LEVANTE";
	codec->owner = THIS_MODULE;
	codec->dai = levante_dai;
	codec->num_dai = ARRAY_SIZE(levante_dai);
	codec->write = levante_soc_write;
	codec->read = levante_soc_read;
	/* codec->dapm_event = levante_dapm_event; */
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	/* register pcms */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0)
		goto pcm_err;

	ret = levante_reset(codec, 0);
	if (ret < 0) {
		printk(KERN_ERR "Reset Levante error\n");
		goto reset_err;
	}
	levante_dapm_event(codec, SNDRV_CTL_POWER_D3hot);

	if (machine_is_benzina2() || machine_is_tavorevb() || machine_is_saar() || 
        machine_is_tavor_evb3() || machine_is_bigbrother() || machine_is_saarb())
	{
		printk(KERN_ERR "Audio: Levante SOC Probe TRUE; supported machine type!\n");
		levante_add_direct_access(codec);
		levante_id_access(codec);
	}
	else
	{
		printk(KERN_ERR "Audio: Levante SOC Probe FALSE; non-supported machine type! ==> ABORT!\n");
		levante_add_controls(codec);
		levante_add_widgets(codec);
	}

	ret = snd_soc_register_card(socdev);
	if (ret < 0)
		goto reset_err;

	levante_proc_entry = create_proc_entry("driver/codec", 0, NULL);
	if (levante_proc_entry) {
		levante_proc_entry->data = codec;
		levante_proc_entry->read_proc = levante_proc_read;
		levante_proc_entry->write_proc = levante_proc_write;
	}

	return 0;

reset_err:
	snd_soc_free_pcms(socdev);

pcm_err:
	snd_soc_free_ac97_codec(codec);

cache_err:
	kfree(socdev->codec);
	socdev->codec = NULL;
	return ret;
}

static int levante_soc_remove(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

	if (codec == NULL)
		return 0;

	snd_soc_dapm_free(socdev);
	snd_soc_free_pcms(socdev);
	snd_soc_free_ac97_codec(codec);
	kfree(codec->reg_cache);
	kfree(codec->dai);
	kfree(codec);
	return 0;
}
struct snd_soc_codec_device soc_codec_dev_levante = {
	.probe = 	levante_soc_probe,
	.remove = 	levante_soc_remove,
	.suspend =	levante_soc_suspend,
	.resume = 	levante_soc_resume,
};
EXPORT_SYMBOL_GPL(soc_codec_dev_levante);

MODULE_DESCRIPTION("ASoC Levante driver");
MODULE_AUTHOR("bin.yang@marvell.com");
MODULE_LICENSE("GPL");


