/*
 * linux/sound/soc/pxa/bigbrother.c
 *
 * Copyright (C) 2007 Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/mach-types.h>
#include <linux/io.h>

#include <mach/mfp-pxa9xx.h>
#include <mach/mfp-pxa3xx.h>
#include <mach/pxa-regs.h>
#include <mach/hardware.h>
#include <mach/audio.h>
#include <mach/regs-ssp.h>
#include <mach/ssp.h>
#include <mach/tavor_common.h>
#include <mach/levante_micco_pmic_id.h>
#include <mach/gpio.h>

#include "pxa3xx-pcm.h"
#include "pxa3xx-ssp.h"

static u8 g_pmic_id;

#define PCM_SSP_PORT	3	  /* used as ap pcm port  */
#define I2S_SSP_PORT	2	  /* used as ap hifi port */
#define PCM_CP_PORT	1	  /* used as cp pcm prot  */

#define ARRAY_AND_SIZE(x)       (x), ARRAY_SIZE(x)

void enable_oscc_pout(void);
void disable_oscc_pout(void);

extern struct snd_soc_dai levante_dai[3];
extern struct snd_soc_codec_device soc_codec_dev_levante;

static void set_dvfm_constraint(void) {}
static void unset_dvfm_constraint(void) {}

static int bigbrother_levante_init(struct snd_soc_codec *codec)
{
	return 0;
}

static int bigbrother_levante_hifi_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	unsigned long sscr0;

	printk("bigbrother_hifi_startup \n");

	set_dvfm_constraint();

	cpu_dai->playback.channels_min = 2;
	cpu_dai->playback.channels_max = 2;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |\
		SNDRV_PCM_RATE_16000;

	enable_oscc_pout();

	__raw_writel(0x415000BF, ssp->mmio_base + SSCR0);
	__raw_writel(0x03201DC0, ssp->mmio_base + SSCR1);
	__raw_writel(0x02100004, ssp->mmio_base + SSPSP);

	/* enable packing mode for cpu_dai->hw_params judge*/
	sscr0 = SSCR0_FPCKE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	return 0;
}

static int bigbrother_levante_hifi_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp, sstsa;
	unsigned long ssacd, ssacdd, ssrsa;

	printk("bigbrother_levante_hifi_prepare \n");

	/* Because the internal 13M clock will be 10M in D0CS,
	 * we route SSP_CLK to GPIO126(EXT_CLK) and let SSP select
	 * NETWORK CLK as CLK source.
	 * This workaround need an ECO on Littleton mainboard.
	 */

	sscr0 = 0x415000BF;
	sscr1 = 0x03201DC0;
	sspsp = 0x02100004;
	sstsa = 0x00000003;
	ssrsa = 0x00000003;

	ssacd = 0x60;
	ssacdd = 0x00000040;

	switch (rate) {
	case 48000:
		ssacdd |= (0x659 << 16);
		break;
	case 44100:
		ssacdd |= (0x6E9 << 16);
		break;
	case 16000:
		ssacdd |= (0x130B << 16);
		break;
	default:
		return -EINVAL;
	}

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);
	__raw_writel(sstsa, ssp->mmio_base + SSTSA);
	__raw_writel(ssrsa, ssp->mmio_base + SSRSA);
	__raw_writel(ssacd, ssp->mmio_base + SSACD);
	__raw_writel(ssacdd, ssp->mmio_base + SSACDD);

	return 0;
}

static void bigbrother_levante_hifi_shutdown(struct snd_pcm_substream *substream)
{
	printk("bigbrother_levante_hifi_shutdown \n");

	disable_oscc_pout();
	unset_dvfm_constraint();
}

static int bigbrother_levante_voice_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;

	set_dvfm_constraint();

	printk(KERN_ALERT "----->bigbrother_levante_voice_startup \n");


	cpu_dai->playback.channels_min = 1;
	cpu_dai->playback.channels_max = 1;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | \
		SNDRV_PCM_RATE_32000;

	enable_oscc_pout();

	return 0;
}

static int bigbrother_levante_voice_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp;

	printk(KERN_ALERT "----->bigbrother_levante_voice_prepare \n");

	cpu_dai->dai_ops.set_sysclk(cpu_dai, PXA3XX_SSP_CLK_PLL, 13000000, 0);

	sscr0 = 0x00C0003F;
	sscr1 = 0x00701DC0;
	sspsp = 0x00800085;

	switch (rate) {
	case 8000:
		sscr0 |= (89 << 8);
		break;
	case 16000:
		sscr0 |= (44 << 8);
		break;
	case 32000:
		sscr0 |= (22 << 8);
		break;
	default:
		return -EINVAL;
	}

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		sscr1 &= ~0x00800000;
	} else {
		sscr1 |= 0x00800000;
	}

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);

	return 0;
}

static void bigbrother_levante_voice_shutdown(struct snd_pcm_substream *substream)
{
	disable_oscc_pout();
	unset_dvfm_constraint();
}

static int bigbrother_levante_cp_voice_startup(struct snd_pcm_substream *substream)
{
	enable_oscc_pout();
	return 0;
}
static int bigbrother_levante_cp_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}
static void bigbrother_levante_cp_voice_shutdown(struct snd_pcm_substream *substream)
{
	disable_oscc_pout();
}


/* machine stream operations */
static struct snd_soc_ops bigbrother_machine_ops_levante[] = {
{
	.startup = bigbrother_levante_hifi_startup,
	.prepare = bigbrother_levante_hifi_prepare,
	.shutdown = bigbrother_levante_hifi_shutdown,
},
{
	.startup = bigbrother_levante_voice_startup,
	.prepare = bigbrother_levante_voice_prepare,
	.shutdown = bigbrother_levante_voice_shutdown,
},
{
	.startup = bigbrother_levante_cp_voice_startup,
	.prepare = bigbrother_levante_cp_voice_prepare,
	.shutdown = bigbrother_levante_cp_voice_shutdown,
},
};

/* bigbrother digital audio interface glue - connects codec <--> CPU */
static struct snd_soc_dai_link bigbrother_dai_levante[] = {
{
	.name = "I2S",
	.stream_name = "I2S HiFi",
	.cpu_dai = &pxa3xx_ssp_dai[I2S_SSP_PORT-1],
	.codec_dai = &levante_dai[0],
	.ops = &bigbrother_machine_ops_levante[0],
	.init = bigbrother_levante_init,
},
{
	.name = "PCM",
	.stream_name = "PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[PCM_SSP_PORT-1],
	.codec_dai = &levante_dai[1],
	.ops = &bigbrother_machine_ops_levante[1],
	.init = bigbrother_levante_init,
},
{
	.name = "CP PCM",
	.stream_name = "CP PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[4],
	.codec_dai = &levante_dai[2],
	.ops = &bigbrother_machine_ops_levante[2],
	.init = bigbrother_levante_init,
},
};

/* bigbrother audio machine driver */
static struct snd_soc_machine snd_soc_machine_bigbrother_levante = {
	.name = "bigbrother",
	.dai_link = bigbrother_dai_levante,
	.num_links = ARRAY_SIZE(bigbrother_dai_levante),
};

/* bigbrother audio subsystem */
static struct snd_soc_device bigbrother_snd_devdata_levante = {
	.machine = &snd_soc_machine_bigbrother_levante,
	.platform = &pxa3xx_soc_platform,
	.codec_dev = &soc_codec_dev_levante,
};

static struct platform_device *bigbrother_snd_device_levante;

/*
 * Logic for a Levante as connected on a bigbrother Device
 */

static int __init bigbrother_init(void)
{
	int ret;

	printk("*** Audio: bigbrother_init; machine_arch_type= %d (MACH_TYPE_BIGBROTHER= %d) ***\n", machine_arch_type, MACH_TYPE_BIGBROTHER);
	if (!machine_is_bigbrother())
	{
		printk("*** Audio: bigbrother_init; machine is NOT 'MACH_TYPE_BIGBROTHER' ==> ABORT! ***\n");
		return -ENODEV;
	}

   	g_pmic_id = pmic_id();

	if(g_pmic_id >= 0x40)
	{  /* LEVANTE */
		printk("*** Audio: bigbrother_init; g_pmic_id= 0x%x ==> LEVANTE! ***\n", g_pmic_id);
		bigbrother_snd_device_levante = platform_device_alloc("soc-audio", -1);
		if (!bigbrother_snd_device_levante)
		{
			printk("ERROR in LEVANTE device alloc!");
			return -ENOMEM;
		}

		platform_set_drvdata(bigbrother_snd_device_levante, &bigbrother_snd_devdata_levante);
		bigbrother_snd_devdata_levante.dev = &bigbrother_snd_device_levante->dev;
		ret = platform_device_add(bigbrother_snd_device_levante);

		if(ret)
		{
			printk("ERROR in LEVANTE device add!");
			platform_device_put(bigbrother_snd_device_levante);
		}
	}
	else
	{  /* MICCO */
		printk("*** Audio: bigbrother_init; g_pmic_id= 0x%x ==> MICCO! ==> ERROR!!! Micco is not supported! ***\n", g_pmic_id);
		ret = -ENODEV;
	}

	return ret;
}

static void __exit bigbrother_exit(void)
{
	platform_device_unregister(bigbrother_snd_device_levante);
}

module_init(bigbrother_init);
module_exit(bigbrother_exit);

/* Module information */
MODULE_AUTHOR("mwwang@marvell.com");
MODULE_DESCRIPTION("ALSA SoC LEVANTE BIGBROTHER");
MODULE_LICENSE("GPL");
