/*
 * linux/sound/soc/pxa/fbw_tavorp65.c
 *
 * Copyright (C) 2007 Marvell International Ltd.
 *
 * Copyright (C) 2008 CMCS, Inc. <maskwcchung@tp.cmcs.com.tw>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
//#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/mach-types.h>
#include <asm/io.h>

#include <mach/mfp-pxa9xx.h>
#include <mach/pxa-regs.h>
#include <mach/hardware.h>
#include <mach/pxa3xx-regs.h>
#include <mach/audio.h>
#include <mach/regs-ssp.h>
#include <mach/ssp.h>
#include <mach/micco.h>
#include <mach/gpio.h>

#include "pxa3xx-pcm.h"
#include "pxa3xx-ssp.h"
#if defined(CONFIG_PXA3xx_DVFM)
#include "mach/dvfm.h"
#endif
#include <linux/wakelock.h>

#define SOUND_SSP_DITHER_AUDIO_CLOCK

#define PCM_SSP3_PORT   3  /* used as AP SSP3 port  */
#define I2S_SSP2_PORT   2  /* used as AP SSP2 port */
#define PCM_GSSP1_PORT  1  /* used as CP GSSP1 prot  */

#define ARRAY_AND_SIZE(x)       (x), ARRAY_SIZE(x)

mfp_cfg_t pcm_gssp1_mfp_cfg[4];
mfp_cfg_t pcm_ssp3_mfp_cfg[4];
mfp_cfg_t pcm_off_mfp_cfg[4];
mfp_cfg_t i2s_ssp2_mfp_cfg[3];
mfp_cfg_t i2s_off_mfp_cfg[3];

static unsigned int wake_lock_count = 0;
static struct wake_lock  sound_suspend_lock;

static char * gpio_label[3][4] = 
{
    {"GSSP1 BITCLK", "GSSP1 SYNC", "GSSP1 TXD", "GSSP1 RXD"},
    {"SSP2 BITCLK", "SSP2 SYNC", "SSP2 DOUT", },
    {"SSP3 BITCLK", "SSP3 SYNC", "SSP3 TXD", "SSP3 RXD"},
};

static unsigned int ssp_flag = 0;
static unsigned char is_ssp_open(unsigned int port)
{
    if ( ssp_flag & (1<<port))
    {
        return 1;
    }
    else
    {
        return 0;
    }
}
static int snd_ssp_set(unsigned int ssp, unsigned int on)
{
    static u32 gpio_pin;
    int err, i;

    if (is_ssp_open(ssp) == on) {
        return 0;
    }

    if (on == 0)
    {
        ssp_flag &= ~(1<<ssp);
    }
    else
    {
        ssp_flag |= (1<<ssp);
    }

    if (ssp == PCM_GSSP1_PORT)
    {
        if (!on){
            pxa3xx_mfp_config(ARRAY_AND_SIZE(pcm_off_mfp_cfg));
            for(i = 0; i< ARRAY_SIZE(pcm_off_mfp_cfg); i++)
            {
                gpio_pin = mfp_to_gpio(MFP_CFG_PIN(pcm_off_mfp_cfg[i]));
                err = gpio_request(gpio_pin, gpio_label[ssp-1][i]);
                if (err){
                    printk(KERN_ERR "Request %s GPIO failed!", gpio_label[ssp-1][i]);
                    continue;
                }
                gpio_direction_input(gpio_pin);
                gpio_free(gpio_pin);
            }
        }
        else{
            pxa3xx_mfp_config(ARRAY_AND_SIZE(pcm_gssp1_mfp_cfg));
        }
    }

    if (ssp == I2S_SSP2_PORT)
    {
        if (!on){
            pxa3xx_mfp_config(ARRAY_AND_SIZE(i2s_off_mfp_cfg));
            for(i = 0; i< ARRAY_SIZE(i2s_off_mfp_cfg); i++)
            {
                gpio_pin = mfp_to_gpio(MFP_CFG_PIN(i2s_off_mfp_cfg[i]));
                err = gpio_request(gpio_pin, gpio_label[ssp-1][i]);
                if (err){
                    printk(KERN_ERR "Request %s GPIO failed!", gpio_label[ssp-1][i]);
                    continue;
                }
                gpio_direction_input(gpio_pin);
                gpio_free(gpio_pin);
            }
        }
        else{
            pxa3xx_mfp_config(ARRAY_AND_SIZE(i2s_ssp2_mfp_cfg));
        }
    }

    if (ssp == PCM_SSP3_PORT)
    {
        if (!on){
            pxa3xx_mfp_config(ARRAY_AND_SIZE(pcm_off_mfp_cfg));
            for(i = 0; i< ARRAY_SIZE(pcm_off_mfp_cfg); i++)
            {
                gpio_pin = mfp_to_gpio(MFP_CFG_PIN(pcm_off_mfp_cfg[i]));
                err = gpio_request(gpio_pin, gpio_label[ssp-1][i]);
                if (err){
                    printk(KERN_ERR "Request %s GPIO failed!", gpio_label[ssp-1][i]);
                    continue;
                }
                gpio_direction_input(gpio_pin);
                gpio_free(gpio_pin);
            }
        }
        else{
            pxa3xx_mfp_config(ARRAY_AND_SIZE(pcm_ssp3_mfp_cfg));
        }
    }
	
    return 0;
}

static void fbwp65_wake_lock(unsigned int on)
{
    if (on){
        if(!wake_lock_count++)
        {
            wake_lock(&sound_suspend_lock);
        }
    }
    else{
        if(!wake_lock_count)
        {
            printk(KERN_ERR "fbwp65_wake_lock count had been zero\n");
        }
        else if(!--wake_lock_count) 
        {
            wake_unlock(&sound_suspend_lock);
        }
    }
}

extern void enable_oscc_pout(void);
extern void disable_oscc_pout(void);

extern struct snd_soc_dai micco_dai[3];
extern struct snd_soc_codec_device soc_codec_dev_micco;

#if defined(CONFIG_PXA3xx_DVFM)

static int dvfm_dev_idx = 0;

static void set_dvfm_constraint(unsigned int port)
{
    if (port == I2S_SSP2_PORT){
        dvfm_disable_op_name("D1", dvfm_dev_idx);
	    dvfm_disable_op_name("D2", dvfm_dev_idx);
        if (cpu_is_pxa935())
            dvfm_disable_op_name("CG", dvfm_dev_idx);
    }
    else if (port == PCM_SSP3_PORT){
        dvfm_disable_op_name("D0CS", dvfm_dev_idx);
        dvfm_disable_op_name("D1", dvfm_dev_idx);
        dvfm_disable_op_name("D2", dvfm_dev_idx);
        if (cpu_is_pxa935())
            dvfm_disable_op_name("CG", dvfm_dev_idx);
    }
}

static void unset_dvfm_constraint(unsigned int port)
{
    if (port == I2S_SSP2_PORT){
        if (is_ssp_open(PCM_SSP3_PORT)){
            return;
        }
        else{
            dvfm_enable_op_name("D1", dvfm_dev_idx);
	        dvfm_enable_op_name("D2", dvfm_dev_idx);
            if (cpu_is_pxa935())
                dvfm_enable_op_name("CG", dvfm_dev_idx);
        }
    }
    else if (port == PCM_SSP3_PORT){
        if (is_ssp_open(I2S_SSP2_PORT)){
            dvfm_enable_op_name("D0CS", dvfm_dev_idx);
        }
        else{
            dvfm_enable_op_name("D0CS", dvfm_dev_idx);
            dvfm_enable_op_name("D1", dvfm_dev_idx);
            dvfm_enable_op_name("D2", dvfm_dev_idx);
            if (cpu_is_pxa935())
                dvfm_enable_op_name("CG", dvfm_dev_idx);
        }
    }
}
#else
static void set_dvfm_constraint(int) {}
static void unset_dvfm_constraint(int) {}
#endif

static int fbwp65_micco_init(struct snd_soc_codec *codec);

static int fbwp65_micco_hifi_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	unsigned long sscr0;

    snd_ssp_set(I2S_SSP2_PORT, 1);
	set_dvfm_constraint(I2S_SSP2_PORT);
    fbwp65_wake_lock(1);

	cpu_dai->playback.channels_min = 2;
	cpu_dai->playback.channels_max = 2;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |\
		SNDRV_PCM_RATE_16000;

	enable_oscc_pout();

	/* enable packing mode for cpu_dai->hw_params judge*/
	sscr0 = SSCR0_FPCKE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	return 0;
}

static int fbwp65_micco_hifi_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp, sstsa;
	unsigned long ssacd, ssacdd, ssrsa;

	/* Because the internal 13M clock will be 10M in D0CS,
	 * we set Tavor as SSP slaver.
	 */

	sscr0 = 0xE1C0003F;
	sscr1 = 0x02401DC0;
	sspsp = 0x31A00080;
	sstsa = 0x00000003;
	ssrsa = 0x00000003;

#ifdef SOUND_SSP_DITHER_AUDIO_CLOCK
	ssacd = 0x60;
	ssacdd = 0x00000040;

	switch (rate) {
	case 48000:
		ssacdd |= (0x659 << 16);
		break;
	case 44100:
		ssacdd |= (0x6E9 << 16);
		break;
	case 16000:
		ssacdd |= (0x130B << 16);
		break;
	default:
		return -EINVAL;
	}
#else
	ssacd = 0x0;
	ssacdd = 0x0;

	switch (rate) {
	case 48000:
		ssacd |= (0 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (2 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	case 44100:
		ssacd |= (0 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (1 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	case 22050:
		ssacd |= (0 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (0 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	case 16000:
		ssacd |= (3 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (4 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	case 11025:
		ssacd |= (1 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (0 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	case 8000:
		ssacd |= (4 << 0) /* ACDS */
			| (0 << 3) /* SCDB */
			| (4 << 4) /* ACPS */
			| (0 << 7); /* SCDX8 */
		break;
	default:
		return -EINVAL;
	}
#endif
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);
	__raw_writel(sstsa, ssp->mmio_base + SSTSA);
	__raw_writel(ssrsa, ssp->mmio_base + SSRSA);
	__raw_writel(ssacd, ssp->mmio_base + SSACD);
	__raw_writel(ssacdd, ssp->mmio_base + SSACDD);

	return 0;
}

static void fbwp65_micco_hifi_shutdown(struct snd_pcm_substream *substream)
{
	disable_oscc_pout();

	snd_ssp_set(I2S_SSP2_PORT, 0);
    unset_dvfm_constraint(I2S_SSP2_PORT);

    fbwp65_wake_lock(0);
}

static int fbwp65_micco_voice_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
    
    snd_ssp_set(PCM_GSSP1_PORT, 0);
    snd_ssp_set(PCM_SSP3_PORT, 1);
	set_dvfm_constraint(PCM_SSP3_PORT);
    fbwp65_wake_lock(1);

	cpu_dai->playback.channels_min = 1;
	cpu_dai->playback.channels_max = 1;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | \
		SNDRV_PCM_RATE_32000;

	enable_oscc_pout();

	return 0;
}

static int fbwp65_micco_voice_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp;

	cpu_dai->dai_ops.set_sysclk(cpu_dai, PXA3XX_SSP_CLK_PLL, 13000000, 0);

	sscr0 = 0x00C0003F;
	sscr1 = 0x00701DC0;
	sspsp = 0x00800085;

	switch (rate) {
	case 8000:
		sscr0 |= (89 << 8);
		break;
	case 16000:
		sscr0 |= (44 << 8);
		break;
	case 32000:
		sscr0 |= (22 << 8);
		break;
	default:
		return -EINVAL;
	}

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		sscr1 &= ~0x00800000;
	} else {
		sscr1 |= 0x00800000;
	}

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);

	return 0;
}

static void fbwp65_micco_voice_shutdown(struct snd_pcm_substream *substream)
{
	disable_oscc_pout();

    snd_ssp_set(PCM_SSP3_PORT, 0);
    unset_dvfm_constraint(PCM_SSP3_PORT);

    fbwp65_wake_lock(0);
}

static int fbwp65_micco_cp_voice_startup(struct snd_pcm_substream *substream)
{
    snd_ssp_set(PCM_SSP3_PORT, 0);
    snd_ssp_set(PCM_GSSP1_PORT, 1);

	enable_oscc_pout();
	return 0;
}
static int fbwp65_micco_cp_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}
static void fbwp65_micco_cp_voice_shutdown(struct snd_pcm_substream *substream)
{
    disable_oscc_pout();

    snd_ssp_set(PCM_GSSP1_PORT, 0);
}

/* machine stream operations */
static struct snd_soc_ops fbwp65_machine_ops[] = {
{
	.startup = fbwp65_micco_hifi_startup,
	.prepare = fbwp65_micco_hifi_prepare,
	.shutdown = fbwp65_micco_hifi_shutdown,
},
{
	.startup = fbwp65_micco_voice_startup,
	.prepare = fbwp65_micco_voice_prepare,
	.shutdown = fbwp65_micco_voice_shutdown,
},
{
	.startup = fbwp65_micco_cp_voice_startup,
	.prepare = fbwp65_micco_cp_voice_prepare,
	.shutdown = fbwp65_micco_cp_voice_shutdown,
},
};

/* fbwp65 digital audio interface glue - connects codec <--> CPU */
static struct snd_soc_dai_link fbwp65_dai[] = {
{
	.name = "I2S",
	.stream_name = "I2S HiFi",
	.cpu_dai = &pxa3xx_ssp_dai[I2S_SSP2_PORT-1],
	.codec_dai = &micco_dai[0],
	.ops = &fbwp65_machine_ops[0],
	.init = fbwp65_micco_init,
},
{
	.name = "PCM",
	.stream_name = "PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[PCM_SSP3_PORT-1],
	.codec_dai = &micco_dai[1],
	.ops = &fbwp65_machine_ops[1],
	.init = fbwp65_micco_init,
},
{
	.name = "CP PCM",
	.stream_name = "CP PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[4],
	.codec_dai = &micco_dai[2],
	.ops = &fbwp65_machine_ops[2],
	.init = fbwp65_micco_init,
},
};

/* fbwp65 audio machine driver */
static struct snd_soc_machine snd_soc_machine_fbwp65 = {
	.name = "fbw_tavorp65",
	.dai_link = fbwp65_dai,
	.num_links = ARRAY_SIZE(fbwp65_dai),
};

/* fbwp65 audio subsystem */
static struct snd_soc_device fbwp65_snd_devdata = {
	.machine = &snd_soc_machine_fbwp65,
	.platform = &pxa3xx_soc_platform,
	.codec_dev = &soc_codec_dev_micco,
};

static struct platform_device *fbwp65_snd_device;

/*
 * Logic for a Micco as connected on a fbwp65 Device
 */
static int fbwp65_micco_init(struct snd_soc_codec *codec)
{
	snd_ssp_set(I2S_SSP2_PORT, 0);
    return 0;
}

static int __init fbwp65_init(void)
{
	int ret;

	if (!machine_is_benzina() && !machine_is_benzglb() && !machine_is_blaze() && !machine_is_spark())
		return -ENODEV;

	fbwp65_snd_device = platform_device_alloc("soc-audio", -1);
	if (!fbwp65_snd_device)
		return -ENOMEM;

	platform_set_drvdata(fbwp65_snd_device, &fbwp65_snd_devdata);
	fbwp65_snd_devdata.dev = &fbwp65_snd_device->dev;
	ret = platform_device_add(fbwp65_snd_device);

	if (ret)
		platform_device_put(fbwp65_snd_device);

	pxa_set_cken(CKEN_SSP2, 0);
	pxa_set_cken(CKEN_SSP3, 0);

#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("Sound", &dvfm_dev_idx);
#endif

    wake_lock_init(&sound_suspend_lock, WAKE_LOCK_SUSPEND, "sound");

	return ret;
}

static void __exit fbwp65_exit(void)
{
	platform_device_unregister(fbwp65_snd_device);

    wake_lock_destroy(&sound_suspend_lock);

#ifdef CONFIG_PXA3xx_DVFM
   	dvfm_unregister("Sound", &dvfm_dev_idx);
#endif

}

module_init(fbwp65_init);
module_exit(fbwp65_exit);

/* Module information */
MODULE_AUTHOR("shawn.bow@borqs.com, Shawn Bow <shawn.bow@borqs.com>");
MODULE_DESCRIPTION("ALSA SoC MICCO FBW TavorP65");
MODULE_LICENSE("GPL");
