/*
 *  Copyright (C) 2008-2009 Foxconn K.H. Fan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/firmware.h>
#include <asm/io.h>
#include <mach/gpio.h>
#include <mach/pxa3xx-aec-fm2010.h>
#include <linux/mutex.h>
// major & minor define area
#define AEC_MAJOR       18                   /* dynamic major by default */
#define AEC_MINOR       5

#define SYNC1			0xfc
#define SYNC2			0xf3
#define MEM_WRITE		0x3B
#define MEM_READ		0x37
#define REG_READ		0x60
#define REG_WRITE		0x6A

#ifdef AEC_MAJOR
static int aec_major =   AEC_MAJOR;
#else
static int aec_major =   0;
#endif

#ifdef AEC_MINOR
static int aec_minor =   AEC_MINOR;
#else
static int aec_minor =   0;
#endif

static int aec_nr_dev = 1;
static int g_mode = AEC_NORMAL_MODE;

extern void enable_oscc_pout(void);
extern void disable_oscc_pout(void);
static int aec_devPwrCtrl(int on);

static struct delayed_work g_resetWork;
static struct workqueue_struct *g_pWqueue;
static char g_aec_restartOn = 0;

static DEFINE_MUTEX(g_mutex);

/* firmware */
/* for handsfree */
static const u8 fw_normal_mode[] = {
	0x1E,0x32,0x00,0x01,
	0x1E,0x34,0x00,0x88,
	0x1E,0x3D,0x03,0x00,
	0x1E,0x41,0x01,0x01,
	0x1E,0x45,0x03,0x0F,
	0x1E,0x46,0x00,0x11,
	0x1E,0x47,0x20,0x00,
	0x1E,0x48,0x09,0x00,
	0x1E,0x49,0x09,0x00,
	0x1E,0x4A,0x00,0x0E,
	0x1E,0x4D,0x01,0x00,
	0x1E,0x60,0x00,0x00,
	0x1E,0x63,0x00,0x01,
	0x1E,0x70,0x05,0xC0,
	0x1E,0x86,0x00,0x08,
	0x1E,0x87,0x00,0x03,
	0x1E,0x88,0x58,0x00,
	0x1E,0x89,0x00,0x10,
	0x1E,0x8B,0x02,0x80,
	0x1E,0x8C,0x00,0x20,
	0x1E,0x91,0x00,0x02,
	0x1E,0x92,0x20,0x00,
	0x1E,0x93,0x20,0x00,
	0x1E,0xC7,0x18,0x00,
	0x1E,0xC8,0x48,0x00,
	0x1E,0xF8,0x18,0x00,
	0x1E,0xF9,0x03,0x00,
	0x1E,0xFF,0x52,0x00,
	0x1E,0x3A,0x00,0x00,
	0x1e, 0x3a, 0x00, 0x00  /* end */
};
#if 0
/* for handset and recording */
static const u8 fw_bypass_mode[] = {
	0x1e, 0x34, 0x00, 0x0e,
	0x1e, 0x41, 0x01, 0x01,
	0x1e, 0x4a, 0x00, 0x0e,
	0x1e, 0x60, 0x00, 0x00,
	0x1e, 0x70, 0x05, 0xc0,
	0x1e, 0x3a, 0x00, 0x00  /* end */
};
#endif
#if 1 //def CTA
static const u8 fw_bypass_mode[] = {
	0x1e, 0x34, 0x00, 0x98,
	0x1e, 0x41, 0x01, 0x01,
	0x1e, 0x4a, 0x00, 0x0E,
	0x1e, 0x4c, 0x00, 0x1F,
	0x1e, 0x60, 0x00, 0x00,
	0x1e, 0x70, 0x05, 0xc0,
	0x1e, 0x3a, 0x00, 0x00  /* end */
};
#endif

static int fw_bypass_len = sizeof(fw_bypass_mode)>>2;
static int fw_normal_len = sizeof(fw_normal_mode)>>2;

static void ghost_DevRelease(struct device *dev)
{
	
}

static struct device ghost_device = {
	.bus_id    = "aec_param",
	.release   = ghost_DevRelease,
};

struct i2c_client *gAecClient = NULL;

static int aec_softReset(void)
{
	char buf[10];
	int ret;

	printk("[AEC]aec softReset\n");

	buf[0] = 0xfc;
	buf[1] = 0xf3;
	buf[2] = 0x6a;
	buf[3] = 0x2a;
	buf[4] = 0x00;
	buf[5] = 0x30;

	ret = i2c_master_send(gAecClient, buf, 6);
	if (ret !=6)
	{
		printk("aec send data error\n");
		return -EIO;
	}

	msleep(100);

	return 0;
}


/*
	read one byte from the aec chip's register(0x25 or 0x26), the register is 1byte align;
	return 0, success
*/

static int aec_data_read(u8 reg, u8 *data)
{
	char buf[10];
	int ret;

	if(reg != 0x25 && reg != 0x26){
		printk(KERN_WARNING"error register addr %x\n", reg);	
		return -EIO;
	}
	buf[0] = 0xfc;
	buf[1] = 0xf3;
	buf[2] = 0x60;
	buf[3] = reg;
	ret = i2c_master_send(gAecClient, buf, 4);
	if (ret !=4) {
		printk("aec send data error\n");
		return -EIO;
	}
	ret = i2c_master_recv (gAecClient,  buf, 1);
	if (ret <=0) {
		printk("aec send data error\n");
		return -EIO;
	}
	*data = buf[0];
	return 0;
}

/*
	send two byte to the aec chip's memory, the memory is 2bytes align;
	return 0, success
	-1, fails
*/
static int aec_mem_write(unsigned short mem_addr, unsigned short data)
{
	char aec_buf[10];
	int ret;
/*
	if ((mem_addr < 0x1e30) || (mem_addr > 0x1e40))
		return -1;
*/
	//two sync byte
	aec_buf[0] = 0xfc;
	aec_buf[1] = 0xf3; 
	//command byte
	aec_buf[2] = 0x3b;

	//fm2010's memory addr
	aec_buf[3] = (mem_addr>>8) & 0xff;
	aec_buf[4] = mem_addr & 0xff;

	//the data to the memory
	aec_buf[5] = (data>>8) & 0xff;
	aec_buf[6] = data & 0xff;
	
	ret = i2c_master_send(gAecClient, aec_buf, 7);
	if (ret != 7)
	{
        printk("aec aec_mem_write fail, ret=%d\n", ret);
		return -EIO;
    }
	return 0;
}

static int find_fw_max_len(void)
{
	int tmp1, tmp2;

	tmp1 = sizeof(fw_bypass_mode)>>2;
	tmp2 = sizeof(fw_normal_mode)>>2;

	return ((tmp1 > tmp2) ? tmp1 : tmp2);
}

static void construct_command(u8 *buf, const u8 *data, int len)
{
	int i;

	while (len>0) {
		*(buf++) = SYNC1;
		*(buf++) = SYNC2;
		*(buf++) = MEM_WRITE;
		for (i = 0; i < 4; ++i,++buf,++data)
			*buf = *data;
		len--;
	}
}

static int fm2018_write(u8 *buf, int len)
{
	int err;

	if (!buf)
		return -EINVAL;
	err = i2c_master_send(gAecClient, buf, len);
	if (err < 0)
    {
        printk(KERN_ERR "[FM2018] fm2018 write error!");
		return err;
    }

	return 0;
}

/*
	read two byte to the aec chip's memory, the memory is 2bytes align;
	return 1, success
	0, fails
*/
static int aec_mem_read(unsigned short mem_addr, unsigned short *data)
{
	char aec_buf[10];
	char data_buf[2] = {0};
	int ret;
/*
	if ((mem_addr < 0x1e30) || (mem_addr > 0x1e40))
		return -EIO;
*/
	//two sync byte
	aec_buf[0] = 0xfc;
	aec_buf[1] = 0xf3; 
	//command byte
	aec_buf[2] = 0x37;

	//fm2010's memory addr
	aec_buf[3] = (mem_addr>>8) & 0xff;
	aec_buf[4] = mem_addr & 0xff;
	
	ret = i2c_master_send(gAecClient, aec_buf, 5);
	if (ret != 5) {
		return -EIO;
	}
	
	ret = aec_data_read(0x26, &data_buf[0]);
	if (ret) {
		printk("aec reg %x read error\n", 0x26);
		return -EIO;
	}
	ret = aec_data_read(0x25, &data_buf[1]);
	if (ret) {
		printk("aec reg %x read error\n", 0x25);
		return -EIO;
	}

	*data = data_buf[0];
	*data = (*data << 8) | data_buf[1];
	return  0;
}

static void aec_resetWork(struct work_struct *work)
{
    mutex_lock(&g_mutex);
    printk("[AEC] aec_resetWork.\n");
    if(g_aec_restartOn)
    {
        if(aec_devPwrCtrl(1)<0)
        {
            aec_devPwrCtrl(0);
            queue_delayed_work(g_pWqueue, &g_resetWork, msecs_to_jiffies(1000));
        }
    }
    mutex_unlock(&g_mutex);
}

static int aec_set_regs_default(int mode)
{
    u8 *data;
	int ret = 0;
    int len = -1;
    int err = -1;
	unsigned int cnt = 0;
    unsigned short val;

    len = find_fw_max_len();
	data = kmalloc(len*7, GFP_KERNEL);
	if (!data) 
    {
		err = -ENOMEM;
        printk(KERN_ERR "[FM2018] kmalloc error\n");
        return err;
	}

    cnt = 0;
    do{
        val = 0xffff;
        ret = aec_mem_read(0x1E3A, &val);
        cnt++;
        msleep(10);
        if(cnt>=10){
            printk(KERN_ERR "[AEC]Device is not ready, cnt = %d\n", cnt);
            goto err;
        }    
        printk("[AEC]Ready status = %hx\n", val);
    }while((val&0x8000) == 0x0000);
    
    cnt = 0;    

	printk("[FM2018] mode = %s, len = %d\n", ((mode == AEC_BYPASS_MODE) ? "BYPASS" : "NORMAL"), len);   

    if (AEC_BYPASS_MODE == mode)
    {
        construct_command(data, fw_bypass_mode, fw_bypass_len);
        err = fm2018_write(data, fw_bypass_len * 7);
    }
    else
    {
        construct_command(data, fw_normal_mode, fw_normal_len);
        err = fm2018_write(data, fw_normal_len * 7);
    }

	if (err)
    {
        printk(KERN_ERR "[FM2018] i2c write error!\n");
		goto err;
    }

    kfree(data);

    msleep(200);        	
	return 0;

err:
    kfree(data);
    return -1;
}

static unsigned char aec_devTest(struct aec_platform_data *pdata)
{
	unsigned short test_data;
	unsigned short org_data;
	unsigned short dsp_test;
	int cnt;
	static unsigned short randData;
	
	randData += 3;
	/*I2C check*/
	aec_mem_read(0x1e9f, &org_data);	
	aec_mem_write(0x1e9f, randData);
	aec_mem_read(0x1e9f, &test_data);
	printk("read from memory 0x1e9f, res:%x\n", test_data);
    if(test_data != randData)
        return 0;
    
    aec_mem_write(0x1e9f, org_data);
   	/*Bypass check*/
   	aec_mem_read(0x3fe1, &test_data);
   	printk("read from memory 0x3fe1, res:%x\n", test_data);
   	if(test_data & 0x0020)
   	{
   	    printk("AEC in bypass mode\n");
   	    aec_mem_read(0x3fc0, &test_data);
   	    aec_mem_read(0x1e4a, &dsp_test);
   	    if(test_data != dsp_test)
   	    {
   	        printk("[AEC] bypass mode param not load\n");
   	        return 0;
   	    }
   	    pdata->set_mode(AEC_NORMAL_MODE);
   	    aec_mem_read(0x3fe1, &test_data);
   	    if(test_data & 0x0020)
   	        return 0;
   	    pdata->set_mode(AEC_BYPASS_MODE);
   	}
   	else
   	{
   	    printk("AEC in normal mode\n");
   	    aec_mem_read(0x3fc0, &test_data);
   	    aec_mem_read(0x1e34, &dsp_test);
   	    if(test_data != dsp_test)
   	    {
   	        printk("[AEC] normal mode param not load\n");
   	        return 0;
   	    }
   	    /*In/Out test*/
   	    cnt=10;

    	aec_mem_read(0x3fc4, &test_data);
    	printk("read from memory 0x3fc4, res1:%x\n", test_data);
RETRY1:
    	msleep(1);
    	aec_mem_read(0x3fc4, &dsp_test);
    	printk("read from memory 0x3fc4, res2:%x\n", dsp_test);
    	if(test_data == dsp_test)
    	{
    	    if(cnt>0)
    	    {
    	        cnt--;
    	        goto RETRY1;
    	    }
    	    else
    	        return 0;


    	}
    	cnt=10;

    	aec_mem_read(0x3fc9, &test_data);
    	printk("read from memory 0x3fc9, res1:%x\n", test_data);
RETRY2:
    	msleep(1);
    	aec_mem_read(0x3fc9, &dsp_test);
    	printk("read from memory 0x3fc9, res2:%x\n", dsp_test);
    	if(test_data == dsp_test)
    	{
    	    if(cnt>0)
    	    {
    	        cnt--;
    	        goto RETRY2;
    	    }
    	    else
    	        return 0;

    	}
   	    pdata->set_mode(AEC_BYPASS_MODE);
   	    aec_mem_read(0x3fe1, &test_data);
   	    if(!(test_data & 0x0020))
   	        return 0;
   	    pdata->set_mode(AEC_NORMAL_MODE);
   	}
   	
   	/*DSP check*/
	printk("Start DSP check\n");
	aec_mem_read(0x1e65, &test_data);
	printk("read from memory 0x1e65, res1:%x\n", test_data);
	msleep(10);
	aec_mem_read(0x1e65, &dsp_test);
	printk("read from memory 0x1e65, res2:%x\n", dsp_test);
	if(test_data == dsp_test)
	    return 0;
	
	return 1;	
}

static int aec_devPwrCtrl2(int on)
{
	struct aec_platform_data *pdata;
	unsigned char cnt;
	unsigned short val1, val2;

	if(!gAecClient)
		return -EPERM;

	pdata = gAecClient->dev.platform_data;

	if(!pdata)
		return -EPERM;

	if(on == 1)
	{   
		cnt = 0;
		/* Perform cold start sequence. */
		do
		{
			if (0 != cnt) {
				pdata->sleep();
				disable_oscc_pout();
				msleep(500);
			}
			pdata->wakeup();
			enable_oscc_pout();
			msleep(10);
			pdata->reset();
			msleep(20);
			aec_set_regs_default(g_mode);
			pdata->set_mode(g_mode);
			/* This is workaround that performs more wakeup process if it fails to wakeup AEC. */
			msleep(50);            
			aec_mem_read(0x1e65, &val1);
			msleep(50);
			aec_mem_read(0x1e65, &val2);
			cnt++;
			printk("[AEC]Power on. Reset count=%d.\n", cnt);
		}while(val1==val2 && cnt<5);
	}
	else
	{
		pdata->sleep();
		printk("[AEC]Power off.\n");
		disable_oscc_pout();
	}

	return 0;
}

void aec_reset_dev(void)
{
	printk("reset aec.....\n");
	aec_devPwrCtrl2(0);
	aec_devPwrCtrl2(1);
	printk("reset aec..... done\n");
}

static int aec_suspend(struct i2c_client * aec, pm_message_t state)
{
    struct aec_platform_data *pdata;

    enable_oscc_pout();
    	
	pdata = aec->dev.platform_data;
	
	pdata->sleep();

    disable_oscc_pout();

	return 0;
}

static int aec_resume(struct i2c_client * aec)
{
	struct aec_platform_data *pdata;

    enable_oscc_pout();
    	
	pdata = aec->dev.platform_data;
	
	pdata->wakeup();

    disable_oscc_pout();

	return 0;
}


static int aec_devPwrCtrl(int on)
{
    struct aec_platform_data *pdata;
    unsigned char cnt;
    unsigned short dspCnt0, dspCnt1;
    unsigned short micCnt0, micCnt1;
    unsigned short outCnt0, outCnt1;
    char fail = 0;

    if(!gAecClient)
        return -EPERM;

    pdata = gAecClient->dev.platform_data;

    if(!pdata)
        return -EPERM;

    if(on == 1)
    {   
        cnt = 0;
        /* Perform cold start sequence. */
        do
        {
            fail = 0;
			if(cnt == 0) {
            pdata->wakeup();
            msleep(20);
            enable_oscc_pout();
			}
		    msleep(20);
            pdata->reset();
		    msleep(20);

            dspCnt0 = 0;
            dspCnt1 = 0;
            micCnt0 = 0;
            micCnt1 = 0;
            outCnt0 = 0;
            outCnt1 = 0;
            if(aec_set_regs_default(g_mode) == 0)
            {
		        msleep(20);
                pdata->set_mode(g_mode);
                /* This is workaround that performs more wakeup process if it fails to wakeup AEC. */
        	    aec_mem_read(0x1e65, &dspCnt0);
        	    aec_mem_read(0x3fc4, &micCnt0);
        	    aec_mem_read(0x3fc9, &outCnt0);
        	    msleep(50);
        	    aec_mem_read(0x1e65, &dspCnt1);
        	    aec_mem_read(0x3fc4, &micCnt1);
        	    aec_mem_read(0x3fc9, &outCnt1);
        	    
        	    switch(g_mode)
        	    {
        	        case AEC_BYPASS_MODE:
        	            if(dspCnt0==dspCnt1)
        	                fail = 1;
        	            break;
        	        case AEC_NORMAL_MODE:
        	        default:
        	            if(dspCnt0==dspCnt1)
        	                fail = 1;
        	            if(micCnt0==micCnt1)
        	                fail = 1;
        	            if(outCnt0==outCnt1)
        	                fail = 1;
        	            break;
        	    }
    	    }
    	    else
    	        fail = 1;
    	    cnt++;
            printk("[AEC]Power on. Reset count=%d, [%d, %d] [%d, %d] [%d, %d]\n", cnt, dspCnt0, dspCnt1, micCnt0, micCnt1, outCnt0, outCnt1);
        }while(fail==1 && cnt<2);
        if(fail==1 && cnt==2){
            printk("[AEC]Start work queue.\n");
            return -1;
        }
    }
    else
    {
        pdata->sleep();
        printk("[AEC]Power off.\n");
        disable_oscc_pout();
    }
    return 0;
}

static ssize_t aec_reg_store(struct device *dev,
					  struct device_attribute *attr,
					  const char *buf, size_t count)
{
	int err = -EINVAL;
	int res;
	char cmd;
	u32 reg, val;
    struct aec_platform_data *pdata;
    unsigned char status;
    unsigned short value;

    if(!gAecClient)
        return -EPERM;

    pdata = gAecClient->dev.platform_data;

    if(!pdata)
        return -EPERM;
	
	err =0;
	res = sscanf(buf, "%c %x %x", &cmd, &reg, &val);	

	switch(cmd)
	{
	    case 'r':
        	if (res != 2)
        	{
        		err =  -EINVAL;
        		goto out;
        	}
	        aec_mem_read(reg, &value);
	        printk("[AEC]Read reg. :%04x, val :%04x\n", reg, value);
	        break;
	    case 'w':
        	if (res != 3)
        	{
        		err =  -EINVAL;
        		goto out;
        	}
	        aec_mem_write(reg, val);
	        printk("[AEC]Write reg. :%04x, val :%04x\n", reg, val);
	        break;
	    case 'm':  /* AEC mode */
        	if (res != 2)
        	{
        		err =  -EINVAL;
        		goto out;
        	}
        	
            switch(reg)
            {
                case AEC_NORMAL_MODE:
                    printk(KERN_INFO "[AEC] set normal mode\n");
                    break;
                case AEC_BYPASS_MODE:
                    printk(KERN_INFO "[AEC] set bypass mode\n");
                    break;
                default:
                    printk(KERN_INFO "[AEC] Invalid argument\n");
                    reg = AEC_NORMAL_MODE;
                    break;
            }
            
            g_mode = reg;
            
	        aec_softReset();
	        aec_set_regs_default(g_mode);
	        pdata->set_mode(g_mode);
	        break;
	    case 's': /* Sleep & Wake up */
        	if (res != 2)
        	{
        		err =  -EINVAL;
        		goto out;
        	}
        	if(reg)
        	{
	            pdata->sleep();
	            printk("[AEC]Sleep.\n");
        	}
        	else
        	{
        	    pdata->wakeup();
        	    printk("[AEC]Wakeup.\n");
	        }
	        break;
	    case 'p': /* Power on & Power off */
        	if (res != 2)
        	{
        	    printk("[AEC]Power on/off, wrong parameter.\n");
        		err =  -EINVAL;
        		goto out;
        	}

            mutex_lock(&g_mutex);
            if(reg){
                if(aec_devPwrCtrl(1)<0)
                {
                    aec_devPwrCtrl(0);
                    g_aec_restartOn = 1;
                    queue_delayed_work(g_pWqueue, &g_resetWork, msecs_to_jiffies(1000));
                }
            }
            else{
                g_aec_restartOn = 0;
                cancel_delayed_work_sync(&g_resetWork);
                aec_devPwrCtrl(0);
            }
            mutex_unlock(&g_mutex);
                
	        break;
	    case 't': /*AEC test*/
	        status = aec_devTest(pdata);
	        if(status == 0)
	            printk("[AEC]Test fail\n");
	        else
	            printk("[AEC]Test pass\n");
	        break;
	    case 'x':
	        aec_softReset();
	        break;
	    case 'd': /*Debug*/
	        aec_mem_read(0x1e65, &value);
	        printk("DSP count val:%x\n", value);
	        aec_mem_read(0x3fc4, &value);
	        printk("read from memory 0x3fc4, val:%x\n", value);
	        aec_mem_read(0x3fc9, &value);
	        printk("read from memory 0x3fc9, val:%x\n", value);
	        break;
	    default:
	        printk("[AEC]unsupported command.\n");
	        goto out;
	}

out:
	return err ? err : count;
}

/* codec register dump */
static ssize_t aec_reg_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	int count = 0;
	unsigned short val;
	unsigned short special_reg;
	aec_reg_t *ptr;
	struct aec_platform_data *pdata;
	
	count += sprintf(buf, "aec registers\n");

    pdata = gAecClient->dev.platform_data;
    if(!pdata)
        return 0;
        	
	ptr = (aec_reg_t *)pdata->parmGet(AEC_NORMAL_MODE);
	if(!ptr)
	    return 0;
	    
	while(ptr->reg != 0x0000)
	{
		aec_mem_read(ptr->reg, &val);
		count += sprintf(buf + count, "%4x: %4x\n", ptr->reg, val);
		ptr++;
	}
	
	special_reg = 0x1E65;
	aec_mem_read(special_reg, &val);
	count += sprintf(buf + count, "%4x: %4x\n", special_reg, val);

/* only for test sword*/
	special_reg = 0x3FC4;
	aec_mem_read(special_reg, &val);
	count += sprintf(buf + count, "%4x: %4x\n", special_reg, val);

	special_reg = 0x3FC9;
	aec_mem_read(special_reg, &val);
	count += sprintf(buf + count, "%4x: %4x\n", special_reg, val);
	
	return count;
}
static DEVICE_ATTR(aec_reg, 0644, aec_reg_show, aec_reg_store);

static int aec_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct aec_platform_data *pdata;
	int ret;
    id = id;

	printk("AEC: aec_probe\n");

	pdata = client->dev.platform_data;

	pdata->init();

	gAecClient = client;

	ret = device_create_file(&gAecClient->dev, &dev_attr_aec_reg);
	if (ret < 0)
		printk(KERN_WARNING "AEC: failed to add aec chip sysfs entries\n");

	if(device_register(&ghost_device) < 0) {
		printk("[AEC]device_register error\n");	    
	}

	//aec_devPwrCtrl(1);
	//aec_devPwrCtrl(0);

    INIT_DELAYED_WORK(&g_resetWork, aec_resetWork);
    
    g_pWqueue = create_singlethread_workqueue("resetWork");
    if (!g_pWqueue) {
    	printk("[AEC] Fail to create delay work.\n");
    	ret = -1;
    	goto ERROR_HANDLE;
    }
    
ERROR_HANDLE:
	return ret;
}

static int aec_remove(struct i2c_client *client)
{  
	printk("AEC: aec_remove\n");
	device_unregister(&ghost_device);
	device_remove_file(&gAecClient->dev, &dev_attr_aec_reg);    
    gAecClient = NULL;
	return 0;
}

static const struct i2c_device_id fm2018_id[] = {
	{ "fm2018", 0 },
	{ }
};

/* corgi i2c codec control layer */
struct i2c_driver aec_i2c_driver = {
	.driver = {
		.name = "fm2018",		
	},
	.probe      = aec_probe,
	.remove     = aec_remove,
	.suspend    = aec_suspend,
	.resume     = aec_resume,	
    .id_table   = fm2018_id,
};

// device structure
struct aec_dev {	
	struct cdev cdev;
};

struct aec_dev *aec_device;	

/*  File ops */
static int aec_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int aec_release(struct inode *inode, struct file *file)
{
	return 0;
}

static int aec_ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
    void __user *argp = (void __user *)arg;
	int i;
	int ret = 0;
	unsigned char status;	
    struct aec_platform_data *pdata;

	
	printk(KERN_EMERG "aec ioctl cmd %x \n", cmd);	
	printk(KERN_EMERG "aec arg %lx \n", arg);

	if(_IOC_TYPE(cmd) != AEC_IOC_MAGIC)
		return -ENOTTY;	

    if(!gAecClient)
        return -EPERM;

    pdata = gAecClient->dev.platform_data;

    if(!pdata)
        return -EPERM;


	switch (cmd) {
	    case AEC_IOC_SET_REGISTER:
	        {
	            S_AEC_REG_CMD_PARM parm;
	            aec_reg_t *pReg;
	            
    	        if (copy_from_user(&parm, argp, sizeof(parm)))
		            return -EFAULT;
    	        
    	        if(parm.size == 0)
    	            break;
    	            
    	        pReg = kmalloc(parm.size * sizeof(aec_reg_t), GFP_KERNEL);
    	        if (!pReg)
    	            return -ENOMEM;
    	            
    	        if (copy_from_user(pReg, parm.pReg, parm.size * sizeof(aec_reg_t)))
    	        {
    	            kfree(pReg);
		            return -EFAULT;
		        }
	
    	        for (i = 0; i < parm.size; i++){
            		ret = aec_mem_write(pReg[i].reg, pReg[i].val);
            		if (ret < 0){
            			printk(KERN_ERR"cannnot initialize aec chip\n");
            			kfree(pReg);
            			return -EIO;
            		}			

                    printk(KERN_EMERG "aec register addr: 0x%hx, set value: 0x%hx\n", pReg[i].reg, pReg[i].val);
            	}

            	kfree(pReg);
            }
	        break;
	    case AEC_IOC_CMD_TEST:
	        status = aec_devTest(pdata);
	        {
	            unsigned long rtn;
	            rtn=copy_to_user((void *)arg, &status, sizeof(unsigned char));
	        }
	        break;	
	    case AEC_IOC_SET_MODE:
            switch(arg)
            {
                case AEC_NORMAL_MODE:
                    printk(KERN_INFO "[AEC] set normal mode\n");
                    break;
                case AEC_BYPASS_MODE:
                    printk(KERN_INFO "[AEC] set bypass mode\n");
                    break;
                default:
                    printk(KERN_INFO "[AEC] Invalid argument\n");
                    arg = AEC_NORMAL_MODE;
                    break;
            }

            if(g_mode != arg) {
    	        g_mode = arg;
                pdata->reset();
    	        aec_set_regs_default(g_mode);
    	        pdata->set_mode(g_mode);
    	    } else
    	        printk(KERN_INFO "[AEC] New mode is same as active mode\n");

	        break;
	    case AEC_IOC_RESET:	        	            	            
            if(gAecClient)
            {
                aec_set_regs_default(g_mode);
            }               	        
	        break;
	    case AEC_IOC_SLEEP:	        
	        if(gAecClient)
            {
                printk("[AEC] AEC_IOC_SLEEP arg=%lx.\n", arg);
                mutex_lock(&g_mutex);
                if(arg == 1){
                    g_aec_restartOn = 0;
                    cancel_delayed_work_sync(&g_resetWork);
                    aec_devPwrCtrl(0);
                }
                else{
                    if(aec_devPwrCtrl(1)<0)
                    {
                        aec_devPwrCtrl(0);
                        queue_delayed_work(g_pWqueue, &g_resetWork, msecs_to_jiffies(1000));
                    }
                    g_aec_restartOn = 1;
                }
                mutex_unlock(&g_mutex);
            }               	        
	        break;
		default:
			return -ENOTTY;
	}    
	return ret;
}

static struct file_operations aec_fops = {
	.owner =    THIS_MODULE,
	.ioctl =    aec_ioctl,
	.open =     aec_open,
	.release =  aec_release,	
};

static int setupCdev(struct aec_dev *dev)
{
	int err = 0;
	int devno = MKDEV(aec_major, aec_minor);
    
	cdev_init(&dev->cdev, &aec_fops);
	dev->cdev.owner = THIS_MODULE;
	dev->cdev.ops = &aec_fops;
	err = cdev_add (&dev->cdev, devno, 1);
	printk(KERN_EMERG "setupCdev result %d \n", err);
	/* Fail gracefully if need be */
	if (err)
		printk(KERN_EMERG "setupCdev Error %d", err);
		
	return err;
}

static int __init aec_init(void)
{
	int result = -1;
	dev_t dev = 0;
	
	printk(KERN_EMERG "AEC: aec_init start ...\n");
	
	result = i2c_add_driver(&aec_i2c_driver);
	if (result) {
		printk(KERN_ERR "AEC: can't add i2c driver");							
	}
	
	/*
     * Get a range of minor numbers to work with, asking for a dynamic
     * major unless directed otherwise at load time.
     */    
     
    
    // prepare device id
	if (aec_major) {
		dev = MKDEV(aec_major, aec_minor);
		result = register_chrdev_region(dev, aec_nr_dev, AEC_DEV_FILE);
	} else {
		result = alloc_chrdev_region(&dev, aec_minor, aec_nr_dev,
				AEC_DEV_FILE);
		aec_major = MAJOR(dev);
	}
	if (result < 0) {
		printk(KERN_WARNING "scull: can't get major %d\n", aec_major);
		return result;
	}
	printk(KERN_EMERG "[AEC] :major %d minor %d\n", aec_major, aec_minor);

    aec_device = kmalloc(sizeof(struct aec_dev), GFP_KERNEL);
	if (!aec_device) {
		result = -ENOMEM;
		goto fail_malloc;
	}
	memset(aec_device, 0, sizeof(struct aec_dev));
	
	// setup cdev
	result = setupCdev(aec_device);
	if (result < 0) {
		result = -ENOMEM;
		goto fail_setup;
	}	
	
	printk(KERN_EMERG "[AEC] aec_init complete >> \n");
	
	return (0);

fail_setup:
	kfree(aec_device);
	
fail_malloc:
	unregister_chrdev_region(dev, aec_nr_dev);
	return result;		
}

static void __exit aec_exit(void)
{	
	dev_t devno = MKDEV(aec_major, aec_minor);
	cdev_del(&aec_device->cdev);
	kfree(aec_device);
	
	/* cleanup_module is never called if registering failed */
	unregister_chrdev_region(devno, aec_nr_dev);
	
	i2c_del_driver(&aec_i2c_driver);
}

module_init(aec_init);
module_exit(aec_exit);

MODULE_AUTHOR("K.H. Fan <kh.fan@foxconn.com>");
MODULE_DESCRIPTION("Foxconn manafacture AEC");
MODULE_LICENSE("GPL");

