/*
 * linux/sound/soc/pxa/saarb.c
 *
 * Copyright (C) 2007 Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/mach-types.h>
#include <linux/io.h>

#include <mach/mfp-pxa9xx.h>
#include <mach/mfp-pxa3xx.h>
#include <mach/pxa-regs.h>
#include <mach/hardware.h>
#include <mach/audio.h>
#include <mach/regs-ssp.h>
#include <mach/ssp.h>
#include <mach/micco.h>
#include <mach/tavor_common.h>
#include <mach/levante_micco_pmic_id.h>
#include <mach/gpio.h>

#include "saarb.h"
#include "pxa3xx-pcm.h"
#include "pxa3xx-ssp.h"
#if defined(CONFIG_PXA3xx_DVFM)
#include "mach/dvfm.h"
#endif

static u8 g_pmic_id;

#define PCM_SSP_PORT	3	  /* used as ap pcm port  */
#define I2S_SSP_PORT	2	  /* used as ap hifi port */
#define PCM_CP_PORT	1	  /* used as cp pcm prot  */

#define ARRAY_AND_SIZE(x)       ((x), ARRAY_SIZE(x))

/*void enable_oscc_pout(void);
void disable_oscc_pout(void);*/

#if defined(CONFIG_PXA3xx_DVFM)

static int dvfm_dev_idx;

static void set_dvfm_constraint(void)
{
	/* Disable Lowpower mode */
	dvfm_disable_op_name("D1", dvfm_dev_idx);
	dvfm_disable_op_name("D2", dvfm_dev_idx);
	dvfm_disable_op_name("CG", dvfm_dev_idx);
}

static void unset_dvfm_constraint(void)
{
	/* Disable Lowpower mode */
	dvfm_enable_op_name("D1", dvfm_dev_idx);
	dvfm_enable_op_name("D2", dvfm_dev_idx);
	dvfm_enable_op_name("CG", dvfm_dev_idx);
}
#else
static void set_dvfm_constraint(void) {}
static void unset_dvfm_constraint(void) {}
#endif

static int saarb_micco_init(struct snd_soc_codec *codec);

static int ssp2_mfp(int enable)
{
	static u32 gpio_bitclk, gpio_sync, gpio_dataout;
	int err, gpio_req_step = 0;

	gpio_bitclk = mfp_to_gpio(MFP_PIN_GPIO85);
	gpio_sync = mfp_to_gpio(MFP_PIN_GPIO87);
	gpio_dataout = mfp_to_gpio(MFP_PIN_GPIO88);

	if (!enable) {
		err = gpio_request(gpio_bitclk, "SSP2 BITCLK");
		if (err)
			goto out;
		gpio_req_step = 1;
		err = gpio_request(gpio_sync, "SSP2 SYNC");
		if (err)
			goto out;
		gpio_req_step = 2;
		err = gpio_request(gpio_dataout, "SSP2 DATAOUT");
		if (err)
			goto out;
		gpio_req_step = 3;

		gpio_direction_input(gpio_bitclk);
		gpio_direction_input(gpio_sync);
		gpio_direction_input(gpio_dataout);
		gpio_free(gpio_bitclk);
		gpio_free(gpio_sync);
		gpio_free(gpio_dataout);
	}

	return 0;
out:
	if (gpio_req_step > 0)
		gpio_free(gpio_bitclk);
	if (gpio_req_step > 1)
		gpio_free(gpio_sync);
	if (gpio_req_step > 2)
		gpio_free(gpio_dataout);
	printk(KERN_ERR "Request GPIO failed!!");
	return err;
}

static int saarb_micco_hifi_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	unsigned long sscr0;

	set_dvfm_constraint();
	cpu_dai->playback.channels_min = 2;
	cpu_dai->playback.channels_max = 2;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |\
		SNDRV_PCM_RATE_16000;

	ssp2_mfp(1);
	/*enable_oscc_pout();*/

	/* enable packing mode for cpu_dai->hw_params judge*/
	sscr0 = SSCR0_FPCKE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	return 0;
}

static int saarb_micco_hifi_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp, sstsa;
	unsigned long ssacd, ssacdd, ssrsa;

	/*
	 * Bit clk is comming from micco
	 * Tavor is slave
	 */

	sscr0 = 0xE1C0003F;
	sscr1 = 0x02401DC0;
	sspsp = 0x31A00080;
	sstsa = 0x00000003;
	ssrsa = 0x00000003;

	ssacd = 0x60;
	ssacdd = 0x00000040;

	switch (rate) {
	case 48000:
		ssacdd |= (0x659 << 16);
		break;
	case 44100:
		ssacdd |= (0x6E9 << 16);
		break;
	case 16000:
		ssacdd |= (0x130B << 16);
		break;
	default:
		/*return -EINVAL;*/
		break;
	}
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);
	__raw_writel(sstsa, ssp->mmio_base + SSTSA);
	__raw_writel(ssrsa, ssp->mmio_base + SSRSA);
	__raw_writel(ssacd, ssp->mmio_base + SSACD);
	__raw_writel(ssacdd, ssp->mmio_base + SSACDD);

	return 0;
}

static void saarb_micco_hifi_shutdown(struct snd_pcm_substream *substream)
{
	/*disable_oscc_pout();*/
	ssp2_mfp(0);
	unset_dvfm_constraint();
}

static int saarb_micco_voice_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	set_dvfm_constraint();

	printk(KERN_ALERT "----->saarb_micco_voice_startup \n");

	cpu_dai->playback.channels_min = 1;
	cpu_dai->playback.channels_max = 1;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | \
		SNDRV_PCM_RATE_32000;

	/*enable_oscc_pout();*/
	return 0;
}

static int saarb_micco_voice_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp;

	printk(KERN_ALERT "----->saarb_micco_voice_prepare \n");

	cpu_dai->dai_ops.set_sysclk(cpu_dai, PXA3XX_SSP_CLK_PLL, 13000000, 0);

	sscr0 = 0x00C0003F;
	sscr1 = 0x00701DC0;
	sspsp = 0x00800085;

	switch (rate) {
	case 8000:
		sscr0 |= (89 << 8);
		break;
	case 16000:
		sscr0 |= (44 << 8);
		break;
	case 32000:
		sscr0 |= (22 << 8);
		break;
	default:
		return -EINVAL;
	}

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		sscr1 &= ~0x00800000;
	else
		sscr1 |= 0x00800000;

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);

	return 0;
}

static void saarb_micco_voice_shutdown(struct snd_pcm_substream *substream)
{
	/*disable_oscc_pout();*/
	unset_dvfm_constraint();
}

static int saarb_micco_cp_voice_startup(struct snd_pcm_substream *substream)
{
	/*enable_oscc_pout();*/
	return 0;
}
static int saarb_micco_cp_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}
static void saarb_micco_cp_voice_shutdown(struct snd_pcm_substream *substream)
{
	/*disable_oscc_pout();*/
}

/* machine stream operations */
static struct snd_soc_ops saarb_machine_ops[] = {
{
	.startup = saarb_micco_hifi_startup,
	.prepare = saarb_micco_hifi_prepare,
	.shutdown = saarb_micco_hifi_shutdown,
},
{
	.startup = saarb_micco_voice_startup,
	.prepare = saarb_micco_voice_prepare,
	.shutdown = saarb_micco_voice_shutdown,
},
{
	.startup = saarb_micco_cp_voice_startup,
	.prepare = saarb_micco_cp_voice_prepare,
	.shutdown = saarb_micco_cp_voice_shutdown,
},
};

/* saarb digital audio interface glue - connects codec <--> CPU */
static struct snd_soc_dai_link saarb_dai[] = {
{
	.name = "I2S",
	.stream_name = "I2S HiFi",
	.cpu_dai = &pxa3xx_ssp_dai[I2S_SSP_PORT-1],
	.codec_dai = &micco_dai[0],
	.ops = &saarb_machine_ops[0],
	.init = saarb_micco_init,
},
{
	.name = "PCM",
	.stream_name = "PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[PCM_SSP_PORT-1],
	.codec_dai = &micco_dai[1],
	.ops = &saarb_machine_ops[1],
	.init = saarb_micco_init,
},
{
	.name = "CP PCM",
	.stream_name = "CP PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[4],
	.codec_dai = &micco_dai[2],
	.ops = &saarb_machine_ops[2],
	.init = saarb_micco_init,
},
};

/* saarb audio machine driver */
static struct snd_soc_machine snd_soc_machine_saarb = {
	.name = "saarb",
	.dai_link = saarb_dai,
	.num_links = ARRAY_SIZE(saarb_dai),
};

/* saarb audio subsystem */
static struct snd_soc_device saarb_snd_devdata = {
	.machine = &snd_soc_machine_saarb,
	.platform = &pxa3xx_soc_platform,
	.codec_dev = &soc_codec_dev_micco,
};

static struct platform_device *saarb_snd_device_micco;

/*
 * Logic for a Micco as connected on a saarb Device
 */
static int saarb_micco_init(struct snd_soc_codec *codec)
{
	ssp2_mfp(0);
	return 0;
}

static int saarb_levante_init(struct snd_soc_codec *codec)
{
	return 0;
}

static int saarb_levante_hifi_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	unsigned long sscr0;

	printk(KERN_INFO "saarb_levante_hifi_startup \n");

	set_dvfm_constraint();

	cpu_dai->playback.channels_min = 2;
	cpu_dai->playback.channels_max = 2;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |\
		SNDRV_PCM_RATE_16000;

	/*enable_oscc_pout();*/

	__raw_writel(0x415000BF, ssp->mmio_base + SSCR0);
	__raw_writel(0x03201DC0, ssp->mmio_base + SSCR1);
	__raw_writel(0x02100004, ssp->mmio_base + SSPSP);

	/* enable packing mode for cpu_dai->hw_params judge*/
	sscr0 = SSCR0_FPCKE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	return 0;
}


static int saarb_levante_hifi_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp, sstsa;
	unsigned long ssacd, ssacdd, ssrsa;

	printk(KERN_INFO "saarb_levante_hifi_prepare \n");

	/* Because the internal 13M clock will be 10M in D0CS,
	 * we route SSP_CLK to GPIO126(EXT_CLK) and let SSP select
	 * NETWORK CLK as CLK source.
	 * This workaround need an ECO on Littleton mainboard.
	 */

	sscr0 = 0x415000BF;
	sscr1 = 0x03201DC0;
	sspsp = 0x02100004;
	sstsa = 0x00000003;
	ssrsa = 0x00000003;

	ssacd = 0x60;
	ssacdd = 0x00000040;

	switch (rate) {
	case 48000:
		ssacdd |= (0x659 << 16);
		break;
	case 44100:
		ssacdd |= (0x6E9 << 16);
		break;
	case 16000:
		ssacdd |= (0x130B << 16);
		break;
	default:
		/*return -EINVAL;*/
		break;
	}

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);
	__raw_writel(sstsa, ssp->mmio_base + SSTSA);
	__raw_writel(ssrsa, ssp->mmio_base + SSRSA);
	__raw_writel(ssacd, ssp->mmio_base + SSACD);
	__raw_writel(ssacdd, ssp->mmio_base + SSACDD);

	return 0;
}

static void saarb_levante_hifi_shutdown(struct snd_pcm_substream *substream)
{
	printk(KERN_INFO "saarb_levante_hifi_shutdown \n");

	/*disable_oscc_pout();*/
	unset_dvfm_constraint();
}

static int saarb_levante_voice_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;

	set_dvfm_constraint();

	printk(KERN_ALERT "----->saarb_levante_voice_startup \n");


	cpu_dai->playback.channels_min = 1;
	cpu_dai->playback.channels_max = 1;
	cpu_dai->playback.formats = SNDRV_PCM_FMTBIT_S16_LE;
	cpu_dai->playback.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | \
		SNDRV_PCM_RATE_32000;

	/*enable_oscc_pout();*/

	return 0;
}

static int saarb_levante_voice_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *machine = rtd->dai;
	struct snd_soc_dai *cpu_dai = machine->cpu_dai;
	struct ssp_device *ssp = cpu_dai->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned long rate = runtime->rate;
	unsigned long sscr0, sscr1, sspsp;

	printk(KERN_ALERT "----->saarb_levante_voice_prepare \n");

	cpu_dai->dai_ops.set_sysclk(cpu_dai, PXA3XX_SSP_CLK_PLL, 13000000, 0);

	sscr0 = 0x00C0003F;
	sscr1 = 0x00701DC0;
	sspsp = 0x00800085;

	switch (rate) {
	case 8000:
		sscr0 |= (89 << 8);
		break;
	case 16000:
		sscr0 |= (44 << 8);
		break;
	case 32000:
		sscr0 |= (22 << 8);
		break;
	default:
		return -EINVAL;
	}

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		sscr1 &= ~0x00800000;
	else
		sscr1 |= 0x00800000;

	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	__raw_writel(sspsp, ssp->mmio_base + SSPSP);

	return 0;
}

static void saarb_levante_voice_shutdown(struct snd_pcm_substream *substream)
{
	/*disable_oscc_pout();*/
	unset_dvfm_constraint();
}

static int saarb_levante_cp_voice_startup(struct snd_pcm_substream *substream)
{
	/*enable_oscc_pout();*/
	return 0;
}
static int saarb_levante_cp_voice_prepare(struct snd_pcm_substream *substream)
{
	return 0;
}
static void saarb_levante_cp_voice_shutdown(struct snd_pcm_substream *substream)
{
	/*disable_oscc_pout();*/
}


/* machine stream operations */
static struct snd_soc_ops saarb_machine_ops_levante[] = {
{
	.startup = saarb_levante_hifi_startup,
	.prepare = saarb_levante_hifi_prepare,
	.shutdown = saarb_levante_hifi_shutdown,
},
{
	.startup = saarb_levante_voice_startup,
	.prepare = saarb_levante_voice_prepare,
	.shutdown = saarb_levante_voice_shutdown,
},
{
	.startup = saarb_levante_cp_voice_startup,
	.prepare = saarb_levante_cp_voice_prepare,
	.shutdown = saarb_levante_cp_voice_shutdown,
},
};

/* saarb digital audio interface glue - connects codec <--> CPU */
static struct snd_soc_dai_link saarb_dai_levante[] = {
{
	.name = "I2S",
	.stream_name = "I2S HiFi",
	.cpu_dai = &pxa3xx_ssp_dai[I2S_SSP_PORT-1],
	.codec_dai = &levante_dai[0],
	.ops = &saarb_machine_ops_levante[0],
	.init = saarb_levante_init,
},
{
	.name = "PCM",
	.stream_name = "PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[PCM_SSP_PORT-1],
	.codec_dai = &levante_dai[1],
	.ops = &saarb_machine_ops_levante[1],
	.init = saarb_levante_init,
},
{
	.name = "CP PCM",
	.stream_name = "CP PCM Voice",
	.cpu_dai = &pxa3xx_ssp_dai[4],
	.codec_dai = &levante_dai[2],
	.ops = &saarb_machine_ops_levante[2],
	.init = saarb_levante_init,
},
};

/* saarb audio machine driver */
static struct snd_soc_machine snd_soc_machine_saarb_levante = {
	.name = "saarb",
	.dai_link = saarb_dai_levante,
	.num_links = ARRAY_SIZE(saarb_dai_levante),
};

/* saarb audio subsystem */
static struct snd_soc_device saarb_snd_devdata_levante = {
	.machine = &snd_soc_machine_saarb_levante,
	.platform = &pxa3xx_soc_platform,
	.codec_dev = &soc_codec_dev_levante,
};

static struct platform_device *saarb_snd_device_levante;

/*
 * Logic for a Levante as connected on a saarb Device
 */

static int __init saarb_init(void)
{
	int ret;

	printk(KERN_INFO \
	"**Audio: saarb_init; machine_arch_type= %d (MACH_TYPE_SAARB= %d) **\n"\
	, machine_arch_type, MACH_TYPE_SAARB);
	if (!machine_is_saarb()) {
		printk(KERN_CRIT \
	"**Audio: saarb_init; machine is NOT 'MACH_TYPE_SAARB' ==> ABORT!**\n");
		return -ENODEV;
	}

	g_pmic_id = pmic_id();

	if (g_pmic_id >= 0x40) {  /* LEVANTE */
		printk(KERN_INFO\
		"*** Audio: saarb_init; g_pmic_id= 0x%x ==> LEVANTE! () ***\n"\
		, g_pmic_id);
		saarb_snd_device_levante = \
			platform_device_alloc("soc-audio", -1);
		if (!saarb_snd_device_levante) {
			printk(KERN_ALERT "ERROR in LEVANTE device alloc!");
			return -ENOMEM;
		}

		platform_set_drvdata(saarb_snd_device_levante,\
					&saarb_snd_devdata_levante);
		saarb_snd_devdata_levante.dev = &saarb_snd_device_levante->dev;
		ret = platform_device_add(saarb_snd_device_levante);

		if (ret) {
			printk(KERN_ALERT "ERROR in LEVANTE device add!");
			platform_device_put(saarb_snd_device_levante);
		}
	} else {  /* MICCO */
		printk("** Audio: saarb_init; g_pmic_id= 0x%x ==> MICCO! **\n"\
			, g_pmic_id);
		saarb_snd_device_micco = platform_device_alloc("soc-audio", -1);
		if (!saarb_snd_device_micco)
			return -ENOMEM;

		platform_set_drvdata(saarb_snd_device_micco, \
					&saarb_snd_devdata);
		saarb_snd_devdata.dev = &saarb_snd_device_micco->dev;
		ret = platform_device_add(saarb_snd_device_micco);

		if (ret)
			platform_device_put(saarb_snd_device_micco);

	}

#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("Sound", &dvfm_dev_idx);
#endif

	return ret;
}

static void __exit saarb_exit(void)
{
	platform_device_unregister(saarb_snd_device_micco);
	platform_device_unregister(saarb_snd_device_levante);

#ifdef CONFIG_PXA3xx_DVFM
	dvfm_unregister("Sound", &dvfm_dev_idx);
#endif

}

module_init(saarb_init);
module_exit(saarb_exit);

/* Module information */
MODULE_AUTHOR("mwwang@marvell.com");
MODULE_DESCRIPTION("ALSA SoC MICCO LEVANTE SAARB");
MODULE_LICENSE("GPL");
