/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for MT9P111
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/
#include "camera.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
//#include <asm/arch/mfp.h>
//#include <asm/arch/gpio.h>
//#include <asm/hardware.h>
#include <asm/uaccess.h>
#include <linux/err.h>
#include <linux/proc_fs.h>

#include <linux/clk.h>
#include <linux/i2c.h>

/*****************************************************************************
 * Constants & Registers
 *****************************************************************************/
/* CICLK for QCI */
#define CICLK	1300

/* Revision constants */
#define PID_VERH		0x2880
#define PID_ADRRH            0x0000

#define AF_SETTING		0
/* Output Size & Format */
#define MT9P111_SIZE_QVGA	1		//320x240
#define MT9P111_SIZE_VGA		2		//640x480
#define MT9P111_SIZE_SVGA	3		//800x600
#define MT9P111_SIZE_D1		4		//864x600
#define MT9P111_SIZE_XGA		5		//1024x768
#define MT9P111_SIZE_SXGA	6		//1280x1024
#define MT9P111_SIZE_UXGA	7		//1600X1200
#define MT9P111_SIZE_QXGA	8		//2048x1536
#define MT9P111_SIZE_QSXGA	9		//2592X1944
#define MT9P111_SIZE_YGA		10		//480X360
#define MT9P111_SIZE_2560X1920		11

#define MT9P111_SIZE_SQCIF	100		//128x96
#define MT9P111_SIZE_QCIF	101		//176x144
#define MT9P111_SIZE_CIF		102		//352x288
#define MT9P111_SIZE_JPEG_VGA	103		//640x480
#define MT9P111_SIZE_QCIF_H	104		//144x176
#define MT9P111_SIZE_256X192	200		//256X192 for 4X zoom
#define MT9P111_SIZE_JPEG_QVGA	201		//320X240
#define MT9P111_SIZE_JPEG_256X192 202		//256X192

#define MT9P111_FORMAT_YUV_422	( 1 << 8)
#define MT9P111_FORMAT_RAW10	( 2 << 8)
#define MT9P111_FORMAT_RAW8	( 3 << 8)
#define MT9P111_FORMAT_JPG	( 4 << 8)
#define MT9P111_FORMAT_RGB565	( 5 << 8)

#define MT9P111_QVGA_WIDTH	320
#define MT9P111_QVGA_HEIGHT	240
#define MT9P111_VGA_WIDTH	640
#define MT9P111_VGA_HEIGHT	480
#define MT9P111_SVGA_WIDTH	800
#define MT9P111_SVGA_HEIGHT	600
#define MT9P111_YGA_WIDTH	480
#define MT9P111_YGA_HEIGHT	360
#define MT9P111_SQCIF_WIDTH	128
#define MT9P111_SQCIF_HEIGHT	96
#define MT9P111_QCIF_WIDTH	176
#define MT9P111_QCIF_HEIGHT	144
#define MT9P111_CIF_WIDTH	352
#define MT9P111_CIF_HEIGHT	288
#define MT9P111_XGA_WIDTH	1024
#define MT9P111_XGA_HEIGHT	768
#define MT9P111_UXGA_WIDTH	1600
#define MT9P111_UXGA_HEIGHT	1200
#define MT9P111_QXGA_WIDTH	2048
#define MT9P111_QXGA_HEIGHT	1536
#define MT9P111_QSXGA_WIDTH	2592
#define MT9P111_QSXGA_HEIGHT	1944

#define MT9P111_REGEND		(0xFFFF)

#define AF_STATE_STEP_STATE_NO		0x80
#define AF_STATE_MASK_MODECHANGE	0x40
#define AF_STATE_MASK_CAPTURECMD	0x20
#define AF_STATE_RESERVED_INVALID	0x10
#define AF_STATE_MODE_IDLE		0x00
#define AF_STATE_MODE_SINGLE		0x04
#define AF_STATE_MODE_CONTINUE		0x08
#define AF_STATE_MODE_STEP		0x0c
#define AF_STATE_MODE_STEP_INSTRUCTION	0x00
#define AF_STATE_MODE_STEP_FOCUSING	0x01
#define AF_STATE_MODE_STEP_FOCUSED	0x02
#define AF_STATE_MODE_STEP_CAPTURE	0x03

#define	AF_STATE_INF		(AF_STATE_MODE_IDLE|AF_STATE_MODE_STEP_INSTRUCTION)
#define AF_STATE_SINGLE		(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSING|AF_STATE_MASK_MODECHANGE|AF_STATE_MASK_CAPTURECMD)
#define AF_STATE_SUCCESS_S	(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSED|AF_STATE_MASK_MODECHANGE)
#define AF_STATE_FAIL_S		(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_FOCUSED|AF_STATE_MASK_MODECHANGE|AF_STATE_STEP_STATE_NO)
#define AF_STATE_CAPTURE_S	(AF_STATE_MODE_SINGLE|AF_STATE_MODE_STEP_CAPTURE|AF_STATE_MASK_MODECHANGE)

#define SENSOR_EV_N_2_0       -199
#define SENSOR_EV_N_1_7       -169
#define SENSOR_EV_N_1_3       -129
#define SENSOR_EV_N_1_0       -99
#define SENSOR_EV_N_0_7       -69
#define SENSOR_EV_N_0_3       -29
#define SENSOR_EV_0            0
#define SENSOR_EV_0_3          30
#define SENSOR_EV_0_7          70
#define SENSOR_EV_1_0          100
#define SENSOR_EV_1_3          130
#define SENSOR_EV_1_7          170
#define SENSOR_EV_2_0          200

/* Unique ID allocation */
static struct i2c_client *g_client;

#define I2C_DEBUG
//#undef I2C_DEBUG

#ifdef  I2C_DEBUG
#define ovI2C_dbg(fmt, arg...)    printk(KERN_INFO "%s(line %d): " fmt "\n",  \
                __FUNCTION__, __LINE__, ##arg)
#else 
#define ovI2C_dbg(fmt, arg...)        do {} while (0)
#endif

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

/****************************************************************************

   Gloable variable
   
*****************************************************************************/
struct frame_sizes_t {
	u32	width;
	u32	height;
};
static int G_mt9p111_mode_change = CAMERA_MODE_VIDEO;
static int g_mt9p111_size = -1;
static int g_mt9p111_format = -1;
#if 0
static u8  mt9p111_gaincode = 0;
static int  mt9p111_expcode = 0;
static u8  mt9p111_recRgain =0,mt9p111_recBgain=0,mt9p111_recGgain=0;
#define QXGAMAXEXPLIN 1568
#define XGAMAXEXPLIN    779
const static u8 gSensorSlaveAddr = 0x3c;
#endif

#define FLASHLED_PROC_FILE	"driver/flashled"
#define MT9P111_PROC_FILE	"driver/mt9p111"

static struct proc_dir_entry *flashled_proc_file;
static struct proc_dir_entry *mt9p111_proc_file;
static int flashled_flag= 0;
static int flashled_on = 0;
static int flashled_fire = 0;
static struct timeval flashled_fire_time;
static int af_flag= 0;
static int reg_format= 0, reg_size= 0, reg_preview_size = 0;

static int mt9p111_single_af_enabled = 0;
static int mt9p111set_fps = 0;

static struct delayed_work flash_work;
static struct workqueue_struct *flash_wqueue;
static DECLARE_MUTEX(mt9p111_flash_lock);
static struct delayed_work autofocus_work;
static struct workqueue_struct *autofocus_wqueue;
static DECLARE_MUTEX(mt9p111_af_lock);
static DECLARE_MUTEX(mt9p111_i2c_lock);

static const struct frame_sizes_t mt9p111_yuv422_frame_sizes[] = 
{
	{480, 360},
	{432, 320},
	{320, 240},
	{160, 120},
	{128, 96},
	{352, 288},
	{176, 144}
};

static const struct frame_sizes_t mt9p111_jpeg_frame_sizes[] = 
{
	{2560, 1920},
	{2048, 1536},
	{1600, 1200},
	{1280, 960},
	{1024, 768},
	{640, 480},
	{320, 240}
};

static inline int read_sensor_reg( const u16 subAddress, u16 *bufP );
static inline int write_sensor_reg( const u16 subAddress, u16 bufP );

static unsigned int atoi( char* str)
{
	unsigned int i= 0, radius= 10, val= 0;

	// hex
	if( ( str[ i] == '0') && ( str[ i+ 1] == 'x'))
	{
		radius= 16;
		i+= 2;
	}

	while( str[ i] != '\0')
	{
		val*= radius;

		if( ( str[ i] >= '0') && ( str[ i] <= '9'))
			val+= str[ i]- '0';
		else if( ( str[ i] >= 'A') && ( str[ i] <= 'F'))
			val+= 10+ str[ i]- 'A';
		else if( ( str[ i] >= 'a') && ( str[ i] <= 'f'))
			val+= 10+ str[ i]- 'a';
		else			// not a recognized char
		{
			val/= radius;
			break;
		}
		i++;
	}

	return val;
}

/*****************************************************************************

 *  Register Settings Function

 *****************************************************************************/

/* include AF firmware register tabs */

typedef struct{
	u16	func;
	u16	reg;
	u16	val;
} MT9P111_REG_TABLE;

//part 1
static const MT9P111_REG_TABLE MT9P111_init1[] =
{
//Section 2
	{0, 0x0010, 0x0138},  //0x0134
	{0, 0x0012, 0x00D0},  //0x00C0
	{0, 0x0014, 0x2025},
	{0, 0x001E, 0x0777},
	{0, 0x0022, 0x00F0},
	{0, 0x002A, 0x7F8C},  //0x7F8A
	{0, 0x002C, 0x0000},
	{0, 0x002E, 0x0000},
	{0, 0x0018, 0x4008},  // STANDBY_CONTROL_AND_STATUS
};

static const MT9P111_REG_TABLE MT9P111_init2[] =
{
	{0, 0x0018, 0x4008},  // STANDBY_CONTROL_AND_STATUS
	{1, 0, 50},
};

static const MT9P111_REG_TABLE MT9P111_init3[] =
{	
	{0, 0x098E, 0x483A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_Y_ADDR_START]
	{0, 0x0990, 0x000C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x483C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_X_ADDR_START]
	{0, 0x0990, 0x0018},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x483E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_Y_ADDR_END]
	{0, 0x0990, 0x07B1},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4840},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_X_ADDR_END]
	{0, 0x0990, 0x0A45},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4842},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_ROW_SPEED]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4844},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_SKIP_X_CORE]
	{0, 0x0990, 0x0103},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4846},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_SKIP_Y_CORE]
	{0, 0x0990, 0x0103},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4848},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_SKIP_X_PIPE]
	{0, 0x0990, 0x0103},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x484A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_SKIP_Y_PIPE]
	{0, 0x0990, 0x0103},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x484C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_POWER_MODE]
	{0, 0x0990, 0x0080},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x484E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_BIN_MODE]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC850},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_ORIENTATION]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC851},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_PIXEL_ORDER]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4852},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_FINE_CORRECTION]
	{0, 0x0990, 0x019C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4854},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_FINE_ITMIN]
	{0, 0x0990, 0x0732},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4858},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_COARSE_ITMIN]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x485A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_COARSE_ITMAX_MARGIN]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x485C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_MIN_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x0423},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x485E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_MAX_FRAME_LENGTH_LINES]
	{0, 0x0990, 0xFFFF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4860},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_BASE_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x0423},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4862},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_MIN_LINE_LENGTH_PCLK]
	{0, 0x0990, 0x1A25},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4864},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_MAX_LINE_LENGTH_PCLK]
	{0, 0x0990, 0xFFFE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4866},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_P4_5_6_DIVIDER]
	{0, 0x0990, 0x7F8C},  // MCU_VARIABLE_DATA0  0x7F8A
	{0, 0x098E, 0x4868},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x0423},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x486A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_LINE_LENGTH_PCK]
	{0, 0x0990, 0x1A25},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x486C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_OUTPUT_SIZE_WIDTH]
	{0, 0x0990, 0x0518},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x486E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x03D4},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4870},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_RX_FIFO_TRIGGER_MARK]
	{0, 0x0990, 0x0014},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4858},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_COARSE_ITMIN]
	{0, 0x0990, 0x0002},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4856},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_ANALOG_CONTROL2]
	{0, 0x0990, 0x2113},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48B8},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_JPEG_CONTROL]
	{0, 0x0990, 0x0004},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AE},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_OUTPUT_FORMAT]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
};

//part 2
static const MT9P111_REG_TABLE preview_to_HVGA[] =
{
	{0, 0x098E, 0x48AA}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_WIDTH]
	{0, 0x0990, 0x01B0}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AC}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_HEIGHT]
	{0, 0x0990, 0x0140}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0             
};

//part 2
static const MT9P111_REG_TABLE preview_to_QCIF_H[] =
{
	{0, 0x098E, 0x483A},
	{0, 0xC83A, 0x004E},
	{0, 0xC83C, 0x0294},
	{0, 0xC83E, 0x0D8D},
	{0, 0xC840, 0x0A45},
	{0, 0xC86C, 0x028E},
	{0, 0xC86E, 0x034F},
	{0, 0xC8AA, 0x0090},
	{0, 0xC8AC, 0x00B0},
	{0, 0xDC0A, 0x0E04},
	{0, 0x098E, 0x8404},
	{0, 0x0990, 0x0600},
};

//part 3
static const MT9P111_REG_TABLE MT9P111_init4[] =
{
	{0, 0x098E, 0xC8ED},  // LOGICAL_ADDRESS_ACCESS [CAM_RESERVED1]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC8ED},  // LOGICAL_ADDRESS_ACCESS [CAM_RESERVED1]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4872},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_Y_ADDR_START]
	{0, 0x0990, 0x0010},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4874},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_X_ADDR_START]
	{0, 0x0990, 0x001C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4876},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_Y_ADDR_END]
	{0, 0x0990, 0x07AF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4878},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_X_ADDR_END]
	{0, 0x0990, 0x0A43},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x487A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ROW_SPEED]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x487C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_SKIP_X_CORE]
	{0, 0x0990, 0x0101},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x487E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_SKIP_Y_CORE]
	{0, 0x0990, 0x0101},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4880},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_SKIP_X_PIPE]
	{0, 0x0990, 0x0101},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4882},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_SKIP_Y_PIPE]
	{0, 0x0990, 0x0101},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4884},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_POWER_MODE]
	{0, 0x0990, 0x0064},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4886},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_BIN_MODE]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC888},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC889},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_PIXEL_ORDER]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x488A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_FINE_CORRECTION]
	{0, 0x0990, 0x009C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x488C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_FINE_ITMIN]
	{0, 0x0990, 0x034A},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4890},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_COARSE_ITMIN]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4892},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_COARSE_ITMAX_MARGIN]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4894},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_MIN_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x082F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4896},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_MAX_FRAME_LENGTH_LINES]
	{0, 0x0990, 0xFFFF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4898},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_BASE_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x082F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x489A},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_MIN_LINE_LENGTH_PCLK]
	{0, 0x0990, 0x1D52},  // MCU_VARIABLE_DATA0  0x1252
	{0, 0x098E, 0x489C},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_MAX_LINE_LENGTH_PCLK]
	{0, 0x0990, 0xFFFE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x489E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_P4_5_6_DIVIDER]
	{0, 0x0990, 0x7F8C},  // MCU_VARIABLE_DATA0  0x7F8A
	{0, 0x098E, 0x48A0},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_FRAME_LENGTH_LINES]
	{0, 0x0990, 0x082F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48A2},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_LINE_LENGTH_PCK]
	{0, 0x0990, 0x1D52},  // MCU_VARIABLE_DATA0  0x1252
	{0, 0x098E, 0x48A4},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_WIDTH]
	{0, 0x0990, 0x0A28},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48A6},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48A8},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_RX_FIFO_TRIGGER_MARK]
	{0, 0x0990, 0x0021},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4890},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_COARSE_ITMIN]
	{0, 0x0990, 0x0002},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x488E},  // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ANALOG_CONTROL2]
	{0, 0x0990, 0x2113},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48CE},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_JPEG_CONTROL]
	{0, 0x0990, 0x0015},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48C4},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_OUTPUT_FORMAT]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48C0},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]
	{0, 0x0990, 0x0A20},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48C2},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]
	{0, 0x0990, 0x0798},  // MCU_VARIABLE_DATA0

#ifdef OLD_FLICK_SETTING
	{0, 0x098E, 0xA00E},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_NUM_AUTOCOR_FUNC_VALUES_TO_CHECK]
	{0, 0x0990, 0x3200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2018},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED50HZ_FLICKER_PERIOD_IN_CONTEXT_A]
	{0, 0x0990, 0x00A7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201A},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED50HZ_FLICKER_PERIOD_IN_CONTEXT_B]
	{0, 0x0990, 0x011F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201C},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED60HZ_FLICKER_PERIOD_IN_CONTEXT_A]
	{0, 0x0990, 0x00A7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201E},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED60HZ_FLICKER_PERIOD_IN_CONTEXT_B]
	{0, 0x0990, 0x011F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2010},  // LOGICAL_ADDRESS_ACCESS [FD_MIN_EXPECTED50HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x009D},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2012},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_EXPECTED50HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x00B1},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2014},  // LOGICAL_ADDRESS_ACCESS [FD_MIN_EXPECTED60HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x0081},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2016},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_EXPECTED60HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x0095},  // MCU_VARIABLE_DATA0
#else
	{0, 0x098E, 0x2018},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED50HZ_FLICKER_PERIOD_IN_CONTEXT_A]
	{0, 0x0990, 0x00A7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201A},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED50HZ_FLICKER_PERIOD_IN_CONTEXT_B]
	{0, 0x0990, 0x011F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201C},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED60HZ_FLICKER_PERIOD_IN_CONTEXT_A]
	{0, 0x0990, 0x00A7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x201E},  // LOGICAL_ADDRESS_ACCESS [FD_EXPECTED60HZ_FLICKER_PERIOD_IN_CONTEXT_B]
	{0, 0x0990, 0x011F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2010},  // LOGICAL_ADDRESS_ACCESS [FD_MIN_EXPECTED50HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x009D},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2012},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_EXPECTED50HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x00B1},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2014},  // LOGICAL_ADDRESS_ACCESS [FD_MIN_EXPECTED60HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x0081},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2016},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_EXPECTED60HZ_FLICKER_PERIOD]
	{0, 0x0990, 0x0095},  // MCU_VARIABLE_DATA0
#endif
	{0, 0x098E, 0xA000},  // LOGICAL_ADDRESS_ACCESS [FD_STATUS]
	{0, 0x0990, 0x1400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8417},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_FD]
	{0, 0x0990, 0x0400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xA00F},  // LOGICAL_ADDRESS_ACCESS [FD_FLICKER_PERSISTENCE_TEST]
	{0, 0x0990, 0x1200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xA00E},  // LOGICAL_ADDRESS_ACCESS [FD_MAX_NUM_AUTOCOR_FUNC_VALUES_TO_CHECK]
	{0, 0x0990, 0x2800},  // MCU_VARIABLE_DATA0

//Section 3
	{0, 0x0982, 0x0000},
	{0, 0x098A, 0x0838},
	{0, 0x0990, 0xC0F1},
	{0, 0x0992, 0xC5E1},
	{0, 0x0994, 0x246A},
	{0, 0x0996, 0x1280},
	{0, 0x0998, 0xC4E1},
	{0, 0x099A, 0xD30E},
	{0, 0x099C, 0x2069},
	{0, 0x099E, 0x0000},
	{0, 0x098A, 0x0848},
	{0, 0x0990, 0x1305},
	{0, 0x0992, 0x0084},
	{0, 0x0994, 0x1730},
	{0, 0x0996, 0x7005},
	{0, 0x0998, 0xD801},
	{0, 0x099A, 0x8B41},
	{0, 0x099C, 0xD900},
	{0, 0x099E, 0x0D0A},
	{0, 0x098A, 0x0858},
	{0, 0x0990, 0x0064},
	{0, 0x0992, 0x8B63},
	{0, 0x0994, 0xE809},
	{0, 0x0996, 0xD509},
	{0, 0x0998, 0xF004},
	{0, 0x099A, 0x1504},
	{0, 0x099C, 0x1400},
	{0, 0x099E, 0x7840},
	{0, 0x098A, 0x0868},
	{0, 0x0990, 0xD007},
	{0, 0x0992, 0x0DFB},
	{0, 0x0994, 0x9004},
	{0, 0x0996, 0xC4C1},
	{0, 0x0998, 0x2029},
	{0, 0x099A, 0x0300},
	{0, 0x099C, 0x0191},
	{0, 0x099E, 0x0684},
	{0, 0x098A, 0x0878},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x089C},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0838},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x0888},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x089C},
	{0, 0x098A, 0x0888},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x1184},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x10F0},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x0F14},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x11F0},
	{0, 0x098A, 0x0898},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x08A4},
	{0, 0x0994, 0x0008},
	{0, 0x0996, 0x0002},
	{0, 0x0998, 0x0000},
	{0, 0x099A, 0x0000},
	{0, 0x099C, 0xC0F1},
	{0, 0x099E, 0x08EA},
	{0, 0x098A, 0x08A8},
	{0, 0x0990, 0x06A4},
	{0, 0x0992, 0xDA30},
	{0, 0x0994, 0xD7DB},
	{0, 0x0996, 0xD6DC},
	{0, 0x0998, 0x75E9},
	{0, 0x099A, 0x70A9},
	{0, 0x099C, 0x0F2E},
	{0, 0x099E, 0x0664},
	{0, 0x098A, 0x08B8},
	{0, 0x0990, 0x8620},
	{0, 0x0992, 0xD0DA},
	{0, 0x0994, 0xA705},
	{0, 0x0996, 0x0139},
	{0, 0x0998, 0x06A4},
	{0, 0x099A, 0xA6A0},
	{0, 0x099C, 0xC5E1},
	{0, 0x099E, 0xC6E1},
	{0, 0x098A, 0x08C8},
	{0, 0x0990, 0xD3D7},
	{0, 0x0992, 0x8B38},
	{0, 0x0994, 0x2340},
	{0, 0x0996, 0x05C2},
	{0, 0x0998, 0x7030},
	{0, 0x099A, 0x003C},
	{0, 0x099C, 0x002C},
	{0, 0x099E, 0xD900},
	{0, 0x098A, 0x08D8},
	{0, 0x0990, 0x8BB2},
	{0, 0x0992, 0x62A9},
	{0, 0x0994, 0x080B},
	{0, 0x0996, 0x0042},
	{0, 0x0998, 0x6D29},
	{0, 0x099A, 0x792F},
	{0, 0x099C, 0xF014},
	{0, 0x099E, 0xD901},
	{0, 0x098A, 0x08E8},
	{0, 0x0990, 0xF003},
	{0, 0x0992, 0xE101},
	{0, 0x0994, 0x792F},
	{0, 0x0996, 0x633A},
	{0, 0x0998, 0x8A58},
	{0, 0x099A, 0x090D},
	{0, 0x099C, 0x0363},
	{0, 0x099E, 0x633E},
	{0, 0x098A, 0x08F8},
	{0, 0x0990, 0x08F3},
	{0, 0x0992, 0x8083},
	{0, 0x0994, 0x8E77},
	{0, 0x0996, 0x7A02},
	{0, 0x0998, 0x2002},
	{0, 0x099A, 0x00CD},
	{0, 0x099C, 0x0A09},
	{0, 0x099E, 0x0342},
	{0, 0x098A, 0x0908},
	{0, 0x0990, 0xB961},
	{0, 0x0992, 0x792F},
	{0, 0x0994, 0x7028},
	{0, 0x0996, 0xC6C1},
	{0, 0x0998, 0x7FE0},
	{0, 0x099A, 0xC5C1},
	{0, 0x099C, 0xC0F1},
	{0, 0x099E, 0x0852},
	{0, 0x098A, 0x0918},
	{0, 0x0990, 0x0684},
	{0, 0x0992, 0xC1A4},
	{0, 0x0994, 0xD0C0},
	{0, 0x0996, 0xD5C3},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x8009},
	{0, 0x099C, 0x75FA},
	{0, 0x099E, 0x7860},
	{0, 0x098A, 0x0928},
	{0, 0x0990, 0x2780},
	{0, 0x0992, 0x2B06},
	{0, 0x0994, 0xE081},
	{0, 0x0996, 0x04DE},
	{0, 0x0998, 0x0009},
	{0, 0x099A, 0xD6BD},
	{0, 0x099C, 0x2540},
	{0, 0x099E, 0x1F10},
	{0, 0x098A, 0x0938},
	{0, 0x0990, 0x2554},
	{0, 0x0992, 0x1DD9},
	{0, 0x0994, 0x2640},
	{0, 0x0996, 0x15C1},
	{0, 0x0998, 0xC140},
	{0, 0x099A, 0x963D},
	{0, 0x099C, 0x2104},
	{0, 0x099E, 0x0F81},
	{0, 0x098A, 0x0948},
	{0, 0x0990, 0x0000},
	{0, 0x0992, 0xFF00},
	{0, 0x0994, 0xB948},
	{0, 0x0996, 0x8E46},
	{0, 0x0998, 0xC141},
	{0, 0x099A, 0xC242},
	{0, 0x099C, 0x96A0},
	{0, 0x099E, 0x0A31},
	{0, 0x098A, 0x0958},
	{0, 0x0990, 0x0051},
	{0, 0x0992, 0xD7B5},
	{0, 0x0994, 0xDA01},
	{0, 0x0996, 0xBDA4},
	{0, 0x0998, 0xBDAD},
	{0, 0x099A, 0xAF40},
	{0, 0x099C, 0xBDAF},
	{0, 0x099E, 0xB6A0},
	{0, 0x098A, 0x0968},
	{0, 0x0990, 0xE886},
	{0, 0x0992, 0xD800},
	{0, 0x0994, 0x1E3E},
	{0, 0x0996, 0x1002},
	{0, 0x0998, 0x1E40},
	{0, 0x099A, 0x1002},
	{0, 0x099C, 0x258A},
	{0, 0x099E, 0x1008},
	{0, 0x098A, 0x0978},
	{0, 0x0990, 0xD800},
	{0, 0x0992, 0xAF04},
	{0, 0x0994, 0xAF45},
	{0, 0x0996, 0x1E3F},
	{0, 0x0998, 0x1042},
	{0, 0x099A, 0xF019},
	{0, 0x099C, 0x2586},
	{0, 0x099E, 0x17F0},
	{0, 0x098A, 0x0988},
	{0, 0x0990, 0x258C},
	{0, 0x0992, 0x9004},
	{0, 0x0994, 0x25CC},
	{0, 0x0996, 0x9F82},
	{0, 0x0998, 0x0000},
	{0, 0x099A, 0x0200},
	{0, 0x099C, 0xF40F},
	{0, 0x099E, 0xD0A7},
	{0, 0x098A, 0x0998},
	{0, 0x0990, 0x8000},
	{0, 0x0992, 0x1638},
	{0, 0x0994, 0x1104},
	{0, 0x0996, 0x8663},
	{0, 0x0998, 0x710A},
	{0, 0x099A, 0x80F3},
	{0, 0x099C, 0x702B},
	{0, 0x099E, 0x7F60},
	{0, 0x098A, 0x09A8},
	{0, 0x0990, 0x8642},
	{0, 0x0992, 0xD1A1},
	{0, 0x0994, 0xA904},
	{0, 0x0996, 0xD800},
	{0, 0x0998, 0xA905},
	{0, 0x099A, 0x9619},
	{0, 0x099C, 0x1630},
	{0, 0x099E, 0x1113},
	{0, 0x098A, 0x09B8},
	{0, 0x0990, 0xC043},
	{0, 0x0992, 0xB61A},
	{0, 0x0994, 0x8602},
	{0, 0x0996, 0x8623},
	{0, 0x0998, 0x2005},
	{0, 0x099A, 0x807E},
	{0, 0x099C, 0x1E32},
	{0, 0x099E, 0x14C4},
	{0, 0x098A, 0x09C8},
	{0, 0x0990, 0xF223},
	{0, 0x0992, 0xDF00},
	{0, 0x0994, 0x254A},
	{0, 0x0996, 0x2640},
	{0, 0x0998, 0x773A},
	{0, 0x099A, 0x775A},
	{0, 0x099C, 0x2153},
	{0, 0x099E, 0x0054},
	{0, 0x098A, 0x09D8},
	{0, 0x0990, 0x7782},
	{0, 0x0992, 0x20F4},
	{0, 0x0994, 0x2484},
	{0, 0x0996, 0x7FF0},
	{0, 0x0998, 0x2C05},
	{0, 0x099A, 0x053E},
	{0, 0x099C, 0x2100},
	{0, 0x099E, 0x2E51},
	{0, 0x098A, 0x09E8},
	{0, 0x0990, 0x0A9A},
	{0, 0x0992, 0x06A4},
	{0, 0x0994, 0xDA02},
	{0, 0x0996, 0x2542},
	{0, 0x0998, 0x2055},
	{0, 0x099A, 0x2240},
	{0, 0x099C, 0x2052},
	{0, 0x099E, 0x0DE1},
	{0, 0x098A, 0x09F8},
	{0, 0x0990, 0xA075},
	{0, 0x0992, 0x222F},
	{0, 0x0994, 0x2487},
	{0, 0x0996, 0x702A},
	{0, 0x0998, 0x0846},
	{0, 0x099A, 0x06A4},
	{0, 0x099C, 0x71E9},
	{0, 0x099E, 0x222F},
	{0, 0x098A, 0x0A08},
	{0, 0x0990, 0x2008},
	{0, 0x0992, 0xB618},
	{0, 0x0994, 0xF008},
	{0, 0x0996, 0xD08A},
	{0, 0x0998, 0x101A},
	{0, 0x099A, 0x0112},
	{0, 0x099C, 0xDF00},
	{0, 0x099E, 0x779A},
	{0, 0x098A, 0x0A18},
	{0, 0x0990, 0x1E30},
	{0, 0x0992, 0x1484},
	{0, 0x0994, 0x1606},
	{0, 0x0996, 0x1096},
	{0, 0x0998, 0x0E17},
	{0, 0x099A, 0x2051},
	{0, 0x099C, 0x8E18},
	{0, 0x099E, 0x2014},
	{0, 0x098A, 0x0A28},
	{0, 0x0990, 0x0500},
	{0, 0x0992, 0x7710},
	{0, 0x0994, 0x214A},
	{0, 0x0996, 0x2040},
	{0, 0x0998, 0x21CA},
	{0, 0x099A, 0x202B},
	{0, 0x099C, 0xF009},
	{0, 0x099E, 0xD081},
	{0, 0x098A, 0x0A38},
	{0, 0x0990, 0x8824},
	{0, 0x0992, 0x8805},
	{0, 0x0994, 0x214A},
	{0, 0x0996, 0x2040},
	{0, 0x0998, 0x7110},
	{0, 0x099A, 0x21C2},
	{0, 0x099C, 0x244A},
	{0, 0x099E, 0xE580},
	{0, 0x098A, 0x0A48},
	{0, 0x0990, 0x03C2},
	{0, 0x0992, 0x0001},
	{0, 0x0994, 0xD07B},
	{0, 0x0996, 0x162C},
	{0, 0x0998, 0x1094},
	{0, 0x099A, 0x1004},
	{0, 0x099C, 0x0095},
	{0, 0x099E, 0x258C},
	{0, 0x098A, 0x0A58},
	{0, 0x0990, 0x9002},
	{0, 0x0992, 0x8829},
	{0, 0x0994, 0x163B},
	{0, 0x0996, 0x1098},
	{0, 0x0998, 0xF242},
	{0, 0x099A, 0x006C},
	{0, 0x099C, 0x0009},
	{0, 0x099E, 0xE5A0},
	{0, 0x098A, 0x0A68},
	{0, 0x0990, 0x022E},
	{0, 0x0992, 0x0001},
	{0, 0x0994, 0xE5C0},
	{0, 0x0996, 0x0394},
	{0, 0x0998, 0x0002},
	{0, 0x099A, 0x8E15},
	{0, 0x099C, 0x0A11},
	{0, 0x099E, 0x2003},
	{0, 0x098A, 0x0A78},
	{0, 0x0990, 0x9601},
	{0, 0x0992, 0x080D},
	{0, 0x0994, 0x025E},
	{0, 0x0996, 0x9600},
	{0, 0x0998, 0xB88D},
	{0, 0x099A, 0xB600},
	{0, 0x099C, 0xD06E},
	{0, 0x099E, 0x710A},
	{0, 0x098A, 0x0A88},
	{0, 0x0990, 0x0CAA},
	{0, 0x0992, 0x0664},
	{0, 0x0994, 0xDA32},
	{0, 0x0996, 0xD06B},
	{0, 0x0998, 0x8804},
	{0, 0x099A, 0xAE07},
	{0, 0x099C, 0x8E06},
	{0, 0x099E, 0x080D},
	{0, 0x098A, 0x0A98},
	{0, 0x0990, 0x0413},
	{0, 0x0992, 0x2700},
	{0, 0x0994, 0x200C},
	{0, 0x0996, 0xAC00},
	{0, 0x0998, 0x9602},
	{0, 0x099A, 0x0817},
	{0, 0x099C, 0x0410},
	{0, 0x099E, 0x8E26},
	{0, 0x098A, 0x0AA8},
	{0, 0x0990, 0x090F},
	{0, 0x0992, 0x0453},
	{0, 0x0994, 0xD064},
	{0, 0x0996, 0x9658},
	{0, 0x0998, 0xB862},
	{0, 0x099A, 0x7834},
	{0, 0x099C, 0xB040},
	{0, 0x099E, 0xD800},
	{0, 0x098A, 0x0AB8},
	{0, 0x0990, 0x96A2},
	{0, 0x0992, 0xAE06},
	{0, 0x0994, 0xD05C},
	{0, 0x0996, 0xD902},
	{0, 0x0998, 0xBD24},
	{0, 0x099A, 0xBDC0},
	{0, 0x099C, 0xA820},
	{0, 0x099E, 0x033D},
	{0, 0x098A, 0x0AC8},
	{0, 0x0990, 0x0020},
	{0, 0x0992, 0xBD07},
	{0, 0x0994, 0x258C},
	{0, 0x0996, 0x9004},
	{0, 0x0998, 0xF231},
	{0, 0x099A, 0x258C},
	{0, 0x099C, 0x9008},
	{0, 0x099E, 0x032C},
	{0, 0x098A, 0x0AD8},
	{0, 0x0990, 0x0002},
	{0, 0x0992, 0xD055},
	{0, 0x0994, 0x8804},
	{0, 0x0996, 0x088F},
	{0, 0x0998, 0x0010},
	{0, 0x099A, 0xF03E},
	{0, 0x099C, 0xD053},
	{0, 0x099E, 0x8000},
	{0, 0x098A, 0x0AE8},
	{0, 0x0990, 0x1636},
	{0, 0x0992, 0x1104},
	{0, 0x0994, 0x8663},
	{0, 0x0996, 0x710A},
	{0, 0x0998, 0x104C},
	{0, 0x099A, 0x0005},
	{0, 0x099C, 0x702B},
	{0, 0x099E, 0x2023},
	{0, 0x098A, 0x0AF8},
	{0, 0x0990, 0x0140},
	{0, 0x0992, 0x8642},
	{0, 0x0994, 0xE8B0},
	{0, 0x0996, 0xD04D},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x1636},
	{0, 0x099C, 0x1104},
	{0, 0x099E, 0x8663},
	{0, 0x098A, 0x0B08},
	{0, 0x0990, 0x710A},
	{0, 0x0992, 0x80F3},
	{0, 0x0994, 0xD04C},
	{0, 0x0996, 0x7F60},
	{0, 0x0998, 0x8642},
	{0, 0x099A, 0xE8A6},
	{0, 0x099C, 0xC101},
	{0, 0x099E, 0x163F},
	{0, 0x098A, 0x0B18},
	{0, 0x0990, 0x1080},
	{0, 0x0992, 0x085D},
	{0, 0x0994, 0x0042},
	{0, 0x0996, 0x163E},
	{0, 0x0998, 0x1080},
	{0, 0x099A, 0x7013},
	{0, 0x099C, 0x02DC},
	{0, 0x099E, 0x000B},
	{0, 0x098A, 0x0B28},
	{0, 0x0990, 0xD801},
	{0, 0x0992, 0x02D9},
	{0, 0x0994, 0x0020},
	{0, 0x0996, 0x1E40},
	{0, 0x0998, 0x1002},
	{0, 0x099A, 0xD03F},
	{0, 0x099C, 0x8804},
	{0, 0x099E, 0xE80D},
	{0, 0x098A, 0x0B38},
	{0, 0x0990, 0x163E},
	{0, 0x0992, 0x1080},
	{0, 0x0994, 0x7013},
	{0, 0x0996, 0x02C4},
	{0, 0x0998, 0x000A},
	{0, 0x099A, 0xE001},
	{0, 0x099C, 0x780F},
	{0, 0x099E, 0x02BD},
	{0, 0x098A, 0x0B48},
	{0, 0x0990, 0x0020},
	{0, 0x0992, 0x1E3E},
	{0, 0x0994, 0x1002},
	{0, 0x0996, 0xD800},
	{0, 0x0998, 0x1E3F},
	{0, 0x099A, 0x1002},
	{0, 0x099C, 0x02AD},
	{0, 0x099E, 0x0020},
	{0, 0x098A, 0x0B58},
	{0, 0x0990, 0x258A},
	{0, 0x0992, 0x1008},
	{0, 0x0994, 0xD800},
	{0, 0x0996, 0x1E3E},
	{0, 0x0998, 0x1002},
	{0, 0x099A, 0x02A1},
	{0, 0x099C, 0x0020},
	{0, 0x099E, 0x258A},
	{0, 0x098A, 0x0B68},
	{0, 0x0990, 0x1004},
	{0, 0x0992, 0xC101},
	{0, 0x0994, 0x163F},
	{0, 0x0996, 0x1080},
	{0, 0x0998, 0x0811},
	{0, 0x099A, 0x0043},
	{0, 0x099C, 0xE001},
	{0, 0x099E, 0x780F},
	{0, 0x098A, 0x0B78},
	{0, 0x0990, 0x0289},
	{0, 0x0992, 0x0020},
	{0, 0x0994, 0x1E3F},
	{0, 0x0996, 0x1002},
	{0, 0x0998, 0xD02B},
	{0, 0x099A, 0x8805},
	{0, 0x099C, 0xE805},
	{0, 0x099E, 0xD900},
	{0, 0x098A, 0x0B88},
	{0, 0x0990, 0x1E3F},
	{0, 0x0992, 0x1042},
	{0, 0x0994, 0xF016},
	{0, 0x0996, 0x163E},
	{0, 0x0998, 0x1080},
	{0, 0x099A, 0x080D},
	{0, 0x099C, 0x0602},
	{0, 0x099E, 0xD801},
	{0, 0x098A, 0x0B98},
	{0, 0x0990, 0x1E40},
	{0, 0x0992, 0x1002},
	{0, 0x0994, 0xD025},
	{0, 0x0996, 0x8000},
	{0, 0x0998, 0x163C},
	{0, 0x099A, 0x1104},
	{0, 0x099C, 0x8663},
	{0, 0x099E, 0x710A},
	{0, 0x098A, 0x0BA8},
	{0, 0x0990, 0x104C},
	{0, 0x0992, 0x0005},
	{0, 0x0994, 0xD024},
	{0, 0x0996, 0x2023},
	{0, 0x0998, 0x0140},
	{0, 0x099A, 0x8642},
	{0, 0x099C, 0xD11E},
	{0, 0x099E, 0xA905},
	{0, 0x098A, 0x0BB8},
	{0, 0x0990, 0x9622},
	{0, 0x0992, 0x090F},
	{0, 0x0994, 0x015E},
	{0, 0x0996, 0xD800},
	{0, 0x0998, 0x0241},
	{0, 0x099A, 0x0020},
	{0, 0x099C, 0xAE06},
	{0, 0x099E, 0x08CF},
	{0, 0x098A, 0x0BC8},
	{0, 0x0990, 0x0010},
	{0, 0x0992, 0x8E46},
	{0, 0x0994, 0x202F},
	{0, 0x0996, 0x0500},
	{0, 0x0998, 0xE281},
	{0, 0x099A, 0x21CC},
	{0, 0x099C, 0x8221},
	{0, 0x099E, 0xF20A},
	{0, 0x098A, 0x0BD8},
	{0, 0x0990, 0x8E78},
	{0, 0x0992, 0x2314},
	{0, 0x0994, 0x0503},
	{0, 0x0996, 0x73B2},
	{0, 0x0998, 0xDB01},
	{0, 0x099A, 0x23CA},
	{0, 0x099C, 0x0029},
	{0, 0x099E, 0x733A},
	{0, 0x098A, 0x0BE8},
	{0, 0x0990, 0x0959},
	{0, 0x0992, 0x2010},
	{0, 0x0994, 0x8EF8},
	{0, 0x0996, 0x2714},
	{0, 0x0998, 0x1503},
	{0, 0x099A, 0x0B65},
	{0, 0x099C, 0x0543},
	{0, 0x099E, 0xC300},
	{0, 0x098A, 0x0BF8},
	{0, 0x0990, 0x8EB2},
	{0, 0x0992, 0x63AB},
	{0, 0x0994, 0x238C},
	{0, 0x0996, 0x8FC3},
	{0, 0x0998, 0x23CC},
	{0, 0x099A, 0x8541},
	{0, 0x099C, 0xF406},
	{0, 0x099E, 0xB861},
	{0, 0x098A, 0x0C08},
	{0, 0x0990, 0x2502},
	{0, 0x0992, 0x200F},
	{0, 0x0994, 0x7FF0},
	{0, 0x0996, 0xF024},
	{0, 0x0998, 0x2517},
	{0, 0x099A, 0x250F},
	{0, 0x099C, 0xF01F},
	{0, 0x099E, 0x78E0},
	{0, 0x098A, 0x0C18},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x1360},
	{0, 0x0994, 0x8000},
	{0, 0x0996, 0x0124},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x0914},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x0574},
	{0, 0x098A, 0x0C28},
	{0, 0x0990, 0x8000},
	{0, 0x0992, 0x02E0},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x1348},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x0150},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x07C8},
	{0, 0x098A, 0x0C38},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x0440},
	{0, 0x0994, 0x8000},
	{0, 0x0996, 0x0424},
	{0, 0x0998, 0xC000},
	{0, 0x099A, 0x8E72},
	{0, 0x099C, 0x606F},
	{0, 0x099E, 0x2717},
	{0, 0x098A, 0x0C48},
	{0, 0x0990, 0x1500},
	{0, 0x0992, 0x0D0D},
	{0, 0x0994, 0x2003},
	{0, 0x0996, 0x2514},
	{0, 0x0998, 0x250F},
	{0, 0x099A, 0x7FF0},
	{0, 0x099C, 0x090B},
	{0, 0x099E, 0x0411},
	{0, 0x098A, 0x0C58},
	{0, 0x0990, 0xD801},
	{0, 0x0992, 0xAE06},
	{0, 0x0994, 0xF007},
	{0, 0x0996, 0x17E0},
	{0, 0x0998, 0xF000},
	{0, 0x099A, 0x9638},
	{0, 0x099C, 0xB862},
	{0, 0x099E, 0x7854},
	{0, 0x098A, 0x0C68},
	{0, 0x0990, 0xB020},
	{0, 0x0992, 0x17D0},
	{0, 0x0994, 0xF000},
	{0, 0x0996, 0x8E46},
	{0, 0x0998, 0x8824},
	{0, 0x099A, 0x2742},
	{0, 0x099C, 0x2040},
	{0, 0x099E, 0x6058},
	{0, 0x098A, 0x0C78},
	{0, 0x0990, 0xA820},
	{0, 0x0992, 0xD166},
	{0, 0x0994, 0x8120},
	{0, 0x0996, 0x8121},
	{0, 0x0998, 0xDD20},
	{0, 0x099A, 0x7960},
	{0, 0x099C, 0x78EF},
	{0, 0x099E, 0x8E06},
	{0, 0x098A, 0x0C88},
	{0, 0x0990, 0x70E2},
	{0, 0x0992, 0xA8E0},
	{0, 0x0994, 0xD802},
	{0, 0x0996, 0xAE06},
	{0, 0x0998, 0xF0B8},
	{0, 0x099A, 0xDD80},
	{0, 0x099C, 0xF0B6},
	{0, 0x099E, 0x090B},
	{0, 0x098A, 0x0C98},
	{0, 0x0990, 0x2010},
	{0, 0x0992, 0x0E1F},
	{0, 0x0994, 0x2090},
	{0, 0x0996, 0x1790},
	{0, 0x0998, 0xF002},
	{0, 0x099A, 0x8A02},
	{0, 0x099C, 0x0813},
	{0, 0x099E, 0x0582},
	{0, 0x098A, 0x0CA8},
	{0, 0x0990, 0x8A01},
	{0, 0x0992, 0x2200},
	{0, 0x0994, 0x2012},
	{0, 0x0996, 0x222F},
	{0, 0x0998, 0x2488},
	{0, 0x099A, 0x1E30},
	{0, 0x099C, 0x1484},
	{0, 0x099E, 0x0B49},
	{0, 0x098A, 0x0CB8},
	{0, 0x0990, 0x2482},
	{0, 0x0992, 0x0A29},
	{0, 0x0994, 0x24C3},
	{0, 0x0996, 0x214C},
	{0, 0x0998, 0xA000},
	{0, 0x099A, 0x214A},
	{0, 0x099C, 0x2040},
	{0, 0x099E, 0x21C2},
	{0, 0x098A, 0x0CC8},
	{0, 0x0990, 0x2442},
	{0, 0x0992, 0x2100},
	{0, 0x0994, 0x0500},
	{0, 0x0996, 0xC102},
	{0, 0x0998, 0x7810},
	{0, 0x099A, 0xE182},
	{0, 0x099C, 0x25CA},
	{0, 0x099E, 0x1F89},
	{0, 0x098A, 0x0CD8},
	{0, 0x0990, 0x0000},
	{0, 0x0992, 0x0040},
	{0, 0x0994, 0xDA00},
	{0, 0x0996, 0xF02A},
	{0, 0x0998, 0xC002},
	{0, 0x099A, 0x0823},
	{0, 0x099C, 0x00D2},
	{0, 0x099E, 0xDD40},
	{0, 0x098A, 0x0CE8},
	{0, 0x0990, 0x214C},
	{0, 0x0992, 0xA000},
	{0, 0x0994, 0x214A},
	{0, 0x0996, 0x2040},
	{0, 0x0998, 0x21C2},
	{0, 0x099A, 0x2442},
	{0, 0x099C, 0x2100},
	{0, 0x099E, 0x0500},
	{0, 0x098A, 0x0CF8},
	{0, 0x0990, 0x7810},
	{0, 0x0992, 0xF1F2},
	{0, 0x0994, 0x9601},
	{0, 0x0996, 0x080B},
	{0, 0x0998, 0x01DF},
	{0, 0x099A, 0x708A},
	{0, 0x099C, 0xF1EC},
	{0, 0x099E, 0xC003},
	{0, 0x098A, 0x0D08},
	{0, 0x0990, 0x08FB},
	{0, 0x0992, 0x84C3},
	{0, 0x0994, 0x2202},
	{0, 0x0996, 0x24C0},
	{0, 0x0998, 0x2885},
	{0, 0x099A, 0x0901},
	{0, 0x099C, 0xC003},
	{0, 0x099E, 0x2302},
	{0, 0x098A, 0x0D18},
	{0, 0x0990, 0x2001},
	{0, 0x0992, 0x0D2E},
	{0, 0x0994, 0x0664},
	{0, 0x0996, 0x702F},
	{0, 0x0998, 0x162E},
	{0, 0x099A, 0x1081},
	{0, 0x099C, 0x7810},
	{0, 0x099E, 0x09DF},
	{0, 0x098A, 0x0D28},
	{0, 0x0990, 0x8003},
	{0, 0x0992, 0x162D},
	{0, 0x0994, 0x1080},
	{0, 0x0996, 0xDA01},
	{0, 0x0998, 0x0927},
	{0, 0x099A, 0x2010},
	{0, 0x099C, 0x1618}, 
	{0, 0x099E, 0x1090},
	{0, 0x098A, 0x0D38},
	{0, 0x0990, 0x2000},
	{0, 0x0992, 0x2001},
	{0, 0x0994, 0xEA04},
	{0, 0x0996, 0x71B2},
	{0, 0x0998, 0x20CA},
	{0, 0x099A, 0x0509},
	{0, 0x099C, 0x2000},
	{0, 0x099E, 0x2001},
	{0, 0x098A, 0x0D48},
	{0, 0x0990, 0x0D2F},
	{0, 0x0992, 0x2042},
	{0, 0x0994, 0x2502},
	{0, 0x0996, 0x2010},
	{0, 0x0998, 0x202F},
	{0, 0x099A, 0x2408},
	{0, 0x099C, 0xF012},
	{0, 0x099E, 0x2000},
	{0, 0x098A, 0x0D58},
	{0, 0x0990, 0x0541},
	{0, 0x0992, 0x8EF2},
	{0, 0x0994, 0xC300},
	{0, 0x0996, 0x2332},
	{0, 0x0998, 0x03D0},
	{0, 0x099A, 0xEA05},
	{0, 0x099C, 0x7112},
	{0, 0x099E, 0x20CA},
	{0, 0x098A, 0x0D68},
	{0, 0x0990, 0x0509},
	{0, 0x0992, 0x70A2},
	{0, 0x0994, 0x7810},
	{0, 0x0996, 0x7012},
	{0, 0x0998, 0x20CA},
	{0, 0x099A, 0x200C},
	{0, 0x099C, 0xF6C2},
	{0, 0x099E, 0xDD40},
	{0, 0x098A, 0x0D78},
	{0, 0x0990, 0x9602},
	{0, 0x0992, 0x0819},
	{0, 0x0994, 0x0211},
	{0, 0x0996, 0x0E15},
	{0, 0x0998, 0x2453},
	{0, 0x099A, 0x17AF},
	{0, 0x099C, 0xF600},
	{0, 0x099E, 0xB862},
	{0, 0x098A, 0x0D88},
	{0, 0x0990, 0x2014},
	{0, 0x0992, 0x0580},
	{0, 0x0994, 0x1800},
	{0, 0x0996, 0x0484},
	{0, 0x0998, 0x9601},
	{0, 0x099A, 0x2051},
	{0, 0x099C, 0x8080},
	{0, 0x099E, 0x202F},
	{0, 0x098A, 0x0D98},
	{0, 0x0990, 0x0407},
	{0, 0x0992, 0xF227},
	{0, 0x0994, 0xE5C0},
	{0, 0x0996, 0xF425},
	{0, 0x0998, 0x9622},
	{0, 0x099A, 0x0947},
	{0, 0x099C, 0x0211},
	{0, 0x099E, 0xFEC8},
	{0, 0x098A, 0x0DA8},
	{0, 0x0990, 0xD908},
	{0, 0x0992, 0xAE07},
	{0, 0x0994, 0x17A0},
	{0, 0x0996, 0xF600},
	{0, 0x0998, 0xA823},
	{0, 0x099A, 0xD802},
	{0, 0x099C, 0xB602},
	{0, 0x099E, 0x179A},
	{0, 0x098A, 0x0DB8},
	{0, 0x0990, 0xF600},
	{0, 0x0992, 0x8000},
	{0, 0x0994, 0x8007},
	{0, 0x0996, 0x7840},
	{0, 0x0998, 0x244A},
	{0, 0x099A, 0x7640},
	{0, 0x099C, 0xD800},
	{0, 0x099E, 0x20A8},
	{0, 0x098A, 0x0DC8},
	{0, 0x0990, 0x0400},
	{0, 0x0992, 0x218A},
	{0, 0x0994, 0x0FCF},
	{0, 0x0996, 0x2114},
	{0, 0x0998, 0x300C},
	{0, 0x099A, 0xB420},
	{0, 0x099C, 0x71CF},
	{0, 0x099E, 0x8000},
	{0, 0x098A, 0x0DD8},
	{0, 0x0990, 0x0424},
	{0, 0x0992, 0xDA00},
	{0, 0x0994, 0x7914},
	{0, 0x0996, 0xE001},
	{0, 0x0998, 0xB140},
	{0, 0x099A, 0x780F},
	{0, 0x099C, 0xF013},
	{0, 0x099E, 0xD10B},
	{0, 0x098A, 0x0DE8},
	{0, 0x0990, 0x8120},
	{0, 0x0992, 0x8121},
	{0, 0x0994, 0x7940},
	{0, 0x0996, 0x8E06},
	{0, 0x0998, 0x080B},
	{0, 0x099A, 0x0413},
	{0, 0x099C, 0x70E2},
	{0, 0x099E, 0x1800},
	{0, 0x098A, 0x0DF8},
	{0, 0x0990, 0x0402},
	{0, 0x0992, 0x8E06},
	{0, 0x0994, 0xE001},
	{0, 0x0996, 0xAE06},
	{0, 0x0998, 0x9600},
	{0, 0x099A, 0x2086},
	{0, 0x099C, 0x080F},
	{0, 0x099E, 0x651D},
	{0, 0x098A, 0x0E08},
	{0, 0x0990, 0xB6A0},
	{0, 0x0992, 0x039D},
	{0, 0x0994, 0x0664},
	{0, 0x0996, 0xC0A4},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x00A0},
	{0, 0x099C, 0xD049},
	{0, 0x099E, 0x218A},
	{0, 0x098A, 0x0E18},
	{0, 0x0990, 0x0FFD},
	{0, 0x0992, 0x9067},
	{0, 0x0994, 0xD048},
	{0, 0x0996, 0x9046},
	{0, 0x0998, 0xD048},
	{0, 0x099A, 0x7350},
	{0, 0x099C, 0xDA00},
	{0, 0x099E, 0x22CA},
	{0, 0x098A, 0x0E28},
	{0, 0x0990, 0x0F8C},
	{0, 0x0992, 0x0000},
	{0, 0x0994, 0x0080},
	{0, 0x0996, 0x06E9},
	{0, 0x0998, 0x0244},
	{0, 0x099A, 0x78E0},
	{0, 0x099C, 0xC0F1},
	{0, 0x099E, 0x0B56},
	{0, 0x098A, 0x0E38},
	{0, 0x0990, 0x0644},
	{0, 0x0992, 0x0A56},
	{0, 0x0994, 0x0544},
	{0, 0x0996, 0xD042},
	{0, 0x0998, 0xD53E},
	{0, 0x099A, 0x88D2},
	{0, 0x099C, 0x0D6A},
	{0, 0x099E, 0x0064},
	{0, 0x098A, 0x0E48},
	{0, 0x0990, 0x8D04},
	{0, 0x0992, 0x7EDB},
	{0, 0x0994, 0x218A},
	{0, 0x0996, 0x0FCF},
	{0, 0x0998, 0x2805},
	{0, 0x099A, 0x03BE},
	{0, 0x099C, 0x2941},
	{0, 0x099E, 0x7140},
	{0, 0x098A, 0x0E58},
	{0, 0x0990, 0x208C},
	{0, 0x0992, 0x8FCF},
	{0, 0x0994, 0x21CA},
	{0, 0x0996, 0x0005},
	{0, 0x0998, 0xD03A},
	{0, 0x099A, 0xD63B},
	{0, 0x099C, 0x182E},
	{0, 0x099E, 0x0064},
	{0, 0x098A, 0x0E68},
	{0, 0x0990, 0x16E3},
	{0, 0x0992, 0x108F},
	{0, 0x0994, 0x9507},
	{0, 0x0996, 0x165E},
	{0, 0x0998, 0x1101},
	{0, 0x099A, 0x1660},
	{0, 0x099C, 0x1102},
	{0, 0x099E, 0x73E9},
	{0, 0x098A, 0x0E78},
	{0, 0x0990, 0x2386},
	{0, 0x0992, 0x03FC},
	{0, 0x0994, 0xBB44},
	{0, 0x0996, 0x0D26},
	{0, 0x0998, 0x0064},
	{0, 0x099A, 0x2753},
	{0, 0x099C, 0x10C4},
	{0, 0x099E, 0x16E2},
	{0, 0x098A, 0x0E88},
	{0, 0x0990, 0x108F},
	{0, 0x0992, 0x2840},
	{0, 0x0994, 0x0210},
	{0, 0x0996, 0x9507},
	{0, 0x0998, 0x165E},
	{0, 0x099A, 0x1101},
	{0, 0x099C, 0x73E9},
	{0, 0x099E, 0x1660},
	{0, 0x098A, 0x0E98},
	{0, 0x0990, 0x1102},
	{0, 0x0992, 0x2386},
	{0, 0x0994, 0x03FC},
	{0, 0x0996, 0xBB44},
	{0, 0x0998, 0x0D02},
	{0, 0x099A, 0x0064},
	{0, 0x099C, 0x2753},
	{0, 0x099E, 0x10C4},
	{0, 0x098A, 0x0EA8},
	{0, 0x0990, 0xD128},
	{0, 0x0992, 0xB80C},
	{0, 0x0994, 0x2005},
	{0, 0x0996, 0x0400},
	{0, 0x0998, 0x1932},
	{0, 0x099A, 0x0024},
	{0, 0x099C, 0xFFD8},
	{0, 0x099E, 0x0339},
	{0, 0x098A, 0x0EB8},
	{0, 0x0990, 0x0644},
	{0, 0x0992, 0x78E0},
	{0, 0x0994, 0xC0F1},
	{0, 0x0996, 0xC5E1},
	{0, 0x0998, 0x0FA6},
	{0, 0x099A, 0x0564},
	{0, 0x099C, 0x7548},
	{0, 0x099E, 0x0D13},
	{0, 0x098A, 0x0EC8},
	{0, 0x0990, 0x1F81},
	{0, 0x0992, 0x0000},
	{0, 0x0994, 0x328A},
	{0, 0x0996, 0xD01E},
	{0, 0x0998, 0x8813},
	{0, 0x099A, 0x0827},
	{0, 0x099C, 0x001F},
	{0, 0x099E, 0x0D13},
	{0, 0x098A, 0x0ED8},
	{0, 0x0990, 0x1F81},
	{0, 0x0992, 0x0000},
	{0, 0x0994, 0x328E},
	{0, 0x0996, 0xD01A},
	{0, 0x0998, 0x8813},
	{0, 0x099A, 0x0817},
	{0, 0x099C, 0x005F},
	{0, 0x099E, 0x0D2D},
	{0, 0x098A, 0x0EE8},
	{0, 0x0990, 0x1F81},
	{0, 0x0992, 0x0000},
	{0, 0x0994, 0x328C},
	{0, 0x0996, 0xD016},
	{0, 0x0998, 0x8813},
	{0, 0x099A, 0x0821},
	{0, 0x099C, 0x009E},
	{0, 0x099E, 0xD417},
	{0, 0x098A, 0x0EF8},
	{0, 0x0990, 0x2574},
	{0, 0x0992, 0x9300},
	{0, 0x0994, 0x2053},
	{0, 0x0996, 0x0101},
	{0, 0x0998, 0x0911},
	{0, 0x099A, 0x07D1},
	{0, 0x099C, 0x2004},
	{0, 0x099E, 0x0F80},
	{0, 0x098A, 0x0F08},
	{0, 0x0990, 0xFFFF},
	{0, 0x0992, 0x03FF},
	{0, 0x0994, 0x1D00},
	{0, 0x0996, 0x1024},
	{0, 0x0998, 0x02F5},
	{0, 0x099A, 0x0644},
	{0, 0x099C, 0xC0F1},
	{0, 0x099E, 0x0A7E},
	{0, 0x098A, 0x0F18},
	{0, 0x0990, 0x0664},
	{0, 0x0992, 0xDA38},
	{0, 0x0994, 0xD10E},
	{0, 0x0996, 0xD50F},
	{0, 0x0998, 0x76A9},
	{0, 0x099A, 0x0812},
	{0, 0x099C, 0x0664},
	{0, 0x099E, 0x70C9},
	{0, 0x098A, 0x0F28},
	{0, 0x0990, 0xD00D},
	{0, 0x0992, 0xA503},
	{0, 0x0994, 0xD00D},
	{0, 0x0996, 0xA50D},
	{0, 0x0998, 0xD00D},
	{0, 0x099A, 0x02CD},
	{0, 0x099C, 0x0664},
	{0, 0x099E, 0xA0C0},
	{0, 0x098A, 0x0F38},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x07C8},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x1350},
	{0, 0x0998, 0x0000},
	{0, 0x099A, 0x3EE6},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x05D4},
	{0, 0x098A, 0x0F48},
	{0, 0x0990, 0xFF00},
	{0, 0x0992, 0x3390},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x06E4},
	{0, 0x0998, 0xFF00},
	{0, 0x099A, 0x0000},
	{0, 0x099C, 0x0000},
	{0, 0x099E, 0xF170},
	{0, 0x098A, 0x0F58},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x1390},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0E34},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x0EBC},
	{0, 0x099C, 0x8000},
	{0, 0x099E, 0x014C},
	{0, 0x098A, 0x0F68},
	{0, 0x0990, 0xC0F1},
	{0, 0x0992, 0x0A2A},
	{0, 0x0994, 0x0644},
	{0, 0x0996, 0xD56C},
	{0, 0x0998, 0x157E},
	{0, 0x099A, 0x190E},
	{0, 0x099C, 0x0EE2},
	{0, 0x099E, 0x0184},
	{0, 0x098A, 0x0F78},
	{0, 0x0990, 0xD06A},
	{0, 0x0992, 0xD16B},
	{0, 0x0994, 0x1D52},
	{0, 0x0996, 0x1024},
	{0, 0x0998, 0x208A},
	{0, 0x099A, 0x0004},
	{0, 0x099C, 0x1D9C},
	{0, 0x099E, 0x1024},
	{0, 0x098A, 0x0F88},
	{0, 0x0990, 0x8901},
	{0, 0x0992, 0x081D},
	{0, 0x0994, 0x005E},
	{0, 0x0996, 0x8101},
	{0, 0x0998, 0x080D},
	{0, 0x099A, 0x01DE},
	{0, 0x099C, 0xD901},
	{0, 0x099E, 0xD065},
	{0, 0x098A, 0x0F98},
	{0, 0x0990, 0xA820},
	{0, 0x0992, 0xF006},
	{0, 0x0994, 0x0809},
	{0, 0x0996, 0x019E},
	{0, 0x0998, 0x1D7E},
	{0, 0x099A, 0x13A4},
	{0, 0x099C, 0x0259},
	{0, 0x099E, 0x0644},
	{0, 0x098A, 0x0FA8},
	{0, 0x0990, 0xC0F1},
	{0, 0x0992, 0x09E6},
	{0, 0x0994, 0x0644},
	{0, 0x0996, 0xD75E},
	{0, 0x0998, 0x970E},
	{0, 0x099A, 0x0813},
	{0, 0x099C, 0x015E},
	{0, 0x099E, 0xD15E},
	{0, 0x098A, 0x0FB8},
	{0, 0x0990, 0x814E},
	{0, 0x0992, 0x9239},
	{0, 0x0994, 0xB734},
	{0, 0x0996, 0x925A},
	{0, 0x0998, 0xF005},
	{0, 0x099A, 0xD25C},
	{0, 0x099C, 0x922A},
	{0, 0x099E, 0x924B},
	{0, 0x098A, 0x0FC8},
	{0, 0x0990, 0xB734},
	{0, 0x0992, 0xD555},
	{0, 0x0994, 0xD65A},
	{0, 0x0996, 0x2053},
	{0, 0x0998, 0x8040},
	{0, 0x099A, 0x1D40},
	{0, 0x099C, 0x1064},
	{0, 0x099E, 0x1D42},
	{0, 0x098A, 0x0FD8},
	{0, 0x0990, 0x10A4},
	{0, 0x0992, 0xB755},
	{0, 0x0994, 0xF220},
	{0, 0x0996, 0x0B36},
	{0, 0x0998, 0x01A4},
	{0, 0x099A, 0xD801},
	{0, 0x099C, 0xD800},
	{0, 0x099E, 0xAF00},
	{0, 0x098A, 0x0FE8},
	{0, 0x0990, 0x1D3C},
	{0, 0x0992, 0x1024},
	{0, 0x0994, 0x8600},
	{0, 0x0996, 0x8009},
	{0, 0x0998, 0x7860},
	{0, 0x099A, 0xDF00},
	{0, 0x099C, 0x8600},
	{0, 0x099E, 0x8004},
	{0, 0x098A, 0x0FF8},
	{0, 0x0990, 0x7840},
	{0, 0x0992, 0x8600},
	{0, 0x0994, 0x8005},
	{0, 0x0996, 0x7840},
	{0, 0x0998, 0x8600},
	{0, 0x099A, 0x8006},
	{0, 0x099C, 0x7840},
	{0, 0x099E, 0xD047},
	{0, 0x098A, 0x1008},
	{0, 0x0990, 0x1D52},
	{0, 0x0992, 0x1024},
	{0, 0x0994, 0x208A},
	{0, 0x0996, 0x0004},
	{0, 0x0998, 0x1D9C},
	{0, 0x099A, 0x1024},
	{0, 0x099C, 0x8600},
	{0, 0x099E, 0x8003},
	{0, 0x098A, 0x1018},
	{0, 0x0990, 0x7840},
	{0, 0x0992, 0xF021},
	{0, 0x0994, 0x8600},
	{0, 0x0996, 0x8003},
	{0, 0x0998, 0x7840},
	{0, 0x099A, 0xD800},
	{0, 0x099C, 0x153C},
	{0, 0x099E, 0x1901},
	{0, 0x098A, 0x1028},
	{0, 0x0990, 0x2153},
	{0, 0x0992, 0x00C2},
	{0, 0x0994, 0x0A25},
	{0, 0x0996, 0x0153},
	{0, 0x0998, 0x695C},
	{0, 0x099A, 0xBAC3},
	{0, 0x099C, 0x0A1D},
	{0, 0x099E, 0x0153},
	{0, 0x098A, 0x1038},
	{0, 0x0990, 0x2942},
	{0, 0x0992, 0x0202},
	{0, 0x0994, 0xBAC3},
	{0, 0x0996, 0xE284},
	{0, 0x0998, 0x29C2},
	{0, 0x099A, 0x032C},
	{0, 0x099C, 0x21D3},
	{0, 0x099E, 0x00EC},
	{0, 0x098A, 0x1048},
	{0, 0x0990, 0x21CC},
	{0, 0x0992, 0x812C},
	{0, 0x0994, 0x20CA},
	{0, 0x0996, 0x006C},
	{0, 0x0998, 0xE86A},
	{0, 0x099A, 0xDF00},
	{0, 0x099C, 0x0ABE},
	{0, 0x099E, 0x01A4},
	{0, 0x098A, 0x1058},
	{0, 0x0990, 0x70E9},
	{0, 0x0992, 0xD134},
	{0, 0x0994, 0x8900},
	{0, 0x0996, 0xE80B},
	{0, 0x0998, 0x157E},
	{0, 0x099A, 0x1900},
	{0, 0x099C, 0xE080},
	{0, 0x099E, 0xD800},
	{0, 0x098A, 0x1068},
	{0, 0x0990, 0x20CA},
	{0, 0x0992, 0x0061},
	{0, 0x0994, 0x1D7E},
	{0, 0x0996, 0x1024},
	{0, 0x0998, 0xA9E0},
	{0, 0x099A, 0x8600},
	{0, 0x099C, 0x8007},
	{0, 0x099E, 0x7840},
	{0, 0x098A, 0x1078},
	{0, 0x0990, 0x017D},
	{0, 0x0992, 0x0644},
	{0, 0x0994, 0xC0F1},
	{0, 0x0996, 0x0B7E},
	{0, 0x0998, 0x0184},
	{0, 0x099A, 0xD229},
	{0, 0x099C, 0x920E},
	{0, 0x099E, 0x2051},
	{0, 0x098A, 0x1088},
	{0, 0x0990, 0x8180},
	{0, 0x0992, 0x9212},
	{0, 0x0994, 0xF205},
	{0, 0x0996, 0x204F},
	{0, 0x0998, 0x01C1},
	{0, 0x099A, 0xB232},
	{0, 0x099C, 0xF004},
	{0, 0x099E, 0xB8A7},
	{0, 0x098A, 0x1098},
	{0, 0x0990, 0x7910},
	{0, 0x0992, 0xB212},
	{0, 0x0994, 0xD027},
	{0, 0x0996, 0x1804},
	{0, 0x0998, 0x0064},
	{0, 0x099A, 0xC0D1},
	{0, 0x099C, 0x7EE0},
	{0, 0x099E, 0x78E0},
	{0, 0x098A, 0x10A8},
	{0, 0x0990, 0xC0F1},
	{0, 0x0992, 0xC5E1},
	{0, 0x0994, 0xD21E},
	{0, 0x0996, 0x8A01},
	{0, 0x0998, 0x9234},
	{0, 0x099A, 0x9275},
	{0, 0x099C, 0x0815},
	{0, 0x099E, 0x009E},
	{0, 0x098A, 0x10B8},
	{0, 0x0990, 0x8A31},
	{0, 0x0992, 0x8A12},
	{0, 0x0994, 0x8A73},
	{0, 0x0996, 0xB908},
	{0, 0x0998, 0x7905},
	{0, 0x099A, 0x8A14},
	{0, 0x099C, 0xBB08},
	{0, 0x099E, 0x7B05},
	{0, 0x098A, 0x10C8},
	{0, 0x0990, 0x8A00},
	{0, 0x0992, 0x92AE},
	{0, 0x0994, 0xB8A2},
	{0, 0x0996, 0xAA00},
	{0, 0x0998, 0x0D13},
	{0, 0x099A, 0x105E},
	{0, 0x099C, 0xDD05},
	{0, 0x099E, 0xBD09},
	{0, 0x098A, 0x10D8},
	{0, 0x0990, 0x0D0B},
	{0, 0x0992, 0x1043},
	{0, 0x0994, 0xB882},
	{0, 0x0996, 0xAA00},
	{0, 0x0998, 0x71A9},
	{0, 0x099A, 0xD017},
	{0, 0x099C, 0x1808},
	{0, 0x099E, 0x0064},
	{0, 0x098A, 0x10E8},
	{0, 0x0990, 0x180A},
	{0, 0x0992, 0x00E4},
	{0, 0x0994, 0x0119},
	{0, 0x0996, 0x0644},
	{0, 0x0998, 0xC0F1},
	{0, 0x099A, 0x08A2},
	{0, 0x099C, 0x0664},
	{0, 0x099E, 0xDA28},
	{0, 0x098A, 0x10F8},
	{0, 0x0990, 0xD112},
	{0, 0x0992, 0xD513},
	{0, 0x0994, 0x76A9},
	{0, 0x0996, 0x0E36},
	{0, 0x0998, 0x0624},
	{0, 0x099A, 0x70C9},
	{0, 0x099C, 0xD011},
	{0, 0x099E, 0xA507},
	{0, 0x098A, 0x1108},
	{0, 0x0990, 0xD011},
	{0, 0x0992, 0xA502},
	{0, 0x0994, 0xD011},
	{0, 0x0996, 0xA503},
	{0, 0x0998, 0xD011},
	{0, 0x099A, 0xA509},
	{0, 0x099C, 0xD008},
	{0, 0x099E, 0x00E9},
	{0, 0x098A, 0x1118},
	{0, 0x0990, 0x0664},
	{0, 0x0992, 0xA0C0},
	{0, 0x0994, 0xFF00},
	{0, 0x0996, 0x3BC4},
	{0, 0x0998, 0x0000},
	{0, 0x099A, 0xFFFF},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x0290},
	{0, 0x098A, 0x1128},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x134E},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0340},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x02D0},
	{0, 0x099C, 0x8000},
	{0, 0x099E, 0x0070},
	{0, 0x098A, 0x1138},
	{0, 0x0990, 0xFF00},
	{0, 0x0992, 0x3C3C},
	{0, 0x0994, 0xFF00},
	{0, 0x0996, 0x3C00},
	{0, 0x0998, 0x0000},
	{0, 0x099A, 0xE758},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x13C8},
	{0, 0x098A, 0x1148},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x0F68},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0FA8},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x107C},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x10A8},
	{0, 0x098A, 0x1158},
	{0, 0x0990, 0xC0F1},
	{0, 0x0992, 0xC5E1},
	{0, 0x0994, 0x0ABE},
	{0, 0x0996, 0xFFA3},
	{0, 0x0998, 0x7508},
	{0, 0x099A, 0xD111},
	{0, 0x099C, 0x79B6},
	{0, 0x099E, 0x8910},
	{0, 0x098A, 0x1168},
	{0, 0x0990, 0x080D},
	{0, 0x0992, 0x0091},
	{0, 0x0994, 0xD20F},
	{0, 0x0996, 0x9201},
	{0, 0x0998, 0xB886},
	{0, 0x099A, 0xB201},
	{0, 0x099C, 0x890F},
	{0, 0x099E, 0x080D},
	{0, 0x098A, 0x1178},
	{0, 0x0990, 0x0111},
	{0, 0x0992, 0xD00D},
	{0, 0x0994, 0xD905},
	{0, 0x0996, 0xB021},
	{0, 0x0998, 0x0085},
	{0, 0x099A, 0x0644},
	{0, 0x099C, 0xC0F1},
	{0, 0x099E, 0x080A},
	{0, 0x098A, 0x1188},
	{0, 0x0990, 0x0664},
	{0, 0x0992, 0xDA34},
	{0, 0x0994, 0xD709},
	{0, 0x0996, 0xD60A},
	{0, 0x0998, 0x75E9},
	{0, 0x099A, 0x70A9},
	{0, 0x099C, 0x0E4E},
	{0, 0x099E, 0x0624},
	{0, 0x098A, 0x1198},
	{0, 0x0990, 0x8620},
	{0, 0x0992, 0xD008},
	{0, 0x0994, 0xA701},
	{0, 0x0996, 0x0059},
	{0, 0x0998, 0x0664},
	{0, 0x099A, 0xA6A0},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x01B0},
	{0, 0x098A, 0x11A8},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x05E8},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0698},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x13F0},
	{0, 0x099C, 0x8000},
	{0, 0x099E, 0x0000},
	{0, 0x098A, 0x11B8},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x1158},
	{0, 0x0994, 0xC0F1},
	{0, 0x0996, 0x0E2E},
	{0, 0x0998, 0x0304},
	{0, 0x099A, 0xD015},
	{0, 0x099C, 0x806E},
	{0, 0x099E, 0xD115},
	{0, 0x098A, 0x11C8},
	{0, 0x0990, 0x934E},
	{0, 0x0992, 0x193A},
	{0, 0x0994, 0x00A4},
	{0, 0x0996, 0x2040},
	{0, 0x0998, 0x0F81},
	{0, 0x099A, 0x10F1},
	{0, 0x099C, 0x0082},
	{0, 0x099E, 0x0B09},
	{0, 0x098A, 0x11D8},
	{0, 0x0990, 0x0040},
	{0, 0x0992, 0x7A5D},
	{0, 0x0994, 0xBAC0},
	{0, 0x0996, 0xBA06},
	{0, 0x0998, 0xD00F},
	{0, 0x099A, 0x0B36},
	{0, 0x099C, 0x0264},
	{0, 0x099E, 0x218A},
	{0, 0x098A, 0x11E8},
	{0, 0x0990, 0x0FFE},
	{0, 0x0992, 0xC0D1},
	{0, 0x0994, 0x7EE0},
	{0, 0x0996, 0x78E0},
	{0, 0x0998, 0xC0F1},
	{0, 0x099A, 0x0F9E},
	{0, 0x099C, 0x0624},
	{0, 0x099E, 0xDA44},
	{0, 0x098A, 0x11F8},
	{0, 0x0990, 0xD00A},
	{0, 0x0992, 0xD70B},
	{0, 0x0994, 0x80C0},
	{0, 0x0996, 0x75E9},
	{0, 0x0998, 0x70A9},
	{0, 0x099A, 0x0DE2},
	{0, 0x099C, 0x0624},
	{0, 0x099E, 0x862D},
	{0, 0x098A, 0x1208},
	{0, 0x0990, 0xD008},
	{0, 0x0992, 0xA70C},
	{0, 0x0994, 0x07E9},
	{0, 0x0996, 0x0624},
	{0, 0x0998, 0xA6AD},
	{0, 0x099A, 0x78E0},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x0340},
	{0, 0x098A, 0x1218},
	{0, 0x0990, 0xFF00},
	{0, 0x0992, 0x3138},
	{0, 0x0994, 0x0000},
	{0, 0x0996, 0x3EDC},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x0094},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x1424},
	{0, 0x098A, 0x1228},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x11BC},
	{0, 0x0994, 0xC0F1},
	{0, 0x0996, 0x0F62},
	{0, 0x0998, 0x0604},
	{0, 0x099A, 0xD62E},
	{0, 0x099C, 0x8E01},
	{0, 0x099E, 0x0883},
	{0, 0x098A, 0x1238},
	{0, 0x0990, 0x011E},
	{0, 0x0992, 0xD72D},
	{0, 0x0994, 0xD52D},
	{0, 0x0996, 0x173A},
	{0, 0x0998, 0x1080},
	{0, 0x099A, 0x173B},
	{0, 0x099C, 0x1081},
	{0, 0x099E, 0x1502},
	{0, 0x098A, 0x1248},
	{0, 0x0990, 0x1084},
	{0, 0x0992, 0xB808},
	{0, 0x0994, 0x7825},
	{0, 0x0996, 0x9524},
	{0, 0x0998, 0x9545},
	{0, 0x099A, 0x090E},
	{0, 0x099C, 0x0064},
	{0, 0x099E, 0x8D60},
	{0, 0x098A, 0x1258},
	{0, 0x0990, 0x1E3C},
	{0, 0x0992, 0x1002},
	{0, 0x0994, 0x173A},
	{0, 0x0996, 0x1080},
	{0, 0x0998, 0x173B},
	{0, 0x099A, 0x1081},
	{0, 0x099C, 0x1503},
	{0, 0x099E, 0x1084},
	{0, 0x098A, 0x1268},
	{0, 0x0990, 0xB808},
	{0, 0x0992, 0x7825},
	{0, 0x0994, 0x9524},
	{0, 0x0996, 0x9545},
	{0, 0x0998, 0x08EE},
	{0, 0x099A, 0x0064},
	{0, 0x099C, 0x8D61},
	{0, 0x099E, 0x1E3D},
	{0, 0x098A, 0x1278},
	{0, 0x0990, 0x1002},
	{0, 0x0992, 0x173A},
	{0, 0x0994, 0x1080},
	{0, 0x0996, 0x173B},
	{0, 0x0998, 0x1081},
	{0, 0x099A, 0x1506},
	{0, 0x099C, 0x1084},
	{0, 0x099E, 0xB808},
	{0, 0x098A, 0x1288},
	{0, 0x0990, 0x7825},
	{0, 0x0992, 0x9524},
	{0, 0x0994, 0x9545},
	{0, 0x0996, 0x08D2},
	{0, 0x0998, 0x0064},
	{0, 0x099A, 0x8D64},
	{0, 0x099C, 0x1E3E},
	{0, 0x099E, 0x1002},
	{0, 0x098A, 0x1298},
	{0, 0x0990, 0x173A},
	{0, 0x0992, 0x1080},
	{0, 0x0994, 0x173B},
	{0, 0x0996, 0x1081},
	{0, 0x0998, 0x1507},
	{0, 0x099A, 0x1084},
	{0, 0x099C, 0xB808},
	{0, 0x099E, 0x7825},
	{0, 0x098A, 0x12A8},
	{0, 0x0990, 0x9524},
	{0, 0x0992, 0x9545},
	{0, 0x0994, 0x08B2},
	{0, 0x0996, 0x0064},
	{0, 0x0998, 0x8D65},
	{0, 0x099A, 0x1E3F},
	{0, 0x099C, 0x1002},
	{0, 0x099E, 0x0F56},
	{0, 0x098A, 0x12B8},
	{0, 0x0990, 0x0444},
	{0, 0x0992, 0x073D},
	{0, 0x0994, 0x0604},
	{0, 0x0996, 0x78E0},
	{0, 0x0998, 0xC0F1},
	{0, 0x099A, 0x0ED2},
	{0, 0x099C, 0x0624},
	{0, 0x099E, 0xDA2C},
	{0, 0x098A, 0x12C8},
	{0, 0x0990, 0xD10B},
	{0, 0x0992, 0xD60C},
	{0, 0x0994, 0x75C9},
	{0, 0x0996, 0x0C66},
	{0, 0x0998, 0x0624},
	{0, 0x099A, 0x70A9},
	{0, 0x099C, 0xD00A},
	{0, 0x099E, 0xD107},
	{0, 0x098A, 0x12D8},
	{0, 0x0990, 0xA603},
	{0, 0x0992, 0xD00A},
	{0, 0x0994, 0xA0A0},
	{0, 0x0996, 0xD00A},
	{0, 0x0998, 0x071D},
	{0, 0x099A, 0x0624},
	{0, 0x099C, 0xB035},
	{0, 0x099E, 0x78E0},
	{0, 0x098A, 0x12E8},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x05E8},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x0534},
	{0, 0x0998, 0xFF80},
	{0, 0x099A, 0x1350},
	{0, 0x099C, 0x0000},
	{0, 0x099E, 0xF074},
	{0, 0x098A, 0x12F8},
	{0, 0x0990, 0xFF80},
	{0, 0x0992, 0x1468},
	{0, 0x0994, 0xFF80},
	{0, 0x0996, 0x122C},
	{0, 0x0998, 0x8000},
	{0, 0x099A, 0x0130},
	{0, 0x099C, 0xFF80},
	{0, 0x099E, 0x0118},
	{0, 0x098A, 0x1308},
	{0, 0x0990, 0xEA0A},
	{0, 0x0992, 0x242F},
	{0, 0x0994, 0x7089},
	{0, 0x0996, 0x78E0},
	{0, 0x0998, 0x20A8},
	{0, 0x099A, 0x0180},
	{0, 0x099C, 0x1002},
	{0, 0x099E, 0x0D02},
	{0, 0x098A, 0x1318},
	{0, 0x0990, 0x1902},
	{0, 0x0992, 0x0094},
	{0, 0x0994, 0x7FE0},
	{0, 0x0996, 0x7028},
	{0, 0x0998, 0x7308},
	{0, 0x099A, 0x1000},
	{0, 0x099C, 0x0900},
	{0, 0x099E, 0x7904},
	{0, 0x098A, 0x1328},
	{0, 0x0990, 0x7947},
	{0, 0x0992, 0x1B00},
	{0, 0x0994, 0x0064},
	{0, 0x0996, 0x7EE0},
	{0, 0x0998, 0xEA0A},
	{0, 0x099A, 0x242F},
	{0, 0x099C, 0x7089},
	{0, 0x099E, 0x78E0},
	{0, 0x098A, 0x1338},
	{0, 0x0990, 0x20A8},
	{0, 0x0992, 0x0180},
	{0, 0x0994, 0x1102},
	{0, 0x0996, 0x0502},
	{0, 0x0998, 0x1802},
	{0, 0x099A, 0x00B4},
	{0, 0x099C, 0x7FE0},
	{0, 0x099E, 0x7028},
	{0, 0x098A, 0x1348},
	{0, 0x0990, 0x0002},
	{0, 0x0992, 0x0100},
	{0, 0x0994, 0x0000},
	{0, 0x0996, 0x0000},
	{0, 0x0998, 0x08F0},
	{0, 0x099A, 0x5152},
	{0, 0x099C, 0x08F0},
	{0, 0x099E, 0x5556},
	{0, 0x098A, 0x1358},
	{0, 0x0990, 0x0096},
	{0, 0x0992, 0x0014},
	{0, 0x0994, 0xFFFF},
	{0, 0x0996, 0x0000},
	{0, 0x098E, 0x0016},  // LOGICAL_ADDRESS_ACCESS [MON_ADDRESS_LO]
	{0, 0x0990, 0x0838},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x0002},  // LOGICAL_ADDRESS_ACCESS [MON_CMD]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0           
};

static const MT9P111_REG_TABLE MT9P111_init5[] =
{
//Char
	{0, 0x30D4, 0x9080},  // COLUMN_CORRECTION
	{0, 0x316E, 0xC400},  // DAC_ECL
	{0, 0x305E, 0x10A0},  // GLOBAL_GAIN
	{0, 0x3E00, 0x0010},  // SAMP_CONTROL
	{0, 0x3E02, 0xED02},  // SAMP_ADDR_EN
	{0, 0x3E04, 0xC88C},  // SAMP_RD1_SIG
	{0, 0x3E06, 0xC88C},  // SAMP_RD1_SIG_BOOST
	{0, 0x3E08, 0x700A},  // SAMP_RD1_RST
	{0, 0x3E0A, 0x701E},  // SAMP_RD1_RST_BOOST
	{0, 0x3E0C, 0x00FF},  // SAMP_RST1_EN
	{0, 0x3E0E, 0x00FF},  // SAMP_RST1_BOOST
	{0, 0x3E10, 0x00FF},  // SAMP_RST1_CLOOP_SH
	{0, 0x3E12, 0x0000},  // SAMP_RST_BOOST_SEQ
	{0, 0x3E14, 0xC78C},  // SAMP_SAMP1_SIG
	{0, 0x3E16, 0x6E06},  // SAMP_SAMP1_RST
	{0, 0x3E18, 0xA58C},  // SAMP_TX_EN
	{0, 0x3E1A, 0xA58E},  // SAMP_TX_BOOST
	{0, 0x3E1C, 0xA58E},  // SAMP_TX_CLOOP_SH
	{0, 0x3E1E, 0xC0D0},  // SAMP_TX_BOOST_SEQ
	{0, 0x3E20, 0xEB00},  // SAMP_VLN_EN
	{0, 0x3E22, 0x00FF},  // SAMP_VLN_HOLD
	{0, 0x3E24, 0xEB02},  // SAMP_VCL_EN
	{0, 0x3E26, 0xEA02},  // SAMP_COLCLAMP
	{0, 0x3E28, 0xEB0A},  // SAMP_SH_VCL
	{0, 0x3E2A, 0xEC01},  // SAMP_SH_VREF
	{0, 0x3E2C, 0xEB01},  // SAMP_SH_VBST
	{0, 0x3E2E, 0x00FF},  // SAMP_SPARE
	{0, 0x3E30, 0x00F3},  // SAMP_READOUT
	{0, 0x3E32, 0x3DFA},  // SAMP_RESET_DONE
	{0, 0x3E34, 0x00FF},  // SAMP_VLN_CLAMP
	{0, 0x3E36, 0x00F3},  // SAMP_ASC_INT
	{0, 0x3E38, 0x0000},  // SAMP_RS_CLOOP_SH_R
	{0, 0x3E3A, 0xF802},  // SAMP_RS_CLOOP_SH
	{0, 0x3E3C, 0x0FFF},  // SAMP_RS_BOOST_SEQ
	{0, 0x3E3E, 0xEA10},  // SAMP_TXLO_GND
	{0, 0x3E40, 0xEB05},  // SAMP_VLN_PER_COL
	{0, 0x3E42, 0xE5C8},  // SAMP_RD2_SIG
	{0, 0x3E44, 0xE5C8},  // SAMP_RD2_SIG_BOOST
	{0, 0x3E46, 0x8C70},  // SAMP_RD2_RST
	{0, 0x3E48, 0x8C71},  // SAMP_RD2_RST_BOOST
	{0, 0x3E4A, 0x00FF},  // SAMP_RST2_EN
	{0, 0x3E4C, 0x00FF},  // SAMP_RST2_BOOST
	{0, 0x3E4E, 0x00FF},  // SAMP_RST2_CLOOP_SH
	{0, 0x3E50, 0xE38D},  // SAMP_SAMP2_SIG
	{0, 0x3E52, 0x8B0A},  // SAMP_SAMP2_RST
	{0, 0x3E58, 0xEB0A},  // SAMP_PIX_CLAMP_EN
	{0, 0x3E5C, 0x0A00},  // SAMP_PIX_PULLUP_EN
	{0, 0x3E5E, 0x00FF},  // SAMP_PIX_PULLDOWN_EN_R
	{0, 0x3E60, 0x00FF},  // SAMP_PIX_PULLDOWN_EN_S
	{0, 0x3E90, 0x3C01},  // RST_ADDR_EN
	{0, 0x3E92, 0x00FF},  // RST_RST_EN
	{0, 0x3E94, 0x00FF},  // RST_RST_BOOST
	{0, 0x3E96, 0x3C00},  // RST_TX_EN
	{0, 0x3E98, 0x3C00},  // RST_TX_BOOST
	{0, 0x3E9A, 0x3C00},  // RST_TX_CLOOP_SH
	{0, 0x3E9C, 0xC0E0},  // RST_TX_BOOST_SEQ
	{0, 0x3E9E, 0x00FF},  // RST_RST_CLOOP_SH
	{0, 0x3EA0, 0x0000},  // RST_RST_BOOST_SEQ
	{0, 0x3EA6, 0x3C00},  // RST_PIX_PULLUP_EN
	{0, 0x3ED8, 0x3057},  // DAC_LD_12_13
	{0, 0x316C, 0xB44F},  // DAC_TXLO
	{0, 0x316E, 0xC6FF},  // DAC_ECL
	{0, 0x3ED2, 0xEA0A},  // DAC_LD_6_7
	{0, 0x3ED4, 0x0004},  // DAC_LD_8_9
	{0, 0x3EDC, 0x6020},  // DAC_LD_16_17
	{0, 0x3EE6, 0xA541},  // DAC_LD_26_27
	{0, 0x31E0, 0x0000},  // PIX_DEF_ID
	{0, 0x3ED0, 0x2409},  // DAC_LD_4_5
	{0, 0x3EDE, 0x0A49},  // DAC_LD_18_19
	{0, 0x3EE0, 0x4910},  // DAC_LD_20_21
	{0, 0x3EE2, 0x09D2},  // DAC_LD_22_23
	{0, 0x30B6, 0x000B},  // AUTOLR_CONTROL 0x0006
};

static const MT9P111_REG_TABLE MT9P111_init6[] =
{
//Section 4
//used  for BYD module
	{0,0x364A, 0x02F0},
	{0,0x364C, 0xA30C},
	{0,0x364E, 0x1E50},
	{0,0x3650, 0x778E},
	{0,0x3652, 0xC970},
	{0,0x368A, 0x370D},
	{0,0x368C, 0x1EAD},
	{0,0x368E, 0x114D},
	{0,0x3690, 0xB76E},
	{0,0x3692, 0x81EF},
	{0,0x36CA, 0x1311},
	{0,0x36CC, 0x17EA},
	{0,0x36CE, 0xA493},
	{0,0x36D0, 0x896F},
	{0,0x36D2, 0x2F54},
	{0,0x370A, 0x5ECE},
	{0,0x370C, 0xE8ED},
	{0,0x370E, 0xFB8F},
	{0,0x3710, 0x0DB0},
	{0,0x3712, 0x08D0},
	{0,0x374A, 0xF2F1},
	{0,0x374C, 0x8ECD},
	{0,0x374E, 0x63F4},
	{0,0x3750, 0x4B2D},
	{0,0x3752, 0xB655},
	{0,0x3640, 0x04B0},
	{0,0x3642, 0x602D},
	{0,0x3644, 0x1331},
	{0,0x3646, 0xB6C9},
	{0,0x3648, 0x9131},
	{0,0x3680, 0xDC89},
	{0,0x3682, 0xB62E},
	{0,0x3684, 0xECCB},
	{0,0x3686, 0x3A8F},
	{0,0x3688, 0x1CCC},
	{0,0x36C0, 0x4D71},
	{0,0x36C2, 0x144F},
	{0,0x36C4, 0xC953},
	{0,0x36C6, 0xF62F},
	{0,0x36C8, 0x2EB4},
	{0,0x3700, 0x02EF},
	{0,0x3702, 0x7FEE},
	{0,0x3704, 0x52CE},
	{0,0x3706, 0xA610},
	{0,0x3708, 0xC1B1},
	{0,0x3740, 0x84F2},
	{0,0x3742, 0xB550},
	{0,0x3744, 0x3D54},
	{0,0x3746, 0x6BD1},
	{0,0x3748, 0xBC94},
	{0,0x3654, 0x0370},
	{0,0x3656, 0x39ED},
	{0,0x3658, 0x08D0},
	{0,0x365A, 0xA4ED},
	{0,0x365C, 0xACF0},
	{0,0x3694, 0x8C2E},
	{0,0x3696, 0x014E},
	{0,0x3698, 0x508F},
	{0,0x369A, 0xF80E},
	{0,0x369C, 0xE0CF},
	{0,0x36D4, 0x1BF1},
	{0,0x36D6, 0x352F},
	{0,0x36D8, 0x9853},
	{0,0x36DA, 0xF02F},
	{0,0x36DC, 0x1DB4},
	{0,0x3714, 0xDC4E},
	{0,0x3716, 0xCF2F},
	{0,0x3718, 0x2C50},
	{0,0x371A, 0x1AB1},
	{0,0x371C, 0x8AD1},
	{0,0x3754, 0xE8B1},
	{0,0x3756, 0xE8D0},
	{0,0x3758, 0x5094},
	{0,0x375A, 0x1CD2},
	{0,0x375C, 0x8B75},        
	{0,0x365E, 0x0130},
	{0,0x3660, 0xA90D},
	{0,0x3662, 0x0D11},
	{0,0x3664, 0x016F},
	{0,0x3666, 0x90D1},
	{0,0x369E, 0xFACB},
	{0,0x36A0, 0xED2E},
	{0,0x36A2, 0x528E},
	{0,0x36A4, 0x586F},
	{0,0x36A6, 0xF78F},
	{0,0x36DE, 0x5A11},
	{0,0x36E0, 0x29ED},
	{0,0x36E2, 0xBBB3},
	{0,0x36E4, 0xE24F},
	{0,0x36E6, 0x2074},
	{0,0x371E, 0x8E0D},
	{0,0x3720, 0x68EE},
	{0,0x3722, 0x4B6F},
	{0,0x3724, 0xC46F},
	{0,0x3726, 0x604D},
	{0,0x375E, 0x8312},
	{0,0x3760, 0xA28E},
	{0,0x3762, 0x1E14},
	{0,0x3764, 0x3850},
	{0,0x3766, 0xEA73}, 
	{0,0x3784, 0x0508},
	{0,0x3782, 0x0418}, 
//end
};

static const MT9P111_REG_TABLE MT9P111_init7[] =
{
	{0, 0x3210, 0x49B8},  // COLOR_PIPELINE_CONTROL


//Section 5
	{0, 0x098E, 0xAC01},  // LOGICAL_ADDRESS_ACCESS [AWB_MODE]
	{0, 0x0990, 0xBB00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB83E},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_WINDOW_POS_X]
	{0, 0x0990, 0x0F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB83F},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_WINDOW_POS_Y]
	{0, 0x0990, 0x0F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB840},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_WINDOW_SIZE_X]
	{0, 0x0990, 0xE500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB841},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_WINDOW_SIZE_Y]
	{0, 0x0990, 0xE500},  // MCU_VARIABLE_DATA0
#ifdef USING_OLD_AWB_SETTING
	{0, 0x098E, 0x2C46},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_0]
	{0, 0x0990, 0x023D},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C48},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_1]
	{0, 0x0990, 0xFEC2},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4A},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_2]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4C},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_3]
	{0, 0x0990, 0xFFBF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4E},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_4]
	{0, 0x0990, 0x0154},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C50},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_5]
	{0, 0x0990, 0xFFED},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C52},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_6]
	{0, 0x0990, 0xFFC3},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C54},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_7]
	{0, 0x0990, 0xFE85},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C56},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_8]
	{0, 0x0990, 0x02B7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C58},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_R2BRATIO]
	{0, 0x0990, 0x00F3},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5C},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_0]
	{0, 0x0990, 0x0212},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5E},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_1]
	{0, 0x0990, 0xFF49},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C60},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_2]
	{0, 0x0990, 0xFFA4},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C62},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_3]
	{0, 0x0990, 0xFFB7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C64},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_4]
	{0, 0x0990, 0x014C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C66},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_5]
	{0, 0x0990, 0xFFFD},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C68},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_6]
	{0, 0x0990, 0xFFFE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6A},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_7]
	{0, 0x0990, 0xFF56},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6C},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_8]
	{0, 0x0990, 0x01AC},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6E},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_R2BRATIO]
	{0, 0x0990, 0x0066},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3842},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_GRAY_CHECKER_OFFSET_X]
	{0, 0x0990, 0x0043},  // MCU_VARIABLE_DATA0 0x003A
	{0, 0x098E, 0x3844},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_GRAY_CHECKER_OFFSET_Y]
	{0, 0x0990, 0x0021},  // MCU_VARIABLE_DATA0 0x0047
	{0, 0x3240, 0x0024},  // AWB_XY_SCALE
	{0, 0x3240, 0x0024},  // AWB_XY_SCALE
	{0, 0x3242, 0xD05C},  // AWB_WEIGHT_R0
	{0, 0x3244, 0xDEB5},  // AWB_WEIGHT_R1
	{0, 0x3246, 0x4D54},  // AWB_WEIGHT_R2
	{0, 0x3248, 0x6DC4},  // AWB_WEIGHT_R3
	{0, 0x324A, 0x731D},  // AWB_WEIGHT_R4
	{0, 0x324C, 0x7A90},  // AWB_WEIGHT_R5
	{0, 0x324E, 0x73E4},  // AWB_WEIGHT_R6
	{0, 0x3250, 0xA2E4},  // AWB_WEIGHT_R7
	{0, 0x3266, 0x0096},  // AWB_WEIGHT_TH
	{0, 0x098E, 0xAC3C},  // LOGICAL_ADDRESS_ACCESS [AWB_MIN_ACCEPTED_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x2C00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3D},  // LOGICAL_ADDRESS_ACCESS [AWB_MAX_ACCEPTED_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x7100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3E},  // LOGICAL_ADDRESS_ACCESS [AWB_MIN_ACCEPTED_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x1800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3F},  // LOGICAL_ADDRESS_ACCESS [AWB_MAX_ACCEPTED_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x5C00},  // MCU_VARIABLE_DATA0
#endif

//Section 6
	{0, 0x098E, 0xD80F},  // LOGICAL_ADDRESS_ACCESS [JPEG_QSCALE_0]
	{0, 0x0990, 0x0400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xD810},  // LOGICAL_ADDRESS_ACCESS [JPEG_QSCALE_1]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC8D2},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_JPEG_QSCALE_0]
	{0, 0x0990, 0x0C00},  // MCU_VARIABLE_DATA0 0x0400
	{0, 0x098E, 0xC8D3},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_JPEG_QSCALE_1]
	{0, 0x0990, 0x0C00},  // MCU_VARIABLE_DATA0 0x0A00
	{0, 0x098E, 0xC8BC},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_JPEG_QSCALE_0]
	{0, 0x0990, 0x0400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC8BD},  // LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_JPEG_QSCALE_1]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x301A, 0x10F4},  // RESET_REGISTER
	{0, 0x301E, 0x0080},  // DATA_PEDESTAL
	{0, 0x301A, 0x10FC},  // RESET_REGISTER
	{0, 0x098E, 0xDC33},  // LOGICAL_ADDRESS_ACCESS [SYS_FIRST_BLACK_LEVEL]
	{0, 0x0990, 0x2000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC35},  // LOGICAL_ADDRESS_ACCESS [SYS_UV_COLOR_BOOST]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	{0, 0x326E, 0x00A4},  // LOW_PASS_YUV_FILTER
	{0, 0x098E, 0xDC37},  // LOGICAL_ADDRESS_ACCESS [SYS_BRIGHT_COLORKILL]
	{0, 0x0990, 0x6200},  // MCU_VARIABLE_DATA0
	{0, 0x35A4, 0x0596},  // BRIGHT_COLOR_KILL_CONTROLS
	{0, 0x35A2, 0x009B},  // DARK_COLOR_KILL_CONTROLS
	{0, 0x098E, 0xDC36},  // LOGICAL_ADDRESS_ACCESS [SYS_DARK_COLOR_KILL]
	{0, 0x0990, 0x3300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C0E},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_CURVE_ADJ_START_POS]
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C10},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_CURVE_ADJ_MID_POS]
	{0, 0x0990, 0x0002},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C12},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_CURVE_ADJ_END_POS]
	{0, 0x0990, 0x03E8},  // MCU_VARIABLE_DATA0 0x0320
	{0, 0x098E, 0xBC51},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_CURVE_SELECTOR]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C14},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_FADE_TO_BLACK_START_POS]
	{0, 0x0990, 0xFBF4},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C16},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_FADE_TO_BLACK_END_POS]
	{0, 0x0990, 0xFFFF},  // MCU_VARIABLE_DATA0
	
#ifdef USING_OLD_GAMMA_SETTING	
	{0, 0x098E, 0xBC2B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_0]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_1]
	{0, 0x0990, 0x0D00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_2]
	{0, 0x0990, 0x2300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2E},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_3]
	{0, 0x0990, 0x4D00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2F},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_4]
	{0, 0x0990, 0x7800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC30},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_5]
	{0, 0x0990, 0x9200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC31},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_6]
	{0, 0x0990, 0xA600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC32},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_7]
	{0, 0x0990, 0xB600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC33},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_8]
	{0, 0x0990, 0xC300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC34},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_9]
	{0, 0x0990, 0xCE00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC35},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_10]
	{0, 0x0990, 0xD800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC36},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_11]
	{0, 0x0990, 0xE100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC37},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_12]
	{0, 0x0990, 0xE900},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC38},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_13]
	{0, 0x0990, 0xEF00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC39},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_14]
	{0, 0x0990, 0xF400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3A},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_15]
	{0, 0x0990, 0xF700},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_16]
	{0, 0x0990, 0xFA00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_17]
	{0, 0x0990, 0xFD00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_18]
	{0, 0x0990, 0xFF00},  // MCU_VARIABLE_DATA0
#endif

#ifdef USING_LEALI_GAMMA_SETTING
//leali [Gamma tune]
	{0, 0x098E, 0xBC2B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_0] 
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC2C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_1] 
	{0, 0x0990, 0x1C00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC2D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_2] 
	{0, 0x0990, 0x3900},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC2E},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_3] 
	{0, 0x0990, 0x5A00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC2F},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_4] 
	{0, 0x0990, 0x7800},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC30},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_5] 
	{0, 0x0990, 0x9200},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC31},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_6] 
	{0, 0x0990, 0xA600},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC32},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_7] 
	{0, 0x0990, 0xB600},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC33},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_8] 
	{0, 0x0990, 0xC300},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC34},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_9] 
	{0, 0x0990, 0xCE00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC35},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_10]
	{0, 0x0990, 0xD800},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC36},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_11]
	{0, 0x0990, 0xE100},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC37},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_12]
	{0, 0x0990, 0xE900},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC38},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_13]
	{0, 0x0990, 0xEF00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC39},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_14]
	{0, 0x0990, 0xF400},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC3A},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_15]
	{0, 0x0990, 0xF700},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC3B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_16]
	{0, 0x0990, 0xFA00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC3C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_17]
	{0, 0x0990, 0xFD00},  // MCU_VARIABLE_DATA0                                
	{0, 0x098E, 0xBC3D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_18]
	{0, 0x0990, 0xFF00},  // MCU_VARIABLE_DATA0                                
#endif

//Lisa
	{0, 0x098E, 0xBC2B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_0]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_1]
	{0, 0x0990, 0x1300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_2]
	{0, 0x0990, 0x2700},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2E},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_3]
	{0, 0x0990, 0x4D00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC2F},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_4]
	{0, 0x0990, 0x7800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC30},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_5]
	{0, 0x0990, 0x9200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC31},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_6]
	{0, 0x0990, 0xA600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC32},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_7]
	{0, 0x0990, 0xB600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC33},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_8]
	{0, 0x0990, 0xC300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC34},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_9]
	{0, 0x0990, 0xCE00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC35},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_10]
	{0, 0x0990, 0xD800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC36},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_11]
	{0, 0x0990, 0xE100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC37},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_12]
	{0, 0x0990, 0xE900},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC38},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_13]
	{0, 0x0990, 0xEF00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC39},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_14]
	{0, 0x0990, 0xF400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3A},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_15]
	{0, 0x0990, 0xF700},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3B},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_16]
	{0, 0x0990, 0xFA00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3C},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_17]
	{0, 0x0990, 0xFD00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC3D},  // LOGICAL_ADDRESS_ACCESS [LL_GAMMA_NEUTRAL_CURVE_18]
	{0, 0x0990, 0xFF00},  // MCU_VARIABLE_DATA0
//ends

	{0, 0x3540, 0x0000},  // ENABLE_TONAL_CURVE
	{0, 0x098E, 0x3C02},  // LOGICAL_ADDRESS_ACCESS [LL_ALGO]
	{0, 0x0990, 0x01F6},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB801},  // LOGICAL_ADDRESS_ACCESS [STAT_MODE]
	{0, 0x0990, 0xE000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB862},  // LOGICAL_ADDRESS_ACCESS [STAT_BMTRACKING_SPEED]
	{0, 0x0990, 0x0400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2818},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_TARGET_INT_TIME_ROWS]
	{0, 0x0990, 0x0640},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x281A},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MAX_INT_TIME_ROWS]
	{0, 0x0990, 0x0FA0},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xA401},  // LOGICAL_ADDRESS_ACCESS [AE_RULE_MODE]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2402},  // LOGICAL_ADDRESS_ACCESS [AE_RULE_ALGO]
	{0, 0x0990, 0x0004},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xA409},  // LOGICAL_ADDRESS_ACCESS [AE_RULE_BASE_TARGET]
	{0, 0x0990, 0x3800},  // MCU_VARIABLE_DATA0 0x3200
	{0, 0x098E, 0xA801},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MODE]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2802},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_ALGO]
	{0, 0x0990, 0x0007},  // MCU_VARIABLE_DATA0

#ifdef LEALI_OK_BUT_NOISE
	{0, 0x098E, 0x281E},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_TARGET_AGAIN]
	{0, 0x0990, 0x0060},  // MCU_VARIABLE_DATA0 0x0080  0x0060
	{0, 0x098E, 0x2820},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MAX_AGAIN]
	{0, 0x0990, 0x0060},  // MCU_VARIABLE_DATA0 0x0100 0x00C0
	{0, 0x098E, 0x281C},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MIN_AGAIN]
	{0, 0x0990, 0x0040},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2822},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MIN_DGAIN]
	{0, 0x0990, 0x0080},  // MCU_VARIABLE_DATA0 0x0080 0x0060
	{0, 0x098E, 0x2824},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MAX_DGAIN]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0 0x0100 0x0060
#endif

//Gain setting
	{0, 0x098E, 0x281E},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_TARGET_AGAIN]
	{0, 0x0990, 0x0080},  // MCU_VARIABLE_DATA0   0x0060
	{0, 0x098E, 0x2820},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MAX_AGAIN]
	{0, 0x0990, 0x00C0},  // MCU_VARIABLE_DATA0 0x0080
	{0, 0x098E, 0x281C},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MIN_AGAIN]
	{0, 0x0990, 0x0040},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2822},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MIN_DGAIN]
	{0, 0x0990, 0x0080},  // MCU_VARIABLE_DATA0 0x0080 0x0060
	{0, 0x098E, 0x2824},  // LOGICAL_ADDRESS_ACCESS [AE_TRACK_MAX_DGAIN]
	{0, 0x0990, 0x0180},  // MCU_VARIABLE_DATA0 0x00C0
//end

	{0, 0x098E, 0xB820},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_WINDOW_POS_X]
	{0, 0x0990, 0x2600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB821},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_WINDOW_POS_Y]
	{0, 0x0990, 0x2600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB822},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_WINDOW_SIZE_X]
	{0, 0x0990, 0xB500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB823},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_WINDOW_SIZE_Y]
	{0, 0x0990, 0xB500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB824},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_NUM_ZONES_X]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB825},  // LOGICAL_ADDRESS_ACCESS [STAT_AE_NUM_ZONES_Y]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB829},  // LOGICAL_ADDRESS_ACCESS [STAT_LL_BRIGHTNESS_METRIC_DIVISOR]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB863},  // LOGICAL_ADDRESS_ACCESS [STAT_BM_MUL]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C52},  // LOGICAL_ADDRESS_ACCESS [LL_START_BRIGHTNESS_METRIC]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C54},  // LOGICAL_ADDRESS_ACCESS [LL_END_BRIGHTNESS_METRIC]
	{0, 0x0990, 0x0A28},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C58},  // LOGICAL_ADDRESS_ACCESS [LL_START_GAIN_METRIC]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C5A},  // LOGICAL_ADDRESS_ACCESS [LL_END_GAIN_METRIC]
	{0, 0x0990, 0x12C0},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C5E},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_GAIN_BM]
	{0, 0x0990, 0x00FA},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C60},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_GAIN_BM]
	{0, 0x0990, 0x0258},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C66},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_GM]
	{0, 0x0990, 0x00FA},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C68},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_GM]
	{0, 0x0990, 0x0258},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C86},  // LOGICAL_ADDRESS_ACCESS [LL_START_FFNR_GM]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C88},  // LOGICAL_ADDRESS_ACCESS [LL_END_FFNR_GM]
	{0, 0x0990, 0x0640},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CBC},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_START_GAIN]
	{0, 0x0990, 0x0040},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CBE},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_END_GAIN]
	{0, 0x0990, 0x01FC},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CCC},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_START_MAX_GM]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CCE},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_END_MAX_GM]
	{0, 0x0990, 0x0640},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CAA},  // LOGICAL_ADDRESS_ACCESS [LL_CDC_THR_ADJ_START_POS]
	{0, 0x0990, 0x044C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CAC},  // LOGICAL_ADDRESS_ACCESS [LL_CDC_THR_ADJ_MID_POS]
	{0, 0x0990, 0x00AF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CAE},  // LOGICAL_ADDRESS_ACCESS [LL_CDC_THR_ADJ_END_POS]
	{0, 0x0990, 0x0009},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CD8},  // LOGICAL_ADDRESS_ACCESS [LL_PCR_START_BM]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CDA},  // LOGICAL_ADDRESS_ACCESS [LL_PCR_END_BM]
	{0, 0x0990, 0x0A28},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3C02},  // LOGICAL_ADDRESS_ACCESS [LL_ALGO]
	{0, 0x0990, 0x01F6},  // MCU_VARIABLE_DATA0
	{0, 0x3380, 0x0504},  // KERNEL_CONFIG
	{0, 0x098E, 0xBCB2},  // LOGICAL_ADDRESS_ACCESS [LL_CDC_DARK_CLUS_SLOPE]
	{0, 0x0990, 0x2800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCB3},  // LOGICAL_ADDRESS_ACCESS [LL_CDC_DARK_CLUS_SATUR]
	{0, 0x0990, 0x5F00},  // MCU_VARIABLE_DATA0
	{0, 0x3380, 0x0584},  // KERNEL_CONFIG
	{0, 0x3380, 0x0586},  // KERNEL_CONFIG
	{0, 0x3380, 0x0505},  // KERNEL_CONFIG
	{0, 0x33B0, 0x2A16},  // FFNR_ALPHA_BETA
	{0, 0x098E, 0xBC8A},  // LOGICAL_ADDRESS_ACCESS [LL_START_FF_MIX_THRESH_Y]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC8B},  // LOGICAL_ADDRESS_ACCESS [LL_END_FF_MIX_THRESH_Y]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC8C},  // LOGICAL_ADDRESS_ACCESS [LL_START_FF_MIX_THRESH_YGAIN]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC8D},  // LOGICAL_ADDRESS_ACCESS [LL_END_FF_MIX_THRESH_YGAIN]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC8E},  // LOGICAL_ADDRESS_ACCESS [LL_START_FF_MIX_THRESH_GAIN]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC8F},  // LOGICAL_ADDRESS_ACCESS [LL_END_FF_MIX_THRESH_GAIN]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC0},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_RAMP_START]
	{0, 0x0990, 0x1000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC1},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_RAMP_STOP]
	{0, 0x0990, 0x0300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC2},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SLOPE_START]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC3},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SLOPE_STOP]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC4},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_THSTART]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC5},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_THSTOP]
	{0, 0x0990, 0x0E00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CBA},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_CONFIG]
	{0, 0x0990, 0x0009},  // MCU_VARIABLE_DATA0

//Section 7
	{0, 0x33BA, 0x0086},  // APEDGE_CONTROL
	{0, 0x33BE, 0x0000},  // UA_KNEE_L
	{0, 0x098E, 0xBCE2},  // LOGICAL_ADDRESS_ACCESS [LL_UA_KP1]
	{0, 0x0990, 0x1000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCE3},  // LOGICAL_ADDRESS_ACCESS [LL_UA_KN1]
	{0, 0x0990, 0x2000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC62},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_KPGAIN]
	{0, 0x0990, 0x0900},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC63},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_KPGAIN]
	{0, 0x0990, 0x0F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC64},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_KNGAIN]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC65},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_KNGAIN]
	{0, 0x0990, 0x0F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC6A},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_INTEGER_GAIN]
	{0, 0x0990, 0x0300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC6B},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_INTEGER_GAIN]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC6C},  // LOGICAL_ADDRESS_ACCESS [LL_START_APERTURE_EXP_GAIN]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC6D},  // LOGICAL_ADDRESS_ACCESS [LL_END_APERTURE_EXP_GAIN]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x3284, 0x0102},  // APERTURE_KNEE_VALUES
	{0, 0x3210, 0x49B8},  // COLOR_PIPELINE_CONTROL
	{0, 0x098E, 0xBC56},  // LOGICAL_ADDRESS_ACCESS [LL_START_CCM_SATURATION]
	{0, 0x0990, 0x6E00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBC57},  // LOGICAL_ADDRESS_ACCESS [LL_END_CCM_SATURATION]
	{0, 0x0990, 0x0A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCDE},  // LOGICAL_ADDRESS_ACCESS [LL_START_SYS_THRESHOLD]
	{0, 0x0990, 0x0300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCDF},  // LOGICAL_ADDRESS_ACCESS [LL_STOP_SYS_THRESHOLD]
	{0, 0x0990, 0x5000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCE0},  // LOGICAL_ADDRESS_ACCESS [LL_START_SYS_GAIN]
	{0, 0x0990, 0x0800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCE1},  // LOGICAL_ADDRESS_ACCESS [LL_STOP_SYS_GAIN]
	{0, 0x0990, 0x0300},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CD0},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SOBEL_FLAT_START]
	{0, 0x0990, 0x000A},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CD2},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SOBEL_FLAT_STOP]
	{0, 0x0990, 0x00FE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CD4},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SOBEL_SHARP_START]
	{0, 0x0990, 0x001E},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3CD6},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SOBEL_SHARP_STOP]
	{0, 0x0990, 0x00FF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC6},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SHARPENING_START]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC7},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_SHARPENING_STOP]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC8},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_FLATNESS_START]
	{0, 0x0990, 0x2000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCC9},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_FLATNESS_STOP]
	{0, 0x0990, 0x4000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCCA},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_TRANSITION_START]
	{0, 0x0990, 0x0400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xBCCB},  // LOGICAL_ADDRESS_ACCESS [LL_SFFB_TRANSITION_STOP]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB42F},  // LOGICAL_ADDRESS_ACCESS [AS_LL_SFFB_ZERO_ENABLE]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB42F},  // LOGICAL_ADDRESS_ACCESS [AS_LL_SFFB_ZERO_ENABLE]
	{0, 0x0990, 0x0300},  // MCU_VARIABLE_DATA0
//AWB
//[CCM TS]
	{0, 0x098E, 0x2C46},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_0]
	{0, 0x0990, 0x02AE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C48},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_1]
	{0, 0x0990, 0xFE44},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4A},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_2]
	{0, 0x0990, 0x000E},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4C},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_3]
	{0, 0x0990, 0xFFC0},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4E},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_4]
	{0, 0x0990, 0x0168},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C50},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_5]
	{0, 0x0990, 0xFFD9},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C52},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_6]
	{0, 0x0990, 0xFFCD},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C54},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_7]
	{0, 0x0990, 0xFE7C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C56},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_8]
	{0, 0x0990, 0x02B7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C58},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_R2BRATIO]
	{0, 0x0990, 0x00C8},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5C},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_0]
	{0, 0x0990, 0x0266},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5E},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_1]
	{0, 0x0990, 0xFF08},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C60},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_2]
	{0, 0x0990, 0xFF92},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C62},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_3]
	{0, 0x0990, 0xFFAE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C64},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_4]
	{0, 0x0990, 0x0153},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C66},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_5]
	{0, 0x0990, 0xFFFF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C68},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_6]
	{0, 0x0990, 0xFFFD},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6A},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_7]
	{0, 0x0990, 0xFF6F},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6C},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_8]
	{0, 0x0990, 0x0194},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6E},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_R2BRATIO]
	{0, 0x0990, 0x005A},  // MCU_VARIABLE_DATA0

//[AWB TS]
	{0, 0x098E, 0x3842},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_GRAY_CHECKER_OFFSET_X]
	{0, 0x0990, 0x004C},  // MCU_VARIABLE_DATA0  0x0052 0x004E
	{0, 0x098E, 0x3844},  // LOGICAL_ADDRESS_ACCESS [STAT_AWB_GRAY_CHECKER_OFFSET_Y]
	{0, 0x0990, 0x0063},  // MCU_VARIABLE_DATA0  0x0068 0x0060
	{0, 0x3240, 0x0023},  // AWB_XY_SCALE
	{0, 0x3240, 0x0013},  // AWB_XY_SCALE
	{0, 0x3242, 0x5503},  // AWB_WEIGHT_R0
	{0, 0x3244, 0xE082},  // AWB_WEIGHT_R1
	{0, 0x3246, 0xE2C1},  // AWB_WEIGHT_R2
	{0, 0x3248, 0x3E10},  // AWB_WEIGHT_R3
	{0, 0x324A, 0x217A},  // AWB_WEIGHT_R4
	{0, 0x324C, 0x5731},  // AWB_WEIGHT_R5
	{0, 0x324E, 0x58D3},  // AWB_WEIGHT_R6
	{0, 0x3250, 0x604E},  // AWB_WEIGHT_R7
	{0, 0x3266, 0x0096},  // AWB_WEIGHT_TH

//[Pre_AWB TS]
	{0, 0x098E, 0xAC3C},  // LOGICAL_ADDRESS_ACCESS [AWB_MIN_ACCEPTED_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x3200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3D},  // LOGICAL_ADDRESS_ACCESS [AWB_MAX_ACCEPTED_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x5A00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3E},  // LOGICAL_ADDRESS_ACCESS [AWB_MIN_ACCEPTED_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x1E00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3F},  // LOGICAL_ADDRESS_ACCESS [AWB_MAX_ACCEPTED_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x5000},  // MCU_VARIABLE_DATA0 0x4B00 0x5500


//setting zoom init
	{0, 0x098E, 0xDC0A},	// LOGICAL_ADDRESS_ACCESS [SYS_SCALE_MODE]
	{0, 0x0990, 0x0200},	// MCU_VARIABLE_DATA0 0x0A00


	{0, 0x098E, 0x6004},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x1040},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x6002},  // LOGICAL_ADDRESS_ACCESS [IO_ALGO]
	{0, 0xE002, 0x0104},  // MCU_VARIABLE_DATA0     

//EXIF header
	{0, 0x098E, 0x981C},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_0]
	{0, 0x0990, 0x5600},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x981D},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_1]
	{0, 0x0990, 0x6900},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x981E},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_2]
	{0, 0x0990, 0x6500},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x981F},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_3]
	{0, 0x0990, 0x7700},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9820},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_4]
	{0, 0x0990, 0x5300},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9821},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_5]
	{0, 0x0990, 0x6F00},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9822},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_6]
	{0, 0x0990, 0x6E00},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9823},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_7]
	{0, 0x0990, 0x6900},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9824},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_8]
	{0, 0x0990, 0x6300},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9825},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_9]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9826},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_10]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9827},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_11]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9828},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_12]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9829},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_13]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982A},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_14]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982B},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MAKE_PSZ_15]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982C},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_0]
	{0, 0x0990, 0x5600},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982D},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_1]
	{0, 0x0990, 0x3900},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982E},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_2]
	{0, 0x0990, 0x3000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x982F},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_3]
	{0, 0x0990, 0x3100},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9830},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_4]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9831},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_5]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9832},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_6]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9833},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_7]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9834},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_8]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9835},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_9]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9836},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_10]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9837},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_11]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9838},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_12]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x9839},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_13]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x983A},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_14]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x983B},	// LOGICAL_ADDRESS_ACCESS [EXIF_HOST_CAMERA_MODEL_PSZ_15]
	{0, 0x0990, 0x0000},	// MCU_VARIABLE_DATA0

//DPI
	{0, 0x098E, 0x1864},	// 
	{0, 0x0990, 0x0060},	//96 DPI
	{0, 0x098E, 0x1868},	// 
	{0, 0x0990, 0x0060},	// 96 DPI
	
//Section 8
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
	//{1, 0, 200},
};

static const MT9P111_REG_TABLE MT9P111_init8[] =
{
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	//{1, 0, 100},
};

static const MT9P111_REG_TABLE MT9P111_init9[] =
{
	{0, 0x098E, 0x8001},  // LOGICAL_ADDRESS_ACCESS [MON_MODE]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x0018, 0x2008},  // STANDBY_CONTROL_AND_STATUS
	//{1, 0, 300},
};

////////////////////////////////////////AF///////////////////////////////////
static const MT9P111_REG_TABLE AF_init[]=
{
	{0, 0x098E, 0xB854}, 	// LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_POS_X]
	{0, 0x0990, 0x5F00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB855}, 	// LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_POS_Y]
	{0, 0x0990, 0x4C00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB856}, 	// LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_SIZE_X]
	{0, 0x0990, 0x4300}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB857}, 	// LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_SIZE_Y]
	{0, 0x0990, 0x6A00}, 	// MCU_VARIABLE_DATA0
	{0, 0x34D0, 0x0400}, 	// DAF_CONFIG_1
	{0, 0x098E, 0xB02C}, 	// LOGICAL_ADDRESS_ACCESS [AF_HC_STEP]
	{0, 0x0990, 0x1000}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3002}, 	// LOGICAL_ADDRESS_ACCESS [AF_MODE]
	{0, 0x0990, 0x0201}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x440C}, 	// LOGICAL_ADDRESS_ACCESS [AFM_POS_MAX]
	{0, 0x0990, 0x00C8}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3036}, 	// LOGICAL_ADDRESS_ACCESS [AF_MD_TH_FOR_STARTING_MOT_EVAL]
	{0, 0x0990, 0x2420}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB015}, 	// LOGICAL_ADDRESS_ACCESS [AF_TH_AVE_DIFF_SHA]
	{0, 0x0990, 0x2300}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC400}, 	// LOGICAL_ADDRESS_ACCESS [AFM_ALGO]
	{0, 0x0990, 0x8800}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0500}, 	// MCU_VARIABLE_DATA0
};
static const MT9P111_REG_TABLE AF_full_scan[]=
{
	{0, 0x098E, 0x3004}, 	// LOGICAL_ADDRESS_ACCESS [AF_ALGO]
	{0, 0x0990, 0x0008}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8419}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AF]
	{0, 0x0990, 0x0500}, 	// MCU_VARIABLE_DATA0
};
static const MT9P111_REG_TABLE AF_pretrigger[] =
{
	{0, 0x098E, 0x3004}, 	// LOGICAL_ADDRESS_ACCESS [AF_ALGO]
	{0, 0x0990, 0x0008}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8419}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AF]
	{0, 0x0990, 0x0500}, 	// MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0xC400}, 	// LOGICAL_ADDRESS_ACCESS [AFM_ALGO]
	{0, 0x0990, 0x8800}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0500}, 	// MCU_VARIABLE_DATA0
};
static const MT9P111_REG_TABLE AF_trigger[] =
{
	{0, 0x098E, 0xB006}, 	// LOGICAL_ADDRESS_ACCESS [AF_PROGRESS]
	{0, 0x0990, 0x0100}, 	// MCU_VARIABLE_DATA0
};
static const MT9P111_REG_TABLE AF_init_and_scan[] = 
{
	{0, 0x098E, 0xB854},  // LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_POS_X]
	{0, 0x0990, 0x3700},  // MCU_VARIABLE_DATA0 0x5F00
	{0, 0x098E, 0xB855},  // LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_POS_Y]
	{0, 0x0990, 0x3B00},  // MCU_VARIABLE_DATA0 0x4C00
	{0, 0x098E, 0xB856},  // LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_SIZE_X]
	{0, 0x0990, 0x8F00},  // MCU_VARIABLE_DATA0 0x4300
	{0, 0x098E, 0xB857},  // LOGICAL_ADDRESS_ACCESS [STAT_SM_WINDOW_SIZE_Y]
	{0, 0x0990, 0x9C00},  // MCU_VARIABLE_DATA0 0x6A00
	{0, 0x34D0, 0x0400},  // DAF_CONFIG_1
	{0, 0x098E, 0xB012},  // LOGICAL_ADDRESS_ACCESS [AF_FS_NUM_STEPS]
	{0, 0x0990, 0x1100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB018},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_0]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB019},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_1]
	{0, 0x0990, 0x1000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01A},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_2]
	{0, 0x0990, 0x2000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01B},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_3]
	{0, 0x0990, 0x3000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01C},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_4]
	{0, 0x0990, 0x4000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01D},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_5]
	{0, 0x0990, 0x5000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01E},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_6]
	{0, 0x0990, 0x6000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB01F},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_7]
	{0, 0x0990, 0x7000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB020},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_8]
	{0, 0x0990, 0x8000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB021},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_9]
	{0, 0x0990, 0x9000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB022},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_10]
	{0, 0x0990, 0xA000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB023},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_11]
	{0, 0x0990, 0xB000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB024},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_12]
	{0, 0x0990, 0xC000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB025},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_13]
	{0, 0x0990, 0xD000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB026},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_14]
	{0, 0x0990, 0xE000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB027},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_15]
	{0, 0x0990, 0xF000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xB028},  // LOGICAL_ADDRESS_ACCESS [AF_FS_POS_16]
	{0, 0x0990, 0xFF00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3002},  // LOGICAL_ADDRESS_ACCESS [AF_MODE]
	{0, 0x0990, 0x0301},  // MCU_VARIABLE_DATA0 0x0305---second, 0x0301--once
//	{0, 0x098E, 0xC408},  // LOGICAL_ADDRESS_ACCESS [AFM_MACRO_POS]
//	{0, 0x0990, 0x0700},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xC400},  // LOGICAL_ADDRESS_ACCESS [AFM_ALGO]
	{0, 0x0990, 0x8800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	//{1, 0, 100},
};
static const MT9P111_REG_TABLE AF_init_and_scan_1[] = 
{
	{0, 0x098E, 0x8419},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AF]
	{0, 0x0990, 0x0500},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
};


//normal status, 16fps
static const MT9P111_REG_TABLE normal_preview_Max_Fps[] =
{
	{0,0x098E, 0x485C},	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_MIN_FRAME_LENGTH_LINES]
	{0,0x0990, 0x0423}, 	// 24fps    MCU_VARIABLE_DATA0, default 1059--->27fps:13M:52M
//	{0,0x0990, 0x0623}, 	// 17fps      MCU_VARIABLE_DATA0, default 1059--->27fps:13M:52M
//	{0,0x0990, 0x0723}, 	// 14fps      MCU_VARIABLE_DATA0, default 1059--->27fps:13M:52M
};

static const MT9P111_REG_TABLE MT9P111_video_YUV422[] = 
{
    {0, 0x098E, 0x843C}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_5_MAX_FRAME_CNT]
    {0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
    {0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
    {0, 0x0990, 0x0100}, 	// MCU_VARIABLE_DATA0
//    {0, 0x0016, 0x0447}, 	// CLOCKS_CONTROL

};

// add by xiaochuqin
static const MT9P111_REG_TABLE MT9P111_video_VGA[] =
{
	{0, 0x098E, 0x486E},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x03D4},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48AA}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_WIDTH]
	{0, 0x0990, 0x0280}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AC}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_HEIGHT]
	{0, 0x0990, 0x01E0}, 	// MCU_VARIABLE_DATA0          
};

static const MT9P111_REG_TABLE MT9P111_video_QVGA[] =
{
	{0, 0x098E, 0x48AA}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_WIDTH]
	{0, 0x0990, 0x0140}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AC}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_HEIGHT]
	{0, 0x0990, 0x00F0}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0             
};

static const MT9P111_REG_TABLE MT9P111_video_QCIF[] = 
{
	{0, 0x098E, 0x48AA}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_WIDTH]
	{0, 0x0990, 0x00B0}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AC}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_HEIGHT]
	{0, 0x0990, 0x0090}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0             
};

//cyj, using QCIF value temporarily
static
const MT9P111_REG_TABLE MT9P111_video_SQCIF[]=
{
	{0, 0x098E, 0x48AA}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_WIDTH]
	{0, 0x0990, 0x00B0}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48AC}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_0_IMAGE_HEIGHT]
	{0, 0x0990, 0x0090}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0             
};

static const MT9P111_REG_TABLE MT9P111_capture_JPEG[] = 
{
    {0, 0x098E, 0x843c}, 	// MCU_ADDRESS [PRI_B_NUM_OF_FRAMES_RUN]
    {0, 0x0990, 0xFF00}, 	// MCU_DATA_0
    {0, 0x098E, 0x8404}, 	// MCU_ADDRESS [SEQ_CMD]
    {0, 0x0990, 0x0200}, 	// MCU_DATA_0
};

static const MT9P111_REG_TABLE MT9P111_capture_QSXGA[] =
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0A20}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x0798}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_2560X1920[] =
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0A00}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x0780}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_QXGA[] =
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0800}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_UXGA[] =
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0640}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x04B0}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

// 1024*768
static const MT9P111_REG_TABLE MT9P111_capture_XGA[] =
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0400}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_VGA[] = 
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0280}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x01E0}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_QVGA[] = 
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0140}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x00F0}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE MT9P111_capture_SVGA[] = 
{
	{0, 0x098E, 0x48A6},    // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_HEIGHT]
	{0, 0x0990, 0x07A0},    // MCU_VARIABLE_DATA0
	
	{0, 0x098E, 0x48C0}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_WIDTH]                                   
	{0, 0x0990, 0x0320}, 	// MCU_VARIABLE_DATA0                                   
	{0, 0x098E, 0x48C2}, 	// LOGICAL_ADDRESS_ACCESS [CAM_OUTPUT_1_IMAGE_HEIGHT]                                   
	{0, 0x0990, 0x0258}, 	// MCU_VARIABLE_DATA0
//	{0, 0x098E, 0xC888}, 	// LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_ORIENTATION]
//	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
//refresh
//	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
//	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};
static const MT9P111_REG_TABLE MT9P111_capture_256X192[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0xc4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xc0, 0x00},
};

static const MT9P111_REG_TABLE PreSetting_for_QVGA_and_multi[] = 
{
	{0, 0x098E, 0xDC0A},     // LOGICAL_ADDRESS_ACCESS [SYS_SCALE_MODE]
	{0, 0x0990, 0x0200},     // MCU_VARIABLE_DATA0  0x0600
	{0, 0x098E, 0x483C},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_X_ADDR_START]
	{0, 0x0990, 0x0018},     // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4840},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_X_ADDR_END]
	{0, 0x0990, 0x0A45},     // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4874},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_X_ADDR_START]
	{0, 0x0990, 0x001C},     // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x4878},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_X_ADDR_END]
	{0, 0x0990, 0x0A43},     // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x486C},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_A_OUTPUT_SIZE_WIDTH]
	{0, 0x0990, 0x0518},     // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x48A4},     // LOGICAL_ADDRESS_ACCESS [CAM_CORE_B_OUTPUT_SIZE_WIDTH]
	{0, 0x0990, 0x0A28},     // MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE Refresh_CMD_600[] = {
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

////////////////////////////////////// AEC ////////////////////////////////////
static
const MT9P111_REG_TABLE AEC_M2_0EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x1000},
	
};

static
const MT9P111_REG_TABLE AEC_M1_7EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x1200},
};

static
const MT9P111_REG_TABLE AEC_M1_3EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x1400},
};

static
const MT9P111_REG_TABLE AEC_M1_0EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x1600},
};

static
const MT9P111_REG_TABLE AEC_M0_7EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x1800},
};

static
const MT9P111_REG_TABLE AEC_M0_3EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x2800},
};

static
const MT9P111_REG_TABLE AEC_default_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x3800},
};

static
const MT9P111_REG_TABLE AEC_0_3EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x4800},
};

static
const MT9P111_REG_TABLE AEC_0_7EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x5800},
};

static
const MT9P111_REG_TABLE AEC_1_0EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x6800},
};

static
const MT9P111_REG_TABLE AEC_1_3EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x7800},
};

static
const MT9P111_REG_TABLE AEC_1_7EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0x8800},
};

static
const MT9P111_REG_TABLE AEC_2_0EV_tab[] =
{
	{ 0, 0x098e, 0xa409},
	{ 0, 0x0990, 0xc000},
};

/////////////////////////////////// Contrast //////////////////////////////////

static const MT9P111_REG_TABLE contrast_M4_tab[] =
{
	{0, 0x337E, 0x0000},
};

static const MT9P111_REG_TABLE contrast_M3_tab[] =
{
	{0, 0x337E, 0x0000},
};

static const MT9P111_REG_TABLE contrast_M2_tab[] =
{
	{0, 0x337E, 0x1000},
};

static const MT9P111_REG_TABLE contrast_M1_tab[] =
{
	{0, 0x337E, 0x1000},
};

static const MT9P111_REG_TABLE contrast_default_tab[] =
{
	{0, 0x337E, 0x2000},
};
static const MT9P111_REG_TABLE contrast_1_tab[] =
{
	{0, 0x337E, 0x3000},
};
static const MT9P111_REG_TABLE contrast_2_tab[] =
{
	{0, 0x337E, 0x3000},
};
static const MT9P111_REG_TABLE contrast_3_tab[] =
{
	{0, 0x337E, 0x4000},
};
static const MT9P111_REG_TABLE contrast_4_tab[] =
{
	{0, 0x337E, 0x4000},
};
////////////////////////////////// Brightness /////////////////////////////////

static
const MT9P111_REG_TABLE brightness_4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x40, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const MT9P111_REG_TABLE brightness_3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x30, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const MT9P111_REG_TABLE brightness_2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x20, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const MT9P111_REG_TABLE brightness_1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x10, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const MT9P111_REG_TABLE brightness_default_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x00, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const MT9P111_REG_TABLE brightness_M1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x10, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const MT9P111_REG_TABLE brightness_M2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x20, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const MT9P111_REG_TABLE brightness_M3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x30, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const MT9P111_REG_TABLE brightness_M4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x40, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

////////////////////////////////////// ZOOM ////////////////////////////////////

//normal
static const MT9P111_REG_TABLE zoom_x1[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x0064}, 	// 0x0064 MCU_DATA_0      100             
};

// 1.2x
static const MT9P111_REG_TABLE zoom_x2[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x0078}, 	// 0x0078 MCU_DATA_0      120     
};

// 1.5x
static const MT9P111_REG_TABLE zoom_x3[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x008C}, 	// 0x0096 MCU_DATA_0      140                              
};

// 2.0x
static const MT9P111_REG_TABLE zoom_x4[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x00A0}, 	// 0x00C8 MCU_DATA_0       160                           
};

// 2.5x
static const MT9P111_REG_TABLE zoom_x5[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x00B4}, 	//0x00FA  MCU_DATA_0        180                      
};

// 3.0x
static const MT9P111_REG_TABLE zoom_x6[] = 
{
    {0, 0x098E, 0x5C18}, 	// MCU_ADDRESS [SYS_ZOOM_FACTOR]
    {0, 0x0990, 0x00C8}, 	//0x012C  MCU_DATA_0       200                        
};

////////////////////////////////////// AWB ////////////////////////////////////
static const MT9P111_REG_TABLE wb_auto_tab[] =
{
#ifdef USING_OLD_AWB_SETTING
	{0, 0x098E, 0x2C46},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x023D},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C48},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFEC2},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4A},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x0001},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4C},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFBF},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C4E},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x0154},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C50},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFED},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C52},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFC3},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C54},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFE85},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C56},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x02B7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5C},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x0212},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C5E},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFF49},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C60},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFA4},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C62},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFB7},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C64},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x014C},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C66},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFFD},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C68},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFFFE},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6A},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0xFF56},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x2C6C},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x01AC},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3C},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x2C00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3D},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x7100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3E},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x1800},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC3F},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x5C00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3842},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x0043},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x3844},  // LOGICAL_ADDRESS_ACCESS
	{0, 0x0990, 0x0021},  // MCU_VARIABLE_DATA0
	{0, 0x3240, 0x0024},  // AWB_XY_SCALE
	{0, 0x3240, 0x0024},  // AWB_XY_SCALE
	{0, 0x3242, 0xD05C},  // AWB_WEIGHT_R0
	{0, 0x3244, 0xDEB5},  // AWB_WEIGHT_R1
	{0, 0x3246, 0x4D54},  // AWB_WEIGHT_R2
	{0, 0x3248, 0x6DC4},  // AWB_WEIGHT_R3
	{0, 0x324A, 0x731D},  // AWB_WEIGHT_R4
	{0, 0x324C, 0x7A90},  // AWB_WEIGHT_R5
	{0, 0x324E, 0x73E4},  // AWB_WEIGHT_R6
	{0, 0x3250, 0xA2E4},  // AWB_WEIGHT_R7
#endif	
	{0, 0x098E, 0x8410},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_0_AWB]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8418},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AWB]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8420},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_2_AWB]
	{0, 0x0990, 0x0200},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC44},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x0000},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC45},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
//Delay=100
	{1, 0, 100},  
};


static const MT9P111_REG_TABLE wb_manual_default_tab[] = 
{
 	{0, 0x098E, 0x8410},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_0_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8418},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8420},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_2_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC44},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC45},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC08},  // LOGICAL_ADDRESS_ACCESS [AWB_CUR_CCM_POS]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
//Delay=100
	{1, 0, 500},  
	{0, 0x098E, 0xAC04},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x2F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC05},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x3E00},  // MCU_VARIABLE_DATA0

};

static const MT9P111_REG_TABLE wb_manual_sunshine_tab[] = 
{
	{0, 0x098E, 0xAC01}, 	// LOGICAL_ADDRESS_ACCESS [AWB_MODE]
	{0, 0x0990, 0xEB00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC9A}, 	// LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_TINT_COEF_FOR_CCM_ROW_0]
	{0, 0x0990, 0xA700}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC9B}, 	// LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_TINT_COEF_FOR_CCM_ROW_1]
	{0, 0x0990, 0x8000}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC9C}, 	// LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_TINT_COEF_FOR_CCM_ROW_2]
	{0, 0x0990, 0x8000}, 	// MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE wb_manual_cloudy_tab[] = 
{
 	{0, 0x098E, 0x8410},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_0_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8418},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8420},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_2_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC44},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC45},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC08},  // LOGICAL_ADDRESS_ACCESS [AWB_CUR_CCM_POS]
	{0, 0x0990, 0x7F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
//Delay=100
	{1, 0, 500},  
	{0, 0x098E, 0xAC04},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x3400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC05},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x3F00},  // MCU_VARIABLE_DATA0};
};

static const MT9P111_REG_TABLE wb_manual_fluorescence_tab[] = 
{
 	{0, 0x098E, 0x8410},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_0_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8418},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8420},  // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_2_AWB]
	{0, 0x0990, 0x0100},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC44},  // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x3F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC45},  // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x3F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC08},  // LOGICAL_ADDRESS_ACCESS [AWB_CUR_CCM_POS]
	{0, 0x0990, 0x3F00},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},  // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},  // MCU_VARIABLE_DATA0
//Delay=100
	{1, 0, 500},  
	{0, 0x098E, 0xAC04},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x3400},  // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC05},  // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x4100},  // MCU_VARIABLE_DATA0
};

static const MT9P111_REG_TABLE wb_manual_incandescence_tab[] = 
{
	{0, 0x098E, 0x8410},      // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_0_AWB]
	{0, 0x0990, 0x0100},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8418},      // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_1_AWB]
	{0, 0x0990, 0x0100},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8420},      // LOGICAL_ADDRESS_ACCESS [SEQ_STATE_CFG_2_AWB]
	{0, 0x0990, 0x0100},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC44},     // LOGICAL_ADDRESS_ACCESS [AWB_LEFT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x0000},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC45},     // LOGICAL_ADDRESS_ACCESS [AWB_RIGHT_CCM_POS_RANGE_LIMIT]
	{0, 0x0990, 0x0000},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC08},     // LOGICAL_ADDRESS_ACCESS [AWB_CUR_CCM_POS]
	{0, 0x0990, 0x0000},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404},      // LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600},      // MCU_VARIABLE_DATA0
//Delay=100
	{1, 0, 500},  
	{0, 0x098E, 0xAC04},     // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_R2G_RATIO]
	{0, 0x0990, 0x4800},      // MCU_VARIABLE_DATA0
	{0, 0x098E, 0xAC05},     // LOGICAL_ADDRESS_ACCESS [AWB_PRE_AWB_B2G_RATIO]
	{0, 0x0990, 0x2500},      // MCU_VARIABLE_DATA0
};

///////////////////////////////// effect ////////////////////////////////////////////////
static MT9P111_REG_TABLE effect_mono_tab[] = 
{
	{0, 0x098E, 0x5C02}, 	// LOGICAL_ADDRESS_ACCESS [SYS_ALGO]
	{0, 0x0990, 0x106E}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC38}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SELECT_FX]
	{0, 0x0990, 0x0100}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0    
};

static MT9P111_REG_TABLE effect_sepia_tab[] = 
{
	{0, 0x098E, 0x5C02}, 	// LOGICAL_ADDRESS_ACCESS [SYS_ALGO]
	{0, 0x0990, 0x106E}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC38}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SELECT_FX]
	{0, 0x0990, 0x0200}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC3A}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SEPIA_CR]
	{0, 0x0990, 0x1A00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC3B}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SEPIA_CB]
	{0, 0x0990, 0xCB00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0                   
};

static MT9P111_REG_TABLE effect_neg_tab[] = 
{
 	{0, 0x098E, 0x5C02}, 	// LOGICAL_ADDRESS_ACCESS [SYS_ALGO]
	{0, 0x0990, 0x106E}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC38}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SELECT_FX]
	{0, 0x0990, 0x0300}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0                         
};

static MT9P111_REG_TABLE effect_sola_tab[] = 
{
	{0, 0x098E, 0xDC39}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SOLARIZATION_TH]
	{0, 0x0990, 0x6B00}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x5C02}, 	// LOGICAL_ADDRESS_ACCESS [SYS_ALGO]
	{0, 0x0990, 0x106E}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC38}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SELECT_FX]
	{0, 0x0990, 0x0400}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};

static MT9P111_REG_TABLE effect_off_tab[] = 
{
	{0, 0x098E, 0x5C02}, 	// LOGICAL_ADDRESS_ACCESS [SYS_ALGO]
	{0, 0x0990, 0x102E}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0xDC38}, 	// LOGICAL_ADDRESS_ACCESS [SYS_SELECT_FX]
	{0, 0x0990, 0x0000}, 	// MCU_VARIABLE_DATA0
	{0, 0x098E, 0x8404}, 	// LOGICAL_ADDRESS_ACCESS [SEQ_CMD]
	{0, 0x0990, 0x0600}, 	// MCU_VARIABLE_DATA0
};
////////////////////////////////////// Gamma ////////////////////////////////////

static
const MT9P111_REG_TABLE gamma_2_tab[] =
{
	{ 0x331b, 0x0e, 0x00},
	{ 0x331c, 0x1a, 0x00},
	{ 0x331d, 0x31, 0x00},
	{ 0x331e, 0x5a, 0x00},
	{ 0x331f, 0x69, 0x00},
	{ 0x3320, 0x75, 0x00},
	{ 0x3321, 0x7e, 0x00},
	{ 0x3322, 0x88, 0x00},
	{ 0x3323, 0x8f, 0x00},
	{ 0x3324, 0x96, 0x00},
	{ 0x3325, 0xa3, 0x00},
	{ 0x3326, 0xaf, 0x00},
	{ 0x3327, 0xc4, 0x00},
	{ 0x3328, 0xd7, 0x00},
	{ 0x3329, 0xe8, 0x00},
	{ 0x332a, 0x20, 0x00},
};

static
const MT9P111_REG_TABLE gamma_1_tab[] =
{
	{ 0x331b, 0x0F, 0x00},
	{ 0x331c, 0x1E, 0x00},
	{ 0x331d, 0x3A, 0x00},
	{ 0x331e, 0x62, 0x00},
	{ 0x331f, 0x74, 0x00},
	{ 0x3320, 0x82, 0x00},
	{ 0x3321, 0x8E, 0x00},
	{ 0x3322, 0x98, 0x00},
	{ 0x3323, 0xA0, 0x00},
	{ 0x3324, 0xA7, 0x00},
	{ 0x3325, 0xB5, 0x00},
	{ 0x3326, 0xC0, 0x00},
	{ 0x3327, 0xD2, 0x00},
	{ 0x3328, 0xE1, 0x00},
	{ 0x3329, 0xED, 0x00},
	{ 0x332a, 0x19, 0x00},
};

static
const MT9P111_REG_TABLE gamma_default_tab[] =
{
	{ 0x331b, 0x03, 0x00},
	{ 0x331c, 0x12, 0x00},
	{ 0x331d, 0x2c, 0x00},
	{ 0x331e, 0x55, 0x00},
	{ 0x331f, 0x67, 0x00},
	{ 0x3320, 0x78, 0x00},
	{ 0x3321, 0x86, 0x00},
	{ 0x3322, 0x93, 0x00},
	{ 0x3323, 0x9e, 0x00},
	{ 0x3324, 0xa8, 0x00},
	{ 0x3325, 0xb5, 0x00},
	{ 0x3326, 0xc1, 0x00},
	{ 0x3327, 0xd2, 0x00},
	{ 0x3328, 0xe1, 0x00},
	{ 0x3329, 0xec, 0x00},
	{ 0x332a, 0x1a, 0x00},
};

const static u16 mt9p111_reg_flash_strobe_on[] =
{
	/* Enable strobe flash: */
	0x307a, 0x83,
	MT9P111_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 mt9p111_reg_flash_strobe_off[] =
{
	/* Disable strobe flash: */
	0x307a, 0x00,
	MT9P111_REGEND, 0x00	    /* End of list delimiter */
};

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_on)
		pdata->power_on(pdata->id);
}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_off)
		pdata->power_off(pdata->id);
}

static void sensor_hw_reset(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->reset)
		pdata->reset(pdata->id);
}

static void cam_flash_fire(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(1);
}

static void cam_flash_torch(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(0);
}

static
void mt9p111hw_power_down( u8 powerMode )
{
	/* MT9P111 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();
	msleep(10);
}

static
void mt9p111hw_reset(void)
{
	printk(KERN_ERR"[%s:%d]sensor reset\n", __FILE__, __LINE__);
	write_sensor_reg(0x3012, 0x80);	
	mdelay(5);
	return;
}

static void mt9p111hw_resume_return(void)
{
#if 0
	write_sensor_reg(0x3086, 0x08);
	write_sensor_reg(0x308D, 0x14);
	wite_sensor_reg(0x300E, 0x32);
#endif
}

static
inline void mt9p111_write_regs( const MT9P111_REG_TABLE *tab, const unsigned len)
{
	unsigned i= 0;

	for( i= 0; i< len; i++)
	{
		if (1 == tab[ i].func)
		{
			msleep(tab[ i].val);
		}
		else
		{
			write_sensor_reg( tab[ i].reg, tab[ i].val);
		}
	}
}

static void RefreshCMD(void)
{
	mt9p111_write_regs( ARRAY_AND_SIZE( Refresh_CMD_600));
	return;
}

static
int mt9p111_set_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != MT9P111_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		status = (write_sensor_reg( regP[curReg << 1], regVal) == 0) ?
			0 : -EIO;
		curReg++;
	}

	return status;
}

static int init_AF( void)
{
	mt9p111_write_regs(ARRAY_AND_SIZE( AF_init ));
    mdelay(100);
    //refresh_check();
    mt9p111_write_regs(ARRAY_AND_SIZE( AF_full_scan ));
    mdelay(50);
    //refresh_check();
	return 0;
}
static int trigger_AF(void)
{
    mt9p111_write_regs(ARRAY_AND_SIZE( AF_pretrigger));	
    mdelay(100);
    mt9p111_write_regs(ARRAY_AND_SIZE( AF_trigger));

	return 0;
}

static int status_check_AF(void)
{
	u16 reg_value = 0;
	int num = 15;
	mdelay(50);

    while (num--)
    {    
        if (write_sensor_reg(0x098E, 0x3000) != 0)
        {
            return -1;
        }

        if (read_sensor_reg(0x0990, &reg_value) != 0)
        {
            return -1;
        }

        if (0x0010 == reg_value)
        {
            break;
        }
        
    }
    mdelay(100);
	return 0;
}
static
void mt9p111hw_update_setting(void)
{
	u16 value1=0, value2=0;
	int carry= 0;
	read_sensor_reg( 0x3002, &value1);
	read_sensor_reg( 0x3003, &value2);

	if( value2 > 127)
		carry= 1;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x3002, value1);
	write_sensor_reg( 0x3003, value2);

	read_sensor_reg( 0x302d, &value1);
	read_sensor_reg( 0x302e, &value2);

	if( value2 > 127)
		carry= 1;
	else
		carry= 0;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x302d, value1);
	write_sensor_reg( 0x302e, value2);
}

static
int mt9p111_sensor_reg_init(int format)
{	
//	if( ( format == reg_format) || ( format == reg_size))
//		return 0;

	switch(format)
	{
	case MT9P111_SIZE_QVGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_QVGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_QVGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_YGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_YGA\n", __FILE__, __LINE__);
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_YGA));
		reg_preview_size = MT9P111_SIZE_YGA;
		reg_size= format;
		break;
	case MT9P111_SIZE_256X192:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_256X192\n", __FILE__, __LINE__);
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_256X192));
		reg_size= format;
		break;
	case MT9P111_SIZE_VGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_VGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_VGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_SQCIF:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_SQCIF\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_SQCIF));
		reg_size= format;
		break;
	case MT9P111_SIZE_QCIF:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_QCIF\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_QCIF));
		reg_size= format;
		break;
	case MT9P111_SIZE_QCIF_H:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_QCIF_H\n", __FILE__, __LINE__);
		reg_preview_size = MT9P111_SIZE_QCIF_H;
		mt9p111_write_regs( ARRAY_AND_SIZE( preview_to_QCIF_H));
		reg_size= format;
		break;
	case MT9P111_SIZE_CIF:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_CIF\n", __FILE__, __LINE__);
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_CIF));
		reg_size= format;
		break;
	case MT9P111_SIZE_QSXGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_QSXGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_QSXGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_2560X1920:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_2560X1920\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_2560X1920));
		reg_size= format;
		break;
	case MT9P111_SIZE_QXGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_QXGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_QXGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_UXGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_UXGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_UXGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_XGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_XGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_XGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_SVGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_XGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_SVGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_JPEG_VGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_JPEG_VGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_VGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_JPEG_QVGA:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_JPEG_QVGA\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_QVGA));
		reg_size= format;
		break;
	case MT9P111_SIZE_JPEG_256X192:
printk(KERN_ERR"[%s:%d] MT9P111_SIZE_JPEG_256X192\n", __FILE__, __LINE__);
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_256X192));
		reg_size= format;
		break;

	case MT9P111_FORMAT_JPG:
printk(KERN_ERR"[%s:%d] MT9P111_FORMAT_JPG\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_JPEG));
		//mt9p111hw_update_setting();
		reg_format= format;
		break;
	case MT9P111_FORMAT_YUV_422:
printk(KERN_ERR"[%s:%d] MT9P111_FORMAT_YUV_422\n", __FILE__, __LINE__);
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_YUV422));
		reg_format= format;
		break;

	case AF_SETTING:
		init_AF();
		break;
	default:
		printk("No such init table\n");
		return -1;
	}
	return 0;
}

static
int mt9p111_sensorID_detect(void)
{
	u16 val_h= 0;
	read_sensor_reg(PID_ADRRH, &val_h);
	
	printk(KERN_ERR"[%s:%d] val_h= 0x%02X\n", __FILE__, __LINE__, val_h);
		
	if( (val_h == PID_VERH ))
		return 0;
	else
		return -1;
}

static
int mt9p111hw_read_sensor_reg( const u16 subAddress, u16 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

static
int mt9p111hw_write_sensor_reg( const u16 subAddress, u16 *bufP )
{
	return write_sensor_reg(subAddress, *bufP);
}

static void mt9p111hw_flash_torch(void)
{
	down(&mt9p111_flash_lock);
	if(!flashled_on){
		//printk(KERN_INFO "mt9p111: flash led on\n");
		cam_flash_torch();
		flashled_fire = 0;
		mt9p111_set_regs( mt9p111_reg_flash_strobe_on );
		flashled_on = 1;

		/* start 5s timeout to off flash */
		queue_delayed_work(flash_wqueue, &flash_work, msecs_to_jiffies(5000));
	}
	else
		printk(KERN_ERR "mt9p111: error, flash led already was on\n");
	up(&mt9p111_flash_lock);
}

static void mt9p111hw_flash_fire(void)
{
	struct timeval now;
        unsigned int ms;

	down(&mt9p111_flash_lock);
	if(flashled_on){
		do_gettimeofday(&now);
		ms = 1000 * (now.tv_sec - flashled_fire_time.tv_sec);
		ms += (now.tv_usec - flashled_fire_time.tv_usec)/1000;
		if(ms<=5000) {
			printk(KERN_ERR "mt9p111: warning, flash led off %d ms, must be off within 5s.\n", ms);
			up(&mt9p111_flash_lock);
			return;
		}

		printk(KERN_INFO "mt9p111: flash led fire\n");
		cam_flash_fire();
		flashled_fire = 1;
	}
	else
		printk(KERN_ERR "mt9p111: fire error, flash led is not on\n");
	up(&mt9p111_flash_lock);
}

static void mt9p111hw_flash_off(void)
{
	down(&mt9p111_flash_lock);
	if(flashled_on){
		cam_flash_torch();
		/* save flash fire off time */
		if(flashled_fire)
			do_gettimeofday(&flashled_fire_time);
		flashled_fire = 0;

		mt9p111_set_regs( mt9p111_reg_flash_strobe_off);
		flashled_on = 0;
		printk(KERN_INFO "mt9p111: flash led off\n");
	}
	up(&mt9p111_flash_lock);
}

static void mt9p111hw_flash_work_off(void)
{
	cancel_delayed_work(&flash_work);
	if(flashled_on){
		mt9p111hw_flash_off();
	}
}

static void mt9p111_flash_work_handler(struct work_struct *work)
{
	printk(KERN_INFO "mt9p111: flashing timeout to be off\n");
	mt9p111hw_flash_off();
}

/*****************************************************************************
 *  Settings
 *****************************************************************************/
static
int mt9p111hw_version_revision(u16 * pCmRevision)
{
	read_sensor_reg( PID_ADRRH, pCmRevision );
	return 0;
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
static int mt9p111hw_af_command(int cmd)
{
#if 0
	u8 reg = 0;
	int loop;

	/* Send auto focus firmware command to AF module */
	write_sensor_reg(0x3f00, cmd);

	loop = 100;
	do {
		read_sensor_reg(0x3f00, &reg);
		if(reg==0)
			return 0;
		msleep(10);
	} while(loop--);

	printk(KERN_ERR "mt9p111: af command timeout, cmd=0x%02x, reg(0x%02x)\n", cmd, reg);
	return -ETIME;
#else
	/* Send auto focus firmware command to AF module */
	write_sensor_reg(0x3f00, cmd);
	return 0;
#endif
}

static int mt9p111hw_af_capture(void)
{
	/* disable af overlay window */
	write_sensor_reg(0x3300, 0x13);
	return 0;
}

static int mt9p111hw_af_finish(void)
{
	u16 state = 0;
	int loop;

	/* Send finish command for AF module */
	mt9p111hw_af_command(0x08);

	/* Read out state register value state_current. 
	  If(state_current == STATE_INF)finish OK and focus on infinity */
	loop = 50;
	do {
		msleep(10);

		read_sensor_reg(0x3f01, &state);
		printk(KERN_INFO "mt9p111: af state(0x%02x)\n", state);

		if(state == AF_STATE_INF)
		{
			mt9p111_single_af_enabled = 0;
			printk(KERN_INFO "mt9p111: af finished, ok\n");
			return 0;
		}
		else if(state == AF_STATE_SUCCESS_S)
		{
			/* Send finish command again */
			printk(KERN_INFO "mt9p111: af not finished, send finish command again! state(0x%02x)\n", state);
			mt9p111hw_af_command(0x08);
		}
	} while(loop--);

	printk(KERN_ERR "mt9p111: af finish timeout, state(0x%02x)\n", state);
	return -ETIME;
}

static int mt9p111hw_af_work_finish(void)
{
	cancel_delayed_work(&autofocus_work);
	if(mt9p111_single_af_enabled)
		return mt9p111hw_af_finish();
	return 0;
}

static int mt9p111hw_single_af(void)
{
	u16 state = 0;
	int loop;
	int ret = 0;

	mt9p111hw_af_work_finish();

	read_sensor_reg(0x3f01, &state);

	if(state != AF_STATE_INF)
	{
		printk(KERN_ERR "mt9p111: af state(0x%02x) is not STATE_INF\n", state);
		mt9p111hw_af_finish();
		return -EINVAL;
	}
#if 0//cyj
	if(flashled_flag == 1){
		mt9p111hw_flash_torch();
	}
#endif
	/* Send single focus command */
	mt9p111hw_af_command(0x03);

	loop = 200;
	do {
		msleep(10);

		read_sensor_reg(0x3f01, &state);
		printk(KERN_INFO "mt9p111: af state(0x%02x)\n", state);

		if(state == AF_STATE_SINGLE)
			continue;
		else if(state == AF_STATE_FAIL_S)
		{
			printk(KERN_INFO "mt9p111: single af fail.\n");
			ret = -EIO;
			goto err;
		}
		else if(state == AF_STATE_SUCCESS_S)
		{
			printk(KERN_INFO "mt9p111: single af ok.\n");
			mt9p111_single_af_enabled = 1;

			/* start 5s timeout to disable focus frame */
			queue_delayed_work(autofocus_wqueue, &autofocus_work, msecs_to_jiffies(5000));

			/* wait about 1 frame to display last frame */
			msleep(100);
			return 0;
		}
	} while(loop--);
	ret = -ETIME;
	printk(KERN_ERR "mt9p111: single af timeout, af state(0x%02x)\n", state);

err:
	if(flashled_on)
		mt9p111hw_flash_work_off();

	mt9p111hw_af_finish();
	return ret;
}

static void mt9p111_af_work_handler(struct work_struct *work)
{
	if( mt9p111_single_af_enabled )
	{
		printk(KERN_INFO "mt9p111: af overlay window timeout to be disable\n");
		/* disable af overlay window */
		write_sensor_reg(0x3300, 0x13);
	}
}

static int mt9p111_sensor_stop_preview(void)
{
	u16 reg = 0;
#if 0//cyj
	if(flashled_flag == 1){
		if(flashled_on) {
			mt9p111hw_flash_fire();
			msleep(700);
		}
		else {
			mt9p111hw_flash_torch();
			mt9p111hw_flash_fire();
			msleep(1000);
		}
	}
#endif
	/* Stop AG/AE */
	read_sensor_reg(0x3013,&reg);
	reg &= 0xfa;
	write_sensor_reg(0x3013,reg);

	printk("mt9p111_sensor_stop_preview\n");
	return 0;
}

static int mt9p111_sensor_prepare_snap(void)
{
	mt9p111hw_update_setting();

	if( mt9p111_single_af_enabled )
	{
		cancel_delayed_work(&autofocus_work);
		mt9p111hw_af_capture();
	}

	printk("mt9p111_sensor_prepare_snap\n");
	return 0;
}

static int mt9p111_sensor_stop_snap(void)
{
	if(flashled_on) {
		mt9p111hw_flash_work_off();
	}

	mt9p111hw_af_work_finish();
	printk("mt9p111_sensor_stop_snap\n");
	return 0;
}

static int mt9p111_sensor_back2preview(void)
{
	if (MT9P111_SIZE_QCIF_H == reg_preview_size)
		mt9p111_write_regs( ARRAY_AND_SIZE( preview_to_QCIF_H));
	else
		mt9p111_write_regs( ARRAY_AND_SIZE( preview_to_HVGA));
		
	printk("mt9p111_sensor_back2preview\n");
	return 0;
}

/*****************************************************************************
 *  Format
 *****************************************************************************/

static int mt9p111hw_set_framerate(int fps)
{
	mt9p111_write_regs( ARRAY_AND_SIZE(normal_preview_Max_Fps) );
	printk(KERN_INFO "mt9p111: set video default fps.\n");
	
	return 0;
}

static
int mt9p111hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{
	if(G_mt9p111_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			mt9p111_sensor_stop_preview();
		else
			mt9p111_sensor_stop_snap();
	}

#if 1
	mt9p111_sensor_reg_init( colorFormat);
	mt9p111_sensor_reg_init( captureSizeFormat);
#else
	mt9p111_sensor_reg_init( captureSizeFormat);
	mt9p111_sensor_reg_init( colorFormat);
#endif

	if( colorFormat == MT9P111_FORMAT_YUV_422)
	{
		//printk("wait focus-----------\n");
		//mt9p111_sensor_reg_init( AF_SETTING);//af function
		//printk("wait focus done-----------\n");
		mt9p111hw_set_framerate(mt9p111set_fps);
	}
/*
	{
	u8 value;
	int addr;
	addr= 0x3000;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	}
*/
	if(G_mt9p111_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			mt9p111_sensor_prepare_snap();
		else
			mt9p111_sensor_back2preview();
		G_mt9p111_mode_change = mode;
	}

	return 0;
}

/*
int mode: sensor output mode, MODE_STILL for QXGA, MODE_VIDEO for XGA;
int iZoomLevel: zoom level, 100-1X, 110-1.1X, ... etc;
unsigned int iWidthOut: output width of the sensor;
unsigned int iHeightOut: output height of the sensor;
*/
static int mt9p111hw_set_resolution_zoom(u32 iZoomLevel)
{
	u32 uZoomnumber;

	uZoomnumber = (iZoomLevel/30) * 30;
	//printk("mt9p111hw_set_resolution_zoom in, uZoomnumber = %d", uZoomnumber);

	write_sensor_reg(0x098E, 0x5c18);
	write_sensor_reg(0x0990, uZoomnumber);

	return 0;
}

static
int mt9p111hw_set_gamma( u32 value)
{
	switch( value)
	{
	case 0:
		mt9p111_write_regs( ARRAY_AND_SIZE( gamma_default_tab));
		break;
	case 1:
		mt9p111_write_regs( ARRAY_AND_SIZE( gamma_1_tab));
		break;
	case 2:
		mt9p111_write_regs( ARRAY_AND_SIZE( gamma_2_tab));
		break;
	default:
		break;
	}

	return 0;
}


//add for colortone
static
int mt9p111hw_set_effect( u32 value)
{
	//printk("[%s] value = %d\n", __FUNCTION__, value);
	switch( value)
	{
       
		case 0:
			mt9p111_write_regs( ARRAY_AND_SIZE( effect_off_tab));//normal
			break;
		case 1:
			mt9p111_write_regs( ARRAY_AND_SIZE( effect_mono_tab));
			break;
		case 2:
			mt9p111_write_regs( ARRAY_AND_SIZE( effect_neg_tab));
			break;	
		case 3:
			mt9p111_write_regs( ARRAY_AND_SIZE( effect_sola_tab));
			break;
		case 4:
			mt9p111_write_regs( ARRAY_AND_SIZE( effect_sepia_tab));
			break;
		default:
			break;
	}
	
	return 0;
}

static
int mt9p111hw_set_brightness( u32 value)
{
	switch( value)
	{
	case 0:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_default_tab));
		break;
	case 1:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_M4_tab));
		break;
	case 2:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_M3_tab));
		break;
	case 3:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_M2_tab));
		break;
	case 4:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_M1_tab));
		break;
	case 5:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_1_tab));
		break;
	case 6:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_2_tab));
		break;
	case 7:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_3_tab));
		break;
	case 8:
		mt9p111_write_regs( ARRAY_AND_SIZE( brightness_4_tab));
		break;
	}
	
	return 0;
}

static
int mt9p111hw_set_contrast( u32 value)
{
	switch( value)
	{
	case 0:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_default_tab));
		break;
	case 1:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_M4_tab));
		break;
	case 2:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_M3_tab));
		break;
	case 3:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_M2_tab));
		break;
	case 4:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_M1_tab));
		break;
	case 5:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_1_tab));
		break;
	case 6:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_2_tab));
		break;
	case 7:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_3_tab));
		break;
	case 8:
		mt9p111_write_regs( ARRAY_AND_SIZE( contrast_4_tab));
		break;
	}

	return 0;
}

///////////////////////////////EV//////////////////////////////////////////
static int mt9p111hw_set_EVoffset( u32 value)
{
 
	switch( value)
	{
	case 0:				// -2.0 - auto
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M2_0EV_tab));
		break;

	case 1:				// -1.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_7EV_tab));
		break;

	case 2:				// -1.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_3EV_tab));
		break;

	case 3:				// -1.0
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_0EV_tab));
		break;

	case 4:				// -0.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M0_7EV_tab));
		break;

	case 5:				// -0.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M0_3EV_tab));
		break;

	case 6:				// 0
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_default_tab));
		//write_sensor_reg(0x098e, 0xa409);
		//read_sensor_reg(0x0990, &af_target);
		//printk(KERN_INFO "byd--------------->af_target = %x\n",af_target);
		break;

	case 7:				// 0.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_0_3EV_tab));
		break;

	case 8:				// 0.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_0_7EV_tab));
		break;

	case 9:				// 1.0
        	mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_0EV_tab));
		break;

	case 10:			// 1.3
        	mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_3EV_tab));
		break;

	case 11:			// 1.7
	        mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_7EV_tab));
		break;

	case 12:			// 2.0
        	mt9p111_write_regs( ARRAY_AND_SIZE( AEC_2_0EV_tab));
		break;

	default:
	        break;
	}

	return 0;
}

#if 0
static
int mt9p111hw_set_exposure( u32 value)
{
	switch( value)
	{
	case 0:				// 0 - auto
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_default_tab));
		break;
	case 1:				// -1.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_7EV_tab));
		break;
	case 2:				// -1.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_3EV_tab));
		break;
	case 3:				// -1
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M1_0EV_tab));
		break;
	case 4:				// -0.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M0_7EV_tab));
		break;
	case 5:				// -0.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_M0_3EV_tab));
		break;
	case 6:				// 0.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_0_3EV_tab));
		break;
	case 7:				// 0.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_0_7EV_tab));
		break;
	case 8:				// 1.0
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_0EV_tab));
		break;
	case 9:				// 1.3
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_3EV_tab));
		break;
	case 10:			// 1.7
		mt9p111_write_regs( ARRAY_AND_SIZE( AEC_1_7EV_tab));
		break;
	}
	return 0;
}
#endif

/*****************************************************************************
 * Auto White Balance
 *****************************************************************************/
static
int mt9p111hw_set_white_balance( u32 value)
{
	switch( value)
	{
	case 0:				// auto
		mt9p111_write_regs( ARRAY_AND_SIZE( wb_auto_tab));
		break;
	case 1:				// Incandescence
		mt9p111_write_regs( ARRAY_AND_SIZE( wb_manual_incandescence_tab));
		break;
	case 2:				// daylight
		mt9p111_write_regs( ARRAY_AND_SIZE( wb_manual_sunshine_tab));
		break;
	case 3:				// Fluorescence
		mt9p111_write_regs( ARRAY_AND_SIZE( wb_manual_fluorescence_tab));
		break;
	case 4:				// cloudy
		mt9p111_write_regs( ARRAY_AND_SIZE( wb_manual_cloudy_tab));
		break;
	default:			// default
		break;
	}
	
	return 0;
}

static
int mt9p111_poll_MCU_register(u16 reg, u16 mask, u16 value, int times)
{
	int ret = 0;
	u16 sensor_value;

	while(times > 0)
	{
		ret = read_sensor_reg(reg, &sensor_value);
		//printk("polling reg = 0x%x value = 0x%x\n", reg, sensor_value);
		if (value == (sensor_value & mask))
			break;
		msleep(10);
		times--;
	}
	
	if (0 == times)
	{
		printk(KERN_ERR"[mt9p111_poll_MCU_register] failed to poll reg: 0x%x, value = 0x%x\n", reg, sensor_value);
	}
	return ret;
}

static
int mt9p111_poll_sensor_register(u16 reg, u16 mask, u16 value, int times)
{
	int ret = 0;
	u16 sensor_value;

	while(times > 0)
	{
		ret = write_sensor_reg(0x098E, reg);
		ret = read_sensor_reg(0x0990, &sensor_value);
		//printk("polling reg = 0x%x value = 0x%x\n", reg, sensor_value);
		if (value == (sensor_value & mask))
			break;
		msleep(10);
		times--;
	}
	
	if (0 == times)
	{
		printk(KERN_ERR"[mt9p111_poll_sensor_register] failed to poll reg: 0x%x, value = 0x%x\n", reg, sensor_value);
	}
	return ret;
}

static int mt9p111hw_sensor_init(void)
{
	int ret = 0;
	G_mt9p111_mode_change = CAMERA_MODE_VIDEO;
	g_mt9p111_size = -1;
	g_mt9p111_format = -1;

	flashled_flag = 0;
	flashled_on = 0;
	flashled_fire = 0;
	flashled_fire_time.tv_sec = 0;
	flashled_fire_time.tv_usec = 0;

	mt9p111_single_af_enabled = 0;
	mt9p111set_fps = 0;
	reg_preview_size = MT9P111_SIZE_YGA;

	printk(KERN_INFO "mt9p111: sensor init\n");
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init1));
	ret = mt9p111_poll_MCU_register(0x0018, 0xFFFF, 0x2008, 200);

	//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init2));
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init3));
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init4));
	ret = mt9p111_poll_sensor_register(0x000A, 0xFFFF, 0, 20);

	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init5));
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init6));
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init7));
	ret = mt9p111_poll_sensor_register(0x8404, 0xFFFF, 0x3, 200);

	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init8));
	ret = mt9p111_poll_sensor_register(0x8404, 0xFFFF, 0x3, 200);
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_init9));
	ret = mt9p111_poll_MCU_register(0x0018, 0xFFFF, 0x2008, 200);

	mt9p111_write_regs( ARRAY_AND_SIZE( AF_init_and_scan));
	ret = mt9p111_poll_sensor_register(0x8404, 0xFFFF, 0x3, 200);

	mt9p111_write_regs( ARRAY_AND_SIZE( AF_init_and_scan_1));
	//mt9p111_sensor_reg_init( AF_SETTING);//af function
	mt9p111_write_regs( ARRAY_AND_SIZE( normal_preview_Max_Fps));
	mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_YUV422));
	mt9p111_write_regs( ARRAY_AND_SIZE( PreSetting_for_QVGA_and_multi));
	mt9p111_write_regs( ARRAY_AND_SIZE( preview_to_HVGA));
	//RefreshCMD();
	return 0;
}

/*****************************************************************************
 * MT9P111 Functions
 *****************************************************************************/
static
int mt9p111_init( p_camera_context_t camera_context )
{
	u16 cm_pid;
	u32 timeout;
	int status;
	
	printk("[MT9P111] mt9p111_init in\n");

	printk(KERN_ERR"mt9p111_init\n");

	af_flag= reg_format= reg_size= 0;

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE | /* SENSOR_CAP_MANUAL_BRIGHTNESS | */
		SENSOR_CAP_MANUAL_EXPOSURE | SENSOR_CAP_MANUAL_ZOOM;

	/* Configure CI according to MT9P111's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);
	
	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 48MHZ
	 */
	/* software work around*/
	ci_set_clock(1, 1, CICLK);

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* MT9P111 Power on sequence
	 * Take out of Power down mode (GPIO_176), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	mt9p111hw_power_down(CAMERA_POWER_FULL);
	mt9p111hw_resume_return();
	//mt9p111hw_reset();

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = 0;
		status = mt9p111hw_version_revision(&cm_pid);

		/* Check to make sure we are working with an MT9P111 */
		if( cm_pid != PID_VERH  )
		{
			printk("mt9p111: pid(%x) is not correct, try again!\n", cm_pid);
			mt9p111hw_power_down( CAMERA_POWER_OFF );
			mt9p111hw_power_down( CAMERA_POWER_FULL );
			sensor_hw_reset();
			mdelay(1);
		}
		else
		{
			printk("mt9p111: pid(%x) is correct!\n", cm_pid);			
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_VERH );

	camera_context->mode[CAMERA_MODE_VIDEO].width = 432;
	camera_context->mode[CAMERA_MODE_VIDEO].height = 320;

	/* default preview framerate */
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_numerator = 15;
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_denominator = 2;
	/* default still capture width/height */
	camera_context->mode[CAMERA_MODE_STILL].width = 2560;
	camera_context->mode[CAMERA_MODE_STILL].height = 1920;
	camera_context->mode[CAMERA_MODE_STILL].input_format = V4L2_PIX_FMT_JPEG;
	camera_context->mode[CAMERA_MODE_STILL].output_format = V4L2_PIX_FMT_JPEG;

	mt9p111hw_sensor_init();

	return 0;
}

static
int mt9p111_deinit(p_camera_context_t camera_context)
{
	/* off flash if it is on */
	mt9p111hw_flash_work_off();

	/* do af finish if the af is not finished */
	mt9p111hw_af_work_finish();

	/* power off the external module */
	mt9p111hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

static
int mt9p111_sleep(  p_camera_context_t camera_context )
{
	return mt9p111_deinit( camera_context );
}

static
int mt9p111_wake(  p_camera_context_t camera_context )
{
	return mt9p111_init( camera_context );
}

static
int mt9p111_enum_framesizes(p_camera_context_t camera_context, int capture_mode, u32 format, struct v4l2_frmsizeenum *frm_sizes)
{
	int index = frm_sizes->index;

	frm_sizes->type = V4L2_FRMSIZE_TYPE_DISCRETE;

	//printk("cyj test 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n", frm_sizes->index, frm_sizes->pixel_format, frm_sizes->type, capture_mode, format);
	if (CAMERA_MODE_VIDEO == capture_mode)
	{
		if (V4L2_PIX_FMT_YUV422P != format) {
			printk(KERN_ERR "mt9p111: only support YUV422 on video mode\n");
			return -EINVAL;
		}
		if (index < ARRAY_SIZE(mt9p111_yuv422_frame_sizes)) {
			frm_sizes->discrete.width = mt9p111_yuv422_frame_sizes[index].width;
			frm_sizes->discrete.height = mt9p111_yuv422_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	else if (CAMERA_MODE_STILL == capture_mode)
	{
		if (V4L2_PIX_FMT_JPEG != format) {
			printk(KERN_ERR "mt9p111: only support JPEG on still capture mode\n");
			return -EINVAL;
		}
		if (index < ARRAY_SIZE(mt9p111_jpeg_frame_sizes)) {
			frm_sizes->discrete.width = mt9p111_jpeg_frame_sizes[index].width;
			frm_sizes->discrete.height = mt9p111_jpeg_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	else
	{
		return -EINVAL;
	}

	return 0;
}

static
int mt9p111_enum_frameintervals(p_camera_context_t camera_context, u32 format, struct v4l2_frmivalenum *frm_intervals)
{
	if (NULL == frm_intervals)
		return -EINVAL;

	return 0;
}

static
int mt9p111_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 mtSizeFormat, mtFormat;
	int capture_width, capture_height;
	int zoom, zoom_width, zoom_height;

	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
//			printk(KERN_ERR"[%s:%d] output format yuv422\n", __FILE__, __LINE__);
			mtFormat = MT9P111_FORMAT_YUV_422;
			break;
			
		case V4L2_PIX_FMT_SRGGB10:
//			printk(KERN_ERR"[%s:%d] output format raw10\n", __FILE__, __LINE__);
			mtFormat = MT9P111_FORMAT_RAW10;		
			break;
			
		case V4L2_PIX_FMT_SRGGB8:
//			printk(KERN_ERR"[%s:%d] output format raw8\n", __FILE__, __LINE__);
			mtFormat = MT9P111_FORMAT_RAW8;		
			break;

			
		case V4L2_PIX_FMT_JPEG:
//			printk(KERN_ERR"[%s:%d] output format jpeg\n", __FILE__, __LINE__);
			mtFormat = MT9P111_FORMAT_JPG;
			break;
		
		case V4L2_PIX_FMT_RGB565X:
//			printk(KERN_ERR"[%s:%d] output format rgb565\n", __FILE__, __LINE__);
			 mtFormat = MT9P111_FORMAT_RGB565;
		
			 break;
		default:
//			printk(KERN_ERR"[%s:%d] output format default yuv422\n", __FILE__, __LINE__);
			mtFormat = MT9P111_FORMAT_YUV_422;
			break;
	}

	capture_width	 = camera_context->capture_input_width;
	capture_height = camera_context->capture_input_height;

	zoom = camera_context->sensor_status.digital_zoom_value;
	if(zoom<100)
		zoom = 100;

	printk(KERN_INFO "mt9p111: capture %d:%d zoom %d\n", capture_width, capture_height, zoom);

	if( mtFormat == MT9P111_FORMAT_JPG )
	{	/* JPEG */
		zoom_width = 2560*100/zoom;
		zoom_height = 1920*100/zoom;

		if(capture_width>zoom_width)
			capture_width = zoom_width;
		if(capture_height>zoom_height)
			capture_height = zoom_height;

		if( (capture_width >= MT9P111_QSXGA_WIDTH) && (capture_height >= MT9P111_QSXGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_QSXGA;
			capture_width = MT9P111_QSXGA_WIDTH;
			capture_height = MT9P111_QSXGA_HEIGHT;
		}
		else if( (capture_width >= 2560) && (capture_height >= 1920) ) {
			mtSizeFormat = MT9P111_SIZE_2560X1920;
			capture_width = 2560;
			capture_height = 1920;
		}
		else if( (capture_width >= MT9P111_QXGA_WIDTH) && (capture_height >= MT9P111_QXGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_QXGA;
			capture_width = MT9P111_QXGA_WIDTH;
			capture_height = MT9P111_QXGA_HEIGHT;
		}
		else if( (capture_width >= MT9P111_UXGA_WIDTH) && (capture_height >= MT9P111_UXGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_UXGA;
			capture_width = MT9P111_UXGA_WIDTH;
			capture_height = MT9P111_UXGA_HEIGHT;
		}
		else if( (capture_width >= MT9P111_XGA_WIDTH) && (capture_height >= MT9P111_XGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_XGA;
			capture_width = MT9P111_XGA_WIDTH;
			capture_height = MT9P111_XGA_HEIGHT;
		}
		else if( (capture_width >= MT9P111_SVGA_WIDTH) && (capture_height >= MT9P111_SVGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_SVGA;
			capture_width = MT9P111_SVGA_WIDTH;
			capture_height = MT9P111_SVGA_HEIGHT;
		}
		else if( (capture_width >= MT9P111_VGA_WIDTH) && (capture_height >= MT9P111_VGA_HEIGHT) ) {
			mtSizeFormat = MT9P111_SIZE_JPEG_VGA;
			capture_width = MT9P111_VGA_WIDTH;
			capture_height = MT9P111_VGA_HEIGHT;
		}
		else if( (capture_width >= 320) && (capture_height>= 240) ) {
			mtSizeFormat = MT9P111_SIZE_JPEG_QVGA;
			capture_width = 320;
			capture_height = 240;
		}
		else if( (capture_width >= 256) && (capture_height>= 192) ) {
			mtSizeFormat = MT9P111_SIZE_JPEG_256X192;
			capture_width = 256;
			capture_height = 192;
		}
		else {
			mtSizeFormat = MT9P111_SIZE_JPEG_VGA;
			capture_width = MT9P111_VGA_WIDTH;
			capture_height = MT9P111_VGA_HEIGHT;
			printk(KERN_ERR "mt9p111: can not capture %d:%d at zoom %d\n", capture_width, capture_height, zoom);
		}

		printk(KERN_DEBUG "mt9p111: jpeg capture %d:%d, zoom=%d, zoomed %d:%d\n",
			capture_width, capture_height, zoom, zoom_width, zoom_height);

		camera_context->capture_input_width = capture_width;
		camera_context->capture_input_height = capture_height;

	} else /* YUV422 */
	if( ( camera_context->capture_input_width <= MT9P111_SQCIF_WIDTH) 
			&& (camera_context->capture_input_height <= MT9P111_SQCIF_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_SQCIF;
	} else if( ( camera_context->capture_input_width <= MT9P111_QCIF_WIDTH)
			&& (camera_context->capture_input_height <= MT9P111_QCIF_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_QCIF;
	} else if( ( camera_context->capture_input_width <= MT9P111_QCIF_HEIGHT)
			&& (camera_context->capture_input_height <= MT9P111_QCIF_WIDTH)) {
		mtSizeFormat = MT9P111_SIZE_QCIF_H;
	} else if( ( camera_context->capture_input_width <= 256)
			&& (camera_context->capture_input_height <= 192)) {
		mtSizeFormat = MT9P111_SIZE_256X192;
	} else if( ( camera_context->capture_input_width <= MT9P111_QVGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9P111_QVGA_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_QVGA;
	} else if( ( camera_context->capture_input_width <= MT9P111_CIF_WIDTH)
			&& (camera_context->capture_input_height <= MT9P111_CIF_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_CIF;
	} else if( ( camera_context->capture_input_width <= MT9P111_YGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9P111_YGA_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_YGA;
	} else if( ( camera_context->capture_input_width <= MT9P111_VGA_WIDTH)
			&& (camera_context->capture_input_height <= MT9P111_VGA_HEIGHT)) {
		mtSizeFormat = MT9P111_SIZE_VGA;
	} else {
		mtSizeFormat = MT9P111_SIZE_VGA;
	}

	if( (mtSizeFormat!=g_mt9p111_size)||(mtFormat!=g_mt9p111_format)||(camera_context->capture_mode!=G_mt9p111_mode_change) ) {
		status = mt9p111hw_set_format( mtSizeFormat, mtFormat, camera_context->capture_mode);
		g_mt9p111_size = mtSizeFormat;
		g_mt9p111_format = mtFormat;
	}

	/* set digital zoom after size changed */
	mt9p111hw_set_resolution_zoom(zoom);

	if( mtFormat == MT9P111_FORMAT_YUV_422)
		/* Update Focus Window, When Host update the sensor’s registers
		   about image resolution, this command must be sent. */
		mt9p111hw_af_command(0x09);
		//mt9p111hw_af_command(0x12);

	timing.BFW = 0x00;
	timing.BLW = 0x00;

	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);
	return 0;
}

static
int mt9p111_start_capture( p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;
	if(camera_context->capture_mode == CAMERA_MODE_STILL)
		/* wait about 2 frame before capture */
		camera_context->capture_drop_frames = 2;
	else
		camera_context->capture_drop_frames = 0;

	return 0;
}

static
int mt9p111_stop_capture(  p_camera_context_t camera_context )
{
	(void) camera_context;
	if(flashled_on && camera_context->capture_mode == CAMERA_MODE_STILL) {
		mt9p111hw_flash_work_off();
	}
	return 0;
}

static
int mt9p111_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	mt9p111hw_power_down( mode );
	return 0;
}

static
int mt9p111_read_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;

	mt9p111hw_read_sensor_reg( reg_addr, reg_val );
	return 0;
}

static
int mt9p111_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	u16 buffer;
	int status;

	(void) camera_context;

	buffer = reg_val;
	status = (mt9p111hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

static int mt9p111_set_digital_zoom( p_camera_context_t camera_context, u32 value)
{
	if ((value<100) || (value>699))
		return -EINVAL;

	/* set digital zoom */
	return mt9p111hw_set_resolution_zoom(value);
}

/**get digital zoom max level*/
static int mt9p111_get_digital_zoom_max( p_camera_context_t cam_ctx, u32 *value)
{
	int w, h, mw, mh;

	w = cam_ctx->mode[CAMERA_MODE_VIDEO].width;
	h = cam_ctx->mode[CAMERA_MODE_VIDEO].height;

	if( (w<=0) || (h<=0) )
		return -EINVAL;

	mw = 1280 * 100 / w;
	mh = 960 * 100 / h;
	if (mw<mh)
		*value = mw;
	else
		*value = mh;
	return 0;
}

/**get digital zoom step*/
static int mt9p111_get_digital_zoom_step( p_camera_context_t camera_context, u32 *value)
{
	*value = 30;
	return 0;
}

static
int mt9p111_set_gamma( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 2)
		value= 2;

	return mt9p111hw_set_gamma( value);
}

static
int mt9p111_set_brightness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 8)
		value= 8;

	return mt9p111hw_set_brightness( value);
}

static
int mt9p111_try_contrast( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_CONTRAST_LOWEST) ||
	    (value == SENSOR_CONTRAST_LOW)   ||
	    (value == SENSOR_CONTRAST_MIDDLE) ||
	    (value == SENSOR_CONTRAST_HIGH) ||
	    (value == SENSOR_CONTRAST_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static
int mt9p111_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int mtindex;
	(void) camera_context;

	if( value == SENSOR_CONTRAST_LOWEST )
		mtindex = 1;	/* M4 */
	else if( value == SENSOR_CONTRAST_LOW )
		mtindex = 3;	/* M2 */
	else if( value == SENSOR_CONTRAST_MIDDLE )
		mtindex = 0;	/* default */
	else if( value == SENSOR_CONTRAST_HIGH )
		mtindex = 6;	/* 2 */
	else if( value == SENSOR_CONTRAST_HIGHEST )
		mtindex = 8;	/* 4 */
	else
		mtindex = 0;

	return mt9p111hw_set_contrast( mtindex);
}

#if 0
static
int mt9p111_set_exposure( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int mtindex;
	(void) camera_context;

	{
		if( value == SENSOR_EXPOSURE_LOWEST )
			mtindex = 1;	/* -1.7EV */
		else if( value == SENSOR_EXPOSURE_LOW )
			mtindex = 3;	/* -1.0EV */
		else if( value == SENSOR_EXPOSURE_MIDDLE )
			mtindex = 0;	/* default */
		else if( value == SENSOR_EXPOSURE_HIGH )
			mtindex = 8;	/* 1.0EV */
		else if( value == SENSOR_EXPOSURE_HIGHEST )
			mtindex = 10;	/* 1.7EV */
		else
			mtindex = 0;
	}

	return mt9p111hw_set_exposure( mtindex);
}
#endif

//////////////////////////////ev///////////////////////////
static int mt9p111_set_evoffset( p_camera_context_t camera_context, int value)
{
	(void) camera_context;
	int mtindex;

	mtindex = 0;
	if (value==0)			/* 0 EV */
		mtindex = 6;
	else if (value>0) {
		if (value<=33)		/* 0.3 EV */
			mtindex = 7;
		else if (value<=70)	/* 0.7 EV */
			mtindex = 8;
		else if (value<=100)	/* 1.0 EV */
			mtindex = 9;
		else if (value<=133)	/* 1.3 EV */
			mtindex = 10;
		else if (value<=170)	/* 1.7 EV */
			mtindex = 11;
		else if (value<=200)	/* 2.0 EV */
			mtindex = 12;
		else
			mtindex = 6;
	}
	else {
		if (value>=-33)		/* -0.3 EV */
			mtindex = 5;
		else if (value>=-70)	/* -0.7 EV */
			mtindex = 4;
		else if (value>=-100)	/* -1.0 EV */
			mtindex = 3;
		else if (value>=-133)	/* -1.3 EV */
			mtindex = 2;
		else if (value>=-170)	/* -1.7 EV */
			mtindex = 1;
		else if (value>=-200)	/* -2.0 EV */
			mtindex = 0;
		else
			mtindex = 6;
	}

    //printk("%s mtindex = %d\n", __FUNCTION__, mtindex);
	return mt9p111hw_set_EVoffset( mtindex);
}

/**white balance**/
static int mt9p111_try_white_balance( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_WHITEBALANCE_AUTO) ||
	    (value == SENSOR_WHITEBALANCE_INCANDESCENT) ||
	    (value == SENSOR_WHITEBALANCE_DAYLIGHT) ||
	    (value == SENSOR_WHITEBALANCE_FLUORESCENT) ||
	    (value == SENSOR_WHITEBALANCE_CLOUDY) )
		return 0;
	else
		return -ENOSYS;
}

static
int mt9p111_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
        int mtindex = 0;
	(void) camera_context;

	{
		if( value == SENSOR_WHITEBALANCE_AUTO )
			mtindex = 0;
		else if( value == SENSOR_WHITEBALANCE_INCANDESCENT )
			mtindex = 1;
		else if( value == SENSOR_WHITEBALANCE_DAYLIGHT )
			mtindex = 2;
		else if( value == SENSOR_WHITEBALANCE_FLUORESCENT )
			mtindex = 3;
		else if( value == SENSOR_WHITEBALANCE_CLOUDY )
			mtindex = 4;
		else
			mtindex = 0;
	}

	return mt9p111hw_set_white_balance( mtindex);
}

static int mt9p111_set_framerate(u32 numerator, u32 denominator)
{
	int fps;
	if (0<denominator && 0<=numerator) {
		fps = numerator / denominator;

		if( fps<=10 )
			mt9p111set_fps = 0;
		else if( fps>25 )
			mt9p111set_fps = 30;
		else if( fps>15 )
			mt9p111set_fps = 25;
		else
			mt9p111set_fps = 15;

		return mt9p111hw_set_framerate(mt9p111set_fps);
	}
	return 0;
}

/**flash**/
static int mt9p111_try_flashmode( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if ( (value==CAMERA_FLASH_OFF) ||
	     (value==CAMERA_FLASH_ON) )
		return 0;
	else
		return -ENOSYS;
}

static int mt9p111_set_flashmode( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	//printk("mt9p111: set_flashmode(%d, %d)\n", mode, value);
	flashled_flag = value;
	return 0;
}

#if 0
/**saturation**/
static int mt9p111_try_saturation( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SATURATION_LOWEST) ||
	    (value == SENSOR_SATURATION_LOW)   ||
	    (value == SENSOR_SATURATION_STANDARD) ||
	    (value == SENSOR_SATURATION_HIGH) ||
	    (value == SENSOR_SATURATION_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int mt9p111_set_saturation( p_camera_context_t camera_context, u32 value)
{
	//cyj to do
	return 0;
}
#endif

/*colort tone*/
static int mt9p111_try_colortone( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_COLORTONE_NORMAL) ||
	    (value == SENSOR_COLORTONE_COOL)   ||
	    (value == SENSOR_COLORTONE_BW)   ||
	    (value == SENSOR_COLORTONE_SEPIA)   ||
	    (value == SENSOR_COLORTONE_WARM) )
		return 0;
	else
		return -ENOSYS;
}

static int mt9p111_set_colortone( p_camera_context_t camera_context, u32 value)
{
	//cyj to do
	return 0;
}

#if 0
/*sharpness*/
static int mt9p111_try_sharpness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SHARPNESS_LOWEST) ||
	    (value == SENSOR_SHARPNESS_LOW)   ||
	    (value == SENSOR_SHARPNESS_STANDARD) ||
	    (value == SENSOR_SHARPNESS_HIGH) ||
	    (value == SENSOR_SHARPNESS_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int mt9p111_set_sharpness( p_camera_context_t camera_context, u32 value)
{
	//cyj to do
	return 0;
}
#endif

/**sensor jpeg quality**/
static int mt9p111_set_jpeg_quality( p_camera_context_t camera_context, u32 value)
{
	//cyj to do
	return 0;
}

/*effect*/
static int mt9p111_try_effect( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_EFFECT_NONE) ||
	    (value == SENSOR_EFFECT_MONO)   ||
	    (value == SENSOR_EFFECT_NEGATIVE)   ||
	    (value == SENSOR_EFFECT_SOLARIZE)   ||
	    (value == SENSOR_EFFECT_SEPIA) )
		return 0;
	else
		return -ENOSYS;
}

static int mt9p111_set_effect( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;
	//printk("[%s] value = %d\n", __FUNCTION__, value);
	if( value > 4)
		value = 4;

	return mt9p111hw_set_effect( value);
}

static int mt9p111_set_autofocus( p_camera_context_t camera_context, u32 value)
{
	int ret;
	(void) camera_context;

#if 0
	if (value==SENSOR_AUTOFOCUS_IDLE) {
		ret = mt9p111hw_af_work_finish();
	}
	else if (value==SENSOR_AUTOFOCUS_SINGLE) {
		ret = mt9p111hw_single_af();
	}
	else
		return -EINVAL;
#endif
#if 1	
    //move to camera init setting, we just do it one time
	if (init_AF() == -1)
	{
		return -1;
	}
			
	if (trigger_AF() == -1)
	{	
		return -1;
	}

    //wait focus finish.
    ret = status_check_AF();
    mdelay(300);
#endif
	return ret;
}

static char name[] = "Micron_mt9p111";

static p_camera_function_t camera_functions;

static int format_list[] = {

	V4L2_PIX_FMT_YUV422P,
	/* V4L2_PIX_FMT_SRGGB10, */
	
	/* V4L2_PIX_FMT_SRGGB8, */
	/* V4L2_PIX_FMT_RGB565X, */
	V4L2_PIX_FMT_JPEG, 
	
	-1	
};

static
int mt9p111_register(int id)
{
	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max = 2560;
	camera_functions->height_max = 1920;
	camera_functions->width_min = 2;
	camera_functions->height_min = 2;

	camera_functions->v_power = 0; //2800;
	camera_functions->v_io = 0; //1800;
		

	camera_functions->format_list = format_list;

	camera_functions->init = mt9p111_init;     	
	camera_functions->deinit = mt9p111_deinit;     
	camera_functions->set_capture_format = mt9p111_set_capture_format; 	
	camera_functions->start_capture = mt9p111_start_capture;	 		
	camera_functions->stop_capture = mt9p111_stop_capture; 			
	camera_functions->sleep = mt9p111_sleep;	
	camera_functions->wakeup = mt9p111_wake;	
	camera_functions->read_8bit = NULL;			
	camera_functions->write_8bit = NULL;			
	camera_functions->read_16bit = mt9p111_read_16bit;		
	camera_functions->write_16bit = mt9p111_write_16bit;		
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode = mt9p111_set_power_mode;			
	camera_functions->set_contrast = mt9p111_set_contrast;
	camera_functions->try_contrast = mt9p111_try_contrast;
	camera_functions->set_whitebalance = mt9p111_set_white_balance;	
	camera_functions->try_whitebalance = mt9p111_try_white_balance;	
	camera_functions->set_gamma= mt9p111_set_gamma;
	camera_functions->set_zoom = NULL;

	camera_functions->set_framerate =	mt9p111_set_framerate;
	camera_functions->set_autofocus =	mt9p111_set_autofocus;
	//camera_functions->set_flashmode =	mt9p111_set_flashmode;
	//camera_functions->try_flashmode =	mt9p111_try_flashmode;
	//camera_functions->set_saturation =	mt9p111_set_saturation;
	//camera_functions->try_saturation =	mt9p111_try_saturation;
	//camera_functions->set_scene =	mt9p111_set_scene;
	//camera_functions->try_scene =	mt9p111_try_scene;
	camera_functions->set_digital_zoom = 	mt9p111_set_digital_zoom;
	camera_functions->get_digital_zoom_max 	= mt9p111_get_digital_zoom_max;
	camera_functions->get_digital_zoom_step	= mt9p111_get_digital_zoom_step;
	//camera_functions->set_colortone =	mt9p111_set_colortone;
	//camera_functions->try_colortone =	mt9p111_try_colortone;
	//camera_functions->set_sharpness =	mt9p111_set_sharpness;
	//camera_functions->try_sharpness =	mt9p111_try_sharpness;
	//camera_functions->set_jpeg_quality_mode =	mt9p111_set_jpeg_quality_mode;
	//camera_functions->try_jpeg_quality_mode =	mt9p111_try_jpeg_quality_mode;
	//camera_functions->set_jpeg_quality_percent =	mt9p111_set_jpeg_quality_percent;
	//camera_functions->try_jpeg_quality_percent =	mt9p111_try_jpeg_quality_percent;

	camera_functions->set_ev_offset =	mt9p111_set_evoffset;
	//camera_functions->try_ev_offset =	mt9p111_try_evoffset;
	//camera_functions->set_isospeed =	mt9p111_set_isospeed;
	//camera_functions->try_isospeed =	mt9p111_try_isospeed;
	//camera_functions->set_focus_mode =	mt9p111_set_focus_mode;
	//camera_functions->try_focus_mode =	mt9p111_try_focus_mode;
	//camera_functions->set_focus_macro =	mt9p111_set_focus_macro;
	//camera_functions->try_focus_macro =	mt9p111_try_focus_macro;
	//camera_functions->set_flash_torch =	mt9p111_set_flash_torch;
	//camera_functions->try_flash_torch =	mt9p111_try_flash_torch;
	camera_functions->set_effect =	mt9p111_set_effect;
	camera_functions->try_effect =	mt9p111_try_effect;
	//camera_functions->set_power_line_frequency =	mt9p111_set_power_line_frequency;
	//camera_functions->try_power_line_frequency =	mt9p111_try_power_line_frequency;
	camera_functions->enum_frame_sizes =	mt9p111_enum_framesizes;
	camera_functions->enum_frame_intervals = mt9p111_enum_frameintervals;

	camera_functions->name = name;
	camera_functions->id = id;

	/*register mt9p111 as high resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static
int mt9p111_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *                              MT9P111 I2C Client Driver
 *****************************************************************************/
static
int read_sensor_reg_16bit_dat_16bit(u16 addr, u16 *pvalue)
{    
	int ret;
	u8 dataH;
	u8 dataL;
	u8 address[2];
	if( g_client == NULL )	/* Not initialize the client */	
	{
		ovI2C_dbg("not initialize the client\n");
		return -1;
	}
	address[0] = addr>>8;	
	address[1] = addr;
	ret = i2c_smbus_write_byte_data(g_client,address[0],address[1]);
	if(ret) {
		printk(KERN_ERR "mt9p111: read reg [%04x], write address failed %d\n", addr, ret);
		return ret;
	}
	ret = i2c_smbus_read_byte(g_client);
	if(ret<0) {
		printk(KERN_ERR "mt9p111: read reg [%04x], read byte failed %d\n", addr, ret);
		return ret;
	}
	dataH = (u8)ret;
	ret = i2c_smbus_read_byte(g_client);
	if(ret<0) {
		printk(KERN_ERR "mt9p111: read reg [%04x], read byte failed %d\n", addr, ret);
		return ret;
	}
	dataL = (u8)ret;
	*pvalue = (dataH << 8) | dataL;

	return 0; 
}

static
int write_sensor_reg_16bit_dat_16bit(u16 addr, u16 *value)
{
	int ret;
	u8 data[4];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;			
	data[0] = addr>>8;
	data[1] = addr;	
	data[2]=  (*value) >> 8;
	data[3]=  (*value) & 0xFF;
	

	ret = i2c_master_send(g_client, data, 4);
	if(ret<0) {
		printk(KERN_ERR "mt9p111: write reg [%04x:%02x] failed %d\n", addr, *value, ret);
		return ret;
	}
	
	return 0;
}

static
int write_sensor_reg_16bit_dat_16bit_m(u16 addr, u16 value, u16 mask)
{
	int ret = 0;
	u16 regval;
	
	read_sensor_reg_16bit_dat_16bit(addr, &regval);
	
	regval  &=  (~mask);
	value   &=  mask;
	
	regval|=value;
	
	ret = write_sensor_reg_16bit_dat_16bit(addr,&regval);

	return ret;
}

static int write_sensor_reg(u16 addr, u16 value)
{
	int ret;
	down(&mt9p111_i2c_lock);
	ret = write_sensor_reg_16bit_dat_16bit(addr, &value);
	up(&mt9p111_i2c_lock);
	return ret;
}

static int read_sensor_reg(u16 addr, u16 *pvalue)
{
	int ret;
	down(&mt9p111_i2c_lock);
	ret = read_sensor_reg_16bit_dat_16bit(addr,pvalue);
	up(&mt9p111_i2c_lock);
	return ret;
}

static int mt9p111_create_workqueue(void)
{
	INIT_DELAYED_WORK(&flash_work, mt9p111_flash_work_handler);
	flash_wqueue = create_workqueue("mt9p111_flash");
	if(!flash_wqueue){
		printk(KERN_ERR "mt9p111: create_workqueue(flash) failed\n");
		return -ENOMEM;
	}
	INIT_DELAYED_WORK(&autofocus_work, mt9p111_af_work_handler);
	autofocus_wqueue = create_workqueue("mt9p111_af");
	if(!autofocus_wqueue){
		printk(KERN_ERR "mt9p111: create_workqueue(autofocus) failed\n");
		destroy_workqueue(flash_wqueue);
		flash_wqueue = NULL;
		return -ENOMEM;
	}
	return 0;
}

static void mt9p111_destroy_workqueue(void)
{
	if(flash_wqueue){
		cancel_rearming_delayed_workqueue(flash_wqueue, &flash_work);
		destroy_workqueue(flash_wqueue);
		flash_wqueue = NULL;
	}
	if(autofocus_wqueue){
		cancel_rearming_delayed_workqueue(autofocus_wqueue, &autofocus_work);
		destroy_workqueue(autofocus_wqueue);
		autofocus_wqueue = NULL;
	}
}

static int __devinit sensor_probe(struct i2c_client *client, const struct i2c_device_id *dev_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	int id = 0;
	
	pdata = client->dev.platform_data;
	id = pdata->id;
	printk(KERN_ERR"[%s:%d] id= %d\n", __FILE__, __LINE__, id);

	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		printk("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);
	
	ci_set_clock(1, 1, CICLK);
	
	g_client = client;

	if( pdata->power_on && (pdata->power_on( id) != 0) )
		printk(KERN_ERR"[%s:%d] Camera power on failed!!!!!!!\n", __FILE__, __LINE__);

	if (mt9p111_sensorID_detect()<0)
		goto err;

	if(mt9p111_register(id)<0)
		goto err;

	printk("MT9P111 detected.\n");

	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	clk_disable(clk);
	clk_put(clk);

	if(mt9p111_create_workqueue())
		return -1;

	return 0;
err:
	printk("MT9P111 detect failed.\n");
	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	printk("[MT9P111] sensor_remove in\n");
	
	mt9p111_unregister(pdata->id);
	g_client = NULL;

	mt9p111_destroy_workqueue();
	return 0;
}

static const struct i2c_device_id mt9p111_id[] = {
	{"sensor_mt9p111", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "sensor_mt9p111",
	},
	.id_table	= mt9p111_id,
	.probe		= sensor_probe,
	.remove		= sensor_remove,
};

static ssize_t flashled_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	printk(KERN_INFO "%d\n", flashled_flag);

	return 0;
}

static ssize_t flashled_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char msg[ 3];

	if (len > 3)
		len = 3;

	if (copy_from_user(msg, buff, len))
		return -EFAULT;

	switch( msg[ 0])
	{
	case '0':
		flashled_flag= 0;
		break;
	case '1':
		flashled_flag= 1;
		break;
	}
	
	return len;
}

static struct file_operations flashled_proc_fops = {
	.owner		= THIS_MODULE,
	.read		= flashled_proc_read,
	.write		= flashled_proc_write,
};

static int create_flashled_proc_file(void)
{
	flashled_proc_file = create_proc_entry( FLASHLED_PROC_FILE, 0644, NULL);
	if (!flashled_proc_file) {
		printk(KERN_INFO "Create proc file for PXA Backlight failed\n");
		return -ENOMEM;
	}

	printk( KERN_INFO "PXA Backlight proc OK\n" );
	flashled_proc_file->proc_fops = &flashled_proc_fops;
	return 0;
}

static ssize_t mt9p111_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	printk( KERN_INFO "ccc\n");

	return 0;
}

static ssize_t mt9p111_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char msg[ 4096], tmp[ 10];
	unsigned int addr= 0, i= 0;
	u16 value= 0;

	if( len > 4096)
		len = 4096;

	if( copy_from_user( msg, buff, len))
		return -EFAULT;

	if( strncmp( msg, "QVGA", 4) == 0)
	{
		printk("set QVGA\n");
//		mt9p111_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		mt9p111_write_regs( ARRAY_AND_SIZE( XGA_QVGA_tab));
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_QVGA));
		i= len;					// force to leave the function

	}
	else if( strncmp( msg, "YGA2", 4) == 0)
	{
		printk("set YGA2\n");
//		mt9p111_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		mt9p111_write_regs( ARRAY_AND_SIZE( XGA_YGA2_tab));
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_YGA2));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "YGA", 3) == 0)
	{
		printk("set YGA\n");
//		mt9p111_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		mt9p111_write_regs( ARRAY_AND_SIZE( XGA_YGA_tab));
		//mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_YGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "VGA", 3) == 0)
	{
		printk("set VGA\n");
//		mt9p111_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		mt9p111_write_regs( ARRAY_AND_SIZE( XGA_VGA_tab));
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_video_VGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "QXGA", 4) == 0)
	{
		printk("set QXGA\n");
//		mt9p111_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		mt9p111_write_regs( ARRAY_AND_SIZE( QXGA_tab));
		mt9p111_write_regs( ARRAY_AND_SIZE( MT9P111_capture_QXGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "AF", 2) == 0)
	{
		printk("Auto Focus\n");
		mt9p111_sensor_reg_init( AF_SETTING);//af function
		i= len;					// force to leave the function
	}

	while( i< len)
	{
		// get address
		while( i< len)
		{
			if( ( msg[ i] == '0') && ( msg[ i+ 1] == 'x'))
			{
				strncpy( tmp, msg+ i, 6);
				tmp[ 6]= '\0';
				i+= 6;
				addr= atoi( tmp);
				break;
			}
			i++;
		}
	
		// get value
		while( i<= len)
		{
			// get value, not set value
			if( ( msg[ i] == '\n') || ( i == len))
			{
				read_sensor_reg( addr, &value);
				printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
				break;
			}
			else if( ( msg[ i] == '0') && ( msg[ i+ 1] == 'x'))
			{
				strncpy( tmp, msg+ i, 6);
				tmp[ 4]= '\0';
				i+= 4;
				value= atoi( tmp);
printk("Write to addr= 0x%04X, val= 0x%02X\n", addr, value);
				write_sensor_reg( addr, value);
				break;
			}
	
			i++;
		}
	}
	
	return len;
}

static struct file_operations mt9p111_proc_fops = {
	.owner		= THIS_MODULE,
	.read		= mt9p111_proc_read,
	.write		= mt9p111_proc_write,
};

static int create_mt9p111_proc_file(void)
{
	mt9p111_proc_file = create_proc_entry( MT9P111_PROC_FILE, 0644, NULL);
	if (!mt9p111_proc_file) {
		printk(KERN_INFO "Create proc file for PXA MT9P111 camera failed\n");
		return -ENOMEM;
	}

	printk( KERN_INFO "PXA Camera MT9P111 proc OK\n" );
	mt9p111_proc_file->proc_fops = &mt9p111_proc_fops;
	return 0;
}

static int __init sensor_init(void)
{
	int ret = 0;
	create_flashled_proc_file();
	create_mt9p111_proc_file();

	//mt9p111hw_reset();

	ret = i2c_add_driver(&sensor_driver);
	printk( KERN_INFO "cyj test ret = 0x%x\n", ret);
	return ret;
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}

module_init(sensor_init);
module_exit(sensor_exit);

MODULE_DESCRIPTION("MT9P111 I2C Client driver");
MODULE_LICENSE("GPL");
