/**************************************************************************
 *                                                                         
 *         Copyright (c) 2008 by Ommnivision Technology Co., Ltd.            
 *                                                                         
 *  This software is copyrighted by and is the property of Ommnivision       
 *  Technology Co., Ltd. All rights are reserved by Ommnivision Technology   
 *  Co., Ltd. This software may only be used in accordance with the        
 *  corresponding license agreement. Any unauthorized use, duplication,    
 *  distribution, or disclosure of this software is expressly forbidden.   
 *                                                                         
 *  This Copyright notice MUST not be removed or modified without prior    
 *  written consent of Ommnivision Technology Co., Ltd.                      
 *                                                                         
 *  Ommnivision Technology Co., Ltd. reserves the right to modify this       
 *  software without notice.                                                                                                                                             
 *  Author: Jacky Wang  
 *                                                                                                                                
 **************************************************************************/
#include "camera.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <mach/mfp.h>
#include <mach/gpio.h>
#include <mach/hardware.h>
#include <linux/err.h>
#include <linux/time.h>
#include <linux/clk.h>
#include <linux/i2c.h>




#define OV5642_DEBUG
#undef  OV5642_DEBUG

#ifdef  OV5642_DEBUG
#define	ov5642_dbg(fmt, arg...)    printk(KERN_DEBUG "%s(line %d): " fmt, 	\
		__FUNCTION__, __LINE__, ##arg)
#else
#define	ov5642_dbg(fmt, arg...)	do {} while (0)
#endif
/*****************************************************************************
 * Constants & Registers
 *****************************************************************************/
/* CICLK for QCI */
#define CICLK	2600

/* Revision constants */
#define PID_VERH		0x56
#define PID_VERL	       0x40
#define PID_ADRRH            0x300a
#define PID_ADRRL            0x300b

/* Output Size & Format */
#define OV5642_SIZE_CIF				1          //352x288
#define OV5642_SIZE_QCIF			2          //176X144
#define OV5642_SIZE_SQCIF			3          //128X96
#define OV5642_SIZE_QVGA			4          //320x240
#define OV5642_SIZE_432X320		5         //432x320
#define OV5642_SIZE_480X360		6        // 480x360
#define OV5642_SIZE_VGA			7	   //640x480
#define OV5642_SIZE_XGA			8	   //1024X768
#define OV5642_SIZE_1280X960		9	  // 1280x960
#define OV5642_SIZE_UXGA			10         //1600X1200
#define OV5642_SIZE_QXGA			11        //2048x1536
#define OV5642_SIZE_QSXGA			12      //2592X1944
#define OV5642_SIZE_854X480		13     //854x480
#define OV5642_SIZE_1280X720           14     //1280x720
#define OV5642_SIZE_1600X900		15	//1600x900
#define OV5642_SIZE_2048X1152		16	//2048x1152
#define OV5642_SIZE_2592X1458		17	//2592x1458
#define OV5642_SIZE_2560X1920		18
#define OV5642_SIZE_160X120		19          //160X120


#define OV5642_FORMAT_YUV_422		1
#define OV5642_FORMAT_RAW10		2
#define OV5642_FORMAT_RAW8		3
#define OV5642_FORMAT_JPG			4
#define OV5642_FORMAT_RGB565		5

#define OV5642_REGEND		(0xFFFF)

/*Structure*/
struct OV5642_SETTING
{
	int   format;
	int   resolution;
	const u16 *setting;
};

/* Unique ID allocation */
static struct i2c_client *g_client;

#define I2C_DEBUG
//#undef I2C_DEBUG

#ifdef  I2C_DEBUG
#define ovI2C_dbg(fmt, arg...)    printk(KERN_INFO "%s(line %d): " fmt "\n",  \
                __FUNCTION__, __LINE__, ##arg)
#else 
#define ovI2C_dbg(fmt, arg...)        do {} while (0)
#endif

/****************************************************************************

   Gloable variable

*****************************************************************************/
struct frame_sizes_t {
	u32	width;
	u32	height;
};

/* preview YUV422 size list, supported preview frame sizes "preview-size-values" on Android */
static const struct frame_sizes_t ov5642_yuv422_frame_sizes[] = {
	{480, 360},
	{432, 320},
	{320, 240},
	{160, 120},
	{128, 96},
	{352, 288},
	{176, 144}
};

/* capture JPEG size list, supported capture frame sizes "picture-size-values" on Android */
static const struct frame_sizes_t ov5642_jpeg_frame_sizes[] = {
	{2560, 1920},
	{2048, 1536},
	{1600, 1200},
	{1280, 960},
	{1024, 768},
	{640, 480},
	{320, 240}
};

/* frame interval list,  frame rate = 1 / frame interval,
   supported frame rates "preview-frame-rate-values" on Android */
static const struct v4l2_fract ov5642_yuv422_frame_intervals[] = {
	{2, 15},	/* 2/15 seconds, 7.5fps, the frame rate could be down to 7.5 due to night mode enabled */
	{1, 15},	/* 1/15 seconds, 15fps */
	{1, 25},	/* 1/25 seconds, 25fps */
	{1, 30}		/* 1/30 seconds, 30fps */
};

static u16	g_ov5642_mode_change = CAMERA_MODE_VIDEO;
static int	g_ov5642_size = -1;
static int	g_ov5642_format = -1;
static u32	g_preview_line_width = 0;
static u32	g_preview_exposure = 0;
static u8    g_preview_gain = 0;
//static u8   ov5642_recRgain_high =0,ov5642_recBgain_high=0,ov5642_recGgain_high=0;
//static u8   ov5642_recRgain_low =0,  ov5642_recBgain_low=0, ov5642_recGgain_low=0;
static int 	g_initial_flag = 0;
static int 	g_af_fw_flag = 0;
static int read_sensor_reg( const u16 subAddress, u8 *bufP );
static int write_sensor_reg( const u16 subAddress, u8 bufP );
static int write_sensor_block_regs(u8 *databuf ,int lenth);

static int flashled_flag= 0;
static int flashled_on = 0;
static int flashled_fire = 0;
static struct timeval flashled_fire_time;

static struct delayed_work flash_work;
static struct workqueue_struct *flash_wqueue;
static DECLARE_MUTEX(ov5642_flash_lock);

static DECLARE_MUTEX(ov5642_i2c_lock);

/*****************************************************************************

 *  Register Settings Function

 *****************************************************************************/
 const static u16 g_preview_iq_init_tab[]=
{

	//// Saturation x1.1
	0x5001, 0xFF,    
	0x5583, 0x40,
	0x5584, 0x40,
	0x5580, 0x06,
	
	0x5581, 0x7b,
	0x5582, 0x21,
	0x558a, 0x01,

	//// BLC
	0x4007, 0x10,

	//// Pregain 
	0x3a13, 0x38,

	//// banding Manual 50hz
	0x3c01,0x80,
	0x3c00,0x04,

	//// Auto sharpness
	0x530A,0x00,
	0x530c,0x00,
	0x530d,0x0c,
	0x5312,0x40,

	//// Lenc 
	0x5000,0xcf,

	0x528a, 0x02,
	0x528b, 0x06,
	0x528c, 0x20,
	0x528d, 0x30,
	0x528e, 0x40,
	0x528f, 0x50,
	0x5290, 0x60,
	0x5292, 0x00,
	0x5293, 0x02,
	0x5294, 0x00,
	0x5295, 0x04,
	0x5296, 0x00,
	0x5297, 0x08,
	0x5298, 0x00,
	0x5299, 0x10,
	0x529a, 0x00,
	0x529b, 0x20,
	0x529c, 0x00,
	0x529d, 0x28,
	0x529e, 0x00,
	0x529f, 0x30,
	0x5282, 0x00,

	OV5642_REGEND, 0x00	    /* End of list delimiter */
	
};

static u8 g_preview_gamma[58] = 
{
	0x54,
	0x80,
	0x0b,
	0x18,
	0x2e,
	0x50,
	0x61,
	0x6e,
	0x7b,
	0x86,
	0x90,
	0x9a,
	0xaa,
	0xb8,
	0xcd,
	0xdd,
	0xea,
	0x1d,
	0x05,
	0x00,
	0x04,
	0x20,
	0x03,
	0x60,
	0x02,
	0xb8,
	0x02,
	0x86,
	0x02,
	0x5b,
	0x02,
	0x3b,
	0x02,
	0x1c,
	0x02,
	0x04,
	0x01,
	0xed,
	0x01,
	0xc5,
	0x01,
	0xa5,
	0x01,
	0x6c,
	0x01,
	0x41,
	0x01,
	0x20,
	0x00,
	0x16,
	0x01,
	0x20,
	0x01,
	0x40,
	0x00,
	0xf0,
	0x01,
	0xdf	
};

static u8 g_preview_cmx[23] = 
{ 
	0x53,
	0x80,
	0x01,
	0x00,
	0x00,
	0x55,
	0x00,
	0x0f,
	0x00,
	0x00,
	0x01,
	0x6d,
	0x00,
	0x1b,
	0x00,
	0x00,
	0x00,
	0x1c,
	0x00,
	0xd5,
	0x00,
	0xa6,
	0x08
};

static u8 g_preview_awb[33] = 
{
	0x51,
	0x80,
	0xff,
	0x50,
	0x00,
	0x00,
	0x25,
	0xff,
	0x16,
	0x15,
	0x18,
	0x85,
	0x5a,
	0xef,
	0xaa,
	0x2f,
	0x2c,
	0x6e,
	0x44,
	0xf8,
	0x04,
	0x70,
	0xf0,
	0xf0,
	0x03,
	0x01,
	0x06,
	0x28,
	0x04,
	0x00,
	0x04,
	0xcd,
	0x00
};

static unsigned char g_preview_lenc[138] = 
{
	0x58,
	0x00,
	0x42,
	0x26,
	0x1a,
	0x14,
	0x14,
	0x1a,
	0x26,
	0x3e,
	0x1d,
	0x12,
	0x0c,
	0x0b,
	0x0b,
	0x0d,
	0x12,
	0x1b,
	0x11,
	0x09,
	0x07,
	0x03,
	0x04,
	0x07,
	0x0b,
	0x0e,
	0x0d,
	0x08,
	0x03,
	0x00,
	0x00,
	0x03,
	0x07,
	0x0b,
	0x0c,
	0x07,
	0x02,
	0x01,
	0x01,
	0x02,
	0x08,
	0x0c,
	0x10,
	0x0a,
	0x06,
	0x05,
	0x05,
	0x06,
	0x0b,
	0x0e,
	0x20,
	0x14,
	0x0d,
	0x0a,
	0x0a,
	0x0d,
	0x14,
	0x1e,
	0x45,
	0x28,
	0x20,
	0x1b,
	0x1b,
	0x1e,
	0x28,
	0x40,
	0x0b,
	0x0a,
	0x0b,
	0x0b,
	0x0c,
	0x08,
	0x10,
	0x0d,
	0x0e,
	0x0e,
	0x0e,
	0x0c,
	0x0f,
	0x0f,
	0x11,
	0x10,
	0x10,
	0x0b,
	0x10,
	0x0f,
	0x10,
	0x10,
	0x10,
	0x0b,
	0x0e,
	0x0d,
	0x0e,
	0x0d,
	0x0d,
	0x0a,
	0x0f,
	0x08,
	0x09,
	0x0a,
	0x0b,
	0x0a,
	0x19,
	0x1b,
	0x1d,
	0x1c,
	0x1d,
	0x1f,
	0x19,
	0x17,
	0x15,
	0x15,
	0x18,
	0x1c,
	0x18,
	0x14,
	0x11,
	0x10,
	0x15,
	0x1b,
	0x1a,
	0x14,
	0x10,
	0x10,
	0x14,
	0x1c,
	0x1a,
	0x18,
	0x16,
	0x16,
	0x18,
	0x1c,
	0x16,
	0x1b,
	0x1e,
	0x1d,
	0x1b,
	0x1b
};			

static u8 g_capture_gamma_indoor[58] = 
{

};

static u8 g_capture_gamma_outdoor[58] = 
{

};

static u8 g_capture_gamma_middle[58] = 
{

};		

const static u16 g_yuv2jpg_tab[]=
{
	0x3503, 0x07,
	//0x3000, 0x00,
	//0x3001, 0x00,
	0x3002, 0x00,
	0x3003, 0x00,
	0x3005, 0xff,
	0x3006, 0xff,
	0x3007, 0x3f,
	0x350c, 0x07,
	0x350d, 0xd0,
	0x3602, 0xe4,
	0x3612, 0xac,
	0x3613, 0x44,
	0x3621, 0x27,
	0x3622, 0x08,
	0x3623, 0x22,
	0x3604, 0x60,
	0x3705, 0xda,
	0x370a, 0x80,
	0x3801, 0x8a,
	0x3803, 0x0a,
	0x3804, 0x0a,
	0x3805, 0x20,
	0x3806, 0x07,
	0x3807, 0x98,
	0x3808, 0x0a,
	0x3809, 0x20,
	0x380a, 0x07,
	0x380b, 0x98,
	0x380c, 0x0c,
	0x380d, 0x80,
	0x380e, 0x07,
	0x380f, 0xd0,
	0x3810, 0xc2,
	0x3815, 0x44,
	0x3818, 0xc8,
	0x3824, 0x01,
	0x3827, 0x0a,
	0x3a00, 0x78,
	0x3a0d, 0x10,
	0x3a0e, 0x0d,
	0x3a00, 0x78,
	0x460b, 0x35,
	0x471d, 0x00,
	0x4713, 0x03,
	0x471c, 0x50,
	0x5682, 0x0a,
	0x5683, 0x20,
	0x5686, 0x07,
	0x5687, 0x98,
	0x5001, 0x7f,
	0x589b, 0x00,
	0x589a, 0xc0,
	0x4402, 0x90,
	0x4407, 0x04,
	0x589b, 0x00,
	0x589a, 0xc0,
	0x3002, 0x0c,
	0x3002, 0x00,
	0x380c ,0x0d,
	0x380d ,0x8b,
	// for vysc and clock
	0x4740 ,0x21,//
	0x4602 ,0x04,//

	// for af board
	0x3030  ,0x0b,
	
	0x3819  ,0x81,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */
	
};
const static u16 g_jpg2yuv_tab[]=
{
	
	0x3503, 0x00, 
	//0x3000, 0xd8,
	//0x3001, 0x08,
	0x3002, 0x5c,
	0x3003, 0x02,
	0x3005, 0xf7,
	0x3006, 0x43,
	0x3007, 0x37,
	0x350c, 0x03,
	0x350d, 0xe8,
	0x3602, 0xfc,
	0x3612, 0xff,
	0x3613, 0x00,
	0x3621, 0xc7,
	0x3622, 0x00,
	0x3623, 0x00,
	0x3604, 0x40,
	0x3705, 0xdb,
	0x370a, 0x81,
	0x3801, 0x50,
	0x3803, 0x08,
	0x3804, 0x05,
	0x3805, 0x00,
	0x3806, 0x03,
	0x3807, 0xc0,
	0x3808, 0x02,
	0x3809, 0x80,
	0x380a, 0x01,
	0x380b, 0xe0,
	0x380c, 0x0c,
	0x380d, 0x80,
	0x380e, 0x03,
	0x380f, 0xe8,
	0x3810, 0xc0,
	0x3815, 0x04,
	0x3818, 0xc1,
	0x3824, 0x11,
	0x3827, 0x08,
	0x3a00, 0x78,
	0x3a0d, 0x08,
	0x3a0e, 0x06,
	0x460b, 0x37,
	0x471d, 0x05,
	0x4713, 0x02,
	0x471c, 0xd0,
	0x5682, 0x05,
	0x5683, 0x00,
	0x5686, 0x03,
	0x5687, 0xc0,
	0x5001, 0x7f,
	0x589b, 0x04,
	0x589a, 0xc5,
	0x4407, 0x0c,
	0x589b, 0x04,
	0x589a, 0xc5,
	0x380c, 0x0d,
	0x380d, 0x8b,
	0x3030, 0x0b,
	
	0x3819, 0x80,

	OV5642_REGEND, 0x00	    /* End of list delimiter */
	
};

const static u16 gYUV_initialize_tab[] = 
{
	0x3103 ,0x93,
	0x3008 ,0x02,
	0x3017 ,0x7f,
	0x3018 ,0xfc,
	0x3810 ,0xc2,
	0x3615 ,0xf0,
	0x3000 ,0x00,
	0x3001 ,0x00,
	0x3002 ,0x00,
	0x3003 ,0x00,
	0x3001 ,0x00,
	0x3002 ,0x5c,
	0x3003 ,0x02,
	0x3004 ,0x07,
	0x3005 ,0xf7,
	0x3006 ,0x43,
	0x3007 ,0x37,
	0x3011 ,0x08,
	0x3010 ,0x10,
	0x460c ,0x22,
	0x3815 ,0x04,
	0x370d ,0x06,
	0x370c ,0xa0,
	0x3602 ,0xfc,
	0x3612 ,0xff,
	0x3634 ,0xc0,
	0x3613 ,0x00,
	0x3605 ,0x7c,
	0x3621 ,0x09,
	0x3622 ,0x00,
	0x3604 ,0x40,
	0x3603 ,0xa7,
	0x3603 ,0x27,
	0x4000 ,0x21,
	0x401d ,0x02,
	0x3600 ,0x54,
	0x3605 ,0x04,
	0x3606 ,0x1b,
	0x3c01 ,0x80,
	0x5000 ,0x4f,
	0x5020 ,0x04,
	0x5181 ,0x79,
	0x5182 ,0x00,
	0x5185 ,0x22,
	0x5197 ,0x01,
	0x5001 ,0xff,
	0x5500 ,0x0a,
	0x5504 ,0x00,
	0x5505 ,0x7f,
	0x5080 ,0x08,
	0x300e ,0x18,
	0x4610 ,0x00,
	0x471d ,0x05,
	0x4708 ,0x06,
	0x3710 ,0x10,
	0x3632 ,0x41,
	0x3702 ,0x40,
	0x3620 ,0x37,
	0x3631 ,0x01,
	0x3808 ,0x02,
	0x3809 ,0x80,
	0x380a ,0x01,
	0x380b ,0xe0,
	0x380e ,0x07,
	0x380f ,0xd0,
	0x501f ,0x00,
	0x5000 ,0x4f,
	0x4300 ,0x30,
	0x3503 ,0x07,
	0x3501 ,0x73,
	0x3502 ,0x80,
	0x350b ,0x00,
	0x3503 ,0x07,
	0x3824 ,0x11,
	0x3501 ,0x1e,
	0x3502 ,0x80,
	0x350b ,0x7f,
	0x380c ,0x0d,
	0x380d ,0x8b,
	0x380e ,0x03,
	0x380f ,0xe8,
	0x3a0d ,0x04,
	0x3a0e ,0x03,
	0x3818 ,0xc1,
	0x3705 ,0xdb,
	0x370a ,0x81,
	0x3801 ,0x80,
	0x3621 ,0xc7,
	0x3801 ,0x50,
	0x3803 ,0x08,
	0x3827 ,0x08,
	0x3810 ,0xc0,
	0x3804 ,0x05,
	0x3805 ,0x00,
	0x5682 ,0x05,
	0x5683 ,0x00,
	0x3806 ,0x03,
	0x3807 ,0xc0,
	0x5686 ,0x03,
	0x5687 ,0xc0,
	0x3a00 ,0x78,
	0x3a1a ,0x04,
	0x3a13 ,0x30,
	0x3a18 ,0x00,
	0x3a19 ,0x7c,
	0x3a08 ,0x12,
	0x3a09 ,0xc0,
	0x3a0a ,0x0f,
	0x3a0b ,0xa0,
	0x3004 ,0xff,
	0x350c ,0x07,
	0x350d ,0xd0,
	0x3500 ,0x00,
	0x3501 ,0x00,
	0x3502 ,0x00,
	0x350a ,0x00,
	0x350b ,0x00,
	0x3503 ,0x00,
	
	0x3a0f ,0x38,
	0x3a10 ,0x30,
	0x3a1b ,0x38,
	0x3a1e ,0x30,
	0x3a11 ,0x61,
	0x3a1f ,0x10,
	0x3a02 ,0x00,
	0x3a03 ,0x7d,
	0x3a04 ,0x00,
	0x3a14 ,0x00,
	0x3a15 ,0x7d,
	0x3a16 ,0x00,
	0x3a00 ,0x7c,
	0x3a08 ,0x09,
	0x3a09 ,0x60,
	0x3a0a ,0x07,
	0x3a0b ,0xd0,
	0x3a0d ,0x08,
	0x3a0e ,0x06,
	
	0x5193 ,0x70,
	0x3620 ,0x57,
	0x3703 ,0x98,
	0x3704 ,0x1c,
	0x589b ,0x04,
	0x589a ,0xc5,
	
	0x5300 ,0x00,
	0x5301 ,0x20,
	0x5302 ,0x00,
	0x5303 ,0x7c,
	0x530c ,0x00,
	0x530d ,0x0c,
	0x530e ,0x20,
	0x530f ,0x80,
	0x5310 ,0x20,
	0x5311 ,0x80,
	0x5308 ,0x20,
	0x5309 ,0x40,
	0x5304 ,0x00,
	0x5305 ,0x30,
	0x5306 ,0x00,
	0x5307 ,0x80,
	0x5314 ,0x08,
	0x5315 ,0x20,
	0x5319 ,0x30,
	0x5316 ,0x10,
	0x5317 ,0x08,
	0x5318 ,0x02,

	0x3406 ,0x00,
	
	0x5025 ,0x82,

	0x3633 ,0x07,
	0x3702 ,0x10,
	0x3703 ,0xb2,
	0x3704 ,0x18,
	0x370b ,0x40,
	0x370d ,0x02,
	0x3620 ,0x52,
	0x3808 ,0x01,
	0x3809 ,0x40,
	0x380a ,0x00,
	0x380b ,0xf0,
	0x3a00 ,0x7c,
	
	0x380c ,0x0d,
	0x380d ,0x8b,
	// add for AF module
	0x3030, 0x0b,

    0x302c, 0xc2,	/* IO drive capbility */
	OV5642_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 gYUV_CIF_initialize_tab[] = 
{
	0x3800 ,0x1 , 
	0x3801 ,0x50, 
	0x3802 ,0x0 , 
	0x3803 ,0x8 , 
	0x3804 ,0x4 , 
	0x3805 ,0x96,  
	0x3806 ,0x3 , 
	0x3807 ,0xc0, 
	0x3808 ,0x1 , 
	0x3809 ,0x60, 
	0x380a ,0x1 , 
	0x380b ,0x20,
	0x380c ,0xc , 
	0x380d ,0x80, 
	0x380e ,0x3 , 
	0x380f ,0xe8, 
	0x5001 ,0x7f, 
	0x5680 ,0x0 , 
	0x5681 ,0x0 , 
	0x5682 ,0x4 , 
	0x5683 ,0x96,  
	0x5684 ,0x0 , 
	0x5685 ,0x0 , 
	0x5686 ,0x3 , 
	0x5687 ,0xc0, 
	
	0x302c,0xc2,//capbility
	0x4300,0x32,//vyuy

	OV5642_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 gYUV_QCIF_initialize_tab[]=
{
	0x3800 ,0x1 , 
	0x3801 ,0x50, 
	0x3802 ,0x0 , 
	0x3803 ,0x8 , 
	0x3804 ,0x4 , 
	0x3805 ,0x96,  
	0x3806 ,0x3 , 
	0x3807 ,0xc0, 
	0x3808 ,0x0 , 
	0x3809 ,0xb0, 
	0x380a ,0x0 , 
	0x380b ,0x90,
	0x380c ,0xc , 
	0x380d ,0x80, 
	0x380e ,0x3 , 
	0x380f ,0xe8, 
	0x5001 ,0x7f, 
	0x5680 ,0x0 , 
	0x5681 ,0x0 , 
	0x5682 ,0x4 , 
	0x5683 ,0x96,  
	0x5684 ,0x0 , 
	0x5685 ,0x0 , 
	0x5686 ,0x3 , 
	0x5687 ,0xc0, 
	0x302c,0xc2,//capbility
	0x4300,0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_SQCIF_initialize_tab[]=
{
	0x3824	,0x11,
	0x3825	,0xB4,
	0x3826	,0x0 ,
	0x3827	,0x8 ,
	0x3804	,0x5 ,
	0x3805	,0x0 ,
	0x3806	,0x3 ,
	0x3807	,0xC0,
	0x3808	,0x0 ,
	0x3809	,0x80,
	0x380a	,0x0 ,
	0x380b	,0x60,
	0x380c	,0xc ,
	0x380d	,0x80,
	0x380e	,0x3 ,
	0x380f	,0xe8,
	0x5001	,0x7f,
	0x5680	,0x0 ,
	0x5681	,0x0 ,
	0x5682	,0x5 ,
	0x5683	,0x0 ,
	0x5684	,0x0 ,
	0x5685	,0x0 ,
	0x5686	,0x3 ,
	0x5687	,0xC0,
	0x3815	,0x4 ,
	0x302c	,0xc2,//capbility
	0x4300	,0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_160X120_initialize_tab[]=
{
	0x3824	,0x11,
	0x3825	,0xAA,
	0x3826	,0x0 ,
	0x3827	,0x8 ,
	0x3804	,0x5 ,
	0x3805	,0xE ,
	0x3806	,0x3 ,
	0x3807	,0xCC,
	0x3808	,0x0 ,
	0x3809	,0xA0,
	0x380a	,0x0 ,
	0x380b	,0x78,
	0x380c	,0xC ,
	0x380d	,0x80,
	0x380e	,0x3 ,
	0x380f	,0xE8,
	0x5001	,0xF7,
	0x5680	,0x0 ,
	0x5681	,0x0 ,
	0x5682	,0x5 ,
	0x5683	,0xE ,
	0x5684	,0x0 ,
	0x5685	,0x0 ,
	0x5686	,0x3 ,
	0x5687	,0xCC,
	0x3815	,0x4 ,
	0x302c	,0xc2,//capbility
	0x4300	,0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_432X320_initialize_tab[]=
{
	0x3824	,0x11,
	0x3825	,0xB4,
	0x3826	,0x0 ,
	0x3827	,0x8 ,
	0x3804	,0x5 ,
	0x3805	,0x0 ,
	0x3806	,0x3 ,
	0x3807	,0xC0,
	0x3808	,0x1 ,
	0x3809	,0xB0,
	0x380a	,0x1 ,
	0x380b	,0x40,
	0x380c	,0xc ,
	0x380d	,0x80,
	0x380e	,0x3 ,
	0x380f	,0xe8,
	0x5001	,0x7f,
	0x5680	,0x0 ,
	0x5681	,0x0 ,
	0x5682	,0x5 ,
	0x5683	,0x0 ,
	0x5684	,0x0 ,
	0x5685	,0x0 ,
	0x5686	,0x3 ,
	0x5687	,0xC0,
	0x3815	,0x4 ,
	0x302c	,0xc2,//capbility
	0x4300	,0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_480X360_initialize_tab[]=
{
	0x3824	,0x11,
	0x3825	,0xB4,
	0x3826	,0x0 ,
	0x3827	,0x8 ,
	0x3804	,0x5 ,
	0x3805	,0x0 ,
	0x3806	,0x3 ,
	0x3807	,0xC0,
	0x3808	,0x1 ,
	0x3809	,0xE0,
	0x380a	,0x1 ,
	0x380b	,0x68,
	0x380c	,0xc ,
	0x380d	,0x80,
	0x380e	,0x3 ,
	0x380f	,0xe8,
	0x5001	,0x7f,
	0x5680	,0x0 ,
	0x5681	,0x0 ,
	0x5682	,0x5 ,
	0x5683	,0x0 ,
	0x5684	,0x0 ,
	0x5685	,0x0 ,
	0x5686	,0x3 ,
	0x5687	,0xC0,
	0x3815	,0x4 ,
	0x302c	,0xc2,//capbility
	0x4300	,0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_QVGA_initialize_tab[]=
{
	0x3824, 0x11,
	0x3825, 0xB4,
	0x3826, 0x0 ,
	0x3827, 0x8 ,
	0x3804, 0x5 ,
	0x3805, 0x0 ,
	0x3806, 0x3 ,
	0x3807, 0xC0,
	0x3808, 0x1 ,
	0x3809, 0x40,
	0x380A, 0x0 ,
	0x380B, 0xF0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x3 ,
	0x380F, 0xE8,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0x5 ,
	0x5683, 0x0 ,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x3 ,
	0x5687, 0xC0,

	0x4300 , 0x32,//vyuy
	OV5642_REGEND ,0x00
};

const static u16 gYUV_VGA_initialize_tab[]=
{
	0x3824, 0x11,
	0x3825, 0xB4,
	0x3826, 0x0 ,
	0x3827, 0x8 ,
	0x3804, 0x5 ,
	0x3805, 0x0 ,
	0x3806, 0x3 ,
	0x3807, 0xC0,
	0x3808, 0x2 ,
	0x3809, 0x80,
	0x380A, 0x1 ,
	0x380B, 0xE0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x3 ,
	0x380F, 0xE8,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0x5 ,
	0x5683, 0x0 ,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x3 ,
	0x5687, 0xC0,
	
	0x4300 , 0x32,//vyuy
	OV5642_REGEND ,0x00
};
const static u16 gYUV_854X480_initialize_tab[]=
{
	0x3824,0x11,
	0x3825,0xB4,
	0x3826,0x0 ,
	0x3827,0x8 ,
	0x3804,0x5 ,
	0x3805,0x0 ,
	0x3806,0x3 ,
	0x3807,0xC0,
	0x3808,0x3 ,
	0x3809,0x60,
	0x380a,0x1 ,
	0x380b,0xE0,
	0x380c,0xd ,
	0x380d,0x8b,
	0x380e,0x3 ,
	0x380f,0xe8,
	0x5001,0x7f,
	0x5680,0x0 ,
	0x5681,0x0 ,
	0x5682,0x5 ,
	0x5683,0x0 ,
	0x5684,0x0 ,
	0x5685,0x0 ,
	0x5686,0x3 ,
	0x5687,0xC0,
	0x3815,0x2 ,

	0x4300 , 0x32,//vyuy
	OV5642_REGEND ,0x00
};
const static u16 gYUV_1280X720_initialize_tab[]=
{

	OV5642_REGEND ,0x00
};

const static u16 gJPG_QVGA_initialize_tab[]=
{
	
	// qvga set
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x1 ,
	0x3809, 0x40,
	0x380A, 0x0 ,
	0x380B, 0xF0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
		
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_VGA_initialize_tab[]=
{
	
	// vga set
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x2 ,
	0x3809, 0x80,
	0x380A, 0x1 ,
	0x380B, 0xE0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 gJPG_854X480_initialize_tab[]=
{
	
	//  864 480 
	0x3808 ,0x03,
	0x3809 ,0x60,
	0x380a ,0x01,
	0x380b ,0xe0,
	0x5001 ,0x7f,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 gJPG_XGA_initialize_tab[]=
{
	
	// xga set 1024 768
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x4 ,
	0x3809, 0x0 ,
	0x380A, 0x3 ,
	0x380B, 0x0 ,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,

	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_1280X960_initialize_tab[]=
{
	
	// 1280x960 set 1280 960
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x5 ,
	0x3809, 0x0 ,
	0x380A, 0x3 ,
	0x380B, 0xC0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_1600X900_initialize_tab[]=
{
	// uxga set  1600 896
	
	0x3800 ,0x1 ,
	0x3801 ,0x8A,
	0x3802 ,0x0 ,
	0x3803 ,0xA ,
	0x3804 ,0xA ,
	0x3805 ,0x20,
	0x3806 ,0x4 ,
	0x3807 ,0x74,
	0x3808 ,0x6 ,
	0x3809 ,0x40, //1600
	0x380a ,0x3 , 
	0x380b ,0x80, //896
	0x380c ,0xc ,
	0x380d ,0x80,
	0x380e ,0x7 ,
	0x380f ,0xd0,
	0x5001 ,0x7f,
	0x5680 ,0x0 ,
	0x5681 ,0x0 ,
	0x5682 ,0xA ,
	0x5683 ,0x20,
	0x5684 ,0x0 ,
	0x5685 ,0x0 ,
	0x5686 ,0x7 ,
	0x5687 ,0x98,

	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_UXGA_initialize_tab[]=
{
	// uxga set  1600 1200
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x6 ,
	0x3809, 0x40,
	0x380A, 0x4 ,
	0x380B, 0xB0,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 gJPG_2048X1152_initialize_tab[]=
{
	// qxga set 2048 1152
	0x3808 ,0x08,
	0x3809 ,0x00,
	0x380a ,0x04,
	0x380b ,0x80,
	0x5001 ,0x7f,
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_QXGA_initialize_tab[]=
{
	// qxga set 2048 1536
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0x8 ,
	0x3809, 0x0 ,
	0x380A, 0x6 ,
	0x380B, 0x0 ,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};
const static u16 gJPG_2592X1458_initialize_tab[]=
{
	0x3808 ,0x0a,
	0x3809 ,0x20,
	0x380a ,0x05,
	0x380b ,0xb0,
	0x5001 ,0x7f,
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 gJPG_2560X1920_initialize_tab[]=
{
	0x3800, 0x1 ,
	0x3801, 0x8A,
	0x3802, 0x0 ,
	0x3803, 0xA ,
	0x3804, 0xA ,
	0x3805, 0x20,
	0x3806, 0x7 ,
	0x3807, 0x98,
	0x3808, 0xA ,
	0x3809, 0x0 ,
	0x380A, 0x7 ,
	0x380B, 0x80,
	0x380C, 0xC ,
	0x380D, 0x80,
	0x380E, 0x7 ,
	0x380F, 0xD0,
	0x3815, 0x4 ,
	0x5001, 0xF7,
	0x5680, 0x0 ,
	0x5681, 0x0 ,
	0x5682, 0xA ,
	0x5683, 0x20,
	0x5684, 0x0 ,
	0x5685, 0x0 ,
	0x5686, 0x7 ,
	0x5687, 0x98,
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

const static u16 gJPG_QSXGA_initialize_tab[]=
{
	
	OV5642_REGEND, 0x00	    /* End of list delimiter */

};

//TDO: add format here
struct OV5642_SETTING g_ov5642_setting_table[] =
{	
	
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_SQCIF, 		gYUV_SQCIF_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_160X120, 		gYUV_160X120_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_QCIF , 		gYUV_QCIF_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_CIF  , 		gYUV_CIF_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_QVGA, 		gYUV_QVGA_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_432X320 , 	gYUV_432X320_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_480X360  , 	gYUV_480X360_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_VGA, 		gYUV_VGA_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_854X480, 	gYUV_854X480_initialize_tab},
	{OV5642_FORMAT_YUV_422, OV5642_SIZE_1280X720,	gYUV_1280X720_initialize_tab},
	
	{OV5642_FORMAT_JPG, OV5642_SIZE_QVGA,			gJPG_QVGA_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_VGA,			gJPG_VGA_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_854X480,		gJPG_854X480_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_XGA,			gJPG_XGA_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_1280X960  , 		gJPG_1280X960_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_1600X900  , 		gJPG_1600X900_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_UXGA  , 			gJPG_UXGA_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_2048X1152  , 	gJPG_2048X1152_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_QXGA  , 			gJPG_QXGA_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_2592X1458  , 	gJPG_2592X1458_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_2560X1920  , 	gJPG_2560X1920_initialize_tab},
	{OV5642_FORMAT_JPG, OV5642_SIZE_QSXGA  , 		gJPG_QSXGA_initialize_tab},	
	{0, 0, 0}
} ;

//TDO: Image Quality

//Contrast
const u16 g_Contrast_Standard_Settings[]=
{
	0x5001 ,0xff,
	0x5587 ,0x20,
	0x5588 ,0x20, 
	0x558a ,0x00, 
	OV5642_REGEND ,0x00
};
const u16 g_Contrast_Highest_Settings[]=
{
	0x5001 ,0xff,
	0x5587 ,0x30,
	0x5588 ,0x30,
	0x558a ,0x00,
	OV5642_REGEND ,0x00
};
const u16 g_Contrast_High_Settings[]=
{
	0x5001 ,0xff,
	0x5587 ,0x28,
	0x5588 ,0x28,
	0x558a ,0x00,
	OV5642_REGEND ,0x00
};
const u16 g_Contrast_Low_Settings[]=
{
	0x5001 ,0xff,
	0x5587 ,0x18,
	0x5588 ,0x18,
	0x558a ,0x00, 

	OV5642_REGEND ,0x00
};
const u16 g_Contrast_Lowest_Settings[]=
{
	0x5001 ,0xff,
	0x5587 ,0x10,
	0x5588 ,0x10,
	0x558a ,0x00,
	OV5642_REGEND ,0x00
};

//Sharpness
const u16 g_Sharpness_Standard_Settings[]=
{
	0x530A ,0x00,

	OV5642_REGEND ,0x00
};
const u16 g_Sharpness_Highest_Settings[]=
{
	0x530A ,0x08,
	0x531e ,0x1f,
	0x531f ,0x1f,
	OV5642_REGEND ,0x00
};
const u16 g_Sharpness_High_Settings[]=
{
	0x530A ,0x08,
	0x531e ,0x0f,
	0x531f ,0x0f,
	OV5642_REGEND ,0x00
};
const u16 g_Sharpness_Low_Settings[]=
{
	0x530A ,0x08,
	0x531e ,0x08,
	0x531f ,0x08,
	OV5642_REGEND ,0x00
};
const u16 g_Sharpness_Lowest_Settings[]=
{
	0x530A ,0x08,
	0x531e ,0x04,
	0x531f ,0x04,
	OV5642_REGEND ,0x00
};

//Satuation
const u16 g_Satuation_Standard_Settings[]=
{
	0x5001 ,0xff, 
	0x5583 ,0x40,
	0x5584 ,0x40,

	OV5642_REGEND ,0x00
};
const u16 g_Satuation_Highest_Settings[]=
{
	0x5001 ,0xff, 
	0x5583 ,0x80,
	0x5584 ,0x80,

	OV5642_REGEND ,0x00
};
const u16 g_Satuation_High_Settings[]=
{
	0x5001 ,0xff, 
	0x5583 ,0x60,
	0x5584 ,0x60,

	OV5642_REGEND ,0x00
};
const u16 g_Satuation_Low_Settings[]=
{
	0x5001 ,0xff,
	0x5583 ,0x20,
	0x5584 ,0x20,

	OV5642_REGEND ,0x00
};
const u16 g_Satuation_Lowest_Settings[]=
{
	0x5001 ,0xff,
	0x5583 ,0x00,
	0x5584 ,0x00,

	OV5642_REGEND ,0x00
};

//White Balance
const u16 g_WB_Auto_Settings[]=
{
	0x3406 ,0x00,

	0x5180, 0xff,  
	0x5181, 0x50,  
	0x5182, 0x00,  
	0x5183, 0x00,  
	0x5184, 0x25,  
	0x5185, 0xff,  
	0x5186, 0x16,  
	0x5187, 0x15,  
	0x5188, 0x18,  
	0x5189, 0x85,  
	0x518a, 0x5a,  
	0x518b, 0xef,  
	0x518c, 0xaa,  
	0x518d, 0x2f,  
	0x518e, 0x2c,  
	0x518f, 0x6e,  
	0x5190, 0x44,  
	0x5191, 0xf8,  
	0x5192, 0x04,  
	0x5193, 0x70,  
	0x5194, 0xf0,  
	0x5195, 0xf0,  
	0x5196, 0x03,  
	0x5197, 0x01,  
	0x5198, 0x06,  
	0x5199, 0x28,  
	0x519a, 0x04,  
	0x519b, 0x00,  
	0x519c, 0x04,  
	0x519d, 0xcd,  
	0x519e, 0x00,  

	OV5642_REGEND ,0x00
};

const u16 g_WB_Sunny_Settings[]=
{
	0x3406, 0x01,
    0x3400, 0x07,
    0x3401, 0x82,
    0x3402, 0x04,
    0x3403, 0x00,
    0x3404, 0x05,
    0x3405, 0x15,
    
	OV5642_REGEND ,0x00
};

const u16 g_WB_Cloudy_Settings[]=
{
	0x3406, 0x01,
	0x3400, 0x07,
	0x3401, 0x82,
	0x3402, 0x04,
	0x3403, 0x00,
	0x3404, 0x06,
	0x3405, 0x15,
	
	OV5642_REGEND ,0x00
};

const u16 g_WB_Home_Settings[]=
{
	0x3406, 0x01,
    0x3400, 0x03,
    0x3401, 0x00,
    0x3402, 0x03,
    0x3403, 0x00,
    0x3404, 0x06,
    0x3405, 0x80,
    
	OV5642_REGEND ,0x00
};
const u16 g_WB_Office_Settings[]=
{
	0x3406, 0x01,
    0x3400, 0x04,
    0x3401, 0x00,
    0x3402, 0x03,
    0x3403, 0x00,
    0x3404, 0x06,
    0x3405, 0x00,

	OV5642_REGEND ,0x00
};

//Color Tone
const u16 g_ColorTone_Off_Settings[]=
{
	0x5001 ,0x7f,
	0x5580 ,0x06,
	OV5642_REGEND ,0x00
};
const u16 g_ColorTone_Cool_Settings[]=
{
	0x5001 ,0xff,
	0x5580 ,0x18,
	0x5585 ,0xa0,
	0x5586 ,0x40,
	OV5642_REGEND ,0x00
};

const u16 g_ColorTone_Warm_Settings[]=
{
	0x5001 ,0xff,
	0x5580 ,0x18,
	0x5585 ,0x80,
	0x5586 ,0xc0,
	OV5642_REGEND ,0x00
};
const u16 g_ColorTone_BW_Settings[]=
{
	0x5001 ,0xff,
	0x5580 ,0x18,
	0x5585 ,0x80,
	0x5586 ,0x80,
	OV5642_REGEND ,0x00
};
const u16 g_ColorTone_Sepia_Settings[]=
{
	0x5001 ,0xff,
	0x5580 ,0x18,
	0x5585 ,0x40,
	0x5586 ,0xa0,
	OV5642_REGEND ,0x00
};

//Scene Mode
const u16 g_Scene_Auto_Settings[]=
{
	OV5642_REGEND ,0x00
};
const u16 g_Scene_CloseUp_Settings[]=
{
	OV5642_REGEND ,0x00
};
const u16 g_Scene_Portrait_Settings[]=
{
	OV5642_REGEND ,0x00
};
const u16 g_Scene_Landscape_Settings[]=
{
	OV5642_REGEND ,0x00
};
const u16 g_Scene_Sports_Settings[]=
{
	OV5642_REGEND ,0x00
};
const u16 g_Scene_Night_Settings[]=
{
	0x3010, 0x30,
	0x3a00, 0x7c,
	0x3a17, 0x89,
	0x3a02, 0x01,
	0x3a03, 0x7d,

	OV5642_REGEND ,0x00
};

//JPEG Quality
const u16 g_JpegQuality_Standard_Settings[]=
{
	0x4407, 0x06,
	OV5642_REGEND ,0x00
};
const u16 g_JpegQuality_Fine_Settings[]=
{
	0x4407, 0x04,
	OV5642_REGEND ,0x00
};
const u16 g_JpegQuality_Superfine_Settings[]=
{
	0x4407, 0x02,
	OV5642_REGEND ,0x00
};
const u16 g_JpegQuality_Low_Settings[]=
{
	0x4407, 0x08,
	OV5642_REGEND ,0x00
};

//EV compensation
const u16 g_EVCompensat_0_Setting[]=
{
					
	// -1.7EV
	0x3a0f ,0x48,
	0x3a10 ,0x40,
	0x3a1b ,0x48,
	0x3a1e ,0x40,
	0x3a11 ,0x80,
	0x3a1f ,0x10,
	OV5642_REGEND ,0x00
};               

const u16 g_EVCompensat_1_Setting[]=
{
	  //-1.3EV          
	0x3a0f ,0x50,
	0x3a10 ,0x48,
	0x3a1b ,0x50,
	0x3a1e ,0x48,
	0x3a11 ,0x80,
	0x3a1f ,0x10,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_2_Setting[]=
{
	 // -1.0EV          
	0x3a0f ,0x58,
	0x3a10 ,0x50,
	0x3a1b ,0x58,
	0x3a1e ,0x50,
	0x3a11 ,0x80,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_3_Setting[]=
{
	 // -0.7EV          
	0x3a0f ,0x60,
	0x3a10 ,0x58,
	0x3a1b ,0x60,
	0x3a1e ,0x58,
	0x3a11 ,0x90,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_4_Setting[]=
{
	//-0.3EV          
	0x3a0f ,0x68,
	0x3a10 ,0x60,
	0x3a1b ,0x68,
	0x3a1e ,0x60,
	0x3a11 ,0xa0,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_5_Setting[]=
{
	  //default         
	0x3a0f ,0x70,
	0x3a10 ,0x68,
	0x3a1b ,0x70,
	0x3a1e ,0x68,
	0x3a11 ,0xb0,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_6_Setting[]=
{
	 //0.3EV           
	0x3a0f ,0x78,
	0x3a10 ,0x70,
	0x3a1b ,0x78,
	0x3a1e ,0x70,
	0x3a11 ,0xb0,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_7_Setting[]=
{
	// 0.7EV           
	0x3a0f ,0x80,
	0x3a10 ,0x78,
	0x3a1b ,0x80,
	0x3a1e ,0x78,
	0x3a11 ,0xc0,
	0x3a1f ,0x20,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_8_Setting[]=
{
	  //1.0EV           
	0x3a0f ,0x88,
	0x3a10 ,0x80,
	0x3a1b ,0x88,
	0x3a1e ,0x80,
	0x3a11 ,0xc0,
	0x3a1f ,0x30,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_9_Setting[]=
{
	//1.3EV           
	0x3a0f ,0x90,
	0x3a10 ,0x88,
	0x3a1b ,0x90,
	0x3a1e ,0x88,
	0x3a11 ,0xc0,
	0x3a1f ,0x30,
	OV5642_REGEND ,0x00
};
const u16 g_EVCompensat_10_Setting[]=
{
	 // 1.7EV           
	0x3a0f ,0x98,
	0x3a10 ,0x90,
	0x3a1b ,0x98,
	0x3a1e ,0x90,
	0x3a11 ,0xd0,
	0x3a1f ,0x40,
	OV5642_REGEND ,0x00
};

const u16 g_Framerate_15_Night_Setting[]=
{
	0x300F ,0x06,
	0x3010 ,0x10,
	0x3011 ,0x08,
	0x3012 ,0x00,
	0x380C ,0x0D,
	0x380D ,0x8B,
	0x3815 ,0x44,
	0x460C ,0x22,
	0x3615 ,0x00,
	0x3A08 ,0x09,
	0x3A09 ,0x60,
	0x3A0E ,0x06,
	0x3A0A ,0x07,
	0x3A0B ,0xD0,
	0x3A0D ,0x08,

	0x3a00 ,0x7c,	/* Night on */
	OV5642_REGEND ,0x00
};
                 
const u16 g_Framerate_15_Setting[]=
{
	0x300F ,0x06,
	0x3010 ,0x10,
	0x3011 ,0x08,
	0x3012 ,0x00,
	0x380C ,0x0D,
	0x380D ,0x8B,
	0x3815 ,0x44,
	0x460C ,0x22,
	0x3615 ,0x00,
	0x3A08 ,0x09,
	0x3A09 ,0x60,
	0x3A0E ,0x06,
	0x3A0A ,0x07,
	0x3A0B ,0xD0,
	0x3A0D ,0x08,

	0x3a00 ,0x78,	/* Night off */
	OV5642_REGEND ,0x00
};
                 
const u16 g_Framerate_25_Setting[]=
{
	0x300F ,0x06, 
	0x3010 ,0x00, 
	0x3011 ,0x06, 
	0x3012 ,0x00, 
	0x380C ,0x0C, 
	0x380D ,0x30,
	0x3815 ,0x44,
	0x460C ,0x22,
	0x3615 ,0x00, 
	0x3A08 ,0x0F, 
	0x3A09 ,0xA0,
	0x3A0E ,0x04, 
	0x3A0A ,0x0D, 
	0x3A0B ,0x00, 
	0x3A0D ,0x04, 

	0x3a00 ,0x78,	/* Night off */
	OV5642_REGEND ,0x00
};

const u16 g_Framerate_30_Setting[]=
{
	0x300F ,0x06, 
	0x3010 ,0x00, 
	0x3011 ,0x08, 
	0x3012 ,0x00, 
	0x380C ,0x0D, 
	0x380D ,0x8B,
	0x3815 ,0x48,
	0x460C ,0x22,
	0x3615 ,0x00, 
	0x3A08 ,0x12,
	0x3A09 ,0xC0,
	0x3A0E ,0x03, 
	0x3A0A ,0x0F, 
	0x3A0B ,0xA0,
	0x3A0D ,0x04, 

	0x3a00 ,0x78,	/* Night off */
	OV5642_REGEND ,0x00
};

#define AUTO_FOCUS_FIRMWARE_TAB1_COUNTX 22
#define AUTO_FOCUS_FIRMWARE_TAB1_COUNTY 258
static unsigned char gAutoFocusFirmwareInitTab1[AUTO_FOCUS_FIRMWARE_TAB1_COUNTX][AUTO_FOCUS_FIRMWARE_TAB1_COUNTY] =
{
	{
		0x80,
		0x00,
		0x02,
		0x00,
		0x06,
		0x02,
		0x0b,
		0xce,
		0x78,
		0x7f,
		0xe4,
		0xf6,
		0xd8,
		0xfd,
		0x75,
		0x81,
		0x7e,
		0x02,
		0x14,
		0x61,
		0x00,
		0x02,
		0x14,
		0x17,
		0xe0,
		0xf5,
		0x72,
		0xa3,
		0xe0,
		0xf5,
		0x73,
		0xae,
		0x6a,
		0xe4,
		0x85,
		0x6b,
		0x55,
		0x8e,
		0x54,
		0xf5,
		0x53,
		0xf5,
		0x52,
		0xab,
		0x55,
		0xaa,
		0x54,
		0xa9,
		0x53,
		0xa8,
		0x52,
		0xaf,
		0x2c,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x55,
		0x8e,
		0x54,
		0x8d,
		0x53,
		0x8c,
		0x52,
		0xaf,
		0x55,
		0xae,
		0x54,
		0xad,
		0x53,
		0xac,
		0x52,
		0x8f,
		0x2b,
		0x8e,
		0x2a,
		0x8d,
		0x29,
		0x8c,
		0x28,
		0xae,
		0x6c,
		0xe4,
		0x85,
		0x6d,
		0x55,
		0x8e,
		0x54,
		0xf5,
		0x53,
		0xf5,
		0x52,
		0xab,
		0x55,
		0xaa,
		0x54,
		0xa9,
		0x53,
		0xa8,
		0x52,
		0xaf,
		0x2d,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x55,
		0x8e,
		0x54,
		0x8d,
		0x53,
		0x8c,
		0x52,
		0xe5,
		0x2b,
		0x25,
		0x55,
		0xf5,
		0x2b,
		0xe5,
		0x2a,
		0x35,
		0x54,
		0xf5,
		0x2a,
		0xe5,
		0x29,
		0x35,
		0x53,
		0xf5,
		0x29,
		0xe5,
		0x28,
		0x35,
		0x52,
		0xf5,
		0x28,
		0xae,
		0x6e,
		0xe4,
		0x85,
		0x6f,
		0x55,
		0x8e,
		0x54,
		0xf5,
		0x53,
		0xf5,
		0x52,
		0xab,
		0x55,
		0xaa,
		0x54,
		0xa9,
		0x53,
		0xa8,
		0x52,
		0xaf,
		0x2e,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x55,
		0x8e,
		0x54,
		0x8d,
		0x53,
		0x8c,
		0x52,
		0xe5,
		0x2b,
		0x25,
		0x55,
		0xf5,
		0x2b,
		0xe5,
		0x2a,
		0x35,
		0x54,
		0xf5,
		0x2a,
		0xe5,
		0x29,
		0x35,
		0x53,
		0xf5,
		0x29,
		0xe5,
		0x28,
		0x35,
		0x52,
		0xf5,
		0x28,
		0xae,
		0x70,
		0xe4,
		0x85,
		0x71,
		0x55,
		0x8e,
		0x54,
		0xf5,
		0x53,
		0xf5,
		0x52,
		0xab,
		0x55,
		0xaa,
		0x54,
		0xa9,
		0x53,
		0xa8,
		0x52,
		0xaf,
		0x2f,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x55,
		0x8e,
		0x54,
		0x8d,
		0x53,
		0x8c,
		0x52,
		0xe5,
		0x2b,
		0x25,
		0x55,
		0xf5,
		0x2b,
		0xe5,
		0x2a,
		0x35,
		0x54,
		0xf5,
		0x2a,
		0xe5,
		0x29,
		0x35,
		0x53,
		0xf5,
		0x29,
		0xe5,
	},
	{
		0x81,
		0x00,
		0x28,
		0x35,
		0x52,
		0xf5,
		0x28,
		0xae,
		0x72,
		0xe4,
		0x85,
		0x73,
		0x55,
		0x8e,
		0x54,
		0xf5,
		0x53,
		0xf5,
		0x52,
		0xab,
		0x55,
		0xaa,
		0x54,
		0xa9,
		0x53,
		0xa8,
		0x52,
		0xaf,
		0x30,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x55,
		0x8e,
		0x54,
		0x8d,
		0x53,
		0x8c,
		0x52,
		0xe5,
		0x2b,
		0x25,
		0x55,
		0xf5,
		0x2b,
		0xe5,
		0x2a,
		0x35,
		0x54,
		0xf5,
		0x2a,
		0xe5,
		0x29,
		0x35,
		0x53,
		0xf5,
		0x29,
		0xe5,
		0x28,
		0x35,
		0x52,
		0xf5,
		0x28,
		0x22,
		0xab,
		0x0d,
		0xaa,
		0x0c,
		0xa9,
		0x0b,
		0xa8,
		0x0a,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x0d,
		0x8e,
		0x0c,
		0x8d,
		0x0b,
		0x8c,
		0x0a,
		0x7b,
		0x40,
		0xe4,
		0xfa,
		0xf9,
		0xf8,
		0x12,
		0x07,
		0x1e,
		0x8f,
		0x0d,
		0x8e,
		0x0c,
		0x8d,
		0x0b,
		0x8c,
		0x0a,
		0x22,
		0xd2,
		0x29,
		0x90,
		0x30,
		0x1b,
		0xe5,
		0x25,
		0xf0,
		0x22,
		0x85,
		0x49,
		0x82,
		0x85,
		0x48,
		0x83,
		0xe5,
		0x7c,
		0x75,
		0xf0,
		0x02,
		0x12,
		0x07,
		0xdf,
		0xe4,
		0x93,
		0xfe,
		0x74,
		0x01,
		0x93,
		0xff,
		0x85,
		0x49,
		0x82,
		0x85,
		0x48,
		0x83,
		0xe4,
		0x93,
		0xfc,
		0x74,
		0x01,
		0x93,
		0xfd,
		0xc3,
		0xef,
		0x9d,
		0xff,
		0xee,
		0x9c,
		0x22,
		0xe5,
		0x0b,
		0x24,
		0x01,
		0xff,
		0xe4,
		0x33,
		0xfe,
		0x22,
		0xaf,
		0x2b,
		0xae,
		0x2a,
		0xad,
		0x29,
		0xac,
		0x28,
		0x78,
		0x06,
		0x12,
		0x07,
		0xb0,
		0x8f,
		0x2b,
		0x8e,
		0x2a,
		0x8d,
		0x29,
		0x8c,
		0x28,
		0xd3,
		0xe5,
		0x29,
		0x94,
		0x00,
		0xe5,
		0x28,
		0x94,
		0x00,
		0x22,
		0x12,
		0x06,
		0x93,
		0x8f,
		0x69,
		0x8e,
		0x68,
		0x8d,
		0x67,
		0x8c,
		0x66,
		0xaf,
		0x69,
		0xae,
		0x68,
		0xad,
		0x67,
		0xac,
		0x66,
		0x22,
		0xe0,
		0x44,
		0x01,
		0xf0,
		0xe0,
		0x44,
		0x02,
		0xf0,
		0xe0,
		0x44,
		0x04,
		0xf0,
		0x22,
		0xd2,
		0x09,
		0x90,
		0x30,
		0x18,
		0xe5,
		0x21,
		0xf0,
		0x22,
		0xe4,
		0x85,
		0x11,
		0x0d,
		0x85,
		0x10,
		0x0c,
		0xf5,
		0x0b,
		0xf5,
		0x0a,
		0xab,
		0x0d,
		0xaa,
		0x0c,
		0xa9,
		0x0b,
		0xa8,
	},
	{
		0x82,
		0x00,
		0x0a,
		0x22,
		0x90,
		0x30,
		0x42,
		0xe0,
		0xf5,
		0x22,
		0x75,
		0x51,
		0x0a,
		0x22,
		0x85,
		0x49,
		0x82,
		0x85,
		0x48,
		0x83,
		0x22,
		0xf5,
		0x82,
		0xe4,
		0x3a,
		0xf5,
		0x83,
		0x02,
		0x07,
		0xc3,
		0x8f,
		0x0a,
		0x74,
		0x4a,
		0x2f,
		0xf8,
		0xe6,
		0x22,
		0xc2,
		0x07,
		0xc2,
		0x06,
		0xc2,
		0x02,
		0xc2,
		0x01,
		0xc2,
		0x00,
		0xc2,
		0x03,
		0xd2,
		0x04,
		0x22,
		0xe5,
		0x16,
		0x25,
		0xe0,
		0x25,
		0xe0,
		0x22,
		0x12,
		0x07,
		0x1e,
		0x8f,
		0x69,
		0x8e,
		0x68,
		0x8d,
		0x67,
		0x8c,
		0x66,
		0x22,
		0xe4,
		0x85,
		0x0f,
		0x0d,
		0x85,
		0x0e,
		0x0c,
		0xf5,
		0x0b,
		0xf5,
		0x0a,
		0x22,
		0x90,
		0x0f,
		0x27,
		0xe4,
		0x93,
		0xff,
		0x90,
		0x30,
		0x0a,
		0xe0,
		0x22,
		0xc2,
		0x02,
		0xc2,
		0x01,
		0xd2,
		0x00,
		0xc2,
		0x03,
		0xc2,
		0x04,
		0x22,
		0x75,
		0xf0,
		0x02,
		0x02,
		0x07,
		0xdf,
		0xe4,
		0x93,
		0xf5,
		0x7c,
		0xa3,
		0xe4,
		0x93,
		0xf5,
		0x32,
		0x22,
		0xd2,
		0x02,
		0xd2,
		0x01,
		0xc2,
		0x00,
		0x22,
		0xd3,
		0xe5,
		0x0b,
		0x94,
		0x80,
		0xe5,
		0x0a,
		0x94,
		0x00,
		0x22,
		0xd3,
		0xe5,
		0x69,
		0x94,
		0xff,
		0xe5,
		0x68,
		0x94,
		0x03,
		0x22,
		0x30,
		0x18,
		0x50,
		0x20,
		0x19,
		0x4d,
		0x75,
		0x0a,
		0x02,
		0x74,
		0x4a,
		0x25,
		0x0a,
		0xf8,
		0xe6,
		0xff,
		0xe5,
		0x4a,
		0xd3,
		0x9f,
		0x40,
		0x04,
		0x7f,
		0x00,
		0x80,
		0x02,
		0xaf,
		0x0a,
		0x12,
		0x02,
		0x1c,
		0xff,
		0xe5,
		0x4b,
		0xd3,
		0x9f,
		0x40,
		0x04,
		0x7f,
		0x01,
		0x80,
		0x02,
		0xaf,
		0x0a,
		0x12,
		0x02,
		0x1c,
		0xff,
		0xe5,
		0x4d,
		0xd3,
		0x9f,
		0x40,
		0x04,
		0x7f,
		0x03,
		0x80,
		0x02,
		0xaf,
		0x0a,
		0x12,
		0x02,
		0x1c,
		0xff,
		0xe5,
		0x4e,
		0xd3,
		0x9f,
		0x40,
		0x04,
		0x7f,
		0x04,
		0x80,
		0x02,
		0xaf,
		0x0a,
		0x12,
		0x02,
		0x1c,
		0xf5,
		0x0b,
		0x80,
		0x06,
		0x85,
		0x79,
		0x0a,
		0x85,
		0x4f,
		0x0b,
		0x7f,
		0x01,
		0xe4,
		0xfe,
		0x74,
		0x4a,
		0x25,
		0x0a,
		0xf8,
		0xe6,
		0xfd,
		0xe5,
		0x0b,
		0xc3,
		0x9d,
		0x50,
		0x04,
		0x7d,
		0x01,
		0x80,
	},
	{
		0x83,
		0x00,
		0x02,
		0x7d,
		0xff,
		0xac,
		0x0b,
		0xe5,
		0x4e,
		0xb5,
		0x0b,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1f,
		0xe5,
		0x4d,
		0xb5,
		0x0b,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1e,
		0xe5,
		0x4c,
		0xb5,
		0x0b,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1d,
		0xe5,
		0x4b,
		0xb5,
		0x0b,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1c,
		0xe5,
		0x4a,
		0xb5,
		0x0b,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1b,
		0xe5,
		0x30,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1f,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1f,
		0xe5,
		0x2f,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1e,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1e,
		0xe5,
		0x2e,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1d,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1d,
		0xe5,
		0x2d,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1c,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1c,
		0xe5,
		0x2c,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1b,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1b,
		0xe5,
		0x23,
		0x54,
		0xf8,
		0x70,
		0x43,
		0xbf,
		0x01,
		0x08,
		0xed,
		0xf4,
		0x04,
		0xfd,
		0x7f,
		0x02,
		0x80,
		0x06,
		0xbf,
		0x02,
		0x02,
		0x7f,
		0x01,
		0x0e,
		0xd3,
		0xed,
		0x64,
		0x80,
		0x94,
		0x80,
		0x40,
		0x12,
		0xec,
		0x2e,
		0xf5,
		0x0b,
		0xc3,
		0x95,
		0x7c,
		0x50,
		0x03,
		0x02,
		0x03,
		0x05,
		0x7d,
		0xff,
		0xe4,
		0xff,
		0x80,
		0x11,
		0xec,
		0xd3,
		0x9e,
		0x50,
		0x0b,
		0x7d,
		0x01,
		0xe4,
		0xff,
		0xec,
		0x2e,
		0xf5,
		0x0b,
		0x02,
		0x03,
		0x05,
		0xc3,
		0xec,
		0x9e,
		0xf5,
		0x0b,
		0x02,
		0x03,
		0x05,
		0x12,
		0x01,
		0x9c,
		0xe5,
		0x4e,
		0xb5,
		0x07,
		0x07,
		0xe4,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1f,
		0x92,
		0x1f,
		0xe5,
		0x4d,
		0xb5,
		0x07,
		0x07,
		0xe4,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1e,
		0x92,
		0x1e,
		0x12,
		0x01,
		0x9c,
		0xe5,
		0x4c,
		0xb5,
		0x07,
		0x07,
		0xe4,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1d,
		0x92,
	},
	{
		0x84,
		0x00,
		0x1d,
		0xe5,
		0x4b,
		0xb5,
		0x07,
		0x07,
		0xe4,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1c,
		0x92,
		0x1c,
		0x12,
		0x01,
		0x9c,
		0xe5,
		0x4a,
		0xb5,
		0x07,
		0x07,
		0xe4,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1b,
		0x92,
		0x1b,
		0xe5,
		0x4e,
		0x12,
		0x01,
		0x9e,
		0xad,
		0x0b,
		0x7c,
		0x00,
		0xef,
		0xb5,
		0x05,
		0x07,
		0xec,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1f,
		0x92,
		0x1f,
		0xe5,
		0x4d,
		0x12,
		0x01,
		0x9e,
		0xef,
		0xb5,
		0x05,
		0x07,
		0xee,
		0xb5,
		0x04,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1e,
		0x92,
		0x1e,
		0xe5,
		0x4c,
		0x12,
		0x01,
		0x9e,
		0xad,
		0x0b,
		0x7c,
		0x00,
		0xef,
		0xb5,
		0x05,
		0x07,
		0xec,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1d,
		0x92,
		0x1d,
		0xe5,
		0x4b,
		0x12,
		0x01,
		0x9e,
		0xef,
		0xb5,
		0x05,
		0x07,
		0xee,
		0xb5,
		0x04,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1c,
		0x92,
		0x1c,
		0xe5,
		0x4a,
		0x12,
		0x01,
		0x9e,
		0x7c,
		0x00,
		0xef,
		0xb5,
		0x0b,
		0x07,
		0xec,
		0xb5,
		0x06,
		0x03,
		0xd3,
		0x80,
		0x02,
		0xa2,
		0x1b,
		0x92,
		0x1b,
		0xe5,
		0x30,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1f,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1f,
		0xe5,
		0x2f,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1e,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1e,
		0xe5,
		0x2e,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1d,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1d,
		0xe5,
		0x2d,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1c,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1c,
		0xe5,
		0x2c,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x04,
		0xa2,
		0x1b,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x1b,
		0x85,
		0x0a,
		0x79,
		0xe5,
		0x7d,
		0xb5,
		0x0b,
		0x03,
		0x02,
		0x15,
		0xa9,
		0x85,
		0x0b,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0xd2,
		0x02,
		0xc2,
		0x01,
		0xd2,
		0x00,
		0x75,
		0x31,
		0x03,
		0x22,
		0x90,
		0x38,
		0x04,
		0xe0,
		0xfe,
		0xa3,
		0xe0,
		0xfd,
		0xed,
		0xff,
		0xee,
		0x54,
		0x0f,
	},
	{
		0x85,
		0x00,
		0xf5,
		0x0e,
		0x8f,
		0x0f,
		0xa3,
		0xe0,
		0xfe,
		0xa3,
		0xe0,
		0xfd,
		0xed,
		0xff,
		0xee,
		0x54,
		0x07,
		0xf5,
		0x10,
		0x8f,
		0x11,
		0xe5,
		0x0e,
		0xc4,
		0xf8,
		0x54,
		0xf0,
		0xc8,
		0x68,
		0xf5,
		0x0e,
		0xe5,
		0x0f,
		0xc4,
		0x54,
		0x0f,
		0x48,
		0xf5,
		0x0f,
		0xe5,
		0x10,
		0xc4,
		0xf8,
		0x54,
		0xf0,
		0xc8,
		0x68,
		0xf5,
		0x10,
		0xe5,
		0x11,
		0xc4,
		0x54,
		0x0f,
		0x48,
		0xf5,
		0x11,
		0xe4,
		0xf5,
		0x17,
		0x75,
		0x16,
		0x04,
		0x12,
		0x02,
		0x33,
		0x24,
		0x34,
		0xf8,
		0xe6,
		0xf5,
		0x12,
		0x12,
		0x02,
		0x33,
		0x24,
		0x35,
		0xf8,
		0xe6,
		0xf5,
		0x14,
		0x12,
		0x02,
		0x33,
		0x24,
		0x36,
		0xf8,
		0xe6,
		0xf5,
		0x13,
		0x12,
		0x02,
		0x33,
		0x24,
		0x37,
		0xf8,
		0xe6,
		0xf5,
		0x15,
		0x12,
		0x02,
		0x46,
		0xaf,
		0x12,
		0x12,
		0x01,
		0x42,
		0x8f,
		0x12,
		0x12,
		0x02,
		0x46,
		0xaf,
		0x13,
		0x12,
		0x01,
		0x42,
		0x8f,
		0x13,
		0x12,
		0x01,
		0xee,
		0xaf,
		0x14,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x12,
		0x01,
		0x61,
		0x7b,
		0x30,
		0x12,
		0x01,
		0x5a,
		0x8f,
		0x14,
		0x12,
		0x01,
		0xee,
		0xaf,
		0x15,
		0xfc,
		0xfd,
		0xfe,
		0x12,
		0x06,
		0x93,
		0x12,
		0x01,
		0x61,
		0xe4,
		0x7b,
		0x30,
		0x12,
		0x01,
		0x5b,
		0x8f,
		0x15,
		0xc3,
		0xe5,
		0x13,
		0x95,
		0x12,
		0xff,
		0x0f,
		0xef,
		0xc3,
		0x13,
		0xff,
		0xc3,
		0x94,
		0x04,
		0x50,
		0x27,
		0xe5,
		0x12,
		0x9f,
		0x40,
		0x06,
		0xe5,
		0x12,
		0x9f,
		0xfe,
		0x80,
		0x02,
		0x7e,
		0x00,
		0x8e,
		0x12,
		0xef,
		0xfd,
		0xe5,
		0x13,
		0x2d,
		0xfd,
		0xe4,
		0x33,
		0xfc,
		0xc3,
		0xed,
		0x95,
		0x0f,
		0xec,
		0x95,
		0x0e,
		0x50,
		0x02,
		0x80,
		0x02,
		0xad,
		0x0f,
		0x8d,
		0x13,
		0xc3,
		0xe5,
		0x15,
		0x95,
		0x14,
		0xff,
		0xc3,
		0x94,
		0x04,
		0x50,
		0x29,
		0xe5,
		0x14,
		0x9f,
		0x40,
		0x06,
		0xe5,
		0x14,
		0x9f,
		0xfe,
		0x80,
		0x02,
		0x7e,
		0x00,
		0x8e,
		0x14,
		0xef,
		0xfd,
		0xe5,
		0x15,
		0x2d,
		0xfd,
		0xe4,
		0x33,
		0xfc,
		0xc3,
		0xed,
		0x95,
		0x11,
		0xec,
		0x95,
	},
	{
		0x86,
		0x00,
		0x10,
		0x50,
		0x04,
		0xaf,
		0x05,
		0x80,
		0x02,
		0xaf,
		0x11,
		0x8f,
		0x15,
		0xe5,
		0x15,
		0xd3,
		0x95,
		0x17,
		0x40,
		0x04,
		0xaf,
		0x15,
		0x80,
		0x02,
		0xaf,
		0x17,
		0x8f,
		0x17,
		0xd3,
		0xe5,
		0x16,
		0x64,
		0x80,
		0x94,
		0x80,
		0x40,
		0x04,
		0xaf,
		0x15,
		0x80,
		0x02,
		0xaf,
		0x17,
		0x8f,
		0x15,
		0xe5,
		0x16,
		0xfd,
		0x33,
		0x95,
		0xe0,
		0xfc,
		0xed,
		0xae,
		0x04,
		0x78,
		0x02,
		0xc3,
		0x33,
		0xce,
		0x33,
		0xce,
		0xd8,
		0xf9,
		0xff,
		0x24,
		0x01,
		0xfb,
		0xee,
		0x34,
		0x60,
		0x8b,
		0x82,
		0xf5,
		0x83,
		0xe5,
		0x12,
		0xf0,
		0xef,
		0x24,
		0x02,
		0xff,
		0xee,
		0x34,
		0x60,
		0x8f,
		0x82,
		0xf5,
		0x83,
		0xe5,
		0x14,
		0xf0,
		0xed,
		0xae,
		0x04,
		0x78,
		0x02,
		0xc3,
		0x33,
		0xce,
		0x33,
		0xce,
		0xd8,
		0xf9,
		0xff,
		0x24,
		0x03,
		0xfd,
		0xee,
		0x34,
		0x60,
		0x8d,
		0x82,
		0xf5,
		0x83,
		0xe5,
		0x13,
		0xf0,
		0xef,
		0x24,
		0x04,
		0xff,
		0xee,
		0x34,
		0x60,
		0x8f,
		0x82,
		0xf5,
		0x83,
		0xe5,
		0x15,
		0xf0,
		0x15,
		0x16,
		0xc3,
		0xe5,
		0x16,
		0x64,
		0x80,
		0x94,
		0x80,
		0x40,
		0x03,
		0x02,
		0x05,
		0x3d,
		0xc2,
		0x30,
		0x22,
		0xe8,
		0x8f,
		0xf0,
		0xa4,
		0xcc,
		0x8b,
		0xf0,
		0xa4,
		0x2c,
		0xfc,
		0xe9,
		0x8e,
		0xf0,
		0xa4,
		0x2c,
		0xfc,
		0x8a,
		0xf0,
		0xed,
		0xa4,
		0x2c,
		0xfc,
		0xea,
		0x8e,
		0xf0,
		0xa4,
		0xcd,
		0xa8,
		0xf0,
		0x8b,
		0xf0,
		0xa4,
		0x2d,
		0xcc,
		0x38,
		0x25,
		0xf0,
		0xfd,
		0xe9,
		0x8f,
		0xf0,
		0xa4,
		0x2c,
		0xcd,
		0x35,
		0xf0,
		0xfc,
		0xeb,
		0x8e,
		0xf0,
		0xa4,
		0xfe,
		0xa9,
		0xf0,
		0xeb,
		0x8f,
		0xf0,
		0xa4,
		0xcf,
		0xc5,
		0xf0,
		0x2e,
		0xcd,
		0x39,
		0xfe,
		0xe4,
		0x3c,
		0xfc,
		0xea,
		0xa4,
		0x2d,
		0xce,
		0x35,
		0xf0,
		0xfd,
		0xe4,
		0x3c,
		0xfc,
		0x22,
		0x75,
		0xf0,
		0x08,
		0x75,
		0x82,
		0x00,
		0xef,
		0x2f,
		0xff,
		0xee,
		0x33,
		0xfe,
		0xcd,
		0x33,
		0xcd,
		0xcc,
		0x33,
		0xcc,
		0xc5,
		0x82,
		0x33,
		0xc5,
		0x82,
		0x9b,
		0xed,
		0x9a,
		0xec,
		0x99,
		0xe5,
		0x82,
	},
	{
		0x87,
		0x00,
		0x98,
		0x40,
		0x0c,
		0xf5,
		0x82,
		0xee,
		0x9b,
		0xfe,
		0xed,
		0x9a,
		0xfd,
		0xec,
		0x99,
		0xfc,
		0x0f,
		0xd5,
		0xf0,
		0xd6,
		0xe4,
		0xce,
		0xfb,
		0xe4,
		0xcd,
		0xfa,
		0xe4,
		0xcc,
		0xf9,
		0xa8,
		0x82,
		0x22,
		0xb8,
		0x00,
		0xc1,
		0xb9,
		0x00,
		0x59,
		0xba,
		0x00,
		0x2d,
		0xec,
		0x8b,
		0xf0,
		0x84,
		0xcf,
		0xce,
		0xcd,
		0xfc,
		0xe5,
		0xf0,
		0xcb,
		0xf9,
		0x78,
		0x18,
		0xef,
		0x2f,
		0xff,
		0xee,
		0x33,
		0xfe,
		0xed,
		0x33,
		0xfd,
		0xec,
		0x33,
		0xfc,
		0xeb,
		0x33,
		0xfb,
		0x10,
		0xd7,
		0x03,
		0x99,
		0x40,
		0x04,
		0xeb,
		0x99,
		0xfb,
		0x0f,
		0xd8,
		0xe5,
		0xe4,
		0xf9,
		0xfa,
		0x22,
		0x78,
		0x18,
		0xef,
		0x2f,
		0xff,
		0xee,
		0x33,
		0xfe,
		0xed,
		0x33,
		0xfd,
		0xec,
		0x33,
		0xfc,
		0xc9,
		0x33,
		0xc9,
		0x10,
		0xd7,
		0x05,
		0x9b,
		0xe9,
		0x9a,
		0x40,
		0x07,
		0xec,
		0x9b,
		0xfc,
		0xe9,
		0x9a,
		0xf9,
		0x0f,
		0xd8,
		0xe0,
		0xe4,
		0xc9,
		0xfa,
		0xe4,
		0xcc,
		0xfb,
		0x22,
		0x75,
		0xf0,
		0x10,
		0xef,
		0x2f,
		0xff,
		0xee,
		0x33,
		0xfe,
		0xed,
		0x33,
		0xfd,
		0xcc,
		0x33,
		0xcc,
		0xc8,
		0x33,
		0xc8,
		0x10,
		0xd7,
		0x07,
		0x9b,
		0xec,
		0x9a,
		0xe8,
		0x99,
		0x40,
		0x0a,
		0xed,
		0x9b,
		0xfd,
		0xec,
		0x9a,
		0xfc,
		0xe8,
		0x99,
		0xf8,
		0x0f,
		0xd5,
		0xf0,
		0xda,
		0xe4,
		0xcd,
		0xfb,
		0xe4,
		0xcc,
		0xfa,
		0xe4,
		0xc8,
		0xf9,
		0x22,
		0xe8,
		0x60,
		0x0f,
		0xec,
		0xc3,
		0x13,
		0xfc,
		0xed,
		0x13,
		0xfd,
		0xee,
		0x13,
		0xfe,
		0xef,
		0x13,
		0xff,
		0xd8,
		0xf1,
		0x22,
		0xe4,
		0x93,
		0xfc,
		0x74,
		0x01,
		0x93,
		0xfd,
		0x74,
		0x02,
		0x93,
		0xfe,
		0x74,
		0x03,
		0x93,
		0xff,
		0x22,
		0xec,
		0xf6,
		0x08,
		0xed,
		0xf6,
		0x08,
		0xee,
		0xf6,
		0x08,
		0xef,
		0xf6,
		0x22,
		0xa4,
		0x25,
		0x82,
		0xf5,
		0x82,
		0xe5,
		0xf0,
		0x35,
		0x83,
		0xf5,
		0x83,
		0x22,
		0xd0,
		0x83,
		0xd0,
		0x82,
		0xf8,
		0xe4,
		0x93,
		0x70,
		0x12,
		0x74,
		0x01,
		0x93,
		0x70,
		0x0d,
		0xa3,
		0xa3,
		0x93,
		0xf8,
		0x74,
		0x01,
		0x93,
	},
	{
		0x88,
		0x00,
		0xf5,
		0x82,
		0x88,
		0x83,
		0xe4,
		0x73,
		0x74,
		0x02,
		0x93,
		0x68,
		0x60,
		0xef,
		0xa3,
		0xa3,
		0xa3,
		0x80,
		0xdf,
		0x75,
		0x0f,
		0x0a,
		0xa2,
		0xaf,
		0x92,
		0x32,
		0xc2,
		0xaf,
		0xc2,
		0x33,
		0x12,
		0x01,
		0x6a,
		0x12,
		0x01,
		0xe5,
		0x12,
		0x01,
		0x6a,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xc2,
		0x29,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xc2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0x75,
		0x10,
		0x18,
		0x12,
		0x09,
		0x51,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x12,
		0x02,
		0x02,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xc2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0x30,
		0x11,
		0x03,
		0x02,
		0x09,
		0x08,
		0x12,
		0x01,
		0x6a,
		0x12,
		0x01,
		0xe5,
		0xe5,
		0x0d,
		0xf5,
		0x10,
		0x12,
		0x09,
		0x51,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x12,
		0x02,
		0x02,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xc2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0x30,
		0x11,
		0x04,
		0x15,
		0x0f,
		0x80,
		0x45,
		0x12,
		0x01,
		0x6a,
		0x12,
		0x01,
		0xe5,
		0x85,
		0x0e,
		0x10,
		0x12,
		0x14,
		0xf2,
		0xc2,
		0x09,
		0x12,
		0x01,
		0xe7,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x12,
		0x02,
		0x02,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xc2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xaf,
		0x51,
	},
	{
		0x89,
		0x00,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0x30,
		0x11,
		0x06,
		0x15,
		0x0f,
		0xd2,
		0x33,
		0x80,
		0x03,
		0xe4,
		0xf5,
		0x0f,
		0x12,
		0x01,
		0x6a,
		0x12,
		0x01,
		0xe5,
		0xc2,
		0x29,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0x12,
		0x01,
		0x6a,
		0x75,
		0x51,
		0x05,
		0xaf,
		0x51,
		0x15,
		0x51,
		0xef,
		0x70,
		0xf9,
		0xa2,
		0x32,
		0x92,
		0xaf,
		0xe5,
		0x0f,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x03,
		0x02,
		0x08,
		0x18,
		0x22,
		0x12,
		0x14,
		0xf2,
		0xc2,
		0x09,
		0x90,
		0x30,
		0x18,
		0xe5,
		0x21,
		0xf0,
		0x22,
		0xe5,
		0x33,
		0x70,
		0x03,
		0x02,
		0x0a,
		0x97,
		0xc2,
		0xaf,
		0xaf,
		0x33,
		0xe4,
		0xf5,
		0x33,
		0xd2,
		0xaf,
		0x90,
		0x30,
		0x25,
		0xe0,
		0xf5,
		0x7e,
		0x90,
		0x50,
		0x82,
		0xe0,
		0xf5,
		0x66,
		0xa3,
		0xe0,
		0xf5,
		0x67,
		0xa3,
		0xe0,
		0xf5,
		0x68,
		0xa3,
		0xe0,
		0xf5,
		0x69,
		0xef,
		0x12,
		0x07,
		0xeb,
		0x09,
		0xb1,
		0x03,
		0x09,
		0xcf,
		0x04,
		0x09,
		0xf0,
		0x05,
		0x0a,
		0x1a,
		0x06,
		0x0a,
		0x08,
		0x08,
		0x0a,
		0x27,
		0x10,
		0x0a,
		0x3b,
		0x12,
		0x0a,
		0x40,
		0x20,
		0x0a,
		0x4e,
		0x21,
		0x0a,
		0x53,
		0x30,
		0x0a,
		0x7c,
		0x50,
		0x0a,
		0x5e,
		0xd8,
		0x00,
		0x00,
		0x0a,
		0x89,
		0x20,
		0x05,
		0x03,
		0x02,
		0x0a,
		0x89,
		0x30,
		0x00,
		0x03,
		0x02,
		0x0a,
		0x89,
		0x20,
		0x07,
		0x06,
		0x20,
		0x06,
		0x03,
		0xc3,
		0x80,
		0x01,
		0xd3,
		0x92,
		0x36,
		0xd2,
		0x07,
		0xc2,
		0x06,
		0x80,
		0x1c,
		0x20,
		0x05,
		0x03,
		0x02,
		0x0a,
		0x89,
		0x30,
		0x00,
		0x03,
		0x02,
		0x0a,
		0x89,
		0x20,
		0x07,
		0x06,
		0x20,
		0x06,
		0x03,
		0xc3,
		0x80,
		0x01,
		0xd3,
		0x92,
		0x36,
		0xd2,
		0x07,
		0xd2,
		0x06,
		0x12,
		0x02,
		0x5d,
		0x80,
		0x24,
		0x20,
		0x05,
		0x03,
		0x02,
		0x0a,
		0x89,
		0x30,
		0x00,
		0x03,
		0x02,
		0x0a,
		0x89,
		0xc2,
		0x07,
		0xd2,
		0x06,
	},
	{
		0x8a,
		0x00,
		0x12,
		0x02,
		0x78,
		0xc2,
		0x04,
		0x02,
		0x0a,
		0x89,
		0x12,
		0x02,
		0x24,
		0x30,
		0x05,
		0x06,
		0xe4,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0xc2,
		0x31,
		0xd2,
		0x34,
		0x80,
		0x6f,
		0x30,
		0x07,
		0x6c,
		0x30,
		0x06,
		0x69,
		0x12,
		0x02,
		0x5d,
		0xd2,
		0x31,
		0x80,
		0x62,
		0x20,
		0x07,
		0x03,
		0x30,
		0x06,
		0x09,
		0xe5,
		0x7e,
		0x64,
		0x0e,
		0x70,
		0x56,
		0x20,
		0x00,
		0x53,
		0x12,
		0x0a,
		0x98,
		0x80,
		0x4e,
		0x12,
		0x04,
		0xf3,
		0x80,
		0x49,
		0x30,
		0x05,
		0x46,
		0x20,
		0x07,
		0x43,
		0x20,
		0x06,
		0x40,
		0x12,
		0x15,
		0xc5,
		0x80,
		0x3b,
		0x12,
		0x11,
		0x1c,
		0x80,
		0x36,
		0x20,
		0x07,
		0x33,
		0x20,
		0x06,
		0x30,
		0x12,
		0x16,
		0x30,
		0x80,
		0x2b,
		0xe5,
		0x7e,
		0x64,
		0x01,
		0x70,
		0x25,
		0xd2,
		0x35,
		0x90,
		0x50,
		0x82,
		0xe5,
		0x74,
		0xf0,
		0xa3,
		0xe5,
		0x75,
		0xf0,
		0xa3,
		0xe5,
		0x76,
		0xf0,
		0xa3,
		0xe5,
		0x77,
		0xf0,
		0xc2,
		0x35,
		0x80,
		0x0d,
		0x90,
		0x50,
		0x82,
		0x30,
		0x33,
		0x05,
		0x74,
		0x55,
		0xf0,
		0x80,
		0x02,
		0xe4,
		0xf0,
		0x20,
		0x07,
		0x06,
		0x30,
		0x06,
		0x03,
		0x30,
		0x04,
		0x05,
		0x90,
		0x30,
		0x25,
		0xe4,
		0xf0,
		0x22,
		0xe5,
		0x7e,
		0x12,
		0x07,
		0xeb,
		0x0a,
		0xdd,
		0x01,
		0x0a,
		0xf1,
		0x02,
		0x0a,
		0xfd,
		0x03,
		0x0b,
		0x0e,
		0x0b,
		0x0b,
		0x22,
		0x0c,
		0x0b,
		0x33,
		0x0d,
		0x0b,
		0x47,
		0x0e,
		0x0b,
		0x76,
		0x80,
		0x0b,
		0x7f,
		0x81,
		0x0b,
		0x84,
		0x82,
		0x0b,
		0x94,
		0x90,
		0x0b,
		0x94,
		0x91,
		0x0b,
		0x94,
		0x92,
		0x0b,
		0x94,
		0x93,
		0x0b,
		0x94,
		0x94,
		0x0b,
		0xa9,
		0xa0,
		0x0b,
		0xa9,
		0xa1,
		0x0b,
		0xa9,
		0xa2,
		0x0b,
		0xa9,
		0xa3,
		0x0b,
		0xa9,
		0xa4,
		0x00,
		0x00,
		0x0b,
		0xb5,
		0xe4,
		0xf5,
		0x0a,
		0x12,
		0x10,
		0xb2,
		0x75,
		0x79,
		0x02,
		0x12,
		0x0b,
		0xc7,
		0xd2,
		0x18,
		0xd2,
		0x19,
		0xc2,
		0x3c,
		0x80,
		0x14,
		0x75,
		0x0a,
		0x01,
		0x12,
		0x0b,
		0xb6,
		0xc2,
		0x3c,
		0xd2,
		0x3b,
		0x80,
		0x0a,
		0x75,
		0x0a,
		0x02,
	},
	{
		0x8b,
		0x00,
		0x12,
		0x0b,
		0xb6,
		0xd2,
		0x3c,
		0xc2,
		0x3b,
		0xd2,
		0x1a,
		0xd2,
		0x38,
		0xd2,
		0x30,
		0x22,
		0x30,
		0x38,
		0x0c,
		0x20,
		0x3c,
		0x03,
		0x30,
		0x3b,
		0x06,
		0xe4,
		0xf5,
		0x0a,
		0x12,
		0x13,
		0xc8,
		0xd2,
		0x18,
		0xc2,
		0x19,
		0x22,
		0x30,
		0x38,
		0x09,
		0x20,
		0x3c,
		0x03,
		0x30,
		0x3b,
		0x03,
		0x12,
		0x0b,
		0xc7,
		0xd2,
		0x18,
		0xd2,
		0x19,
		0x22,
		0x30,
		0x38,
		0x0c,
		0x20,
		0x3c,
		0x03,
		0x30,
		0x3b,
		0x06,
		0x75,
		0x0a,
		0x02,
		0x12,
		0x13,
		0xc8,
		0xc2,
		0x18,
		0xd2,
		0x19,
		0x22,
		0x20,
		0x3c,
		0x03,
		0x30,
		0x3b,
		0x68,
		0xe5,
		0x79,
		0xd3,
		0x94,
		0x03,
		0x40,
		0x04,
		0x7f,
		0x00,
		0x80,
		0x04,
		0xe5,
		0x79,
		0x04,
		0xff,
		0x8f,
		0x79,
		0x30,
		0x18,
		0x09,
		0x30,
		0x19,
		0x06,
		0x75,
		0x0a,
		0x01,
		0x12,
		0x13,
		0xc8,
		0x20,
		0x18,
		0x48,
		0x30,
		0x19,
		0x45,
		0x75,
		0x0a,
		0x02,
		0x02,
		0x13,
		0xc8,
		0xc2,
		0x38,
		0xc2,
		0x1a,
		0xc2,
		0x18,
		0xc2,
		0x19,
		0x22,
		0xd2,
		0x1a,
		0xd2,
		0x30,
		0x22,
		0xe5,
		0x69,
		0xd3,
		0x94,
		0x04,
		0x40,
		0x04,
		0xaf,
		0x79,
		0x80,
		0x02,
		0xaf,
		0x69,
		0x8f,
		0x79,
		0x22,
		0x20,
		0x1a,
		0x1e,
		0xe5,
		0x7e,
		0x24,
		0x70,
		0xf9,
		0x12,
		0x01,
		0xcf,
		0x25,
		0xe0,
		0x25,
		0xe0,
		0x24,
		0x34,
		0xf8,
		0x02,
		0x07,
		0xd3,
		0x20,
		0x1a,
		0x09,
		0xe5,
		0x7e,
		0x24,
		0x60,
		0x24,
		0x2c,
		0xf8,
		0xa6,
		0x69,
		0x22,
		0x12,
		0x10,
		0xb2,
		0x75,
		0x79,
		0x02,
		0xe4,
		0xf5,
		0x0a,
		0x12,
		0x13,
		0xc8,
		0xd2,
		0x18,
		0xc2,
		0x19,
		0x22,
		0x75,
		0x0a,
		0x01,
		0x12,
		0x13,
		0xc8,
		0x22,
		0xc0,
		0xe0,
		0xc0,
		0xf0,
		0xc0,
		0x83,
		0xc0,
		0x82,
		0xc0,
		0xd0,
		0x75,
		0xd0,
		0x00,
		0xc0,
		0x00,
		0xc0,
		0x01,
		0xc0,
		0x02,
		0xc0,
		0x03,
		0xc0,
		0x04,
		0xc0,
		0x05,
		0xc0,
		0x06,
		0xc0,
		0x07,
		0x90,
		0x3f,
		0x0c,
		0xe0,
		0xf5,
		0x08,
		0xe5,
		0x08,
		0x20,
		0xe3,
		0x03,
		0x02,
		0x0c,
		0x7f,
		0x30,
		0x35,
		0x03,
		0x02,
		0x0c,
		0x7f,
		0x90,
	},
	{
		0x8c,
		0x00,
		0x60,
		0x16,
		0xe0,
		0xf5,
		0x6a,
		0xa3,
		0xe0,
		0xf5,
		0x6b,
		0x90,
		0x60,
		0x1e,
		0xe0,
		0xf5,
		0x6c,
		0xa3,
		0xe0,
		0xf5,
		0x6d,
		0x90,
		0x60,
		0x26,
		0xe0,
		0xf5,
		0x6e,
		0xa3,
		0xe0,
		0xf5,
		0x6f,
		0x90,
		0x60,
		0x2e,
		0xe0,
		0xf5,
		0x70,
		0xa3,
		0xe0,
		0xf5,
		0x71,
		0x90,
		0x60,
		0x36,
		0x12,
		0x00,
		0x16,
		0x12,
		0x01,
		0xa5,
		0x40,
		0x06,
		0x75,
		0x2a,
		0xff,
		0x75,
		0x2b,
		0xff,
		0x85,
		0x2a,
		0x74,
		0x85,
		0x2b,
		0x75,
		0x90,
		0x60,
		0x1a,
		0xe0,
		0xf5,
		0x6a,
		0xa3,
		0xe0,
		0xf5,
		0x6b,
		0x90,
		0x60,
		0x22,
		0xe0,
		0xf5,
		0x6c,
		0xa3,
		0xe0,
		0xf5,
		0x6d,
		0x90,
		0x60,
		0x2a,
		0xe0,
		0xf5,
		0x6e,
		0xa3,
		0xe0,
		0xf5,
		0x6f,
		0x90,
		0x60,
		0x32,
		0xe0,
		0xf5,
		0x70,
		0xa3,
		0xe0,
		0xf5,
		0x71,
		0x90,
		0x60,
		0x3a,
		0x12,
		0x00,
		0x16,
		0x12,
		0x01,
		0xa5,
		0x40,
		0x06,
		0x75,
		0x2a,
		0xff,
		0x75,
		0x2b,
		0xff,
		0x85,
		0x2a,
		0x76,
		0x85,
		0x2b,
		0x77,
		0xd2,
		0x3d,
		0xe5,
		0x08,
		0x30,
		0xe5,
		0x41,
		0x90,
		0x56,
		0x90,
		0xe0,
		0xf5,
		0x55,
		0xe5,
		0x7b,
		0x12,
		0x01,
		0x9e,
		0xad,
		0x55,
		0xc3,
		0xef,
		0x9d,
		0x74,
		0x80,
		0xf8,
		0x6e,
		0x98,
		0x50,
		0x02,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x27,
		0xaf,
		0x55,
		0xef,
		0x24,
		0x01,
		0xff,
		0xe4,
		0x33,
		0xfe,
		0xc3,
		0xef,
		0x95,
		0x7b,
		0x74,
		0x80,
		0xf8,
		0x6e,
		0x98,
		0x50,
		0x02,
		0x80,
		0x02,
		0xa2,
		0x27,
		0x92,
		0x27,
		0x30,
		0x27,
		0x04,
		0xaf,
		0x55,
		0x80,
		0x02,
		0xaf,
		0x7b,
		0x8f,
		0x7b,
		0xe5,
		0x08,
		0x30,
		0xe1,
		0x08,
		0x90,
		0x30,
		0x24,
		0xe0,
		0xf5,
		0x33,
		0xe4,
		0xf0,
		0x90,
		0x3f,
		0x0c,
		0xe5,
		0x08,
		0xf0,
		0xd0,
		0x07,
		0xd0,
		0x06,
		0xd0,
		0x05,
		0xd0,
		0x04,
		0xd0,
		0x03,
		0xd0,
		0x02,
		0xd0,
		0x01,
		0xd0,
		0x00,
		0xd0,
		0xd0,
		0xd0,
		0x82,
		0xd0,
		0x83,
		0xd0,
		0xf0,
		0xd0,
		0xe0,
		0x32,
		0x30,
		0x04,
		0x03,
		0x02,
		0x0d,
		0xd3,
		0xd2,
		0x04,
		0xe5,
		0x7e,
		0xb4,
		0x01,
		0x06,
	},
	{
		0x8d,
		0x00,
		0x12,
		0x15,
		0xf2,
		0x02,
		0x0d,
		0xcc,
		0xe5,
		0x7e,
		0xb4,
		0x02,
		0x06,
		0x12,
		0x16,
		0x03,
		0x02,
		0x0d,
		0xcc,
		0xe5,
		0x7e,
		0xb4,
		0x03,
		0x05,
		0xe4,
		0xf5,
		0x0c,
		0x80,
		0x08,
		0xe5,
		0x7e,
		0xb4,
		0x04,
		0x09,
		0x85,
		0x7c,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0x02,
		0x0d,
		0xcc,
		0xe5,
		0x7e,
		0x64,
		0x0f,
		0x70,
		0x15,
		0x12,
		0x02,
		0x89,
		0x40,
		0x06,
		0x7e,
		0x03,
		0x7f,
		0xff,
		0x80,
		0x04,
		0xae,
		0x68,
		0xaf,
		0x69,
		0x12,
		0x0d,
		0xd4,
		0x02,
		0x0d,
		0xcc,
		0xe5,
		0x7e,
		0x64,
		0x10,
		0x60,
		0x03,
		0x02,
		0x0d,
		0xcc,
		0xf5,
		0x66,
		0xf5,
		0x67,
		0xf5,
		0x68,
		0xab,
		0x69,
		0xaa,
		0x68,
		0xa9,
		0x67,
		0xa8,
		0x66,
		0x12,
		0x01,
		0x73,
		0xfe,
		0xe4,
		0xfc,
		0xfd,
		0x12,
		0x01,
		0xc4,
		0xe4,
		0x7b,
		0xff,
		0xfa,
		0xf9,
		0xf8,
		0x12,
		0x02,
		0x3a,
		0x12,
		0x02,
		0x0c,
		0xe4,
		0x93,
		0xfe,
		0x74,
		0x01,
		0x93,
		0xff,
		0xe4,
		0xfc,
		0xfd,
		0xe5,
		0x69,
		0x2f,
		0xf5,
		0x69,
		0xe5,
		0x68,
		0x3e,
		0xf5,
		0x68,
		0xed,
		0x35,
		0x67,
		0xf5,
		0x67,
		0xec,
		0x35,
		0x66,
		0xf5,
		0x66,
		0x12,
		0x02,
		0x89,
		0x40,
		0x06,
		0x7e,
		0x03,
		0x7f,
		0xff,
		0x80,
		0x04,
		0xae,
		0x68,
		0xaf,
		0x69,
		0x12,
		0x0d,
		0xd4,
		0xe4,
		0xf5,
		0x67,
		0xf5,
		0x67,
		0xe5,
		0x67,
		0xd3,
		0x95,
		0x7c,
		0x50,
		0x1c,
		0x12,
		0x02,
		0x0c,
		0xaf,
		0x67,
		0x75,
		0xf0,
		0x02,
		0xef,
		0x12,
		0x07,
		0xdf,
		0xc3,
		0x74,
		0x01,
		0x93,
		0x95,
		0x69,
		0xe4,
		0x93,
		0x95,
		0x68,
		0x50,
		0x04,
		0x05,
		0x67,
		0x80,
		0xdd,
		0x85,
		0x67,
		0x7d,
		0x90,
		0x30,
		0x25,
		0xe4,
		0xf0,
		0xd2,
		0x34,
		0x22,
		0x8e,
		0x68,
		0x8f,
		0x69,
		0x85,
		0x68,
		0x64,
		0x85,
		0x69,
		0x65,
		0xe5,
		0x69,
		0xc4,
		0xf8,
		0x54,
		0x0f,
		0xc8,
		0x68,
		0xf5,
		0x69,
		0xe5,
		0x68,
		0xc4,
		0x54,
		0xf0,
		0x48,
		0xf5,
		0x68,
		0x85,
		0x68,
		0x0d,
		0x85,
		0x69,
		0x0e,
		0x12,
		0x08,
		0x11,
		0x22,
		0x12,
		0x02,
		0x52,
		0xb5,
		0x07,
		0x03,
	},
	{
		0x8e,
		0x00,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x40,
		0x03,
		0x02,
		0x0e,
		0x99,
		0x90,
		0x30,
		0x04,
		0xe0,
		0x44,
		0x20,
		0xf0,
		0xa3,
		0xe0,
		0x44,
		0x40,
		0xf0,
		0x90,
		0x50,
		0x25,
		0xe0,
		0x44,
		0x04,
		0xf0,
		0x90,
		0x50,
		0x03,
		0xe0,
		0x54,
		0xfd,
		0xf0,
		0x90,
		0x50,
		0x27,
		0xe0,
		0x44,
		0x01,
		0xf0,
		0x90,
		0x50,
		0x31,
		0xe4,
		0xf0,
		0x90,
		0x50,
		0x33,
		0xf0,
		0x90,
		0x30,
		0x1e,
		0x12,
		0x01,
		0xd8,
		0x90,
		0x30,
		0x18,
		0x12,
		0x01,
		0xd8,
		0x90,
		0x30,
		0x1b,
		0x12,
		0x01,
		0xd8,
		0xe0,
		0xf5,
		0x25,
		0x90,
		0x30,
		0x18,
		0xe0,
		0xf5,
		0x21,
		0x90,
		0x60,
		0x00,
		0x74,
		0xf5,
		0xf0,
		0x90,
		0x3f,
		0x01,
		0xe4,
		0xf0,
		0xa3,
		0xf0,
		0x90,
		0x3f,
		0x01,
		0xe0,
		0x44,
		0x08,
		0xf0,
		0xe0,
		0x44,
		0x20,
		0xf0,
		0x90,
		0x3f,
		0x05,
		0x74,
		0x30,
		0xf0,
		0xa3,
		0x74,
		0x24,
		0xf0,
		0x90,
		0x3f,
		0x0b,
		0xe0,
		0x44,
		0x0f,
		0xf0,
		0x90,
		0x3f,
		0x01,
		0xe0,
		0x44,
		0x02,
		0xf0,
		0xc2,
		0x8c,
		0x75,
		0x89,
		0x03,
		0x75,
		0xa8,
		0x07,
		0x75,
		0xb8,
		0x04,
		0xe4,
		0xf5,
		0xd8,
		0xf5,
		0xe8,
		0x90,
		0x30,
		0x01,
		0xe0,
		0x44,
		0x40,
		0xf0,
		0xe0,
		0x54,
		0xbf,
		0xf0,
		0x22,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x20,
		0x06,
		0xae,
		0x56,
		0xaf,
		0x57,
		0x80,
		0x04,
		0xae,
		0x6a,
		0xaf,
		0x6b,
		0x8e,
		0x56,
		0x8f,
		0x57,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x21,
		0x06,
		0xae,
		0x58,
		0xaf,
		0x59,
		0x80,
		0x04,
		0xae,
		0x6c,
		0xaf,
		0x6d,
		0x8e,
		0x58,
		0x8f,
		0x59,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x22,
		0x06,
		0xae,
		0x5a,
		0xaf,
		0x5b,
		0x80,
		0x04,
		0xae,
		0x6e,
		0xaf,
		0x6f,
		0x8e,
		0x5a,
		0x8f,
		0x5b,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x23,
		0x06,
		0xae,
		0x5c,
		0xaf,
		0x5d,
		0x80,
		0x04,
		0xae,
		0x70,
		0xaf,
		0x71,
		0x8e,
		0x5c,
		0x8f,
		0x5d,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x24,
		0x06,
		0xae,
		0x5e,
		0xaf,
		0x5f,
		0x80,
		0x04,
		0xae,
		0x72,
		0xaf,
		0x73,
		0x8e,
		0x5e,
		0x8f,
		0x5f,
		0x30,
		0x3e,
	},
	{
		0x8f,
		0x00,
		0x09,
		0x30,
		0x25,
		0x06,
		0xae,
		0x60,
		0xaf,
		0x61,
		0x80,
		0x04,
		0xae,
		0x74,
		0xaf,
		0x75,
		0x8e,
		0x60,
		0x8f,
		0x61,
		0x30,
		0x3e,
		0x09,
		0x30,
		0x26,
		0x06,
		0xae,
		0x62,
		0xaf,
		0x63,
		0x80,
		0x04,
		0xae,
		0x76,
		0xaf,
		0x77,
		0x8e,
		0x62,
		0x8f,
		0x63,
		0x22,
		0x56,
		0x0d,
		0x03,
		0x00,
		0x00,
		0x00,
		0xc8,
		0x01,
		0x2c,
		0x01,
		0x5e,
		0x01,
		0x8b,
		0x01,
		0xb8,
		0x01,
		0xe5,
		0x02,
		0x12,
		0x02,
		0x3f,
		0x02,
		0x6c,
		0x02,
		0x99,
		0x02,
		0xc6,
		0x02,
		0xf3,
		0x03,
		0x20,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x09,
		0x00,
		0x02,
		0x12,
		0x02,
		0x30,
		0x02,
		0x4e,
		0x02,
		0x6c,
		0x02,
		0x8a,
		0x02,
		0xa8,
		0x02,
		0xc6,
		0x02,
		0xe4,
		0x03,
		0x02,
		0x03,
		0x20,
		0x07,
		0x20,
		0x12,
		0x28,
		0x1e,
		0x18,
		0x18,
		0x28,
		0x1e,
		0x18,
		0x12,
		0x28,
		0x1e,
		0x18,
		0x12,
		0x28,
		0x18,
		0x18,
		0x12,
		0x20,
		0x18,
		0x28,
		0x1c,
		0x30,
		0x24,
		0x10,
		0x1c,
		0x18,
		0x24,
		0x1c,
		0x14,
		0x24,
		0x1c,
		0x28,
		0x0c,
		0x30,
		0x14,
		0x10,
		0x0c,
		0x18,
		0x14,
		0x1c,
		0x20,
		0x24,
		0x28,
		0x0c,
		0x14,
		0x14,
		0x1c,
		0x1c,
		0x14,
		0x24,
		0x1c,
		0x2c,
		0x14,
		0x34,
		0x1c,
		0x1c,
		0x08,
		0x24,
		0x10,
		0x19,
		0x19,
		0x1c,
		0x19,
		0x19,
		0x10,
		0x10,
		0x10,
		0x10,
		0x10,
		0x00,
		0x00,
		0x00,
		0x00,
		0x00,
		0x30,
		0x36,
		0x4e,
		0x12,
		0x10,
		0x2c,
		0xe5,
		0x32,
		0xd3,
		0x95,
		0x7d,
		0x40,
		0x0c,
		0xe5,
		0x32,
		0x04,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0xd2,
		0x37,
		0x80,
		0x44,
		0xe5,
		0x7c,
		0xb5,
		0x7d,
		0x04,
		0x7f,
		0x01,
		0x80,
		0x02,
		0x7f,
		0x00,
		0xef,
		0x24,
		0xfb,
		0xd3,
		0x64,
		0x80,
		0xf8,
		0xe5,
		0x7d,
		0x64,
		0x80,
		0x98,
		0x40,
		0x12,
		0xc2,
		0x37,
		0xe5,
		0x7c,
		0xb5,
		0x7d,
		0x04,
		0x7f,
		0x01,
		0x80,
		0x02,
		0x7f,
		0x00,
		0xef,
		0x24,
		0xfa,
		0x80,
		0x06,
		0xd2,
		0x37,
		0xe5,
		0x7d,
		0x24,
		0x02,
		0xf5,
		0x0c,
	},
	{
		0x90,
		0x00,
		0x12,
		0x12,
		0xb4,
		0x80,
		0x0c,
		0xe5,
		0x7d,
		0x70,
		0x0f,
		0x12,
		0x10,
		0x2c,
		0xc2,
		0x03,
		0x12,
		0x15,
		0xf2,
		0xd2,
		0x02,
		0xd2,
		0x01,
		0xd2,
		0x00,
		0x22,
		0x30,
		0x03,
		0x08,
		0xc2,
		0x03,
		0xc2,
		0x04,
		0x12,
		0x02,
		0x78,
		0x22,
		0xe4,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0xd2,
		0x03,
		0x22,
		0x12,
		0x15,
		0xde,
		0xc2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0xc2,
		0x3e,
		0x12,
		0x12,
		0x51,
		0x22,
		0xd3,
		0xe5,
		0x57,
		0x95,
		0x6b,
		0xe5,
		0x56,
		0x95,
		0x6a,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x20,
		0xd3,
		0xe5,
		0x59,
		0x95,
		0x6d,
		0xe5,
		0x58,
		0x95,
		0x6c,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x21,
		0xd3,
		0xe5,
		0x5b,
		0x95,
		0x6f,
		0xe5,
		0x5a,
		0x95,
		0x6e,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x22,
		0xd3,
		0xe5,
		0x5d,
		0x95,
		0x71,
		0xe5,
		0x5c,
		0x95,
		0x70,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x23,
		0xd3,
		0xe5,
		0x5f,
		0x95,
		0x73,
		0xe5,
		0x5e,
		0x95,
		0x72,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x24,
		0xd3,
		0xe5,
		0x61,
		0x95,
		0x75,
		0xe5,
		0x60,
		0x95,
		0x74,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x25,
		0xd3,
		0xe5,
		0x63,
		0x95,
		0x77,
		0xe5,
		0x62,
		0x95,
		0x76,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x92,
		0x26,
		0x22,
		0xe5,
		0x0a,
		0x70,
		0x04,
		0x7a,
		0x0f,
		0x7b,
		0x69,
		0xe5,
		0x0a,
		0xb4,
		0x01,
		0x04,
		0x7a,
		0x0f,
		0x7b,
		0x7d,
		0xe5,
		0x0a,
		0xb4,
		0x02,
		0x04,
		0x7a,
		0x0f,
		0x7b,
		0x91,
		0x8b,
		0x82,
		0x8a,
		0x83,
		0x12,
		0x07,
		0xc3,
		0x8f,
		0x37,
		0x8e,
		0x36,
		0x8d,
		0x35,
		0x8c,
		0x34,
		0xe5,
		0x82,
		0x24,
		0x04,
		0xf5,
		0x82,
		0xe4,
		0x35,
		0x83,
		0xf5,
		0x83,
		0x12,
		0x07,
		0xc3,
		0x8f,
		0x3b,
		0x8e,
		0x3a,
		0x8d,
		0x39,
		0x8c,
		0x38,
		0xeb,
		0x24,
		0x08,
		0x12,
		0x02,
		0x13,
		0x8f,
		0x3f,
		0x8e,
		0x3e,
		0x8d,
		0x3d,
		0x8c,
		0x3c,
		0xeb,
	},
	{
		0x91,
		0x00,
		0x24,
		0x0c,
		0x12,
		0x02,
		0x13,
		0x8f,
		0x43,
		0x8e,
		0x42,
		0x8d,
		0x41,
		0x8c,
		0x40,
		0xeb,
		0x24,
		0x10,
		0x12,
		0x02,
		0x13,
		0x8f,
		0x47,
		0x8e,
		0x46,
		0x8d,
		0x45,
		0x8c,
		0x44,
		0x22,
		0xe5,
		0x7e,
		0x64,
		0x01,
		0x70,
		0x50,
		0x12,
		0x02,
		0x0c,
		0xe5,
		0x7d,
		0x12,
		0x01,
		0x7b,
		0xfe,
		0xe4,
		0x8f,
		0x69,
		0x8e,
		0x68,
		0xf5,
		0x67,
		0xf5,
		0x66,
		0x12,
		0x01,
		0xcf,
		0x7b,
		0xff,
		0xfa,
		0xf9,
		0xf8,
		0x12,
		0x01,
		0xc4,
		0xc0,
		0x04,
		0xc0,
		0x05,
		0xc0,
		0x06,
		0xc0,
		0x07,
		0x12,
		0x01,
		0x73,
		0xab,
		0x07,
		0xfa,
		0xe4,
		0xf9,
		0xf8,
		0xd0,
		0x07,
		0xd0,
		0x06,
		0xd0,
		0x05,
		0xd0,
		0x04,
		0x12,
		0x02,
		0x3a,
		0x85,
		0x69,
		0x66,
		0x85,
		0x7d,
		0x67,
		0x12,
		0x02,
		0x0c,
		0xe5,
		0x7d,
		0x12,
		0x02,
		0x68,
		0xe4,
		0x93,
		0xf5,
		0x68,
		0x74,
		0x01,
		0x93,
		0xf5,
		0x69,
		0x90,
		0x50,
		0x82,
		0xe5,
		0x66,
		0xf0,
		0xa3,
		0xe5,
		0x67,
		0xf0,
		0xa3,
		0xe5,
		0x68,
		0xf0,
		0xa3,
		0xe5,
		0x69,
		0xf0,
		0x22,
		0x12,
		0x10,
		0x3a,
		0x30,
		0x36,
		0x40,
		0xe5,
		0x24,
		0x54,
		0x1f,
		0xff,
		0x60,
		0x0e,
		0x64,
		0x1f,
		0x60,
		0x0a,
		0xe5,
		0x7c,
		0x65,
		0x7d,
		0x60,
		0x04,
		0xe5,
		0x7d,
		0x70,
		0x18,
		0xbf,
		0x1f,
		0x02,
		0xb2,
		0x37,
		0xe5,
		0x7c,
		0xb5,
		0x7d,
		0x02,
		0xc2,
		0x37,
		0xe5,
		0x7d,
		0x70,
		0x02,
		0xd2,
		0x37,
		0xc2,
		0x02,
		0xd2,
		0x01,
		0xd2,
		0x00,
		0xc2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0xc2,
		0x3e,
		0x12,
		0x12,
		0x51,
		0x30,
		0x37,
		0x03,
		0x02,
		0x15,
		0xf2,
		0x02,
		0x16,
		0x03,
		0xd2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0xd2,
		0x3e,
		0x12,
		0x12,
		0x51,
		0x12,
		0x15,
		0xf2,
		0xe5,
		0x32,
		0xd3,
		0x95,
		0x7d,
		0x40,
		0x05,
		0xe4,
		0x95,
		0x7d,
		0x40,
		0x06,
		0xc2,
		0x02,
		0xd2,
		0x01,
		0xd2,
		0x00,
		0x22,
		0x12,
		0x10,
		0x3a,
		0xc3,
		0x30,
		0x25,
		0x0c,
		0xe5,
		0x61,
		0x95,
		0x75,
		0xff,
		0xe5,
		0x60,
		0x95,
		0x74,
		0xfe,
		0x80,
		0x0a,
		0xe5,
		0x75,
	},
	{
		0x92,
		0x00,
		0x95,
		0x61,
		0xff,
		0xe5,
		0x74,
		0x95,
		0x60,
		0xfe,
		0x8e,
		0x0a,
		0x8f,
		0x0b,
		0x30,
		0x03,
		0x26,
		0x12,
		0x02,
		0x7f,
		0x50,
		0x03,
		0x30,
		0x27,
		0x07,
		0xc2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0x80,
		0x2f,
		0x05,
		0x31,
		0xe5,
		0x31,
		0xd3,
		0x94,
		0x02,
		0x40,
		0x29,
		0xe4,
		0xf5,
		0x31,
		0xc2,
		0x03,
		0xc2,
		0x02,
		0xc2,
		0x01,
		0xd2,
		0x00,
		0xd2,
		0x34,
		0x22,
		0x12,
		0x02,
		0x7f,
		0x50,
		0x03,
		0x30,
		0x27,
		0x04,
		0x05,
		0x31,
		0x80,
		0x03,
		0xe4,
		0xf5,
		0x31,
		0xe5,
		0x31,
		0xd3,
		0x94,
		0x02,
		0x40,
		0x05,
		0xd2,
		0x03,
		0xe4,
		0xf5,
		0x31,
		0x22,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x20,
		0x04,
		0xaf,
		0x4a,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4a,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x21,
		0x04,
		0xaf,
		0x4b,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4b,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x22,
		0x04,
		0xaf,
		0x4c,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4c,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x23,
		0x04,
		0xaf,
		0x4d,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4d,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x24,
		0x04,
		0xaf,
		0x4e,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4e,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x25,
		0x04,
		0xaf,
		0x4f,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x4f,
		0x30,
		0x3e,
		0x07,
		0x30,
		0x26,
		0x04,
		0xaf,
		0x50,
		0x80,
		0x02,
		0xaf,
		0x7d,
		0x8f,
		0x50,
		0x22,
		0xe5,
		0x0c,
		0xd3,
		0x95,
		0x7c,
		0x40,
		0x01,
		0x22,
		0x12,
		0x02,
		0x0c,
		0xe5,
		0x0c,
		0x12,
		0x02,
		0x68,
		0xe4,
		0x93,
		0xfe,
		0x74,
		0x01,
		0x93,
		0xff,
		0x4e,
		0x60,
		0x21,
		0x8e,
		0x64,
		0x8f,
		0x65,
		0xef,
		0xc4,
		0xf8,
		0x54,
		0x0f,
		0xc8,
		0x68,
		0xff,
		0xee,
		0xc4,
		0x54,
		0xf0,
		0x48,
		0xfe,
		0x43,
		0x07,
		0x0d,
		0x8e,
		0x0d,
		0x8f,
		0x0e,
		0x12,
		0x08,
		0x11,
		0x30,
		0x33,
		0x22,
		0xc3,
		0x22,
		0x75,
		0x0d,
		0x00,
		0x75,
		0x0e,
		0x0d,
		0x12,
		0x08,
		0x11,
		0x30,
		0x33,
		0x02,
		0xc3,
		0x22,
		0x75,
		0x0d,
		0x00,
	},
	{
		0x93,
		0x00,
		0x75,
		0x0e,
		0x64,
		0x12,
		0x15,
		0x87,
		0x75,
		0x0d,
		0x80,
		0x75,
		0x0e,
		0x00,
		0x12,
		0x08,
		0x11,
		0x85,
		0x0c,
		0x7d,
		0xd3,
		0x22,
		0x12,
		0x10,
		0x3a,
		0xd2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0xd2,
		0x3e,
		0x12,
		0x12,
		0x51,
		0x30,
		0x36,
		0x26,
		0x30,
		0x37,
		0x06,
		0xe5,
		0x7c,
		0x65,
		0x7d,
		0x60,
		0x3e,
		0x20,
		0x37,
		0x04,
		0xe5,
		0x7d,
		0x60,
		0x37,
		0xe5,
		0x24,
		0x54,
		0x1f,
		0xff,
		0xbf,
		0x1f,
		0x06,
		0x30,
		0x25,
		0x03,
		0x20,
		0x26,
		0x29,
		0x30,
		0x37,
		0x02,
		0x80,
		0x21,
		0x02,
		0x16,
		0x03,
		0xe5,
		0x7c,
		0xd3,
		0x95,
		0x7d,
		0x40,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x50,
		0x14,
		0x20,
		0x39,
		0x0e,
		0xe5,
		0x24,
		0x54,
		0x1f,
		0xff,
		0xbf,
		0x1f,
		0x06,
		0x30,
		0x25,
		0x03,
		0x20,
		0x26,
		0x03,
		0x02,
		0x15,
		0xf2,
		0x12,
		0x02,
		0x93,
		0x22,
		0xc2,
		0x34,
		0x20,
		0x05,
		0x05,
		0x75,
		0x0a,
		0xff,
		0x80,
		0x36,
		0x20,
		0x07,
		0x08,
		0x20,
		0x06,
		0x05,
		0xe4,
		0xf5,
		0x0a,
		0x80,
		0x2b,
		0x20,
		0x07,
		0x08,
		0x30,
		0x06,
		0x05,
		0x75,
		0x0a,
		0x20,
		0x80,
		0x20,
		0x30,
		0x00,
		0x05,
		0x75,
		0x0a,
		0x01,
		0x80,
		0x18,
		0xe5,
		0x20,
		0x54,
		0x07,
		0xff,
		0xbf,
		0x06,
		0x0d,
		0x30,
		0x31,
		0x04,
		0x7f,
		0x12,
		0x80,
		0x02,
		0x7f,
		0x02,
		0x8f,
		0x0a,
		0x80,
		0x03,
		0x75,
		0x0a,
		0xfe,
		0x90,
		0x30,
		0x27,
		0xe5,
		0x0a,
		0xf0,
		0xe5,
		0x23,
		0x54,
		0xf8,
		0xf5,
		0x0a,
		0xe5,
		0x79,
		0x25,
		0x0a,
		0xf5,
		0x0a,
		0x90,
		0x30,
		0x26,
		0xe5,
		0x0a,
		0xf0,
		0x22,
		0xe5,
		0x0a,
		0x70,
		0x04,
		0x7e,
		0x0f,
		0x7f,
		0xa5,
		0xe5,
		0x0a,
		0xb4,
		0x01,
		0x04,
		0x7e,
		0x0f,
		0x7f,
		0xaa,
		0xe5,
		0x0a,
		0xb4,
		0x02,
		0x04,
		0x7e,
		0x0f,
		0x7f,
		0xaf,
		0x8f,
		0x82,
		0x8e,
		0x83,
		0xe4,
		0x93,
		0xf5,
		0x2c,
		0x74,
		0x01,
		0x93,
		0xf5,
		0x2d,
		0x74,
		0x02,
		0x93,
		0xf5,
		0x2e,
		0x74,
		0x03,
		0x93,
		0xf5,
		0x2f,
		0x74,
		0x04,
		0x93,
		0xf5,
		0x30,
		0xe5,
		0x0a,
	},
	{
		0x94,
		0x00,
		0xb4,
		0x01,
		0x07,
		0x74,
		0x2c,
		0x25,
		0x79,
		0xf8,
		0x76,
		0x40,
		0xe5,
		0x0a,
		0xb4,
		0x02,
		0x07,
		0x74,
		0x2c,
		0x25,
		0x79,
		0xf8,
		0x76,
		0x80,
		0x22,
		0xc0,
		0xe0,
		0xc0,
		0x83,
		0xc0,
		0x82,
		0xc0,
		0xd0,
		0x90,
		0x3f,
		0x0d,
		0xe0,
		0xf5,
		0x09,
		0xe5,
		0x09,
		0x30,
		0xe0,
		0x2e,
		0xe5,
		0x7a,
		0xb4,
		0x01,
		0x09,
		0x90,
		0x3a,
		0x00,
		0xe0,
		0xf5,
		0x78,
		0x44,
		0x01,
		0xf0,
		0xe5,
		0x7a,
		0xb4,
		0x03,
		0x09,
		0x90,
		0x3a,
		0x00,
		0xe0,
		0xf5,
		0x78,
		0x54,
		0xfe,
		0xf0,
		0xe5,
		0x7a,
		0xb4,
		0x03,
		0x05,
		0x75,
		0x7a,
		0x00,
		0x80,
		0x02,
		0x05,
		0x7a,
		0x90,
		0x3f,
		0x0d,
		0x74,
		0x01,
		0xf0,
		0xd0,
		0xd0,
		0xd0,
		0x82,
		0xd0,
		0x83,
		0xd0,
		0xe0,
		0x32,
		0x90,
		0x30,
		0x27,
		0x74,
		0xfa,
		0xf0,
		0x12,
		0x0d,
		0xfa,
		0x12,
		0x15,
		0x63,
		0xe4,
		0xf5,
		0x33,
		0xd2,
		0xaf,
		0x12,
		0x09,
		0x5d,
		0x30,
		0x30,
		0x03,
		0x12,
		0x04,
		0xf3,
		0x30,
		0x34,
		0x03,
		0x12,
		0x13,
		0x6f,
		0x30,
		0x3d,
		0xee,
		0xc2,
		0x3d,
		0xd2,
		0x35,
		0x30,
		0x00,
		0x05,
		0x12,
		0x15,
		0x36,
		0x80,
		0x17,
		0x30,
		0x07,
		0x0b,
		0x30,
		0x06,
		0x08,
		0x20,
		0x31,
		0x0e,
		0x12,
		0x11,
		0xeb,
		0x80,
		0x09,
		0x20,
		0x07,
		0x06,
		0x30,
		0x06,
		0x03,
		0x12,
		0x0c,
		0xf3,
		0xc2,
		0x35,
		0x80,
		0xc7,
		0x90,
		0x50,
		0x27,
		0xe0,
		0x44,
		0x01,
		0xf0,
		0x90,
		0x50,
		0x34,
		0x74,
		0x80,
		0xf0,
		0xa3,
		0x74,
		0x2a,
		0xf0,
		0xa3,
		0x74,
		0x14,
		0xf0,
		0x90,
		0x50,
		0x30,
		0xe4,
		0xf0,
		0xa3,
		0x74,
		0x02,
		0xf0,
		0xa3,
		0xe4,
		0xf0,
		0xa3,
		0x74,
		0x80,
		0xf0,
		0xe4,
		0xf5,
		0x0a,
		0x12,
		0x10,
		0xb2,
		0x75,
		0x79,
		0x02,
		0x75,
		0x0a,
		0x01,
		0x12,
		0x13,
		0xc8,
		0xd2,
		0x18,
		0xd2,
		0x19,
		0xc2,
		0x3c,
		0xc2,
		0x3b,
		0xd2,
		0x1a,
		0xd2,
		0x38,
		0xd2,
		0x30,
		0xc2,
		0x35,
		0xc2,
		0x3d,
		0x22,
		0x85,
		0x10,
		0x11,
		0x7f,
		0x08,
		0xe5,
		0x11,
		0x30,
		0xe7,
		0x04,
		0xd2,
		0x29,
		0x80,
		0x02,
	},
	{
		0x95,
		0x00,
		0xc2,
		0x29,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xae,
		0x51,
		0x15,
		0x51,
		0xee,
		0x70,
		0xf9,
		0xe5,
		0x11,
		0x25,
		0xe0,
		0xf5,
		0x11,
		0xd2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x0a,
		0xae,
		0x51,
		0x15,
		0x51,
		0xee,
		0x70,
		0xf9,
		0xc2,
		0x28,
		0x12,
		0x01,
		0x6c,
		0x75,
		0x51,
		0x05,
		0xae,
		0x51,
		0x15,
		0x51,
		0xee,
		0x70,
		0xf9,
		0xdf,
		0xc2,
		0x22,
		0xe5,
		0x20,
		0x54,
		0x07,
		0xff,
		0xbf,
		0x01,
		0x03,
		0x02,
		0x0f,
		0xb4,
		0xe5,
		0x20,
		0x54,
		0x07,
		0xff,
		0xbf,
		0x07,
		0x03,
		0x02,
		0x11,
		0x85,
		0xe5,
		0x20,
		0x54,
		0x07,
		0xff,
		0xbf,
		0x03,
		0x03,
		0x02,
		0x13,
		0x14,
		0xe5,
		0x20,
		0x54,
		0x07,
		0xff,
		0xbf,
		0x05,
		0x03,
		0x12,
		0x16,
		0x3f,
		0x22,
		0x12,
		0x14,
		0xab,
		0x12,
		0x16,
		0x4d,
		0x50,
		0x02,
		0xd2,
		0x05,
		0x12,
		0x02,
		0x24,
		0xc2,
		0x39,
		0xc2,
		0x36,
		0xc2,
		0x31,
		0xd2,
		0x34,
		0x12,
		0x02,
		0x52,
		0xb5,
		0x07,
		0x03,
		0xd3,
		0x80,
		0x01,
		0xc3,
		0x40,
		0x02,
		0xc2,
		0x05,
		0x22,
		0xe4,
		0xff,
		0xfe,
		0xc3,
		0xef,
		0x95,
		0x0e,
		0xee,
		0x95,
		0x0d,
		0x50,
		0x15,
		0x7d,
		0x8a,
		0x7c,
		0x02,
		0xed,
		0x1d,
		0xaa,
		0x04,
		0x70,
		0x01,
		0x1c,
		0x4a,
		0x70,
		0xf6,
		0x0f,
		0xbf,
		0x00,
		0x01,
		0x0e,
		0x80,
		0xe2,
		0x22,
		0xc2,
		0x03,
		0xd2,
		0x04,
		0x12,
		0x02,
		0x78,
		0x30,
		0x07,
		0x05,
		0x30,
		0x06,
		0x02,
		0xd2,
		0x36,
		0xc2,
		0x3e,
		0x12,
		0x0e,
		0x9a,
		0xc2,
		0x3e,
		0x12,
		0x12,
		0x51,
		0xd2,
		0x34,
		0x22,
		0xe5,
		0x7e,
		0xb4,
		0x01,
		0x05,
		0x12,
		0x16,
		0x12,
		0x80,
		0x08,
		0xe5,
		0x7e,
		0xb4,
		0x02,
		0x09,
		0x12,
		0x16,
		0x21,
		0xe4,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0x22,
		0xe4,
		0xff,
		0xef,
		0x25,
		0xe0,
		0x24,
		0x56,
		0xf8,
		0xe4,
		0xf6,
		0x08,
		0xf6,
		0x0f,
		0xbf,
		0x07,
		0xf2,
		0x53,
		0x24,
		0x80,
		0x22,
		0xe5,
		0x7d,
		0xc3,
		0x95,
		0x7c,
		0x40,
		0x01,
		0x22,
		0xe5,
		0x7d,
		0x04,
		0xf5,
		0x0c,
		0x12,
	}
};
#define AUTO_FOCUS_FIRMWARE_TAB2_COUNT 89
static unsigned char gAutoFocusFirmwareInitTab2[AUTO_FOCUS_FIRMWARE_TAB2_COUNT] =
{
		0x96,
		0x00,
		0x12,
		0xb4,
		0x22,
		0xe5,
		0x7d,
		0x70,
		0x02,
		0xc3,
		0x22,
		0xe5,
		0x7d,
		0x14,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0x22,
		0x75,
		0x48,
		0x0f,
		0x75,
		0x49,
		0x2a,
		0x90,
		0x0f,
		0x28,
		0x12,
		0x02,
		0x6e,
		0xc2,
		0x3a,
		0x22,
		0x75,
		0x48,
		0x0f,
		0x75,
		0x49,
		0x54,
		0x90,
		0x0f,
		0x52,
		0x12,
		0x02,
		0x6e,
		0xd2,
		0x3a,
		0x22,
		0xe5,
		0x7e,
		0x24,
		0xfe,
		0x60,
		0x06,
		0x04,
		0x70,
		0x05,
		0xd2,
		0x39,
		0x22,
		0xc2,
		0x39,
		0x22,
		0xe5,
		0x31,
		0xd3,
		0x94,
		0x00,
		0x40,
		0x03,
		0x15,
		0x31,
		0x22,
		0x12,
		0x15,
		0xa9,
		0x22,
		0x12,
		0x16,
		0x12,
		0xe4,
		0xf5,
		0x0c,
		0x12,
		0x12,
		0xb4,
		0x22,
};

const static u16 ov5642_reg_flash_strobe_on[] =
{
	/* Enable strobe flash: */
	//0x3b00, 0x83,
	0x3016, 0x02,
	0x301c, 0x02,
	0x3019, 0x02,
	OV5642_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 ov5642_reg_flash_strobe_off[] =
{
	/* Disable strobe flash: */
	//0x3b00, 0x00,
	0x3016, 0x00,
	0x301c, 0x00,
	0x3019, 0x00,
	OV5642_REGEND, 0x00	    /* End of list delimiter */
};

static int __attribute__ ((unused)) ov5642_sensor_irq_subfunc(unsigned int int_state)
{	
	
	return 0;		
}

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_on(pdata->id);

}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	pdata->power_off(pdata->id);
}

static void cam_flash_fire(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(1);
}

static void cam_flash_torch(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(0);
}

static void ov5642hw_power_down( u8 powerMode )
{
	/* OV5642 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();

	msleep(100);
}

static void ov5642hw_reset(void)
{
	ov5642_dbg("ov5642 hw reset\n");
	return;
}

static int ov5642_set_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0;//,regCal =0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != OV5642_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		//mdelay(1);
		status = (write_sensor_reg( regP[curReg << 1], regVal) == 0) ?
			0 : -EIO;
		//read_sensor_reg(regP[curReg << 1],&regCal);
		//if(regCal != regVal)
		//	ov5642_dbg("not match adr=%x, val=%02x, reg=%02x\n", regP[curReg << 1], regVal, regCal);
		curReg++;
	}

	return status;
}

static int __attribute__ ((unused)) ov5642_check_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0,regCal =0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != OV5642_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		read_sensor_reg(regP[curReg << 1],&regCal);
		if(regCal != regVal)
			ov5642_dbg("not match adr=%x, val=%02x, reg=%02x\n", regP[curReg << 1], regVal, regCal);
		curReg++;
	}

	return status;
}

static int ov5642_sensorID_detect(void)
{
	u8 val_h,val_l;
	read_sensor_reg(PID_ADRRH, &val_h);
	read_sensor_reg(PID_ADRRL, &val_l);
	
	ov5642_dbg("ov5642_sensorID_detect\n");
	ov5642_dbg("val_h=%x, val_l=%x\n", val_h, val_l);
	
	if( (val_h == PID_VERH ))
		return 0;
	else
		return -1;
}

static int ov5642hw_read_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

static int ov5642hw_write_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return write_sensor_reg(subAddress, *bufP);
}

static void ov5642hw_flash_torch(void)
{
	down(&ov5642_flash_lock);
	if(!flashled_on){
		printk(KERN_DEBUG "ov5642: flash led on\n");
		cam_flash_torch();
		flashled_fire = 0;
		ov5642_set_regs( ov5642_reg_flash_strobe_on );
		flashled_on = 1;

		/* start 5s timeout to off flash */
		queue_delayed_work(flash_wqueue, &flash_work, msecs_to_jiffies(5000));
	}
	else
		printk(KERN_ERR "ov5642: error, flash led already was on\n");
	up(&ov5642_flash_lock);
}

static void ov5642hw_flash_fire(void)
{
	struct timeval now;
        unsigned int ms;

	down(&ov5642_flash_lock);
	if(flashled_on){
		do_gettimeofday(&now);
		ms = 1000 * (now.tv_sec - flashled_fire_time.tv_sec);
		ms += (now.tv_usec - flashled_fire_time.tv_usec)/1000;
		if(ms<=5000) {
			printk(KERN_ERR "ov5642: warning, flash led off %d ms, must be off within 5s.\n", ms);
			up(&ov5642_flash_lock);
			return;
		}

		printk(KERN_DEBUG "ov5642: flash led fire\n");
		cam_flash_fire();
		flashled_fire = 1;
	}
	else
		printk(KERN_ERR "ov5642: fire error, flash led is not on\n");
	up(&ov5642_flash_lock);
}

static void ov5642hw_flash_off(void)
{
	down(&ov5642_flash_lock);
	if(flashled_on){
		cam_flash_torch();
		/* save flash fire off time */
		if(flashled_fire)
			do_gettimeofday(&flashled_fire_time);
		flashled_fire = 0;

		ov5642_set_regs( ov5642_reg_flash_strobe_off);
		flashled_on = 0;
		printk(KERN_DEBUG "ov5642: flash led off\n");
	}
	up(&ov5642_flash_lock);
}

static void ov5642hw_flash_work_off(void)
{
	cancel_delayed_work(&flash_work);
	if(flashled_on){
		ov5642hw_flash_off();
	}
}

static void ov5642_flash_work_handler(struct work_struct *work)
{
	printk(KERN_INFO "ov5642: flashing timeout to be off\n");
	ov5642hw_flash_off();
}

/*****************************************************************************
 *  Settings
 *****************************************************************************/
static int ov5642hw_version_revision(u8 * pCmRevision, u8 *pSensorRevision)
{
	read_sensor_reg( PID_ADRRH, pCmRevision );
	read_sensor_reg( PID_ADRRL, pSensorRevision );
	return 0;
}

static void __attribute__ ((unused)) ov5642hw_set_hsync(void)
{

}

static void __attribute__ ((unused)) ov5642hw_auto_function_on(void)
{
	
}

static void __attribute__ ((unused)) ov5642hw_auto_function_off(void)
{
	
}

/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
static int __attribute__ ((unused)) ov5642hw_viewfinder_on(void)
{

	return 0;
}

static int ov5642hw_viewfinder_off(void)
{
	

	return 0;
}

 static int __attribute__ ((unused))  ov5642hw_format_check(u32 captureSizeFormat, u32 colorFormat)
{
	const u16   *defaultDataP;
	struct OV5642_SETTING  *ov5642_setting;
	defaultDataP  = NULL;

		
	for(ov5642_setting = g_ov5642_setting_table;
		ov5642_setting->format != 0;ov5642_setting++) {
		if ((colorFormat == ov5642_setting->format) &&
			(captureSizeFormat == ov5642_setting->resolution)) {
			defaultDataP = ov5642_setting->setting;
			break;
		}
	}
	
	if (defaultDataP == NULL)
	{
		ov5642_dbg("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
		ov5642_dbg("captureSizeFormat =%x colorFormat =%x\n",captureSizeFormat,colorFormat);
		ov5642_dbg("_5m_format_check has no such format\n");
		ov5642_dbg("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
		return -EPERM;
	}
	ov5642_set_regs( defaultDataP );
	return 0;
}

#define Capture_Framerate 750
#define Preview_FrameRate 1500
static int ov5642_parameter_record(void)
{
	//int reg = 0x519f;
	u8 ret_l,ret_m,ret_h;
	int tp_l,tp_m,tp_h;
	write_sensor_reg(0x3503, 0x07);
 	ov5642_dbg("************************************\n");
  	ov5642_dbg("******preview parameter record******\n");
	ov5642_dbg("************************************\n");
 	read_sensor_reg(0x3500,&ret_h);
 	read_sensor_reg(0x3501, &ret_m);
	read_sensor_reg(0x3502, &ret_l);
	ov5642_dbg("0x3500=%x 0x3501=%x 0x3502=%x\n",ret_h,ret_m,ret_l);
	tp_l = ret_l;
	tp_m = ret_m;
	tp_h = ret_h;
	g_preview_exposure = (tp_h<<12)+(tp_m<<4)+(tp_l>>4);
	ov5642_dbg("g_preview_exposure=%x\n",g_preview_exposure);
	read_sensor_reg(0x350c, &ret_h);
	read_sensor_reg(0x350d, &ret_l);
	ov5642_dbg("350c=%x 350d=%x\n",ret_h,ret_l);
	 g_preview_line_width = ret_h & 0xff;
	 g_preview_line_width = g_preview_line_width*256 +ret_l;
 	ov5642_dbg("g_preview_line_width=%x\n",g_preview_line_width);
 	//Read back AGC Gain for preview
 	read_sensor_reg(0x350b, &g_preview_gain);
	ov5642_dbg("0x350b gain =%x\n",g_preview_gain);
	/**need fine tune later**/
	//read_sensor_reg(reg,    &ov5642_recRgain_high);				
	//read_sensor_reg(reg+1,&ov5642_recRgain_low);				
	//read_sensor_reg(reg+2,&ov5642_recBgain_high);
	//read_sensor_reg(reg+3,&ov5642_recBgain_low);	
	//read_sensor_reg(reg+4,&ov5642_recGgain_high);
	//read_sensor_reg(reg+5,&ov5642_recGgain_low);	
	//ov5642_dbg("ov5642recode the awb value\n");	

	return 0;
}

static int ov5642hw_ae_transfer(void)
{
	unsigned char ExposureLow;

	unsigned char ExposureMid; 

	unsigned char ExposureHigh; 
	
	long ulCapture_Exposure;

	long ulCapture_Exposure_Gain;

	long iCapture_Gain;

	int Lines_10ms;

	bool m_60Hz = 0;
	
	u8 reg_l = 0,reg_h =0;

	int Preview_Maxlines = g_preview_line_width;

	unsigned char Gain = g_preview_gain;
 
	int Capture_MaxLines;

	int uEV;


	ov5642_dbg("************************************\n");
  	ov5642_dbg("******ov5642hw_ae_transfer******\n");
	ov5642_dbg("************************************\n");
	read_sensor_reg(0x350c, &reg_h);
	read_sensor_reg(0x350d, &reg_l);
	Capture_MaxLines = reg_h & 0xff;
	Capture_MaxLines = Capture_MaxLines*256+ reg_l;

 	ov5642_dbg("0x350c=%x  0x350d=%x\n",reg_h,reg_l);
	ov5642_dbg("Capture_MaxLines=%x\n",Capture_MaxLines);
	 if(m_60Hz== 1)
	 {
	 	 Lines_10ms = Capture_Framerate * Capture_MaxLines/12000;
	 }
	 else
	 {
	 	 Lines_10ms = Capture_Framerate * Capture_MaxLines/10000;
	 } 

	//Lines_10ms = Lines_10ms * 13 /12;
	if(Preview_Maxlines == 0)
	{
		Preview_Maxlines = 1;
		ov5642_dbg("can not divide zero\n");
	}
	 ulCapture_Exposure = 
 	 (g_preview_exposure*(Capture_Framerate)*(Capture_MaxLines))/(((Preview_Maxlines)*(Preview_FrameRate)));
 
	 iCapture_Gain = (Gain & 0x0f) + 16;
 
	 if (Gain & 0x10)
	 {
	  iCapture_Gain = iCapture_Gain << 1;
	 }
 
	 if (Gain & 0x20)
	 {
	  iCapture_Gain = iCapture_Gain << 1;
	 }
 
	 if (Gain & 0x40)
	 {
	  iCapture_Gain = iCapture_Gain << 1;
	 }
 
	 if (Gain & 0x80)
	 {
	  iCapture_Gain = iCapture_Gain << 1;
	 }
 
	 ulCapture_Exposure_Gain = ulCapture_Exposure * iCapture_Gain * 3 / 2;
	 
	 if(ulCapture_Exposure_Gain < ((u32)(Capture_MaxLines)*16))
	 {
	  ulCapture_Exposure = ulCapture_Exposure_Gain/16;
	 
	  if (ulCapture_Exposure > Lines_10ms)
	  {
		   ulCapture_Exposure /= Lines_10ms;
		   ulCapture_Exposure *= Lines_10ms;
	  }
	 }
	 else
	 {
	 	 ulCapture_Exposure = Capture_MaxLines;
	 }
	 
	 if(ulCapture_Exposure == 0)
	 {
	  	ulCapture_Exposure = 1;
	 }
 
	iCapture_Gain = (ulCapture_Exposure_Gain*2/ulCapture_Exposure + 1)/2;

	ExposureLow = ((unsigned char)ulCapture_Exposure)<<4;

	ExposureMid = (unsigned char)(ulCapture_Exposure >> 4) & 0xff;

	ExposureHigh = (unsigned char)(ulCapture_Exposure >> 12);
	ov5642_dbg("ulCapture_Exposure =%lx \n",ulCapture_Exposure);
	ov5642_dbg("ExposureLow =%x \n",ExposureLow);
	ov5642_dbg("ExposureMid =%x \n",ExposureMid);
	ov5642_dbg("ExposureHigh =%x \n",ExposureHigh);
 
	//ulCapture_Exposure_end=ulCapture_Exposure;

	//iCapture_Gain_end=iCapture_Gain;


	Gain = 0;

	if (iCapture_Gain > 31)
	{
	Gain |= 0x10;

	iCapture_Gain = iCapture_Gain >> 1;
	}
 
	if (iCapture_Gain > 31)
	{
	Gain |= 0x20;
	iCapture_Gain = iCapture_Gain >> 1;
	}

	if (iCapture_Gain > 31)
	{
	Gain |= 0x40;
	iCapture_Gain = iCapture_Gain >> 1;
	}

	if (iCapture_Gain > 31)
	{
	Gain |= 0x80;

	iCapture_Gain = iCapture_Gain >> 1;
	}

	if (iCapture_Gain > 16)
	{
	Gain |= ((iCapture_Gain -16) & 0x0f);

	}

	if(Gain == 0x10)
	{
	Gain = 0x11;
	}
 
	// write the gain and exposure to 0x350* registers

	//m_iWrite0x350b=Gain;
	write_sensor_reg(0x350b, Gain);

	//m_iWrite0x3502=ExposureLow;
	write_sensor_reg(0x3502, ExposureLow);

	//m_iWrite0x3501=ExposureMid;
	write_sensor_reg(0x3501, ExposureMid);

	//m_iWrite0x3500=ExposureHigh;
	write_sensor_reg(0x3500, ExposureHigh);

 	uEV = ulCapture_Exposure * iCapture_Gain;
	ov5642_dbg("*******final calculate result ************\n");
	ov5642_dbg("ulCapture_Exposure = %ld iCapture_Gain =%ld uEV =%d\n",ulCapture_Exposure,iCapture_Gain,uEV);
	if(uEV < 200)
	{
		write_sensor_block_regs(g_capture_gamma_outdoor, 58);
		ov5642_dbg("capture gamma outdoor\n");
	}
	else if(uEV < 600)
	{
		write_sensor_block_regs(g_capture_gamma_middle, 58);
		ov5642_dbg("capture gamma middle\n");
	}
	else
	{
		write_sensor_block_regs(g_capture_gamma_indoor, 58);
		ov5642_dbg("capture gamma middle\n");
	}
 
 	return 0;
}
 
static int ov5642hw_preview2snap(void)
{
	//int reg = 0x5198;
	ov5642hw_ae_transfer();
	/**need fine tune later**/
	//write_sensor_reg(0x5196,0x83);//set manual mode
	//write_sensor_reg(reg,		ov5642_recRgain_high);
	//write_sensor_reg(reg+1,		ov5642_recRgain_low);
	//write_sensor_reg(reg+2,		ov5642_recBgain_high);
	//write_sensor_reg(reg+3,		ov5642_recBgain_low);
	//write_sensor_reg(reg+4,		ov5642_recGgain_high);
	//write_sensor_reg(reg+5,		ov5642_recGgain_low);

	ov5642_dbg("preview2snap****\n");
	return 0;
}
 
static int ov5642hw_snap2preview(void)
{
	/**need fine tune later**/
	write_sensor_block_regs(g_preview_gamma, 58);
	ov5642_dbg("preview gamma \n");
	ov5642_dbg("snap2preview****\n");
	msleep(100);	/* workaround: next preview frames may not good */
	return 0;
}
 
static int ov5642hw_check_flash(void)
{
	if(flashled_flag == 1) {
		if(flashled_on) {
			ov5642hw_flash_fire();
			msleep(700);
		}
		else {
			ov5642hw_flash_torch();
			ov5642hw_flash_fire();
			msleep(1000);
		}
	}
	return 0;
}

static int ov5642hw_mode_check(u32 mode)
{
	if(g_ov5642_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			ov5642hw_preview2snap();
		else {
			if(flashled_on) {
				ov5642hw_flash_work_off();
			}
			ov5642hw_snap2preview();
		}
		
		g_ov5642_mode_change = mode;
		
	}
	ov5642_dbg("mode check\n");
	return 0;
}

static int ov5642hw_auto_focus_idle(void)
{
	write_sensor_reg(0x3024, 0x08);

    if(flashled_on){
        ov5642hw_flash_work_off();
	 }
	
	return 0;
}

static int ov5642hw_auto_focus(void)
{
	u8 ret;
	int cnt = 3000;

	write_sensor_reg(0x3024, 0x03);

	ret = 0;
	while(cnt --)
	{
		read_sensor_reg(0x3027, &ret);
		if(ret  == 0x02)
			break;
	}

	if(cnt < 10) {
		ov5642_dbg("ov5642hw_auto_focus failed ret=%d\n",ret);
		return -ETIME;
	}
	else
		ov5642_dbg("ov5642hw_auto_focus ok \n");
	return 0;
}

static int ov5642hw_af_firmware(void)
{
	int i;
	write_sensor_reg(0x3000,0x20);

	
	for(i = 0; i < AUTO_FOCUS_FIRMWARE_TAB1_COUNTX; i ++)
		i2c_master_send(g_client,gAutoFocusFirmwareInitTab1[i],AUTO_FOCUS_FIRMWARE_TAB1_COUNTY);
	
	i2c_master_send(g_client,gAutoFocusFirmwareInitTab2,AUTO_FOCUS_FIRMWARE_TAB2_COUNT);
	
	write_sensor_reg(0x3024, 0x00);
	write_sensor_reg(0x3025, 0x00);
	write_sensor_reg(0x5082, 0x00);
	write_sensor_reg(0x5083, 0x00);
	write_sensor_reg(0x5084, 0x00);
	write_sensor_reg(0x5085, 0x00);
	write_sensor_reg(0x3026, 0x00);
	write_sensor_reg(0x3027, 0xFF);
	
	write_sensor_reg(0x3000,0x00);
	return 0;
	
}

/*****************************************************************************
 *  Format
 *****************************************************************************/
static int ov5642hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{
	
	if(g_initial_flag == 0)
	{
		
		ov5642_set_regs(gYUV_initialize_tab);
		ov5642_set_regs(g_preview_iq_init_tab);
		write_sensor_block_regs(g_preview_cmx,23);
		write_sensor_block_regs(g_preview_gamma,58);
		write_sensor_block_regs(g_preview_lenc,138);
		write_sensor_block_regs(g_preview_awb,33);
		
		ov5642_set_regs( g_Contrast_Standard_Settings );
		ov5642_set_regs( g_JpegQuality_Standard_Settings );
		ov5642_set_regs( g_Sharpness_Standard_Settings );
		ov5642_set_regs( g_Satuation_Standard_Settings );
		ov5642_set_regs( g_EVCompensat_5_Setting );
		ov5642_set_regs( g_WB_Auto_Settings );

		ov5642_dbg("init sensor first time\n");		
		
	}else
	{
		ov5642_parameter_record();
	}

	if(g_ov5642_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			ov5642hw_check_flash();
	}

	if(colorFormat == OV5642_FORMAT_JPG)
	{
		ov5642_set_regs(g_yuv2jpg_tab);
		switch(captureSizeFormat)
		{
			case OV5642_SIZE_QSXGA:
				ov5642_set_regs(gJPG_QSXGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_QSXGA\n");
				break;
			case OV5642_SIZE_2560X1920:
				ov5642_set_regs(gJPG_2560X1920_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_2560X1920\n");
				break;	
			case OV5642_SIZE_2592X1458:
				ov5642_set_regs(gJPG_2592X1458_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_2592X1458\n");
				break;
			case OV5642_SIZE_QXGA:
				ov5642_set_regs(gJPG_QXGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_QXGA\n");
				break;
			case OV5642_SIZE_2048X1152:
				ov5642_set_regs(gJPG_2048X1152_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_2048X1152\n");
				break;
			case OV5642_SIZE_UXGA:
				ov5642_set_regs(gJPG_UXGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_UXGA\n");
				break;
			case OV5642_SIZE_1600X900:
				ov5642_set_regs(gJPG_1600X900_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_1600X900\n");
				break;
			case OV5642_SIZE_1280X960:
				ov5642_set_regs(gJPG_1280X960_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_1280X960\n");
				break;
			case OV5642_SIZE_XGA:
				ov5642_set_regs(gJPG_XGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_XGA\n");
				break;
			case OV5642_SIZE_854X480:
				ov5642_set_regs(gJPG_854X480_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_854X480\n");
				break;
			case OV5642_SIZE_VGA:
				ov5642_set_regs(gJPG_VGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_VGA\n");
				break;
			case OV5642_SIZE_QVGA:
				ov5642_set_regs(gJPG_QVGA_initialize_tab);
				ov5642_dbg(" jpg OV5642_SIZE_QVGA\n");
				break;
			default:
				ov5642_set_regs(gJPG_QSXGA_initialize_tab);
				ov5642_dbg(" jpg default QSXGA\n");
				break;
			
		}
		
	}
	else
	{
		if(g_af_fw_flag == 0)
		{
			ov5642_dbg("download af firmware pls wait-----\n");
			ov5642hw_af_firmware();
			ov5642_dbg(" download af firmware complete\n");
			g_af_fw_flag = 1;
		}
		if(g_initial_flag == 1)
		{
			ov5642_set_regs(g_jpg2yuv_tab);
			ov5642_dbg("**soft reset**\n");
		}
		
		switch(captureSizeFormat)
		{
			case OV5642_SIZE_854X480:
				ov5642_set_regs(gYUV_854X480_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_854X480\n");
				break;
			case OV5642_SIZE_VGA:
				ov5642_set_regs(gYUV_VGA_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_VGA\n");
				break;
			case OV5642_SIZE_1280X720:
				ov5642_set_regs(gYUV_1280X720_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_1280X720\n");
				break;
			case OV5642_SIZE_QVGA:
				ov5642_set_regs(gYUV_QVGA_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_QVGA\n");
				break;
			case OV5642_SIZE_SQCIF:
				ov5642_set_regs(gYUV_SQCIF_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_SQCIF\n");
				break;				
			case OV5642_SIZE_160X120:
				ov5642_set_regs(gYUV_160X120_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_160X120\n");
				break;				
			case OV5642_SIZE_QCIF:
				ov5642_set_regs(gYUV_QCIF_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_QCIF\n");
				break;				
			case OV5642_SIZE_CIF:
				ov5642_set_regs(gYUV_CIF_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_CIF\n");
				break;
			case OV5642_SIZE_432X320:
				ov5642_set_regs(gYUV_432X320_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_432X320\n");
				break;
			case OV5642_SIZE_480X360:
				ov5642_set_regs(gYUV_480X360_initialize_tab);
				ov5642_dbg(" YUV OV5642_SIZE_480X360\n");
				break;				
		}
		
	}
	if(g_initial_flag == 1)
	{
		ov5642hw_mode_check(mode);	
	}
	g_initial_flag = 1;
	
	return 0;
}

/*****************************************************************************
 * OV5642 Functions
 *****************************************************************************/

/***********************************************************
*** Digital Continue Zoom Function ***
s_format: 0 for SXGA, 1 for QSXGA
dvp_w: Output Width
dvp_h: Output Height
zoom_level: 0 for 1.0 zoom, 1 for 1.1 zoom, 2 for 1.2 zoom, etc. 
The maximum zoom level depends on the resolution of both preview and capture 
***********************************************************/
//static int ov5642_resolution_zoom(unsigned char s_format, unsigned int dvp_w, unsigned int dvp_h, unsigned char zoom_level)
static int ov5642_resolution_zoom(p_camera_context_t camera_context, u32 value)
{
	
	unsigned char Reg3808 = 0, Reg3809 = 0, Reg380a = 0, Reg380b = 0; 
	unsigned char Reg3824 = 0, Reg3825 = 0, Reg3826 = 0, Reg3827 = 0;
	unsigned char Reg3800 = 0, Reg3801 = 0, Reg3802 = 0, Reg3803 = 0;
	unsigned char Reg3804 = 0, Reg3805 = 0, Reg3806 = 0, Reg3807 = 0;
	unsigned char Reg5682 = 0, Reg5683 = 0, Reg5686 = 0, Reg5687 = 0;

	unsigned int MinLevel=0, MaxLevel=0;
	unsigned int windw=0, windh=0, sx=0, sy=0;
	unsigned int zoom =0;
	unsigned int zoom_level =0;
	
	unsigned char s_format=0;
	unsigned int dvp_w;
	unsigned int dvp_h;
	

	if(camera_context->capture_input_format == V4L2_PIX_FMT_YUV422P) 
	{
		s_format = 0;
	}
	else if(camera_context->capture_input_format == V4L2_PIX_FMT_JPEG)
	{
		s_format = 1;
	} 
	
	dvp_w = camera_context->capture_input_width;
	dvp_h = camera_context->capture_input_height;
	
	zoom_level = (unsigned char)value;
		
	zoom = zoom_level + 10;
	MinLevel = 10;
	
	if(s_format == 0)
	{
		MaxLevel = (1280*10)/dvp_w;
	}	
	else if(s_format == 1)
	{
		MaxLevel = (2592*10)/dvp_w;
	}
	
	if(zoom < MinLevel)
	{
		zoom = MinLevel;
	}
	else if(zoom > MaxLevel)
	{
		zoom = MaxLevel;
	}
	if(zoom<10)
		zoom = 10;

	if( (s_format==0) && (zoom!=MaxLevel) )
		dvp_h += 1; /* plus extra 1 line on preview, fix bad top line issue */
	ov5642_dbg("ov5642 zoom: dvp %d:%d, level=%d zoom=%d:%d:%d, format=%d\n", dvp_w, dvp_h, zoom_level, zoom, MinLevel, MaxLevel, s_format);

	Reg3808 = (int)(dvp_w/256);
	Reg3809 = dvp_w - Reg3808*256;
	Reg380a = (int)(dvp_h/256);
	Reg380b = dvp_h - Reg380a*256;
	
	if(s_format == 0)
	{
		windw = (int)(1280/zoom*10);
		windh = (int)(960/zoom*10);
		sx = 436 + (int)((1280 - windw)/2);
	    	sy = 8 + (int)((960 - windh)/2);
		Reg3824 = (int)(sx/256);
		Reg3825 = sx - Reg3824*256;
		Reg3826 = (int)(sy/256);
		Reg3827 = sy - Reg3826*256;
	}	
	else if(s_format == 1)
	{
		windw = (int)(2592/zoom*10);
		windh = (int)(1944/zoom*10);
		sx = 394 + (int)((2592 - windw)/2);
	    	sy = 10 + (int)((1944 - windh)/2);
		Reg3800 = (int)(sx/256);
		Reg3801 = sx - Reg3824*256;
		Reg3802 = (int)(sy/256);
		Reg3803 = sy - Reg3826*256;
	}
	
	Reg3804 = (int)(windw/256);
	Reg3805 = windw - Reg3804*256;
	Reg3806 = (int)(windh/256);
	Reg3807 = windh - Reg3806*256;		

	Reg5682 = (int)(windw/256);
	Reg5683 = windw - Reg5682*256;
	Reg5686 = (int)(windh/256);       
	Reg5687 = windh - Reg5686*256;
	
	if(s_format == 0)
	{
		Reg3824 = Reg3824 | 0x10;
		write_sensor_reg(0x3824, Reg3824);
		write_sensor_reg(0x3825, Reg3825);
		write_sensor_reg(0x3826, Reg3826);
		write_sensor_reg(0x3827, Reg3827);
		write_sensor_reg(0x3804, Reg3804);
		write_sensor_reg(0x3805, Reg3805);
		write_sensor_reg(0x3806, Reg3806);
		write_sensor_reg(0x3807, Reg3807);
		write_sensor_reg(0x3808, Reg3808);
		write_sensor_reg(0x3809, Reg3809);
		write_sensor_reg(0x380a, Reg380a);
		write_sensor_reg(0x380b, Reg380b);
		write_sensor_reg(0x380c, 0x0c);
		write_sensor_reg(0x380d, 0x80);
		write_sensor_reg(0x380e, 0x03);
		write_sensor_reg(0x380f, 0xe8);
		write_sensor_reg(0x5001, 0xf7);
		write_sensor_reg(0x5680, 0x00);
		write_sensor_reg(0x5681, 0x00);
		write_sensor_reg(0x5682, Reg5682);
		write_sensor_reg(0x5683, Reg5683);
		write_sensor_reg(0x5684, 0x00);
		write_sensor_reg(0x5685, 0x00);
		write_sensor_reg(0x5686, Reg5686);
		write_sensor_reg(0x5687, Reg5687);
	}	
	else if(s_format == 1)
	{
		write_sensor_reg(0x3800, Reg3800);
		write_sensor_reg(0x3801, Reg3801);
		write_sensor_reg(0x3802, Reg3802);
		write_sensor_reg(0x3803, Reg3803);
		write_sensor_reg(0x3804, Reg3804);
		write_sensor_reg(0x3805, Reg3805);
		write_sensor_reg(0x3806, Reg3806);
		write_sensor_reg(0x3807, Reg3807);
		write_sensor_reg(0x3808, Reg3808);
		write_sensor_reg(0x3809, Reg3809);
		write_sensor_reg(0x380a, Reg380a);
		write_sensor_reg(0x380b, Reg380b);
		write_sensor_reg(0x380c, 0x0c);
		write_sensor_reg(0x380d, 0x80);
		write_sensor_reg(0x380e, 0x07);
		write_sensor_reg(0x380f, 0xd0);
		write_sensor_reg(0x5001, 0xf7);
		write_sensor_reg(0x5680, 0x00);
		write_sensor_reg(0x5681, 0x00);
		write_sensor_reg(0x5682, Reg5682);
		write_sensor_reg(0x5683, Reg5683);
		write_sensor_reg(0x5684, 0x00);
		write_sensor_reg(0x5685, 0x00);
		write_sensor_reg(0x5686, Reg5686);
		write_sensor_reg(0x5687, Reg5687);
	}
    /* must update af window after zoom setting */
	write_sensor_reg(0x3024, 0x12);
		
	return 0;
}

static inline int ov5642hw_get_zoom_level( int value )
{
	return value / 10 - 10;
}

/**contrast**/
static int ov5642_try_contrast( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_CONTRAST_LOWEST) ||
	    (value == SENSOR_CONTRAST_LOW)   ||
	    (value == SENSOR_CONTRAST_MIDDLE) ||
	    (value == SENSOR_CONTRAST_HIGH) ||
	    (value == SENSOR_CONTRAST_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) {
		case SENSOR_CONTRAST_LOWEST:
			regP = g_Contrast_Lowest_Settings;
			break;
		case SENSOR_CONTRAST_LOW:
			regP = g_Contrast_Low_Settings;
			break;
		case SENSOR_CONTRAST_MIDDLE:
			regP = g_Contrast_Standard_Settings;
			break;
		case SENSOR_CONTRAST_HIGH:
			regP = g_Contrast_High_Settings;
			break;
		case SENSOR_CONTRAST_HIGHEST:
			regP = g_Contrast_Highest_Settings;
			break;
		default:
			regP = g_Contrast_Standard_Settings;
			break;
	}

	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_contrast\n");
	return 0;
}


/**sensor jpeg quality**/
static int ov5642_set_jpeg_quality( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	if(value>=85)
		regP = g_JpegQuality_Superfine_Settings;
	else if(value>=75)
		regP = g_JpegQuality_Fine_Settings;
	else if(value>=65)
		regP = g_JpegQuality_Standard_Settings;
	else
		regP = g_JpegQuality_Low_Settings;

	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_jpegquality\n");
	return 0;

}

/*colort tone*/
static int ov5642_try_colortone( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_COLORTONE_NORMAL) ||
	    (value == SENSOR_COLORTONE_COOL)   ||
	    (value == SENSOR_COLORTONE_BW)   ||
	    (value == SENSOR_COLORTONE_SEPIA)   ||
	    (value == SENSOR_COLORTONE_WARM) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_colortone( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) {
		case SENSOR_COLORTONE_NORMAL:
			regP = g_ColorTone_Off_Settings;
			break;
		case SENSOR_COLORTONE_COOL:	
			regP = g_ColorTone_Cool_Settings;
			break;
		case SENSOR_COLORTONE_BW:		
			regP = g_ColorTone_BW_Settings;
			break;
		case SENSOR_COLORTONE_SEPIA:		
			regP = g_ColorTone_Sepia_Settings;
			break;
		case SENSOR_COLORTONE_WARM:		
			regP = g_ColorTone_Warm_Settings;
			break;
		
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_colortone\n");
	return 0;

}

/*effect*/
static int ov5642_try_effect( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_EFFECT_NONE) ||
	    (value == SENSOR_EFFECT_MONO)   ||
	    (value == SENSOR_EFFECT_SEPIA) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_effect( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) {
		case SENSOR_EFFECT_NONE:
			regP = g_ColorTone_Off_Settings;
			break;
		case SENSOR_EFFECT_MONO:		
			regP = g_ColorTone_BW_Settings;
			break;
		case SENSOR_EFFECT_SEPIA:		
			regP = g_ColorTone_Sepia_Settings;
			break;
		
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_colortone\n");
	return 0;

}

/*sharpness*/
static int ov5642_try_sharpness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SHARPNESS_LOWEST) ||
	    (value == SENSOR_SHARPNESS_LOW)   ||
	    (value == SENSOR_SHARPNESS_STANDARD) ||
	    (value == SENSOR_SHARPNESS_HIGH) ||
	    (value == SENSOR_SHARPNESS_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_sharpness( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) 
	{
		case SENSOR_SHARPNESS_LOWEST: 
			regP = g_Sharpness_Lowest_Settings;
			break;
		case SENSOR_SHARPNESS_LOW:	
			regP = g_Sharpness_Low_Settings;
			break;
		case SENSOR_SHARPNESS_STANDARD:		
			regP = g_Sharpness_Standard_Settings;
			break;
		case SENSOR_SHARPNESS_HIGH:		
			regP = g_Sharpness_High_Settings;
			break;
		case SENSOR_SHARPNESS_HIGHEST:		
			regP = g_Sharpness_Highest_Settings;
			break;
		
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_sharpness\n");
	return 0;

}
/**saturation**/
static int ov5642_try_saturation( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SATURATION_LOWEST) ||
	    (value == SENSOR_SATURATION_LOW)   ||
	    (value == SENSOR_SATURATION_STANDARD) ||
	    (value == SENSOR_SATURATION_HIGH) ||
	    (value == SENSOR_SATURATION_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_saturation( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) 
	{
		case SENSOR_SATURATION_STANDARD: 
			regP = g_Satuation_Standard_Settings;
			break;
		case SENSOR_SATURATION_LOW:	
			regP = g_Satuation_Low_Settings;
			break;
		case SENSOR_SATURATION_HIGH:		
			regP = g_Satuation_High_Settings;
			break;
		case SENSOR_SATURATION_HIGHEST:		
			regP = g_Satuation_Highest_Settings;
			break;
		case SENSOR_SATURATION_LOWEST:		
			regP = g_Satuation_Lowest_Settings;
			break;
		
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_saturation\n");
	return 0;

}

/**ev compesate**/
static int ov5642_set_ev_offset( p_camera_context_t camera_context, int value)
{
	const u16   *regP;
	regP = NULL;
	if (value==0)			/* 0 EV */
		regP = g_EVCompensat_5_Setting;
	else if (value>0) {
		if (value<=33)		/* 0.3 EV */
			regP = g_EVCompensat_6_Setting;
		else if (value<=70)	/* 0.7 EV */
			regP = g_EVCompensat_7_Setting;
		else if (value<=100)	/* 1.0 EV */
			regP = g_EVCompensat_8_Setting;
		else if (value<=133)	/* 1.3 EV */
			regP = g_EVCompensat_9_Setting;
		else			/* 1.7 EV */
			regP = g_EVCompensat_10_Setting;
	}
	else {
		if (value>=-33)		/* -0.3 EV */
			regP = g_EVCompensat_4_Setting;
		else if (value>=-70)	/* -0.7 EV */
			regP = g_EVCompensat_3_Setting;
		else if (value>=-100)	/* -1.0 EV */
			regP = g_EVCompensat_2_Setting;
		else if (value>=-133)	/* -1.3 EV */
			regP = g_EVCompensat_1_Setting;
		else			/* -1.7 EV */
			regP = g_EVCompensat_0_Setting;
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_ev_offset\n");
	return 0;
}

/**white balance**/
static int ov5642_try_white_balance( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_WHITEBALANCE_AUTO) ||
	    (value == SENSOR_WHITEBALANCE_DAYLIGHT) ||
	    (value == SENSOR_WHITEBALANCE_INCANDESCENT) ||
	    (value == SENSOR_WHITEBALANCE_FLUORESCENT) ||
	    (value == SENSOR_WHITEBALANCE_CLOUDY) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_white_balance( p_camera_context_t camera_context, u8 mode, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value) {
		case SENSOR_WHITEBALANCE_AUTO:		/* Auto */
			regP = g_WB_Auto_Settings;
			break;
		case SENSOR_WHITEBALANCE_CLOUDY:	/* Cloudy */
			regP = g_WB_Cloudy_Settings;
			break;
		case SENSOR_WHITEBALANCE_DAYLIGHT:		/* daylight */
			regP = g_WB_Sunny_Settings;
			break;
		case SENSOR_WHITEBALANCE_INCANDESCENT:	/* Home */
			regP = g_WB_Home_Settings;
			break;
		case SENSOR_WHITEBALANCE_FLUORESCENT:	/* Office */
			regP = g_WB_Office_Settings;
			break;
		default:
			break;
	}
	if (regP)
		ov5642_set_regs( regP );
	ov5642_dbg("ov5642_set_white_balance\n");
	return 0;
}

/** auto focus control **/
static int ov5642_set_autofocus( p_camera_context_t camera_context, u32 value)
{
	int ret;

	if (value==SENSOR_AUTOFOCUS_IDLE) {
		ret = ov5642hw_auto_focus_idle();
	}
	else if (value==SENSOR_AUTOFOCUS_SINGLE) {
		if(flashled_flag == 1){
			ov5642hw_flash_torch();
		}
		ret = ov5642hw_auto_focus();
		if(ret && flashled_on)
			ov5642hw_flash_work_off();
	}
	else
		return -EINVAL;
	return ret;
}

/**flash**/
static int ov5642_try_flashmode( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if ( (value==CAMERA_FLASH_OFF) ||
	     (value==CAMERA_FLASH_ON) )
		return 0;
	else
		return -ENOSYS;
}

static int ov5642_set_flashmode( p_camera_context_t camera_context, u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value)
	{
	case CAMERA_FLASH_ON:
		flashled_flag = 1;
		break;
	case CAMERA_FLASH_OFF:
		flashled_flag = 0;
		break;
	}
	if (regP)
		ov5642_set_regs( regP );	
	ov5642_dbg("ov5642_set_flash %d\n", value);
	return 0;
}

/**flash torch**/
static int __attribute__ ((unused)) ov5642_set_flash_torch( p_camera_context_t camera_context, u32 value)
{
	switch(value)
	{
	case CAMERA_FLASHTORCH_ON:
		ov5642hw_flash_torch();
		break;
	case CAMERA_FLASHTORCH_OFF:
		ov5642hw_flash_off();
		break;
	}
	ov5642_dbg("ov5642_set_flash_torch %d\n", value);
	return 0;
}

/**frame**/
static int ov5642hw_set_framerate(u32 value)
{
	
	const u16   *regP;
	regP = NULL;
	switch(value)
	{
	case 15:
		regP = g_Framerate_15_Setting;
		break;
	case 25:
		regP = g_Framerate_25_Setting;
		break;
	case 30:
		regP = g_Framerate_30_Setting;
		break;
	default:
		regP = g_Framerate_15_Night_Setting;
		break;
	}
	if (regP)
		ov5642_set_regs( regP );	
	ov5642_dbg("ov5642hw_set_framerate %d\n", value);
	return 0;
}
//control function over

/**set frame rate**/
static int ov5642_set_framerate(u32 numerator, u32 denominator)
{
	int fps;
	if (0<denominator && 0<=numerator) {
		fps = numerator / denominator;

		if( fps>25 )
			fps = 30;
		else if( fps>15 )
			fps = 25;
		else if( fps>10 )
			fps = 15;
		else
			fps = 7;

		return ov5642hw_set_framerate(fps);
	}
	return 0;
}

/**set digital zoom*/
static int ov5642_set_digital_zoom( p_camera_context_t camera_context, u32 value)
{
	int level;
	if ((value<100) || (value>800))
		return -EINVAL;

	level = ov5642hw_get_zoom_level(value);
	printk(KERN_DEBUG "ov5642: set zoom %d, level %d\n", value, level);

	return ov5642_resolution_zoom(camera_context, level);
}

/**get digital zoom max level*/
static int ov5642_get_digital_zoom_max( p_camera_context_t cam_ctx, u32 *value)
{
	int w, h, mw, mh;

	w = cam_ctx->mode[CAMERA_MODE_VIDEO].width;
	h = cam_ctx->mode[CAMERA_MODE_VIDEO].height;

	if( (w<=0) || (h<=0) )
		return -EINVAL;

	mw = 1280 * 100 / w;
	mh = 960 * 100 / h;
	if (mw<mh)
		*value = mw;
	else
		*value = mh;
	return 0;
}

/**get digital zoom step*/
static int ov5642_get_digital_zoom_step( p_camera_context_t camera_context, u32 *value)
{
	*value = 10;
	return 0;
}

/* set autofocus windows config */
static int ov5642_set_af_windows_config( p_camera_context_t camera_context, int count, struct v4l2_ext_af_window *wins)
{
	int i;
	printk(KERN_DEBUG "ov5642: set af windows count=%d\n", count);
	for(i=0; i<count; i++)
		printk(KERN_DEBUG "ov5642: set af window %d: %d,%d,%d,%d,%d\n", count,
			wins[i].win.left, wins[i].win.top, wins[i].win.width, wins[i].win.height, wins[i].weight);
	/* TODO: ov5642 af config */
	return 0;
}

static int ov5642_enum_frame_sizes(p_camera_context_t camera_context, int mode, u32 format,
				   struct v4l2_frmsizeenum *frm_sizes)
{
	int index = frm_sizes->index;

	frm_sizes->type = V4L2_FRMSIZE_TYPE_DISCRETE;

	if (mode==CAMERA_MODE_VIDEO) {
		if (V4L2_PIX_FMT_YUV422P!=format) {
			printk(KERN_ERR "ov5642: only support YUV422 on video mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ov5642_yuv422_frame_sizes)) {
			frm_sizes->discrete.width = ov5642_yuv422_frame_sizes[index].width;
			frm_sizes->discrete.height = ov5642_yuv422_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	else {
		if (V4L2_PIX_FMT_JPEG!=format) {
			printk(KERN_ERR "ov5642: only support JPEG on still capture mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ov5642_jpeg_frame_sizes)) {
			frm_sizes->discrete.width = ov5642_jpeg_frame_sizes[index].width;
			frm_sizes->discrete.height = ov5642_jpeg_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	return 0;
}

static int ov5642_enum_frame_intervals(p_camera_context_t camera_context, u32 format,
					struct v4l2_frmivalenum *frm_ivals)
{
	int index = frm_ivals->index;

	frm_ivals->type = V4L2_FRMIVAL_TYPE_DISCRETE;

	if (V4L2_PIX_FMT_YUV422P!=format) {
		printk(KERN_ERR "ov5642: only YUV422 have frame interval\n");
		return -EINVAL;
	}
	if (frm_ivals->index < ARRAY_SIZE(ov5642_yuv422_frame_intervals)) {
		frm_ivals->discrete.numerator = ov5642_yuv422_frame_intervals[index].numerator;
		frm_ivals->discrete.denominator = ov5642_yuv422_frame_intervals[index].denominator;
	}
	else
		return -EINVAL;

	return 0;
}

static int ov5642_init( p_camera_context_t camera_context )
{
	u8 cm_rev, cm_pid;
	u32 timeout;
	int status;

	ov5642_dbg("ov5642_init\n");

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to OV5642's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);
	
	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	/* software work around*/
	ci_set_clock(1, 1, CICLK);

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* OV5642 Power on sequence
	 * Take out of Power down mode (GPIO_103), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	ov5642hw_power_down(CAMERA_POWER_FULL);
	ov5642hw_reset();

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = cm_rev = 0;
		status = ov5642hw_version_revision(&cm_pid, &cm_rev);

		/* Check to make sure we are working with an OV5642 */
		if( cm_pid != PID_VERH  )
		{
			ov5642hw_power_down( CAMERA_POWER_OFF );
			ov5642hw_power_down( CAMERA_POWER_FULL );
			ov5642hw_reset();
			mdelay(1);
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_VERH );

	/* initialize default camera parameters */
	/* default preview width/height, 480x360 on 640x360 LCD */
	camera_context->mode[CAMERA_MODE_VIDEO].width = 480;
	camera_context->mode[CAMERA_MODE_VIDEO].height = 360;
	/* default preview framerate */
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_numerator = 15;
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_denominator = 2;
	/* default still capture width/height */
	camera_context->mode[CAMERA_MODE_STILL].width = 2560;
	camera_context->mode[CAMERA_MODE_STILL].height = 1920;
	camera_context->mode[CAMERA_MODE_STILL].input_format = V4L2_PIX_FMT_JPEG;
	camera_context->mode[CAMERA_MODE_STILL].output_format = V4L2_PIX_FMT_JPEG;

	/* turn sensor output off */
	ov5642hw_viewfinder_off();
	g_initial_flag = 0;

	g_ov5642_mode_change = CAMERA_MODE_VIDEO;
	g_ov5642_size = -1;
	g_ov5642_format = -1;

	flashled_flag = 0;
	flashled_on = 0;
	flashled_fire = 0;
	flashled_fire_time.tv_sec = 0;
	flashled_fire_time.tv_usec = 0;

	return 0;
}

static int ov5642_deinit(p_camera_context_t camera_context)
{
	/* off flash if it is on */
	ov5642hw_flash_work_off();

	/* power off the external module */

	ov5642hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

static int ov5642_sleep(  p_camera_context_t camera_context )
{
	return ov5642_deinit( camera_context );
}

static int ov5642_wake(  p_camera_context_t camera_context )
{
	return ov5642_init( camera_context );
}


static unsigned int  t_get_time(void) 
{
	struct timeval tv;
	
	do_gettimeofday(&tv);
	
	return (tv.tv_sec * 1000 + tv.tv_usec / 1000 );
	
}

static int ov5642_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	unsigned int sendtime;	
	int status;
	int ovSizeFormat, ovFormat;
	int capture_width,capture_height;
	int zoom, zoom_width, zoom_height;

	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
			ov5642_dbg("para V4L2_PIX_FMT_YUV422P\n");
			ovFormat = OV5642_FORMAT_YUV_422;
			break;

		case V4L2_PIX_FMT_JPEG:
			ovFormat = OV5642_FORMAT_JPG;
			ov5642_dbg("para V4L2_PIX_FMT_JPEG\n");
			break;
		
		default:
			ovFormat = OV5642_FORMAT_YUV_422;
			ov5642_dbg("para defaut yuv 422\n");
			break;
	}

	ov5642_dbg("camera_context->capture_input_width=%d\n", camera_context->capture_input_width);
	ov5642_dbg("camera_context->capture_input_height=%d\n", camera_context->capture_input_height);
	
	capture_width	 = camera_context->capture_input_width;
	capture_height = camera_context->capture_input_height;	

	zoom = camera_context->sensor_status.digital_zoom_value;
	if((camera_context->capture_mode==CAMERA_MODE_STILL) && (camera_context->capture_input_format==V4L2_PIX_FMT_JPEG)) {
		zoom_width = 2592*100/zoom;
		zoom_height = 1944*100/zoom;
		if(capture_width>zoom_width)
			capture_width = zoom_width;
		if(capture_height>zoom_height)
			capture_height = zoom_height;

		if( (capture_width >= 2592) && (capture_height>= 1944) ) {
			ovSizeFormat = OV5642_SIZE_QSXGA;
			capture_width = 2592;
			capture_height = 1944;
		}
		else if( (capture_width >= 2560) && (capture_height>= 1920) ) {
			ovSizeFormat = OV5642_SIZE_2560X1920;
			capture_width = 2560;
			capture_height = 1920;
		}
		else if( (capture_width >= 2048) && (capture_height>= 1536) ) {
			ovSizeFormat = OV5642_SIZE_QXGA;
			capture_width = 2048;
			capture_height = 1536;
		}
		else if( (capture_width >= 1600) && (capture_height>= 1200) ) {
			ovSizeFormat = OV5642_SIZE_UXGA;
			capture_width = 1600;
			capture_height = 1200;
		}
		else if( (capture_width >= 1280) && (capture_height>= 960) ) {
			ovSizeFormat = OV5642_SIZE_1280X960;
			capture_width = 1280;
			capture_height = 960;
		}
		else if( (capture_width >= 1024) && (capture_height>= 768) ) {
			ovSizeFormat = OV5642_SIZE_XGA;
			capture_width = 1024;
			capture_height = 768;
		}
		else if( (capture_width >= 640) && (capture_height>= 480) ) {
			ovSizeFormat = OV5642_SIZE_VGA;
			capture_width = 640;
			capture_height = 480;
		}
		else if( (capture_width >= 320) && (capture_height>= 240) ) {
			ovSizeFormat = OV5642_SIZE_QVGA;
			capture_width = 320;
			capture_height = 240;
		}
		else {
			ovSizeFormat = OV5642_SIZE_QVGA;
			capture_width = 320;
			capture_height = 240;
		}
		ov5642_dbg("ov5642: jpeg capture %d:%d, zoom=%d, zoomed %d:%d\n",
			capture_width, capture_height, zoom, zoom_width, zoom_height);
		camera_context->capture_input_width = capture_width;
		camera_context->capture_input_height = capture_height;
	}
	else if( (capture_width == 2592) && (capture_height== 1944) )
		ovSizeFormat = OV5642_SIZE_QSXGA;
	else if( (capture_width == 2560) && (capture_height== 1920) )
		ovSizeFormat = OV5642_SIZE_2560X1920;
	else if( (capture_width == 2592) && (capture_height== 1456) )
		ovSizeFormat = OV5642_SIZE_2592X1458;
	else if( (capture_width == 2048) && (capture_height== 1536) )
		ovSizeFormat = OV5642_SIZE_QXGA;
	else if( (capture_width == 2048) && (capture_height== 1152) )
		ovSizeFormat = OV5642_SIZE_2048X1152;
	else if( (capture_width == 1600) && (capture_height== 1200) )
		ovSizeFormat = OV5642_SIZE_UXGA;
	else if( (capture_width == 1600) && (capture_height== 896) )
		ovSizeFormat = OV5642_SIZE_1600X900;
	else if( (capture_width == 1280) && (capture_height== 960) )
		ovSizeFormat = OV5642_SIZE_1280X960;
	else if( (capture_width == 1280) && (capture_height== 720) )
		ovSizeFormat = OV5642_SIZE_1280X720;
	else if( (capture_width == 1024) && (capture_height== 768) )
		ovSizeFormat = OV5642_SIZE_XGA;
	else if( (capture_width == 854) && (capture_height== 480) )
		ovSizeFormat = OV5642_SIZE_854X480;
	else if( (capture_width == 640) && (capture_height== 480) )
		ovSizeFormat = OV5642_SIZE_VGA;
	else if( (capture_width == 480) && (capture_height== 360) )
		ovSizeFormat = OV5642_SIZE_480X360;
	else if( (capture_width == 432) && (capture_height== 320) )
		ovSizeFormat = OV5642_SIZE_432X320;
	else if( (capture_width == 320) && (capture_height== 240) )
		ovSizeFormat = OV5642_SIZE_QVGA;
	else if( (capture_width == 160) && (capture_height== 120) )
		ovSizeFormat = OV5642_SIZE_160X120;
	else if( (capture_width == 352) && (capture_height== 288) )
		ovSizeFormat = OV5642_SIZE_CIF;
	else if( (capture_width == 176) && (capture_height== 144) )
		ovSizeFormat = OV5642_SIZE_QCIF;
	else if( (capture_width == 128) && (capture_height== 96) )
		ovSizeFormat = OV5642_SIZE_SQCIF;
	else
		ovSizeFormat = OV5642_SIZE_QVGA;

	if( (ovSizeFormat!=g_ov5642_size)||(ovFormat!=g_ov5642_format)||(camera_context->capture_mode!=g_ov5642_mode_change) ) {
		sendtime = t_get_time();
		status = ov5642hw_set_format( ovSizeFormat,ovFormat, camera_context->capture_mode);
		sendtime = t_get_time() - sendtime;
		//ov5642_dbg("-------------------\n");
		//ov5642_dbg("sendtime =%d\n",sendtime);
		//ov5642_dbg("-------------------\n");
		printk(KERN_DEBUG "ov5642: mode %d size %d:%d sendtime =%d\n", camera_context->capture_mode, capture_width, capture_height, sendtime);

		g_ov5642_size = ovSizeFormat;
		g_ov5642_format = ovFormat;
	}

	/* set digital zoom after size changed */
	ov5642_resolution_zoom(camera_context, ov5642hw_get_zoom_level(zoom));

	timing.BFW = 0x01;	/* FIXME: green line */
	timing.BLW = 0x00;
	
	ci_configure_mp(camera_context->capture_input_width-1,camera_context->capture_input_height-1, &timing);
	return 0;
}

static int ov5642_start_capture(p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;
	if(camera_context->capture_mode == CAMERA_MODE_STILL)
		/* wait about 2 frame before capture */
		camera_context->capture_drop_frames = 2;
	else
		camera_context->capture_drop_frames = 0;
	printk(KERN_DEBUG "ov5642: start capture, mode %d\n", camera_context->capture_mode);

	return 0;
}

static int ov5642_stop_capture(  p_camera_context_t camera_context )
{
	(void) camera_context;
	if(flashled_on && camera_context->capture_mode == CAMERA_MODE_STILL) {
		ov5642hw_flash_work_off();
	}
	printk(KERN_DEBUG "ov5642: stop capture, mode %d\n", camera_context->capture_mode);
	return 0;
}

static int ov5642_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	ov5642hw_power_down( mode );
	return 0;
}

static int ov5642_read_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;

	ov5642hw_read_sensor_reg( reg_addr, (u8 *)reg_val );
	return 0;
}

static int ov5642_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	u8 buffer;
	int status;

	(void) camera_context;

	buffer = (u8)reg_val;
	status = (ov5642hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

static char name[] = "Omnivision5642";

static p_camera_function_t camera_functions;

static int format_list[] = {
	V4L2_PIX_FMT_YUV422P,
	V4L2_PIX_FMT_JPEG, 
	-1	
};

static int ov5642_register(int id)
{

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		ov5642_dbg("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->name 			= name;
	camera_functions->id 			= id;

	camera_functions->width_max = 2592;
	camera_functions->height_max = 1944;
	camera_functions->width_min = 2;
	camera_functions->height_min = 2;

	camera_functions->v_power = 0; //2800;
	camera_functions->v_io = 0; //1800;

	camera_functions->format_list = format_list;

	camera_functions->init 			= ov5642_init;
	camera_functions->deinit 		= ov5642_deinit;
	camera_functions->set_capture_format 	= ov5642_set_capture_format;
	camera_functions->start_capture 	= ov5642_start_capture;
	camera_functions->stop_capture 		= ov5642_stop_capture;
	camera_functions->sleep 		= ov5642_sleep;
	camera_functions->wakeup 		= ov5642_wake;
	camera_functions->read_8bit 		= NULL;
	camera_functions->write_8bit 		= NULL;
	camera_functions->read_16bit		= ov5642_read_16bit;
	camera_functions->write_16bit 		= ov5642_write_16bit;
	camera_functions->read_32bit 		= NULL;
	camera_functions->write_32bit 		= NULL;
	camera_functions->set_power_mode 	= ov5642_set_power_mode;
	camera_functions->set_contrast 		= ov5642_set_contrast;
	camera_functions->try_contrast 		= ov5642_try_contrast;
	camera_functions->set_whitebalance 	= ov5642_set_white_balance;
	camera_functions->try_whitebalance 	= ov5642_try_white_balance;
	camera_functions->set_saturation	= ov5642_set_saturation;
	camera_functions->try_saturation	= ov5642_try_saturation;
	camera_functions->set_colortone		= ov5642_set_colortone;
	camera_functions->try_colortone		= ov5642_try_colortone;
	camera_functions->set_sharpness		= ov5642_set_sharpness;
	camera_functions->try_sharpness		= ov5642_try_sharpness;
	camera_functions->set_flashmode		= ov5642_set_flashmode;
	camera_functions->try_flashmode		= ov5642_try_flashmode;
	//camera_functions->set_scene		= ov5642_set_scene;
	camera_functions->set_jpeg_quality_percent= ov5642_set_jpeg_quality;
	camera_functions->set_digital_zoom 	= ov5642_set_digital_zoom;
	camera_functions->set_ev_offset		= ov5642_set_ev_offset;
	camera_functions->set_autofocus		= ov5642_set_autofocus;
	camera_functions->set_framerate		= ov5642_set_framerate;
	camera_functions->get_digital_zoom_max 	= ov5642_get_digital_zoom_max;
	camera_functions->get_digital_zoom_step	= ov5642_get_digital_zoom_step;
	//camera_functions->set_flash_torch	= ov5642_set_flash_torch;
	camera_functions->set_af_windows_config	= ov5642_set_af_windows_config;
	camera_functions->set_effect		= ov5642_set_effect;
	camera_functions->try_effect		= ov5642_try_effect;

	camera_functions->enum_frame_sizes	= ov5642_enum_frame_sizes;
	camera_functions->enum_frame_intervals	= ov5642_enum_frame_intervals;

	/*register 5642 as high resolution */
	if(sensor_register(camera_functions, id) < 0){
		ov5642_dbg("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static int ov5642_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *                              OV5642 I2C Client Driver
 *****************************************************************************/
static int read_sensor_reg_16bit(u16 addr, u8 *pvalue)
{    
	u8 data;        
	u8 address[2];
	if( g_client == NULL )	/* Not initialize the client */	
	{
		ovI2C_dbg("not initialize the client\n");
		return -1;
	}
	address[0] = addr>>8;	
	address[1] = addr;
	i2c_smbus_write_byte_data(g_client,address[0],address[1]);
	data = i2c_smbus_read_byte(g_client);
	*pvalue = data;   
	//printk("ov5642: i2c r [%08x]=%02x\n", addr, data);

	return 0;
}
static int write_sensor_reg_16bit(u16 addr, u8 *value)
{
	u8 data[3];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;			
	data[0] = addr>>8;
	data[1] = addr;	
	data[2]=  *value;
	

	//printk("ov5642: i2c w [%08x]=%02x\n", addr, *value);
	i2c_master_send(g_client, data, 3);
	
	return 0;
}


static int write_sensor_reg(u16 addr, u8 value)
{
	int ret;
	down(&ov5642_i2c_lock);
	ret = write_sensor_reg_16bit(addr, &value);
	up(&ov5642_i2c_lock);
	return ret;
}

static int read_sensor_reg(u16 addr, u8 *pvalue)
{
	int ret;
	down(&ov5642_i2c_lock);
	ret = read_sensor_reg_16bit(addr,pvalue);
	up(&ov5642_i2c_lock);
	return ret;
}
static int write_sensor_block_regs(u8 *databuf ,int lenth)
{
	int ret;
	//printk("ov5642: w block [%08x]:%d\n", (databuf[0] << 8) + databuf[1], lenth-2);
	down(&ov5642_i2c_lock);
	ret = i2c_master_send(g_client,databuf,lenth);
	up(&ov5642_i2c_lock);
	return ret;
}
static int __attribute__ ((unused)) write_sensor_block_regs_16bit(u16 st_addr, u8 *databuf ,int lenth)
{
	u8 adr_high_byte;
	u8 adr_low_byte;
	u8 realbuf[17];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;
	if(lenth > 16)
	{
		ov5642_dbg("write block failed ::lenth is too big\n");
		return -1;
	}
	adr_high_byte = (st_addr>>8)&0xff;
	adr_low_byte  = (st_addr) & 0xff;
	realbuf[0] = adr_low_byte;
	memcpy((realbuf+1),databuf,lenth);
	
	
	//printk("ov5642: w block [%08x]:%d\n", st_addr, lenth);
	down(&ov5642_i2c_lock);
	i2c_smbus_write_i2c_block_data(g_client,adr_high_byte,lenth+1, realbuf);
	up(&ov5642_i2c_lock);

	return 0;
}

static int ov5642_create_workqueue(void)
{
	INIT_DELAYED_WORK(&flash_work, ov5642_flash_work_handler);
	flash_wqueue = create_workqueue("ov5642_flash");
	if(!flash_wqueue){
		printk(KERN_ERR "ov5642: create_workqueue(flash) failed\n");
		return -ENOMEM;
	}
	return 0;
}

static void ov5642_destroy_workqueue(void)
{
	if(flash_wqueue){
		cancel_rearming_delayed_workqueue(flash_wqueue, &flash_work);
		destroy_workqueue(flash_wqueue);
		flash_wqueue = NULL;
	}
}

static int __devinit sensor_probe(struct i2c_client *client, const struct i2c_device_id *_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	int id = 0;
	
	pdata = client->dev.platform_data;
	id = pdata->id;
	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		ov5642_dbg("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);
	
	ci_set_clock(1, 1, CICLK);
	
	g_client = client;

	pdata->power_on(id);

	mdelay(1);
	
	if (ov5642_sensorID_detect()<0){
		goto err;
	}

	if(ov5642_register(id)<0)
		goto err;


	ov5642_dbg("OV5642 detected.\n");
	
	ci_set_clock(0, 0, CICLK);
	pdata->power_off(id);

	clk_disable(clk);
	clk_put(clk);

	if(ov5642_create_workqueue())
		return -1;

	return 0;
err:
	ov5642_dbg("OV5642 detect failed.\n");
	ci_set_clock(0, 0, CICLK);
	pdata->power_off(id);

	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	ov5642_unregister(pdata->id);
	g_client = NULL;

	ov5642_destroy_workqueue();
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "sensor_ov5642", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "sensor_ov5642",
	},
	.probe		= sensor_probe,
	.remove		= sensor_remove,
	.id_table	= sensor_id,
};

static int __init sensor_init(void)
{
	return i2c_add_driver(&sensor_driver);
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}

module_init(sensor_init);
module_exit(sensor_exit);

MODULE_DESCRIPTION("OV5642 I2C Client driver");
MODULE_LICENSE("GPL");

